(* ---                           develop.v                              --- *)
(*                                                                          *)
(* Author: Pierre Casteran.                                                 *)
(*    LABRI, URA CNRS 1304,                                                 *)
(*    Departement d'Informatique, Universite Bordeaux I,                    *)
(*    33405 Talence CEDEX,                                                  *)
(*    e-mail:  casteran@labri.u-bordeaux.fr                                 *)

(* 

  This file contains some important lemmas used in the "generation" module.
  It tells how to compose abstract machine code (with append) to build
  code for Call_M, Call_C and Call_K calls (see the module "spec")
*)

AddPath ".".
Require Constants.
Require monoid.
Require spec.
Require machine.
Require Wf_compl.
Require Plus.
Require Mult.
Require Lt.
Require Mult_compl.
Require euclid.
Require two_power.


(* A sequence for computing the cube of some m in M *)
Lemma M3:(gencode (Call_M three)).
(******************************)
Proof.
 Realizer (seq PUSH (seq SQR (seq MUL End))).
 (* ------------------------------*)
 Program_all. 
 Constructor 1.
 Simpl;Unfold power three;Intros.
 Rewrite (u_neutral_r M MO).
 Rewrite <- (point_assoc M MO);Auto.
 Qed .




(* A sequence for computing the (2^q)-th power of some m in M;
 this is a sequence of q SQRs  *)



Lemma M2q:(q:nat)(gencode (Call_M (two_power q))).
(**********************************************)
Proof.
 Realizer [q:nat]<Code>Match q with End
                                    ([_,c:?](seq SQR c))
                                    end.
 (* -------------------------------------------------- *)
 Program_all.
 Constructor 1.
 Unfold power ;Simpl.
 Intros;Rewrite -> (u_neutral_r M MO);Auto.
 Constructor 1;Auto.
 Intros;Simpl.
 Inversion_clear s.
 Rewrite  H.
 Apply Config_inv;Auto.
 Rewrite <- plus_n_O.
 Apply a2x;Auto.
Qed.


(* code generation  for (Call_M p) from the code for (Call_C p q) 
   (it's the same!) *)

Lemma C2M:(p,q:nat)(gencode (Call_C p q ))->
                   (gencode  (Call_M p)).
(****************************************)
Proof.
 Realizer [p,q:nat][c:Code]c.
 (* ----------------------- *)
 Program_all.
 Constructor 1.
 Inversion_clear s.
 Auto.
Qed.



(*  code generation for  (Call_C qp p) from the codes for (Call_M p) and (Call_M q)
    just append !                                                          *)
Lemma MMC:(p,q:nat)(gencode  (Call_M p))->
                   (gencode (Call_M q))->
                   (gencode (Call_C (mult q p) p)).
(***********************************************)
Proof.
 Realizer [p,q:nat][c,c':Code](app c  c').
 (* ----------------------------------- *)
 Program_all.
 Constructor 2.
 Intros;Rewrite  Exec_app.
 Inversion_clear s.
 Rewrite  H.
 Inversion_clear s0.
 Rewrite  (H0 M MO).
 Rewrite power_mult;Auto.
Qed.


(* code generation for  (Call_C qp+r p) from cpr (code for (Call_K p r))
                                      and cq (code for (Call_M q))   
 (it's cpr, followed by cq then a MUL statement                       *)

Lemma KMC:(p,q,r:nat)
            (gencode (Call_K p r))->
            (gencode (Call_M q))->
            (gencode (Call_C (plus (mult q p) r) p)).
(***********************************************************)
Proof.
 Realizer [p,q,r:nat][cpr,cq:Code](app cpr (app  cq (seq MUL   End))).
 (* -------------------------------------------------------------- *)
 Program_all.
 Inversion_clear s;Inversion_clear s0.
 Constructor 2.
 Intros;Rewrite Exec_app.
 Rewrite (H M MO  a).
 Rewrite (Exec_app M MO);Simpl;Rewrite H0.
 Simpl.
 Apply Config_inv.
 Rewrite power_eucl;Auto.
 Auto.
Qed.


(* Code generation for (Call_K qp p) from cp (code for (Call_M p))
                                   and cq (code for (Call_M q))
   It's cp, followed by a PUSH, then cq                       *)

Lemma MMK: (p,q:nat)(gencode (Call_M p))->
                    (gencode (Call_M q))->
                    (gencode  (Call_K (mult q p) p)).
(**************************************************)
Proof.
 Realizer [p,q:nat][cp,cq:Code](app cp  (seq PUSH  cq)).
 (* ------------------------------------------------ *)
 Program_all. 
 Inversion_clear s;Inversion_clear s0.
 Constructor 3.
 Intros;Rewrite Exec_app. 
 Rewrite H;Simpl.
 Rewrite H0;Simpl.
 Apply Config_inv;Auto.
 Rewrite <- power_mult;Auto.
Qed.


(* Code generation for (Call_K qp+r p) from kpr (code for (Call_K p r))
                                     and mq (code for (Call_M q))
  it's kpr, followed by a sequence (PUSH;SWAP), then mq and a MUL statement *)

Lemma KMK:(p,q,r:nat)(gencode (Call_K p r))->
                     (gencode (Call_M q))->
                     (gencode (Call_K (plus (mult q p) r) p)).
(***********************************************************)
Proof.
 Realizer [p,q,r:nat][kpr,mq:Code]
    (app kpr  (seq PUSH (seq  SWAP  ( app mq  (seq MUL  End))))).
 (* ----------------------------------------------- *)
 Program_all.
 Inversion_clear s;Inversion_clear s0.
 Constructor 3.
 Intros;Rewrite Exec_app.
 Rewrite H;Simpl.
 Rewrite Exec_app.
 Rewrite H0;Simpl.Auto.
 Rewrite power_eucl;Auto.
Qed.








