
(* ----                 generation.v ----                                   *)

(*                                                                          *)
(* Author: Pierre Casteran.                                                 *)
(*    LABRI, URA CNRS 1304,                                                 *)
(*    Departement d'Informatique, Universite Bordeaux I,                    *)
(*    33405 Talence CEDEX,                                                  *)
(*    e-mail:  casteran@labri.u-bordeaux.fr                                 *)

(*

  This file uses the results of "develop" to validate the chain 
  generation algorithm; see also the "spec" module.

*)
  
AddPath "$COQTOP/theories/PROGRAMS".
AddPath ".".
Require monoid.
Require spec.
Require Constants.
Require machine.
Require Euclid_proof.
Require Le_lt_compl.
Require euclid.
Require shift.
Require two_power.
Require log2_spec.
Require develop.
Require Arith.
Require Peano_dec.
Require Compare_dec.
Require strategies.
Require Mult_compl.
Require Wf_nat.
Require Wf_compl.


(* termination of the addition chain algorithm *)
(***********************************************)

(* termination order (based on recursive calls) *)

Inductive Call_lt:Call->Call->Prop :=
   M_M_lt:(p,q:nat)(lt p q)->(Call_lt (Call_M p) (Call_M q))
 | C_M_lt:(p,q:nat)(lt p q)->(Call_lt (Call_C q p) (Call_M q))
 | M_C_lt:(p,q,r:nat)(lt p r)->(Call_lt (Call_M p) (Call_C r q))
 | M_K_lt:(p,q,r:nat)(le p r)->(Call_lt (Call_M p) (Call_K r q))
 | K_C_lt:(p,q,r,s:nat)(lt q s)->(Call_lt (Call_K q p) (Call_C s r))
 | K_K_lt:(p,q,r,s:nat)(lt q s)->(Call_lt (Call_K q p) (Call_K s r)).

(* In order to prove well-foundedness of Call_lt, we use an
   auxiliary ad-hoc measure *)


Definition Call_measure:=[c:Call]
                          <nat>Case c of
                                 (* Call_M n *)
                                 [n:nat](S (mult three n))
                                 (* Call_C n p *)
                                 [n,p:nat](mult three n)
                                 (* Call_K n p *)
                                 [n,p:nat](S (S (mult three n)))
                              end.


Lemma  measure_compat : 
   (coarser Call Call_lt (ltof Call Call_measure)).
(**************************************************)
Proof.
 Unfold coarser ltof.
 Induction 1;Unfold Call_measure. 
 Intros;Apply lt_n_S;Apply mult_lt_l;Auto.
 Unfold three;Auto.
 Auto.
 Intros.
 Apply lt_le_trans with (mult three (S p)).
 Simpl.
 Apply lt_n_S.
 Rewrite <- plus_n_O.
 Apply lt_reg_l.
 Rewrite <- plus_Snm_nSm.
 Simpl;Auto.
 Auto.
 Auto.
 Intros.
 Apply lt_le_trans with (mult three (S q)) .
 Replace (mult three (S q)) with (S (S (S (mult three q)))).
 Auto.
 Rewrite <- mult_n_Sm.
 Rewrite plus_sym;Simpl;Auto.
 Auto.
 Intros;Do 2 Apply lt_n_S;Apply mult_lt_l;Auto.
 Unfold three;Auto.
Qed.



Lemma Wf_Call_lt: (well_founded Call Call_lt).
(********************************************)
Proof.
 Apply wf_coarser with (ltof Call Call_measure).
 Exact measure_compat.
 Apply well_founded_ltof.
Qed.



(* code generation algorithm *)
(*****************************)

Section generation.
  Variable gamma:strategy.
  Variable log2_r:(n:nat)(lt O n)->(log2_spec n).

  (* treatment of the 3 cases for (Call_M p)
       - p = 3
       - p is a power of 2 (q is the exponent)
       - we use (gamma p) as an intermediate value
  *)
             

  Lemma chain_cases:(p:nat)(lt O p)->
                          {q:nat|(lt q p) /\ (le two q)}+
                          {l:nat|p=(two_power l)}+
                          {p=three}.
  (*******************************************************)
  Proof.
  Realizer [p:nat]<sumor (sum nat nat)>if (eq_nat_dec p three)
                  then (inright (sum nat nat))
                  else <(sumor (sum nat nat))> 
                    let (l:nat;b:bool)=(log2_r p) in
                    <(sumor (sum nat nat))>if b 
                                           then
                                             (inleft
                                                 (sum nat nat)
                                                 (inr nat nat l))
                                           else (inleft (sum nat nat)
                                                 (inl nat nat
                                                      (gamma p))).

  (* ----------------------------------------------------------- *)
  Program_all.
  Case (le_or_lt (S four) p);[Auto|Intro H2].
  Case (enum4 p H2);Intro.
  Absurd (lt O p);[(Rewrite H0;Auto)|Auto]. 
  Elim H0;Intro H3.
  Absurd p=one;[Replace one with (two_power O) | Auto].
  Apply holes_in_powers with l;Elim a;Auto.
  Simpl;Auto.
  Elim H3;Intro H4.  
  Absurd p=two;[Replace two with (two_power one) | Auto].
  Apply holes_in_powers with l;Elim a;Auto.
  Simpl;Auto. 
  Elim H4;Intro H5.
  Elim (n H5).
  Absurd p=four;[Replace four with (two_power two) | Auto].
  Apply holes_in_powers with l;Elim a;Auto.
  Simpl;Auto. 
 Qed.


(* rougly speaking, COND3 can be used as a specialized control
   structure:
   In Real programs,
   (Cond3 A p case3
              case_2q
              case_gamma)
   is equivalent to :
  "if p = 3 
   then case3
   else if p is the q_th power of 2, 
        then (case_2q q)
        else (case_gamma (gamma p))"
*)


 Lemma COND3:(A:nat->Set)(p:nat)
       (p=three ->(A p))-> 
       ((q:nat)(p=(two_power q))->(A p))->
       ((q:nat)(lt q p )->(le two q)->(A p))->
       (lt O p)->
       (A p).
(*********************************************)
 Proof.
  Realizer [A:Set][p:nat]
                 [case3:A]
                 [case_2n:nat->A]
                 [case_gamma:nat->A]
                 <A>Match (chain_cases p) with
                    [b:(sum nat nat)]
                        <A >Match b with
                            [q:nat](case_gamma q)
                            [q:nat](case_2n q) end
                    case3  end.
 (* ----------------------------------------- *)
  Program_all.
  Elim a;Auto.
  Elim a;Auto.
 Qed.


 (* the code generation algorithm  *)
 (* the labels A,B,C,A1,A2,A3,B1,B2,C1,C2, in the proof comments
    correspond to the following places in the program:

     Match c with
       A: (Call_M p)
         (COND3 
           A1: (p=3) ...
           A2: (p=2^q) ...
           A3: (q=gamma(p)) ...)
       B: (Call_C n n0)
           ... if (zerop r)
               then B1: ...
               else B2: ...

       C: (Call_K n n0)
           ... if (zerop r)
               then C1: ...
               else C2: ...
     end.
*)

    
 Lemma chain_algo: (c:Call)(conform c)->(gencode c).
 (******************************************)
 Proof.
 Realizer <Code>rec REC :: :: {Call_lt}
     [c:Call]
       <Code>Case c of
        (* Call_M p *)
         [p:nat]
         (COND3 Code p M3:: ::{(gencode (Call_M three))} 
                       [q:nat](M2q q)
                       [q:nat] (C2M p q (REC (Call_C p q))))
        (* Call_C p n0 *)
         [p,n0:nat]
         <Code>let (q,r:nat)=(eucl_dev n0 p) in
               <Code>if (zerop r) 
                     then 
                      (MMC n0 q (REC (Call_M n0)) (REC (Call_M q)))
                     else 
                      (KMC n0 q r (REC (Call_K n0 r)) (REC (Call_M q)))
       (* Call_K p n0 *)
        [p,n0:nat]
           <Code>let (q,r:nat)= (eucl_dev n0 p) in
               <Code>if (zerop r) 
                     then
                      (MMK n0 q (REC (Call_M n0)) (REC (Call_M q)))
                     else
                      (KMK n0 q r (REC (Call_K n0 r)) (REC (Call_M q)))
        end.
  (* ---------------------------------------------------------------- *)
  Do 2 Program.
  Exact Wf_Call_lt.
  Program.
  Program.
  Program.
  Program.
  Program.
  (* 
  A:  code generation for (Call_M p)) *)

  Intro H.
  Program_all.

  (* A1: p=3 *)
  Intros H' eq3;Rewrite eq3;Program_all.

  (* A2: p is q power of 2 *)
  (Rewrite -> H0;Auto).

  (* A3: Call to (Call_ C p q) ( 2<= q < p)  *)
  (*  (Call_lt (Call_C p q) (Call_M p)) *)
  Constructor 2;Auto.
  (*  (conform (Call_C p q)) *)
  Simpl;Auto.


  (* 
  B: code generation for (Call_C n n0) *)

  Do 3 Program.
  Intro H0.
  Simpl in H0;Elim H0;Intros H1 H2.
  Program.
  Program.
  Unfold gt ;Apply lt_trans with (S O) ;Auto.
  Program.
  Program.
  Program.
  Intros e0.
 
 (* B1:  n0 divides n *)
  
  Cut  p= (mult q n0).
  Intros eg.
  Cut (lt O q).
  Intro H'.
 
  Program_all.
 
  (* (Call_lt (Call_M n0) (Call_C p n0)) *)
  Constructor 3;Auto.

  (* (conform (Call_M n0)) *)
  Simpl;(Elim e0;Auto).
 
  (* (Call_lt (Call_M q) (Call_C p n0)) *)
  Constructor 3. 
  Rewrite -> eg;Rewrite -> mult_sym;Apply mult_p_lt_qp;Auto.
 
  (*  (Spec (Call_C p n0) c0) *)
  Rewrite -> eg;Auto.
 
  (* lt O q *)
  Apply quotient_positive with p n0.
  Auto.
  Apply le_lt_trans with n0;Auto.   
  Rewrite -> e;Rewrite -> e0;Auto.

  (* B2: n=qn0+r (r>0)  *)
  
  Program_all.
 
  (* (Call_lt (Call_K n0 r) (Call_C p n0)) *)
  Constructor 5;Auto.
   
  (*  (conform (Call_K n0 r)) *)
  Simpl;Auto.
 
  (*  (Call_lt (Call_M q) (Call_C p n0)) *)
  Constructor 3.
  Rewrite -> e;Auto.
 
  (*  (conform (Call_M q)) *)
  Simpl.
  Apply lt_O_q with n0 r ;Auto.
  Elim e;Auto.
  Rewrite -> e;Auto.

  (* 
  C:  code generation for (Call_K p n0) *)
 
 
   Do 3 Program.
   Intros H0. (* (conform (Call_K p n0)) *)
   Elim H0;Intros.
   Program.
   Program_all.
   Do 3 Program.
   Intros e0.
   
  (* C1: p=qn0 *)
   
  Cut  p=(mult q n0).
  Intros eg.
  Cut (lt O q).
  Intro H'.
  Do 2 Program.
  
  (* (Call_lt (Call_M n0) (Call_K p n0)) *)
  Constructor 4;Auto.
  
  (* (conform (Call_M n0)) *)
  Simpl;Auto.
 
  (* (Call_lt (Call_M q) (Call_K p n0)) *)
  Constructor 4.
  Rewrite -> eg;Apply mult_p_le_pq;Auto.
 
  (* conform (Call_M q) *)
  Simpl;Auto.
  Rewrite -> e;Rewrite -> e0;Auto.
  Program_all.
  Rewrite <- plus_n_O;Auto.

  (* lt O q *)
  Apply quotient_positive with p n0.
  Auto.
  Apply le_lt_trans with n0;Auto.   
 
  Rewrite -> e;Rewrite -> e0;Rewrite <- plus_n_O;Auto.

  (* C2: p=qn0+r (r>0)  *)
 
  Program_all.

  (* (Call_lt (Call_K n0 r) (Call_K p n0)) *)
  Constructor 6;Auto.
 
  (* (conform (Call_K n0 r)) *)
  Simpl;Auto.
 
  (* (Call_lt (Call_M q) (Call_K p n0)) *)
  Constructor 4;Auto.
  Rewrite -> e;Apply le_q_qbr;Auto.
 
  (* (conform (Call_M q)) *)
  Simpl.

  (* lt O q *)
  Apply lt_O_q with n0 r ;Elim e;Elim H0;Auto.
  Rewrite -> e;Auto.
 Qed.


 
 Lemma power_algo:(n:nat)(lt O n)->(addchain_spec n).
 (***************************************************)
 Proof.
  Intros;Split.
  Apply chain_algo.
  Simpl;Auto.
 Qed.
 

End generation.






















