(* -----                  log2_implementation.v                        ---- *)
(*                                                                          *)
(* Author: Pierre Casteran.                                                 *)
(*    LABRI, URA CNRS 1304,                                                 *)
(*    Departement d'Informatique, Universite Bordeaux I,                    *)
(*    33405 Talence CEDEX,                                                  *)
(*    e-mail:  casteran@labri.u-bordeaux.fr                                 *)

(* 

This file contains an algorithm to compute the integer logarithm
 (base 2) of a given positive integer.
The  algorithm we give first  not only computes this
 logarithm, but returns also an information on it's exactness.

 Since we are interested in using Coq for low-levels programs,
 we choosed to realize  the specification of log2 with
 a while-loop using shift (multiplication by 2) and unshift
 (quotient by 2, and parity check).

*)


AddPath ".".
AddPath "$COQTOP/theories/PROGRAMS".
Require Arith.
Require Peano_dec.
Require Constants.
Require Mult_compl.
Require Wf_nat. 
Require euclid.
Require Le_lt_compl.
Require Euclid_proof.
Require shift.
Require Compare_dec.
Require imperative.
Require while.
Require two_power.
Require log2_spec.


Section log2_computation.
 Variable n0:nat.
 Hypothesis pos_n0:(lt O n0).
 Hint pos_n0.

 (* specification 
 *****************)
 (*
    It is easy, by a sequence of unshifts (right shifts) to
    compute, not only the integer logarithm (base 2 ) of a positive
    integer, but also to know  wether this logarithm is exact;
    the additionnal cost is a boolean variable, initialized with
    true, and set to false the first time a division by 2 returns a
    non null rest.
*)


 (*
  Sketch of the (imperative) program:
 
  n := n0;
  p:= O;
  b := true; /* n0=2^p * n  */

  while (n>O)  
  {
    if (b)  /* n0=2^p * n  */
      {
         if(n is even) 
            { p:=p+1;
              n:=unshift(n);
              b:=true}
         else
             {p:=p+1;
              n:=unshift(n);
              b:=false}
      }
     else   /* 2^p * n < n0 < 2^p(n+1) /\ n>0 */
      {
       p:=p+1;
       n:=unshift(n);
       b:=false}
    }
    return (p,b);
 *)

 
 
 (* computation states  
 **************************)
 
  Record state:Set :=
     mkstate {
        state_n:nat;
        state_p:nat;
        state_b:bool}.
 
  Inductive ok:state->Prop:=
      oktrue: (p:nat)n0=(two_power p)->
                      (ok (mkstate one p true))
    | okfalse: (p:nat)(lt (two_power p) n0)->
                      (lt n0 (two_power (S p))) ->
                      (ok (mkstate one p false)).

  Local init:=[s:state](state_n s)=n0 /\ (state_p s)=O /\
                       (state_b s)=true.





(* loop parameters *)
 (* exit predicate  (when to stop looping)
 ************************************************)
 
  Inductive final:state->Prop :=
    mkfinal:(p:nat)(b:bool)(final (mkstate one p b)).
 
 (* loop invariant 
 *****************)
  (* in a state (mkstate n p b), the boolean b expresses
     the exactness of the logarithm to compute *)

  Inductive invar:state->Prop :=
     exact:(n,p:nat)
             n0=(mult (two_power p) n)->
             (lt O n)->
             (invar (mkstate n p true))
   |inexact:(n,p:nat)
                (lt (mult (two_power p)n) n0)->
                (lt n0 (mult (two_power p) (S n)))->
                (lt O n)->
                (invar (mkstate n p false)).
 
 
 (* termination order  (hope it's well founded !)
 ************************************************)
 
  Definition stat_order:=(ltof state state_n).
 
 
 (* technical lemmas 
  ******************)

 (* About states 
 ******************)

 Remark decomp:(s:state)
                s=(mkstate (state_n s)(state_p s)(state_b s)).
 Proof.
   Induction s;Simpl;Auto.
 Qed.

 (* About the invariant 
 ***********************)



 Remark invar_inv1:(n,p:nat)(invar (mkstate n p true))->
                           n0=(mult (two_power p) n).
 Proof.
  Intros;Inversion_clear H;Auto.
 Qed.

 Remark invar_inv2: (n,p:nat)(invar (mkstate n p false))->
                            (lt (mult (two_power p)n) n0)  /\
                            (lt n0 (mult (two_power p) (S n))).
 Proof.
  Intros.
  Inversion_clear H; Split;Auto.
 Qed.

 Remark invar_inv3: (n,p:nat)(b:bool)(invar (mkstate n p b))->
                                   (lt O n).
 Proof.
  Intros.
  Inversion_clear H;Auto.
 Qed.


 (* About loop exiting 
 *********************)


 Remark final_inv:(n,p:nat)(b:bool)(final (mkstate n p b))->n=one.
 Proof.
  Intros.
  Inversion_clear H;Auto.
 Qed.




 Remark not_final:(n,p:nat)(b:bool)(invar (mkstate n p b))->
                                  ~(final (mkstate n p b))->
                                  (lt one n).
 Proof.
  Intros.
  Case (le_or_lt n one);[Intro|Auto].
  Cut (lt O n);[Intro|(Apply invar_inv3 with p b;Auto)].
  Case (enum1 n).
  Unfold two;Auto.
  Intro;Absurd (lt O n);Auto.
  Rewrite -> H3;Auto.
  Intro;Absurd (final (mkstate n p b));[Auto|Idtac].
  Rewrite -> H3;Split;Auto.
 Qed.



(* termination of the loop *)
(***************************)

 Remark wf_stat_order:(well_founded state stat_order).
 Proof.
  Unfold stat_order;Apply well_founded_ltof.
 Qed.
 Hint wf_stat_order.

 Lemma log2_impl:(log2_spec n0).
 (******************************)
 Proof.
  Apply Imperative with state init ok;[
        
     (* :init *)
           Realizer (mkstate n0 O true) |
     (* :body *)
     (Intros  a H;
       Apply while_not with init invar stat_order final a;[
        (* :test *)  
               Realizer [s:state]
               <bool>if (le_lt_eq_dec (S O) (state_n s))
                     then false
                     else true  |
        (* :body *)
              Realizer [s:state]
              <state>if (state_b s) 
               then  
                <state>let (m:nat;b':bool)=(Unshift (state_n s))
                       in
                      <state> if b' 
                           then (mkstate m (S (state_p s)) true)
                           else (mkstate m (S (state_p s)) false)
               else
               <state>let (m:nat;b:bool)=(Unshift (state_n s))
                      in   (mkstate m (S (state_p s)) false) |
                Idtac | Idtac | Idtac | Idtac]
       )

         |
     (* :return *)
      Realizer [a:state]
          <(prod nat bool)>Case (state_b a) of
             ((state_p a),true)
             ((state_p a),false)
             end
 ].

(* :init *)
Program_all.
Hint Unfold init.
Auto.                       

 (*
  n0 : nat
  pos_n0 : (lt O n0)
  ============================
   (a:state)(init a)->{a':state | (ok a')}
*)
 (* termination test : *)
   
      Program_all.
      Simpl.
      Change (lt O state_n0).
      Apply invar_inv3 with state_p0 state_b0;Auto.
      Unfold final;Simpl in l;Auto.
      Red;Intro.
      Absurd (lt (S O) one);[Auto|Idtac].
      Rewrite <- (final_inv ? ? ? H1);Exact l.
      Simpl in e;Rewrite <- e;Split;Auto.

  (* one loop step: *)

  Program.
  Do 3 Program.
  Intros  no_final state_ok .
  Cut (lt one state_n0).
  Intro n_ge_1.
  2:Apply not_final with state_p0 state_b0;Auto.
  Program.
  Program.
  Program. 

  (*  b (state_b0) is true ,
   which means that until now all the quotients by 2 were exact,
   or more precisely:  *)
  Cut n0=(mult (two_power state_p0) state_n0).
  Intro n0_2pn.
  2:Apply invar_inv1;Elim H;Auto.
  Simpl in H.
  Simpl.
  Program.
  Program.
  Program.
  Simpl;Split.
  Left.


  (* m is exactly the half of state_n0n *)
  (**************************************)  


  Rewrite  two_power_S.
  Rewrite (mult_sym two (two_power state_p0)).
  Rewrite <- mult_assoc_l;Auto.
  Unfold shift in e;Elim e;Auto.
  Apply quotient_positive with state_n0 two .
  Rewrite mult_sym;Exact e.
  Apply lt_trans with one ;Auto.
  Unfold stat_order ltof.
  Apply half_lt.
  Apply lt_trans with one ;Auto.
  Auto.


  (* state_n0=S(shift m) *)
  (***********************)
  Program_all.
  Simpl;Split.
  Right.
  Rewrite two_power_S.
  Rewrite -> (mult_sym two (two_power state_p0)).
  Rewrite <- mult_assoc_l;Auto.
  Rewrite  (invar_inv1 state_n0 state_p0).
  Rewrite e.
  Unfold shift .
  Auto.
  Auto.
  Elim H0;Auto.
  Rewrite two_power_S.
  Rewrite (mult_sym two (two_power state_p0)).
  Rewrite <- mult_assoc_l.
  Rewrite  (invar_inv1 state_n0 state_p0).
  Rewrite e.
  Unfold shift.
  Apply mult_lt_l.
  Auto.
  Simpl;Auto.
  Elim H0;Auto.
  Case (lt_or_Zero m);Auto.
  Intro.
  Absurd (lt one state_n0);Auto.
  Rewrite e.
  Rewrite -> H1;Simpl;Auto.
  Unfold stat_order ltof;Apply half_lt.
  Simpl.
  Apply lt_trans with one;Auto.
  Auto.
  Left.
  Apply invar_inv1.
  Rewrite <- H0;Auto.
  Apply lt_trans with one;Auto.

 (* b (state_b0) is false , and will be false for ever,
     the logarithm will be inexact; the situation is 
     expressed by the four following cuts
  *)
  Program.
  Elim (invar_inv2 state_n0 state_p0).
  Program.
  Program.
  Program.

  Cut (lt (mult (two_power state_p0) state_n0) n0).
  Intro n0_left.

  Cut (lt n0 (mult (two_power state_p0) (S state_n0))).
  Intro n0_right.
 
  Cut (lt (mult (two_power (S state_p0)) m) n0).
  Intro n0_left'.

  Cut (lt n0 (mult (two_power (S state_p0)) (S m))).
  Intro n0_right'.

  Cut (lt m state_n0).
  Intro m_little.

  Simpl;Split.
  Right;Auto.
  Case (lt_or_Zero m);Auto.
  Intro.
  Absurd (lt one state_n0);Auto.
  Simpl in b;Elim b;Intros eg;Rewrite -> eg.
  Rewrite H3;Simpl;Auto.
  Rewrite H3;Simpl;Auto.
  Simpl.
  Exact m_little.
 

  (* proof of m_little (last cut) *)

  Apply half_lt.
  Apply lt_trans with one ;Auto.
  Simpl in b;Elim b;Auto.

 (* Proof of n0_right':
     (lt n0 (mult (two_power (S state_p0)) (S m))). *)

  Simpl in b;Elim b;Intros eg;Rewrite -> eg.
  Apply lt_trans with (mult (two_power state_p0)
                            (S state_n0)) ;Auto.
  Rewrite two_power_S.
  Rewrite <- mult_assoc_l;Auto.
  Apply mult_lt_l.
  Auto.
  Rewrite -> eg.
  Unfold shift ;Simpl;Auto.
  Replace (mult (two_power (S state_p0)) (S m))  
   with   (mult (two_power state_p0) (S state_n0)).
  Elim (invar_inv2 state_n0 state_p0);Auto.
  Elim H0;Auto.
  Rewrite two_power_S.
  Rewrite <- mult_assoc_l.
  Simpl.
  Unfold shift ;Simpl.
  Rewrite <- plus_n_Sm.
  Rewrite eg;Auto.

 (* proof of n0_left':(lt (mult (two_power (S state_p0)) m) n0). *)

  Apply le_lt_trans with (mult (two_power state_p0) 
                                state_n0);Auto.
  Simpl in b;Elim b;Intro eg.
  Rewrite eg.
  Rewrite  two_power_S.
  Rewrite <- mult_assoc_l;Auto.
  Rewrite -> eg.
  Rewrite -> two_power_S.
  Rewrite <- mult_assoc_l;Auto.


  (*   Proof of  n0_right:
          (lt n0 (mult (two_power state_p0) (S state_n0))) *)

   Elim (invar_inv2 state_n0 state_p0).
   Auto.
   Elim H0;Auto.
  
 (* Proof of n0_left:(lt (mult (two_power state_p0) state_n0) n0). *)

   Elim (invar_inv2 state_n0 state_p0).
   Auto.
   Elim H0;Auto.

   Elim H0;Auto.


  (* (s:state)(final s)->(invar s)->(ok s) *)
  Induction s.
  Induction state_b.
  Intros.
  Rewrite (final_inv state_n0 state_p0 true).
  
  Left.
  Rewrite (invar_inv1 ? ? H1).
  Rewrite (final_inv ? ? ? H0).
  Auto.
  Auto.
  
  Intros.
  Rewrite (final_inv state_n0 state_p0 false).
  Right.
  (* (lt (two_power state_p0) n0) *)
  Elim (invar_inv2 ? ? H1).
  Rewrite (final_inv ? ? ? H0).
  Intros.
  Rewrite <- (mult_n_1 (two_power state_p0));Auto.
  (*  (lt n0 (two_power (S state_p0))) *)
  
  Elim (invar_inv2 ? ? H1).
  Rewrite (final_inv ? ? ? H0).
  Intros.
  Rewrite two_power_S.
  Exact H3.
  Auto.
  
  (*  (s:state)(init s)->(invar s) *)
  Induction s.
  Induction state_b.
  Intro.
  Inversion_clear H0.
  Elim H2.
  Simpl;Intros.
  Left.
  Rewrite H0;Simpl.
  Rewrite <- H1;Auto.
  Simpl in H1;Rewrite  H1;Auto.
  
  Intro.
  Inversion_clear H0.
  Elim H2.
  Intros.
  Simpl in H3.
  Discriminate H3.
  
  
  Auto.
  Auto.
  (* exit *)
  Do 6 Program.
  Simpl in H0.
  Simpl.
  Cut (ok (mkstate state_n0 state_p0 true)).
  Intro H'.
  Inversion_clear H'.
  Auto.
  Elim H0;Auto.
  Simpl in H0;Simpl.
  Cut (ok (mkstate state_n0 state_p0 false)).
  Intro H'.
  Inversion_clear H'.
  Auto.
  Elim H0;Auto.
Qed. 

End log2_computation.
