
(* ---                        monoid.v                                  --- *)
(*                                                                          *)
(* Author: Pierre Casteran.                                                 *)
(*    LABRI, URA CNRS 1304,                                                 *)
(*    Departement d'Informatique, Universite Bordeaux I,                    *)
(*    33405 Talence CEDEX,                                                  *)
(*    e-mail:  casteran@labri.u-bordeaux.fr                                 *)


(* 
 A monoid on some M:Set is a structure composed of a binary
 associative operation o:M->M->M, and a neutral element u:M
 for o.
*)


AddPath ".".
Require Arith.
Require Constants.


Section monoid_def.

Variable M:Set.

Record monoid:Set := 
     mkmonoid {
        u:M;
        o:M->M->M;
        point_assoc:(a,b,c:M)      
                 (o a (o b  c))=(o(o a  b)  c);
        u_neutral_l:(a:M) (o u  a) =a;
        u_neutral_r:(a:M) (o a u)=a }.

Hint point_assoc u_neutral_r u_neutral_l.

Variable MO:monoid.
Local u' := (u MO).
Local o' := (o MO).
Remark o'_assoc:(a,b,c:M)      
                 (o' a (o' b  c))=(o' (o' a  b)  c).
Exact (point_assoc MO).
Qed.
Hint o'_assoc.

Remark u'_neutral_l:(a:M) (o' u'  a) =a.
Exact (u_neutral_l MO).
Qed.
Hint u'_neutral_l.

Remark u'_neutral_r:(a:M) (o' a  u') =a.
Exact (u_neutral_r MO).
Qed.
Hint u'_neutral_r.


(* powering 
************)

Recursive Definition power[x:M]:nat->M:=
                            O => u'
                        |   (S n) =>(o'  x  (power  x  n)).




(* some lemmas on powering *)
(***************************)

Lemma  power_u:(m:nat)(power u'  m)=u'.
(******************************)
Proof.
 Induction m;Simpl.
 Auto.
 Intros.
 Rewrite -> H;Auto.
Qed.
Hint power_u.


Lemma power_plus:(a:M)(n,p:nat)(power a  (plus n p))=
                               (o'  (power  a  n)  (power a  p)).
(****************************************************)
Proof.
 Induction n;Simpl;[Auto|Intros].
 Rewrite -> (H p);Auto.
Qed.
Hint power_plus.



Lemma power_1:(a:M)(power a  (S O))=a.
(***********************************)
Proof.
 Intro;Simpl;Auto.
Qed.

Lemma  power_sym:(a:M)(n,p:nat)(o' (power a  n)  (power a  p))=
                               (o' (power  a  p)  (power a  n)).
(*************************************************************)
Proof.
 Intros.
 Rewrite <- power_plus.
 Rewrite <- power_plus.
 Rewrite -> plus_sym;Auto.
Qed.
Hint power_sym.


Lemma power_mult:(a:M)(n,p:nat)(power  a  (mult p n))=
                               (power  (power a  n)  p).
(*****************************************************)
Proof.
 Induction p;Simpl.
 Auto.
 Intros n0 H.
 Rewrite -> power_plus.
 Elim H;Auto.
Qed.
Hint power_mult.



Lemma a2x:(a:M)(x:nat)(power (o'  a  a)   x)=(power a  (plus x x)).
(***************************************************************)
Proof.
 Induction x;Simpl.
 Auto.
 Intros;Simpl.
 Rewrite <- plus_n_Sm.
 Simpl.
 Rewrite -> o'_assoc;Elim H;Auto.
Qed.

Lemma power_eucl:(m:M)(b,q,r:nat)( power m  (plus (mult q b) r))=
                                 (o' (power (power m b) q) (power m  r)).
(*************************************************************)
Proof.
 Intros;Rewrite -> power_plus.
 Rewrite -> power_mult.
 Auto.
Qed.
Hint power_eucl.


End monoid_def.












