(* ---                           spec.v                                 --- *)
(*                                                                          *)
(* Author: Pierre Casteran.                                                 *)
(*    LABRI, URA CNRS 1304,                                                 *)
(*    Departement d'Informatique, Universite Bordeaux I,                    *)
(*    33405 Talence CEDEX,                                                  *)
(*    e-mail:  casteran@labri.u-bordeaux.fr                                 *)


(* 
  This file presents the notations used by the "develop",
  "generation" and "main"  modules.

*)

AddPath ".".
AddPath "$COQTOP/theories/PROGRAMS".
Require monoid.
Require machine.
Require Constants.



(* The addition-chain algorithm uses three mutually  recursive
    functions for generating abstract machine code.
    In order to use the Program tactic, we treat them at the same time.
   First, we define the type of "procedure calls":
*)

Inductive Call:Set :=
   Call_M:nat->Call
 | Call_C:nat->nat->Call
 | Call_K:nat->nat->Call.
 

(* We define constraints on the parameters of procedure calls *)

Definition conform:Call->Prop:=
     [c:Call]
     <Prop>Case c of
         (* Call_M n *)    [n:nat](lt O n)
         (* Call_C n p *)  [n,p:nat](and (lt (S O) p) (lt p n))
         (* Call_K n p *)  [n,p:nat](and (lt O p) (lt p n))
       end.

(*  The proposition (Spec C c) means
    "The code c is correct wrt the call C" 

     That is to say:
       A code for (Call_M p) leaves the stack unchanged,
       and replaces a value x in the register X by x^p.

       A code for (Call_C p q) does the same thing as (Call_M p),
       but we know that the value x^q is computed before x^p
       (2 <= q < p )
    
       A code for (Call_K p q) does  almost the same thing as
       (Call_C p q), except that the computation pushes x^q
       at the top of the stack.
*)


Inductive Spec:Call->Code->Prop :=
     m_spec:(n:nat)(c:Code)
              ((M:Set)(MO:(monoid M))
              ((a:M)(s:(Stack M))
               (Exec M MO  c (config M a s))=
               (config M (power M MO  a n) s))) ->
            (Spec (Call_M n) c )

 |   c_spec:(p,q:nat)(c:Code)
               ((M:Set)(MO:(monoid M))
               ((a:M)(s:(Stack M))
                 (Exec M MO c (config M a s))=
                 (config M (power M MO a p) s))) ->
             (Spec (Call_C p q) c)

 |   k_spec:(p,q:nat)(c:Code)
                ((M:Set)(MO:(monoid M))
                ((a:M)(s:(Stack M))
                  (Exec M MO c (config M a s))=
                  (config M (power M MO a p) 
                            (push M (power M MO a q) s)))) ->
              (Spec (Call_K p q) c).


(* code generation  for each call
*********************************)

Inductive gencode[c:Call]:Set :=
   gencode_intro:(co:Code)(Spec c co)->(gencode c).


(* specification of the powering algorithm 
******************************************)

Inductive  addchain_spec[n:nat]:Set :=
  addchain_spec_intro:(gencode (Call_M n))->(addchain_spec n).




