(*  ---                        two_power.v                             ---  *)
(*                                                                          *)
(* Author: Pierre Casteran.                                                 *)
(*    LABRI, URA CNRS 1304,                                                 *)
(*    Departement d'Informatique, Universite Bordeaux I,                    *)
(*    33405 Talence CEDEX,                                                  *)
(*    e-mail:  casteran@labri.u-bordeaux.fr                                 *)



(* 
   The integer powers of 2 are fundamental in computer science;
   so we wrote this module, the results of which are used mainly
   in the "log2_spec" module.
*)



AddPath ".".
Require Arith.
Require Peano_dec.
Require Constants.
Require Mult_compl.
Require Le_lt_compl.


(* 2 at the n *)
(**************)

Recursive Definition two_power:nat->nat:=
            O => (S O)
          | (S n) => (mult two (two_power n)).


Lemma zero_lt_pow:(a:nat)(lt O (two_power a)).
(********************************************)
Proof.
 Induction a;Simpl.
 Auto.
 Intros.
 Apply lt_plus_trans;Auto.
Qed.
Hint zero_lt_pow.


Lemma two_power_S:(n:nat)(two_power (S n))=
                         (mult  (two_power n) two).
(**************************************************)
Proof.
 Intros;Unfold 1 two_power;Rewrite mult_sym;Auto.
Qed.


Lemma two_power_S_lt:(n:nat)(lt (two_power n) (two_power (S n))).
(***************************************************************)
Proof.
 Intro;Rewrite two_power_S.
 Rewrite mult_sym;Apply mult_p_lt_qp;Auto.
Qed.


Lemma two_power_mono:(n,p:nat)(lt n p)->
                              (lt (two_power n) (two_power p)).
(*************************************************************)
Proof.
 Induction 1.
 Intros;Apply two_power_S_lt;Auto.
 Intros.
 Apply lt_trans with (two_power m);Auto.
 Apply two_power_S_lt;Auto.
Qed.


Lemma two_power_le:(n,p:nat)(le n p)->
                              (le (two_power n) (two_power p)).
(**************************************************************)
Proof.
 Intros n p H;Case (le_lt_or_eq ? ? H);Intro H1.
 Apply lt_le_weak;Apply two_power_mono;Auto.
 Rewrite H1;Auto.
Qed.

Lemma two_power_monoR:(n,p:nat)(lt (two_power n) (two_power p))->
                               (lt n p).
(****************************************************************)
Proof.
 Intros.
 Apply not_le_lt.
 Unfold not;Intro.
 Elim (lt_not_le (two_power n) (two_power p) H).
 Case (le_lt_or_eq p n H0);Intro.
 Apply lt_le_weak.
 Apply two_power_mono;Auto.
 Auto.
 Rewrite -> H1;Auto.
Qed.


(* if 2^n < z <2^(n+1), then z cannot be a power of 2 ! *)
(*********************************************************)

Lemma holes_in_powers:(n:nat)(z:nat)(lt (two_power n) z)->
                                    (lt z (two_power (S n)))->
		                    (p:nat)~z=(two_power p).
(**************************************************************)
Proof.
 Unfold not;Intros.
 Cut (lt n p).
 Intro.
 Cut (lt p (S n)).
 Intro.
 Case (lt_not_lt n p);Auto.
 Apply two_power_monoR.
 Elim H1;Auto.
 Apply two_power_monoR.
 Elim H1;Auto.
Qed.


(* the intervals ]2^n,2^(n+1)[ are pairwise disjoint *)
(*****************************************************)

Lemma log2_unicity:(n,p:nat)(z:nat)(lt (two_power n) z)->
                                   (lt z (two_power (S n)))->
		                   (lt (two_power p) z)->
                                   (lt z (two_power (S p)))->
                                    n=p.
Proof.
 Intros.
 Case (le_or_lt n p);Intros.		                     
 Case (le_lt_or_eq n p H3);Intros.
 Absurd (lt z z).
 Auto.
 Apply lt_trans with (two_power (S n)).
 Auto.
 Apply le_lt_trans with (two_power p).
 Apply two_power_le;Auto.
 Auto.
 Auto.
 Absurd (lt z z).
 Auto.
 Apply lt_trans with (two_power (S p)).
 Auto.
 Apply le_lt_trans with (two_power n).
 Apply two_power_le;Auto.
 Auto.
Qed.
