(****************************************************************************)
(* This contribution was updated for Coq V5.10 by the COQ workgroup.        *)
(* January 1995                                                             *)
(****************************************************************************)
(*   script coq V5.8.3                                                      *)
(*                                                                          *)
(*                                                                          *)
(*   An exercise on groups                                                  *)
(*                                                                          *)
(*                                                                          *)
(*   Pierre Casteran.                                                       *)
(*   LaBRI [casteran@labri.u-bordeaux.fr]                                   *)
(*                                                                          *)
(*                                                                          *)
(*  This file presents a development in Coq of the following lemma:         *)
(*  "Let E be a non-empty set, o an associative law on E,                   *)
(*   such that for each x:E, the functions [y:E](o x y) and                 *)
(*   [y:E](o y x) are onto.                                                 *)
(*   Then E is a group"                                                     *)
(*                                                                          *)
(*   We use in this development two Skolem functions, in order to           *)
(*   "extract" the neutral element and the symetric function from           *)
(*   the hypothesis, and use them as global variables e and s in the        *)
(*   rest of the proof.                                                     *)
(*                                                                          *)
(****************************************************************************)
(*                                 Groups.v                                 *)
(****************************************************************************)

(*****************************
 Preliminaries : Skolemization
 ******************************)


Section Skolem0.

 Variable A:Set.
 Variable P:A->Prop.
 Definition skolem0:=[x0:{x:A|(P x)}]
                    (<A>Case x0 of [x:A][q:(P x )]x end).

(* Usage:
  Let us consider a lemma L:{x:A|(P x)}, you can directly
  type:
  Definition x :=(skolem0 A P L).
  to define x as a global variable for further usage
*)

(*****************************************************)
Lemma skolem0_lemma: (x0:{x:A|(P x)})(P (skolem0 x0)).
(*****************************************************)
Intro x0. Elim x0. Auto.
Save.
Hint skolem0_lemma.

End Skolem0.



Section Skolem1.
 Variable A:Set.
 Variable B:Set.
 Variable P:A->B->Prop.
 Variable x:A.

Definition skolem1:=[y0:{y:B|(P x y)}]
                  (<B>Case y0 of [y:B][q:(P x y)]y end).

(* Usage:
  Let us consider a lemma L:(x:A){y:B|(P x y)}, you can directly
  type:
  Definition f :=[x:A](skolem1 A B P x (L x)).
  to define f as a global variable for further usage
*)

(********************************************************)
Lemma skolem1_lemma: (y0:{y:B|(P x y)})(P x (skolem1 y0)).
(********************************************************)
Intro y0. Elim y0. Auto.
Save.

End Skolem1.


(* Basic Definitions and results on groups  *)


Section Groups.

Variable E:Set.
Variable o:E->E->E. 
Definition assoc := (x,y,z:E)(o x (o y z))=(o (o x y) z).
Variable s:E->E.
Variable e:E.
Definition left_neutral := (x:E)(o e x)=x.
Definition right_neutral := (x:E)(o x e)=x.
Definition left_sym := (x:E)(o (s x) x)=e.
Definition right_sym := (x:E)(o x (s x))=e.
Definition group := left_neutral/\right_neutral/\assoc/\left_sym/\right_sym.
Hint Unfold left_neutral right_neutral assoc left_sym right_sym.

(********************************************************)
Lemma unicsym: group -> (x,y:E)(o x y)=e -> y=(s x).
(********************************************************)

Intro G;Elim G.Intro left_n.Intro H0;Elim H0.Intro right_n.Intro H1;Elim H1.
Intro ass.Intro H2;Elim H2;Intros left_s right_s x y H.Clear G H0 H1 H2.
Transitivity (o e y).Symmetry;Auto.Transitivity (o (s x) (o x y)).
Transitivity (o (o (s x) x) y).Replace (o (s x) x) with e;Auto.
Auto.Rewrite H;Auto.
Save.


(***************************************)
Lemma ssx: group -> (x:E)(s (s x))=x.
(***************************************)
Unfold group.
Unfold left_neutral right_neutral assoc left_sym right_sym.
Intro G.Elim G.Intro left_n.Intro H0;Elim H0.Intro right_n.Intro H1;Elim H1.
Intro ass.Intro H2;Elim H2;Intro left_s.Intros right_s x.Clear  H0 H1 H2.
Symmetry;Apply unicsym;Auto.
Save.

End Groups.



(****************************************
  Now we are ready  to define the problem
******************************************)

(*
Please notice that the definition of "onto" uses the "{x:E " contruct
and NOT the <E>Ex(...) existential. This is necessary for the further
extraction of e and s.

*)

Parameter E:Set.
Definition onto := [f:E->E](y:E){x:E | (f x)=y}. 
Parameter a:E. (* so E is non empty *)
Parameter o:E->E->E.
Axiom o_assoc:(assoc E o). 
Axiom o_onto_r:(x:E)(onto [y:E](o x y)).
Axiom o_onto_l:(x:E)(onto  [y:E](o y x)).
Hint o_assoc o_onto_r o_onto_l.



(**********************************
 Building the neutral element ...
************************************)


(*****************************)
Lemma Ea: {ea:E | (o a ea)=a}.
(*****************************)

Elim (o_onto_r a a); Intros x p; Exists x ;Auto.
Save.

(*******************************)
Lemma E'a: {e'a:E | (o e'a a)=a}.
(*******************************)

Elim (o_onto_l a a); Intros x p; Exists x; Auto.
Save.


(* so we have a left_neutral element ...
***************************************)
Lemma N_left: {e:E | (left_neutral E o e)}.
(**************************************)

Elim E'a.Unfold left_neutral;Intros e'a eg.
 Exists e'a;Intro x.Elim (o_onto_r a x).Intros u eg'.
Elim eg'.Elim eg'.
Transitivity (o (o e'a a) u).Auto.
Replace (o e'a a) with a;Auto.
Save.

(* same stuff for ea  (at the right side ) 
******************************************)
Lemma N_right: {e:E | (right_neutral E o e)}.
(*****************************************)

Elim Ea.Unfold right_neutral;Intros ea eg.
 Exists ea;Intro x.Elim (o_onto_l a x).Intros u eg'.
Elim eg'.Elim eg'.
Transitivity (o u (o a ea));Auto.
Replace (o a ea) with a;Auto.
Save.


(* So we deduce that there exists a neutral element 
**************************************************************)
Lemma Neutral: {e:E | (left_neutral E o e)/\(right_neutral E o e)}.
(*************************************************************)

Elim N_left.Intros e He.Elim N_right;Intros e' He'.
Exists e;Split;Auto.Cut <E>e=e'.Intro eg;Auto.
Rewrite eg;Auto.Transitivity (o e e');Auto.
Save.


(* we now apply the skolem function skolem0
**********************************************************************)
Definition e := (skolem0 E [x:E](left_neutral E o x)/\(right_neutral E o x)
                       Neutral).
(**********************************************************************)


(*********************************************************)
Lemma e_neutral: (left_neutral E o e)/\(right_neutral E o e).
(*********************************************************)

Unfold e.Pattern 1 2 (skolem0 E [x:E](left_neutral E o x)/\
                                     (right_neutral E o x) Neutral).
Apply skolem0_lemma.
Save.


(***********************************************************
 we can now use the new element e to study the symmetrical 
***********************************************************)


(*********************************)
Lemma lsym: (x:E){y:E | (o y x)=e}.
(*********************************)

Intro x. Elim (o_onto_l x e). Intros u H;Exists u;Auto.
Save.

(*********************************)
Lemma rsym: (x:E){y:E | (o x y)=e}.
(*********************************)

Intro x. Elim (o_onto_r x e). Intros u H;Exists u;Auto.
Save.


(**************************************************************)
Lemma sym_construction: (x:E){y:E | (o x y)=e/\(o y x)=e}.
(**************************************************************)

Intro x.Elim (rsym x).Intros x' eg'.Elim (lsym x);Intros x'' eg''.
Exists x'.Pattern 2 x';Replace x' with x'';Auto.
Transitivity (o e x'). Replace e with (o x'' x).
Transitivity (o x'' (o x x'));Auto.
Rewrite eg'. Elim e_neutral;Intros H H0.
 Symmetry;Elim (H0 x'');Auto.
Elim e_neutral;Intros H H0.
 Elim (H x');Auto.
Save.


(**********************************************
  We can now define the function s (with skolem1)
***********************************************)

Definition s:=[x:E](skolem1 E E 
                    [x0:E][y:E]
                     ((o x0 y)=e)/\((o y x0)=e)
                    x (sym_construction x)).

 

(******************************************************)
 Lemma s_thm: (x:E)(o x (s x))=e/\(o (s x) x)=e.
(******************************************************)

Proof 
[x:E]
 (skolem1_lemma E E [x0:E][y:E]((o x0 y)=e)/\((o y x0)=e) x
    (sym_construction x)). 



(************************************************************
The final result (Please note that we need the symbols s and e
to express that we have a group *)

Theorem E_is_a_group: (group E o s e).
(************************************)

Unfold group.Split. Elim e_neutral;Auto.Split.Elim e_neutral;Auto.
Split.Exact o_assoc.Split.Unfold left_sym.Intro x. Elim (s_thm x).Intro.Auto.
Unfold right_sym.Intro x. Elim (s_thm x).Intro.Auto.
Save.
