(****************************************************************************)
(*                                                                          *)
(*                                                                          *)
(* BINARY SEARCH TREES    (V5.10)                                           *)
(*                                                                          *)
(* Pierre Casteran                                                          *)
(*                                                                          *)
(* LaBRI, Universite Bordeaux-I                                             *)
(* 351 Cours de la Liberation                                               *)
(* F-33405 TALENCE Cedex                                                    *)
(* France                                                                   *)
(*                                                                          *)
(* casteran@labri.u-bordeaux.fr                                             *)
(*                                                                          *)
(*                                                                          *)
(*     I Introduction                                                       *)
(*                                                                          *)
(*                                                                          *)
(*                                                                          *)
(*      In this file, we prove some basic algorithms on binary search       *)
(*      trees labeled with natural integers: searching, insertion, and      *)
(*      deletion.                                                           *)
(*      The main features used here are: dependent types of the form        *)
(*      {x:A|(P x)}, the Program tactic, and specification "a la Prolog"    *)
(*      of the preceeding operations.                                       *)
(*                                                                          *)
(*                                                                          *)
(*                                                                          *)
(****************************************************************************)

Require Peano_dec. (* for some decision lemmas *)
Require Compare_dec. (* idem *)
Require Lt.


(*       II The binary tree data structure
******************************************
******************************************
 

   We define here the "bintree" inductive data type.
   In the remaining of this file, the term "binary tree" will
   denote a complete binary tree, the internal nodes of which are
   labeled by natural integers.
*)

Inductive bintree : Set:=
   NIL : bintree
 | bin : nat->bintree->bintree->bintree.

(*
    Example:


         3
       /  \
      /    \
     7      8          Coq representation:
    / \    / \
   NIL \  NIL NIL      (bin three (bin seven NIL (bin nine NIL NIL))
        9                         (bin eight NIL NIL))
       / \             where three is the term (S (S (S O))) and so on ...
      /   \
     NIL  NIL
*)


(* II.1   Some definitions and lemmas 
   ************************************)


(* (binp t) means "t is not NIL"  *)

Inductive binp : bintree->Prop :=
 bin_bin : (n:nat)(t1,t2:bintree)(binp (bin n t1 t2)).

Hint bin_bin.


Lemma NIL_not_bin : ~(binp NIL).
(******************************)
Proof.
Unfold not; Intros H.
Inversion_clear H.
Qed.
Hint NIL_not_bin.


Lemma not_bin_NIL : (t:bintree)~(binp t)->t=NIL.
(**********************************************)
Proof.
Induction t.
Auto.
Intros.
Absurd (binp (bin n b b0)); Auto.
Qed.


Lemma diff_nil_bin : (n:nat)(t1,t2:bintree)~(bin n t1 t2)=NIL.
(************************************************************)
Proof.
Intros; Discriminate.
Qed.

Hint diff_nil_bin.


(* binp is decidable *)
Lemma binp_dec : (t:bintree){(binp t)}+{~(binp t)}.
(*************************************************)
Proof.
Induction t.
Right; Auto.
Intros; Left; Auto.
Qed.


(* II.2 membership
*******************************************
(has t p) means: "t has an occurrence of p" *)

Inductive  has : bintree->nat->Prop :=
   has_root : (n:nat)(t1,t2:bintree)(has (bin n t1 t2) n)
 | has_l    : (n,p:nat)(t1,t2:bintree)(has t1 p)->(has (bin n t1 t2) p)
 | has_r    : (n,p:nat)(t1,t2:bintree)(has t2 p)->(has (bin n t1 t2) p).

Hint has_root has_l has_r.

Derive Inversion_clear HAS_INV with 
    (n,p:nat)(t1,t2:bintree)(has (bin n t1 t2) p).

(* Coq < Check HAS_INV.
HAS_INV : (P:nat->nat->bintree->bintree->Prop)
        (n,p:nat)(t1,t2:bintree)
            (P p p t1 t2)
             ->((has t1 p)->(P n p t1 t2))
                ->((has t2 p)->(P n p t1 t2))
                   ->(has (bin n t1 t2) p)->(P n p t1 t2) *)

Lemma has_inv : (n,p:nat)(t1,t2:bintree)(has (bin n t1 t2) p) ->
                n=p \/ (has t1 p) \/ (has t2 p).
(**********************************************)
Proof.
Intros.
Use Inversion H HAS_INV; Auto.
Qed.

Hint has_inv.

Lemma not_has_nil : (p:nat)~(has NIL p).
(**************************************)
Proof.
Unfold not; Intros p H.
Inversion_clear H.
Qed.

Hint not_has_nil.


(* Decidability of predicate "has".
   Please notice that the following program realizes the decidability of
   has for general binary trees. In the case of search trees, we have a 
   more efficient program (search_has_dec). *)


Lemma bin_has_dec : (t:bintree)(n:nat){(has t n)}+{~(has t n)}.
(*************************************************************)
Proof.
Realizer [t:bintree][n:nat]
           (<bool> Match t with false
                                [p:nat][t1:bintree][b1:bool]
                                       [t2:bintree][b2:bool]
                          <bool> if (eq_nat_dec n p)
                                 then true
                                 else <bool> if b1 then true else b2 end).

Program_all. 
Rewrite e; Auto. 
Unfold not; Intro H.
Elim (has_inv ? ? ? ? H).
Auto.
Induction 1; Auto.
Qed.



(*   III  Binary search trees 
***********************************
***********************************)

(* III.1  Auxiliary definitions 
*******************************)

(* p is less than every member of t *)

Inductive min [p:nat; t:bintree] : Prop :=
  min_intro : ((q:nat)(has t q)->(lt p q)) -> (min p t).

Hint min_intro.

(* p is greater than every member of t *)

Inductive maj [p:nat; t:bintree] : Prop :=
  maj_intro : ((q:nat)(has t q)->(lt q p)) -> (maj p t).

Hint maj_intro.


Lemma min_nil : (p:nat)(min p NIL).
(*********************************)
Proof.
Intro p; Apply min_intro.
Intros q H.
Inversion_clear H.
Qed.

Hint min_nil.


Lemma maj_nil : (p:nat)(maj p NIL).
(*********************************)
Proof.
Intro p; Apply maj_intro.
Intros q H.
Inversion_clear H.
Qed.

Hint maj_nil.


Lemma maj_not_has : (p:nat)(t:bintree)(maj p t)->~(has t p).
(**********************************************************)
Proof.
Intros p t H.
Unfold not; Intro H'. 
Elim H.
Intros.
Absurd (lt p p); Auto.
Qed.

Hint maj_not_has.

Lemma min_not_has : (p:nat)(t:bintree)(min p t)->~(has t p).
(**********************************************************)
Proof.
Intros p t H.
Unfold not; Intro H'. 
Elim H.
Intros.
Absurd (lt p p); Auto.
Qed.

Hint min_not_has.
 

(* III.2 Definition of being a search tree 
******************************************)
Inductive search : bintree->Prop :=
          nil_search : (search NIL)
        | bin_search : (n:nat)(t1,t2:bintree)
                (search t1) ->
                (search t2) ->
                (maj n t1) ->
                (min n t2) ->
                (search (bin n t1 t2)).

Hint nil_search bin_search.


(* III.3 Some useful lemmas about a given search tree (bin n t1 t2) 
*******************************************************************)

Section search_tree_basic_properties.

     Variable n:nat.
     Variable t1,t2:bintree.
     Hypothesis se:(search (bin n t1 t2)).

     Lemma search_l : (search t1).
     (**************************)
     Proof.
     Inversion_clear se; Auto.
     Qed.

     Hint search_l.

     Lemma search_r : (search t2).
     (***************************)
     Proof.
     Inversion_clear se; Auto.
     Qed.
     Hint search_r.

     Lemma maj_l : (maj n t1).
     (***********************)
     Proof.
     Inversion_clear se; Auto.
     Qed.

     Hint maj_l.

     Lemma min_r : (min n t2).
     (***********************)
     Proof.
     Inversion_clear se; Auto.
     Qed.
     Hint min_r.


     (* Exclusion lemmas *)

     Lemma not_right : (p:nat)(le p n)->~(has t2 p).
     (*********************************************)
     Proof.
     Intros p H.
     Elim min_r.
     Unfold not; Intros H0 H1.
     Absurd (lt n p); Auto.
     Qed.

     Lemma not_left : (p:nat)(le n p)->~(has t1 p).
     (*********************************************)
     Proof.
     Intros p H.
     Elim maj_l.
     Unfold not; Intros. 
     Absurd (lt p n); Auto.
     Qed.
     
     (* Lemmas for controling the research of p:nat *)

     Lemma go_left : (p:nat)(has (bin n t1 t2) p)->(lt p n)->(has t1 p).
     (*****************************************************************)
     Proof.
     Intros p H H0.  Elim (has_inv ? ? ? ? H).
     Induction 1; Absurd (lt p n); Auto. 
     Rewrite H1; Auto.
     Induction 1; Auto.
     Intro H2.Absurd (has t2 p); Auto.
     Apply not_right; Auto.
     Qed.

     Lemma go_right : (p:nat)(has (bin n t1 t2) p)->(lt n p)->(has t2 p).
     (******************************************************************)
     Proof.
     Intros p H H0. 
     Elim (has_inv n p t1 t2).
     Intro eg.
     Absurd (lt n p); Auto.
     Elim eg; Auto.
     Induction 1.
     Intro.
     Absurd (has t1 p); Auto.
     Apply not_left; Auto.
     Auto.
     Auto.
     Qed.
     
     
End search_tree_basic_properties.
(*******************************)


(* Inversion lemma:
   Usage: Apply search_inv with n t1 t2
***************************************)

Lemma search_inv : (n:nat)(t1,t2:bintree)(P:Prop)
            ((search t1)->(search t2)->(maj n t1)->(min n t2)->P) ->
            (search (bin n t1 t2)) -> P.
(****************************************************************)
Proof.
Intros n t1 t2 P H H0. 
Apply H.
Apply search_l with n t2; Auto.
Apply search_r with n t1; Auto.
Apply maj_l with t2; Auto.
Apply min_r with t1; Auto.
Qed.


(*    IV: Searching  in a binary search tree 
*********************************************
*********************************************)

(* we give the validation of the classical searching algorithm,
   which uses the go_left and go_right lemmas *)

Theorem search_has_dec : (p:nat)(t:bintree)
                         (search t) -> {(has t p)}+{~(has t p)}.
(**************************************************************)
Proof.
Realizer [p:nat][t:bintree]<bool>Match t with
                     (* NIL *) false
                     (* bin n t1 t2 *)
                      [n:nat][t1:bintree][b1:bool][t2:bintree][b2:bool]
                       <bool> if (le_gt_dec p n)
                           then <bool> if (le_lt_eq_dec p n)
                                       then b1 else true
                           else b2 end.
Proof.
Program_all. 
Apply search_l with n b1; Auto.
Unfold not; Intro H2.
Apply n0.
Apply go_left with n b1; Auto.
Rewrite e; Auto.
Apply search_r with n b0; Auto.
Unfold not; Intro H2.
Apply n0.
Apply go_right with n b0; Auto.
Qed.


(* V:        INSERTION 
**********************************
**********************************


V.1   Definition of an INSERT predicate (a la  Prolog)
***************************************************


We begin with the definition of a predicate:

(INSERT n t t') if t' is a binary search tree containing exactly
  n and the elements of t *)


Inductive INSERT[n:nat; t,t':bintree] : Prop :=
 insert_intro: ((p:nat)(has t p) -> (has t' p)) ->
               (has t' n) ->
               ((p:nat)(has t' p) -> (has t p) \/ n=p) ->
               (search t') ->
               (INSERT n t t').

Hint insert_intro.


Lemma insert_nil: (n:nat)(INSERT n NIL (bin n NIL NIL)).
(******************************************************)
Proof.
Intro n; Apply insert_intro; Auto.
Intros p H; Inversion_clear H; Auto.
Qed.

Hint insert_nil.



(* Inserting in the left son *)

Lemma insert_l : (n,p:nat)(t1,t'1,t2:bintree)
                   (lt n p) ->
                   (search (bin p t1 t2)) ->
                   (INSERT n t1 t'1) ->
                   (INSERT n (bin p t1 t2) (bin p t'1 t2)).
(**********************************************************)
Proof.
Intros n p t1 t'1 t2 H H0 H1.
Apply insert_intro.
Intros p0 H2.
Inversion_clear H2.
Auto.
Elim H1; Auto.
Auto.
Cut (has t'1 n).
Auto. 
Elim H1; Auto.
Intros p0 H2.
Inversion_clear H2.
Auto.
Elim H1; Intros. 
Elim (H5 p0); Auto.
Auto.
Apply bin_search. 
Elim H1; Auto.
Inversion_clear H0; Auto.
Elim H1; Intros.
Apply maj_intro. 
Intros q H6; Elim (H4 q).
Intro; Cut (maj p t1).
Induction 1; Auto.
Apply maj_l with t2; Auto.
Induction 1; Auto.
Auto.
Apply min_r with t1; Auto.
Qed.

(*  inserting in the right son *)

Lemma insert_r : (n,p:nat)(t1,t2,t'2:bintree)
                 (lt p n) ->
                 (search (bin p t1 t2)) ->
                 (INSERT n t2 t'2) ->
                 (INSERT n (bin p t1 t2) (bin p t1 t'2)).
(*******************************************************)
Proof.
Intros n p t1 t2 t'2 H H0 H1.
Apply insert_intro.
Intros p0 H2.
Inversion_clear H2.
Auto.
Elim H1; Auto.
Auto.
Apply has_r.
Elim H1; Auto. 
Apply has_r.
Elim H1; Auto.
Intros.
Elim (has_inv ? ? ? ? H2). 
Induction 1; Auto. 
Induction 1; Intro. 
Auto.
Elim H1; Intros. 
Elim (H7 p0); Auto.
Apply bin_search. 
Apply search_l with p t2; Auto. 
Elim H1; Auto.
Inversion_clear H0; Auto.
Apply min_intro.
Intros; Elim H1; Intros. 
Elim (H5 q).
Intro.Elim (min_r p t1 t2); Auto.
Induction 1; Auto.
Auto.
Qed.

(* no need for insertion  ! *)

Lemma insert_eq : (n:nat)(t1,t2:bintree)
                 (search (bin n t1 t2)) ->
                 (INSERT n (bin n t1 t2) (bin n t1 t2)).
(******************************************************)
Proof.
Auto.
Qed.

Hint insert_l insert_r insert_eq.


(* V.2   The insertion algorithm
*******************************)

(* specification *)

Definition INSERT_SPEC := [n:nat][t:bintree]
           {t':bintree | (INSERT n t t')}.

(* realization *)

Lemma insert_algo : (n:nat)(t:bintree)(search t)->(INSERT_SPEC n t).
Proof.
Realizer [n:nat][t:bintree]<bintree>Match t with
                (* NIL *) (bin n NIL NIL )
                (* bin  p t1 t2 *)
                 [p:nat][t1,r1,t2,r2:bintree]
                  <bintree> if (le_gt_dec n p)
                            then <bintree> if (le_lt_eq_dec n p)
                                           then  (bin p r1 t2)
                                           else (bin n t1 t2)
                            else (bin p t1 r2) end.

(******************************************************************)

Program_all.
Inversion_clear H; Auto.
Rewrite e; Auto.
Inversion_clear H; Auto.
Qed.
Hint insert_algo.


(* V.3  Construction of a binary search tree from a given list 
     of natural numbers 
**************************************************************)

(*  This section has an interesting computational content: 
    the predicate search (being of sort Prop) is never tested, 
    and the standard way of building search trees is to insert
     one by one the elements of a given list *)


(* we begin with a definition of lists of natural numbers ... *)

Inductive  natlist : Set :=
   emptylist:natlist
 | natcons: nat -> natlist -> natlist.


Inductive  member: nat -> natlist -> Prop :=
   incar : (n:nat)(l:natlist)(member n (natcons n l))
 | incdr : (n,p:nat)(l:natlist)(member n l)->(member n (natcons p l)).

Hint incar incdr.


Lemma not_member_empty : (p:nat)~(member p emptylist).
(****************************************************)
Proof.
Unfold not; Intros p H; Inversion_clear H.
Qed.

Hint not_member_empty.


Lemma member_inv : (n,p:nat)(l:natlist)
     (member p (natcons n l)) -> n=p \/ (member p l).
(***************************************************)
Proof.
Intros n p l H.
Inversion_clear H; Auto.
Qed.

Hint member_inv.

(* construction of a search tree containing all elements
of a given list of natural numbers 
********************************************************)

(* In order to have a tail recursive program , we need an auxiliary function *)

Lemma const_aux : (l:natlist)(t:bintree)(search t) ->
                  {t':bintree | (search t' ) /\ 
                  (p:nat) (iff (or (member p l) (has t p))
                               (has t' p))}.

(* Computation scheme: 
   (const_aux () t)= t
   (const_aux (a::l) t)=(const_aux l (insert_algo a t))
*)
Proof.
Realizer [l:natlist]<bintree->bintree>Match l with
      (* emptylist *) [t:bintree]t
      (* hd . tl *)   [hd:nat][tl:natlist][R:bintree->bintree]
		      [t:bintree](R (insert_algo hd t)) end.
Program_all.
Split.
Auto.
Intro p; Unfold iff; Split; Intro H0.
Elim H0; Auto.
Intro; Absurd (member p emptylist); Auto.
Auto.
Elim y; Auto.
Split; Elim a; Auto.
Intros.
Unfold iff.
Split. 
Intros.
Elim H2; Intros.
Inversion_clear H3.
Elim (H1 n0); Intros.
Apply H3. Right; Elim y; Auto.
Elim (H1 p); Auto.
Elim (H1 p); Intros.
Apply H4.
Right; Elim y; Auto.
Intro H2.
Elim (H1 p); Intros.
Elim (H4 H2); Intros. 
Left; Auto. 
Elim y; Intros.
Elim (H8 p).
Auto.
Induction 1; Auto.
Auto.
Qed.

Theorem const_tail :
  (l:natlist) {t:bintree | (search  t) /\
                           (p:nat)(iff (member p l) (has t p))}.
Proof.
Realizer [l:natlist](const_aux l NIL).

Program_all.
Elim a; Split; [Auto|Idtac].
Intro p0; Unfold iff; Split; Intros.
Elim (H0 p0); Intros. 
Apply H2; Auto.
Elim (H0 p0); Intros. 
Elim (H3 H1); Auto.
Intro; Absurd (has NIL p0); Auto.
Qed.


(* VI    Deletion.
******************
*******************


  Deleting an element from a binary search tree is a little more
  complex than inserting or searching.
  The difficult case is the deletion of the root of a tree; we have
  to reconstruct a search tree. To solve this problem, we define
  an auxiliary operation: deleting the greatest element of a non-empty
  binary search tree.



VI.1  Removing the greatest element of a given search tree
***********************************************************


  t' is obtained by removing n: the greatest element ot t 

*)

Inductive RMAX [t,t':bintree; n:nat] : Prop :=
 rmax_intro : 
             (has t n) -> 
             ((p:nat)(has t p)->(le p n)) ->
             ((q:nat) (has t' q)->(has t q)) ->
             ((q:nat)(has t q)->(has t' q) \/ n=q) ->
             (~(has t' n)) ->
	     (search t') ->
             (RMAX t t' n).

Hint rmax_intro.


(* base cases *)

Lemma rmax_nil_nil : (n:nat)(RMAX (bin n NIL NIL) NIL n).
(*******************************************************)
Proof.
Intro n; Apply rmax_intro; Auto.
Intros p H; Inversion_clear H; Auto.
Absurd (has NIL p); Auto.
Absurd (has NIL p); Auto.
Intros q H; Inversion_clear H; Auto.
Qed.


Lemma rmax_t_NIL : (t:bintree)(n:nat)(search (bin n t NIL)) ->
                                     (RMAX (bin n t NIL) t n).
(************************************************************)
Proof.
Intros t n H; Apply rmax_intro; Auto. 
Intros p H0.
Elim (has_inv n p t NIL H0); Intro H1.
Elim H1; Auto. 
Elim H1; Intro H2.
Apply lt_le_weak. 
Elim (maj_l n t NIL H); Auto.
Absurd (has NIL p); Auto.
Intros q H1; Elim (has_inv n q t NIL H1); Intros; Auto.
Elim H0.
Auto.
Intro H'; Absurd (has NIL q); Auto.
Apply not_left with n NIL; Auto.
Apply search_l with n NIL; Auto.
Qed. 

Hint rmax_t_NIL.


(* We study the case of a search tree (bin n t1 (bin p t2 t3))  *)

Section RMAX_np.
     Variable n,p,q : nat.
     Variable t1,t2,t3,t' : bintree.
     Hypothesis S1 : (search (bin n t1 (bin p t2 t3))).
     Hypothesis R1:(RMAX (bin p t2 t3) t' q).
     Hint S1 R1.
     
     Lemma rmax_1 : (has (bin n t1 (bin p t2 t3)) q).
     (**********************************************)
     Proof.
     Elim R1; Auto.
     Qed.

     Hint rmax_1.

     Lemma rmax_2 : (lt n p).
     (**********************)
     Proof.
     Elim (min_r ? ?  ? S1); Auto.
     Qed.

     Hint rmax_2.
     
     Lemma rmax_3 : (min n t'). 
     (************************)
     Proof.
     Apply min_intro.
     Intros q' H.
     Elim R1; Intros.
     Elim (min_r ? ? ? S1); Auto.
     Qed.
     Hint rmax_3.
     
     Lemma rmax_4 : (search (bin n t1 t')).
     (************************************)
     Proof.
     Apply bin_search.
     Apply search_l with n (bin p t2 t3); Auto. 
     Elim R1; Auto.
     Apply maj_l with (bin p t2 t3); Auto.
     Auto.
     Qed. 

     Hint rmax_4.
     
     Lemma rmax_5 : (lt n q).
     (**********************)
     Proof.
     Elim R1; Intros; Apply lt_le_trans with p; Auto.
     Qed.

     Hint rmax_5.

     Lemma rmax_6 : (p0:nat)(has (bin n t1 (bin p t2 t3)) p0)->(le p0 q).
     (******************************************************************)
     Proof.
     Intros p0 H.
     Elim R1.
     Intros H0 H1 H2 H3 H4 H5.
     Elim (has_inv ? ? ? ? H); Intro H6.
     Elim H6; Auto.
     Elim H6; Intro H7.
     Elim (maj_l ? ? ? S1). 
     Intro H8.
     Cut (lt p0 n); Auto.
     Intro.Apply lt_le_weak.
     Apply lt_trans with n; Auto.
     Elim (min_r ? ? ? S1); Auto.
     Qed.

     Hint rmax_6.
     
     Lemma rmax_7 : (q':nat)(has (bin n t1 t') q') ->
                            (has (bin n t1 (bin p t2 t3)) q').
     (*******************************************************)
     Proof.
     Intros q' H.Elim (has_inv ? ? ? ? H); Intro H0.
     Elim H0; Auto.
     Elim H0; Auto. 
     Intro H1 .Elim R1; Auto.
     Qed.

     Hint rmax_7.
     
     Lemma rmax_8 : ~(has (bin n t1 t') q).
     (************************************)
     Proof.
     Unfold not; Intro F.
     Elim (has_inv ? ? ? ? F).
     Intro eg; Absurd (lt n q); Auto.
     Elim eg; Auto.Induction 1; Intro H1.
     Absurd (has t1 q); Auto. 
     Apply not_left with n (bin p t2 t3); Auto.
     Elim R1; Auto.
     Qed.

     Hint rmax_8.

     Lemma rmax_9 : (q0:nat)
        (has (bin n t1 (bin p t2 t3)) q0) ->
        ((has (bin n t1 t') q0) \/ q=q0).
     (****************************************)
     Proof.
     Intros q0 H.
     Elim (has_inv ? ? ? ? H).
     Induction 1; Left; Auto.
     Induction 1; Intro H'.
     Left; Auto.
     Elim R1; Intros H1 H2 H3 H4 H5 H6.
     Elim (H4 ? H'); Auto.
     Qed.

     Hint rmax_9.

     Lemma rmax_t1_t2t3 : (RMAX (bin n t1 (bin p t2 t3)) (bin n t1 t') q).
     (*******************************************************************)
     Proof.
     Apply rmax_intro; Auto.
     Qed.

End RMAX_np.

Hint rmax_t1_t2t3.


Lemma rmax_algo : (t:bintree)
        (search t) ->
        (binp t) ->
        {q:nat & {t':bintree | (RMAX t t' q)}}.
(*********************************************)
Proof.
Realizer [t:bintree]<nat*bintree>Match t with
                      (* NIL *) (O,NIL) (* failure *)
                      (* n t1 t2 *)
                         [n:nat][t1:bintree][r1:nat*bintree]
                                [t2:bintree]
                         <nat*bintree->nat*bintree>Match t2 with 
                             [r2:nat*bintree](n,t1)
                             [n':nat]
                             [t'1:bintree][r'1:nat*bintree->nat*bintree]
                             [t'2:bintree][r'2:nat*bintree->nat*bintree]
			     [r2:nat*bintree]
			      <nat*bintree>let (num,tr) = r2 in
					(num,(bin n t1 tr)) end
                         end.
Program_all. 
Absurd (binp NIL); Auto.
Apply search_r with n b0; Auto.
Qed.


(* VI.2  Deletion in general
****************************

  We are now ready to study the remove operation in it's generality:
 (RM n t t') if t' is a search tree obtained by removing n from t *)


Inductive  RM [n:nat; t,t':bintree] : Prop :=
   rm_intro : ~(has t' n) ->
              ((p:nat)(has t' p) -> (has t p)) ->
              ((p:nat)(has t p) -> (has t' p) \/ n=p) ->
              (search t') -> (RM n t t').

Hint rm_intro.


(* base cases *)

Lemma RM_0 : (n:nat)(RM n NIL NIL).
(*********************************)
Proof.
Intro n; Apply rm_intro; Auto.
Qed.

Hint RM_0.


Lemma RM_1 : (n:nat)(RM n (bin n NIL NIL) NIL).
(*********************************************)
Proof.
Intros; Apply rm_intro; Auto.
Intros p H; Elim (has_inv n ? ? ? H); Auto.
Tauto.
Qed.

Hint RM_1.


(* deleting in the left son *)

Lemma rm_left : (n,p:nat)(t1,t2,t':bintree)
                (lt p n) ->
                (search (bin n t1 t2)) ->
                (RM p t1 t') ->
                (RM p (bin n t1 t2) (bin n t' t2)).
(*************************************************)
Proof.
Intros n p t1 t2 t' H H0 H1.
Apply rm_intro.Unfold not; Intro H2.
Elim (has_inv n p t' t2).
Intro eg; Absurd (lt p p); Auto.
Pattern 2 p; Elim eg; Auto.
Intro D; Elim D; Intro H3. 
Elim H1; Auto.
Absurd (has t2 p); Auto.Apply not_right with n t1; Auto.
Auto.
Intros p0 H2.
Elim (has_inv n p0 t' t2).
Induction 1; Auto.
Induction 1; Auto.
Intro; Elim H1; Auto.
Auto.
Intros.
Elim (has_inv n p0 t1 t2).
Induction 1; Auto.
Induction 1; Intro H4.
Elim H1.
Intros H5 H6 H7 H8.
Elim (H7 p0 H4); Auto.
Auto.
Auto.
Apply bin_search.
Elim H1; Auto.
Apply search_r with n t1; Auto.
Apply maj_intro; Intros q H2.
Cut (has t1 q).
Intro; Elim (maj_l n t1 t2 H0); Intros; Auto.
Auto.
Elim H1; Auto.
Apply min_r with t1; Auto.
Qed.

Hint rm_left.


(* deleting in the right son *)

Lemma rm_right : (n,p:nat)(t1,t2,t':bintree)
                 (lt n p) ->
                 (search (bin n t1 t2)) ->
                 (RM p t2 t') ->
                 (RM p (bin n t1 t2) (bin n t1 t')).
(**************************************************)
Proof.
Intros n p t1 t2 t' H H0 H1.
Apply rm_intro.
Unfold not; Intro H2.
Elim (has_inv n p t1 t').
Intro eg; Absurd (lt p p); Auto.
Pattern 1 p; Elim eg; Auto.
Intro D; Elim D; Intro H3. 
Elim H1; Auto.
Absurd (has t1 p).
Apply not_left with n t2; Auto.
Auto.
Elim H1; Auto.
Auto.
Intros p0 H2.
Elim (has_inv n p0 t1 t').
Induction 1; Auto.
Induction 1; Auto.
Intro; Elim H1; Auto.
Auto.
Intros.
Elim (has_inv n p0 t1 t2).
Induction 1; Auto.
Induction 1; Auto.
Intro H4.
Elim H1; Intros H5 H6 H7 H8.
Elim (H7 p0 H4); Auto.
Auto.
Apply bin_search.
Apply search_l with n t2; Auto.
Elim H1; Auto.
Apply maj_l with t2; Auto.
Apply min_intro; Intros q H2.
Cut (has t2 q).
Intro.
Elim (min_r n t1 t2 H0); Auto. 
Elim H1; Auto.
Qed.

Hint rm_right.


(* base case for deleting the root *)

Lemma rm_NILt : (n:nat)(t:bintree)(search (bin n NIL t)) ->
                                  (RM n (bin n NIL t) t).
(*******************************************************)
Proof.
Intros; Apply rm_intro.
Apply not_right with n NIL; Auto.
Auto.
Intros p H1; Elim (has_inv n p NIL t H1); Intro H2.
Right; Auto.
Elim H2; Intro.
Absurd (has NIL p); Auto.
Left; Auto.
Apply search_r with n NIL; Auto.
Qed.

Hint rm_NILt.


(* General case: we use the RMAX predicate *)

Section rm_t1t2t3.
     Variable n,p:nat.
     Variable t1,t2,t3:bintree.
     Hypothesis S: (search (bin n (bin p t1 t2) t3)).
     Variable q:nat.
     Variable t': bintree.
     Hypothesis R: (RMAX  (bin p t1 t2) t' q).
     Hint S.
     
     Lemma rm_2 : (lt q n).
     (********************)
     Proof.
     Elim R.
     Intros.
     Elim (maj_l n (bin p t1 t2) t3).
     Auto.
     Auto.
     Qed.

     Hint rm_2.
     
     Lemma rm_3 : ~(has (bin q t' t3) n).
     (**********************************)
     Proof.
     Unfold not; Intro H.
     Elim (has_inv q n t' t3).
     Intro eg; Absurd (lt q q); Auto.
     Pattern 2 q; Rewrite eg; Auto.
     Intro D; Elim D; Intro H'.
     Elim R; Intros H0 H1 H2 H3 H4 H5.
     Absurd (has (bin p t1 t2) n); Auto.
     Apply not_left with n t3; Auto.
     Absurd (has t3 n); Auto.
     Apply not_right with n (bin p t1 t2); Auto.
     Auto.
     Qed.

     Hint rm_3.
     

     Lemma rm_4 : (p0:nat)
         (has (bin q t' t3) p0) -> (has (bin n  (bin p t1 t2) t3) p0).
     (***************************************************************)
     Proof. 
     Intros p0 H.
     Elim (has_inv q p0 t' t3 H).  
     Intro eg.
     Elim R; Rewrite eg; Auto.
     Induction 1; Auto.
     Intro H'.Elim R; Auto.
     Qed.

     Hint rm_4.
     

     Lemma rm_5 : (p0:nat)
                  (has (bin n (bin p t1 t2) t3) p0) ->
                  ((has (bin q t' t3) p0) \/ n=p0).
     (********************************************)
     Proof.
     Intros p0 H.
     Elim (has_inv n p0  (bin p t1 t2) t3). 
     Induction 1; Auto.
     Induction 1.
     Intro H1.
     Elim R; Intros H2 H3 H4 H5 H6 H7.
     Elim (H5 p0 H1).Intro; Left; Auto.
     Induction 1; Left; Auto.
     Intro; Left; Auto.
     Auto.
     Qed.

     Hint rm_5.
     

     Lemma rm_6 : (search (bin q t' t3)).
     (**********************************)
     Proof.
     Apply bin_search.
     Elim R; Auto.
     Apply search_r with n (bin p t1 t2); Auto.
     Elim R; Intros H H0 H1 H2 H3 H4.
     Apply maj_intro.
     Intros q0 H5; Elim (le_lt_or_eq q0 q (H0 q0 (H1 q0 H5))).
     Auto.
     Intro eg; Absurd (has  t' q0).
     Rewrite eg; Auto.
     Auto.
     Apply min_intro.
     Intros q0 H. 
     Apply lt_trans with n. 
     Elim R; Auto.
     Elim (min_r n (bin p t1 t2) t3). 
     Auto. 
     Auto.
     Qed.

     Hint rm_6.
     

     Lemma rm_t1t2t3_lemma : (RM n (bin n (bin p t1 t2) t3) (bin q t' t3)).
     (********************************************************************)
     Proof.
     Apply rm_intro; Auto.
     Qed.
     
End rm_t1t2t3.


(* The final algorithm *)

Theorem rm_algo : (t:bintree)(n:nat)
      (search t) -> {t':bintree | (RM n t t')}.
(*********************************************)
Proof.
Realizer [t:bintree][n:nat]
      (<bintree> Match t with
                NIL
                [p:nat][t1,r1,t2,r2:bintree]
	          <bintree> if (le_gt_dec n p)
			    then <bintree> if (le_lt_eq_dec n p)
				then (bin p r1 t2)
				else (<bintree> Case t1 of
                                   (* NIL *) t2
                                   (* q t'1 t'2 *)
                                   [p':nat][t'1,t'2:bintree]
                                    <bintree>let (q,t')= 
                                                 (rmax_algo (bin p' t'1 t'2))
                                             in (bin q t' t2) end)
                            else (bin p t1 r2) end).

Program_all.
Apply search_inv with n0 b0 b1; Auto.
Rewrite e; Apply rm_NILt; Auto.
Apply search_l with n0 b1; Auto.
Rewrite e; Apply rm_t1t2t3_lemma; Auto.
Apply search_r with n0 b0; Auto.
Qed.

