(****************************************************************************)
(*                                                                          *)
(* Jean-Francois.Monin@lannion.cnet.fr                                      *)
(* February 1995, Coq V5.10                                                 *)
(*                                                                          *)
(****************************************************************************)
(*                              end_term_unif.v                             *)
(****************************************************************************)

Require Arith.
Require nat_complements.
Require nat_term_eq_quasiterm.
Require is_in_quasiterm_term_subst.
Require listv_is_in_lv.
Require deb_term_unif.

(*****************************************************************)
(***************** Specification of the problem of the unification ******)
(*****************************************************************)
(********************* Unif defines the minimal unifiers.*************)
(*****************************************************************)

(* Only the success part *)
Inductive Unification_s[t1,t2:quasiterm]:Set:=
Unif_succeed_def:(f:quasisubst)
     (unif f t1 t2)
     ->(idempotent f)
     ->(over f (ConsArg t1 t2))
     ->(under f (ConsArg t1 t2))
     ->(min_unif f t1 t2)
(*------------ Verifications on the terms:begin -------------*)
(**) ->((l:list_nat)(L_TERM l t1)->(L_TERM l t2)
(**)     ->((Length t1)=(Length t2))->(term_subst l f))
(*------------ Verifications on the terms:end --------------*)
     ->(Unification_s t1 t2).

(* Failure part : only a Prop *)
Inductive Unification_f[t1,t2:quasiterm]:Prop:=
  Unif_fail_def:((f:quasisubst)~(Subst f t1)=(Subst f t2))
     ->(Unification_f t1 t2).

(* The following is equivalent to the original "Unification"  *)
(* an will be used only once (in unif_call) at the end of the *)
(* development.                                               *)
Inductive Unification_or_fail[t1,t2:quasiterm]:Set:=
   Unification_or_fail_succeed: 
        (Unification_s t1 t2) ->(Unification_or_fail t1 t2)
 | Unification_or_fail_fail: 
        (Unification_f t1 t2)   ->(Unification_or_fail t1 t2).

(*****************************************************************)
(************** Failure in the unification with the head symbol: *******)
(*****************************************************************)

Inductive head_diff:quasiterm->quasiterm->Prop:=
Fail_hd1:(l:fun)(t1,t2:quasiterm)(head_diff (C l) (ConsArg t1 t2))
|Fail_hd2:(l:fun)(t1,t2:quasiterm)(head_diff (ConsArg t1 t2) (C l))
|Fail_hd3:(l,l0:fun)(t:quasiterm)(head_diff (C l) (Root l0 t))
|Fail_hd4:(l,l0:fun)(t:quasiterm)(head_diff (Root l0 t) (C l))
|Fail_hd5:(t1,t2,t3:quasiterm)(l:fun)
            (head_diff (ConsArg t1 t2)(Root l t3))
|Fail_hd6:(t1,t2,t3:quasiterm)(l:fun)
            (head_diff (Root l t3)(ConsArg t1 t2)).

Require user_comarg.
Require gram1.
Require cps_Nx.

Section algorithmes.

Load DeclareNx.

(* New version of Unification with exceptions *)
Definition Unification := 
   [t1,t2:quasiterm] (Unification_f t1 t2) V+ (Unification_s t1 t2). 

(* An instance of identity, useful only for its type, cf. unif_call *)
Definition Unification_build :
  (t1,t2:quasiterm)
  (Unification t1 t2)->
     ((Unification_f t1 t2)->P)->((Unification_s t1 t2)->X)->X :=
  [t1,t2:quasiterm](Nx_handle (Unification_f t1 t2) (Unification_s t1 t2)).
   

Definition Nxunif_unit : 
  (t1,t2:quasiterm)(Unification_s t1 t2)->(Unification t1 t2) :=
  [t1,t2:quasiterm](Nx_unit (Unification_f t1 t2) (Unification_s t1 t2)).

Definition Nxunif_raise :
  (t1,t2:quasiterm)(Unification_f t1 t2)->(Unification t1 t2) :=
  [t1,t2:quasiterm](Nx_raise (Unification_f t1 t2) (Unification_s t1 t2)).

(* New version of Unif_succeed *)
Definition Unif_succeed :=
  [t1,t2:quasiterm][f:quasisubst]
     [u:(unif f t1 t2)][i:(idempotent f)][o:(over f (ConsArg t1 t2))]
     [u0:(under f (ConsArg t1 t2))][m:(min_unif f t1 t2)]
     [t:((l:list_nat)(L_TERM l t1)->(L_TERM l t2)
        ->((Length t1)=(Length t2))->(term_subst l f))]
   (Nxunif_unit t1 t2 (Unif_succeed_def t1 t2 f u i o u0 m t)).

(* New version of Unif_fail, to be used when raising an exception. *)
(* Warning: Unif_fail_def has to be used for transmitting an       *)
(* information "an exception must have been raised".               *)
Definition Unif_fail :=
  [t1,t2:quasiterm][e:((f:quasisubst)~(Subst f t1)=(Subst f t2))]
    (Nxunif_raise t1 t2 (Unif_fail_def t1 t2 e)).


Definition Unification_elim :
  (t1,t2,t1',t2':quasiterm)
    ((Unification_s t1 t2)->(Unification t1' t2'))->
    ((Unification_f t1 t2)->(Unification_f t1' t2'))->
        (Unification t1 t2)->(Unification t1' t2') :=
  [t1,t2,t1',t2':quasiterm]
    (Nx_elim_Nx (Unification_s t1 t2) (Unification_s t1' t2')
                (Unification_f t1 t2) (Unification_f t1' t2')).

(* Plays the role of Unification_rec;                *)
(* "Elim H" becomes "Elimunif H",                    *)
(* at least when H is not a product                  *)
(* In the development H is a product in only one     *)
(* instance, namely (H0 q1) in proof_unif;           *)
(* we can use a "Cut" in such situations.            *)
Goal
  (t1,t2,t1',t2':quasiterm)
    ((f:quasisubst)(u:(unif f t1 t2))
       (i:(idempotent f))(o:(over f (ConsArg t1 t2)))
         (u0:(under f (ConsArg t1 t2)))(m:(min_unif f t1 t2))
            (t:(l:list_nat)
                (L_TERM l t1)->
                 (L_TERM l t2)->
                  ((Length t1)=(Length t2))->(term_subst l f))
            (Unification t1' t2'))->
    ((n:(f:quasisubst)(~((Subst f t1)=(Subst f t2))))
            (Unification_f t1' t2'))->
        (Unification t1 t2)->(Unification t1' t2').
  Intros t1 t2 t1' t2' inh_s inh_f nu.
  Apply Unification_elim with 3:=nu.
    Intro u; Elim u; Exact inh_s.
    Intro u; Elim u; Exact inh_f.
Save Unif_elim.
Transparent Unif_elim.

Grammar tactic tactic1 :=
   [ "Elimunif" user_comarg($c) ] -> 
   [$0=<:tactic:<Apply Unif_elim with 3:=$c>>].

(* *)


Goal (t1,t2:quasiterm)(head_diff t1 t2)->(Unification t1 t2).
Intros;Apply Unif_fail;Elim H;Intros.
Apply Diff with BC;Simpl;Auto.
Apply Diff with BConsArg;Simpl;Auto.
Apply Diff with BC;Simpl;Auto.
Apply Diff with BRoot;Simpl;Auto.
Apply Diff with BConsArg;Simpl;Auto.
Apply Diff with BRoot;Simpl;Auto.
Save Decomp_fail.

Hint Fail_hd1 Fail_hd2 Fail_hd3 Fail_hd4 Fail_hd5 Fail_hd6 Decomp_fail.

(*****************************************************************)
(*************** Symetry of the problem of the unification : **********)
(*****************************************************************)

Goal (t1,t2:quasiterm)(Unification t1 t2)->(Unification t2 t1).
Intros;Elimunif H;Unfold idempotent over under min_unif unif;Intros.
Apply Unif_succeed with f;
Unfold idempotent over under min_unif unif;Auto;Intros.
Apply o;Unfold not;Intros h;Elim H0;Simpl;Elim h;Auto.
Simpl;Elim (u0 x y);Auto.
Apply Unif_fail_def;Intros.
Unfold not;Intros;Elim (n f);Auto.
Save sym_Unification.

(*****************************************************************)
(*********** Unification when one of the terms is a variable : ********)
(*****************************************************************)

Goal (n:var)(t:quasiterm)(Unification (V n) t).
Intros;Elim (IS_IN_decS n t);Intros.
Elim (quasiterm_eq_decS (V n) t);Intros.
(*Case (IS_IN n t) and <quasiterm>(V n)=t*)
Elim y0;Apply Unif_succeed with [x:var](V x);
Unfold unif idempotent over under;Simpl;Auto;Intros.
Absurd (V y1)=(V y1);Auto.
Unfold min_unif unif;Intros.
Apply less_subst_init with g;Auto.
(*------------ Verifications on the terms:begin -------------*)
Apply term_subst_init;Intros;Apply term_init;Simpl;Auto.
(*------------ Verifications on the terms:end --------------*)
(*Case (IS_IN n t) and ~<quasiterm>(V n)=t*)
Apply Unif_fail;Intros;Apply SUB_diff.
Simpl;Apply IS_IN_SUB;Auto.
(*Case ~(IS_IN n t)*)
Elim (sig_elem_subst n t);Intros.
Apply Unif_succeed with x;
Unfold unif idempotent over under;Simpl;Elim y0;Intros.
Replace (x n) with t;Auto.
Apply elem_subst_conserve with n;Auto.
Elim (var_eq_decP n x0);Intros.
Elim (H0 x0);Auto.
Apply elem_subst_conserve with n;Auto.
Elim (H x0);Simpl;Auto.
Elim (var_eq_decP x0 n);Intros.
Elim H1;Auto.
Elim (H x0);Auto.
(*BOGUS: Unfold not;Intros;Elim H2;Auto.*)
Elim (var_eq_decP n y1);Intros.
Replace t with (x y1);Auto.
Symmetry;Auto.
Absurd (V y1)=(x y1);Auto.
Unfold min_unif unif;Intros.
Apply less_subst_init with g.
Intros;Elim (var_eq_decP n x0);Intros.
Elim H2;Elim (H0 n);Auto.
Elim (H x0);Auto.
(*------------ Verifications on the terms:begin -------------*)
Apply term_subst_init.
Intros;Elim (var_eq_decP x0 n);Intros.
Elim (H0 x0);Auto.
Apply Length_SO_term;Auto.
Elim (H x0).
Apply term_init;Simpl;Auto.
Unfold not;Intros;Elim H4;Auto.
(*------------ Verifications on the terms:end --------------*)
Save UnifV1.

Goal (t:quasiterm)(x:var)(Unification t (V x)).
Intros;Apply sym_Unification;Apply UnifV1.
Save UnifV2.

Hint UnifV1 UnifV2.

(*****************************************************************)
(********** Unification when one of the terms is a constant : ********)
(*****************************************************************)

Goal (t:quasiterm)(l:fun)(Unification (C l) t).
Induction t;Auto.
Intros;Elim (fun_eq_decS l f);Intros.
Elim y;Apply Unif_succeed with V;
Unfold unif idempotent over under;Simpl;Auto.
(*Intros;Absurd <quasiterm>(V y)=(V y);Auto.*)
Unfold min_unif unif;Intros.
Apply less_subst_init with g ;Auto.
(*------------ Verifications on the terms:begin -------------*)
Intros;Apply term_subst_init;Intros;Apply term_init;Simpl;Auto.
(*------------ Verifications on the terms:end --------------*)
Apply Unif_fail;Intros;Simpl.
Apply C_diff_C;Auto.
Save UnifC1.

Goal (t:quasiterm)(l:fun)(Unification t (C l)).
Intros;Apply sym_Unification;Apply UnifC1.
Save UnifC2.

Hint UnifC1 UnifC2.

(*****************************************************************)
(********** Link between the unification of  t1 and t2 ****************)
(****** and the unification of (Root l1 t1) and (Root l2 t2) : *********)
(*****************************************************************)

Goal (t1,t2:quasiterm)(l1,l2:fun)
     (Unification t1 t2)
     ->(Unification (Root l1 t1) (Root l2 t2)).
Intros;Elim (fun_eq_decS l1 l2);Intros.
Elim y;Elimunif H;Unfold unif idempotent over under;Intros.
Apply Unif_succeed with f;
Unfold unif idempotent over under;Simpl;Auto.
Elim u;Auto.
Unfold min_unif unif;Simpl;Intros.
Apply m;Unfold unif.
Apply proj_Root2 with l1 l1;Auto.
Intros.
Elim H0;Elim H1;Intros;Apply t;Auto.
Elim H4;Auto.
Apply Unif_fail_def;Intros.
Simpl;Apply Root_diff_Root;Auto.
Apply Unif_fail;Intros.
Simpl;Apply Root_diff_Root;Auto.
Save UnifRoot.

Hint UnifRoot.

(********************************************************************)
(********** The failure of the unification of t1 and t3 implies ************)
(* the failure of the unification of (ConsArg t1 t2) and (ConsArg t3 t4) : **)
(********************************************************************)

Goal (t1,t2,t3,t4:quasiterm)
     ((f:quasisubst)(~<quasiterm>(Subst f t1)=(Subst f t3)))
     ->(Unification_f (ConsArg t1 t2) (ConsArg t3 t4)).
Intros;Apply Unif_fail_def;Intros.
Simpl;Apply ConsArg_diff_ConsArg;Auto.
Save UnifConsArgfail1.

Hint UnifConsArgfail1.

(*****************************************************************)
(********** The unification of the ground terms : ********************)
(*****************************************************************)

Goal (t1,t2:quasiterm)
     (DIFFELNB (list_var (ConsArg t1 t2)) O)
     ->(Unification t1 t2).
Intros;Elim (quasiterm_eq_decS t1 t2);Intros.
Elim y.
Apply Unif_succeed with V;
Unfold unif idempotent over under min_unif;Auto.
Intros.
Absurd (V y0)=(V y0);Auto.
Intros;Apply less_subst_init with g;Auto.
Intros;Apply term_subst_init;Intros;Apply term_init;Simpl;Auto.
Apply Unif_fail;Intros.
Elim (clossubst t1 f).
Elim (clossubst t2 f);Auto.
Apply closConsArg2 with t1;Apply DIFFELNB_O_clos;Auto.
Apply closConsArg1 with t2;Apply DIFFELNB_O_clos;Auto.
Save Unif_DIFFELNB_O.

(*****************************************************************)
(********* The number of variables of (ConsArg t1 t3)  **************)
(************ and the one of the variables of (ConsArg t2 t4) ********)
(********* are less or equal to the number of the variables of *********)
(********** (ConsArg (ConsArg t1 t2)(ConsArg t3 t4)). **************) 
(*****************************************************************)

Goal (t1,t2,t3,t4:quasiterm)(n:nat)
     (DIFFELNB (list_var (ConsArg (ConsArg t1 t2)(ConsArg t3 t4))) n)
     ->(n0:nat)(DIFFELNB (list_var (ConsArg t2 t4)) n0)
     ->({(le (S n0) n)}+{<nat>n0=n}).
Intros;Elim (le_decS n0 n);Intros.
Apply (le_S_eqS n0 n y).
Absurd (le n0 n);Auto.
Apply inclv_le with (list_var (ConsArg t2 t4))
(list_var (ConsArg (ConsArg t1 t2) (ConsArg t3 t4)));Auto.
Apply inclv_init;Intros;Apply IS_IN_IS_IN_LV;Simpl.
Elim (IS_IN_LV_IS_IN (ConsArg t2 t4) y0 H1);Auto.
Save DIFFELNB_ConsArg_ConsArg24_le.
 
Goal (t1,t2,t3,t4:quasiterm)(n:nat)
     (DIFFELNB (list_var (ConsArg (ConsArg t1 t2) (ConsArg t3 t4))) n)
     ->(n0:nat)(DIFFELNB (list_var (ConsArg t1 t3)) n0)
     ->({(le (S n0) n)}+{<nat>n0=n}).
Intros;Elim (le_decS n0 n);Intros h.
Apply (le_S_eqS n0 n h).
Absurd (le n0 n);Auto.
Apply inclv_le with (list_var (ConsArg t1 t3))
(list_var (ConsArg (ConsArg t1 t2) (ConsArg t3 t4)));Auto.
Apply inclv_init;Intros;Apply IS_IN_IS_IN_LV;Simpl.
Elim (IS_IN_LV_IS_IN (ConsArg t1 t3) y H1);Auto.
Save DIFFELNB_ConsArg_ConsArg13_le.

Hint eq_V_stab.

(*****************************************************************)
(******************** If f unifies t1 and u1 ************************)
(********** and if g unifies (Subst f t2) (Subst f u2), ****************)
(************** then [x:var](Subst g (f x)) unifies ********************)
(************** (ConsArg t1 t2) et (ConsArg u1 u2). ****************)
(*****************************************************************)

Goal (t1,t2,u1,u2:quasiterm)(f,g:quasisubst)
     (unif f t1 u1)
     ->(unif g (Subst f t2) (Subst f u2))
     ->(unif [x:var](Subst g (f x)) (ConsArg t1 t2) (ConsArg u1 u2)).
Unfold unif;Simpl;Intros.
Elim (comp_subst f g t2);Elim (comp_subst f g u2);
Elim (comp_subst f g t1);Elim (comp_subst f g u1).
Elim H;Elim H0;Auto.
Save unif_comp.

Hint unif_comp.

(*****************************************************************)
(**************** If f is a minimal unifier of t1 and u1 *************)
(********** and if g is a one of (Subst f t2) (Subst f u2), then *******)
(********************** [x:var](Subst g (f x)) ***********************) 
(************ is a one of (ConsArg t1 t2) and (ConsArg u1 u2) : ****)
(*****************************************************************)

Goal (t1,t2,u1,u2:quasiterm)(f,g:quasisubst)
     (min_unif f t1 u1)
     ->(min_unif g (Subst f t2) (Subst f u2))
     ->(min_unif [x:var](Subst g (f x)) (ConsArg t1 t2) (ConsArg u1 u2)).
Unfold min_unif unif;Simpl;Intros.
Elim (H g0);Intros.
Elim (H0 h);Intros.
Apply less_subst_init with h0.
Intros;Elim (H2 x).
Elim (exp_comp_subst g h0 (f x)).
Apply eq_restriction_s_t;Auto.
Elim (exp_comp_subst f h t2);Elim (exp_comp_subst f h u2).
Transitivity (Subst g0 t2).
Apply eq_restriction_s_t;Auto.
Transitivity (Subst g0 u2).
Apply proj_ConsArg2 with (Subst g0 t1) (Subst g0 u1);Auto;Intros.
Apply eq_restriction_s_t;Auto.
Apply proj_ConsArg1 with (Subst g0 t2) (Subst g0 u2);Auto;Intros.
Save min_unif_comp.

Hint min_unif_comp.

(*------------ Verifications on the terms:begin -------------*)


Goal (t1,t2,t3,t4:quasiterm)(f,g:quasisubst)(l:list_nat)
     (L_TERM l (ConsArg t1 t2))
     ->(L_TERM l (ConsArg t3 t4))
     ->(<nat>(Length (ConsArg t1 t2))=(Length (ConsArg t3 t4)))
     ->((l:list_nat)(L_TERM l t1)->(L_TERM l t3)
         ->(<nat>(Length t1)=(Length t3))->(term_subst l f))
     ->((l:list_nat)(L_TERM l (Subst f t2))->(L_TERM l (Subst f t4))
         ->(<nat>(Length (Subst f t2))=(Length (Subst f t4)))
         ->(term_subst l g))
     ->(term_subst l [x:var](Subst g (f x))).
Simpl;Intros.
Elim H;Elim H0;Intros.
Elim H5;Elim H7;Intros.
Cut (term_subst l f).
Intros hyp;Apply term_subst_init.
Intros;Simpl;Apply term_term_subst.
Elim hyp;Auto.
Apply H3;Try Apply L_TERM_term_subst;Auto.
Elim (term_subst_eq_Length l f t2);Elim (term_subst_eq_Length l f t4);
Auto.
Cut (plus (Length t1) (Length t2))=(plus (Length t3) (Length t4));Auto.
Replace  (Length t1) with (S O);Try Apply SIMPLE_SO;Auto.
Replace  (Length t3) with (S O);Try Apply SIMPLE_SO;Auto.
Apply (H2 l H8 H10).
Replace  (Length t1) with (S O);Try Apply SIMPLE_SO;Auto.
Save term_subst_comp.
(*------------ Verifications on the terms:end --------------*)

(*****************************************************************)
(********** If f is a minimal unifier of t1 and t3 *******************)
(********** whose domain and image are in (ConsArg t1 t3) *********)
(**************** and if g is a minimal unifier *********************)
(**************** of (Subst f t2) and (Subst f  t4) ******************)
(******** then [x:var](Subst g (f x)) is also a minimal unifier *********)
(*********** of (ConsArg t1 t2) and (ConsArg t3 t4) : ***************)
(*****************************************************************)

Hint over_comp under_comp.

Goal (t1,t2,t3,t4:quasiterm)(f,g:quasisubst)
     (unif f t1 t3)
     ->(idempotent f)
     ->(over f (ConsArg t1 t3))
     ->(under f (ConsArg t1 t3))
     ->(min_unif f t1 t3)
     ->((l:list_nat)(L_TERM l t1)->(L_TERM l t3)
         ->(<nat>(Length t1)=(Length t3))->(term_subst l f))
     ->(unif g (Subst f t2) (Subst f t4))
     ->(idempotent g)
     ->(over g (ConsArg (Subst f t2) (Subst f t4)))
     ->(under g (ConsArg (Subst f t2) (Subst f t4)))
     ->(min_unif g (Subst f t2) (Subst f t4))
     ->((l:list_nat)(L_TERM l (Subst f t2))->(L_TERM l (Subst f t4))
         ->(<nat>(Length (Subst f t2))=(Length (Subst f t4)))
         ->(term_subst l g))
     ->(Unification (ConsArg t1 t2) (ConsArg t3 t4)).
Intros;Apply Unif_succeed with [x:var](Subst g (f x));Intros;Auto.
(*Apply unif_comp;over_comp;under_comp;min_unif_comp*)
Apply (idempotent_Fondamental (ConsArg t2 t4));Auto.
(*------------ Verifications on the terms:begin -------------*)
Apply term_subst_comp with t1 t2 t3 t4;Auto.
(*------------ Verifications on the terms:end --------------*)
Save two_succes.

(*****************************************************************)
(************** If f is a minimal unifier of t1 and t3 ***************)
(*********** whose domain and image are in (ConsArg t1 t3) ********)
(*********************** and if the unification  *********************)
(***************** of (Subst f t2) and (Subst f  t4) fails *************)
(**** then the unification of (ConsArg t1 t2) and (ConsArg t3 t4) *****)
(**************************** fails also. ***************************)
(*****************************************************************)

Goal (t1,t2,t3,t4:quasiterm)(f:quasisubst)
     (unif f t1 t3)
     ->(idempotent f)
     ->(min_unif f t1 t3)
     ->((g:quasisubst)~<quasiterm>(Subst g (Subst f t2))
                                =(Subst g (Subst f t4)))(*1*)
     ->(Unification_f (ConsArg t1 t2) (ConsArg t3 t4)).
Unfold unif idempotent over under min_unif;Intros;Apply Unif_fail_def;Intros.
Unfold not;Simpl;Intros.
Elim (H2 f0).
Elim (H1 f0);Unfold unif;Intros.
(**)
Elim (exp_comp_subst f f0 t2);Elim (exp_comp_subst f f0 t4).
Replace (Subst [x:var](Subst f0 (f x)) t2) with (Subst f0 t2).
Replace (Subst [x:var](Subst f0 (f x)) t4) with (Subst f0 t4).
Apply proj_ConsArg2 with (Subst f0 t1) (Subst f0 t3);Auto.
(**)
Apply eq_restriction_s_t;Intros.
Elim (H4 x).
Replace (Subst h (f x)) with (Subst h (Subst f (f x))).
Elim (exp_comp_subst f h (f x));Apply eq_restriction_s_t;Auto. 
Elim H0;Auto.
(**)
Apply eq_restriction_s_t;Intros.
Elim (H4 x).
Replace (Subst h (f x)) with (Subst h (Subst f (f x))).
Elim (exp_comp_subst f h (f x));Apply eq_restriction_s_t;Auto. 
Elim H0;Auto.
(**)
Apply proj_ConsArg1 with (Subst f0 t2) (Subst f0 t4);Auto.
Save one_only_succes.

(*****************************************************************)
(*********** If t1 and t3 can be unified by the identity, **************)
(******** then (ConsArg t1 t2) and (ConsArg t3 t4) can this also *****) 
(**************** if and only if t2 and t4 can this . ****************)
(*****************************************************************)

Goal (t1,t2,t3,t4:quasiterm)(f:quasisubst)
     ((x:var)<quasiterm>(V x)=(f x))
     ->(unif f t1 t3)
     ->(Unification t2 t4)
     ->(Unification (ConsArg t1 t2) (ConsArg t3 t4)).
Unfold unif;Intros.
Elimunif H1;Unfold unif over under idempotent min_unif;Intros.
(*Success*)
Apply two_succes with f f0;
Replace (Subst f t2) with t2;
Replace (Subst f t4) with t4;
Unfold unif over under idempotent dom;Simpl;Auto.
(*Apply eq_V_stab*)
Intros;Absurd <quasiterm>(V y)=(f y);Auto.
Unfold min_unif;Intros.
Apply less_subst_init with g.
Intros;Elim (H x);Auto.
(*------------ Verifications on the terms:begin -------------*)
(**)Intros;Apply term_subst_init;Intros;Apply term_init;
(**)Elim H;Simpl;Auto.
(*------------ Verifications on the terms:end --------------*)
(*Fail*)
Apply Unif_fail_def.
Simpl;Intros;Unfold not;Intros.
Elim (n f0);
Apply proj_ConsArg2 with (Subst f0 t1) (Subst f0 t3);Auto.
Save eq_V_stab3.

(*****************************************************************)
(******************** For any terms u1 and u2, *******************)
(********** either there exists a minimal unifier of u1 and u2 ********)
(*********** whose domain and image are in (ConsArg u1 u2) *******)
(************** or u1 and u2 do not admit any unifier. *************)
(*****************************************************************)

Goal (n:nat)(u1,u2:quasiterm)
     (DIFFELNB (list_var (ConsArg u1 u2)) n)
     ->(Unification u1 u2).
Intros n;Pattern n;Apply (ind_leS n).


   (***************************************************************)
   (********************* Case without variable : ********************)
   (***************************************************************)
Intros;Apply Unif_DIFFELNB_O;Auto.
   (***************************************************************)
   (************** Case with p+1 different variables, *******************)
   (**************** induction on u1, then on u2 : *********************)
   (***************************************************************)
Induction u1;Auto.(*Apply UnifV1;UnifC1*)
(*Case u1 = (Root ...) : *)
Induction u2;Auto.(*Apply UnifV2;UnifC2;UnifRoot;Decomp_fail;Fail_hd6*)
(*Case u1 = (ConsArg ...) : *)
Induction u2;Auto.(*Apply UnifV2;UnifC2;Decomp_fail;Fail_hd5*)
Intros.
   (**      1 subgoal                                                *)
   (**      (Unification (ConsArg q q0) (ConsArg q1 q2))              *)
   (**       ============================            *)
   (**      number of different variables in (ConsArg q q1) :           *)
Elim (DIFFELNBor (list_var (ConsArg q q1)));Intros p0 H5.

   (***************************************************************)
   (******** Comparison of this number named p0 with p+1, **********)
   (*********** one knows that p0 < p+1 or p0 = p+1 : *************)
   (***************************************************************)
Elim (DIFFELNB_ConsArg_ConsArg13_le q q0 q1 q2 (S p) H4 p0 H5);Intros.
   (******************************************************************)
   (*  DIFFELNB_ConsArg_ConsArg13_le:(t1,t2,t3,t4:quasiterm)(n:nat)      *)
   (*(DIFFELNB (list_var (ConsArg (ConsArg t1 t2) (ConsArg t3 t4))) n) *)
   (*      ->(n0:nat)(DIFFELNB (list_var (ConsArg t1 t3)) n0          *)
   (*               ->({(le (S n0) n)}+{<nat>n0=n})                 *)
   (******************************************************************)

   (***************************************************************)
   (***************************************************************)
   (************************ Case p0 < p+1. **********************)
   (***************************************************************)
   (***************************************************************)

   (***************************************************************)
   (*********** one applies the recurrence hypothesis H : *************)
   (* (q:nat)(le q p)->(u1,u2:quasiterm)                              *)
   (* (DIFFELNB (list_var (ConsArg u1 u2)) q)->(Unification u1 u2)  *)
   (***************************************************************)
Elimunif (H p0 (le_S_n p0 p y) q q1 H5);Intros.
2 : Auto.
(*Apply UnifConsArgfail1*)

   (***************************************************************)
   (******** the unifier f of f and y1 is either the identity, ***********) 
   (***************** or different of the identity : *******************)
   (***************************************************************)
Elim (ident_or_notS (ConsArg q q1) f o);Intros.

   (***************************************************************)
   (************************ Case f non identity ********************)
   (***************************************************************)
Elim y0;Intros x H6.

   (***************************************************************)
   (********* one supposes (Unification (Subst f q0) (Subst f q2)) ******)
   (***************************************************************)
Cut (Unification (Subst f q0) (Subst f q2)).
Intros H7;Elimunif H7;Intros.
Apply two_succes with f f0;Auto.
Apply one_only_succes with f;Auto.

   (***************************************************************)
   (********* proof of (Unification (Subst f q0) (Subst f q2)) : *********)
   (********* 1) (list_var (ConsArg (Subst f q0) (Subst f q2))) *********)
   (******************** counts x1 different variables, ****************)
   (********** 2) one applies the recurrence hypothesis H *************)
   (********** 3) for this, one uses the fact that the unifier f *********)
   (********** makes the number of different variables decreasing. ******)
   (***************************************************************)
Elim (DIFFELNBor (list_var (ConsArg (Subst f q0) (Subst f q2))));
Intros p1 H7.
Apply (H p1);Auto.
Apply le_S_n;
Apply (f_n_id_minus q q0 q1 q2 (S p)) with f;Auto.
   (****************************************************************)
   (***** f_n_id_minus:(t1,t2,t3,t4:quasiterm)(n:nat) *********************)
   (* (DIFFELNB **************************************************)
   (*** (list_var (ConsArg (ConsArg t1 t2) (ConsArg t3 t4))) n) ********)
   (* ->(f:quasisubst)(idempotent f)->(over f (ConsArg t1 t3)) ********)
   (* ->(under f (ConsArg t1 t3)) **********************************)
   (* ->((Ex [x:var](~<quasiterm>(V x)=(f x)))) ****************)
   (* ->(n0:nat)(DIFFELNB  **************************************)
   (*********** (list_var (ConsArg (Subst f t2) (Subst f t4))) n0) *******)
   (* ->(le (S n0) n) **********************************************)
   (****************************************************************)
Exists x;Auto.

   (****************************************************************)
   (**************** Case f = identity extentionnelly ******************)
   (** 1) eq_V_stab3 reduces the problem to (Unification q0 q2), **********)
   (** 2) (list_var (ConsArg q0 q2)) counts p0 different variables, ********)
   (***** 3) si p0 < p+1, one applies the recurrence hypothesis H *******)
   (***** 4) si p0 = p+1, one applies the recurrence hypothesis H1. *****)
   (****************************************************************)
Apply eq_V_stab3 with f;Auto.
Elim (DIFFELNBor (list_var (ConsArg q0 q2)));Intros p1 H6.
Elim (DIFFELNB_ConsArg_ConsArg24_le q q0 q1 q2 (S p) H4 p1 H6); Intros.
   (***************************************************************)
   (* DIFFELNB_ConsArg_ConsArg24_le:(t1,t2,t3,t4:quasiterm)(n:nat) ****)
   (* (DIFFELNB *************************************************)
   (* ***(list_var (ConsArg (ConsArg t1 t2) (ConsArg t3 t4))) n) ******)
   (* ->(n0:nat)(DIFFELNB (list_var (ConsArg t2 t4)) n0) ***********)
   (* ->({(le (S n0) n)}+{<nat>n0=n}) ****************************)
   (***************************************************************)
Apply (H p1 (le_S_n p1 p y1) q0 q2);Auto.
Apply (H1 q2 );Elim y1;Auto.

   (***************************************************************)
   (***************************************************************)
   (********************* Case p0 = p+1 **************************)
   (***************************************************************)
   (***************************************************************)

   (***************************************************************)
   (******** 1) one applies the recurrence hypothesis H0 **************)
   (********** which proves (Unification y y1), **********************)
   (********2) one proceeds after exactly like above. ******************)
   (***************************************************************)
Cut (Unification q q1);[Intro H0'|Apply (H0 q1);Elim y;Auto].
Elimunif H0';Auto;Intros.(*Apply UnifConsArgfail1*)
Elim (ident_or_notS (ConsArg q q1) f o);Intros.
Elim y0;Intros x H6.

   (***************************************************************)
   (********* one supposes (Unification (Subst f q0) (Subst f q2))******)
   (***************************************************************)
Cut (Unification (Subst f q0) (Subst f q2)).
Intros H7;Elimunif H7;Intros.
Apply two_succes with f f0;Auto.
Apply one_only_succes with f;Auto.

   (***************************************************************)
Elim (DIFFELNBor (list_var (ConsArg (Subst f q0) (Subst f q2))));
Intros p1 H7.
Apply (H p1);Auto.
Apply le_S_n;Apply (f_n_id_minus q q0 q1 q2 (S p)) with f;Auto.
Exists x;Auto.

   (***************************************************************)
Apply eq_V_stab3 with f;Auto.
Elim (DIFFELNBor (list_var (ConsArg q0 q2)));Intros p1 H6.
Elim (DIFFELNB_ConsArg_ConsArg24_le q q0 q1 q2 (S p) H4 p1 H6);Intros.
Apply (H p1);Auto.(*Apply Apply le_S_n*)
Apply (H1 q2 );Elim y1;Auto.
Save proof_unif.

Goal (t,u:quasiterm)(Unification t u).
Intros;Elim (DIFFELNBor (list_var (ConsArg t u)));Intros n0 p;
Apply proof_unif with n0;Auto.
Save unif_proof.

End algorithmes.

Goal (t,u:quasiterm)(Unification_or_fail t u).
  Intros; Apply Unification_build with (Unification_f t u) t u.
    Apply unif_proof. Intro c; Apply Unification_or_fail_fail; Exact c.
    Intro c; Exact c.
    Intro res; Apply Unification_or_fail_succeed; Exact res.
Save unif_call.
