#open "hashtbl";;

(* UTILITIES *)

type 'a option = SOME of 'a | NONE;;

(* filter a list by a predicate and split it into two lists *)
let filter p = 
  let rec loop (p_true,p_false) = function
     (x::l) ->
       loop (if p x then (x::p_true,p_false) else (p_true,x::p_false)) l
   | [] -> (rev p_true, rev p_false) in
  loop ([],[]);;

(* push an element on top of a mutable list *)
let push v l = l := v :: !l;;

(* pgcd *)
let rec pgcd x y = if y = 0 then x else pgcd y (x mod y);;

let pgcd_l = function
   [] -> failwith "pgcd_l"
 | (x::l) -> it_list pgcd x l;;

(* Computes the floor of the rational a / b *)
let floor_div a b =
  match a>=0 ,b>0 with
     true,true -> a / b
   | false,false -> a / b
   | true, false -> (a-1) / b - 1
   | false,true  -> (a+1) / b - 1;;

(* BASIC TYPES *)

type coeff = {c: int ; v: int};;
type linear == coeff list;;
type afine = { 
  (* a number uniquely identifying the equation *)
  id: int ; 
  (* a boolean true for an eq, false for an ineq (Sigma a_i x_i >= 0) *)
  kind: bool; 
  (* the variables and their coefficient *)
  body: coeff list; 
  (* a constant *)
  constant: int };;

exception UNSOLVABLE;;

type action =
   DIVIDE_AND_APPROX of afine * afine * int * int
 | NOT_EXACT_DIVIDE of afine * int
 | FORGET_C of int
 | EXACT_DIVIDE of afine * int
 | SUM of int * (int * afine) * (int * afine)
 | STATE of afine * afine * afine * int * int
 | HYP of afine
 | FORGET   of int * int
 | FORGET_I of int * int
 | CONTRADICTION of afine * afine
 | MERGE_EQ of int * afine * int 
 | CONSTANT_NOT_NUL of int * int
 | CONSTANT_NEG of int * int
 | WEAKEN of int * int;;

let new_id = let cpt = ref 0 in (fun () -> incr cpt; ! cpt);;

let debug = ref false;;

(* Pour le parsing des formules *)

let new_var = 
  let cpt = ref 0 in fun () -> incr cpt; "WW" ^ string_of_int !cpt;;

let intern_id,unintern_id =
  let cpt = ref 0 in
  let table = new 7 and co_table = new 7 in
  (fun (name:string) -> 
     try find table name
     with Not_found -> 
       let idx = !cpt in
       add table name idx; add co_table idx name;
       incr cpt; idx),
  (fun idx -> try find co_table idx with _ -> 
     let v = new_var () in 
     add table v idx; add co_table idx v; v)
;;

(* Pretty printing for debugging *)

let string_of_var v =
  try unintern_id v
  with Not_found -> 
    "?" ^ make_string  1 (char_of_int (v mod 26+65)) ^ 
    (if v < 27 then "" else string_of_int (v / 26));;

let display_eq (l,e) =
  it_list (fun not_first f ->
             print_string (if f.c < 0 then "- " 
                           else if not_first then "+ " else "");
             let c = abs f.c in
	     if c = 1 then printf__printf "%s " (string_of_var f.v)
             else printf__printf "%d %s " c (string_of_var f.v); true)
    false l;
  if e > 0 then printf__printf "+ %d " e 
  else if e < 0 then printf__printf "- %d " (abs e);;

let display_system l = 
  do_list (fun { kind=b; body=e; constant=c; id=id} -> 
      	     print_int id; print_string ": ";
             display_eq (e,c); print_string (if b then "= 0\n" else ">= 0\n"))
          l;
  print_string "------------------------\n\n";;

let display_inequations l = 
  do_list (fun e -> display_eq e; print_string ">= 0\n") l;
  print_string "------------------------\n\n";;

let rec display_action = function
  (act::l) -> 
    begin match act with
       DIVIDE_AND_APPROX (e1,e2,k,d) ->
         printf__printf 
           "Inequation E%d is divided by %d and the constant coefficient is \
            rounded by substracting %d.\n" e1.id k d
     | NOT_EXACT_DIVIDE (e,k) ->
         printf__printf
           "Constant in equation E%d is not divisible by the pgcd \
            %d of its other coefficients.\n" e.id k
     | EXACT_DIVIDE (e,k) ->
         printf__printf
           "Equation E%d is divided by the pgcd \
            %d of its coefficients.\n" e.id k
     | WEAKEN (e,k) ->
         printf__printf 
           "To ensure a solution in the dark shadow \
            the equation E%d is weakened by %d.\n" e k
     | SUM (e,(c1,e1),(c2,e2)) -> 
         let k b = if b then "equation" else "inequation" in
         printf__printf
	   "We state %s E%d = %d %s E%d + %d %s E%d.\n" 
            (k (e1.kind & e2.kind)) e c1 (k e1.kind) e1.id c2 (k e2.kind) e2.id
     | STATE (e,_,_,x,_) ->
         printf__printf "We define a new equation %d :" e.id; 
         display_eq (e.body,e.constant); 
         print_string (if e.kind then "= 0\n" else ">= 0\n")
     | HYP e ->   
         printf__printf "We define %d :" e.id; 
         display_eq (e.body,e.constant); 
         print_string (if e.kind then "= 0\n" else ">= 0\n")
     | FORGET_C e ->  printf__printf "E%d is trivially satisfiable.\n" e
     | FORGET (e1,e2) -> printf__printf "E%d subsumes E%d.\n" e1 e2
     | FORGET_I (e1,e2) -> printf__printf "E%d subsumes E%d.\n" e1 e2
     | MERGE_EQ (e,e1,e2) ->
         printf__printf "E%d and E%d can be merged into E%d.\n" e1.id e2 e
     | CONTRADICTION (e1,e2) -> 
         printf__printf
           "equations E%d and E%d implie a contradiction on their \
            constant factors.\n" e1.id e2.id
     | CONSTANT_NOT_NUL (e,k) -> 
         printf__printf "equation E%d states %d=0.\n" e k
     | CONSTANT_NEG(e,k) -> 
         printf__printf "equation E%d states %d >= 0.\n" e k
    end; display_action l
 | [] -> flush std_out;;

(* history of the moves made *)
let add_event, history, clear_history =
  let accu = ref [] in
  (fun (v:action) -> if !debug then display_action [v]; push v accu), 
  (fun () -> !accu),
  (fun () -> accu := []);;

(* normalize equations by sorting *)
let nf_linear = sort__sort (fun x y -> x.v > y.v);;
let nf ((b:bool),(e,(x:int))) = (b,(nf_linear e,x));;

(* maps on equation (linear or afine) *)
let map_eq_linear f = 
  let rec loop = fun
     (x::l) -> let c = f x.c in if c=0 then loop l else {v=x.v; c=c} :: loop l
   | [] -> [] in
  loop;;

let map_eq_afine f e = 
  { id = e.id; kind = e.kind; body = map_eq_linear f e.body; 
    constant = f e.constant };;

(* negation of an equation *)
let negate_eq = map_eq_afine (fun x -> -x);;

(* sum up two linear equations *)
let rec sum = fun
   [] l -> l | l [] -> l
 | ((x1::l1) as l1') ((x2::l2) as l2') -> 
     if x1.v = x2.v then (
       let c=x1.c+x2.c in
       if c = 0 then sum l1 l2 else {v=x1.v;c=c}::sum l1 l2)
     else if x1.v > x2.v then x1 :: sum l1 l2'
     else x2 :: sum l1' l2;;

(* sum up two equations *)
let sum_afine eq1 eq2 = 
  {kind = eq1.kind & eq2.kind; id = new_id ();
   body = sum eq1.body eq2.body; constant = eq1.constant + eq2.constant };;

(* PART I : BANERJEE TEST *)

exception FACTOR1;;

let rec chop_factor_1 = function
   (x::l) -> if abs x.c = 1 then x,l
             else let (c',l') = chop_factor_1 l in (c',x::l')
 | [] -> raise FACTOR1;;

exception CHOPVAR;;

let rec chop_var v = function
   (f::l) -> if f.v = v then f,l else let (f',l') = chop_var v l in (f',f::l')
 | [] -> raise CHOPVAR;;

let normalize ({id=id; kind=eq_flag; body=e; constant =x} as eq) =
  if e = [] then 
    if eq_flag then 
      if x =0 then eq
      else begin
        add_event (CONSTANT_NOT_NUL(id,x)); raise UNSOLVABLE
      end
    else 
      if x >= 0 then eq 
      else begin
        add_event (CONSTANT_NEG(id,x)); raise UNSOLVABLE
      end
  else
    let gcd = pgcd_l (map (fun f -> abs f.c) e) in
    if eq_flag & x mod gcd <> 0 then begin
      add_event (NOT_EXACT_DIVIDE (eq,gcd)); raise UNSOLVABLE
    end else if gcd <> 1 then begin
      let c = floor_div x gcd in
      let d = x - c * gcd in
      let new_eq = {id=id; kind=eq_flag; constant=c; 
                    body=map_eq_linear (fun c -> c / gcd) e} in
      add_event (if eq_flag then EXACT_DIVIDE(eq,gcd)
                 else DIVIDE_AND_APPROX(eq,new_eq,gcd,d));
      new_eq
    end else eq;;

(* Given a variable v defined in function of eq2, eliminate this
 * variable v in eq1 *)

let eliminate_with_in {v=v;c=c_unite} eq2
                        ({body=e1; constant=c1;_} as eq1) =
  try
    let (f,_) = chop_var v e1 in 
    let coeff = if c_unite=1 then -f.c else if c_unite= -1 then f.c 
                else failwith "eliminate_with_in" in
    let res = sum_afine eq1 (map_eq_afine (fun c -> c * coeff) eq2) in
    add_event (SUM (res.id,(1,eq1),(coeff,eq2))); res
  with CHOPVAR -> eq1;;

(* The bizarre mod of the Banerjee test *)

let omega_mod a b = a - b * floor_div (2 * a + b) (2 * b);;

let new_var = let cpt = ref 1000 in (fun () -> incr cpt; !cpt);;

(* One banerjee step. Reduce all coefs bigger than the base by a factor 
   of 2/3 at least. Substitution in the other equations is performed *)

let banerjee_step original l1 l2 = 
  let e = original.body in
  let sigma = new_var () in
  let smallest,var =
    try
      it_list (fun (v,p) c ->  if v > (abs c.c) then abs c.c,c.v else (v,p))
              (abs (hd e).c, (hd e).v) (tl e)
    with Failure "tl" -> display_system [original] ; failwith "TL" in
  let m = smallest + 1 in
  let new_eq =
    { constant = omega_mod original.constant m;
      body = {c= -m;v=sigma} :: 
             map_eq_linear (fun a -> omega_mod a m) original.body;
      id = new_id (); kind = true } in
  let definition =
    { constant = - floor_div (2 * original.constant + m) (2 * m);
      body = map_eq_linear (fun a -> - floor_div (2 * a + m) (2 * m))
                             original.body;
      id = new_id (); kind = true } in
  add_event (STATE (new_eq,definition,original,m,sigma));
  let new_eq = normalize new_eq in
  let eliminated_var, def = 
    try chop_var var new_eq.body 
    with e ->
     print_endline (unintern_id var);display_system [original; new_eq];
     raise e in
  let other_equations = 
    map (fun e -> normalize (eliminate_with_in eliminated_var new_eq e)) l1 in
  let inequations = 
    map (fun e -> normalize (eliminate_with_in eliminated_var new_eq e)) l2 in
  let original' = eliminate_with_in eliminated_var new_eq original in
  let mod_original = map_eq_afine (fun c -> c / m) original' in
  add_event (EXACT_DIVIDE (original',m));
  normalize mod_original,other_equations,inequations;;


let rec eliminate_one_equation (e,other,ineqs) = 
  (* look for an equation *)
  if !debug then display_system (e::other);
  try
    let v,def = chop_factor_1 e.body in
    (map (fun e' -> normalize (eliminate_with_in v e e')) other,
     map (fun e' -> normalize (eliminate_with_in v e e')) ineqs)
  with FACTOR1 -> eliminate_one_equation (banerjee_step e other ineqs);;

(* Apply the banerjee test to eliminate equations *)

(* We assume that the system is normalized *)

let rec banerjee (sys_eq,sys_ineq) =
  let rec fst_eq_1 = fun
     (eq::l) -> 
        if exists (fun x -> abs x.c = 1) eq.body then eq,l
        else let (eq',l') = fst_eq_1 l in (eq',eq::l')
   | [] -> raise Not_found in
  match sys_eq with
     [] -> if !debug then display_system sys_ineq; sys_ineq
   | (e1::rest) -> 
       let eq,other = try fst_eq_1 sys_eq with Not_found -> (e1,rest) in
       if eq.body = [] then 
         if eq.constant = 0 then begin
           add_event (FORGET_C eq.id); banerjee (other,sys_ineq)
         end else begin
	   add_event (CONSTANT_NOT_NUL(eq.id,eq.constant)); raise UNSOLVABLE
	 end
       else banerjee (eliminate_one_equation (eq,other,sys_ineq));;

(* PART II : REDONDANCY ELIMINATION *)

type kind = INVERTED | NORMAL;;

(* From now on, we only treat inequalities *)

let redundancy_elimination system =
  (* ensures that a set of factors has a unique representation *)
  let normal = function
     ({body=f::_;_} as e) when f.c < 0 ->  negate_eq e, INVERTED
   | e -> e,NORMAL in
  let table = new 7 in
  do_list
    (fun e -> 
       let ({body=ne;_} as nx) ,kind = normal e in
       if ne = [] then
         if nx.constant < 0 then begin
           add_event (CONSTANT_NEG(nx.id,nx.constant)); raise UNSOLVABLE
	 end else add_event (FORGET_C nx.id)
       else
       try 
         let (optnormal,optinvert) = find table ne in
         let final =
           if kind = NORMAL then begin
	     match optnormal with 
                SOME v -> 
		  let kept =
                    if v.constant < nx.constant 
                    then begin add_event (FORGET (v.id,nx.id));v end
                    else begin add_event (FORGET (nx.id,v.id));nx end in
                  (SOME(kept),optinvert)
              | NONE -> SOME nx,optinvert
           end else begin
	     match optinvert with 
                SOME v ->
		  let kept =
                    if v.constant > nx.constant 
                    then begin add_event (FORGET_I (v.id,nx.id));v end
                    else begin add_event (FORGET_I (nx.id,v.id));nx end in
                  (optnormal,SOME(if v.constant > nx.constant then v else nx))
              | NONE -> optnormal,SOME nx
	   end in
         begin match final with
            (SOME high, SOME low) -> 
              if high.constant < low.constant then begin
	        add_event(CONTRADICTION (high,negate_eq low));
                raise UNSOLVABLE
              end
          | _ -> () end;
         remove table ne;
         add table ne final
       with Not_found ->
         add table ne 
           (if kind = NORMAL then (SOME nx,NONE) else (NONE,SOME nx)))
    system;
  let accu_eq = ref [] in
  let accu_ineq = ref [] in
  do_table 
    (fun e (SOME x, SOME y) when x.constant = -y.constant -> 
      	   let id=new_id () in
      	   add_event (MERGE_EQ(id,x,y.id));
      	   push {id=id; kind=true; body=x.body; constant=x.constant} accu_eq
       | e (optnorm,optinvert) ->
          begin match optnorm with 
             SOME x -> push x accu_ineq | _ -> () end;
          begin match optinvert with 
             SOME x -> push (negate_eq x) accu_ineq | _ -> () end)
    table;
  !accu_eq,!accu_ineq;;

(* PART III : OMEGA PROJECTION (FOURIER MOTZKIN) *)

(* We want to select a variable (to eliminate) such that it has the lowest
 * coefficients. May be we should then select on the lowest number of
 * occurences *)

exception SOLVED_SYSTEM;;

let select_variable system =
  let table = new 7 in
  let push v c= 
    try let r = find table v in r := max !r (abs c)
    with Not_found -> add table v (ref (abs c)) in
  do_list (fun {body=l;_} -> do_list (fun f -> push f.v f.c) l) system;
  let vmin,cmin = ref (-1), ref 0 in
  let var_cpt = ref 0 in
  do_table 
    (fun v (ref c) ->
       incr var_cpt;
       if c < !cmin or !vmin = (-1) then begin vmin := v; cmin := c end)
    table;
  if !var_cpt < 1 then raise SOLVED_SYSTEM;
  !vmin;;

(* then we extract the coefficient related to this variable from each
   inequation and we classify them in three category: not occur, positive, 
   negative *)

let classify v system =
  it_list 
    (fun (not_occ,below,over) eq ->
       try let f,eq' = chop_var v eq.body in
       	   if f.c >= 0 then (not_occ,((f.c,eq) :: below),over)
	   else (not_occ,below,((-f.c,eq) :: over))
       with CHOPVAR -> (eq::not_occ,below,over))
    ([],[],[]) system;;

(* We build the cartesian product of positive and negative inequations. 
 * dark_shadow is a boolean telling if we must use the smaller projection
 * insuring the existence of an integer in it *)

let product dark_shadow low high =
  it_list
    (fun accu (a,eq1) ->
       it_list
         (fun accu (b,eq2) ->
	    let eq = 
	      sum_afine (map_eq_afine (fun c -> c * b) eq1)
	                (map_eq_afine (fun c -> c * a) eq2) in
	    add_event(SUM(eq.id,(b,eq1),(a,eq2)));
	    let eq = normalize eq in
	    let final_eq =
  	      if dark_shadow then 
	        let delta = (a - 1) * (b - 1) in
		add_event(WEAKEN(eq.id,delta));
                {id = eq.id; kind=false; body = eq.body; 
                 constant = eq.constant - delta} 
              else eq
	    in final_eq :: accu)
         accu high)
    [] low;;

(* GLOBAL SIMPLIFICATION PROCEDURE *)

let simplify dark_shadow system =
  clear_history ();
  do_list (fun e -> add_event (HYP e)) system;
  (* Initial simplification phase *)
  let system = map normalize system in
  let eqs,ineqs = filter (fun e -> e.kind) system in
  let simp_eq,simp_ineq = redundancy_elimination ineqs in
  let system = (eqs @ simp_eq,simp_ineq) in

  let rec loop1a system =
    let sys_ineq = banerjee system in 
    loop1b sys_ineq 
  and loop1b sys_ineq =
    let simp_eq,simp_ineq = redundancy_elimination sys_ineq in
    if simp_eq = [] then simp_ineq
    else loop1a (simp_eq,simp_ineq) in
  let rec loop2 system =
    try
      let v = select_variable system in
      let (ineq_out, ineq_low,ineq_high) = classify v system in
      let expanded = ineq_out @ product dark_shadow ineq_low ineq_high in
      if !debug then display_system expanded;
      let without_redundancy = loop1b expanded in
      loop2 without_redundancy
    with SOLVED_SYSTEM -> if !debug then display_system system; system in
  loop2 (loop1a system);;

(* EXTRACTION OF RESULTS: CONTRADICTION *)

let rec depend relie_on accu = function
  (act::l) -> 
    begin match act with
       DIVIDE_AND_APPROX (e,_,_,_) ->
         if mem e.id relie_on then depend relie_on (act::accu) l
	 else depend relie_on accu l
     | EXACT_DIVIDE (e,_) ->
         if mem e.id relie_on then depend relie_on (act::accu) l
	 else depend relie_on accu l
     | WEAKEN (e,_) ->
         if mem e relie_on then depend relie_on (act::accu) l
	 else depend relie_on accu l
     | SUM (e,(_,e1),(_,e2)) -> 
         if mem e relie_on then depend (e1.id::e2.id::relie_on) (act::accu) l
	 else depend relie_on accu l
     | STATE (e,_,_,_,_) ->
         if mem e.id relie_on then depend relie_on (act::accu) l
	 else depend relie_on accu l
     | HYP e ->   
         if mem e.id relie_on then depend relie_on (act::accu) l
	 else depend relie_on accu l
     | FORGET_C _ -> depend relie_on accu l
     | FORGET _ -> depend relie_on accu l
     | FORGET_I _ -> depend relie_on accu l
     | MERGE_EQ (e,e1,e2) ->
         if mem e relie_on then depend (e1.id::e2::relie_on) (act::accu) l
	 else depend relie_on accu l
     | NOT_EXACT_DIVIDE (e,_) ->  depend (e.id::relie_on) (act::accu) l
     | CONTRADICTION (e1,e2) -> depend (e1.id::e2.id::relie_on) (act::accu) l
     | CONSTANT_NOT_NUL (e,_) -> depend (e::relie_on) (act::accu) l
     | CONSTANT_NEG (e,_) -> depend (e::relie_on) (act::accu) l
    end
 | [] -> accu;;

let solve system =
  try simplify false system;
      failwith "no contradiction"
  with UNSOLVABLE -> display_action (depend [] [] (history ()));;
