(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              July 1st 1995                               *)
(*                                                                          *)
(****************************************************************************)
(****************************************************************)
(*      Proof of a multiplier circuit                           *)
(*      C. Paulin-Mohring, June 95                              *)
(****************************************************************)
(*      The Multiplier proof                                    *)
(****************************************************************)

Require Arith.
Require Peano_dec.
Require Bool.
Require Orb.
Require Zerob.
Require IfProp.

Require Streams.
Require Circ.

(* Definition of each combinational part of the circuit *)

Definition Mux : bool->nat->nat->nat
      := [b:bool][n,m:nat]<nat>Case b of n m end.

(* Extra lemmas *)

Lemma pred_SO : (n:nat)(pred n)=O -> ~(n=O) -> n=(S O).
Destruct n; Auto.
Intros H1 H2; Elim H2; Auto.
Qed.

Lemma plus_mult_pred : 
   (n,m:nat)(~n=O)->(plus (mult (pred n) m) m)=(mult n m).
Destruct n; Simpl; Intros ; [Elim H; Trivial | Auto].
Qed.

(* The updating function *)

Definition upd1 := [i1,r1:nat][r3:bool](Mux r3 i1 (pred r1)).
Definition upd2 := [i1,i2,r2:nat][r3:bool]
      (Mux r3 (Mux (zerob i1) O i2) (plus (Mux (zerob i1) O i2) r2)).
Definition upd3 := [i1,i2,r1:nat][r3:bool]
      (orb (zerob (Mux r3 (pred i1) (pred (pred r1)))) (zerob i2)).

Lemma upd3_true : (i1,i2,r1:nat)
  (upd3 i1 i2 r1 true)= (orb (zerob (pred i1)) (zerob i2)).
Trivial.
Qed.

Lemma upd3_false : (i1,i2,r1:nat)
  (upd3 i1 i2 r1 false)= (orb (zerob (pred (pred r1))) (zerob i2)).
Trivial.
Qed.

Record TR : Set := reg {reg1 : nat; reg2 : nat; reg3 : bool}.

Record TI : Set := {in1 : nat; in2 : nat}.

Record TO : Set := out {res : nat; done : bool}.

Definition update : TI->TR->TR := 
   [i][r](reg (upd1 (in1 i) (reg1 r) (reg3 r))
              (upd2 (in1 i) (in2 i) (reg2 r) (reg3 r))
              (upd3 (in1 i) (in2 i) (reg1 r) (reg3 r))).

Definition output : TI->TR->TO := [i][r](out (reg2 r) (reg3 r)).

Section Multiplier_circuit.

Variables ri1,ri2 : nat.
Definition init : TR := (reg ri1 ri2 true).

Definition circ_mult : (Str TI)->(Str TO) := (Circ output update init).

Variable i0 : (Str TI).
Definition X:= (in1 (Hd i0)).
Definition Y:= (in2 (Hd i0)).
Definition XY := (Build_TI X Y).

Lemma nth_i0_O : XY=(Nth i0 O).
Replace (Nth i0 O) with (Hd i0).
Unfold XY X Y; Elim (Hd i0); Trivial.
Trivial.
Qed.
Hint nth_i0_O.

Definition stable : nat -> Prop
        := [n:nat](between [q:nat]XY=(Nth i0 q) O n).
Hint Unfold stable.

Lemma stable_O : (stable O).
Auto.
Qed.
Hint stable_O.

Lemma stable_S : (n:nat)(stable (S n))->(stable n).
Intros n H; Inversion H; Auto.
Qed.
Immediate stable_S.

Lemma stable_Sn : (n:nat)(stable (S n))->XY=(Nth i0 n).
Intros n H; Inversion H; Trivial.
Qed.
Immediate stable_Sn.

Definition Q : nat->TO->Prop := 
  [n:nat][o:TO](stable n) -> ~(n=O) -> (done o)=true -> (res o)=(mult X Y).

Definition InvM : nat->TR->Prop :=
[n:nat][r:TR](stable n)->
 (IfProp ~(n=O) -> (reg2 r)=(mult X Y)
         ((~(pred (reg1 r))=O) /\ 
            ~(X=O) /\ (plus (mult (pred (reg1 r)) Y) (reg2 r))=(mult X Y))
         (reg3 r)).

Lemma InvM_init : (InvM O init).
Red; Intro; Unfold reg3; Simpl.
Left; Destruct 1; Auto.
Qed.

Lemma InvM_stable : 
(n:nat)(r:TR)(InvM n r)->
    (Q n (output (Nth i0 n) r)) /\ (InvM (S n) (update (Nth i0 n) r)).

Split.
(* Proof of (Q n (output (Nth i0 n) r)) *)
Red; Simpl.
Intros H0 H1 H2.
Elim (Iftrue_inv ? ? ? (H H0)); Auto.

(* Proof of (InvM (S n) (update (Nth i0 n) r)) *)
Red; Simpl; Intro.
Replace (Nth i0 n) with XY; Simpl.
(* Proof of (Nth i0 n)=XY *) 
2: Auto.
(* Cases on the value of (reg3 r) *)
Elim H; Simpl; Intros.
(* Proof that (stable n) *)
3:Auto.
(* Case (reg3 r)=true *)
Rewrite upd3_true.
(* Case analysis wether X=0 *)
Case X.
(* X=0 *)
Simpl; Auto.
(* X'<>0 *)
(* Case analysis wether Y=0 *)
Intro X'; Case Y; Simpl.
(* Y = O *)
Rewrite orb_b_true; Auto.
(* Y=(S Y')<>0 *)
(* Case analysis wether X'=0 *)
Intro Y'; Case X'; Simpl.
(* X'=O *)
Auto.
(* X' = (S X'') <> 0 *)
Intro X''.
Right; Split;[Auto | Split].
Auto.
Apply eq_S; Elim plus_n_Sm; Elim plus_n_Sm; Auto.
(* Case (reg3 r)=false *)
Rewrite upd3_false.
Case H1; Clear H1; Intros.
Case H2; Clear H2; Intros.
Rewrite (zerob_false_intro X); Trivial; Simpl.
(* Case analysis whether Y=0 *)
Generalize H3; Clear H3; Case Y; Simpl.
(* Y=O *)
Do 3 Elim mult_n_O; Rewrite orb_b_true; Auto.
(* Y=(S Y') *)
Intro Y'; Elim (eq_nat_dec (pred (pred (reg1 r))) O); Intro H3.
(* (pred (pred (reg1 r))=O *)
Rewrite H3; Simpl.
Rewrite (pred_SO (pred (reg1 r))); Trivial; Simpl.
Elim plus_n_O; Auto.
(* (pred (pred (reg1 r)))<>O *)
Intros; Rewrite zerob_false_intro; Trivial; Simpl.
Right; Intros.
Split; Trivial.
Split; Trivial.
Elim H4.
Replace (S (plus Y' (reg2 r))) with (plus (S Y') (reg2 r)); Trivial.
Rewrite plus_assoc_l.
Rewrite (plus_mult_pred (pred (reg1 r)) (S Y')); Auto.
Save.

Theorem circ_mult_proof :
   (n:nat)(stable n)->~(n=O)->
   (done (Nth (circ_mult i0) n))=true
    -> (res (Nth (circ_mult i0) n))=(mult X Y).
Exact (Circ_property ? ? ? output update i0 init Q InvM InvM_init InvM_stable).
Qed.

End Multiplier_circuit.
