(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              July 1st 1995                               *)
(*                                                                          *)
(****************************************************************************)
(****************************************************************)
(*		        Proof of a multiplier circuit		*)
(*		        C. Paulin-Mohring, June 95		*)
(****************************************************************)
(*	                Streams.v			   	*)
(*			Representing Streams	        	*)
(****************************************************************)

Require GFP.
Require Linear.

Section Streams.
Variable A : Set.

Local F :=  [X:Set]A*X.
Definition Fmon : (X,Y:Set)(X->Y) -> (F X) -> (F Y).
Unfold F; Intros.
Split; Case H0; Auto.
Defined.

Definition Str := (nu F).

Definition STROUT : Str->A*Str := (Out F Fmon).
Local StrOut := STROUT.

Definition HD : Str->A   := [s:Str](Fst (StrOut s)).
Definition TL : Str->Str := [s:Str](Snd (StrOut s)).

Local Hd := HD.
Local Tl := TL.

Definition STRCOIT : (X:Set)(X->A*X)->X->Str := (CoIt F).
Local StrCoIt := STRCOIT.

Definition STRIT : (X:Set)(X->A)->(X->X)->X->Str
  := [X:Set][h:X->A][t:X->X][x:X](StrCoIt X [y:X]((h y),(t y)) x).
Local StrIt := STRIT.

Lemma Hd_StrIt : (X:Set)(h:X->A)(t:X->X)(x:X)(Hd (StrIt X h t x))=(h x).
Trivial.
Qed.

Lemma Tl_StrIt : (X:Set)(h:X->A)(t:X->X)(x:X)(Tl (StrIt X h t x))=(StrIt X h t (t x)).
Trivial.
Qed.

Fixpoint NTHTL [s:Str;n:nat] : Str := <Str> Case n of s [p:nat](Tl (NTHTL s p)) end.
Local NthTl:= NTHTL.

Definition NTH : Str->nat->A := [s:Str][n:nat](Hd (NthTl s n)).
Local Nth := NTH.

(* A generic principle to prove that the property Q n (Nthtl n s) holds for 
   any n *)

Section Invariant_Stream.

Variable s : Str.
Variable Inv : nat->Str->Prop.
Hypothesis Inv_init : (Inv O s).
Hint Inv_init.

Section Generic.
Variable Q : nat->Str->Prop.
Hypothesis Inv_out : (n:nat)(s:Str)(Inv n s)->(Q n s)/\(Inv (S n) (Tl s)).

Theorem All_Q_Nthtl : (n:nat)(Q n (NthTl s n)).
Intro n.
Cut (Inv n (NthTl s n)).
Linear with Inv_out.
Elim n; Trivial; Intros.
Simpl; Linear with H Inv_out.
Save.
End Generic.

(* Principle to prove that the property Q n holds for the elements of s *)

Variable Q : nat->A->Prop.
Lemma All_Q_Nth : 
  ((n:nat)(s:Str)(Inv n s)->(Q n (Hd s)) /\ (Inv (S n) (Tl s)))
      ->  (n:nat)(Q n (Nth s n)).
Proof (All_Q_Nthtl [n:nat][s:Str](Q n (Hd s))).

End Invariant_Stream.

(* Other lemma working on the implementation *)

Section Invariant_Implementation.
Variable X : Set.
Variable h : X->A.
Variable t : X->X.
Variable x : X.
Variable Inv : nat->X->Prop.
Hypothesis Inv_init : (Inv O x).
Hint Inv_init.

Section Nth_Implementation.
Variable Q : nat -> A -> Prop.
Hypothesis Inv_out : (n:nat)(x:X)(Inv n x)->(Q n (h x))/\(Inv (S n) (t x)).

Theorem All_Q_Nth_Imp : (n:nat)(Q n (Nth (StrIt X h t x) n)).
Intro n; Apply All_Q_Nth 
         with Inv:=[n:nat][s:Str](Ex [x:X]((s=(StrIt X h t x)) /\ (Inv n x))).
Linear with refl_equal Inv_init.
Destruct 1; Destruct 1; Intro.
Rewrite H1; Rewrite Hd_StrIt; Rewrite Tl_StrIt.
Linear with refl_equal Inv_out.
(***
Intro; Elim (Inv_out n0 x0); Trivial.
Split; [Trivial | Exists (t x0); Auto].
***)
Qed.
End Nth_Implementation.

End Invariant_Implementation.

End Streams.

Syntactic Definition StrOut := (STROUT ?).
Syntactic Definition Nth := (NTH ?).
Syntactic Definition NthTl := (NTHTL ?).
Syntactic Definition Hd := (HD ?).
Syntactic Definition Tl := (TL ?).
Syntactic Definition StrCoIt := (STRCOIT ? ?).
Syntactic Definition StrIt := (STRIT ? ?).

Syntax constr StrOut <<(STROUT $1 $t1)>> 1  <$t:"term":L> 
  with $t:= (APPLIST StrOut $t1).

Syntax constr Nth1 <<(NTH $1 $t1)>> 1  <$t:"term":L> 
  with $t:= (APPLIST Nth $t1).

Syntax constr Nth <<(NTH $1 $t1 $t2)>> 1  <$t:"term":L> 
  with $t:= (APPLIST Nth $t1 $t2).

Syntax constr Nthtl <<(NTHTL $1 $t1 $t2)>> 1 <$t:"term":L> 
    with $t:= (APPLIST NthTl $t1 $t2).

Syntax constr Hd <<(HD $1 $t1)>> 1  <$t:"term":L> 
  with $t:= (APPLIST Hd $t1).

Syntax constr Tl <<(TL $1 $t1)>>  1  <$t:"term":L> 
  with $t:= (APPLIST Tl $t1).

Syntax constr StrCoIt3 <<(STRCOIT $1 $2 $t1 $t2 $t3)>> 1  <$t:"term":L> 
  with $t:= (APPLIST StrCoIt $t1 $t2 $t3).

Syntax constr StrCoIt2 <<(STRCOIT $1 $2 $t1 $t2)>> 1  <$t:"term":L> 
  with $t:= (APPLIST StrCoIt $t1 $t2).

Syntax constr StrIt3 <<(STRIT $1 $2 $t1 $t2 $t3)>> 1  <$t:"term":L> 
  with $t:= (APPLIST StrIt $t1 $t2 $t3).

Syntax constr StrIt2 <<(STRIT $1 $2 $t1 $t2)>> 1   <$t:"term":L> 
  with $t:= (APPLIST StrIt $t1 $t2).




