(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 29th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                Alter.v                                   *)
(****************************************************************************)




Section Alter_I.

(*****************************************************************)
(* A proof of the extensional equality of two different versions *)
(* of the stream that infinitely alternates the values true and  *)
(* false, using F. Leclerc and C. Paulin representation.         *)
(*****************************************************************)


Require Bool.

Inductive St : Set  := build : (X:Set)(X->bool*X)->X->St.

Definition hd := 
  [H:St]
 <[_:St]bool>Case H of
    ([X:Set][p:X->bool*X][y:X](fst bool X (p y)))
    end.

Definition tl := 
[H:St]
 <[_:St]St>Case H of
    ([X:Set][p:X->bool*X][y:X](build X p (snd bool X (p y))))
    end.

(* The quality between streams is the gfp of this operator T *)

Definition T := 
 [R:St->St->Prop]
   [z1:St][z2:St](((hd z1)=(hd z2))/\ (R (tl z1) (tl z2))).

Inductive EqSt : St->St->Prop :=
       buildeq : (R:St->St->Prop)
		 ((x:St)(y:St)(R x y)->(T R x y))->
		 (s1:St)
		 (s2:St)
		 (R s1 s2) ->
		 (EqSt s1 s2).

(* Two process generating the same stream *)

Local gen1 := (build bool [b:bool](pair bool bool b       (neg b))).
Local gen2 := (build bool [b:bool](pair bool bool (neg b) (neg b))).

Local alter1 := (gen1 true).
Local alter2 := (gen2 false).

(* An invariant relation in the generation *)

Definition Invariant := 
[s1:St][s2:St] <bool>Ex([b:bool]s1=(gen1 (neg b))/\s2=(gen2 b)).

(* The proof that both processes generate the same stream *)

Theorem eqalters_I : (EqSt alter1 alter2).
Apply (buildeq Invariant).

Intros.
Case H.
Intros.
Case H0.
Intros.
(Rewrite -> H1;Rewrite -> H2).
Simpl.
Split.
Try Trivial.
Unfold Invariant .
Exists (neg x0).
Unfold gen1 .
Unfold gen2 .
Split.
Trivial.
Trivial.

Unfold Invariant .
Exists false.
Split.
Trivial.
Trivial.
Qed. 

(* An easier way of definig the same invariant relation presented above *)

Inductive SimplerInvariant : St->St->Prop :=
  makeinv : (b:bool)(SimplerInvariant (gen1 (neg b)) (gen2 b)).

Theorem eqalters_I_bis : (EqSt alter1 alter2).
Apply (buildeq SimplerInvariant).
Intros.
Case H.
Unfold T.
Intros.
Split.
Trivial.
Apply (makeinv (neg b)).
Apply (makeinv false).
Qed. 

(***********************************************************************)
(*  A different proof using the method of translation described in the *) 
(*  paper "Codifying guarded definitions with recursive schemes"       *)
(***********************************************************************)


(* The invariant relation *)

Inductive Cls : St->St->Prop := cls : (Cls alter1 alter2).

Inductive EqSt1 : St -> St -> Prop :=
  eqst1 : (z1,z2:St)
            ((hd z1)=(hd z2))
                ->((Cls (tl z1) (tl z2))\/(EqSt1 (tl z1) (tl z2)))
                    ->(EqSt1 z1 z2).


(* The streams alter1 and alter2 verify the relation EqSt1 *)

Lemma eqalters_body : (EqSt1 alter1 alter2).
Apply eqst1.
Trivial.
Right.
Apply eqst1.
Trivial.
Left.
Apply cls.
Qed.


(* EqSt1 is a  bisimulation *)

Lemma  H : (z1:St)(z2:St)(EqSt1 z1 z2)->(T EqSt1 z1 z2). 
Intros z1 z2 x.
Red.
Case x.
Intros z3 z4 H0 H1.
Split.
Assumption.
Case H1.
Intro H2.
Case H2.
Apply eqalters_body.
Intros.
Assumption.
Qed. 

(* The proof of the extensional equality of alter1 and alter2 *)

Theorem eqalters_II : (EqSt alter1 alter2).
Apply (buildeq EqSt1 H).
Apply eqalters_body.
Qed.


End Alter_I.


Section Alter_II. 
(* The same proof but using the new co-inductive types  *)

Require Streams.

CoFixpoint gen1 : bool->(Stream bool) := 
[b:bool](cons b (gen1 (neg b))).

CoFixpoint gen2 : bool->(Stream bool) := 
[b:bool](cons (neg b) (gen2 (neg b))).

Local alter1 := (gen1 false).
Local alter2 := (gen2 true ).


Theorem eqalters_III : (EqSt bool alter1 alter2).
Cofix u.
Apply eqst.
Trivial.
Apply eqst.
Trivial.
Assumption.
Qed.

End Alter_II.


