(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10.14                              *)
(*                                 April 1995                               *)
(*                                                                          *)
(****************************************************************************)
(*                                Streams.v                                 *)
(****************************************************************************)
(*       An example of the new implementation of co-inductive types         *)
(*                       by Eduardo Gimenez                                 *)
(****************************************************************************)

Require Plus.
Require Streams.


Section Examples. (* Introducing some streams *)


Section Map.

Variables A,B:Set.
Variable f:A->B.

CoFixpoint map : (Stream A)->(Stream B) :=
  [s:(Stream A)]
    <(Stream B)>Case s of
	          [a:A][s:(Stream A)](cons (f a) (map s))
                end.
End Map.


CoFixpoint zeros : (Stream nat) := (cons O zeros).

CoFixpoint alter : (Stream nat) := (cons O (cons (S O) alter)).

CoFixpoint alter1 : (Stream nat) := (cons O     alter2)
with       alter2 : (Stream nat) := (cons (S O) alter1).

CoFixpoint from : nat->(Stream nat) := [n:nat](cons n (from (S n))).
Definition allnats   : (Stream nat)      := (from O).


Lemma all_greater_than_m__are_here : (n,m:nat)(nth nat n (from m))=(plus n m).
Induction n.
Trivial.
Intros.
Rewrite -> plus_Snm_nSm.
Simpl.
Apply H.
Qed.

CoFixpoint mapS : (Stream nat)->(Stream nat) :=
  [s:(Stream nat)]
    <(Stream nat)>Case s of
	            [n:nat][s1:(Stream nat)](cons (S n) (mapS s1))
  	          end.


(* A different way of defining the set of streams *)

Inductive Set Times [A,B:Set] := 
                           times : A->B->(Times A B).


CoInductive Set St2  := cons2 : (Times nat St2)->St2.


CoFixpoint mapS2 :St2->St2 := 
[x:St2]
  <St2>Case x of 
         [y:(Times nat St2)]
              <St2>Case y of
                     [x1:nat][x2:St2](cons2 (times nat St2 
			                           (S x1) 
						   (mapS2 x2) ) )
                   end
         end.



(* Some examples of  wrong definitions *)
(*
(* 1: Un unguarded recursive call *)

CoFixpoint filter : (nat->bool)->(Stream nat)->(Stream nat) :=
  [p:nat->bool]
     [x:(Stream nat)]<(Stream nat)>Case x of
	           [n:nat][s:(Stream nat)]
                       <(Stream nat)>if  (p n) then (cons n (filter p s)) 
					       else (filter p s)
	       end. 

(* 2: A recursive call in the argument of a Case *)

CoFixpoint wrong : (Stream nat) := (cons O (tl nat wrong)).

(* 3: Nested recursive calls *)


CoFixpoint nested : (Stream nat)->(Stream nat) := [s:(Stream nat)](cons O (nested (nested s))).


(* An inductive definition on a coinductive set *)

Fixpoint wrong_mapS [s:(Stream nat)] : (Stream nat) :=
  <(Stream nat)>Case s of 
	 [n:nat][s1:(Stream nat)](cons (S n) (wrong_mapS s1))
      end.


(* A Cofixpoint definition defining an infinite element of an inductive set *)

CoFixpoint u:nat := (S u).


(* A recursive call in a non-recursive argument of a constructor *)

Mutual CoInductive V : Prop := consv : (A:Prop)A->(not A)->V.

Definition V_is_false := 
    [x:V]<False>Case x of 
	            [A:Prop][a:A][h:(not A)](h a)
	        end.

CoFixpoint V_is_true : V := (consv V  V_is_true V_is_false).


(* A sound definition which does not verfy the guarded condition *)

CoFixpoint good_from : (Stream nat) := (cons O (mapS good_from)).

*)

(* Unfolding the definition of a stream *)


Theorem zeros_unfold : zeros=(cons O zeros).
Apply (unfold_Stream nat zeros).
Qed.

End Examples.



Section Stream_Equalities. 

Variable A : Set.


(* The definition given in theories/Streams.v 
   is equivalent to require the elements at each position to be equal *)

Theorem Equiv1 :(s1,s2:(Stream A))(EqSt A s1 s2)->(n:nat)(nth A n s1)=(nth A n s2).
Fix u / 4.
Case H.
Intros.
Case n.
Assumption.
Intros.
Simpl.
Apply u.
Assumption.
Qed.

Theorem Equiv2 : (s1,s2:(Stream A))((n:nat)(nth A n s1)=(nth A n s2))->(EqSt A s1 s2).
Cofix u.
Intros.
Apply eqst.
Apply (H O).
Apply u.
Intros n.
Apply (H (S n)).
Qed.


Section Example.

Variable f:A->A.

CoFixpoint iter : A->(Stream A) := [x:A](cons (f x) (iter (f x))).

Lemma map_iter_eq : (x:A)(EqSt A (iter (f x)) (map A A f (iter x))).
Cofix u.
Intros.
Apply eqst.
Try Trivial.
Simpl.
Apply u.
Qed.

End Example.


(* Extensional Equality between two streams (opus II) *)


CoInductive EqSt2 [A:Set] : (Stream A)->(Stream A)->Prop := 
		   eqst2 : (s1,s2:(Stream A))
			   (a : A)
			   (EqSt2 A s1 s2)
			     ->(EqSt2 A (cons a s1) (cons a s2)).



(* The proof that this description  is equivalent to the one in the
   directory theories *)

(* The proof that EqSt2 is an equivalence relation *)


Theorem EqSt2_reflex : (s:(Stream A))(EqSt2 A s s).
Cofix u.
Intros.
Case s.
Intros.
Apply eqst2.
Apply u.
Qed.


Theorem EqSt2_sym : (s1:(Stream A))(s2:(Stream A))(EqSt2 A s1 s2)->(EqSt2 A s2 s1).
Cofix u.
Intros.
Case H.
Intros.
Apply eqst2.
Apply u.
Assumption.
Qed.


Theorem EqSt2_trans : (s1,s2,s3:(Stream A))(EqSt2 A s1 s2)->(EqSt2 A  s2 s3)->(EqSt2 A s1 s3).
Cofix u.
Intros s1 s2 s3.
Intro.
Inversion_clear H.
Intro.
Inversion_clear H.
Apply eqst2.
Apply (u s0 s4 s6).
Assumption.
Assumption.
Qed.


Definition derived_eqst :(s1,s2:(Stream A))(a:A)(EqSt A s1 s2)->(EqSt A (cons a s1) (cons a s2)) := 
[s1,s2:(Stream A)]
 [a:A]
  [H:(EqSt A s1 s2)]
   (eqst A (cons a s1) (cons a s2)
     (refl_equal A (hd (cons a s2))) H).

Definition derived_eqst2 : 
  (s1,s2:(Stream A))((hd s1)=(hd s2))->(EqSt2 A (tl s1) (tl s2))->(EqSt2 A s1 s2) :=
[s1,s2:(Stream A)]
 <([s:(Stream A)]
    (hd s)=(hd s2)->(EqSt2 A (tl s) (tl s2))->(EqSt2 A s s2))>
   Case s1 of
    <([s:(Stream A)]
       (a:A)
        (s0:(Stream A))
         (hd (cons a s0))=(hd s)
          ->(EqSt2 A (tl (cons a s0)) (tl s))
             ->(EqSt2 A (cons a s0) s))>
      Case s2 of
       ([a:A]
         [s:(Stream A)]
          [a0:A]
           [s0:(Stream A)]
            [H:(hd (cons a0 s0))=(hd (cons a s))]
             [H0:(EqSt2 A (tl (cons a0 s0)) (tl (cons a s)))]
              <([a1:A](EqSt2 A (cons a0 s0) (cons a1 s)))>Case H of
                 (eqst2 A s0 s a0 H0)
                 end)
       end
    end.


End Stream_Equalities.

