(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                               Process.v                                  *)
(****************************************************************************)
(*      An interpreter for a Calculus of Broadcasting Systems               *)
(*                                                                          *)
(*                     by Eduardo Gimenez                                   *)
(*                                                                          *)
(*              Developped in V5.10  June 1995                              *)
(****************************************************************************)


Require Streams.

(*************************)
   Section  Interpreter.
(*************************)

Variable A : Set.

CoInductive Set Process := 
	TALK   :  A->Process->(A->Process)->Process |
	LISTEN :  (A->Process)->Process              |
	PAR    :  Process->Process->Process.



Inductive Set Signal :=
        Noise :    Signal |
        Clear  : A->Signal.


Inductive Set KofA := Fail : KofA | Succeed : KofA.

Inductive Set Oracle :=
  Here  : Oracle         |
  Left  : Oracle->Oracle |
  Right : Oracle->Oracle.


Definition Pointer := KofA*Oracle.

CoInductive Set Answer :=
  Evolve : Process->Process->Signal->Answer->Answer |
  Refuse : Process->Answer->Answer.


Section listen_program.

Variable v:A.

CoFixpoint listen : Process->Process :=
[p:Process]
  <Process>Case p of
                 [_:A][_:Process][f:A->Process](f v) 

	         [f:A->Process](f v)  

                 [p:Process][q:Process](PAR (listen p) (listen q))
	   end.

End listen_program.


Section Say_and_Become_Program.

Definition Result := (Exc (Signal*Process)).

Definition analysewith := 
  [w:Result]
    [f:Signal->Process->Result]
       <Result>Case w of 
                 [p:Signal*Process]<Result>Case p of [w:Signal][p:Process](f w p) end
                 Error 
                end.


Variable r:KofA.

Fixpoint  say [p:Process;o:Oracle] : Result :=
  <Result>Case p of
	   [v:A][q:Process][_:A->Process]
                             <Result>if r then (Value ( Noise,   q))
					  else (Value ((Clear v),q))
	   [_:A->Process]Error

           [p:Process][q:Process]
              <Result>Case o  of
	 	               Error
                               [ol:Oracle](analysewith 
                                            (say p ol)
			                    [w:Signal][p1:Process]
                                              <Result>Case w of
      	                                                 (Value (Noise,(PAR p1 q)))
                                                          [v:A](Value ((Clear v),(PAR p1 (listen v q))))
						      end)
			       [or:Oracle](analysewith 
                                            (say q or)
			                    [w:Signal][q1:Process]
                                              <Result>Case w of
      	                                                  (Value  (Noise,(PAR p q1)))
                                                          [v:A](Value ((Clear v),(PAR (listen v p) q1)))
						      end)
                      end
	  end.

End Say_and_Become_Program.

CoFixpoint simulator : Process->(Stream Pointer)->Answer :=
       [p:Process]
        [sp:(Stream Pointer)]
         <Answer>let (r,o) = (hd sp) in
            <Answer>Case (say r p o) of
                  [wq:Signal*Process]
                    <Answer>let (w,q)= wq in
                       (Evolve p q w (simulator q (tl sp)))
                      
                  (Refuse p (simulator p (tl sp)))
                 end.



(**************************)
(*     Verification       *)
(**************************)

(* Semantics of processes *)

Inductive Set  Action  :=
	Transmit : Signal->Action |
	Receive  : A->Action.


Inductive UnrelProcTrans : Process ->Action->Process->Prop :=

	Uttalk1 : (f:A ->Process)
		  (v:A)
		  (p:Process)
	          (UnrelProcTrans (TALK v p f) (Transmit (Clear v)) p) |

	Uttalk2 : (f:A->Process)
		  (v,w:A)
		  (p1,p2:Process)
		  ((f v)=p2)->
		  (UnrelProcTrans (TALK w p1 f) (Receive v) p2) |

	Uttalk3 : (f:A ->Process)
		  (v:A)
		  (p:Process)
	          (UnrelProcTrans (TALK v p f) (Transmit Noise)  p) |


	Utlisten1: (v:A)
		   (f:A->Process)
	           (p:Process)
		   ((f v)=p)->
 		   (UnrelProcTrans (LISTEN f) (Receive v) p) |

	Utpar1  : (p1,p2,p3,p4:Process)
		  (v:A)
	          (UnrelProcTrans  p1 (Transmit (Clear v))  p2)->
		  (UnrelProcTrans  p3 (Receive  v)  p4)->
		  (UnrelProcTrans (PAR p1 p3) (Transmit (Clear v)) (PAR p2 p4)) |

	Utpar2  : (p1,p2,p3,p4:Process)
		  (v:A)
	          (UnrelProcTrans  p1 (Receive  v)  p2)->
		  (UnrelProcTrans  p3 (Transmit (Clear v))  p4)->
		  (UnrelProcTrans (PAR p1 p3) (Transmit (Clear v)) (PAR p2 p4)) |

	Utpar3  : (p1,p2,p3,p4:Process)
		  (v:A)
		  (UnrelProcTrans  p1 (Receive v)  p2)->
		  (UnrelProcTrans  p3 (Receive v)  p4)->
		  (UnrelProcTrans (PAR p1 p3) (Receive v) (PAR p2 p4)) |

      	Utpar5  : (p1,p2,p3:Process)
		  (UnrelProcTrans  p1 (Transmit Noise)  p2)->
		  (UnrelProcTrans (PAR p1 p3) (Transmit Noise) (PAR p2 p3))        |
 
     	Utpar6  : (p1,p3,p4:Process)
		  (UnrelProcTrans  p3 (Transmit Noise)  p4)->
		  (UnrelProcTrans (PAR p1 p3) (Transmit Noise) (PAR p1 p4)).




(* Properties of processes *)

Inductive IsFinite : Process->Prop :=
  tfinite : (f:A ->Process)
            (v:A)
            (p:Process)
	    (IsFinite (TALK v p f)) |

  lfinite : (f:A ->Process)
            (IsFinite (LISTEN f))   |

      
  pfinite : (p,q:Process)
            (IsFinite p)->
	    (IsFinite q)->
	    (IsFinite (PAR p q)).


Inductive BadPointer : Process->Oracle->Prop :=
        lsncase  : (f:A ->Process)
	           (o:Oracle)
                   (BadPointer (LISTEN f) o) |

	parcase  : (p1,p2:Process)
		   (BadPointer (PAR p1 p2) Here) |

	deeperr  : (p1,p2:Process)
		   (o:Oracle)
		   (BadPointer p2 o)-> 
                   (BadPointer (PAR p1 p2)  (Right o)) |

      	deeperl  : (p1,p2:Process)
		   (o:Oracle)
		   (BadPointer p1 o)-> 
                   (BadPointer (PAR p1 p2)  (Left o)).


CoInductive ForAlwaysFinite :Process -> Prop :=
 Faf : (p:Process)
       (IsFinite p)->
       ((w:Signal)(q:Process)(UnrelProcTrans p (Transmit w) q)->(ForAlwaysFinite q))->
       (ForAlwaysFinite p).


CoInductive Trace : Process->Set :=
 Tevolve  : (p,q:Process)
            (w:Signal)
            (UnrelProcTrans p (Transmit w) q)->
 	    (Trace q)->
	    (Trace p)    |

 Trefuses : (p:Process)
       	    (Ex [o:Oracle](BadPointer p o))->
	    (Trace p)->
            (Trace p).



(* Unfolding lemmas *)

Lemma unfold_process : (p:Process)(p=(<Process>Case p of TALK LISTEN PAR end)).
Intro p.
Case  p.
Trivial.
Trivial.
Trivial.
Qed.

Lemma unfold_listen1 :
(v:A)(a:A)(q:Process)(f:A->Process)
  (listen v (TALK a q f)) =(f v).
Intros.
Rewrite -> (unfold_process (listen v (TALK a q f))).
Simpl.
Symmetry.
Apply unfold_process.
Qed.

Lemma unfold_listen2 :
(v:A)(f:A->Process)
  (listen v (LISTEN  f)) = (f v).
Intros.
Rewrite -> (unfold_process (listen v (LISTEN f))).
Simpl.
Symmetry.
Apply unfold_process.
Qed.

Lemma unfold_listen3 :
(v:A)(p,q:Process)
  (listen v (PAR p q)) = (PAR (listen v p) (listen v q)).
Intros.
Rewrite -> (unfold_process (listen v (PAR p q))).
Trivial.
Qed.


(* Verification of listen *) 

Lemma listen_lemma :
(v:A)
 (p:Process)
   (IsFinite p)
       -> (UnrelProcTrans p (Receive v) (listen v p)).
Intros.
Elim H.
Intros.
Rewrite unfold_listen1.
Apply Uttalk2; Auto.
Intros.
Rewrite unfold_listen2.
Apply Utlisten1; Auto.
Intros.
Rewrite unfold_listen3.
Apply Utpar3; Auto.
Qed.


Section Verification_of_say.

Variable r:KofA.

Lemma say_specification  : 
 (p:Process)
   (o:Oracle)
     (IsFinite p)->
       {w:Signal & { q:Process | (UnrelProcTrans p (Transmit w) q)}} + {(BadPointer p o)}.

Realizer [p:Process][o:Oracle](say r p o).
Do 7 Program.
Do 4 Program.
(Do 2 Program;Apply Uttalk3).
(Do  2 Program;Apply Uttalk1).
(Do  2 Program;Apply lsncase).
 Do  2 Program. Program. Program. Apply parcase.
 Do  2 Program. Inversion_clear H.Assumption.
 Do 10 Program. Apply Utpar5;Assumption.
 Program. 
 Apply Utpar1. 
 Assumption. 
 Apply listen_lemma.
 Inversion_clear H;Assumption.
 Program. Apply deeperl. Assumption.
 Do 2 Program. (Inversion_clear H;Assumption).
 Do 10 Program. Apply Utpar6. Assumption.
 Program. Apply Utpar2. Apply listen_lemma. (Inversion_clear H;Assumption). Assumption.
 Program. Apply deeperr. Assumption.
Qed. 


End Verification_of_say.


(* Verification of the simulator *)


Lemma ConcludeIsFinite : 
 (p:Process)
   (ForAlwaysFinite p)->(IsFinite p).
Intros.
Case H.
Auto.
Qed.


Theorem simulator_specification  : 
 (p:Process)
   (ForAlwaysFinite p)->
      (Stream (KofA*Oracle))->(Trace p).
Realizer 
CoFix simulator{simulator : Process->(Stream Pointer)->Answer :=
       [p:Process]
        [sp:(Stream Pointer)]
         <Answer>let (r,o) = (hd sp) in
            <Answer>Case (say r p o) :: ::{ {w:Signal & 
                                               {q:Process | (UnrelProcTrans p (Transmit w) q)}}
                                             + {(BadPointer p o)} } of
                  [wq:Signal*Process]
                    <Answer>let (w,q)= wq in
                       (Evolve p q w (simulator q (tl sp)))
                      

                  (Refuse p (simulator p (tl sp)))
                 end}.
Program.
Program.
Program.
Program.
Program.
Program.
Program.
Program.

Intros.
Apply (say_specification r p0 o0).
Apply ConcludeIsFinite.
Assumption.
 
Program.
Program.
Program.
Program.
Program.
Program.

Assumption. 
Inversion_clear H.
Apply (H2 w q0).
Assumption.
 
Program.

Exists o.
Assumption.
Assumption.
Qed.

End Interpreter.
