(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                  Zadd.v                                  *)
(****************************************************************************)

Require Arith.
Require Nat_complements.
Require Lci.
Require groups.
Require rings.
Require Zbase.
Require Z_succ_pred.

(* Addition on Z, (Z, +) is a group *)

Recursive Definition addZ : Z -> Z -> Z := 
       OZ        y => y
|   (pos O)      y => (succZ y)
|   (pos (S n1)) y => (succZ (addZ (pos n1) y))
|   (neg O)      y => (predZ y)
|   (neg (S n1)) y => (predZ (addZ (neg n1) y)).


(************* was
Definition addZ := 
  [x, y: Z] (<Z> Match x with
    (* OZ *)    y
    (* pos n *) [n: nat] (<Z> Match n with 
       (* O *)    (succZ y)
       (* S n1 *) [n1: nat] [q: Z] (succZ q) end)
    (* neg n *) [n: nat] (<Z> Match n with 
       (* O *)    (predZ y)
       (* S n1 *) [n1: nat] [q: Z] (predZ q) end) end).
**************)

(***************)
Lemma succ_addZ_l: (x, y: Z) (addZ (succZ x) y) = (succZ (addZ x y)).

Intros; Elim x.
(* OZ *)
Reflexivity.
(* pos n *)
Trivial.
(* neg n *)
Destruct n.
(* neg O *)
Simpl; Symmetry; Exact (succ_predZ y). 
(* neg (S n0) *)
Intros; Symmetry; Rewrite addZ_eq5.
Apply succ_predZ.
Save.

(***************)
Lemma pred_addZ_l: (x, y: Z) (addZ (predZ x) y) = (predZ (addZ x y)).

Intros; Elim x.
(* OZ *)
Reflexivity.
(* pos n *)
Destruct n.
Simpl; Rewrite pred_succZ; Trivial.
Intros; Rewrite addZ_eq3; Rewrite pred_succZ; Trivial.
Trivial.
Save.

(**********************)
Lemma tech_add_pos_succZ :
  (x: nat) (y: Z) (addZ (pos (S x)) y) = (succZ (addZ (pos x) y)).
Proof addZ_eq3.

(**********************)
Lemma tech_add_neg_predZ: 
   (x: nat) (y: Z) (addZ (neg (S x)) y) = (predZ (addZ (neg x) y)).
Proof addZ_eq5.

(***************)
Lemma succ_addZ_r: (x, y: Z) (addZ x (succZ y)) = (succZ (addZ x y)).

Intros; Elim x.
(* OZ *)
Reflexivity.
(* pos n *)
Induction n.
(* pos O *)
Reflexivity.
(* pos y0 *)
Intros.
Do 2 Rewrite (tech_add_pos_succZ n0).
Elim H; Reflexivity.
(* neg n *)
Induction n.
(* neg O *)
Simpl; Symmetry; Apply succ_pred_pred_succZ.
(* neg y0 *)
Intros.
Do 2  Rewrite (tech_add_neg_predZ n0).
Rewrite H.
Symmetry; Apply succ_pred_pred_succZ.
Save.

(***************)
Lemma pred_addZ_r: (x, y: Z) (addZ x (predZ y)) = (predZ (addZ x y)).

Intros; Elim x.
(* OZ *)
Reflexivity.
(* (pos n) *)
Induction n.
(* (pos O) *)
Simpl; Apply succ_pred_pred_succZ.
(* (pos (S y0)) *)
Intros.
Do 2 Rewrite (tech_add_pos_succZ n0).
Rewrite H; Apply succ_pred_pred_succZ.
(* neg n *)
Induction n.
(* neg O *)
Reflexivity.
(* neg (S y0) *)
Intros.
Do 2  Rewrite (tech_add_neg_predZ n0).
Elim H; Reflexivity.
Save.

(***********)
Lemma add_OZ: (x: Z) (addZ x OZ) = x.

Induction x.
(* OZ *)
Reflexivity.
(* pos n *)
Induction n.
Reflexivity.
Intros; Rewrite tech_add_pos_succZ; Rewrite H; Reflexivity.
(* neg n *)
Induction n.
Reflexivity.
Intros; Rewrite tech_add_neg_predZ; Rewrite H; Reflexivity.
Save.

(*****************)
Lemma add_IZ_succZ: (x: Z) (addZ x IZ) = (succZ x).

Intros.
Cut ((succZ OZ) = IZ); Intros. Elim H.
Rewrite (succ_addZ_r x OZ); Rewrite (add_OZ x); Reflexivity.
Reflexivity.
Save.

(*****************)
Lemma add_mIZ_predZ: (x: Z) (addZ x (neg O)) = (predZ x).

Intros.
Cut ((predZ OZ) = (neg O)); Intros. Elim H.
Rewrite (pred_addZ_r x OZ); Rewrite (add_OZ x); Reflexivity.
Reflexivity.
Save.

(*************************)
Theorem addZ_commutativity: (commutativity Z addZ).

Unfold commutativity; Intros; Elim x.
(* OZ *)
Simpl; Symmetry; Exact (add_OZ y).
(* pos n *)
Induction n.
(* pos O *)
Simpl; Symmetry; Exact (add_IZ_succZ y).
(* pos (S y0) *)
Intros; Rewrite (tech_add_pos_succZ n0 y). Rewrite H.
Cut ((succZ (pos n0)) = (pos (S n0))); Intros. Elim H0.
Rewrite (succ_addZ_r y (pos n0)); Reflexivity.
Reflexivity.
(* neg n *)
Induction n.
(* neg O *)
Simpl; Symmetry; Exact (add_mIZ_predZ y).
(* neg (S y0) *)
Intros; Rewrite (tech_add_neg_predZ n0 y). Rewrite H.
Cut ((predZ (neg n0)) = (neg (S n0))); Intros. Elim H0.
Rewrite (pred_addZ_r y (neg n0)); Reflexivity.
Reflexivity.
Save.

(*************************)
Lemma tech_add_pos_neg_posZ: (n1, n2: nat) 
  (lt n2 n1) -> (addZ (pos n1) (neg n2)) = (pos (minus n1 (S n2))).

Induction n2.
(* O *)
Intros; Elim (addZ_commutativity (neg O) (pos n1)).
Rewrite addZ_eq4.
Elim minus_n_Sm; Trivial.
Elim minus_n_O.
Apply tech_pred_posZ; Trivial.
(* S y *)
Intros; Elim (addZ_commutativity (neg (S n)) (pos n1)).
Rewrite tech_add_neg_predZ.
Elim (addZ_commutativity (pos n1) (neg n)).
Rewrite H; Auto.
Elim (minus_n_Sm n1 (S n) H0).
Apply tech_pred_posZ.
Apply lt_minus2; Trivial.
Save.

(*************************)
Theorem addZ_associativity: (associativity Z addZ).

Unfold associativity; Intros; Elim x.
(* OZ *)
Unfold addZ; Reflexivity.
(* pos n *)
Intros; Elim n.
(* pos O *)
Simpl; Symmetry; Exact (succ_addZ_l y z).
(* pos (S y0) *)
Intros.
Do 2 Rewrite (tech_add_pos_succZ n0).
Rewrite (succ_addZ_l (addZ (pos n0) y) z); Elim H; Reflexivity.
(* neg n *)
Induction n.
(* neg O *)
Simpl; Symmetry; Exact (pred_addZ_l y z).
(* neg (S n0) *)
Intros.
Do 2 Rewrite (tech_add_neg_predZ n0).
Rewrite (pred_addZ_l (addZ (neg n0) y) z); Elim H; Reflexivity.
Save.

(************)
Definition IdZ [x: Z] True.

(*******************)
Theorem addZ_neutral: (neutral Z IdZ addZ OZ).

Unfold neutral; Intros.
Split. Exact I.
Intros. Split. Exact (add_OZ x). Unfold addZ; Reflexivity.
Save.

(*************)
Definition oppZ :=
    [x: Z] (<Z> Case x of
    (* OZ *)    OZ
    (* pos n *) [n: nat] (neg n)
    (* neg n *) [n: nat] (pos n)
    end).

(*************)
Lemma opp_succZ: (x: Z) (oppZ (succZ x)) = (predZ (oppZ x)).

Destruct x.
(* OZ *)
Reflexivity.
(* pos n *)
Intros; Reflexivity.
(* neg n *)
Destruct n; Intros; Reflexivity. 
Save.

(*************)
Lemma opp_predZ: (x: Z) (oppZ (predZ x)) = (succZ (oppZ x)).

Destruct x.
(* OZ *)
Reflexivity.
(* pos n *)
Destruct n; Intros; Reflexivity. 
(* neg n *)
Intros; Reflexivity.
Save.

(*********************)
Lemma tech_add_pos_negZ: (n: nat) (addZ (pos n) (neg n)) = OZ.

Induction n.
(* O *)
Reflexivity.
(* S y *)
Intros; Rewrite (tech_add_pos_succZ n0).
Elim succ_addZ_r; Exact H.
Save.

(*********************)
Lemma tech_add_neg_posZ: (n: nat) (addZ (neg n) (pos n)) = OZ.

Intros; Elim (addZ_commutativity (pos n) (neg n)); Exact (tech_add_pos_negZ n).
Save.

(*********************)
Lemma tech_add_pos_posZ: (n: nat) (m: nat) 
   (addZ (pos n) (pos m)) = (pos (S (plus n m))).

Intros; Elim n.
(* O *)
Reflexivity.
(* S n0 *)
Intros; Rewrite (tech_add_pos_succZ n0 (pos m)); Rewrite H; Reflexivity.
Save.

(*********************)
Lemma tech_add_neg_negZ: (n: nat) (m: nat) 
   (addZ (neg n) (neg m)) = (neg (S (plus n m))).

Induction n.
(* O *)
Reflexivity.
(* S n0 *)
Intros; Rewrite (tech_add_neg_predZ n0 (neg m)); Rewrite H; Reflexivity.
Save.

(********************)
Theorem addZ_opposite: (opposite Z IdZ addZ OZ oppZ).

Repeat Split; Trivial.
(* -> *)
Case x.
(* OZ *)
Reflexivity.
(* pos n *)
Intros; Exact (tech_add_pos_negZ n).
(* neg n *)
Intros; Exact (tech_add_neg_posZ n).
(* <- *)
Case x.
(* OZ *)
Reflexivity.
(* pos n *)
Intros; Exact (tech_add_neg_posZ n).
(* neg n *)
Intros; Exact (tech_add_pos_negZ n).
Save.

(**************)
Theorem Z_group: (is_group Z IdZ addZ OZ oppZ).

Split.
Red; Trivial.
Split. Exact addZ_associativity.
Split. Exact addZ_neutral.
Exact addZ_opposite.
Save.

(********************)
Theorem abs_eq_or_oppZ: (x: Z) {(absZ x) = x}+{(absZ x) = (oppZ x)}.

Destruct x; Auto.
Save.

(*********************)
Lemma tech_opp_pos_negZ: (n: nat) 
  ((oppZ (pos n)) = (neg n)) /\ ((oppZ (neg n)) = (pos n)).

Induction n; Auto.
Save.
