(****************************************************************************)
(*                                                                          *)
(*                                                                          *)
(*  Mini_ML.v                                                               *)
(*                                                                          *)
(*  This file contains the definitions of the Mini-ML and the               *)
(*  Categotical Abstract Machine (C.A.M) natural semantics                  *)
(*  It also contains the definition of the translation from                 *)
(*  Mini-ML to the CAM and the proof that this translation is               *)
(*  correct                                                                 *)
(*                                                                          *)
(*                                                                          *)
(*  Samuel Boutin                                                           *)
(*  Coq V5.10                                                               *)
(*  December  1994                                                          *)
(*                                                                          *)
(*                                                                          *)
(****************************************************************************)
(*                                Mini_ML.v                                 *)
(****************************************************************************)

(*****************************************************************************)
(*                CORRECTNESS PROOF OF A MINI_ML COMPILER                    *)
(*                             Samuel Boutin                                 *)
(*                             December 1994                                 *)
(*                              Coql  V5.10                                  *)
(*****************************************************************************)

(*****************************************************************************)
(* First we define a Set "OP" of variable which is a representation of the   *)
(* arithmetic functions *,+,-... that we will not describe explicitely       *)
(* We take the natural numbers to code the variables (but of course it has no*)
(* relation with the De bruijn indexes)                                      *)
(*****************************************************************************)

Parameter OP:Set.

Parameter eval_op:OP->nat->nat->nat.

Definition  Pat:=nat.

(*****************************************************************************)
(*              Defining the syntax of ML expressions                        *)
(*****************************************************************************)

Inductive MLexp : Set :=

  Bool: bool -> MLexp
 |Num: nat -> MLexp
 |op: OP -> MLexp
 |id: Pat -> MLexp
 |appl: MLexp -> MLexp -> MLexp
 |mlpair: MLexp -> MLexp -> MLexp
 |lambda: Pat -> MLexp -> MLexp
 |let': Pat -> MLexp -> MLexp -> MLexp
 |letrec: Pat -> Pat -> MLexp -> MLexp -> MLexp
 |ite: MLexp -> MLexp -> MLexp -> MLexp.

(*****************************************************************************)
(*     ML values and environment are defined with the following mutual       *)
(*     recursive definition                                                  *)
(*****************************************************************************)

Mutual Inductive 
MLval : Set :=
   num: nat -> MLval
 | boolean : bool -> MLval
 | valpair: MLval -> MLval -> MLval
 | OP_clos: OP -> MLval
 | Clos: Pat -> MLexp -> MLenv -> MLval
 | Clos_rec: Pat -> MLexp -> Pat -> MLenv -> MLval
with
MLenv : Set :=
   Enil: MLenv
 | Econs: Pat -> MLval -> MLenv -> MLenv.


(*****************************************************************************)
(* Definition of VAL_OF which decide if a variable belongs to an environment *)
(*****************************************************************************)

Inductive VAL_OF:MLenv->Pat->MLval->Prop:=

 ELT:(e:MLenv)(I:Pat)(a:MLval)
             (VAL_OF (Econs I a e) I a)

 |CHG:(e:MLenv)(X,I:Pat)(a,b:MLval)
         (VAL_OF e I a)->
                    ( ~(<Pat> X=I))->
             (VAL_OF (Econs X b e) I a).

(*****************************************************************************)
(*         The following lemma shows us that VAL_OF is deterministic         *)
(*****************************************************************************)

Lemma determ_VAL_OF :  (e:MLenv)(i:Pat)(V,V':MLval)
                (VAL_OF e i V')->
                    (VAL_OF e i V)->
                       (<MLval> V=V').

Induction 1.
Intros.
Inversion_clear H0;Auto.
Elim H2;Auto.
Intros.
Apply H1.
Inversion H3.
Elim H2.
Injection H4.
Intros.
Rewrite <- H5.
Elim H9;Auto.
Rewrite -> H7.
Injection H6.
Intro HH.
Rewrite -> HH.
Rewrite -> H8.
Intros;Auto.
Save.

(*****************************************************************************)
(*                   We define now the Mini-ML semantics                     *)
(*In this semantics we want to distinguish the places where application is   *)
(*used in a recursive scheme from the other places.                          *)
(*That's why we have a special rule for application in the case where the    *)
(*function applied has as value a recursive closure (this correspond in fact *)
(*to a test during application                                               *)
(*****************************************************************************)

Inductive ML_DS :MLenv->MLexp->MLval->Prop:=

 BOOL:(b:bool)(e:MLenv)(ML_DS e (Bool b) (boolean b))

 |NUM:(n:nat)(e:MLenv)(ML_DS e (Num n) (num n))

 |Sem_OP:(c:OP)(e:MLenv)(ML_DS e (op c) (OP_clos c))

 |LAMBDA:(e:MLenv)(P:Pat)(E:MLexp)
               (ML_DS e (lambda P E) (Clos P E e))

 |IDENT:(e:MLenv)(v:MLval)(I:Pat)
               (VAL_OF e I v)->
                  (ML_DS e (id I) v)

 |ITE1:(e:MLenv)(E1,E2,E3:MLexp)(v:MLval)
               (ML_DS e E1 (boolean true))->
                  (ML_DS e E2 v)->
                     (ML_DS e (ite E1 E2 E3) v)

 |ITE2:(e:MLenv)(E1,E2,E3:MLexp)(v:MLval)
               (ML_DS e E1 (boolean false))->
                  (ML_DS e E3 v)->
                     (ML_DS e (ite E1 E2 E3) v)
 
 |MLPAIR:(e:MLenv)(E1,E2:MLexp)(u,v:MLval)
               (ML_DS e E1 u)->
                  (ML_DS e E2 v)->
                     (ML_DS e (mlpair E1 E2) (valpair u v))

 |APPml1:(e,e1:MLenv)(P:Pat)(E,E1,E2:MLexp)(u,v:MLval)
               (ML_DS e E1 (Clos P E e1))->
                 (ML_DS e E2 u)->
                   (ML_DS (Econs P u e1) E v)->
                      (ML_DS e (appl E1 E2) v)

 |APPml2:(e,e1:MLenv)(x,P:Pat)(E,E1,E2:MLexp)(u,v:MLval)
               (ML_DS e E1 (Clos_rec x E P e1))-> 
                 (ML_DS e E2 u)->
                   (ML_DS (Econs x u (Econs P (Clos_rec x E P e1) e1)) E v)->
                     (ML_DS e (appl E1 E2) v)

 |APPml_op:(e:MLenv)(E1,E2:MLexp)(n,m:nat)(c:OP)
           (ML_DS e E1 (OP_clos c))->
               (ML_DS e E2 (valpair (num n) (num m)))->
                     (ML_DS e (appl E1 E2) (num (eval_op c n m)))

 |Sem_let:(e:MLenv)(P:Pat)(E1,E2:MLexp)(u,v:MLval)
               (ML_DS e E2 u)->
                  (ML_DS (Econs P u e) E1 v)->
                     (ML_DS e (let' P E2 E1) v)

 |Sem_letrec:(e:MLenv)(P:Pat)(x:Pat)(E,E2:MLexp)(u:MLval)
               (ML_DS (Econs P (Clos_rec x E P e) e) E2 u)->
                  (ML_DS e (letrec P x E E2) u).   


(*****************************************************************************)
(* This predicate can be seen as an injective partial function associating a *)
(*  value to an environment and an expression:                               *)
(*****************************************************************************)


Lemma ML_DS_determ :   
  (e:MLenv)(E:MLexp)(V:MLval)(ML_DS e E V)->
              (V':MLval)
                 (ML_DS e E V')->
                    (<MLval> V=V').


Induction 1;Intros.

Inversion_clear H0;Auto.
Inversion_clear H0;Auto.
Inversion_clear H0;Auto.
Inversion_clear H0;Auto.
Inversion_clear H1;Auto. Apply (determ_VAL_OF e0 I);Auto.

Inversion_clear H4.Exact (H3 V' H6).
Cut (boolean true)=(boolean false).
Intro HH;Discriminate HH.
Exact (H1 (boolean false) H5).

Inversion_clear H4.
Cut (boolean false)=(boolean true).
Intro HH;Discriminate HH.
Exact (H1 (boolean true) H5).
Exact (H3 V' H6).

Inversion_clear H4.
Elim (H1 u0 H5);Elim (H3 v0 H6);Trivial.

Inversion_clear H6.
Cut (Clos P E0 e1)=(Clos P0 E3 e3).
Intro HH;Injection HH.
Intros.
Cut u=u0.
Intro.
Apply H5.
Rewrite -> H6;Rewrite -> H10;Rewrite -> H11;Rewrite -> H12.
Exact H9.
Exact (H3 u0 H8).
Exact (H1 (Clos P0 E3 e3) H7).
Cut (Clos P E0 e1)=(Clos_rec x E3 P0 e3).
Intro HH;Discriminate HH.
Exact (H1 (Clos_rec x E3 P0 e3) H7).
Cut (Clos P E0 e1)=(OP_clos c).
Intro HH;Discriminate HH.
Exact (H1 (OP_clos c) H7).

Inversion_clear H6.
Cut (Clos_rec x E0 P e1)=(Clos P0 E3 e3).
Intro HH;Discriminate HH.
Exact (H1 (Clos P0 E3 e3) H7).
Cut (Clos_rec x E0 P e1)=(Clos_rec x0 E3 P0 e3).
Intro HH ; Injection HH.
Cut u=u0.
Intros.
Apply H5.
Rewrite->H6;Rewrite->H10;Rewrite->H11;Rewrite->H12;Rewrite->H13.
Exact H9.
Exact (H3 u0 H8).
Exact (H1 (Clos_rec x0 E3 P0 e3) H7).
Cut (Clos_rec x E0 P e1)=(OP_clos c).
Intro HH;Discriminate HH.
Exact (H1 (OP_clos c) H7).

Inversion_clear H4.
Cut (OP_clos c)=(Clos P E0 e2).
Intro HH;Discriminate HH.
Exact (H1 (Clos P E0 e2) H5).
Cut (OP_clos c)=(Clos_rec x E0 P e2).
Intro HH;Discriminate HH.
Exact (H1 (Clos_rec x E0 P e2) H5).
Cut (valpair (num n) (num m))=(valpair (num n0) (num m0)).
Intro HH;Injection HH.
Intros.
Cut (OP_clos c)=(OP_clos c0).
Intro HH1;Injection HH1.
Intro.
Rewrite->H4;Rewrite->H7;Rewrite->H8.
Trivial.
Exact (H1 (OP_clos c0) H5).
Exact (H3 (valpair (num n0) (num m0)) H6).

Inversion_clear H4.
Cut u=u0.
Intro.
Apply H3.
Rewrite->H4.
Exact H6.
Exact (H1 u0 H5).

Inversion_clear H2.
Apply H1.
Exact H3.

Save.

(*****************************************************************************)
(*We present now the definitions that concerns the Categorical Abstract      *)
(*Machine. The Set Commande defines its syntax                               *)
(*****************************************************************************)

Inductive Value : Set :=
null:Value
|elem:bool->Value
|int:nat->Value
|def_op:OP->Value.

(*****************************************************************************)

Inductive Commande : Set :=
quote:Value->Commande
 |car:Commande
 |cdr:Commande
 |cons:Commande
 |push:Commande
 |swap:Commande
 |branch:Commande->Commande->Commande
 |cur:Commande->Commande
 |cur_rec:Commande->Commande
 |app:Commande
 |o:Commande->Commande->Commande.

(*****************************************************************************)

Inductive CSem_val : Set :=
 val:Value->CSem_val
 |Cam_pair:CSem_val->CSem_val->CSem_val
 |Cam_clos:Commande->CSem_val->CSem_val 
 |Cam_clos_rec:Commande->CSem_val->CSem_val  
 |Cam_nil:CSem_val.

(*****************************************************************************)

Inductive Etat : Set :=
  nil:Etat
  |ETcons:CSem_val->Etat->Etat.
           
(*****************************************************************************)
(*               We provide now the Semantics of the CAM                     *)
(*****************************************************************************)


Inductive CAM_DS : Etat->Commande->Etat->Prop := 


  QUO:(s:Etat)(a:CSem_val)(b:Value)
        (CAM_DS (ETcons a s) (quote b) (ETcons (val b) s))

 |CAR:(s:Etat)(a,b:CSem_val)
        (CAM_DS (ETcons (Cam_pair a b) s) car (ETcons a s))

 |CDR:(s:Etat)(a,b:CSem_val)
        (CAM_DS (ETcons (Cam_pair a b) s) cdr (ETcons b s))

 |CONS:(s:Etat)(a,b:CSem_val)
        (CAM_DS (ETcons b (ETcons a s)) cons (ETcons (Cam_pair a b) s))

 |PUSH:(s:Etat)(a:CSem_val)
        (CAM_DS (ETcons a s) push (ETcons a (ETcons a s)))
 
 |SWAP:(s:Etat)(a,b:CSem_val)
        (CAM_DS (ETcons a (ETcons b s)) swap (ETcons b (ETcons a s)))
          
 |BRANCHT:(s,s1:Etat)(c1,c2:Commande)
        (CAM_DS s c1 s1)->
            (CAM_DS (ETcons (val (elem true)) s) (branch c1 c2) s1)
 
 |BRANCHF:(s,s2:Etat)(c1,c2:Commande)
        (CAM_DS s c2 s2)->
            (CAM_DS (ETcons (val (elem false)) s) (branch c1 c2) s2)

 |CUR:(s:Etat)(a:CSem_val)(c:Commande)
        (CAM_DS (ETcons a s) (cur c) (ETcons (Cam_clos c a) s))

 |APPcam1:(s,s1:Etat)(a,b:CSem_val)(c:Commande)
        (CAM_DS (ETcons (Cam_pair a b) s) c s1)->
            (CAM_DS (ETcons (Cam_pair (Cam_clos c a) b) s) app s1)
 
 |APPcam2:(s,s1:Etat)(a,b:CSem_val)(c:Commande)
    (CAM_DS (ETcons (Cam_pair (Cam_pair b (Cam_clos_rec c b)) a) s) c s1)->
            (CAM_DS (ETcons (Cam_pair (Cam_clos_rec c b) a) s) app s1)

 |APPcam_op:(s:Etat)(n,m:nat)(oper:OP)
     (CAM_DS (ETcons (Cam_pair (val (def_op oper))
       (Cam_pair (val (int n)) (val (int m)))) s) app 
           (ETcons (val (int (eval_op oper n m))) s)) 

 |CUR_REC:(s:Etat)(a:CSem_val)(c:Commande)
      (CAM_DS (ETcons a s) (cur_rec c) (ETcons (Cam_clos_rec c a) s))

 |o_DS:(s,s1,s2:Etat)(c1,c2:Commande)
       (CAM_DS s c1 s1)->
           (CAM_DS s1 c2 s2)->
               (CAM_DS s (o c1 c2) s2).
             

(*****************************************************************************)
(*We now define the "Squelette"of an environment which is the intuitive      *)
(*corresponding list of Debruijn indexes of an environment                   *)
(*****************************************************************************)


Inductive Squelette : Set :=
nil_squelette:Squelette
|cons_squelette:Pat->Squelette->Squelette.

(*****************************************************************************)

Inductive Habite : MLenv->Squelette->Prop :=

triv_habite:(Habite Enil nil_squelette)

|cons_habite:(x:Pat)(u:MLval)(e:MLenv)(s:Squelette)
              (Habite e s)->
               (Habite (Econs x u e) (cons_squelette x s)).

(*****************************************************************************)
(*                Habite is an injective w.r.t the Squelette argument        *)
(*****************************************************************************)

Lemma Habite_inject:
   (e:MLenv)(s1,s2:Squelette)
              (Habite e s1) ->
                (Habite e s2) ->
                   s1=s2.

Induction e.
Intros.
Inversion_clear H;Inversion_clear H0;Trivial.
Intros.
Inversion_clear H0;Inversion_clear H1.
Elim (H s s0 H2 H0);Trivial.
Save.

(*****************************************************************************)
(*  The Acces predicate is defined using the notion of Squelette             *)
(* It describes how to reach an identifier in an environment                 *)
(*****************************************************************************)

Inductive Access:Pat->Squelette->Commande->Prop:=


  Rule1:(P:Pat)(s:Squelette)
                (Access P (cons_squelette P s) cdr)

  |Rule2:(P,T:Pat)(s:Squelette)(C:Commande)
               (~<Pat> P=T)->
                (Access P s C)->
                 (Access P (cons_squelette T s) (o car C)).

(*****************************************************************************)
(*   The following small lemma shows that Access is injective                *)
(*****************************************************************************)

Lemma Access_inject :  (x:Pat)(s:Squelette)(C,C':Commande)
            (Access x s C')->
              (Access x s C)->
                (<Commande> C=C').

Intro.
Induction s;Intros.
Inversion_clear H0.
Simple Inversion H0.
Simple Inversion H1.
Elim H4;Elim H7;Trivial.
Injection H8;Injection H3.
Intros.
Elim H12.
Rewrite->H7;Rewrite<-H2;Rewrite->H6;Rewrite<-H11;Auto.
Rewrite->H4.
Simple Inversion H1.
Intros.
Elim H8.
Injection H5;Injection H3;Intros.
Rewrite<-H2;Rewrite->H11;Rewrite<-H13;Auto.
Rewrite->H7.
Injection H8;Injection H5.
Intros HH1 HH2 HH3.
Rewrite->HH1;Rewrite->HH3.
Rewrite<-H6;Rewrite<-H9.
Intros.
Elim (H C1 C0 H12 H10).
Trivial.
Save.


(*****************************************************************************)

Lemma Squelet :   (e:MLenv)((Ex [s:Squelette](Habite e s))).

Intro.
Pattern e.
Apply MLenv_ind.
Exists nil_squelette.
Exact triv_habite.
Intros.
Elim H;Intro s;Intro.
Exists (cons_squelette p s).
Exact (cons_habite p m m0 s H0).
Save.

(*****************************************************************************)
(*It's now time to define the translation of ML code onto CAM code           *)
(*****************************************************************************)


Inductive Traduction:Squelette->MLexp->Commande->Prop:=
            

Bool_Trad:(b:bool)(S:Squelette)
              (Traduction S (Bool b) (quote (elem b)))

|Trad_num:(n:nat)(S:Squelette)
              (Traduction S (Num n) (quote (int n)))

|Trad_clos:(c:OP)(S:Squelette)
              (Traduction S (op c) (quote (def_op c)))

|Trad_var:(p:Pat)(S:Squelette)(C:Commande)
              (Access p S C)->(Traduction S (id p) C)


|Trad_ite:(S:Squelette)(E1,E2,E3:MLexp)(C1,C2,C3:Commande)
              (Traduction S E1 C1)->
                (Traduction S E2 C2)->
                  (Traduction S E3 C3)->
  (Traduction S (ite E1 E2 E3) (o push (o C1 (branch C2 C3))))


|Trad_pair:(S:Squelette)(E1,E2:MLexp)(C1,C2:Commande)
               (Traduction S E1 C1)->
                 (Traduction S E2 C2)->
  (Traduction S (mlpair E1 E2) (o push (o C1 (o swap (o C2 cons)))))


|Trad_app:(S:Squelette)(E1,E2:MLexp)(C1,C2:Commande)
               (Traduction S E1 C1)->
                 (Traduction S E2 C2)->
  (Traduction S (appl E1 E2) (o push (o C1 (o swap (o C2 (o cons app))))))


|Trad_let:(p:Pat)(S:Squelette)(E1,E2:MLexp)(C1,C2:Commande)
              (Traduction S E1 C1)->
                 (Traduction (cons_squelette p S) E2 C2)->
  (Traduction S (let' p E1 E2) (o push (o C1 (o cons C2))))


|Trad_let_rec:(p,x:Pat)(S:Squelette)(E,E2:MLexp)(C,C2:Commande)
 (Traduction (cons_squelette x (cons_squelette p S)) E C)->
                 (Traduction (cons_squelette p S) E2 C2)->
  (Traduction S (letrec p x E E2) (o push (o (cur_rec C) (o cons C2))))


|Trad_lambda:(S:Squelette)(p:Pat)(E:MLexp)(C:Commande)
               (Traduction (cons_squelette p S) E C)->
                 (Traduction S (lambda p E) (cur C)). 


(*****************************************************************************)
(* The Traduction Predicate can be seen as a partial injective function      *)
(* that given a state and an ML expression yields a CAM sentence:            *)
(*****************************************************************************)

Lemma Traduction_inject :   (E:MLexp)(C,C':Commande)(s:Squelette)
              (Traduction s E C)->
                 (Traduction s E C')->
                    (<Commande> C=C').



Induction E;Intros.

Inversion_clear H.
Inversion_clear H0.
Trivial.

Inversion_clear H.
Inversion_clear H0.
Trivial.

Inversion_clear H.
Inversion_clear H0.
Trivial.

Inversion_clear H.
Inversion_clear H0.
Exact (Access_inject p s C C' H H1).

Inversion_clear H1.
Inversion_clear H2.
Elim (H C1 C0 s H3 H1).
Elim (H0 C2 C3 s H4 H5).
Trivial.

Inversion_clear H1.
Inversion_clear H2.
Elim (H C1 C0 s H3 H1).
Elim (H0 C2 C3 s H4 H5).
Trivial.

Inversion_clear H0.
Inversion_clear H1.
Elim (H C0 C1 (cons_squelette p s) H2 H0).
Trivial.

Inversion_clear H1.
Inversion_clear H2.
Elim (H C1 C0 s H3 H1).
Elim (H0 C2 C3 (cons_squelette p s) H4 H5).
Trivial.

Inversion_clear H1.
Inversion_clear H2.
Elim (H C0 C1 (cons_squelette p0 (cons_squelette p s)) H3 H1).
Elim (H0 C2 C3 (cons_squelette p s) H4 H5).
Trivial.

Inversion_clear H2.
Inversion_clear H3.
Elim (H C1 C0 s H4 H2).
Elim (H0 C2 C4 s H5 H7).
Elim (H1 C3 C5 s H6 H8).
Trivial.

Save.


(*****************************************************************************)
(*We can now define an equivalence between ML values and CAM values with the *)
(*the help of the predicate Traduction                                       *)
(*****************************************************************************)


Mutual Inductive
Equiv_val:MLval->CSem_val->Prop:=


Eqbool:(b:bool)(Equiv_val (boolean b) (val (elem b)))

|Eqnum:(n:nat)(Equiv_val (num n) (val (int n)))

|Eq_op:(c:OP)(Equiv_val (OP_clos c) (val (def_op c)))

|Eqpair:(V1,V2:MLval)(Cval1,Cval2:CSem_val)
         (Equiv_val V1 Cval1)->
           (Equiv_val V2 Cval2)->
              (Equiv_val (valpair V1 V2) (Cam_pair Cval1 Cval2))

|Eqclos:(p:Pat)(E:MLexp)(C:Commande)(e:MLenv)(CV:CSem_val)(s:Squelette)
              (Equiv_env e CV)->
                (Habite e s)->  
                 (Traduction (cons_squelette p s) E C)->
                  (Equiv_val (Clos p E e) (Cam_clos C CV))

|Eqclos_rec:(p,x:Pat)(E:MLexp)(e:MLenv)(C:Commande)(CV:CSem_val)
       (s:Squelette)
          (Equiv_env e CV)->
              (Habite e s)->
                (Traduction (cons_squelette x (cons_squelette p s)) E C)->
   (Equiv_val (Clos_rec x E p e) (Cam_clos_rec C CV))

with
Equiv_env:MLenv->CSem_val->Prop:=

 Eqenv1:(Equiv_env Enil Cam_nil)

|Eqenv2:(p:Pat)(E:MLenv)(CV0:CSem_val)
            (Equiv_env E CV0)->
              (V:MLval)(CV:CSem_val)(Equiv_val V CV)->
                (Equiv_env (Econs p V E) (Cam_pair CV0 CV)).



(*****************************************************************************)
(*                 We can now give a formulation of the proof                *)
(*****************************************************************************)

Inductive compilation[E:MLexp]:Prop:=
preuve_compilation:
  ((e:MLenv)(V:MLval)(ML_DS e E V)->     
         (s:Squelette)(C:Commande)(Traduction s E C)->
           (Habite e s)->
               (CV:CSem_val)(Equiv_env e CV)->
                 ((Ex [CV1:CSem_val]
                   ((Equiv_val V CV1)
                  /\ (s:Etat)(CAM_DS (ETcons CV s) C (ETcons CV1 s))))))->
                          (compilation E).
                 



(*****************************************************************************)
(*This formulation permits us to make the proof for the different cases      *)
(*       where we don't need induction assumptions.                          *)
(*****************************************************************************)


(*****************************************************************************)
(*                  Case where E is a boolean                                *)
(*****************************************************************************)

Lemma Proof_bool : (b:bool)(compilation (Bool b)).


Intro b.
Apply preuve_compilation.
Intros e V ML_b s C Trad_b hab CV Eq.
Inversion_clear ML_b.
Exists (val (elem b)).
Split.
Exact (Eqbool b).
Inversion_clear Trad_b.
Intro s0.
Exact (QUO s0 CV (elem b)).
Save.

(*****************************************************************************)
(*                  Case where E is an integer                               *)
(*****************************************************************************)

Lemma  Proof_int :  (n:nat)(compilation (Num n)).

Intro n.
Apply preuve_compilation.
Intros e V ML_n s C Trad_n hab CV Eq.
Inversion_clear ML_n.
Exists (val (int n)).
Split.
Exact (Eqnum n).
Inversion_clear Trad_n.
Intro s0.
Exact (QUO s0 CV (int n)).
Save.

(*****************************************************************************)
(*               Case where E is a predefinite operator                      *)
(*****************************************************************************)


Lemma  Proof_op :  (c:OP)(compilation (op c)).


Intro c.
Apply preuve_compilation.
Intros e V ML_c s C Trad_c hab CV Eq.
Inversion_clear ML_c.
Exists (val (def_op c)).
Split.
Exact (Eq_op c).
Inversion_clear Trad_c.
Intro s0.
Exact (QUO s0 CV (def_op c)).
Save.

(*****************************************************************************)
(*                    Case where E is an identifier                          *)
(* This case is special in the sens that we cannot use the predicate         *)
(*compilation because the variables of the formulation must be introduced    *)
(* in a very precise order.                                                  *)
(* Of course we could avoid a new definition using a the Cut tactic          *)
(* The point is that in this case only we must first do an induction on the  *)
(* ML environment in which we evaluate our expression (an identifier!)       *)
(*****************************************************************************)


Inductive compilation_id[E:MLexp]:Prop:=

preuve_compilation_id: (
               (e:MLenv)(s:Squelette)(Habite e s)->
                (V:MLval)(ML_DS e E V)->
                 (C:Commande)(Traduction s E C)->
                  (CV:CSem_val)(Equiv_env e CV)->
                   ((Ex [CV1:CSem_val]
                   ( (Equiv_val V CV1)
                  /\ (s:Etat)(CAM_DS (ETcons CV s) C (ETcons CV1 s))))))->
                          (compilation_id E).


Lemma Proof_ident :  (x:Pat)(compilation_id (id x)).

Intro.
Apply preuve_compilation_id.
Induction e.
Intros.
Inversion_clear H0.
Inversion_clear H3.
Do 6 Intro.
Inversion_clear H0.
Intro;Intro.
Inversion_clear H0.
Simple Inversion H2.
Injection H0;Do 3 Intro.
Rewrite<-H3;Rewrite->H7.
Do 2 Intro.
Inversion_clear H8.
Inversion_clear H9.
Do 2 Intro.
Inversion_clear H8.
Rewrite<-H4;Rewrite->H6.
Exists CV1.
Split.
Exact H10.
Intro;Apply CDR.

Elim H8;Trivial.
Injection H4;Do 3 Intro.
Rewrite->H5;Rewrite->H7;Rewrite->H6;Rewrite->H0.
Do 4 Intro.
Inversion_clear H10.

Simple Inversion H11.
Injection H12;Rewrite->H10;Do 2 Intro.
Elim H9;Elim H15;Auto.

Rewrite->H13;Injection H14;Do 2 Intro;Rewrite->H12;Rewrite->H10.
Rewrite<-H15.
Do 4 Intro.
Inversion_clear H18.
Cut (ML_DS m0 (id x) V).
Cut (Traduction s0 (id x) C0).
Intros Hyp_Trad Hyp_ML.
Elim (H s0 H1 V Hyp_ML C0 Hyp_Trad CV0 H19).
Do 2 Intro.
Elim H18;Do 2 Intro.
Exists x0.
Split.
Exact H21.
Intro.
Apply (o_DS  (ETcons (Cam_pair CV0 CV1) s2) (ETcons CV0 s2) (ETcons x0 s2)
                 car C0).
Apply CAR.
Exact (H22 s2).
Apply Trad_var.
Exact H17.
Apply IDENT.
Exact H8.

Save.




(*****************************************************************************)
(*                     Case where E is an abstraction                        *)
(*****************************************************************************)


Lemma Proof_abstraction :   (E:MLexp)(p:Pat)(compilation (lambda p E)).

Intros E p.
Apply preuve_compilation.
Intros e V ML_lambda.
Inversion_clear ML_lambda.
Intros s C Trad_lam.
Inversion_clear Trad_lam.
Intros hab CV Eq_e_CV.
Exists (Cam_clos C0 CV).
Split.
Apply (Eqclos p E C0 e CV s Eq_e_CV hab H).
Intro.
Apply CUR.

Save.

(*****************************************************************************)
(*          We have now all the tools we need to make the proof              *)
(* We want to prove that the following diagram commutes:                     *)
(*                                                                           *)
(*                                                                           *)
(*                                                                           *)
(*                           "Traduction"                                    *)
(*         Mini_ML terms ---------------------> CAM terms                    *)
(*             |                                  |                          *)
(*             |                                  |                          *)
(*             |                                  |                          *)
(*             |                                  |                          *)
(*      "ML_DS"|                                  |"CAM_DS"                  *)
(*             |                                  |                          *)
(*             |                                  |                          *)
(*             |                                  |                          *)
(*             V              "Equiv"             V                          *)
(*      Mini_ML values ---------------------> CAM values                     *)
(*                                                                           *)
(*                                                                           *)
(*                                                                           *)
(* This means that having an MLexp E its ML value and its translation onto   *)
(*CAM code we can find a  CAM value such that the diagram commutes           *)
(*                                                                           *)
(*      The proof is made by induction on the Predicate ML_DS.               *)
(*****************************************************************************)
    
Lemma final_proof :  (E:MLexp)(e:MLenv) 
       (V:MLval)(ML_DS e E V)->     
         (s:Squelette)(C:Commande)(Traduction s E C)->
           (Habite e s)->
             (V:MLval)(ML_DS e E V)->
               (CV:CSem_val)(Equiv_env e CV)->
                 ((Ex [CV1:CSem_val]
                   ((Equiv_val V CV1)
                  /\ (s:Etat)(CAM_DS (ETcons CV s) C (ETcons CV1 s))))).

Intros E e V.
Induction 1.

(*****************************************************************************)
(*                Simple cases have already been treated                     *)
(*****************************************************************************)

Intros b e0 s C Trad_b hab V0 ML_b CV Eq.
Cut (compilation (Bool b)).
2:Exact (Proof_bool b).
Intro comp.
Elim comp;Intro hyp.
Exact (hyp e0 V0 ML_b s C Trad_b hab CV Eq).

Intros n e0 s C Trad_int hab V0 ML_int CV Eq.
Cut (compilation (Num n)).
2:Exact (Proof_int n).
Intro comp.
Elim comp;Intro hyp.
Exact (hyp e0 V0 ML_int s C Trad_int hab CV Eq).

Intros c e0 s C Trad_op hab V0 ML_op CV Eq.
Cut (compilation (op c)).
2:Exact (Proof_op c).
Intro comp.
Elim comp;Intro hyp.
Exact (hyp e0 V0 ML_op s C Trad_op hab CV Eq).

Intros e0 P E0 s C Trad_lam hab V0 ML_lambda CV Eq.
Cut  (compilation (lambda P E0)).
2:Exact (Proof_abstraction  E0 P).
Intro comp.
Elim comp;Intro hyp.
Exact (hyp e0 V0 ML_lambda s C Trad_lam hab CV Eq).

Intros e0 v I VAL_I s C Trad_pat hab V0 ML_pat CV Eq.
Cut (compilation_id (id I)).
2:Exact (Proof_ident I).
Intro comp.
Elim comp;Intro hyp.
Exact (hyp e0 s hab V0 ML_pat C Trad_pat CV Eq).


(*****************************************************************************)
(*We have now to treat the cases where "compilation" cannot be used          *)
(*****************************************************************************)


(*****************************************************************************)
(*                       Case of the " if then else"                         *)
(*****************************************************************************)


                     (*   Case where E1 evaluates to true    *)


Intros e0 E1 E2 E3 v ML_E1 hyp_E1 ML_E2 hyp_E2 s C Trad_if.
Inversion_clear Trad_if.
Intros hab V_final ML_DS_if. 
Inversion_clear ML_DS_if.
Intros CV0 eq_e0_CV0.
Elim (hyp_E2 s C2 H1 hab V_final H4 CV0 eq_e0_CV0).
Intros CV_final HH;Elim HH.
Intros Eq_Vf_CVf CAM_C2.
Exists CV_final.
Split.
Exact Eq_Vf_CVf.
Intro s0.
Cut (CAM_DS (ETcons CV0 (ETcons CV0 s0))
               (o C1 (branch C2 C3))
                 (ETcons CV_final s0)).
Cut (CAM_DS (ETcons CV0 s0)  push (ETcons CV0 (ETcons CV0 s0))).
Exact (o_DS  (ETcons CV0 s0)
             (ETcons CV0 (ETcons CV0 s0))
             (ETcons CV_final s0)
                  push 
                  (o C1 (branch C2 C3))).
Exact (PUSH s0 CV0).
Elim (hyp_E1 s C1 H0 hab (boolean true) ML_E1 CV0 eq_e0_CV0).
Intros CV1 HH';Elim HH'.
Intro Eq_1;Inversion_clear Eq_1.
Intro CAM_C1.
Apply (o_DS   (ETcons CV0 (ETcons CV0 s0)) 
              (ETcons (val (elem true)) (ETcons CV0 s0))
              (ETcons CV_final s0)
              C1
              (branch C2 C3)).
Exact (CAM_C1 (ETcons CV0 s0)).
Apply (BRANCHT (ETcons CV0 s0) 
               (ETcons CV_final s0)
               C2 C3).
Exact (CAM_C2 s0).
Cut ((boolean false)=(boolean true)).
Intro HH.
Discriminate HH.
Exact (ML_DS_determ e0 E1 (boolean false) H3 (boolean true) ML_E1).


               (*   Case where E1 evaluates to false    *)


Intros e0 E1 E2 E3 v ML_E1 hyp_E1 ML_E3 hyp_E3 s C Trad_if.
Inversion_clear Trad_if.
Intros hab V_final ML_DS_if. 
Inversion_clear ML_DS_if.
Cut ((boolean false)=(boolean true)).
Intro HH.
Discriminate HH.
Exact (ML_DS_determ e0 E1 (boolean false) ML_E1 (boolean true) H3).
Intros CV0 eq_e0_CV0.
Elim (hyp_E3 s C3 H2 hab V_final H4 CV0 eq_e0_CV0).
Intros CV_final HH;Elim HH.
Intros Eq_Vf_CVf CAM_C3.
Exists CV_final.
Split.
Exact Eq_Vf_CVf.
Intro s0.
Cut (CAM_DS (ETcons CV0 (ETcons CV0 s0))
               (o C1 (branch C2 C3))
                 (ETcons CV_final s0)).
Cut (CAM_DS (ETcons CV0 s0)  push (ETcons CV0 (ETcons CV0 s0))).
Exact (o_DS  (ETcons CV0 s0)
             (ETcons CV0 (ETcons CV0 s0))
             (ETcons CV_final s0)
                  push 
                  (o C1 (branch C2 C3))).
Exact (PUSH s0 CV0).
Elim (hyp_E1 s C1 H0 hab (boolean false) ML_E1 CV0 eq_e0_CV0).
Intros CV1 HH';Elim HH'.
Intro Eq_1;Inversion_clear Eq_1.
Intro CAM_C1.
Apply (o_DS   (ETcons CV0 (ETcons CV0 s0)) 
              (ETcons (val (elem false)) (ETcons CV0 s0))
              (ETcons CV_final s0)
              C1
              (branch C2 C3)).
Exact (CAM_C1 (ETcons CV0 s0)).
Apply (BRANCHF (ETcons CV0 s0) 
               (ETcons CV_final s0)
               C2 C3).
Exact (CAM_C3 s0).


(*****************************************************************************)
(*                            the pair case                                  *)
(*****************************************************************************)


Intros e0 E1 E2 u v ML_E1 hyp_E1 ML_E2 hyp_E2 s C Trad_p.
Inversion_clear Trad_p.
Intros hab V_final ML_DS_pair.
Inversion_clear ML_DS_pair.
Intros CV0 eq_e0_CV0.
Elim (hyp_E1 s C1 H0 hab u0 H2 CV0 eq_e0_CV0).
Intros CV_E1 HH;Elim HH.
Intros eq_u0_CV_E1 CAM_C1.
Elim (hyp_E2 s C2 H1 hab v0 H3 CV0 eq_e0_CV0).
Intros CV_E2 HH';Elim HH'.
Intros eq_v0_CV_E2 CAM_C2.
Exists (Cam_pair CV_E1 CV_E2).
Split.
Apply (Eqpair u0 v0 CV_E1 CV_E2 eq_u0_CV_E1 eq_v0_CV_E2).
Intro s0.
Apply (o_DS (ETcons CV0 s0)
             (ETcons CV0 (ETcons CV0 s0))
             (ETcons (Cam_pair CV_E1 CV_E2) s0)
             push
             (o C1 (o swap (o C2 cons)))).
Apply (PUSH s0 CV0).
Apply (o_DS  (ETcons CV0 (ETcons CV0 s0))
             (ETcons CV_E1 (ETcons CV0 s0))
             (ETcons (Cam_pair CV_E1 CV_E2) s0)
              C1
              (o swap (o C2 cons))).
Apply CAM_C1.
Apply (o_DS  (ETcons CV_E1 (ETcons CV0 s0))
             (ETcons CV0 (ETcons CV_E1 s0)) 
             (ETcons (Cam_pair CV_E1 CV_E2) s0)
              swap
              (o C2 cons)).
Apply SWAP.
Apply (o_DS  (ETcons CV0 (ETcons CV_E1 s0)) 
             (ETcons CV_E2 (ETcons CV_E1 s0)) 
             (ETcons (Cam_pair CV_E1 CV_E2) s0)
              C2
              cons).
Apply CAM_C2.
Apply CONS.

(*****************************************************************************)
(*We have now to solve the case of the application (It is important to notice*)
(*that we were not able to make the proof on induction on E because of the   *)
(*application case                                                           *)
(*****************************************************************************)


Intros e0 e1 P E0 E1 E2 u v ML_E1 hyp_E1 ML_E2 hyp_E2 ML_E0 hyp_E0
s C Trad_appl.
Inversion_clear Trad_appl.
Intros hab V0 ML_appl.
Inversion_clear ML_appl.
Cut (Clos P E0 e1)=(Clos P0 E3 e3).
Intro HH1;Injection HH1.
Intros eq1 eq2 eq3.
Cut u=u0.
Intro eq4.
Cut (ML_DS (Econs P0 u0 e3) E3 V0).
Rewrite<-eq1;Rewrite<-eq2;Rewrite<-eq3;Rewrite<-eq4.
Intro HH2;Cut v=V0.
Intro eq5;Rewrite<-eq5.
Intros CV0 eq_e0_CV0.
Elim (hyp_E2 s C2 H1 hab u ML_E2 CV0 eq_e0_CV0).
Intros CVal_E2 HH3;Elim HH3.
Intros Eq_u_Cval_E2 CAM_C2.
Elim (hyp_E1 s C1 H0 hab (Clos P E0 e1) ML_E1 CV0 eq_e0_CV0).
Intros Closure HH4;Elim HH4.
Intro eq_clos.
Inversion_clear eq_clos.
Intro CAM_C1.
Cut (Habite (Econs P u e1) (cons_squelette P s0)).
Intro hab_P_e1.
Cut (Equiv_env (Econs P u e1) (Cam_pair CV CVal_E2)).
Intro eq_env_closure.
Elim (hyp_E0 (cons_squelette P s0) C0 H7 hab_P_e1 v ML_E0 
   (Cam_pair CV CVal_E2)  eq_env_closure).
Intros CV_final HH5;Elim HH5.
Intros eq_v_CV_final CAM_C0.
Exists CV_final.
Split.
Exact eq_v_CV_final. 
Intro sq.
Apply (o_DS (ETcons CV0 sq)
            (ETcons CV0 (ETcons CV0 sq))
            (ETcons CV_final sq)
            push
            (o C1 (o swap (o C2 (o cons app))))).
Apply PUSH.
Apply (o_DS (ETcons CV0 (ETcons CV0 sq))
            (ETcons (Cam_clos C0 CV) (ETcons CV0 sq))
            (ETcons CV_final sq)
            C1 
            (o swap (o C2 (o cons app)))).
Apply CAM_C1.
Apply (o_DS (ETcons (Cam_clos C0 CV) (ETcons CV0 sq))
            (ETcons CV0 (ETcons (Cam_clos C0 CV) sq))
            (ETcons CV_final sq)
            swap
            (o C2 (o cons app))).
Apply SWAP.
Apply (o_DS (ETcons CV0 (ETcons (Cam_clos C0 CV) sq))
            (ETcons CVal_E2 (ETcons (Cam_clos C0 CV) sq))
            (ETcons CV_final sq)
            C2
            (o cons app)).
Apply CAM_C2.
Apply  (o_DS (ETcons CVal_E2 (ETcons (Cam_clos C0 CV) sq))
             (ETcons (Cam_pair (Cam_clos C0 CV) CVal_E2) sq)
             (ETcons CV_final sq)
            cons
            app).
Apply CONS.
Apply (APPcam1 sq  
               (ETcons CV_final sq)
                CV
                CVal_E2
                C0).
Apply CAM_C0.
Apply (Eqenv2 P e1 CV H5 u CVal_E2 Eq_u_Cval_E2).
Apply (cons_habite P u e1 s0 H6).
Apply (ML_DS_determ (Econs P u e1) E0 v ML_E0 V0 HH2).
Exact H4.
Apply (ML_DS_determ  e0 E2 u ML_E2 u0 H3).
Apply (ML_DS_determ  e0 E1 (Clos P E0 e1) ML_E1 (Clos P0 E3 e3) H2).
Cut (Clos P E0 e1)=(Clos_rec x E3 P0 e3).
Intro hyp_not;Discriminate hyp_not.
Apply (ML_DS_determ  e0 E1 (Clos P E0 e1) ML_E1 (Clos_rec x E3 P0 e3) H2).
Cut (Clos P E0 e1)=(OP_clos c).
Intro hyp_not;Discriminate hyp_not.
Apply (ML_DS_determ  e0 E1 (Clos P E0 e1) ML_E1 (OP_clos c) H2).


(*****************************************************************************)
(*                         Case of a recursive closure                       *)
(*****************************************************************************)


Intros e0 e1 x P E0 E1 E2 u v ML_E1 hyp_E1 ML_E2 hyp_E2 ML_E0 hyp_E0
s C Trad_appl.
Inversion_clear Trad_appl.
Intros hab V0 ML_app.
Inversion_clear ML_app.
Cut (Clos_rec x E0 P e1)=(Clos P0 E3 e3). 
Intro HH1;Discriminate HH1.
Apply (ML_DS_determ e0 E1 (Clos_rec x E0 P e1) ML_E1 (Clos P0 E3 e3) H2).
Intros CV0 eq_e0_CV0.
Cut (Clos_rec x E0 P e1)=(Clos_rec x0 E3 P0 e3).
Intro HH1;Injection HH1.
Intros eq1 eq2 eq3 eq4.
Cut (ML_DS (Econs x0 u0 (Econs P0 (Clos_rec x0 E3 P0 e3) e3)) E3 V0).
Rewrite<-eq1;Rewrite<-eq2;Rewrite<-eq3;Rewrite<-eq4.
Cut u=u0.
Intro eq5;Rewrite<-eq5.
Intro ML_E0'.
Cut v=V0.
Intro eq6;Rewrite<-eq6.
Elim (hyp_E2 s C2 H1 hab u ML_E2 CV0 eq_e0_CV0).
Intros CV_E2 HH3;Elim HH3.
Intros eq_u_CV_E2 CAM_C2.
Elim (hyp_E1 s C1 H0 hab (Clos_rec x E0 P e1) ML_E1 CV0 eq_e0_CV0).
Intros CV_E1 HH2;Elim  HH2.
Intros eq_V1_CV_E1 CAM_C1.
Simple Inversion eq_V1_CV_E1.
Discriminate H5.
Discriminate H5.
Discriminate H5.
Discriminate H7.
Discriminate H8.
Injection H8.
Intros eq7 eq8 eq9 eq10.
Rewrite->eq7;Rewrite->eq8;Rewrite->eq9;Rewrite->eq10.
Intros.
Cut (Habite (Econs x u (Econs P (Clos_rec x E0 P e1) e1))
         (cons_squelette x (cons_squelette P s0))).
Intro hab0.
Cut (Equiv_env (Econs x u (Econs P (Clos_rec x E0 P e1) e1))
             (Cam_pair (Cam_pair CV CV_E1) CV_E2)).
Intro Eq_env0.
Elim (hyp_E0 (cons_squelette x (cons_squelette P s0))
      C0 H7 hab0 v ML_E0 (Cam_pair (Cam_pair CV CV_E1) CV_E2) Eq_env0).

Intros CV_E0 HH4;Elim HH4.
Intros eq_v_CV_E0 CAM_C0.
Exists CV_E0.
Split.
Exact eq_v_CV_E0.
Intro S.
Apply (o_DS (ETcons CV0 S)
            (ETcons CV0 (ETcons CV0 S))
            (ETcons CV_E0 S)
            push
            (o C1 (o swap (o C2 (o cons app))))).
Apply PUSH.
Apply (o_DS (ETcons CV0 (ETcons CV0 S))
            (ETcons CV_E1 (ETcons CV0 S))
            (ETcons CV_E0 S)
            C1
            (o swap (o C2 (o cons app)))).
Apply CAM_C1.
Apply (o_DS  (ETcons CV_E1 (ETcons CV0 S))
             (ETcons CV0 (ETcons CV_E1 S))
             (ETcons CV_E0 S) 
             swap
             (o C2 (o cons app))).
Apply SWAP.
Apply (o_DS (ETcons CV0 (ETcons CV_E1 S))
            (ETcons CV_E2 (ETcons CV_E1 S))
            (ETcons CV_E0 S)
            C2
            (o cons app)).
Apply CAM_C2.
Apply (o_DS (ETcons CV_E2 (ETcons CV_E1 S))
            (ETcons (Cam_pair CV_E1 CV_E2 ) S)
            (ETcons CV_E0 S)
            cons
            app).
Apply CONS.
Cut  (s:Etat)
(CAM_DS (ETcons (Cam_pair (Cam_pair CV CV_E1) CV_E2) s) C0 (ETcons CV_E0 s)).
2:Auto.
Rewrite<-H9.
Intro Good_CAM_C0.
Apply (APPcam2 S (ETcons CV_E0 S) CV_E2 CV C0 (Good_CAM_C0 S)).
Apply Eqenv2;Auto.
Apply Eqenv2;Auto.
Apply cons_habite;Auto.
Apply cons_habite;Auto.
Apply (ML_DS_determ (Econs x u (Econs P (Clos_rec x E0 P e1) e1))
                    E0 v ML_E0 V0 ML_E0').
Apply (ML_DS_determ e0 E2 u ML_E2 u0 H3).
Auto.
Apply (ML_DS_determ e0 E1 (Clos_rec x E0 P e1) ML_E1 
                                 (Clos_rec x0 E3 P0 e3) H2).
Cut (OP_clos c)=(Clos_rec x E0 P e1).
Intro Hyp_false;Discriminate Hyp_false.
Apply (ML_DS_determ e0 E1 (OP_clos c) H2
                           (Clos_rec x E0 P e1) ML_E1). 

                             

(*****************************************************************************)
(*            Case where we apply a predefined operator                      *)
(*****************************************************************************)



Intros e0 E1 E2 n m c ML_E1 hyp_E1 ML_E2 hyp_E2 s C Trad_appl.
Inversion_clear Trad_appl.
Intros hab V0 ML_appl.
Inversion_clear ML_appl.
Intros CV0 eq_e0_CV0.
Cut (OP_clos c) = (Clos P E0 e2).
2: Exact (ML_DS_determ e0 E1 (OP_clos c) ML_E1 (Clos P E0 e2) H2). 
Intro HH;Discriminate HH.
Cut (OP_clos c) = (Clos_rec x E0 P e2).
2: Exact (ML_DS_determ e0 E1 (OP_clos c) ML_E1 (Clos_rec x E0 P e2) H2). 
Intro HH;Discriminate HH.
Cut (valpair (num n) (num m)) = (valpair (num n0) (num m0)).
2: Exact (ML_DS_determ e0 E2 (valpair (num n) (num m)) ML_E2
                   (valpair (num n0) (num m0)) H3).
Intro HH;Injection HH.
Intros eq1 eq2.
Cut (OP_clos c)=(OP_clos c0).
Intro HH1;Injection HH1.
Intro eq3.
Rewrite<-eq1;Rewrite<-eq2;Rewrite<-eq3.
Intros CV0 eq_e0_CV0.
Elim (hyp_E1 s C1 H0 hab (OP_clos c) ML_E1 CV0 eq_e0_CV0).
Intro HH2;Elim HH2.
Intros CV_c HH3.
Elim HH3.
Intros eq_c_CV_c CAM_C1.
Elim (hyp_E2 s C2 H1 hab (valpair (num n) (num m)) ML_E2 CV0 eq_e0_CV0).
Intros CV_nm HH4.
Elim HH4.
Intro eq_nm_CV_nm.
Inversion_clear eq_nm_CV_nm.
Inversion_clear H4.
Inversion_clear H5.
Intro CAM_C2.
Exists (val (int (eval_op c n m))).
Split.
Exact (Eqnum (eval_op c n m)).
Intro S.
Apply (o_DS (ETcons CV0 S)
            (ETcons CV0 (ETcons CV0 S))
            (ETcons (val (int (eval_op c n m))) S)
            push
            (o C1 (o swap (o C2 (o cons app))))).
Apply PUSH.
Apply (o_DS (ETcons CV0 (ETcons CV0 S))
            (ETcons (val CV_c) (ETcons CV0 S))
            (ETcons (val (int (eval_op c n m))) S)
            C1
            (o swap (o C2 (o cons app)))).
Apply CAM_C1.
Apply (o_DS  (ETcons (val CV_c) (ETcons CV0 S))
             (ETcons CV0 (ETcons (val CV_c) S))
             (ETcons (val (int (eval_op c n m))) S)
             swap
             (o C2 (o cons app))).
Apply SWAP.
Apply (o_DS  (ETcons CV0 (ETcons (val CV_c) S))
             (ETcons (Cam_pair (val (int n)) (val (int m)))
                                         (ETcons (val CV_c) S))
             (ETcons (val (int (eval_op c n m))) S)
             C2
             (o cons app)).
Apply CAM_C2.
Cut (s:Etat)(CAM_DS (ETcons CV0 s) C1 (ETcons (val CV_c) s)).
2:Auto.
Inversion_clear eq_c_CV_c.
Intro U_CAM_C1.
Apply (o_DS (ETcons (Cam_pair (val (int n)) (val (int m)))
                                              (ETcons (val (def_op c)) S))
             (ETcons (Cam_pair (val (def_op c))
                            (Cam_pair (val (int n)) (val (int m)))) S)
             (ETcons (val (int (eval_op c n m))) S)
             cons
             app).
Apply CONS.
Apply (APPcam_op S n m c).  
Intros.
Elim H6.
Intro HH5;Inversion_clear HH5.
Intros.
Elim H5.
Intro HH5;Inversion_clear HH5.
Intros.
Elim H5.
Intro HH5;Inversion_clear HH5.
Intros.
Elim H4.
Intro HH4; Inversion_clear HH4.
Exact (ML_DS_determ e0 E1 (OP_clos c) ML_E1 (OP_clos c0) H2).


(*****************************************************************************)
(*                        Case where E is a let'                             *)
(*****************************************************************************)


Intros e0 P E1 E2 u v ML_E2 hyp_E2 ML_E1 hyp_E1 s C Trad_let'.
Inversion_clear Trad_let'.
Intros hab V0 ML_let.
Inversion_clear ML_let.
Intros CV0 eq_e0_CV0.
Cut u=u0;Cut v=V0.
Intros eq1 eq2.
Rewrite<-eq1;Rewrite<-eq2.
Elim (hyp_E2 s C1 H0 hab u ML_E2 CV0 eq_e0_CV0).
Intros CV_u HH.
Elim HH.
Intros eq_u_CV_u CAM_C1.
Elim (hyp_E1 (cons_squelette P s) C2 H1 
         (cons_habite P u e0 s hab) v ML_E1 
             (Cam_pair CV0 CV_u) 
            (Eqenv2 P e0 CV0 eq_e0_CV0 u CV_u eq_u_CV_u)).
Intros CV_v HH2.        
Elim HH2.
Intros eq_v_CV_v CAM_C2.
Exists CV_v.
Split.
Exact eq_v_CV_v.
Intro S.
Apply (o_DS (ETcons CV0 S)
            (ETcons CV0 (ETcons CV0 S))
            (ETcons CV_v S)
            push
            (o C1 (o cons C2))).
Apply PUSH.      
Apply (o_DS (ETcons CV0 (ETcons CV0 S))
            (ETcons CV_u (ETcons CV0 S))
            (ETcons CV_v S)
            C1
            (o cons C2)).
Apply CAM_C1.        
Apply (o_DS (ETcons CV_u (ETcons CV0 S))
            (ETcons (Cam_pair CV0 CV_u) S)
            (ETcons CV_v S)
            cons
            C2).
Apply CONS.
Apply CAM_C2.
Cut u=u0.
Intro eq3.
Cut (ML_DS (Econs P u0 e0) E1 V0).
2: Exact H3.
Rewrite<-eq3.
Intro ML_E1'.
Apply (ML_DS_determ (Econs P u e0) E1 v ML_E1 V0 ML_E1').
Apply (ML_DS_determ e0 E2 u ML_E2 u0 H2).
Intro;Apply (ML_DS_determ e0 E2 u ML_E2 u0 H2).
Cut u=u0.
Intro eq3.
Cut (ML_DS (Econs P u0 e0) E1 V0).
2: Exact H3.
Rewrite<-eq3.
Intro ML_E1'.
Apply (ML_DS_determ (Econs P u e0) E1 v ML_E1 V0 ML_E1').
Apply (ML_DS_determ e0 E2 u ML_E2 u0 H2).


(*****************************************************************************)
(*                     Case where E is a let_rec                             *)
(*****************************************************************************)

Intros e0 P x E0 E2 u ML_E2 hyp_E2 s C Trad_letrec.
Inversion_clear Trad_letrec.
Intros hab V0 ML_letrec.
Inversion_clear ML_letrec.
Intros CV0 eq_e0_CV0.
Cut u=V0.
Intro eq1.
Rewrite<-eq1.
Elim (hyp_E2 (cons_squelette P s) C2 H1 
       (cons_habite P (Clos_rec x E0 P e0) e0 s hab)
             u ML_E2 
             (Cam_pair CV0 (Cam_clos_rec C0 CV0))
  (Eqenv2 P e0 CV0 eq_e0_CV0 (Clos_rec x E0 P e0)
          (Cam_clos_rec C0 CV0)
            (Eqclos_rec P x E0 e0 C0 CV0 s eq_e0_CV0 hab H0))).   
Intros CV_u HH.
Elim HH.Intros eq_u_CV_u CAM_C2.
Exists CV_u.
Split.
Exact eq_u_CV_u.
Intro S.
Apply (o_DS (ETcons CV0 S) 
            (ETcons CV0 (ETcons CV0 S))
            (ETcons CV_u S)
            push
            (o (cur_rec C0) (o cons C2))).
Apply PUSH.
Apply (o_DS  (ETcons CV0 (ETcons CV0 S))
             (ETcons (Cam_clos_rec C0 CV0) (ETcons CV0 S))
             (ETcons CV_u S)
             (cur_rec C0)
             (o cons C2)).
Apply CUR_REC.
Apply (o_DS  (ETcons (Cam_clos_rec C0 CV0) (ETcons CV0 S))
             (ETcons (Cam_pair CV0 (Cam_clos_rec C0 CV0)) S)
             (ETcons CV_u S)
             cons C2).
Apply CONS.
Apply CAM_C2.
Apply (ML_DS_determ (Econs P (Clos_rec x E0 P e0) e0) E2 u ML_E2 V0 H2). 


Save.



(*****************************************************************************)

