(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)

(****************************************************************************)
(*                                  Cube.v                                  *)
(*                                                                          *)
(*                                Gerard Huet                               *)
(*                                                                          *)
(* Developed in V5.8  January 1993                                          *)
(* Ported    to V5.10 January 1995                                          *)
(****************************************************************************)

Require Arith.
Require Terms.
Require Reduction.
Require Redexes.
Require Test.
Require Substitution.
Require Residuals.

(*****************)
(* Prism Theorem *)
(*****************)

(* Auxiliary notion : compat 
   Used to generate the right simultaneous induction on U, V and W *)

(* (compat U V W) iff U,V,W are compatible markings, and (sub V U) *)
Inductive compat : redexes -> redexes -> redexes -> Prop :=
    Compat_Var : (n:nat)(compat (Var n) (Var n) (Var n))
  | Compat_Fun : (U,V,W:redexes)(compat U V W) 
                             -> (compat (Fun U) (Fun V) (Fun W))
  | Compat_Ap1 : (U1,V1,W1:redexes)(compat U1 V1 W1) -> 
        (U2,V2,W2:redexes)(compat U2 V2 W2) -> (b:bool) 
      (compat (Ap false U1 U2) (Ap false V1 V2) (Ap b W1 W2))
  | Compat_Ap2 : (U1,V1,W1:redexes)(compat U1 V1 W1) -> 
        (U2,V2,W2:redexes)(compat U2 V2 W2) -> (b,b':bool)
      (compat (Ap true (Fun U1) U2) (Ap b (Fun V1) V2) (Ap b' (Fun W1) W2)).

Lemma compat_intro : (U,W,WU:redexes)
   (residuals W U WU)->(V,WV:redexes)(R2:(residuals W V WV))(S:(sub V U))
   (compat U V W).
Proof.
(* Remark use of name R2 for uniform command below *)
Induction 1; Intros; Generalize S; Inversion_clear R2; Intros; Inversion_clear S.
Apply Compat_Var.
Apply Compat_Fun.
Inversion_clear S0; Apply H1 with W1; Auto.
Inversion_clear S0; Apply Compat_Ap1; Auto.
Apply H1 with W0; Auto.
Apply H3 with W3; Auto.
Inversion_clear S0.
Inversion_clear S0; Generalize H4; Inversion_clear H8.
Intro H11; Inversion_clear H11; Apply Compat_Ap2; Auto.
Apply H1 with W4; Auto.
Apply H3 with W3; Auto.
Generalize H4; Inversion_clear S0.
Inversion_clear H8.
Intro H11; Inversion_clear H11; Apply Compat_Ap2; Auto.
Apply H1 with W4; Auto.
Apply H3 with W3; Auto.
Inversion_clear S0; Apply Compat_Ap2; Inversion_clear H8; Auto.
Apply H1 with W0; Auto.
Apply H3 with W3; Auto.
Inversion_clear S0; Apply Compat_Ap2; Inversion_clear H8; Auto.
Apply H1 with W0; Auto.
Apply H3 with W3; Auto.
Qed.

Lemma prism0 : (U,V,W:redexes)(compat U V W) -> 
     (UV:redexes)(R1:(residuals U V UV))(WU,WV:redexes)
       (R2:(residuals W U WU))(R3:(residuals W V WV)) (residuals WV UV WU).
Proof.
Induction 1.
Intros; Inversion_clear R1; Inversion_clear R2; Inversion_clear R3; Auto.
Intros; Inversion_clear R1; Inversion_clear R2; Inversion_clear R3; Auto.
Intros; Inversion_clear R1; Inversion_clear R2; Inversion_clear R3; Auto.
Induction b; Intros; Inversion_clear R1; Inversion_clear R2; Inversion_clear R3; Auto.
Apply commutation; Auto.
Inversion_clear H4; Inversion_clear H8; Auto.
Qed.

(* Remark - It should be possible to get rid of auxiliary compat
   and to combine prism0 with compat_intro in one lemma *)


(*****************************************************************)
(* Theorem prism : (U,V,W:redexes)(sub V U) ->                   *)
(*     (UV:redexes)(residuals U V UV) ->                         *)
(*     (WV:redexes)(residuals W V WV) ->                         *)
(*     (WU:redexes)(residuals W U WU) <-> (residuals WV UV WU).  *)
(*****************************************************************)

Lemma prism1 : (U,V,W:redexes)(sub V U) -> (UV:redexes)(residuals U V UV) ->
     (WV:redexes)(residuals W V WV) -> 
     (WU:redexes)(residuals W U WU) -> (residuals WV UV WU).
Proof.
Intros; Apply prism0 with U V W; Auto.
Apply compat_intro with WU WV; Trivial.
Qed.

(* Converse of prism1 but needs regularity of U *)
Lemma prism2 : 
      (U,V,W:redexes)(sub V U) -> (regular U) -> 
      (UV:redexes)(residuals U V UV) ->
      (WV:redexes)(residuals W V WV) -> 
      (WU:redexes)(residuals WV UV WU) -> (residuals W U WU).
Proof.
Intros.
Elim (residuals_intro W U); Trivial.
Intros WU' R; Elim (residuals_function WV UV WU) with WU'; Trivial.
Apply prism1 with U V W; Trivial.
Apply comp_trans with V.
Apply residuals_comp with WV; Trivial.
Apply comp_sym; Apply residuals_comp with UV; Trivial.
Qed.

Theorem prism : (U,V,W:redexes)(sub V U) -> 
                (UV:redexes)(residuals U V UV) ->
                (WV:redexes)(residuals W V WV) -> 
   ((WU:redexes)(residuals W U WU) <-> (regular U) /\ (residuals WV UV WU)).
Proof.
Intros; Unfold iff; Split.
Intro; Split.
Apply residuals_regular with W WU; Trivial.
Apply prism1 with U V W; Trivial.
Induction 1; Intros; Apply prism2 with V UV WV; Trivial.
Qed.

(**************************************************************************)
(*  Levy's cube lemma :                                                   *)
(*  (U,V,UV,VU:redexes)  (residuals U V UV) -> (residuals V U VU) ->      *)
(*  (W,WU,WV,WUV:redexes)(residuals W U WU) -> (residuals WU VU WUV) ->   *)
(*                       (residuals W V WV) -> (residuals WV UV WUV).     *)
(**************************************************************************)


Lemma cube :
  (U,V,UV,VU:redexes)  (residuals U V UV) -> (residuals V U VU) ->
  (W,WU,WV,WUV:redexes)(residuals W U WU) -> (residuals WU VU WUV) ->
                       (residuals W V WV) -> (residuals WV UV WUV).
Proof.
Intros.
Cut (comp U V).
2: Apply residuals_comp with UV; Trivial.
Intro C; Elim (union_defined U V C); Intros T UVT.
Apply prism1 with T V W; Trivial.
Apply union_r with U; Trivial.
Apply preservation with U; Trivial.
Apply prism2 with U VU WU; Trivial.
Apply union_l with V; Trivial.
Apply union_preserve_regular with U V; Trivial.
Apply residuals_regular with V VU; Trivial.
Apply residuals_regular with U UV; Trivial.
Apply preservation with V; Trivial.
Apply union_sym; Trivial.
Qed.


(* 3-dimensional paving diagram *)
Lemma paving : 
  (U,V,W,WU,WV:redexes)(residuals W U WU) -> (residuals W V WV) ->
  (Ex [UV:redexes](Ex [VU:redexes](Ex [WUV:redexes]
           ((residuals WU VU WUV) /\ (residuals WV UV WUV))))).
Proof.
Intros; Elim (residuals_intro U V).
Intros UV R1; Exists UV.
Elim (residuals_intro V U).
Intros VU R2; Exists VU.
Elim (residuals_intro WU VU).
Intros WUV R3; Exists WUV.
Split; Trivial.
Apply cube with U V VU W WU; Trivial.
Apply mutual_residuals_comp with U W V; Trivial.
Apply residuals_preserve_regular with V U; Trivial.
Apply residuals_regular with U UV; Trivial.
Apply comp_sym; Apply residuals_comp with UV; Trivial.
Apply residuals_regular with W WU; Trivial.
Apply comp_trans with W.
Apply comp_sym; Apply residuals_comp with WU; Trivial.
Apply residuals_comp with WV; Trivial.
Apply residuals_regular with W WV; Trivial.
Qed.

