(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                Redexes.v                                 *)
(*                                                                          *)
(*                                Gerard Huet                               *)
(*                                                                          *)
(* Developed in V5.8  January 1993                                          *)
(* Ported    to V5.10 January 1995                                          *)
(****************************************************************************)

Require Arith.
Require Test.
Require Terms.
Require Reduction.

(************************************)
(* Lambda terms with marked redexes *)
(************************************)

Inductive redexes : Set :=
    Var : nat -> redexes
  | Fun : redexes -> redexes
  | Ap  : bool -> redexes -> redexes -> redexes.

(* A redex is marked as (Ap true (Fun M) N) *)

(* The Boolean algebra of sets of redexes *)

Inductive sub : redexes -> redexes -> Prop :=
    Sub_Var : (n:nat)(sub (Var n) (Var n))
  | Sub_Fun : (U,V:redexes)(sub U V) -> (sub (Fun U) (Fun V))
  | Sub_Ap1 : (U1,V1:redexes)(sub U1 V1) -> (U2,V2:redexes)(sub U2 V2) -> 
                    (b:bool)(sub (Ap false U1 U2) (Ap b V1 V2))
  | Sub_Ap2 : (U1,V1:redexes)(sub U1 V1) -> (U2,V2:redexes)(sub U2 V2) -> 
                    (b:bool)(sub (Ap true U1 U2) (Ap true V1 V2)).

Definition bool_max := [b,b':bool]<bool>Case b of true b' end.

Lemma max_false : (b:bool)(bool_max b false)=b.
Proof.
Induction b; Simpl; Trivial.
Qed.

Inductive union : redexes -> redexes -> redexes -> Prop :=
    Union_Var : (n:nat)(union (Var n) (Var n) (Var n))
  | Union_Fun : (U,V,W:redexes)(union U V W) -> (union (Fun U) (Fun V) (Fun W))
  | Union_Ap : (U1,V1,W1:redexes)(union U1 V1 W1) -> 
               (U2,V2,W2:redexes)(union U2 V2 W2) -> 
       (b1,b2:bool)(union (Ap b1 U1 U2) (Ap b2 V1 V2) (Ap (bool_max b1 b2) W1 W2)).

Lemma union_l : (U,V,W:redexes)(union U V W) -> (sub U W).
Proof.
Induction 1; Intros.
Apply Sub_Var.
Apply Sub_Fun; Trivial.
Elim b1.
Elim b2; Simpl; Apply Sub_Ap2; Trivial.
Elim b2; Simpl; Apply Sub_Ap1; Trivial.
Qed.

Lemma union_r : (U,V,W:redexes)(union U V W) -> (sub V W).
Proof.
Induction 1; Intros.
Apply Sub_Var.
Apply Sub_Fun; Trivial.
Elim b2.
Elim b1; Simpl; Apply Sub_Ap2; Trivial.
Elim b1; Simpl; Apply Sub_Ap1; Trivial.
Qed.

Lemma bool_max_Sym : (b,b':bool)(bool_max b b')=(bool_max b' b).
Proof.
Induction b; Induction b'; Simpl; Trivial.
Qed.

Lemma union_sym : (U,V,W:redexes)(union U V W) -> (union V U W).
Proof.
Induction 1; Intros.
Apply Union_Var; Trivial.
Apply Union_Fun; Trivial.
Rewrite (bool_max_Sym b1 b2); Apply Union_Ap; Trivial.
Qed.

(* Compatibility *)
(* (comp U V) iff (unmark U)=(unmark V) *)

Inductive comp : redexes -> redexes -> Prop :=
    Comp_Var : (n:nat)(comp (Var n) (Var n))
  | Comp_Fun : (U,V:redexes)(comp U V) -> (comp (Fun U) (Fun V))
  | Comp_Ap : (U1,V1:redexes)(comp U1 V1) -> 
               (U2,V2:redexes)(comp U2 V2) -> 
       (b1,b2:bool)(comp (Ap b1 U1 U2) (Ap b2 V1 V2)).

Hint Comp_Var Comp_Fun Comp_Ap.

Lemma comp_refl : (U:redexes)(comp U U).
Proof.
Induction U; Auto.
Qed.

Lemma comp_sym : (U,V:redexes)(comp U V) -> (comp V U).
Proof.
Induction 1; Auto.
Qed.

Lemma comp_trans : 
      (U,V:redexes)(comp U V) -> (W:redexes)(CVW:(comp V W))(comp U W).
Induction 1; Intros; Inversion_clear CVW; Auto.
Qed.

Lemma union_defined : 
      (U,V:redexes)(comp U V) -> (Ex [W:redexes](union U V W)).
Proof.
Induction 1.
Intro n; Exists (Var n); Apply Union_Var.
Induction 2; Intros W0 H2; Exists (Fun W0); Apply Union_Fun; Trivial.
Intros U1 V1 H1 E1 U2 V2 H2 E2; Elim E1; Elim E2.
Intros W2 A W1 B b1 b2; Exists (Ap (bool_max b1 b2) W1 W2). 
Apply Union_Ap; Trivial.
Qed.


(* A element of type redexes is said to be regular if its true marks label
   redexes *)

Fixpoint regular [U:redexes] : Prop := <Prop>Case U of
  (* Var *) [n:nat]True
  (* Fun *) [V:redexes](regular V)
  (* Ap  *) [b:bool][V:redexes][W:redexes]<Prop>Case b of
        (* true *) <Prop>Case V of
  (* Var *) [n:nat]False
  (* Fun *) [U1:redexes]((regular V) /\ (regular W))
  (* Ap  *) [b':bool][V':redexes][W':redexes]False end
        (* false *) (regular V) /\ (regular W) end end.

(* System Error (Anomaly) cherche_instance 
Recursive Definition regular : redexes -> Prop :=
   (Var n)                   => True
 | (Fun V)                   => (regular V)
 | (Ap true  (Var n)      W) => False
 | (Ap true  (Fun V1)     W) => (regular V1) /\ (regular W)
 | (Ap true  (Ap b V1 V2) W) => False
 | (Ap false V W)            => (regular V) /\ (regular W). *)

Lemma union_preserve_regular : 
    (U,V,W:redexes)(union U V W) -> (regular U) -> (regular V) -> (regular W).
Proof.
Induction 1; Simpl; Trivial.
Induction b1; Induction b2; Simpl.
Generalize H1.
Elim H0; Try Contradiction.
Intros; Elim H7; Elim H8; Auto.
Generalize H1.
Elim H0; Try Contradiction.
Intros; Elim H7; Elim H8; Auto.
Induction 1.
Generalize H1.
Elim H0; Try Contradiction.
Intros; Elim H10; Auto.
Induction 1; Intros O1 O2; Induction 1; Auto.
Qed.

