(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)

(****************************************************************************)
(*                               Residuals.v                                *)
(*                                                                          *)
(*                                Gerard Huet                               *)
(*                                                                          *)
(* Developed in V5.8  January 1993                                          *)
(* Ported    to V5.10 January 1995                                          *)
(****************************************************************************)

Require Arith.
Require Terms.
Require Reduction.
Require Redexes.
Require Test.
Require Substitution.

(*************************************************)
(* Parallel beta reduction with residual tracing *)
(*************************************************)

(* (residuals U V W) means W are residuals of redexes U by step V *)

Inductive residuals : redexes -> redexes -> redexes -> Prop :=
   Res_Var : (n:nat)(residuals (Var n) (Var n) (Var n))
 | Res_Fun : (U,V,W:redexes)(residuals U V W) -> 
                                (residuals (Fun U) (Fun V) (Fun W))
 | Res_Ap : (U1,V1,W1:redexes)(residuals U1 V1 W1) -> 
            (U2,V2,W2:redexes)(residuals U2 V2 W2) -> 
            (b:bool)(residuals (Ap b U1 U2) (Ap false V1 V2) (Ap b W1 W2))
 | Res_redex : (U1,V1,W1:redexes)(residuals U1 V1 W1) -> 
               (U2,V2,W2:redexes)(residuals U2 V2 W2) -> 
  (b:bool)(residuals (Ap b (Fun U1) U2) (Ap true (Fun V1) V2) (subst_r W2 W1)).


Hint Res_Var Res_Fun Res_Ap Res_redex.

Lemma residuals_function : 
    (U,V,W:redexes)(residuals U V W) -> (W':redexes)
    (R:(residuals U V W'))W'=W.
Proof.
(* Remark use of name R necessary for uniform expression of next line *)
Induction 1; Intros; Inversion R; Auto.
Elim H1 with W1; Trivial.
Elim H1 with W0; Elim H3 with W3; Trivial.
Elim H1 with W0; Elim H3 with W3; Trivial.
Qed.

(* Commutation theorem *)

Lemma residuals_lift_rec : (U1,U2,U3:redexes)
   (residuals U1 U2 U3) -> (k,n:nat)
   (residuals (lift_rec_r U1 n k) (lift_rec_r U2 n k) (lift_rec_r U3 n k)).
Proof.
Induction 1; Simpl; Intros; Auto.
Rewrite lift_subst; Auto.
Qed.

Lemma residuals_lift : (U1,U2,U3:redexes)
   (residuals U1 U2 U3) -> (k:nat)
   (residuals (lift_r k U1) (lift_r k U2) (lift_r k U3)).
Proof.
Unfold lift_r; Intros; Apply residuals_lift_rec; Trivial.
Qed.
Hint residuals_lift.

Lemma residuals_subst_rec : (U1,U2,U3,V1,V2,V3:redexes)
   (residuals U1 U2 U3) -> (residuals V1 V2 V3) -> (k:nat)
   (residuals (subst_rec_r U1 V1 k) 
              (subst_rec_r U2 V2 k) (subst_rec_r U3 V3 k)).
Proof.
Induction 1; Simpl; Auto.
Intros n R k; Unfold insert_Var; Elim (compare k n); Auto.
Induction y; Auto.
Intros; Rewrite substitution; Auto.
Qed.
Hint residuals_subst_rec.

(***************************)
(* The Commutation Theorem *)
(***************************)

Theorem commutation : (U1,U2,U3,V1,V2,V3:redexes)
   (residuals U1 U2 U3) -> (residuals V1 V2 V3) -> 
   (residuals (subst_r V1 U1) (subst_r V2 U2) (subst_r V3 U3)).
Proof.
Unfold subst_r; Auto.
Qed.

Lemma residuals_comp : (U,V,W:redexes)(residuals U V W) -> (comp U V).
Proof.
Induction 1; Simpl; Auto.
Qed.

Lemma preservation1 : (U,V,UV:redexes)(residuals U V UV) ->
                      (T:redexes)(UVT:(union U V T))(residuals T V UV).
Proof.
(* Remark use of name UVT for uniform command below *)
Induction 1; Induction T; Intros; Inversion UVT; Auto.
Rewrite (max_false b); Auto.
Inversion H8; Auto. 
Qed.

Lemma preservation : (U,V,W,UV:redexes)(union U V W) ->
                     (residuals U V UV) -> (residuals W V UV).
Proof.
Intros; Apply preservation1 with U; Auto.
Qed.

Lemma mutual_residuals_comp : (W,U,UW:redexes)(RU:(residuals U W UW))
                  (V,VW:redexes)(RV:(residuals V W VW)) (comp UW VW).
Proof.
Induction W.
Intros; Inversion_clear RU; Inversion_clear RV; Trivial.
Intros; Inversion_clear RU; Inversion_clear RV; Apply Comp_Fun; Apply H with U0 U1; Auto.
Induction b; Intros; Generalize RU H; Inversion_clear RV.
Intro RU1; Inversion_clear RU1; Intros.
Apply subst_preserve_comp.
Cut (comp (Fun W0) (Fun W1)).
Intro CF; Inversion_clear CF; Trivial.
Apply H5 with (Fun U0) (Fun U1); Auto.
Apply H0 with U3 U2; Auto.
Intros; Inversion_clear RU; Apply Comp_Ap.
Apply H with U0 U1; Auto.
Apply H0 with U3 U2; Auto.
Qed.

(* We take residuals only by regular redexes *)

Lemma residuals_regular : (U,V,W:redexes)(residuals U V W) -> (regular V).
Proof.
Induction 1; Simpl; Auto.
Qed.

(* Conversely, residuals by compatible regular redexes always exist 
   (and are unique by residuals_function lemma above) *)

Lemma residuals_intro : (U,V:redexes)(comp U V) -> (regular V) ->
                        (Ex[W:redexes] (residuals U V W)).
Proof.
Induction 1; Simpl.
Intro n; Exists (Var n); Trivial.
Intros U0 V0 C E O; Elim (E O); Intros W0 R; Exists (Fun W0); Auto.
Induction b2.
Generalize H1; Elim H0; Try Contradiction.
Intros; Elim H7; Intros H8 H9; Elim (H6 H8); Intros FW0 R.
Inversion_clear R.
Elim (H3 H9); Intros W2 R2.
EApply ex_intro; Apply Res_redex.
Apply H10.
Apply R2.
Induction 1; Intros O1 O2; Elim (H1 O1); Intro W1; Elim (H3 O2); Intro W2.
Intros; Exists (Ap b1 W1 W2); Auto.
Qed.

(* Residuals preserve regularity *)

Lemma residuals_preserve_regular : 
      (U,V,W:redexes)(residuals U V W) -> (regular U) -> (regular W).
Proof.
Induction 1; Simpl; Auto.     
Induction b.
Generalize H1; Elim H0; Try Contradiction.
Intros; Elim H7; Split; Auto.
Induction 1; Split; Auto.
Induction b; Intros; Apply subst_preserve_regular; Elim H4; Auto.
Qed.

