(****************************************************************************)
(*                                                                          *)
(*                                                                          *)
(*  INTEGER.v                                                               *)
(*                                                                          *)
(*  Definition of integers from natural numbers using quotient types        *)
(*                                                                          *)
(*                                                                          *)
(*  Samuel Boutin                                                           *)
(*  Coq V5.10                                                               *)
(*  June  1995                                                              *)
(*                                                                          *)
(*                                                                          *)
(****************************************************************************)
(*                             INTEGER.v                                    *)
(****************************************************************************)


Require Arith.
Require quotient.
Require HS.
Require AC.

(*****************************************************************************)
(*                          Some basic stuff                                 *)
(*****************************************************************************)

Lemma eq_pair : (A,B:Set)(x,y:A*B)(Fst x)=(Fst y) -> (Snd x)=(Snd  y) -> x=y.
Intros A B.
Induction x; Intros x1 x2.
Induction y; Intros y1 y2.
Simpl.
Intros Hx Hy.
Elim Hx.
Elim Hy.
Auto.
Save.


(*****************************************************************************)
(*                   Definition of Z as a quotient type                      *)
(*****************************************************************************)

Section Integers.

Require nat.
Require productSyntax.
Require intSyntax.
Require Nat_in_com.
Require quoSyntax.

Definition Z_typ := nat*nat.

(* Z is define as the symetric completion of the semi-group N *)

Definition Z_rel := [x,y:nat*nat] [| x.1 + y.2 = y.1 + x.2 |].

Hint Unfold Z_rel.

Definition Z := Z_typ/Z_rel.

(*****************************************************************************
 unary_minus : Z -> Z 
*****************************************************************************)

Definition swap : nat*nat -> Z
                := [x:nat*nat] |[| (x.2,x.1) |]|z.

Lemma Compat_swap : (Compat Z_typ Z_rel Z swap).

Red.
Induction x; Intros x1 x2.
Induction y; Intros y1 y2.
Unfold Z_rel swap ; Simpl.
Intro H.
Apply (From_R_to_L Z_typ Z_rel).
Red;Simpl.
Rewrite -> (plus_sym x2 y1).
Rewrite <- H.
ACNa.
Save.

Definition unary_minus : Z -> Z
                       := (lift Z_typ Z_rel  Z swap Compat_swap).

(*****************************************************************************
 plusZ : Z -> Z -> Z
*****************************************************************************)

Definition plusZ_XXX : Z_typ -> Z_typ -> Z_typ
                     := [x,y:Z_typ] 
           [|( x.1 + y.1 , x.2 + y.2 )|].

Definition plusZ_XX : Z_typ -> Z_typ -> Z
                    := [x,y:Z_typ] |(plusZ_XXX x y)|z.

Lemma Compat_plusZ_XX: 
       (x:Z_typ) (Compat Z_typ Z_rel Z (plusZ_XX x)).

Induction x; Intros x1 x2.
Red.
Induction x; Intros y1 y2.
Induction y; Intros z1 z2.
Unfold Z_rel plusZ_XX plusZ_XXX;  Simpl.
Intro H.
Apply (From_R_to_L Z_typ Z_rel).
Red; Simpl.
ANa.
HSNa.
HSNa.
Assumption.
Save.

Definition plusZ_X : Z_typ -> Z -> Z
                   := [x:Z_typ]
              (lift Z_typ Z_rel Z (plusZ_XX x) (Compat_plusZ_XX x)).

Lemma Compat_plusZ_X: 
       (Compat Z_typ Z_rel (Z->Z) plusZ_X ).

Red.
Induction x; Intros x1 x2.
Induction y; Intros y1 y2.
Unfold Z_rel;Simpl.
Intro H.
Apply Ext.
Intro z.
Pattern z.
Apply (Closure_prop Z_typ Z_rel z).
Induction x; Intros a1 a2.
Unfold plusZ_X.
Repeat Rewrite -> Reduce.
Unfold plusZ_XX.
Apply (From_R_to_L Z_typ Z_rel).
Unfold Z_rel plusZ_XXX; Simpl.
ANa.
HSNa.
HSNa.
Assumption.
Save.

Definition plusZ : Z -> Z -> Z
                 := (lift Z_typ Z_rel (Z->Z) plusZ_X Compat_plusZ_X ).


(*****************************************************************************)
(*                         multZ : Z -> Z -> Z                               *)
(*****************************************************************************)

Definition multZ_XXX
          : Z_typ -> Z_typ -> Z_typ
          := [x,y :Z_typ] [|(x.1*y.1 + x.2*y.2, x.1*y.2 + x.2*y.1)|].

Definition multZ_XX
          : Z_typ -> Z_typ -> Z
          := [x,y :Z_typ] |(multZ_XXX x y)|z.

Lemma Compat_multZ_XX: 
       (x:Z_typ) (Compat Z_typ Z_rel Z (multZ_XX x)).

Induction x; Intros x1 x2.
Red.
Induction x; Intros y1 y2.
Induction y; Intros z1 z2.
Unfold Z_rel multZ_XX multZ_XXX;Simpl.
Intro H.
Apply (From_R_to_L Z_typ Z_rel).
Red.
Simpl.
ANa.
(*
  x : Z_typ
  x1 : nat
  x2 : nat
  x0 : Z_typ
  y1 : nat
  y2 : nat
  y : Z_typ
  z1 : nat
  z2 : nat
  H : y1 + z2=z1 + y2
  ============================
   x1*y1 + x2*y2 + x1*z2 + x2*z1=x1*z1 + x2*z2 + x1*y2 + x2*y1
*)
Rewrite -> (plus_permute [| x2*y2 |]).
Rewrite -> (plus_permute [| x2*z2 |]).
Repeat Rewrite <- mult_plus_distr_l.
Rewrite -> (plus_sym y2).
Rewrite <- H.
Rewrite -> (plus_sym y1).
HSNa.
Repeat Rewrite <- mult_plus_distr_l.
Elim H;Auto.
Save.

Definition multZ_X : Z_typ -> Z -> Z
                 := [x:Z_typ]
              (lift Z_typ Z_rel Z (multZ_XX x) (Compat_multZ_XX x)).

Lemma Compat_multZ_X : (Compat Z_typ Z_rel (Z->Z) multZ_X).

Red.
Induction x; Intros x1 x2.
Induction y; Intros y1 y2.
Unfold Z_rel;  Simpl.
Intro H.
Apply Ext.
Intro z.
Pattern z.
Apply (Closure_prop Z_typ Z_rel z).
Induction x; Intros a1 a2.
Unfold multZ_X.
Repeat Rewrite -> Reduce.
Unfold multZ_XX.
Apply (From_R_to_L Z_typ Z_rel).
Unfold Z_rel multZ_XXX; Simpl.
ANa.
Repeat Rewrite -> (plus_permute [| x1*a1 |]).
Repeat Rewrite -> (plus_permute [| y1*a1 |]).
Repeat Rewrite <- mult_plus_distr.
Elim H.
HSNa.
Repeat Rewrite <- mult_plus_distr.
Rewrite -> plus_sym.
Rewrite <- H.
Rewrite -> plus_sym.
Auto.
Save.

Definition multZ : Z -> Z -> Z
                 := (lift Z_typ Z_rel (Z->Z) multZ_X Compat_multZ_X).


End Integers.

