(****************************************************************************)
(*                                                                          *)
(*                                                                          *)
(*  NATSYNTAX.v                                                             *)
(*                                                                          *)
(*  Definition of an LL1 grammar for natural numbers using Coq facilities   *)
(*  for dynamic extension of the grammar                                    *)
(*  This includes the definition of a new entry: ``natural'' ,denoting      *)
(*  the grammar of naturals                                                 *)
(*  All the non terminals are enumerated in head of the file                *)
(*  The nonterminal formula matches elementary predicates on natural        *)
(*  numbers while the non-terminal expr for instance matches every          *)
(*  arithmetic expression where * has a bigger precedence as + (as usual)   *)
(*  When defining the grammar rules, we always factorise as much as         *)
(*  possible in order to facilitate sound extensions                        *)
(*                                                                          *)
(*                                                                          *)
(*  Samuel Boutin                                                           *)
(*  Coq V5.10                                                               *)
(*  June  1995                                                              *)
(*                                                                          *)
(*                                                                          *)
(****************************************************************************)
(*                            NATSYNTAX.v                                   *)
(****************************************************************************)


(*
coqc -I $COQ/theories/INIT -I "$QUO/Tactics/ML_struct" -I "$QUO/Tactics/AC" -I "$QUO/Tactics/HS" -I "."  NATSYNTAX
*)

(* Declaration of non-terminal symbols *)

Grammar natural ident := [ prim:ident($id) ] -> [$0 = $id].

(* formula is the non-terminal for predicates on natural numbers: look
at its definition to see the predicates for whom syntax facilities
are defined *)
Grammar natural formula :=.
Grammar natural atom[$p] :=.
(* expr is the non terminal for arithmetic expressions on natural numbers *)
Grammar natural expr :=.
Grammar natural exprsum[$p] :=.
Grammar natural fact :=.
Grammar natural factprod[$p]:=.
Grammar natural final :=.
Grammar natural application[$p] :=.
Grammar natural lexpr[$p]:=.
Grammar natural proj[$c]:=.

(* LL1 Grammar for atomic formulas *)

Grammar natural formula :=
  [ expr($c) atom[[$0 = $c]]($r) ] -> [$0 = $r].

Token "<=".

Grammar natural atom[$p]:=
  [ ] -> [$0 = $p]
| [ "=" expr($c)] -> [$0 = <<(eq nat $p $c)>>]
| [ "<=" expr($c)] -> [$0 = <<(le $p $c)>>]
| [ "<" expr($c)] -> [$0 = <<(lt $p $c)>>]
| [ "<" ">" expr($c)] -> [$0 = <<~(eq nat $p $c)>>].

(* LL1 Grammar rules for arithmetic*)

Grammar natural expr :=
  [ fact($c) exprsum[[$0 = $c]]($r)] -> [$0 = $r].

Grammar natural exprsum[$p] := 
  [ ] -> [$0 = $p]
| [ "+" fact($c) exprsum[[$0 =<<(plus $p $c)>>]]($r)] -> [$0 = $r].

Grammar natural fact :=
  [ final($c) factprod[[$0 = $c]]($r) ] -> [$0 = $r].

Grammar natural factprod[$p] :=
  [ ] -> [$0 = $p]
| [ "*" final($c) factprod[[$0 =<<(mult $p $c)>>]]($r)] -> [$0 = $r].

Grammar natural final :=
  [ ident($c) ] -> [$0 = ($VAR $c)]
| [ "[" command:command($c) "]" ] -> [$0 = $c]
| [ "(" expr($a) application[[$0 = $a]]($r) ] -> [$0 = $r]
| [ "0" ] -> [$0 = <<O>>]
| [ "1" ] -> [$0 = <<(S O)>>]
| [ "2" ] -> [$0 = <<(S (S O))>>]
| [ "3" ] -> [$0 = <<(S (S (S O)))>>]
| [ "4" ] -> [$0 = <<(S (S (S (S O))))>>]
| [ ident($c) proj[[$0=($VAR $c)]]($r) ] -> [$0 = $r].

Grammar natural proj[$c]:=
  [ "." "1"] -> [$0 = <<(fst nat nat $c)>>]
| [ "." "2"] -> [$0 = <<(snd nat nat $c)>>].

Grammar natural lexpr[$p] :=
  [ ")" ] -> [$0 = $p]
| [ expr($c1)  lexpr[[$0 = <<($p $c1)>>]]($r) ] -> [$0 = $r].

Grammar natural application[$p] :=
  [ lexpr[[$0 = $p]]($r)] -> [$0 = $r]
| [ "," expr($c) ")"] -> [$0 = <<($p,$c)>>].

(* Associated pretty printing rules *)

Syntax constr plus <<(plus $n1 $n2)>> 1
                 <$n1 : "left" : E> " + " <$n2 : "left" : E>  .

Syntax constr mult <<(mult $n1 $n2)>> 1
                 <$n1 : "left" : E> "*" <$n2 : "left" : E>  .

Syntax constr minus <<(minus $n1 $n2)>> 1
                 <$n1 : "left" : E> " - " <$n2 : "left" : E>  .

Syntax constr le <<(le $n1 $n2)>> 1
                 <$n1 : "left" : E> "<=" <$n2 : "left" : E>  .

Syntax constr lt <<(lt $n1 $n2)>> 1
                 <$n1 : "left" : E> "<" <$n2 : "left" : E>  .

Syntax constr fst <<(fst nat nat $c)>> 1
                 <$c : "left" : E> ".1".

Syntax constr snd <<(snd nat nat $c)>> 1
                 <$c : "left" : E> ".2".

Syntax constr zero <<O>> 1 "0".
Syntax constr one <<(S O)>> 1 "1".
Syntax constr two <<(S (S O))>> 1 "2".
Syntax constr three <<(S (S (S O)))>> 1 "3".
Syntax constr four <<(S (S (S (S O))))>> 1 "4".

