(****************************************************************************)
(*                                                                          *)
(*                                                                          *)
(*  NATURAL  .v                                                             *)
(*                                                                          *)
(* Defines (infix)  syntax for operations on nat                            *)
(* Defines tactics for addition and multiplication                          *)
(* Defines some additional properties for addition and multiplication       *)
(*                                                                          *)
(*                                                                          *)
(*                                                                          *)
(*  Samuel Boutin                                                           *)
(*  Coq V5.10                                                               *)
(*  June  1995                                                              *)
(*                                                                          *)
(*                                                                          *)
(****************************************************************************)
(*                              NATURAL.v                                   *)
(****************************************************************************)


(*
coqc -I $COQ/theories/INIT -I "$QUO/Tactics/ML_struct" -I "$QUO/Tactics/AC" -I "$QUO/Tactics/HS" -I "."  NATURAL
*)

Require Export Arith.
Require Export Plus.
Require Export Mult.
Require AC.
Require HS.

Section NAT.

Require NATSYNTAX.
Require Nat_in_com.

(* space around eq *)

Syntax constr equal <<(eq $_  $t1  $t2)>> 7
           [<hov 0>  <$t1:"term":E> [0 1]  " = " <$t2:"term":E>].



(*************************************************************
PROPERTIES OF nat
*************************************************************

eq_S         : (n,m:nat)                   n=m->(S n)=(S m)
pred_Sn      : (m:nat)                              m=(pred (S m))

plus_sym     : (n,m:nat)                       n + m = m + n
plus_assoc_l : (n,m,p:nat)               n + (m + p) = (n + m) + p
plus_permute : (n,m,p:nat)               n + (m + p) = m + (n + p)
plus_assoc_r : (n,m,p:nat)               (n + m) + p = n + (m + p)
plus_n_O     : (n:nat)                             n = n + O
plus_n_Sm    : (n,m:nat)                   (S n + m) = n + (S m)
simpl_plus_l : (n,m,p:nat)       n + m = n + p =>  m = p

mult_sym         : (n,m:nat)                   n * m = m * n
mult_assoc_l     : (n,m,p:nat)           n * (m * p) = (n * m) * p
mult_permute     : (n,m,p:nat)           n * (m * p) = m * (n * p)
mult_assoc_r     : (n,m,p:nat)           (n * m) * p = n * (m * p)
mult_n_O         : (n:nat)                         O = n * O
mult_n_Sm        : (n,m:nat)             (n * m) + n = n * (S m)

mult_plus_distr   : (n,m,p:nat)           (n + m) * p = (n * p) + (m * p)
mult_plus_distr_l : (n,m,p:nat)           p * (n + m) = (p * n) + (p * m).
mult_minus_distr : (n,m,p:nat)           (n - m) * p = (n * p) - (m * p)

minus_n_O        : (n:nat)                         n = n - O
minus_n_n        : (n:nat)                         O = n - n

minus_plus_simpl : (n,m,p:nat)                 n - m = (p + n) - (p + m)
plus_minus       : (n,m,p:nat)        n = m + p => p = n - m

*************************************************************)


(* Decision AC for addition and multiplication on N*)

Lemma plus_simpl_l : (n,m,p:nat)  [| m = p |] -> [| n+m = n+p |].
Intros.
Elim H;Auto.
Save.

Lemma mult_simpl_l : (n,m,p:nat) [| m = p |] -> [| n*m = n*p |].
Intros.
Elim H;Auto.
Save.

Lemma mult_permute : (n,m,p:nat) [| n*(m*p) = m*(n*p) |].
Intros.
Elim (mult_sym p n).
Elim (mult_assoc_r m p n).
Elim (mult_sym n).
Auto.
Save.

Lemma mult_plus_distr_l : (n,m,p:nat) [| p*(n+m) = p*n + p*m |].
Intros.
Rewrite -> (mult_sym p).
Rewrite -> (mult_plus_distr n m p).
Repeat Elim (mult_sym p);Auto.
Save.

End NAT.

Grammar tactic simple_tactic :=
  ["CNa"] ->
  [$0 = <:tactic:<Intros; AC_OF eq plus plus_sym plus_permute; Auto>>].

Grammar tactic simple_tactic :=
  ["ANa"] ->
  [$0 = <:tactic:<Intros;Repeat Elim plus_assoc_l>>].

Grammar tactic simple_tactic :=
  ["ACNa"] ->
  [$0 = <:tactic:<Intros;Repeat Elim plus_assoc_l;
                   AC_OF eq plus plus_sym plus_permute; Auto>>].

Grammar tactic simple_tactic :=
  ["HSNa"] ->
  [$0 = <:tactic:<Intros; HS_OF eq plus plus_sym plus_permute plus_simpl_l>>].

Grammar tactic simple_tactic :=
  ["CNm"] ->
  [$0 = <:tactic:< Intros;
                   AC_OF eq mult mult_sym mult_permute;
                   Auto>>].

Grammar tactic simple_tactic :=
  ["ANm"] ->
  [$0 = <:tactic:<Intros;
                  Repeat Elim mult_assoc_l>>].

Grammar tactic simple_tactic :=
  ["ACNm"] ->
  [$0 = <:tactic:< Intros;
                   Repeat Elim mult_assoc_l;
                   AC_OF eq mult mult_sym mult_permute;
                   Auto>>].

Grammar tactic simple_tactic :=
  ["HSNm"] ->
  [$0 = <:tactic:<Intros;
                  HS_OF eq mult mult_sym mult_permute mult_simpl_u>>].
