(****************************************************************************)
(*                                                                          *)
(*                                                                          *)
(*  subset.v                                                                *)
(*                                                                          *)
(*  Definition of subsets as types                                          *)
(*                                                                          *)
(*                                                                          *)
(*  Samuel Boutin                                                           *)
(*  Coq V5.10                                                               *)
(*  June  1995                                                              *)
(*                                                                          *)
(*                                                                          *)
(****************************************************************************)
(*                              subset.v                                    *)
(****************************************************************************)


(* 
We define subset as types in the same way as quotients:
like in the deliverable model where a subset is a type
and a predicate on that type.
The alternative consisting of using a strong sum to encode
this feature suffers from the fact that we need to
use proof irrelevance if we want two
objects (a,pa) (a,pb) with pa,pb:(P a) to be identified.
*)

(* Definition of a subset: *)

Parameter MK_SUBSET : (A:Set)(R:A->Prop)Set.

(* An element of a subset is an element of the underlying type
with a proof that this element satisfies the underlying predicate *)

Parameter In_subset : (A:Set)(R:A->Prop)(a:A)(R a)->(MK_SUBSET A R).

(* Every element of a subset is an element of the underlying type *)

Parameter Out_subset : (A:Set)(R:A->Prop)(MK_SUBSET A R)->A.

(* Every element of the subset should satisfy the underlying predicate *)

Parameter proof : (A:Set)(R:A->Prop)(t:(MK_SUBSET A R))(R (Out_subset A R t)).

(* Rewrite rules explaining coercion principles between the subset and its 
underlying type *)

Axiom In_Out : (A:Set)(R:A->Prop)(t:(MK_SUBSET A R))(p:(R (Out_subset A R t)))
                 (In_subset A R (Out_subset A R t) p) = t.

Axiom Out_In : (A:Set)(R:A->Prop)(t:A)(p:(R t))
                 (Out_subset A R (In_subset A R t p)) = t.

(* Without proof irrelevance we should identify two element of a subset
as soon as their projection on the ground type are leibniz equal *)

Axiom Canonic : (A:Set)(R:A->Prop)(t:A)(p,p':(R t))
               (In_subset A R t p) = (In_subset A R t p').

(* Closure axioms defining the canonical form of objects of a subset type *)

Axiom Prop_closure: (A:Set)(R:A->Prop)(P:(MK_SUBSET A R)->Prop)
                       (x:(MK_SUBSET A R))
             ((y:A)(p:(R y))(P (In_subset A R y p)))->
                         (P x).

Axiom Set_closure: (A:Set)(R:A->Prop)(P:(MK_SUBSET A R)->Set)
                       (x:(MK_SUBSET A R))
             ((y:A)(p:(R y))(P (In_subset A R y p)))->
                         (P x).


(* Notation for subsets: *)

Token "!".

Grammar command command1 := [ command0($c) "!" command0($c2) ] ->
                  [$0 = <<(MK_SUBSET $c $c2)>>].

Syntax constr MK_SUBSET <<(MK_QUO $c $c1)>> 1 
                 <$c : "The set" : L> "!" <$c1 : "The predicate" : L>  .

