(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                                                                          *)
(****************************************************************************)
(*                                 Words.v                                  *)
(****************************************************************************)
(* G. Huet - V5.8  Nov. 1994 *)
(*    ported V5.10 June 1995 *)

Require Bool "$COQTH/BOOL/Bool".

(*****************)
(* Boolean words *)
(*****************)

Inductive word : Set := empty : word 
                      | bit : bool -> word -> word.

(* Remark : word ~ bool list *)

(* word concatenation : logical definition *)
Inductive conc : word -> word -> word -> Prop :=
  conc_empty : (v:word)(conc empty v v)
| conc_bit   : (u,v,w:word)(b:bool)(conc u v w)->(conc (bit b u) v (bit b w)).

(* word concatenation : functional definition *)
Recursive Definition append : word -> word -> word :=
    empty     v => v
  | (bit b w) v => (bit b (append w v)).

(* Abbreviation for: 

Fixpoint append [u:word] : word -> word :=
    [v:word]<word>Case u of
    (* empty *)   v
    (* bit b w *) [b:bool][w:word](bit b (append w v)) end. *)

(* Relating the two definitions; unused below *)
Lemma conc_append : (u,v,w:word)(conc u v w) -> w=(append u v).
Proof.
Induction 1; Simpl; Trivial.
Induction 2; Trivial.
Qed.

(* Associativity of append; unused below *)
Lemma assoc_append : 
      (u,v,w:word)(append u (append v w))=(append (append u v) w).
Proof.
Induction u; Simpl; Intros; Auto.
Rewrite H; Trivial.
Qed.


(**************)
(* Singletons *)
(**************)

Definition single := [b:bool](bit b empty).


(*********************)
(* Parities of words *)
(*********************)

Mutual Inductive odd : word -> Prop :=
  even_odd : (w:word)(even w)->(b:bool)(odd (bit b w))

            with even : word -> Prop :=
  even_empty : (even empty)
| odd_even : (w:word)(odd w)->(b:bool)(even (bit b w)).

Hint odd_even even_empty even_odd.

Lemma not_odd_empty : ~(odd empty).
Proof.
Unfold not; Intro od.
Inversion od.
Qed.

Hint not_odd_empty.

Lemma inv_odd : (w:word)(b:bool)(odd (bit b w))->(even w).
Proof.
Intros w b od.
Inversion od; Trivial.
Qed.

Lemma inv_even : (w:word)(b:bool)(even (bit b w))->(odd w).
Proof.
Intros w b ev.
Inversion ev; Trivial.
Qed.

(**********************)
(* (odd w) + (even w) *)
(**********************)

Lemma odd_or_even : (w:word)(odd w) \/ (even w).
Proof.
Induction w; Auto.
Induction 1; Intros.
Right; Auto.
Left; Auto.
Qed.

Lemma not_odd_and_even : (w:word)(odd w)->(even w)->False.
Proof.
Induction w; Intros.
Elim not_odd_empty; Trivial.
Apply H.
Apply inv_even with b; Trivial.
Apply inv_odd with b; Trivial.
Qed.

(************************)
(* Parities of subwords *)
(************************)

Lemma odd_even_conc : (u,v,w:word)(conc u v w) ->
           (odd w) /\
                   ( (odd u) /\ (even v)
                  \/ (even u) /\ (odd v))
        \/ (even w) /\
                   ( (odd u) /\ (odd v)
                  \/ (even u) /\ (even v)).
Proof.
Induction 1; Intros.
Elim (odd_or_even v0); Auto.
Elim H1; [Right | Left]; Intuition.
Qed.

Lemma even_conc : (u,v,w:word)(conc u v w) -> (even w) ->
                    (odd u) /\ (odd v)
                 \/ (even u) /\ (even v).
Proof.
Intros u v w c e; Elim odd_even_conc with u v w; Intros.
Elim H; Intro o; Elim not_odd_and_even with w; Auto.
Elim H; Intros; Trivial.
Trivial.
Qed.

