(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                           sur_les_relations.v                            *)
(****************************************************************************)

(*****************************************************************************)
(*          Projet Coq  - Calculus of Inductive Constructions V5.8           *)
(*****************************************************************************)
(*                                                                           *)
(*      Meta-theory of the explicit substitution calculus lambda-env         *)
(*      Amokrane Saibi                                                       *)
(*                                                                           *)
(*      September 1993                                                       *)
(*                                                                           *)
(*****************************************************************************)


                  (*  Generalites sur les relations  *)


(**************)
(* Relations  *)
(**************)

Section Rels.

Variable A: Set.

(* R* fermeture reflexive-transitive d une relation binaire R *)

Inductive explicit_star [R:A->A->Prop]: A->A->Prop
:=  star_refl: (x:A)(explicit_star R x x)
 | star_trans1 : (x,y,z:A)(R x y)->(explicit_star R y z)->(explicit_star R x z).

(* composition de deux relations *)

Inductive explicit_comp_rel [R1,R2:A->A->Prop]: A->A->Prop
:= comp_2rel: (x,y,z:A)(R1 x y)->(R2 y z)->(explicit_comp_rel R1 R2 x z).

(* R+ frmeture transitive de R *)

Inductive explicit_rel_plus [R:A->A->Prop]:A->A->Prop
:=  relplus_1step: (x,y:A)(R x y)->(explicit_rel_plus R x y)
 | relplus_trans1: (x,y,z:A)(R x y)->(explicit_rel_plus R y z)->(explicit_rel_plus R x z). 

End Rels.

Hint star_refl.
Hint relplus_1step.

Syntactic Definition star := (explicit_star ?).
Syntax constr star3 <<(explicit_star $1 $t1 $t2 $t3)>> 10
  [<hov 0> "star " <$t1:"term":L> " " <$t2:"term":L>  " " <$t3:"term":L>].
Syntax constr star2 <<(explicit_star $1 $t1 $t2)>> 10
  [<hov 0> "star " <$t1:"term":L> " " <$t2:"term":L>].
Syntax constr star1 <<(explicit_star $1 $t1)>> 10
  [<hov 0> "star " <$t1:"term":L>].

Syntactic Definition comp_rel := (explicit_comp_rel ?).
Syntax constr comp_rel4 <<(explicit_comp_rel $1 $t1 $t2 $t3 $t4)>> 10
  [<hov 0> "comp_rel " <$t1:"term":L> " " <$t2:"term":L>
                " " <$t3:"term":L> " " <$t4:"term":L>     ].
Syntax constr comp_rel2 <<(explicit_comp_rel $1 $t1 $t2)>> 10
  [<hov 0> "comp_rel " <$t1:"term":L> " " <$t2:"term":L>].

Syntactic Definition rel_plus := (explicit_rel_plus ?).
Syntax constr rel_plus3 <<(explicit_rel_plus $1 $t1 $t2 $t3)>> 10
  [<hov 0> "rel_plus " <$t1:"term":L> " " <$t2:"term":L> " " <$t3:"term":L>].
Syntax constr rel_plus2 <<(explicit_rel_plus $1 $t1 $t2)>> 10
  [<hov 0> "rel_plus " <$t1:"term":L> " " <$t2:"term":L> ].
Syntax constr rel_plus1 <<(explicit_rel_plus $1 $t1)>> 10
  [<hov 0> "rel_plus " <$t1:"term":L> ].



(**************)
(* proprietes *)


Section rels_prop.

Variable A: Set.
Variable R: A->A->Prop.

(* R confluente *)

Definition confluence_en := [x:A](y,z:A)(star R x y)->(star R x z)->
                            (Ex ([u:A](star R y u)/\(star R z u))).

Definition explicit_confluence := (x:A)(confluence_en x).

(* R localement confluente *)

Definition local_confluence_en := [x:A](y,z:A)(R x y)->(R x z)->
                         (Ex( ([u:A](star R y u)/\(star R z u)))).

Definition explicit_local_confluence := (x:A)(local_confluence_en x).

(* R fortement confluente *)

Definition strong_confluence_en := [x:A](y,z:A)(R x y)->(R x z)->
                              (Ex ([u:A](R y u)/\(R z u))).

Definition explicit_strong_confluence := (x:A)(strong_confluence_en x).

End rels_prop.

Syntactic Definition confluence := (explicit_confluence ?).
Syntax constr confluence <<(explicit_confluence $1 $t1)>> 10
  [<hov 0> "confluence " <$t1:"term":L> ].

Syntactic Definition local_confluence := (explicit_local_confluence ?).
Syntax constr local_confluence <<(explicit_local_confluence $1 $t1)>> 10
  [<hov 0> "local_confluence " <$t1:"term":L> ].

Syntactic Definition strong_confluence := (explicit_strong_confluence ?).
Syntax constr strong_confluence <<(explicit_strong_confluence $1 $t1)>> 10
  [<hov 0> "strong_confluence " <$t1:"term":L> ].

(* inclusion de relations binaires, R1 inclus dans R2: R1 incl R2*)

Definition explicit_inclus := [A:Set][R1,R2:A->A->Prop]
      (x,y:A)(R1 x y)->(R2 x y).

Syntactic Definition inclus := (explicit_inclus ?).
Syntax constr inclus <<(explicit_inclus $1 $t1 $t2)>> 10
  [<hov 0> "inclus " <$t1:"term":L> " " <$t2:"term":L>].


Section relations_noetherian.

Variable U:Set.

Variable R:U->U->Prop.

(* Sets as characteristic predicates over universe U *)
Definition a_set := U->Prop.

(* A is a subset of B *)
Definition sub := [A,B:a_set](x:U)(A x)->(B x).

(* The full universe *)
Definition universal := [A:a_set](x:U)(A x).

(* Adjoint map *)
Definition adjoint : a_set->a_set := [A:a_set][x:U](sub (R x) A).

Definition hereditary := [A:a_set](sub (adjoint A) A).
(*  i.e  (hereditary A) <-> (x:A)(sub (R x) A)->(A x) *)

Definition explicit_noetherian := (A:a_set)(hereditary A)->(universal A).

End relations_noetherian.

Syntactic Definition noetherian := (explicit_noetherian ?).
Syntax constr noetherian <<(explicit_noetherian $1 $t1)>> 10
  [<hov 0> "noetherian " <$t1:"term":L> ].

(**********************)
(* quelques resultats *)

(* sur le Ex *)

Goal (A:Set)(P,Q:A->Prop)(Ex( ([u:A](P u)/\(Q u))))->(Ex ([u:A](Q u)/\(P u))).
Induction 1;Intros u1 H1.
Elim H1;Intros H2 H3.
Exists u1;Split;Assumption.
Save Ex_PQ.
Hint Ex_PQ.
 
(* sur les constructions de relations *)

Lemma star_trans:(A:Set)(R:A->A->Prop)(x,y,z:A)
                 (star R x y)->(star R y z)->(star R x z).
Intros A R x y z H;Elim H.
Intros x0 H1;Assumption.
Intros x0 y0 z0 H1 H2 H3 H4;Apply star_trans1 with y0.
Assumption.
Exact (H3 H4).
Save.

Goal (A:Set)(R:A->A->Prop)(x,y:A)(R x y)->(star R x y).
Intros;Apply star_trans1 with y.
Assumption.
Apply star_refl.
Save star_step1.

Hint star_step1.

Goal (A:Set)(R1,R2:A->A->Prop)(M,N:A)(comp_rel R1 R2 M N)->
                                  (Ex [u:A] (R1 M u)/\(R2 u N)).  
Intros A R1 R2 M N H;Elim H.
Intros x y z H1 H2;Exists y;Split;Assumption.
Save comp_case.

Goal (A:Set)(R:A->A->Prop)(x,y:A)(comp_rel R (star R) x y)->
                                 (rel_plus R x y).
Intros A R x y H;Elim H.
Intros a b c H1 H2;Generalize H1;Generalize a.
Elim H2.
Intros;Apply relplus_1step;Assumption.
Intros x0 y0 z H3 H4 H5 a0 H6;Apply relplus_trans1 with x0.
Assumption.
Apply H5;Assumption.
Save comp_relplus.

Goal (A:Set)(R:A->A->Prop)(M,N:A)(star R M N)->((M=N) \/
                                  (Ex([u:A] (R M u)/\(star R u N)))).
Intros A R M N H;Elim H.
Intros x;Left;Trivial.
Intros x y z H1 H2 H3;Right;Exists y;Split;Trivial.
Save star_case.

Goal (A:Set)(R:A->A->Prop)(x,y,z:A)(rel_plus R x y)-> 
     (rel_plus R y z)->(rel_plus R x z).
Induction 1.
Intros;Apply relplus_trans1 with y0;Trivial.
Intros;Apply relplus_trans1 with y0;Auto.
Save Rplus_transitive.

Goal (A:Set)(R:A->A->Prop)(x,y:A)(rel_plus R x y)->(star R x y).
Induction 1;Intros.
Auto.
Apply star_trans1 with y0;Auto.
Save Rplus_Rstar.

Hint Rplus_Rstar.

Goal (A:Set)(R:A->A->Prop)
      (x,y,z:A)(star R x y)->(rel_plus R y z)-> 
           (Ex([u:A](R x u) /\ (star R u z))).
Induction 1;Intros.
Elim H0;Intros.
Exists y0;Auto.
Exists y0;Auto.
Exists y0; 
Split;Trivial.
Apply star_trans with z0;Auto.
Save Rstar_Rplus_R.

(* sur les relations noetheriennes *)

Goal (A:Set)(R:A->A->Prop)(noetherian R)-> 
     (A1:(a_set A))(hereditary A (rel_plus R) A1) -> 
     (universal A (adjoint A (star R) A1)).
Unfold explicit_noetherian;Unfold hereditary;Unfold universal; 
 Unfold sub;Intros A R N A1 H x.
Apply (N (adjoint A (star R) A1)).
Unfold adjoint;Unfold sub;Intros.
Apply H;Unfold adjoint;Unfold sub;Intros.
Elim Rstar_Rplus_R with A R x0 x1 x2;Trivial.
Intro z;Induction 1;Intros C1 C2;Apply H0 with z;Trivial.
Save noetherian_course_of_values.

Lemma plus_preserves_noetherian :
  (A:Set)(R:A->A->Prop)(noetherian R)->(noetherian (rel_plus R)).
Generalize noetherian_course_of_values.
Unfold adjoint;Unfold universal;Unfold sub;Intros.
Unfold explicit_noetherian;Unfold universal;Unfold sub;Intros.
Apply (H A R H0 A0 H1 x x).
Auto.
Save.

Lemma noetherian_induction1: (A:Set)(R:A->A->Prop)(noetherian R)->
    (x:A)(P:A->Prop)((y:A)((z:A)(R y z)->(P z))->(P y))->(P x).
Unfold explicit_noetherian;Unfold universal;Unfold hereditary;
  Unfold adjoint;Unfold sub;Unfold a_set;Intros.
Pattern x;Apply H;Exact H0.
Save.

Lemma noetherian_induction: (A:Set)(R:A->A->Prop)(noetherian R)->
    (x:A)(P:A->Prop)((y:A)((z:A)(rel_plus R y z)->(P z))->(P y))->(P x).
Intros;Pattern x;Apply noetherian_induction1 with A (rel_plus R).
Apply plus_preserves_noetherian;Assumption.
Exact H0.
Save.

Lemma noether_inclus: (A:Set)(R,R':A->A->Prop)
                      (noetherian R)->
                      ((x,y:A)(R' x y)->(R x y))->
                      (noetherian R').
Intros;Unfold explicit_noetherian;Unfold universal;Unfold hereditary;
  Unfold adjoint;Unfold sub;Unfold a_set;Intros.
Pattern x;Apply (noetherian_induction1 A R H);Auto.
Save.

(* sur l'inclusion *)

Goal (A:Set)(R,S:A->A->Prop)(inclus R (star S))->(inclus (star R) (star S)).
Intros A R S H;Red;Induction 1.
Auto.
Intros x0 y0 z H1 H2 H3;Apply star_trans with y0;Auto.
Save inclus_star.

Goal (A:Set)(R,S:A->A->Prop)(inclus R S)->(inclus (star R) (star S)).
Unfold explicit_inclus;Induction 2.
Auto.
Intros x0 y0 z H1 H2 H3;Apply star_trans1 with y0.
Apply (H x0 y0 H1).
Assumption.
Save inclus_reg_star.
Hint inclus_reg_star.

Goal (A:Set)(R1,R2,S:A->A->Prop)(inclus R1 S)->(inclus R2 S)->
     (* S transitive *)((x,y,z:A)(S x y)->(S y z)->(S x z))->
     (inclus (comp_rel R1 R2) S).   
Intros A R1 R2 S H H0 H1;Red;Induction 1.
Intros x0 y0 z H3 H4;Apply H1 with y0;Auto.
Save inclus_comp.
Hint inclus_comp.

(* sur la confluence *)

Goal (A:Set)(R:A->A->Prop)(strong_confluence R)->(confluence R).
Intros A R H;Red;Red.
Intros x y z H1;Generalize z;Elim H1.
Intros x0 z0 H2;Exists z0;Split;Auto.
Intros x0 y0 y1 H2 H3 H4 z0 H5.
Cut (Ex [u:A](star R y0 u)/\(R z0 u)).
Intro H6;Elim H6;Intros z1 H7;Elim H7;Intros H8 H9.
Elim (H4 z1 H8);Intros u H10;Elim H10;Intros H11 H12.
Exists u;Split.
Assumption.
Apply star_trans1 with z1;Assumption.
Generalize H2;Generalize y0;Elim H5.
Intros x1 y2 H6;Exists y2;Split;Auto.
Intros x1 y2 z1 H6 H7 H8 y3 H9;Elim (H x1 y3 y2).
Intros x2 H10;Elim H10;Intros H11 H12.
Elim (H8 x2 H12);Intros u H13;Elim H13;Intros H14 H15.
Exists u;Split;[Apply star_trans1 with x2;Assumption|Assumption];Trivial.
Assumption.
Assumption.
Save strong_conf_conf.

Goal (A:Set)(R,S:A->A->Prop)(inclus R S)->(inclus S (star R))->
     (confluence S)->(confluence R).
Red;Red;Intros A R S H H0 H1 x y z H2 H3.
Cut (inclus (star R) (star S)).
2:Auto.
Intro H4;Elim (H1 x y z (H4 x y H2) (H4 x z H3)).
Intros x' H5;Elim H5;Intros H6 H7.
Exists x';Split.
Exact (inclus_star A S R H0 y x' H6).
Exact (inclus_star A S R H0 z x' H7).
Save inclus_conf.







