(****************************************************************************)
(*                                                                          *)
(*                          Constructive Geometry                           *)
(*                             according to                                 *)
(*                            Jan von Plato                                 *)
(*                                                                          *)
(*                               Coq V5.10                                  *)
(*									    *)
(*			        Gilles Kahn 				    *)
(*				   INRIA                                    *)
(*		              Sophia-Antipolis		 	            *)
(*									    *)
(*                                                                          *)
(* Acknowledgments: This work started in Turin in May 95 with Yves Bertot.  *)
(* Thanks to the Newton Institute for providing an exceptional work         *)
(* environment in Summer 1995 so that I could complete this.                *)
(****************************************************************************)


(****************************************************************************)
(*                              basis.v                                     *)
(*                                                                          *)
(* Basic definitions, axioms, figures                                       *)
(****************************************************************************)

Definition Relation :=  [U: Set] U -> U -> Prop.

Definition Reflexive : (U: Set) (Relation U) -> Prop :=
   [U: Set] [R: (Relation U)] (x: U) (R x x).

Definition Transitive : (U: Set) (Relation U) -> Prop :=
   [U: Set] [R: (Relation U)] (x, y, z: U) (R x y) -> (R y z) -> (R x z).

Definition Symmetric : (U: Set) (Relation U) -> Prop :=
   [U: Set] [R: (Relation U)] (x, y: U) (R x y) -> (R y x).

Inductive Equivalence [U:Set; R:(Relation U)] : Prop :=
      Definition_of_equivalence:
        (Reflexive U R) -> (Symmetric U R) -> (Transitive U R) ->
         (Equivalence U R).

Definition Negation : (U: Set) (Relation U) -> (Relation U) :=
   [U: Set] [R: (Relation U)] [x, y: U] ~ (R x y).
Hint Unfold Negation.

Theorem Sym_imp_NegSym:
  (U: Set) (R: (Relation U)) (Symmetric U R) -> (Symmetric U (Negation U R)).
Unfold Symmetric Negation.
(Intros U R H' x y H'0; Red; Intro H'1); Auto.
Qed.
Hint Sym_imp_NegSym.

Definition Irreflexive : (U: Set) (Relation U) -> Prop :=
   [U: Set] [R: (Relation U)] (x: U) ~ (R x x).

Definition Separating : (U: Set) (Relation U) -> Prop :=
   [U: Set] [R: (Relation U)] (x, y, z: U) (R x y) -> (R x z) \/ (R y z).

Inductive Apartness [U:Set; R:(Relation U)]: Prop :=
      Definition_of_apartness:
        (Irreflexive U R) -> (Separating U R) -> (Apartness U R).
Hint Unfold Reflexive Irreflexive Symmetric Transitive.
Hint Definition_of_equivalence Definition_of_apartness.

(****************************************************************)

Parameter Point: Set.
Parameter Line: Set.
Parameter DiPt: Point -> Point -> Prop.
Parameter DiLn: Line -> Line -> Prop.
Parameter ConLn: Line -> Line -> Prop.

Axiom apart_dipt: (Apartness Point DiPt).

Axiom apart_diln: (Apartness Line DiLn).

Axiom apart_con: (Apartness Line ConLn).

(****************************************************************)
Hint apart_dipt apart_diln apart_con.

Theorem Apart_imp_Sym:
  (U: Set) (R: (Relation U)) (Apartness U R) -> (Symmetric U R).
Proof.
Intros U R H'; Elim H'.
Intros H'0 H'1; Red.
Intros x y H'2; Red in H'1.
Elim (H'1 x y x); Trivial.
Intro H'3; Elim (H'0 x); Trivial.
Qed.
Hint Apart_imp_Sym.

Theorem sym_DiPt: (x, y: Point) (DiPt x y) -> (DiPt y x).
Proof.
Intros x y H'; Cut (Symmetric Point DiPt); Auto.
Qed.

Theorem sym_DiLn: (x, y: Line) (DiLn x y) -> (DiLn y x).
Proof.
Intros x y H'; Cut (Symmetric Line DiLn); Auto.
Qed.

Theorem sym_ConLn: (x, y: Line) (ConLn x y) -> (ConLn y x).
Proof.
Intros x y H'; Cut (Symmetric Line ConLn); Auto.
Qed.
Immediate sym_DiPt sym_DiLn sym_ConLn.

Theorem Neg_apart_equiv:
  (U: Set) (R: (Relation U)) (Apartness U R) -> (Equivalence U (Negation U R)).
Proof.
Intros U R H'; Elim H'.
Constructor 1; Auto.
Unfold Transitive Negation.
Intros x y z H'2 H'3; Red; Intro H'4.
Red in H0.
Elim (H0 x z y); Auto.
Cut (Symmetric U R); Auto.
Qed.
Hint Neg_apart_equiv.

Definition EqPt :=  (Negation Point DiPt).

Definition EqLn :=  (Negation Line DiLn).

Definition Par :=  (Negation Line ConLn).

Theorem equiv_EqPt: (Equivalence Point EqPt).
Proof.
Unfold EqPt; Auto.
Qed.
Hint equiv_EqPt.

Theorem equiv_EqLn: (Equivalence Line EqLn).
Proof.
Unfold EqLn; Auto.
Qed.
Hint equiv_EqLn.

Theorem equiv_Par: (Equivalence Line Par).
Proof.
Unfold Par; Auto.
Qed.
Hint equiv_Par.

Theorem sym_EqPt: (x, y: Point) (EqPt x y) -> (EqPt y x).
Proof.
Intros x y H'; Cut (Symmetric Point EqPt); Auto.
Unfold 1 EqPt; Auto.
Qed.

Theorem sym_EqLn: (x, y: Line) (EqLn x y) -> (EqLn y x).
Proof.
Intros x y H'; Cut (Symmetric Line EqLn); Auto.
Unfold 1 EqLn; Auto.
Qed.

Theorem sym_Par: (x, y: Line) (Par x y) -> (Par y x).
Proof.
Intros x y H'; Cut (Symmetric Line Par); Auto.
Unfold 1 Par; Auto.
Qed.
Immediate sym_EqPt sym_EqLn sym_Par.

(****************************************************************)
Parameter Apart: Point -> Line -> Prop.
(****************************************************************)

Definition Incident :=  [a: Point] [l: Line] ~ (Apart a l).

Record Segment : Set := Seg {
   origin: Point;
   extremity: Point;
   Seg_cond: (DiPt origin extremity) }.

Record Twolines : Set := Twol {
   line1: Line;
   line2: Line;
   Twol_cond: (ConLn line1 line2) }.

(****************************************************************)

Axiom line:
   (x: Segment)
    {l: Line | (Incident (origin x) l) /\ (Incident (extremity x) l)}.

Axiom point:
   (x: Twolines) {a: Point | (Incident a (line1 x)) /\ (Incident a (line2 x))}.

(****************************************************************)

Definition ln: Segment -> Line.
Proof.
Intro x; Elim (line x); Intros x0 H'; Exact x0.
Defined.

Definition pt: Twolines -> Point.
Proof.
Intro x; Elim (point x); Intros x0 H'; Exact x0.
Defined.

Theorem inc_ln1: (x: Segment) (Incident (origin x) (ln x)).
Proof.
Intro x; Elim x.
Intros a b d; Unfold ln; Simpl.
Elim (line (Seg a b d)); Simpl.
Tauto.
Qed.

Theorem inc_ln2: (x: Segment) (Incident (extremity x) (ln x)).
Proof.
Intro x; Elim x.
Intros a b d; Unfold ln; Simpl.
Elim (line (Seg a b d)); Simpl.
Tauto.
Qed.

Theorem inc_pt1: (x: Twolines) (Incident (pt x) (line1 x)).
Proof.
Intro x; Elim x.
Intros a b d; Unfold pt; Simpl.
Elim (point (Twol a b d)); Simpl.
Tauto.
Qed.

Theorem inc_pt2: (x: Twolines) (Incident (pt x) (line2 x)).
Proof.
Intro x; Elim x.
Intros a b d; Unfold pt; Simpl.
Elim (point (Twol a b d)); Simpl.
Tauto.
Qed.
Hint inc_ln1 inc_ln2 inc_pt1 inc_pt2.

(****************************************************************)

Axiom el_ax:
   (x: Segment) (l, m: Line) (DiLn l m) ->
    ((Apart (origin x) l) \/ (Apart (extremity x) l)) \/
    ((Apart (origin x) m) \/ (Apart (extremity x) m)).

Axiom cmp_apt_dipt:
   (a, b: Point) (l: Line) (Apart a l) -> (DiPt a b) \/ (Apart b l).

Axiom cmp_apt_diln:
   (a: Point) (l, m: Line) (Apart a l) -> (DiLn l m) \/ (Apart a m).

Axiom cmp_con_diln: (l, m, n: Line) (ConLn l m) -> (DiLn m n) \/ (ConLn l n).

(****************************************************************)

Record Triangle : Set := Tri {
   summit: Point;
   base: Segment;
   Tri_cond: (Apart summit (ln base)) }.

Theorem Triangle_def: (t: Triangle) (Apart (summit t) (ln (base t))).
Proof.
Intro t; Elim t.
Simpl.
Intros s b H'; Exact H'.
Qed.
Hint Triangle_def.

Definition SPar : (Relation Line) :=  [l, m: Line] (Par l m) /\ (DiLn l m).

Record Parallelogram : Set := Pgram {
   side1: Segment;
   side2: Segment;
   side3: Segment;
   side4: Segment;
   connect1:
      (origin side3) = (origin side1) /\
      (extremity side3) = (origin side2);
   connect2:
      (origin side4) = (extremity side1) /\
      (extremity side4) = (extremity side2);
   parsides_i: (SPar (ln side1) (ln side2));
   parsides_ii: (SPar (ln side3) (ln side4)) }.



