(****************************************************************************)
(*                                                                          *)
(*                          Constructive Geometry                           *)
(*                             according to                                 *)
(*                            Jan von Plato                                 *)
(*                                                                          *)
(*                               Coq V5.10                                  *)
(*									    *)
(*			        Gilles Kahn 				    *)
(*				   INRIA                                    *)
(*		              Sophia-Antipolis		 	            *)
(*									    *)
(*                                                                          *)
(* Acknowledgments: This work started in Turin in May 95 with Yves Bertot.  *)
(* Thanks to the Newton Institute for providing an exceptional work         *)
(* environment in Summer 1995 so that I could complete this.                *)
(****************************************************************************)

(****************************************************************************)
(*                       orthogonality.v                                    *)
(****************************************************************************)

Require Import basis.
Require Import part1.
Require Import part2.
Require Import part3.
Require Import affinity.

Parameter Unort : Line -> Line -> Prop.

Axiom O1 : (l, m: Line) (ConLn l m) \/ (Unort l m).

Axiom O2 : (Separating Line [l, m: Line] (ConLn l m) /\ (Unort l m)).

Definition Ort :=  (Negation Line Unort).

Axiom constructed_orthogonal :
   (l: Line) (a: Point) {l': Line | (Ort l' l) /\ (Incident a l')}.

Definition ort : Line -> Point -> Line.
Proof.
Intros l a; Elim (constructed_orthogonal l a).
Intros x H'; Exact x.
Defined.

Axiom constructive_uniqueness_for_orthogonals :
   (l, m, n: Line) (a: Point) (DiLn l m) ->
    ((Apart a l) \/ (Apart a m)) \/ ((Unort l n) \/ (Unort m n)).

Theorem O3_i : (l: Line) (a: Point) (Ort (ort l a) l).
Proof.
Intros l a.
Unfold 1 ort.
Elim (constructed_orthogonal l a).
Simpl; Tauto.
Qed.

Theorem O3_ii : (l: Line) (a: Point) (Incident a (ort l a)).
Proof.
Intros l a.
Unfold 1 ort.
Elim (constructed_orthogonal l a).
Simpl; Tauto.
Qed.
Hint O3_i O3_ii.

Theorem O4' :
  (l, m, n: Line)
  (a: Point) ((Incident a l) /\ (Incident a m)) /\ ((Ort l n) /\ (Ort m n)) ->
   (EqLn l m).
Proof.
Unfold Incident Ort EqLn Negation.
Intros l m n a.
Generalize (constructive_uniqueness_for_orthogonals l m n a).
Tauto.
Qed.

Theorem Uniqueness_of_orthogonality :
  (l, m: Line) (a: Point) (Incident a l) /\ (Ort l m) -> (EqLn l (ort m a)).
Proof.
Intros l m a H'; Elim H'; Intros H'0 H'1; Clear H'.
Apply O4' with n := m a := a; Auto.
Qed.

Theorem Unort_reflexive : (Reflexive Line Unort).
Proof.
Red.
Intro l.
Generalize (O1 l l); Intro H'.
Elim H'; [Intro H'0; Clear H' | Trivial].
Elim apart_con.
Intro H'; Elim (H' l); Auto.
Qed.

Theorem cmp_unort_con :
  (l, m, n: Line) (Unort l m) -> (ConLn l n) \/ (Unort m n).
Proof.
Intros l m n H'.
Generalize (O1 m n); Intro H'1.
Elim H'1; [Intro H'0; Clear H'1 | Auto].
Elim apart_con.
Intros H'1 H'2; Red in H'2.
LApply (H'2 m n l); [Intro H'6 | Assumption].
Elim H'6; [Intro H'3; Clear H'6 | Auto].
Generalize O2.
Intro H'4; Red in H'4.
LApply (H'4 l m n); Intuition.
Qed.

Theorem cmp_unort_diln :
  (l, m, n: Line) (Unort l m) -> (DiLn l n) \/ (Unort m n).
Proof.
Intros l m n H'.
LApply (cmp_unort_con l m n); [Intro H'3 | Assumption]; Auto.
Intuition.
Qed.

Theorem Unort_symmetric : (Symmetric Line Unort).
Proof.
Unfold 1 Symmetric.
Intros l m H'.
LApply (cmp_unort_con l m l);
 [Intro H'3; Elim H'3; [Intro H'4; Clear H'3 | Trivial] | Idtac]; Auto.
Cut (Irreflexive Line ConLn); Auto.
Intro H'0; Red in H'0.
Elim (H'0 l); Auto.
Elim apart_con; Auto.
Qed.

Theorem thm8_6 : (l, m, n: Line) (ConLn l m) -> (Unort l n) \/ (Unort m n).
Proof.
Intros l m n H'.
LApply (Convergent_imp_distinct l m); [Intro H'2 | Assumption].
LApply (constructive_uniqueness_for_orthogonals l m n (pt (Twol l m H')));
 [Intro H'5 | Assumption].
Elim H'5; [Intro H'0; Clear H'5 | Trivial].
Generalize (inc_pt1 (Twol l m H')); Generalize (inc_pt2 (Twol l m H')).
Simpl; Unfold Incident Negation; Tauto.
Qed.

Definition Oblique : (Relation Line) :=
   [l, m: Line] (ConLn l m) /\ (Unort l m).

Theorem apart_obl : (Apartness Line Oblique).
Proof.
Apply Definition_of_apartness.
Unfold Irreflexive Negation Oblique.
Intro l; Red; Intro H'; Elim H'; Intros H'0 H'1; Clear H'.
Elim apart_con.
Intros H' H'2; Apply (H' l); Assumption.
Unfold Oblique; Exact O2.
Qed.

Theorem ort_ort_like_par_i : (l: Line) (a: Point) (Incident a (ort (ort l a) a)).
Proof.
Auto.
Qed.

Theorem thm8_8 :
  (l: Line) (a, b: Point) (Incident b (ort l a)) -> (EqLn (ort l a) (ort l b)).
Proof.
Intros l a b H'.
Apply O4' with n := l a := b; Auto.
Qed.

Section delicate.
Variable a: Point.
Variable l: Line.
Variable H : (ConLn l (ort (ort l a) a)).

Local t : Twolines :=  (Twol l (ort (ort l a) a) H).

Local b : Point :=  (pt t).

Theorem thm8_9_aux : False.
Proof.
Generalize (inc_pt1 t); Intro H'0; Simpl in H'0.
Generalize (inc_pt2 t); Intro H'1; Simpl in H'1.
LApply (Convergent_imp_distinct l (ort (ort l a) a)); [Intro H'4 | Assumption].
LApply (thm8_8 (ort l a) a b); [Intro H'6 | Assumption].
LApply (cong_eqln_diln l (ort (ort l a) a) (ort (ort l a) b));
 [Intro H'7; LApply H'7; [Intro H'8; Clear H'7 | Clear H'7] | Idtac]; Auto.
LApply (constructive_uniqueness_for_orthogonals l (ort (ort l a) b) (ort l a) b);
 [Intro H'9 | Assumption].
Elim H'9;
 [Intro H'2; Elim H'2; [Intro H'3; Clear H'2 H'9 | Intro H'3; Clear H'2 H'9] |
   Intro H'2; Clear H'9].
Unfold 1 b in H'3; Elim H'0; Auto.
Elim (O3_ii (ort l a) b); Auto.
Generalize Unort_symmetric.
Intro H'; Red in H'.
Elim H'2; (Intro H'3; Clear H'2).
Elim (O3_i l a); Auto.
Elim (O3_i (ort l a) b); Auto.
Qed.

End delicate.

Theorem thm8_9 : (l: Line) (a: Point) (Par l (ort (ort l a) a)).
Proof.
Intros l a; Unfold Par Negation.
Red; Intro H'; Apply (thm8_9_aux a l); Trivial.
Qed.
Hint thm8_9.

Theorem thm8_10 : (l: Line) (a: Point) (EqLn (par l a) (ort (ort l a) a)).
Proof.
Intros l a; Apply sym_EqLn; Auto.
Qed.



