(****************************************************************************)
(*                                                                          *)
(*                           Group Theory in Coq                            *)
(*                                                                          *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*									    *)
(*			         Gilles Kahn 				    *)
(*									    *)
(*                                  INRIA                                   *)
(*                             Sophia-Antipolis                             *)
(*									    *)
(*				 January 1996				    *)
(*                                                                          *)
(****************************************************************************)

Require Import Ensembles.
Require Import Laws.
Require Import Group_definitions.
Require Import gr.
Parameter U: Type.
Parameter Gr: (Group U).

Local G: (Ensemble U) := (G_ U Gr).

Local star: U ->U ->U := (star_ U Gr).

Local inv: U ->U := (inv_ U Gr).

Local e: U := (e_ U Gr).

Definition G0': (a, b:U) (In U G a) ->(In U G b) ->(In U G (star a b)) :=
   (G0 U Gr).

Definition G1': (a, b, c:U)(star a (star b c)) == (star (star a b) c) :=
   (G1 U Gr).

Definition G2a': (In U G e) := (G2a U Gr).

Definition G2b': (a:U)(star e a) ==a := (G2b U Gr).

Definition G2c': (a:U)(star a e) ==a := (G2c U Gr).

Definition G3a': (a:U) (In U G a) ->(In U G (inv a)) := (G3a U Gr).

Definition G3b': (a:U)(star a (inv a)) ==e := (G3b U Gr).

Definition G3c': (a:U)(star (inv a) a) ==e := (G3c U Gr).
Hint G1'.
Hint G2a' G2b' G2c'.
Hint G3a' G3b' G3c'.
Hint G0'.

Definition triv1': (a, b:U)(star (inv a) (star a b)) ==b := (triv1 U Gr).

Definition triv2': (a, b:U)(star (star b a) (inv a)) ==b := (triv2 U Gr).

Definition resolve': (a, b:U) (star b a) ==e ->b == (inv a) := (resolve U Gr).

Definition self_inv': e == (inv e) := (self_inv U Gr).

Definition inv_star': (a, b:U)(star (inv b) (inv a)) == (inv (star a b)) :=
   (inv_star U Gr).

Definition cancellation': (a, b:U) (star a b) ==a ->b ==e :=
   (cancellation U Gr).

Definition inv_involution': (a:U)a == (inv (inv a)) := (inv_involution U Gr).
Hint triv1' triv2' inv_star' resolve' inv_involution'.

Section Elements.
Variable H: (Group U).
Variable sub: (subgroup U H Gr).

Lemma l1: (Included U (G_ U H) G).
Proof.
Elim sub; Auto.
Qed.
Hint l1.

Lemma eH_in_G: (In U G (e_ U H)).
Proof.
Elim sub.
Elim H; Auto.
Qed.
Hint eH_in_G.

Lemma starH_is_star: (star_ U H) ==star.
Proof.
Elim sub; Auto.
Qed.
Hint starH_is_star.

Lemma eh_is_e: (e_ U H) ==e.
Proof.
Apply cancellation' with a := (e_ U H).
Rewrite <- starH_is_star; Auto.
Qed.
Hint eh_is_e.

Theorem invH_is_inv: (a:U) (In U (G_ U H) a) ->(inv_ U H a) == (inv a).
Proof.
Intros a H'.
Apply resolve'; Auto.
Rewrite <- starH_is_star.
Rewrite <- eh_is_e.
Generalize H'; Clear H'.
Elim H; Auto.
Qed.

Theorem Subgroup_inhabited: (Inhabited U (G_ U H)).
Proof.
Apply Inhabited_intro with x := (e_ U H); Auto.
Qed.

Theorem star_endo: (endo_operation U (G_ U H) star).
Proof.
Rewrite <- starH_is_star; Auto.
Qed.

Theorem inv_endo: (endo_function U (G_ U H) inv).
Proof.
Red; Intros a H'; Rewrite <- (invH_is_inv a); Auto.
Qed.

End Elements.

Section Premier.

Variable H: (Ensemble U).
Variable witness: U.
Variable inhabited: (In U H witness).
Variable subset: (Included U H G).
Variable hstar: (endo_operation U H star).
Variable hinv: (endo_function U H inv).
Hint inhabited subset hstar hinv.

Local assoc: (associative U star).
Proof.
Auto.
Qed.
Hint assoc.

Local eH: U := (star (inv witness) witness).

Local eH_in_H: (In U H eH).
Proof.
Unfold 1 eH; Auto.
Qed.

Local eH_left_neutral: (left_neutral U star eH).
Proof.
Unfold eH left_neutral.
Rewrite (G3c' witness); Auto.
Qed.

Local eH_right_neutral: (right_neutral U star eH).
Proof.
Unfold eH left_neutral.
Rewrite (G3c' witness); Auto.
Qed.

Local inv_left_inverse: (left_inverse U star inv eH).
Proof.
Unfold eH left_inverse.
Intro x.
Rewrite (G3c' x); Auto.
Qed.

Local inv_right_inverse: (right_inverse U star inv eH).
Proof.
Unfold eH right_inverse.
Intro x.
Rewrite (G3b' x); Auto.
Qed.

Local GrH: (Group U) :=
   (group
      U
      H
      star
      inv
      eH
      hstar
      assoc
      eH_in_H
      eH_left_neutral
      eH_right_neutral
      hinv
      inv_right_inverse
      inv_left_inverse).
Hint Definition_of_subgroup.

Theorem T_1_6_2: (Setsubgroup U H Gr).
Proof.
Unfold 1 Setsubgroup; Simpl.
Exists GrH.
Unfold GrH; Simpl; Auto.
Qed.

End Premier.

Require Import Zbase.
Require Import Z_succ_pred.
Require Import Zadd.

Definition exp: Z ->U ->U.
Proof.
Intros n a.
Elim n; Clear n.
Exact e.
Intro n; Elim n; Clear n.
Exact a.
Intros n valrec.
Exact (star a valrec).
Intro n; Elim n; Clear n.
Exact (inv a).
Intros n valrec.
Exact (star (inv a) valrec).
Defined.

Theorem exp_endo: (a:U) (n:Z) (In U G a) -> (In U G (exp n a)).
Proof.
Intros a n; Elim n; Simpl; Auto.
Intro n0; Elim n0; Simpl; Auto.
Intro n0; Elim n0; Simpl; Auto.
Qed.
Hint exp_endo.

Lemma exp_unfold_pos:
  (a:U) (n:nat) (In U G a) -> (exp (pos (S n)) a) == (star a (exp (pos n) a)).
Proof.
Auto.
Qed.

Lemma exp_unfold_neg:
  (a:U) (n:nat) (In U G a) ->
        (exp (neg (S n)) a) == (star (inv a) (exp (neg n) a)).
Proof.
Auto.
Qed.

Lemma exp_l1:
  (a:U) (n:nat) (In U G a) -> (star a (exp (neg (S n)) a)) == (exp (neg n) a).
Proof.
Intros a n H'; Try Assumption.
Rewrite (exp_unfold_neg a); Trivial.
Rewrite (inv_involution' a); Trivial.
Cut (inv (inv (inv a))) == (inv a).
Intro H'0; Rewrite H'0; Apply triv1'; Auto.
Rewrite <- (inv_involution' a); Auto.
Qed.
Hint exp_l1.

Lemma exp_l2:
  (a:U) (n:Z) (In U G a) -> (star a (exp n a)) == (exp (succZ n) a).
Proof.
Intros a n; Elim n; Auto.
Intro n0; Elim n0; Auto.
Simpl; Auto.
Intros n1 H' H'0.
Rewrite (exp_l1 a n1); Auto.
Qed.

Lemma exp_l2':
  (a:U) (n:Z) (In U G a) -> (star (inv a) (exp n a)) == (exp (predZ n) a).
Proof.
Intros a n; Elim n; Auto.
Intro n0; Elim n0; Auto.
Intros n1 H' H'0.
Rewrite (exp_unfold_pos a n1); Trivial.
Qed.
Hint exp_l2 exp_l2' exp_unfold_pos exp_unfold_neg.
Immediate sym_eqT.

Theorem add_exponents:
  (a:U) (m, n:Z) (In U G a) -> (star (exp m a) (exp n a)) == (exp (addZ m n) a).
Proof.
Intros a m; Elim m; Auto.
Intro n; Elim n; Auto.
Intros n0 H' n1 H'0.
Rewrite (tech_add_pos_succZ n0 n1).
Rewrite <- (exp_l2 a (addZ (pos n0) n1)); Trivial.
Rewrite (exp_unfold_pos a n0); Trivial.
Rewrite <- (H' n1); Trivial.
Auto.
Intro n; Elim n.
Simpl.
Intros n0 H'.
Apply exp_l2'; Auto.
Intros n0 H' n1 H'0.
Rewrite (tech_add_neg_predZ n0 n1).
Rewrite <- (exp_l2' a (addZ (neg n0) n1)); Trivial.
Rewrite <- (H' n1); Trivial.
Rewrite (exp_unfold_neg a n0); Trivial.
Auto.
Qed.

Lemma exp_commut1:
  (a:U) (m:Z) (In U G a) ->(star (exp m a) a) == (star a (exp m a)).
Proof.
Intros a m H'.
Change (star (exp m a) (exp IZ a)) == (star (exp IZ a) (exp m a)).
Rewrite (add_exponents a m IZ); Trivial.
Rewrite (add_exponents a IZ m); Trivial.
Rewrite (addZ_commutativity IZ m); Trivial.
Qed.

Lemma tech_opp_pos_negZ1: (n:nat)(oppZ (pos n)) =(neg n).
Proof.
Intro n; Elim n; Auto.
Qed.

Lemma tech_opp_pos_negZ2: (n:nat)(oppZ (neg n)) =(pos n).
Proof.
Intro n; Elim n; Auto.
Qed.

Theorem change_exponent_sign:
  (a:U) (m:Z) (In U G a) ->(inv (exp m a)) == (exp (oppZ m) a).
Proof.
Intros a m; Elim m.
Simpl; Auto.
Simpl; Auto.
Intro H'; Symmetry; Auto.
Intros n H'.
Rewrite (tech_opp_pos_negZ1 n).
Elim n; Auto.
Intros n0 H'0.
Rewrite (exp_unfold_pos a n0); Trivial.
Rewrite (exp_unfold_neg a n0); Trivial.
Rewrite <- (exp_commut1 a (pos n0)); Trivial.
Rewrite <- (inv_star' (exp (pos n0) a) a); Auto.
Rewrite H'0; Trivial.
Intros n H'.
Rewrite (tech_opp_pos_negZ2 n).
Elim n.
Simpl; Symmetry; Auto.
Intros n0 H'0.
Rewrite (exp_unfold_pos a n0); Trivial.
Rewrite (exp_unfold_neg a n0); Trivial.
Rewrite <- (exp_commut1 a (pos n0)); Trivial.
Rewrite <- (inv_star' (inv a) (exp (neg n0) a)); Auto.
Rewrite H'0; Trivial.
Rewrite <- (inv_involution' a); Trivial.
Qed.

Inductive powers[a:U]: (Ensemble U) :=
   In_powers: (m:Z) (x:U) x == (exp m a) -> (In U (powers a) x).

Theorem powers_of_one_element: (a:U) (In U G a) -> (Setsubgroup U (powers a) Gr).
Proof.
Intros a H'.
Apply T_1_6_2 with witness := a.
Apply In_powers with m := IZ; Auto.
Red.
Intros x H'0; Elim H'0.
Intros m x0 H'1; Rewrite H'1; Auto.
Red.
Intros x y H'0; Elim H'0.
Intros m x0 H'1 H'2; Elim H'2.
Intros m0 x1 H'3; Rewrite H'3.
Rewrite H'1.
Rewrite (add_exponents a m m0); Trivial.
Apply In_powers with m := (addZ m m0); Trivial.
Red.
Intros x H'0; Elim H'0.
Intros m x0 H'1.
Apply In_powers with m := (oppZ m); Trivial.
Rewrite H'1.
Apply change_exponent_sign; Trivial.
Qed.

Section Second.

Variable H: (Ensemble U).
Variable witness: U.
Variable h0: (In U H witness).
Variable h1: (Included U H G).
Variable h2: (a, b:U) (In U H a) ->(In U H b) ->(In U H (star a (inv b))).

Local eH: U := (star witness (inv witness)).

Theorem T_1_6_3: (Setsubgroup U H Gr).
Proof.
Cut (In U H eH).
Intro H'.
Apply T_1_6_2 with witness := witness; Auto.
Red; Intros a b H'0 H'1; Try Assumption.
LApply (h2 eH b);
 [Intro H'4; LApply H'4; [Intro H'5; Try Exact H'5; Clear H'4 | Clear H'4] |
   Idtac]; Auto.
Unfold 1 eH in H'5.
Generalize H'5; Clear H'5.
Rewrite (G3b' witness).
Rewrite (G2b' (inv b)).
Intro H'4.
LApply (h2 a (inv b));
 [Intro H'5; LApply H'5; [Intro H'6; Generalize H'6; Clear H'5 | Clear H'5] |
   Idtac]; Auto.
Rewrite <- (inv_involution' b); Auto.
Red; Intros a H'0; Try Assumption.
Rewrite <- (G2b' (inv a)).
Apply h2; Auto.
Unfold 1 eH in H'.
Rewrite <- (G3b' witness); Auto.
Unfold 1 eH; Auto.
Qed.

End Second.

Theorem Ex1: (Setsubgroup U (Singleton U e) Gr).
Proof.
Apply T_1_6_2 with witness := e; Auto.
Red; Intros x H'; Elim H'; Auto.
Red; Intros a b H' H'0.
Elim H'.
Elim H'0.
Rewrite (G2c' e); Auto.
Red; Intros a H'; Elim H'.
Rewrite <- (resolve' e e); Auto.
Qed.

Theorem Ex2: (Setsubgroup U (Singleton U e) Gr).
Proof.
Apply T_1_6_3 with witness := e; Auto.
Red.
Intros x H'; Elim H'; Auto.
Intros a b H'; Elim H'.
Intro H'0; Elim H'0.
Rewrite (G3b' e); Auto.
Qed.

Lemma Ex3: (n:Z) (exp n e) == e.
Proof.
Intro n; Elim n; Auto.
Intro n0; Elim n0; Auto.
Intros n1 H'.
Rewrite (exp_unfold_pos e n1); Auto.
Rewrite H'; Auto.
Intro n0; Elim n0; Auto.
Simpl.
Symmetry; Auto.
Intros n1 H'.
Rewrite (exp_unfold_neg e n1); Auto.
Rewrite H'; Auto.
Qed.

Lemma Ex4: (powers e) == (Singleton U e).
Proof.
Apply Extensionality_Ensembles; Split; Red; Auto.
Intros x H'; Elim H'.
Intros m x0 H'0; Rewrite H'0.
Rewrite (Ex3 m); Auto.
Intros x H'; Elim H'.
Apply In_powers with m := IZ; Auto.
Qed.

Theorem Ex5: (Setsubgroup U (Singleton U e) Gr).
Proof.
Rewrite <- Ex4.
Apply powers_of_one_element; Auto.
Qed.

