(****************************************************************************)
(*                                                                          *)
(*                           Group Theory in Coq                            *)
(*                                                                          *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*									    *)
(*			         Gilles Kahn 				    *)
(*									    *)
(*                                  INRIA                                   *)
(*                             Sophia-Antipolis                             *)
(*									    *)
(*				 January 1996				    *)
(*                                                                          *)
(****************************************************************************)

Require Import Ensembles.
Require Import Zbase.
Require Import Z_succ_pred.
Require Import Zadd.
Require Import Zle.
Require Import Classical_Prop.
Require Import Relations_1.
Require Import Relations_1_facts.
Require Import Partial_Order.
Require Import Cpo.
Require Import Powerset.
Require Import Powerset_facts.
Require Import Gt.
Require Import Lt.
Require Import Compare.
Require Import Arith.
Require Import Finite_sets.
Require Import Finite_sets_facts.
Require Import Image.
Require Import Infinite_sets.
Require Import Integers.
Require Import Laws.
Require Import Group_definitions.
Require Export gr.
Require Export g1.

Section Cinq.
Variable H: (Ensemble U).
Variable H_inhabited: (Inhabited U H).
Variable H_included_in_G: (Included U H G).
Variable stability: (endo_operation U H star).
Variable H_Finite: (Finite U H).

Local h: (x, y: U) (In U H x) -> (In U H y) -> (In U H (star x y)).
Proof.
Auto.
Qed.
Hint h.

Definition phi : U -> nat -> U :=  [a: U] [n: nat] (exp (pos n) a).

Lemma phi_unfold:
  (a: U) (n: nat) (In U G a) ->  (phi a (S n)) == (star a (phi a n)).
Proof.
Unfold phi; Auto.
Qed.

Lemma positive_powers: (a: U) (n: nat) (In U H a) -> (In U H (phi a n)).
Proof.
Intros a n; Elim n; Auto.
Intros n0 H' H'0.
Rewrite (phi_unfold a n0); Auto.
Qed.

Lemma tech_exp:
  (a: U) (n: nat) (In U G a) -> (star (phi a n) a) == (phi a (S n)).
Proof.
Intros a n; Elim n; Auto.
Intros n0 H' H'0.
Rewrite (phi_unfold a n0); Auto.
Rewrite <- (G1' a (phi a n0) a).
Rewrite H'; Auto.
Qed.

Lemma tech_exp': (n: nat) (phi e n) == e.
Proof.
Intro n; Elim n; Auto.
Intros n0 H'.
Rewrite <- (tech_exp e n0); Auto.
Rewrite H'; Auto.
Qed.

Lemma phi_exp:
  (a: U) (n: nat) (m: nat) (In U G a) ->
   (star (phi a n) (phi a m)) == (phi a (S (plus n m))).
Proof.
Unfold phi.
Intros a n m H'.
Rewrite (add_exponents a (pos n) (pos m)); Trivial.
Rewrite (tech_add_pos_posZ n m); Trivial.
Qed.

Lemma powers_repeat:
  (a: U) (n, m: nat) (In U G a) -> (phi a n) == (phi a (S (S (plus n m)))) ->
   (phi a m) == (inv a).
Proof.
Intros a n m H' H'0.
Apply resolve'.
Apply cancellation' with a := (phi a n).
Rewrite (tech_exp a m); Trivial.
Rewrite (phi_exp a n (S m)); Trivial.
Rewrite <- (plus_n_Sm n m); Auto.
Qed.

Definition psi: U -> Nat -> U.
Proof.
Intros a H'0; Elim H'0; Intro n; Exact (phi a n).
Defined.

Lemma psi_not_inj: (a: U) (In U H a) -> ~ (injective Nat U (psi a)).
Proof.
Intros a H'; Try Assumption.
Apply Pigeonhole_bis with A := Integers.
Exact Integers_infinite.
Apply Finite_downward_closed with A := H; Auto.
Red.
Intros x H'0; Elim H'0.
Intro x0; Elim x0.
Intros n H'1 y H'2; Rewrite H'2.
Unfold 1 psi; Simpl.
Apply positive_powers; Auto.
Qed.

Theorem remaining:
  (a: U) (In U H a) ->
   (Ex [r: nat] (Ex [m: nat] (phi a r) == (phi a (S (S (plus r m)))))).
Proof.
Intros a ainH.
LApply (not_injective_elim Nat U (psi a)); [Intro H'2 | Apply psi_not_inj; Auto].
Elim H'2; Clear H'2.
Intros x H'.
Elim H'; Clear H'.
Intros x0 H'0; Elim H'0; Clear H'0.
Elim x; Elim x0.
Intros n n0 H'0 H'1.
Unfold psi in H'0; Simpl in H'0.
Cut ~ n0 = n.
2:Red; Intro H'4; Apply H'1; Rewrite <- H'4; Auto.
Intro H'.
Elim (nat_total_order n0 n).
Clear H'1 H'.
Intro H'.
LApply (discrete_nat n0 n); [Intro H'4 | Assumption].
Elim H'4; Intro H'1; Clear H'4.
Exists n0; Exists O.
Generalize H'0.
Rewrite <- H'1.
Rewrite <- (tech_exp a n0).
Intro H'3.
LApply (sym_eqT U (phi a n0) (star (phi a n0) a)); [Intro H'9 | Assumption].
LApply (cancellation' (phi a n0) a); [Intro H'8 | Assumption].
Rewrite H'8.
Rewrite (tech_exp' n0).
Rewrite (tech_exp' (S (S (plus n0 O)))); Auto.
Auto.
Elim H'1; Intros r E; Clear H'1.
Exists n0; Exists r.
Rewrite <- E; Auto.
Clear H'1 H'.
Intro H'.
LApply (discrete_nat n n0); [Intro H'4 | Assumption].
Elim H'4; Intro H'1; Clear H'4.
Exists n; Exists O.
Generalize H'0.
Rewrite <- H'1.
Rewrite <- (tech_exp a n).
Intro H'3.
LApply (cancellation' (phi a n) a); [Intro H'8 | Assumption].
Rewrite H'8.
Rewrite (tech_exp' n).
Rewrite (tech_exp' (S (S (plus n O)))); Auto.
Auto.
Elim H'1; Intros r E; Clear H'1.
Exists n; Exists r.
Rewrite <- E; Auto.
Auto.
Qed.

Theorem T_1_6_4: (Setsubgroup U H Gr).
Proof.
Elim H_inhabited.
Intros witness inH.
Apply T_1_6_2 with witness := witness; Trivial.
Red.
Intros a H'.
Cut (Ex [n: nat] (inv a) == (phi a n)).
Intro H'0; Elim H'0; Intros n E; Rewrite E; Clear H'0.
Apply positive_powers; Trivial.
Cut (Ex [r: nat] (ex nat [m: nat] (phi a r) == (phi a (S (S (plus r m)))))).
Intro H'0; Elim H'0; Intros r E; Elim E; Intros m E0; Try Exact E0; Clear E H'0.
Cut (inv a) == (phi a m).
Intro H'0; Rewrite H'0.
Exists m; Trivial.
Symmetry.
Apply powers_repeat with n := r; Trivial.
Apply H_included_in_G; Auto.
Apply remaining; Auto.
Qed.

End Cinq.



