(****************************************************************************)
(*                                                                          *)
(*                  Binary Decision Trees (BDTs)                            *)
(*               MacMillan's "Proof" of canonicity of BDDs                  *)
(*                                                                          *)
(****************************************************************************)
(*                                  BDTs.v                                  *)
(****************************************************************************)

Require Boolean_functions.
Require Vars.

Inductive BDT [n:nat] : Set :=
    zero : (BDT n)
 |  one :  (BDT n)
 |  node : (Var n) -> (BDT n) -> (BDT n) -> (BDT n).

Definition Dimension : (n:nat)(BDT n)->nat :=
  [n:nat][bdt:(BDT n)]<nat>Case bdt of
  (* zero  *)            O
  (* one   *)            O
  (* (node i bl bh) *) [xi:(Var n)][bl,bh:(BDT n)](Order n xi) end.

Inductive dimension [n:nat] : (BDT n)->nat->Prop :=
   dim_zero: (dimension n (zero n) O)
 | dim_one : (dimension n (one n) O)
 | dim_node: (xi:(Var n))(bl,bh:(BDT n))
             (dimension n (node n xi bl bh) (Order n xi)).
Hint dim_zero dim_one dim_node.


Lemma dim_functional : (n,d:nat)(bdt:(BDT n))(dimension n bdt d) ->
           d = (Dimension n bdt).
Proof.
Induction 1; Simpl; Auto.
Qed.
Hint dim_functional.


Inductive node_decomposition [n:nat; bdt:(BDT n); d:nat] : Prop :=
   node_exist : (x:(Var n))(bl,bh:(BDT n))
          (Order n x) = d -> (node n x bl bh) = bdt
                          -> (node_decomposition n bdt d).

Local dim_inv : (n:nat)(BDT n)->nat->Prop :=
   [n:nat][bdt:(BDT n)][d:nat]<Prop>Case d of
      (* O *)  ((zero n)=bdt \/ (one n)=bdt)
      (* S k *) [k:nat](node_decomposition n bdt d) end.

Lemma p_dim_inv : (n,d:nat)(bdt:(BDT n))(dimension n bdt d)->(dim_inv n bdt d).
Proof.
Induction 1; Simpl; Auto.
Intros.
Elim (Order_S n xi).
Intros x E; Rewrite -> E; Simpl.
Apply node_exist with xi bl bh; Auto.
Qed.
Hint p_dim_inv.

Lemma dim_Dim : (n:nat)(bdt:(BDT n))(dimension n bdt (Dimension n bdt)).
Proof.
Intro n.
Induction bdt; Simpl; Auto.
Qed.
Hint dim_Dim.

Lemma Exist_dim : (n:nat)(bdt:(BDT n))(Ex [d:nat](dimension n bdt d)).
Proof.
Intros n bdt.
Exists (Dimension n bdt); Auto.
Qed.

Lemma Dim_le_n : (n,d:nat)(bdt:(BDT n))(le (Dimension n bdt) n).
Proof.
Induction bdt; Simpl; Auto.
Qed.
Hint Dim_le_n.


Inductive BDT_eq [n:nat] : (BDT n)->(BDT n)->Prop :=
    BDT_eq_0: (BDT_eq n (zero n) (zero n))
  | BDT_eq_1: (BDT_eq n (one n)  (one n))
  | BDT_eq_node: (bl1,bl2,bh1,bh2:(BDT n))(xi1,xi2:(Var n))
            (BDT_eq n bl1 bl2) -> (BDT_eq n bh1 bh2) -> (eq_Var n xi1 xi2)
         -> (BDT_eq n (node n xi1 bl1 bh1) (node n xi2 bl2 bh2)).
Hint BDT_eq_0 BDT_eq_1 BDT_eq_node.

Definition inv_BDT_eq : (n:nat)(BDT n)->(BDT n)->Prop :=
  [n:nat][bdt1:(BDT n)]<(BDT n)->Prop>Case bdt1 of
(* zero *)           [bdt2:(BDT n)]<Prop>Case bdt2 of
                      (* zero *)     True
                      (* one  *)     False
                      (* node *) [x2:(Var n)][b2l,b2h:(BDT n)]False end
(* one  *)           [bdt2:(BDT n)]<Prop>Case bdt2 of
                      (* zero *)     False
                      (* one  *)     True
                      (* node *) [x2:(Var n)][b2l,b2h:(BDT n)]False end
(* node *) [x1:(Var n)][b1l,b1h:(BDT n)]
                     [bdt2:(BDT n)]<Prop>Case bdt2 of
                      (* zero *)     False
                      (* one  *)     False
                      (* node *) [x2:(Var n)][b2l,b2h:(BDT n)]
      ((BDT_eq n b1l b2l) /\ (BDT_eq n b1h b2h)) /\ (eq_Var n x1 x2) end 
                                      end.

Lemma p_inv_BDT_eq : (n:nat)(bdt1,bdt2:(BDT n))(BDT_eq n bdt1 bdt2) ->
                     (inv_BDT_eq n bdt1 bdt2). 
Proof.
Induction 1; Simpl; Auto.
Qed.
Hint p_inv_BDT_eq.

Lemma ne_zero_one : (n:nat)~(BDT_eq n (zero n) (one n)).
Proof.
Unfold not.
Intros n H_absurde.
Change (inv_BDT_eq n (zero n) (one n)); Auto.
Qed.
Hint ne_zero_one.

Lemma ne_zero_node : (n:nat)(x2:(Var n))
      (b2l,b2h:(BDT n))~(BDT_eq n (zero n) (node n x2 b2l b2h)).
Proof.
Unfold not.
Intros n x2 b2l b2h H_absurde.
Change (inv_BDT_eq n (zero n) (node n x2 b2l b2h)); Auto.
Qed.
Hint ne_zero_node.

Lemma ne_one_zero : (n:nat)~(BDT_eq n (one n) (zero n)).
Proof.
Unfold not.
Intros n H_absurde.
Change (inv_BDT_eq n (one n) (zero n)); Auto.
Qed.
Hint ne_one_zero.

Lemma ne_one_node : (n:nat)(x2:(Var n))(b2l,b2h:(BDT n))
      ~(BDT_eq n (one n) (node n x2 b2l b2h)).
Proof.
Unfold not.
Intros n x2 b2l b2h H_absurde.
Change (inv_BDT_eq n (one n) (node n x2 b2l b2h)); Auto.
Qed.
Hint ne_one_node.

Lemma inv_BDT_eq_node : (n:nat)(x1,x2:(Var n))
       (b1l,b1h,b2l,b2h:(BDT n))
         (BDT_eq n (node n x1 b1l b1h) (node n x2 b2l b2h)) ->
              ((BDT_eq n b1l b2l) /\ (BDT_eq n b1h b2h)) /\ (eq_Var n x1 x2).
Proof.
Intros n x1 x2 b1l b1h b2l b2h node1_eq_node2.
Change (inv_BDT_eq n (node n x1 b1l b1h) (node n x2 b2l b2h)); Auto.
Qed.

Lemma ne_node_zero : (n:nat)(x1:(Var n))(b1l,b1h:(BDT n))
      ~(BDT_eq n (node n x1 b1l b1h) (zero n)).
Proof.
Unfold not.
Intros n x1 b1l b1h H_absurde.
Change (inv_BDT_eq n (node n x1 b1l b1h) (zero n)); Auto.
Qed.
Hint ne_node_zero.

Lemma ne_node_one : (n:nat)(x1:(Var n))(b1l,b1h:(BDT n))
     ~(BDT_eq n (node n x1 b1l b1h) (one n)).
Proof.
Unfold not.
Intros n x1 b1l b1h H_absurde.
Change (inv_BDT_eq n (node n x1 b1l b1h) (one n)); Auto.
Qed.
Hint ne_node_one.

Inductive OBDT [n:nat] : (BDT n)->Prop :=
      order_0: (OBDT n (zero n))
    | order_1: (OBDT n (one n))
    | order_node: (xi:(Var n))(bl,bh:(BDT n))
                  (OBDT n bl)->(OBDT n bh)->
                  (gt (Order n xi) (Dimension n bl))->
                  (gt (Order n xi) (Dimension n bh))->
                  (~(BDT_eq n bl bh))
                              ->(OBDT n (node n xi bl bh)).

Hint order_0 order_1 order_node.

Definition inv_OBDT_Order : (n:nat)(BDT n)->Prop :=
        [n:nat][bdt:(BDT n)]<Prop>Case bdt of
      (* zero *)          True
      (* one  *)          True
      (* node xi bl bh *) [xi:(Var n)][bl,bh:(BDT n)]
        (gt (Order n xi) (Dimension n bl)) 
     /\ (gt (Order n xi) (Dimension n bh)) end.

Lemma p_inv_OBDT_order :
  (n:nat)(bdt:(BDT n))(OBDT n bdt) -> (inv_OBDT_Order n bdt).
Proof.
Induction 1; Simpl; Auto.
Qed.
Hint p_inv_OBDT_order.

Definition inv_OBDT_eq : (n:nat)(BDT n)->Prop :=
        [n:nat][bdt:(BDT n)]<Prop>Case bdt of
      (* zero *)          True
      (* one  *)          True
      (* node xi bl bh *) [xi:(Var n)][bl,bh:(BDT n)] ~(BDT_eq n bl bh) end.

Lemma p_inv_OBDT_eq : (n:nat)(bdt:(BDT n))(OBDT n bdt) -> (inv_OBDT_eq n bdt).
Proof.
Induction 1; Simpl; Auto.
Qed.
Hint p_inv_OBDT_eq.

Definition inv_OBDT_bl : (n:nat)(BDT n)->Prop :=
        [n:nat][bdt:(BDT n)]<Prop>Case bdt of
      (* zero *)          True
      (* one  *)          True
      (* node xi bl bh *) [xi:(Var n)][bl,bh:(BDT n)](OBDT n bl) end.

Lemma p_inv_OBDT_bl : (n:nat)(bdt:(BDT n))(OBDT n bdt)->(inv_OBDT_bl n bdt).
Proof.
Induction 1;Simpl;Auto.
Qed.
Hint p_inv_OBDT_bl.


Definition inv_OBDT_bh:(n:nat)(BDT n)->Prop :=
        [n:nat][bdt:(BDT n)]<Prop>Case bdt of
      (* zero *)          True
      (* one  *)          True
      (* node xi bl bh *) [xi:(Var n)][bl,bh:(BDT n)](OBDT n bh) end.

Lemma p_inv_OBDT_bh : (n:nat)(bdt:(BDT n))(OBDT n bdt)->(inv_OBDT_bh n bdt).
Proof.
Induction 1;Simpl;Auto.
Qed.
Hint p_inv_OBDT_bh.

Lemma  Dim_node_Dim_bh : (n:nat)(bl,bh:(BDT n))(xi:(Var n))
      (OBDT n (node n xi bl bh)) ->
      (dn,dh:nat)(dimension n (node n xi bl bh) dn) -> 
      (dimension n bh dh) -> (gt dn dh).
Proof.
Intros n bl bh xi H_OBDT dn dh Def_dn Def_dh.
Replace dn with (Dimension n (node n xi bl bh)); Try Symmetry; Auto.
Replace dh with (Dimension n bh); Try Symmetry; Auto.
Simpl.
Cut (inv_OBDT_Order n (node n xi bl bh)); Auto.
Simpl; Induction 1; Auto.
Qed.

Lemma Dim_node_Dim_bl : (n:nat)(bl,bh:(BDT n))(xi:(Var n))
      (OBDT n (node n xi bl bh)) ->
      (dn,dl:nat)(dimension n (node n xi bl bh) dn) ->
      (dimension n bl dl) -> (gt dn dl).
Proof.
Intros n bl bh xi H_OBDT dn dl Def_dn Def_dl.
Replace dn with (Dimension n (node n xi bl bh)); Try Symmetry; Auto.
Replace dl with (Dimension n bl); Try Symmetry; Auto.
Simpl.
Cut (inv_OBDT_Order n (node n xi bl bh)); Auto.
Simpl; Induction 1; Auto.
Qed.


(********************         From BDT to BF             *******************)

Definition Fun : (n:nat)(BDT n)->(BF n) := 
    [n:nat][bdt:(BDT n)]<(BF n)>Match bdt with
(* zero  *)           [A:(Assign n)]false
(* one   *)           [A:(Assign n)]true
(* node xi bl bh *)   [xi:(Var n)]
		      [bl:(BDT n)][fbl:(BF n)][bh:(BDT n)][fbh:(BF n)]
                           [A:(Assign n)]<bool>Case (A xi) of
                                               (* true *)  (fbh A)
                                               (* false *) (fbl A) end end.

Lemma ne_Fzero_Fone : (n:nat)~(BF_eq n (Fun n (zero n)) (Fun n (one n))).
Proof.
Intro n.
Simpl.
Unfold not BF_eq.
Intro H_absurd.
Absurd false=true; Auto.
Apply H_absurd; Exact [x:(Var n)]true.
Qed.
Hint ne_Fzero_Fone.
 
Lemma ne_Fone_Fzero : (n:nat)~(BF_eq n (Fun n (one n)) (Fun n (zero n))).
Proof.
Intro n.
Unfold not;Intro H1.
Cut ~(BF_eq n (Fun n (zero n)) (Fun n (one n)));Auto.
Qed.
Hint ne_Fone_Fzero.

Lemma Restr_node_Restr_low : (n:nat)(x:(Var n))(bl,bh:(BDT n))
        (BF_eq n (Restr n (Fun n (node n x bl bh)) x false)
                 (Restr n (Fun n bl) x false)).
Proof.
Intros n x bl bh.
Unfold BF_eq.
Intro A.
Unfold 1 Restr; Simpl.
Unfold 1 Upd; Simpl.
Elim (eq_Var_dec n x x).
Intro x_eq_x.
Change (Restr n (Fun n bl) x false A) = (Restr n (Fun n bl) x false A); Auto.
Intro H_absurde.
Absurd (eq_Var n x x); Auto.
Qed.
Hint Restr_node_Restr_low.

Lemma Restr_node_Restr_high : (n:nat)(x:(Var n))(bl,bh:(BDT n))
        (BF_eq n (Restr n (Fun n (node n x bl bh)) x true)
                 (Restr n (Fun n bh) x true)).
Proof.
Intros n x bl bh.
Unfold BF_eq.
Intro A.
Unfold 1 Restr; Simpl.
Unfold 1 Upd; Simpl.
Elim (eq_Var_dec n x x).
Intro x_eq_x.
Change (Restr n (Fun n bh) x true A) = (Restr n (Fun n bh) x true A); Auto.
Intro H_absurde.
Absurd (eq_Var n x x);Auto.
Qed.
Hint Restr_node_Restr_high.

(*---------------------------------------------------------------------------*)
Lemma vars_sup_dim_non_dep :
  (n:nat)(bdt:(BDT n))(OBDT n bdt) ->        
  (x:(Var n))(gt (Order n x) (Dimension n bdt)) ->
  (b:bool)(BF_eq n (Restr n (Fun n bdt) x b) (Fun n bdt)).
(*---------------------------------------------------------------------------*)
Proof.
Cut (n,k:nat)(bdt:(BDT n))(OBDT n bdt) -> (le (Dimension n bdt) k) ->
    (x:(Var n))(gt (Order n x) (Dimension n bdt)) ->
    (b:bool)(BF_eq n (Restr n (Fun n bdt) x b) (Fun n bdt)).
Intros Coupure n bdt bdt_OBDT x x_gt_Dim_bdt b.
Apply Coupure with k:= n; Auto.
Induction k.
       (* k = O *)
Intros bdt bdt_OBDT Dim_O x gt_Orderx_Dim b.
Cut (dim_inv n bdt O).
Unfold dim_inv.
Do 2 Induction 1; Simpl; Unfold BF_eq; Auto.
(* Use : Intro ; Apply refl_equal
   Use : Intro ; Apply refl_equal et pas Restr_cst *)
Apply p_dim_inv.
ElimType (Dimension n bdt)=O; Try Apply sym_equal; Auto.
     (* k -> (S k) *)
Intros l Hl bdt H_OBDT H_dim x H_order_x b.
Elim (le_nSm_le_nm_or_eq l (Dimension n bdt)); Try Assumption;
   Intro H_dim_bdt; Auto. (*  Use : Apply Hl *)
Cut (node_decomposition n bdt (S l)).
Induction 1.
Intros xi bl bh Order_xi decomp_bdt.
Unfold BF_eq.
Intro A.
Elim decomp_bdt; Simpl.
Unfold Restr; Simpl.
ElimType (A xi)=((Upd n A x b) xi).
Elim (A xi).
(*---- (Fun n bh (Upd n A x b))=(Fun n bh A) ---*)
Change (Restr n (Fun n bh) x b A)=(Fun n bh A).
Cut (A:(Assign n))(Restr n (Fun n bh) x b A)=(Fun n bh A); Auto.
Change (BF_eq n (Restr n (Fun n bh) x b) (Fun n bh)).
Apply Hl.
Change (inv_OBDT_bh n (node n xi bl bh));Rewrite ->decomp_bdt; Auto.
Apply gt_S_le.
Replace (S l) with (Dimension n bdt).
Apply Dim_node_Dim_bh with n bl bh xi; Try Rewrite -> decomp_bdt; Auto.
Apply gt_trans with m:=(Dimension n bdt); Auto.
Apply Dim_node_Dim_bh with n bl bh xi; Try Rewrite -> decomp_bdt; Auto.
(*---- (Fun n bl (Upd n A x b))=(Fun n bl A) ----*)
Change (Restr n (Fun n bl) x b A) = (Fun n bl A).
Cut (A:(Assign n))(Restr n (Fun n bl) x b A) = (Fun n bl A); Auto.
Change (BF_eq n (Restr n (Fun n bl) x b) (Fun n bl)).
Apply Hl.
Change (inv_OBDT_bl n (node n xi bl bh)); Rewrite -> decomp_bdt; Auto.
Apply gt_S_le.
Replace (S l) with (Dimension n bdt).
Apply Dim_node_Dim_bl with n bl bh xi; Try Rewrite -> decomp_bdt; Auto.
Apply gt_trans with m:=(Dimension n bdt); Auto.
Apply Dim_node_Dim_bl with n bl bh xi; Try Rewrite -> decomp_bdt; Auto.
(*---- (A xi)=(Upd n A x b xi) -----*)
Symmetry; Apply Upd_xi_xj.
Unfold eq_Var not; Intro E.
Generalize H_order_x; Rewrite H_dim_bdt; Elim E; Rewrite Order_xi.
Intro; Apply (gt_antirefl (S l)); Trivial.
(*---- (node_decomposition n bdt (S l)) -----*)
Change (dim_inv n bdt (S l)).
Rewrite <- H_dim_bdt; Auto.
Qed.

Lemma vars_sup_dim_non_dep2 : (n:nat)(bdt:(BDT n))(OBDT n bdt) ->        
  (x:(Var n))(gt (Order n x) (Dimension n bdt)) -> ~(Dep_Var n (Fun n bdt) x).
Proof.
Intros n bdt On x H_Order_x.
Unfold not.
Intro H_absurde.
Cut ~(BF_eq n (Restr n (Fun n bdt) x true) 
              (Restr n (Fun n bdt) x false)); Auto.
Unfold not.
Intro H; Apply H.
Apply BF_eq_trans with f2:=(Fun n bdt).
Apply vars_sup_dim_non_dep; Auto.
Apply BF_eq_sym; Apply vars_sup_dim_non_dep; Auto.
Qed.

(*--------------------------------------------------------------------------*)
  Lemma BDT_iso_BDF_eq : (n:nat)(bdt1,bdt2:(BDT n))
 (BDT_eq n bdt1 bdt2)->(BF_eq n (Fun n bdt1) (Fun n bdt2)).
(*--------------------------------------------------------------------------*)
Proof.
Induction 1; Auto.
Intros bl1 bl2 bh1 bh2 xi1 xi2 bl1_iso_bl2 Fbl1_eq_Fbl2 
       bh1_iso_bh2 Fbh1_eq_Fbh2 xi1_eq_xi2.
Unfold BF_eq; Intro A; Simpl.
ElimType (A xi1)=(A xi2).
Unfold BF_eq in Fbh1_eq_Fbh2.
Unfold BF_eq in Fbl1_eq_Fbl2.
Elim (A xi1); Auto.
Pattern xi2; Apply Proof_irrelevance with n xi1; Auto.
Qed.
Hint BDT_iso_BDF_eq.

(*--------------------------------------------------------------------------*)
  Lemma Fun_eq_OBDT_eq: (n:nat)(bdt1,bdt2:(BDT n))
  (OBDT n bdt1) -> (OBDT n bdt2) -> 
  (BF_eq n (Fun n bdt1) (Fun n bdt2)) -> (BDT_eq n bdt1 bdt2).
(*--------------------------------------------------------------------------*)
Proof.
Intros n bdt1 bdt2 H H0 H1.
Cut  (n,k:nat)(bdt1,bdt2:(BDT n))(le (Dimension n bdt1) k)
    ->(le (Dimension n bdt2) k)->(OBDT n bdt1)->(OBDT n bdt2)
    ->(BF_eq n (Fun n bdt1) (Fun n bdt2))->(BDT_eq n bdt1 bdt2).
Intro Fun_eq_OBDT_eq_dimk.
Apply Fun_eq_OBDT_eq_dimk with k:=n; Auto.
(* Cut *)
Clear H1 H0 H bdt2 bdt1 n.
Intro n.
Induction k.
(*------ k= O ------*)
Intros bdt1 bdt2 H_dim1 H_dim2 H_OBDT1 H_OBDT2.
Cut (dim_inv n bdt1 O).
Cut (dim_inv n bdt2 O).
Unfold dim_inv.
Do 4 Induction 1; Auto.
Intro Fone_eq_Fzero.
Unfold BF_eq Fun in Fone_eq_Fzero.
Absurd true=false; Auto.
Apply Fone_eq_Fzero; Exact [x:(Var n)]true.
Intro Fzero_eq_Fone.
Unfold BF_eq Fun in Fzero_eq_Fone.
Absurd false=true; Auto.
Apply Fzero_eq_Fone; Exact [x:(Var n)]true.
(* (dim_inv n bdt2 O) *)
Cut O=(Dimension n bdt2); Try Apply le_n_O_eq; Try Assumption.
Intro Dim_bdt2.
Apply p_dim_inv; Rewrite -> Dim_bdt2; Auto.
(* (dim_inv n bdt1 O) *)
Cut O=(Dimension n bdt1); Try Apply le_n_O_eq; Try Assumption.
Intro Dim_bdt1.
Apply p_dim_inv; Rewrite -> Dim_bdt1; Auto.
(*------  k -> (S k) -------*)
Intros l Hl.
Intros bdt1 bdt2 Dim_bdt1 Dim_bdt2 OBDT1 OBDT2 Fun1_eq_Fun2.
Elim (le_nSm_le_nm_or_eq l (Dimension n bdt1)); Try Assumption;
Elim (le_nSm_le_nm_or_eq l (Dimension n bdt2)); Try Assumption;
Auto.
(*--- Case Dim(bdt1)<=l Dim(bdt2)=(S l) ---*)
Intros Dim2_eq_Sl Dim1_le_l.
Cut (node_decomposition n bdt2 (S l)).
Induction 1.
Intros x b2l b2h Order_x decomp_bdt2.
Absurd (BDT_eq n b2l b2h). 
    (* ~(BDT_eq n b2l b2h) *)
Change (inv_OBDT_eq n (node n x b2l b2h)).
Replace (node n x b2l b2h) with bdt2;Auto.
    (* (BDT_eq n b2l b2h) *)
Cut (OBDT n bdt2); Try Assumption.
Rewrite <- decomp_bdt2.
Intro OBDT_node_b2.
Cut (OBDT n b2l).
Intro OBDT_b2l.
Cut (OBDT n b2h).
Intro OBDT_b2h.
Apply Hl; Auto.
      (* (le (Dimension n b2l) l) *)
Apply gt_S_le.
Replace (S l) with (Dimension n bdt2).
Apply Dim_node_Dim_bl with n b2l b2h x; Auto.
Rewrite -> decomp_bdt2;Auto.
      (* (le (Dimension n b2h) l) *)
Apply gt_S_le.
Replace (S l) with (Dimension n bdt2).
Apply Dim_node_Dim_bh with n b2l b2h x; Auto.
Rewrite -> decomp_bdt2;Auto.
      (* (BF_eq n (Fun n b2l) (Fun n b2h)) *)
Apply BF_eq_trans with (Fun n bdt1).
Apply BF_eq_trans with (Restr n (Fun n bdt1) x false).
Apply BF_eq_trans with (Restr n (Fun n bdt2) x false); Auto.
Apply BF_eq_trans with (Restr n (Fun n (node n x b2l b2h)) x false).
  (* (BF_eq n (Fun n b2l) (Restr n (Fun n (node n x b2l b2h)) x false)) *)
Apply  BF_eq_trans with (Restr n (Fun n b2l) x false); Auto.
(* Use: Restr_node_Restr_low *)
Apply BF_eq_sym.
Apply vars_sup_dim_non_dep; Auto.
  (* (gt (Order n x) (Dimension n b2l)) *)
Rewrite -> Order_x.
Rewrite <- Dim2_eq_Sl.
Rewrite <- decomp_bdt2.
Apply Dim_node_Dim_bl with n:=n bl:=b2l bh:=b2h xi:=x; Auto. 
Rewrite ->decomp_bdt2; Auto.
  (*  (BF_eq n (Restr n (Fun n bdt1) x false) (Fun n bdt1)) *)
Apply vars_sup_dim_non_dep; Auto.
Rewrite -> Order_x.
Apply gt_le_trans with n:=(S l) m:=l p:=(Dimension n bdt1); Auto.
  (*  (BF_eq n (Fun n bdt1) (Fun n b2h)) *)
Apply BF_eq_trans with (Restr n (Fun n bdt1) x true).
Apply BF_eq_sym; Apply vars_sup_dim_non_dep; Auto.
Rewrite -> Order_x.
Apply gt_le_trans with n:=(S l) m:=l p:=(Dimension n bdt1); Auto.
  (* (BF_eq n (Restr n (Fun n bdt1) x true) (Fun n b2h)) *)
Apply BF_eq_trans with (Restr n (Fun n bdt2) x true); Auto.
(* Use : Apply BF_eq_Restr *)
Apply BF_eq_trans with (Restr n (Fun n (node n x b2l b2h)) x true). 
Rewrite -> decomp_bdt2; Auto.
  (*  (BF_eq n (Restr n (Fun n (node n x b2l b2h)) x true) (Fun n b2h)) *)
Apply  BF_eq_trans with (Restr n (Fun n b2h) x true); Auto.
(* Use: Restr_node_Restr_high *)
Apply vars_sup_dim_non_dep; Auto.
  (* (gt (Order n x) (Dimension n b2h)) *)
Rewrite -> Order_x.
Rewrite <- Dim2_eq_Sl.
Rewrite <- decomp_bdt2.
Apply Dim_node_Dim_bh with n:=n bl:=b2l bh:=b2h xi:=x; Auto. 
      (* (OBDT n b2h) *)
Change (inv_OBDT_bh n (node n x b2l b2h)); Auto.
      (* (OBDT n b2l) *)
Change (inv_OBDT_bl n (node n x b2l b2h)); Auto.
    (* (node_decomposition n bdt2 (S l)) *)
Change (dim_inv n bdt2 (S l)); Rewrite <- Dim2_eq_Sl; Auto.
(*--- Case Dim(bdt1)=Sl Dim(bdt2)<=l ---*) 
Intros Dim2_le_l Dim1_eq_Sl.
Cut (node_decomposition n bdt1 (S l)).
Induction 1.
Intros x b1l b1h Order_x decomp_bdt1.
Absurd (BDT_eq n b1l b1h).
    (* ~(BDT_eq n b1l b1h) *)
Change (inv_OBDT_eq n (node n x b1l b1h)).
Replace (node n x b1l b1h) with bdt1; Auto.
    (* (BDT_eq n b1l b1h) *)
Cut (OBDT n bdt1); Try Assumption.
Rewrite <- decomp_bdt1.
Intro OBDT_node_b1.
Cut (OBDT n b1l).
Intro OBDT_b1l.
Cut (OBDT n b1h).
Intro OBDT_b1h.
Apply Hl; Auto.
      (* (le (Dimension n b1l) l) *)
Apply gt_S_le.
Replace (S l) with (Dimension n bdt1).
Apply Dim_node_Dim_bl with n b1l b1h x; Auto.
Rewrite -> decomp_bdt1; Auto.
      (* (le (Dimension n b1h) l) *)
Apply gt_S_le.
Replace (S l) with (Dimension n bdt1).
Apply Dim_node_Dim_bh with n b1l b1h x; Auto.
Rewrite -> decomp_bdt1; Auto.
      (* (BF_eq n (Fun n b1l) (Fun n b1h)) *)
Apply BF_eq_trans with (Fun n bdt2).
Apply BF_eq_trans with (Restr n (Fun n bdt2) x false).
Apply BF_eq_trans with (Restr n (Fun n bdt1) x false); Auto.
Apply BF_eq_trans with (Restr n (Fun n (node n x b1l b1h)) x false).
  (* (BF_eq n (Fun n b1l) (Restr n (Fun n (node n x b1l b1h)) x false)) *)
Apply  BF_eq_trans with (Restr n (Fun n b1l) x false); Auto.
(* Use: Restr_node_Restr_low *)
Apply BF_eq_sym.
Apply vars_sup_dim_non_dep; Auto.
  (* (gt (Order n x) (Dimension n b1l)) *)
Rewrite -> Order_x.
Rewrite <- Dim1_eq_Sl.
Rewrite <- decomp_bdt1.
Apply Dim_node_Dim_bl with n:=n
bl:=b1l bh:=b1h xi:=x; Auto. 
Rewrite -> decomp_bdt1; Auto.
  (* (BF_eq n (Restr n (Fun n bdt2) x false) (Fun n bdt2)) *)
Apply vars_sup_dim_non_dep; Auto.
Rewrite -> Order_x.
Apply gt_le_trans with n:=(S l) m:=l p:=(Dimension n bdt2); Auto.
  (* (BF_eq n (Fun n bdt2) (Fun n b1h)) *)
Apply BF_eq_trans with (Restr n (Fun n bdt2) x true).
Apply BF_eq_sym; Apply vars_sup_dim_non_dep; Auto.
Rewrite -> Order_x.
Apply gt_le_trans with n:=(S l) m:=l p:=(Dimension n bdt2); Auto.
  (* (BF_eq n (Restr n (Fun n bdt2) x true) (Fun n b1h)) *)
Apply BF_eq_trans with (Restr n (Fun n bdt1) x true); Auto.
(* Use : Apply BF_eq_Restr *)
Apply BF_eq_trans with (Restr n (Fun n (node n x b1l b1h)) x true). 
Rewrite ->decomp_bdt1; Auto.
  (*  (BF_eq n (Restr n (Fun n (node n x b1l b1h)) x true) (Fun n b1h)) *)
Apply  BF_eq_trans with (Restr n (Fun n b1h) x true); Auto.
(* Use: Restr_node_Restr_high *)
Apply vars_sup_dim_non_dep; Auto.
  (* (gt (Order n x) (Dimension n b1h)) *)
Rewrite -> Order_x.
Rewrite <- Dim1_eq_Sl.
Rewrite <- decomp_bdt1.
Apply Dim_node_Dim_bh with n:=n bl:=b1l bh:=b1h xi:=x; Auto. 
      (* (OBDT n b1h) *)
Change (inv_OBDT_bh n (node n x b1l b1h)); Auto.
      (* (OBDT n b1l) *)
Change (inv_OBDT_bl n (node n x b1l b1h)); Auto.
    (* (node_decomposition n bdt1 (S l)) *)
Change (dim_inv n bdt1 (S l)); Rewrite <- Dim1_eq_Sl; Auto.
(*-----    Case Dim(bdt1)=Sl Dim(bdt2)=Sl    ------*) 
Intros Dim2_eq_Sl Dim1_eq_Sl.
Cut (node_decomposition n bdt1 (S l)).
Induction 1.
Intros x1 b1l b1h Order_x1 decomp_bdt1.
Cut (node_decomposition n bdt2 (S l)).
Induction 1.
Intros x2 b2l b2h Order_x2 decomp_bdt2.
Cut (OBDT n bdt1); Try Assumption.
Rewrite <- decomp_bdt1.
Intro OBDT_node_b1.
Cut (OBDT n bdt2); Try Assumption.
Rewrite <- decomp_bdt2.
Intro OBDT_node_b2.
Cut (eq_Var n x1 x2).
Intro x1_eq_x2.
Apply BDT_eq_node; Auto.
   (* (BDT_eq n b1l b2l) *)
Cut (OBDT n b1l).
Intro OBDT_b1l.
Cut (OBDT n b2l).
Intro OBDT_b2l.
Apply Hl; Auto. 
       (* (le (Dimension n b1l) l) *)
Apply gt_S_le.
Replace (S l) with (Dimension n bdt1).
Apply Dim_node_Dim_bl with n b1l b1h x1; Auto.
Rewrite -> decomp_bdt1; Auto.
       (* (le (Dimension n b2l) l) *)
Apply gt_S_le.
Replace (S l) with (Dimension n bdt2).
Apply Dim_node_Dim_bl with n b2l b2h x2; Auto.
Rewrite -> decomp_bdt2;Auto.
       (* (BF_eq n (Fun n b1l) (Fun n b2l)) *)
Apply BF_eq_trans with (Restr n (Fun n (node n x1 b1l b1h)) x1 false).
Apply BF_eq_sym.
Apply  BF_eq_trans with (Restr n (Fun n b1l) x1 false); Auto.
(* Use : Apply Restr_node_Restr_low *)
Apply vars_sup_dim_non_dep; Auto.
  (* (gt (Order n x1) (Dimension n b1l)) *)
Rewrite -> Order_x1.
Rewrite <- Dim1_eq_Sl.
Rewrite <- decomp_bdt1.
Apply Dim_node_Dim_bl with n:=n bl:=b1l bh:=b1h xi:=x1; Auto.
  (*  (BF_eq n (Restr n (Fun n (node n x1 b1l b1h)) x1 false) (Fun n b2l)) *)
Apply BF_eq_trans with (Restr n (Fun n (node n x2 b2l b2h)) x2 false).
Rewrite -> decomp_bdt1; Rewrite -> decomp_bdt2.
Pattern x2.
Apply Proof_irrelevance with n:=n x:=x1; Auto.
(* Use: Apply BF_eq_Restr ; Assumption *)
Apply  BF_eq_trans with (Restr n (Fun n b2l) x2 false); Auto.
(* Use : Apply Restr_node_Restr_low *)
Apply vars_sup_dim_non_dep; Auto.
  (* (gt (Order n x2) (Dimension n b2l)) *)
Rewrite -> Order_x2.
Rewrite <- Dim2_eq_Sl.
Rewrite <- decomp_bdt2.
Apply Dim_node_Dim_bl with n:=n bl:=b2l bh:=b2h xi:=x2; Auto.
       (*  (OBDT n b2l) *)
Change (inv_OBDT_bl n (node n x2 b2l b2h)); Auto.
       (*  (OBDT n b1l) *)
Change (inv_OBDT_bl n (node n x1 b1l b1h)); Auto.
  (* (BDT_eq n b1h b2h) *)
Cut (OBDT n b1h).
Intro OBDT_b1h.
Cut (OBDT n b2h).
Intro OBDT_b2h.
Apply Hl; Auto.
       (* (le (Dimension n b1h) l) *)
Apply gt_S_le.
Replace (S l) with (Dimension n bdt1).
Apply Dim_node_Dim_bh with n b1l b1h x1; Auto.
Rewrite -> decomp_bdt1; Auto.
       (* (le (Dimension n b2h) l) *)
Apply gt_S_le.
Replace (S l) with (Dimension n bdt2).
Apply Dim_node_Dim_bh with n b2l b2h x2; Auto.
Rewrite -> decomp_bdt2;Auto.
       (* (BF_eq n (Fun n b1h) (Fun n b2h)) *)
Apply BF_eq_trans with (Restr n (Fun n (node n x1 b1l b1h)) x1 true).
Apply BF_eq_sym.
Apply  BF_eq_trans with (Restr n (Fun n b1h) x1 true); Auto.
(* Use : Apply Restr_node_Restr_high *)
Apply vars_sup_dim_non_dep; Auto.
  (* (gt (Order n x1) (Dimension n b1h)) *)
Rewrite -> Order_x1.
Rewrite <- Dim1_eq_Sl.
Rewrite <- decomp_bdt1.
Apply Dim_node_Dim_bh with n:=n bl:=b1l bh:=b1h xi:=x1; Auto.
  (*  (BF_eq n (Restr n (Fun n (node n x1 b1l b1h)) x1 true) (Fun n b2h)) *)
Apply BF_eq_trans with (Restr n (Fun n (node n x2 b2l b2h)) x2 true).
Rewrite -> decomp_bdt1; Rewrite -> decomp_bdt2.
Pattern x2.
Apply Proof_irrelevance with n:=n x:=x1; Auto.
(* Use: Apply BF_eq_Restr ; Assumption *)
Apply  BF_eq_trans with (Restr n (Fun n b2h) x2 true); Auto.
(* Use : Apply Restr_node_Restr_high *)
Apply vars_sup_dim_non_dep; Auto.
  (* (gt (Order n x2) (Dimension n b2h)) *)
Rewrite -> Order_x2.
Rewrite <- Dim2_eq_Sl.
Rewrite <- decomp_bdt2.
Apply Dim_node_Dim_bh with n:=n bl:=b2l bh:=b2h xi:=x2; Auto.
       (*  (OBDT n b2h) *)
Change (inv_OBDT_bh n (node n x2 b2l b2h)); Auto.
       (*  (OBDT n b1h) *)
Change (inv_OBDT_bh n (node n x1 b1l b1h)); Auto.
      (* (eq_Var n x1 x2) *)
Apply same_order_eq_Var with k:=(S l); Auto.
    (* (node_decomposition n bdt2 (S l)) *)
Change (dim_inv n bdt2 (S l)); Rewrite <- Dim2_eq_Sl; Auto.
    (* (node_decomposition n bdt1 (S l)) *)
Change (dim_inv n bdt1 (S l)); Rewrite <- Dim1_eq_Sl; Auto.
Qed.

(*---------------------------------------------------------------------------*)
  Lemma Dim_is_Greatest_Dep_Var :
  (n:nat)(bdt:(BDT n))(OBDT n bdt) -> (gt (Dimension n bdt) O) ->
      (Ex [x:(Var n)]((Order n x) = (Dimension n bdt) /\
                      (Greatest_Dep_Var n (Fun n bdt) x))).
(*---------------------------------------------------------------------------*)
Proof.
Intros n bdt On Dim.
Cut (node_decomposition n bdt (Dimension n bdt)).
Induction 1.
Intros x bl bh Order_x decomp_bdt.
Exists x; Split; Auto.
Apply Def_GreatestDV.
(*-----  (Dep_Var n (Fun n bdt) x)----- *)
Unfold Dep_Var.
Cut ~(BF_eq n (Fun n bl) (Fun n bh)).
Intro H1.
Unfold not in H1.
Unfold not.
Intro H_absurde.
Apply H1.
    (* (BF_eq n (Fun n bl) (Fun n bh)) *)
Apply BF_eq_trans with n:=n f2:=(Restr n (Fun n bh) x true).
Apply BF_eq_trans with n:=n f2:=(Restr n (Fun n (node n x bl bh)) x true); Auto. 
Apply  BF_eq_trans with n:=n f2:=(Restr n (Fun n (node n x bl bh)) x false);
   [Idtac | Apply BF_eq_sym; Rewrite->decomp_bdt; Auto].
Apply  BF_eq_trans with n:=n f2:=(Restr n (Fun n bl) x false);
    [Idtac | Apply BF_eq_sym; Auto].
Apply BF_eq_sym.
 (*  (BF_eq n (Restr n (Fun n bl) x false) (Fun n bl)) *)
Apply vars_sup_dim_non_dep.
Change (inv_OBDT_bl n (node n x bl bh)).
Rewrite -> decomp_bdt; Auto.
Rewrite -> Order_x.
Apply Dim_node_Dim_bl with n:=n bl:=bl bh:=bh xi:=x; Rewrite -> decomp_bdt; Auto.
 (*  (BF_eq n (Restr n (Fun n bh) x true) (Fun n bh))  *)
Apply vars_sup_dim_non_dep.
Change (inv_OBDT_bh n (node n x bl bh)).
Rewrite -> decomp_bdt; Auto.
Rewrite -> Order_x.
Apply Dim_node_Dim_bh with n:=n bl:=bl bh:=bh xi:=x; Rewrite -> decomp_bdt; Auto.
 (* ~(BF_eq n (Fun n bl) (Fun n bh)) *)
Apply Contra with Q:=(BDT_eq n bl bh).
Intro Fbl_eq_Fbh; Apply Fun_eq_OBDT_eq.
Change (inv_OBDT_bl n (node n x bl bh)); Rewrite -> decomp_bdt; Auto.
Change (inv_OBDT_bh n (node n x bl bh)); Rewrite -> decomp_bdt; Auto.
Auto.
Change (inv_OBDT_eq n (node n x bl bh)); Rewrite -> decomp_bdt; Auto.
 (*----- (xj:(Var n))(Dep_Var n (Fun n bdt) xj)->(le_Var n xj x) ----- *)
Intros xj xj_Dep.
Elim (le_or_gt_Var n xj x); Auto.
Intro H_absurde.
Absurd (Dep_Var n (Fun n bdt) xj); Auto.
Apply vars_sup_dim_non_dep2; Auto.
Elim Order_x.
Change (gt_Var n xj x); Assumption.
 (*  (node_decomposition n bdt (Dimension n bdt)) *)
Cut (dim_inv n bdt (Dimension n bdt)); Auto.
Generalize Dim; Elim (Dimension n bdt); Trivial.
Intros; Absurd (gt O O); Auto.
Qed.

(*****************    Denotation of a function by a BDT   ******************)

Definition Denote := [n:nat][f:(BF n)][bdt:(BDT n)]
                    (OBDT n bdt) /\ (BF_eq n f (Fun n bdt)).

Lemma Denote_inv_eq : 
  (n:nat)(f:(BF n))(bdt:(BDT n))(Denote n f bdt) -> (BF_eq n f (Fun n bdt)).
Proof.
Intros n f bdt.
Unfold Denote.
Induction 1;Auto.
Qed.
Hint Denote_inv_eq.

Lemma Denote_inv_OBDT : 
    (n:nat)(f:(BF n))(bdt:(BDT n))(Denote n f bdt) -> (OBDT n bdt).
Proof.
Intros n f bdt.
Unfold Denote.
Induction 1; Auto.
Qed.

