(****************************************************************************)
(*                                                                          *)
(*                                                                          *)
(*            Lemmas on sets, finite sets, totally ordered sets             *)
(*                            Coq V5.10                                     *)
(*                                                                          *)
(*   Some definitions and lemmas are borrowed from Floyd.v and Dijkstra.v   *)
(*                        in $COQTH/PROGRAMS                                *)
(*                                                                          *)
(*       E. Ledinot                                                         *)
(*       Dassault-Aviation                                                  *)
(*       March 1993                                                         *)
(*                                                                          *)
(****************************************************************************)
(*                               Finite_sets.v                              *)
(****************************************************************************)

(* Warning : Two axioms are introduced:
    Axiom V_decidable : (x,y:V){x=y}+{~x=y}.
    Axiom extensionality: (V,W:set)(set_eq V W) -> (P:set->Prop)(P V) -> (P W).
*)

Require Prelude0.
Require Gt.
Require Lt.

Chapter sets.

Variable V : Set. (* The Universe of discourse *)

Axiom V_decidable : (x,y:V){x=y}+{~x=y}.

Definition set := V->Prop.

Definition In := [P:set][x:V](P x).

Definition incl := [P,Q:set](x:V)(In P x) -> (In Q x).

Definition incl_st := [P,Q:set](incl P Q) /\ (Ex [x:V]~(In P x) /\ (In Q x)).


Lemma incl_refl : (U:set)(incl U U).
Proof.
Intro U; Unfold incl In; Auto.
Qed.
Hint incl_refl.


Lemma incl_trans : (W,X,Y:set)(incl W X)->(incl X Y)->(incl W Y).
Proof.
Unfold incl; Auto.
Qed.


Inductive set_eq : set->set->Prop :=
  Def_set_eq:(U,W:set)(incl U W)->(incl W U)->(set_eq U W).
Hint Def_set_eq.


Lemma set_eq_inv : (U,W:set)(set_eq U W)->((incl U W) /\ (incl W U)).
Proof.
Induction 1; Auto.
Qed.


Lemma set_eq_refl : (U:set)(set_eq U U).
Proof.
Auto.
Qed.
Hint set_eq_refl.


Lemma set_eq_sym : (V1,V2:set)(set_eq V1 V2) -> (set_eq V2 V1).
Proof.
Induction 1; Auto.
Qed.
Immediate set_eq_sym.


Lemma set_eq_trans : 
    (V1,V2,V3:set)(set_eq V1 V2) -> (set_eq V2 V3) -> (set_eq V1 V3).
Proof.
Intros V1 V2 V3 V1_eq_V2 V2_eq_V3.
Cut (incl V1 V2) /\ (incl V2 V1); Try (Apply set_eq_inv; Trivial).
Cut (incl V2 V3) /\ (incl V3 V2); Try (Apply set_eq_inv; Trivial).
Induction 1; Intros H23 H32.
Induction 1; Intros H12 H21.
Apply Def_set_eq; Apply incl_trans with X:=V2; Trivial.
Qed.


Axiom extensionality:
      (U,W:set)(set_eq U W) -> (P:set->Prop)(P U) -> (P W).

Inductive inhabited [S:set] : Prop :=
  Def_inh:(v:V)(In S v) -> (inhabited S).

Definition empty := [x:V]False : set.

Definition all_ := [x:V]True : set.


Section Add.

Variable Q : set.
Variables x : V.

Definition add := [y:V](In Q y) \/ x=y : set.

Variable y : V.

Lemma in_Q_add : (In Q y) -> (In add y).
Proof.
Intro.
Unfold In add.
Auto.
Qed.

Lemma eq_add : x=y -> (In add y).
Proof.
Intro.
Unfold In add.
Auto.
Qed.

End Add.
Hint in_Q_add eq_add.


Lemma singlx : (x,y:V)(In (add empty x) y)->x=y.
Proof.
Intros x y.
Unfold add empty In.
Induction 1; Auto.
Induction 1; Auto.
Qed.



(****************************)
(* Lemmas about sets        *)
(****************************)


Lemma in_empty : (x:V)~(In empty x).
Proof.
Unfold not In empty; Trivial.
Qed.
Hint in_empty.

Lemma in_allV : (x:V)(In all_ x).
Proof.
Unfold In all_; Trivial.
Qed.
Hint in_allV.

Lemma incl_allV : (W:set)(incl W all_).
Proof.
Red; Auto.
Qed.
Hint incl_allV.

Lemma incl_empty : (W:set)(incl empty W).
Proof.
Red; Intros W x H.
Elim H.
Qed.
Hint incl_empty.

Lemma incl_Q_add : (Q:set)(x:V)(incl Q (add Q x)).
Proof.
Exact in_Q_add.
Qed.
Hint incl_Q_add.

Lemma in_add_x : (Q:set)(x:V)(In (add Q x) x).
Proof.
Unfold In add; Auto.
Qed.
Hint in_add_x.

Lemma add_in_or_eq : (Q:set)(x,y:V)(In (add Q x) y) -> (In Q y) \/ x=y.
Proof.
Auto.
Qed.

Lemma incl_add : (Q,Q':set)(y:V)(incl Q Q')->(incl (add Q y) (add Q' y)).
Proof.
Red; Intros.
Red; Red.
Elim H0; Auto.
Qed.
Hint incl_add.

Lemma In_not_In : (Q:set)(x,y:V)(In Q x) -> (~(In Q y)) -> ~x=y.
Proof.
Red; Intros.
Apply H0; Elim H1; Auto.
Qed.

Lemma In_not_In_sym : (Q:set)(x,y:V)(In Q x) -> ~(In Q y) -> ~y=x.
Proof.
Intros; Apply sym_not_equal; Apply (In_not_In Q); Auto.
Qed.

Lemma add_inh : (S:set)(x:V)(inhabited (add S x)).
Proof.
Intros S x.
Apply Def_inh with x; Auto.
Qed.
Hint add_inh.


Lemma incl_add_x : (S1,S2:set)(x:V)
     ~(In S1 x) -> (incl (add S1 x) (add S2 x)) -> (incl S1 S2).
Proof.
Intros S1 S2 x x_out_S1.
Unfold incl add In.
Intros H y y_in_S1.
Elim (H y); Auto.
Intro H_abs.
Unfold In in x_out_S1.
Absurd (S1 x); Auto.
Rewrite -> H_abs; Trivial.
Qed.


Lemma incl_st_add_x : (S1,S2:set)(x:V)
      ~(In S1 x) -> (incl_st (add S1 x) (add S2 x)) -> (incl_st S1 S2).
Proof.
Intros S1 S2 x x_out_S1.
Unfold incl_st.
Induction 1.
Intro H_incl.
Induction 1.
Intros y Def_y.
Split.
Apply incl_add_x with x:=x; Trivial.
Exists y.
Elim Def_y; Intros H1 H2.
Split.
Apply Contra with (In (add S1 x) y); Auto.
Unfold add 1 In in H2.
Elim H2; Auto.
Intro H_abs.
Absurd (In (add S1 x) x); Auto.
Pattern 2 x; Rewrite -> H_abs; Trivial.
Qed.


(********************************)
(*     Operations on sets       *)
(********************************)
    
Definition inter := [P,Q:set][x:V] (In P x) /\ (In Q x).

Definition union := [P,Q:set][x:V] (In P x) \/ (In Q x).

Definition soustr := [Q:set][x,y:V] (In Q y) /\ ~x=y.


Lemma inter_incl_l : (P,Q:set) (incl (inter P Q) P).
Proof.
Red; Intros P Q x H; Elim H; Auto.
Qed.
Hint inter_incl_l.

Lemma inter_incl_r : (P,Q:set) (incl (inter P Q) Q).
Proof.
Red; Intros P Q x H; Elim H; Auto.
Qed.
Hint inter_incl_r.

Lemma inter_intro : (P,Q:set)(x:V)(In P x) -> (In Q x) -> (In (inter P Q) x).
Proof.
Red; Red; Auto.
Qed.
Hint inter_intro.

Lemma notP_notinterPQ : (P,Q:set)(x:V)(~(In P x)) -> ~(In (inter P Q) x).
Proof.
Intros P Q x.
Unfold not inter In; Auto.
Intros H1 H2; Apply H1; Elim H2; Auto.
Qed.
Hint notP_notinterPQ.

Lemma notQ_notinterPQ : (P,Q:set)(x:V)(~(In Q x)) -> ~(In (inter P Q) x).
Proof.
Intros P Q x.
Unfold not inter In; Auto.
Intros H1 H2; Apply H1; Elim H2; Auto.
Qed.
Hint notQ_notinterPQ.

Lemma union_sym : (P,Q:set)(set_eq (union P Q) (union Q P)).
Proof.
Intros P Q.
Apply Def_set_eq; Unfold union incl In; Induction 1; Auto.
Qed.
Immediate union_sym.

Lemma P_union_empty : (P:set)(set_eq (union P empty) P).
Proof.
Intro P.
Apply Def_set_eq; Unfold union incl In; Try Induction 1; Auto.
Unfold empty; Contradiction.
Qed.
Hint P_union_empty.

Lemma empty_union_P : (P:set)(set_eq (union empty P) P).
Proof.
Intro P.
Apply Def_set_eq; Unfold union incl In; Try Induction 1; Auto.
Unfold empty; Contradiction.
Qed.
Hint empty_union_P.

Lemma union_add : (P,Q:set)(x:V)
      (set_eq (union (add P x) Q) (add (union P Q) x)).
Proof.
Intros P Q x.
Apply Def_set_eq; Unfold incl union add In; 
Induction 1; Try Induction 1; Auto.
Qed.

Lemma incl_soustr_in : (M:set)(y:V)(In M y) -> (incl (soustr M y) M).
Proof.
Red; Intros.
Elim H0; Auto.
Qed.
Hint incl_soustr_in.

Lemma incl_soustr : 
      (M,M':set)(y:V)(incl M M') -> (incl (soustr M y) (soustr M' y)).
Proof.
Unfold soustr; Red; Red; Intros.
Elim H0; Intros.
Apply conj; Auto.
Qed.
Hint incl_soustr.

Lemma not_In_soustr : (M:set)(y,z:V)(~(In M y)) -> ~(In (soustr M z) y).
Proof.
Red; Intros.
Elim H0; Intros; Apply H; Auto.
Qed.
Hint not_In_soustr.

Lemma not_In_soustr_eq: (M:set)(y:V)~(In (soustr M y) y).
Proof.
Red; Intros.
Elim H; Intros.
Apply H1; Auto.
Qed.
Hint not_In_soustr_eq.

Lemma still_in_soustr: 
(M:set)(x:V)(In M x)->(y:V)~(x=y) -> (In (soustr M y) x).
Proof.
Intros M x 
x_in_M y x_ne_y.
Unfold In soustr; Auto.
Qed.

Lemma incl_soustr_add_l: (M:set)(y:V)(incl (soustr (add M y) y) M).
Proof.
Red; Intros.
Elim H; Intros.
Elim H0; Intros; Auto.
Absurd y=x; Auto.
Qed.
Hint incl_soustr_add_l.

Lemma incl_soustr_add_r: (M:set)(y:V)
      (~(In M y)) -> (incl M (soustr (add M y) y)).
Proof.
Red; Intros.
Red; Red.
Apply conj; Auto.
Red; Intro; Absurd (In M x); Auto.
Elim H1; Auto.
Qed.
Hint incl_soustr_add_r.

Lemma add_soustr_2: (Q:set) (y:V) (In Q y) -> (incl Q (add (soustr Q y) y)).
Proof.
Intros.
Red; Intros.
Red; Red.
Unfold In soustr.
Elim (V_decidable y x); Intro; Auto.
Qed.
Hint add_soustr_2.

Lemma add_soustr_1: (Q:set) (y:V) (In Q y) -> (incl (add (soustr Q y) y) Q).
Proof.
Intros.
Red; Intros.
Elim H0; Intros.
Elim H1; Auto.
Elim H1; Assumption.
Qed.
Hint add_soustr_1.

Lemma add_soustr_xy:  (Q:set)(x,y:V)~(x=y) ->
                      (set_eq (soustr (add Q x) y) (add (soustr Q y) x)).
Proof.
Intros Q x y x_diff_y.
Apply Def_set_eq; Unfold incl soustr add In; 
Intros z Def_z.
Elim Def_z; Intros H1 H2.
Elim H1; Auto.
Elim Def_z; Try (Induction 1; Auto).
Qed.
Hint add_soustr_xy.

Lemma incl_st_add_soustr:
    (E1,E2:set)(x:V)~(In E1 x) -> (incl_st (add E1 x) E2)
                               -> (incl_st E1 (soustr E2 x)).
Proof.
Intros E1 E2 x 
x_not_in_E1 H.
Unfold incl_st.
Unfold incl_st in H.
Elim H; Intros H_incl H_Ex.
Split.
Apply incl_trans with (soustr (add E1 x) x); Auto.
Elim H_Ex; Intros y Def_y.
Exists y.
Split; Elim Def_y; Intros H1 H2.
Apply Contra with (In (add E1 x) y); Auto.
Apply still_in_soustr; Auto.
Cut ~(In E1 y)/\~(y=x).
Induction 1; Auto.
Apply Morgan_and_not.
Apply Contra with (In (add E1 x) y); Auto; 
Induction 1; Auto.
Qed.

Lemma addQxinterS_addQinterSx : (Q,S:set)(x:V)
      ~(In Q x) -> (In S x) -> (incl (inter (add Q x) S) (add (inter Q S) x)).
Proof.
Intros Q S x x_not_in_Q x_in_S.
Unfold incl inter add In.
Intros y H_y.
Elim H_y; Intros Qy_or_xeqy y_in_S.
Elim Qy_or_xeqy; Auto.
Qed.
Hint addQxinterS_addQinterSx.

Lemma addQinterSx_addQxinterS : (Q,S:set)(x:V)
    (~(In Q x)) -> (In S x) -> (incl (add (inter Q S) x) (inter (add Q x) S)).
Proof.
Intros Q S x x_not_in_Q x_in_S.
Unfold incl inter add In.
Intros y H_y.
Elim H_y.
Intro Qy_and_Sy; Split; Elim Qy_and_Sy; Auto.
Induction 1; Auto.
Qed.
Hint addQinterSx_addQxinterS.

Lemma addQxinterS_QinterS : (Q,S:set)(x:V)
      ~(In Q x) -> ~(In S x) -> (incl (inter (add Q x) S) (inter Q S)).
Proof.
Intros Q S x x_not_in_Q x_not_in_S.
Unfold incl inter add In.
Intro y; Induction 1.
Induction 1; Auto.
Intros Exy Sy; Split; Auto.
Absurd (S y); Trivial.
Elim Exy; Trivial.
Qed.
Hint addQxinterS_QinterS.

Lemma QinterS_addQxinterS : (Q,S:set)(x:V)
      ~(In Q x) -> ~(In S x) -> (incl (inter Q S) (inter (add Q x) S)).
Proof.
Intros Q S x x_not_in_Q x_not_in_S.
Unfold incl inter add In.
Intros y.
Induction 1; Intros.
Split; Auto.
Qed.
Hint QinterS_addQxinterS.


(**************************)
(*   Image by functions   *)
(**************************)

Definition Im : set->(V->V)->set :=
           [Q:set][f:V->V][y:V] (Ex [x:V](In Q x) /\ (f x)=y).

Lemma Im_add : (Q:set)(y:V)(f:V->V)
               (set_eq (Im (add Q y) f) (add (Im Q f) (f y))).
Proof.
Intros Q y f.
Apply Def_set_eq; Unfold incl Im add In; 
Intro fx.
Induction 1; Intro x.
Do 2 Induction 1; Intros.
Elim H1. (* H1 : (Q x)\/y=x *)
Left.
Exists x; Auto.
Right.
Rewrite -> H4; Trivial. (* H4 : y=x *)
Right.
Rewrite -> H2; Trivial. (* H2 : y=x *)
Do 2 Induction 1.
Intro x.
Induction 1.
Intros.
Exists x; Auto.
Exists y; Auto.
Qed.


(*********************)
(*    Finite sets    *)
(*********************)

Inductive finite : set -> Prop :=
  finite_empty : (finite empty)
| finite_add   : (y:V)(Q:set)(~(In Q y)) -> (finite Q) -> (finite (add Q y)).
Hint finite_empty finite_add.

Inductive cardinal : set->nat->Prop :=
  card_empty: (cardinal empty O)
| card_add:   (E:set)(k:nat)(x:V)
              (cardinal E k)-> (~(In E x)) -> (cardinal (add E x) (S k)).
Hint card_empty card_add.

Local inv_card : set->nat->Prop :=
      [E:set][k:nat]<Prop>Match k with
                      (* O  *) (set_eq E empty)
                      (* Sp *) [p:nat][Pp:Prop]True end.

Lemma cardinal_inv : (E:set)(n:nat)(cardinal E n) -> (inv_card E n).
Proof.
Induction 1; 
Unfold inv_card; Auto.
Qed.

Lemma cardinalO_empty : (E:set)(cardinal E O) -> (set_eq E empty).
Proof.
Intros E CardO.
Change (inv_card E O).
Apply cardinal_inv; Trivial.
Qed.

Lemma inh_card_gt_O : (E:set)(inhabited E ) -> 
                      (c:nat)(cardinal E c) -> (gt c O).
Intros E HE.
Elim HE.
Intros e Def_e c Def_c.
Elim (gt_O_eq c); Auto.
Intro c_eq_O.
Absurd (In E e); Auto.
Pattern E.
Apply extensionality with empty; Try (Unfold empty In not; Contradiction).
Apply set_eq_sym; Apply cardinalO_empty.
Rewrite -> c_eq_O; Trivial.
Qed.

Inductive decidable [S:set] : Prop := 
   Def_dec : ((x:V)((In S x) \/ ~(In S x))) -> (decidable S).

Inductive decidable_on [Q,S:set] : Prop := 
   Def_dec_on : ((x:V)(In Q x) -> 
                ((In S x) \/ ~(In S x))) -> (decidable_on Q S).

Lemma inv_dec_on : 
   (Q,S:set)(decidable_on Q S) -> (x:V)(In Q x) -> ((In S x) \/ ~(In S x)).
Proof.
Induction 1; Auto.
Qed.

Lemma dec_addQx_dec_Q : 
   (Q:set)(P:V->Prop)(x:V)(decidable_on (add Q x) P) -> (decidable_on Q P).
Proof.
Intros Q P x.
Induction 1.
Intro P_dec_on_addQx.
Apply Def_dec_on; Auto.
Qed.

Lemma finite_cardinal : (S:set)(finite S) -> (Ex [k:nat](cardinal S k)).
Proof.
Intros E HE.
Elim HE.
Exists O; Trivial.
Intros y Q y_not_inQ Q_finite Ex_CardQ.
Elim Ex_CardQ.
Intros k Def_k.
Exists (S k); Auto.
Qed.

Lemma cardinal_finite : (S:set)(c:nat)(cardinal S c) -> (finite S).
Proof.
Intros S c.
Induction 1; Auto.
Qed.

Lemma decidable_finite : (finite all_) -> (S:set)(decidable S) -> (finite S).
Proof.
Intros V_finite S S_dec.
      (* Apply extensionality with U:=(inter all_ S). PB *)
Apply (extensionality (inter all_ S) S). 
Apply Def_set_eq; Auto.
Unfold incl inter In all_; Simpl; Auto.
Elim V_finite.
      (* Apply extensionality with empty; Auto. *)
Apply (extensionality empty (inter empty S)); Auto. 
Intros x Q x_not_in_Q Q_finite Q_inter_S_finite.
Elim S_dec.
Intro Def_S_decidable.
Elim (Def_S_decidable x).
(*   Case x in S  *)
Intro x_in_S.
      (* Apply extensionality with (add (inter Q S) x); Auto. PB *)
Apply (extensionality (add (inter Q S) x) (inter (add Q x) S)); Auto.
(*   Case x out of S *)
Intro x_not_in_S.
      (* Apply extensionality with (inter Q S); Auto. PB *)
Apply (extensionality (inter Q S) (inter (add Q x) S)); Auto.
Qed.

Lemma finite_decidable: (S:set)(finite S) -> (decidable S).
Proof.
Intro S.
Induction 1.
Apply Def_dec.
Unfold empty; Auto.
Intros y Q y_not_in_Q Q_finite Q_dec.
Apply Def_dec.
Unfold In add.
Elim Q_dec.
Intros Def_Q_dec x.
Elim (Def_Q_dec x); Auto.
Intro not_in_Q_x.
Elim (V_decidable x y); Auto.
Intro x_ne_y.
Right.
Apply Morgan_or; Auto.
Qed.

Lemma card_soustr_1 : 
    (Q:set)(n:nat)(cardinal Q n) -> 
    (y:V)(In Q y) -> (cardinal (soustr Q y) (pred n)).
Proof.
Intros Q n.
Induction 1.
Intro y; Unfold empty In; Contradiction.
Clear H n Q.
Intros Q n x Def_n H_n x_not_in_Q y y_in_addQx.
Cut (In Q y) \/ ~(In Q y).
Induction 1.
(* subcase (In Q y) *)
Intro H_in.
Pattern (soustr (add Q x) y).
(* Apply extensionality with (add (soustr Q y) x). *)
Apply (extensionality (add (soustr Q y) x) (soustr (add Q x) y)).
   (* Proof of set_eq *)
Elim (V_decidable x y).
Intro x_eq_y.
Absurd (In Q y); Auto.
Elim x_eq_y; Auto.
Intro x_diff_y.
Apply set_eq_sym; Auto.
Cut (S (pred n))=(pred (S n)).
Intro H_pred.
Elim H_pred.
Apply card_add; Auto.
(* Use: H_n not_In_soustr *)
Unfold 2 pred; Symmetry.
Apply S_pred with m:=O.
Change (gt n O).
Apply inh_card_gt_O with Q; Auto.
Apply Def_inh with y; Trivial.
(* subcase ~(In Q y) *)
Intro H_out.
Elim (V_decidable x y).
Intro x_eq_y.
Elim x_eq_y.
Pattern (soustr (add Q x) x).
Apply extensionality with Q; Auto.
Intro x_diff_y.
Absurd (In (add Q x) y); Auto.
Unfold add 1 In.
Apply Morgan_or; Trivial.
  (* (In Q y)\/(~(In Q y)) *)
Cut (decidable Q).
Induction 1; Auto.
Apply finite_decidable.
Apply cardinal_finite with n; Trivial.
Qed.

Lemma incl_st_card_lt:
  (E1:set)(c1:nat)(cardinal E1 c1) ->
  (E2:set)(c2:nat)(cardinal E2 c2) -> (incl_st E1 E2) -> (gt c2 c1).
Proof.
Induction 1.
(* E1 empty *)
Induction 1.
   (* E2 empty *)
Unfold incl_st.
Induction 1.
Intro empty_incl_empty.
Induction 1.
Intro e.
Induction 1; Intros; Absurd (In empty e); Auto.
   (* E2 -> (add E2 x) *)
Auto.
(* E1 -> (add E1 x) *)
Clear H E1 c1.
Intros E1 c1 x1 Def_c1 H_E1 x1_not_in_E1 E2 c2 Def_c2.
Elim Def_c2.
   (* E2 empty *)
Unfold incl_st.
Induction 1; Intro H_abs1.
Induction 1; Intro e.
Induction 1; Intros; Absurd (In empty e); Auto.
   (*  E2 -> (add E2 x) *)
Clear Def_c2 c2 E2.
Intros E2 c2 x2 Def_c2 H_E1x x2_not_in_E2 H.
ElimType (In E2 x1) \/ ~(In E2 x1); Try Intro Hx1.
(* subcase (In E2 x1) *)
Apply gt_n_S.
Apply H_E1 with E2:=(soustr (add E2 x2) x1); 
Try (Apply incl_st_add_soustr; Auto).
ElimType (pred (S c2))=c2; Auto.
Apply card_soustr_1; Auto.
(* subcase ~(In E2 x1) *)
Elim (V_decidable x1 x2); Intro Hx1x2.
   (* x1 = x2 *)
Apply gt_n_S.
Apply H_E1 with E2:=E2; Try Trivial.
Apply incl_st_add_x with x:=x1; Auto.
Pattern 2 x1; Rewrite -> Hx1x2; Trivial.
    (* x1 =/= x2 *)
Absurd (In (add E2 x2) x1).
Unfold add 1 In.
Apply Morgan_or; Auto.
Unfold incl_st in H.
Elim H.
Intros H_incl H_Ex.
Unfold incl in H_incl; Auto.
(* Proof of Cut (In E2 x1) \/ ~(In E2 x1) *)
Cut (decidable E2).
Induction 1; Auto.
Apply finite_decidable.
Apply cardinal_finite with c:=c2; Trivial.
Qed.

Lemma finite_union : (E1,E2:set)(finite E1) -> (finite E2) ->
                                (finite (union E1 E2)).
Proof.
Intros E1 E2.
Induction 1.
Intro E2_finite.
(* Apply extensionality with U:=E2; Auto. PB *)
Apply (extensionality E2 (union empty E2)); Auto.
Intros y Q y_not_in_Q Q_finite HQ E2_finite.
(* Apply extensionality with V:=(add (union Q E2) y); PB *)
Apply (extensionality (add (union Q E2) y) (union (add Q y) E2)); 
   Try (Apply set_eq_sym; Apply union_add; Auto).
ElimType (In E2 y) \/~ (In E2 y).
(* (In E2 y)->(finite (add (union Q E2) y)) *)
Intro y_in_E2.
(* Apply extensionality with V:=(union Q E2); Auto. PB *)
Apply (extensionality (union Q E2) (add (union Q E2) y)); Auto.
Apply Def_set_eq; Unfold incl union add In; Auto.
Unfold In in y_in_E2.
Do 3 (Induction 1; Auto).
(*  (~(In E2 y))->(finite (add (union Q E2) y)) *)
Intro y_notin_E2.
Apply finite_add; Auto.
Unfold union In.
Apply Morgan_or; Auto.
Cut (decidable E2); Try (Apply finite_decidable; Auto).
Induction 1; Auto.
Qed.

Lemma finite_image : (E:set)(f:V->V)(finite E) -> (finite (Im E f)).
Proof.
Intros E f.
Induction 1.
(* (finite (Im empty f)) *)
(* Apply extensionality with U:=empty; Auto. PB *)
Apply (extensionality empty (Im empty f)); Auto.
Apply Def_set_eq; Unfold Im empty In; Auto.
Unfold incl In; Do 2 Induction 1; Auto.
(* (finite (Im Q f))->(finite (Im (add Q y) f)) *)
Intros y Q y_notin_Q Q_finite ImQ_finite.
(* Apply extensionality with U:=(add (Im Q f) (f y)); PB *)
Apply (extensionality (add (Im Q f) (f y)) (Im (add Q y) f)); 
  Try (Apply set_eq_sym; Apply Im_add).
Cut (decidable (Im Q f)); Try (Apply finite_decidable; Auto).
Induction 1.
Intro ImQ_dec.
Elim (ImQ_dec (f y)); Auto.
Intro fy_in_ImQ.
(* Apply extensionality with U:=(Im Q f); Auto. PB *)
Apply (extensionality (Im Q f) (add (Im Q f) (f y))); Auto.
Apply Def_set_eq; Unfold incl add; Auto.
Unfold 2 In; Auto.
Intro x.
Unfold 1 In; Induction 1; Auto; 
Induction 1; Auto.
Qed.

Lemma inv_inh_addQx :  (Q:set)(x:V)
   (finite Q) -> (inhabited (add Q x)) -> ((inhabited Q) \/ (set_eq Q empty)).
Proof.
Intros Q' x'.
Induction 1; Auto.
Qed.

Lemma notall_addQx_notall_Q : (Q:set)(P:V->Prop)
       (x:V)(P x) -> (~(y:V)(In (add Q x) y)->(P y)) ->
                     ~((y:V)(In Q y) -> (P y)).
Proof.
Intros Q P x H_Px.
Unfold not.
Intros H1 H2.
Apply H1.
Intros y; Unfold In add; Induction 1; Auto.
Induction 1; Assumption.
Qed.

Lemma notall_existnot_singleton : (P:V->Prop)(x:V)
    (decidable_on (add empty x) P) -> (~(y:V)(In (add empty x) y) -> (P y))
     -> ~(P x).
Proof.
Intros P x singlx_dec singlx_notall.
Cut (In P x)\/~(In P x); Try Apply inv_dec_on with (add empty x); Auto.
Induction 1; Unfold In; Auto.
Intro Px.
Absurd (y:V)(In (add empty x) y)->(P y); Auto.
Intros y y_in_singlx.
Elim (singlx x y); Auto.
Qed.

Lemma onQ_not_All_Exist_not : (Q:set)(P:V->Prop)
       (finite Q) -> (inhabited Q) -> (decidable_on Q P) ->
       (~(x:V)(In Q x) -> (P x)) -> (Ex [x:V](In Q x) /\ ~(P x)).
Proof.
Intros Q' P'.
Induction 1. (*---- Q = empty ------*)
Intros H_inh H_dec H_notall.
Absurd (inhabited empty); Auto.
Unfold not; Induction 1; Unfold In; Auto.
(*---- Q -> (add Q x) -----*)
Intros x Q x_not_in_Q Q_finite H_Q  addQx_inh addQx_dec addQx_notall.
Elim (inv_inh_addQx Q x); Auto.
      (* Q inhabited  *)
Intro Q_inh.
ElimType (P' x)\/~(P' x).
         (* case  (P' x) *)
Intro H_P'x.
Elim H_Q; Auto.
Intro t; Induction 1.
Intros t_in_Q not_P't.
Exists t; Auto.
Apply dec_addQx_dec_Q with x; Trivial.
Apply notall_addQx_notall_Q with x; Trivial.
         (* case ~(P' x) *)
Intro H_not_P'x.
Exists x; Auto.
   (* Proof of (P' x)\/~(P' x) *)
Change (In P' x) \/ ~(In P' x); 
Apply inv_dec_on with (add Q x); Trivial.
(*-----  Q uninhabited  -----*)
Intro Q_eq_empty.
Exists x; Split; Auto.
Apply notall_existnot_singleton; 
Pattern empty; Apply extensionality with Q; Auto; Elim Q_eq_empty; Auto.
Qed.

(*---------------------------------------------------------------------------*)
  Lemma not_Exist_All_not :   (*  Trivial for any set V even infinite *)
   (P:V->Prop) ~(Ex [x:V](P x)) -> (x:V)~(P x).
(*---------------------------------------------------------------------------*)
Proof.
Unfold not; Intros P not_exist x H_Px.
Apply not_exist; Exists x; Auto.
Qed.

(*---------------------------------------------------------------------------*)
  Lemma not_All_Exist_not:
  (P:V->Prop)(finite all_) -> (inhabited all_) ->
   (decidable P)-> (~(x:V)(P x)) -> (Ex [x:V]~(P x)).
(*---------------------------------------------------------------------------*)
Proof.
Intros P V_finite V_inhabited P_dec not_all_P.
Elim (onQ_not_All_Exist_not all_ P); Auto.
Intro x; Induction 1.
Intros Triv not_Px.
Exists x; Auto.
Apply Def_dec_on.
Change (x:V)True -> ((In P x) \/ ~(In P x)).
Elim P_dec; Auto.
Change ~(x:V)True->(P x).
Unfold not.
Unfold not in not_all_P.
Intro H.
Apply not_all_P; Auto.
Qed.

Lemma P_dec_on_Q_AllP_dec_on_Q : (Q:set)(finite Q)->(P:V->Prop)
      (decidable_on Q P) -> ((x:V)(In Q x) -> (P x)) 
                        \/ ~((x:V)(In Q x) -> (P x)).
Proof.
Induction 1.
(*  Q = empty  *)
Intros P P_dec.
Left.
Intro x; Unfold In empty; Contradiction.
(*  Q -> (add Q x) *)
Clear H Q.
Intros x Q x_not_in_Q
Q_finite H_Q P P_dec_addQx.
Elim (H_Q P).
   (* Case P(x) for all x in Q *)
Intro AllP_onQ.
ElimType (P x)\/~(P x); [Idtac | Idtac | Change ((In P x) \/ ~(In P x)); 
             Apply inv_dec_on with Q:=(add Q x); Auto].
Intro Px.
Left.
Intro v.
Unfold add In; Induction 1; Auto.
Induction 1; Trivial.
Right.
Unfold not.
Intro H_absurde.
Absurd (P x); Auto.
   (* Case not P(x) for all x in Q *)
Intro not_AllP_onQ.
Right.
Unfold not in not_AllP_onQ.
Unfold not.
Intro H.
Apply not_AllP_onQ.
Intros v v_in_Q.
Apply H; Auto.
(* (decidable_on Q P) *)
Apply dec_addQx_dec_Q with x:=x; Trivial.
Qed.

(*--------------------------------------------------------------------------*)
 Lemma P_dec_AllP_dec:
    (finite all_) -> (P:V->Prop)(decidable P) -> ((x:V)(P x)) \/ ~(x:V)(P x).
(*--------------------------------------------------------------------------*)
Proof.
Intros V_finite P P_dec.
Elim (P_dec_on_Q_AllP_dec_on_Q all_ V_finite P); Auto.
Intro H.
Right.
Unfold not.
Unfold not in H.
Intro H'.
Apply H; Auto.
(* (decidable_on all_ P) *)
Apply Def_dec_on.
Change (x:V)True->((In P x) \/ ~(In P x)).
Elim P_dec; Auto.
Qed.


(*****************************)
(*       Ordered sets        *)
(*****************************)

Variable    R : V->V->Prop.
Variable    V_eq : V->V->Prop.
Definition reflexivity  := (x:V)(R x x).
Definition anti_symetry := (x,y:V)(R x y)->(R y x)->(V_eq x y).
Definition transitivity := (x,y,z:V)(R x y)->(R y z)->(R x z).

Definition order := reflexivity /\ anti_symetry /\ transitivity.

Definition totality    := (x,y:V)(R x y)\/(R y x).
Definition total_order := order /\ totality.

Inductive greatest [E:set; g:V] : Prop :=
  Def_greatest: (In E g) -> ((y:V)(In E y) -> (R y g)) -> (greatest E g).

Lemma inv_greatest_In : (E:set)(g:V)(greatest E g) -> (In E g).
Proof.
Induction 1; Auto.
Qed.
Hint inv_greatest_In.

Lemma inv_greatest_R : (E:set)(g:V)(greatest E g) -> (y:V)(In E y) -> (R y g).
Proof.
Induction 1; Auto.
Qed.

Section Greatest.
 Hypothesis R_total_order : total_order.

(*---------------------------------------------------------------------------*)
  Lemma Card_gtO_exist_greatest :
       (E:set)(c:nat)(cardinal E c) -> (gt c O) -> (Ex [g:V](greatest E g)).
(*---------------------------------------------------------------------------*)
Proof.
Induction 1.
(* c = O *)
Intro Habs.
Absurd (gt O O); Auto.
(* E ->(add E x) *)
Clear H c E.
Intros E cE x Def_cE H_E x_not_in_E H_triv.
Elim (gt_O_eq cE).
  (* subcase cE > O *)
Intro cE_gt_O.
Elim H_E; Auto.
Intros gE Def_gE.
Elim R_total_order.
Intros R_order R_total.
Unfold totality in R_total.
Elim (R_total x gE).
(* (R x gE) -> (Ex([g:V](greatest (add E x) g))) *)
Intro x_R_gE.
Exists gE; Apply Def_greatest.
Unfold add In; Left; Change (In E gE); Auto.
Intro y; Unfold add; Unfold 1 In.
Induction 1.
Intro y_in_E; Apply inv_greatest_R with E:=E; Auto.
Induction 1; Trivial.
(*  (R gE x) -> (Ex([g:V](greatest (add E x) g))) *)
Intro gE_R_x.
Exists x; Apply Def_greatest; Auto.
Intro y; Unfold add; Unfold 1 In.
Unfold order in R_order.
Elim R_order.
Intros R_refl Rasym_and_trans.
Unfold reflexivity in R_refl.
Elim Rasym_and_trans.
Intros R_asym R_trans.
Unfold transitivity in R_trans.
Induction 1; Try Induction 1; Auto.
Intro y_in_E.
Apply R_trans with y:=gE; Auto.
Apply inv_greatest_R with E:=E; Auto.
  (* subcase cE=O *)
Intro cE_eq_O.
Exists x.
Apply Def_greatest; Auto.
Pattern E; Apply extensionality with U:=empty.
Apply set_eq_sym.
Change (inv_card E O).
Apply cardinal_inv.
Rewrite -> cE_eq_O; Trivial.
Intro y.
Unfold add empty In.
Elim R_total_order.
Intros R_order R_total.
Elim R_order.
Intros Rrefl Rasym_and_trans.
Unfold reflexivity in Rrefl.
Induction 1; [Contradiction | Induction 1; Auto].
Qed.

End Greatest.
End sets.

Hint in_empty in_allV incl_allV incl_empty incl_Q_add
     in_add_x finite_empty finite_add set_eq_refl.

