(****************************************************************************)
(*                                                                          *)
(*                                                                          *)
(*             Sets of variables indexed by natural numbers                 *)
(*                     representing their order                             *)
(*                                                                          *)
(*                                                                          *)
(*                                                                          *)
(****************************************************************************)
(*                                  Vars.v                                  *)
(****************************************************************************)

Require Peano_dec.
Require Compare_dec.
Require Compare.
Require Le.
Require Gt.
Require Prelude1.
Require Finite_sets.

Section Intervals.

Variable n:nat.

Inductive Var [n:nat] : Set := 
  Var_intro : (k:nat)(gt k O)->(le k n)->(Var n).

Definition Order : (Var n)->nat :=
    [xi:(Var n)]<nat>Match xi with [i:nat][pi:(gt i O)][qi:(le i n)]i end.

Lemma Ordering_surj : (k:nat)(gt k O)->(le k n)->(Ex [x:(Var n)]k=(Order x)).
Proof.
Intros k H1 H2.
Exists (Var_intro n k H1 H2); Auto.
Qed.
Hint Ordering_surj.

Definition eq_Var := [xi,xj:(Var n)](Order xi)=(Order xj).
Definition le_Var := [xi,xj:(Var n)](le (Order xi) (Order xj)).
Definition gt_Var := [xi,xj:(Var n)](gt (Order xi) (Order xj)).

Lemma same_order_eq_Var : (k:nat)(xi,xj:(Var n))
       (Order xi)=k -> (Order xj)=k -> (eq_Var xi xj).
Proof.
Intros k xi xj Ind_xi Ind_xj.
Unfold eq_Var.
Replace (Order xi) with k.
Replace (Order xj) with k.
Trivial.
Qed.

Lemma eq_Var_same_order : (xi,xj:(Var n))(eq_Var xi xj)->(Order xi)=(Order xj).
Proof.
Unfold eq_Var;Auto.
Qed.
Hint eq_Var_same_order.

Lemma eq_Var_refl : (x:(Var n))(eq_Var x x).
Proof.
Induction x.
Unfold eq_Var;Auto.
Qed.
Hint eq_Var_refl.

Lemma  eq_Var_sym : (x1,x2:(Var n))(eq_Var x1 x2) -> (eq_Var x2 x1).
Proof.
Intros 
x1 x2 eq_x1_x2.
Apply same_order_eq_Var with k:=(Order x2);Auto.
Qed.
Immediate eq_Var_sym.

Lemma eq_Var_trans : (x1,x2,x3:(Var n))(eq_Var x1 x2) -> 
      (eq_Var x2 x3) -> (eq_Var x1 x3).
Proof.
Intros
x1 x2 x3
eq_12 eq_13.
Cut (Order x1)=(Order x2); Auto.
Cut (Order x2)=(Order x3); Auto.
Intros H23 H12.
Apply same_order_eq_Var with k:= (Order x2); Auto.
Qed.

Lemma eq_Var_dec : (xi,xj:(Var n)){(eq_Var xi xj)}+{~(eq_Var xi xj)}.
Proof.
Unfold eq_Var.
Intros xi xj.
Apply eq_nat_dec.
Qed.
Hint eq_Var_dec.

Lemma le_or_gt_Var : (xi,xj:(Var n))(le_Var xi xj) \/ (gt_Var xi xj).
Proof.
Intros xi xj.
Elim (le_gt_dec (Order xi) (Order xj)); Intro.
Left.
Unfold le_Var; Trivial.
Right.
Unfold gt_Var; Trivial.
Qed.

Lemma Order_le_n : (x:(Var n))(le (Order x) n).
Proof.
Induction x; Simpl; Auto.
Qed.
Hint Order_le_n.

Lemma Order_gt_O : (x:(Var n))(gt (Order x) O).
Proof.
Induction x; Simpl; Auto.
Qed.
Hint Order_gt_O.

Lemma Order_S : (x:(Var n))(Ex [m:nat](Order x)=(S m)).
Proof.
Intros.
Generalize (Order_gt_O x).
Elim (Order x).
Intro; Absurd (gt O O); Auto.
Intros p H1 H2; Exists p; Trivial.
Qed.

Lemma le_Var_total_order : (total_order (Var n) le_Var eq_Var).
Proof.
Unfold total_order.
Split.
(*  (order (Var n) le_Var eq_Var) *)
Unfold order.
Split; Try Split.
    (* (reflexivity (Var n) le_Var) *)
Unfold reflexivity le_Var; Auto.
    (* (anti_symetry (Var n) le_Var eq_Var) *)
Unfold  anti_symetry le_Var eq_Var; Auto.
    (* (transitivity (Var n) le_Var) *)
Unfold transitivity le_Var.
Intros xi xj xk xi_le_xj xj_le_xk.
Apply le_trans with m := (Order xj); Auto.
(*  (totality (Var n) le_Var)     *)
Unfold totality le_Var; Auto.
Qed.
Hint le_Var_total_order.

(*-----------------       Sets of indexed variables      ------------------*)


Definition Varset := [n:nat](set (Var n)).


(*-------------------------------------------------------------------------*)

Axiom Proof_irrelevance :
  (x:(Var n))(P:(Varset n))(P x)->(a:(Var n))(eq_Var x a)->(P a).

(*-------------------------------------------------------------------------*)

Definition Var_le : nat->(Var n)->Prop 
  := [k:nat][x:(Var n)](le (Order x) k).


Lemma Var_k_finite : (k:nat)(le k n)->(finite (Var n) (Var_le k)).
Proof.
Induction k.
(* k = O *)
Intro H_triv.
(* Apply (extensionality with U:=(Var n) V0:=(empty (Var n)); Auto. PB *)
Apply (extensionality (Var n) (empty (Var n)) (Var_le O)); Auto.
Apply Def_set_eq; Unfold incl empty Var_le In; Try Contradiction.
Intros x H_abs.
Absurd (le (Order x) O); Auto.
(* k -> (S k) *)
Clear k.
Intros k Hk Sk_le_n.
ElimType (Ex [x:(Var n)](S k)=(Order x)); Auto.
Intros x Order_x.
(* Apply extensionality with U:=(Var n) V0:=(add (Var n) (Var_le k) x). PB *)
Apply (extensionality (Var n) (add (Var n) (Var_le k) x) (Var_le (S k))).
  (* (set_eq (Var n) (add (Var n) (Var_le k) x) (Var_le (S k))) *)
Apply Def_set_eq; Unfold incl empty Var_le add In.
Intro xi.
Induction 1; Auto.
Induction 1; Elim Order_x; Auto.
Intros xi Oxi_le_Sk.
Elim (le_nSm_le_nm_or_eq  k (Order xi)); Auto.
Intro Order_xi.
Right.
Pattern xi.
Apply Proof_irrelevance with x:=x; Auto.
Apply same_order_eq_Var with k:=(S k); Trivial.
Symmetry; Trivial.
  (* (finite (Var n) (add (Var n) (Var_le k) x)) *)
Apply finite_add; Auto.
Unfold In Var_le.
Elim Order_x; Auto.
Qed.

Lemma Var_n_finite : (finite (Var n) (all_ (Var n))).
Proof.
(* Apply extensionality with V:=(Var n) V0:=(Var_le n). PB *)
Apply (extensionality (Var n) (Var_le n) (all_ (Var n))).
Apply Def_set_eq; Unfold incl Var_le all_ In; Auto.
Apply Var_k_finite; Auto.
Qed.

End Intervals.

Hint Ordering_surj eq_Var_same_order
     eq_Var_refl eq_Var_dec Order_le_n Order_gt_O 
     le_Var_total_order Var_n_finite.

Immediate eq_Var_sym.




