(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                environ.ml                                *)
(****************************************************************************)


(* The Coq theory module entries comprise now :
   TOKEN	   tokens
   CONSTANT        global constants
   MUTUALINDUCTIVE       global inductive definitions
   SYNTAXCONSTANT  parsing macros
   AUTOHINT        hints for tactic Auto
   PPSYNTAX        syntax tables for prettyprinting
   GRAMMAR         LL1 grammars
   ABSTRACTION     abstract syntax macros, declared in abstraction.ml 

The module environ instanciates libobject for entries CONSTANT, INDUCTIVE,
SYNTAXCONSTANT, AUTOHINT, PPSYNTAX and GRAMMAR *)

#infix "o";;
#open "pp";;
#open "stdpp";;
#open "lomsg";;
#open "std";;
#open "names";;
#open "more_util";;
#open "term";;
#open "constrtypes";;
#open "libobject";;
#open "reintern0";;
#open "reintern1";;
#open "summary";;


(* CONSTANT *)

let cache_constant = function
    (sp,(cmap,str,uni)) ->
    (nametab__rollback o constraintab__rollback)
    (fun () ->
    listmap__app
    (fun (k,lobj) -> nametab__push (basename sp) (coerce_path k sp))
    cmap;
    constraintab__push uni)
    ()
  | _ -> invalid_arg "environ__cache_constant"
;;

let reintern_constant (cmap,n,uni) =
    (map (fun (pk,cb) -> (pk,ri_cb cb)) cmap,n,
     ri_universes uni)
;;

let print_constant_object = fun
    _ (cmap,n,uni) ->
    [< 'S"#<constant(" ;
       prlist_with_sep (fun () -> [< 'S"," >])
                       (fun (k,_) -> [< 'S(string_of_kind k) >]) cmap ;
       'S")" ;
       (match n with
        NeverDischarge -> [< 'S"(global)" >]
      | DischargeAt sp -> [< 'S"(disch@" ; 'S(string_of_path sp) >]) ;
       'S"(" ; 'INT(impuniv__num_universes uni) ; 'S" uni,";
               'INT(impuniv__num_edges uni) ; 'S"edges)>" >]
;;

let specification_constant (cmap,n,uni) =
    (specification_of_constant cmap,n,uni);;

let (inConstant,outConstant) =
    declare_object ("CONSTANT",
                    {cache_function = cache_constant;
                     reintern_function = reintern_constant;
                     specification_function = specification_constant});;


let cache_mutual_inductive (sp,(mimap,uni)) =
    (nametab__rollback o constraintab__rollback)
    (fun () ->
    listmap__app
    (fun (k,mib) ->
         do_vect
         (fun mip ->
              begin
              nametab__push mip.MINDTYPENAME (coerce_path k sp);
              do_vect (fun cn -> nametab__push cn (coerce_path k sp)) 
                      mip.MINDCONSNAMES
              end)
         mib.MINDPACKETS)
     mimap;
    constraintab__push uni)
    ()
;;

let reintern_mutual_inductive ((mimap,uni):mutual_inductive_object) =
    ((map (fun (pk,mib) -> (pk,ri_mib mib)) mimap,
     ri_universes uni):mutual_inductive_object)
;;


let specification_minductive (imap,uni) =
    (specification_of_minductive imap,uni);;

let (inMutualInductive,outMutualInductive) =
    declare_object ("MUTUALINDUCTIVE",
                    {cache_function = cache_mutual_inductive;
                    reintern_function = reintern_mutual_inductive;
            specification_function = specification_minductive});;

let cache_variable (_,(id,vbody,str,sticky,uni)) =
    (vartab__rollback o constraintab__rollback)
    (fun () ->
    vartab__push id sticky vbody;
    constraintab__push uni)
    ()
;;

let reintern_variable (id,(c,copt),n,b,uni) =
    (ri_id id,(ri_constr c,option_app ri_constr copt),n,b,
     ri_universes uni)
;;

let specification_variable (id,_,_,_,_) =
    errorlabstrm "specification_variable"
    [< 'S"Cannot extract a specification from a variable declaration " ;
       'S (string_of_id id) >]
;;

let (inVariable,outVariable) =
    declare_object ("VARIABLE",
                    {cache_function = cache_variable;
                     reintern_function = reintern_variable;
                     specification_function = specification_variable});;

let cache_syntaxconstant = function
    (sp,cmap) ->
    listmap__app
    (fun (k,lobj) -> nametab__push (basename sp) (coerce_path k sp))
    cmap
  | _ -> invalid_arg "environ__cache_syntaxconstant"
;;

let reintern_syntaxconstant cmap =
    (map (fun (pk,c) -> (pk,ri_constr c)) cmap)
;;

let specification_syntaxconstant cmap = cmap;;

let (inSyntaxConstant,outSyntaxConstant) =
    declare_object ("SYNTAXCONSTANT",
                    {cache_function = cache_syntaxconstant;
                     reintern_function = reintern_syntaxconstant;
                     specification_function = specification_syntaxconstant})
;;

let cache_ppsyntax x = syntax__add x;;
let reintern_ppsyntax se = syntax__ri_syntax_entry se;;
let specification_ppsyntax x = x;;

let (inPPSyntax,outPPSyntax) =
    declare_object ("PPSYNTAX",
                    {cache_function = cache_ppsyntax;
                     reintern_function = reintern_ppsyntax;
                     specification_function = specification_ppsyntax});;

let const_of_path sp =
  match lib__leaf_object_tag (objsp_of sp) with
    "CONSTANT" ->
      let (cmap,_,_) = outConstant(lib__map_leaf (objsp_of sp)) in
      (sp,listmap__map cmap (kind_of_path sp))

  | _ -> invalid_arg "const_of_path called with non-constant"
;;

let const_option_of_path sp =
try SOME(const_of_path sp)
with Not_found -> NONE
;;

let mind_of_path sp =
    match lib__leaf_object_tag (objsp_of sp) with
    "MUTUALINDUCTIVE" ->
    let (cmap,_) = outMutualInductive (lib__map_leaf (objsp_of sp))
    in (sp,listmap__map cmap (kind_of_path sp))
  | _ -> invalid_arg "mind_of_path called with non-mutual-inductive"
;;

let mind_option_of_path sp =
try SOME(mind_of_path sp)
with Not_found -> NONE
;;

let search_synconst pk str =
let sp = nametab__sp_of_id CCI str
in (match lib__leaf_object_tag(objsp_of sp) with
    "SYNTAXCONSTANT" ->
    let cmap = outSyntaxConstant(lib__map_leaf (objsp_of sp))
    in listmap__map cmap pk
  | _ -> raise Not_found)
;;

(* $Id: environ.ml,v 1.7 1996/02/12 15:37:34 cpaulin Exp $ *)
