(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                impuniv.ml                                *)
(****************************************************************************)

(* Universes are stratified by a partial ordering >=.
   Let ~ be the associated equivalence. We also have a strict ordering
   > between equivalence classes, and we maintain that > is acyclic,
   and contained in >= in the sense that [U]>[V] implies U>=V.

   At every moment, we have a finite number of universes, and we
   maintain the ordering in the presence of assertions U>V and U>=V.

   The equivalence ~ is represented by a tree structure, as in the
   union-find algorithm. The assertions > and >= are represented by
   adjacency lists *)

#open "std";;
#open "names";;
#open "pp";;
#open "stdpp";;
#open "more_util";;

let uni_ord u1 u2 =
    let sp_bit = sp_ord u1.u_sp u2.u_sp
    in if sp_bit < 0 then -1
       else if sp_bit > 0 then 1
       else u1.u_num - u2.u_num
;;

(* in Arc(u,Greater(b,v,r))::arcs, we have u>v if b, and u>=v if not b, 
   and r is the next relation pertaining to u; this relation may be
   Greater or Terminal. *)

(* points to the first arc pertaining to u 

let arc_of u = arc_of_u where rec arc_of_u = function
    [] -> anomaly "Wrong universe structure"
  | (Arc(v,_) as arc)::l -> if eq(u,v) then arc else arc_of_u l;;

A completely applicative structure would search the graph structure
with (arc_of u graph). We prefer, for efficiency, give a direct access
as ARC.(Int i), where (Universe i) = u. *)

(* Fast access to first arc pertaining to given universe *)
let ARC = (fmavm__new (uni_ord,17) : (universe,arc) fmavm__t);;
let empty_universes = fmavm__freeze ARC;;

(* The latest universe *)
let U = ref 0;;
let SP = ref (NONE : section_path option);;

library__declare_summary "Universes_numbering"
{summary__freeze_function = (fun () -> !U);
 summary__unfreeze_function = (fun x -> (*if x>!U then*) U:=x);
 summary__init_function = (fun () -> U:=0);
 summary__sync_function = (fun () -> false)}
;;

let enter_arc a = match a with
   Arc(i,_) -> (if fmavm__in_dom ARC i then
                    fmavm__remap ARC i a
                else fmavm__add ARC (i,a))
 | Undef -> anomaly "Undefined universe"
;;

let declare_univ u =
    if not (fmavm__in_dom ARC u) then 
    fmavm__add ARC (u,Arc(u,Terminal))
;;

(* Every universe has a unique canonical arc representative *)

(* repr : universe -> arc *)
(* canonical representative : we follow the Equiv links *)
let rec repr u = 
try
  match fmavm__map ARC u with 
    Arc(_,Equiv(v)) -> repr v
  | Undef -> anomaly "Undefined universe"
  | arc -> arc
with Not_found -> Undef
;;

let can = map repr;;

(* transitive closure : we follow the Greater links *)
(* close : relation -> universe list * universe list *)
let close = closerec ([],[]) 
where rec closerec ((U,V) as pair) = function
    Terminal           -> pair
  | Greater(true,v,r)  -> closerec (v::U,V) r
  | Greater(false,v,r) -> closerec (U,v::V) r
  | _ -> anomaly "Wrong universe structure";;

(* reprgeq : arc -> arc list *)
(* All canonical arcv such that arcv>=arcu with arcv#arcu *)
let reprgeq = function
    Undef -> anomaly "Undefined universe"
  | (Arc(_,ru) as arcu) -> let (_,V) = close ru
                 in searchrec [] V
where rec searchrec W = function
     [] -> W
   | v::V -> let arcv = repr v 
             in if memq arcv W or eq(arcu,arcv) then searchrec W V
                else searchrec (arcv::W) V;;

(* collect : arc -> arc list * arc list *)
(* collect u = (V,W) iff V={v canonical | u>v} W={w canonical | u>=w}-V *)
(* i.e. collect does the transitive closure of what is known about u *)
let collect u = coll_rec [] [] ([],[u])
where rec coll_rec V W = function
    [],[] -> (V,subtractq W V)
  | (Arc(_,rv) as arcv)::V',W' -> if memq arcv V then coll_rec V W (V',W')
                                  else let (gt,geq) = close rv
                             in coll_rec (arcv::V) W (can(gt@geq)@V',W')
  | [],(Arc(_,rw) as arcw)::W' -> 
          if (memq arcw V) or (memq arcw W) then coll_rec V W ([],W')
          else let (gt,geq) = close rw 
               in coll_rec V (arcw::W) (can(gt),can(geq)@W')
  | _ -> anomaly "Undefined universe";;

type order = EQ | GT | GE | NGE;;

(* compare : universe -> universe -> order *)
let compare(u,v) = 
    let arcu = repr(u) and arcv = repr(v)
    in if eq(arcu,arcv) then EQ
       else let (V,W) = collect(arcu)
            in if memq arcv V then GT
               else if memq arcv W then GE
               else NGE;;

(* Invariants : compare(u,v) = EQ <=> compare(v,u) = EQ
                compare(u,v) = GT or GE => compare(v,u) = NGE
                compare(u,v) = NGE => compare(v,u) = NGE or GE or GT

   Adding u>=v is consistent iff compare(v,u) # GT 
    and then it is redundant iff compare(u,v) # NGE
   Adding u>v is consistent iff compare(v,u) = NGE 
    and then it is redundant iff compare(u,v) = GT *)


(* between : universe -> arc -> arc list *)
(* we assume  compare(u,v) = GE with v canonical    *)
(* between u v = {w|u>=w>=v, w canonical}          *)     
let between u arcv = snd (explore ([(arcv,[arcv])],[]) (repr u))
    where rec explore (memo,l) arcu = 
       try (memo,unionq (assq arcu memo) l) (* when memq arcu memo *)
       with Not_found -> let W = reprgeq arcu
         in let (memo',sols)=it_list explore (memo,[]) W
            in let sols'=if sols=[] then l else arcu::(unionq sols l)
               in ((arcu,sols')::memo',sols');;
(* Note: hd(between u v) = repr u  *)
(* between is the most costly operation *)

(* setgt : universe -> universe -> unit *)
(* forces u > v *)
let setgt u v = match repr(u) with
     Arc(u',ru) -> enter_arc(Arc(u',Greater(true,v,ru))); ()
   | Undef -> anomaly "Undefined universe";;

(* checks that non-redondant *)
let setgt_if u v = match compare(u,v) with
    GT -> ()
  | _ -> setgt u v;;

(* setgeq : universe -> universe -> unit *)
(* forces u >= v *)
let setgeq u v = match repr(u) with
    Arc(u',ru) -> enter_arc(Arc(u',Greater(false,v,ru))); ()
  | Undef -> anomaly "Undefined universe";;

(* checks that non-redondant *)
let setgeq_if u v = match compare(u,v) with
    NGE -> setgeq u v
  | _ -> ();;

(* merge : universe -> universe -> unit *)
(* we assume  compare(u,v) = GE *)
(* merge u v  forces u ~ v with repr u as canonical repr *)
let merge u v = match between u (repr v) with
   Arc(u',_)::V -> let redirect (W,W') = function
        Arc(v',rv) -> let V,V' = close rv
                      in (enter_arc (Arc(v',Equiv(u'))); (unionq V W,V'@W'))
      | Undef -> anomaly "Undefined universe"
                   in let (W,W') = it_list redirect ([],[]) V
                      in (map (setgt_if u') W; map (setgeq_if u') W'; ())
 | Undef::_ -> anomaly "Undefined universe"
 | [] -> anomaly "between";;

(* merge_disc : universe -> universe -> unit *)
(* we assume  compare(u,v) = compare(v,u) = NGE *)
(* merge_disc u v  forces u ~ v with repr u as canonical repr *)
let merge_disc u v = match repr u with
        Arc(u',_) -> (match repr v with
          Arc(v',rv) -> let V,V' = close rv
                        in (enter_arc (Arc(v',Equiv(u')));
                            map (setgt_if u') V;
                            map (setgeq_if u') V'; ())
        | Undef -> anomaly "Undefined universe")
      | Undef -> anomaly "Undefined universe";;


(* enforcegeq : universe -> universe -> unit *)
(* enforcegeq u v will force u>=v if possible, will fail otherwise *)
let enforcegeq u v =
if !SP = NONE then () else
    (declare_univ u;declare_univ v;
     match compare(u,v) with
        NGE -> (match compare(v,u) with
                  GT -> errorlabstrm "universe"
                        [< 'S"Universe Inconsistency" >]
                | GE -> merge v u
                | NGE -> setgeq u v
                | EQ -> anomaly "compare")
      | _ -> ());;

(* enforceq : universe -> universe -> unit *)
(* enforceq u v will force u=v if possible, will fail otherwise *)
let enforceq u v =
if !SP = NONE then () else
    (declare_univ u; declare_univ v;
     match compare(u,v) with
        EQ -> ()
      | GT -> errorlabstrm "universe"
              [< 'S"Universe Inconsistency" >]
      | GE -> merge u v
      | NGE -> match compare(v,u) with
                    GT -> errorlabstrm "universe"
                          [< 'S"Universe Inconsistency" >]
                  | GE -> merge v u
                  | NGE -> merge_disc u v
                  | EQ -> anomaly "compare");;

(* enforcegt u v will force u>v if possible, will fail otherwise *)
let enforcegt u v =
if !SP = NONE then () else
    (declare_univ u; declare_univ v;
    match compare(u,v) with
        GT -> ()
      | GE -> setgt u v
      | EQ -> errorlabstrm "universe"
              [< 'S"Universe Inconsistency" >]
      | NGE -> (match compare(v,u) with
                     NGE -> setgt u v
                   | _ -> errorlabstrm "universe"
                          [< 'S"Universe Inconsistency" >]));;

let dummy_sp = make_path(PA__intern ["univ"]) (id_of_string"dummy") OBJ;;
let Dummy_univ = {u_sp = dummy_sp; u_num = 0};; (* for prover terms *)


let next_univ () = (U:=succ_int !U; !U);;

let mk_univ () = {u_sp = outSOME !SP; u_num = (next_univ())};;

let New_univ () = 
    if !SP = NONE then Dummy_univ else
    let v = mk_univ() in enter_arc(Arc(v,Terminal)); v
and Super u = 
    if !SP = NONE then Dummy_univ else
    (declare_univ u;
     let v = mk_univ() in enter_arc(Arc(v,Greater(true,u,Terminal))); v);;

(* UNUSED 
let Supereq u = 
    let v = mk_univ() in enter_arc(Arc(v,Greater(false,u,Terminal))); v;;
*)

let sup(u,v) = 
    if !SP = NONE then Dummy_univ else
        (declare_univ u; declare_univ v;
         match compare(u,v) with
         NGE -> (match compare(v,u) with
                 NGE -> let w = New_univ() in (setgeq w u; setgeq w v; w)
               | _   -> v)
       | _   -> u);;

let set_uniarc universes = fmavm__unfreeze universes ARC;;

let get_uniarc () = fmavm__freeze ARC;;

let pr_uni u = 
    [< 'S(string_of_path u.u_sp) ; 'S"." ; 'INT u.u_num >]
;;

let pr_reln u r = prec r
    where rec prec = function
    Greater(true,v,r) -> [< pr_uni u ; 'S">" ; pr_uni v ; 'FNL ; prec r >]
  | Greater(false,v,r) -> [< pr_uni u ; 'S">=" ; pr_uni v ; 'FNL ; prec r >]
  | Equiv v -> [< pr_uni u ; 'S"=" ; pr_uni v >]
  | Terminal -> [< >]
;;

(* Debug *)
let pr_universes u =
let am = avm__new uni_ord in
let am' = avm__unfreeze u am in
let graph = avm__toList am
in prlist_with_sep (fun () -> [< 'FNL >])
    (fun (_,Arc(u,r)) -> pr_reln u r
       | (_,Undef) -> anomaly "UNI") graph
;;

let prop_sp = make_path(PA__intern ["univ"]) (id_of_string"prop_univ") OBJ;;

let with_universes f (sp,u,a) =
if !SP <> NONE then
    anomaly "already inside a with_universes"
else
    try (SP := SOME sp;
         set_uniarc u;
         let r = f a in
         let u' = get_uniarc ()
         in (SP := NONE;
             (u',r)))
    with reraise -> (SP:=NONE;raise reraise)
;;

let (initial_universes,(prop_univ,prop_univ_univ,prop_univ_univ_univ)) =
    with_universes
        (fun () ->
             let u = New_univ() in
             let su = Super u in
             let ssu = Super su in
             (u,su,ssu))
        (prop_sp,empty_universes,())
;;

let Super u = 
    if !SP = NONE then Dummy_univ else
    if u = prop_univ then prop_univ_univ 
    else if u = prop_univ_univ then prop_univ_univ_univ else
    (declare_univ u;
     let v = mk_univ() in enter_arc(Arc(v,Greater(true,u,Terminal))); v);;



let uni_pred P a =
let u = get_uniarc()
in try if P a then true
       else (set_uniarc u;false)
   with reraise -> (set_uniarc u;raise reraise)
;;

let uni_or P Q =
let u = get_uniarc()
in try if P() then true
       else
           (set_uniarc u;
            if Q() then true
            else (set_uniarc u;false))
   with reraise -> (set_uniarc u;raise reraise)
;;
let enforce_relation u = enfrec
    where rec enfrec = function
    Terminal -> ()
  | Equiv v -> enforceq u v
  | Greater(false,v,r) -> (enforcegeq u v; enfrec r)
  | Greater(true,v,r) -> (enforcegt u v; enfrec r)
;;

let merge_universes sp U1 U2 =
let ouni = get_uniarc() and
    am = avm__unfreeze U2 (avm__new uni_ord)
in fst(with_universes
        (fun () ->
             avm__app
             (fun (_,Arc(u,r)) -> enforce_relation u r
                | (_,Undef) -> anomaly "merge_universes")
             am)
        (sp,U1,()))
;;

let num_universes U =
let am = avm__unfreeze U (avm__new uni_ord)
in length (avm__dom am)
;;

let num_edges U =

let am = avm__unfreeze U (avm__new uni_ord) in

let reln_len = lenrec 0
    where rec lenrec n = function
    Terminal -> n
  | Equiv _ -> n+1
  | Greater(_,_,r) -> lenrec (n+1) r in

let edges = ref 0

in avm__app (fun (_,Arc(_,r)) -> edges := !edges + (reln_len r)) am;
   !edges
;;

(* $Id: impuniv.ml,v 1.2 1996/02/12 15:37:38 cpaulin Exp $ *)
