(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                 sosub.ml                                 *)
(****************************************************************************)
#open "std";;
#open "names";;
#open "vectops";;
#open "more_util";;
#open "pp";;
#open "stdpp";;
#open "generic";;
#open "term";;

(* Given a term with variables in it, and second-order substitution,
   this function will apply the substitution.  The special
   operator "XTRA[$SOAPP]" is used to represent the second-order
   substitution.

   (1) second-order substitutions: $SOAPP(SOLAM;... soargs ...)
       This operation will take the term SOLAM, which must be
       of the form [x1][x2]...[xn]B and replace the xi with the
       soargs.  If any of the soargs are variables, then we will
       rename the binders for these variables to the name of the
       xi.
 *)

let list_assign_1 l n e = assrec n l
    where rec assrec = fun
    1 (h::t) -> e::t
  | n (h::t) -> h::(assrec (n-1) t)
  | _ [] -> failwith "list_assign_1";;

(* [propagate_name spine_map argt new_na]
 * if new_na is a real name, and argt is a (Rel idx)
 * then the entry for idx in the spine_map
 * is replaced with new_na.  In any case, a new spine_map is returned.
 *)
let propagate_name smap = fun
    (Rel idx) (Name _ as na) ->
    (try list_assign_1 smap idx na
     with Failure _ -> smap)
  | _ _ -> smap
;;

let is_soapp_operator = function
    (DOPN(XTRA("$SOAPP",[]),cl)) -> true
  | (DOP2(XTRA("$SOAPP",[]),_,_)) -> true
  | _ -> false;;

let dest_soapp_operator = function
    (DOPN(XTRA("$SOAPP",[]),cl)) -> (hd_vect cl,list_of_tl_vect cl)
  | (DOP2(XTRA("$SOAPP",[]),c1,c2)) -> (c1,[c2])
  | _ -> failwith "dest_soapp_operator"
;;

let solam_names = namrec []
    where rec namrec acc = function
    DLAM(na,b) -> namrec (na::acc) b
  | _ -> rev acc
;;

let do_propsoapp smap (solam,soargs) =
    let rec propsolam smap = fun
        (DLAM(na,b)) (arg::argl) ->
        propsolam (propagate_name smap arg na) b argl
      | (DLAM _) [] -> error "malformed soapp"
      | _ (_::_) -> error "malformed soapp"
      | _ [] -> smap
    in propsolam smap solam soargs
;;

let propsoapp smap t =
    if is_soapp_operator t then
        do_propsoapp smap (dest_soapp_operator t)
    else smap
;;

(* [propagate_names spine_map t]
 * walks t, looking for second-order applications.
 * Each time it finds one, it tries to propagate
 * backwards any names that it can for arguments of
 * the application which are debruijn references.
 *)
let propagate_names = proprec
    where rec proprec smap t =
    let (smap,t) = 
        (match t with
        DOP0 oper -> (smap,t)
      | DOP1(op,c) ->
        let (smap',c') = proprec smap c
        in (smap',DOP1(op,c'))

      | DOP2(op,c1,c2) ->
        let (smap',c1') = proprec smap c1 in
        let (smap'',c2') = proprec smap c2
        in (smap'',DOP2(op,c1',c2'))
      | DOPN(op,cl) ->
        let cllist = list_of_vect cl in
        let (smap',cl'list) =
            list_it (fun c (smap,acc) ->
                         let (smap',c') = proprec smap c
                         in (smap',c'::acc))
            cllist (smap,[])
        in (smap',DOPN(op,vect_of_list cl'list))

      | DOPL(op,cl) ->
        let cllist = cl in
        let (smap',cl'list) =
            list_it (fun c (smap,acc) ->
                         let (smap',c') = proprec smap c
                         in (smap',c'::acc))
            cllist (smap,[])
        in (smap',DOPL(op,cl'list))

      | DLAM(na,c) ->
        let (na'::smap',c') = proprec (na::smap) c
        in (smap',DLAM(na',c'))

      | DLAMV(na,cl) ->
        let cllist = list_of_vect cl in
        let (na'::smap',cl'list) =
            list_it (fun c (smap,acc) ->
                         let (smap',c') = proprec smap c
                         in (smap',c'::acc))
            cllist (na::smap,[])
        in (smap',DLAMV(na',vect_of_list cl'list))

      | Rel i -> (smap,t)
      | VAR id -> (smap,t))
    in (propsoapp smap t,t) 
;;

let socontract = lamrec [] where rec lamrec sigma = fun
    (h::t) (DLAM(_,c)) -> lamrec (h::sigma) t c
  | []     t           -> (substl sigma t)
  | (_::_) _           -> error "second-order app with mismatched arguments"
;;

(* [soeval t]
 * evaluates t, by contracting any second-order redexes
 *)
let soeval = evrec
    where rec evrec t =
    match t with
        DOP0 _ -> t
      | Rel _ -> t
      | VAR _ -> t
      | DOP1(op,c) -> DOP1(op,evrec c)
      | DLAM(na,c) -> DLAM(na,evrec c)
      | DLAMV(na,cl) -> DLAMV(na,map_vect evrec cl)

      | DOP2(op,c1,c2) ->
        let c1' = evrec c1 and
            c2' = evrec c2
        in if is_soapp_operator t then
            socontract [c2'] c1'
           else DOP2(op,c1',c2')

      | DOPN(op,cl) ->
        let cl' = map_vect evrec cl
        in if is_soapp_operator t then
            let (lam::args) = list_of_vect cl'
            in socontract args lam
           else DOPN(op,cl')

      | DOPL(op,cl) ->
        let cl' = map evrec cl
        in if is_soapp_operator t then
            let (lam::args) = cl'
            in socontract args lam
           else DOPL(op,cl')
;;

let try_soeval = evrec
    where rec evrec t =
    match t with
        DOP0 _ -> t
      | Rel _ -> t
      | VAR _ -> t
      | DOP1(op,c) -> DOP1(op,evrec c)
      | DLAM(na,c) -> DLAM(na,evrec c)
      | DLAMV(na,cl) -> DLAMV(na,map_vect evrec cl)

      | DOP2(op,c1,c2) ->
        let c1' = evrec c1 and
            c2' = evrec c2
        in if is_soapp_operator t then
            try socontract [c2'] c1'
            with (Failure _ | UserError _) -> DOP2(op,c1',c2')
           else DOP2(op,c1',c2')

      | DOPN(op,cl) ->
        let cl' = map_vect evrec cl
        in if is_soapp_operator t then
            let (lam::args) = list_of_vect cl'
            in try socontract args lam
               with (Failure _ | UserError _) -> DOPN(op,cl')
           else DOPN(op,cl')

      | DOPL(op,cl) ->
        let cl' = map evrec cl
        in if is_soapp_operator t then
            let (lam::args) = cl'
            in try socontract args lam
               with (Failure _ | UserError _) -> DOPL(op,cl')
           else DOPL(op,cl')
;;

let soexecute t =
let (_,t) = propagate_names [] t
in soeval t
;;


let try_soexecute t =
let (_,t) = try propagate_names [] t
            with (Failure _ | UserError _) -> ([],t)
in try_soeval t
;;

(* $Id: sosub.ml,v 1.2 1996/02/12 15:37:55 cpaulin Exp $ *)
