(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                 term.ml                                  *)
(****************************************************************************)

(* This module instanciates the structure of generic deBruijn terms to Coq *)

#infix "o";;
#open "std";;
#open "names";;
#open "more_util";;
#open "vectops";;
#open "impuniv";;
#open "generic";;
#open "pp";;
#open "stdpp";;


(********************************************************************)
(*   various utility functions for implementing terms with bindings *)
(********************************************************************)

(* l is a list of pairs (n:nat,x:constr), env is a stack of (na:name,T:constr)
   push_and_lift adds a component to env and lifts l one step *)
let push_and_lift (na,T) env l =
    ((na,T)::env,
     map (fun (n,x) -> (n+1,x)) l)
;;

(* if T is not (x1:A1)(x2:A2)....(xn:An)T' then (push_and_liftl n env T l)
raises an error else it gives ([x1,A1 ; x2,A2 ; ... ; xn,An]@env,T',l')
where l' is l lifted n steps *)
let push_and_liftl n env T l = pushrec n T (env,l)
    where rec pushrec = fun
    0 T (env,l) -> (env,T,l)
  | n (DOP2(Prod,T,DLAM(na,B))) (env,l) 
                -> pushrec (n-1) B (push_and_lift (na,T) env l)
  | n (DOP2(Cast,T,_)) (env,l) -> pushrec n T (env,l)
  | _ _ _ -> error "push_and_liftl"
;;

(* l is a list of pairs (n:nat,x:constr), tlenv is a stack of
(na:name,T:constr), B : constr, na : name
(prod_and_pop ((na,T)::tlenv) B l) gives (tlenv, (na:T)B, l')
where l' is l lifted down one step *)
let prod_and_pop = fun
    [] B l -> error "prod_and_pop"
  | ((na,T)::tlenv) B l ->
    (tlenv,DOP2(Prod,T,DLAM(na,B)),
     map (fun (0,x) -> (0,lift (-1) x)
            | (n,x) -> (n-1,x)) l)
;;

(* recusively applies prod_and_pop :
if env = [na1:T1 ; na2:T2 ; ... ; nan:Tn]@tlenv
then
(prod_and_popl n env T l) gives (tlenv,(nan:Tn)...(na1:Ta1)T,l') where
l' is l lifted down n steps *)
let prod_and_popl n env T l = poprec n (env,T,l)
    where rec poprec = fun
    0 (env,B,l) -> (env,B,l)
  | n ([],_,_) -> error "prod_and_popl"
  | n (env,B,l) -> poprec (n-1) (prod_and_pop env B l)
;;

(* similar to prod_and_pop, but gives [na:T]B intead of (na:T)B *)
let lam_and_pop = fun
    [] B l -> error "lam_and_pop"
  | ((na,T)::tlenv) B l ->
    (tlenv,DOP2(Lambda,T,DLAM(na,B)),
     map (fun (0,x) -> (0,lift (-1) x)
            | (n,x) -> (n-1,x)) l)
;;

(* similar to prod_and_popl but gives [nan:Tan]...[na1:Ta1]B instead of
 * (nan:Tan)...(na1:Ta1)B *)
let lam_and_popl n env T l = poprec n (env,T,l)
    where rec poprec = fun
    0 (env,B,l) -> (env,B,l)
  | n ([],_,_) -> error "lam_and_popl"
  | n (env,B,l) -> poprec (n-1) (lam_and_pop env B l)
;;

(* (nb_lam [na1:T1]...[nan:Tan]c) where c is not an abstraction
 * gives n (casts are ignored) *)
let nb_lam = nbrec 0
    where rec nbrec n = function
    DOP2(Lambda,_,DLAM(_,c)) -> nbrec (n+1) c
  | DOP2(Cast,c,_) -> nbrec n c
  | _ -> n
;;

(* similar to nb_lam, but gives the number of products instead *)
let nb_prod = nbrec 0
    where rec nbrec n = function
    DOP2(Prod,_,DLAM(_,c)) -> nbrec (n+1) c
  | DOP2(Cast,c,_) -> nbrec n c
  | _ -> n
;;

let extract_lifted (n,x) = lift n x;;
let insert_lifted x = (0,x);;

let j_lift n j = {VAL = lift n j.VAL;
                  TYP = lift n j.TYP;
                  KIND = lift n j.KIND};;

let j_pop j = {VAL = pop j.VAL;
               TYP = pop j.TYP;
               KIND = pop j.KIND};;

let j_subst_var id j =
    {VAL = subst_var id j.VAL;
     TYP = subst_var id j.TYP;
     KIND = subst_var id j.KIND};;

let mkVAR id = VAR id;;

let sort_hdchar = function
    Prop(_) -> "P"
  | Type(_,_) -> "T"
;;

let sort_eq_x = fun
    (Prop c1) (Prop c2) -> c1 = c2
  | (Type(c1,u1)) (Type(c2,u2)) -> c1 = c2
  | _ _ -> false;;

(********************************************************************)
(*     Simple term destructuring primitives                         *)
(********************************************************************)

let rec cast_type = function
    DOP2(Cast,c,t) -> t
  | _ -> anomaly "found a type which did not contain a cast"
;;

let rec contents_of_kind = function
    DOP0(Sort(Prop cts)) -> cts
  | DOP0(Sort(Type(cts,u))) -> cts
  | DOP2(Cast,c,t) -> contents_of_kind c
  | _ -> failwith "contents_of_kind not handed a kind"
;;

let rec strip_outer_cast = function
    DOP2(Cast,c,_) -> strip_outer_cast c
  | c -> c
;;

let rec isprop = function
    DOP0(Sort(Prop _)) -> true
  | DOP2(Cast,c,_) -> isprop c
  | _ -> false;;

let rec istype = function
    DOP0(Sort(Type _)) -> true
  | DOP2(Cast,c,_) -> istype c
  | _ -> false;;

let iskind c = isprop c or istype c;;

let same_kind c1 c2 = (isprop c1 & isprop c2) or (istype c1 & istype c2);;


let rec hd_of_prod prod =
  match strip_outer_cast prod with
    (DOP2(Prod,C,DLAM(n,T'))) -> hd_of_prod T'
  |  T -> T
;;

(* Level comparison for information extraction : Prop <= Type *)
let le_kind l m = (isprop l) or (istype m);;

let path_of_abst = function
    (DOPN(Abst sp,_)) -> sp
  | _ -> anomaly "path_of_abst called with bad args"
;;

let args_of_abst = function
    (DOPN(Abst _,args)) -> args
  | _ -> anomaly "args_of_abst called with bad args"
;;

let path_of_const = function
    (DOPN(Const sp,_)) -> sp
  | _ -> anomaly "path_of_const called with bad args"
;;

let args_of_const = function
    (DOPN(Const _,args)) -> args
  | _ -> anomaly "args_of_const called with bad args"
;;

let op_of_mind = function
    (DOPN(MutInd sp,_)) -> sp
  | _ -> anomaly "op_of_mind called with bad args"
;;

let args_of_mind = function
    (DOPN(MutInd _,args)) -> args
  | _ -> anomaly "args_of_mind called with bad args"
;;

let op_of_mconstr = function
    (DOPN(MutConstruct sp,_)) -> sp
  | _ -> anomaly "op_of_mconstr called with bad args"
;;

let args_of_mconstr = function
    (DOPN(MutConstruct _,args)) -> args
  | _ -> anomaly "args_of_mconstr called with bad args"
;;

let isMETA = function (DOP0(Meta _)) -> true | _ -> false;;

let (prop:sorts) = (Prop(Null)) 
and (spec:sorts) = (Prop(Pos)) 
and (typep:sorts) = (Type(Null,Dummy_univ))
and (types:sorts) =  (Type(Pos,Dummy_univ));;


(******************************************************************)
(* Flattening and unflattening of embedded applications and casts *)
(******************************************************************)

(* TODO: Cleanup whd_app *)

let is_appl = function DOPN(AppL,cl) -> true | _ -> false;;

let rec applist = function
    (f,[]) -> f
  | (DOPN(AppL,cl),l2) -> let c = hd_vect cl in 
       if is_appl c then  applist(c,app_tl_vect cl l2)
       else DOPN(AppL,concat_vect cl (vect_of_list l2))
  | (f,l) -> DOPN(AppL,vect_of_list(f::l))

and applistc f l = applist(f,l);;

let rec appvect (f,v) = if vect_length v = 0 then f else match f with
   (DOPN(AppL,cl)) -> let c = hd_vect cl in 
       if is_appl c then  appvect(c,concat_vect (tl_vect cl) v)
       else DOPN(AppL,concat_vect cl v)
  | _ -> DOPN(AppL,cons_vect f v)

and appvectc f l = appvect(f,l);;

let ensure_appl = function
    DOPN(AppL,_) as t -> t
  | t -> DOPN(AppL,[|t|])
;;

(* unflattens application lists *)
let rec telescope_appl = function
    DOPN(AppL,cl) ->
            it_vect_from 1 (fun c e -> DOPN(AppL,[|c;e|])) (hd_vect cl) cl
  | c -> c
;;

(* flattens application lists *)
let rec collapse_appl = function
    DOPN(AppL,cl) -> 
    (collapse_rec(hd_vect cl,list_of_tl_vect cl)
     where rec collapse_rec = function
     (DOPN(AppL,cl),l2) -> collapse_rec(hd_vect cl,app_tl_vect cl l2)
  |  (DOP2(Cast,DOPN(AppL,cl),t),l) -> collapse_rec(DOPN(AppL,cl),l)
  | (f,[]) -> f
  | (f,l) -> let v = vect_of_list (f::l)
             in DOPN(AppL,v))
  | c -> c
;;

(* strips head casts and flattens head applications *)
let strip_head_cast = function
    DOPN(AppL,cl) -> 
    (collapse_rec(hd_vect cl,app_tl_vect cl [])
     where rec collapse_rec = function
     (DOPN(AppL,cl),l2) -> collapse_rec(hd_vect cl,app_tl_vect cl l2)
  |  (DOP2(Cast,c,t),l) -> collapse_rec(c,l)
  | (f,[]) -> f
  | (f,l) -> let v = vect_of_list (f::l)
             in DOPN(AppL,v))
  | c -> c
;;

(* pseudo-reduction rule:
 * [prod_app  s (Prod(_,B)) N --> B[N]
 * with an strip_outer_cast on the first argument to produce a product.
 * if this does not work, then we use the string S as part of our
 * error message.
 *)

let prod_app s T N =
    match strip_outer_cast T with
    DOP2(Prod,_,B) -> SAPP B N
  | _ ->
    errorlabstrm s [< 'S"Needed a product, but didn't find one in " ;
                      'S s ; 'FNL >]
;;
 
(* prod_appvect s T [| a1 ; ... ; an |] -> (T a1 ... an) *)

let prod_appvect s T NL = it_vect (prod_app s) T NL;;

(* prod_applist s T [ a1 ; ... ; an ] -> (T a1 ... an) *)

let prod_applist s T NL = it_list (prod_app s) T NL;;


(* TODO compatibility code, perhaps obsolete *)

(* (lamdba name typ c) abstracts the variable of name "name"
 * and type "typ" in c.

 * (lambda name typ c) -> [name:typ]c

 * (produit name typ c) -> (name:typ)c

 * (apply c c') -> (c c')

 * (occur_const (s:section_path) c) -> true if constant s occurs in c,
 * false otherwise
 *)

let lambda name typ c = DOP2(Lambda,typ,DLAM(Name name,subst_var name c))
and produit name typ c = DOP2(Prod,typ,DLAM(Name name,subst_var name c))
and apply c c' = DOPN(AppL,[|c;c'|])
;;

let occur_const s = occur_opern (Const s);;


(* let sigma be a finite function mapping sections paths to 
   constants represented as (identifier list * constr) option.
   (replace_consts sigma M) unfolds the constants from sigma in term M

   - if (sp,NONE) is in sigma then the constant should not appear in
   term M otherwise replace_consts raises an anomaly ;

   - if (sp,SOME (idl,c)), then the constant sp is replaced by
   c in which the variables given by idl are replaced by the arguments
   of (Const sp), if the number of variables and arguments are not equal
   an anomaly is raised ;

   - if no (sp,_) appears in sigma, then sp is not unfolded.

   NOTE : the case of DOPL is not handled...
*)
let replace_consts const_alist =
  if const_alist = [] then function x -> x else substrec
  where rec substrec = function
    DOPN(Const sp,cl) as c ->
      let cl' = map_vect substrec cl in
      (try (match assoc sp const_alist with
              SOME (hyps,body) ->
                 if length hyps <> vect_length cl then
                   anomaly "found a constant with a bad number of args"
                 else
      	       	   replace_vars
      	       	       (combine(hyps,map_vect_list make_substituend cl')) body
            | NONE -> anomaly ("a constant which was never"^
      	       	       	       " supposed to appear has just appeared"))
       with Not_found -> DOPN(Const sp,cl'))

   | DOP1(i,c)         -> DOP1(i,substrec c)
   | DOPN(oper,cl)     -> DOPN(oper,map_vect substrec cl)
   | DOP2(oper,c1,c2)  -> DOP2(oper,substrec c1,substrec c2)
   | DLAM(na,c)        -> DLAM(na,substrec c)
   | DLAMV(na,v)       -> DLAMV(na,map_vect substrec v)
   | x                 -> x
;;

(* NOTE : the case of DOPL is not handled by whd_castapp_stack *)
let whd_castapp_stack = whrec
  where rec whrec x stack =
    match x with
    DOPN(AppL,cl)  -> whrec (hd_vect cl) (app_tl_vect cl stack)
  | DOP2(Cast,c,_) -> whrec c stack
  | x              -> x,stack
;;

(* whd flattens embedded applications
   (whd_castapp ((((a b) c d) e f g) h)) -> (a b c d e f g h)
   even if some casts exist in ((((a b) c d) e f g) h))
 *)
let whd_castapp x = applist(whd_castapp_stack x []);;


(***************************************)
(*  alpha and eta conversion functions *)                         
(***************************************)

(* alpha conversion : ignore print names and casts *)
let rec eq_constr_rec (M,N) = if eq(M,N) then true else
  match (strip_head_cast M,strip_head_cast N) with
     (DOP2(Cast,c1,_),c2) 	       -> eq_constr_rec(c1,c2)
   | (c1,DOP2(Cast,c2,_))              -> eq_constr_rec(c1,c2)
   | (Rel p1,Rel p2)                   -> p1=p2
   | (DOPN(oper1,cl1),DOPN(oper2,cl2)) ->
      	       	 oper1=oper2 & for_all2eq_vect (curry eq_constr_rec) cl1 cl2
   | (DOP0 oper1,DOP0 oper2)           -> oper1=oper2
   | (DOP1(i,c1),DOP1(j,c2))           -> (i=j) & eq_constr_rec(c1,c2)
   | (DOP2(i,c1,c1'),DOP2(j,c2,c2'))   ->
      	       	 (i=j) & eq_constr_rec(c1,c2) & eq_constr_rec(c1',c2')
   | (DLAM(_,c1),DLAM(_,c2)) 	       -> eq_constr_rec(c1,c2)
   | (DLAMV(_,cl1),DLAMV(_,cl2))       -> 
      	       	 for_all2eq_vect (curry eq_constr_rec) cl1 cl2
   | _ 				       -> false
;;

let eq_constr ((M:constr),(N:constr)) = eq_constr_rec(M,N);;

(****** BUG !!??
(* [x:c1]@(f;x;a1...an) --> @(f;a1...an) *)
let rec eta_reduce c =
    match c with
    DOP2(Lambda,c1,DLAM(_,DOPN(AppL,cl))) ->
    (match cl.(1) with
     Rel 1 -> (if not (dependent (Rel 1) cl.(0)) then
                   eta_reduce (applist(pop cl.(0),
                               map_vect_list pop (cddr_vect cl)))
               else c)
   | _ -> c)
      | _ -> c
;;
******)

(* [x:c1]@(f;a1...an;x) --> @(f;a1...an) *)
let rec eta_reduce c =
  match c with
    DOP2(Lambda,c1,DLAM(_,DOPN(AppL,cl))) ->
      let lastn = (vect_length cl) -1 in 
      if lastn < 1 then c else
      	(match cl.(lastn) with
       	   Rel 1 -> let c' = if lastn=1 then cl.(0) 
                       	                else DOPN(AppL,sub_vect cl 0 lastn) in
              	    if not (dependent (Rel 1) c') then eta_reduce (pop c')
               	       	       	       	       	  else c
         | _     -> c)
  | _ -> c
;;

(* alpha-eta conversion : ignore print names and casts *)

let rec eta_eq_constr pair =
let pair = (eta_reduce (strip_head_cast (fst pair)),eta_reduce (strip_head_cast (snd pair)))
in eq(pair) or match pair with
     (DOP2(Cast,c1,_),c2) -> eta_eq_constr(c1,c2)
   | (c1,DOP2(Cast,c2,_)) -> eta_eq_constr(c1,c2)
   | (Rel p1,Rel p2)                   -> p1=p2
   | (DOPN(oper1,cl1),DOPN(oper2,cl2))         -> oper1=oper2 & for_all2eq_vect (curry eta_eq_constr) cl1 cl2
   | (DOP0 oper1,DOP0 oper2)                 -> oper1=oper2
   | (DOP1(i,c1),DOP1(j,c2)) -> (i=j) & eta_eq_constr(c1,c2)
   | (DOP2(i,c1,c1'),DOP2(j,c2,c2')) -> (i=j) & eta_eq_constr(c1,c2) & eta_eq_constr(c1',c2')
   | (DLAM(_,c1),DLAM(_,c2)) -> eta_eq_constr(c1,c2)
   | (DLAMV(_,cl1),DLAMV(_,cl2)) -> for_all2eq_vect (curry eta_eq_constr) cl1 cl2
   | _ -> false
;;

(***************************)
(*  substitution functions *)                         
(***************************)

(* First utilities for avoiding telescope computation for subst_term *)

let prefix_application k (c:constr) (t:constr) = 
  match (whd_castapp c,whd_castapp t) with
    ((DOPN(AppL,cl1)),DOPN(AppL,cl2)) ->
    let l1 = vect_length cl1
    and l2 = vect_length cl2
    in if l1 <= l2
        & eq_constr (DOPN(AppL,cl1),
                     DOPN(AppL,sub_vect cl2 0 l1)) then
        SOME(DOPN(AppL,cons_vect (Rel k) (sub_vect cl2 l1 (l2 - l1))))
       else NONE
  | (_,_) -> NONE
;;

let prefix_application_eta k (c:constr) (t:constr) = 
  match (whd_castapp c,whd_castapp t) with
    ((DOPN(AppL,cl1)),DOPN(AppL,cl2)) ->
    let l1 = vect_length cl1
    and l2 = vect_length cl2
    in if l1 <= l2
        & eta_eq_constr (DOPN(AppL,cl1),
                     DOPN(AppL,sub_vect cl2 0 l1)) then
        SOME(DOPN(AppL,cons_vect (Rel k) (sub_vect cl2 l1 (l2 - l1))))
       else NONE
  | (_,_) -> NONE
;;

(* Recognizing occurrences of a given subterm in a term for Pattern :
   (subst_term c t) substitutes (Rel 1) for all occurrences of term c 
   in a (closed) term t *)

let subst_term c t = substrec 1 c t
where rec substrec k c t =
match prefix_application k c t with
  SOME x -> x
| NONE ->
    (if eq_constr(t,c) then Rel(k) else match t with
     DOPN(Const sp,cl) -> t
  |  DOPN(MutInd sp,cl) -> t
  |  DOPN(MutConstruct sp,cl) -> t
  |  DOPN(oper,tl)     -> DOPN(oper,map_vect (substrec k c) tl)
  |  DOP1(i,t)         -> DOP1(i,substrec k c t)
  |  DOP2(oper,c1,c2)  -> DOP2(oper,substrec k c c1,substrec k c c2)
  |  DLAM(na,t)        -> DLAM(na,substrec (k+1) (lift 1 c) t)
  |  DLAMV(na,v)       -> DLAMV(na,map_vect (substrec (k+1) (lift 1 c)) v)
  |  _                 -> t);;


(* same as subst_term, but modulo eta *)

let subst_term_eta_eq c t = substrec 1 c t
where rec substrec k c t =
match prefix_application_eta k c t with
  SOME x -> x
| NONE ->
  (if eta_eq_constr(t,c) then Rel(k) else match t with
     DOPN(Const sp,cl) -> t
  |  DOPN(oper,tl)     -> DOPN(oper,map_vect (substrec k c) tl)
  |  DOP1(i,t)         -> DOP1(i,substrec k c t)
  |  DOP2(oper,c1,c2)  -> DOP2(oper,substrec k c c1,substrec k c c2)
  |  DLAM(na,t)        -> DLAM(na,substrec (k+1) (lift 1 c) t)
  |  DLAMV(na,v)       -> DLAMV(na,map_vect (substrec (k+1) (lift 1 c)) v)
  |  _                 -> t);;


(* Substitute only a list of locations locs, the empty list is interpreted
   as substitute all, if 0 is in the list then no substitution is done *)

let subst_term_occ locs c t = if locs = [] then subst_term c t
                              else if mem 0 locs then t
                              else snd (substcheck 1 1 c t)
where rec substcheck k occ c t =
        if exists (function u -> u>=occ) locs then substrec k occ c t
        else (occ,t)
and substrec k occ c t =
   if eq_constr(t,c) then if mem occ locs then (occ+1,Rel(k))
                          else (occ+1,t)
   else match t with
     DOPN(Const sp,tl) -> occ,t
  |  DOPN(MutConstruct sp,tl) -> occ,t
  |  DOPN(MutInd sp,tl) -> occ,t

  |  DOPN(i,cl) -> let (occ',cl') =   
                        it_vect (fun (nocc',lfd) f ->
                         let (nocc'',f') = substcheck k nocc' c f
                         in (nocc'',f'::lfd)) (occ,[]) cl
                        in (occ',DOPN(i,vect_of_list (rev cl')))

  |  DOP2(i,t1,t2) -> let (nocc1,t1')=substrec k occ c t1 in
                        let (nocc2,t2')=substcheck k nocc1 c t2
                        in nocc2,DOP2(i,t1',t2')
  |  DOP1(i,t)  -> let (nocc,t')= substrec k occ c t in
                        nocc,DOP1(i,t')

  |  DLAM(n,t)        -> let (occ',t') = substcheck (k+1) occ (lift 1 c) t
                        in (occ',DLAM(n,t'))

  |  DLAMV(n,cl)      -> let (occ',cl') =   
                        it_vect (fun (nocc',lfd) f ->
                         let (nocc'',f') = substcheck (k+1) nocc' (lift 1 c) f
                         in (nocc'',f'::lfd)) (occ,[]) cl
                        in (occ',DLAMV(n,vect_of_list (rev cl') ))
  |  _               -> occ,t
  ;;

(***************************)
(* occurs check functions  *)                         
(***************************)


let rec occur_meta = function
   DOP2(Prod,t,DLAM(_,c))   -> (occur_meta t) or (occur_meta c)
 | DOP2(Lambda,t,DLAM(_,c)) -> (occur_meta t) or (occur_meta c)
 | DOPN(_,cl)    -> (exists_vect occur_meta cl)
 | DOP2(Cast,c,t)     -> occur_meta c or occur_meta t
 | DOP0(Meta(_))       -> true
 | _             -> false;;


let assumptions_for_print lna = ENVIRON(([],[]),map (fun na -> (na,())) lna);;

let rel_vect = (generic__rel_vect : int -> int -> constr vect);;

(* $Id: term.ml,v 1.6 1996/02/12 15:38:01 cpaulin Exp $ *)
