(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                 clenv.ml                                 *)
(****************************************************************************)

#infix "o";;
#open "std";;
#open "initial";;
#open "names";;
#open "vectops";;
#open "generic";;
#open "term";;
#open "himsg";;
#open "proof";;
#open "proof_trees";;
#open "reduction";;
#open "tacmach";;
#open "termenv";;
#open "pp";;
#open "stdpp";;
#open "more_util";;

let new_existential_in_sign hyps =
    let ev_sp = make_path(PA__intern["evar"]) (make_ident(SA__intern "?") (newMETA())) CCI
    in DOPN(Const ev_sp,
            vect_of_list (map (fun id -> VAR id) (ids_of_sign hyps)))
;;

let occur_existential = occrec
    where rec occrec = function
    DOPN(Const sp,_) as K -> is_existential K
  | DOPN(_,cl) -> exists_vect occrec cl
  | DOPL(_,cl) -> exists occrec cl
  | DOP2(_,c1,c2) -> occrec c1 or occrec c2
  | DOP1(_,c) -> occrec c
  | DOP0 _ -> false
  | DLAM(_,c) -> occrec c
  | DLAMV(_,cl) -> exists_vect occrec cl
  | Rel _ -> false
  | VAR _ -> false
;;  

let error_conflict(m,n) =
errorlabstrm "simpl" [< 'S"Impossible to unify"; 'SPC ; pTERM m;
                        'SPC ; 'S"With"; 'SPC ; pTERM n >];;

let rec whd_deltat wc = function
    (DOPN(Const _,_) as K) ->
    if translucent_const (w_Underlying wc) K then
        whd_deltat wc (w_const_value wc K)
    else collapse_appl K
  | t -> collapse_appl t
;;

let ApplyHead n c wc = apprec n c (w_type_of wc c) wc
    where rec apprec n c cty wc =
    if n = 0 then (wc,c)
    else match w_whd_betadeltaiota wc cty with
        DOP2(Prod,c1,c2) ->
        let c1ty = w_type_of wc c1 in
        let (DOPN(Const evar_sp,_)) as evar = new_existential_in_sign (w_hyps wc)
        in ((apprec (n-1) (applist(c,[evar])) (SAPP c2 evar))
            o (w_Declare evar_sp (DOP2(Cast,c1,c1ty)))) wc

      | _ -> error "Apply_Head_Then"
;;

let mimick_evar hdc nargs sp wc =
    (w_Focusing_THEN sp
        (ApplyHead nargs hdc)
        (fun c wc -> w_Define sp c wc)) wc
;;



let rec unify mc wc M N =
    if not(occur_meta M) & w_conv_x wc M N then (mc,[])
    else unirec_rec wc (mc,[]) M N
and unirec_rec wc substn M N =
    let cM = whd_deltat wc M
    and cN = whd_deltat wc N
    in try match (cM,cN) with
    DOP2(Cast,c,_),t2 -> (unirec_rec wc) substn c t2
  | t1,DOP2(Cast,c,_) -> (unirec_rec wc) substn t1 c
  | DOP0(Meta k),_    -> (k,cN)::(fst substn),(snd substn)
  | cM,DOP0(Meta(k))    -> (k,cM)::(fst substn),(snd substn)
  | DOP2(Lambda,t1,DLAM(_,c1)),DOP2(Lambda,t2,DLAM(_,c2)) ->
    (unirec_rec wc) ((unirec_rec wc) substn t1 t2) c1 c2
  | DOP2(Prod,t1,DLAM(_,c1)),DOP2(Prod,t2,DLAM(_,c2)) ->
    (unirec_rec wc) ((unirec_rec wc) substn t1 t2) c1 c2

  | DOPN(AppL,cl1),DOPN(AppL,cl2) ->
      let len1 = vect_length cl1
      and len2 = vect_length cl2
      in
      if len1 = len2 then
          it_vect2 (unirec_rec wc) substn cl1 cl2

      else if len1 < len2 then
          let extras,restcl2 = chop_vect ((len2-len1)+1) cl2
          in it_vect2 (unirec_rec wc) ((unirec_rec wc) substn (hd_vect cl1) (DOPN(AppL,extras)))
                                (tl_vect cl1) restcl2

      else let extras,restcl1 = chop_vect ((len1-len2)+1) cl1
           in it_vect2 (unirec_rec wc) ((unirec_rec wc) substn (DOPN(AppL,extras)) (hd_vect cl2))
                                 restcl1 (tl_vect cl2)

    | DOPN(MutCase,cl1),DOPN(MutCase,cl2) ->
      if vect_length cl1 = vect_length cl2 then
          it_vect2 (unirec_rec wc) substn cl1 cl2
      else error_conflict(M,N)

    | DOPN(MutConstruct _,_),DOPN(MutConstruct _,_) ->
      if w_conv_x wc cM cN then substn else (error_conflict(M,N))

    | DOPN(MutInd _,_),DOPN(MutInd _,_) ->
       if w_conv_x wc cM cN then substn else (error_conflict(M,N))

    | (DOPN(Const _,_)),(DOPN(Const _,_)) ->
      if is_existential cM then (fst substn),((cM,cN)::(snd substn))
      else if is_existential cN then (fst substn),((cN,cM)::(snd substn))
      else if w_conv_x wc cM cN then substn
      else (error_conflict(M,N))

    | (DOPN(Const _,_)),_ ->
      if is_existential cM then (fst substn),((cM,cN)::(snd substn))
      else if w_conv_x wc cM cN then substn
      else (error_conflict(M,N))

    | _,(DOPN(Const _,_)) ->
      if is_existential cN then (fst substn),((cN,cM)::(snd substn))
      else if not(occur_meta cM) & w_conv_x wc cM cN then substn
      else (error_conflict(M,N))

    | _ -> (error_conflict(M,N))
 with UserError s -> 
      if not(occur_meta cM) & w_conv_x wc cM cN then substn
      else raise (UserError s);;

let whd_castappdeltat_stack sigma = whrec
    where rec whrec x l =
    match x with
    (DOPN(Const _,_)) as c      ->
    if translucent_const sigma c then
           whrec (const_value sigma c) l
    else x,l
  | DOP2(Cast,c,_) -> whrec c l
  | DOPN(AppL,cl)    -> whrec (hd_vect cl) (app_tl_vect cl l)
  | x             -> x,l;;

let whd_castappdeltat sigma c = applist(whd_castappdeltat_stack sigma c []);;

let w_whd wc c = whd_castappdeltat (w_Underlying wc) c;;

#infix "ORELSE";;
#infix "THENS";;
#infix "THEN";;

(* Unification
 *
 * Procedure:
 * (1) The function [unify mc wc M N] produces two lists:
 *     (a) a list of bindings Meta->RHS
 *     (b) a list of bindings EVAR->RHS
 *
 * The Meta->RHS bindings cannot themselves contain
 * meta-vars, so they get applied eagerly to the other
 * bindings.  This may or may not close off all RHSs of
 * the EVARs.  For each EVAR whose RHS is closed off,
 * we can just apply it, and go on.  For each which
 * is not closed off, we need to do a mimick step -
 * in general, we have something like:
 *
 *      ?X == (c e1 e2 ... ei[Meta(k)] ... en)
 *
 * so we need to do a mimick step, converting ?X
 * into
 *
 *      ?X -> (c ?z1 ... ?zn)
 *
 * of the proper types.  Then, we can decompose the
 * equation into
 *
 *      ?z1 --> e1
 *          ...
 *      ?zi --> ei[Meta(k)]
 *          ...
 *      ?zn --> en
 *
 * and keep on going.  Whenever we find that a R.H.S.
 * is closed, we can, as before, apply the constraint
 * directly.  Whenever we find an equation of the form:
 *
 *      ?z -> Meta(n)
 *
 * we can reverse the equation, put it into our metavar
 * substitution, and keep going.
 *
 * The most efficient mimick possible is, for each
 * Meta-var remaining in the term, to declare a
 * new EVAR of the same type.  This is supposedly
 * determinable from the clausale form context -
 * we look up the metavar, take its type there,
 * and apply the metavar substitution to it, to
 * close it off.  But this might not always work,
 * since other metavars might also need to be resolved.
 *)

#infix "w_ORELSE";;

let rec w_Unify M N mc wc =
let (mc',ec') = unify mc wc M N
in w_resrec mc' ec' wc

and w_resrec metas evars wc =
    match evars with
    [] -> (wc,metas)

  | (lhs,(DOP0(Meta k) as rhs))::t -> w_resrec ((k,lhs)::metas) t wc

  | (DOPN(Const evsp,_) as evar,rhs)::t ->
    if w_defined_const wc evar then
        let (wc',metas') = w_Unify rhs evar metas wc
        in w_resrec metas' t wc'
    else
        try w_resrec metas t (w_Define evsp rhs wc)
        with UserError _ ->
            (match rhs with
             DOPN(AppL,cl) ->
             (match cl.(0) with
              DOPN(Const sp,_) ->
              let wc' = mimick_evar cl.(0) ((vect_length cl) - 1) evsp wc
              in w_resrec metas evars wc'
            | _ -> error "w_Unify")
           | _ -> error "w_Unify")

;;

let UnifyTerms M N = Walking (fst o (w_Unify M N []));;

let Unify M gls =
let N = pf_concl gls
in UnifyTerms M N gls
;;

let pr_clenv clenv =
let pr_name mv =
    try let id = intavm__map clenv.namenv mv
        in [< 'S"[" ; print_id id ; 'S"]" >]
    with Not_found -> [< >]
in
    [< 'S"TEMPL: " ; pTERM clenv.templval.rebus ;
       'S" : " ; pTERM clenv.templtyp.rebus ; 'FNL ;
       (prlist
        (fun (mv,CLTYP b) -> HOV 0 [< 'INT mv ; pr_name mv ; 'S " : " ; pTERM b.rebus ; 'FNL >]
           | (mv,CLVAL(b,_)) -> HOV 0 [< 'INT mv ; pr_name mv ; 'S " := " ; pTERM b.rebus ; 'FNL >])
        (intavm__toList clenv.env)) >];;

(* collects all metavar occurences, in left-to-right order, preserving
 * repetitions and all.
 *)
let collect_metas c = collrec c []
    where rec collrec c acc =
    match c with
    DOP0(Meta mv)  -> mv::acc
  | (DOP1(oper,c)) -> collrec c acc
  | (DOP2(oper,c1,c2)) -> collrec c1 (collrec c2 acc)
  | DOPN(oper,cl)      -> vect_it collrec cl acc
  | DLAM(_,c)     ->  collrec c acc
  | DLAMV(_,v)    ->  vect_it collrec v acc
  | _                     -> acc;;

let metavars_of c =  collrec c (intavs__mt)
    where rec collrec c acc =
    match c with
    DOP0(Meta mv)  -> intavs__add acc mv
  | (DOP1(oper,c)) -> collrec c acc
  | (DOP2(oper,c1,c2)) -> collrec c1 (collrec c2 acc)
  | DOPN(oper,cl)      -> vect_it collrec cl acc
  | DLAM(_,c)     ->  collrec c acc
  | DLAMV(_,v)    ->  vect_it collrec v acc
  | _                     -> acc;;


let mk_freelisted c =
    {rebus = c;freemetas = metavars_of c}
;;

(* [mentions clenv mv0 mv1] is true if mv1 is defined and mentions
 * mv0, or if one of the free vars on mv1's freelist mentions
 * mv0
 *)
let mentions clenv mv0 = menrec
    where rec menrec mv1 =
    try
    (match intavm__map clenv.env mv1 with
    CLVAL (b,_) -> (intavs__memb b.freemetas mv0 or intavs__exists menrec b.freemetas)
  | CLTYP _ -> false)
    with Not_found -> false
;;

(* Creates a new clause-environment, whose template has a given
 * type, CTY.  This is not all that useful, since not very often
 * does one know the type of the clause - one usually only has
 * a clause which one wants to backchain thru.
 *)
let mk_clenv wc cty =
let mv = newMETA() in
let cty_fls = mk_freelisted cty
in {templval = mk_freelisted(DOP0(Meta mv));
    templtyp = cty_fls;
    namenv = intavm__new ();
    env =  intavm__add (intavm__new ()) (mv,CLTYP cty_fls);
    hook = wc};;

let clenv_environments bound c = 
 let rec clrec (ne,e,mvs) = fun
     0 hd -> (ne,e,rev mvs,hd)
   | n (DOP2(Cast,c,_)) -> clrec (ne,e,mvs) (n-1) c
   | n (DOP2(Prod,c1,DLAM(na,c2))) ->
     let mv = newMETA() and
         dep = dependent (Rel 1) c2 in

     let ne' =
         if dep then
             (match na with
              Anonymous -> ne
            | Name id ->
              if intavm__in_dom ne mv then
                  (warning ("Cannot put metavar "^(string_of_int mv)^" in name-environment twice");
                   ne)
              else intavm__add ne (mv,id))
         else ne and

     e' = intavm__add e (mv,CLTYP (mk_freelisted c1))
             
     in clrec (ne',e',mv::mvs) (n-1) (if dep then (SAPP (DLAM(na,c2)) (DOP0(Meta mv)))
                        else c2)
   | n hd -> (ne,e,rev mvs,hd)
         
 in clrec (intavm__new(),intavm__new(),[]) bound c
;;

let mk_clenv_from wc (c,cty) =
let (namenv,env,args,concl) = clenv_environments (-1) cty
in {templval = mk_freelisted(DOPN(AppL,vect_of_list (c::(map (fun mv -> DOP0(Meta mv)) args))));
    templtyp = mk_freelisted concl;
    namenv = namenv;
    env = env;
    hook = wc}
;;

let bounded_mk_clenv_from bound wc (c,cty) =
let (namenv,env,args,concl) = clenv_environments bound cty
in {templval = mk_freelisted(DOPN(AppL,vect_of_list (c::(map (fun mv -> DOP0(Meta mv)) args))));
    templtyp = mk_freelisted concl;
    namenv = namenv;
    env = env;
    hook = wc}
;;

let connect_clenv wc clenv =
    {templval = clenv.templval;
     templtyp = clenv.templtyp;
     namenv = clenv.namenv;
     env = clenv.env;
     hook = wc}
;;

let mk_clenv_hnf_constr_type_of wc t =
    mk_clenv_from wc (t,w_hnf_constr wc (w_type_of wc t));;

(****
let mk_clenv_rename_type_of wc t =
    mk_clenv_from wc (t,rename_bound_var (ids_of_sign (w_hyps wc)) (w_type_of wc t));;


let mk_clenv_rename_hnf_constr_type_of wc t =
    mk_clenv_from wc (t,rename_bound_var (ids_of_sign (w_hyps wc)) (w_hnf_constr wc (w_type_of wc t)));;
****)

let mk_clenv_rename_from wc (c,t) 
    = mk_clenv_from wc (c,rename_bound_var [] t);;

let mk_clenv_rename_type_of wc t =
    mk_clenv_from wc (t,rename_bound_var [] (w_type_of wc t));;


let mk_clenv_rename_hnf_constr_type_of wc t =
    mk_clenv_from wc (t,rename_bound_var [] (w_hnf_constr wc (w_type_of wc t)));;

let mk_clenv_type_of wc t = mk_clenv_from wc (t,w_type_of wc t);;

let mk_clenv_printable_type_of = mk_clenv_type_of;;

let clenv_assign mv rhs clenv =
let rhs_fls = mk_freelisted rhs
in if intavs__exists (mentions clenv mv) rhs_fls.freemetas then
    error "clenv__assign: circularity in unification"
   else
   try
   (match intavm__map clenv.env mv with
   CLVAL _ -> error "clenv_assign"
 | CLTYP bty -> 
   {templval = clenv.templval;
    templtyp = clenv.templtyp;
    namenv = clenv.namenv;
    env = intavm__remap clenv.env mv (CLVAL(rhs_fls,bty));
    hook = clenv.hook})
   with Not_found -> error "clenv_assign"
;;

let clenv_val_of clenv mv = valrec mv
    where rec valrec mv =
    try
    (match intavm__map clenv.env mv with
    CLTYP _ -> DOP0(Meta mv)
  | CLVAL(b,_) ->
    instance (map (fun mv' -> (mv',valrec mv')) (intavs__toList b.freemetas)) b.rebus)
    with Not_found -> DOP0(Meta mv)
;;

let clenv_instance clenv b =
let c_sigma = map (fun mv -> (mv,clenv_val_of clenv mv)) (intavs__toList b.freemetas)
in instance c_sigma b.rebus;;

let clenv_instance_term clenv c =
    clenv_instance clenv (mk_freelisted c);;

(* put casts around meta variables with closed types *)

let clenv_cast_meta clenv = crec where rec crec = function 
    (DOP0(Meta mv) as u) -> 
      (try match intavm__map clenv.env mv with
                 CLTYP b -> let b' = clenv_instance clenv b in 
                 if occur_meta b' then u else DOP2(Cast,DOP0(Meta mv),b')
               | CLVAL(_) -> u
       with Not_found -> u)
  | DOP2(Cast,DOP0(Meta _),_) as u -> u
  | DOPN(c,cl)                     -> DOPN(c,map_vect crec cl)
  | DOPL(c,cl)                     -> DOPL(c,map crec cl)
  | DOP1(c,t)                      -> DOP1(c,crec t)
  | DOP2(c,t1,t2)                  -> DOP2(c,crec t1,crec t2)
  | DLAM(n,c)                      -> DLAM(n,crec c)
  | DLAMV(n,cl)                    -> DLAMV(n,map_vect crec cl)
  | x                              -> x
;;

(* [clenv_pose (na,mv,cty) clenv]
 * returns a new clausenv which has added to it the metavar MV,
 * with type CTY.  the name NA, if it is not ANONYMOUS, will
 * be entered into the name-map, as a way of accessing the new
 * metavar.
 *)
let clenv_pose (na,mv,cty) clenv =
    {templval = clenv.templval;
     templtyp = clenv.templtyp;
     env = intavm__add clenv.env (mv,CLTYP (mk_freelisted cty));
     namenv = (match na with
               Anonymous -> clenv.namenv
             | Name id -> intavm__add clenv.namenv (mv,id));
     hook = clenv.hook}
;;

let clenv_defined clenv mv =
    match intavm__map clenv.env mv with
    CLVAL _ -> true
  | CLTYP _ -> false
;;

let clenv_value clenv mv =
    match intavm__map clenv.env mv with
    CLVAL(b,_) -> b
  | CLTYP _ -> failwith "clenv_value";;
    
let clenv_type clenv mv =
    match intavm__map clenv.env mv with
    CLTYP b -> b
  | CLVAL(_,b) -> b;;

let clenv_template clenv = clenv.templval;;
let clenv_template_type clenv = clenv.templtyp;;

let clenv_instance_value clenv mv =
    clenv_instance clenv (clenv_value clenv mv);;

let clenv_instance_type clenv mv =
    clenv_instance clenv (clenv_type clenv mv);;

let clenv_instance_template clenv =
    clenv_instance clenv (clenv_template clenv);;

let clenv_instance_template_type clenv =
    clenv_instance clenv (clenv_template_type clenv);;

let clenv_wtactic wt clenv =
    {templval = clenv.templval;
     templtyp = clenv.templtyp;
     namenv = clenv.namenv;
     env = clenv.env;
     hook = wt clenv.hook}
;;

(* [clenv_unify gls M N clenv]
 * peforms a unification of M and N, generating a bunch of
 * unification constraints in the process.  These constraints
 * are processed, one-by-one - they may either generate new
 * bindings, or, if there is already a binding, new unifications,
 * which themselves generate new constraints.  This continues
 * until we get failure, or we run out of constraints.
 *)

let rec clenv_unify M N clenv =
let (mc,ec) = unify [] clenv.hook M N
in clenv_resrec mc ec clenv

and clenv_resrec metas evars clenv =
    match (evars,metas) with
    ([],[]) -> clenv

  | ((lhs,(DOP0(Meta k) as rhs))::t,metas) ->
    clenv_resrec ((k,lhs)::metas) t clenv

  | ((DOPN(Const evsp,_) as evar,rhs)::t,metas) ->
    if w_defined_const clenv.hook evar then
        let (metas',evars') = unify [] clenv.hook rhs evar
        in clenv_resrec (metas'@metas) (evars'@t) clenv
    else
        (try clenv_resrec metas t (clenv_wtactic (w_Define evsp rhs) clenv)
        with UserError _ ->
            (match rhs with
             DOPN(AppL,cl) ->
             (match cl.(0) with
              (DOPN(Const _,_) | DOPN(MutConstruct _,_))
                   ->
              clenv_resrec metas evars (clenv_wtactic (mimick_evar cl.(0) ((vect_length cl) - 1) evsp) clenv)
            | _ -> error "w_Unify")

           | _ -> error "w_Unify"))

  | ([],(mv,N)::t) ->
    if clenv_defined clenv mv then
        let (metas',evars') =
            unify [] clenv.hook (clenv_value clenv mv).rebus N
        in clenv_resrec (metas'@t) evars' clenv
    else clenv_resrec t [] (clenv_assign mv N clenv)
;;


(* [clenv_bchain mv clenv' clenv]
 *
 * Resolves the value of "mv" (which must be undefined) in clenv to be
 * the template of clenv' be the value "c", applied to "n" fresh
 * metavars, whose types are chosen by destructing "clf", which should
 * be a clausale forme generated from the type of "c".  The process of
 * resolution can cause unification of already-existing metavars, and
 * of the fresh ones which get created.  This operation is a composite
 * of operations which pose new metavars, perform unification on
 * terms, and make bindings.  *)

let clenv_bchain mv subclenv clenv =
(* Add the metavars of [subclenv] to [clenv], with
   their name-environment *)
let clenv' = {templval = clenv.templval;
              templtyp = clenv.templtyp;
               namenv =
               it_list (fun ne (mv,id) ->
                            if clenv_defined subclenv mv then ne
                            else if (intavm__in_dom ne mv) then
                                (warning ("Cannot put metavar "^(string_of_int mv)^" in name-environment twice");ne)
                            else intavm__add ne (mv,id))
                       clenv.namenv (intavm__toList subclenv.namenv);
               env = it_list intavm__add clenv.env (intavm__toList subclenv.env);
               hook = clenv.hook} in

(* unify the type of the template of [subclenv] with the type of [mv] *)
let clenv'' = clenv_unify (clenv_instance clenv' (clenv_template_type subclenv))
                          (clenv_instance_type clenv' mv)
                          clenv' in

(* assign the metavar *)
let clenv''' = clenv_assign mv (clenv_instance clenv' (clenv_template subclenv)) clenv''

in clenv'''
;;

(* swaps the "hooks" in [clenv1] and [clenv2], so we can then use
   backchain to hook them together *)

let clenv_swap clenv1 clenv2 =
let clenv1' = {templval = clenv1.templval;
               templtyp = clenv1.templtyp;
               namenv = clenv1.namenv;
               env = clenv1.env;
               hook = clenv2.hook} and

    clenv2' = {templval = clenv2.templval;
               templtyp = clenv2.templtyp;
               namenv = clenv2.namenv;
               env = clenv2.env;
               hook = clenv1.hook}

in (clenv1',clenv2')
;;

let clenv_fchain mv nextclenv clenv =
    let (clenv',nextclenv') = clenv_swap clenv nextclenv
    in clenv_bchain mv clenv' nextclenv'
;;

let clenv_refine KONT clenv gls =
    (KONT clenv.hook
     THEN (Refine (clenv_instance_template clenv))) gls;;


let clenv_refine_cast KONT clenv gls =
    (KONT clenv.hook
     THEN (Refine (clenv_cast_meta clenv (clenv_instance_template clenv))))
     gls;;


(* takes a substitution s, an open term op and a closed term cl
   try to find a subterm of cl which matches op, if op is just a Meta
   FAIL because we cannot find a binding
 *)

(****
let constrain_clenv_to_subterm clause =
    function (DOP0(Meta(_)) as op),_ -> clause,op
            | op,cl            -> matchrec cl
    where rec matchrec cl =
    let cl = strip_outer_cast cl
    in try if closed cl then clenv_unify op cl clause,cl
           else error "Bound 1"
    with UserError(_)->
        match telescope_appl cl with DOPN(AppL,tl) ->
        if vect_length tl = 2 then
            let c1 = tl.(0) and c2 = tl.(1) in
            (try matchrec c1 with UserError(_) -> matchrec c2)
        else error "Match_subterm"
                    | DOP2(Prod,t,DLAM(_,c)) ->
            (try matchrec t with UserError(_) -> matchrec c)
                    | DOP2(Lambda,t,DLAM(_,c)) ->
            (try matchrec t with UserError(_) -> matchrec c)
                    | _ -> error "Match_subterm"
;;
*****)


let constrain_clenv_to_subterm clause =
    function (DOP0(Meta(_)) as op),_ -> error "Match_subterm"
            | op,cl            -> matchrec cl
    where rec matchrec cl =
    let cl = strip_outer_cast cl
    in try if closed cl then clenv_unify op cl clause,cl
           else error "Bound 1"
    with UserError(_)->
        match telescope_appl cl with DOPN(AppL,tl) ->
        if vect_length tl = 2 then
            let c1 = tl.(0) and c2 = tl.(1) in
            (try matchrec c1 with UserError(_) -> matchrec c2)
        else error "Match_subterm"
                    | DOP2(Prod,t,DLAM(_,c)) ->
            (try matchrec t with UserError(_) -> matchrec c)
                    | DOP2(Lambda,t,DLAM(_,c)) ->
            (try matchrec t with UserError(_) -> matchrec c)
                    | _ -> error "Match_subterm"
;;

let constrain_clenv_using_subterm_list allow_K clause oplist t = 
    list_it (fun op (clause,l) -> 
                if occur_meta op then 
                    let (clause',cl) =
                        try constrain_clenv_to_subterm clause (strip_outer_cast op,t)
                        with (UserError _ as e) ->
                            if allow_K then (clause,op) else raise e
                    in (clause',cl::l)
                    else (clause,op::l)) oplist (clause,[]);;


(* [clenv_metavars clenv mv]
 * returns a list of the metavars which appear in the type of
 * the metavar mv.  The list is unordered.
 *)
let clenv_metavars clenv mv =
    match intavm__map clenv.env mv with
    CLVAL(_,b) -> b.freemetas
  | CLTYP b -> b.freemetas
;;

let clenv_template_metavars clenv = clenv.templval.freemetas;;

(* [clenv_dependent clenv cval ctyp]
 * returns a list of the metavariables which appear in the term cval,
 * and which are dependent, This is computed by taking the metavars in cval,
 * in right-to-left order, and collecting the metavars which appear
 * in their types, and adding in all the metavars appearing in ctyp, the
 * type of cval.
 *)

let dependent_metas clenv mvs conclmetas =
    list_it
    (fun mv deps ->
         intavs__union deps (clenv_metavars clenv mv))
    mvs conclmetas
;;

let clenv_dependent clenv (cval,ctyp) =
let mvs = collect_metas (clenv_instance clenv cval) in
let ctyp_mvs = metavars_of (clenv_instance clenv ctyp) in
let deps = dependent_metas clenv mvs ctyp_mvs
in filter (intavs__memb deps) mvs
;;

(* [clenv_independent clenv (cval,ctyp)]
 * returns a list of metavariables which appear in the term cval,
 * and which are not dependent.  That is, they do not appear in
 * the types of other metavars which are in cval, nor in the type
 * of cval, ctyp.
 *)
let clenv_independent clenv (cval,ctyp) =
let mvs = collect_metas (clenv_instance clenv cval) in
let ctyp_mvs = metavars_of (clenv_instance clenv ctyp) in
let deps = dependent_metas clenv mvs ctyp_mvs
in filter (fun mv -> not(intavs__memb deps mv)) mvs
;;

(* [clenv_missing clenv (cval,ctyp)]
 * returns a list of the metavariables which appear in the term cval,
 * and which are dependent, and do NOT appear in ctyp.
 *)
let clenv_missing clenv (cval,ctyp) =
let mvs = collect_metas (clenv_instance clenv cval) in
let ctyp_mvs = metavars_of (clenv_instance clenv ctyp) in
let deps = dependent_metas clenv mvs ctyp_mvs
in filter (fun n -> intavs__memb deps n
               & not(intavs__memb ctyp_mvs n))
    mvs
;;

let clenv_constrain_missing_args mlist clause =
if mlist = [] then clause else
let occlist = clenv_missing clause (clenv_template clause,
                                   (clenv_template_type clause))
in if length occlist = length mlist then
    it_list2 (fun clenv occ m -> clenv_unify (DOP0(Meta occ)) m clenv)
             clause occlist mlist
   else error "Not the right number of missing arguments"
;;

let clenv_constrain_dep_args mlist clause =
if mlist = [] then clause else
let occlist = clenv_dependent clause (clenv_template clause,
                                      (clenv_template_type clause))
in if length occlist = length mlist then
    it_list2 (fun clenv occ m -> clenv_unify (DOP0(Meta occ)) m clenv)
             clause occlist mlist
   else error "clenv_constrain_dep_args: Not the right number of dependent arguments"
;;


let clenv_constrain_dep_args_of mv mlist clause =
if mlist = [] then clause else
let occlist = clenv_dependent clause (clenv_value clause mv,
                                          clenv_type clause mv)
in if length occlist = length mlist then
    it_list2 (fun clenv occ m -> clenv_unify (DOP0(Meta occ)) m clenv)
             clause occlist mlist
   else error "clenv_constrain_dep_args_of: Not the right number of dependent arguments"
;;

let clenv_lookup_name clenv id =
let numbers = intavm__inv clenv.namenv id
in if length numbers = 0 then
    errorlabstrm "clenv_lookup_name" [< 'S"No such variable" >]
   else numbers
;;

let clenv_match_args s clause =
let mvs = clenv_independent clause (clenv_template clause,
                                    clenv_template_type clause)
in matchrec clause s
    where rec matchrec clause = function
    [] -> clause
  | (b,c)::t ->
    let k = (match b with
             Dep s -> let l = clenv_lookup_name clause s
                      in if length l > 1 then
                          error "Multiple instances of this name"
                         else hd l
           | NoDep n -> (try nth mvs n
                        with Failure "item" -> errorlabstrm "clenv_match_args"
                             [< 'S"No such binder" >])
           | Com      -> errorlabstrm "clenv_match_args"
                [< 'S"Cannot mix binding and non binding associations" >])
    in 
    let k_typ = w_hnf_constr clause.hook (clenv_instance_type clause k)
    and c_typ = w_hnf_constr clause.hook (w_type_of clause.hook c)
    in matchrec (clenv_assign k c (clenv_unify k_typ c_typ clause)) t
;;

let rec subterm = fun
    c [] -> c
  | (DOPN(AppL,cl)) (n::t) -> if n < vect_length cl then subterm cl.(n) t
                            else error "subterm"
  | _ _ -> error "subterm"
;;

let rec rev_subterm = fun
    c [] -> c
  | (DOPN(AppL,cl)) (n::t) -> if n < vect_length cl then rev_subterm cl.((vect_length cl - 1) - n) t
                            else error "rev_subterm"
  | _ _ -> error "rev_subterm"
;;

(* [clenv_pose_dependent_evars clenv]
 * For each dependent evar in the clause-env which does not have a value,
 * pose a value for it by constructing a fresh evar.  We do this in
 * left-to-right order, so that every evar's type is always closed w.r.t.
 * metas.
 *)
let clenv_pose_dependent_evars clenv =
let dep_mvs = clenv_dependent clenv (clenv_template clenv,
                                     clenv_template_type clenv)
in it_list
    (fun clenv mv ->
         let (DOPN(Const evar_sp,_)) as evar = new_existential_in_sign (w_hyps clenv.hook) in
         let TY = clenv_instance_type clenv mv in
         let TYty = w_type_of clenv.hook TY in
         let clenv' = clenv_wtactic (w_Declare evar_sp (DOP2(Cast,TY,TYty))) clenv
         in clenv_assign mv evar clenv')
    clenv
    dep_mvs
;;

let clenv_add_sign (id,sign) clenv =
  {templval = clenv.templval;
   templtyp = clenv.templtyp;
   namenv = clenv.namenv;
   env = clenv.env;
   hook = w_add_sign (id,sign) clenv.hook}
;;

let clenv_type_of ce c =
let metamap =
    map (fun (n,CLVAL(_,typ)) -> (n,typ.rebus)
           | (n,CLTYP typ) -> (n,typ.rebus)) (intavm__toList ce.env) in
    (mach__fexemeta (w_Underlying ce.hook) metamap (GLOB(w_hyps ce.hook)) c).TYP
;;

let clenv_instance_type_of ce c =
  clenv_instance ce (mk_freelisted (clenv_type_of ce c));;

(* [clenv_typed_unify gls M N clenv]
 * peforms a unification of M and N, generating a bunch of
 * unification constraints in the process.  These constraints
 * are processed, one-by-one - they may either generate new
 * bindings, or, if there is already a binding, new unifications,
 * which themselves generate new constraints.  This continues
 * until we get failure, or we run out of constraints.
 *)

let rec clenv_typed_unify M N clenv =
let (mc,ec) = unify [] clenv.hook M N
in clenv_resrec mc ec clenv

and clenv_resrec metas evars clenv =
    match (evars,metas) with
    ([],[]) -> clenv

  | ((lhs,(DOP0(Meta k) as rhs))::t,metas) ->
    clenv_resrec ((k,lhs)::metas) t clenv

  | ((DOPN(Const evsp,_) as evar,rhs)::t,metas) ->
    if w_defined_const clenv.hook evar then
        let (metas',evars') = unify [] clenv.hook rhs evar
        in clenv_resrec (metas'@metas) (evars'@t) clenv
    else
        (try clenv_resrec metas t (clenv_wtactic (w_Define evsp rhs) clenv)
        with UserError _ ->
            (match rhs with
             DOPN(AppL,cl) ->
             (match cl.(0) with
              (DOPN(Const _,_) |  DOPN(MutConstruct _,_))
               ->
              clenv_resrec metas evars (clenv_wtactic (mimick_evar cl.(0) ((vect_length cl) - 1) evsp) clenv)
            | _ -> error "w_Unify")
           | _ -> error "w_Unify"))

  | ([],(mv,N)::t) ->
    if clenv_defined clenv mv then
        let (metas',evars') =
            unify [] clenv.hook (clenv_value clenv mv).rebus N
        in clenv_resrec (metas'@t) evars' clenv
    else 
    let Nty = clenv_type_of clenv (clenv_instance clenv (mk_freelisted N)) in
    let (mc,ec) = (try unify [] clenv.hook (clenv_instance_type clenv mv) Nty 
                   with UserError _ -> ([],[])) in
      clenv_resrec (mc@t) ec (clenv_assign mv N clenv)
;;

(* The unique unification algorithm works like this: If the pattern is
   flexible, and the goal has a lambda-abstraction at the head, then
   we do a first-order unification.

   If the pattern is not flexible, then we do a first-order
   unification, too.

   If the pattern is flexible, and the goal doesn't have a
   lambda-abstraction head, then we second-order unification.
 *)

let clenv_fo_resolver clenv gls =
    clenv_unify (clenv_instance_template_type clenv) (pf_concl gls) clenv
;;

let clenv_typed_fo_resolver clenv gls =
    clenv_typed_unify (clenv_instance_template_type clenv) (pf_concl gls) clenv
;;


let args_typ gls = decrec []
    where rec decrec l c = match pf_whd_betadeltaiota gls c with
             DOP2(Prod,A,DLAM(n,B)) -> decrec ((named_hd A n,A)::l) B
           | x           -> l;;

(* if lname_type typ is [xn,An;..;x1,A1] and lists is a list of terms,
gives [x1:A1]..[xn:An]c' such that c converts to ([x1:A1]..[xn:An]c' l) *)

let abstract_scheme c l lname_typ =
    it_list2 (fun t (locc,a) (na,ta) ->
    if occur_meta ta then error "cannot find a type for the generalisation"
    else if occur_meta a then DOP2(Lambda,ta,DLAM(na,t))
         else DOP2(Lambda,ta,DLAM(na,subst_term_occ locc a t))) c (rev l) lname_typ;;

let abstract_list_all gls typ c l =
  let P = abstract_scheme c (map (function a -> [],a) l) (args_typ gls typ)
  in try if pf_conv_x gls (pf_type_of gls P) typ then P else 
     error "abstract_list_all"
     with UserError(_)->
     errorlabstrm "abstract_list_all"
      [< 'S "cannot find a generalisation of the goal with type : ";
          pTERMINENV (GLOB (pf_hyps gls),typ) >];;

let SecondOrderAbstraction allow_K gl p oplist clause =
    let (clause',cllist) = constrain_clenv_using_subterm_list allow_K clause oplist (pf_concl gl) in
    let typp = clenv_instance_type clause' p
    in clenv_unify (DOP0(Meta p)) (abstract_list_all gl typp (pf_concl gl) cllist) clause'
;;

let clenv_so_resolver allow_K clause gl =
    match whd_castapp_stack (clenv_instance_template_type clause) [] with
    (DOP0(Meta(p)),oplist) ->
    let clause' = SecondOrderAbstraction allow_K gl p oplist clause
    in clenv_fo_resolver clause' gl

  | _ -> error "clenv_so_resolver"
;;

let clenv_old_resolver allow_K clenv gls =
    match whd_castapp_stack (clenv_instance_template_type clenv) [] with
    (DOP0(Meta _) as pathd,_) ->
    try let cl = clenv_fo_resolver clenv gls in 
        (warning "FOMATCH";cl)
    with UserError _ -> clenv_so_resolver allow_K clenv gls
  | _ -> clenv_fo_resolver clenv gls
;;


let clenv_unique_resolver allow_K clenv gls =
    match (whd_castapp_stack (clenv_instance_template_type clenv) [],
           whd_castapp_stack (pf_concl gls) []) with
    ((DOP0(Meta _) as pathd,_),(DOP2(Lambda,_,_) as glhd,_)) ->
    if conv_x (w_Underlying clenv.hook) (clenv_instance_type_of clenv pathd) (clenv_type_of clenv glhd) then
        clenv_fo_resolver clenv gls
    else clenv_so_resolver allow_K clenv gls
  | ((DOP0(Meta _),_),_) -> (try clenv_so_resolver allow_K clenv gls
                            with UserError (t1,t2) -> 
                             try clenv_typed_fo_resolver clenv gls
                            with UserError (s1,s2) -> 
                             error 
                            "Cannot solve a second-order unification problem")
  | _ -> clenv_fo_resolver clenv gls
;;

let res_pf KONT clenv gls =
    clenv_refine KONT (clenv_unique_resolver false clenv gls) gls
;;

let res_pf1 KONT clenv gls =
    clenv_refine KONT (clenv_old_resolver false clenv gls) gls
;;

let res_pf_cast KONT clenv gls =
    clenv_refine_cast KONT (clenv_unique_resolver false clenv gls) gls
;;

let elim_res_pf KONT clenv gls =
    clenv_refine KONT (clenv_unique_resolver true clenv gls) gls
;;

let elim_res_pf1 KONT clenv gls =
    clenv_refine KONT (clenv_old_resolver true clenv gls) gls
;;

let elim_res_pf_cast KONT clenv gls =
    clenv_refine_cast KONT (clenv_unique_resolver true clenv gls) gls
;;

let e_res_pf KONT clenv gls =
    clenv_refine KONT (clenv_pose_dependent_evars (clenv_unique_resolver false clenv gls)) gls
;;

(* $Id: clenv.ml,v 1.9 1996/02/12 15:38:31 cpaulin Exp $ *)

(* $Id: clenv.ml,v 1.9 1996/02/12 15:38:31 cpaulin Exp $ *)
