(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                command.ml                                *)
(****************************************************************************)
#infix "o";;
#open "std";;
#open "pp";;
#open "stdpp";;
#open "himsg";;
#open "initial";;
#open "more_util";;
#open "generic";;
#open "term";;
#open "termenv";;
#open "reduction";;
#open "mach";;
#open "proof";;
#open "proof_constraints";;
#open "declare";;
#open "discharge";;
#open "trad";;
#open "names";;
#open "pp";;
#open "stdpp";;
#open "ast";;
#open "library";;
#open "libobject";;

let mkAppLC(f,l) = OP(("APPLIST",[]),f::l);;
let mkLambdaC(x,A,B) = OP(("LAMBDA",[]),[A;SLAM(SOME x,B)]);;
let mkElimC(P,c,l) = OP(("REC",[]),P::c::l);;
let mkRefC id = ASTnvar id;;
let mkProdC (x,A,B) = OP(("PROD",[]),[A;SLAM(SOME x,B)]);;
let mkArrowC(A,B) = OP(("PROD",[]),[A;SLAM(NONE,B)]);;
let mkCastC(c,t) = OP(("CAST",[]),[c;t]);;
let mkProdCit = list_it (fun (x,A) B -> mkProdC(x,A,B));;
let mkLambdaCit = list_it (fun (x,A) B -> mkLambdaC(x,A,B));;


(* Commands of the interface *)

(* 2| Constant definitions *)

let definition_body ident n com = 
let tycom = (match com with
          ASTnode("CAST",[c;t]) -> SOME t
        | _ -> NONE) in
let c = constr_of_com mt_evc (GLOB(initial_sign())) com in
let ty = option_app (constr_of_com mt_evc (GLOB(initial_sign()))) tycom in
let c = (match ty with
         NONE -> c
       | SOME ty -> DOP2(Cast,c,ty))
in machine_constant (initial_assumptions()) ((ident,false,n),c)
;;

let syntax_definition ident com =
let c = raw_constr_of_com mt_evc (GLOB (initial_sign())) com
in machine_syntax_constant ident [(CCI,c);(FW,c)]
;;

let abstraction_definition ident arity com =
let c = (sosub__try_soexecute o
         (astterm__dbize_cci mt_evc (termast__unitize_env (GLOB nil_sign)))) com
in machine_abstraction ident arity c
;;

(* 3.5| Variable definitions *)

let parameter_def_var ident c = 
      machine_parameter (initial_assumptions())
          (id_of_string ident,
           constr_of_com mt_evc (GLOB(initial_sign())) c);;

(* 3| Variable definitions *)

let hypothesis_def_var ident n c = 
    match n with

    NeverDischarge -> parameter_def_var ident c
    
  | DischargeAt disch_sp ->
    if is_section_p disch_sp then
        (machine_variable (initial_assumptions())
        (id_of_string ident,n,false,
         constr_of_com mt_evc (GLOB(initial_sign())) c);
        if pfedit__refining () then 
         (MSGERRNL [< 'S"Warning: Variable " ; 'S ident ; 
                    'S" is not visible from current goals" >]))
    else (MSGERRNL [< 'S"Warning: Declaring " ; 'S ident ; 
                    'S" as a parameter rather than a variable " ;
                    'S" because it is at a global level" >];
          parameter_def_var ident c);;

(* 4| Mutual Inductive definitions *)

let minductive_message = 
  function []  -> anomaly "no inductive definition"
         | [X] -> [< print_id X; 'SPC; 'S "is defined">]
         | l   -> HOV 0 
                  [< prlist_with_sep (function () -> [<'S ", ">]) print_id l;
                     'SPC; 'S "are defined">];;

let recursive_message = 
  function []  -> anomaly "no recursive definition"
         | [X] -> [< print_id X; 'SPC; 'S "is recursively defined">]
         | l   -> HOV 0 
                  [< prlist_with_sep (function () -> [<'S ", ">]) print_id l;
                     'SPC; 'S "are recursively defined">];;

let corecursive_message = 
  function []  -> anomaly "no corecursive definition"
         | [X] -> [< print_id X; 'SPC; 'S "is corecursively defined">]
         | l   -> HOV 0 
                  [< prlist_with_sep (function () -> [<'S ", ">]) print_id l;
                     'SPC; 'S "are corecursively defined">];;

let build_mutual lparams lnamearconstrs =

let lrecnames = list__map (fun (_,x,_,_)->x)  lnamearconstrs and
    lfinite   = vect_of_list (list__map (fun (x,_,_,_)->x) lnamearconstrs) and

     nparams = length lparams and

     sign0 = initial_sign() in

let (ind_sign,arityl) = it_list 
    (fun (sign,arl) (_,recname,arityc,_) -> 
      let arityj = fconstruct mt_evc sign0 (mkProdCit lparams arityc) in
      let arity = j_val arityj in
      (add_sign (recname,arity) sign, (arity::arl)))
      (sign0,[]) lnamearconstrs in

let mispecvec = vect_of_list 
    (list__map2 (fun ar (_,name,_,lname_constr) -> 
    let consconstrl = list__map 
     (fun (_,constr) -> constr_of_com_sign mt_evc ind_sign (mkProdCit lparams constr))
                      lname_constr in
        (name,Anonymous,ar,
        put_DLAMSV_subst (rev lrecnames) (vect_of_list consconstrl),
              vect_of_list (list__map fst lname_constr)))
    (rev arityl) lnamearconstrs)

in machine_minductive (initial_assumptions ()) nparams mispecvec lfinite;
   PPNL(minductive_message lrecnames);;


(* try to find non recursive definitions *)
let collect_non_rec  = searchrec [] 
where rec searchrec lnonrec lnamerec ldefrec larrec nrec = 
  try let i = 
      try_find_i (fun i f -> if for_all (fun def -> not (occur_var f def)) 
                                  ldefrec then i else failwith "try_find_i")
      0 lnamerec
      in let (lf1,f::lf2) = chop_list i lnamerec 
         and (ldef1,def::ldef2) = chop_list i ldefrec 
         and (lar1,ar::lar2) = chop_list i larrec 
         and newlnv = (try (match chop_list i nrec with 
                             (lnv1,_::lnv2) -> (lnv1@lnv2) ) 
                       with _ -> []) (* nrec=[] for cofixpoints *)
      in searchrec ((f,DOP2(Cast,def,ar))::lnonrec) 
                      (lf1@lf2) (ldef1@ldef2) (lar1@lar2) newlnv
  with Failure "try_find_i" -> (lnonrec, (lnamerec,ldefrec,larrec,nrec));;


let build_recursive lnameargsardef = 

let lrecnames = map (fun (f,_,_,_) -> f) lnameargsardef and
(sign0,_ as env) = initial_assumptions() and
nv = vect_of_list (map (fun (_,la,_,_) -> (length la) -1) lnameargsardef) in

let (rec_sign,arityl) = it_list 
    (fun (sign,arl) (recname,lparams,arityc,_) -> 
      let arityj = fconstruct mt_evc sign0 (mkProdCit lparams arityc) in
      let arity = j_val arityj in
      (add_sign (recname,arity) sign, (arity::arl)))
      (sign0,[]) lnameargsardef in 

let recdef = list__map (fun (_,lparams,_,def) -> 
       constr_of_com_sign mt_evc rec_sign (mkLambdaCit lparams def)) 
       lnameargsardef 

in let (lnonrec,(lnamerec,ldefrec,larrec,nvrec)) = 
          collect_non_rec lrecnames recdef (rev arityl) (list_of_vect nv)
in let n = NeverDischarge in 
if lnamerec <> [] then 
   (let recvec = concat_vect (vect_of_list larrec)
                    [|put_DLAMSV_subst (rev lnamerec) (vect_of_list ldefrec)|]
   in let rec declare i = function 
     (fi::lf) -> machine_constant env 
               ((fi,false,n),DOPN(Fix(vect_of_list nvrec,i),recvec));
                 declare (i+1) lf
   | _        -> () in 
(* declare the recursive definitions *)
    declare 0 lnamerec; PPNL(recursive_message lnamerec));
(* The others are declared as normal definitions *)
   do_list (fun (f,def) -> 
   machine_constant env ((f,false,n),astterm__absolutize_cci mt_evc (GLOB sign0) def);
   warning ((string_of_id f)^" is non-recursively defined")) lnonrec;;


let build_corecursive lnameardef = 

let lrecnames = map (fun (f,_,_) -> f) lnameardef and
(sign0,_ as env) = initial_assumptions() in

let (rec_sign,arityl) = it_list 
    (fun (sign,arl) (recname,arityc,_) -> 
      let arityj = fconstruct mt_evc sign0 arityc in
      let arity = j_val arityj in
      (add_sign (recname,arity) sign, (arity::arl)))
      (sign0,[]) lnameardef in 

let recdef = list__map (fun (_,_,def) -> 
       constr_of_com_sign mt_evc rec_sign  def) 
       lnameardef 

in let (lnonrec,(lnamerec,ldefrec,larrec,_)) = 
            collect_non_rec lrecnames recdef (rev arityl) []

in let n = NeverDischarge in 

if lnamerec <> [] then 
   (let recvec = if lnamerec = [] then [||] 
                else concat_vect (vect_of_list larrec)
                    [|put_DLAMSV_subst (rev lnamerec) (vect_of_list ldefrec)|]

   in let rec declare i = function 
     (fi::lf) -> machine_constant env ((fi,false,n),DOPN(CoFix(i),recvec));
                 declare (i+1) lf
   | _        -> () in 

declare 0 lnamerec; PPNL(corecursive_message lnamerec));
  do_list (fun (f,def) -> 
   machine_constant env ((f,false,n),astterm__absolutize_cci mt_evc (GLOB sign0) def);
   warning ((string_of_id f)^" is non-recursively defined")) lnonrec;;


let build_scheme lnamedepindsort = 
let lrecnames = list__map (fun (f,_,_,_) -> f) lnamedepindsort and
    sign0 = initial_sign() in
let lrecspec = list__map (fun (_,dep,ind,sort) -> 
     let indc = constr_of_com_sign mt_evc sign0 ind
     and (DOP0(Sort s)) = constr_of_com_sign mt_evc sign0 sort in
                  (indc,dep,s)) lnamedepindsort
and env = initial_assumptions()
and n = NeverDischarge in 
let listdecl = indrec__build_indrec mt_evc lrecspec in 
let rec declare i = function 
     (fi::lf) -> machine_constant env ((fi,false,n),listdecl.(i));
                 declare (i+1) lf
   | _        -> () in 

declare 0 lrecnames; PPNL(recursive_message lrecnames);;





(* $Id: command.ml,v 1.7 1996/02/12 15:38:36 cpaulin Exp $ *)
