(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                               discharge.ml                               *)
(****************************************************************************)
#infix "o";;
#open "std";;
#open "vectops";;
#open "names";;
#open "impuniv";;
#open "generic";;
#open "term";;
#open "constrtypes";;
#open "himsg";;
#open "pp";;
#open "stdpp";;
#open "machops";;
#open "more_util";;
#open "proof";;
#open "libobject";;
#open "library";;
#open "environ";;
#open "termenv";;
#open "dischcore";;
#open "variables";;
#open "constants";;
#open "indtypes";;
#open "indrec";;
#open "abstraction";;
#open "reduction";;


(********************************************************************)
(*                Declaring Variables and Constants                 *)
(********************************************************************)


let is_global id = 
  try let osp = nametab__sp_of_id CCI id in
      prefix_of (rev (dirpath osp)) (rev (lib__cwd()))
  with Not_found -> false
;;

let declare_variable (sticky,stre) (name,vbody,u) =
    (match stre with
     NeverDischarge ->
     error "Cannot declare a variable which will appear at top-level - use Axiom"
   | DischargeAt disch_sp ->
     if not(is_section_p disch_sp) then
         error "Cannot declare a variable which will appear at top-level - use Axiom");
    if mem_sign (initial_sign()) name then 
        error ("Cannot redeclare global name "^(string_of_id name))
    else add_named_object (name,OBJ) (inVariable(name,vbody,stre,sticky,u))
;;

let declare_constant id (cmap,strength,u) =
    try (let osp = nametab__sp_of_id CCI id
         in if prefix_of (rev (dirpath osp)) (rev (lib__cwd())) then
             error("Clash with previous constant " ^ (string_of_id id))
            else raise Not_found)
   (* we forbid to hide a previous constant with a constant *)
   with Not_found ->
       add_named_object (id,OBJ) (inConstant(cmap,strength,u))
;;

let declare_minductive (idv,cmap,u) =
    let cwd = rev (lib__cwd()) in 
    do_vect 
     (fun id -> try (let osp = nametab__sp_of_id CCI id
         in if prefix_of (rev (dirpath osp)) cwd then
             error("Clash with previous constant " ^ (string_of_id id))
            else ())  with Not_found -> ()) idv;
   (* we forbid to hide a previous constant with a constant *)
       add_named_object (idv.(0),OBJ) (inMutualInductive (cmap,u))
;;

let declare_syntax_constant id cmap =
try nametab__sp_of_id CCI id;
    error("Clash with previous name " ^ (string_of_id id))
with Not_found ->
    add_named_object (id,OBJ) (inSyntaxConstant cmap)
;;

let declare_abstraction id ao =
try nametab__sp_of_id ao.abs_kind id;
    error("Clash with previous name " ^ (string_of_id id))
with Not_found ->
    add_named_object (id,OBJ) (inAbstraction ao)
;;

(* Var intro, Naming the hypothesis *)
let machine_variable (sign,fsign) (name,stre,sticky,T) =
let sp = lib__make_path OBJ name in
let (u,vbody) = with_universes (infexecute_variable (sign,fsign))
                           (sp,initial_universes,T)
in declare_variable (sticky,stre) (name,vbody,u)
;;

let machine_constant (hyps,fhyps) ((name,opacity,str),c) =
let sp = lib__make_path OBJ name in
let (u,cmap) = with_universes (infexecute_constant (hyps,fhyps) opacity)
                          (sp,initial_universes,c)
in declare_constant name (cmap,str,u);;

let machine_constant_verbose  env ((name,_,_),_ as const) = 
machine_constant env const; PPNL [<print_id name; 'S " is defined" >]
;;

let machine_parameter (hyps,fhyps) (name,cty) =
let sp = lib__make_path OBJ name in
let (u',cmap) = with_universes (infexecute_parameter (hyps,fhyps))
                          (sp,initial_universes,cty)
in declare_constant name (cmap,NeverDischarge,u')
;;

(* Construction of a mutual recursive definition for compatibility *)

let declare_eliminations mindid = 
let sp = nametab__sp_of_id CCI mindid in
let mind = global_reference (GLOB(initial_sign())) sp mindid in
let redmind = minductype_spec mt_evc mind in
let mindstr = string_of_id mindid in
let declare na c = machine_constant_verbose
                   (initial_assumptions()) ((na,false,NeverDischarge),c) in 
let mispec = mind_specif_of_mind redmind in 
let elim_scheme = strong strip_outer_cast 
                  (mis_make_indrec mt_evc [] mispec).(0) in
let npars = mis_nparams mispec in
let make_elim s = instanciate_indrec_scheme s npars elim_scheme in
let kd = mis_kd mispec and kn = mis_kn mispec
in  if (mem prop kd) or (mem prop kn) then
            declare (id_of_string(mindstr^"_ind")) (make_elim prop);
    if (mem spec kd) or (mem spec kn) then
        declare (id_of_string(mindstr^"_rec")) (make_elim spec);
    if (mem typep kd) or (mem typep kn)  then
        declare (id_of_string(mindstr^"_rect")) (make_elim (Type(Null,New_univ())))
(*** does not declare automatically the strong informative elimination 
;
  if (mem types kd) or (mem types kn)  then
  declare (id_of_string(mindstr^"_recs")) (make_elim (Type(Pos,New_univ())))
***)
;;

let machine_minductive (hyps,fhyps) nparams mispecvec lfinite = 
if vect_length mispecvec = 0 then anomaly "machine_minductive" 
else 
 let mindidvec = 
   map_vect (fun  (indid,_,_,_,_) -> indid) mispecvec and 
 arlcvec = 
   map_vect (fun  (_,_,indarity,indconstructors,_) -> (indarity,indconstructors)) mispecvec and 
 namesvec = 
   map_vect (fun  (indid,indstamp,_,_,consnames) -> 
   if distinct (list_of_vect consnames) then (indstamp,indid,consnames)
   else error "Two constructors have the same name") mispecvec in 
let sp = lib__make_path OBJ mindidvec.(0) in
let (u',mimap) = with_universes 
    (infexecute_minductive (hyps,fhyps) nparams namesvec lfinite)
                               (sp,initial_universes,arlcvec)
in 
(declare_minductive (mindidvec,mimap,u');
    let rec decl_elim i = if i=vect_length lfinite then () 
     else ( (if lfinite.(i) 
             then declare_eliminations mindidvec.(i));
            decl_elim (i+1) )
    in decl_elim 0)
;;

let machine_syntax_constant id cmap =
    declare_syntax_constant id cmap;;

let machine_abstraction id arity rhs =
    declare_abstraction id (execute_abstraction CCI arity rhs);;

let set_transparent sp =
  if cookable_constant sp then
    let (_,cb) = const_of_path sp
    in (cook_constant sp;
        cb.CONSTOPAQUE <- false)
  else
    errorlabstrm "discharge__set_transparent" [< 'S"Can not set transparent."
       ; 'SPC ; 'S"It is a constant from a required module or a parameter." >]
;;

let set_opaque sp =
  if not(cookable_constant sp) then
    errorlabstrm "discharge__set_tranparent"
            [< 'S"Can not set opaque." ; 'SPC;
               'S"It is a constant from a required module or a parameter." >] ;
  let (_,cb) = const_of_path sp
  in cb.CONSTOPAQUE <- true
;;

(* $Id: declare.ml,v 1.9 1996/02/12 15:38:47 cpaulin Exp $ *)
