(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                               discharge.ml                               *)
(****************************************************************************)
#infix "o";;
#open "std";;
#open "vectops";;
#open "names";;
#open "impuniv";;
#open "generic";;
#open "term";;
#open "constrtypes";;
#open "himsg";;
#open "pp";;
#open "stdpp";;
#open "more_util";;
#open "proof";;
#open "libobject";;
#open "library";;
#open "environ";;
#open "termenv";;
#open "dischcore";;
#open "variables";;
#open "constants";;
#open "indtypes";;
#open "indrec";;
#open "abstraction";;
#open "reduction";;
#open "declare";;

(* C3| Section closing *)

let make_expands sp cmap =
    map (fun (k,_) -> (Const(coerce_path k sp),DO_REPLACE)) cmap
;;

let process_object sec_sp (ids_to_discard,work_alist) (sp,lobj) =
let tag = object_tag lobj
in match tag with

    "VARIABLE" ->
    let (id,(a,inf),stre,sticky,_) = outVariable lobj in
    if stre = (DischargeAt sec_sp) or ids_to_discard <> [] then
        (id::ids_to_discard,work_alist)

    else let expmod_a = expmod_constr work_alist a in
         let expmod_inf = option_app(expmod_constr work_alist) inf in
         let obj' = inVariable(id,(expmod_a,expmod_inf),stre,sticky,initial_universes)
         in add_named_object (id,OBJ) obj';
            (ids_to_discard,work_alist)

  | "CONSTANT" ->
    let (cmap,stre,_) = outConstant lobj in
    if stre = (DischargeAt sec_sp) then
        (ids_to_discard,(make_expands sp cmap)@work_alist)

    else let ((_,spid,spk)) = repr_path sp in
         let newsp = lib__make_path CCI spid in
         let (mods,cmap') = process_recipe_cmap (sp,newsp) (ids_to_discard,work_alist) cmap in
         let obj' = inConstant(cmap',stre,initial_universes)
         in add_named_object (spid,OBJ) obj';
            do_list (fun (k,cb) ->
                         if not cb.CONSTOPAQUE then
                             set_transparent (coerce_path k newsp)) cmap;
            (ids_to_discard,mods@work_alist)

  | "MUTUALINDUCTIVE" ->
    let (mimap,_) = outMutualInductive lobj in
    let ((_,spid,spk)) = repr_path sp in
    let newsp = lib__make_path CCI spid in
    let (mods,mimap') = process_mimap (sp,newsp) (ids_to_discard,work_alist) mimap in
    let obj' = inMutualInductive (mimap',initial_universes)
    in add_named_object (spid,OBJ) obj';
        (ids_to_discard,mods@work_alist)

  | _ -> (ids_to_discard,work_alist)
;;

let process_item sec_sp acc = function
    (sp,lib__LEAF lobj) -> process_object sec_sp acc (sp,lobj)
  | (_,lib__Import(s,isp,false)) -> (library__export_module s;acc)
  | (_,lib__Import(s,isp,true)) -> (library__import_module s;acc)
  | (_,_) -> acc
;;

let update sec_sp decls =
let (ids_to_discard,work_alist) =
    it_list (process_item sec_sp) ([],[]) decls in
let modified = map_succeed
    (fun (_,(DO_ABSTRACT(_,[]))) -> failwith "caught"
       | (Const sp,(DO_ABSTRACT(_,l))) -> Const sp
       | (MutInd spt,(DO_ABSTRACT(_,l))) -> MutInd spt
       | _ -> failwith "caught") work_alist
in (ids_to_discard,modified)
;;

let pop_path s sp =
let ((path_t,id,k)) = repr_path sp in
let sl = PA__un_intern path_t
in if sl <> [] & s = (hd sl) then (make_path(PA__intern (tl sl)) id k)
   else anomaly "pop_path"
;;

(* we now combine everything into a close_section function *)
(* eliminates all variables and constants local to the section *)

(* Only displays modified terms in CCI *)

let display_modified secid (discarded,modified) =
    MSG (prlist (fun (Const sp) ->
                    let sp = (pop_path secid sp) in
                    if (kind_of_path sp) = CCI then
                    let (_,cb) = const_of_path sp
                    in [< 'S"Constant " ; 'S(nametab__print_name sp) ; 'S":" ;
                       'FNL ; 'S"  " ; pTERM cb.CONSTTYPE ; 'FNL >]
                    else [<>]
                  | (MutInd (sp,tyi)) ->
                    let sp = (pop_path secid sp) in
                    if (kind_of_path sp) = CCI then
                    let (_,mib) = mind_of_path sp in 
                    let nameind = mib.MINDPACKETS.(tyi).MINDTYPENAME
                    in [< 'S "Type "; 
                       'S(nametab__print_name_mind sp nameind) ; 'S":" ;
                        pTERM mib.MINDPACKETS.(tyi).MINDARITY ; 'FNL >]
                    else [<>])
             (rev modified))
;;

let close_section str =
let sp = try library__find_dir (SOME str)
         with library__unknown_dir -> error ("Section "^str^
                                   " does not exist (or is already closed")
in
if (library__find_dir NONE) <> sp then
   error("Section " ^str^ " is not the innermost section");
let sec_ctxt = lib__contents_after (SOME sp)
in lib__close_dir sp {lib__exports=[]; lib__export_fn = NONE};
   backtrack_caches();
   display_modified str (update sp (rev sec_ctxt))
;;

let close_section_or_module s =
if pfedit__refining () then
   errorlabstrm "discharge__close_section_or_module"
         [< 'S"proof editing in progress" ; (pfedit__msg_proofs false) >];
let sp = try library__find_dir (SOME s)
         with library__unknown_dir -> error ("Section "^s^
                                   " does not exist (or is already closed)")
in match lib__map sp with
    lib__OpenDir(_,{lib__module_p=true;_}) ->
    let names =
        map_succeed (fun (sp,lib__LEAF _) -> sp
                       | (sp,lib__Import(_,_,false)) -> sp
                       | _ -> failwith "caught") (lib__contents_after (SOME sp))
    in (library__close_module sp names;
        backtrack_caches())
  | lib__OpenDir(_,{lib__module_p=false;_}) -> close_section s
  | _ -> error "Not a module"
;;

(* $Id: discharge.ml,v 1.4 1996/02/12 15:38:50 cpaulin Exp $ *)
