(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                parsing.ml                                *)
(****************************************************************************)
#infix "o";;
#open "std";;
#open "pp";;
#open "stdpp";;
#open "lomsg";;
#open "more_util";;
#open "names";;
#open "lexer";;
#open "termact";;
#open "pureparse";;
#open "ast";;
#open "pureparse";;

let ident_gram = function
    [< 'Tident id >] -> FRAGprim(TOKid(id_of_string id))
;;

let number_gram = function
    [< 'Tint n >] -> FRAGprim(TOKnum n)
;;

let string_gram = function
    [< 'Tstring s >] -> FRAGprim(TOKstr s)
;;

let unparsing_gram = function
    [< unparsings__unparsing u >] ->
    FRAGprim(TOKdyn(unparsings__inUNPARSING u))
;;

let plist_trap_error msg elem =
let rec aux acc strm =
    try aux ((elem strm)::acc) strm
    with Parse_error -> (warning msg;acc)
       | Parse_failure -> acc
in function strm -> rev(aux [] strm)
;;

let interp_gram whatfor id =
try interpc_gram [] [] (id_of_string whatfor,id_of_string id)
with Not_found -> error "Grammar Interpreter error"
;;

let annotation ts =
    match interp_gram "annote" "command" ts with
    FRAGast t -> t
  | _ -> failwith "annote"
;;

let annotation_gram ts =
    interp_gram "annote" "command" ts
;;

let command ts =
    match interp_gram "command" "command" ts with
    FRAGast t -> t
  | _ -> failwith "command"
;;

let command_gram ts =
    interp_gram "command" "command" ts
;;

let spat_parser = function
    [< parse_spat u >] -> u
  | [< 'Tlessless ; command_gram c ; 'Tgreatergreater >] ->
    (match c with
     FRAGast c -> spat_of_ast c
   | _ -> error "Parser: malformed antiquotation")

  | [< 'Tless ; 'Tcolon ;
       begin function
       [< 'Tident"command" >] -> "command"
     | [< 'Tident "tactic" >] -> "tactic"
     | [< 'Tident "vernac" >] -> "vernac"
       end gramname ;
       'Tcolon ; 'Tless ;
       (interp_gram gramname gramname) rslt ; 'Tgreatergreater >] ->
    (match rslt with
     FRAGast c -> spat_of_ast c
   | _ -> error "Parser: malformed antiquotation")
;;

let spat_gram = function
    [< spat_parser c >] -> FRAGprim(TOKdyn(ast__inSPAT c));;

let raw_command_gram = function
    [< parse_ast c >] -> FRAGast c
;;

let action_parser = function
    [< (termact__parse_act spat_parser) a >] -> a;;

let grammar_parser whatfor = function
    [< (pureparse__parse_grammar_entry action_parser whatfor) g >] -> g;;


let grammar_entry_gram = function
    [< 'Tident id ;
       (grammar_parser (id_of_string id)) ge >]
    -> FRAGprim(TOKdyn(pureparse__inGRAMMAR_ENTRY ge))
;;

let lcommand ts =
    match interp_gram "command" "lcommand" ts with
    FRAGast t -> t
  | _ -> failwith "lcommand"
;;

let lcommand_gram ts =
    interp_gram "command" "lcommand" ts
;;

let constructl = function
   [< 'Tlbrace;
      (plist_with_sep (function [< 'Tbar >] -> ()) command) lc;
      'Trbrace >] -> lc
;;

let tactic_com_list ts =
    match interp_gram "tactic" "tactic_com_list" ts with
    FRAGast t ->  t
  | _ -> failwith "tactic_com_list"
;;

let tactic_com_list_gram ts =
    interp_gram "tactic" "tactic_com_list" ts
;;

let vernac ts =
    match interp_gram "vernac" "vernac" ts with
    FRAGast t -> t
  | _ -> failwith "vernac"
;;

let mkLITERAL t = LITERAL(string_of_token vkey__rev_vernac_keywords t,t);;

let declare_prim_gram (whatfor,na) firsts pfun =
    if mhm__in_dom prim_grams (id_of_string whatfor,id_of_string na) then
        anomalylabstrm "parsing__declare_prim_gram"
        [< 'S"Cannot declare the primitive grammar " ; 'S whatfor ;
           'S":" ; 'S na ; 'S" more than once" >];
    mhm__add prim_grams ((id_of_string whatfor,id_of_string na),(pfun,firsts))
;;

let init_grammar () =
    mhm__empty prim_grams;
    (do_list
    (fun whatfor ->
    (mhm__add prim_grams ((whatfor,id_of_string "ident"),
                          (ident_gram,[PRIMITIVE(whatfor,id_of_string"ident")]));
    mhm__add prim_grams ((whatfor,id_of_string "number"),
                         (number_gram,[PRIMITIVE(whatfor,id_of_string"number")]));
    mhm__add prim_grams ((whatfor,id_of_string "unparsing"),
                         (unparsing_gram,[mkLITERAL Tlbracket;
                                          PRIMITIVE(whatfor,id_of_string"number")]));
    mhm__add prim_grams ((whatfor,id_of_string "grammar_entry"),
                         (grammar_entry_gram,[PRIMITIVE(whatfor,id_of_string"ident")]));
    mhm__add prim_grams ((whatfor,id_of_string "spat"),
                         (spat_gram,[mkLITERAL Tlbracket;mkLITERAL Tlessless;
                                           PRIMITIVE(whatfor,id_of_string"ident")]));
    mhm__add prim_grams ((whatfor,id_of_string "raw_command"),
                         (raw_command_gram,[mkLITERAL Tlbracket;
                                           PRIMITIVE(whatfor,id_of_string"ident")]));
    mhm__add prim_grams ((whatfor,id_of_string "string"),
                         (string_gram,[PRIMITIVE(whatfor,id_of_string"string")]))
    ))
   [id_of_string "prim"])
;;

(* $Id: parsing.ml,v 1.2 1996/02/12 15:38:54 cpaulin Exp $ *)
