(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                pfedit.ml                                 *)
(****************************************************************************)

#infix "o";;
#open "std";;
#open "lomsg";;
#open "names";;
#open "generic";;
#open "term";;
#open "declare";;
#open "trad";;
#open "mach";;
#open "more_util";;
#open "pp";;
#open "stdpp";;
#open "termenv";;
#open "tacmach";;
#open "proof";;
#open "proof_trees";;
#open "library";;
#open "libobject";;

let PROOFS = (edit__mt() : (string,pftreestate,proof_topstate) edit__t);;

let undo_default = 12;;
let undo_limit = ref undo_default;;

let set_undo n = if n>=0 then undo_limit:=n+1
                 else error"Cannot set a negative undo limit";;

let unset_undo () = set_undo undo_default;;

let list_proofs () = edit__dom PROOFS;;

let msg_proofs use_resume =
  match edit__dom PROOFS
  with
    [] -> [< 'SPC ; 'S"(No proof-editing in progress)." >]
  | l ->  [< 'S"." ; 'FNL ; 'S"Proofs currently edited:" ; 'SPC ;
             (prlist_with_sep (fun () -> [< 'SPC >])
                               (fun x -> [< 'S x >])
                               (list_proofs ())) ; 'S"." ;
            (if use_resume then [< 'FNL ; 'S"Use \"Resume\" first." >]
                           else [< >])
           >]
;;

let get_state () =
    match edit__read PROOFS with
    NONE -> errorlabstrm "pfedit__get_state" [< 'S"No focused proof";
                                                (msg_proofs true) >]
  | SOME(_,pfs,ts) -> (pfs,ts)
;;

let get_topstate () = snd(get_state());;
let get_pftreestate () = fst(get_state());;

let set_proof s = 
   try edit__focus PROOFS s
   with Invalid_argument "edit__focus" ->
            errorlabstrm "pfedit__set_proof"
                   [< 'S"No such proof" ; (msg_proofs false) >]
;;

let resume_last () = match (edit__last_focused PROOFS)
                     with
                       NONE -> errorlabstrm "resume_last"
                                     [< 'S"No proof-editing in progress." >]
                     | p    -> edit__focus PROOFS p;;

let get_proof () =
    match edit__read PROOFS with
    NONE -> errorlabstrm "pfedit__get_proof" [< 'S"No focused proof" ;
                                                (msg_proofs true) >]
  | SOME(na,_,_) -> na
;;


let add_proof (na,pfs,ts) =
    edit__create PROOFS (na,pfs,ts,SOME !undo_limit)
;;

let del_proof na =
  try edit__delete PROOFS na
  with (UserError ("edit__delete",_)) -> errorlabstrm "pfedit__del_proof"
                                        [< 'S"No such proof" ;
                                                (msg_proofs false)>]
;;

let init_proofs () = edit__empty PROOFS;;

let mutate F =
    try edit__mutate PROOFS (fun _ pfs -> F pfs)
    with
    Invalid_argument "edit__mutate" -> errorlabstrm "pfedit__mutate"
                                                    [< 'S"No focused proof" ;
                                                (msg_proofs true) >]
;;

let rev_mutate F =
    try edit__mutate PROOFS (fun _ pfs -> F pfs)
    with
    Invalid_argument "edit__rev_mutate" -> errorlabstrm "pfedit__rev_mutate"
                                                    [< 'S"No focused proof";
                                                (msg_proofs true) >]
;;

let start (na,ts) =
let pfs = mk_pftreestate ts.top_goal
in add_proof(na,pfs,ts)
;;

let restart () =
    match edit__read PROOFS with
    NONE -> errorlabstrm "pf_edit__restart"
                         [< 'S"No focused proof to restart" ;
                                                (msg_proofs true)>]
  | SOME(na,_,ts) -> (del_proof na;
                      start (na,ts);
                      set_proof (SOME na))
;;

let proof_prompt () =
    match edit__read PROOFS with
    NONE -> "Coq < "
  | SOME(na,_,_) -> na^" < "
;;

let undo n =
  try edit__undo PROOFS n
  with
  (Invalid_argument "edit__undo") -> errorlabstrm "pfedit__undo"
                                               [< 'S"No focused proof"; 
                                                (msg_proofs true)>]
;;


let proof_term () =
    extract_pftreestate (get_pftreestate())
;;

(*********************************************************************)
(*      Saving  functions                                    *)
(*********************************************************************)

let save_named opacity =
let (pfs,ts) = get_state() and
    ident = get_proof() in
let (sign,fsign) = ts.top_hyps and
    {concl=concl;_} = ts.top_goal and
    strength = ts.top_strength in
let pfterm = extract_pftreestate pfs

in (machine_constant (sign,fsign) ((id_of_string ident,opacity,strength),
                                   DOP2(Cast,pfterm,concl));
    del_proof ident;
    message(ident ^ " is defined"))
;;
    
let save_anonymous opacity save_ident n =
let (pfs,ts) = get_state() and
    ident = get_proof() in
let (sign,fsign) = ts.top_hyps and
    {concl=concl;_} = ts.top_goal and
    strength = ts.top_strength in
let pfterm = extract_pftreestate pfs

in if ident = "Unnamed_thm" then
    machine_constant (sign,fsign) ((id_of_string save_ident,opacity,strength),
                                   DOP2(Cast,pfterm,concl))
   else begin
       message("Overriding name " ^ ident ^ " and using " ^ save_ident);
       machine_constant (sign,fsign) ((id_of_string save_ident,opacity,strength),
                                   DOP2(Cast,pfterm,concl))
       end;
       del_proof ident;
       message(save_ident ^ " is defined")
;;

let save_anonymous_thm opacity id =
    save_anonymous opacity id NeverDischarge;;

let save_anonymous_remark opacity id =
    save_anonymous opacity id
    (make_strength (safe_cddr (library__cwd())));;

(*********************************************************************)
(*              Abort   functions                                    *)
(*********************************************************************)
 
let refining () = [] <> (edit__dom PROOFS);;

let abort_goal pn = del_proof pn; message ("Goal "^pn^" aborted")
;;

let abort_cur_goal () =
  del_proof (get_proof ()) ;
  message "Current goal aborted"
;;

let abort_goals () =
    if refining() then
        (init_proofs();
         message "Current goals aborted")
    else error "No proof-editing in progress";;

let abort_refine f x = 
    if refining() then (abort_goals(); f(x))
       (* used to be: error "Must save or abort current goal first" *)
    else f(x);;

let reset_name = abort_refine reset_to
and reset_keeping_name = abort_refine reset_keeping
and save_state = abort_refine states__raw_save_state
and restore_state = abort_refine states__raw_restore_state
and restore_last_saved_state=abort_refine states__raw_restore_last_saved_state
and reset_all = abort_refine reset_library
and reset_section = abort_refine raw_reset_section
;;

let reset_prelude () = restore_state "Prelude"
and reset_initial () = restore_state "Initial";;

    
(*********************************************************************)
(*              Modifying the current prooftree                      *)
(*********************************************************************)

let start_proof na str concl_com =
let concl = fconstruct mt_evc (initial_sign()) concl_com in
match concl.TYP with 
 DOP0(Sort _ ) ->
   (let top_goal = mkGOAL mt_lc mt_ctxt (initial_sign()) (j_val concl) in
    let ts = {top_hyps = initial_assumptions();
          top_goal = top_goal;
          top_strength = str}
    in start(na,ts);   set_proof (SOME na))
 | _ -> error "A goal should be a type"
;;

let solve_nth k tac =
  if not(listset__memb (cursor_of_pftreestate(get_pftreestate())) (-1))
  then  mutate (solve_nth_pftreestate k tac)
(* ;
   let pfs =get_pftreestate() in
   let pf = proof_of_pftreestate pfs in
   let (pfterm,metamap) = refiner__extract_open_proof pf.goal.hyps pf
   in (try typing__control_only_guard mt_evc pfterm
       with e -> (undo 1; raise e));
   ()) 
*)
  else error "cannot apply ``n:'' when we are descended behind a tactic-node"
;;

let by tac = mutate (solve_pftreestate tac);;

let traverse n = rev_mutate (traverse n);;
 
(* [traverse_to path]

   Traverses the current proof to get to the location specified by
   [path].

   ALGORITHM:

   If the cursor is equal to the path, we are done.

   If the cursor is longer than the path, then traverse to the parent

   If the cursor is equal to the tail of the path, then traverse to
   the head of the path, and we are done.

   Otherwise, traverse to the parent, traverse to the tail of the
   path, then traverse to the path.

*)
(* [traverse_to path]

   Traverses the current proof to get to the location specified by
   [path].

   ALGORITHM:

   If the cursor is equal to the path, we are done.

   If the cursor is longer than the path, then traverse to the parent

   If the cursor is equal to the tail of the path, then traverse to
   the head of the path, and we are done.

   Otherwise, traverse to the parent, traverse to the tail of the
   path, then traverse to the path.

*)

let rec nth_cdr = function
    0 -> (function l -> l)
  | n -> tl o (nth_cdr (n - 1));;

let rec common_ancestor_equal_length =
  fun (a::l1) (b::l2) ->
        let (prefx,n) as result = (common_ancestor_equal_length l1 l2) in
          if result = ([],0) then    
             if a = b then
               result
             else 
               (a::prefx),(n + 1)
          else 
             (a::prefx),(n + 1)
   | [] [] -> [],0
   | _ _ -> anomaly "common_ancestor_equal_length";;

let common_ancestor l1 l2 =
    let diff_lengths = (length l1) - (length l2) in
    if diff_lengths > 0 then
       let head,tail = chop_list diff_lengths l1 in
       let (prefx,n) =  common_ancestor_equal_length
              tail l2
       in (head@prefx),n
    else if diff_lengths < 0 then
       let (prefx,n) = common_ancestor_equal_length
              l1 (nth_cdr (- diff_lengths) l2)
       in prefx,(n - diff_lengths)
    else common_ancestor_equal_length l1 l2;;

let rec traverse_up = function
    0 -> (function pf -> pf)
  | n -> (function pf -> tacmach__traverse 0 (traverse_up (n - 1) pf));;

let rec traverse_down = function
    [] -> (function pf -> pf)
  | (n::l) -> (function pf -> tacmach__traverse n (traverse_down l pf));;

let traverse_to path =
    let up_and_down path pfs =
      let cursor = cursor_of_pftreestate pfs in
      let down_list, up_count = common_ancestor path cursor
      in (traverse_down down_list (traverse_up up_count pfs))
    in
       rev_mutate (up_and_down path)
;;



(* $Id: pfedit.ml,v 1.5 1996/02/12 15:38:58 cpaulin Exp $ *)
