(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                pretty.ml                                 *)
(****************************************************************************)

(* A Pretty-Printer for the Calculus of Inductive Constructions *)

#open "std";;
#open "vectops";;
#open "names";;
#open "more_util";;
#open "generic";;
#open "term";;
#open "constrtypes";;
#open "reduction";;
#open "environ";;
#open "termenv";;
#open "machops";;
#open "mach";;
#open "himsg";;
#open "pp_control";;
#open "pp";;
#open "stdpp";;
#open "proof";;
#open "libobject";;


let print_closed_sections = ref false;;

let print_typed_value (val,typ) =
  [< pTERM val ; 'FNL ; 'S "     : "; pTERM typ ; 'FNL >];;

let print_recipe = function
    SOME{v=COOKED c} -> pTERM c
  | SOME{v=RECIPE _} -> [< 'S"<recipe>" >]
  | NONE -> [< 'S"<uncookable>" >]
;;

let fprint_recipe = function
    SOME{v=COOKED c} -> pFTERM c
  | SOME{v=RECIPE _} -> [< 'S"<recipe>" >]
  | NONE -> [< 'S"<uncookable>" >]
;;

let print_typed_recipe (val,typ) =
  [< print_recipe val ; 'FNL ; 'S "     : "; pTERM typ ; 'FNL >];;


(* To be improved; the type should be used to provide the types in the
   abstractions. This should be done recursively inside pTERM, so that
   the pretty-print of a proposition (P:(nat->nat)->Prop)(P [u]u)
   synthesizes the type nat of the abstraction on u *)

let string_of = string_of_id;;

let print_var name typ =
    [< 'S "*** [" ; 'S name ; 'S " : ";
     pTERM typ; 
     'S "]"; 'FNL >];;

let print_env pk = 
    let pterminenv = if pk = CCI then pTERMINENV else pFTERMINENV in
    let rec prec env = function [] -> [<>] 
             |[Name id as name,C] ->  [<print_id id; 'S":"; pTERMINENV(env,C)>]
             |(Name id as name,C)::rest -> 
               [<print_id id; 'S":"; pTERMINENV(env,C); 'S";";
                prec (add_rel (name,C) env) rest>]
    in prec (GLOB nil_sign);;

let print_constructors_with_sep pk fsep mip = 
let pterm,pterminenv = if pk = CCI then (pTERM,pTERMINENV)
                     else (pFTERM,pFTERMINENV) in
let (lna,lC) = decomp_all_DLAMV_name mip.MINDLC in
let ass_name = assumptions_for_print lna in
let lidC = list_of_vect 
          (map2_vect (fun id C -> (id,C)) mip.MINDCONSNAMES lC) in
prlist_with_sep fsep 
   (function (id,C) -> [< pID id; 'S ":"; pterminenv (ass_name,C) >]) lidC;;


let print_mutual pk mib = 
let pterm,pterminenv = if pk = CCI then (pTERM,pTERMINENV)
                     else (pFTERM,pFTERMINENV) in
 let {MINDPACKETS=mipv; MINDNPARAMS=nparams} = mib in 
 let (lpars,_) = decomp_n_prod mt_evc nparams mipv.(0).MINDARITY in
 let lparsname = map fst lpars in
 let lparsprint = assumptions_for_print lparsname in

 let print_constructors mip = 
      let (lna,lC) = decomp_all_DLAMV_name mip.MINDLC in
      let ass_name = assumptions_for_print (lparsname@lna) in
      let lidC = list_of_vect 
          (map2_vect (fun id C -> (id,snd (decomp_n_prod mt_evc nparams C)))
                      mip.MINDCONSNAMES lC) in
      (HOV 0
      (prlist_with_sep (fun () -> [<'CUT; 'S " | " >]) 
   (function (id,C) -> (HOV 0 [< pID id; 'S " : "; pterminenv (ass_name,C) >]))
        lidC)) in

 let print_oneind mip = 
     let  sfinite = if mip.MINDFINITE then "Inductive " else "CoInductive " in
     let (_,arity) = decomp_n_prod mt_evc nparams mip.MINDARITY in 
     (HOV 0 [< 'S sfinite ; pID mip.MINDTYPENAME ;  'S " : "; 
                            pterminenv (lparsprint,arity); 'S " := "; 'FNL;
                            print_constructors mip >] ) in 

 if vect_length mipv = 1 then 
 let mip = mipv.(0) in
 let (_,arity) = decomp_n_prod mt_evc nparams mip.MINDARITY in 
 let  sfinite = if mip.MINDFINITE then "Inductive " else "CoInductive " in
 (HOV 0 [< 'S sfinite ; pID mip.MINDTYPENAME ;
    if nparams = 0 then [<>] 
                   else [< 'S" ["; print_env pk (rev lpars); 'S "]">];
    'BRK(1,5); 'S ": "; pterminenv (lparsprint,arity); 'S " := "; 'BRK(0,4);
         print_constructors mip >] )
 else
 (HOV 0 [< 'S"Mutual " ; 
     if nparams = 0 then [<>] else 
     [< 'S"["; print_env pk (rev lpars); 'S "]">]; 'BRK(0,4);
        prlist_with_sep (fun () -> [<'FNL; 'S " with " >]) print_oneind 
        (list_of_vect mipv);
        'FNL >]);;

let print_extracted_mutual sp = let (_,mib) = mind_of_path (ccisp_of sp) 
  in match mind_singl mib with 
        NONE -> print_mutual FW (snd (mind_of_path (fwsp_of sp)))
      | SOME A -> pFTERM A;;


let print_leaf_entry with_values sep (spopt,lobj) =
  let tag = object_tag lobj in
  match (spopt,tag) with
    (_,"VARIABLE") ->
    let (name,(typ,_),_,_,_) = outVariable lobj in
    [< print_var (string_of name) typ >]
 
  | (sp,"CONSTANT") ->
    let (cmap,_,_) = outConstant lobj in
    let {CONSTBODY=val;CONSTTYPE=typ;_} = listmap__map cmap CCI
    in (match val with NONE -> 
    [< 'S"*** [ "; 'S (nametab__print_name (ccisp_of sp));  'S " : "; 'CUT ;
        pTERM typ ; 'S" ]"; 'FNL >] 
         | _ -> 
       [< 'S(nametab__print_name (ccisp_of sp)) ; 'S sep; 'CUT ;
       if with_values then print_typed_recipe (val,typ) 
                       else [< pTERM typ ; 'FNL >] >])

  | (sp,"MUTUALINDUCTIVE") ->
    let (cmap,_) = outMutualInductive lobj in
    [< print_mutual CCI (listmap__map cmap CCI); 'FNL >]

  | (_,"AUTOHINT") -> [< 'S" Add Auto Marker" ; 'FNL >]

  | (_,"GRAMMAR") -> [< 'S" Grammar Marker" ; 'FNL >]

  | (sp,"SYNTAXCONSTANT") -> 
         [< 'S" Syntax Macro " ; 
      	    'S(string_of_id(basename sp)) ; 'S sep;  
            if with_values then 
              let cmap = outSyntaxConstant lobj in 
      	      [< pTERM (listmap__map cmap CCI) >]
            else [<>]; 'FNL >]

  | (_,"PPSYNTAX") -> [< 'S" Syntax Marker" ; 'FNL >]

  | (_,"TOKEN") -> [< 'S" Token Marker" ; 'FNL >]

  | (sp,s) -> [< 'S(string_of_path sp) ; 'S" : " ;
                 'S"Unrecognized object " ; 'S s ; 'FNL >]
;;


let rec print_library_entry with_values ent = 
    let sep = if with_values then " = " else " : "
in match ent with
    (sp,lib__LEAF lobj) -> [< print_leaf_entry with_values sep (sp,lobj) >]
  | (s,lib__ClosedDir(_,_,_,ctxt)) ->
    [< 'S"Closed Section " ; 'S (string_of_path s) ; 'FNL ;
       if !print_closed_sections then
           [< 'WS 1 ; HOV 0 [< print_context with_values ctxt >] >]
       else [< >] >]

  | (sp,lib__OpenDir(str,_)) -> 
        [< 'S(" >>>>>>> Section " ^ str) ; 'FNL >]

  | (_,lib__Import(s,_,true)) ->
    [< 'S" >>>>>>> Import " ; 'S s ; 'FNL >]

  | (_,lib__Import(s,_,false)) ->
    [< 'S" >>>>>>> Export " ; 'S s ; 'FNL >]

and print_context with_values = prec
    where rec prec = function
    h::rest -> [< prec rest ; print_library_entry with_values h >]
  | [] -> [< >]
;;

let print_full_context () = print_context true (lib__contents_after NONE);;

let print_full_context_typ () = print_context false (lib__contents_after NONE);;

(* For printing an inductive definition with
   its constructors and elimination,
   assume that the declaration of constructors and eliminations
   follows the definition of the inductive type *)

let rec head_const c = match strip_outer_cast c with
      DOP2(Prod,_,DLAM(_,c)) -> head_const c
    | DOPN(AppL,cl)    -> head_const (hd_vect cl)
    | def         -> def;;

let list_filter_vec f vec = frec (vect_length vec -1) []
where rec frec n lf = if n < 0 then lf 
       else if f vec.(n) then frec (n-1) (vec.(n)::lf)
            else frec (n-1) lf;;

let is_rel = function Rel _ -> true | _ -> false;;
 (* The four functions print_constructors_head, print_all_constructors_head,
   print_constructors_rel and crible implement the behavior needed for the
   Coq Search command.  These functions take as first argument the procedure
   that will be called to treat each entry.  This procedure receives the name
   of the object, the assumptions that will make it possible to print its type,
   and the constr term that represent its type. *)

let print_constructors_head  (fn:string -> unit assumptions -> constr -> unit) c mip = 
let (lna,lC) = decomp_all_DLAMV_name mip.MINDLC in
let ass_name = assumptions_for_print lna in
let lidC = map2_vect (fun id C -> (id,C)) mip.MINDCONSNAMES lC in
let flid = list_filter_vec (fun (_,C) -> head_const C = c) lidC in
 do_list
    (function (id,C) -> fn (string_of_id id) ass_name C) flid;;

let print_all_constructors_head fn c mib = 
    let mipvec = mib.MINDPACKETS and n = mib.MINDNTYPES in 
    prec 0 where rec prec i = if i = n then MSG([< >])
     else (print_constructors_head fn c mipvec.(i); prec (i+1));;

let print_constructors_rel fn mip = 
let (lna,lC) = decomp_all_DLAMV_name mip.MINDLC in
let ass_name = assumptions_for_print lna in
let lidC = map2_vect (fun id C -> (id,C)) mip.MINDCONSNAMES lC in
let flid = list_filter_vec (fun (_,C) -> is_rel (head_const C)) lidC in
do_list (function (id,C) -> fn (string_of_id id) ass_name C) flid;;

let crible (fn:string -> unit assumptions -> constr -> unit) name =
  let imported = library__search_imports() in
  let const = global (GLOB(initial_sign())) name
  in let rec crible_rec = function
      (spopt,lib__LEAF lobj)::rest ->

      (match (spopt,object_tag lobj) with
       (_,"VARIABLE") ->
       let (namec,(typ,_),_,_,_) = outVariable lobj in 
              (if (head_const typ)=const then  
                  fn (string_of_id namec) (GLOB(initial_sign())) typ);
              crible_rec rest

     | (sp,"CONSTANT") ->
       let (_,{CONSTTYPE=typ;_}) = const_of_path (ccisp_of sp) in
            (if (head_const typ)=const then 
                fn (nametab__print_name sp) (GLOB(initial_sign())) typ);
                crible_rec rest

     | (sp,"MUTUALINDUCTIVE") -> 
           let (_,mib) = mind_of_path (ccisp_of sp) in 
           (match const with (DOPN(MutInd(sp',tyi),_)) 
                  -> if sp = objsp_of sp' then print_constructors_rel fn
                                      (mind_nth_type_packet mib tyi) 
               | _ -> print_all_constructors_head fn const mib); 
           crible_rec rest
           
     | _ -> crible_rec rest)
  | (spopt,lib__OpenDir _)::rest -> crible_rec rest
  | (spopt,lib__ClosedDir (_,_,_,libseg))::rest -> 
             (if mem spopt imported then crible_rec libseg);
              crible_rec rest
  | (spopt,lib__Import _)::rest -> crible_rec rest
  | []      -> ()

  in try crible_rec (lib__contents_after NONE)
     with Not_found -> error ((string_of_id name) ^ " not declared");;

  let print_crible name =
      crible (fun str ass_name constr -> 
               MSG[<'S str; 'S":"; pTERMINENV(ass_name,constr); 'FNL >]) name;;

let read_sec_context sec =
    let rec get_cxt in_cxt = function
        ((sp,lib__OpenDir(str,_)) as hd)::rest ->
        if str = sec then (hd::in_cxt)
        else get_cxt (hd::in_cxt) rest
      | [] -> []
      | hd::rest -> get_cxt (hd::in_cxt) rest in
    let cxt = (lib__contents_after NONE) in
    rev (get_cxt [] cxt);;

let print_sec_context sec = print_context true (read_sec_context sec);;

let print_sec_context_typ sec = print_context false (read_sec_context sec);;

let print_val j =
     let {VAL=val;TYP=typ;_} = j in print_typed_value(val,typ);;

let print_name name = let str = string_of_id name in 
    try let sp = nametab__sp_of_id CCI name 
    in  let lobj = lib__map_leaf (objsp_of sp) in
    print_leaf_entry true " = " (sp,lobj)
    with Not_found -> (try 
            let typ = snd(lookup_glob name (GLOB (initial_sign ())))
            in print_var str typ 
            with  Not_found | Invalid_argument _
                  -> error (str ^ " not a defined object"))
  | Invalid_argument _ -> error (str ^ " not a defined object");;


let print_opaque_name name = 
    try (match (global (GLOB(initial_sign())) name) with
        DOPN(Const sp,_) as x ->
         if evaluable_const mt_evc x then
             print_typed_value(const_value mt_evc x,const_type mt_evc x)
         else
             (declare__set_transparent sp;
              let valu = const_value mt_evc x
              in (declare__set_opaque sp;
                  print_typed_value(valu,const_type mt_evc x)))

      | DOPN(MutInd (sp,_),_) as x -> print_mutual CCI (snd (mind_of_path sp))
      | DOPN(MutConstruct _,_) as x -> 
                            print_typed_value(x,type_of mt_evc (initial_sign()) x)
      | VAR id ->
         let a = snd(lookup_sign id (initial_sign()))
         in print_var (string_of_id id) (a)
      | _ -> failwith "print_name")
    with Not_found -> error ((string_of_id name) ^ " not declared");;


let print_type j =
    let {VAL=val;TYP=typ;_} = j in
        print_typed_value (val,strong whd_betaiota typ);;

let print_evaluate j =
    let {VAL=val;TYP=typ;_} = j in
        [< 'S "     = ";print_typed_value (strong whd_betaiota val,
                                           strong whd_betaiota typ) >];;

let print_compute j = 
    let {VAL=val;TYP=typ;_} = j in 
        [< 'S "     = ";
 print_typed_value (strong (whd_betadeltaiota mt_evc) val,
                     strong (whd_betadeltaiota mt_evc) typ) >];;

let print_simplify j = 
    let {VAL=val;TYP=typ;_} = j in 
        [< 'S "     = ";print_typed_value (nf mt_evc val, nf mt_evc typ) >];;

let print_local_context () =
let env = lib__contents_after NONE in
   [< print_var_rec env;  print_last_const env >]
where rec print_var_rec = function 
    [] -> [< >]
  | ((_,lib__LEAF lobj))::rest ->
    if "VARIABLE" = object_tag lobj then
        let (name,(typ,_),_,_,_) = outVariable lobj
        in [< print_var_rec rest;
            print_var (string_of name) typ >]
    else print_var_rec rest

 | _::rest -> print_var_rec rest

and print_last_const = function
   ((sp,lib__LEAF lobj))::rest -> (match object_tag lobj with
     "CONSTANT" -> 
       let (_,{CONSTBODY=val;CONSTTYPE=typ;_}) = const_of_path (ccisp_of sp) in
           [< print_last_const rest;
            'S(nametab__print_name sp) ;'S" = ";
            print_typed_recipe (val,typ) >]
    |  "MUTUALINDUCTIVE" -> let (_,mib) = mind_of_path (ccisp_of sp) in 
                            [< print_last_const rest;
                               print_mutual CCI mib; 'FNL >]
    |  "VARIABLE" ->  [< >]
    |  _          ->  print_last_const rest)

 | (sp,lib__ClosedDir _)::rest -> print_last_const rest
 | _ -> [< >];;

let fprint_var name typ =
    [< 'S ("*** [" ^ name ^ " :");
       pFTERM typ; 
       'S "]"; 'FNL >];;

let fprint_judge j = [< pFTERM j.VAL; 'S" : " ; pFTERM j.TYP >];;

let unfold_head_fconst sigma = unfrec
    where rec unfrec = function
    DOPN(Const _,_) as K -> const_value sigma K 
  | DOP2(Lambda,T,DLAM(na,B)) -> DOP2(Lambda,T,DLAM(na,unfrec B))
  | DOPN(AppL,v) -> DOPN(AppL,cons_vect (unfrec (hd_vect v)) (tl_vect v))
  | x -> x
;;

let print_extracted_name name =
    try (match (global (GLOB(initial_sign())) name) with
        DOPN(Const _,_) as x ->
         let cont = snd(infexecute mt_evc (initial_assumptions()) x)
         in 
         (match cont with
               Inf(d) ->
               [< 'S (string_of_id name); 
                  if defined_const mt_evc x then 
                  if opaque_const mt_evc x then 
                      [< 'S " ==> <_>" >]
                  else  [< 'S " ==>";
                        pFTERM (unfold_head_fconst mt_evc d.VAL); 'FNL>]
                  else [< >]; 'S " :  ";
                  pFTERM (d.TYP); 'FNL >]
             | _                  -> error "Non informative term")
      | VAR id ->
        let a = snd(lookup_sign id (initial_sign())) in
        let cont = snd(infexecute mt_evc (initial_assumptions()) a)
        in (match cont with
               Inf(t) -> fprint_var (string_of_id name) t.VAL
             | _                   -> error "Non informative term")
      
      | DOPN(MutInd (sp,_),_) as x ->
         let cont = snd(infexecute mt_evc (initial_assumptions()) x)
         in (match cont with
               Inf(_) ->
               [< 'S ((string_of_id name) ^" ==> ");
                print_extracted_mutual sp >]
             | _                  -> error "Non informative term")
      | DOPN(MutConstruct _ ,_) as x ->
         let cont = snd(infexecute mt_evc (initial_assumptions()) x)
         in (match cont with
               Inf(d) ->
               [< 'S ((string_of_id name) ^" ==> ");
                   fprint_judge d ; 'FNL >]
             | _                  -> error "Non informative term")
      | _ -> anomaly "should be a variable or constant")
    with Not_found -> error ((string_of_id name) ^ " not declared");;

let print_extraction () = [< print_rec (lib__contents_after NONE); 'FNL >]
 where rec print_rec = function
    (sp,lib__LEAF lobj)::rest ->
    (match (sp,object_tag lobj) with
     (sp,"CONSTANT") ->
    (try let (_,{CONSTBODY=d;_}) = const_of_path (fwsp_of sp)
         in [< print_rec rest;
             'S(nametab__print_name sp) ; 'S" ==> ";  fprint_recipe d; 'FNL >]
     with Not_found -> print_rec rest)

  | (_,"VARIABLE") ->
    let (name,(_,cont),_,_,_) = outVariable lobj in
        [< print_rec rest;
         (match cont with SOME(t) ->
                            fprint_var (string_of name) t
                       | _      -> [< >]) >]
  | (sp,"MUTUALINDUCTIVE") ->
     [< print_rec rest; 
      (try [< print_extracted_mutual sp ; 'FNL >] with Not_found -> [<>]) >]

  | _ -> print_rec rest)

  | (sp,lib__ClosedDir _)::rest -> print_rec rest

  | _::rest -> print_rec rest
     
  | [] -> [< 'FNL >];;

let print_extracted_context () =
let env = lib__contents_after NONE in [< print_var_rec env; print_last_constants env >]
where rec print_var_rec = function 
    ((_,lib__LEAF lobj))::rest ->
    if "VARIABLE" = object_tag lobj then
        let (name,(typ,cont),_,_,_) = outVariable lobj in
        [< print_var_rec rest ; 'FNL;
           match cont with
             SOME(t) -> fprint_var (string_of name) t
            | _      -> [< >] >]
    else print_var_rec rest

  |  _::rest -> print_var_rec rest
  | [] -> [< 'FNL >]

and print_last_constants = function 
    ((sp,lib__LEAF lobj))::rest ->
    (match object_tag lobj with 
       "CONSTANT" -> 
    let (_,{CONSTBODY=c;CONSTTYPE=typ;_}) = const_of_path (fwsp_of sp)
    in
         [< print_last_constants rest;
         let s=nametab__print_name sp
         in (try let (_,{CONSTBODY = d}) = const_of_path (fwsp_of sp)
                 in [< 'S (s ^" ==> ");
                     fprint_recipe d; 'FNL >]
             with Not_found -> [< >]) >]
     | "MUTUALINDUCTIVE" -> 
     [< print_last_constants rest; 
        try print_extracted_mutual sp with Not_found -> [<>] >]

     | "VARIABLE" -> let (_,(_,cont),_,_,_) = outVariable lobj in
              (match cont with SOME _ -> [<>]
                             | NONE -> print_last_constants rest)
     | _ ->  print_last_constants rest)

 |  (_,lib__ClosedDir _)::rest -> print_last_constants rest
 | _ -> [< >];;

dflt_gp.ellipsis <- " ..<elided>.. ";;
dflt_gp.format_ellipsis <- false;; (* Pour ne pas avoir une bardee de .... *)
dflt_gp.limit_depth <- 1000;; (* So that the printer does not abort *)
dflt_gp.max_depth <- 80;;



let print_extracted_vars () =
let env = lib__contents_after NONE in print_var_rec env
where rec print_var_rec = function
    ((_,lib__LEAF lobj))::rest ->
    if "VARIABLE" = object_tag lobj then
        let (name,(_,cont),_,_,_) = outVariable lobj in
        [< print_var_rec rest ; 'FNL;
           match cont with
             SOME(t) -> fprint_var (string_of name) t
            | _      -> [< >] >]
    else print_var_rec rest

  |  _::rest -> print_var_rec rest
  | [] -> [< 'FNL >]
;;

(* for debug *)
let inspect depth = 
    let rec inspectrec n res env = if n=0 or env=[] then res
                                   else inspectrec (n-1) (hd env::res) (tl env)
    in let items = rev (inspectrec depth [] (lib__contents_after NONE))
       in print_context false items;;

(* $Id: pretty.ml,v 1.12 1996/02/12 15:39:02 cpaulin Exp $ *)
