(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                              tacentries.ml                               *)
(****************************************************************************)

(* Concrete syntax of the mathematical vernacular MV V2.6 *)

#infix "o";;
#open "std";;
#open "names";;
#open "tacmach";;
#open "ast";;
#open "lexer";;
#open "himsg";;
#open "more_util";;
#open "pp";;
#open "stdpp";;
#open "proof";;
#open "proof_trees";;
#open "tactics";;


let vernac_absurd =
    register_comarg_tactic "Absurd" absurd;;

let contradict =
    register_atomic_tactic "Contradiction"
    (THEN intros (THEN (elim_type (ASTnvar (id_of_string "False"))) assumption));;

(**************************************************************)
(*    Introduction of conjonction, disjunction                *)
(*    existencial quantifier and equality tactics             *)
(**************************************************************)

let vernac_idtac = register_atomic_tactic "IDTAC" IDTAC;;

let vernac_intro = register_atomic_tactic "Intro" intro;;

let vernac_intros_with =
let gentac =
    register_tactic "IntrosWith"
    (fun al -> intros_with (map (fun (IDENTIFIER id) -> id) al))
    (fun _ _ (_,al) ->
         [< 'S"Intros" ; 'SPC ;
             prlist_with_sep (fun () -> [< 'SPC >])
                             (fun (IDENTIFIER id) -> print_id id)
                             al >])
in fun ids -> gentac (map (fun id -> IDENTIFIER id) ids)
;;

let vernac_fix = 
let gentac = 
    register_tactic "FixPoint" 
    (fun [IDENTIFIER id;INTEGER n] -> fix id n)
    (fun _ _ (_,[IDENTIFIER id; INTEGER n]) ->
	  [< 'S"Fix" ; 'SPC ; print_id id; 'S"/"; 'INT n >])
in fun id n -> gentac [IDENTIFIER id;INTEGER n];;

let vernac_intros_until =
let gentac =
    register_tactic "IntrosUntil"
    (fun [IDENTIFIER id] -> intros_until id)
    (fun _ _ (_,[IDENTIFIER id]) ->
         [< 'S"Intros" ; 'SPC ; 'S"Until" ; 'SPC ; print_id id >])
in fun id -> gentac [IDENTIFIER id]
;;

let vernac_intros = register_atomic_tactic "Intros" intros;;

let intros_split_tac =
    register_bindinglist_tactic "Split" split_with_bindings;;

let intros_right_tac =
     register_bindinglist_tactic "Right" right_with_bindings;;

let intros_left_tac =
    register_bindinglist_tactic "Left" left_with_bindings;;

let intros_exists_tac =
    register_bindinglist_tactic "Exists" split_with_bindings;;

let var_of_string c = ASTnvar (id_of_string c);;

let replace_tac_vernac =
let gentac =
    register_tactic "Replace"
    (fun [COMMAND c1;COMMAND c2] -> 
         replace_tac (var_of_string "eq") (var_of_string "eqT") c1 c2 
         (var_of_string "sym_equal") (var_of_string "sym_eqT"))
    (fun sigma goal (_,[COMMAND c1;COMMAND c2]) ->
         [< 'S"Replace"; 'SPC ; pr_com sigma goal c1 ; 'SPC ; 'S"with" ;
            'SPC ; pr_com sigma goal c2 >])
in fun c1 c2 -> gentac [COMMAND c1;COMMAND c2]
;;

let rewrite_tac_LR =
    register_comarg_tactic "RewriteLR" 
   (fun c -> (equality_tac 
   (var_of_string "eq",var_of_string "eq_ind_r",var_of_string "eq_rec_r") 
   (var_of_string "eqT",var_of_string "eqT_ind_r",var_of_string "eqT_rec_r") 
                c))
;;

let rewrite_tac_RL =
    register_comarg_tactic "RewriteRL" elim;;


let reflexivity_tac gl =
   match string_head (pf_concl gl) with
     "eq" -> resolve (var_of_string "refl_equal") gl
  |  "eqT" -> resolve (var_of_string "refl_eqT") gl
  |    _   -> error "Not a predefined equality";;
let intros_reflexivity_tac =
    register_atomic_tactic "Reflexivity" (THEN intros reflexivity_tac);;

let symmetry_tac gl =
   match string_head (pf_concl gl) with
     "eq" -> resolve (var_of_string "sym_equal") gl
  |  "eqT" -> resolve (var_of_string "sym_eqT") gl
  |    _   -> error "Not a predefined equality";;
let intros_symmetry_tac =
    register_atomic_tactic "Symmetry" (THEN intros symmetry_tac);;

let transitivity_tac n gl =
   match string_head (pf_concl gl) with
     "eq" -> resolve_with [var_of_string "trans_equal";n] gl
  |  "eqT" -> resolve_with [var_of_string "trans_eqT";n] gl
  |    _   -> error "Not a predefined equality";;
let intros_transitivity_tac =
    register_comarg_tactic "Transitivity"
    (fun n -> THEN intros (transitivity_tac n));;

(* Named tactics *)
let vernac_assumption = register_atomic_tactic "Assumption" assumption;;
let vernac_give_exact_com = register_comarg_tactic "Exact" give_exact_com;;
let vernac_red_hyp = register_ident_tactic "RedHyp" red_hyp;;
let vernac_red = register_atomic_tactic "Red" red;;
let vernac_simpl_hyp = register_ident_tactic "SimplHyp" simpl_hyp;;
let vernac_simpl_tac = register_atomic_tactic "Simpl" simpl_tac;;
let vernac_simpl_tac = register_atomic_tactic "Compute" compute_tac;;
let vernac_hnf_tac_hyp = register_ident_tactic "HnfHyp" hnf_tac_hyp;;
let vernac_hnf_tac = register_atomic_tactic "Hnf" hnf_tac;;

let vernac_change_hyp =
let gentac = register_tactic "ChangeHyp"
    (fun [COMMAND c;IDENTIFIER id] -> change_hyp c id)
    (fun sigma goal (_,[COMMAND c;IDENTIFIER id]) ->
         [< 'S"Change" ; 'SPC ; pr_com sigma goal c ; 'SPC ;
            'S"in" ; 'SPC ; print_id id >])
in fun c id -> gentac [COMMAND c;IDENTIFIER id]
;;

let vernac_change = register_comarg_tactic "Change" change;;

let pr_pattern sigma goal nl com =
    [< prlist_with_sep
           (fun () -> [< 'SPC >])
           (fun n -> [< 'INT n >])
           nl;
       pr_com sigma goal com >];;

let vernac_pattern =
let gentac = register_tactic "Pattern"
    (fun patl -> pattern (map (function (PATTERN x) -> x) patl))
    (fun sigma goal (_,patl) ->
         [< 'S"Pattern" ; 'SPC ;
            prlist_with_sep
                (fun () -> [< 'SPC >])
                (fun (PATTERN(nl,com)) -> pr_pattern sigma goal nl com)
                patl >])
in fun patl -> gentac (map (function x -> (PATTERN x)) patl)
;;

let vernac_resolve_with_bindings = 
    register_combindinglist_tactic "ApplyWithBindings" resolve_with_bindings;;


let vernac_constructor_with_bindings = 
let gentac =
    register_tactic "ConstructorWithBindings"
    (fun ((INTEGER i)::al) -> 
             constructor_with_bindings i (map (fun (BINDING b) -> b) al))
    (fun sigma goal (_,((INTEGER i)::al)) ->
         [< 'S"Constructor" ; 'SPC ; 'INT i ; 'SPC ; 
               pr_with_bindings_list sigma goal al >])
in fun n ids -> gentac ((INTEGER n)::(map (fun b -> BINDING b) ids))
;;

let vernac_specialize_with_num =
let gentac = register_tactic "SpecializeWithNum"
    (fun ((INTEGER n)::(COMMAND com)::bindl) ->
         new_hyp_with_num n com (map (function (BINDING b) -> b) bindl))
    (fun sigma goal (_,(INTEGER n)::(COMMAND com)::bindl) ->
         [< 'S"Specialize" ; 'SPC ; 'INT n ; 'SPC ; pr_com sigma goal com ; 
         'SPC ; pr_with_bindings_list sigma goal bindl >])
in fun n com bindl -> gentac ((INTEGER n)::(COMMAND com)::(map (fun b -> BINDING b) bindl))
;;

let vernac_specialize_with =
let gentac = register_tactic "SpecializeWith"
    (fun ((COMMAND com)::bindl) ->
         new_hyp_with com (map (function (BINDING b) -> b) bindl))
    (fun sigma goal (_,(COMMAND com)::bindl) ->
         [< 'S"Specialize" ; 'SPC ; 'SPC ; pr_com sigma goal com ; 'SPC ;
        pr_with_bindings_list sigma goal bindl >])
in fun com bindl -> gentac ((COMMAND com)::(map (fun b -> BINDING b) bindl))
;;

let vernac_specialize = register_comarg_tactic "Specialize" new_hyp;;

let vernac_generalize =
let gentac =
    register_tactic "Generalize"
    (fun al -> generalize (map (fun (COMMAND com) -> com) al))
    (fun sigma goal (_,al) ->
         let hdcom::coml = map (fun (COMMAND com) -> com) al
         in [< 'S"Generalize" ; 'SPC ; pr_com sigma goal hdcom ; 'SPC ;
             prlist_with_sep (fun () -> [< 'SPC >])
                             (pr_com sigma goal)
                             coml >])
in fun ids -> gentac (map (fun id -> COMMAND id) ids)
;;

let vernac_elim_with_bindings = 
     register_combindinglist_tactic "ElimWithBindings" elim_with_bindings;;

let collect_bindings = colrec [] where rec 
 colrec flb = function (BINDING _ as b)::rest -> colrec (b::flb) rest
                     | rest              -> (rev flb,rest);;

let vernac_elim_with_using =
let gentac =
    register_tactic "ElimWithUsing"
    (fun ((COMMAND com)::l) -> 
     let (lbc,(COMMAND elim)::lbe) = (collect_bindings l) in 
     elim_with_using (com,map (fun (BINDING b) -> b) lbc) 
                     (elim,map (fun (BINDING b) -> b) lbe))
    (fun sigma goal (_,(COMMAND com)::l) ->
         let (lb,(COMMAND elim)::lbe) = collect_bindings l in
         [< 'S"Elim" ; 'SPC ; pr_com sigma goal com ;'SPC ; 
                pr_with_bindings_list sigma goal lb;
               'SPC ; 'S"Using" ; 'SPC;  pr_com sigma goal elim;'SPC ; 
             pr_with_bindings_list sigma goal lbe >])
in fun (cid,cbl) (eid,ebl) -> 
   gentac ((COMMAND cid)::(map (fun b -> BINDING b) cbl)@
                                 (COMMAND eid::(map (fun b -> BINDING b) ebl)))
;;

let vernac_elim = register_comarg_tactic "Elim" elim;;
let vernac_elim1 = register_comarg_tactic "Elim1" elim1;;
let vernac_elim_type = register_comarg_tactic "ElimType" elim_type;;

let vernac_case_with_bindings =
 register_combindinglist_tactic "CaseWithBindings" case_with_bindings;;

let (vernac_induct,vernac_induct_nodep) =
let gentac = register_tactic "Induction"
    (fun [IDENTIFIER id] -> induct id
       | [INTEGER n] -> induct_nodep n)
    (fun _ _ (_,[IDENTIFIER id]) -> [< 'S"Induction" ; 'SPC ; print_id id >]
       | _ _ (_,[INTEGER n]) -> [< 'S"Induction" ; 'SPC ; 'INT n >])
in ((fun id -> gentac [IDENTIFIER id]),
    (fun id -> gentac [INTEGER id]))
;;

let (vernac_destruct,vernac_destruct_nodep) =
let gentac = register_tactic "Destruct"
    (fun [IDENTIFIER id] -> destruct id
       | [INTEGER n] -> destruct_nodep n)
    (fun _ _ (_,[IDENTIFIER id]) -> [< 'S"Destruct" ; 'SPC ; print_id id >]
       | _ _ (_,[INTEGER n]) -> [< 'S"Destruct" ; 'SPC ; 'INT n >])
in ((fun id -> gentac [IDENTIFIER id]),
    (fun n -> gentac [INTEGER n]))
;;


let vernac_cut = register_comarg_tactic "Cut" cut;;

let vernac_imp_elim = register_comarg_tactic "ImpElim" imp_elim_tac;;

let pr_unfold nl id =
    [< prlist_with_sep
           (fun () -> [< 'SPC >])
           (fun n -> [< 'INT n >])
           nl; 'SPC ;
       print_id id >];;

let vernac_unfold_nth =
let gentac = register_tactic "Unfold"
    (fun unfl -> unfold_nth_tac (map (fun (UNFOLD x) -> x) unfl))
    (fun _ _ (_,unfl) ->
         [< 'S"Unfold"; 'SPC ;
            prlist_with_sep
                (fun () -> [< 'SPC >])
                (fun (UNFOLD(nl,id)) -> pr_unfold nl id)
                unfl >])
in fun l -> gentac (map (fun x -> UNFOLD x) l)
;;

let vernac_unfold_nth_hyp =
let gentac = register_tactic "UnfoldHyp"
    (fun ((IDENTIFIER id)::unfl) -> unfold_nth_hyp (map (fun (UNFOLD x) -> x) unfl) id)
    (fun _ _ (_,(IDENTIFIER id)::unfl) ->
         [< 'S"Unfold"; 'SPC ;
            prlist_with_sep
                (fun () -> [< 'SPC >])
                (fun (UNFOLD(nl,id)) -> pr_unfold nl id)
                unfl ;
            'SPC ; 'S"in" ; 'SPC ; print_id id >])
in fun l id -> gentac ((IDENTIFIER id)::(map (fun x -> UNFOLD x) l))
;;

let vernac_clear_hyp =
let gentac =
    register_tactic "Clear"
    (fun al -> clear_hyp (map (fun (IDENTIFIER id) -> id) al))
    (fun _ _ (_,al) ->
         [< 'S"Clear" ; 'SPC ;
             prlist_with_sep (fun () -> [< 'SPC >])
                             (fun (IDENTIFIER id) -> print_id id)
                             al >])
in fun ids -> gentac (map (fun id -> IDENTIFIER id) ids)
;;

let vernac_interp =
let gentac =
    register_tactic "Interpret"
    (fun [TACEXP com] gl -> tacinterp__interp com gl)
    (fun _ _ (_,[TACEXP com]) ->
         [< pTACTIC com >])
in fun com -> gentac ([TACEXP com])
;;

(* $Id: tacentries.ml,v 1.6 1996/02/12 15:39:07 cpaulin Exp $ *)
