(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                               tacinterp.ml                               *)
(****************************************************************************)

#infix "o";;
#open "std";;
#open "names";;
#open "more_util";;
#open "proof";;
#open "proof_trees";;
#open "pp";;
#open "stdpp";;
#open "tacmach";;
#open "ast";;
#open "himsg";;

let cvt_arg = function
    ASTnvar id -> IDENTIFIER id
  | ASTprim(TOKstr s) -> QUOTED_STRING s
  | ASTprim(TOKnum n) -> INTEGER n
  | ASTnode("COMMAND",[c]) -> COMMAND c

  | ASTnode("BINDING",
            [ASTprim(TOKnum n);
             ASTnode("COMMAND",[c])]) -> BINDING(NoDep n,c)

  | ASTnode("BINDING",
            [ASTnvar id;
             ASTnode("COMMAND",[c])]) -> BINDING(Dep id,c)

  | ASTnode("BINDING",[ASTnode("COMMAND",[c])]) -> BINDING(Com,c)

  | ASTnode("PATTERN",argl) ->
    let (com::nums) = argl in
    let intl = (map (function (ASTprim(TOKnum n)) -> n
                          | _ -> error "malformed pattern-expression")
                nums)
    in (match com with
        (ASTnode("COMMAND",[c])) -> PATTERN(intl,c)
      | _ -> error "malformed pattern-expression")

  | ASTnode("UNFOLD",argl) ->
    let (com::nums) = argl in
    let intl = (map (function (ASTprim(TOKnum n)) -> n
                          | _ -> error "malformed unfold-expression")
                nums)
    in (match com with
        (ASTnvar id) -> UNFOLD(intl,id)
      | _ -> error "malformed unfold-expression")

  | ASTnode("TACTIC",[ast]) -> TACEXP ast

  | ASTnode("FIXEXP",
      [ASTnvar id; ASTprim(TOKnum n);ASTnode("COMMAND",[c])]) ->
           (FIXEXP (id,n,c))

  | ASTnode("COFIXEXP",
      [ASTnvar id; ASTnode("COMMAND",[c])]) ->
           (COFIXEXP (id,c))

  | _ -> failwith "cvt_arg"
;;

#infix "ORELSE";;
#infix "THEN";;
#infix "THENS";;

let tactic_tab = (mhm__new 17 : (string, tactic_arg list -> tactic) mhm__t);;

let tacinterp_add (s,f) =
    try mhm__add tactic_tab (s,f)
    with Failure _ ->
        errorlabstrm "tacinterp_add"
        [< 'S"Cannot add the tactic " ; 'S s ; 'S" twice" >]
;;

let overwriting_tacinterp_add (s,f) =
    if mhm__in_dom tactic_tab s then
        (mhm__rmv tactic_tab s;
         warning ("Overwriting definition of tactic interpreter command "^s));
    mhm__add tactic_tab (s,f)
;;

let tacinterp_map s =
    try mhm__map tactic_tab s
    with Not_found -> errorlabstrm "Tactic Interpreter"
                      [< 'S"Cannot find tactic " ; 'S s >];;
let tacinterp_init () = mhm__empty tactic_tab;;

let rec interp = function
    ASTnode(opn,tl) as ast ->
    (match opn with
        "TACTICLIST" -> interp_semi_list IDTAC tl
      | "TACLIST" -> error "malformed tactic-expression"

      | "DO" ->
        if list_length tl = 2 then
            (try
             (match cvt_arg (car tl) with
            (INTEGER n) -> DO n (interp (cadr tl))
          | _ -> error "malformed tactic-expression")
             with Failure "cvt_varg" ->
                 errorlabstrm "tacinterp__interp"
                 [< 'S"Malformed DO: " ; print_ast ast >])

        else error "malformed tactic-expression"

      | "TRY" ->
        if list_length tl = 1 then
            TRY (interp (car tl))
        else error "malformed tactic-expression"

      | "REPEAT" ->
        if list_length tl = 1 then
            REPEAT (interp (car tl))
        else error "malformed tactic-expression"

      | "ORELSE" ->
        if list_length tl = 2 then
            (interp (car tl)) ORELSE (interp (cadr tl))
        else error "malformed tactic-expression"

      | _ -> (try tacinterp_map opn (map cvt_arg tl)
              with Failure _ | UserError _ ->
              try Tactic(opn,map cvt_arg tl)
              with Failure "cvt_varg" ->
                  errorlabstrm "tacinterp__interp"
                  [< 'S"Malformed atomic tactic expression " ; 'SPC ;
                     print_ast ast >]))

and interp_semi_list acc = function
    (ASTnode("TACLIST",seq))::l ->
    interp_semi_list (acc THENS (map interp seq)) l
  | t::l -> interp_semi_list (acc THEN (interp t)) l
  | [] -> acc
;;

let interp e gl = interp e gl;;

(* $Id: tacinterp.ml,v 1.5 1996/02/12 15:39:11 cpaulin Exp $ *)
