(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                tacmach.ml                                *)
(****************************************************************************)

#infix "o";;
#open "std";;
#open "initial";;
#open "names";;
#open "vectops";;
#open "generic";;
#open "term";;
#open "reduction";;
#open "himsg";;
#open "termenv";;
#open "machops";;
#open "mach";;
#open "trad";;
#open "more_util";;
#open "proof";;
#open "proof_trees";;
#open "pp";;
#open "stdpp";;
#open "logic";;
#open "evar_refiner";;
#open "refiner";;

type 'a sigma == 'a refiner__sigma;;
let sig_it = sig_it;;
let Project = refiner__Project;;

let StartWalk = StartWalk;;
let Walking_THEN = Walking_THEN;;
let Walking = Walking;;
let w_Focusing_THEN = w_Focusing_THEN;;
let w_Declare = w_Declare;;
let w_Declare_At = w_Declare_At;;
let w_Define = w_Define;;
let w_Underlying = w_Underlying;;
let w_hyps = w_hyps;;
let w_type_of = w_type_of;;
let w_IDTAC = w_IDTAC;;
let w_ORELSE = w_ORELSE;;
let w_add_sign = w_add_sign;;
let ctxt_type_of = ctxt_type_of;;

let IDTAC = IDTAC;;
let ORELSE = ORELSE;;
let THEN =  THEN;;
let THEN_i =  THEN_i;;
let THENL =  THENL;;
let THENS =  THENS;;
let REPEAT =  REPEAT;;
let FIRST =  FIRST;;
let TRY =  TRY;;
let COMPLETE =  COMPLETE;;
let AT_LEAST_ONE =  AT_LEAST_ONE;;
let FAILTAC =  FAILTAC;;
let DO =  DO;;
let PROGRESS =  PROGRESS;;

let register_tactic = register_tactic;;
let overwriting_register_tactic = overwriting_register_tactic;;

let unTAC = unTAC;;
let Tactic = Tactic;;
let Context = Context;;

let proof_of_pftreestate =  proof_of_pftreestate;;
let cursor_of_pftreestate =  cursor_of_pftreestate;;
let evc_of_pftreestate =  evc_of_pftreestate;;

let traverse =  traverse;;
let solve_nth_pftreestate =  solve_nth_pftreestate;;
let solve_pftreestate =  solve_pftreestate;;
let mk_pftreestate =  mk_pftreestate;;
let extract_pftreestate =  extract_pftreestate;;
let first_unproven = first_unproven;;
let last_unproven  = last_unproven;;
let node_prev_unproven = node_prev_unproven;;
let node_next_unproven = node_next_unproven;;
let next_unproven = next_unproven;;
let prev_unproven = prev_unproven;;
let top_of_tree = top_of_tree;;

let frontier =  frontier;;

let pr_gls gls =
    HOV 0 [< pDECLS (sig_sig gls) ; 'FNL ; pSEQ (sig_it gls) >]
;;

let pr_glls glls =
    HOV 0 [< pDECLS (sig_sig glls) ; 'FNL ;
             prlist_with_sep (fun () -> [< 'FNL >]) pSEQ (sig_it glls) >]
;;

let pr_rule sigma goal = function
    (PRIM r) -> pr_prim_rule r

  | (TACTIC texp) -> pr_tactic sigma goal texp

  | CONTEXT ctxt -> pCTXT ctxt

  | LOCAL_CONSTRAINTS lc -> [< 'S"Local constraint change" >]

;;

let print_rule sigma goal r = pr_rule sigma goal r;;

let thin_sign osign nsign =
    let com_nsign = combine nsign
    in split(map_succeed (fun (id,ty) ->
                              if not(mem_sign osign id)
                                  or (id,ty) <> (lookup_sign id osign) then
                                  (id,ty)
                              else failwith "caught") com_nsign)
;;

let rec print_proof sigma osign pf =
let {hyps=sign;concl=cl;ctxt=ctxt;lc=lc} = pf.goal in
let sign' = thin_sign osign sign
in match pf.ref with

    NONE -> HOV 0 [< pSEQ {hyps=sign';concl=cl;ctxt=ctxt;lc=lc} >]

  | SOME(r,spfl) ->
    HOV 0 [< HOV 0 (pSEQ {hyps=sign';concl=cl;ctxt=ctxt;lc=lc});
            'SPC ; 'S" BY ";
            HOV 0 [< pr_rule sigma pf.goal r >]; 'FNL ;
            'WS 2 ; HOV 0 (prlist_with_sep (fun () -> [< 'FNL >])
                                           (print_proof sigma sign) spfl)
            >]
;;

let pr_change gl = [< 'S"Change " ; pTERM gl.concl ; 'S"." ; 'FNL>];;

let rec print_script nochange sigma osign pf =
let {hyps=sign;concl=cl;_} = pf.goal in
let sign' = thin_sign osign sign
in match pf.ref with

    NONE -> if nochange then [< 'S"<Your Tactic Text here>" >]
            else pr_change pf.goal

  | SOME(r,spfl) ->
    HOV 0 [< (if nochange then [< >] else (pr_change pf.goal));
             pr_rule sigma pf.goal r ; 'S"." ; 'FNL ;
             prlist_with_sep (fun () -> [< 'FNL >])
                             (print_script nochange sigma sign) spfl >]
;;

let rec print_treescript sigma osign pf =
let {hyps=sign;concl=cl;_} = pf.goal in
let sign' = thin_sign osign sign
in match pf.ref with

    NONE -> [< >]

  | SOME(r,spfl) ->
    HOV 0 [< pr_rule sigma pf.goal r ; 'S"." ; 'FNL ;
             'S(if length spfl > 1 then "  "  else "");
             HOV 0 [< prlist_with_sep (fun () -> [< 'FNL >])
                             (print_treescript sigma sign) spfl >] >]
;;

let refiner = refiner;;

let introduction id pf =
    refiner (PRIM{name="INTRO";newids=[id];
                  hypspecs=[];terms=[];params=[]}) pf;;

let intro_after whereid id pf = 
    refiner (PRIM{name="INTRO-AFTER";newids=[id];
                  hypspecs=[whereid];terms=[];params=[]}) pf;;

let intro_replacing whereid pf = 
    refiner (PRIM{name="INTRO-REPLACING";newids=[];
                  hypspecs=[whereid];terms=[];params=[]}) pf;;

let Refine c pf = refiner (PRIM{name="REFINE";terms=[c];
                                hypspecs=[];newids=[];params=[]}) pf;;
let convert_concl c pf = refiner (PRIM{name="CONVERT_CONCL";terms=[c];
                                       hypspecs=[];newids=[];params=[]}) pf;;
let convert_hyp id c pf = refiner (PRIM{name="CONVERT_HYP";hypspecs=[id];
                                        terms=[c];newids=[];params=[]}) pf;;
let thin ids gl = refiner (PRIM{name="THIN";hypspecs=ids;
                                terms=[];newids=[];params=[]}) gl;;

let fix f n pf =  refiner (PRIM{name="FIX";newids=[f];
                  hypspecs=[];terms=[];params=[ast__TOKnum n]}) pf;;

let mutual_fix lf ln lar pf = refiner (PRIM{name="FIX";newids=lf;
                  hypspecs=[];terms=lar;
                  params=map (fun n -> ast__TOKnum n) ln}) pf;;

let cofix f pf =  refiner (PRIM{name="COFIX";newids=[f];
                  hypspecs=[];terms=[];params=[]}) pf
;;

let mutual_cofix lf lar pf = 
                refiner (PRIM{name="COFIX";newids=lf;
                  hypspecs=[];terms=lar; params=[]}) pf
;;

(* This renames bound variablew with fresh and distinct names *)
(* in such a way that the printer doe not generate new names  *)
(* and therefore that printed names are the intern names      *)
(* In this way, tactics such as Induction works well          *)

let rec rename_bound_var l = function
   DOP2(Prod,c1,DLAM(Name(s),c2))  -> if dependent (Rel 1) c2 then
               let s' = next_ident_away s (global_vars c2@l)
                in DOP2(Prod,c1,DLAM(Name(s'),rename_bound_var (s'::l) c2))
                     else DOP2(Prod,c1,DLAM(Name(s),rename_bound_var l c2))
 | DOP2(Prod,c1,DLAM(Anonymous,c2)) -> DOP2(Prod,c1,DLAM(Anonymous,rename_bound_var l c2))
 | DOP2(Cast,c,t) -> DOP2(Cast,rename_bound_var l c,t)
 |  x -> x;;

let rename_bound_var_goal gls =
let ({hyps=sign;concl=cl;_} as gl) = (sig_it gls)
in convert_concl (rename_bound_var (ids_of_sign sign) cl) gls
;;

let pf_type_of gls c =
    type_of (Project gls) (sig_it gls).hyps c
;;

let hnf_type_of gls = (whd_betadeltaiota (Project gls)) o (pf_type_of gls);;

let pf_constr_of_com gls c =
let evc = Project gls
in constr_of_com evc (GLOB (evc__get_hyps evc)) c
;;

let pf_global gls id = global (GLOB (sig_it gls).hyps) id;;

let pf_fexecute gls =
let evc = Project gls
in fexecute evc (evc__get_hyps evc);;

let pf_infexecute gls =
let evc = Project gls
in infexecute evc (FSIGN evc (evc__get_hyps evc));;

let pf_whd_betadeltaiota gls c = whd_betadeltaiota (Project gls) c;;
let pf_whd_betadeltaiota_stack gls c stack = whd_betadeltaiota_stack (Project gls) c stack;;
let pf_hnf_constr gls c = hnf_constr (Project gls) c;;
let pf_red_product gls c = red_product (Project gls) c;;
let pf_nf gls c = nf (Project gls) c;;
let pf_compute gls c = compute (Project gls) c;;
let pf_unfoldn gls = unfoldn (Project gls);;
let pf_conv_x gls M N = conv_x (Project gls) M N;;
let pf_hyps gls = (sig_it gls).hyps;;
let pf_concl gls = (sig_it gls).concl;;
let pf_const_value gls = const_value (Project gls);;
let pf_one_step_reduce gls = one_step_reduce (Project gls);;

let pf_h gls n = nth_sign (pf_hyps gls) n;;
let pf_ctxt gls = (sig_it gls).ctxt;;

let w_defined_const wc K = defined_const (w_Underlying wc) K;;
let w_const_value wc = const_value (w_Underlying wc);;
let w_conv_x wc M N = conv_x (w_Underlying wc) M N;;
let w_whd_betadeltaiota wc c = whd_betadeltaiota (w_Underlying wc) c;;
let w_hnf_constr wc c = hnf_constr (w_Underlying wc) c;;


let register_atomic_tactic s tac =
    (add_tactic s ((fun [] -> tac),(fun _ _ (s,[]) -> [< 'S s >]));
     Tactic(s,[]))
;;


let overwriting_register_atomic_tactic s tac =
    (overwriting_add_tactic s ((fun [] -> tac),(fun _ _ (s,[]) -> [< 'S s >]));
     Tactic(s,[]))
;;


let pr_com sigma goal com =
    pTERM (rename_bound_var (ids_of_sign goal.hyps) (constr_of_com sigma (GLOB goal.hyps) com));;

let register_comarg_tactic s tac =
let tacfun = fun [COMMAND com] -> tac com
and prfun = fun sigma gl (s,[COMMAND com]) -> [< 'S s ; 'WS 1 ; pr_com sigma gl com >]
in (add_tactic s (tacfun,prfun);
    fun com -> Tactic(s,[COMMAND com]))
;;

let overwriting_register_comarg_tactic s tac =
let tacfun = fun [COMMAND com] -> tac com
and prfun = fun sigma gl (s,[COMMAND com]) -> [< 'S s ; 'WS 1 ; pr_com sigma gl com >]
in (overwriting_add_tactic s (tacfun,prfun);
    fun com -> Tactic(s,[COMMAND com]))
;;

let register_numarg_tactic s tac =
let tacfun = fun [INTEGER n] -> tac n
and prfun = fun _ _ (s,[INTEGER n]) -> [< 'S s ; 'WS 1 ; 'INT n >]
in (add_tactic s (tacfun,prfun);
    fun n -> Tactic(s,[INTEGER n]))
;;

let overwriting_register_numarg_tactic s tac =
let tacfun = fun [INTEGER n] -> tac n
and prfun = fun _ _ (s,[INTEGER n]) -> [< 'S s ; 'WS 1 ; 'INT n >]
in (overwriting_add_tactic s (tacfun,prfun);
    fun n -> Tactic(s,[INTEGER n]))
;;

let register_ident_tactic s tac =
let tacfun = fun [IDENTIFIER id] -> tac id
and prfun = fun _ _ (s,[IDENTIFIER id]) -> [< 'S s ; 'WS 1 ; print_id id >]
in (add_tactic s (tacfun,prfun);
    fun id -> Tactic(s,[IDENTIFIER id]))
;;

let overwriting_register_ident_tactic s tac =
let tacfun = fun [IDENTIFIER id] -> tac id
and prfun = fun _ _ (s,[IDENTIFIER id]) -> [< 'S s ; 'WS 1 ; print_id id >]
in (overwriting_add_tactic s (tacfun,prfun);
    fun id -> Tactic(s,[IDENTIFIER id]))
;;

let register_string_tactic s tac =
let tacfun = fun [QUOTED_STRING str] -> tac str
and prfun = fun _ _ (s,[QUOTED_STRING str]) -> [< 'S s ; 'WS 1 ; 'QS str >]
in (add_tactic s (tacfun,prfun);
    fun str -> Tactic(s,[QUOTED_STRING str]))
;;

let overwriting_register_string_tactic s tac =
let tacfun = fun [QUOTED_STRING str] -> tac str
and prfun = fun _ _ (s,[QUOTED_STRING str]) -> [< 'S s ; 'WS 1 ; 'QS str >]
in (overwriting_add_tactic s (tacfun,prfun);
    fun str -> Tactic(s,[QUOTED_STRING str]))
;;


let register_comarglist_tactic s sep1 tac = 
let tacfun = fun al -> tac (map (fun (COMMAND com) -> com) al)
and prfun = fun sigma goal (s,al) ->
         let coml = map (fun (COMMAND com) -> com) al in
         if coml = [] then  [< 'S s >] 
         else [< 'S s ; 'SPC ; pr_com sigma goal (hd coml) ; 
                        'SPC ; 'S sep1 ; 
             prlist_with_sep (fun () -> [< 'SPC >])
                             (pr_com sigma goal)
                             (tl coml) >]
in (add_tactic s (tacfun,prfun);
    fun ids -> Tactic(s,(map (fun id -> COMMAND id) ids)))
;;

let overwriting_register_comarglist_tactic s sep1 tac = 
let tacfun = fun al -> tac (map (fun (COMMAND com) -> com) al)
and prfun = fun sigma goal (s,al) ->
         let coml = map (fun (COMMAND com) -> com) al in
         if coml = [] then  [< 'S s >] 
         else [< 'S s ; 'SPC ; pr_com sigma goal (hd coml) ; 
                        'SPC ; 'S sep1 ; 
             prlist_with_sep (fun () -> [< 'SPC >])
                             (pr_com sigma goal)
                             (tl coml) >]
in (overwriting_add_tactic s (tacfun,prfun);
    fun ids -> Tactic(s,(map (fun id -> COMMAND id) ids)))
;;

let pr_binding sigma goal = function
    (Dep id,com)  -> [< print_id id ; 'S":=" ; pr_com sigma goal com >]
  | (NoDep n,com) -> [< 'INT n ; 'S":=" ; pr_com sigma goal com >]
  | (Com,com)     -> [< pr_com sigma goal com >]
;;

let pr_bindings_list sigma goal lbe =
  let lb = map (fun (BINDING b) -> b) lbe in 
  let prf = pr_binding sigma goal in 
  [< prlist_with_sep (fun () -> [<'SPC>]) prf lb >];;

let pr_with_bindings_list sigma goal lbe = if lbe = [] then [<>]
  else  [< 'S"with"; 'SPC ; pr_bindings_list sigma goal lbe >];;

let register_bindinglist_tactic s tac = 
let tacfun = fun al -> tac (map (fun (BINDING b) -> b) al)
and prfun = fun sigma goal (s,bl) -> [< 'S s ; 'SPC ; pr_bindings_list sigma goal bl >]
in (add_tactic s (tacfun,prfun);
    fun ids -> Tactic(s,(map (fun b -> BINDING b) ids)))
;;

let overwriting_register_bindinglist_tactic s tac = 
let tacfun = fun al -> tac (map (fun (BINDING b) -> b) al)
and prfun = fun sigma goal (s,bl) -> [< 'S s ; 'SPC ; pr_bindings_list sigma goal bl >]
in (overwriting_add_tactic s (tacfun,prfun);
    fun ids -> Tactic(s,(map (fun b -> BINDING b) ids)))
;;

let register_combindinglist_tactic s tac = 
let tacfun = fun ((COMMAND com)::al) -> tac (com,(map (fun (BINDING b) -> b) al))
and prfun = fun sigma goal (s,(COMMAND com)::bl) -> 
            [< 'S s ; pr_com sigma goal com; 'SPC ; pr_with_bindings_list sigma goal bl >]
in (add_tactic s (tacfun,prfun);
    fun (com,bindl) -> Tactic(s,(COMMAND com)::(map (fun b -> BINDING b) bindl)))
;;

let overwriting_register_combindinglist_tactic s tac = 
let tacfun = fun ((COMMAND com)::al) -> tac (com,(map (fun (BINDING b) -> b) al))
and prfun = fun sigma goal (s,(COMMAND com)::bl) -> 
            [< 'S s ; pr_com sigma goal com; 'SPC ; pr_with_bindings_list sigma goal bl >]
in (overwriting_add_tactic s (tacfun,prfun);
    fun (com,bindl) -> Tactic(s,(COMMAND com)::(map (fun b -> BINDING b) bindl)))
;;

(* $Id: tacmach.ml,v 1.6 1996/02/12 15:39:15 cpaulin Exp $ *)
