(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                tactics.ml                                *)
(****************************************************************************)

#infix "o";;
#open "std";;
#open "initial";;
#open "names";;
#open "vectops";;
#open "generic";;
#open "ast";;
#open "term";;
#open "reduction";;
#open "himsg";;
#open "environ";;
#open "termenv";;
#open "machops";;
#open "mach";;
#open "trad";;
#open "more_util";;
#open "pp";;
#open "stdpp";;
#open "tacmach";;
#open "proof";;
#open "proof_trees";;
#open "clenv";;
#open "discharge";;
#open "declare";;

#infix "ORELSE";;
#infix "NEXT";;
#infix "THEN";;
#infix "THENF";;
#infix "THENL";;
#infix "THENS";;


(*********************************************************************)
(*                             Tactics                   *)
(*********************************************************************)

let rec decomp_app c =
    match collapse_appl c with
     DOPN(AppL,cl) -> (hd_vect cl,list_of_tl_vect cl)
 | DOP2(Cast,c,t) -> decomp_app c
 | c -> (c,[])
;;

let args_typ gls = decrec []
    where rec decrec l c = match pf_whd_betadeltaiota gls c with
             DOP2(Prod,A,DLAM(n,B)) -> decrec ((named_hd A n,A)::l) B
           | x           -> l;;

(* if lname_type typ is [xn,An;..;x1,A1] and lists is a list of terms,
gives [x1:A1]..[xn:An]c' such that c converts to ([x1:A1]..[xn:An]c' l) *)

let abstract_scheme c l lname_typ =
    it_list2 (fun t (locc,a) (na,ta) ->
    if occur_meta ta then error "cannot find a type for the generalisation"
    else if occur_meta a then DOP2(Lambda,ta,DLAM(na,t))
         else DOP2(Lambda,ta,DLAM(na,subst_term_occ locc a t))) c (rev l) lname_typ;;

let change_pf c gl = convert_concl c gl;;

let pattern_list l gl =
    let lname_typ = map (fun (_,c) ->
                             let t = pf_type_of gl c
                             in (named_hd t Anonymous,t))
                        (rev l)
    in change_pf (applist(abstract_scheme (pf_concl gl) l lname_typ,map snd l)) gl ;;

let pattern l gl =
    let loc_co = map (function (n,com) -> (n, pf_constr_of_com gl com)) l
    in pattern_list loc_co gl;;

let abstract_list_all gls typ c l =
  let P = abstract_scheme c (map (function a -> [],a) l) (args_typ gls typ)
  in try if pf_conv_x gls (pf_type_of gls P) typ then P else
     error "cannot find a generalisation of the goal with the appropriate type"
     with UserError(_)->
     error "Cannot find a well-typed generalisation of the goal";;

(****************************************)
(* General functions                    *)
(****************************************)

let (tactic_com : (constr -> tactic) -> command -> tactic) =
    fun tac t x -> tac (pf_constr_of_com x t) x;;

let (tactic_com_list : (constr list -> tactic) -> command list -> tactic) =
    fun tac tl x -> let translate = pf_constr_of_com x in 
                    tac (map translate tl) x;;

let (tactic_bind_list : ((BindOcc * constr) list -> tactic) 
                      -> (BindOcc * command) list -> tactic) =
    fun tac tl x -> let translate = pf_constr_of_com x in 
                    tac (map (fun (b,c)->(b,translate c)) tl) x;;

let (tactic_com_bind_list : (constr * (BindOcc * constr) list -> tactic)
                      -> (command * (BindOcc * command) list -> tactic)) =
    fun tac (c,tl) x -> let translate = pf_constr_of_com x in 
                 tac (translate c,map (fun (b,c')->(b,translate c')) tl) x;;

(**********************************************************************)
(* apply a tactic to the nth element of the signature             *)
(**********************************************************************)

let NTH m (tac:constr->tactic) gl =
  tac (try VAR(fst(nth_sign (pf_hyps gl) m)) with Failure _ -> error "No such assumption") gl;;

let LAST = NTH 1;;

(*******************************************)
(*         Introduction tactics        *)
(*******************************************)

(*******************************
(* Seems complicated with the gensym *)

let new_string_signl l0 name sign =
  let l = (ids_of_sign sign)@l0
  in next_ident_away name l
;;

let new_string_sign = new_string_signl [];;


let intro_use s1 name gl =
 let identifier = match name with
     Anonymous -> s1
   | Name s2   -> s2
 in introduction (new_string_sign identifier (pf_hyps gl)) gl;;

let Fresh id na tac gl =
 let identifier = match na with
     Anonymous -> id
   | Name s2   -> s2
 in tac (new_string_sign identifier (pf_hyps gl)) gl;;

(* Create a name if needed *)
let (intro:tactic) = fun gl ->
  match strip_outer_cast (pf_concl gl) with
   DOP2(Prod,c1,DLAM(name,c2)) -> (match pf_whd_betadeltaiota gl (pf_type_of gl c1) with
          DOP0(Sort(Prop _))    -> intro_use (id_of_string "H") name gl
        | DOP0(Sort(Type(_,_)))    -> intro_use (id_of_string "X") name gl
        | _         -> anomaly "Wrong sort") 
 | _ -> error "Intro needs a product";;

let intro_using id gls =
    introduction (new_string_sign id (pf_hyps gls)) gls
;;
**********************************************)


let next_global_ident_from  id avoid = next_rec id
where rec next_rec id =
  let id = next_ident_away_from  id avoid
  in  if not(is_global id) then id
      else  next_rec (lift_ident id) ;;

let next_global_ident_away id avoid =
  let id  = next_ident_away id avoid
  in if not(is_global id) then id
      else  next_global_ident_from (lift_ident id) avoid ;;

let intro_use s1 name gl =
 let identifier = match name with
     Anonymous -> s1
   | Name s2   -> s2 in
 let avoid = ids_of_sign (pf_hyps gl)
 in introduction (next_global_ident_away  identifier avoid) gl;;

let Fresh id na tac gl =
 let identifier = match na with
     Anonymous -> id
   | Name s2   -> s2  in
 let avoid = ids_of_sign (pf_hyps gl) 
 in tac (next_global_ident_away identifier avoid) gl;;

(* Create a name if needed *)
let (intro:tactic) = fun gl ->
  match strip_outer_cast (pf_concl gl) with
   DOP2(Prod,c1,DLAM(name,c2)) -> (match pf_whd_betadeltaiota gl (pf_type_of gl c1) with
          DOP0(Sort(Prop _))    -> intro_use (id_of_string "H") name gl
        | DOP0(Sort(Type(_,_)))    -> intro_use (id_of_string "X") name gl
        | _         -> anomaly "Wrong sort") 
 | _ -> error "Intro needs a product";;

let intro_using id gls =
    let avoid = ids_of_sign (pf_hyps gls)
    in introduction (next_global_ident_away id avoid )gls
;;


let intros = REPEAT intro;;

(**********************************)
(* Intros with explicit names     *)
(**********************************)
(***
let intro_with str gl = introduction str gl;;
***)
let intro_with str gl = 
  if not ((is_global str) or (mem str (ids_of_sign (pf_hyps gl)))) 
   then introduction str gl
   else error ("name "^(string_of_id str)^" is already bound") ;;

let rec intros_with  =
    function []      -> IDTAC
       | str::l -> (intro_with str) THEN intros_with l;;



(********
let intros_until s g = 
  let depth = lookup 1 (rename_bound_var (ids_of_sign (pf_hyps g)) (pf_concl g))
    where rec lookup n = function
           DOP2(Prod,_,DLAM(name,c')) -> (match name with
                  Name(s')  -> if s'=s then n
                               else lookup (n+1) c'
                | Anonymous -> lookup (n+1) c')
         | DOP2(Cast,c,_) -> lookup n c
         | _ -> error "No such hypothesis in current goal"
  in DO depth intro g;;

**************)

let intros_until s g = 
  let depth = lookup 1 (rename_bound_var (global_vars (pf_concl g)) (pf_concl g))
    where rec lookup n = function
           DOP2(Prod,_,DLAM(name,c')) -> (match name with
                  Name(s')  -> if s'=s then n
                               else lookup (n+1) c'
                | Anonymous -> lookup (n+1) c')
         | DOP2(Cast,c,_) -> lookup n c
         | _ -> error "No such hypothesis in current goal"
  in DO depth intro g;;

let intros_do n g = 
  let depth = lookup 1 1 (pf_concl g)
    where rec lookup all nodep = function
           DOP2(Prod,_,DLAM(name,c'))  -> (match name with
                  Name(s')  -> if dependent (Rel 1) c' then lookup (all+1) nodep c'
                               else if nodep=n then all
                               else lookup (all+1) (nodep+1) c'
                | Anonymous -> if nodep=n then all
                               else lookup (all+1) (nodep+1) c')
         | DOP2(Cast,c,_) -> lookup all nodep c
         | _ -> error "No such hypothesis in current goal"
  in DO depth intro g;;

(********************************************************)
(*  Resolution tactics                              *)
(********************************************************)

(* Refinement tactic: unification against the head of the head normal form
   of the type of a term *)

let apply_type hdcty argl gl =
    Refine (DOPN(AppL,vect_of_list
                ((DOP2(Cast,DOP0(Meta(newMETA())),hdcty))::argl))) gl
;;

let apply_term hdc argl gl =
    Refine (DOPN(AppL,vect_of_list(hdc::argl))) gl
;;

let exact c gl = Refine c gl;;

(*********************************************************)
(*  Resolve with missing arguments                       *)
(*********************************************************)

let collect_com lbind = 
  map_succeed (function (Com,c)->c | _ -> failwith "Com") lbind;;

let make_clenv_binding_apply wc (c,t) lbind = 
     let largs = collect_com lbind 
     in let lcomargs = length largs 
     in if lcomargs = length lbind then 
               let clause = mk_clenv_from wc (c,t) in
               clenv_constrain_missing_args largs clause
     else if lcomargs = 0 
          then let clause = mk_clenv_rename_from wc (c,t) in
               clenv_match_args lbind clause
          else errorlabstrm "make_clenv_bindings"
          [<'S "Cannot mix bindings and free associations">];;

let resolve_with_bindings_tac  (c,lbind) gl = 
    let (wc,KONT) = StartWalk gl in
    let t = w_hnf_constr wc (w_type_of wc c) in 
    let clause = make_clenv_binding_apply wc (c,t) lbind
    in res_pf KONT clause gl;;

let resolve_with_bindings = tactic_com_bind_list resolve_with_bindings_tac;;

let resolve c = resolve_with_bindings (c,[]);;
let resolve_with (c::l) 
   = resolve_with_bindings (c,map (fun com ->(Com,com)) l);;

(* no reduction on the type *)
let resolve_term c gl = 
    let (wc,KONT) = StartWalk gl in
    let clause = mk_clenv_type_of wc c in res_pf KONT clause gl;;

let refinew_scheme KONT clause gl = res_pf KONT clause gl;;

(* Cut tactics *)

let cut_tac c gl =
  match hnf_type_of gl c with
    (DOP0(Sort _)) ->
    apply_type (DOP2(Prod,c,DLAM(Anonymous,(pf_concl gl)))) 
               [DOP0(Meta (newMETA()))] gl
  | _  -> error "Not a proposition or a type";;

let cut = tactic_com cut_tac;;

let generalize_tac lconstr gl = 
    let newcl = it_list 
      (fun cl' ci -> let t = pf_type_of gl ci in
            match ci with (VAR _) -> produit (id_of_var ci) t cl'
                        | _       -> DOP2(Prod,t,DLAM(Anonymous,cl')))
      (pf_concl gl) (rev lconstr)
    in apply_type newcl lconstr gl;;

let generalize = tactic_com_list generalize_tac;;

(********************************************************************)
(* A more atomic imp_elim tactic (Yves Bertot)                      *)
(********************************************************************)

let imp_elim c t gl =
let goal_constr = pf_concl gl
in match t with
    DOP2(Prod,c1,DLAM(_,c2)) ->
    if not dependent (Rel 1) c2 then
    ((apply_type (DOP2(Prod,c2,DLAM(Anonymous,goal_constr))) [DOP0(Meta(newMETA()))])
    THENS [IDTAC;apply_term c [DOP0(Meta(newMETA()))]])
    gl
    else error "Imp_elim needs a non-dependent product"

  | _ -> error "Imp_elim needs a non-dependent product";;

let imp_elim_tac =
   tactic_com (fun c gl -> imp_elim c (pf_hnf_constr gl (pf_type_of gl c)) gl);;

(* TODO: The Name(_) is bogus and needs to be killed off *)
let (prod_elim: constr -> constr -> constr -> tactic) = fun
    c t v gl ->
(* no check is done to know whether the value actually has the right type *)
let goal_constr = pf_concl gl
in match t with
      DOP2(Prod,c1,DLAM(Name(_),c2)) ->
      if pf_conv_x gl c1 (pf_type_of gl v) then
          apply_type (DOP2(Prod,subst1 v c2,DLAM(Anonymous,goal_constr)))
          [applistc c [v]] gl

       else error "The type of the argument does not match"
      | _ -> error "prod_elim needs a dependent product";;

let (prod_elim_tac: command list -> tactic) = fun com_list x ->
    match com_list with
      [prod_command;val_command] ->
      let prod = (pf_constr_of_com x prod_command)
      in prod_elim prod (pf_hnf_constr x (pf_type_of x prod))
          (pf_constr_of_com x val_command) x

    | _ -> error "prod_elim_tac needs two arguments";;

(********************************************************************)
(*               Exact tactics                                      *)
(********************************************************************)

let (assumption:tactic) = fun gl ->
let concl =  pf_concl gl in 
let rec arec = function
   ([],[])          -> error "No such assumption"
 | (s::sl,a::al) -> if pf_conv_x gl concl a then 
                    exact (VAR(s)) gl else arec (sl,al)
in arec (pf_hyps gl);;


(* Give exact proof term *)
let give_exact c gl = let concl = (pf_concl gl)
in 
if pf_conv_x gl (pf_concl gl) (pf_type_of gl c) then exact c gl 
        else error "Not an exact proof";;

let give_exact_com = tactic_com (fun c gl -> give_exact c gl);;

(**************************************************************)
(*             Adding new hypotheses                          *)
(**************************************************************)

let new_hyp_tac c gl =
    ((cut_tac (pf_type_of gl c)) THENL (give_exact c)) gl;;

let new_hyp = tactic_com new_hyp_tac;;

let new_hyp_with_tac blist m (KONT,clause) g =
    let blist = map (fun (b,c) -> (b,pf_constr_of_com g c)) blist in
    let clause' = clenv_match_args blist clause in
    let (thd,tstack) = whd_castapp_stack (clenv_instance_template clause') [] in
    let nargs = length tstack in
    let cut_pf = applist(thd,if m < nargs then firstn m tstack else tstack)
    in (KONT clause.hook THEN
        (cut_tac (pf_type_of g cut_pf)) THENL
        (give_exact cut_pf)) g;;

let new_hyp_with com blist g =
    let c = (pf_constr_of_com g com)
    in let (wc,KONT) = StartWalk g
    in let clf = mk_clenv_printable_type_of wc c
    in let nargs = length (snd (whd_castapp_stack (clenv_template clf).rebus []))
    in new_hyp_with_tac blist nargs (KONT,clf) g;;

let new_hyp_with_num m com blist g =
    let c = (pf_constr_of_com g com)
    in let (wc,KONT) = StartWalk g
    in let clf = mk_clenv_printable_type_of wc c
    in new_hyp_with_tac blist m (KONT,clf) g;;


(**************************************************************)
(*             Conversion tactics                             *)
(**************************************************************)

let change  = tactic_com change_pf;;

let red gl = convert_concl (pf_red_product gl (pf_concl gl)) gl;;

let hnf_tac gl = convert_concl (pf_hnf_constr gl (pf_concl gl)) gl;;

let simpl_tac gl = convert_concl (pf_nf gl (pf_concl gl)) gl;;

let compute_tac gl = convert_concl (pf_compute gl (pf_concl gl)) gl;;

let whd_expand_const sigma name = whrec
    where rec whrec = function
  (DOPN(Const _,_) as x) ->
  let name' = (path_of_const x)
  in if name = name' then
      if not(evaluable_const sigma x) then
          error "Constant is opaque"
      else
          (const_value sigma x)
     else x

 | (DOPN(Abst sp,cl) as x) ->
   let name' = (path_of_abst x)
   in if name = name' then
      if not(evaluable_abst x) then
          error "Abstraction is opaque"
      else
          whrec (abst_value x)
     else x
 | x -> x
;;

let expand_const sigma name c = strong (whd_expand_const sigma name) c;;

let pf_expand_const gls = expand_const (Project gls);;

let unfold_nth loccname gl = 
let cl' = strong whd_betaiota
    (it_list 
     (fun c (lo,name) ->
          match lo with
          [] -> pf_expand_const gl name c
        | _  -> pf_unfoldn gl lo name c) (pf_concl gl) loccname)
in convert_concl cl' gl
;;

let unfold_nth_tac loccname gl =
    unfold_nth (map (fun (l,na) ->
                         try let nasp = nametab__sp_of_id CCI na
                             in (l,nasp)
                         with Not_found ->
                             error "unfold_nth") loccname) gl
;;

(**********************************************************************)
(*  Simplication of local context                                     *)
(**********************************************************************)

(* function clear_hyp: command list -> tactic *)
(* This tactic enables the user to remove hypotheses from the signature *)
(* so that these hypotheses do not appear in the assumption list.  Some *)
(* care is taken to prevent him from removing variables that are        *)
(* subsequently used in other hypotheses or in the conclusion of the    *)
(* goal.  The same result could be achieved in a safer way if the       *)
(* signature used to perform type checking and the actual list of       *)
(* hypotheses displayed on the screen were distinct data structures.    *)

let clear_hyp ids gl = thin ids gl;;

(**********************************************************************)
(*  Conversion tactics in an hypothesis                               *)
(**********************************************************************)

let apply_in_hyp reductor id gl =
let ty = try (snd(lookup_sign id (pf_hyps gl)))
         with Not_found -> error ("No such hypothesis : " ^ (string_of_id id))
in convert_hyp id (under_casts (reductor gl) ty) gl
;;

let red_hyp = apply_in_hyp pf_red_product;;

let hnf_tac_hyp  = apply_in_hyp pf_hnf_constr;;

let change_hyp_constr c gl =
    apply_in_hyp (fun gls ch -> if pf_conv_x gls c ch then c else
                      error "Not convertible")
    gl;;

let Change = change_pf;;
let ChangeHyp c hid gl = change_hyp_constr c hid gl;;

let change_hyp c nh gl =
  change_hyp_constr (pf_constr_of_com gl c) nh gl;;

let simpl_hyp = apply_in_hyp pf_nf;;

let unfold_nth_hyp loccname =
  apply_in_hyp (fun gls ch -> strong whd_betaiota
   (it_list 
    (fun c (lo,name) ->
         let sp = nametab__sp_of_id CCI name
         in match lo with
        [] -> pf_expand_const gls sp c
      | _  -> pf_unfoldn gls lo sp c) 
   ch loccname));;

(**********************************************************************)
(*  Elimination tactics                           *)
(**********************************************************************)

(* elimination on the inductive definition t, generate a new subgoal t
(the last one), enough for non-dependent elimination *)

let elim_scheme_type elim t gl =
    let (wc,KONT) = StartWalk gl in
    let clause = mk_clenv_type_of wc elim
    in match rev_subterm (clenv_template clause).rebus [0] with
        DOP0(Meta mv) ->
        let clause' = clenv_unify (clenv_instance_type clause mv) t clause
        in elim_res_pf_cast KONT clause' gl
      | _ -> anomaly "elim_scheme_type"
;;

(* put t as t'=(x1:A1)..(xn:An)B with B an inductive definition of name name
 return name, B and t' *)

let prod_it = it_list (fun c (n,t)  ->  DOP2(Prod,t,DLAM(n,c)));;

let reduce_to_ind gl t = elimrec t []
    where rec elimrec t l = match whd_castapp_stack t [] with
         (DOPN(MutInd (sp,_),_) as mind,_) -> ((mind_path mind),t,prod_it t l)
       | (DOPN(Const _,_),_) -> 
                 (try let t' = strong whd_betaiota (pf_one_step_reduce gl t)
                      in elimrec t' l
                  with UserError _ -> errorlabstrm "tactics__reduce_to_ind"
                                         [< 'S"Not an inductive product" >])
       | (DOPN(MutCase,_),_) ->
             (try let t' = strong whd_betaiota (pf_one_step_reduce gl t)
                      in elimrec t' l
                  with UserError _ -> errorlabstrm "tactics__reduce_to_ind"
                                         [< 'S"Not an inductive product" >])
       | (DOP2(Cast,c,_),[]) -> elimrec c l
       | (DOP2(Prod,ty,DLAM(n,t')),[]) -> elimrec t' ((n,ty)::l)
       | _ -> error "Not an inductive product";;

let reduce_to_mind gl t = elimrec t []
    where rec elimrec t l = match whd_castapp_stack t [] with
               (DOPN(MutInd sp,_) as mind,_) -> (mind,prod_it t l)
             | (DOPN(Const _,_),_) -> 
                 (try let t' = strong whd_betaiota (pf_one_step_reduce gl t)
                      in elimrec t' l
                  with UserError _ -> errorlabstrm "tactics__reduce_to_mind"
                                         [< 'S"Not an inductive product" >])
             | (DOPN(MutCase,_),_) ->
             (try let t' = strong whd_betaiota (pf_one_step_reduce gl t)
                      in elimrec t' l
                  with UserError _ -> errorlabstrm "tactics__reduce_to_mind"
                                           [< 'S"Not an inductive product" >])              | (DOP2(Cast,c,_),[]) -> elimrec c l
             | (DOP2(Prod,ty,DLAM(n,t')),[]) -> elimrec t' ((n,ty)::l)
             | _ -> error "Not an inductive product";;

let constructor_tac nconstropt i lbind gl 
    = let cl = pf_concl gl in 
    let (DOPN(MutInd(sp),args) as mind,redcl) = reduce_to_mind gl cl in 
        let nconstr = mis_nconstr (mind_specif_of_mind mind) and
        sigma = Project gl in
         (match nconstropt with SOME expnconstr -> 
                if expnconstr <> nconstr then
                error "Not the expected number of constructors"
                             | _ -> ());
        if i > nconstr then error "Not enough Constructors" 
        else let c = DOPN(MutConstruct(sp,i),args) 
        in let resolve_tac = resolve_with_bindings_tac (c,lbind) in
         (change_pf redcl THEN intros THEN resolve_tac) gl
;;

let left_with_bindings = tactic_bind_list (constructor_tac (SOME 2) 1);;
let right_with_bindings = 
        tactic_bind_list (constructor_tac (SOME 2) 2);;
let split_with_bindings = tactic_bind_list (constructor_tac (SOME 1) 1);;
let constructor_with_bindings i =  tactic_bind_list (constructor_tac NONE i);;


(************************************************)
(*      Eliminations tactics                    *)
(************************************************)

let sort_of_goal gl = match hnf_type_of gl (pf_concl gl) with 
     DOP0(Sort s) -> s
  |  _            -> anomaly "goal should be a type";;

(* Find the right elimination suffix corresponding to the sort of the goal *)
 let suff gl cl = match hnf_type_of gl cl with
                  DOP0(Sort(Type(Null,_))) -> "_rect"
                | DOP0(Sort(Type(Pos,_))) -> "_recs"
                | DOP0(Sort(Prop(Null)))   -> "_ind"
                | DOP0(Sort(Prop(Pos)))    -> "_rec"
                | _            -> anomaly "goal should be a type";;

(* c should be of type A1->.. An->B with B an inductive definition *)

let lookup_eliminator sign path suff =
let path_elim = append_to_path path suff
in try global_reference (GLOB sign) path_elim (basename path_elim)
   with UserError _ ->
        global (GLOB sign) (basename path_elim)
;;

let elimination_clause_scheme KONT wc elimclause indclause gl = 
  let indmv = (match rev_subterm (clenv_template elimclause).rebus [0] with
               DOP0(Meta mv) -> mv
             | _ -> errorlabstrm "elimination_clause"
       [< 'S "The type of elimination clause is not well-formed" >])
  in 
  let elimclause' = clenv_fchain indmv elimclause indclause
  in elim_res_pf KONT elimclause' gl;;

let elimination_scheme elim largs c t gl = 
 let (wc,KONT) = StartWalk gl in
  let indclause = mk_clenv_from wc (c,t) in
  let indclause' = clenv_constrain_dep_args largs indclause in
  let elimclause = mk_clenv_from () (elim,w_type_of wc elim) in
  elimination_clause_scheme KONT wc elimclause indclause' gl;;

let elimination largs c gl =
  let (path_name,_,t) = reduce_to_ind gl (pf_type_of gl c) in
  let elim = lookup_eliminator (pf_hyps gl) path_name (suff gl (pf_concl gl))
  in  elimination_scheme elim largs c t gl
;;

let make_clenv_binding wc (c,t) lbind = 
     let largs = collect_com lbind 
     in let lcomargs = length largs 
     in if lcomargs = length lbind then 
               let clause = mk_clenv_from wc (c,t) in
               clenv_constrain_dep_args largs clause
     else if lcomargs = 0 
          then let clause = mk_clenv_rename_from wc (c,t) in
               clenv_match_args lbind clause
          else errorlabstrm "make_clenv_bindings"
          [<'S "Cannot mix bindings and free associations">];;

let elimination_general_scheme  (c,lbindc) (elim,lbindelim) gl = 
 let (wc,KONT) = StartWalk gl in
 let (_,_,t) = reduce_to_ind gl (pf_type_of gl c) in
 let indclause = make_clenv_binding wc (c,t) lbindc  in
 let elimt = w_type_of wc elim in
 let elimclause =  make_clenv_binding wc (elim,elimt) lbindelim in
 elimination_clause_scheme KONT wc elimclause indclause gl;;


let elim_with_using = fun (c,lbindc) (elim,lbinde) gl -> 
   let translate = pf_constr_of_com gl
 in let translate_bind = map (fun (b,com)->(b,translate com))
 in elimination_general_scheme (translate c,translate_bind lbindc)
 (translate elim,translate_bind lbinde) gl;;

let elimination_with_bindings lbindc c gl = 
  let (path_name,_,t) = reduce_to_ind gl (pf_type_of gl c) in
  let elim = lookup_eliminator (pf_hyps gl) path_name (suff gl (pf_concl gl))
  in  elimination_general_scheme (c,lbindc) (elim,[]) gl
;;

let elim_with_bindings = fun (c,lbindc) gl -> 
   let translate = pf_constr_of_com gl
 in let translate_bind = map (fun (b,com)->(b,translate com)) 
 in elimination_with_bindings (translate_bind lbindc)  (translate c) gl;;
  

(* elim is the elimination combinator which should end by a clause
   (x:I)(P .. ) with P a bound variable.
    c is of type t which is a product ending with a type matching I
    largs are the expected for c arguments *)


let elimination1_scheme elim largs c t gl = 
 let (wc,KONT) = StartWalk gl in
  let indclause = mk_clenv_from wc (c,t) in
  let indclause' = clenv_constrain_dep_args largs indclause in
  let elimclause = mk_clenv_from () (elim,w_type_of wc elim) in
  let indmv = (match rev_subterm (clenv_template elimclause).rebus [0] with
               DOP0(Meta mv) -> mv
             | _ -> anomaly "elimination") in
  let elimclause' = clenv_fchain indmv elimclause indclause'
  in elim_res_pf1 KONT elimclause' gl
;;

let elimination1 largs c gl =
  let (path_name,_,t) = reduce_to_ind gl (pf_type_of gl c) in
  let elim = lookup_eliminator (pf_hyps gl) path_name (suff gl (pf_concl gl))
  in  elimination1_scheme elim largs c t gl
;;

let elimination_type t gl =
  let (path_name,tind,t) = reduce_to_ind gl t in
  let elim = lookup_eliminator (pf_hyps gl) path_name (suff gl (pf_concl gl))
  in match t with DOP2(Prod,_,_) -> error "Not an inductive definition"
                | _       -> elim_scheme_type elim tind gl
;;

let elim = tactic_com (fun c gl -> elimination [] c gl);;
let elim1 = tactic_com (fun c gl -> elimination1 [] c gl);;

let elim_with = fun
  (c::lc) gl ->
    let translate = pf_constr_of_com gl
     in elimination (map translate lc) (translate c) gl
| _ _ -> anomaly "elim_with";;

let elim_last  = LAST (fun c gl -> elimination [] c gl);;

let elim_type = tactic_com elimination_type;;

(* Induction tactics *)
let induct s = (intros_until s) THEN elim_last;;

let induct_nodep n = (intros_do n) THEN elim_last;;

let case_analysis lbindc c gl =
  let (mind,_) = reduce_to_mind gl (pf_type_of gl c) in
  let sigma = Project gl in 
  let sort = sort_of_goal gl in
  let elim = indrec__make_case_gen sigma mind sort in 
  elimination_general_scheme (c,lbindc) (elim,[]) gl;;

let case_with_bindings = fun (c,lbindc) gl -> 
   let translate = pf_constr_of_com gl
 in let translate_bind = map (fun (b,com)->(b,translate com)) 
 in case_analysis (translate_bind lbindc)  (translate c) gl;;
 

let case_analysis_type t gl =
  let (mind,t) = reduce_to_mind gl t in
  match t with DOP2(Prod,_,_) -> error "Not an inductive definition"
                | _       -> 
  let sigma = Project gl in 
  let sort = sort_of_goal gl in
  let elim = indrec__make_case_gen sigma mind sort in 
  elim_scheme_type elim t gl;;

let case = tactic_com (fun c gl -> case_analysis [] c gl);;

(***
let case_with = fun
  (c::lc) gl ->
    let translate = pf_constr_of_com gl
     in case_analysis (map translate lc) (translate c) gl
| _ _ -> anomaly "elim_with";;
****)

let case_type = tactic_com case_analysis_type;;

let case_last  = LAST (fun c gl -> case_analysis [] c gl);;

(* Induction tactics *)
let destruct s = (intros_until s) THEN case_last;;

let destruct_nodep n = (intros_do n) THEN case_last;;


(* Tactics "Macro" for treatment of equality : Need the definition of *)
(* Leibniz equality on Set and on Type as an inductive type .        *)
(* exact names of the constants are in vernac.ml *)

(* eq=RefC("eq") eqt=RefC("eqt") *)
(* sym_equal=RefC("sym_equal") sym_eqt=RefC("sym_eqT") *)

let replace_tac eq eqt c2 c1 sym_equal sym_eqt gl =
  let cc1 = pf_constr_of_com gl c1 in
  let cc2 = pf_constr_of_com gl c2 in
  let t1 = pf_type_of gl cc1 
  and t2 = pf_type_of gl cc2 in
  if pf_conv_x gl t1 t2 then
  let (e,sym) = match hnf_type_of gl t1 with 
     DOP0(Sort(Prop(Pos))) -> (eq,sym_equal)
   | DOP0(Sort(Type(_,_))) -> (eqt,sym_eqt)
   | _         -> error "replace_tac"
 in (elimination_type (applist (constr_of_com mt_evc (GLOB nil_sign) e, [t1;cc1;cc2]))
  THENL (assumption 
     ORELSE TRY (resolve sym THEN assumption))) gl
  else error "terms does not have convertible types";;

(* This applies Elim on a term of conclusion an equality but *)
(* with inverting the sense of the equality *)

(******
let equality_tac eq eqt c sym_equal sym_eqt gl =
  let cc = (pf_constr_of_com gl c) in
    let rec decomp_prodrec k l typ = match (collapse_appl typ) with
      (DOP2(Prod,t,DLAM(n,c)))    -> DOP2(Lambda,t,DLAM(n,decomp_prodrec (k+1) (Rel(k)::l) c))
   | DOP2(Cast,c,_) -> decomp_prodrec k l c
   |  DOPN(AppL,tl) ->
      if vect_length tl = 4 then
          let [e;t;cc2;cc1] = list_of_vect tl in
         let sym = if (e=(constr_of_com mt_evc (GLOB nil_sign) eq)) then sym_equal
              else if (e=(constr_of_com mt_evc (GLOB nil_sign) eqt)) then sym_eqt
              else error "No equality here"
          in applist(constr_of_com mt_evc (GLOB nil_sign) sym,[t;cc2;cc1;applist(cc,l)])
      else error "No equality here"
   |  _ -> error "No equality here"
   in let ccc = decomp_prodrec 1 [] (pf_type_of gl cc)
    in elimination [] ccc gl;;

*****)


let equality_tac (eq,eq_ind,eq_rec) (eqt,eqt_ind,eqt_rec) c gl =
  let cc = (pf_constr_of_com gl c) in
    let rec find_elim_r typ = match (collapse_appl typ) with
      (DOP2(Prod,t,DLAM(n,c)))    -> find_elim_r c
   | DOP2(Cast,c,_) -> find_elim_r c
   |  DOPN(AppL,tl) ->
      if vect_length tl = 4 then
          let [e;t;cc2;cc1] = list_of_vect tl in
          let s = sort_of_goal gl in
          let elim_th = 
           if (e=(constr_of_com mt_evc (GLOB nil_sign) eq)) then 
                if s=Prop(Null) then eq_ind else 
                if s=Prop(Pos) then eq_rec else 
                error "Equality elimination cannot be done with a goal of sort Type" 
           else if (e=(constr_of_com mt_evc (GLOB nil_sign) eqt)) then 
                if s=Prop(Null) then eqt_ind else 
                if s=Prop(Pos) then eqt_rec else 
                error "Equality elimination cannot be done with a goal of sort Type" 
           else error "No equality here" in
        constr_of_com mt_evc (GLOB nil_sign) elim_th
      else error "No equality here"
   |  _ -> error "No equality here" in
  let ccc = find_elim_r (pf_type_of gl cc) in
  elimination_general_scheme (cc,[]) (ccc,[]) gl;;

let absurd_tac c gls =
    (elim_type (ASTnvar (id_of_string "False"))
    THEN cut_tac c
    THENS [cut_tac (applist(pf_global gls (id_of_string "not"),[c]))
           THENS [intros THEN
                  (fun gl ->
                       let (ida,_) = pf_h gl 1
                       and (idna,_) = pf_h gl 2
                       in exact (applist(VAR idna,[VAR ida])) gl);
                   IDTAC];
           IDTAC]) gls
;;
let absurd = tactic_com absurd_tac;;

(* Auxiliary functions *)
(*exception BOUND;;*)

let rec string_head_bound = function 
     DOP2(Prod,_,DLAM(_,c2))            -> string_head_bound c2
  |  DOPN(AppL,cl)               -> string_head_bound (hd_vect cl)
  |  DOP2(Cast,c,_)     -> string_head_bound c
  |  DOPN(Const _,_) as x -> string_of_id (basename (path_of_const x))
  |  DOPN(MutInd _,_) as x -> let mispec = mind_specif_of_mind x in 
     string_of_id (mis_typename mispec)
  |  DOPN(MutConstruct ((sp,tyi),i),_) ->
     let (_,mib) = mind_of_path sp
     in string_of_id
        (mib.constrtypes__MINDPACKETS.(tyi).constrtypes__MINDCONSNAMES.(i-1))

  |  VAR(_) as x   -> string_of_id (id_of_var x)
  |  _                       -> raise BOUND;;

let string_head c = try string_head_bound c
             with BOUND -> error "Bound head variable";;

let rec head_constr_bound t =
let t = strip_outer_cast(collapse_appl t)
in match t with
     DOP2(Prod,_,DLAM(_,c2)) -> head_constr_bound c2
  |  DOPN(AppL,cl) -> head_constr_bound (hd_vect cl)
  |  DOPN(Const _,_) as x -> x
  |  DOPN(MutInd _,_) as x -> x
  |  DOPN(MutConstruct _,_) as x -> x
  |  VAR(_) as x   -> x
  |  _                       -> raise BOUND;;

let head_constr c = try head_constr_bound c
             with BOUND -> error "Bound head variable";;


(* $Id: tactics.ml,v 1.9 1996/02/12 15:39:19 cpaulin Exp $ *)
