(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                 trad.ml                                  *)
(****************************************************************************)

#infix "o";;
#open "std";;
#open "names";;
#open "vectops";;
#open "impuniv";;
#open "generic";;
#open "term";;
#open "reduction";;
#open "termenv";;
#open "typing";;
#open "himsg";;
#open "pp";;
#open "stdpp";;
#open "more_util";;
#open "proof_trees";;
#open "proof_constraints";;
#open "proof";;
#open "machops";;

let isVAR = function VAR _ -> true | _ -> false;;
let isRel = function Rel _ -> true | _ -> false;;

(* gives a new section path each time it is called *)
let gensym_exsp =
let existential_ctr = ref 0
in
fun () ->
let n = (incr existential_ctr; !existential_ctr)
in make_path(PA__intern["isevar"]) (make_ident(SA__intern"?") n) CCI;;

(* domain of a product *)
let prod_dom_tycon sigma = function
    NONE -> NONE
  | SOME c -> (match hnf_constr sigma c with
               DOP2(Prod,c1,_) -> SOME c1
             | _ -> NONE)
;;

(* range of a product *)
let prod_rng_tycon sigma = function
    NONE -> NONE
  | SOME c -> (match hnf_constr sigma c with
               DOP2(Prod,_,DLAM(_,c2)) -> SOME c2
             | _ -> NONE)
;;

(* domain of a lamdba *)
let lambda_dom_tycon = prod_dom_tycon
;;

let app_rng_tycon sigma var = function
    NONE -> NONE
  | SOME c -> (match hnf_constr sigma c with
               DOP2(Prod,_,B) -> SOME (SAPP B var)
             | _ -> NONE)
;;

(* isevars : readable_constraints ref *)
let isevars = ref mt_evc;;

let set_ise evc = isevars := evc;;
let get_ise () = !isevars;;

(* declare a new existential of section path sp, type typ, with the initial
signature sign *)
let ise_declare sp sign typ =
    isevars := evc__add !isevars (sp,{evgoal=mkGOAL mt_lc mt_ctxt sign typ;
                                      body = EVAR_EMPTY})
;;

(* declare a new existential in an env env with type typ *)
let new_isevar env typ =
let (ENVIRON(sign,dbenv)) = env in
let t = it_list (fun B (na,D)-> DOP2(Lambda,D,DLAM(na,B))) typ dbenv in
let rec aux sign = fun
    0 T -> (sign,T)
  | n (DOP2(Lambda,D,(DLAM(na,B)))) ->
    let id = next_name_away na (ids_of_sign sign)
    in aux (add_sign (id,D) sign) (n-1) (SAPP (DLAM(na,B)) (VAR id))
  | _ _ -> anomalylabstrm "trad__new_isevar" [< 'S"Bug in" ; 'SPC ;
                                              'S"trad__new_isevar" ;'SPC;
                                              'S"code." >]
in
let (sign',typ') = aux sign (length dbenv) t in
let newsp = gensym_exsp()
in (ise_declare newsp sign' typ';
    DOPN(Const newsp,
         vect_of_list ((rev(rel_list 0 (length dbenv)))
                       @(map (fun id -> VAR id) (ids_of_sign sign)))))
;;

(* define the existential of section path sp as the constr body *)
let ise_define sp body =
let evd = evc__map !isevars sp
in match evd.body with
    EVAR_EMPTY ->
    isevars := evc__remap !isevars sp {evgoal = evd.evgoal;
                                       body = EVAR_DEFINED body}
  | _ -> anomaly "cannot define an isevar twice"
;;

(* say if the section path sp corresponds to an existential *)
let ise_in_dom sp = evc__in_dom !isevars sp;;

(* say if the constr DOPN (Const sp,_) corresponds to a defined
existential or to a non-existential *)

let ise_defined (DOPN(Const sp,_) as K) =
    not(ise_in_dom sp) or
    defined_const (get_ise()) K;;

let unsolved_ise (DOPN(Const sp,_) as K) = (ise_in_dom sp) &
    not (defined_const (get_ise()) K);;

(* map the given section path to the evar_declaration *)
let ise_map sp = evc__map !isevars sp;;

(* try to apply predicate P to a.
If (P a) is false or raises an exception, the previous existential context
is restored *)
let ise_pred P a =
let e = get_ise()
in try if P a then true
       else (set_ise e;false)
   with reraise -> (set_ise e;raise reraise)
;;

(* Try P (), if true OK
             if false then ise_pred Q ()
             if raises an exception then restore the previous existential
             context*)
let ise_or P Q =
let u = get_ise()
in try if P() then true
       else
           (set_ise u;
            if Q() then true
            else (set_ise u;false))
   with reraise -> (set_ise u;raise reraise)
;;

type conv_pb = CONV | CONV_LEQ | CONV_X;;

let conversion_problems = ref ([] : (conv_pb * constr * constr) list);;
let solved_problems = ref ([] : (section_path * evar_declaration) list);;

let add_conv_pb pb = (conversion_problems := pb::!conversion_problems);;

let rec whd_ise = function
    (DOPN(Const sp,_) as K) ->
    if ise_defined K & ise_in_dom sp then
        whd_ise (const_value (get_ise()) K)
    else if ise_in_dom sp then
        failwith "whd_ise"
    else K
  | DOP2(Cast,c,_) -> whd_ise c
  | DOP0(Sort(Type(c,_))) -> DOP0(Sort(Type(c,Dummy_univ)))
  | c -> c
;;

(* This code takes a unification problem, and tries to solve it.
 * If it solves it, then it removes all the conversion problems,
 * and re-runs conversion on each one, in the hopes that the new
 * solution will aid in solving them.
 *
 * The kinds of problems it knows how to solve are those in which
 * the usable arguments of an existential var are all themselves
 * universal variables.
 * The solution to this problem is to do renaming for the Var's,
 * to make them match up with the Var's which are found in the
 * hyps of the existential, to do a "pop" for each Rel which is
 * not an argument of the existential, and a subst1 for each which
 * is, again, with the corresponding variable.
 *
 * Thus, we take the arguments of the existential which we are about
 * to assign, and zip them with the identifiers in the hypotheses.
 * Then, we process all the Var's in the arguments, and sort the
 * Rel's into ascending order.  Then, we just march up, doing
 * subst1's and pop's.
 *
 * NOTE: We can do this more efficiently for the relative arguments,
 * by building a long substituend by hand, but this is a pain in the
 * ass.
 *)



let sort_increasing_snd = sort (fun ((_,Rel m),(_,Rel n)) -> m < n);;


let head_is_exist = hrec where rec hrec = function
    DOPN(Const sp,_) as K -> ise_in_dom sp & not(ise_defined K)
  | DOPN(AppL,cl) -> hrec (hd_vect cl)
  | DOP2(Cast,c,_) -> hrec c
  | _ -> false
;;


let rec is_eliminator = function
    DOPN(AppL,_) -> true
  | DOPN(MutCase,_) -> true
  | DOP2(Cast,c,_) -> is_eliminator c
  | _ -> false
;;

let head_is_embedded_exist c =
    (head_is_exist c) & (is_eliminator c)
;;

let headconstant = hrec where rec hrec = function
    DOPN(Const sp,_) -> sp
  | DOPN(MutCase,cl) -> hrec (hd_vect cl)
  | DOPN(AppL,cl) -> hrec (hd_vect cl)
  | DOP2(Cast,c,_)-> hrec c
  | _ -> failwith "headconstant"
;;

let status_changed sp (pbty,t1,t2) =
    try (sp = headconstant t1) or (sp = headconstant t2)
    with Failure _ ->
    try (sp = headconstant t2)
    with Failure _ -> false
;;

let conv_minductive convfun mind1 mind2 =
    let mispec1 = mind_specif_of_mind mind1 and 
        mispec2 = mind_specif_of_mind mind2 in
    let ar1 = mis_arity mispec1 and
        ar2 = mis_arity mispec2 and
        lc1 = mis_lc mispec1 and
        lc2 = mis_lc mispec2 and
        nparams1 = mis_nparams mispec1 and
        nparams2 = mis_nparams mispec2
    in  nparams1=nparams2
        & convfun ar1 ar2
        & convfun lc1 lc2
;;

let conv_mconstructor convfun
    (DOPN(MutConstruct(indsp1,i1),cl1)) (DOPN(MutConstruct(indsp2,i2),cl2)) =
    i1 = i2
    & (indsp1 = indsp2
       or conv_minductive convfun (DOPN(MutInd indsp1,cl1))
                                  (DOPN(MutInd indsp2,cl2)))
;;

let rec rename_rels curidx sofar = function
    [] -> sofar
  | ((id,Rel n)::tl as l) ->
    if curidx = n then rename_rels (curidx+1) (subst1 (VAR id) sofar) tl
    else rename_rels (curidx+1) (subst1 (DOP0 Implicit) sofar) l
;;

let pop_references l = (make_set (prec l)) where rec prec = function
    1::l -> prec l
  | n::l -> (n-1)::(prec l)
  | [] -> []
;;

let rec free_refs c = frec [] c where rec frec l = function
    Rel(m)      -> m::l
  | DOPN(_,cl)       -> it_vect frec l cl
  | DOPL(_,cl)       -> it_list frec l cl
  | DOP2(_,c1,c2) -> frec (frec l c1) c2
  | DOP1(_,c) -> frec l c

  | DLAM(_,c) -> (pop_references (frec [] c))@l

  | DLAMV(_,cl) -> (pop_references (it_vect frec [] cl))@l

  | VAR _ -> l
  | DOP0 _ -> l
;;

let clean_rhs rhs worklist =
let workvars = filter (isVAR o snd) worklist in
let rhs' = replace_vars (map (fun (id',VAR id) -> (id,CLOSEDTERM (VAR id'))) workvars) rhs in
let workrels = filter (isRel o snd) worklist in
let workrels' = sort_increasing_snd workrels
in rename_rels 1 rhs' workrels'
;;

let evar_hnf c =
    match hnf_constr (get_ise()) c with
    DOPN(AppL,cl) -> (hd_vect cl,list_of_tl_vect cl)
  | c -> (c,[])
;;

let explode_appl M =
let (DOPN(AppL,cl)) = ensure_appl(collapse_appl M)
in (hd_vect cl,list_of_tl_vect cl)
;;

let evar_apprec l c = evar_hnf (applistc c l);;

let has_undefined_isevars c =
    (try (hasrec c ; false) with Failure "caught" -> true)
    where rec hasrec = function
    DOPN(Const sp,cl) as K ->
    if ise_in_dom sp then 
       if defined_const (get_ise()) K then hasrec (const_value (get_ise()) K)
                                      else failwith "caught"
    else do_vect hasrec cl

  | DOP1(_,c) -> hasrec c
  | DOP2(_,c1,c2) -> (hasrec c1; hasrec c2)
  | DOPL(_,l) -> do_list hasrec l
  | DOPN(_,cl) -> do_vect hasrec cl
  | DLAM(_,c) -> hasrec c
  | DLAMV(_,cl) -> do_vect hasrec cl
  | (VAR _|Rel _|DOP0 _) -> ()
;;

let rec do_solve_pb(sp,args,rhs) =
(let evd = ise_map sp in
let hyps= evd.evgoal.hyps in
let worklist = combine (ids_of_sign hyps,args) in
let rhs'' = clean_rhs rhs worklist in
if not(closed rhs'') then error "cannot solve this unification problem";
(* rhs'' should be the completely clean term. *)
ise_define sp rhs'';
solved_problems := (sp,ise_map sp)::!solved_problems;
let pbs = !conversion_problems in
   conversion_problems := [];
   for_all (function pb ->
            if status_changed sp pb then
                rerun_problem pb
            else (add_conv_pb pb;true)) pbs)


and rerun_problem = function
(*    (CONV,t1,t2) -> evar_conv t1 t2
  | (CONV_LEQ,t1,t2) -> evar_conv_leq t1 t2
  |*) (CONV_X,t1,t2) -> evar_conv_x t1 t2


and solve_pb ((pbty,t1,t2) as pb) =
    if t1 = t2 then true else
    match (t1,t2) with
    (DOPN(Const sp1,args1),DOPN(Const sp2,args2)) ->
    if unsolved_ise t1 & ise_in_dom sp2  & ise_defined t2 then
        try_solve_pb(sp1,args1,t2)
    else if unsolved_ise t2 & ise_defined t1 then
        try_solve_pb(sp2,args2,t1)
    else (add_conv_pb pb;true)

  | (DOPN(Const sp1,args1),t2) -> if unsolved_ise t1 
                                     then try_solve_pb(sp1,args1,t2)
                                     else  (add_conv_pb pb;true)

  | (t1,DOPN(Const sp2,args2)) -> if unsolved_ise t2 then 
                                  try_solve_pb(sp2,args2,t1)
                                  else (add_conv_pb pb;true)
    
    
and try_solve_pb(sp,argsv,t2) =
let args = list_of_vect argsv
in if for_all (fun (VAR _) -> true | (Rel _) -> true
                 | (DOP0 Implicit) -> true | _ -> false) args
    & distinct (filter (fun (DOP0 Implicit) -> false | _ -> true) args) then
    (* We can instanciate the existential *)
    do_solve_pb(sp,args,t2)
   else error "We found an existential that we could not instanciate"


and evar_conv_x term1 term2 =
let term1 = strip_outer_cast term1 and
    term2 = strip_outer_cast term2 
in if term1 = term2 then true
   else if not(has_undefined_isevars term1) & not(has_undefined_isevars term2) then
       conv_x (get_ise()) term1 term2
    else if (match (term1,term2) with
        (DOPN(Const sp1,_),DOPN(Const sp2,_)) ->
        (ise_in_dom sp1 & not(ise_defined term1))
     or (ise_in_dom sp2 & not(ise_defined term2))
      | (DOPN(Const sp1,_),_) -> unsolved_ise term1
      | (_,DOPN(Const sp2,_)) -> unsolved_ise term2
      | _ -> false) then
        solve_pb(CONV_X,term1,term2)
    else 
        let (t1,l1) = evar_hnf term1
        and (t2,l2) = evar_hnf term2
        in if (head_is_embedded_exist t1 & not(is_eliminator t2))
            or (head_is_embedded_exist t2 & not(is_eliminator t1)) then
            (add_conv_pb (CONV_X,applist(t1,l1),applist(t2,l2));true)
        else
            evar_eqappr_x ((t1,l1),(t2,l2))

and evar_eqappr_x = function
       ((DOPN(Const sp1,al1) as term1,l1),(DOPN(Const sp2,al2) as term2,l2)) ->
       ((sp1=sp2 & for_all2eq_vect evar_conv_x al1 al2) (* case of two axioms *)       or
       (unsolved_ise term1 or unsolved_ise term2) &
       solve_pb(CONV_X,DOPN(Const sp1,al1),DOPN(Const sp2,al2)))
       & for_all2eq evar_conv_x l1 l2

     | ((DOPN(Const sp1,al1) as term1,l1),(t2,l2)) ->  
       unsolved_ise term1 &
       solve_pb(CONV_X,DOPN(Const sp1,al1),t2)
       & for_all2eq evar_conv_x l1 l2

     | ((t1,l1),(DOPN(Const sp2,al2) as term2,l2))  -> 
       unsolved_ise term2 &
       solve_pb(CONV_X,t1,DOPN(Const sp2,al2))
       & for_all2eq evar_conv_x l1 l2

     | ((DOPN(Abst sp1,al1),l1),(DOPN(Abst sp2,al2),l2)) ->
           (eq(sp1,sp2) & eq(al1,al2) & (length(l1) = length(l2)) & (for_all2 (evar_conv_x) l1 l2))
        or ((evaluable_abst (DOPN(Abst sp2,al2)))
            & evar_eqappr_x ((DOPN(Abst sp1,al1),l1), evar_apprec l2 (abst_value (DOPN(Abst sp2,al2)))))
        or ((not (evaluable_abst (DOPN(Abst sp2,al2))))
            & (evaluable_abst (DOPN(Abst sp1,al1)))
            & evar_eqappr_x  (evar_apprec l1 (abst_value (DOPN(Abst sp1,al1))),(DOPN(Abst sp2,al2),l2)))

     | ((DOPN(Abst sp1,al1),l1),p2) ->  
       (evaluable_abst (DOPN(Abst sp1,al1)))
       & evar_eqappr_x (evar_apprec l1 (abst_value (DOPN(Abst sp1,al1))),p2)

     | (p1,(DOPN(Abst sp2,al2),l2))  -> 
       (evaluable_abst (DOPN(Abst sp2,al2)))
       & evar_eqappr_x (p1,evar_apprec l2 (abst_value (DOPN(Abst sp2,al2))))

     | ((Rel(n),l1),(Rel(m),l2)) ->
           (n=m) & (length(l1) = length(l2)) & (for_all2 evar_conv_x l1 l2)
     | ((DOP2(Cast,c,_),l),appr2) -> evar_eqappr_x ((c,l),appr2)
     | (appr1,(DOP2(Cast,c,_),l)) -> evar_eqappr_x (appr1,(c,l))
     | ((VAR _ as v1,l1),(VAR _ as v2,l2)) ->
           (eq(id_of_var v1,id_of_var v2) & (length(l1) = length(l2)) & (for_all2 evar_conv_x l1 l2))
     | ((DOP0(Meta(n)),l1),(DOP0(Meta(m)),l2)) ->
           ((n=m) & (length(l1) = length(l2)) & (for_all2 evar_conv_x l1 l2))
     | ((DOP0(Sort s1),[]),(DOP0(Sort s2),[])) -> sort_eq_x s1 s2
     | ((DOP2(Lambda,c1,DLAM(_,c2)),[]),(DOP2(Lambda,c'1,DLAM(_,c'2)),[])) -> 
                                          evar_conv_x c1 c'1 & evar_conv_x c2 c'2
     | ((DOP2(Prod,c1,DLAM(_,c2)),[]),(DOP2(Prod,c'1,DLAM(_,c'2)),[])) -> 
                                          evar_conv_x c1 c'1 & evar_conv_x c2 c'2
     | ((DOPN(MutInd indsp1,cl1) as ind1,l'1),(DOPN(MutInd indsp2,cl2) as ind2,l'2)) ->
        (ise_or (fun () -> (indsp1 = indsp2 & for_all2eq_vect evar_conv_x cl1 cl2))
                 (fun () -> conv_minductive evar_conv_x ind1 ind2))
       & for_all2eq evar_conv_x l'1 l'2
         
     | ((DOPN(MutConstruct i1,cl1) as constr1,l1),(DOPN(MutConstruct i2,cl2) as constr2,l2)) ->
        (((i1=i2) & for_all2eq_vect evar_conv_x cl1 cl2)
         or conv_mconstructor evar_conv_x constr1 constr2)
        & for_all2eq evar_conv_x l1 l2

     | ((DOPN(MutCase,cl1),l'1),(DOPN(MutCase,cl2),l'2)) -> 
         (for_all2eq_vect evar_conv_x cl1 cl2) & (for_all2eq evar_conv_x l'1 l'2)

     | ((DOPN(Fix(i1),cl1),l1),(DOPN(Fix(i2),cl2),l2))   ->
       i1=i2 & 
       (for_all2eq_vect evar_conv_x cl1 cl2) &
       (for_all2eq evar_conv_x l1 l2)

     | ((DOPN(CoFix(i1),cl1),l1),(DOPN(CoFix(i2),cl2),l2))   ->
       i1=i2 & 
       (for_all2eq_vect evar_conv_x cl1 cl2) &
       (for_all2eq evar_conv_x l1 l2)

     | (DOP0(Implicit),[]),(DOP0(Implicit),[]) -> true
(* added to compare easily the specification of fixed points *)
     | (DLAM(_,c1),[]),(DLAM(_,c2),[]) -> evar_conv_x c1 c2
     | (DLAMV(_,vc1),[]),(DLAMV(_,vc2),[]) -> for_all2eq_vect evar_conv_x vc1 vc2
     | _ -> false;;


let the_conv_x t1 t2 = conv_x (get_ise()) t1 t2 or evar_conv_x t1 t2;;


let evar_type_fix (DOPN(Fix(_,j),vargs)) vdeft = 
let lt = vect_length vdeft in 
let var = sub_vect vargs 0 lt in 
if for_all2eq_vect
   (fun ari ti ->  the_conv_x (lift lt ari) ti) var vdeft then
   var.(j) 
else error "Implicit Syntax: Ill-typed case recursive call
                             (branches are malformed)";;


let evar_type_cofix (DOPN(CoFix(j),vargs)) vdeft = 
let lt  = vect_length vdeft in 
let var = sub_vect vargs 0 lt in 
if for_all2eq_vect
   (fun ari ti ->  the_conv_x (lift lt ari) ti) var vdeft then
   var.(j) 
else error "Implicit Syntax: (cofix) Ill-typed case recursive call
                             (branches are malformed)";;

(* attempt_coercion will attempt to make [argj.TYP] be convertible
   with [typ].  The methods used are:

   if [argj.TYP] is a product, (x:A)B, and B[?n/x] successfully
   converts with [typ], then we can repalce [argj.VAL] with
   (argj.VAL ?n) and [argj.TYP] with B[?n/x].

   if [argj.TYP] is a product, then we can try eta-converting 

 *)

(* TODO: Verify that the kind is correct here.
   WARNING: I am not certain that this is the correct KIND *)
let attempt_coercion env typ = apprec
    where rec apprec argj =
    match hnf_constr (get_ise()) argj.TYP with
    DOP2(Prod,argc1,argB) ->
    let K = new_isevar env argc1 in
    let argj = {VAL = applist(argj.VAL,[K]);
                TYP = SAPP argB K;
                KIND = argj.KIND}
    in if ise_pred (the_conv_x (SAPP argB K))  typ then
        argj
       else apprec argj

  | _ -> failwith "attempt_coercion"
;;

let apply_rel_list env argjl funj tycon =
let rec apply_rec acc typ = function
    [] -> (match tycon with NONE -> ()
                         | SOME typ' -> let _ = the_conv_x typ typ' in ());
          {VAL=applist(j_val funj,map j_val (rev acc));
           TYP= typ;
           KIND = funj.KIND}
  | hj::restjl ->
  match hnf_constr (get_ise()) typ with
     DOP2(Prod,c1,DLAM(_,c2)) ->
     if the_conv_x hj.TYP c1 then
         apply_rec (hj::acc) (subst1 hj.VAL c2) restjl
        else
            (try let hj = attempt_coercion env c1 hj
                 in apply_rec (hj::acc) (subst1 hj.VAL c2) restjl
             with UserError _ | Failure _ ->
                (errorlabstrm "trad__apply_rel_list"
              [< 'S"Illegal application (Type Error) : "; 'CUT ;
                 apply_error env (funj.VAL,funj.TYP)
                                 (map (fun j -> (j.VAL,j.TYP)) argjl) >]))
   | _ -> errorlabstrm "trad__apply_rel_list"
              [< 'S"Illegal application (Non-functional construction) : ";
                 'CUT ;
                 apply_error env (funj.VAL,funj.TYP)
                                 (map (fun j -> (j.VAL,j.TYP)) argjl) >]
in apply_rec [] funj.TYP argjl;;

let cast_rel env cj tj =
    if the_conv_x cj.TYP tj.VAL then
        {VAL=j_val cj;
         TYP=tj.VAL;
         KIND = hnf_constr (get_ise()) tj.TYP}
    else
        (try attempt_coercion env tj.VAL cj
         with UserError _ | Failure _ ->
             errorlabstrm "cast"
        [< 'S"In environment " ; pENV env ; 'SPC ;
           'S"The term: " ; pTERMINENV(env,cj.VAL) ; 'SPC ; 
           'S"does not have type" ; 'SPC ; pTERMINENV(env,tj.VAL) ; 'S"." ;
           'FNL ; 'S"Actually, it has type " ;
           'SPC ; pTERMINENV(env,cj.TYP) >])
;;

let type_of_existential sigma env (DOPN(Const sp,args) as K) =
if evc__ctxt_access sigma sp then
try let evd = evc__map sigma sp
    in if for_all2eq_vect (fun id (VAR id') -> id = id' | _ _ -> false) (vect_of_list (ids_of_sign evd.evgoal.hyps)) args
        & hyp_seq_ordered_inclusion evd.evgoal.hyps (get_globals env) then
        evd.evgoal.concl
       else 
        const_type sigma K
with Not_found -> error ((string_of_path sp) ^ " not declared")
   | Failure _ -> invalid_arg "the current environment and the constant's env of construction are incompatible"
else error ((string_of_path sp) ^ " not declared")
;;

let type_of_const sigma env (DOPN(Const sp,_) as K) =
let bn = basename sp
in if "?" = SA__un_intern(atompart_of_id bn) then
    type_of_existential sigma env K
   else type_of_constant sigma env K
;;

let let_path = path_of_string "#Core#let.cci";;


let check_branches_message env (c,ct) (explft,lft) = 
let n = vect_length explft and expn = vect_length lft in
if n<>expn then let c = strong whd_ise c and ct = strong whd_ise ct in
errorlabstrm "Case with wrong number of cases"
 (HOV 0 [< 'S "Case on term "; pTERMINENV(env,c); 'CUT ; 'S "of type"; 'WS 1;
           pTERMINENV(env,ct); 'WS 1; 'CUT; 'S "expects " ; 'INT expn; 'S " branches" >]) 
else
let rec check_conv i = 
    if i = n then () else
    if not (the_conv_x lft.(i) (explft.(i))) then 
let c = strong whd_ise c and ct = strong whd_ise ct 
and lfi = strong whd_ise lft.(i) in
errorlabstrm "Ill-formed branches" 
 (HOV 0 [< 'S "In Case expression on term "; pTERMINENV(env,c); 
           'SPC; 'CUT; 'S "the branch  " ; 'INT (i+1); 'S " has type";  'SPC;
         pTERMINENV(env,lfi); 'CUT; 
         'S " which should be :"; 'WS 1; pTERMINENV(env,explft.(i)) >]) 
    else check_conv (i+1) 
in check_conv 0;;

let evar_type_case env ct Pt lft P c =
let (bty,rslty) = type_case_branches env (get_ise()) ct Pt P c
in check_branches_message env (c,ct) (bty,lft); rslty;;



let rec unsafe_fmachine noverify sign constr = 
    ((if not(noverify) then verify_wf_env sign) ; exemeta_rec (NONE,NONE) sign constr)
(* exemeta_rec (valcon,tycon) env constr tries to solve the *)
(* existential variables in constr in environment env with the *)
(* constraint that constr is of type tycon *)
and exemeta_rec (valcon,tycon) env =  function

  DOP0(XTRA("ISEVAR",[])) ->
  (match valcon with
   SOME val -> exemeta_rec (NONE,NONE) env val
 | NONE ->
  (match tycon with
   NONE -> error "unconstrained isevar"
 | SOME ty -> let tyj = exemeta_rec (NONE,NONE) env ty in
              let K = new_isevar env (j_val tyj)
              in exemeta_rec (NONE,NONE) env K))

 | Rel(n)                 ->  relative (get_ise()) n env 
 | (VAR id as val) ->
  let (DOP2(Cast,typ,kind)) = snd(lookup_glob id env)
  (* and str = id_of_var val *)
  in {VAL=val;
      TYP=typ;
      KIND = kind}

| (DOPN(Const sp,_) as val) ->
  (match type_of_const (get_ise()) env val with
   (DOP2(Cast,typ,kind)) -> {VAL=val;TYP=typ;KIND = kind}
 | typ -> {VAL=val;TYP=typ;
           KIND= hnf_constr (get_ise()) (exemeta_rec (NONE,NONE) env typ).TYP})

| (DOPN(Abst sp,cl) as val) ->
  if sp = let_path then
      (match list_of_vect cl with
       [M;DLAM(na,B)] ->
       let Mj = exemeta_rec (NONE,NONE) env M in
       let Bj = exemeta_rec (NONE,NONE) (add_rel (na,DOP2(Cast,Mj.TYP,Mj.KIND)) env) B in
           {VAL = DOPN(Abst sp,[|Mj.VAL;DLAM(na,Bj.VAL)|]);
            TYP = SAPP (DLAM(na,Bj.TYP)) Mj.VAL;
            KIND = pop Bj.KIND}
     | _ -> errorlabstrm "trad__constr_of_com"
            [< 'S"Malformed ``let''" >])
   else if evaluable_abst val then
       exemeta_rec (valcon,tycon) env (abst_value val)
   else error "Cannot typecheck an unevaluable abstraction"

| DOPN(MutInd _,_) as val ->
  let (DOP2(Cast,typ,kind)) = type_of_mind (get_ise())  env val
  in {VAL=val;
      TYP=typ;
      KIND=kind}
 
| DOPN(MutConstruct _,_) as val          -> 
  let (DOP2(Cast,typ,kind)) = type_of_mconstr (get_ise()) env val
  in {VAL=val;
      TYP=typ;
      KIND=kind}
 
| (DOPN(MutCase,cl) as val) ->
  let P = cl.(0)
  and c = cl.(1)
  and lf = sub_vect cl 2 ((vect_length cl) - 2) in
  (let {TYP=ct;VAL=c} as cj = exemeta_rec (NONE,NONE) env c in
   let {TYP=Pt;VAL=P} as Pj = exemeta_rec (NONE,NONE) env P in
   let evalct = strong whd_ise ct and evalPt = strong whd_ise Pt in
   let (bty,rsty) = type_case_branches env (get_ise()) evalct evalPt P c in
   let lfj = map2_vect (fun tyc f -> exemeta_rec (NONE,SOME tyc) env f) bty lf in
   let lft = (map_vect (fun j -> j.TYP) lfj) in
  (check_branches_message env (c,ct) (bty,lft);
   {VAL=DOPN(MutCase,concat_vect [|j_val Pj; j_val cj|] (map_vect j_val lfj));
    TYP= rsty;
    KIND = sort_of_arity (get_ise()) Pj.TYP}))

| (DOPN(Fix(vn,i),cl) as val) ->
  let nbfix = let nv = vect_length cl in 
           if nv < 2 then error "Ill-formed recursive definition" else nv-1 in
  let lar = sub_vect cl 0 nbfix in
  let ldef = last_vect cl in
  let (lfi,vdef) = decomp_DLAMV_name nbfix ldef in
  let newenv = it_vect2 
     (fun env name ar -> let assum = assumption_of_judgement (get_ise())
                         (exemeta_rec (NONE,NONE) env ar) in 
                         (add_rel (name,assum) env))
     env (vect_of_list (rev lfi)) lar in
  let larj = map_vect (exemeta_rec (NONE,NONE) env) lar in
  let vdefj = map_i_vect 
      (fun i def -> exemeta_rec (NONE,SOME(larj.(i).VAL)) newenv def) 0 vdef in
  let vdefv = put_DLAMSV lfi (map_vect j_val vdefj) in
  let larv = map_vect j_val larj in
  let fix = DOPN(Fix(vn,i),concat_vect larv [|vdefv|])
  in check_fix (get_ise()) fix;
     {VAL= fix;
      TYP=evar_type_fix fix (map_vect (fun j -> j.TYP) vdefj);
      KIND=sort_of_arity (get_ise()) larj.(i).TYP}

| (DOPN(CoFix(i),cl) as val) ->
  let nbfix = let nv = vect_length cl in 
           if nv < 2 then error "Ill-formed recursive definition" else nv-1 in
  let lar = sub_vect cl 0 nbfix in
  let ldef = last_vect cl in
  let (lfi,vdef) = decomp_DLAMV_name nbfix ldef in
  let newenv = it_vect2 
     (fun env name ar -> let assum = assumption_of_judgement (get_ise())
                         (exemeta_rec (NONE,NONE) env ar) in 
                         (add_rel (name,assum) env))
     env (vect_of_list (rev lfi)) lar in
  let larj = map_vect (exemeta_rec (NONE,NONE) env) lar in
  let vdefj = map_i_vect 
      (fun i def -> exemeta_rec (NONE,SOME(larj.(i).VAL)) newenv def) 0 vdef in
  let vdefv = put_DLAMSV lfi (map_vect j_val vdefj) in
  let larv = map_vect j_val larj in
  let cofix = DOPN(CoFix(i),concat_vect larv [|vdefv|])
  in check_cofix (get_ise()) cofix;
     {VAL= cofix;
      TYP=evar_type_cofix cofix (map_vect (fun j -> j.TYP) vdefj);
      KIND=sort_of_arity (get_ise()) larj.(i).TYP}

| DOP0(Sort(Prop(c)))                 -> fcn_proposition c
| DOP0(Sort(Type(c,u)))               -> fcn_type_with_universe (c,u)
| DOPN(AppL,tl)           -> 
  let j = exemeta_rec (NONE,NONE) env (hd_vect tl) in
  let jl = rev(snd(it_list (fun (tycon,jl) c ->
                            let dtc = prod_dom_tycon (get_ise()) tycon in
                            let cj = exemeta_rec (NONE,dtc) env c in
                            let rtc = app_rng_tycon (get_ise()) cj.VAL tycon
                            in (rtc,cj::jl)) (SOME j.TYP,[]) (list_of_tl_vect tl)))
  in
      apply_rel_list env jl j tycon

| DOP2(Lambda,c1,DLAM(name,c2))      -> 
  let j = exemeta_rec (lambda_dom_tycon (get_ise()) tycon,NONE) env c1 in
  let assum = assumption_of_judgement (get_ise()) j in
  let var = (name,assum) in
  let j' = exemeta_rec (NONE,prod_rng_tycon (get_ise()) tycon)
                       (add_rel var env) c2 in 
      abs_rel (get_ise()) name assum j'

| DOP2(Prod,c1,DLAM(name,c2))        ->
  let j = exemeta_rec (NONE,NONE) env c1 in
  let assum = assumption_of_judgement (get_ise()) j in
  let var = (name,assum) in
  let j' = exemeta_rec (NONE,NONE) (add_rel var env) c2 in
      gen_rel (get_ise()) env name assum j'

| DOPN(XTRA("REC",[]),cl)  ->
  let P = cl.(0)
  and c = cl.(1)
  and lf = sub_vect cl 2 ((vect_length cl) - 2) in
  let {TYP=ct;_} as cj = exemeta_rec (NONE,NONE) env c and
      Pj = exemeta_rec (NONE,NONE) env P in
  let P = Pj.VAL and
      c = cj.VAL in
  let (bty,rsty) = indrec__type_rec_branches true (get_ise()) env ct Pj.TYP P c in
  let lfj = map2_vect (fun tyc f -> exemeta_rec (NONE,SOME tyc) env f) bty lf in
  let lft = (map_vect (fun j -> j.TYP) lfj) in
  (check_branches_message env (c,ct) (bty,lft);
     let REC = concat_vect [|P; c|] (map_vect j_val lfj) in
      {VAL=indrec__transform_rec env (get_ise()) REC (cj.TYP,Pj.TYP);
       TYP = rsty;
       KIND = sort_of_arity (get_ise()) Pj.TYP})

| DOP2(Cast,c,t) ->
  let tj = exemeta_rec (NONE,NONE) env t in
  let cj = exemeta_rec (NONE,SOME (assumption_of_judgement (get_ise()) tj)) env c
  in cast_rel env cj tj

| _ -> error "Cannot be executed";;

(* constr_of_com takes an environment of typing assumptions,
 * and translates a command to a constr.
 *)

let raw_constr_of_com sigma env com =
((fun c ->
      try sosub__soexecute c
      with Failure _|UserError _ ->
          errorlabstrm "raw_constr_of_com"
          [< 'S"Failed to contract second-order (parser-introduced)"; 'SPC ;
             'S"redexes ; perhaps the term " ; 'SPC ; pCMD(CCI,com) ; 'SPC ;
             'S"is malformed?" >])
o (astterm__dbize_cci sigma (termast__unitize_env env))) com
;;

let raw_fconstr_of_com sigma env com =
((fun c ->
      try sosub__soexecute c
      with Failure _|UserError _ ->
          errorlabstrm "raw_constr_of_com"
          [< 'S"Failed to contract second-order (parser-introduced)"; 'SPC ;
             'S"redexes ; perhaps the term " ; 'SPC ; pCMD(FW,com) ; 'SPC ;
             'S"is malformed?" >])
o (astterm__dbize_fw sigma (termast__unitize_env env))) com
;;

let raw_constr_of_compattern sigma env com =
((fun c ->
      try sosub__try_soexecute c
      with Failure _|UserError _ ->
          errorlabstrm "raw_constr_of_compattern"
          [< 'S"Failed to contract second-order (parser-introduced)"; 'SPC ;
             'S"redexes ; perhaps the pattern" ; 'SPC ; pCMD(CCI,com) ; 'SPC ;
             'S"is malformed?" >])
o (astterm__dbize_cci sigma (termast__unitize_env env))) com
;;

let has_ise = hasrec
    where rec hasrec = function
    DOP0(XTRA("ISEVAR",[])) -> true
  | DOPN(XTRA("REC",[]),_) -> true
  | DOP1(_,c) -> hasrec c
  | DOP2(_,c1,c2) -> hasrec c1 or hasrec c2
  | DOPN(_,cl) -> exists_vect hasrec cl
  | DLAM(_,c) -> hasrec c
  | DLAMV(_,cl) -> exists_vect hasrec cl
  | _ -> false
;;

let is_fmachine noverify sign c =
conversion_problems := [];
solved_problems := [];
set_ise mt_evc;
let j = unsafe_fmachine noverify sign c
in strong whd_ise j.VAL;;

let ise_resolve sigma env c =
    if has_ise c then
    try is_fmachine false env c
    with Failure "whd_ise" ->
        errorlabstrm "ise_resolve"
        [< 'S"The term " ; 'CUT  ; pTERM c ; 'CUT ;
           'S" still has some unresolved implicit syntax" ; 'FNL >]
   else c
;;

let constr_of_com_env sigma hyps com = 
let c = raw_constr_of_com mt_evc hyps com
  in ise_resolve sigma hyps c
;;

let fconstr_of_com_env sigma hyps com = 
let c = raw_fconstr_of_com mt_evc hyps com
  in ise_resolve sigma hyps c
;;
 
let constr_of_com_sign sigma sign com = 
  let hyps = (GLOB sign) in constr_of_com_env sigma hyps com;;
  

let fconstr_of_com_sign sigma sign com =
   let hyps = (GLOB sign) in
  let c = raw_fconstr_of_com mt_evc hyps com
  in ise_resolve sigma hyps c
;;

(* Ignore de Bruijn variables *)
let constr_of_com sigma env com =
  let sign = get_globals env in
  constr_of_com_sign sigma sign com
;;

let fconstr_of_com sigma env com =
  let sign = get_globals env in
  fconstr_of_com_sign sigma sign com
;;

let fconstruct sigma sign com =
let c = constr_of_com sigma (GLOB sign) com
in mach__fexecute sigma sign c;;

let infconstruct sigma (sign,fsign) cmd =
    mach__infexecute sigma (sign,fsign) (constr_of_com sigma (GLOB sign) cmd);;

let try_whd_ise c =
    try whd_ise c
    with Failure _ -> c
;;

let attempt_is_fmachine noverify sign c =
conversion_problems := [];
solved_problems := [];
set_ise mt_evc;
let j = unsafe_fmachine noverify sign c
in strong try_whd_ise j.VAL;;

let attempt_ise_resolution sign c =
    if not(has_ise c) then (mach__fexecute mt_evc sign c; c)
    else let c' = attempt_is_fmachine false (GLOB sign) c
         in strong (fun (DOPN(Const sp,_) as K) ->
                        if ise_in_dom sp then DOP0(XTRA("ISEVAR",[]))
                        else K
                      | c -> c) c'
;;

(* $Id: trad.ml,v 1.12 1996/02/12 15:39:24 cpaulin Exp $ *)
