(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                vernac.ml                                 *)
(****************************************************************************)

(* Concrete syntax of the mathematical vernacular MV V2.6 *)

#open "std";;
#open "initial";;
#open "names";;
#open "system";;
#open "lexer";;
#open "more_util";;
#open "ast";;
#open "pp";;
#open "stdpp";;
#open "vkey";;
#open "library";;
#open "environ";;
#open "pfedit";;
#open "himsg";;
#infix "o";;

let lequal_spat = plist 
  (function
   [< 'Tident var; 'Tcolonequal; ast__parse_spat valu >] -> (id_of_string var,valu)
  )
;;

let pr_equal_spat =
  prlist_with_sep
    (fun () -> [< 'SPC >])
    (fun (v,val) -> [< 'S(string_of_id v) ; 'SPC ;
		       'S(string_of_token rev_vernac_keywords Tcolonequal);
		       ast__print_spat val >])
;;

(* vernac parses the given stream, executes interpfun on the syntax tree it
 * parses, and is verbose on "primitives" commands if verbosely is true *)
let rec vernac interpfun verbosely =
 let MSGve = if verbosely then MSGNL else function _ -> () in
 function
  [< 'Tkw Kgrammar ; 'Tident whatfor ;
     (ne_list_with_sep (function [< 'Tkw Kwith >] -> ())
                       (parsing__grammar_parser (id_of_string whatfor))) l ; 
     'Tdot >] ->
       	 MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	  [< 'Tkw Kgrammar ; 'Tident whatfor >] ;
      	       	  prlist_with_sep
      	       	    (fun () ->
      	       	       [< 'FNL ; 'S(string_of_token vkey__rev_vernac_keywords
      	       	       	       	       	(Tkw Kwith)) ; 'SPC >])
		    (fun g_entry ->
      	       	       pureparse__print_grammar_entry (id_of_string whatfor)
      	       	       	       	       	       	      g_entry)
		    l;
		     'S(string_of_token vkey__rev_vernac_keywords Tdot) >];
  add_anonymous_object (pureparse__inGrammar l)

| [< 'Tkw Kaddpath ; 'Tstring dir ; 'Tdot >] ->
      	 MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	    [< 'Tkw Kaddpath ; 'Tstring dir ; 'Tdot >] >];
      	 add_path dir

| [< 'Tkw Kdelpath ; 'Tstring dir ; 'Tdot >] ->
      	 MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	    [< 'Tkw Kdelpath ; 'Tstring dir ; 'Tdot >] >];
      	 del_path dir
| [< 'Tident "Drop"; 'Tdot >] ->
      	 MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	    [< 'Tident "Drop"; 'Tdot >] >];
	 vernacinterp__drop ()
| [< 'Tident "Quit"; 'Tdot >] ->
      	 MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	    [< 'Tident "Drop"; 'Tdot >] >];
	 sys__exit 0
| [<
    'Tkw Kload;
    begin function
    [< 'Tident "Verbose" >] -> true
  | [< >] -> false
    end verbosely;
    begin function
      [< 'Tident s >] -> s
    | [< 'Tstring s >] -> s
    end s;
    'Tdot
  >] ->  MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	    [< 'Tkw Kload; (if verbosely then [< 'Tident "Verbose" >]
		                    else [<>]);
		       'Tstring s ; 'Tdot>] >];
	 load_vernac verbosely s

| [< 'Tkw Krestore; 'Tkw Kstate; 'Tident name ; 'Tdot >] ->
      	 MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	    [< 'Tkw Krestore; 'Tkw Kstate; 'Tident name ; 'Tdot >] >];
	 restore_state name

| [< 'Tident "Read" ; 'Tident "Module" ; 'Tident name ; 'Tdot >] ->
      	 MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	    [< 'Tident "Read" ;'Tident"Module";'Tident name;'Tdot>]>];
      	 library__intern_module name

| [< 'Tident "Module" ; 'Tident mname ; 'Tdot >] ->
      	MSGve [< fmt_token_stream vkey__rev_vernac_keywords
	       [< 'Tident "Module" ; 'Tident mname ; 'Tdot >] >];
	library__open_module mname

| [< 'Tident "Compile" ;
     begin function
     [< 'Tident "Verbose" >] -> true
   | [< >] -> false
     end verbosely;
     'Tident "Module" ; 'Tident mname ;
     begin function
     [< 'Tstring fname >] -> fname
   | [< >] -> mname
     end fname;
    'Tdot >] ->
      	 MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	[< 'Tident "Compile";(if verbosely then[<'Tident"Verbose">]
		                      else [<>]);
		   'Tident "Module" ; 'Tident mname ; 'Tstring fname;'Tdot>]>];
       	 compile_module verbosely mname fname

| [< 'Tident "Import" ; 'Tident name ; 'Tdot >] ->
      	 MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	    [< 'Tident "Import" ; 'Tident name ; 'Tdot >] >];
      	 library__import_module name

| [< 'Tident "Require" ;
    (function
      [< 'Tident "Implementation" ; 'Tident name ; (function
          [< 'Tstring fname ; 'Tdot >] ->
      	 MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	    [< 'Tident "Require" ;'Tident "Implementation" ;
      	       	       'Tident name ; 'Tstring fname ; 'Tdot >] >];
	 library__require_module_from name fname true

        | [< 'Tdot >] ->
      	 MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	    [< 'Tident "Require" ;'Tident "Implementation" ;
      	       	       'Tident name ; 'Tdot >] >];
	 library__require_module name true ) _ >] -> ()

    | [< 'Tident "Import" ; 'Tident name ; (function
          [< 'Tstring fname ; 'Tdot >] 
               -> (*library__require_module_specification_from*)
                  (* we do not want to deal with module specification
                    any longer... for the moment *)
         MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	    [< 'Tident "Require" ;'Tident "Import" ;
      	       	       'Tident name ; 'Tstring fname ; 'Tdot >] >];
         library__require_module_from name fname true

        | [< 'Tdot >] 
               -> (*library__require_module_specification*)
                  (* we do not want to deal with module specification
                    any longer... for the moment *)
         MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	    [< 'Tident "Require" ;'Tident "Import" ;
      	       	       'Tident name ; 'Tdot >] >];
                  library__require_module name true) _ >] -> ()

    | [< 'Tident "Export" ; 'Tident name ; (function
          [< 'Tstring fname ; 'Tdot >] 
               -> (*library__require_module_specification_from*)
                  (* we do not want to deal with module specification
                    any longer... for the moment *)
         MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	    [< 'Tident "Require" ;'Tident "Export" ;
      	       	       'Tident name ; 'Tstring fname ; 'Tdot >] >];
                  library__require_module_from name fname false

       | [< 'Tdot >] 
              -> (*library__require_module_specification*)
                  (* we do not want to deal with module specification
                    any longer... for the moment *)
         MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	    [< 'Tident "Require" ;'Tident "Export" ;
      	       	       'Tident name ; 'Tdot >] >];
                  library__require_module name false) _ >] -> ()

    | [< 'Tident name ; (function
          [< 'Tstring fname ; 'Tdot >] 
               -> (*library__require_module_specification_from*)
                  (* we do not want to deal with module specification
                    any longer... for the moment *)
         MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	    [< 'Tident "Require" ;
      	       	       'Tident name ; 'Tstring fname ; 'Tdot >] >];
                   library__require_module_from name fname true
        | [< 'Tdot >] 
               -> (*library__require_module_specification*)
                  (* we do not want to deal with module specification
                    any longer... for the moment *)
         MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	    [< 'Tident "Require" ;
      	       	       'Tident name ; 'Tdot >] >];
                   library__require_module name true) _ >] -> ()
    ) _ >] -> ()

| [< 'Tident "Export" ; 'Tident name ; 'Tdot >] ->
         MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	    [< 'Tident "Export" ; 'Tident name ; 'Tdot >] >];
	 library__export_module name

| [< 'Tkw Ksyntax; 'Tident whatfor ; 'Tident s; 
     parsing__spat_gram (FRAGprim(TOKdyn d)) ;
     unparsings__unparsing unp ;
     begin function
     [< 'Tkw Kwith ; lequal_spat consl >] -> consl
   | [< >] -> []
     end consl ;
     'Tdot >] ->
       	MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	    [< 'Tkw Ksyntax; 'Tident whatfor ; 'Tident s >];
	        pridyn__pr_dyn d; 'SPC;
      	       	pridyn__pr_dyn(unparsings__inUNPARSING unp);
		if consl = [] then [<>]
		else
                [<'FNL;
      	       	  'S(string_of_token vkey__rev_vernac_keywords (Tkw Kwith));
		  'SPC;
      	       	  pr_equal_spat consl >];
		'S(string_of_token vkey__rev_vernac_keywords Tdot)>];
  (match pridyn__tag d with
   "SPAT" ->
   let spat = outSPAT d in
   let gt = astterm__spat_adjust_consts (GLOB (termenv__initial_sign())) spat in
   let se = {syntax__syn_id = s;
             syntax__syn_universe = id_of_string whatfor;
             syntax__syn_astpat = gt;
             syntax__syn_recbindings = consl;
             syntax__syn_unparsing = unp}
   in add_anonymous_object (inPPSyntax(se))
     | _ -> error "Malformed pretty-printing spec")

| [<
    'Tkw Kreset;
    begin function
      [< 'Tident "after"; 'Tident s; 'Tdot >] ->
         MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	  [< 'Tkw Kreset; 'Tident "after"; 'Tident s; 'Tdot >] >];  
        reset_keeping_name (id_of_string s)

    | [< 'Tident "Initial"; 'Tdot >] ->
         MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	  [< 'Tkw Kreset; 'Tident "Initial"; 'Tdot >] >];
	 reset_initial()

    | [< 'Tident "Section"; 'Tident s; 'Tdot >] ->
         MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	  [< 'Tkw Kreset; 'Tident "Section" ; 'Tident s; 'Tdot >] >];
	 reset_section s

    | [< 'Tident s; 'Tdot >] ->
         MSGve [< fmt_token_stream vkey__rev_vernac_keywords
      	       	  [< 'Tkw Kreset; 'Tident s; 'Tdot >] >];
      	 reset_name (id_of_string s)
    end x
  >] -> x

| [< parsing__vernac vcom >] -> MSGve (pVERNAC vcom);interpfun vcom

| [< (stream_get) _ >] -> raise Parse_error

and load_vernac verbosely s =
  let interp = if not verbosely
       	       then (* parsing and executing silently *)
                    (fun ast -> silently vernacinterp__interprec ast) 
               else (* parsing and executing verbosely *)
      	       	    (fun ast -> vernacinterp__interprec ast)
  in
    let in_chan = open_with_suffix_from_path open_in (search_paths()) s ".v" in
    try
     let ts = token_stream vkey__vernac_keywords (stream_of_channel in_chan)
     in(* we go out of the following infinite loop when a End_of_file or
        * a Parse_failure is raised, which means that we raised the end
        * of the file being loaded *)
     while true do vernac interp verbosely ts done
    with
    e ->(* whatever the exception, we must close the file first *)
        (try close_in in_chan with _ -> ()) ;
         match e	     (* then we cope with it *)
         with Parse_failure | End_of_file -> ()

            | Parse_error ->
      	        MSGERRNL [< 'FNL; 'S"Parse error during load of " ; 'S s >];
                errorlabstrm "load_vernac" [< lexer__fmt_syntax_error() >]

            | Lexer_error n ->
      	        MSGERRNL [< 'FNL; 'S"Lexical error during load of " ; 'S s >];
                errorlabstrm "load_vernac" [< lexer__fmt_lexer_error n >]

            | e ->
      	        MSGERRNL [< 'FNL; 'S"Load of " ; 'S s ; 'S" failed.">];
	        raise e

and compile_module verbosely mname fname =
 let interp = if not verbosely
      	      then (* parsing and executing silently *)
                   (fun ast -> silently vernacinterp__interprec ast) 
              else (* parsing and executing verbosely *)
      	       	   (fun ast -> vernacinterp__interprec ast)
 in
 library__with_heavy_rollback (* we want to roll back in case of error *)
   (fun () ->
     library__open_module mname; (* creating a new module *)
     let (in_chan,longfname) = open_with_suffix_from_path
                                   (fun x -> open_in x,x)
                                   (search_paths()) fname ".v" in
     try let ts = token_stream vkey__vernac_keywords
      	       	       	       	    (stream_of_channel in_chan) in
      	 (* we go out of the following infinite loop when a End_of_file or
          * a Parse_failure is raised, which means that we raised the end
          * of the file being compiled *)
          while true do vernac interp verbosely ts done
      with e ->(* whatever the exception, we try to close the file first *)
      	       (try close_in in_chan with _ -> ());
	        match e
		with
      	       	  Parse_failure | End_of_file ->
      	       	    (* when the end of the file is reached, we close *)
		    (* the module ... *)
      	       	    discharge__close_section_or_module mname;
		    (* ... and write it on the disk as longfname *)
		    (* without the ".v" suffix *)
		    library__extern_module_to mname
      	       	      (if filename__check_suffix longfname ".v"
		       then filename__chop_suffix longfname ".v"
		       else longfname)

		| Parse_error ->
      	       	    MSGERRNL [< 'FNL; 'S "Parse error during compile of ";
		                'S longfname >];
                    errorlabstrm"compile_module"[< lexer__fmt_syntax_error() >]

		| Lexer_error n ->
      	       	    MSGERRNL [< 'FNL ; 'S"Lexical error during compile of " ;
      	       	       	       	'S longfname >];
                    errorlabstrm "compile_module"[< lexer__fmt_lexer_error n >]

		| e -> MSGERRNL [< 'FNL; 'S"Compile of " ; 'S longfname;
		                   'S" failed.">];
		       raise e) ()
;;

(* parse_vernac : vernac_keyword token stream -> unit
 * parses and executes the vernacular token stream *)
let parse_vernac = vernac vernacinterp__interp false;;

(* $Id: vernac.ml,v 1.5 1996/02/12 15:39:31 cpaulin Exp $ *)
