(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                             vernacentries.ml                             *)
(****************************************************************************)

(* Concrete syntax of the mathematical vernacular MV V2.6 *)

#infix "o";;
#open "std";;
#open "initial";;
#open "names";;
#open "system";;
#open "term";;
#open "mach";;
#open "discharge";;
#open "tacmach";;
#open "pfedit";;
#open "command";;
#open "pretty";;
#open "lexer";;
#open "vkey";;
#open "parsing";;
#open "himsg";;
#open "more_util";;
#open "pp";;
#open "stdpp";;
#open "proof";;
#open "proof_trees";;
#open "tactics";;
#open "library";;
#open "libobject";;
#open "environ";;
#open "termenv";;
#open "states";;
#open "trad";;
#open "ast";;
#open "vernacinterp";;
#open "declare";;

let join_binders binders = 
join (list__map (fun (idl,c) -> list__map (fun id -> (id,c)) idl) binders);;

let add = vinterp_add;;

let show_script () =
    let pts = get_pftreestate () in
    let pf = proof_of_pftreestate pts and
        evc = evc_of_pftreestate pts
    in MSGNL(print_script true (rc_of_gc evc pf.goal) (initial_sign()) pf)
;;

let show_prooftree () =
    let pts = get_pftreestate () in
    let pf = proof_of_pftreestate pts and
        evc = evc_of_pftreestate pts
    in MSG(print_proof (rc_of_gc evc pf.goal) (initial_sign()) pf)
;;

let show_open_subgoals () =
    let pf = proof_of_pftreestate (get_pftreestate ())
    in MSG(printer__pr_subgoals (fst(frontier pf)));;

let show_nth_open_subgoal n =
    let pf = proof_of_pftreestate (get_pftreestate ())
    in MSG(printer__pr_subgoal n (fst(frontier pf)));;

let show_open_subgoals_focused () =
    let pf = proof_of_pftreestate (get_pftreestate ())
    in match focus() with
             0 -> MSG(printer__pr_subgoals (fst(frontier pf)))
            | n -> let gls = fst(frontier pf) in 
    if n > length gls then (make_focus 0; MSG(printer__pr_subgoals gls))
    else MSG(printer__pr_subgoal n gls);;


let show_node () =
    let pts = get_pftreestate () in
    let pf = proof_of_pftreestate pts and
        evc = evc_of_pftreestate pts and
        cursor = cursor_of_pftreestate pts
    in MSG [< prlist_with_sep (fun () -> [< 'S" " >])
                             (fun n -> [< 'INT n >])
                             (rev cursor); 'FNL ;
             printer__prgl pf.goal ; 'FNL ;
             (match pf.ref with
              NONE -> [< 'S"BY <rule>" >]
            | SOME(r,spfl) ->
              [< 'S"BY " ; print_rule (rc_of_gc evc pf.goal) pf.goal r ; 'FNL ;
                 'S"  " ; HOV 0 (prlist_with_sep (fun () -> [< 'FNL >]) printer__prgl
                                 (map (fun p -> p.goal) spfl)) >]) ;
             'FNL >]
;;

let print_loadpath () =
    let l = search_paths () in
        MSG [< 'S"Load Path: " ;
               HOV 0 (prlist_with_sep (fun () -> [< 'FNL >]) 
                      (fun s -> [< 'S s >]) l) ; 'FNL >]
;;

add("ADDPATH",
    fun  [VARG_STRING dir] ->
    (fun () -> add_path dir));;

add("DELPATH",
    fun  [VARG_STRING dir] ->
    (fun () -> del_path dir));;

add("PWD",
    fun [] -> fun () -> print_endline (unix__getcwd()));;

add("DROP",
    fun [] -> fun () -> raise Drop);;

add("QUIT",
    fun [] -> fun () -> sys__exit 0);;

add("CD",
    fun  [VARG_STRING s] ->
      	  (fun () ->
             (try sys__chdir (glob s)
              with sys__Sys_error str -> warning ("Cd failed: " ^ str));
             print_string (unix__getcwd()))
       | [] -> (fun () -> print_string (unix__getcwd()))
);;

add("GOAL",
    fun [VARG_COMMAND com] ->
        (fun () ->
            if not(refining()) then
            (start_proof "Unnamed_thm" NeverDischarge com;
             show_open_subgoals())
        else error "repeated Goal not permitted in refining mode")
      | [] -> (fun () -> ()
   (* warning "An unadorned Goal command is now superfluous" *)
              ));;

add("ABORT",
    fun [VARG_IDENTIFIER id] -> (fun () -> abort_goal (string_of_id id))
      | [] -> (fun () -> abort_cur_goal ()));;

add("ABORTALL",
    fun [] -> (fun () -> abort_goals()));;

add("RESTART",
    fun [] -> (fun () -> (restart();show_open_subgoals())));;

add("PROOF",
    fun [VARG_COMMAND c] ->
        fun () -> (by (tacentries__vernac_give_exact_com c);
                   save_named true));;

add("SOLVE",
    fun [VARG_NUMBER n;VARG_TACTIC tcom] ->
    fun () -> (solve_nth n (tacentries__vernac_interp tcom);
               if not(is_silent()) then 
                       (* AC show_open_subgoals() *)
                       show_open_subgoals_focused() ));;
add("FOCUS",
     fun[] ->
         fun () -> make_focus 1);;

add("UNFOCUS",
     fun [] -> 
        fun () -> make_focus 0);;
          
add("BY",
    fun [VARG_TACTIC tcom] ->
    fun () -> begin
              by (tacentries__vernac_interp tcom);
              show_node()
              end);;



add("SaveNamed",
    fun [] -> fun () -> show_script() ; save_named true);;

add("DefinedNamed",
    fun [] -> fun () -> show_script() ; save_named false);;

add("SaveAnonymousThm",
    fun [VARG_IDENTIFIER id] -> fun () -> show_script() ; save_anonymous_thm true (string_of_id id));;
add("SaveAnonymousRmk",
    fun [VARG_IDENTIFIER id] -> fun () -> show_script() ; save_anonymous_remark true (string_of_id id));;

add("SaveState",
    fun [VARG_IDENTIFIER name;VARG_STRING desc] ->
        fun () -> save_state (string_of_id name) desc);;

add("TRANSPARENT",
    fun [VARG_IDENTIFIER id] -> fun () -> set_transparent (nametab__sp_of_id CCI id));;
add("OPAQUE",
    fun [VARG_IDENTIFIER id] -> fun () -> set_opaque (nametab__sp_of_id CCI id));;

add("WriteModule",
    fun [VARG_IDENTIFIER id] ->
    fun () ->
        (library__extern_module (string_of_id id)
        (* Do not want to deal any longer with module specif...*)
        (*;
         library__extern_module_specification (string_of_id id)*)))
;;

add("WriteModuleTo",
    fun [VARG_IDENTIFIER id;VARG_STRING filename] ->
    fun () ->
        (library__extern_module_to (string_of_id id) filename
        (* Do not want to deal any longer with module specif...*)
        (*;
         library__extern_module_specification_to (string_of_id id) filename*)))
;;

add("WriteStates",
    fun [arg] ->
        fun () -> 
        let s = (match arg with
                 VARG_IDENTIFIER id -> string_of_id id
               | VARG_STRING s -> s)
        in extern_state s);;

add("RemoveState",
    fun [VARG_IDENTIFIER id] ->
        fun () ->
          let str = (string_of_id id) in
          if str = "Initial" then error "Cannot forget Initial"
          else forget_state (is_silent()) str);;

add("UNDO",
    fun [VARG_NUMBER n] ->
        fun () -> (undo n;show_open_subgoals_focused()));;

add("SHOW",
    fun [] -> (fun () -> show_open_subgoals())
      | [VARG_NUMBER n] -> (fun () -> show_nth_open_subgoal n));;

add("ShowNode",
    fun [] -> fun () -> show_node());;


add("Go",
    fun [VARG_NUMBER n] ->
        (fun () -> (traverse n;show_node()))
      | [VARG_STRING"top"] ->
        (fun () -> (mutate top_of_tree;show_node()))
      | [VARG_STRING"next"] ->
        (fun () -> (mutate next_unproven;show_node()))
      | [VARG_STRING"prev"] ->
        (fun () -> (mutate prev_unproven;show_node())));;

add("ShowProof",
    fun [] ->
    fun () ->
    let pts = get_pftreestate () in
    let pf = proof_of_pftreestate pts and
        cursor = cursor_of_pftreestate pts in
    let (pfterm,meta_types) = refiner__extract_open_proof pf.goal.hyps pf in
        MSGNL [< 'S"LOC: " ; prlist_with_sep (fun () -> [< 'S" " >])
                                 (fun n -> [< 'INT n >])
                                 (rev cursor); 'FNL ;

                 'S"Subgoals" ; 'FNL ;
                 prlist
                 (fun (mv,ty) -> [< 'INT mv ; 'S" -> " ; pTERM ty ; 'FNL >])
                 meta_types;
                 'S"Proof: " ;
                 pTERM pfterm >]);;

add("CheckGuard",
    fun [] ->
    fun () ->
    let pts = get_pftreestate () in
    let pf = proof_of_pftreestate pts and
        cursor = cursor_of_pftreestate pts in
    let (pfterm,meta_types) = refiner__extract_open_proof pf.goal.hyps pf in
    let message = (try (typing__control_only_guard mt_evc pfterm; 
                        [<'S"The condition holds up to here">] )
                   with UserError(_,s) -> [<'S("Condition violated : ") ;s>])
    in MSGNL message)
;;


add("ShowTree",
    fun [] ->
    fun () -> show_prooftree());;

add("ShowScript",
    fun [] -> fun () -> show_script());;

add("ExplainProof",
    fun l ->
    let l = map (fun (VARG_NUMBER n) -> n) l
    in fun () ->
        let pts = get_pftreestate() in
        let evc = evc_of_pftreestate pts in
        let rec aux pts = function
            [] -> pts
          | (n::l) -> aux (tacmach__traverse n pts) l in 
        let pts = aux pts (l@[-1]) in
        let pf = proof_of_pftreestate pts
        in MSG(print_script true (rc_of_gc evc pf.goal) (initial_sign()) pf))
;;

add("ExplainProofTree",
    fun l ->
    let l = map (fun (VARG_NUMBER n) -> n) l
    in fun () ->
        let pts = get_pftreestate() in
        let evc = evc_of_pftreestate pts in
        let rec aux pts = function
            [] -> pts
          | (n::l) -> aux (tacmach__traverse n pts) l in 
        let pts = aux pts (l@[-1]) in
        let pf = proof_of_pftreestate pts
        in MSG(print_proof (rc_of_gc evc pf.goal) (initial_sign()) pf))
;;

add("ShowProofs",
    fun [] ->
        fun () ->
      let l = pfedit__list_proofs()
      in MSGNL (prlist_with_sep (fun () -> [< 'SPC >]) (fun s -> [< 'S s >]) l));;

add("PrintAll",
    fun [] -> fun () -> MSG(print_full_context_typ ()));;

add("PRINT",
    fun [] -> fun () -> MSG(print_local_context()));;

add("PrintId",
    fun [VARG_IDENTIFIER id] -> fun () -> MSG(print_name id));;

add("PrintOpaqueId",
    fun [VARG_IDENTIFIER id] -> fun () -> MSG(print_opaque_name id));;

add("PrintSec",
    fun [VARG_IDENTIFIER id] -> fun () -> MSG(print_sec_context_typ (string_of_id id)));;

add("PrintStates",
    fun [] -> 
        fun () ->
        MSG (prlist (fun (n,desc) -> [< 'ID n; 'S" : "; 'S desc ; 'FNL >])
                   (list_saved_states())));;

add("PrintModules",
    fun [] -> 
    fun () ->
        let opened = library__search_imports () and
            loaded = library__search_modules ()
        in 
        MSG [< 'S"Imported (open) Modules: " ;
              HOV 0 (prlist_with_sep (fun () -> [< 'FNL >])
                     (fun sp -> [< 'S(string_of_path sp) >]) opened) ; 
              'FNL ;
              'S"Loaded Modules: " ;
              HOV 0 (prlist_with_sep (fun () -> [< 'FNL >])
                     (fun (s,sp) -> [< 'S s ; 'S" = " ;
                                       'S(string_of_path sp) >]) loaded) ;
              'FNL >])
;;



(*=================
LaTeX output
Claude
*)

add("PrintLaTeXGrammarRules",
    fun [VARG_IDENTIFIER whatfor;VARG_IDENTIFIER gramname] ->
    fun () -> MSG(pureparse__LaTeX_print_nonterminal_definition (whatfor,gramname)));;

add("PrintLaTeXGrammar",
    fun [VARG_IDENTIFIER whatfor] ->
    fun () -> MSG(pureparse__LaTeX_print_grammar_definition (whatfor)));;

(*==============*)

add("PrintGrammar",
    fun [VARG_IDENTIFIER whatfor;VARG_IDENTIFIER gramname] ->
    fun () -> MSG(pureparse__print_nonterminal_definition (whatfor,gramname)));;

add("PrintCompiledGrammar",
    fun [VARG_IDENTIFIER whatfor;VARG_IDENTIFIER gramname] ->
    fun () -> MSG(pureparse__print_nonterminal_code (whatfor,gramname)));;

add("PrintPath",
    fun [] -> fun () -> print_loadpath ());;

add("BeginModule",
    fun [VARG_IDENTIFIER id] ->
    fun () ->
    library__open_module (string_of_id id));;

add("BeginSection",
    fun [VARG_IDENTIFIER id] ->
        fun () ->
        open_section (string_of_id id));;

add("BeginSilent",
    fun [] -> fun () -> make_silent(true));;
add("EndSilent",
    fun [] -> fun () -> make_silent(false));;

add("EndSection",
    fun [VARG_IDENTIFIER id] ->
        fun () ->
          if refining () then
             errorlabstrm "vernacentries : EndSection"
             [< 'S"proof editing in progress" ; (msg_proofs false) ;
                'S"Use \"Abort All\" first or complete proof(s)." >]
          else close_section_or_module (string_of_id id));;

add("StartProof",
    fun [VARG_STRING kind;VARG_IDENTIFIER s;VARG_COMMAND c] ->
        let stre = (match kind with
                 "THEOREM" -> NeverDischarge
               | "LEMMA" -> make_strength(safe_cdddr (library__cwd()))
               | "FACT" -> make_strength(safe_cdr (library__cwd()))
               | "REMARK" -> make_strength(library__cwd())
               | "DEFINITION" -> NeverDischarge
               | "LET" -> make_strength(safe_cddr (library__cwd()))
               | "LOCAL" -> make_strength(library__cwd()))
        in fun () ->
            begin
            start_proof (string_of_id s) stre c;
            show_open_subgoals()
            end)
;;

add("TheoremProof",
    fun [VARG_STRING kind;VARG_IDENTIFIER s;VARG_COMMAND c;
         VARG_VARGLIST coml] ->
        let calls = (map (fun (VCALL(na,l)) -> (na,l)) coml) in
        let (stre,opacity) = (match kind with
                 "THEOREM" -> (NeverDischarge,true)
               | "LEMMA" -> (NeverDischarge,true)
               | "FACT" -> (make_strength(safe_cdr (library__cwd())),true)
               | "REMARK" -> (make_strength(library__cwd()),true)
               | "DEFINITION" -> (NeverDischarge,false)
               | "LET" -> (make_strength(safe_cdr (library__cwd())),false)
               | "LOCAL" -> (make_strength(library__cwd()),false))
        in fun () ->
        try
        library__with_heavy_rollback
        (fun () ->
            (start_proof (string_of_id s) stre c;
             show_open_subgoals();
             do_listLR (vernacinterp__call NONE) calls;
             show_script();
             save_named opacity))
        ()
        with e ->
            if is_unsafe "proof" then
                (MSGNL [< 'S"Error: checking of theorem " ; print_id s ;
                          'SPC ; 'S"failed" ;
                          'S"... converting to Axiom" >];
                 del_proof (string_of_id s);
                 parameter_def_var (string_of_id s) c)
            else (errorlabstrm "vernacentries__TheoremProof"
                     [< 'S"checking of theorem " ; print_id s ; 'SPC ;
                        'S"failed... aborting" >] ))
;;

add("SUSPEND",
    fun [] -> fun () -> set_proof NONE);;
add("RESUME",
    fun [VARG_IDENTIFIER id] ->
                      (fun () -> set_proof (SOME(string_of_id id)))
      | [] -> (fun () -> resume_last ()))
;;

add("DEFINITION",
    fun [VARG_STRING kind;VARG_IDENTIFIER id;VARG_COMMAND c] ->
        let stre = (match kind with
                 "DEFINITION" -> NeverDischarge
               | "LET" -> make_strength(safe_cddr (library__cwd()))
               | "LOCAL" -> make_strength(library__cwd()))
        in fun () -> (definition_body id stre c;
                      if not(is_silent()) then
                          message ((string_of_id id) ^ " is defined")));;

add("VARIABLE",
    fun [VARG_STRING kind;VARG_BINDER (b,c)] ->
        let stre = make_strength(library__cwd ())
(*** inutile ? - CP 
(match kind with
                    "HYPOTHESIS" -> make_strength(library__cwd ()) 
                  | "VARIABLE" -> make_strength(library__cwd ()))
***)
        in fun () -> do_list (fun s -> (hypothesis_def_var (string_of_id s) stre c;
                                        if not(is_silent()) then
                                            message ((string_of_id s) ^ " is assumed"))) b);;

add("PARAMETER",
    fun [VARG_STRING kind;VARG_BINDER (b,c)] ->
    fun () -> do_list (fun s -> (parameter_def_var (string_of_id s) c;
                                 if not(is_silent()) then
                                     message ((string_of_id s) ^ " is assumed"))) b);;

add("CHECK",
    fun [VARG_STRING kind;VARG_COMMAND c] ->
        let f = (match kind with
                 "EVAL" -> MSG o print_evaluate
               | "COMPUTE" -> MSG o print_compute
               | "SIMPLIFY" -> MSG o print_simplify
               | "CHECK" -> MSG o print_val
               | "PRINTTYPE"  -> MSG o print_type)
        in fun () -> f(fconstruct mt_evc (initial_sign()) c))
;;

add("PrintExtractId",
    fun [VARG_IDENTIFIER id] -> fun () -> MSG(print_extracted_name id));;
add("PrintExtract",
    fun [] -> fun () -> MSG(print_extraction ()));;

add("INFO",
    fun [VARG_IDENTIFIER id] -> fun () ->
        error "Doesn't exist anymore; use Print instead");;
add("SEARCH",
    fun [VARG_IDENTIFIER id] -> fun () -> print_crible id);;
add("INSPECT",
    fun [VARG_NUMBER n] -> fun () -> MSG(inspect n));;

add("SETUNDO",
    fun [VARG_NUMBER n] -> fun () -> set_undo n);;

add("UNSETUNDO",
    fun [] -> fun () -> unset_undo ());;

add("SETHYPSLIMIT",
    fun [VARG_NUMBER n] -> fun () -> set_print_hyps_limit n);;

add("UNSETHYPSLIMIT",
    fun [] -> fun () -> unset_print_hyps_limit ());;

add("ONEINDUCTIVE",
    fun [VARG_STRING f;
         VARG_IDENTIFIER s;
         VARG_COMMAND c;
         VARG_BINDERLIST binders;
         VARG_BINDERLIST constructors] ->
        let la = join_binders binders in
        let isfinite = (fun "Inductive"   -> true 
                           | "CoInductive" -> false) in 
        let li = list__map (fun ([id],c) -> (id,c)) constructors
        in fun () -> build_mutual la [(isfinite f,s,c,li)])
;;

add("MUTUALINDUCTIVE",
    fun [VARG_BINDERLIST binders; VARG_VARGLIST indl; VARG_VARGLIST coindl] ->
        let la = join_binders binders in
        let lnamearconstructs = 
         (map (fun (VARG_VARGLIST 
              [VARG_IDENTIFIER arid; VARG_COMMAND arc; VARG_BINDERLIST lconstr])
            -> (true,arid,arc,list__map (fun ([id],c) -> (id,c)) lconstr)) indl)
         @
         (map (fun (VARG_VARGLIST 
              [VARG_IDENTIFIER arid; VARG_COMMAND arc; VARG_BINDERLIST lconstr])
            -> (false,arid,arc,list__map (fun ([id],c) -> (id,c)) lconstr)) coindl)
    in fun () -> build_mutual la lnamearconstructs);;

add("MUTUALRECURSIVE",
    fun [VARG_VARGLIST lmi] -> 
      let lnameargsardef = map
        (fun (VARG_VARGLIST 
  [VARG_IDENTIFIER fid;  VARG_BINDERLIST farg; 
   VARG_COMMAND arf; VARG_COMMAND ardef])
            -> (fid,join_binders farg,arf,ardef)) lmi
    in fun () -> build_recursive lnameargsardef);;

add("MUTUALCORECURSIVE",
    fun [VARG_VARGLIST lmi] -> 
      let lnameardef = map
        (fun (VARG_VARGLIST 
  [VARG_IDENTIFIER fid; 
   VARG_COMMAND arf; VARG_COMMAND ardef])
            -> (fid,arf,ardef)) lmi
    in fun () -> build_corecursive lnameardef);;

add("INDUCTIONSCHEME",
    fun [VARG_VARGLIST lmi] -> 
      let lnamedepindsort = map
        (fun (VARG_VARGLIST 
  [VARG_IDENTIFIER fid;  VARG_STRING depstr; VARG_COMMAND ind; 
       VARG_COMMAND sort]) ->
       let dep = (match depstr with "DEP" -> true | "NODEP" -> false)
       in (fid,dep,ind,sort)) lmi
    in fun () -> build_scheme lnamedepindsort);;


add("SyntaxMacro",
    fun [VARG_IDENTIFIER id;VARG_COMMAND com] ->
        (fun () -> syntax_definition id com;
                   if not(is_silent()) then
                    message ((string_of_id id) ^ " is now a syntax macro"))
      | [VARG_IDENTIFIER id;VARG_COMMAND com; VARG_NUMBER n] ->
        (fun () ->
             let rec aux t = function
                 0 -> t
               | n -> aux (OP(("APPLIST",[]),[t;OP(("XTRA",[TOKstr"ISEVAR"]),[])])) (n-1)
             in syntax_definition id (aux com n);
                if not(is_silent()) then
                    message ((string_of_id id) ^ " is now a syntax macro"))
);;


add("ABSTRACTION",
    fun ((VARG_IDENTIFIER id)::(VARG_COMMAND com)::l) ->
        (fun () ->
             let arity = vect_of_list (map (fun (VARG_NUMBER n) -> n) l)
             in abstraction_definition id arity com;
                 if not(is_silent()) then
                     message ((string_of_id id)^" is now an abstraction")));;

add("RequireImport",
    fun [VARG_IDENTIFIER id] ->
    fun () ->
        library__require_module (string_of_id id) true);;

add("RequireExport",
    fun [VARG_IDENTIFIER id] ->
    fun () ->
        library__require_module (string_of_id id) false);;

add("RequireImportFrom",
    fun [VARG_IDENTIFIER id;VARG_STRING filename] ->
    fun () ->
        library__require_module_from (string_of_id id) filename true);;

add("RequireExportFrom",
    fun [VARG_IDENTIFIER id;VARG_STRING filename] ->
    fun () ->
        library__require_module_from (string_of_id id) filename false);;

(* We do not want to deal with module specification any longer...
for the moment*)
(*add("RequireImportSpec",
    fun [VARG_IDENTIFIER id] ->
    fun () ->
        library__require_module_specification (string_of_id id) true);;

add("RequireExportSpec",
    fun [VARG_IDENTIFIER id] ->
    fun () ->
        library__require_module_specification (string_of_id id) false);;

add("RequireImportSpecFrom",
    fun [VARG_IDENTIFIER id;VARG_STRING filename] ->
    fun () ->
        library__require_module_specification_from (string_of_id id) filename true);;

add("RequireExportSpecFrom",
    fun [VARG_IDENTIFIER id;VARG_STRING filename] ->
    fun () ->
        library__require_module_specification_from (string_of_id id) filename false);;
*)

add("SYNTAX",
    fun ((VARG_IDENTIFIER whatfor)::(VARG_IDENTIFIER s)::
         (VARG_SPAT gt)::(VARG_UNPARSING unp)::l) ->
        let l = list__map (fun (VARG_SPAT_BINDER(id,smc)) -> (id,smc)) l in
        fun () -> let gt = astterm__spat_adjust_consts (GLOB (initial_sign())) gt in
                  let se = {syntax__syn_id = string_of_id s;
                            syntax__syn_universe = whatfor;
                            syntax__syn_astpat = gt;
                            syntax__syn_recbindings = l;
                            syntax__syn_unparsing = unp}
                  in add_anonymous_object (inPPSyntax(se)))
;;

add("GRAMMAR",
    fun [VARG_GRAMMAR_ENTRY ge] ->
    fun () -> add_anonymous_object(pureparse__inGrammar [ge]));;

add("TOKEN",
    fun [VARG_STRING s] ->
      	(fun () -> add_anonymous_object (pureparse__inToken s))
      |	_ -> anomalylabstrm "vernacentries__addTOKEN"
      	       	       	    [< 'S"Bad argument in TOKEN">])
;;

(* infix_grammar_entry : int -> string -> string -> grammar_entry
 *    takes precedence, infix op and prefix op and returns
 *    the corresponding Grammar entry
 *)
let infix_grammar_entry n inf pref =
  let commandn = "command" ^ (string_of_int n) in
  let str = commandn ^ ":=" ^ " [ command" ^ (string_of_int (n-1)) 
            ^ "($c1) \"" ^ inf ^ "\" " ^ commandn ^ "($c2)] -> [ $0 = <<("
            ^ pref ^ " $c1 $c2)>> ]" in 
  let ts = token_stream vernac_keywords (stream_of_string str) in
  grammar_parser (id_of_string "command") ts
;;

(* infix_syntax_entry : int -> string -> string -> syntax_entry
 *    takes precedence, infix op and prefix op and returns
 *    the corresponding Syntax entry
 *)
let infix_syntax_entry n inf pref =
  let str = "<<(" ^ pref ^ " $t1 $t2)>> " ^ (string_of_int n) ^
      	    " [<hov 0> <$t1:\"term\":L> [<hov 0> \" " ^ inf ^ 
      	    " \" <$t2:\"term\":E>]]" in
  let ts = token_stream vernac_keywords (stream_of_string str) in
  match ts with
    [< parsing__spat_gram (FRAGprim(TOKdyn d)) ;
       unparsings__unparsing unp >] -> 
          (match pridyn__tag d
      	   with "SPAT" ->
       	       	 let spat = outSPAT d in
       	       	 let gt = astterm__spat_adjust_consts 
      	       	       	       	 (GLOB (termenv__initial_sign())) spat in
      	       	 {syntax__syn_id = pref;
               	  syntax__syn_universe = id_of_string "constr";
               	  syntax__syn_astpat = gt;
               	  syntax__syn_recbindings = [];
               	  syntax__syn_unparsing = unp}
       	       | _ -> anomalylabstrm "vernacentries__infix"
      	       	       	       	     [< 'S"Malformed syntax entry." >])
  | [< >] -> anomalylabstrm "vernacentries__infix"
      	      	       	    [< 'S"Malformed syntax entry." >]
;;

add("INFIX",
    fun [VARG_NUMBER n; VARG_STRING inf; VARG_IDENTIFIER pref] ->
    (fun () -> 
       (* check the precedence *)
       if n<6 or n>9 then
       	 errorlabstrm "vernacentries__infix_grammar_entry"
      	      	      [< 'S"Precedence must be 6,7,8 or 9." >];

       (* check that the infix operator is correctly lexed *)
       (try token_stream vernac_keywords (stream_of_string inf)
        with Lexer_error _ | Parse_error -> errorlabstrm "vernacentries__infix"
      	       	       	       	       	  [< 'S"Malformed infix operator." >]);

       let pr = string_of_id pref in

       (* register the grammar entry *)
       let ge = infix_grammar_entry n inf pr in
       add_anonymous_object (pureparse__inGrammar [ge]);

       (* register the syntax entry *)
       let se = infix_syntax_entry n inf pr in
       add_anonymous_object (inPPSyntax se) ))
;;

(* $Id: vernacentries.ml,v 1.14 1996/02/12 15:39:35 cpaulin Exp $ *)
