(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                             vernacinterp.ml                              *)
(****************************************************************************)

(* Concrete syntax of the mathematical vernacular MV V2.6 *)

#infix "o";;
#open "std";;
#open "names";;
#open "more_util";;
#open "pp";;
#open "stdpp";;
#open "himsg";;
#open "ast";;
#open "proof_trees";;

let drop () = raise Drop;;(* Back to CAML *)

let vernac_tab = (mhm__new 17 : (string, vernac_arg list -> unit -> unit) mhm__t);;

let vinterp_add (s,f) =
    try mhm__add vernac_tab (s,f)
    with Failure _ ->
        errorlabstrm "vinterp_add"
        [< 'S"Cannot add the vernac command " ; 'S s ; 'S" twice" >]
;;

let overwriting_vinterp_add (s,f) =
    if mhm__in_dom vernac_tab s then
        (mhm__rmv vernac_tab s;
         warning ("Overwriting definition of toplevel command "^s));
    mhm__add vernac_tab (s,f)
;;

let vinterp_map s =
    try mhm__map vernac_tab s
    with Not_found -> errorlabstrm "Vernac Interpreter"
                      [< 'S"Cannot find command " ; 'S s >];;
let vinterp_init () = mhm__empty vernac_tab;;

let rec cvt_varg = function
    ASTnode("VERNACARGLIST",l) -> VARG_VARGLIST (map cvt_varg l)
  | ASTnode("VERNACCALL",(ASTprim(TOKstr na))::l) ->
    VCALL (na,map cvt_varg l)

  | ASTnode("VERNACCALL",(ASTprim(TOKid na))::l) ->
    VCALL (string_of_id na,map cvt_varg l)

  | ASTnvar id -> VARG_IDENTIFIER id
  | ASTprim(TOKstr s) -> VARG_STRING s
  | ASTprim(TOKnum n) -> VARG_NUMBER n
  | ASTnode("NONE",[]) -> VARG_NONE
  | ASTnode("COMMAND",[c]) -> VARG_COMMAND c
  | ASTnode("COMMANDLIST",l) -> VARG_COMMANDLIST l
  | ASTnode("TACTIC",[c]) -> VARG_TACTIC c
  | ASTnode("BINDER",c::idl) ->
    VARG_BINDER(list__map (fun (ASTnvar id) -> id) idl,c)
  | ASTnode("BINDERLIST",l) ->
    VARG_BINDERLIST(list__map ((fun (VARG_BINDER x) -> x) o cvt_varg) l)
  | ASTnode("NUMBERLIST",ln) -> VARG_NUMBERLIST (list__map (fun (ASTprim(TOKnum n)) -> n) ln) 

  | ASTprim(TOKdyn u) ->
    (match pridyn__tag u with
     "SPAT" -> VARG_SPAT(outSPAT u)
   | "UNPARSING" -> VARG_UNPARSING (unparsings__outUNPARSING u)
   | "GRAMMAR_ENTRY" -> VARG_GRAMMAR_ENTRY (pureparse__outGRAMMAR_ENTRY u)
   | _ -> errorlabstrm "vernacinterp__cvt_varg"
        [< 'S"Parser produced a mis-tagged token" >])

  | ASTnode("SPAT_BINDER",[ASTnvar id;ASTprim(TOKdyn c)]) ->
    if pridyn__tag c = "SPAT" then
        VARG_SPAT_BINDER(id,outSPAT c)
    else errorlabstrm "vernacinterp__cvt_varg"
        [< 'S"Parser produced a mis-tagged token" >]
  | ASTnode("ANY",[ast]) -> VARG_ANY ast

  | ASTnode("TACTIC_ARG",[targ]) -> VARG_TACTIC_ARG (tacinterp__cvt_arg targ)
  | _ -> failwith "cvt_varg"
;;

let call cmd (opn,converted_args) =
  let loc = ref "Looking up command" in
    try
      let callback = vinterp_map opn in
       loc:= "Checking arguments";
       let hunk = callback converted_args in
         loc:= "Executing command";
         hunk()

    with
        Drop -> raise Drop
      | e    -> (if !errors__debug then
                    MSGNL [< 'S"Vernac Interpreter " ; 'S !loc >];
                 MSG [< errors__explain_user_exn e ;
                  (match cmd with
                   SOME cmd ->  [< 'S"during command"; 'SPC ;
                                            pVERNAC cmd >]
                 | NONE -> [< 'S"during command" ; 'SPC ; 'S opn ; 'FNL >]) >];
           raise e)
;;

let rec interprec = function
    ASTnode("VERNACLIST",l) ->
    (let tstart = system__timestamp()
     in (do_listLR interprec l;
         library__resynchronize_caches();
         MSGNL [< 'S"Finished transaction in " ;
                 system__fmt_time_difference tstart (system__timestamp()) >]))
  | ASTnode(opn,argl) as cmd ->
    let converted_args =
        try list__map cvt_varg argl
        with e ->
            (MSGNL [< 'S"Vernac Interpreter " ; 'S "Converting arguments" ;
                      'SPC ; errors__explain_user_exn e ;
                      'S"during command"; 'SPC ;  pVERNAC cmd >];
                      raise e)
    in call (SOME cmd) (opn,converted_args)

   | cmd -> errorlabstrm "Vernac Interpreter" 
                         [< 'S"Malformed input : " ; print_ast cmd >]
;;

let interp c =
try
    library__with_heavy_rollback
        (fun () -> (interprec c;
                    library__resynchronize_caches())) ()
with e -> (if !errors__debug then
               MSGNL [< 'S"The last toplevel command" ;
                        'S" caused an error and has been rolled back." >];
           errors__raise_if_debug e)
;;


(* $Id: vernacinterp.ml,v 1.3 1996/02/12 15:39:39 cpaulin Exp $ *)
