(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                 goml.ml                                  *)
(****************************************************************************)

#open "std";;
#open "himsg";;
#open "initial";;
#open "states";;
#open "system";;
#open "pp";;
#open "stdpp";;
#open "coqtoplevel";;
#open "lexer";;
#open "sys";;
#open "version";;
#open "arg_parse";;

prerr_string "Welcome to Coq ";;
prerr_endline version;;

(*
 * Load of rcfile.
 * rcfile is either $HOME/.coqrc.5.10, or $HOME/.coqrc if the first one
 * does not exist.
 *)

let rcfile = ref ((getenv "HOME")^ "/.coqrc");;
let rcfile_specified = ref false;;
let set_rcfile s = rcfile := s; rcfile_specified := true;;
let set_rcuser s = rcfile := "~"^s^"/.coqrc";;

let load_rc = ref true;;
let no_load_rc () = load_rc := false;;

let initialized = ref false;;
let initialize() =
    if not !initialized then
        begin
        initialized := true;
        if !load_rc then
            try (if !rcfile_specified then
                   (if file_readable_p !rcfile then 
                      load_vernacular !rcfile
                    else failwith "caught")
                 else
                   (if file_readable_p ((!rcfile)^".5.10") then
                      load_vernacular ((!rcfile)^".5.10")
                    else if file_readable_p !rcfile then
                      load_vernacular !rcfile
                    else failwith "caught") )
            with _ -> message ("Load of "^(!rcfile)^" failed")
        else message "Skipping rcfile loading"
        end
;;

let intern_restore_state s =
    intern_state s;
    pfedit__restore_last_saved_state ()
;;

let add_include s =
    add_path s
;;

let set_hash_cons () = reintern0__use_hash_cons := true;;

let batch_mode = ref false;;
let set_batch_mode () = batch_mode := true;;

let set_debug () = errors__debug := true;;

let outputstate = ref "";;
let set_outputstate s = outputstate:=s;;
let outputstate () = if !outputstate <> "" then extern_state !outputstate;;

let inputstate = ref "";;
let set_inputstate s = inputstate:= s;;
let inputstate () =if !inputstate <> "" then intern_restore_state !inputstate;;

let load_vernacular_list = ref ([]:string list);;
let add_load_vernacular s =
  load_vernacular_list := s::(!load_vernacular_list);;
let load_vernacular () =
  do_list (fun s -> coqtoplevel__load_vernacular s)
      	  (rev !load_vernacular_list)
;;

let load_vernacular_obj = ref ([]:string list);;
let add_vernac_obj s = load_vernacular_obj := s::(!load_vernacular_obj);;
let load_vernac_obj () = 
  do_list
    (fun s -> library__read_module_from (filename__basename s) s)
    (rev !load_vernacular_obj)
;;

let require_list = ref ([]:string list);;
let add_require s = require_list := s::(!require_list);;
let require () =
  do_list
    (fun s -> library__require_module_from (filename__basename s) s false)
    (rev !require_list)
;;

parsing__init_grammar();;

let start () =
  inputstate ();
  load_vernac_obj ();
  require ();
  initialize();
  load_vernacular ();
  outputstate()
;;

(*
 * Parsing of the command line.
 * Options are :
 *    -I rep             add directory rep in the include path
 *    -include rep         "     "      "      "      "    "
 *    -q                 skip loading of rcfile
 *    -batch             batch mode: exits just after arguments parsing (coqc)
 *    -preload f         add f to the files to load during load_iobase()
 *    -outputstate f     write state in file f.coq
 *    -inputstate f      read state from file f.coq
 *    -load-vernac-source f
 *                       load Coq file f.v (Load f.)
 *    -load-vernac-object f
 *                       load Coq object file f.vo
 *    -require f         load Coq object file f.vo and import it (Require f.)
 *    -unsafe s          add s in the UNSAFE list; if "proof" is given then
 *                       incorrect proofs are converted to axioms
 *    -hash-cons         hash consing
 *    -debug             switch on the debug flag
 *    -init-file f       give the rcfile f
 *    -user u            give the user u
 *    -notactics         cannot load ML object files
 *)

let usage () =
  prerr_endline ("Usage: coql");
  prerr_endline 
  "    [-I directory] [-include directory] [-theories directory] [-is state]";
  prerr_endline
  "    [-nois] [-notactics] [-init-file filename] [-q] [-user name]";
  prerr_endline
  "    [-load-vernac-source filename] [-load-vernac-object filename]";
  prerr_endline
  "    [-require filename] [-batch] [-debug]";
  prerr_endline
  "    [-hash-cons] [-image filename] [-O module-set] [-open module-set]";
  if !batch_mode then prerr_endline "    filename"
;;

try

parse
 ["-I", String add_include;
  "-include", String add_include;
  "-q", Unit no_load_rc;
  "-batch",Unit set_batch_mode;
  "-outputstate",String set_outputstate;
  "-inputstate",String set_inputstate;
  "-load-vernac-source",String add_load_vernacular;
  "-load-vernac-object",String add_vernac_obj;
  "-require",String add_require;
  "-unsafe",String add_unsafe;
  "-hash-cons",Unit set_hash_cons;
  "-debug",Unit set_debug;
  "-init-file",String set_rcfile;
  "-user",String set_rcuser;
  "-notactics",Unit (fun () -> mltop__l_o := mltop__no_load_object;
                               mltop__incl := mltop__no_include)
 ]
 (fun s ->  raise (Bad ("don't know what to do with " ^ s)))
 (sys__getenv "COQ_ARGUMENT")

with Bad s -> (MSGNL [< 'S"Exception raised during Coq argument parsing:";
                        'SPC; 'S s; 'FNL >]; 
               usage();
               sys__exit 1)
   | UserError(_,s) as e -> 
              (try end_of_stream s; sys__exit 1
               with Parse_failure -> MSGNL [< errors__explain_sys_exn e >] ; 
                                     sys__exit 1)
   | e ->     (MSGNL [< errors__explain_sys_exn e >] ; sys__exit 1)
;;

(* $Id: goml.ml,v 1.6 1996/02/12 15:44:16 cpaulin Exp $ *)
