(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                  ast.ml                                  *)
(****************************************************************************)

#infix "o";;
#open "std";;
#open "names";;
#open "more_util";;
#open "pp";;
#open "stdpp";;
#open "lexer";;

let injection = function
    TOKnum _ -> "INT"
  | TOKstr _ -> "STRING"
  | TOKid _ -> "IDENT"
  | TOKpath _ -> "PATH"
  | TOKdyn d -> "DYN:"^(pridyn__tag d)
;;

let parse_tok = function
    [< 'Tint n >] -> TOKnum n
  | [< 'Tident id >] -> TOKid(id_of_string id)
  | [< 'Tstring s >] -> TOKstr s
  | [< qualid sp >] -> TOKpath sp
;;

let print_tok = function
    TOKnum n -> [< 'INT n >]
  | TOKid id -> print_id id
  | TOKstr s -> [< 'QS s >]
  | TOKpath sp -> [< 'S(string_of_path sp) >]
  | TOKdyn d -> [< 'S"#<dyn " ; pridyn__pr_dyn d ; 'S">" >]
;;

let parse_idoption = function
    [< 'Tlessgreater >] -> NONE
  | [< 'Tident id >] -> SOME(id_of_string id)
;;

let rec parse_ast = function
    [< 'Tident id >] -> ASTnvar(id_of_string id)
  | [< 'Tlbrace ; parse_tok tok ; 'Trbrace >] -> ASTprim tok
  | [< 'Tlparen ; 'Tident nname ; (plist parse_ast) l ; 'Trparen >] ->
    ASTnode(nname,l)
  | [< 'Tlbracket ; parse_idoption ido ; 'Trbracket ; parse_ast b >] ->
    ASTslam(ido,b)
  | [< parse_tok tok >] -> ASTprim tok
;;

let rec print_ast = function
    ASTprim tok -> [< 'S"{" ; print_tok tok ; 'S"}" >]
  | ASTnode(nna,l) -> [< 'S"(" ; 'S nna ; 'SPC ;
                         prlist_with_sep (fun () -> [< 'SPC >])
                         print_ast l ; 'S")" >]

  | ASTnvar id -> print_id id
  | ASTslam(ido,ast) -> [< 'S"[";
                           (match ido with
                            NONE -> [< 'S"<>" >]
                          | SOME id -> print_id id) ;
                           'S"]" ; print_ast ast >]
;;

let print_frag = function
    FRAGast ast -> [< 'S"#AST " ; print_ast ast >]
  | FRAGastlist l -> [< 'S"#ASTLIST " ;
                        prlist_with_sep (fun () -> [< 'SPC >])
                        print_ast l >]
  | FRAGprim p -> [< 'S"#PRIM " ; print_tok p >]
;;

(* dest_slam_as_ast will destruct a SLAM of type bound[alpha],
   rendering two ASTs - one which will be of type option[prim[ident]],
   the other of type alpha. *)

let mkNODE nna astl = ASTnode(nna,astl);;

let dest_slam_as_ast = function
    ASTslam(NONE,b) -> (mkNODE "NONE" [],b)
  | ASTslam(SOME id,b) -> (mkNODE "SOME" [ASTprim(TOKid id)],b);;

let mk_slam_as_ast = function
    (ASTnode("SOME",[ASTprim(TOKid id)]),b) -> ASTslam(SOME id,b)
  | (ASTnode("NONE",[]),b) -> ASTslam(NONE,b)
;;

(* 3. Binding management primitives *)

let isNVAR = function
    (ASTnvar _) -> true
  | _ -> false
;;
let mkNVAR id = ASTnvar id;;

let rec drop = function
    [] -> []
  | (NONE::t) -> drop t
  | ((SOME id)::t) -> id::(drop t)
;;

let lifted_iso = function
    (SOME _,SOME _) -> true
  | (NONE,NONE) -> true
  | _ -> false
;;

(* alpha-conversion *)

let alpha_eq(t1,t2) = aux([],t1,[],t2)
    where rec aux = function

    (bl1,ASTnode(op1,tl1),bl2,ASTnode(op2,tl2)) ->
    (op1 = op2) & auxlists(bl1,tl1,bl2,tl2)

  | (bl1,ASTnvar id1,bl2,ASTnvar id2) ->
    if (pure_index(id1,bl1) = -1) then (id1=id2)
    else pure_index(id1,bl1) = pure_index(id2,bl2)

  | (bl1,ASTslam(id1,body1),bl2,ASTslam(id2,body2)) ->
    lifted_iso(id1,id2)
    & aux(revc(drop [id1],bl1),body1,
          revc(drop [id2],bl2),body2)

  | (_,ASTprim p1,_,ASTprim p2) -> p1=p2

  | _ -> false

and auxlists = function
    (bl1,h1::t1,bl2,h2::t2) ->
    aux(bl1,h1,bl2,h2) & auxlists(bl1,t1,bl2,t2)
  | (_,[],_,[]) -> true
  | (_,_,_,_) -> false
;;

let fold curriedf l base = list_it (fun a -> fun b -> curriedf(a,b)) l base
;;

(* free variables in ast t *)
(* free_vars : ast -> identifier listset__t *)

let free_vars t = aux(t,[]) where rec aux = function
  (ASTnvar id,accum) -> listset__add accum id
| (ASTnode(_,tl),accum) -> fold aux tl accum
| (ASTslam(ido,body),accum) ->
        (listset__diff (aux(body,[])) (drop [ido]))@accum
| (ASTprim _,accum) -> accum
;;

(* Matching and Substitution *)

let rec smatch M N = srec [] M N

and srec sigma = fun
    SPATnil (FRAGastlist[]) -> sigma

  | (SPATcons(s1,s2)) (FRAGastlist(h::t)) -> 
    srec (srec sigma s1 (FRAGast h)) s2 (FRAGastlist t)

  | (SPATappend _) _ -> failwith "smatch"

  | (SPATlist sl) (FRAGastlist al) ->
    if length sl = length al then
        it_list2 srec sigma sl (map (fun a -> (FRAGast a)) al)
    else failwith "smatch"

  | (SPATprim(pv,injopt)) (FRAGast(ASTprim tok)) ->
    if (match injopt with
        NONE -> true
      | SOME inj -> inj = injection tok) then
        (pv,FRAGprim tok)::sigma
    else failwith "smatch"

  | (SPATprim(pv,injopt)) (FRAGprim tok) ->
    if (match injopt with
        NONE -> true
      | SOME inj -> inj = injection tok) then
        (pv,FRAGprim tok)::sigma
    else failwith "smatch"

  | (SPATnode(snname,sl)) (FRAGast(ASTnode(nname,l))) ->
    if snname = nname & length sl = length l then
        it_list2 srec sigma sl (map (fun a -> FRAGast a) l)
    else failwith "smatch"

  | (SPATbreaknode(nnapat,lpat)) (FRAGast(ASTnode(nname,l))) ->
    srec (srec sigma nnapat (FRAGprim(TOKid (id_of_string nname))))
         lpat (FRAGastlist l)

  | (SPATslam(pvopt,spb)) (FRAGast(ASTslam(idopt,b))) ->
    (match (pvopt,idopt) with
     (NONE,NONE) -> srec sigma spb (FRAGast b)
   | (SOME pv,SOME id) -> srec ((pv,FRAGprim (TOKid id))::sigma) spb (FRAGast b)
   | _ -> failwith "smatch")

  | (SPATslamlist(spidl,spb)) (FRAGast(ASTslam _ as ast)) ->
    let rec aux acc = function
        ASTslam _ as ast ->
        let (idast,bast) = dest_slam_as_ast ast
        in aux (idast::acc) bast
      | bast -> (rev acc,bast) in
    let (idastl,b) = aux [] ast
    in srec (srec sigma spidl (FRAGastlist idastl)) spb (FRAGast b)

  | (SPATnvar pv) (FRAGast(ASTnvar id)) ->
    (pv,FRAGprim(TOKid id))::sigma

  | (SPATquoteslam(id,spb)) (FRAGast(ASTslam(SOME id',b))) ->
     if id = id' then
         srec sigma spb (FRAGast b)
     else failwith "smatch"

  | (SPATquote ast) (FRAGast ast') ->
    if alpha_eq(ast,ast') then sigma else failwith "smatch"

  | (SPATquote (ASTprim p)) (FRAGprim p') ->
    if p = p' then sigma else failwith "smatch"

  | (SPATmeta pv) (FRAGast ast) ->
    if listmap__in_dom sigma pv then
        (match listmap__map sigma pv with
         FRAGast ast' ->
         if alpha_eq(ast,ast') then
             sigma
         else failwith "smatch"
       | _ -> failwith "smatch")
    else (pv,FRAGast ast)::sigma

  | (SPATmeta pv) f ->
    if listmap__in_dom sigma pv then
        if (listmap__map sigma pv) = f then sigma
        else failwith "smatch"
    else (pv,f)::sigma

  | _ _ -> failwith "smatch"
;;

let ssub sigma = srec
    where rec srec = function
    SPATmeta pv -> listmap__map sigma pv

  | SPATnvar pv ->
    (match listmap__map sigma pv with
     (FRAGast(ASTprim (TOKid id))) -> FRAGast(ASTnvar id)
   | (FRAGprim(TOKid id)) -> FRAGast(ASTnvar id)
   | _ -> failwith "ssub")

  | SPATnil -> FRAGastlist []

  | SPATcons(s1,s2) ->
    (match (srec s1,srec s2) with
     (FRAGast h,FRAGastlist l) -> FRAGastlist(h::l)
   | (FRAGprim p,FRAGastlist l) -> FRAGastlist((ASTprim p)::l)
   | _ -> failwith "ssub")

  | SPATappend l ->
    FRAGastlist
    (join(map
          (fun s ->
               match srec s with
               FRAGastlist l -> l
             | _ -> failwith "ssub")
     l))

  | SPATlist l ->
    FRAGastlist
    (map
     (fun sp ->
          match srec sp with
          FRAGast a -> a
        | FRAGprim p -> ASTprim p
        | _ -> failwith "ssub") l)

  | SPATprim(pv,ido) ->
    let p =
        (match listmap__map sigma pv with
         FRAGast(ASTprim p) -> p
       | FRAGprim p  -> p
       | _ -> failwith "ssub") in
    (match ido with
     NONE -> FRAGprim p
   | SOME inj ->
     if injection p = inj then FRAGprim p else failwith "ssub")

  | SPATnode(nna,spl) ->
    let astl =
        map (fun sp ->
                 match srec sp with
                 FRAGast a -> a
               | FRAGprim p -> ASTprim p
               | _ -> failwith "ssub") spl in
        FRAGast(ASTnode(nna,astl))

  | SPATbreaknode(spnna,lsp) ->
    let (nna,l) =
    (match (srec spnna,srec lsp) with
     (FRAGast(ASTprim (TOKid nna)),FRAGastlist l) -> (nna,l)
   | (FRAGprim (TOKid nna),FRAGastlist l) -> (nna,l)
   | _ -> failwith "ssub")
    in FRAGast(ASTnode(string_of_id nna,l))

  | SPATslam(NONE,sp) ->
    (match srec sp with
     FRAGast a -> FRAGast(ASTslam(NONE,a))
   | _ -> failwith "ssub")

  | SPATslam(SOME pv,spb) ->
    let id =
        (match listmap__map sigma pv with
             FRAGprim (TOKid id) -> id
           | _ -> failwith "ssub")
    in (match srec spb with
        FRAGast a -> FRAGast(ASTslam(SOME id,a))
      | _ -> failwith "ssub")

  | SPATslamlist(spidl,spb) ->
    (match (srec spidl,srec spb) with
     (FRAGastlist astidl,FRAGast b) ->
     let idoptl =
         map (fun (ASTnode("SOME",[ASTprim(TOKid id)])) -> (SOME id)
                | (ASTnode("NONE",[])) -> NONE
                | _ -> failwith "ssub") astidl
     in FRAGast(list_it (fun ido b -> ASTslam(ido,b)) idoptl b)
   | _ -> failwith "ssub")

  | (SPATquoteslam(id,spb)) ->
    (match srec spb with
     FRAGast b -> FRAGast(ASTslam(SOME id,b))
   | _ -> failwith "ssub")

  | (SPATquote (ASTprim p)) -> FRAGprim p
  | (SPATquote ast) -> FRAGast ast
;;

let isMeta id =
    let s = SA__un_intern (atompart_of_id id)
    in string_length s <> 0
       & ("$" = (sub_string s 0 1));;

let spat_of_ast = sprec
    where rec sprec = function
    ASTprim p ->
    (match p with
     TOKid id -> if isMeta id then SPATmeta id
                 else  SPATquote (ASTprim p)
   | _ ->  SPATquote (ASTprim p))

  | ASTnvar id ->
    if isMeta id then
        SPATmeta id
    else SPATquote (ASTnvar id)

  | ASTnode("$VAR",[ASTnvar id]) ->
    if isMeta id then
        SPATnvar id
    else failwith "spat_of_ast"

  | ASTnode("$NIL",[]) -> SPATnil

  | ASTnode("$CONS",[sp1;sp2]) ->
    SPATcons(sprec sp1,sprec sp2)

  | ASTnode("$APPEND",l) ->
    SPATappend(map sprec l)

  | ASTnode("$LIST",l) ->
    SPATlist(map sprec l)

  | ASTnode("$PRIM",
            [ASTnvar id;inj]) ->
    if isMeta id then
        (match inj with
         ASTnode("SOME",[ASTprim(TOKid inj)]) ->
         SPATprim(id,SOME (string_of_id inj))
       | ASTnode("NONE",[]) -> SPATprim(id,NONE)
       | _ -> failwith "spat_of_ast")
    else failwith "spat_of_ast"

  | ASTnode("$PRIM",[ASTnvar id]) ->
    if isMeta id then
        SPATprim(id,NONE)
    else failwith "spat_of_ast"

  | ASTnode("$OPER",[sp1;sp2]) ->
    SPATbreaknode(sprec sp1,sprec sp2)

  | ASTslam(NONE,spb) -> SPATslam(NONE,sprec spb)

  | ASTslam(SOME id,spb) ->
    if isMeta id then
        SPATslam(SOME id,sprec spb)
    else
        SPATquoteslam(id,sprec spb)

  | ASTnode("$SLAML",[spidl;b]) ->
    SPATslamlist(sprec spidl,sprec b)

  | ASTnode("$QUOTE",[ast]) -> SPATquote ast

  | ASTnode(nna,l) -> SPATnode(nna,map sprec l)

;;

let ast_of_spat = astrec
    where rec astrec = function
    SPATquote (ASTprim p) -> ASTprim p
  | SPATquote (ASTnvar id) -> ASTnvar id
  | SPATquote ast -> ASTnode("$QUOTE",[ast])
  | SPATmeta id -> ASTnvar id
  | SPATnvar id -> ASTnode("$VAR",[ASTnvar id])
  | SPATnil -> ASTnode("$NIL",[])
  | SPATcons(sp1,sp2) -> ASTnode("$CONS",[astrec sp1;astrec sp2])
  | SPATappend l -> ASTnode("$APPEND",map astrec l)
  | SPATlist l -> ASTnode("$LIST",map astrec l)
  | SPATprim(id,spinj) ->
    let inj = (match spinj with
               NONE -> ASTnode("NONE",[])
             | SOME inj -> ASTnode("SOME",[ASTprim(TOKid (id_of_string inj))]))
    in ASTnode("$PRIM",[ASTnvar id;inj])

  | SPATnode(nna,l) -> ASTnode(nna,map astrec l)

  | SPATbreaknode(spnna,sp) -> ASTnode("$OPER",[astrec spnna;astrec sp])

  | SPATslam(pvo,spb) -> ASTslam(pvo,astrec spb)

  | SPATquoteslam(id,spb) -> ASTslam(SOME id,astrec spb)

  | SPATslamlist(spidl,spb) -> ASTnode("$SLAML",[astrec spidl;astrec spb])
;;

let print_spat = print_ast o ast_of_spat;;
let parse_spat x = (spat_of_ast o parse_ast) x;;

let free_vars_of_spat = fvrec
    where rec fvrec = function
    SPATmeta _ -> []
  | SPATnvar _ -> []
  | SPATnil -> []
  | SPATcons(s1,s2) -> (fvrec s1)@(fvrec s2)
  | SPATappend l -> join(map fvrec l)
  | SPATlist l -> join(map fvrec l)
  | SPATprim _ -> []
  | SPATnode (_,l) -> join(map fvrec l)
  | SPATbreaknode (_,sp) -> fvrec sp
  | SPATslamlist (_,b) -> fvrec b
  | SPATslam (_,b) -> fvrec b
  | SPATquoteslam (id,b) -> listset__diff (fvrec b) [id]
  | SPATquote ast -> free_vars ast
;;

let raise_id id howmuch =
let (sa,n) = repr_ident id
in make_ident (sa) (n+howmuch)
;;

let raise_ast_bound_indices startwith howmuch T =
    raise startwith T where rec raise = function env -> function
    (ASTnvar id) -> (try ASTnvar(assoc id env) with _ -> (ASTnvar id))
  | (ASTnode(opn,tl)) -> ASTnode(opn,map (raise env) tl)
  | (ASTslam(id,body) as t) ->
    let new_id = option_app (fun x -> raise_id x howmuch) id
    in ASTslam(new_id,raise ((combine(drop [id],drop [new_id]))@env) body)
  | (ASTprim _ as ast) -> ast
;;

let raise_spat_bound_indices T =
let fv = free_vars_of_spat T in
let howmuch = it_list (fun n id -> max n (index_of_id id)) (initial__newMETA()) fv
in raise [] T where rec raise env T = match T with
    SPATmeta _ -> T
  | SPATnvar _ -> T
  | SPATnil -> T
  | SPATcons(sp1,sp2) -> SPATcons(raise env sp1,raise env sp2)
  | SPATappend l -> SPATappend(map (raise env) l)
  | SPATlist l -> SPATlist (map (raise env) l)
  | SPATprim _ -> T
  | SPATnode(nna,l) -> SPATnode(nna,map (raise env) l)
  | SPATbreaknode(spnna,sp) -> SPATbreaknode(spnna,raise env sp)

  | SPATslamlist(sp1,sp2) -> SPATslamlist(sp1,raise env sp2)
  | SPATslam(pvo,sp) -> SPATslam(pvo,raise env sp)
  | SPATquoteslam(id,sp) ->
    let new_id = raise_id id howmuch in
        SPATquoteslam(new_id,raise ((id,new_id)::env) sp)
  | SPATquote ast -> SPATquote(raise_ast_bound_indices env howmuch ast)
;;


(* reinterning *)

#open "reintern0";;
let ri_tok = function
    TOKnum _ as x -> x
  | TOKid id -> TOKid(ri_id id)
  | TOKstr _ as x -> x
  | TOKpath sp -> TOKpath(ri_sp sp)
  | TOKdyn d -> TOKdyn(pridyn__ri_dyn d)
;;

let rec ri_ast = function
    ASTprim tok -> ASTprim(ri_tok tok)
  | ASTnode(nna,l) -> ASTnode(nna,map ri_ast l)
  | ASTnvar id -> ASTnvar (ri_id id)
  | ASTslam(ido,ast) -> ASTslam(option_app ri_id ido,ri_ast ast)
;;

let ri_frag = function
    FRAGast ast -> FRAGast(ri_ast ast)
  | FRAGastlist l -> FRAGastlist (map ri_ast l)
  | FRAGprim tok -> FRAGprim(ri_tok tok)
;;

let rec ri_spat = function
    SPATmeta pv -> SPATmeta(ri_id pv)
  | SPATnvar pv -> SPATnvar (ri_id pv)
  | SPATnil -> SPATnil
  | SPATcons (sp1,sp2) -> SPATcons(ri_spat sp1,ri_spat sp2)
  | SPATappend l -> SPATappend(map ri_spat l)
  | SPATlist l -> SPATlist(map ri_spat l)
  | SPATprim (pv,inj) -> SPATprim(ri_id pv,inj)
  | SPATnode (nna,l) -> SPATnode(nna,map ri_spat l)
  | SPATbreaknode (sp1,sp2) -> SPATbreaknode(ri_spat sp1,ri_spat sp2)

  | SPATslamlist  (sp1,sp2) -> SPATslamlist(ri_spat sp1,ri_spat sp2)
  | SPATslam (pvo,sp) -> SPATslam(option_app ri_id pvo,ri_spat sp)
  | SPATquoteslam (id,sp) -> SPATquoteslam(ri_id id,ri_spat sp)
  | SPATquote ast -> SPATquote (ri_ast ast)
;;

let rec ri_tyexp = function
    TYprim _ as x -> x
  | TYlist e -> TYlist (ri_tyexp e)
  | TYoption e -> TYoption(ri_tyexp e)
  | TYcon _ as x -> x
  | TYbound e -> TYbound(ri_tyexp e)
  | TYboundlist e -> TYboundlist(ri_tyexp e)

;;
let pr_spat spat = [< 'S"<spat>" >];;

let ((inSPAT:spat->pridyn__t),
     (outSPAT:pridyn__t->spat)) =
    pridyn__new ("SPAT",ri_spat,pr_spat)
;;

let OP ((na,pl),tl) = ASTnode(na,(map (fun p -> ASTprim p) pl)@tl);;
let SLAM(ido,ast) = ASTslam(ido,ast);;

let gk_ord (wf1,id1) (wf2,id2) =
    let hi_bit = id_ord wf1 wf2
    in if hi_bit = 0 then id_ord id1 id2
       else hi_bit
;;

let fv_num = function
    LITERAL _ -> 0
  | PRIMITIVE _ -> 1
  | EPSILON -> 2
;;

let fv_ord fv1 fv2 =
    if fv_num fv1 = fv_num fv2 then
        match (fv1,fv2) with
        (LITERAL (s1,_),LITERAL (s2,_)) -> compare_strings s1 s2
      | (PRIMITIVE gk1,PRIMITIVE gk2) -> gk_ord gk1 gk2
      | (EPSILON,EPSILON) -> 0
    else (fv_num fv1) - (fv_num fv2)
;;

(* $Id: ast.ml,v 1.2 1996/02/12 15:44:46 cpaulin Exp $ *)
