(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                 fvset.ml                                 *)
(****************************************************************************)
(* Sets over ordered types *)

#open "eq";;
#open "int";;
#open "baltree";;

(* Sets are represented by AVL trees. *)

type t == ast__first_val baltree__t;;

let order = ast__fv_ord;;

let empty = Empty;;

let is_empty s =
  match s with Empty -> true | _ -> false;;

let mem x s =
  baltree__contains (order x) s;;

let add x s =
  baltree__add (order x) x s;;

let remove x s =
  baltree__remove (order x) s;;

let union s1 s2 =
  let rec union = fun
    Empty t2 -> t2
  | t1 Empty -> t1
  | (Node(l1, v1, r1, _)) t2 ->
      let (l2, _, r2) = baltree__split (order v1) t2 in
      baltree__join (union l1 l2) v1 (union r1 r2) in
  union s1 s2;;

let inter s1 s2 =
  let rec inter = fun
    Empty t2 -> Empty
  | t1 Empty -> Empty
  | (Node(l1, v1, r1, _)) t2 ->
      match baltree__split (order v1) t2 with
        (l2, Nothing, r2) ->
          baltree__concat (inter l1 l2) (inter r1 r2)
      | (l2, Something _, r2) ->
          baltree__join (inter l1 l2) v1 (inter r1 r2) in
  inter s1 s2;;

let diff s1 s2 =
  let rec diff = fun
    Empty t2 -> Empty
  | t1 Empty -> t1
  | (Node(l1, v1, r1, _)) t2 ->
      match baltree__split (order v1) t2 with
        (l2, Nothing, r2) ->
          baltree__join (diff l1 l2) v1 (diff r1 r2)
      | (l2, Something _, r2) ->
          baltree__concat (diff l1 l2) (diff r1 r2) in
  diff s1 s2;;

let compare s1 s2 =
  baltree__compare order s1 s2;;

let equal s1 s2 =
  compare s1 s2 == 0;;

let iter f s =
  let rec iter = function
    Empty -> ()
  | Node(l, v, r, _) -> iter l; f v; iter r
  in iter s;;

let fold f s init =
  let rec fold accu = function
    Empty -> accu
  | Node(l, v, r, _) -> fold (f v (fold accu r)) l
  in fold init s;;

let elements s =
  let rec elements accu = function
    Empty -> accu
  | Node(l, v, r, _) -> elements (v :: elements accu r) l
  in elements [] s;;

(* $Id: fvset.ml,v 1.2 1996/02/12 15:44:59 cpaulin Exp $ *)
