(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                generic.ml                                *)
(****************************************************************************)

#infix "o";;
#open "std";;
#open "names";;
#open "vectops";;
#open "pp";;
#open "stdpp";;
#open "more_util";;

(********************************************************************)
(*           Generic syntax of terms with de Bruijn indexes         *)
(********************************************************************)

(* returns the list of free debruijn indices in a term *)

let free_rels M = uniquize(frec 1 [] M)
    where rec frec depth acc = function
    Rel n         -> if n >= depth then (n-depth+1)::acc else acc
  | DOPN(_,cl)    -> it_vect (frec depth) acc cl
  | DOPL(_,cl)    -> it_list (frec depth) acc cl
  | DOP2(_,c1,c2) -> frec depth (frec depth acc c1) c2
  | DOP1(_,c)     -> frec depth acc c
  | DLAM(_,c)     -> frec (depth+1) acc c
  | DLAMV(_,cl)   -> it_vect (frec (depth+1)) acc cl
  | VAR _         -> acc
  | DOP0 _        -> acc
;;

(* (closedn n M) raises FreeVar if a variable of height greater than n
   occurs in M, returns () otherwise *)

let closedn = let rec closed_rec n = function
    Rel(m)        -> if m>n then raise FreeVar
  | VAR _         -> ()
  | DOPN(_,cl)    -> do_vect (closed_rec n) cl
  | DOPL(_,cl)    -> do_list (closed_rec n) cl
  | DOP2(_,c1,c2) -> closed_rec n c1; closed_rec n c2
  | DOP1(_,c)     -> closed_rec n c
  | DLAM(_,c)     -> closed_rec (n+1) c
  | DLAMV(_,v)    -> do_vect (closed_rec (n+1)) v
  | _             -> ()
in closed_rec;;

(* (closed M) is true iff M is a (deBruijn) closed term *)

let closed term =
    try (closedn 0 term; true) with FreeVar -> false;;

(* (noccurn n M) returns true iff (Rel n) does NOT occur in term M  *)

let noccurn n term = let rec occur_rec n = function
    Rel(m)        -> if m=n then raise Occur
  | VAR _         -> ()
  | DOPN(_,cl)    -> do_vect (occur_rec n) cl
  | DOPL(_,cl)    -> do_list (occur_rec n) cl
  | DOP1(_,c)     -> occur_rec n c
  | DOP2(_,c1,c2) -> occur_rec n c1; occur_rec n c2
  | DLAM(_,c)     -> occur_rec (n+1) c
  | DLAMV(_,v)    -> do_vect (occur_rec (n+1)) v
  | _             -> ()
in try (occur_rec n term; true) with Occur -> false;;

(* (noccur_bet n m M) returns true iff (Rel p) does NOT occur in term M 
  for n <= p < n+m *)

let noccur_bet n m term = let rec occur_rec n = function
    Rel(p)        -> if n<=p & p<n+m then raise Occur
  | VAR _         -> ()
  | DOPN(_,cl)    -> do_vect (occur_rec n) cl
  | DOPL(_,cl)    -> do_list (occur_rec n) cl
  | DOP1(_,c)     -> occur_rec n c
  | DOP2(_,c1,c2) -> occur_rec n c1; occur_rec n c2
  | DLAM(_,c)     -> occur_rec (n+1) c
  | DLAMV(_,v)    -> do_vect (occur_rec (n+1)) v
  | _             -> ()
in try (occur_rec n term; true) with Occur -> false;;

(* Lifting the binding depth across k bindings *)

let liftn k n c = if k=0 then c
    else liftrec n c where rec
    liftrec n = function
     (Rel i as c)     -> if i<n then c else Rel(i+k)
   | VAR id           -> VAR id
   | DOPN(oper,cl)    -> DOPN(oper,map_vect (liftrec n) cl)
   | DOPL(oper,cl)    -> DOPL(oper,map (liftrec n) cl)
   | DOP1(oper,c)     -> DOP1(oper,liftrec n c)
   | DOP2(oper,c1,c2) -> DOP2(oper,liftrec n c1,liftrec n c2)
   | DLAM(na,c)       -> DLAM(na,liftrec (n+1) c)
   | DLAMV(na,v)      -> DLAMV(na,map_vect (liftrec (n+1)) v)
   | x                -> x;;

let lift k c = liftn k 1 c;;
let lift1 c = lift 1 c;;

(* (subst1 M c) substitutes M for Rel(1) in c 
   we generalise it to (substl [M1,...,Mn] c) which substitutes in parallel
   M1,...,Mn for respectively Rel(1),...,Rel(n) in c *)

(* 1st : general case *)

let lift_substituend depth = function
    CLOSEDTERM c -> c
  | OPENTERM c -> lift depth c
  | DANGLINGREF -> failwith "lift_substituend"
;;
let make_substituend c =
    if closed c then CLOSEDTERM c else OPENTERM c;;

let substn_many laml n =
  let nterms = length laml in substrec n
  where rec substrec depth = function 
     (Rel k as x)     -> if k<=depth then x
                         else if k-depth <= nterms then
                                 lift_substituend depth (nth laml (k-depth))
                         else Rel(k-nterms)
   | VAR id           -> VAR id
   | DOPN(oper,cl)    -> DOPN(oper,map_vect (substrec depth) cl)
   | DOPL(oper,cl)    -> DOPL(oper,map (substrec depth) cl)
   | DOP1(oper,c)     -> DOP1(oper,substrec depth c)
   | DOP2(oper,c1,c2) -> DOP2(oper,substrec depth c1,substrec depth c2)
   | DLAM(na,c)       -> DLAM(na,substrec (depth+1) c)
   | DLAMV(na,v)      -> DLAMV(na,map_vect (substrec (depth+1)) v)
   | x                -> x
;;

let substl laml = substn_many (map make_substituend laml) 0;;
let subst1 lam = substl [lam];;
let pop t = lift (-1) t;;

(* Various occurs checks *)

let occur_opern s = occur_rec where rec occur_rec = function
    DOPN(oper,cl) -> s=oper or (exists_vect occur_rec cl)
  | DOPL(oper,cl) -> s=oper or (exists occur_rec cl)
  | VAR _         -> false
  | DOP1(_,c)     -> occur_rec c 
  | DOP2(_,c1,c2) -> (occur_rec c1) or (occur_rec c2)
  | DLAM(_,c)     -> occur_rec c
  | DLAMV(_,v)    -> exists_vect occur_rec v
  | _             -> false;;

let occur_oper0 s = occur_rec where rec occur_rec = function
    DOPN(_,cl)    -> (exists_vect occur_rec cl)
  | DOPL(_,cl)    -> (exists occur_rec cl)
  | DOP0 oper     -> s=oper
  | VAR _         -> false
  | DOP1(_,c)     -> occur_rec c 
  | DOP2(_,c1,c2) -> (occur_rec c1) or (occur_rec c2)
  | DLAM(_,c)     -> occur_rec c
  | DLAMV(_,v)    -> exists_vect occur_rec v
  | _             -> false;;

let occur_var s = occur_rec where rec occur_rec = function
    DOPN(_,cl)    -> (exists_vect occur_rec cl)
  | DOPL(_,cl)    -> (exists occur_rec cl)
  | VAR id        -> s=id
  | DOP1(_,c)     -> occur_rec c 
  | DOP2(_,c1,c2) -> (occur_rec c1) or (occur_rec c2)
  | DLAM(_,c)     -> occur_rec c
  | DLAMV(_,v)    -> exists_vect occur_rec v
  | _             -> false;;

let occur_oper s = occur_rec where rec occur_rec = function
    DOPN(oper,cl)    -> s=oper or (exists_vect occur_rec cl)
  | DOPL(oper,cl)    -> s=oper or (exists occur_rec cl)
  | VAR _            -> false
  | DOP0 oper        -> s=oper
  | DOP1(oper,c)     -> s=oper or occur_rec c 
  | DOP2(oper,c1,c2) -> s=oper or (occur_rec c1) or (occur_rec c2)
  | DLAM(_,c)        -> occur_rec c
  | DLAMV(_,v)       -> exists_vect occur_rec v
  | _                -> false;;

(* TODO Unused ? *)
let process_opers_of_term P f l constr = (filtrec l constr)
  where rec filtrec acc = function
    (DOP0 oper)        -> if P oper then ((f oper)::acc) else acc
  | VAR _              -> acc
  | (DOP1(oper,c))     -> let newacc = filtrec acc c in
                          if P oper then ((f oper)::newacc) else newacc
  | (DOP2(oper,c1,c2)) -> let newacc = filtrec (filtrec acc c1) c2 in
                          if P oper then ((f oper)::newacc) else newacc
  | DOPN(oper,cl)      -> let newacc = (it_vect filtrec acc cl) in
                          if P oper then ((f oper)::newacc) else newacc
  | DOPL(oper,cl)      -> let newacc = (it_list filtrec acc cl) in
                          if P oper then ((f oper)::newacc) else newacc
  | DLAM(_,c)          -> filtrec acc c
  | DLAMV(_,v)         -> it_vect filtrec acc v
  | _                  -> acc
;;

(* Returns the list of global variables in a term *)

let global_varsl l constr = (filtrec l constr) where rec filtrec acc = function
    VAR id             -> id::acc
  | (DOP1(oper,c))     -> filtrec acc c
  | (DOP2(oper,c1,c2)) -> filtrec (filtrec acc c1) c2
  | DOPN(oper,cl)      -> it_vect filtrec acc cl
  | DOPL(oper,cl)      -> it_list filtrec acc cl
  | DLAM(_,c)          ->  filtrec acc c
  | DLAMV(_,v)         ->  it_vect filtrec acc v
  | _                  -> acc;;

let global_vars constr = global_varsl [] constr;;

(* alpha equality for generic terms : checks first if M and M' are equal,
   otherwise checks equality forgetting the name annotation of DLAM and DLAMV*)

let rec eq_term M M' =
    (M = M')
    or (match (M,M') with
        (DOP1(oper1,c1),DOP1(oper2,c2)) ->
        oper1 = oper2 & eq_term c1 c2
      | (DOP2(oper1,c1,t1),DOP2(oper2,c2,t2)) ->
        oper1 = oper2 & eq_term c1 c2 & eq_term t1 t2
      | (DOPN(oper1,cl1),DOPN(oper2,cl2)) ->
        oper1 = oper2 & for_all2eq_vect eq_term cl1 cl2
      | (DOPL(oper1,cl1),DOPL(oper2,cl2)) ->
        oper1 = oper2 & for_all2eq eq_term cl1 cl2
      | (DLAM(_,c1),DLAM(_,c2)) -> eq_term c1 c2
      | (DLAMV(_,cl1),DLAMV(_,cl2)) -> for_all2eq_vect eq_term cl1 cl2
      | (_,_) -> false)
;;

(* (dependent M N) is true iff M is eq_term with a subterm of N 
   M is appropriately lifted through abstractions of N *)
let dependent =
 let rec deprec M = function t ->
     (eq_term M t)
     or (match t with
     VAR _       -> false
   | DOP1(_,c)   -> deprec M c
   | DOP2(_,c,t) -> deprec M c or deprec M t
   | DOPN(_,cl)  -> exists_vect (deprec M) cl
   | DOPL(_,cl)  -> exists (deprec M) cl
   | DLAM(_,c)   -> deprec (lift1 M) c
   | DLAMV(_,v)  -> exists_vect (deprec (lift1 M)) v
   | _ -> false)
 in deprec;;

(* TODO Obsolete ? *)
let id_of_var = function
    VAR id -> id
  | _      -> invalid_arg "id_of_var"
;;

(* (thin_val sigma) removes identity substitutions from sigma *)

let rec thin_val = function
    [] -> []
  | (((id,CLOSEDTERM(VAR id')) as s)::tl) -> if id = id' then thin_val tl
                          else s::(thin_val tl)
  | h::tl -> h::(thin_val tl)
;;

(* (replace_vars sigma M) applies substitution sigma to term M *)
let replace_vars var_alist = 
let var_alist = thin_val var_alist
in if var_alist = [] then (function x -> x) else substrec 0
    where rec substrec n = function
     (VAR(x) as c)    ->
        (try lift_substituend n (assoc x var_alist)
         with Not_found -> c)

   | DOPN(oper,cl)    -> DOPN(oper,map_vect (substrec n) cl)
   | DOPL(oper,cl)    -> DOPL(oper,map (substrec n) cl)
   | DOP1(oper,c)     -> DOP1(oper,substrec n c)
   | DOP2(oper,c1,c2) -> DOP2(oper,substrec n c1,substrec n c2)
   | DLAM(na,c)       -> DLAM(na,substrec (n+1) c)
   | DLAMV(na,v)      -> DLAMV(na,map_vect (substrec (n+1)) v)
   | x                -> x;;

let subst_varn str n = replace_vars [(str,make_substituend (Rel n))];;
(* (subst_var str t) substitute (VAR str) by (Rel 1) in t *)
let subst_var str = subst_varn str 1;;


let id_of_name default = function
    Name s -> s
  | Anonymous -> default
;;

let dunbind default sign ty (DLAM(na,c)) =
let id = next_ident_away (id_of_name default na) (ids_of_sign sign)
in (add_sign (id,ty) sign,
    subst1 (VAR id) c)
;;

let dunbindv default sign ty (DLAMV(na,c)) =
let id = next_ident_away (id_of_name default na) (ids_of_sign sign)
in (add_sign (id,ty) sign,
    map_vect (subst1 (VAR id)) c)
;;

let dbind sign c =
let (id,ty) = hd_sign sign
and sign' = tl_sign sign
in (ty,DLAM(Name id,subst_var id c))
;;

let dbindv sign cl =
let (id,ty) = hd_sign sign
and sign' = tl_sign sign
in (ty,DLAMV(Name id,map_vect (subst_var id) cl))
;;

let SAPP = fun
    (DLAM(na,M)) N -> subst1 N M
  | _ _ -> invalid_arg "SAPP"
;;
let SAPPV = fun
    (DLAMV(na,MV)) N -> map_vect (subst1 N) MV
  | _ _ -> invalid_arg "SAPPV";;

let SAPPVi = fun
    i (DLAMV(na,MV)) N -> subst1 N MV.(i)
  | _ _ _ -> invalid_arg "SAPPVi"
;;

let SAPPList constr = srec [] constr where rec srec stack = fun
   (DLAM(_,c)) (h::t) -> srec (h::stack) c t
 |  c           []    -> substl stack c
 | _             _    -> failwith "SAPPList";;

let SAPPVList constr = srec [] constr where rec srec stack = fun
   (DLAM(_,c)) (h::t) -> srec (h::stack) c t
 | (DLAMV(_,c)) [h]   -> map_vect (substl (h::stack)) c
 | _             _    -> failwith "SAPPVList";;

let SAPPViList i = srec [] where rec srec stack = fun
   (DLAM(_,c)) (h::t) -> srec (h::stack) c t
 | (DLAMV(_,c)) [h]   -> substl (h::stack) c.(i)
 | _             _    -> failwith "SAPPViList";;

let under_dlams f = apprec where rec apprec = function 
   DLAMV(na,c) -> DLAMV(na,map_vect f c)
 | DLAM(na,c)  -> DLAM(na,apprec c)
 | _           -> failwith "under_dlams";;

let interp_modif absfun oper (oper',modif) = interprec modif
    where rec interprec = fun
    [] cl -> DOPN(oper',vect_of_list cl)
  | (ERASE::tlm) (c::cl) -> interprec tlm cl
  | (ABSTRACT::tlm) (c::cl) -> absfun (interprec tlm cl) c
;;

(* utility for discharge *)

let modify_opers replfun absfun oper_alist =
if oper_alist = [] then fun x -> x else substrec
    where rec substrec = function
    (DOPN(oper,cl) as c) ->
    let cl' = map_vect substrec cl in
    (try
    (match assoc oper oper_alist with
     NOT_OCCUR -> 
        anomalylabstrm "generic__modify_opers"
        [< 'S"An oper which was never supposed to appear has just appeared" ;
           'SPC ; 'S"Either this is in system code, and you need to" ; 'SPC;
           'S"report this error," ; 'SPC ;
           'S"Or you are using a user-written tactic which calls" ; 'SPC ;
           'S"generic__modify_opers, in which case the user-written code" ;
           'SPC ; 'S"is broken - this function is an internal system" ;
           'SPC ; 'S"for internal system use only" >]
   | DO_ABSTRACT (oper',modif) ->
     if length modif > vect_length cl then
         anomaly "found a reference with too few args"
     else interp_modif absfun oper (oper',modif) (list_of_vect cl')
   | DO_REPLACE -> substrec (replfun (DOPN(oper,cl'))))
    with Not_found -> DOPN(oper,cl'))

   | DOP1(i,c) -> DOP1(i,substrec c)
   | DOPL(oper,cl) -> DOPL(oper,map substrec cl)
   | DOP2(oper,c1,c2)  -> DOP2(oper,substrec c1,substrec c2)
   | DLAM(na,c) -> DLAM(na,substrec c)
   | DLAMV(na,v) -> DLAMV(na,map_vect substrec v)
   | x              -> x;;

let put_DLAMSV lna lc = 
match lna with  [] -> anomaly "put_DLAM"
             | na::lrest ->  
                 it_list (fun c na -> DLAM(na,c)) (DLAMV(na,lc)) lrest;;

let put_DLAMSV_subst lid lc = 
match lid with  [] -> anomaly "put_DLAM"
             | id::lrest ->  
    it_list (fun c id' -> DLAM(Name id',subst_var id' c)) 
             (DLAMV(Name id,map_vect (subst_var id) lc)) lrest;;


let rec decomp_DLAMV n = function 
       DLAM(_,lc)  -> decomp_DLAMV (n-1) lc
     | DLAMV(_,lc) -> if n = 1 then lc else error "decomp_DLAMV 0"
     | _           -> error "decomp_DLAMV 1";;

let decomp_DLAMV_name n = decomprec [] n where rec decomprec lna n = function 
       DLAM(na,lc)  -> decomprec (na::lna) (pred n) lc
     | DLAMV(na,lc) -> if n = 1 then (na::lna,lc) else error "decomp_DLAMV 0"
     | _           -> error "decomp_DLAMV 1";;

let decomp_all_DLAMV_name constr = decomprec [] constr
  where rec decomprec lna = function 
       DLAM(na,lc)  -> decomprec (na::lna) lc
     | DLAMV(na,lc) -> (na::lna,lc);;

(* [Rel (n+m);...;Rel(n+1)] *)

let rel_vect n m = tabulate_vect (fun i -> Rel(n+m-i)) (m);;
let rel_list n m = reln [] 1
where rec reln l p = if p>m then l else reln (Rel(n+p)::l) (p+1);;


(* $Id: generic.ml,v 1.3 1996/02/12 15:45:03 cpaulin Exp $ *)
