(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                 lexer.ml                                 *)
(****************************************************************************)

(*
 * Lexical analysis :
 *
 *    An identifier is a regular expression of the form
 *     
 *              ( $ _ a..z A..Z ) ( $ _ a..z A..Z ' 0..9 )*
 *
 *    The user can define new tokens with the command Token "<new token>".
 *    New tokens are regular expressions of the form
 *
 *              ( < > / \ - + ; , | ! @ # % ^ & ? * : ~ )
 *              ( < > / \ - + ; , | ! @ # % ^ & ? * : ~ $ _ a..z A..Z ' 0..9)*
 *
 *    The lexical convention is the *Mizar*'s one: Tokens are considered
 *    in parallel, and the longest match wins. For instance,
 *    if "~" and "~~" are tokens, then "~~~" is lexed as "~~" "~".
 *
 *    [ ] ( ) { } "" . are reserved characters.
 *)

#open "std";;
#open "pp";;
#open "stdpp";;
#open "more_util";;
#open "lomsg";;

let cnt = ref 0;;

(* | : := = > >> <> << < -> ; # * and , are predefined tokens. *)
let reserved_tok_list =
       [ "|",Tbar ; ":",Tcolon ; ":=",Tcolonequal ; "=",Tequal ;
      	 ">",Tgreater ; ">>",Tgreatergreater ; "<>",Tlessgreater ;
      	 "<<",Tlessless ; "<",Tless ; "->",Tminusgreater ; ";",Tsemi ;
	 "#",Tsharp ; "*",Tstar ; ",",Tcomma
       ]
;;

let (special_tokens_tab,rev_special_tokens_tab) =
  let t=(hashtbl__new 73)
  and rev_t = (hashtbl__new 73)
  in
   do_list (fun (s,tok) -> hashtbl__add t s tok ; hashtbl__add rev_t tok s)
       	   reserved_tok_list;
   t,rev_t
;;

let special_token_of_string s = hashtbl__find special_tokens_tab s
;;

(* !token_tree : the tree of tokens
 * node (SOME t,tab) indicates that t is a token (and corresponds to the
 * path from the root, and tab is a table associating subtrees to chars.
 *)
type ttree = TR of (string option)*((char,ttree) avm__t);;
type frozen = FT of (string option)*((char*frozen) list);;

let char_ord = (fun c1 c2 -> (int_of_char c1)-(int_of_char c2));;

let token_tree = ref (TR (NONE,avm__new char_ord))
;;

let freeze () = 
  let rec freeze_rec (TR(s,t)) =
    let l = avm__toList t in
    FT(s,map (fun (c,t') -> (c,freeze_rec t')) l)
  in freeze_rec !token_tree
;;

let unfreeze t = 
  let rec unfreeze_rec (FT(s,l)) =
    let t = avm__ofList char_ord (map (fun (c,f) -> (c,unfreeze_rec f)) l) in
    TR(s,t)
  in token_tree := unfreeze_rec t
;;

(* char_list_of_string : string -> char list *)
let char_list_of_string s =
  let l = (string_length s) -1
  in let rec acc_rec acc = fun
     0 -> (match (nth_char s 0) with
      	   `|` | `,` | `;` | `+` | `\\` | `/` | `<` | `>`
      	 | `=` | `!` | `@` | `#` | `%` | `^` | `&` | `*`
      	 | `:` | `?` | `~` | `-` as c -> (c::acc)
	 | c -> errorlabstrm "char_list_of_string"
      	       	       	     [<'S(string_for_read (string_of_char c)) ;
			       'S" : illegal character in tokens.">])

   | n -> (match (nth_char s n) with
      	   `|` | `,` | `;` | `+` | `\\` | `/` | `<` | `>`
      	 | `=` | `!` | `@` | `#` | `%` | `^` | `&` | `*`
      	 | `:` | `?` | `~` | `-` | `$` | `_` | `a`..`z`
      	 | `A`..`Z` | `'` | `0`..`9` as c -> acc_rec (c::acc) (n-1)
	 | c -> errorlabstrm "char_list_of_string"
      	       	       	     [<'S(string_for_read (string_of_char c)) ;
			       'S" : illegal character in tokens.">])
  in
  acc_rec [] l
;;

(* add_tok : char list -> string -> () 
 * add a token in the tree. 
 *)
let branch c t = try avm__map t c
                 with Not_found -> TR(NONE,avm__new char_ord)
;;

let add_tok chl token =
  let rec add_rec = fun 
      	 [] (TR (NONE,t)) -> (TR (SOME token,t))
       | [] ((TR (SOME s,_)) as tt) ->
	       (try special_token_of_string s;
		    errorlabstrm "lexer__add_tok"
                                 [< 'S"This token is a reserved keyword.">]
      	        with Not_found -> warning "This token already exists."; tt)
       | (c::r) (TR (a,t)) ->
               let t'= avm__remap t c (add_rec r (branch c t))
               in TR (a,t')
  in token_tree := add_rec chl !token_tree
;;

let add_token string =
  let l = char_list_of_string string
  in add_tok l string
;;

(* initialization : we insert the predefined tokens in the tree
 * and we add also ? @ :: / <-
 *)
let init_tokens () =
  do_list (fun (s,_) -> add_token s) reserved_tok_list;
  do_list (fun s -> add_token s) [ "?" ; "@" ; "::" ; "/" ; "<-"]
;;

init_tokens ();;

exception Tok_and_clist of string*(char list);;

(* the list of parsed tokens is returned by the mean of the exception
 * Tok_and_clist, and caught by token_stream.
 *)

(* process_chars : char stream -> 'a
 * parses a char list and retuns the corresponding token list (as strings)
 * raise Lexer_error if the argument does not correspond to a token list.
 *)
let process_chars cs = 
  let rec proc_rec nch (TR (_,tab) as tt) lg = function
     [< 'c ; l >] -> (try (match avm__map tab c with
       	       	       	      (TR(NONE,_)) as tt' -> proc_rec (c::nch) tt' lg l
			    | (TR(SOME s,_)) as tt' -> proc_rec [] tt' s l)
		      with Not_found ->
			      if lg="" then
			        raise (Lexer_error !cnt)
		              else
			        raise (Tok_and_clist (lg,rev (c::nch))))

   | [< >] ->  if lg="" then
      	       	 raise (Lexer_error !cnt)
       	       else
      	       	 raise (Tok_and_clist (lg,(rev nch)))
  in proc_rec [] !token_tree "" cs
;;

(* beginning of the "classical" lexer *)
let (add_in_buff, get_buff, fget_buff) =
  let buff = ref (create_string 80) in (
  (* add_in_buff *) (
    fun i x ->
      let len = string_length !buff in
      if i >= len then (buff := !buff ^ (create_string len); ());
      set_nth_char !buff i x;
      succ i
  ),
  (* get_buff *) (
    fun len ->
      sub_string !buff 0 len
  ),
  (* fget_buff *) (
    fun len ->
      set_nth_char !buff len `\000`;
      !buff
  )
)
;;

let rec spec_char = function
  [< '`n` >] -> `\n` 
| [< '`t` >] -> `\t`
| [< '`b` >] -> `\b` 
| [< '`r` >] -> `\r`
| [< '`0`..`9` as c; (spec1 (int_of_char c-int_of_char `0`)) v >] ->
    char_of_int v
| [< 'x >] -> x

and spec1 v = function
  [< '`0`..`9` as c; s >] -> spec1 (10*v+int_of_char c-int_of_char `0`) s
| [< >] -> v
;;

let rec sq_string len = function
  [< '`'` >] -> len
| [< '`\\`;
     (function [< '`\n` >] -> len
             | [< spec_char c >] -> add_in_buff len c) len;
     s >] -> sq_string len s
| [< '`\n` ; s >] -> let c = !cnt in
                     sq_string len s;
                     raise (Lexer_error c)
| [< 'x; s >] -> sq_string (add_in_buff len x) s
;;

let rec string len = function
  [< '`"` >] -> len 
| [< '`\\`;
     (function [< '`\n` >] -> len
           | [< spec_char c >] -> add_in_buff len c) len;
     s >] -> string len s
| [< 'x; s >] -> string (add_in_buff len x) s

and skip_comm = function
  [< '`*`; s >] -> skip1 s
| [< '`(`; skip2 _; s >] -> skip_comm s
| [< '`"`; (string 0) _; s >] -> skip_comm s
| [< '_; s >] -> skip_comm s

and skip1 = function
  [< '`)` >] -> ()
| [< s >] -> skip_comm s

and skip2 = function
  [< '`*`; s >] -> skip_comm s
| [< >] -> ()
;;

let rec number len = function
  [< '`0`..`9` as d; s >] -> number (add_in_buff len d) s
| [< >] -> Tint(int_of_string(fget_buff len))

and neg_number len = function
  [< '`0`..`9` as d; s >] -> neg_number (add_in_buff len d) s
| [< >] -> Tint(-(int_of_string(fget_buff len)))
;;

let char = function [< '`\\`; spec_char c >] -> c | [< 'x >] -> x
;;

(* next_token: takes the keyword table, a stream of chars and returns
 * the first token of that stream, except in the case of special characters
 * in which it raises the exception (Tok_and_clist (s,l)) where s is the
 * parsed token, and l is a list of chars to be put in the stream in order
 * to be re-read.
 *)
let next_token keyw = next_token_loop
where rec next_token_loop = function
  [< '`_` | `$` | `a`..`z` | `A`..`Z` as c; (ident keyw (add_in_buff 0 c)) i >]
     -> i
| [< '`|` | `,` | `;` | `+` | `\\` | `/` | `<` | `>` | `=` |
      `!` | `@` | `#` | `%` | `^` | `&` | `*` | `:` | `?` | `~` as c; s >] ->
      	 process_chars [< 'c ; s ; [<>] >]
      	(* warning : the [<>] above is needed to circumvent a bug of caml *)

| [< '`0`..`9` as d; (number (add_in_buff 0 d)) n >] -> n
| [< '`-`; s >] -> (match s with
      	       	   [< '`0`..`9` as d; (neg_number (add_in_buff 0 d)) n >] -> n
		   | [< >] -> process_chars [< '`-` ; s ; [<>] >])
      	   (* warning : the [<>] above is needed to circumvent a bug of caml *)

| [< '`"`; (string 0) len >] -> Tstring (get_buff len)
| [< '`'`; (fun strm -> get_buff (sq_string 0 strm)) sqs ;
   begin function
      [< '`0`..`9` as d; (number (add_in_buff 0 d)) (Tint n) >] 
         -> Tident("'"^sqs^"'"^(string_of_int n))
    | [< >] -> Tident("'"^sqs^"'")
   end x >] -> x

| [< '`(`; s >] -> lparen s
| [< '`)` >] -> Trparen
| [< '`[` ; (function [< '`{` >] -> Tlbracketlbrace 
                    | [< >] -> Tlbracket) t >] -> t
| [< '`]` >] -> Trbracket
| [< '`{` >] -> Tlbrace
| [< '`}` >] -> Trbrace

| [< '`.` >] -> Tdot
| [< '` ` | `\n` | `\t`; s >] -> next_token_loop s
| [< '_ >] -> raise (Lexer_error !cnt)

and lparen = function
    [< '`*`; skip_comm _; s >] -> next_token_loop s
  | [< '`:` >] -> Tlparencolon
  | [< >] -> Tlparen

and ident keyw = (ident_loop
  where rec ident_loop len = function
    [< '`$` | `a`..`z` | `A`..`Z` | `0`..`9` | `'` | `_` as c; s >] ->
      ident_loop (add_in_buff len c) s
  | [< s >] ->
      let str = get_buff len in
      try Tkw(hash_assoc str keyw) with _ -> Tident str)
;;

let rec reset_lexer = function
    [< '`\n` >] -> ()
  | [< '_; reset_lexer x >] -> x
  | [< >] -> ()
;;

(* the discarding function: reads the tokens until it reaches a dot. *)
let discard_dot keyw cs =
  let rec discrec cs =
      try match next_token keyw cs
          with Tdot -> ()
             | _    -> discrec cs
      with
         Lexer_error _      -> discrec cs
       | Tok_and_clist(_,s) -> discrec [< (stream_of_list s) ; cs >]
  in
  discrec cs
;;

(* token_stream: takes the keyword table, a stream of chars and returns
 * a stream of tokens.
 *)
let token_stream keyw cs =
  let rec tok_rec cs =
    try let tok = next_token keyw cs in [< 'tok ; (tok_rec cs) >]
    with
      	        Lexer_error c -> discard_dot keyw cs; raise (Lexer_error c)
              | Tok_and_clist(sp_id,l) ->
	         let tok = (try special_token_of_string sp_id
		            with Not_found -> Tspecial sp_id) in
		 let s = stream_of_list l in
		 [< 'tok ; (tok_rec [< s ; cs >]) >]
  in [< tok_rec cs >]
;;

let stream_of_string str =
  cnt := -1; str_rec 0
  where rec str_rec i =
    if i >= string_length str then [< >]
    else [< 'nth_char str i; str_rec(i+1) >]
;;

let stream_of_channel ic =
  cnt := 0;
  stream_from (fun _ -> cnt := !cnt+1; input_char ic)
;;

let fmt_lexer_error c =
    [< 'S "Lexical error: illegal character" ;
        (if c>=0 then [< 'S ", char " ; 'INT c >] else [< >]);
        'S "." ; 'FNL >]
;;

let print_lexer_error c =
  prerr_string "Lexical error: illegal character";
  if c >= 0 then (
    prerr_string ", char ";
    prerr_int c
  );
  prerr_endline ""
;;

let fmt_syntax_error () =
    [< 'S"Syntax error" ;
       (if !cnt >= 0 then [< 'S ", char " ; 'INT !cnt >] else [< >]);
       'FNL >]
;;

let print_syntax_error() =
  prerr_string "Syntax error";
  if !cnt >= 0 then (
    prerr_string ", char ";
    prerr_int !cnt
  );
  prerr_endline ""
;;

(* pretty-print of the tokens. *)
let string_of_token rev_keywords = function
    Tdot -> "."
  | Tident s -> s
  | Tint n -> string_of_int n
  | Tkw kw  -> hash_assoc kw rev_keywords
  | Tlbrace -> "{"

  | Tlbracket -> "["
  | Tlparen -> "("

  | Trbrace -> "}"
  | Trbracket -> "]"
  | Trparen -> ")"

  | Tstring s -> "\"" ^ (string_for_read s) ^ "\""
  | Tlparencolon -> "(:"
  | Tlbracketlbrace -> "[{"

  | Tspecial s -> s
  | tok -> hashtbl__find rev_special_tokens_tab tok
 ;;

(* with_tokens: applies the function f on the token stream corresponding
 * to a string s.
 *)
let with_tokens ktab f s =
    let ts = token_stream ktab (stream_of_string s) in
    f ts
;;

let rec fmt_token_stream rev_keywords = function
    [< 'tok ; s >] -> [< 'S(string_of_token rev_keywords tok) ; 'SPC ;
      	       	         (fmt_token_stream rev_keywords s) >]
|   [< >]          -> [< >]
;;

(* $Id: lexer.ml,v 1.5 1996/02/12 15:45:20 cpaulin Exp $ *)
