(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                  lib.ml                                  *)
(****************************************************************************)

#open "std";;
#open "names";;
#open "more_util";;
#open "pp";;
#open "stdpp";;

let lib_stk = ref ([] : (section_path * node) list);;
let lib_tab = (mhm__new 17 : (section_path,node) mhm__t);;
let path_prefix = ref ([] : string list);;

let init () =
    lib_stk := [];
    mhm__empty lib_tab;
    path_prefix := []
;;

let make_path kind str = names__make_path (PA__intern(!path_prefix)) str kind;;

let recalc_path_prefix () = (path_prefix := recalcrec !lib_stk)
  where rec recalcrec = function
    (sp,OpenDir _)::_ -> let (pl,id,_) = repr_path sp
                         in (string_of_id id)::(PA__un_intern pl)
  | _::t -> recalcrec t
  | []   -> []
;;

(* an object SP is in a closed directory when the path-part of the SP
 * is not a (possibly improper) prefix of the current path.
 *)

let add_entry (sp,obj) =
    if mhm__in_dom lib_tab sp then
        errorlabstrm "lib__add_entry"
        [< 'S"Cannot add the object " ; 'S(string_of_path sp) ; 'S": an";
      	   'SPC ; 'S"identically named object already exists" >];
    lib_stk := (sp,obj):: !lib_stk;
    mhm__add lib_tab (sp,obj);
    (sp,obj)
;;

let add_leaf_entry(spopt,obj) = add_entry(spopt,LEAF obj);;

let add ((id,pk),obj) =
  (match obj with
      ClosedDir _ | OpenDir _ -> anomaly "lib__add"
    | _ -> ());
  let sp = make_path pk id
  in add_entry(sp,obj)
;;

let add_leaf (id,obj) = add(id,LEAF obj);;

let split_lib sp = findrec [] !lib_stk
  where rec findrec acc = function
     ((sp',obj) as hd)::t ->
      	       	   if sp = sp' then (acc,hd,t) else findrec (hd::acc) t
   | [] -> errorlabstrm "lib__split_lib" [< 'S"no such entry" >]
;;

let split_lib_P P = findrec [] !lib_stk
  where rec findrec acc = function
     ((sp',obj) as hd)::t ->
             if P hd then (acc,hd,t) else findrec (hd::acc) t
   | [] -> errorlabstrm "lib__split_lib_P" [< 'S"no such entry" >]
;;

let discard_context = discrec
  where rec discrec = function
    (sp,ClosedDir (_,_,_,ctxt))::t -> (mhm__rmv lib_tab sp;
       	       	       	       	       discrec ctxt;
       	       	       	       	       discrec t)
  | (sp,_)::t -> (mhm__rmv lib_tab sp;
       	       	  discrec t)
  | [] -> ()
;;

let restore_context = restrec
  where rec restrec = function
    ((sp,(ClosedDir (_,_,_,ctxt) as obj)))::t -> (mhm__add lib_tab (sp,obj);
       	       	       	       	       	       	  restrec ctxt;
       	       	       	       	       	       	  restrec t)
  | (sp,obj)::t -> (mhm__add lib_tab (sp,obj);
       	       	    restrec t)
  | [] -> ()
;;

(* adds a module right after the latest module entry in the library. *)
let add_module (sp,node) =
  let (pref,hd,tail) = 
    split_lib_P (fun (_,ClosedDir(_,{module_p=true;_},_,_)) -> true
                   | _ -> false) in
  let newlib = rev_append pref ((sp,node)::hd::tail)
  in (lib_stk := newlib;
      restore_context [(sp,node)])
;;

let reset_to sp =
  let pl = dirpath sp in
  if not(prefix_of (rev pl) (rev !path_prefix)) then
    error "cannot reset to an object which is in a closed section";
  if not(mhm__in_dom lib_tab sp) then
    error "cannot reset to a nonexistent object";
  let (discard,ent,save) = split_lib sp
  in lib_stk := save;
     discard_context (ent::discard);
     recalc_path_prefix()
;;

let reset_to_P P =
  let (discard,ent,save) = 
    try split_lib_P P
    with UserError("lib__split_lib_P",_) -> errorlabstrm "lib__reset_to_P"
                    [< 'S"cannot reset to a nonexistent object" >]
  in let sp,n = ent 
  in let pl = dirpath sp
  in if not(prefix_of (rev pl) (rev !path_prefix)) then
       error "cannot reset to an object which is in a closed section";
     (match n with
        ClosedDir(_,{module_p=true},_,_) -> errorlabstrm "lib__reset_to_P"
                        [< 'S"cannot reset to a module" >]
      | Import _ -> anomaly "reset to an 'Import' entry"
      | _ -> lib_stk := save;
             discard_context (ent::discard);
             recalc_path_prefix() )
;;

let reset_keeping sp =
  let pl = dirpath sp in
  if not(prefix_of (rev pl) (rev !path_prefix)) then
    error "cannot reset keeping an object which is in a closed section";
  if not(mhm__in_dom lib_tab sp) then
    error "cannot reset keeping a nonexistent object";
  let (discard,itobj,save) = split_lib sp
  in lib_stk := itobj::save;
     discard_context discard;
     recalc_path_prefix()
;;

let reset_keeping_P P =
  let (discard,itobj,save) = split_lib_P P
  in lib_stk := itobj::save;
     discard_context discard;
     recalc_path_prefix()
;;

let map sp = mhm__map lib_tab sp;;

let map_leaf sp =
  match mhm__map lib_tab sp with
    LEAF x -> x
  | _      -> invalid_arg "lib__map_leaf"
;;

let leaf_object_tag sp =
  let lobj = map_leaf sp
  in libobject__object_tag lobj
;;

let find_entry_P P = findrec !lib_stk
  where rec findrec = function
    (sp,_ as ent)::tl -> (if try P ent with _ -> false then sp else findrec tl)
  | []                -> failwith "lib__find_entry: no such entry"
;;

let find_dir = function
    NONE   -> find_entry_P (fun (_,OpenDir _) -> true | _ -> false)
  | SOME s -> find_entry_P (fun (_,OpenDir(s',_)) -> s = s' | _ -> false)
;;

let open_dir s odc =
  let s_id = id_of_string s in
  let s_sp = make_path OBJ s_id in
  (if mhm__in_dom lib_tab s_sp then
        errorlabstrm "lib__open_dir"
        [< 'S"Section " ; 'S s ; 'S" already exists" >]);
  let rv = add_entry (s_sp,OpenDir(s,odc))
  in path_prefix := s:: !path_prefix; rv
;;

let close_dir sp cdc =
  if not(kind_of_path sp = OBJ) then
    invalid_arg "lib__close_dir";
  if not(mhm__in_dom lib_tab sp) then
    error "directory does not exist";
  match mhm__map lib_tab sp with
    (OpenDir (s,odc)) -> mhm__rmv lib_tab sp;
      	       	       	 let (ctxt,_,rest) = split_lib sp
      	       	       	 in lib_stk := rest;
       	       	       	 add_entry (sp,ClosedDir(s,odc,cdc,ctxt));
       	       	       	 path_prefix := tl !path_prefix
  | _ -> error "is not a directory"
;;

let contents_after = function
   SOME sp -> let (ctxt,_,_) = split_lib sp
              in rev ctxt
 | NONE    -> !lib_stk
;;

let app f = apprec !lib_stk
  where rec apprec = function
     []   -> ()
   | h::t -> apprec t; f h
;;

let freeze () = !lib_stk;;

let unfreeze l =
    lib_stk := l;
    mhm__empty lib_tab;
    restore_context l;
    recalc_path_prefix()
;;

let cwd () = !path_prefix;;

let node_name (sp,_) = basename sp;;

let anon_SA = SA__intern "_";;

let is_anonymous_sp sp = atompart_of_id (basename sp) = anon_SA;;

let next_anonymous_id () = nextrec (contents_after NONE)
  where rec nextrec = function
    [] 		      -> make_ident anon_SA 0
  | (sp,OpenDir _)::_ -> make_ident anon_SA 0
  | (sp,_)::tl 	      -> if is_anonymous_sp sp then
              	       	   make_ident anon_SA (1+(index_of_id(basename sp)))
      	       	       	 else nextrec tl
;;

let last_entry () = 
  match !lib_stk with
    []        -> NONE
  | (sp,_)::_ -> SOME sp
;;

(* $Id: lib.ml,v 1.3 1996/02/12 15:45:23 cpaulin Exp $ *)
