(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                               libobject.ml                               *)
(****************************************************************************)

(* libobject declares persistent objects given with two methods:
   - a caching function specifying how to import the object in the current
   scope of theory modules  
   - a reintern function specifying how to relocate atoms (SA,PA) when
   reading the object from disk *)

#open "std";;
#open "names";;
#open "more_util";;

type object == dyn__t;; (* persistent dynamic objects *)

type dynamic_object_declaration =
    {dyn_cache_function : (section_path * object -> unit);
     dyn_reintern_function : (object -> object);
     dyn_specification_function : object -> object}
;;

let object_tag lobj = dyn__tag lobj;;

let cache_tab = (mhm__new 17 : (string,dynamic_object_declaration) mhm__t);;

let declare_object(na,odecl) =
  let (infun,outfun) = dyn__new na in

  let cacher (spopt,lobj) =
    if dyn__tag lobj = na then odecl.cache_function(spopt,outfun lobj)
    else anomaly "somehow we got the wrong dynamic object in the cachefun"

  and reinterner lobj = infun(odecl.reintern_function(outfun lobj))

  and spec_extractor lobj = infun(odecl.specification_function (outfun lobj))

  in mhm__add cache_tab (na,{dyn_cache_function = cacher;
                            dyn_reintern_function = reinterner;
                            dyn_specification_function = spec_extractor});
     (infun,outfun)
;;

let cache_object (spopt,lobj) =
  let tag = object_tag lobj in
  try let dodecl = mhm__map cache_tab tag
      in dodecl.dyn_cache_function(spopt,lobj)
  with Not_found ->
      	    anomaly ("Cannot find cachefun for an object with tag "^tag)
;;

let reintern_object lobj =
  let tag = object_tag lobj in
  try let dodecl = mhm__map cache_tab tag
      in dodecl.dyn_reintern_function lobj
  with Not_found ->
      	    anomaly ("Cannot find reintern fun for an object with tag "^tag)
;;

let extract_object_specification lobj =
  let tag = object_tag lobj in
  try let dodecl = mhm__map cache_tab tag
      in dodecl.dyn_specification_function lobj
  with Not_found ->
    anomaly ("Cannot find specification extractor for an object with tag "^tag)
;;

let make_strength = function
    [] -> NeverDischarge
  | l  -> DischargeAt (sp_of_wd l)
;;

(* $Id: libobject.ml,v 1.3 1996/02/12 15:45:27 cpaulin Exp $ *)
