(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                library.ml                                *)
(****************************************************************************)

#open "std";;
#open "lomsg";;
#open "names";;
#open "pp";;
#open "stdpp";;
#open "more_util";;
#open "system";;
#open "summary";;
#open "libobject";;

let lookup_module s =
  lib__find_entry_P
    (fun (sp,lib__ClosedDir(s',{lib__module_p=true;_},_,_)) -> s = s'
       | _ -> false)
;;

let module_is_loaded sp =
  try match lib__map sp with
        lib__ClosedDir(_,{lib__module_p=true;_},_,_) -> true
      | _ -> false
  with _ -> false
;;

let module_is_known s =
    try lookup_module s ; true
    with Failure _ -> false
;;

let find_dir s =
    try lib__find_dir s
    with Failure _ -> raise unknown_dir
;;

let import_watchers = (thunk__new() : section_path thunk__t);;
let add_import_watcher (s,f) = thunk__add import_watchers (s,f);;

(* Initializing the environment *)
let reset_library () = lib__init(); init_caches();;

let cache_closeddir sp (s,odc,cdc) = ();;

let rec cache_import sp =
  match lib__map sp with
    lib__ClosedDir(_,odc,cdc,ctxt) ->
      if listset__memb !IMPORTED sp then
        warning ("Module "^(string_of_path sp)^" already imported")
      else if for_all module_is_loaded odc.lib__imports then
             (do_listLR (fun (sp',_ as d) ->
                       if listset__memb cdc.lib__exports sp' then
                           recache_decl d
                       else ()) ctxt;
             IMPORTED := sp:: !IMPORTED;
             thunk__run import_watchers sp)
           else error ("Some needed imports haven't been done for module "^
      	       	       (string_of_path sp))
  | _ -> errorlabstrm "library__cache_import"
      	       	      [< 'S"Cannot import a non-directory" >]

and recache_decl = function
    (_,lib__Import(s,isp,true)) -> if not(module_is_loaded isp) then
              	       	       	     error "Missing imports"
              	       	       	   else ()
  | (sp,lib__OpenDir(s,odc)) -> anomaly "library__recache_decl: OpenDir"

  | (sp,lib__LEAF x) -> cache_object(sp,x)

  | (_,lib__Import(s,isp,false)) -> cache_import isp

  | (sp,lib__ClosedDir(s,odc,cdc,ctxt)) -> cache_closeddir sp (s,odc,cdc)

and recache_ctxt l = do_listRL recache_decl l
;;

let cache_decl = function
    (sp,lib__OpenDir(s,odc)) -> freeze_to_key sp

  | (sp,lib__LEAF x) -> cache_object(sp,x)

  | (sp,lib__Import(s,isp,_)) -> cache_import isp

  | (sp,lib__ClosedDir(s,odc,cdc,ctxt)) -> cache_closeddir sp (s,odc,cdc)
;;

let cache_ctxt l = do_listRL cache_decl l;;

let rebuild_caches () = init_caches(); lib__app cache_decl;;


let init_caches = summary__init_caches;;

(* Protocol for unfreezing states and verifying that the tables are
 * in sync with the library.

 * When we are "out_of_sync", we need to unfreeze the latest valid
 * summary (summary__find_latest), and then re-cache all the entries
 * in the library which follow.  Then we can try to resynchronize, and
 * if this reports that there was something to resync, we save that
 * frozen state under the key of the last lib-entry.

 * If we aren't out of sync, then "sum_sp" contains the latest valid
   frozen summary.  "sp" contains the latest library-entry-name.  If
   the last thing unfrozen wasn't sum_sp, then we treat this as being
   out of sync.

   If the last thing unfrozen _was_ sum_sp, then we still need to
   check if the caches are in sync or not, and if they aren't we need
   to resynchronize them.

 *)

let last_unfrozen = ref (NONE : section_path option);;

let resync_caches out_of_sync = 
  match lib__last_entry() with
     NONE    -> ()
   | SOME sp ->
       try let sum_sp = summary__find_latest () in
           if (SOME sum_sp) <> !last_unfrozen or out_of_sync then begin
              summary__unfreeze_latest sum_sp;
              last_unfrozen := SOME sum_sp;
              cache_ctxt (lib__contents_after (SOME sum_sp)) end;

           if summary__resync_caches() or
                       [] = summary__list_frozen_summaries() then begin
              summary__freeze_to_key sp;
              last_unfrozen := SOME sp;
              MSGNL [< 'S"[Froze to " ; 'S(string_of_path sp) ; 'S"]" >] end

       with Failure _ ->
          if out_of_sync then begin init_caches(); rebuild_caches() end;
          if summary__resync_caches() or
              	       	[] = summary__list_frozen_summaries() then begin
             summary__freeze_to_key sp;
             last_unfrozen := SOME sp;
             MSGNL [< 'S"[Froze to " ; 'S(string_of_path sp) ; 'S"]" >] end
;;

let resynchronize_caches () = resync_caches false;;
let backtrack_caches () = resync_caches true;;

let force_freeze_caches () =
  resynchronize_caches();
  try if not(lib__last_entry () = SOME(summary__find_latest())) then
        failwith "caught"
  with Failure _ ->
        (match lib__last_entry() with
            NONE -> ()
          | SOME sp ->
             (summary__freeze_to_key sp;
              last_unfrozen := SOME sp;
              MSGNL [< 'S"[Froze to " ; 'S(string_of_path sp) ; 'S"]" >]))
;;

let open_section name =
  cache_decl (lib__open_dir name {lib__module_p = false;
                                  lib__imports = []})
;;

(* This function does not rebuild caches; hence, it assumes that if
   something goes wrong, the offending function understands how to
   rollback it's modifs.  Since this is restricted to the functions
   which manipulate the tables, it is a safe assumption.
 *)
let rollback f arg =
  let fs = lib__freeze()
  in try f arg
     with e -> (lib__unfreeze fs; raise e)
;;

let add_anonymous_object obj =
    let anon_id = lib__next_anonymous_id() in
    rollback (fun () -> cache_decl(lib__add_leaf((anon_id,OBJ),obj))) ()
;;

let add_named_object idp obj =
    rollback (fun () -> cache_decl(lib__add_leaf(idp,obj))) ()
;;

let cwd () = lib__cwd();;

(* can only open a module when the only top-level entries around are
   other (closed) modules, and import declarations
 *)
let open_module name =
  let ctxt = lib__contents_after NONE in
  if for_all (fun (_,lib__ClosedDir _) -> true
                | (_,lib__Import _) -> true
                | _ -> false) ctxt then
    cache_decl (lib__open_dir name {lib__module_p = true;
                                    lib__imports = !IMPORTED})
  else errorlabstrm "open_module"
       [< 'S"You cannot open a module" ; 'SPC ;
          'S"when there are things other than" ; 'SPC ;
          'S"Modules and Imports in the context." >]
;;

let rec specification_of_module (lib__ClosedDir(s,odc,cdc,ctxt)) =
  if not odc.lib__module_p then
    error "Only modules can be converted to specifications";
  let ctxt' = map_succeed
    (fun (sp,lib__LEAF obj) -> 
      	       	       	(sp,lib__LEAF (extract_object_specification obj))
       | (_,lib__OpenDir _) -> 
      	       	       	anomaly "specification_of_module: found an OpenDir"
       | (sp,(lib__ClosedDir(_,odc',_,_) as cdir)) ->
               	       	if odc'.lib__module_p then
               	       	  (sp,specification_of_module cdir)
               	       	else failwith "caught"
       | (_,lib__Import _ as x) -> x) ctxt
  in lib__ClosedDir(s,odc,cdc,ctxt')
;;

let import_module s =
  let anon_id = lib__next_anonymous_id() in
  let isp = try lookup_module s
            with Failure _ -> errorlabstrm "import_module"
                                           [< 'S s ; 'S " not loaded." >] in
  if not(listset__memb !IMPORTED isp) then
    cache_decl(lib__add((anon_id,OBJ),lib__Import(s,isp,true)))
  else
    warning (s^" already imported")
;;

let export_module s =
  let anon_id = lib__next_anonymous_id() in
  let isp = lookup_module s in
  if not(listset__memb !IMPORTED isp) then
    cache_decl(lib__add((anon_id,OBJ),lib__Import(s,isp,false)))
  else
    warning (s^" already imported")
;;

(* can only close a module when:
 * (1) all the imported names are currently loaded modules
 *)
let close_module sp exports =
  match lib__map sp with
    lib__OpenDir(s,odc) ->
      if for_all (fun (_,lib__Import(_,isp,_)) ->
                    (try (match lib__map isp with
                            lib__ClosedDir(_,odc,_,_) -> odc.lib__module_p
                          | _ -> false)
                     with Not_found -> false)
                 | _ -> true) (lib__contents_after (SOME sp)) then
        (lib__close_dir sp {lib__exports = exports; lib__export_fn = NONE})
      else 
      	failwith "Cannot close a module right now"
  | _ -> errorlabstrm "close_module"
                  [< 'S"Cannot ``close'' an object which is not a module" >]
;;

let needed_modules = function
    lib__ClosedDir(_,odc,_,ctxt) ->
      let imports = map_succeed (fun (_,lib__Import(_,isp,_)) -> isp
                                   | _ -> failwith "caught") ctxt
      in uniquize(odc.lib__imports@imports)
  | _ -> failwith "library__needed_modules"
;;

let (extern_module,
     extern_module_to,
     (*extern_module_specification,*)
     (*extern_module_specification_to,*)
     intern_module,
     intern_module_from(*,
     intern_module_specification,
     intern_module_specification_from*)) =

let (raw_extern_module, raw_intern_module) =
      system__extern_intern(1001,".vo")

(***
and (raw_extern_module_specification, raw_intern_module_specification) =
      system__extern_intern(1002,".vi")
***)

in

let rec extern_module s = extern_module_to s s

and extern_module_to s filename =
    let sp = lookup_module s in
    match lib__map sp with
      lib__ClosedDir _ as d ->
            raw_extern_module filename ((SA__freeze(),PA__freeze()),(sp,d))
    | _ -> errorlabstrm "library__extern_module_to"
              	       	[< 'S"Can only extern a module-object" >]

(***
and extern_module_specification s = extern_module_specification_to s s
***)

(***
and extern_module_specification_to s filename =
    let sp = lookup_module s in
    match lib__map sp with
      lib__ClosedDir _ as d ->
        let d = specification_of_module d in
        raw_extern_module_specification filename
      	       	       	       	       ((SA__freeze(),PA__freeze()),(sp,d))
    | _ -> errorlabstrm "library__extern_module_specification_to"
              	       	[< 'S"Can only extern a module-object" >]
***)

and intern_module s = intern_module_from s s

and intern_module_from s filename =
    let ((sa,pa),(sp,d)) = raw_intern_module filename in
    reintern0__install_saved_tables(sa,pa);
    MSG [< 'S"[Reinterning " ; 'S s ; 'S" ..." >];
    flush std_out;
    (match reintern0__ri_node d with
       lib__ClosedDir(_,odc,_,_) as d ->
         let sp = reintern0__ri_sp sp in
         let impbefore = needed_modules d in
         MSGNL [< 'S"done]" >];
         reintern0__erase_tables();
         do_listLR (fun sp -> if not(module_is_loaded sp) then
                                intern_module(string_of_id(basename sp)))
        	   impbefore;
         lib__add_module (sp,d);
         cache_decl (sp,d)
     | _ -> failwith "library__intern_module")

(***
and intern_module_specification s = intern_module_specification_from s s

and intern_module_specification_from s filename =
    let ((sa,pa),(sp,d)) = raw_intern_module_specification filename in
    reintern0__install_saved_tables(sa,pa);
    MSG [< 'S"[Reinterning specification " ; 'S s ; 'S" ..." >];
    flush std_out;
    (match reintern0__ri_node d with
       lib__ClosedDir(_,odc,_,_) as d ->
         let sp = reintern0__ri_sp sp in
         let impbefore = needed_modules d in
         MSGNL [< 'S"done]" >];
         reintern0__erase_tables();
         do_listLR (fun sp ->
                      if not(module_is_loaded sp) then
                        intern_module_specification(string_of_id(basename sp)))
      	           impbefore;
         lib__add_module (sp,d);
         cache_decl (sp,d)
     | _ -> failwith "library__intern_module_specification")
***)

in (extern_module,
    extern_module_to,
    (*extern_module_specification,*)
    (*extern_module_specification_to,*)
    intern_module,
    intern_module_from(*,
    intern_module_specification,
    intern_module_specification_from*))
;;

let read_module s =
  if not(module_is_known s) then intern_module s
  else warning ("Module "^s^" already interned")
;;

let read_module_from s filename =
  if not(module_is_known s) then intern_module_from s filename
  else warning ("Module "^s^" already interned")
;;

(*** The following two functions are not needed any more...
let read_module_specification s =
  if not(module_is_known s) then intern_module_specification s
  else warning ("Module "^s^" already interned")
;;

let read_module_specification_from s filename =
  if not(module_is_known s) then intern_module_specification_from s filename
  else warning ("Module "^s^" already interned")
;;
***)

let require_module s locally =
  if not(module_is_known s) then intern_module s;
  if locally then import_module s else export_module s
;;

let require_module_from s filename locally =
  if not(module_is_known s) then intern_module_from s filename;
  if locally then import_module s else export_module s
;;

(*** two following functions not needed any more
let require_module_specification s locally =
  if not(module_is_known s) then intern_module_specification s;
  if locally then import_module s else export_module s
;;

let require_module_specification_from s filename locally =
  if not(module_is_known s) then intern_module_specification_from s filename;
  if locally then import_module s else export_module s
;;
***)

let declare_summary = summary__declare_summary;;
let search_imports = summary__search_imports;;

let with_heavy_rollback f x =
    resynchronize_caches();
    let sum = freeze_summary()
    and flib = lib__freeze() in
    try f x
    with reraise -> (unfreeze_summary sum;
                     lib__unfreeze flib;
                     raise reraise)
;;

let search_modules () =
  let ctxt = lib__contents_after NONE in
  map_succeed (fun (sp,lib__ClosedDir(s,{lib__module_p=true;_},_,_)) -> (s,sp)
                 | _ -> failwith "caught") ctxt
;;

let fmt_module_state () =
  let opened = search_imports ()
  and loaded = search_modules () in
       [< 'S"Imported (open) Modules: " ;
          prlist_with_sep (fun () -> [< 'SPC >])
               (fun sp -> [< 'S(string_of_path sp) >])
      	       opened ;
      	  'FNL ;
          'S"Loaded Modules: " ;
          prlist_with_sep (fun () -> [< 'SPC >])
               (fun (s,sp) -> [< 'S s ; 'S" = " ; 'S(string_of_path sp) >])
      	       loaded ;
      	  'FNL >]
;;

(* D1 | Resetting *)

let raw_reset_section str =
  (try let sp = lib__find_dir (SOME str) in
       lib__reset_to sp
   with Failure _ -> error ("Section "^str^" not declared"));
  backtrack_caches()
;;

let reset_keeping str =
  lib__reset_keeping_P (fun node -> str = (lib__node_name node));
  backtrack_caches ();;

let reset_to str =
  lib__reset_to_P (fun node -> str = (lib__node_name node));
  backtrack_caches ()
;;

let find_nearest_module () =
  lib__find_entry_P (fun (_,lib__OpenDir (_,{lib__module_p=true;_})) -> true
                       | _ -> false)
;;

let find_nearest_section () =
  lib__find_entry_P (fun (_,lib__OpenDir (_,{lib__module_p=false;_})) -> true
                       | _ -> false)
;;

let is_module_p sp =
  match lib__map sp with
    lib__OpenDir(_,{lib__module_p=true;_}) -> true
  | _ -> false
;;

let is_section_p sp =
  match lib__map sp with
    lib__OpenDir(_,{lib__module_p=false;_}) -> true
  | _ -> false
;;

(* $Id: library.ml,v 1.4 1996/02/12 15:45:32 cpaulin Exp $ *)
