(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                 names.ml                                 *)
(****************************************************************************)

#open "std";;
#open "more_util";;
#open "pp";;
#open "stdpp";;
#open "lexer";;

(********************************************************************)
(*                             Names                                *)
(********************************************************************)

type identifier = {mutable atom : SA__t ; index : int};;

type section_path = {mutable dirpath : PA__t ;
                     basename : identifier ;
                     kind : path_kind}
;;

let make_ident sa n = {atom = sa ; index = n};;
let repr_ident {atom=sa; index=n} = (sa,n);;

let make_path pa id k = {dirpath = pa; basename = id; kind = k};;
let repr_path {dirpath=pa; basename=id; kind=k} = (pa,id,k);;

(* type name = Name of identifier | Anonymous;; *)

let explode_id {atom=s;index=n} =
  (explode (SA__un_intern s))@(if n = (-1) then []
                                           else explode(string_of_int n))
;;

let print_id {atom=a;index=n} = match (SA__un_intern a,n) with
    ("",-1) -> [< 'S"[]" >]
  | ("",n)  -> [< 'S"[" ; 'INT n ; 'S"]" >]
  | (s,n)   -> [< 'S s ; (if n = (-1) then [< >] else [< 'INT n >]) >]
;;

let string_of_id {atom=s;index=n} =
    (SA__un_intern s) ^ (if n = (-1) then "" else (string_of_int n))
;;

let id_of_string s =
    let slen = string_length s in
    let rec numpart n =
        if n = 0 then failwith("identifier " ^ s ^ " cannot be split")
        else let c = int_of_char(nth_char s (n-1)) in
            if (int_of_char `0`) <= c & c <= (int_of_char `9`) then
                numpart (n-1)
            else
                n
    in
    let numstart = numpart slen in
    if numstart = slen then {atom=SA__intern s;index=(-1)} else
        {atom=SA__intern(sub_string s 0 numstart);
         index =int_of_string (sub_string s numstart (slen - numstart))}
;;

let atompart_of_id id = id.atom;;
let index_of_id id = id.index;;


let id_ord id1 id2 =
  let ((s1,n1)) = repr_ident id1 and
      ((s2,n2)) = repr_ident id2 in
  let s_bit = SA__ord s1 s2
  in if s_bit = 0 then n1 - n2 else s_bit
;;

let next_ident_away ({atom=str;_} as id) l = name_rec (-1)
  where rec name_rec i =
    let new = if i = (-1) then id else {atom=str;index= i}
    in if mem new l then name_rec (i+1) else new
;;

let lift_ident {atom=str;index=i} = {atom=str;index=i+1};;

let rec next_ident_away_from  id avoid = name_rec id 
  where rec name_rec ({atom=str;index= i} as id)  =
    if mem id avoid then name_rec {atom=str; index=i+1} else id
;;

let next_name_away name l = match name with
   Name(str) ->  next_ident_away str l
 | Anonymous -> id_of_string "Anonymous"
;;

let kind_of_path sp = sp.kind;;

let string_of_kind = function
    CCI -> "cci" | FW -> "fw" | OBJ -> "obj"
;;

let kind_of_string = function
    "cci" -> CCI | "fw" -> FW | "obj" -> OBJ
  | _ -> error "Malformed path-kind"
;;

let coerce_path k {dirpath=p;basename=id;_} =
  {dirpath=p;basename=id;kind=k};;
let ccisp_of_fwsp {dirpath=p;basename=id;kind=FW} =
  {dirpath=p;basename=id;kind=CCI};;
let ccisp_of {dirpath=p;basename=id;_} =
  {dirpath=p;basename=id;kind=CCI};;
let objsp_of {dirpath=p;basename=id;_} =
  {dirpath=p;basename=id;kind=OBJ};;
let fwsp_of_ccisp {dirpath=p;basename=id;kind=CCI} =
  {dirpath=p;basename=id;kind=FW};;
let fwsp_of {dirpath=p;basename=id;_} =
  {dirpath=p;basename=id;kind=FW};;

let sp_ord sp1 sp2 =
  let ((p1,id1,k)) = repr_path sp1 and
      ((p2,id2,k')) = repr_path sp2 in
  if k = k' then
    let p_bit = PA__ord p1 p2
    in if p_bit = 0 then id_ord id1 id2 else p_bit
  else
    compare_strings (string_of_kind k) (string_of_kind k')
;;

let sp_gt (sp1,sp2) =
  if sp_ord sp1 sp2 > 0 then true else false
;;

let string_of_path sp =
  let (sl,id,k) = repr_path sp in
  implode(join((map (fun s -> ["#";s])
    ((rev (PA__un_intern sl))@[string_of_id id]))@[[".";string_of_kind k]]))
;;

let string_of_path_mind sp mindid =
  let ((sl,id,k)) = repr_path sp in
  implode(join((map (fun s -> ["#";s])
   ((rev (PA__un_intern sl))@[string_of_id mindid]))@[[".";string_of_kind k]]))
;;

let qualid = function
    [< 'Tsharp ;
       (ne_list_with_sep (function [< 'Tsharp >] -> ())
                         (function [< 'Tident s >] -> s)) l ;
       'Tdot ; 'Tident pk >] ->
    let (bn,dirp) = sep_last l
    in make_path (PA__intern dirp) (id_of_string bn) (kind_of_string pk)
;;

let path_of_string s =
    try with_tokens vkey__vernac_keywords qualid s
    with Parse_error|Parse_failure|(UserError _) ->
         errorlabstrm "path_of_string" [< 'S s >]
;;

let basename sp = sp.basename;;
let dirpath sp = PA__un_intern sp.dirpath;;

let append_to_path sp str =
  let ((sp,id,k)) = repr_path sp in
  make_path sp (id_of_string((string_of_id id)^str)) k
;;

let sp_of_wd dp =
  let (bn::dp) = dp in
  make_path (PA__intern dp) (id_of_string bn) OBJ
;;

let GLOB hyps = ENVIRON(hyps,[]);;

let ids_of_sign (idl,_) = idl;;
let vals_of_sign (_,vals) = vals;;
let add_sign (id,ty) (idl,tyl) = (id::idl,ty::tyl);;
let sign_it f (idl,tyl) e = list_it2 f idl tyl e;;
let it_sign f e (idl,tyl) = it_list2 f e idl tyl;;
let nil_sign = ([],[]);;
let rev_sign (idl,tyl) = (rev idl,rev tyl);;
let map_sign_typ f (idl,tyl) = (idl,map f tyl);;

let dbenv_it f (ENVIRON(_,dbs)) init =
    list_it (fun (na,t) v -> f na t v) dbs init;;
let it_dbenv f init (ENVIRON(_,dbs)) =
    it_list (fun v (na,t) -> f v na t) init dbs;;

let isnull_sign = function 
    ([],[])     -> true
  | (_::_,_::_) -> false
  | _           -> invalid_arg "isnull_sign"
;;

let isnull_rel_env (ENVIRON(_,dbs)) = (dbs = []);;
let uncons_rel_env (ENVIRON(sign,dbs)) = (hd dbs,ENVIRON(sign,tl dbs));;

let hd_sign = function
    (id::_,ty::_) -> (id,ty)
  | _             -> failwith "hd_sign"
;;

let tl_sign = function
    (_::idl,_::tyl) -> (idl,tyl)
  | _               -> failwith "tl_sign"
;;

let get_globals (ENVIRON(g,_)) = g;;
let get_rels (ENVIRON(_,r)) = r;;

let add_rel (n,x) (ENVIRON(g,r)) = (ENVIRON(g,(n,x)::r));;

let add_glob (id,x) (ENVIRON((dom,rang),r)) = (ENVIRON((id::dom,x::rang),r));;

let lookup_glob id (ENVIRON((dom,rang),_)) = aux dom rang
  where rec aux = fun
     []          []        -> raise Not_found
   | (id'::id'l) (ty::tyl) -> if id' = id then (id',ty) else aux id'l tyl
;;

let lookup_rel n (ENVIRON(_,r)) = lookrec n r
  where rec lookrec = fun
     1 ((na,x)::l) -> (na,x)
   | n (_::l)      -> lookrec (n-1) l
   | _ []          -> raise Not_found
;;

let rec lookup_rel_id id (ENVIRON(_,r)) = lookrec 1 r
  where rec lookrec = fun
     n ((Anonymous,x)::l) -> lookrec (n+1) l
   | n ((Name id',x)::l)  -> if id' = id then (n,x) else lookrec (n+1) l
   | _ []                 -> raise Not_found
;;

let lookup_id id env =
    try RELNAME(lookup_rel_id id env)
    with Not_found -> GLOBNAME(lookup_glob id env)
;;

let map_rel_env f (ENVIRON(g,r)) =
    (ENVIRON(g,map (fun (na,x) -> (na,f x)) r));;

let map_var_env f (ENVIRON((dom,rang),r)) =
  (ENVIRON(list_it2 (fun na x (doml,rangl) -> (na::doml,(f x)::rangl))
      	       	       	       	       	       	       	  dom rang ([],[]),r))
;;

let nth_sign (idl,tyl) n = (nth idl n,nth tyl n);;
let map_sign_graph f (ids,tys) = map2 f ids tys;;

let make_sign = split;;
let do_sign f (ids,tys) = do_list2 f ids tys;;

let uncons_sign = function
    (id::idl,ty::tyl) -> ((id,ty),(idl,tyl))
  | _ -> anomaly "signatures are being manipulated in a non-abstract way"
;;

let lookup_sign id sign = lookup_glob id (GLOB sign);;

let sign_length (idl,tyl) =
  let lenid = list_length idl
  and lenty = list_length tyl in
  if lenid = lenty then lenid
                   else invalid_arg "lookup_sign"
;;

let mem_sign sign id = mem id (ids_of_sign sign);;

let modify_sign id ty = modrec
  where rec modrec = function
     [],[] -> invalid_arg "modify_sign"
   | sign  -> let (id',ty') = hd_sign sign in
              if id = id' then
      	        add_sign (id,ty) (tl_sign sign)
              else
      	       	add_sign (id',ty') (modrec (tl_sign sign))
;;

let add_envp_rel (na,v1,v2) (env1,env2) =
    (add_rel (na,v1) env1,add_rel (na,v2) env2);;

let exists_sign f = exrec
  where rec exrec sign =
    if isnull_sign sign then false
    else let ((id,T),tl) = uncons_sign sign
         in f id T or exrec tl
;;

(* [sign_prefix id sign] returns the signature up to and including id,
   with all later assumptions stripped off.  It is an error to call it
   with a signature not containing id, and that error is generated
   with error. *)

let sign_prefix id sign = prefrec sign
  where rec prefrec sign =
    if isnull_sign sign then
      error "sign_prefix"
    else
      let ((id',T),sign') = uncons_sign sign in
      if id' = id then sign else prefrec sign'
;;

let add_sign_after whereid (id,T) sign = addrec sign
  where rec addrec sign =
    if isnull_sign sign then
      error "add_sign_after"
    else
      let ((id',T'),sign') = uncons_sign sign in
      if id' = whereid then add_sign (id,T) sign
      else add_sign (id',T') (addrec sign')
;;

let add_sign_replacing whereid (id,T) sign = addrec sign
  where rec addrec sign =
    if isnull_sign sign then
      error "add_replacing_after"
    else
      let ((id',T'),sign') = uncons_sign sign in
      if id' = whereid then add_sign (id,T) sign'
      else add_sign (id',T') (addrec sign')
;;

(* [prepend_sign Gamma1 Gamma2]
   prepends Gamma1 to the front of Gamma2, given that their namespaces
   are distinct. *)

let prepend_sign Gamma1 Gamma2 =
  if [] = listset__intersect (ids_of_sign Gamma1) (ids_of_sign Gamma2) then
    let (ids1,vals1) = Gamma1 and
        (ids2,vals2) = Gamma2
    in (ids1@ids2,vals1@vals2)
  else
    failwith "prepend_sign"
;;

(* $Id: names.ml,v 1.4 1996/02/12 15:45:36 cpaulin Exp $ *)
