(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                               pureparse.ml                               *)
(****************************************************************************)

#infix "o";;
#open "std";;
#open "names";;
#open "more_util";;
#open "lexer";;
#open "pp";;
#open "stdpp";;
#open "lomsg";;
#open "termact";;
#open "vkey";;
#open "ast";;


let gk_ord (wf1,id1) (wf2,id2) =
    let hi_bit = id_ord wf1 wf2
    in if hi_bit = 0 then id_ord id1 id2
       else hi_bit
;;

let fv_num = function
    LITERAL _ -> 0
  | PRIMITIVE _ -> 1
  | EPSILON -> 2
;;

let fv_ord fv1 fv2 =
    if fv_num fv1 = fv_num fv2 then
        match (fv1,fv2) with
        (LITERAL (s1,_),LITERAL (s2,_)) -> compare_strings s1 s2
      | (PRIMITIVE gk1,PRIMITIVE gk2) -> gk_ord gk1 gk2
      | (EPSILON,EPSILON) -> 0
    else (fv_num fv1) - (fv_num fv2)
;;

let isMeta id =
    let s = SA__un_intern (atompart_of_id id)
    in string_length s <> 0
       & ("$" = (sub_string s 0 1));;

let lex_string s =
  match stream_to_list (token_stream vernac_keywords (stream_of_string s)) with
    x::_ -> x
  | _    -> raise (Lexer_error (-1))
;;

let parse_bar = function
    [< 'Tbar >] -> ()
;;

let parse_prod_item parse_act whatfor = function
    [< 'Tstring s >] -> TERM(lex_string s)
  | [< 'Tident id ;
       (function
        [< 'Tcolon ; 'Tident id' >] -> (id_of_string id,id_of_string id')
      | [< >] -> (whatfor,id_of_string id)) gk ;
       (function
        [< 'Tlbracket ;
           (plist_with_sep (function [< 'Tsemi >] -> ()) parse_act) al ;
           'Trbracket >] -> al
      | [< >] -> []) al ;
       (function
        [< 'Tlparen; 'Tident at ; 'Trparen >]
        -> let atid = id_of_string at
           in if isMeta atid then NONTERM((gk,al),SOME atid)
           else error ("ident found which was not a metavariable: "^at)
      | [< >] -> NONTERM((gk,al),NONE)) x >] -> x
;;

let parse_production parse_act whatfor = function
    [< 'Tlbracket ; (plist (parse_prod_item parse_act whatfor)) l ; 'Trbracket ;
       'Tminusgreater ; parse_act act >] ->
    (l,act)
;;

let parse_grammar_entry parse_act whatfor = function
    [< 'Tident id ; 
     (function
      [< 'Tlbracket ; 
           (plist_with_sep (function [< 'Tsemi >] -> ())
                           (function [< 'Tident id >] -> 
                     let atid=id_of_string id
                     in if isMeta atid then atid
                        else error ("ident found which was not a metavariable: "^id) )) l ;
         'Trbracket >] -> l
    | [< >] -> []) idl ;
       'Tcolonequals ;
       (plist_with_sep parse_bar (parse_production parse_act whatfor)) pl >] ->
    ((whatfor,id_of_string id),(idl,pl))
;;

let print_bar () = [< 'CUT ; 'S"|"; 'WS 1 >];;

let print_token t =
[< 'S"\"" ; 'S(string_for_read (string_of_token rev_vernac_keywords t)) ;
       'S"\"" >];;

let print_fnl () = [< 'FNL >];;
let print_spc () = [< 'SPC >];;

let print_gk whatfor (id1,id2) =
    if whatfor = id1 then
        [< 'S(string_of_id id2) >]
    else
        [< 'S(string_of_id id1) ; 'S":" ; 'S(string_of_id id2) >]
;;

let print_prod_item whatfor = function
    NONTERM((gk,al),ido) ->
    [< print_gk whatfor gk ;
       (if al = [] then [< >]
        else [< 'S"[" ; prlist_with_sep (fun () -> [< 'S";" >])
                        fmt_act al ;
                'S"]" >]) ;
       (match ido with
        NONE -> [< >]
      | SOME atid -> [< 'S"(" ; 'S(string_of_id atid) ; 'S")" >]) >]

  | TERM t -> print_token t
;;

let print_production whatfor = function
     (l,a) ->
     HOV 0 [< 'S"[" ; prlist_with_sep (fun () -> [< 'SPC >]) (print_prod_item whatfor) l ;
        'S"]" ; 'CUT ;  'S" -> " ; fmt_act a >]
;;

let print_grammar_entry whatfor (id,(idl,pl)) =
    [< print_gk whatfor id ;
       (if idl = [] then [< >]
        else [< 'S"[" ;
                prlist_with_sep (fun () -> [< 'S";" >])
                                (fun id -> [< 'S (string_of_id id) >])
                                idl ;
                'S"]">]) ;
       'S " := " ; HOV 0 (prlist_with_sep print_bar (print_production whatfor) pl) >]
;;

let rec print_cgram = function
    c_ACTION a -> [< 'S" -> " ; fmt_act a >]
  | c_SWITCH(tl,el) ->
    [< 'S"{" ;
    HOV 0 [< prlist_with_sep (fun () -> [< 'S"|" ; 'FNL >])
                             (fun l -> HOV 0 (print_toks_line l)) tl ;
             if el <> [] then [< 'FNL >] else [< >] ;
             prlist_with_sep (fun () -> [< 'S"|" ; 'FNL >])
                             (fun l -> HOV 0 (print_eps_line l)) el >] ;
      'S"}" >]

and print_toks_line(toks,(pgl,cg)) =
    [< 'S"[" ; prlist_with_sep print_spc print_token toks ; 'S"]" ; 'SPC ;
       'S"[" ; prlist_with_sep print_spc print_pgram pgl ; 'S"]" ; 'FNL ;
       HOV 2 (print_cgram cg) >]

and print_eps_line(pgl,cg) =
    [< 'S"[" ; prlist_with_sep print_spc print_pgram pgl ; 'S"]" ; 'FNL ;
       HOV 2 (print_cgram cg) >]

and print_pgram = function
    c_NTERM ((gk,al),ido) ->
    [< print_gk (id_of_string "$BOGUS") gk ;
       (if al = [] then [< >]
        else [< 'S"[" ; prlist_with_sep (fun () -> [< 'S";" >])
                        fmt_act al ;
                'S"]" >]) ;
       (match ido with
        NONE -> [< >]
      | SOME atid -> [< 'S"(" ; 'S(string_of_id atid) ; 'S")" >]) >]
  | c_TERM t -> print_token t
;;

let dollarzero = id_of_string "$0";;

let prim_grams = (mhm__new 7 : (gram_key, (vernac_keyword lexer__token stream -> frag) * first_val list) mhm__t);;

let lookup_prim_gram gk = fst(mhm__map prim_grams gk);;

let grams = (mhs__new(17,NONE) : (gram_key,gram_defn) mhs__t);;

let compiled_grams = (fmavm__new(gk_ord,17) : (gram_key,compiled_grammar_entry) fmavm__t);;

let uncompiled_grams = (mls__mt() : (gram_key*gram_defn) mls__t);;

let freeze () = (mhs__freeze grams,fmavm__freeze compiled_grams,
                 mls__freeze uncompiled_grams);;
let unfreeze (fs,fcs,fug) = (mhs__unfreeze fs grams;
                         fmavm__unfreeze fcs compiled_grams;
                         mls__unfreeze fug uncompiled_grams);;

let init () = (mhs__empty grams; fmavm__empty compiled_grams);;

let lookup_cgram gk =
    try fmavm__map compiled_grams gk
    with Not_found -> errorlabstrm "Grammar Compiler"
                      [< 'S"Undeclared nonterminal " ;
                         'S(string_of_id(fst gk)) ; 'S":" ;
                         'S(string_of_id(snd gk)) >]
;;

let make_new_cgram gk idl prods=
    fmavm__add compiled_grams
    (gk,
     {g_args = idl;
      g_prodns = prods;
      g_compiled = c_SWITCH([],[]);
      g_firsts = fvavs__mt;
      g_mentions = gkavs__mt;
      g_mentioned_by = gkavs__mt})
;;

let update_firsts gk new_firsts =
if mhm__in_dom prim_grams gk then anomaly "update_firsts";
let oge = lookup_cgram gk
in fmavm__remap compiled_grams gk
    {g_args = oge.g_args;
     g_prodns = oge.g_prodns;
     g_compiled = oge.g_compiled;
     g_firsts = fvavs__union oge.g_firsts new_firsts;
     g_mentions = oge.g_mentions;
     g_mentioned_by = oge.g_mentioned_by}
;;

let update_mby gk new_mby =
if mhm__in_dom prim_grams gk then () else
let oge = lookup_cgram gk
in fmavm__remap compiled_grams gk
    {g_args = oge.g_args;
     g_prodns = oge.g_prodns;
     g_compiled = oge.g_compiled;
     g_firsts = oge.g_firsts;
     g_mentions = oge.g_mentions;
     g_mentioned_by = gkavs__union oge.g_mentioned_by new_mby}
;;

let update_mentions gk new_mentions =
if mhm__in_dom prim_grams gk then anomaly "update_mentions";
let oge = lookup_cgram gk
in fmavm__remap compiled_grams gk
    {g_args = oge.g_args;
     g_prodns = oge.g_prodns;
     g_compiled = oge.g_compiled;
     g_firsts = oge.g_firsts;
     g_mentions = gkavs__union oge.g_mentions new_mentions;
     g_mentioned_by = oge.g_mentioned_by}
;;

let update_prods gk new_prods =
if mhm__in_dom prim_grams gk then anomaly "update_mentions";
let oge = lookup_cgram gk
in fmavm__remap compiled_grams gk
    {g_args = oge.g_args;
     g_prodns = new_prods@oge.g_prodns;
     g_compiled = oge.g_compiled;
     g_firsts = oge.g_firsts;
     g_mentions = oge.g_mentions;
     g_mentioned_by = oge.g_mentioned_by}
;;

let update_compiled gk new_compiled =
if mhm__in_dom prim_grams gk then anomaly "update_compiled";
let oge = lookup_cgram gk
in fmavm__remap compiled_grams gk
    {g_args = oge.g_args;
     g_prodns = oge.g_prodns;
     g_compiled = new_compiled;
     g_firsts = oge.g_firsts;
     g_mentions = oge.g_mentions;
     g_mentioned_by = oge.g_mentioned_by}
;;

let nonterminals pl =
    it_list (fun acc (l,_) ->
                 it_list (fun acc (TERM _) -> acc
                            | acc (NONTERM((gk,_),_)) -> gkavs__add acc gk)
                         acc l)
            (gkavs__mt) pl
;;

let productions_of gk =
    join(map (fun (_,pl) -> pl) (mhs__map grams gk))
;;

let productions_of1 gk =
   if fmavm__in_dom compiled_grams gk then
        (lookup_cgram gk).g_prodns
      else errorlabstrm "Grammar Compiler"
        [< 'S"Undeclared nonterminal " ;
           'S(string_of_id(fst gk)) ; 'S":" ;
           'S(string_of_id(snd gk)) >]
;;

(* inutilise'e par "personne" pour l'instant *)
let nonterminals_of gk =
    nonterminals(productions_of gk);;

let params_of gk =
    try (match mhs__map grams gk with
         (idl,_)::_ -> idl
       | _ -> error "No such grammar")
    with Not_found -> error "No such grammar"
;;

let params_of1 gk =
   if fmavm__in_dom compiled_grams gk then
        (lookup_cgram gk).g_args
      else errorlabstrm "Grammar Compiler"
        [< 'S"Undeclared nonterminal " ;
           'S(string_of_id(fst gk)) ; 'S":" ;
           'S(string_of_id(snd gk)) >];;

let firsts_of_key gk =
    if fmavm__in_dom compiled_grams gk then
        (lookup_cgram gk).g_firsts
    else if mhm__in_dom prim_grams gk then
        fvavs__ofList (snd(mhm__map prim_grams gk))
    else errorlabstrm "Grammar Compiler"
        [< 'S"Undeclared nonterminal " ;
           'S(string_of_id(fst gk)) ; 'S":" ;
           'S(string_of_id(snd gk)) >]
;;

let mby_of_key gk =
    if fmavm__in_dom compiled_grams gk then
        (lookup_cgram gk).g_mentioned_by
    else anomaly "asking for mbys on terminal"
;;

let firsts_of_prodn acc (l,_) = firstrec acc l
    where rec firstrec acc = function
    [] -> fvavs__add acc EPSILON
  | (TERM t)::l -> fvavs__add acc (LITERAL (string_of_token rev_vernac_keywords t,t))
  | (NONTERM((gk,_),_))::l ->
    let fs = firsts_of_key gk
    in if fvavs__memb fs EPSILON then
        firstrec (fvavs__union acc (fvavs__rmv fs EPSILON)) l
       else fvavs__union acc fs
;;

let firsts l = it_list firsts_of_prodn (fvavs__mt) l;;

let literals_only (fs,_) =
    for_all (fun (LITERAL _) -> true | _ -> false) (fvavs__toList fs)
    & not(fvavs__is_empty fs)
;;

let prims_plus_literals (fs,_) =
let l = fvavs__toList fs
in  for_all (fun (LITERAL _) -> true
               | (PRIMITIVE _) -> true
               | _ -> false) l
  & exists (fun (PRIMITIVE _) -> true | _ -> false) l
  & l <> []
;;

let epsilon (fs,_) =
    fvavs__memb fs EPSILON & not(fvavs__is_empty fs)
;;

let empties (fs,_) = fvavs__is_empty fs;;

let sort_list l =
let lits = filter literals_only l and
    primlits = filter prims_plus_literals l and
    epss = filter epsilon l and
    emps = filter empties l
in if exists (fun (litset,_) ->
                   exists (fun (primlset,_) ->
                               not(fvavs__is_empty (fvavs__intersect primlset litset)))
                          primlits)
              lits then
      errorlabstrm "Grammar Compiler"
      [< 'S"A production with a literal in its first-set overlaps with" ; 'SPC ;
         'S"a production with a primitive in its first-set" >]
   else (lits,primlits,epss,emps)
;;

(* given 'a list list, produces 'a list * 'a list list
 * where the first list is the common prefix of all the lists in the
 * argument.
 *)
let rec common_prefix acc (ll : production_item list list) =
    if ll = [] then failwith "common_prefix"
    else if hd ll = [] then (rev acc),ll
    else
    let e = hd (hd ll)
    in if for_all (fun (_::_ as l) -> e = hd l
                     | _ -> false) ll then
        common_prefix (e::acc) (map tl ll)
       else (rev acc,ll)
;;

let rec compile_prodl = function
    [] -> c_SWITCH([],[])
  | [([],a)] -> c_ACTION a
  | l ->
    (let fl = map (firsts_of_prodn (fvavs__mt)) l in
     let zipped = combine(fl,l) in
     let (litl,priml,epsl,emptyl) = sort_list zipped
     in (c_SWITCH(compile_literals litl,
                  (compile_prims priml)@
                  (compile_mt emptyl)@
                  (compile_eps epsl))))

(* Given a first-set, FS, and the list of
 * branches, this function will
 * (1) cut out all branches which have matching first-sets, MATCHBL
 * (2) construct the longest common prefix of these branches, PREFL
 *     and the tails, TAILBL
 * (3) recursively process each tail, giving PROC_TAILBL
 * (4) and return (PREFL,PROC_TAILBL)
 *)
and compile_fs fs bl =
let matchbl = map snd (filter (fun (fs',_) -> (fvavs__equal fs fs')) bl)
in if length matchbl = 1 then
    let prefl = map (fun (NONTERM gc) -> c_NTERM gc
                       | (TERM t) -> c_TERM t) (fst(hd matchbl))
    in (prefl,c_ACTION (snd(hd matchbl)))
   else 
    let (prefl,tailbl) =
        (let actl = map snd matchbl in
         let prefl,restl = common_prefix [] (map fst matchbl)
         in (prefl,(combine(restl,actl)))) in
    if prefl = [] then
        errorlabstrm "non-LL branch"
        [< 'S"Tails of productions were : " ; 'FNL ;
           prlist_with_sep (fun () -> [< 'FNL >])
                           (print_production (id_of_string "$foo")) (map snd bl) >];  
    let prefl = map (fun (NONTERM gc) -> c_NTERM gc
                       | (TERM t) -> c_TERM t) prefl in
    let proc_tailbl = compile_prodl tailbl
    in ((prefl: primitive_grammar list),(proc_tailbl:compiled_grammar))

and compile_literals litl =
let fsl = uniquize_eqfun fvavs__equal (map fst litl) in
let compl = list_it (fun fs sofar ->
                         let this_branch = compile_fs fs litl
                         in (fs,this_branch)::sofar)
                    fsl []
in map (fun (fs,(branch:(primitive_grammar list * compiled_grammar))) ->
                 (map (fun (LITERAL(_,t)) -> t) (fvavs__toList fs),branch))
       compl

and compile_prims litl =
let fsl = uniquize_eqfun fvavs__equal (map fst litl) in
let compl = list_it (fun fs sofar ->
                         let this_branch = compile_fs fs litl
                         in (fs,this_branch)::sofar)
                    fsl []
in map snd compl

and compile_mt litl =
    map (fun (fs,b) -> compile_fs fs [(fs,b)]) litl

and compile_eps litl =
let fsl = uniquize_eqfun fvavs__equal (map fst litl) in
let compl = list_it (fun fs sofar ->
                         let this_branch = compile_fs fs litl
                         in (fs,this_branch)::sofar)
                    fsl []
in if length compl > 1 then
    error "too many epsilon branches"
   else map snd compl
;;

let compile_gkey gk =
let pl = productions_of1 gk
in try compile_prodl pl
   with UserError(l,pps) ->
       errorlabstrm "Grammar Compiler"
       [< 'S l;'FNL; 'S"Error while compiling grammar " ;
          'S(string_of_id (fst gk)) ; 'S":" ; 'S(string_of_id (snd gk)) ;'FNL;
          pps >]
      | e -> (MSG(errors__explain_sys_exn e); raise e)
;;

(* To recompile a nonterminal, we first put it on the recompilation
 * list and the re-check list.
 *
 * Then, until the re-check list becomes empty,
 * let L be everything on the re-check list, and empty out the re-check list.
 *   for every nonterminal X on L,
 *     recompute X's firsts, and X's mentions.
 *     Update X's mentions, and also the inverse mapping
 *     if X's firsts have changed, then
 *         put all nonterminals on X's mentioned_by list onto the re-check list
 *         update X's firsts
 *)

let recheck_because_of gkbecause gkrecheckl (gkofirsts,gknfirsts) = ();;

let recompile_nonterminals gkl =
let recomp_list = mls__mt() in
let recheck_list = mls__mt()
in do_list (mls__add recheck_list) gkl;
   do_list (mls__add recomp_list) gkl;
   do_list
   (fun gk ->
        let pl = productions_of1 gk in
        let fs = firsts pl and
            mentions = nonterminals pl
        in update_mentions gk mentions;
            gkavs__app (fun m -> update_mby m (gkavs__ofList [gk])) mentions)
   gkl;
   do_list (fun gk -> gkavs__app (mls__add recheck_list) (mby_of_key gk)) gkl;
      while (not(mls__is_empty recheck_list)) do
       (let L = mls__toList recheck_list
        in mls__empty recheck_list;
           do_list
            (fun gk ->
                 (let pl = productions_of1 gk in
                  let fs = firsts pl
                  in if not(fvavs__equal fs (firsts_of_key gk)) then
                      (recheck_because_of gk (mby_of_key gk) (fs,firsts_of_key gk);
                       gkavs__app (mls__add recheck_list) (mby_of_key gk);
                       gkavs__app (mls__add recomp_list) (mby_of_key gk);
                       update_firsts gk fs)
                     else ())) L)
   done;
let L = mls__toList recomp_list
in MSGNL [< 'S"[Recompiling " ; 'INT(length L) ; 'S" nonterminal(s)...]" >];
    do_list (fun gk -> update_compiled gk (compile_gkey gk)) L
;;


(* a` mettre dans LIB/util, existe de'ja`? *)

let rec all_diffs = function 
     [] -> true
   | a::l -> if mem a l then false else all_diffs l;;

(** renommage "uniforme" des re`gles de grammaire **)

let rec ren_pat index env = function
    SPATquote ast -> index,SPATquote ast,env
  | SPATmeta id -> 
     (try let v=(assoc id env) 
         in index,SPATmeta v,env with      
        Not_found -> 
         let v = id_of_string ("$"^(string_of_int (index+1)))
         in index+1,SPATmeta v,(id,v)::env)
  | SPATnvar id -> (* je pense inutil!!! *)
    (try let v=(assoc id env) 
         in index,SPATnvar v,env with      
        Not_found -> 
         let v = id_of_string ("$"^(string_of_int (index+1)))
         in index+1,SPATnvar v,(id,v)::env)
  | SPATnil -> index,SPATnil,env
  | SPATcons(sp1,sp2) -> 
     let i1,pa1,env1=ren_pat index env sp1
     in let i2,pa2,env2=ren_pat i1 env1 sp2
        in i2,SPATcons(pa1,pa2),env2
  | SPATappend l -> let i',l',env'=ren_lpat index env l
                     in i',SPATappend (rev l'),env'
  | SPATlist l -> let i',l',env'=ren_lpat index env l
                     in i',SPATlist (rev l'),env'
  | SPATprim(id,spinj) -> 
     (try let v=(assoc id env) 
         in index,SPATprim(v,spinj),env with      
      Not_found -> 
        let v = id_of_string ("$"^(string_of_int (index+1)))
        in index+1,SPATprim(v,spinj),(id,v)::env)
  | SPATnode(nna,l) -> let i',l',env'=ren_lpat index env l
                        in i',SPATnode(nna,rev l'),env'
  | SPATbreaknode(spnna,sp) -> 
     let i1,pa1,env1=ren_pat index env spnna
     in let i2,pa2,env2=ren_pat i1 env1 sp
        in i2,SPATbreaknode(pa1,pa2),env2
  | SPATslam(NONE,spb) -> 
     let i1,pa1,env1=ren_pat index env spb 
     in i1,SPATslam(NONE,pa1),env1
  | SPATslam(SOME id,spb) -> 
     let i1,sid1,env1=
         (try let v=(assoc id env) 
              in index,SOME v,env with      
           Not_found -> 
            let v = id_of_string ("$"^(string_of_int (index+1)))
            in index+1,SOME v,(id,v)::env)
     in let i2,pa2,env2=ren_pat i1 env1 spb 
        in i2,SPATslam(sid1,pa2),env2
  | SPATquoteslam(id,spb) -> 
     let i1,pa1,env1=ren_pat index env spb
     in i1,SPATquoteslam(id,pa1),env1
  | SPATslamlist(spidl,spb) -> 
     let i1,pa1,env1=ren_pat index env spidl
     in let i2,pa2,env2=ren_pat i1 env1 spb
        in i2,SPATslamlist(pa1,pa2),env2
  | X -> index,X,env 

and ren_lpat index env lpat=
     it_list (function (i,res,env)->function pat->
             let i1,pa1,env1=ren_pat i env pat
             in i1,pa1::res,env1) (index,[],env) lpat;;

let ren_bindings index env bl=
   it_list (function (i,res,env)->function (id,pat) ->
             let i',pat',env'=ren_pat i env pat
             in try let v=(assoc id env') 
                     in i',(v,pat')::res,env' with      
                  Not_found ->
                   let v = id_of_string ("$"^(string_of_int (i'+1)))
                   in i'+1,(v,pat')::res,(id,v)::env') (index,[],env) bl;;

let rec ren_ac index env ac=
    match ac with 
      ACTL(bl) -> let i',bl',env'=ren_bindings index env bl
                  in i',ACTL(rev bl'),env'
    | CASE(id,alt) -> 
       let id',i,env = 
          try let v=(assoc id env) in v,index,env with
                 Not_found -> 
                  let v = id_of_string ("$"^(string_of_int (index+1)))
                  in v,index+1,(id,v)::env
       in let i',alt',env'=
           it_list (function (i,res,env) ->function (pat,ac) ->
                      let i',pat',env'=ren_pat i env pat
                      in let i'',ac'',env''=ren_ac i' env' ac
                         in i'',(pat',ac'')::res,env'') (i,[],env) alt
          in i',CASE(id',rev alt'),env';;

let ren_acl index env acl=
    it_list (function (i,res,env)->function ac -> 
             let i',ac',env'=ren_ac i env ac 
              in i',ac'::res,env') (index,[],env) acl;;

let ren_mdp index env mdp=
  it_list (function (i,res,env)-> function 
    (TERM t) -> i,(TERM t)::res,env
  | NONTERM((gk,acl),ido)  -> 
     let i',acl',env'=ren_acl i env acl 
     in (match ido with
           NONE -> i',NONTERM((gk,rev acl'),ido)::res,env'
         | SOME(id) -> 
            (try let v=(assoc id env') 
                in i',(NONTERM((gk,rev acl'),SOME(v)))::res,env' with
              Not_found -> 
               let v = id_of_string ("$"^(string_of_int (i'+1)))
               in i'+1,(NONTERM((gk,rev acl'),SOME(v)))::res,(id,v)::env')))
        (index,[],env) mdp;;

let ren_params params=
         it_list (function (i,res,env)->function p->
                         let v = id_of_string ("$"^(string_of_int (i+1)))
                         in i+1,v::res,(p,v)::env) 
                 (0,[],[(dollarzero,dollarzero)]) params;;

let ren_grammar_entry (params,pl) =
   let index1,new_params,env1=ren_params params
   in let new_pl=
       map (function (mdp,ac) -> 
                let index2,mdp2,env2=ren_mdp index1 env1 mdp 
                in let index3,ac3,env3=ren_ac index2 env2 ac
                   in (rev mdp2,ac3)) pl 
      in (rev new_params,new_pl);;

(* fin des fonctions de renommage *)


let add_productions l =
   do_list (fun (gk,(idl,pl)) ->    
    mhs__add grams (gk,(idl,pl));
    let idl1,pl1 = ren_grammar_entry (idl,pl)
    in if not(fmavm__in_dom compiled_grams gk) 
       then make_new_cgram gk idl1 pl1
       else update_prods gk pl1) l;
    recompile_nonterminals (map fst l)
;;

let push l =
let list_uncomp = mls__toList uncompiled_grams in
do_list (fun (gk,(idl,pl)) -> 
    if (not (all_diffs idl))
    then error "Malformed parameter list"
    else (try let (idl0,_) = hd(mhs__map grams gk)
              in if length(idl) <> length(idl0)
                 then error "Malformed parameter list"
          with  Not_found -> (try let (idl0,_)= assoc gk list_uncomp
                                 in if length(idl) <> length(idl0)
                                    then error "Malformed parameter list" 
                             with Not_found -> () )
              | UserError _ as reraise -> raise reraise)) l;
uncompiled_grams := list_uncomp@l
;;

let choose pl =
let rec aux = function
    [] -> (fun strm -> raise Parse_failure)
  | (p::pl) ->
    (function
     [< p rslt >] -> rslt
   | [< (aux pl) rslt >] -> rslt)
in aux pl
;;

let synchronize_grammar_interpreter () =
    if not(mls__is_empty uncompiled_grams) then
       let fs = freeze() in
        try (add_productions (mls__toList uncompiled_grams);
            mls__empty uncompiled_grams;true)
        with UserError _ as reraise -> (unfreeze fs; mls__empty uncompiled_grams;raise reraise)
              | _ -> (unfreeze fs;mls__empty uncompiled_grams;error "Grammar Compiler")        
    else false
;;

let comp_gram_of key =
(* in order to compile a grammar (by synchronize_grammar_interpreter)
 * and freeze the new state, we resync caches *)
  if not(mls__is_empty uncompiled_grams)  
    then library__resynchronize_caches ();
  (lookup_cgram key).g_compiled
;;

let build_call_env env idl al gk=
    if length idl <> length al then
       errorlabstrm "Grammar Compiler"
        [< 'S"Bad number of arguments in the call of ";
         (print_gk (id_of_string "#foo") gk) >]
    else
        map2 (fun id a -> (id,interp_act a env)) idl al
;;

let rec interpc env = function
    c_ACTION a -> (function strm ->interp_act a env)

  | c_SWITCH(kmap,cgseql) ->
    (function
     [< (choose
         ((map (fun (tok,cgseq) strm ->
                    (let _ = stream_get strm in ();
                     interpc_seq env cgseq strm)) kmap)@
          (map (interpc_seq env) cgseql))) r >] -> r)

and interpc_seq env = function
    ([],cg) -> (function [< (interpc env cg) r >] -> r)

  | (gh::gl,cg) ->
    (function
     [< (interpc_pg env gh) env' ;
        (interpc_seq env' (gl,cg)) renv >] -> renv)

and interpc_pg env = function
    c_TERM t ->
    (function
     [< (stream_check (fun t' -> t' = t)) t' >] -> env)

  | c_NTERM ((gk,al),ido) ->
    (function
     [< (interpc_gram env al gk) r >] ->
     (match ido with
      NONE -> env
    | SOME id -> termact__push id r env))

and interpc_gram env al gk = function
    [< (if (fmavm__in_dom compiled_grams gk)  
           or (exists (fun (gk',_) -> gk=gk') (mls__toList uncompiled_grams)) then
             let g = comp_gram_of gk and
                 idl = params_of1 gk
             in (interpc (build_call_env env idl al gk) g)
         else (try lookup_prim_gram gk
               with Not_found -> anomalylabstrm "interpc_gram"
                                 [< 'S"Missing grammar for nonterminal " ;
                                    'S(string_of_id (fst gk)) ; 'S":" ;
                                    'S(string_of_id (snd gk)) >])) r >] -> r
;;

let print_nonterminal_definition gk =
let pl = productions_of gk and
    idl = params_of gk
in print_grammar_entry (fst gk) (gk,(idl,pl))
;;

(*=================
LaTeX output
Claude
*)

let LaTeX_stream s = LaTeX_encode (stream_of_string s) 
where rec LaTeX_encode = function
    [< '`$`|`_` as c ; s >] -> [< 'S"\\" ; 'S(string_of_char c) ; (LaTeX_encode s) >]
  | [< 'c ; s >] -> [< 'S(string_of_char c) ; (LaTeX_encode s) >]
;;


let LaTeX_print_bar () = [< 'CUT ; 
                            'S"\\end{tabular}" ; 'FNL ; 'FNL ;
                            'S"\\begin{tabular}{cp{140mm}}" ; 'FNL ;
                            'S"\\mi&"; 'WS 1 >];;

let LaTeX_print_token t =
[< 'S"{\\tt " ; 'S(string_for_read (string_of_token rev_vernac_keywords t)) ;
       'S"}" >];;

let LaTeX_print_fnl () = [< 'FNL >];;
let LaTeX_print_spc () = [< 'SPC >];;

let LaTeX_print_gk whatfor (id1,id2) =
    if whatfor = id1 then
        [< LaTeX_stream (string_of_id id2) >]
    else
        [< LaTeX_stream (string_of_id id1) ; 'S":" ; LaTeX_stream(string_of_id id2) >]
;;

let LaTeX_print_prod_item whatfor = function
    NONTERM((gk,al),ido) ->
    [< 'S"{\\sl " ; LaTeX_print_gk whatfor gk ;
       (if al = [] then [< >]
        else [< 'S"[" ; prlist_with_sep (fun () -> [< 'S";" >])
                        fmt_act al ;
                'S"]" >]) ;
       (match ido with
        NONE -> [< 'S"}">]
      | SOME atid -> [< 'S"\\rm (" ; LaTeX_stream(string_of_id atid) ; 'S")}" >]) >]

  | TERM t -> LaTeX_print_token t
;;

let LaTeX_print_production whatfor = function
     (l,a) ->
     HOV 0 [< prlist_with_sep (fun () -> [< 'SPC >]) (LaTeX_print_prod_item whatfor) l ;
        'CUT ;  'S" \\\\" ; 'FNL ; 'S" & \\action{" ; fmt_act a ; 'S"}\\\\" >]
;;

let LaTeX_print_grammar_entry whatfor (id,(idl,pl)) =
    [< 'S"\\multicolumn{2}{l}{\\sl " ; LaTeX_print_gk whatfor id ;
       (if idl = [] then [< 'S"} \\\\" >]
        else [< 'S"[" ;
                prlist_with_sep (fun () -> [< 'S";" >])
                                (fun id -> [< 'S (string_of_id id) >])
                                idl ;
                'S"]}\\\\" >]) ;
       'FNL ; 'S"\\ra& " ; (prlist_with_sep LaTeX_print_bar (LaTeX_print_production whatfor) pl) >]
;;

let rec LaTeX_print_cgram = function
    c_ACTION a -> [< 'S" -> " ; fmt_act a >]
  | c_SWITCH(tl,el) ->
    [< 'S"{" ;
    HOV 0 [< prlist_with_sep (fun () -> [< 'S"|" ; 'FNL >])
                             (fun l -> HOV 0 (LaTeX_print_toks_line l)) tl ;
             if el <> [] then [< 'FNL >] else [< >] ;
             prlist_with_sep (fun () -> [< 'S"|" ; 'FNL >])
                             (fun l -> HOV 0 (LaTeX_print_eps_line l)) el >] ;
      'S"}" >]

and LaTeX_print_toks_line(toks,(pgl,cg)) =
    [< 'S"[" ; prlist_with_sep LaTeX_print_spc LaTeX_print_token toks ; 'S"]" ; 'SPC ;
       'S"[" ; prlist_with_sep LaTeX_print_spc LaTeX_print_pgram pgl ; 'S"]" ; 'FNL ;
       HOV 2 (LaTeX_print_cgram cg) >]

and LaTeX_print_eps_line(pgl,cg) =
    [< 'S"[" ; prlist_with_sep LaTeX_print_spc LaTeX_print_pgram pgl ; 'S"]" ; 'FNL ;
       HOV 2 (LaTeX_print_cgram cg) >]

and LaTeX_print_pgram = function
    c_NTERM ((gk,al),ido) ->
    [< LaTeX_print_gk (id_of_string "$BOGUS") gk ;
       (if al = [] then [< >]
        else [< 'S"[" ; prlist_with_sep (fun () -> [< 'S";" >])
                        fmt_act al ;
                'S"]" >]) ;
       (match ido with
        NONE -> [< >]
      | SOME atid -> [< 'S"(" ; 'S(string_of_id atid) ; 'S")" >]) >]
  | c_TERM t -> LaTeX_print_token t
;;


(*==============*)


(*=================
LaTeX output
Claude
*)


let LaTeX_print_nonterminal_definition gk =
let pl = productions_of gk and
    idl = params_of gk
in [< 'S"\\begin{tabular}{cp{140mm}}" ; 'FNL ;
      LaTeX_print_grammar_entry (fst gk) (gk,(idl,pl)) ;
      'FNL ; 'S"\\end{tabular}" ; 'FNL ; 'FNL >]
;;

let LaTeX_print_grammar_definition gr_id =
  do_print (mhs__dom grams)
where rec do_print = function
    [] -> [< >]
  | (g,p)::l -> if g=gr_id 
                   then [< LaTeX_print_nonterminal_definition (g,p) ; do_print l >]
                   else [<do_print l >]
;;

(*==============*)

let print_nonterminal_code gk =
    print_cgram (comp_gram_of gk)
;;

#open "reintern0";;
#open "libobject";;

let ri_token ri_keyword = function
    Tkw kw -> Tkw (ri_keyword kw)
  | x -> x
;;

let ri_gkey (id1,id2) = (ri_id id1,ri_id id2);;

let ri_prod_item = function
    NONTERM((gk,al),ido) -> NONTERM((ri_gkey gk,map ri_action al),
                                    option_app ri_id ido)
  | TERM tok -> TERM(ri_token (fun x -> x) tok)
;;

let ri_prod (l,a) = (map ri_prod_item l,ri_action a);;

let ri_grammar_entry (gk,(idl,pl)) =
    (ri_gkey gk,(map ri_id idl,map ri_prod pl));;

let pr_grammar_entry(gk,_) =
    [< 'S"#<grammar entry " ;
       'S(string_of_id (fst gk)) ;
       'S":" ; 'S(string_of_id (snd gk)) ;
       'S">" >]
;;

let ((inGRAMMAR_ENTRY:grammar_entry -> pridyn__t),
     (outGRAMMAR_ENTRY:pridyn__t -> grammar_entry)) =
    pridyn__new ("GRAMMAR_ENTRY",ri_grammar_entry,pr_grammar_entry)
;;


let cache_grammar (_,x) = pureparse__push x;;
let reintern_grammar gel = (map ri_grammar_entry gel);;
let specification_grammar x = x;;

let (inGrammar,outGrammar) =
    declare_object ("GRAMMAR",
                    {cache_function = cache_grammar;
                     reintern_function = reintern_grammar;
                     specification_function = specification_grammar});;

library__declare_summary "grammar"
{summary__freeze_function = freeze;
 summary__unfreeze_function = unfreeze;
 summary__init_function = init;
 summary__sync_function = synchronize_grammar_interpreter}
;;

(* We declare Tokens module here, instead of constr/environ.ml, to
 * answer a Sophia requirement. *)

(* The action corresponding to add a new token is done by calling add_token *)
let inToken,outToken =
  declare_object("TOKEN",
      	       	 {cache_function = (fun (_,s) -> lexer__add_token s);
		  reintern_function = (fun x -> x);
		  specification_function = (fun x -> x)})
;;

(* !token_tree corresponds exactly to the current state of the lexer *)
library__declare_summary "TOKEN-TABLE"
  {summary__freeze_function = lexer__freeze;
   summary__unfreeze_function = lexer__unfreeze;
   summary__init_function = (fun () -> ());
   summary__sync_function = (fun () -> false)}
;;


(* $Id: pureparse.ml,v 1.4 1996/02/12 15:45:47 cpaulin Exp $ *)
