(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                 stock.ml                                 *)
(****************************************************************************)
#open "std";;
#open "more_util";;
#open "pp";;
#open "stdpp";;
#open "lomsg";;
#open "lexer";;
#open "initial";;
#open "names";;
#open "library";;
#infix "o";;

(* The pattern table for tactics *)

(* The idea is that we want to write tactic files which are only
   "activated" when certain modules are loaded and imported.  Already,
   the question of how to store the tactics is hard, and we will not
   address that here.  However, the question arises of how to store
   the patterns that we will want to use for term-destructuring, and
   the solution proposed is that we will store patterns with a
   "module-marker", telling us which modules have to be open in order
   to use the pattern.  So we will write:

   let mark = make_module_marker ["<module-name>";<module-name>;...];;

   let p1 = put_pat mark "<parseable pattern goes here>";;

   And now, we can use:

             (get p1)

   to get the term which corresponds to this pattern, already parsed
   and with the global names adjusted.

   In other words, we will have the term which we would have had if we
   had done an:

        constr_of_com mt_ctxt (initial_sign()) "<text here>"

   except that it will be computed at module-opening time, rather than
   at tactic-application time.  The ONLY difference will be that
   no implicit syntax resolution will happen.

   So the entries we provide are:

   type module_mark;;

   value make_module_marker : string list -> module_mark;;

   type marked_term;;

   value put_pat : module_mark -> string -> marked_term;;

   value get_pat : marked_term -> constr;;

 *)

type module_mark = MMK of int;;
let path_mark_bij = (mhb__new 17 : (section_path list,module_mark) mhb__t);;

let make_module_marker stock sl =
let spl =
    (try map path_of_string sl
     with _ -> error "make_module_marker: malformed section_path in marker") in
let sorted_spl = sort sp_gt spl
in (try mhb__map path_mark_bij sorted_spl
    with Not_found ->
        let mmk = MMK(newMETA())
        in (mhb__add path_mark_bij (sorted_spl,mmk);
            mmk))
;;

let mark_satisfied mmk =
let spl = mhb__pam path_mark_bij mmk
in listset__subset spl (search_imports())
;;

(* src_tab: for each module_mark, stores the tickets of objects which
   need to be compiled when that mark becomes active.

   obj_tab: for each ticket, stores the (possibly nonexistent)
   compiled object

   ticket_tab: for each ticket, stores its module_mark and the string
   (source)

   string_tab: for each string * module_mark, stores the ticket.

*)
type 'a stock =
    {src_tab : (module_mark,int) mhs__t;
     obj_tab : (int,'a) mhm__t;
     ticket_string_bij : (int,string * module_mark) mhb__t;
     args : 'a stock_args}
;;

type 'a stocked == int;;

let make_stock args =

let stock =
    {src_tab = mhs__new (17,NONE);
     obj_tab = mhm__new 17;
     ticket_string_bij = mhb__new 17;
     args = args}

in stock
;;

let stock stock mmk s =
try mhb__pam stock.ticket_string_bij (s,mmk)
with Not_found ->
let idx = newMETA() 
in (mhs__add stock.src_tab (mmk,idx);
    mhb__add stock.ticket_string_bij (idx,(s,mmk));
    idx)
;;


let pr_mm mm =
let spl = mhb__pam path_mark_bij mm in
    prlist_with_sep (fun () -> [< 'SPC >])
    (fun p -> [< 'S(string_of_path p) >])
    spl
;;

let pr_spc () = [< 'SPC >];;
let pr_fnl () = [< 'FNL >];;
let pr_int n = [< 'INT n >];;

let retrieve stock idx =
try mhm__map stock.obj_tab idx
with Not_found ->
    let (s,mmk) = mhb__map stock.ticket_string_bij idx
    in if mark_satisfied mmk then
        try
        let c = stock.args.proc s
        in (mhm__add stock.obj_tab (idx,c);
            c)
        with e ->
            (MSGNL [< 'S"Processing of the stocked datum " ; 'S s ;
                     'S" failed: " ; errors__explain_user_exn e;
                     library__fmt_module_state() >];
             errors__reraise_user_exn e)
      else errorlabstrm "stock__retrieve"
          [< 'S"The stocked object " ; 'S s ; 'S" was not compilable" ; 'FNL ;
             'S"Its module mark was: " ; pr_mm mmk ; 'FNL ;
             library__fmt_module_state() >]
;;
        


(* $Id: stock.ml,v 1.2 1996/02/12 15:46:14 cpaulin Exp $ *)
