(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                summary.ml                                *)
(****************************************************************************)

#infix "o";;
#open "std";;
#open "pp";;
#open "stdpp";;
#open "names";;
#open "more_util";;

(* Imports Summary *)
let IMPORTED = ref ([] : section_path list);;
let search_imports () = !IMPORTED;;

let SUMMARIES = (mlm__new() : (string,dyn__t summary_declaration) mlm__t);;

type summary_vector == (string,dyn__t) listmap__t;;

let declare_summary sumname sdecl =

  let (infun,outfun) = dyn__new (sumname^"-SUMMARY") in

  let dyn_freeze () = infun(sdecl.freeze_function()) and
      dyn_unfreeze sum = sdecl.unfreeze_function(outfun sum) and
      dyn_init = sdecl.init_function and
      dyn_sync = sdecl.sync_function in

  let ddecl = {freeze_function = dyn_freeze;
               unfreeze_function = dyn_unfreeze;
               init_function = dyn_init;
               sync_function = dyn_sync} in

  try mlm__add SUMMARIES (sumname,ddecl)
  with Failure _ -> anomalylabstrm "summary__declare_summary"
                    [< 'S"Cannot declare a summary twice: " ; 'S sumname >]
;;

let freeze_summary () =
    map (fun (id,sdecl) -> (id,sdecl.freeze_function()))
        (mlm__toList SUMMARIES)
;;

let unfreeze_summary sum_vector =
    do_list (fun (id,sdecl) ->
                 if listmap__in_dom sum_vector id then
                     sdecl.unfreeze_function (listmap__map sum_vector id)
                 else sdecl.init_function())
            (mlm__toList SUMMARIES)
;;

let init_summaries () =
    mlm__app (fun (_,sdecl) -> sdecl.init_function()) SUMMARIES;;

let resync_summaries () =
    let synced =
        map (fun (_,sdecl) -> sdecl.sync_function()) (mlm__toList SUMMARIES)
    in exists (fun x -> x) synced
;;

let SUMMARY_STACK = ref ([] : (section_path * summary_vector) list);;

let init_caches () =
    init_summaries();
    SUMMARY_STACK := [];;

let resync_caches = resync_summaries;;

let discard_to_key_including sp = (SUMMARY_STACK := discrec !SUMMARY_STACK)
  where rec discrec = function
    []                -> anomaly "summary__discard_to_key_including"
  | ((sp',_)::l as L) -> if sp = sp' then l else discrec l
;;

(* Freeze the state of the summaries, storing it under the given key,
   in the summary-stack *)
let freeze_to_key sp =
    if listmap__in_dom !SUMMARY_STACK sp then
        discard_to_key_including sp;
    resync_caches();
    SUMMARY_STACK := (sp,freeze_summary()):: !SUMMARY_STACK
;;

let unfreeze_from_key sp =
  let sum = try (listmap__map !SUMMARY_STACK sp)
            with Not_found -> failwith "summary__unfreeze_from_key"
  in try unfreeze_summary sum
     with e -> anomalylabstrm "summary__unfreeze_from_key"
                         [< 'S"Exception in unfreezing summary: " ; 'CUT ;
               errors__explain_sys_exn e >]
;;

type frozen_t == (section_path * summary_vector) list;;

let get_frozen_summaries() = !SUMMARY_STACK;;
let set_frozen_summaries fs = SUMMARY_STACK := fs;;

let discard_to_key_excluding sp = (SUMMARY_STACK := discrec !SUMMARY_STACK)
  where rec discrec = function
    []                -> anomaly "summary__discard_to_key_excluding"
  | ((sp',_)::l as L) -> if sp = sp' then L else discrec l
;;

let find_latest () =
  let cd = rev (lib__cwd()) in
  let sp = try_find
        (fun (sp,_) ->
             if prefix_of (rev (dirpath sp)) cd then
                 try lib__map sp ; sp
                 with _ -> failwith "caught"
             else failwith "caught")
      	!SUMMARY_STACK
  in sp
;;

(* will unfreeze the latest summary in the stack, discarding any which
   come after it.  We return the section_path of that summary. *)
let unfreeze_latest sp =
    unfreeze_from_key sp;
    discard_to_key_excluding sp
;;

declare_summary "imported"
  {freeze_function = (fun () -> !IMPORTED);
   unfreeze_function = (fun l -> IMPORTED := l);
   init_function = (fun () -> IMPORTED := []);
   sync_function = (fun () -> false)}
;;

let list_frozen_summaries () = map fst !SUMMARY_STACK;;

(* $Id: summary.ml,v 1.3 1996/02/12 15:46:19 cpaulin Exp $ *)
