(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                termact.ml                                *)
(****************************************************************************)
#infix "o";;
#open "std";;
#open "names";;
#open "more_util";;
#open "lexer";;
#open "ast";;
#open "pp";;
#open "stdpp";;
#open "lomsg";;

let mt = [];;

let dollarzero = id_of_string "$0";;

let isMeta id =
    let s = SA__un_intern (atompart_of_id id)
    in string_length s <> 0
       & ("$" = (sub_string s 0 1));;

let push id v stk = (id,v)::stk;;

let lookup_binding b pat =
    try listmap__map b pat
    with Not_found ->
        error ("lookup of "^(string_of_id pat)^" for a term failed")
;;

let strip_to_value b =
    try listmap__map b dollarzero
    with Not_found ->
        errorlabstrm "termact"
        [< 'S"Interpretation of action did not bind $0" >]
;;

let apply_pattern b pat =
    ast__ssub b (raise_spat_bound_indices pat)
;;

let first_ap f = aux
    where rec aux = function
    [] -> error "Parsing failed in case-split"
  | (h::t) -> try f h with Failure _ -> aux t
;;

let parse_semi = function
    [< 'Tsemi >] -> ()
;;

let parse_bar = function
    [< 'Tbar >] -> ()
;;

let parse_binding parse_pat = function
    [< 'Tident at ; 'Tequal ; parse_pat pat >] -> 
      let atid = id_of_string at 
      in if isMeta atid then (atid,pat)
         else error ("ident found which was not a metavariable: "^at)
;;

let parse_act parse_pat =
let rec parse_act = function
    [< 'Tlbracket ;
       (list_with_sep parse_semi (parse_binding parse_pat)) bl ;
       'Trbracket >]
    -> ACTL bl
  | [< 'Tident "case" ; 'Tident at ; 'Tident "of" ;
        (list_with_sep parse_bar parse_case_branch) cases;
        'Tident "esac" >] -> 
     let atid = id_of_string at 
      in if isMeta atid then CASE(atid,cases)
         else error ("ident found which was not a metavariable: "^at)   

  | [< 'Tlparen ; parse_act act ; 'Trparen >] -> act

and parse_case_branch = function
    [< parse_pat pat ; 'Tminusgreater; parse_act act >] -> (pat,act)
in parse_act
;;

let fmt_binding  (at,pat) =
    [< 'S(string_of_id at);
     'S"="; print_spat pat >]
;;

let fmt_act =
let rec fmt_act = function
    ACTL l ->
        HOV 0 [< 'S"[" ;
                 prlist_with_sep (fun () -> [< 'S";" ; 'FNL >]) fmt_binding l;
                 'S"]" >]

  | CASE(at,branches) ->
        [< 'S"case"; 'WS 1 ; 'S(string_of_id at);
           'WS 1; 'S"of" ; 'BRK(1,1);
           HOV 0 (prlist_with_sep (fun () -> [< 'CUT ; 'S"|" ; 'WS 1 >])
                  fmt_case_branch branches);
           'CUT ; 'S"esac" >]

and fmt_case_branch = fun (pat,act) ->
    HOV 0 [< print_spat pat ; 'WS 1 ; 'S"->" ; 'SPC ;
             HOV 0 [< fmt_act act >] >]
in fmt_act
;;

let interp_binding b (id,pat) =
    push id (apply_pattern b pat) b;;

let interp_act a b =
let rec interp b a =
    match a with
        ACTL l -> strip_to_value(it_list interp_binding b l)

      | CASE(id,cases) ->
            let v = lookup_binding b id in
            let (sigma,act) =
                (first_ap (fun (p,act) -> (smatch p v,act)) cases)
            in interp (sigma@b) act in
try
    interp b a
with Failure _ ->
    (MSGERRNL [< 'S"Interpretation of action: "; fmt_act a; 'SPC ;
                 'S"with env " ; 'FNL ;
                 prlist_with_sep (fun () -> [< 'FNL >])
                 (fun (id,f) -> [< print_id id ; 'S" := " ; print_frag f >])
                 b ; 'S" failed" >];
    raise Parse_error)
   | Not_found ->
     (MSGERRNL [< 'S"Interpretation of action: "; fmt_act a; 'SPC ;
                 'S"with env " ; 'FNL ;
                 prlist_with_sep (fun () -> [< 'FNL >])
                 (fun (id,f) -> [< print_id id ; 'S" := " ; print_frag f >])
                 b ; 'FNL;'S" failed - perhaps there is an unbound metavariable" >];
     raise Parse_error)
;;

#open "reintern0";;
let ri_binding (id,smc) = (ri_id id,ri_spat smc);;


let ri_nonterm_map m = map (fun (id,nt) -> (ri_id id,ri_frag nt)) m;;

let rec ri_action = function
    ACTL bindl -> ACTL(map ri_binding bindl)
  | CASE(id,smc_act_l) -> CASE(ri_id id,
                               map (fun (smc,act) -> (ri_spat smc,
                                                      ri_action act))
                               smc_act_l)
;;

(* $Id: termact.ml,v 1.2 1996/02/12 15:46:28 cpaulin Exp $ *)
