(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                              unparsings.ml                               *)
(****************************************************************************)
#open "std";;
#open "names";;
#open "more_util";;
#open "pp";;
#open "stdpp";;
#open "lomsg";;
#open "lexer";;

let pp_ppcmd = function
    ppHB n -> HB n
  | ppHOVB n -> HOVB n
  | ppHVB n -> HVB n
  | ppVB n -> VB n
  | ppCLOSE -> CLOSE
  | ppBRK(n,m) -> BRK(n,m)
  | ppFNL -> FNL
;;

let box_kind = function
    [< 'Tident "h" >] -> (fun n -> ppHB n)
  | [< 'Tident "v" >] -> (fun n -> ppVB n)
  | [< 'Tident "hv" >] -> (fun n -> ppHVB n)
  | [< 'Tident "hov" >] -> (fun n -> ppHOVB n)
;;

let box = function
    [< 'Tless; box_kind bk; 'Tint n; 'Tgreater >] -> bk n;;

let paren_reln = function
    [< 'Tident "L" >] -> L
  | [< 'Tident "E" >] -> E
  | [< 'Tstar >] -> Any
;;

let rec next_simple_hunk l = function
    [< 'Tstring c ; s >] -> next_simple_hunk (c::l) s
  | [< >] -> (implode(rev l))
;;

let rec next_hunks = function
    [< 'Tstring c ; s >] -> [RO(next_simple_hunk [c] s)]
  | [< 'Tless ; 'Tident id ; 'Tcolon ;
       (next_simple_hunk []) r ; 'Tcolon ; paren_reln pr ; 'Tgreater >] ->
    [PH(id_of_string id,"<"^r^">",pr)]
  | [< 'Tlbracket ;
       begin function
        [< box b ; (plist next_hunks) ll ; 'Trbracket >] ->
            [UNP_PPTOKEN b]@(join ll)@[UNP_PPTOKEN ppCLOSE]
      | [< 'Tint n ; 'Tint m; 'Trbracket >] ->
            [UNP_PPTOKEN(ppBRK(n,m))]
    end x>] -> x
 | [< 'Tident "FNL" >] -> [UNP_PPTOKEN ppFNL]
;;

let compare_prec (a1,b1,c1) (a2,b2,c2) =
    match (a1=a2),(b1=b2),(c1=c2) with
    true,true,true -> 0
  | true,true,false -> c1-c2
  | true,false,_ -> b1-b2
  | false,_,_ -> a1-a2
;;

let unparsing = function
    [< 'Tlbracket ; 'Tint a ; 'Tint b ; 'Tint c ; 'Trbracket ;
       (plist next_hunks) ll >] -> {unp = join ll; prec = (a,b,c)}
  | [< 'Tint a ; (plist next_hunks) ll >] -> {unp = join ll; prec = (a,0,0)};;

let apply_unparsing dflt (dfname,unp) submap = 
    (unpcheck unp.unp;
     unprec unp.unp)
    where rec unprec = function
    [] -> [< >]
  | (PH(id,_,pr))::t -> [< (try (listmap__map submap id)((dfname,unp.prec),pr)
                            with Not_found -> dflt);
                           unprec t >]
  | (RO s)::t -> [< 'S s ; unprec t >]
  | (UNP_PPTOKEN ppc)::t -> [< '(pp_ppcmd ppc) ; unprec t >]

    and unpcheck = function
    [] -> ()
  | (PH(id,_,pr))::t ->
    if not listmap__in_dom submap id then
        (WARN [< 'S"The variable " ; 'S(string_of_id id) ; 'S" was not bound in " ; 
           prlist_with_sep (fun () -> [< 'FNL >])
           (fun (id,subpr) -> [< 'S(string_of_id id) ; 'S":=" ;
                                 subpr((dfname,unp.prec),pr) >])
           submap >];
        unpcheck t)
        else unpcheck t
  | (RO s)::t -> unpcheck t
  | (UNP_PPTOKEN ppc)::t -> unpcheck t

;;

#open "lexer";;
type lat_class == string list;;
type lat_chain == lat_class list;;
type lat_level == lat_chain list;;
type lat_levels == lat_level list;;
type lattice == lat_levels;;
let mt_lattice = [];;

type mapping_entry = {class:lat_class;
                      chain:lat_chain;
                      level:lat_level;
                      levels:lat_levels};;

type compiled_lattice == (string * (mapping_entry list ref)) list;;

let new_mapping clatref s =
    if not (in_dom s (!clatref)) then
        clatref := ((s,ref [])::(!clatref))
    else failwith "match"
;;

let init_mappings clatref = (clatref := []);;

let get_mappings clat s = !(assoc s clat);;

let rec push_mapping clatref (s,m) =
    try (let r = assoc s (!clatref) in
        r := m::(!r))
    with _ -> (new_mapping clatref s;push_mapping clatref (s,m))
;;

let comp_lattice levels =
let clatref = ref []
in
     app (function level ->
          app (function chain ->
               app (function class ->
                    app (function addr ->
                         push_mapping clatref
                                     (addr,{class=class;
                                            chain=chain;
                                            level=level;
                                            levels=levels}))
                        class)
                   chain)
              level)
         levels;
    !clatref
;;

let undeclared_dforms clat all_dforms =
    filter (function s -> not(in_dom s clat)) all_dforms
;;

let same_at clat accessor a b =
    exists
    (function map_a ->
     exists
     (function map_b ->
      (accessor map_a) = (accessor map_b))
     (get_mappings clat b))
    (get_mappings clat a)
;;

let different_at clat accessor a b =
    exists
    (function map_a ->
     exists
     (function map_b ->
      (accessor map_a) <> (accessor map_b))
     (get_mappings clat b))
    (get_mappings clat a)
;;

let earlier_in_seq seq a b =
    try (index a seq) < (index b seq) with
    _ -> false
;;

let earlier_at clat accessor parent_accessor a b =
    exists
    (function map_a ->
     exists
     (function map_b ->
      let acc_a = accessor map_a in
      let acc_b = accessor map_b in
      let par_acc_a = parent_accessor map_a in
      let par_acc_b = parent_accessor map_b in
          (par_acc_a = par_acc_b) & (earlier_in_seq par_acc_a acc_a acc_b))
     (get_mappings clat b))
    (get_mappings clat a)
;;

let different_child_of clat accessor parent_accessor a b =
    exists
    (function map_a ->
     exists
     (function map_b ->
      let acc_a = accessor map_a in
      let acc_b = accessor map_b in
      let par_acc_a = parent_accessor map_a in
      let par_acc_b = parent_accessor map_b in
          (par_acc_a = par_acc_b) & (acc_a <> acc_b))
     (get_mappings clat b))
    (get_mappings clat a)
;;

let irrit_eq_rule_1 clat a b =
    same_at clat (function m -> m.class) a b;;

let irrit_lt_rule_2 clat a b =
    earlier_at clat (function m -> m.class) (function m -> m.chain) b a;;

let irrit_unrelated_rule_3 clat a b =
    different_child_of clat (function m -> m.chain) (function m -> m.level) a b;;

let irrit_lt_rule_4 clat a b =
    earlier_at clat (function m -> m.level) (function m -> m.levels) b a;;

(* type irrit_factor = Less | Greater | Equal | Unrelated;;*)

let irrit_factor clat a b =
    if irrit_eq_rule_1 clat a b then Equal
    else if irrit_lt_rule_2 clat a b then Less
    else if irrit_lt_rule_2 clat b a then Greater
    else if irrit_unrelated_rule_3 clat a b then Unrelated
    else if irrit_lt_rule_4 clat a b then Less
    else if irrit_lt_rule_4 clat b a then Greater
    else Unrelated
;;

let make_lattice l undeclared =
    l@[[[undeclared]]]
;;

let parse_lattice =
let p_class =
    function [< 'Tlparen ; (plist (function [< 'Tident s >] -> s)) addresses ;
              'Trparen >] -> addresses in
let p_chain =
    function [< 'Tlbracket ; (plist p_class) classes ;
              'Trbracket >] -> classes in
let p_level =
    (function [< 'Tlparen ; (plist p_chain) chains ;
              'Trparen >] -> chains)
in
    (function [< 'Tlparen ; (plist p_level) levels ;
               'Trparen >] -> levels)
;;

let print_lattice lat =
let space() = [< 'SPC >] 
and newline() = [< 'FNL >]
and two_newlines() = [< 'FNL ; 'FNL >] 
and pr_address s = [< 'S s ; 'CUT >] in
let pr_class cl =
    HOV 0 [< 'S"(" ; prlist_with_sep space pr_address cl ; 'S")" >] in
let pr_chain ch =
    HOV 0 [< 'S"[" ; prlist_with_sep space pr_class ch ; 'S"]" >] in
let pr_level lev =
    HOV 0 [< 'S"(" ; prlist_with_sep newline pr_chain lev ; 'S")" >]
in
    prlist_with_sep two_newlines pr_level lat
;;


let compile_lattice lattice all_dforms =
let clat = comp_lattice lattice in
let lattice' = make_lattice lattice (undeclared_dforms clat all_dforms)
in comp_lattice lattice'
;;

let tolerable_prec parent_prec child_prec = function
    L -> (compare_prec child_prec parent_prec) < 0
  | E -> (compare_prec child_prec parent_prec) <= 0
  | Any -> true
;;

let tolerable clatopt (parent_name,parent_prec) ((child_name,child_prec),paren_reln) =
match clatopt with
NONE -> tolerable_prec parent_prec child_prec paren_reln
|SOME clat ->
try
    let factor = irrit_factor clat parent_name child_name in
        match paren_reln with
        L -> factor = Less
      | E -> factor = Equal or factor = Less
      | Any -> true
with Not_found -> (warning ("[Undeclared dforms: "^parent_name^" "^child_name^"]");
                   true)
;;

#open "reintern0";;
let ri_unparsing {unp=unp;prec=prec} =
    {prec=prec;
     unp= map (fun (PH(id,s,pr)) -> PH(ri_id id,s,pr)
                 | (RO s) -> RO s
                 | (UNP_PPTOKEN ppc) -> UNP_PPTOKEN ppc) unp}
;;

let pr_unparsing unp = [< 'S"<unparsing>" >];;

let ((inUNPARSING:unparsing->pridyn__t),
     (outUNPARSING:pridyn__t->unparsing)) =
    pridyn__new ("UNPARSING",ri_unparsing,pr_unparsing)
;;

(* $Id: unparsings.ml,v 1.2 1996/02/12 15:46:33 cpaulin Exp $ *)
