(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                              extraction.ml                               *)
(****************************************************************************)

#open "more_util";;
#open "std";;
#open "names";;
#open "impuniv";;
#open "generic";;
#open "term";;
#open "reduction";;
#open "typing";;
#open "proof";;
#open "termenv";;
#open "vectops";;

(* The constructive engine can be used for two purposes :
    - checking that a "logical" proof is well-formed.
    - writing a "constructive" proof and extracting its computational meaning
                (an Fomega + Ind + Fix + Implicit program)
Each term c of the Calculus of Construction contains a computational 
information that is :
    Inf(Judge(c',t,k))if c is already typed in Fw and if c' is its translation 
                      as an fterm.
    Logic             if c is a "logical" term (without computational meaning).

The information is initialized at the level of propositional variables :
    A : Prop(Null)   A is a logical propositional variable.
    A : Prop(Pos)   A is a constructive propositional variable.

This file contains the rule for computing the information during an 
application, generalization or abstraction.

When processing a term c equal to  (M N) or (x:N)M or [x:N]M,
three cases are distinguished :
  If M is a logical term then c is also a logical term.
  If M is a constructive term then depending of the level of N and of its
            information, the information of c will be the same as the 
            information of M or computed using the fterm extracted from N.

*)

(* a special constant to solve the problem of extracting informations 
where it is not compatible with the current type checking algorithm *)

let implicit = (DOP0(Implicit): constr);;

let le_kind_implicit k1 k2 = (k1=implicit) or (isprop k1)
 or (k2=implicit) or (istype k2);;

let rec is_implicit = function 
    DOP0(Implicit) -> true 
  | DOPN(AppL,cl) -> is_implicit cl.(0)
  | DOPN(MutCase,cl) -> is_implicit cl.(1)
  | DOP2(Cast,c,_) -> is_implicit c
  | DOP2(Prod,_,c) -> is_implicit c
  | DOP2(Lambda,_,c) -> is_implicit c
  | DLAM(_,c) -> is_implicit c
  | DOPN(Fix(_,i),cl) -> is_implicit_proj i (last_vect cl)
  | DOPN(CoFix(i),cl) -> is_implicit_proj i (last_vect cl)
  | _ -> false
and is_implicit_proj i = function 
    DLAM(_,c) -> is_implicit_proj i c
  | DLAMV(_,c) -> is_implicit c.(i)
;;
  

let inf_app f = function Inf (c) -> Inf (f c) | x -> x;;

let kind_of_inf (Inf j) = j.KIND;;

let inf_var env cts n =
    match cts with
    Null -> Logic
  | _ ->
    let (DOP2(Cast,typ,kind)) = snd(lookup_glob n env)
    in Inf{VAL=VAR n;
           TYP= typ;
           KIND = kind}
;;
let inf_sort = function
    Prop Null -> Logic 
  | (Prop Pos as t)  ->
    Inf{VAL= DOP0(Sort t);
        TYP= DOP0(Sort(Type(Pos,prop_univ)));
        KIND = DOP0(Sort(Type(Pos,prop_univ_univ)))}
  | Type(Null,_) -> Logic
  | Type(Pos,u) as t ->
    Inf{VAL= (DOP0(Sort t));
        TYP= DOP0(Sort(Type(Pos,u)));
        KIND = DOP0(Sort(Type(Pos,u)))}
;;

let inf_apply = fun
    Logic _ -> Logic
  | (Inf _ as funinf) Logic -> funinf
  | (Inf c1 as funinf) (Inf c2) ->
    if le_kind_implicit c1.KIND c2.KIND then
        let app_val = applist(c1.VAL,[c2.VAL])
        and app_ty = let typ = whd_betadeltaiota mt_evc c1.TYP 
                     in (match typ with DOP2(Prod,_,_) ->
                        hnf_prod_app mt_evc "inf_apply" typ c2.VAL
                          |_ -> if is_implicit typ then implicit 
                        else 
                     (* found a non-product type which was not implicit *)
                        hnf_prod_app mt_evc "inf_apply" typ c2.VAL)
        in Inf{VAL= app_val ;
               TYP= app_ty;
               KIND = c1.KIND}
    else funinf
;;

(* Cumulativity: slide type t up if product with second type imposes it *)
let adjust t = function
    DOP0(Sort(Prop(_)))    -> t
  | DOP0(Sort(Type(_,u'))) -> (match t with 
         DOP0(Sort(Type(c,u))) -> let v = sup(u,u') in if eq(u,v) then t else DOP0(Sort(Type(c,v)))
       | _ -> invalid_arg "adjust")
  | _          -> invalid_arg "level_of_kind/adjust";;

let type_of_product sigma domtyp rangtyp =
    match whd_betadeltaiota sigma rangtyp with
    DOP0(Sort(Prop _)) -> rangtyp
  | (DOP0(Sort(Type _))) as typ' -> adjust typ' domtyp
  | (DOP0(Implicit)) -> rangtyp
  | _ -> error "type_of_product: Not a sort"
;;

let type_of_type = function
    DOP0(Sort(Type(cts,u))) -> DOP0(Sort(Type(cts,Super u)))
  | DOP0(Sort(Prop cts)) -> DOP0(Sort(Type(cts,prop_univ)))
  | DOP0(Implicit) -> implicit
  | _ -> invalid_arg "type_of_type called with something which wasn't Type"
;;

let j_val j = DOP2(Cast,j.VAL,j.TYP);;

let rec sort_of_arity sigma c =
     match whd_betadeltaiota sigma c with
     DOP0(Sort(Prop(_))) as c'      -> c'
   | DOP0(Sort(Type(_,_))) as c'      -> c'
   | DOP2(Prod,_,DLAM(_,c2)) -> sort_of_arity sigma c2
   | DOP0(Implicit) -> implicit
   | _            -> error "sort_of_arity: Not an arity";;

let inf_gen_rel name = fun
    Logic _ -> Logic
  | (Inf c) Logic -> Inf(j_pop c)
  | (Inf c) (Inf tv) ->
    if le_kind_implicit c.TYP tv.TYP then
        let res_type = type_of_product mt_evc (whd_betadeltaiota mt_evc tv.TYP) (whd_betadeltaiota mt_evc c.TYP)
        in Inf{VAL = DOP2(Prod,tv.VAL,DLAM(name,c.VAL));
               TYP = res_type;
               KIND = type_of_type res_type}
    else Inf(j_pop c)
;;

let abs_compat = fun
    Logic _ -> anomaly "abs_compat"
  | _ Logic -> false
  | (Inf main) (Inf var) -> le_kind_implicit main.KIND var.TYP
;;

let inf_abs_rel name = fun
    Logic _ -> Logic
  | (Inf c) Logic -> Inf(j_pop c)
  | (Inf c) (Inf tv) ->
    if le_kind_implicit c.KIND tv.TYP then
        let abs_val = (DOP2(Lambda,tv.VAL,DLAM(name,c.VAL)))
        and abs_ty = DOP2(Prod,tv.VAL,DLAM(name,c.TYP))
        and abs_tyty = type_of_product mt_evc (whd_betadeltaiota mt_evc tv.TYP) (whd_betadeltaiota mt_evc c.KIND)
        in Inf{VAL= abs_val;
               TYP= abs_ty;
               KIND = abs_tyty}
    else Inf(j_pop c)
;;

let inf_abs_rel_count name = fun
    Logic _ -> Logic,ERASE
  | (Inf c) Logic -> Inf(j_pop c),ERASE
  | (Inf c) (Inf tv) ->
    if le_kind_implicit c.KIND tv.TYP then
        let abs_val = (DOP2(Lambda,tv.VAL,DLAM(name,c.VAL)))
        and abs_ty = DOP2(Prod,tv.VAL,DLAM(name,c.TYP))
        and abs_tyty = type_of_product mt_evc (whd_betadeltaiota mt_evc tv.TYP) (whd_betadeltaiota mt_evc c.KIND)
        in Inf{VAL= abs_val;
               TYP= abs_ty;
               KIND = abs_tyty},ABSTRACT
    else Inf(j_pop c),ERASE
;;

let inf_abs_var name varty inf =
let varinf = (Inf{VAL=varty;TYP=cast_type varty;
                  KIND = type_of_type (cast_type varty)})
in (inf_abs_rel (Name name) (inf_app (j_subst_var name) inf) varinf,
    if abs_compat inf varinf then
        ABSTRACT
    else ERASE)
;;

let inf_rel env cts d =
    match cts with
    Null -> Logic
  | _ ->
    let (DOP2(Cast,typ,kind)) = lift d (snd(lookup_rel d env))
    in Inf{VAL = Rel d;
           TYP = typ;
           KIND = kind}
;;

(* Inductive types in Fw *)

let extract = function Inf(c) -> c | _ -> anomaly "not informative";;
let extract_val = function Inf(c) -> c.VAL | _ -> anomaly "not informative";;
let extract_typ = function Inf(c) -> c.TYP | _ -> anomaly "not informative";;
let extract_kind = function Inf(c) -> c.KIND | _ -> anomaly "not informative";;

let inf_mutcase env sigma ct Pj cinf lfinf = 
    match cinf with Inf cj -> 
    (try if vect_length lfinf = 1 then 
    (* check if it is a singleton *)
       let (DOPN(MutInd (sp,_),_) as mind,_) = find_mrectype sigma ct in
       let mis = mind_specif_of_mind mind in 
       match mis_singl mis with 
          NONE -> raise Singleton 
        | SOME _ ->  let f = extract_val (hd_vect lfinf) 
            in  Inf{VAL= apply f cj.VAL;
                    TYP= Pj.VAL;
                    KIND = Pj.TYP}
       else raise Singleton
     with Singleton -> 
     let (rec_ty,kind) = 
     if is_implicit Pj.TYP then (implicit,implicit)
     else ((if is_implicit cj.TYP then implicit
           else type_case env sigma cj.TYP Pj.TYP  Pj.VAL cj.VAL),
                sort_of_arity mt_evc Pj.TYP)
     in let val = if istype cj.TYP or istype kind then implicit else 
        DOPN(MutCase,concat_vect [|Pj.VAL;cj.VAL|]
                                   (map_vect extract_val lfinf))
     in   Inf{VAL=val; TYP = rec_ty; KIND = kind})
   | _ -> if vect_length lfinf = 1 
          then let jf = lfinf.(0) in 
              if conv_x mt_evc Pj.VAL (extract_typ jf) then jf 
              else anomaly "The elimination should be on an informative term"
          else anomaly "The elimination should be on an informative term";;

let inf_fix sigma i Aityp lfi inflars infdefs_ind = 
(* collect the informative components, contains at least one of them Ai *)
let rec collect  k (lind,lf,larv,ldefj) = 
    function [] -> (lind,rev lf,larv,ldefj)
         | (fi::lfrest) -> (match inflars.(k) with 
              Inf Ark -> let (infdef,indk) = infdefs_ind.(k) 
                in collect (k-1)
               (indk::lind,fi::lf,Ark.VAL::larv,(extract infdef)::ldefj) lfrest
             | Logic   -> collect (k-1) (lind,lf,larv,ldefj) lfrest)

in let (lind,lf,larv,ldefj) = collect (length lfi -1) ([],[],[],[]) lfi
in let vdefj = vect_of_list ldefj
in let newdef = put_DLAMSV lf (map_vect j_val vdefj) 
in let fix = DOPN(Fix(vect_of_list lind,i),
                  concat_vect (vect_of_list larv) [|newdef|])     
   in Inf{VAL=fix;
          TYP=type_fix sigma fix (map_vect (fun j -> j.TYP) vdefj);
          KIND=Aityp};;
   

let inf_cofix sigma i Aityp lfi inflars infdefs  = 
let rec collect  k (lf,larv,ldefj) = 
    function []           -> (rev lf,larv,ldefj)
             | (fi::lfrest) -> 
                    match inflars.(k) with 
                          Inf Ark ->  collect (k-1) 
                                              (fi::lf,Ark.VAL::larv,
                                                 (extract infdefs.(k))::ldefj) 
                                              lfrest
                       | Logic   -> collect (k-1) (lf,larv,ldefj) lfrest

in let (lf,larv,ldefj) = collect (length lfi -1) ([],[],[]) lfi
in let vdefj = vect_of_list ldefj
in let newdef = put_DLAMSV lf (map_vect j_val vdefj) 
in let cofix = DOPN(CoFix i,concat_vect (vect_of_list larv) [|newdef|])     
   in Inf{VAL=cofix;
          TYP=type_cofix sigma cofix (map_vect (fun j -> j.TYP) vdefj);
          KIND=Aityp};;

(* $Id: extraction.ml,v 1.8 1996/04/03 12:52:48 cpaulin Exp $ *)
