(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                               indtypes.ml                                *)
(****************************************************************************)

#infix "o";;
#open "pp";;
#open "std";;
#open "names";;
#open "vectops";;
#open "impuniv";;
#open "extraction";;
#open "generic";;
#open "term";;
#open "constrtypes";;
#open "reduction";;
#open "typing";;
#open "termenv";;
#open "himsg";;
#open "stdpp";;
#open "more_util";;
#open "proof";;
#open "mach";;
#open "machops";;
#open "dischcore";;

let cast_of_judgement = machops__cast_of_judgement;;

let casted_execute_rec sigma = core_fmachine false (sigma,[]);;

(* WITH INFORMATION *)
let casted_infmachine nocheck (sigma,metamap) sign fsign constr =
let constr_metamap = map (fun (id,(c,_)) -> (id,c)) metamap
and inf_metamap = map (fun (id,(_,i)) -> (id,i)) metamap in
let j = core_fmachine nocheck (sigma,constr_metamap) sign constr in

let inf' = 
if is_info_sort j.KIND then 
    core_infmachine (sigma,(constr_metamap,inf_metamap)) sign fsign (j_val j)
else Logic
in (j,inf')
;;

let casted_infexecute_rec sigma (sign,fsign) c =
    casted_infmachine false (sigma,[]) sign fsign c;;
let casted_infexecute sigma (sign,fsign) c =
    casted_infmachine false (sigma,[]) (GLOB sign) (GLOB fsign) c;;

let casted_execute sigma sign c =
    core_fmachine false (sigma,[]) (GLOB sign) c;;

let extract_of sign fsign c = 
let (_,infc) = infexecute mt_evc (sign,fsign) c in 
match infc with Logic -> anomaly"extract_of"
             | Inf j  -> j_val j;;

(* In this part we can assume the context of existential variables is empty *)

let whd_betadeltaiota_empty = whd_betadeltaiota mt_evc;;

let assumption_of_cast = function 
    DOP2(Cast,val,typ) -> DOP2(Cast,val,whd_betadeltaiota mt_evc typ);;

(* reduce a term to its hnf form keeping casts *)
let assumption_of_cast = function 
    DOP2(Cast,val,typ) -> DOP2(Cast,val,whd_betadeltaiota mt_evc typ);;

let rec decomp_DLAMV_sign sign = function 
     ([(DOP2(Cast,_,_) as typ)],DLAMV(na,lc)) ->  
           let var = assumption_of_cast typ in (add_rel (na,var) sign,lc)
   | ((DOP2(Cast,_,_) as typ)::rest,DLAM(na,lc)) -> 
           let var = assumption_of_cast typ in decomp_DLAMV_sign 
           (add_rel (na,var) sign) (rest,lc)         
   | _ -> error "decomp_DLAMV_sign 1";;

let rec hnf_recast sign c = match whd_betadeltaiota mt_evc c with 
   DOP2(Prod,(DOP2(Cast,_,_) as b),DLAM(name,c)) -> 
        let var = assumption_of_cast b in 
        DOP2(Prod,b,DLAM(name,hnf_recast (add_rel (name,var) sign) c))
 | DOP2(Prod,b,DLAM(name,c)) -> 
        let jb = execute_rec mt_evc sign (strong strip_outer_cast b) in 
        let b' = assumption_of_judgement mt_evc jb in 
        DOP2(Prod,b',DLAM(name,hnf_recast (add_rel (name,b') sign) c))
 | x -> x;;


(**** Check a mutually inductive definition is well-formed ***)


(* takes the arity (c) of an inductive definition with the number of parameters
   and compute the number of parameters of the extracted inductive definition
   plus compute the "lamexp" which is of the form 
    [x1]...[xn]((Rel n+1) xi1 .. xik)
   the idea is that if Constr(i):(y1:A1)...(yn:An)B
*)

let inf_minductive sign fsign nparams arity C checksingl =

let sort = snd (decomp_prod mt_evc arity) in

let infexec sign' fsign' c = let cj = casted_execute_rec mt_evc sign' c 
    in if is_info_sort cj.KIND then 
       core_infmachine (mt_evc,([],[])) sign' fsign' (j_val cj) 
       else Logic
and infexec_cast sign' fsign' = core_infmachine (mt_evc,([],[])) sign' fsign' in

let extract_na_ty_pop c =
        (match whd_betadeltaiota_empty c with
         DOP2(Prod,T,DLAM(na,B)) -> (na,T),B
       | _ -> error "infexecute") in

let rec descend e fe n C = 
 if n = 0 then 
     let singl = if checksingl & isprop sort then 
        let infC = infexec e fe C in
        (match whd_betadeltaiota_empty (j_val (extract infC)) with 
           DOP2(Prod,A,DLAM(_,x)) -> 
               (match whd_betadeltaiota_empty x with 
                 (Rel _ | DOPN(AppL,_)) ->
                   if noccurn (nparams+1) A then 
                   let tA = unsafe_type_of mt_evc fe A in 
                    if conv_x mt_evc tA sort then SOME(A) else NONE
                   else NONE
                | _ -> NONE)
          | _ -> NONE)
                else NONE in (0,[],[],singl)
 else let (na,T),newC = extract_na_ty_pop C in
      let Tj = casted_execute_rec mt_evc e T in
      let infT = if is_info_sort Tj.KIND then infexec_cast e fe (j_val Tj) 
                 else Logic in
      let (inf_nparams,constrlamenv,argsign,singl) =
          descend (add_rel (na,j_val Tj) e) (add_inf_rel (na,infT) fe) (n-1) newC
      in (match infT with
            Logic -> (inf_nparams, Logic::constrlamenv,argsign,singl)
          | (Inf infTj) ->
            if isprop infTj.TYP 
            then (inf_nparams,infT::constrlamenv,ERASE::argsign,singl)
            else (inf_nparams+1,infT::constrlamenv,ABSTRACT::argsign,singl)) in
    
let (inf_nparams,constrlamenv,argsign,singl) = 
        descend sign fsign nparams (strong strip_outer_cast C) in

let constrlambody = match singl with 
          NONE -> applist (Rel (nparams+1),
                          rev(it_list_i (fun i l ERASE -> l
                                           | i l ABSTRACT -> (Rel(nparams+1-i))::l)
                                               1 [] argsign))
       | SOME A -> lambda_create(A,Rel 1) in
let constrlamexp = list_it (fun Logic b ->  (pop b)
                                 | (Inf j) b -> lambda_create(j.VAL,b))
                               constrlamenv constrlambody in 

let newsingl = match singl with NONE -> NONE 
          | SOME A -> SOME(put_inf_lam constrlamenv argsign) where rec 
               put_inf_lam = 
                  fun (Logic::cle) -> pop o (put_inf_lam cle)
                     |(Inf j::cle) -> 
                        (fun (ERASE::l) -> pop (put_inf_lam cle l)
                           |(ABSTRACT::l) -> 
                                lambda_create(j.VAL,put_inf_lam cle l))
                     | [] -> (fun [] -> A)

in (inf_nparams,SOME constrlamexp,argsign,newsingl);;

let inf_mindterm (sign,fsign) = inf_minductive (GLOB sign) (GLOB fsign);;

(* remove the redondant products corresponding to the given
list of ERASE, ABSTRACT information, keep the outer-most cast unchanged *)

let strip_constr larg (DOP2(Cast,c,t)) = DOP2(Cast,striprec (larg,c),t) 
where rec striprec (l,C) = 
  match (l,whd_betadeltaiota_empty C) with 
    (ERASE::larg,DOP2(Prod,_,DLAM(_,C))) -> striprec (larg,pop C)
  | (ABSTRACT::larg,DOP2(Prod,T,DLAM(na,C))) 
                             -> DOP2(Prod,T,DLAM(na,striprec(larg,C)))
  | [], _                               -> C
  | _                                   -> anomaly "strip_constr";;


let mapplied_to_params nbtypes nparams indidx c = 
 (nparams = 0) or (apprec indidx c)
    where rec apprec indidx = function
    DOP2(Cast,c,_) -> apprec indidx c
  | DOPN(AppL,cl) ->
    (* Either the arguments of the application have to look like:
     * AppL(Rel indidx+p; Rel (indidx-1);... Rel(indidx-nparams);...)
     * with p<nbtypes
     * or the subterms must themselves meet this test.
     * If we find a bare "Rel" with the value "indidx+p", it's
     * an error.
     *)
       (vect_length cl >= (nparams+1)
       & (match cl.(0) with (Rel p) -> (indidx <= p) & (p < indidx+nbtypes)
                           | _       -> false)
       & for_all_vect_i (fun k v -> (Rel (indidx - k - 1)) = v) 0 
                        (sub_vect cl 1 nparams))
    or (for_all_vect (apprec indidx) cl)
 
  | Rel i -> i < indidx or i >= indidx+nbtypes
  | VAR _ -> true
  | DOP0 _ -> true
  | DOP1(_,c) -> apprec indidx c
  | DOP2(_,c1,c2) -> apprec indidx c1 & apprec indidx c2
  | DOPN(_,cl) -> for_all_vect (apprec indidx) cl
  | DOPL(_,cl) -> for_all (apprec indidx) cl
  | DLAM(na,c) -> apprec (indidx+1) c
  | DLAMV(na,cl) -> for_all_vect (apprec (indidx+1)) cl
;;
 
let applied_to_params = mapplied_to_params 1;;

let rec decomp_par n c = if n <= 0 then c
       else match whd_betadeltaiota_empty c with 
                  DOP2(Prod,B,DLAM(na,d)) -> decomp_par (n-1) d
                | _                       -> error "decomp_par";;

(* Assume terms are casted and in hnf form *)

let is_small C = match C with 
    DOP2(Cast,_,t) -> isprop t 
  | _              -> anomaly "is_small found no cast";;

let rec small_constr C = 
  match C with 
    DOP2(Prod,B,DLAM(na,D)) -> is_small B & small_constr D 
  | _                       -> true;;

let small_list_constr sign nbpar ltypes indlc = 
  let indsign,lC = decomp_DLAMV_sign sign (ltypes,indlc) in
  for_all_vect 
    (fun c -> small_constr (decomp_par nbpar (hnf_recast indsign c))) lC;;


(* check that all products are logic *)

let is_logic_casted_type C = match C with 
    DOP2(Cast,_,k) -> 
      (match whd_betadeltaiota mt_evc k with 
       (DOP0(Sort(s))) -> 
        (match s with (Prop Null) | Type(Null,_) -> true | _ -> false)
       | _ ->  anomaly "is_logic_casted found a term which is not a type")
  | _              -> anomaly "is_logic_casted found no cast";;

let rec logic_constr C = 
  match C with 
    DOP2(Prod,B,DLAM(na,D)) -> is_logic_casted_type B & logic_constr D 
  | _                       -> true;;

let rec logic_arity C = 
  match C with 
    DOP2(Prod,B,DLAM(na,D)) -> ((is_logic_casted_type B) or (is_small B))
                               & logic_constr D 
  | _                       -> true;;

let unit_constr sign nbpar ltypes arity indlc recargs = 
  length ltypes = 1 
  & not (is_recursive [0] recargs)
  & let (indsign,lC) = decomp_DLAMV_sign sign (ltypes,indlc) in
     vect_length lC = 1  
  & logic_arity (decomp_par nbpar (hnf_recast sign arity) )
  & logic_constr (decomp_par nbpar  (hnf_recast indsign lC.(0)));;


let allowed_msorts issmall isunit s = let sorts = match s with
   DOP0(Sort(Type(Pos,_)))   -> [prop;spec;typep;types]
 | DOP0(Sort(Type(Null,_)))  -> [prop;typep]
 | DOP0(Sort(Prop(Pos))) -> if issmall then 
                                [prop;spec;typep;types]
                            else [prop;spec]
 | DOP0(Sort(Prop(Null))) -> if isunit then [prop;spec]
                            else [prop]
in (sorts,sorts);;

let inf_allowed_msorts s = match s with
   DOP0(Sort(Type(Pos,_)))   -> [types;spec],[types;spec]
 | DOP0(Sort(Type(Null,_)))  -> [],[]
 | DOP0(Sort(Prop(Pos)))     -> [],[types;spec]
 | DOP0(Sort(Prop(Null)))    -> [],[];;

(* check the constructors of the ith- type are well-formed 
in particular is correctly applied to its arguments and compute
the vector of recursive arguments *)

exception NonPos;;
exception NonPar;;

(* Check largs = [|Rel(n-1);..Rel(n-nbpar);a1;..ak|] and Rel p does not occur in ai
for n<=p<=n+nbtyp, raise NonPos or NonPar *)

let correct_par nbpar nbtyp n largs = 
 if  vect_length largs >= nbpar then 
     let (lpar,largs') = chop_vect nbpar largs in
     if for_all_vect_i (fun k v -> (Rel (n-k-1))=(whd_betadeltaiotaeta mt_evc v)) 0 lpar
     then if for_all_vect (noccur_bet n nbtyp) largs' then true else 
          raise NonPos
     else raise NonPar
 else raise NonPar;;

(* Given a specification of a parameterised inductive defintion
[X1]...[Xntyp]
   (a1:A1)...(ak:Ak)(B1(Xi1(a1,..ak)))->..->Bp(Xip(a1,..ak))->(Xi a1..ak t1..tn), 
gives its abstract specification as 
(a1:A1)...(ak:Ak)(B1(Xi1))->..->Bp(Xip))->(Xi t1..tn) *)

let abstract_mind_lc ntyps npars lc = 
let lC = decomp_DLAMV ntyps lc in
if npars = 0 then lC else 
let make_abs = 
tabulate_list (function i -> lambda_implicit_lift npars (Rel (i+1))) ntyps in 
map_vect ((strong whd_beta) o (substl make_abs)) lC;;


let listrec_mconstr nbtyp nbpar i indlc =

(* check the inductive types occur positively in C *)
let rec check_pos n C = match whd_betadeltaiota_empty C with 
     DOP2(Prod,B,DLAM(na,D)) -> if (noccur_bet n nbtyp B) 
                                then check_pos (n+1) D
                                else raise NonPos
   | x -> (match ensure_appl x with 
           DOPN(AppL,cl) -> let hd = hd_vect cl and largs = tl_vect cl 
           in (match hd with 
                  Rel k -> if k >= n & k<n+nbtyp 
                             & (correct_par nbpar nbtyp n largs)
                           then Mrec(n-k+nbtyp-1) 
                           else raise NonPos
                | (DOPN(MutInd(_),_) as mi) -> 
                let mispeci = mind_specif_of_mind mi in
                let auxnpar = mis_nparams mispeci in
                let (lpar,auxlargs) = chop_vect auxnpar largs in 
           if for_all_vect (noccur_bet n nbtyp) auxlargs then 
              let auxlc = mis_lc mispeci 
              and auxntyp = mis_ntypes mispeci in 
              if auxntyp <> 1 then raise NonPos else
              let auxlcvect = abstract_mind_lc auxntyp auxnpar auxlc in
              let newidx = n + auxntyp in
              let lrecvec = map_vect 
                (function C -> 
                   let C' = hnf_prod_appvect mt_evc "check_pos" C 
                      (map_vect (lift auxntyp) lpar)
                   in check_construct false newidx C') auxlcvect
              in Imbr lrecvec
(* Ce n'est pas suffisant il faut prendre en compte les imbriques recursifs *)
           else raise NonPos
                | _ -> raise NonPos)
          | _         -> anomaly "check_pos")

(* check the inductive types occur positively in the products of C, 
   if checkhd=true, also check the head corresponds to a constructor of
   the ith type *)
   
and check_construct check  = check_constr_rec [] 
 where rec check_constr_rec lrec n C = 
   match whd_betadeltaiota_empty C with 
     DOP2(Prod,B,DLAM(na,D)) -> if noccur_bet n nbtyp B
                                then check_constr_rec (Norec::lrec) (n+1) D
                                else let recarg = (check_pos n B) in 
                                check_constr_rec (recarg::lrec) (n+1) D
  |  x -> (match ensure_appl x with 
           DOPN(AppL,cl) -> let hd = hd_vect cl and largs = tl_vect cl in 
           if check then 
             (match hd with 
                  Rel k -> if k = n+nbtyp-i & correct_par nbpar nbtyp n largs 
                           then rev lrec else raise NonPos
                | _ -> raise NonPos)
           else if for_all_vect (noccur_bet n nbtyp) largs then rev lrec
                else raise NonPos)

in let (lna,lC) = decomp_DLAMV_name nbtyp indlc in
map_vect (fun C -> try check_construct true (1+nbpar) (decomp_par nbpar C)
                   with (NonPos|NonPar as err) -> 
                   let ass = assumptions_for_print lna in 
                   errorlabstrm "Ill-formed Inductive Specification"
                   [<'S (if err = NonPos then "Non positive Occurrence in "
                         else "Bad application to parameters in "); 
                     pTERMINENV(ass,C)>]) lC
;;


(* check a mutually inductive definition is well-formed and 
   compute the redondant information *)

let recache_mibody_nolamexp mib =

let kind = mib.MINDKIND and
    hyps = mib.MINDHYPS and
    nbpar = mib.MINDNPARAMS and
    nbtyp = mib.MINDNTYPES in

let sign = GLOB hyps in

let ltypes = map_vect_list (fun mip -> (mip.MINDARITY)) mib.MINDPACKETS in

let recache_mipacket i mip =

  let arity = mip.MINDARITY in

  let _,s = decomp_prod mt_evc arity in

  let recargs = listrec_mconstr nbtyp nbpar i mip.MINDLC in

  let (kd,kn) = if kind = CCI then 
              allowed_msorts (small_list_constr sign nbpar ltypes mip.MINDLC)
              (unit_constr sign nbpar ltypes arity mip.MINDLC recargs) s 
              else inf_allowed_msorts s in

   {MINDSTAMP = mip.MINDSTAMP;
    MINDCONSNAMES = mip.MINDCONSNAMES;
    MINDTYPENAME = mip.MINDTYPENAME;
    MINDARITY = mip.MINDARITY;
    MINDLC = mip.MINDLC;
    MINDLAMEXP = mip.MINDLAMEXP;
    MINDKD = kd;
    MINDKN = kn;
    MINDLISTREC = recargs;
    MINDFINITE = mip.MINDFINITE}

in {MINDKIND = mib.MINDKIND;
    MINDNTYPES = nbtyp;
    MINDSINGL = mib.MINDSINGL;
    MINDNPARAMS = nbpar;
    MINDHYPS = hyps;
    MINDPACKETS = map_i_vect recache_mipacket 1 mib.MINDPACKETS
    }
;;

let make_mipacket_lam (stamp,typename,consnamev) (finite,arity,lc,lamexp) = 
    {MINDSTAMP = stamp;
     MINDCONSNAMES = consnamev;
     MINDARITY = (match arity with
                 DOP2(Cast,c,t) -> DOP2(Cast,c,hnf_constr mt_evc t)
               | _ -> arity);
     MINDLC = lc;
     MINDLAMEXP = lamexp;
     MINDKD = [];
     MINDKN = [];
     MINDTYPENAME = typename;
     MINDLISTREC = [||];
     MINDFINITE = finite};;

let make_mibody_lam (k,hyps,nparams,lpack,singl) =
    {MINDKIND = k; 
     MINDNPARAMS = nparams;
     MINDSINGL = singl;
     MINDNTYPES = vect_length lpack;
     MINDHYPS = hyps;
     MINDPACKETS = lpack}
;;


let change_constr_mip (newar,newlc,newlamexp) mip = 
    {MINDCONSNAMES = mip.MINDCONSNAMES;
     MINDTYPENAME = mip.MINDTYPENAME;
     MINDSTAMP = mip.MINDSTAMP;
     MINDLC = newlc;
     MINDARITY = newar;
     MINDKD = mip.MINDKD;
     MINDKN = mip.MINDKN;
     MINDLISTREC = mip.MINDLISTREC;
     MINDLAMEXP = newlamexp;
     MINDFINITE = mip.MINDFINITE};;

let change_constr_mib (newpar,newhyps,newarlcvec) mib = 

let fhyps = fsign_of_sign mt_evc newhyps in

let change_mip (newar,newlc) mip = 

let newlamexp = (* compute the newlamexp when necessary *)
  if newpar = mib.MINDNPARAMS or (mip.MINDLAMEXP=NONE) then mip.MINDLAMEXP 
  else let (_,newl,_,_) = inf_mindterm (newhyps,fhyps) newpar newar newar false  in newl

in change_constr_mip (newar,newlc,newlamexp) mip in 

if mib.MINDSINGL = NONE or newpar = mib.MINDNPARAMS then 

    {MINDNPARAMS = newpar;
     MINDKIND = mib.MINDKIND;
     MINDNTYPES = mib.MINDNTYPES;
     MINDSINGL = mib.MINDSINGL;
     MINDHYPS = newhyps;
     MINDPACKETS = map2_vect change_mip newarlcvec mib.MINDPACKETS}

else   (* case of singleton types *)  
   let (newar,newlc) = newarlcvec.(0) in
   let (DLAMV(_,lC))= newlc in
   let (_,newl,_,newsingl) = 
   let newenv = add_rel (Anonymous,newar) (GLOB newhyps)
   and newinfar = extract_of newhyps fhyps newar
   in let newfenv = add_rel (Anonymous,newinfar) (GLOB fhyps) in
    inf_minductive newenv newfenv newpar newar lC.(0) true
   in  {MINDNPARAMS = newpar;
     MINDKIND = mib.MINDKIND;
     MINDNTYPES = mib.MINDNTYPES;
     MINDSINGL = newsingl;
     MINDHYPS = newhyps;
     MINDPACKETS = [|change_constr_mip (newar,newlc,newl) mib.MINDPACKETS.(0)|]}
    ;;
   
let expmod_mimap modlist mimap =
    map (fun (k,mib) ->
     (k,change_constr_mib 
           (mib.MINDNPARAMS, 
            map_sign_typ (expmod_constr modlist) mib.MINDHYPS,
            map_vect 
           (fun mip -> (expmod_constr modlist mip.MINDARITY,
(* the under_dlams is essential in order to keep the outermost Cast, 
   necessary for generalisation *)
                  under_dlams (expmod_constr modlist) mip.MINDLC)) mib.MINDPACKETS)
            mib)) mimap
;;


let make_minductive_hyps hyps indspec = 
let ids = (auto_save_variables()) in
let globals = it_vect (fun gl (ar,lc) -> (global_varsl (global_varsl gl lc) ar)) [] indspec in thin_hyps_glob ids (hyps,globals);;

let make_minductive_fhyps fhyps hyps =
let fhyps' = thin_to_match hyps fhyps in thin_to_type_level fhyps';;

let put_DLAMSV_inf lna vecarinf = 
    put o (putfirst lna (vect_length vecarinf-1))
  
  where rec putfirst = fun (na::l) k c -> 
         if k<0 then anomaly "put_DLAM_inf" 
         else if vecarinf.(k)=Logic then  putfirst l (k-1) (map_vect pop c)
              else (l,k-1,DLAMV(na,c))
                       | [] k c -> anomaly"put_DLAM_inf"
  and put (lna,k,c) = match lna with 
          (na::l) -> if k<0 then anomaly "put_DLAM_inf" 
                     else if vecarinf.(k)=Logic then  put (l,k-1,pop c)
                          else put (l,k-1,DLAM(na,c))
         | []     -> if k = -1 then c else anomaly "put_LAM_inf";;

let enforce_type_construct univ (sign,nparams) C = 
    let exrec = core_fmachine false (mt_evc,[]) in 
    let rec enfrec (sign,nparams,C) = 
    if nparams = 0 then 
        let jC = exrec sign C 
        in match whd_betadeltaiota_empty jC.TYP with 
                 (DOP0(Sort(Type(_,uc)))) -> (enforcegeq univ uc;jC)
                | _ -> error "Type of Constructor not well-formed"
    else let (DOP2(Prod,c1,DLAM(name,c2))) = whd_betadeltaiota_empty C in 
         let j = exrec sign c1 in
         let assum = assumption_of_judgement mt_evc j in
         let var = (name,assum) in
         let j' = enfrec (add_rel var sign,(nparams-1),c2)
      in gen_rel mt_evc sign name assum j'
    in let judge = (try enfrec (sign,nparams,C)
     with UserError _ -> 
     errorlabstrm "enforce_type_construct"
     [< 'S "Type of constructor :" ; pTERMINENV(sign,C); 'FNL; 
        'S "not well formed in environment :" ; pENV sign >])
    in cast_of_judgement judge
;;


let check_prop_construct sign C = 
    let jC = (try  core_fmachine false (mt_evc,[]) sign C 
              with UserError (s,msg) -> 
     errorlabstrm "enforce_type_construct"
     [< pTERMINENV(sign,C); 'FNL;  
        'S "can not be the type of a constructor since it is not well typed:"
         ; 'FNL ;msg >])
    in 
    match whd_betadeltaiota_empty jC.TYP with 
                 (DOP0(Sort(_))) -> cast_of_judgement jC 
               | _ -> errorlabstrm "indtypes__check_prop_construct" 
                          [< pTERMINENV(sign,C); 'FNL; 
                             'S "can not be the type of a constructor since it is not a type" >];;

let packets_minductive sign nparams namesvec lfinite arlcvec 
  =  let ntypes = vect_length arlcvec in 
(* At the moment inductive definitions have an empty stamp *)
let na = Anonymous in

let aritj_vec = 
    map_vect (function (ar,_) -> casted_execute mt_evc sign
              (instantiate_universes ar)) arlcvec in

let hyps = (GLOB sign) in
let ind_sign = it_vect 
   (fun isign arj -> 
   (add_rel (na,assumption_of_judgement mt_evc arj) isign)) hyps aritj_vec in

let one_ind finite namesarg arj (_,lc) = 
   let lc = instantiate_universes lc in 
   let (_,DOP0(Sort(sort))) = decomp_prod mt_evc (j_val arj) in
   let ar = cast_of_judgement arj in
   let lna,lC = decomp_DLAMV_name ntypes lc in 
   let LC = match sort with 
                  Type(_,univ) -> 
                  map_vect (enforce_type_construct univ (ind_sign,nparams)) lC
                 |Prop(_) -> map_vect (check_prop_construct ind_sign) lC in
     let lcind = put_DLAMSV lna LC in
              make_mipacket_lam namesarg (finite,ar,lcind,NONE),NONE,NONE

in

let packs = map4_vect one_ind lfinite namesvec aritj_vec arlcvec
in let mhyps = make_minductive_hyps sign arlcvec in
   (mhyps,packs);;

let packets_minductive_extract (sign,fsign)  nparams namesvec lfinite arlcvec 
  =  let ntypes = vect_length arlcvec in 

(* At the moment inductive definitions have an empty stamp *)
let na = Anonymous in

let aritj_infar_vec = 
    map_vect (function (ar,_) -> casted_infexecute mt_evc (sign,fsign) 
              (instantiate_universes ar)) arlcvec in

let vecarinf = map_vect snd aritj_infar_vec in

let hyps = GLOB sign and fhyps = GLOB fsign in

let ind_sign,inf_ind_sign =  
let top=vect_length  namesvec 
  in build_signs 0 (hyps,fhyps) 
  where rec build_signs i res    =
    if i=top then res
    else let  (isign,fisign)= res and 
              (arj,infar)=aritj_infar_vec.(i) and
              (_,id,_)=namesvec.(i) 
         in build_signs (i+1)
              (add_rel ((Name id),assumption_of_judgement mt_evc arj) isign,
                add_inf_rel ((Name id),infar) fisign) 

in

let one_ind checksingl finite namesarg (arj,infar) (_,lc) = 
   let lc = instantiate_universes lc in 
   let (_,DOP0(Sort(sort))) = decomp_prod mt_evc (j_val arj) in
   let ar = cast_of_judgement arj in
   let lna,lC = decomp_DLAMV_name ntypes lc in 
   let LC = match sort with 
                  Type(_,univ) -> 
                  map_vect (enforce_type_construct univ (ind_sign,nparams)) lC
                 |Prop(_) -> map_vect (check_prop_construct ind_sign) lC in
   (match infar with 
     Logic -> let lcind = put_DLAMSV lna LC in
              make_mipacket_lam namesarg (finite,ar,lcind,NONE),NONE,NONE
   | Inf infarj  -> 
     let inflcj = map_vect 
           (fun CC -> let (DOP2(Cast,C,_)) = CC in 
                 core_infmachine (mt_evc,([],[])) ind_sign inf_ind_sign C) LC in
     
    let lcind = put_DLAMSV lna LC in
    let (infnpar,lamexp,arglist,singl) = 
    let (check,C) = if checksingl & (vect_length lC) = 1 then (true,lC.(0))
                    else false,ar in 
    inf_minductive ind_sign inf_ind_sign nparams ar C check in 
    (match singl with 
       NONE -> (* compute both the inductive type and its extraction *)
    let inf_vect = map_vect 
               ((strip_constr arglist) o cast_of_judgement o extract) inflcj
    in let newinflc = put_DLAMSV_inf lna vecarinf inf_vect
    in (make_mipacket_lam namesarg (finite,ar,lcind,lamexp),
       SOME(make_mipacket_lam namesarg  
                 (finite,cast_of_judgement infarj,newinflc,SOME(Rel 1)),infnpar),NONE)
     | _ -> (make_mipacket_lam namesarg (finite,ar,lcind,lamexp),NONE,singl)))

in let checksingl = vect_length vecarinf = 1 in 

let packs = map4_vect (one_ind checksingl) lfinite namesvec aritj_infar_vec arlcvec

in let mhyps = make_minductive_hyps sign arlcvec in
   (mhyps,packs);;

let minductive_decl kind mhyps nparams packets singl = 
  (kind,recache_mibody_nolamexp 
      (make_mibody_lam (kind,mhyps,nparams,packets,singl)));;

let cci_minductive (mhyps,nparams,packs) fsign =

match packs.(0) with 
   (pack,_,(SOME _ as singl)) -> 
     [minductive_decl CCI mhyps nparams [|pack|] singl]
  | _ -> let packets = map_vect fst3 packs 
         and infpackets = map_vect (fun (_,x,_) -> x) packs in 
         let mib = minductive_decl CCI mhyps nparams packets NONE
         and infmib = 
    let infpack,infnpar = 
    it_vect (fun (lpack,n) NONE -> (lpack,n)
               | (lpack,n) (SOME(pack,npar)) -> (pack::lpack,npar)) ([],0)
            infpackets
    in if infpack = [] then []
       else let mfhyps = make_minductive_fhyps fsign mhyps in
       [ minductive_decl FW mfhyps infnpar (vect_of_list (rev infpack)) NONE ]
in mib::infmib;;

let fw_minductive (mhyps,nparams,packs) = 
let packets = map_vect fst3 packs 
 in [minductive_decl FW mhyps nparams packets NONE];;

let execute_minductive sign nparams namesvec lfinite arlcvec = 
    let  (mhyps,packs) = 
    packets_minductive sign nparams namesvec lfinite arlcvec 
    in fw_minductive (mhyps,nparams,packs);; 

let infexecute_minductive (sign,fsign)  nparams namesvec lfinite arlcvec = 
  let  (mhyps,packs) = packets_minductive_extract
   (sign,fsign)  nparams namesvec lfinite arlcvec 
   in cci_minductive (mhyps,nparams,packs) fsign ;;

let casted_generalize id var c =
    let c' = DOP2(Prod,var,DLAM(Name id,subst_var id c)) and
        c'ty = type_of_product mt_evc (cast_type var) (cast_type c)
    in DOP2(Cast,c',c'ty)
;;

let casted_abstract id var c =
    let c' = DOP2(Lambda,var,DLAM(Name id,subst_var id c)) and
        c'ty = DOP2(Prod,var,DLAM(Name id,(cast_type c)))
    in DOP2(Cast,c',c'ty)
;;

(* Abstraction on mutually inductive definitions *)

let abs_var_on_minductive id var arity_lc_vec =
let ntyp = vect_length arity_lc_vec in 
let new_refs = tabulate_list (fun k ->applist(Rel (k+2),[Rel 1])) ntyp in 

let arity_lc_vec' = map_vect (function (arity,lc) -> 
          let arity' = casted_generalize id var arity and
          lc' = under_dlams 
          (fun B -> casted_generalize id var (substl new_refs B)) lc in
          (arity',lc')) arity_lc_vec 
in (arity_lc_vec',ABSTRACT)
;;

let abstract_minductive_once ((hyps,ispec),modl) id =
    if isnull_sign hyps or id <> fst(hd_sign hyps) then
        ((hyps,ispec),modl)
    else
let (hyps_ispec,modif) = abs_var_on_minductive id (snd(hd_sign hyps)) ispec
in ((tl_sign hyps,hyps_ispec),modif::modl)
;;

let abstract_minductive ids_to_abs (hyps,mispec) =
let ((hyps',ispec'),revmodl) =
    it_list abstract_minductive_once ((hyps,mispec),[]) ids_to_abs in
let modl = rev revmodl
in ((hyps',ispec'),modl)
;;

let abstract_mibody (osecsp,nsecsp) ids_to_abs mib =
let miarlc = map_vect (fun mip -> (mip.MINDARITY,mip.MINDLC)) 
             mib.MINDPACKETS in
let ((hyps',arity_lc_vec'),modl) =
    abstract_minductive ids_to_abs (mib.MINDHYPS,miarlc) in
let lmodif_one_mind i = 
    let nbc = vect_length ((mind_nth_type_packet mib i).MINDCONSNAMES)
    in (MutInd(osecsp,i),DO_ABSTRACT(MutInd(nsecsp,i),modl))::
         (tabulate_list (function j -> let j'=j+1 in
   (MutConstruct((osecsp,i),j'),DO_ABSTRACT(MutConstruct((nsecsp,i),j'),modl)))
            nbc) in
let modifs = flat (tabulate_list lmodif_one_mind mib.MINDNTYPES)
in (change_constr_mib (mib.MINDNPARAMS + (length modl),hyps',
    arity_lc_vec') mib,modifs);;


let abstract_mimap (osecsp,nsecsp) ids_to_abs cmap =
let (mods,cmap) =
     it_list (fun (mods,cmap) (k,ib) ->
                 match k with
                 CCI ->
                  let (ib',modifs) = abstract_mibody (ccisp_of osecsp,ccisp_of nsecsp) ids_to_abs ib
                  in (modifs@mods,(k,ib')::cmap)
               | FW ->
                  let (ib',modifs) = abstract_mibody (fwsp_of osecsp,fwsp_of nsecsp)  ids_to_abs ib
                  in (modifs@mods,(k,ib')::cmap))
    ([],[]) cmap
in (mods,cmap)
;;

(**** For the discharge operation ******)

let process_mimap (sp,newsp) (ids_to_discard,work_alist) mimap =
let expmod_mimap = expmod_mimap work_alist mimap
in abstract_mimap (sp,newsp) ids_to_discard expmod_mimap
;;


(* $Id: indtypes.ml,v 1.14 1996/04/03 12:55:28 cpaulin Exp $ *)
