(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                 mach.ml                                  *)
(****************************************************************************)
#infix "o";;
#open "std";;
#open "names";;
#open "vectops";;
#open "impuniv";;
#open "extraction";;
#open "generic";;
#open "term";;
#open "reduction";;
#open "typing";;
#open "termenv";;
#open "pp";;
#open "stdpp";;
#open "more_util";;
#open "machops";;
#open "constrtypes";;

(* j_val { VAL = c ; TYPE = t ; Kind = k } gives the constr c casted
to type t *)
let j_val = machops__j_val;;
let assumption_of_judgement = machops__assumption_of_judgement;;

let vect_lift = map_i_vect lift 0;;

let unsafe_fmachine nofix (nocheck,noverify) (sigma,metamap) sign constr = 
    ((if not(noverify) then verify_wf_env sign) ; exemeta_rec sign constr)
where rec exemeta_rec env =  function
  DOP0(Meta n) -> (try (match assoc n metamap with
                        DOP2(Cast,typ,kind) ->
                        {VAL=DOP0(Meta n);
                         TYP=typ;
                         KIND=kind}
                      | typ ->
                        {VAL=DOP0(Meta n);
                         TYP=typ;
                         KIND=hnf_constr sigma (exemeta_rec (GLOB(get_globals env)) typ).TYP})
                    with Not_found -> error "Metavariable wasn't in the metamap")
| Rel(n)                 -> relative sigma n env
| (VAR id as val) ->
  let (DOP2(Cast,typ,kind)) = snd(lookup_glob id env)
  in {VAL=val;
      TYP=typ;
      KIND=kind}

| (DOPN(Abst _,_) as val) ->
   if evaluable_abst val then
       exemeta_rec env (abst_value val)
   else error "Cannot typecheck an unevaluable abstraction"

| (DOPN(Const sp,_) as val) ->
  let (DOP2(Cast,typ,kind)) = type_of_const sigma env val
  in {VAL=val;
      TYP=typ;
      KIND = kind}

| DOPN(MutInd _,_) as val ->
  let (DOP2(Cast,typ,kind)) = type_of_mind sigma env val
  in {VAL=val;
      TYP=typ;
      KIND=kind}

| DOPN(MutConstruct _,_) as val          -> 
  let (DOP2(Cast,typ,kind)) = type_of_mconstr sigma env val
  in {VAL=val;
      TYP=typ;
      KIND=kind}

| (DOPN(MutCase,cl) as val) ->
  let P = cl.(0)
  and c = cl.(1)
  and lf = sub_vect cl 2 ((vect_length cl) - 2) in
  (let {TYP=ct;_} as cj = exemeta_rec env c in
   let Pj = exemeta_rec env P in
   let lfj = map_vect (exemeta_rec env) lf in
   let lft = map_vect (fun j -> j.TYP) lfj
   in  {VAL=DOPN(MutCase,concat_vect [|j_val Pj; j_val cj|] (map_vect j_val lfj));
        TYP=type_case_elim env sigma ct Pj.TYP lft P c;
        KIND = sort_of_arity sigma Pj.TYP})

| (DOPN(Fix(vn,i),cl) as val) ->
  if nofix & exists_vect (fun n -> n < 0) vn then 
      error "General Fixpoints not allowed";
  let nbfix = let nv = vect_length cl in 
           if nv < 2 then error "Ill-formed recursive definition" else nv-1 in
  let lar = sub_vect cl 0 nbfix in
  let ldef = last_vect cl in
  let (lfi,vdef) = decomp_DLAMV_name nbfix ldef in
  let newenv = it_vect2 
     (fun env name ar -> let assum = assumption_of_judgement sigma 
                         (exemeta_rec env ar) in (add_rel (name,assum) env))
     env (vect_of_list (rev lfi)) (vect_lift lar) in
  let larj = map_vect (exemeta_rec env) lar in
  let vdefj = map_vect (exemeta_rec newenv) vdef in
  let vdefv = put_DLAMSV lfi (map_vect j_val vdefj) in
  let larv = map_vect j_val larj in
  let fix = DOPN(Fix(vn,i),concat_vect larv [|vdefv|])
  in check_fix sigma fix;
  {VAL= fix;
   TYP=type_fix sigma fix (map_vect (fun j -> j.TYP) vdefj);
   KIND=larj.(i).TYP}

| (DOPN(CoFix i,cl) as val) ->
(*  if nofix  
  then error "General Fixpoints not allowed"
  else *)
       let nbfix = let nv = vect_length cl in 
                       if nv < 2 
                       then error "Ill-formed recursive definition" 
                       else nv-1 in
       let lar = sub_vect cl 0 nbfix in
       let ldef = last_vect cl in
       let (lfi,vdef) = decomp_DLAMV_name nbfix ldef in
       let newenv = it_vect2 (fun env name ar -> 
                                  let assum = assumption_of_judgement 
                                                 sigma (exemeta_rec env ar) in
                                  (add_rel (name,assum) env))
                              env (vect_of_list (rev lfi)) (vect_lift lar) in
       let larj  = map_vect (exemeta_rec env) lar in
       let vdefj = map_vect (exemeta_rec newenv) vdef in
       let vdefv = put_DLAMSV lfi (map_vect j_val vdefj) in
       let larv  = map_vect j_val larj in
       let cofix = DOPN(CoFix i,concat_vect larv [|vdefv|])
       in check_cofix sigma cofix;
          {VAL= cofix;
           TYP=type_cofix sigma cofix (map_vect (fun j -> j.TYP) vdefj);
           KIND=larj.(i).TYP}

| DOP0(Sort(Prop(c)))                 -> fcn_proposition c
| DOP0(Sort(Type(c,u)))               -> fcn_type_with_universe (c,u)
| DOPN(AppL,tl)           -> 
  let j = exemeta_rec env (hd_vect tl) in
  let jl = map_vect_list (exemeta_rec env) (tl_vect tl) in
      apply_rel_list sigma env nocheck jl j

| DOP2(Lambda,c1,DLAM(name,c2))      -> 
  let j = exemeta_rec env c1 in
  let assum = assumption_of_judgement sigma j in
  let var = (name,assum) in
  let j' = exemeta_rec (add_rel var env) c2 in 
      abs_rel sigma name assum j'

| DOP2(Prod,c1,DLAM(name,c2))        ->
  let j = exemeta_rec env c1 in
  let assum = assumption_of_judgement sigma j in
  let var = (name,assum) in
  let j' = exemeta_rec (add_rel var env) c2 in
      gen_rel sigma env name assum j'

| DOP2(Cast,c,t) ->
  let cj = exemeta_rec env c in
  let tj = exemeta_rec env t
  in cast_rel sigma env cj tj

| DOP0(Implicit) -> error "cannot execute Implicit"

| _ -> error "Cannot be executed";;

let core_fmachine nocheck = unsafe_fmachine true (nocheck,false);;

(* WITHOUT INFORMATION *)
let fmachine nocheck metamap sign constr =
let j = core_fmachine nocheck metamap sign constr
in {VAL=constr;TYP=strong strip_outer_cast j.TYP;KIND=j.KIND};;

(* allows general fixpoints to appear in the term *)

let fixfexemeta sigma metamap sign constr = 
let j = unsafe_fmachine false (true,false) (sigma,metamap) sign constr
in {VAL=constr;TYP=strong strip_outer_cast j.TYP;KIND=j.KIND};;

let fexemeta sigma metamap = fmachine true (sigma,metamap);;
let execute_rec sigma = fmachine false (sigma,[]);;
let fexecute sigma sign = fmachine false (sigma,[]) (GLOB sign);;

let type_of_rel sigma env c = 
    try strong whd_betaiota (core_fmachine false (sigma,[]) env c).TYP
    with Invalid_argument s -> error ("Invalid arg " ^ s)
;;

let type_of sigma sign c = type_of_rel sigma (GLOB sign) c;;

let unsafe_type_of sigma env c = 
  try strong whd_betaiota (unsafe_fmachine false (true,true) (sigma,[]) env c).TYP
    with Invalid_argument s -> error ("Invalid arg " ^ s);;

let sign_of_sign sigma sign =
    sign_it
    (fun id a sign ->
         match a with
         DOP2(Cast,_,_) -> add_sign (id,a) sign
       | _ -> (add_sign (id,assumption_of_judgement sigma (fexecute sigma sign a)) sign))
         sign nil_sign
;;

let env_of_env sigma env =
    dbenv_it
    (fun na a env ->
         match a with
         DOP2(Cast,_,_) -> add_rel (na,a) env
       | _ -> add_rel (na,assumption_of_judgement sigma (execute_rec sigma env a)) env)
       env (GLOB(get_globals env))
;;

let castify_sign sigma sign = sign_of_sign sigma sign;;

let castify_env sigma env =
let sign = castify_sign sigma (get_globals env)
in env_of_env sigma (ENVIRON(sign,get_rels env))
;;

let add_inf_rel (na,inf) env =
    match inf with
    Logic -> add_rel (na,DOP0(Implicit)) env
  | Inf j -> add_rel (na,j_val j) env
;;


let inf_of_mind_singl sigma (env,fenv) (DOPN(MutInd (sp,_) as gl,args) as K) =
try (let (_,cb) = environ__mind_of_path sp
     in (match mind_singl cb with 
          SOME A -> Inf(unsafe_fmachine true (true,true) (sigma,[]) fenv A),true
        | NONE ->  if informative_reference K then
         let infK = global_reference fenv (fwsp_of sp) (id_of_global gl)
         in Inf(unsafe_fmachine true (true,true) (sigma,[]) fenv infK),false
         else Logic,false))
with Not_found -> error ((string_of_path sp) ^ " not declared")
   | Failure _ -> anomaly "cannot convert an inductive to an finductive"
;;
let inf_of_mind sigma (env,fenv) mind 
 = fst (inf_of_mind_singl sigma (env,fenv) mind);;

let inf_mconstructor (sigma,metamap) sign fsign
                    (DOPN(MutConstruct((sp,_) as msp,i),cl)) =
 let mind = DOPN(MutInd msp,cl) in 
    (match inf_of_mind_singl sigma (sign,fsign) mind with
      Logic,_ -> Logic
   | (Inf infindj),singl -> let (SOME clamexp) = mind_lamexp mind 
      in if singl then Inf(unsafe_fmachine true (true,true) (sigma,[]) fsign clamexp)
         else let (DOPN(MutInd infindsp,infcl)) = infindj.VAL in
     let infmconstructor = DOPN(MutConstruct(infindsp,i),infcl) in
     let infval = subst1 infmconstructor clamexp
  in Inf (unsafe_fmachine true (true,true) (sigma,fst metamap) fsign infval));;

let rec core_infmachine (sigma,metamap) = infexec 
where rec infexec sign fsign = function 
       DOP2(Cast,c,(DOP0(Sort _) as sort)) -> if is_info_sort sort then 
                         infexec sign fsign c else Logic
    |  DOP2(Cast,c,DOP2(Cast,_,sort)) -> if is_info_sort sort then 
                         infexec sign fsign c else Logic
    |  DOP2(Cast,c,_) -> infexec sign fsign c
    |  DOP0(Meta n) ->
       (match assoc n (snd metamap) with
          Logic -> Logic
        | Inf j -> Inf{VAL=DOP0(Meta n);
                       TYP=j.VAL;
                       KIND = hnf_constr sigma j.TYP})

    | Rel n ->
      inf_rel fsign (contents_of_kind (hnf_constr sigma (cast_type (snd (lookup_rel n sign))))) n

    | VAR str ->
      inf_var fsign (contents_of_kind (hnf_constr sigma (cast_type (snd(lookup_glob str sign))))) str

    | DOPN(Const _,_) as val -> inf_of_const sigma (sign,fsign) val

    | DOPN(Abst _,_) as val ->
      if evaluable_abst val then infexec sign fsign (abst_value val)
      else error "cannot extract from an unevaluable abstraction"

    | DOP0(Sort s) -> inf_sort s

    | DOPN(AppL,tl) ->
      let c1 =  (hd_vect tl)
      and cl = tl_vect tl in
      let funinf = infexec sign fsign c1 in
      if funinf = Logic then Logic else 
      let argsinf = map_vect (infexec sign fsign) cl
      in it_vect inf_apply funinf argsinf

    | DOP2(Lambda,c1,DLAM(name,c2)) ->
      let varinf = infexec sign fsign c1 in
      let bodyinf = infexec (add_rel (name,c1) sign) (add_inf_rel (name,varinf) fsign) c2
      in inf_abs_rel name bodyinf varinf

    | DOP2(Prod,c1,DLAM(name,c2)) ->
      let c1inf = infexec sign fsign c1 in
      let c2inf = infexec (add_rel (name,c1) sign) (add_inf_rel (name,c1inf) fsign) c2
      in inf_gen_rel name c2inf c1inf

    | DOPN(MutInd _,_) as val -> inf_of_mind sigma (sign,fsign) val

    | DOPN(MutConstruct _,_) as val -> inf_mconstructor (sigma,metamap) sign fsign val

    | DOPN(Fix(vn,i),cl) -> let lar = sub_vect cl 0 ((vect_length cl) - 1)
      in let inflar = map_vect (infexec sign fsign) lar in
      let infAi = inflar.(i) in 
       (match infAi with 
          Logic   -> Logic 
        | Inf Aij -> let lfi,ldefs = decomp_all_DLAMV_name (last_vect cl) 
          in let new_env,new_fenv = it_vect2
             (fun (ne,nfe) fk Ak -> 
              (add_rel (fk,Ak) ne, 
               let infAk = infexec ne nfe Ak in (add_inf_rel (fk,infAk) nfe)))
               (sign,fsign) (vect_of_list (rev lfi)) (vect_lift lar) in 
(* a special function to localize the recursive index in the extracted term *)
        let rec exec_def ne nfe n def = 
            (match whd_betadeltaiota sigma def with 
             DOP2(Lambda,c1,DLAM(name,c2)) ->
               let varinf = infexec ne nfe c1 in
               let ne' = (add_rel (name,c1) ne) 
               and nfe' = add_inf_rel (name,varinf) nfe in
             if n = 0 then let infc2 = infexec ne' nfe' c2 
                           in let infdef = inf_abs_rel name infc2 varinf
                              and index = if varinf = Logic then -1
                        (* the recursive call was on a non-informative term *) 
                                          else 0 in (infdef,index)
             else let bodyinf,countl = 
                   exec_def (add_rel (name,c1) ne) 
                            (add_inf_rel (name,varinf) nfe) (n-1) c2 in
                 let (infabs,abs) = inf_abs_rel_count name bodyinf varinf
                 in (infabs,if abs = ERASE then countl else 
                            if countl<0 then countl-1 else countl+1)
            | _ -> anomaly "exec_def : should be a lambda")
       in let infdefs_ind = map2_vect (exec_def new_env new_fenv) vn ldefs
          in inf_fix sigma i Aij.TYP lfi inflar infdefs_ind)


    | DOPN(CoFix i,cl) ->
            let lar    = sub_vect cl 0 ((vect_length cl) - 1) in 
            let inflar = map_vect (infexec sign fsign) lar in
            let infAi  = inflar.(i) 
            in ( match infAi with 
                   Logic   -> Logic 
                 | Inf Aij -> 
                     let lfi,ldefs = decomp_all_DLAMV_name (last_vect cl) in 
                     let new_env,new_fenv = 
                          it_vect2 (fun (ne,nfe) fk Ak -> 
                                        (add_rel (fk,Ak) ne, 
                                        let infAk = infexec ne nfe Ak 
                                        in (add_inf_rel (fk,infAk) nfe)))
                                   (sign,fsign) 
                                   (vect_of_list (rev lfi)) (vect_lift lar) in 
                                   
                     let infdefs = map_vect (infexec new_env new_fenv) ldefs 
                     in  inf_cofix sigma i Aij.TYP lfi inflar infdefs)

    | DOPN(MutCase,cl) -> let P = cl.(0) in 
      let Pinf = infexec sign fsign P in 
      match Pinf with Logic  -> Logic 
                    | Inf Pj -> 
         let c = cl.(1)
         and lf = sub_vect cl 2 ((vect_length cl) - 2) in
         let ct = (execute_rec sigma sign c).TYP in
         let cinf = infexec sign fsign c
         and lfinf = map_vect (infexec sign fsign) lf 
         in inf_mutcase fsign sigma ct Pj cinf lfinf

    | _ -> error "Cannot extract information"
;;

(* WITH INFORMATION *)
let infmachine nocheck (sigma,metamap) sign fsign constr =
let constr_metamap = map (fun (id,(c,_)) -> (id,c)) metamap
and inf_metamap = map (fun (id,(_,i)) -> (id,i)) metamap in
let j = core_fmachine nocheck (sigma,constr_metamap) sign constr in
let inf = 
 if is_info_sort j.KIND then (* Case the term is informative *)
 let uniarc = get_uniarc () in
 let infj = core_infmachine (sigma,(constr_metamap,inf_metamap)) sign fsign j.VAL in
 match infj with
     (Inf inf') -> 
    let jtyp = core_fmachine nocheck (sigma,constr_metamap) sign j.TYP in
    let (Inf inft') = core_infmachine (sigma,(constr_metamap,inf_metamap)) sign fsign jtyp.VAL in
    let _ = set_uniarc uniarc in
(** does not check anymore that extracted terms are well-formed 
    let c = DOP2(Cast,inf'.VAL,inft'.VAL) in
    let constrinf_metamap = (map_succeed (fun (mv,Inf ij) -> (mv,j_val ij)
                                            | (_,Logic) -> failwith "caught")
                                         inf_metamap) in
    let _ = fixfexemeta sigma constrinf_metamap fsign c in
**)
    Inf{VAL=strong whd_beta inf'.VAL;
           TYP=strong whd_beta inft'.VAL;
            KIND=inf'.KIND}
 | _ -> Logic
  else Logic 
in   hashcons__hash1
    ({VAL= strong strip_outer_cast j.VAL;
      TYP=strong strip_outer_cast j.TYP;
      KIND=j.KIND},inf);;

let fsign_of_sign sigma sign =
    snd(sign_it
        (fun id a (sign,fsign) ->
             (add_sign (id,a) sign,
              match snd(infmachine true (sigma,[]) (GLOB sign) (GLOB fsign) a) with
              Logic -> fsign
            | Inf ft -> add_sign (id,j_val ft) fsign))
         sign (([],[]),([],[])))
;;

let fenv_of_env sigma env =
let fsign = fsign_of_sign sigma (get_globals env) in
let fenv = snd(dbenv_it
               (fun id a (env,fenv) ->
                    (add_rel (id,a) env,
                     match snd(infmachine true (sigma,[]) env fenv a) with
                     Logic -> fenv
                   | Inf ft -> add_rel (id,j_val ft) fenv))
               env (GLOB(get_globals env),GLOB fsign))
in fenv
;;

let FENV sigma env = (env,fenv_of_env sigma env);;
let FSIGN sigma sign = (sign,fsign_of_sign sigma sign);;
let FGLOB (sign,fsign) = (GLOB sign,GLOB fsign);;

let infexemeta sigma metamap (env,fenv) c =
    infmachine true (sigma,metamap) env fenv c;;
let infexecute_rec sigma (sign,fsign) c =
    infmachine false (sigma,[]) sign fsign c;;
let infexecute sigma (sign,fsign) c =
    infmachine false (sigma,[]) (GLOB sign) (GLOB fsign) c;;

let fvar_type sigma (sign,fsign) (VAR na) =
let varty = (snd(lookup_sign na sign))
in match (snd(infexemeta sigma [] (FGLOB(sign,fsign)) varty)) with
    Inf ft -> ft.VAL
  | Logic -> invalid_arg "Fvar with a non-informative type (1)!"
;;

let whd_instuni = function
    DOP0(Sort(Type(cts,u))) ->
    let u = (if u = Dummy_univ then New_univ() else u)
    in DOP0(Sort(Type(cts,u)))
  | c -> c
;;

let instantiate_universes c = strong whd_instuni c;;

(* $Id: mach.ml,v 1.8 1996/02/12 15:47:49 cpaulin Exp $ *)
