(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                machops.ml                                *)
(****************************************************************************)

#infix "o";;
#open "std";;
#open "names";;
#open "vectops";;
#open "impuniv";;
#open "generic";;
#open "term";;
#open "reduction";;
#open "environ";;
#open "typing";;
#open "termenv";;
#open "himsg";;
#open "pp";;
#open "stdpp";;
#open "more_util";;
#open "proof_constraints";;
#open "proof_trees";;

(* Functions for interpreting a constant, initially in  termenv.ml 
   using eq_constr instead of conv - Christine *)

let search_var env str = VAR(fst(lookup_glob str env));;


(* this function verifies that HL1 is a sub-sequence of HL2, by structural
 * equality on members.  Members must be found in the same order - that
 * is the meaning of "sequence" here.
 *)
let hyp_seq_ordered_inclusion (idl1,tyl1) (idl2,tyl2) =
    aux idl1 tyl1 idl2 tyl2 where rec aux = fun
    [] [] _ _ -> true
  | _ _ [] [] -> false
  | (id1::idl1) (ty1::tyl1) idl2 tyl2 -> (search idl2 tyl2
                                          where rec search = fun
                      [] [] -> false
                    | (id2::idl2) (ty2::tyl2) -> if id1 = id2 then
                               (conv proof__mt_evc ty1 ty2) & aux idl1 tyl1 idl2 tyl2
                                              else search idl2 tyl2
                    | _ _ -> invalid_arg "hyp_seq_ordered_inclusion")
  | _ _ _ _ -> invalid_arg "hyp_seq_ordered_inclusion"
;;

let construct_reference id env (oper,hyps) =
    if hyp_seq_ordered_inclusion hyps (get_globals env) then
        DOPN(oper,vect_of_list(map (fun id -> VAR id) (ids_of_sign hyps)))
    else error ("Hum - cannot instanciate the reference "^(string_of_id id)^
      	       	" because there weren't enough assumptions")
;;

let global_reference env sp id =
    try construct_reference (basename sp) env (global_operator sp id)
    with Not_found -> error ("Cannot find reference "^(string_of_path sp))
;;

(* Supprime par Claude *)
(* search_reference env str gives the constr of name str
 * in the environment env *)
let search_reference env str =
  let sp = (* try *) nametab__sp_of_id CCI str
           (* with Not_found -> error "search_reference" *)
  in global_reference env sp str
;;

let search_freference env str =
  let sp = (* try *) nametab__sp_of_id FW str
           (* with Not_found -> error "search_freference" *)
  in global_reference env sp str
;;

let search env id =
    try search_var env id
    with Not_found -> search_reference env id
;;

let global env id =
    try search env id
    with Not_found -> errorlabstrm "machops__global"
      	       	       	       	[< 'S((string_of_id id) ^ " not declared") >]
;;

let lookup_exist sigma env sp =
try let evd = evc__map sigma sp
in if hyp_seq_ordered_inclusion evd.evgoal.hyps (get_globals env) then
    DOPN(Const sp,vect_of_list(map (fun id -> VAR id) (ids_of_sign evd.evgoal.hyps)))
   else errorlabstrm "termenv__lookup_exist"
       [< 'S"Hum - cannot instanciate the constant";'SPC;
          'S(string_of_path sp);'SPC;
          'S"because there weren't enough assumptions">]
with Not_found -> errorlabstrm "termenv__lookup_exist"
                  [< 'S ((string_of_path sp) ^ " not declared")>]
;;

let type_of_existential sigma env (DOPN(Const sp,args)) =
if evc__ctxt_access sigma sp then
try let evd = evc__map sigma sp
    in if for_all2eq_vect (fun id (VAR id') -> id = id' | _ _ -> false) (vect_of_list (ids_of_sign evd.evgoal.hyps)) args
        & hyp_seq_ordered_inclusion evd.evgoal.hyps (get_globals env) then
        evd.evgoal.concl
       else 
        invalid_arg "An existential was found with a bad argument list"
with Not_found -> error ((string_of_path sp) ^ " not declared")
   | Failure _ -> invalid_arg "the current environment and the constant's env of construction are incompatible"
else error ((string_of_path sp) ^ " not declared")
;;


let cast_fully = ref true;;
let j_val j =
    if !cast_fully then
        DOP2(Cast,j.VAL,DOP2(Cast,j.TYP,j.KIND))
    else j.VAL;;

let id_from_name = function
    Name id -> id | Anonymous -> (id_of_string "X")
;;

let assumption_of_judgement sigma j =
let typ = whd_betadeltaiota sigma j.TYP in
match typ with DOP0(Sort(_)) -> DOP2(Cast,j.VAL,typ)
             | _ -> errorlabstrm "assumption_of_judgement"
        [< 'S "Cannot introduce an hypothesis over the term "; pTERM j.VAL>]
                 
;;

let fcn_proposition cts =
    {VAL=DOP0(Sort(Prop cts));
     TYP=DOP0(Sort(Type(cts,prop_univ)));
     KIND = DOP0(Sort(Type(cts,prop_univ_univ)))};;

let fcn_type_with_universe (cts,u) =
   let v = Super u 
   in let sv = Super v in 
     {VAL=DOP0(Sort(Type(cts,u)));
      TYP=DOP0(Sort(Type(cts,v)));
      KIND = DOP0(Sort(Type(cts,sv)))};;

(* F| Execution of a construction *)

let gen_rel sigma env name var j = 
let jkind = j.KIND
and varkind = cast_type var
in if isprop jkind then error "Proof objects can only be abstracted" 
  else let typ = var in
       let j = {VAL = j.VAL; TYP=whd_betadeltaiota sigma j.TYP; KIND = j.KIND}
       in (match whd_betadeltaiota sigma j.TYP with 
                 DOP0(Sort(Prop(_))) as typ' -> (* quantification : TYP stays Prop *)
                 {VAL=DOP2(Prod,typ,DLAM(name,j_val j));
                  TYP= j.TYP;
                  KIND = jkind}

               | (DOP0(Sort(Type(_,_))) as typ') -> (* product *)
                 let res_type = type_of_product sigma varkind typ' in
                 {VAL=DOP2(Prod,typ,DLAM(name,j_val j));
                  TYP=res_type;
                  KIND = type_of_type res_type}

               | _ -> errorlabstrm "gen_rel"
                 [< 'S"Type Error (generalization)" ; 'SPC ;
                    'S"Should be typed by Prop or Type" ; 'SPC ;
                    'S"In environment " ; pENV env ; 'FNL ;
                    'S"Cannot generalize " ; pTERMINENV(env,var) ; 'SPC ;
                    'S"over " ; 'SPC ;
                    pTERMINENV(add_rel (name,var) env,j.VAL) >]
);;

let abs_rel sigma name var j =
 let domtyp = cast_type var in
 let typ = var in
     {VAL=DOP2(Lambda,typ,DLAM(name,j_val j)); 
      TYP=DOP2(Prod,typ,DLAM(name,j.TYP));
      KIND = type_of_product sigma domtyp j.KIND};;

let abs_var sigma id var j =
    (abs_rel sigma (Name id) var (j_subst_var id j),ABSTRACT);;

let relative sigma n env = 
try
    let (_,typ) = lookup_rel n env in
    let typtyp = cast_type typ
    in {VAL=Rel(n);
        (* We cannot assume, as in the normal machine operation, that the 
         types of variables are closed terms, and thus we must lift them *)
        TYP=lift n typ;
        KIND = lift n (whd_betadeltaiota sigma typtyp)}
with Not_found -> errorlabstrm "relative"
                  [< 'S"Tried to typecheck an open term in a environment which did not define all of its free references" >]
;;

let apply_error env rator randl =
    [< 'S"In environment: " ; pENV env ; 'FNL ;
       'S"The term " ; pTERMINENV(env,fst rator) ; 'SPC ;
       'S": " ; pTERMINENV(env,snd rator); 'SPC ;
       'S"cannot be applied to the terms" ; 'FNL ;
       prlist_with_sep (fun () -> [< 'FNL >])
                       (fun c -> [< pTERMINENV(env,fst c) ; 'SPC ; 'S": " ;
                                    pTERMINENV(env,snd c) >])
                       randl >]
;;

let apply_rel_list sigma env nocheck argjl funj =
let rec apply_rec typ = function
    [] -> {VAL=applist(j_val funj,map j_val argjl);
           TYP=typ;
           KIND = funj.KIND}
  | hj::restjl ->
  match whd_betadeltaiota sigma typ with
     DOP2(Prod,c1,DLAM(_,c2)) ->
     if nocheck or conv_leq sigma hj.TYP c1 then
         apply_rec (subst1 hj.VAL c2) restjl
        else (errorlabstrm "apply"
              [< 'S"Illegal application (Type Error) : "; 'CUT ;
                 apply_error env (funj.VAL,funj.TYP)
                                 (map (fun j -> (j.VAL,j.TYP)) argjl) >])
   | _ -> errorlabstrm "apply"
              [< 'S"Illegal application (Non-functional construction) : ";
                 'CUT ;
                 apply_error env (funj.VAL,funj.TYP)
                                 (map (fun j -> (j.VAL,j.TYP)) argjl) >]
in apply_rec funj.TYP argjl;;

let cast_rel sigma env cj tj =
    if conv_leq sigma cj.TYP tj.VAL then
        {VAL=j_val cj;
         TYP=tj.VAL;
         KIND = whd_betadeltaiota sigma tj.TYP}
    else errorlabstrm "cast"
        [< 'S"In environment " ; pENV env ; 'FNL ;
           'S"The term: " ; pTERMINENV(env,cj.VAL) ; 'SPC ; 
           'S"does not have type"; 'SPC ; pTERMINENV(env,tj.VAL) ; 'S"." ;
           'FNL; 'S"Actually, it has type " ;
           'SPC ; pTERMINENV(env,cj.TYP) >]
;;

let type_of_constant sigma env (DOPN(Const sp,cl) as K) =
    let (DOPN(_,cl') as K') = global_reference env sp (basename sp)
    in if for_all2eq_vect (conv_x sigma) cl cl' then
        const_type sigma K
       else error ("the current environment and the constant's env of construction are incompatible")
;;

let type_of_const sigma env (DOPN(Const sp,_) as K) =
let bn = basename sp
in if "?" = SA__un_intern(atompart_of_id bn) then
    type_of_existential sigma env K
   else type_of_constant sigma env K
;;

let inf_of_const sigma (env,fenv) (DOPN(Const sp as gl,args) as K) =
try (let (_,cb) = environ__const_of_path sp
     in  if informative_constant K then
         let fsp = fwsp_of sp in
         let infK = global_reference fenv fsp (id_of_global gl) in
         let (DOP2(Cast,typ,kind)) = const_type sigma infK
         in Inf{VAL= infK;
                TYP = typ;
                KIND = kind}
         else Logic)
with Not_found -> error ((string_of_path sp) ^ " not declared")
   | Failure _ -> invalid_arg "cannot convert a constant to an fconstant"
;;

let type_of_mconstr sigma env (DOPN(MutConstruct(sp,i) as gl,cl)) =
    let I = DOPN(MutInd sp,cl) in
    let (DOPN(_,cl') as I') = global_reference env (fst sp) (id_of_global gl)
    in if for_all2eq_vect (conv_x sigma) cl cl' then
        type_mconstruct sigma i I
       else error ("the current environment and the constant's env of construction are incompatible")
;;


let type_of_mind sigma env (DOPN(MutInd (sp,_) as gl,cl) as I) =
    let (DOPN(_,cl') as I') = global_reference env sp (id_of_global gl)
    in if for_all2eq_vect (conv_x sigma) cl cl' then
        mind_arity I
       else error ("the current environment and the constant's env of construction are incompatible")
;;

let verify_wf_env env =
    if not(for_all (function (DOP2(Cast,_,_)) -> true | _ -> false) (vals_of_sign (get_globals env)))
        or not(for_all (function (_,DOP2(Cast,_,_)) -> true | _ -> false) (get_rels env))
        then invalid_arg "the constructive engine was fed a malformed environment"
;;

let cast_of_judgement j = 
    let jv = j.VAL in 
   match jv with DOP2(Cast,_,_) -> jv 
               | _              -> DOP2(Cast,jv,j.TYP);;

(* $Id: machops.ml,v 1.6 1996/02/12 15:47:53 cpaulin Exp $ *)
