(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                             proof_trees.mli                              *)
(****************************************************************************)
#open "initial";;
#open "names";;
#open "more_util";;
#open "ast";;
#open "term";;
#open "constrtypes";;
#open "pp";;
#open "stdpp";;
#open "ast";;

(* The module proof declares the structure of proof trees and a few
   utilities on them *)

(* We keep track of whether proofs are completed or not *)
type pf_status = COMPLETE_PROOF | INCOMPLETE_PROOF;;

type BindOcc = 
    Dep of identifier
  | NoDep of int
  | Com
;;

(* prim_rules are the primitive rules of the base theory, here CIC *)
(* Ex: {name="INTRO"; hypspects=params=terms=[]; newids=["x"]} for Intro x.
       {name="EXACT"; hypspects=params=newids=[]; terms=[t]} for Exact t.
       {name="CONVERT_HYP"; hypspects=["h"]; params=newids=[]; terms=[T]} 
          for Change T in h.  *)


type prim_rule = {name : string;
                  hypspecs : identifier list;
                  newids : identifier list;
                  params : ast__tok list;
                  terms : constr list};;

(* A tactic expression (tacname,args) refers to tactic tacname (indexing into
   a table in tacmach) applied to parameters args *)

type tactic_arg = (* reflects the current version of vernac *)
    COMMAND of ast
  | CONSTR of constr
  | IDENTIFIER of identifier
  | INTEGER of int
  | BINDING of BindOcc * ast
  | PATTERN of int list * ast
  | UNFOLD of int list * identifier
  | QUOTED_STRING of string
  | TACEXP of ast
  | FIXEXP of identifier * int * ast
  | COFIXEXP of identifier * ast

and tactic_expression == string * tactic_arg list
;;

(* When the refinement slot ref=SOME(R=TACTIC ... , ...) 
   then subproof = SOME(...) holds the proof produced by the tactic in R; 
   otherwise, subproof = NONE. *)

(* 'a proofty, 'a goalty, 'a rulety and 'a ctxtty will be used only with 'a=spavs__t (spavs__t is a set of section paths) *)

type 'a proofty =
    {status:pf_status; (* the proof is complete or incomplete *)
     goal: 'a goalty; (* the goal *)
     ref:('a rulety * 'a proofty list) option; (* NONE if the goal has still
                to be proved, SOME (r,l) if the rule r was applied to the goal
                and gave l as subproofs to be completed *)
     subproof : 'a proofty option (* (SOME p) if ref=(SOME(TACTIC t,l)) ;
                p is then the proof that the goal can be proven if the goals
                in l are solved *)
     }

and 'a goalty = {ctxt: 'a ctxtty;  (* context of parameters to tactics *)
            lc : 'a;        (* local constraints as section paths sets *)
            hyps:constr signature;
            concl:constr}

and 'a rulety =
    PRIM of prim_rule
  | TACTIC of tactic_expression
  | CONTEXT of 'a ctxtty
  | LOCAL_CONSTRAINTS of 'a (* for changing the lc of current goal *)

and 'a ctxtty = {pgm : constr option;    (* for Program tactic *)
            mimick : 'a proofty option}  (* for a soon to exist mimick tactic *)
;;

value mkGOAL : 'a -> 'a ctxtty -> constr signature -> constr -> 'a goalty;;

value mt_ctxtty : 'a ctxtty;;

value get_pgm : 'a ctxtty -> constr option;;
value set_pgm : constr option -> 'a ctxtty -> 'a ctxtty;;

value get_mimick : 'a ctxtty -> 'a proofty option;;
value set_mimick : 'a proofty option -> 'a ctxtty -> 'a ctxtty;;

value rule_of_proof : 'a proofty -> 'a rulety;;
value ref_of_proof : 'a proofty -> ('a rulety * 'a proofty list);;
value children_of_proof : 'a proofty -> 'a proofty list;;
value goal_of_proof : 'a proofty -> 'a goalty;;
value subproof_of_proof : 'a proofty -> 'a proofty;;
value status_of_proof : 'a proofty -> pf_status;;
value is_complete_proof : 'a proofty -> bool;;
value is_leaf_proof : 'a proofty -> bool;;
value is_tactic_proof : 'a proofty -> bool;;

(* $Id: proof_trees.mli,v 1.5 1996/02/12 15:48:17 cpaulin Exp $ *)
