(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                               reduction.ml                               *)
(****************************************************************************)

#infix "o";;
#open "std";;
#open "names";;
#open "more_util";;
#open "lomsg";;
#open "vectops";;
#open "generic";;
#open "impuniv";;
#open "pp";;
#open "stdpp";;
#open "term";;
#open "constrtypes";;
#open "environ";;
#open "termenv";;
#open "proof";;
#open "himsg";;

(****************************************************************************)
(*                 Term Manipulation Functions                              *)
(****************************************************************************)

let abs_implicit c = DOP2(Lambda,DOP0(Implicit),DLAM(Anonymous,c));;

let rec hdchar = 
let lower_case c = let n=ascii_code(c) in if n<97 then ascii (n+32) else c
in function
     DOP2(Prod,_,DLAM(_,c))     -> hdchar c
   | DOP2(Cast,c,_)             -> hdchar c
   | DOPN(AppL,cl)              -> hdchar (hd_vect cl)
   | DOP2(Lambda,_,DLAM(_,c))         -> hdchar c
   | DOPN(Const _,_) as x -> lower_case(hd(explode_id (basename(path_of_const x))))
   | DOPN(Abst _,_) as x -> lower_case(hd(explode_id (basename(path_of_abst x))))
   | DOPN(MutInd (sp,i) as x,_) -> if i=0 then lower_case(hd(explode_id (basename sp)))
                         else let na = id_of_global x  in lower_case(hd(explode_id na))
   | DOPN(MutConstruct(sp,i) as x,_) ->
     let na = id_of_global x  in lower_case(hd(explode_id na))

   | VAR _ as x  -> lower_case(hd(explode_id (id_of_var x)))
   | DOP0(Sort s)          -> sort_hdchar s
   | _                     -> "y";;
   
let named_hd A = function Anonymous -> Name(id_of_string(hdchar A)) 
                        | x         -> x;;

(* For manipulation of lists of terms *)

let prod_it = it_list (fun c (n,t)  ->  DOP2(Prod,t,DLAM(n,c)))
and arrow c1 c2 = DOP2(Prod,c1,DLAM(Anonymous,c2));;
let arrow_it = list_it arrow;;
let prod_name (n,A,B) = DOP2(Prod,A,DLAM(named_hd A n,B))
and lambda_name (n,A,B) = DOP2(Lambda,A,DLAM(named_hd A n,B))
and prod_create (A,B) = DOP2(Prod,A,DLAM(Name(id_of_string(hdchar A)),B))
and lambda_create (A,B) =  DOP2(Lambda,A,DLAM(Name(id_of_string(hdchar A)),B));;
let lambda_implicit A = DOP2(Lambda,DOP0(Implicit),DLAM(Name(id_of_string"y"),A));;

let lambda_implicit_lift n A = iterate lambda_implicit n (lift n A);;

(****************************************************************************)
(*                   Reduction Functions                                    *)
(****************************************************************************)


let stacklam recfun = lamrec where rec lamrec sigma = fun
    (h::t) (DOP2(Lambda,_,DLAM(_,c))) -> lamrec (h::sigma) t c
  | stack t -> recfun stack (substl sigma t)
;;


let project_fix_body i = projrec
    where rec projrec = fun
    (h::t) (DLAM _ as B) -> projrec t (SAPP B h)
  | [h] (DLAMV _ as B) -> (SAPPVi i B h)
  | _ _ -> failwith "project_fix_body"
;;


let strong whdfun = strec
    where rec strec t =
    match whdfun t with
    DOP0 _ as t -> t
  | DOP1(oper,c) -> DOP1(oper,strec c)
  | DOP2(oper,c1,c2) -> DOP2(oper,strec c1,strec c2)
  | DOPN(oper,cl) -> DOPN(oper,map_vect strec cl)
  | DOPL(oper,cl) -> DOPL(oper,map strec cl)
  | DLAM(na,C) -> DLAM(na,strec C)
  | DLAMV(na,C) -> DLAMV(na,map_vect strec C)
  | VAR _ as t -> t
  | Rel _ as t -> t
;;


let rec under_casts f = function
    DOP2(Cast,c,t) -> DOP2(Cast,under_casts f c, t)
  | c              -> f c
;;

let whd_beta_stack = whrec
  where rec whrec x stack = match x with
    DOP2(Lambda,c1,DLAM(name,c2)) ->
      (match stack with
         []       -> x,[]
       | a1::rest -> stacklam (fun l x -> whrec x l) [a1] rest c2)

  | DOPN(AppL,cl)  -> whrec (hd_vect cl) (app_tl_vect cl stack)
  | DOP2(Cast,c,_) -> whrec c stack
  | x              -> x,stack
;;

let whd_beta x = applist(whd_beta_stack x []);;

let beta_applist (c,l) = stacklam (fun l c -> applist(c,l)) [] l c;;

(***
let whd_betaimplicit_stack = whrec
    where rec whrec x stack =
    match x with
    DOP2(Lambda,DOP0(Implicit),DLAM(name,c2)) ->
    (match stack with
     [] -> x,[]
   | a1::rest -> stacklam (fun l x -> whrec x l)[a1] rest c2)

  | DOPN(AppL,cl)      -> whrec (hd_vect cl) (app_tl_vect cl stack)
  | DOP2(Cast,c,_)       ->  whrec c stack
  | x -> x,stack
;;

let whd_betaimplicit x = applist(whd_betaimplicit_stack x []);;
***)

type miota_args = {mP : constr; (* the result type *)
                   mconstr : constr; (* the constructor *)
                   mcargs : constr list; (* the constructor's arguments *)
                   mlf : constr vect} (* the branch code vector *)
;;


let contract_cofix cofix =
let (DOPN(CoFix(bodynum),bodyvect)) = cofix in
let nbodies = (vect_length bodyvect) -1 in
let make_Fi j = DOPN(CoFix(j),bodyvect) in
    SAPPViList bodynum (last_vect bodyvect) (tabulate_list make_Fi nbodies)                     
;;

let reducible_mind_case c =
   match c with 
      DOPN(MutConstruct _,_) -> true
    | DOPN(CoFix _,_)        -> true
    | _                      -> false
;;


let reduce_mind_case mia =
  match mia.mconstr with 
     (DOPN(MutConstruct((indsp,tyindx),i),_)) ->
       let I = DOPN(MutInd(indsp,tyindx),args_of_mconstr mia.mconstr) in
       let nparams    = mind_nparams I in
       let real_cargs = snd(chop_list nparams mia.mcargs)
       in applist (mia.mlf.(i-1),real_cargs)
   | (DOPN(CoFix _,_)) as cofix ->
       let cofix_def = contract_cofix cofix
       in  DOPN(MutCase,
                 concat_vect [|mia.mP;applist(cofix_def,mia.mcargs)|] mia.mlf)
;;

(* contracts fix==FIX[nl;i](A1...Ak;[F1...Fk]{B1....Bk}) to produce

   Bi[Fj --> FIX[nl;j](A1...Ak;[F1...Fk]{B1...Bk})]

 *)

let contract_fix fix =
let (DOPN(Fix(recindices,bodynum),bodyvect)) = fix in
let nbodies = vect_length recindices in
let make_Fi j = DOPN(Fix(recindices,j),bodyvect) in
    SAPPViList bodynum (last_vect bodyvect) (tabulate_list make_Fi nbodies)
                     
;;

let fix_recarg fix =
let (DOPN(Fix(recindices,bodynum),bodyvect)) = fix in
if vect_length recindices < bodynum then NONE
else SOME(vect_item recindices bodynum)
;;

let list_assign l n e = assrec [] l n
    where rec assrec deb = fun
    (h::t) 0 -> rev_append deb (e::t)
  | (h::t) n -> assrec (h::deb) t (n-1)
  | [] n -> failwith "list_assign"
;;

let reduce_fix whfun fix stack =
try
let (DOPN(Fix(recindices,bodynum),bodyvect)) = fix in
    (match fix_recarg fix with
     NONE -> (false,(fix,stack))
   | SOME recargnum ->
     if recargnum < 0 then (false,(fix,stack)) else
     let recarg = item stack recargnum in
     let (recarg'hd,_ as recarg')= whfun recarg [] in
     let stack' = list_assign stack recargnum (applist recarg') in
         (match recarg'hd with
          DOPN(MutConstruct _,_) -> (true,(contract_fix fix,stack'))
        | _ -> (false,(fix,stack'))))
with Failure _ -> (false,(fix,stack))
;;


let whd_betaiota_stack = whrec
    where rec whrec x stack =
    match x with
    DOP2(Cast,c,_) -> whrec c stack
  | DOPN(AppL,cl)    -> whrec (hd_vect cl) (app_tl_vect cl stack)
  | DOP2(Lambda,_,DLAM(_,c)) ->
    (match stack with
     [] -> x,stack
   | (a::m) -> stacklam (fun x y -> whrec y x) [a] m c)
  | DOPN(MutCase,cl) ->
    let P = hd_vect cl
    and (c,cargs) = whrec (cadr_vect cl) []
    and lf = cddr_vect cl
    in if reducible_mind_case c
       then whrec (reduce_mind_case {mP=P;
                                     mconstr=c;
                                     mcargs=cargs;
                                     mlf=lf}) stack
       else (DOPN(MutCase,concat_vect [|P;applist(c,cargs)|] lf),stack)

  | DOPN(Fix _,_) as fix ->
    let (reduced,(fix,stack)) = reduce_fix whrec fix stack
    in if reduced then whrec fix stack
       else (fix,stack)

  | x             -> x,stack
;;
let whd_betaiota x = applist(whd_betaiota_stack x []);;

let whd_const_stack namelist sigma = whrec
    where rec whrec x l =
    match x with
    (DOPN(Const sp,_)) as c      ->
    if listset__memb namelist sp then
        if evaluable_const sigma c then
            whrec (const_value sigma c) l
        else error "whd_const_stack"
    else x,l

  | (DOPN(Abst sp,_)) as c      ->
    if listset__memb namelist sp then
        if evaluable_abst c then
            whrec (abst_value c) l
        else error "whd_const_stack"
    else x,l

  | DOP2(Cast,c,_) -> whrec c l
  | DOPN(AppL,cl)    -> whrec (hd_vect cl) (app_tl_vect cl l)
  | x             -> x,l;;

let whd_const namelist sigma c = applist(whd_const_stack namelist sigma c []);;

let whd_delta_stack sigma = whrec
    where rec whrec x l =
    match x with
    (DOPN(Const _,_)) as c      ->
    if evaluable_const sigma c then
           whrec (const_value sigma c) l
    else x,l
  | (DOPN(Abst _,_)) as c      ->
    if evaluable_abst c then
           whrec (abst_value c) l
    else x,l

  | DOP2(Cast,c,_) -> whrec c l
  | DOPN(AppL,cl)    -> whrec (hd_vect cl) (app_tl_vect cl l)
  | x             -> x,l;;

let whd_delta sigma c = applist(whd_delta_stack sigma c []);;

let whd_betadelta_stack sigma = whrec
    where rec whrec x l =
    match x with
    (DOPN(Const _,_)) as c      ->
    if evaluable_const sigma c then
           whrec (const_value sigma c) l
    else x,l
  | (DOPN(Abst _,_)) as c      ->
    if evaluable_abst c then
           whrec (abst_value c) l
    else x,l

  | DOP2(Cast,c,_) -> whrec c l
  | DOPN(AppL,cl)    -> whrec (hd_vect cl) (app_tl_vect cl l)
  | DOP2(Lambda,_,DLAM(_,c)) ->
    (match l with
     [] -> x,l
   | (a::m) -> stacklam (fun x y -> whrec y x) [a] m c)
  | x             -> x,l;;

let whd_betadelta sigma c = applist(whd_betadelta_stack sigma c []);;

let whd_betadeltat_stack sigma = whrec
    where rec whrec x l =
    match x with
    (DOPN(Const _,_)) as c      ->
    if translucent_const sigma c then
           whrec (const_value sigma c) l
    else x,l

  | (DOPN(Abst _,_)) as c      ->
    if translucent_abst c then
           whrec (abst_value c) l
    else x,l

  | DOP2(Cast,c,_) -> whrec c l
  | DOPN(AppL,cl)    -> whrec (hd_vect cl) (app_tl_vect cl l)
  | DOP2(Lambda,_,DLAM(_,c)) ->
    (match l with
     [] -> x,l
   | (a::m) -> stacklam (fun x y -> whrec y x) [a] m c)
  | x             -> x,l;;

let whd_betadeltat sigma c = applist(whd_betadeltat_stack sigma c []);;

let whd_betadeltaeta_stack sigma = whrec
    where rec whrec x stack =
    match x with
    (DOPN(Const _,_)) as c      ->
    if evaluable_const sigma c then whrec (const_value sigma c) stack
    else x,stack
  | (DOPN(Abst _,_)) as c      ->
    if evaluable_abst c then whrec (abst_value c) stack
    else x,stack

  | DOP2(Cast,c,_) -> whrec c stack
  | DOPN(AppL,cl)    -> whrec (hd_vect cl) (app_tl_vect cl stack)
  | DOP2(Lambda,_,DLAM(_,c)) ->
    (match stack with
     [] -> (match applist (whrec c []) with 
               (DOPN(AppL,cl)) -> 
                    (match whrec (last_vect cl) [] with 
                       (Rel 1,[]) -> let napp = (vect_length cl) -1 in
                       if napp = 0 then (x,stack) else
                       let lc = sub_vect cl 0 napp in
                       let u = if napp = 1 then lc.(0) else DOPN(AppL,lc) 
                            in if noccurn 1 u then (pop u,[]) else (x,stack)
                        | _ -> (x,stack))
             | _ -> (x,stack))
   | (a::m) -> stacklam (fun x y -> whrec y x) [a] m c)

  | x -> x,stack;;

let whd_betadeltaeta sigma x = applist(whd_betadeltaeta_stack sigma x []);;

let whd_betadeltatiota_stack sigma = whrec
    where rec whrec x stack =
    match x with
    (DOPN(Const _,_)) as c      ->
    if translucent_const sigma c then
        whrec (const_value sigma c) stack
    else x,stack

  | (DOPN(Abst _,_)) as c      ->
    if translucent_abst c then
           whrec (abst_value c) stack
    else x,stack

  | DOP2(Cast,c,_) -> whrec c stack
  | DOPN(AppL,cl)    -> whrec (hd_vect cl) (app_tl_vect cl stack)
  | DOP2(Lambda,_,DLAM(_,c)) ->
    (match stack with
     [] -> x,stack
   | (a::m) -> stacklam (fun x y -> whrec y x) [a] m c)

  | DOPN(MutCase,cl) ->
    let P = hd_vect cl
    and (c,cargs) = whrec (cadr_vect cl) []
    and lf = cddr_vect cl
    in if reducible_mind_case c
       then whrec (reduce_mind_case {mP=P;
                                     mconstr=c;
                                     mcargs=cargs;
                                     mlf=lf}) stack
       else (DOPN(MutCase,concat_vect [|P;applist(c,cargs)|] lf),stack)

  | DOPN(Fix _,_) as fix ->
    let (reduced,(fix,stack)) = reduce_fix whrec fix stack
    in if reduced then whrec fix stack
       else (fix,stack)

  | x             -> x,stack
;;
let whd_betadeltatiota sigma x = applist(whd_betadeltatiota_stack sigma x []);;


let whd_betadeltaiota_stack sigma = whrec
    where rec whrec x stack =
    match x with
    (DOPN(Const _,_)) as c      ->
    if evaluable_const sigma c then
        whrec (const_value sigma c) stack
    else x,stack
  | (DOPN(Abst _,_)) as c      ->
    if evaluable_abst c then
           whrec (abst_value c) stack
    else x,stack

  | DOP2(Cast,c,_) -> whrec c stack
  | DOPN(AppL,cl)    -> whrec (hd_vect cl) (app_tl_vect cl stack)
  | DOP2(Lambda,_,DLAM(_,c)) ->
    (match stack with
     [] -> x,stack
   | (a::m) -> stacklam (fun x y -> whrec y x) [a] m c)

  | DOPN(MutCase,cl) ->
    let P = hd_vect cl
    and (c,cargs) = whrec (cadr_vect cl) []
    and lf = cddr_vect cl
    in if reducible_mind_case c
       then whrec (reduce_mind_case {mP=P;
                                     mconstr=c;
                                     mcargs=cargs;
                                     mlf=lf}) stack
       else (DOPN(MutCase,concat_vect [|P;applist(c,cargs)|] lf),stack)

  | DOPN(Fix _,_) as fix ->
    let (reduced,(fix,stack)) = reduce_fix whrec fix stack
    in if reduced then whrec fix stack
       else (fix,stack)

  | x             -> x,stack
;;
let whd_betadeltaiota sigma x = applist(whd_betadeltaiota_stack sigma x []);;

let compute_stack sigma = whrec
    where rec whrec x stack =
    match x with
    (DOPN(Const _,_)) as c      ->
    if evaluable_const sigma c then
        whrec (const_value sigma c) stack
    else x,(map wh_emp stack)
  | (DOPN(Abst _,_)) as c      ->
    if evaluable_abst c then
           whrec (abst_value c) stack
    else x,(map wh_emp stack)

  | DOP2(Cast,c,_) -> whrec c stack
  | DOPN(AppL,cl)    -> whrec (hd_vect cl) (app_tl_vect cl stack)
  | DOP2(Lambda,_,DLAM(_,c)) ->
    (match stack with
     [] -> x,stack
   | (a::m) -> stacklam (fun x y -> whrec y x) [a] m c)

  | DOPN(MutCase,cl) ->
    let P = hd_vect cl
    and (c,cargs) = whrec (cadr_vect cl) []
    and lf = cddr_vect cl
    in if reducible_mind_case c
       then whrec (reduce_mind_case {mP=P;
                                     mconstr=c;
                                     mcargs=cargs;
                                     mlf=lf}) stack
       else (DOPN(MutCase,
                   concat_vect [|P;applist(c,cargs)|] (map_vect wh_emp lf)),
             map wh_emp stack)

  | DOPN(Fix _,_) as fix ->
    let (reduced,(fix,stack)) = reduce_fix whrec fix stack
    in if reduced then whrec fix stack
       else (fix,map wh_emp stack)

  | x             -> x,map wh_emp stack
and wh_emp x = applist (whrec x [])
;;

let compute sigma x = applist(compute_stack sigma x []);;

let whd_betadeltaiotaeta_stack sigma = whrec
    where rec whrec x stack =
    match x with
    (DOPN(Const _,_)) as c      ->
    if evaluable_const sigma c then
        whrec (const_value sigma c) stack
    else x,stack
  | (DOPN(Abst _,_)) as c      ->
    if evaluable_abst c then
           whrec (abst_value c) stack
    else x,stack

  | DOP2(Cast,c,_) -> whrec c stack
  | DOPN(AppL,cl)    -> whrec (hd_vect cl) (app_tl_vect cl stack)
  | DOP2(Lambda,_,DLAM(_,c)) ->
    (match stack with
     [] -> (match applist (whrec c []) with 
               (DOPN(AppL,cl)) -> 
                    (match whrec (last_vect cl) [] with 
                       (Rel 1,[]) -> let napp = (vect_length cl) -1 in
                       if napp = 0 then (x,stack) else
                       let lc = sub_vect cl 0 napp in
                       let u = if napp = 1 then lc.(0) else DOPN(AppL,lc) 
                            in if noccurn 1 u then (pop u,[]) else (x,stack)
                        | _ -> (x,stack))
             | _ -> (x,stack))
   | (a::m) -> stacklam (fun x y -> whrec y x) [a] m c)

  | DOPN(MutCase,cl) ->
    let P = hd_vect cl
    and (c,cargs) = whrec (cadr_vect cl) []
    and lf = cddr_vect cl
    in if reducible_mind_case c
       then whrec (reduce_mind_case {mP=P;
                                     mconstr=c;
                                     mcargs=cargs;
                                     mlf=lf}) stack
       else (DOPN(MutCase,concat_vect [|P;applist(c,cargs)|] lf),stack)

  | DOPN(Fix _,_) as fix ->
    let (reduced,(fix,stack)) = reduce_fix whrec fix stack
    in if reduced then whrec fix stack
       else (fix,stack)

  | x             -> x,stack
;;
let whd_betadeltaiotaeta sigma x = applist(whd_betadeltaiotaeta_stack sigma x []);;

let redabs = strong whd_beta;;
let redapplist stack x = strong whd_beta (applist(x,stack));;

let find_Inductype sigma c =
let t,l = whd_betadeltaiota_stack sigma c []
in match t with
    DOPN(MutInd _,_) -> t,l
  | _ -> raise Induc
;;

let find_mrectype sigma c =
let t,l = whd_betadeltaiota_stack sigma c []
in match t with
    DOPN(MutInd (sp,i),_) ->  t,l
  | _ -> raise Induc
;;

let find_minductype sigma c =
let t,l = whd_betadeltaiota_stack sigma c []
in match t with
    DOPN(MutInd (sp,i),_) -> 
      let _,mib = mind_of_path sp 
      in if mind_type_finite mib i 
         then  t,l
         else raise Induc
  | _ -> raise Induc
;;

let find_mcoinductype sigma c =
let t,l = whd_betadeltaiota_stack sigma c []
in match t with
    DOPN(MutInd (sp,i),_) -> 
      let _,mib = mind_of_path sp 
      in if mind_type_finite mib i 
         then raise Induc
         else t,l
  | _ -> raise Induc
;;

let minductype_spec sigma c = 
   try let (x,l) = find_minductype sigma c
       in if l<>[] then anomaly "minductype_spec: Not a recursive type 1"
          else x
   with Induc -> anomaly "minductype_spec: Not a recursive type 2"
;;

let mrectype_spec sigma c = 
   try let (x,l) = find_mrectype sigma c
       in if l<>[] then anomaly "mrectype_spec: Not a recursive type 1"
          else x
   with Induc -> anomaly "mrectype_spec: Not a recursive type 2"
;;

let check_mrectype_spec sigma c =
     try let (x,l) = find_mrectype sigma c
         in if l<>[] then error "check_mrectype: Not a recursive type 1" else x
     with Induc -> error "check_mrectype: Not a recursive type 2";;

(* One step of approximation *)

let rec apprec sigma stack c =
let t,stack = whd_betaiota_stack c stack
in match t with
    DOPN(MutCase,cl) ->
    let P          = hd_vect cl and
       (cr,crargs) = whd_betadeltaiota_stack sigma (cadr_vect cl) [] and
       lf          = cddr_vect cl in
    let rslt = DOPN(MutCase,concat_vect [|P;applist(cr,crargs)|] lf)
    in if   reducible_mind_case cr 
       then apprec sigma stack rslt
       else (t,stack)

  | DOPN(Fix(_),_) as fix ->
    let (reduced,(fix,stack)) 
            = reduce_fix (whd_betadeltaiota_stack sigma) fix stack
    in if reduced then apprec sigma stack fix
       else (fix,stack)

  | _ -> (t,stack)
;;

let hnf sigma = apprec sigma [];;

let red_product sigma = (strong whd_betaiota) o redrec
    where rec redrec = function
   DOPN(AppL,cl)      -> DOPN(AppL,concat_vect [|redrec (hd_vect cl)|] (tl_vect cl))
 | DOP2(Prod,c1,DLAM(s,c2))   -> DOP2(Prod,c1,DLAM(s,redrec c2))
 | DOPN(Const _,_) as x        ->
   if evaluable_const sigma x then
       const_value sigma x
   else x
  | (DOPN(Abst _,_)) as c      ->
    if evaluable_abst c then
           abst_value c
    else c

 | DOP2(Cast,c,_)       -> redrec c
 | _               -> error "Term not reducible";;



(* pseudo-reduction rule:
 * [hnf_prod_app sigma s (Prod(_,B)) N --> B[N]
 * with an HNF on the first argument to produce a product.
 * if this does not work, then we use the string S as part of our
 * error message.
 *)
let hnf_prod_app sigma s T N =
    match whd_betadeltaiota sigma T with
    DOP2(Prod,_,B) -> SAPP B N
  | _ ->
    errorlabstrm s [< 'S"Needed a product, but didn't find one in " ;
                      'S s ; 'FNL >]
;;

let hnf_prod_appvect sigma s T NL = it_vect (hnf_prod_app sigma s) T NL;;
let hnf_prod_applist sigma s T NL = it_list (hnf_prod_app sigma s) T NL;;

let hnf_lam_app sigma s T N =
    match whd_betadeltaiota sigma T with
    DOP2(Lambda,_,B) -> SAPP B N
  | _ ->
    errorlabstrm s [< 'S"Needed a product, but didn't find one in " ;
                      'S s ; 'FNL >]
;;
let hnf_lam_appvect sigma s T NL = it_vect (hnf_lam_app sigma s) T NL;;
let hnf_lam_applist sigma s T NL = it_list (hnf_lam_app sigma s) T NL;;

let splay_prod sigma = decrec []
  where rec decrec m c =
   match whd_betadelta sigma c with
     DOP2(Prod,A,DLAM(n,C))            -> decrec ((n,A)::m) C
   | T -> m,T
;;

let decomp_prod sigma = decrec 0
  where rec decrec m c =
   match whd_betadelta sigma c with
     (DOP0(Sort _) as x) -> m,x
   | DOP2(Prod,A,DLAM(n,C))            -> decrec (m+1) C
   | _                      -> error "decomp_prod: Not a product";;

let decomp_n_prod sigma n = decrec n []
  where rec decrec m ln c = if m = 0 then (ln,c) else 
   match whd_betadelta sigma c with
     DOP2(Prod,A,DLAM(n,C)) -> decrec (m-1) ((n,A)::ln) C
   | _                      -> error "decomp_n_prod: Not enough products";;

(********************************************************************)
(*                         Conversion                               *)
(********************************************************************)

let sort_eq = fun
    (Prop c1) (Prop c2) -> c1 = c2
  | (Type(c1,u1)) (Type(c2,u2)) -> c1 = c2 & (impuniv__enforceq u1 u2;true)
  | _ _ -> false;;

let sort_leq = fun
    (Prop c1) (Prop c2) -> c1 = c2
  | (Type(c1,u1)) (Type(c2,u2)) -> c1 = c2 & (impuniv__enforcegeq u2 u1;true)
  | _ _ -> false;;

(****
let conv_minductive convfun
    (DOPN(MutInd (sp1,i1),cl1)) (DOPN(MutInd (sp2,i2),cl2))  =
    let _,mib1 = mind_of_path sp1 and
        _,mib2 = mind_of_path sp2 in
  (*  let finite1 = mind_type_finite mib1 i1 and
        finite2 = mind_type_finite mib2 i2 
    in  finite1=finite2 & *) sp1=sp2 & i1=i2 & for_all2eq_vect convfun cl1 cl2
;;

let conv_mconstructor convfun
    (DOPN(MutConstruct(indsp1,i1),cl1)) (DOPN(MutConstruct(indsp2,i2),cl2)) =
    i1 = i2
    & (indsp1 = indsp2
       or conv_minductive convfun (DOPN(MutInd indsp1,cl1))
                                  (DOPN(MutInd indsp2,cl2)))
;;
****)

(* equality of terms with universe adjustment *)
let rec conv sigma term1 term2 =
      term1 = term2 or eqappr sigma (hnf sigma term1,hnf sigma term2)
and eqappr sigma = function
       ((VAR v1,l1),(VAR v2,l2)) ->
           (eq(v1,v2) & (length(l1) = length(l2)) & (for_all2 (conv sigma) l1 l2))
     | ((DOP2(Cast,c,_),l),appr2) -> eqappr sigma ((c,l),appr2)
     | (appr1,(DOP2(Cast,c,_),l)) -> eqappr sigma (appr1,(c,l))
     | ((Rel(n),l1),((Rel(m),l2))) ->
           ((n=m) & (length(l1) = length(l2)) & (for_all2 (conv sigma) l1 l2))
     | ((DOP0(Meta(n)),l1),(DOP0(Meta(m)),l2)) ->
           ((n=m) & (length(l1) = length(l2)) & (for_all2 (conv sigma) l1 l2))
     | ((DOP0(Sort s1),[]),(DOP0(Sort s2),[])) -> sort_eq s1 s2
     | ((DOP2(Prod,c1,DLAM(_,c2)),[]),(DOP2(Prod,c'1,DLAM(_,c'2)),[])) -> (conv sigma) c1 c'1 & (conv sigma) c2 c'2
     | ((DOP2(Lambda,c1,DLAM(_,c2)),[]),(DOP2(Lambda,c'1,DLAM(_,c'2)),[])) ->
                                          (conv sigma) c1 c'1 & (conv sigma) c2 c'2
     | (((DOPN(Const sp1,al1),l1) as appr1),(DOPN(Const sp2,al2),l2 as appr2)) ->
       uni_or
       (fun () -> (* try first intensional equality *)
           (eq(sp1,sp2) & eq(al1,al2) & (length(l1) = length(l2)) & (for_all2 (conv sigma) l1 l2)))
       (fun () -> (* else expand the second occurrence (arbitrary heuristic) *)
            (if (evaluable_const sigma (DOPN(Const sp2,al2))) then
                 (eqappr sigma (appr1, apprec sigma l2 (const_value sigma (DOPN(Const sp2,al2)))))
             else if (evaluable_const sigma (DOPN(Const sp1,al1))) then
                 (eqappr sigma (apprec sigma l1 (const_value sigma (DOPN(Const sp1,al1))), appr2))
             else false))

     | ((DOPN(Const sp1,al1),l1),appr2)      ->
       (evaluable_const sigma (DOPN(Const sp1,al1)))
       & eqappr sigma (apprec sigma l1 (const_value sigma (DOPN(Const sp1,al1))), appr2)
     | (appr1,(DOPN(Const sp2,al2),l2))      ->
       (evaluable_const sigma (DOPN(Const sp2,al2)))
       & eqappr sigma (appr1, apprec sigma l2 (const_value sigma (DOPN(Const sp2,al2))))

     | (((DOPN(Abst sp1,al1),l1) as appr1),(DOPN(Abst sp2,al2),l2 as appr2)) ->
       uni_or
       (fun () -> (* try first intensional equality *)
           (eq(sp1,sp2) & eq(al1,al2) & (length(l1) = length(l2)) & (for_all2 (conv sigma) l1 l2)))
       (fun () -> (* else expand the second occurrence (arbitrary heuristic) *)
            (if (evaluable_abst (DOPN(Abst sp2,al2))) then
                 (eqappr sigma (appr1, apprec sigma l2 (abst_value (DOPN(Abst sp2,al2)))))
             else
                 (eqappr sigma (apprec sigma l1 (abst_value (DOPN(Abst sp1,al1))), appr2))))

     | ((DOPN(Abst sp1,al1),l1),appr2)      ->
       (evaluable_abst (DOPN(Abst sp1,al1)))
       & eqappr sigma (apprec sigma l1 (abst_value (DOPN(Abst sp1,al1))), appr2)
     | (appr1,(DOPN(Abst sp2,al2),l2))      ->
       (evaluable_abst (DOPN(Abst sp2,al2)))
       & eqappr sigma (appr1, apprec sigma l2 (abst_value (DOPN(Abst sp2,al2))))
     | ((DOPN(MutInd indsp1,cl1) as mind1,l'1),
        (DOPN(MutInd indsp2,cl2) as mind2,l'2)) ->
    indsp1 = indsp2 & for_all2eq_vect (conv sigma) cl1 cl2
    & for_all2eq (conv sigma) l'1 l'2
(***
       (uni_or
        (fun () -> (indsp1 = indsp2 & for_all2eq_vect (conv sigma) cl1 cl2))
        (fun () -> conv_minductive (conv sigma) mind1 mind2))
       & for_all2eq (conv sigma) l'1 l'2
***)
         
     | ((DOPN(MutConstruct i1,cl1) as mconstr1,l1),
        (DOPN(MutConstruct i2,cl2) as mconstr2,l2)) ->
        (i1=i2) & for_all2eq_vect (conv sigma) cl1 cl2 
      & for_all2eq (conv sigma) l1 l2
(**
        (uni_or
         (fun () -> ((i1=i2) & for_all2eq_vect (conv sigma) cl1 cl2))
         (fun () -> conv_mconstructor (conv sigma) mconstr1 mconstr2))
        & for_all2eq (conv sigma) l1 l2
**)

     | ((DOPN(Fix(i1),cl1),l1),(DOPN(Fix(i2),cl2),l2))   ->
       i1=i2 & 
       (for_all2eq_vect (conv sigma) cl1 cl2) &
       (for_all2eq (conv sigma) l1 l2)

     | ((DOPN(CoFix(i1),cl1),l1),(DOPN(CoFix(i2),cl2),l2))   ->
       i1=i2 & 
       (for_all2eq_vect (conv sigma) cl1 cl2) &
       (for_all2eq (conv sigma) l1 l2)

     | ((DOPN(MutCase,cl1),l1),(DOPN(MutCase,cl2),l2)) -> 
       (for_all2eq_vect (conv sigma) cl1 cl2) &
       (for_all2eq (conv sigma) l1 l2)

     | (DOP0(Implicit),[]),(DOP0(Implicit),[]) -> true

     | (DLAM(_,c1),[]),(DLAM(_,c2),[]) -> conv sigma c1 c2
     | (DLAMV(_,vc1),[]),(DLAMV(_,vc2),[]) -> for_all2eq_vect (conv sigma) vc1 vc2
     | _ -> false

(* we check that term1 <= term2 in the cumulative sense *)
and conv_leq sigma term1 term2 =
      term1 = term2 or eqappr_leq sigma (hnf sigma term1,hnf sigma term2)
and eqappr_leq sigma = function
       ((DOP2(Prod,c1,DLAM(_,c2)),[]),(DOP2(Prod,c'1,DLAM(_,c'2)),[])) -> 
                       (conv sigma) c1 c'1 & (conv_leq sigma) c2 c'2  (* Luo's system *)
     | ((DOP0(Sort s1),[]),(DOP0(Sort s2),[])) -> sort_leq s1 s2
     | ((DOP2(Cast,c,_),l),appr2) -> eqappr_leq sigma ((c,l),appr2)
     | (appr1,(DOP2(Cast,c,_),l)) -> eqappr_leq sigma (appr1,(c,l))

     | (((DOPN(Const sp1,al1),l1) as appr1),(DOPN(Const sp2,al2),l2 as appr2)) ->
       uni_or
       (fun () ->(* try first intensional equality *)
            (eq(sp1,sp2) & eq(al1,al2) & (length(l1) = length(l2)) & (for_all2 (conv sigma) l1 l2)))
       (fun () ->(* else expand the second occurrence (arbitrary heuristic) *)
            (if (evaluable_const sigma (DOPN(Const sp2,al2))) then
                 (eqappr_leq sigma (appr1,apprec sigma l2 (const_value sigma (DOPN(Const sp2,al2)))))
             else if (evaluable_const sigma (DOPN(Const sp1,al1))) then
                 (eqappr_leq sigma (apprec sigma l1 (const_value sigma (DOPN(Const sp1,al1))),appr2))
             else false))

     | ((DOPN(Const sp1,al1),l1),appr2)         ->
       (evaluable_const sigma (DOPN(Const sp1,al1)))
       & eqappr_leq sigma (apprec sigma l1 (const_value sigma (DOPN(Const sp1,al1))),appr2)
     | (appr1,(DOPN(Const sp2,al2),l2))         ->
       (evaluable_const sigma (DOPN(Const sp2,al2)))
       & eqappr_leq sigma (appr1,apprec sigma l2 (const_value sigma (DOPN(Const sp2,al2))))

     | (((DOPN(Abst sp1,al1),l1) as appr1),(DOPN(Abst sp2,al2),l2 as appr2)) ->
       uni_or
       (fun () ->(* try first intensional equality *)
            (eq(sp1,sp2) & eq(al1,al2) & (length(l1) = length(l2)) & (for_all2 (conv sigma) l1 l2)))
       (fun () ->(* else expand the second occurrence (arbitrary heuristic) *)
            (if (evaluable_abst (DOPN(Abst sp2,al2))) then
                 (eqappr_leq sigma (appr1,apprec sigma l2 (abst_value (DOPN(Abst sp2,al2)))))
             else
                 (eqappr_leq sigma (apprec sigma l1 (abst_value (DOPN(Abst sp1,al1))),appr2))))

     | ((DOPN(Abst sp1,al1),l1),appr2)         ->
       (evaluable_abst (DOPN(Abst sp1,al1)))
       & eqappr_leq sigma (apprec sigma l1 (abst_value (DOPN(Abst sp1,al1))),appr2)
     | (appr1,(DOPN(Abst sp2,al2),l2))         ->
       (evaluable_abst (DOPN(Abst sp2,al2)))
       & eqappr_leq sigma (appr1,apprec sigma l2 (abst_value (DOPN(Abst sp2,al2))))

     | x -> eqappr sigma x
;;

let conv_leq sigma M N = uni_pred (conv_leq sigma M) N;;

(* equality of terms without universe adjustment for synthesis *)
let rec conv_x sigma term1 term2 =
      term1 = term2 or eqappr_x sigma (hnf sigma term1,hnf sigma term2)
and eqappr_x sigma = function
       ((Rel(n),l1),(Rel(m),l2)) ->
           (n=m) & (length(l1) = length(l2)) & (for_all2 (conv_x sigma) l1 l2)
     | ((DOP2(Cast,c,_),l),appr2) -> eqappr_x sigma ((c,l),appr2)
     | (appr1,(DOP2(Cast,c,_),l)) -> eqappr_x sigma (appr1,(c,l))
     | ((VAR _ as v1,l1),(VAR _ as v2,l2)) ->
           (eq(id_of_var v1,id_of_var v2) & (length(l1) = length(l2)) & (for_all2 (conv_x sigma) l1 l2))
     | ((DOP0(Meta(n)),l1),(DOP0(Meta(m)),l2)) ->
           ((n=m) & (length(l1) = length(l2)) & (for_all2 (conv_x sigma) l1 l2))
     | ((DOP0(Sort s1),[]),(DOP0(Sort s2),[])) -> sort_eq_x s1 s2
     | ((DOP2(Lambda,c1,DLAM(_,c2)),[]),(DOP2(Lambda,c'1,DLAM(_,c'2)),[])) -> 
                                          (conv_x sigma) c1 c'1 & (conv_x sigma) c2 c'2
     | ((DOP2(Prod,c1,DLAM(_,c2)),[]),(DOP2(Prod,c'1,DLAM(_,c'2)),[])) -> 
                                          (conv_x sigma) c1 c'1 & (conv_x sigma) c2 c'2

     | ((DOPN(Const sp1,al1),l1),(DOPN(Const sp2,al2),l2)) ->
           (eq(sp1,sp2) & eq(al1,al2) & (length(l1) = length(l2)) & (for_all2 (conv_x sigma) l1 l2))
        or ((evaluable_const sigma (DOPN(Const sp2,al2)))
            & eqappr_x sigma  ((DOPN(Const sp1,al1),l1), apprec sigma l2 (const_value sigma (DOPN(Const sp2,al2)))))
        or ((not (evaluable_const sigma (DOPN(Const sp2,al2))))
            & (evaluable_const sigma (DOPN(Const sp1,al1)))
            & eqappr_x sigma  (apprec sigma l1 (const_value sigma (DOPN(Const sp1,al1))),(DOPN(Const sp2,al2),l2)))
     | ((DOPN(Const sp1,al1),l1),p2) ->  
       (evaluable_const sigma (DOPN(Const sp1,al1)))
       & eqappr_x sigma (apprec sigma l1 (const_value sigma (DOPN(Const sp1,al1))),p2)
     | (p1,(DOPN(Const sp2,al2),l2))  -> 
       (evaluable_const sigma (DOPN(Const sp2,al2)))
       & eqappr_x sigma (p1,apprec sigma l2 (const_value sigma (DOPN(Const sp2,al2))))

     | ((DOPN(Abst sp1,al1),l1),(DOPN(Abst sp2,al2),l2)) ->
           (eq(sp1,sp2) & eq(al1,al2) & (length(l1) = length(l2)) & (for_all2 (conv_x sigma) l1 l2))
        or ((evaluable_abst (DOPN(Abst sp2,al2)))
            & eqappr_x sigma  ((DOPN(Abst sp1,al1),l1), apprec sigma l2 (abst_value (DOPN(Abst sp2,al2)))))
        or ((not (evaluable_abst (DOPN(Abst sp2,al2))))
            & (evaluable_abst (DOPN(Abst sp1,al1)))
            & eqappr_x sigma  (apprec sigma l1 (abst_value (DOPN(Abst sp1,al1))),(DOPN(Abst sp2,al2),l2)))
     | ((DOPN(Abst sp1,al1),l1),p2) ->  
       (evaluable_abst (DOPN(Abst sp1,al1)))
       & eqappr_x sigma (apprec sigma l1 (abst_value (DOPN(Abst sp1,al1))),p2)
     | (p1,(DOPN(Abst sp2,al2),l2))  -> 
       (evaluable_abst (DOPN(Abst sp2,al2)))
       & eqappr_x sigma (p1,apprec sigma l2 (abst_value (DOPN(Abst sp2,al2))))

     | ((DOPN(MutInd indsp1,cl1) as mind1,l'1),
        (DOPN(MutInd indsp2,cl2) as mind2,l'2)) ->
       ((indsp1 = indsp2 & for_all2eq_vect (conv_x sigma) cl1 cl2)
       (* or conv_minductive (conv_x sigma) mind1 mind2 *))
       & for_all2eq (conv_x sigma) l'1 l'2
         
     | ((DOPN(MutConstruct i1,cl1) as mconstr1,l1),
        (DOPN(MutConstruct i2,cl2) as mconstr2,l2)) ->
        (((i1=i2) & for_all2eq_vect (conv_x sigma) cl1 cl2)
        (* or conv_mconstructor (conv_x sigma) mconstr1 mconstr2 *))
        & for_all2eq (conv_x sigma) l1 l2

     | ((DOPN(MutCase,cl1),l'1),(DOPN(MutCase,cl2),l'2)) -> 
       (for_all2eq_vect (conv_x sigma) cl1 cl2) &
       (for_all2eq (conv_x sigma) l'1 l'2)

     | ((DOPN(Fix i1,cl1),l'1),(DOPN(Fix i2,cl2),l'2)) -> 
       i1 = i2 &
       (for_all2eq_vect (conv_x sigma) cl1 cl2) &
       (for_all2eq (conv_x sigma) l'1 l'2)

    | ((DOPN(CoFix i1,cl1),l'1),(DOPN(CoFix i2,cl2),l'2)) -> 
       i1 = i2 &
       (for_all2eq_vect (conv_x sigma) cl1 cl2) &
       (for_all2eq (conv_x sigma) l'1 l'2)

     | (DOP0(Implicit),[]),(DOP0(Implicit),[]) -> true
     | (DLAM(_,c1),[]),(DLAM(_,c2),[]) -> conv_x sigma c1 c2
     | (DLAMV(_,vc1),[]),(DLAMV(_,vc2),[]) -> for_all2eq_vect (conv_x sigma) vc1 vc2
     | _ -> false;;


(********************************************************************)
(*             Special-Purpose Reduction                            *)
(********************************************************************)

let whd_meta mvmap = function
    (DOP0(Meta p) as u) -> (try assoc p mvmap with Not_found -> u)
  | x -> x
;;

let instance s c = if s=[] then c else ((strong whd_betaiota) o irec) c
where rec irec = function
   ((DOP0(Meta p)) as u)     -> (try assoc p s with Not_found -> u)
  | DOP2(Lambda,c1,DLAM(s,c2))    -> DOP2(Lambda,irec c1,DLAM(s,irec c2))
  | DOP2(Prod,c1,DLAM(s,c2))      -> DOP2(Prod,irec c1,DLAM(s,irec c2))
  | DOPN(oper,cl)        -> DOPN(oper,map_vect irec cl)
  | DOPL(oper,cl)        -> DOPL(oper,map irec cl)
  | DOP2(Cast,c,t)        -> DOP2(Cast,irec c,irec t)
  | DOP1(oper,c)     -> DOP1(oper,irec c)
  | DOP2(oper,c1,c2) -> DOP2(oper,irec c1,irec c2)
  | x                  -> x;;


let contract_cofix_use_function F cofix =
  let (DOPN(CoFix(bodynum),bodyvect)) = cofix in
  let nbodies =  (vect_length bodyvect) -1 in
  let make_Fi j = DOPN(CoFix(j),bodyvect) in
  let lbodies = list_assign (tabulate_list make_Fi nbodies) bodynum F 
  in  SAPPViList bodynum (last_vect bodyvect) lbodies
;;


let reduce_mind_case_use_function F mia =
  match mia.mconstr with 
     (DOPN(MutConstruct((indsp,tyindx),i),_)) ->
       let I = DOPN(MutInd(indsp,tyindx),args_of_mconstr mia.mconstr) in
       let nparams    = mind_nparams I in
       let real_cargs = snd(chop_list nparams mia.mcargs)
       in applist (mia.mlf.(i-1),real_cargs)
   | (DOPN(CoFix _,_)) as cofix ->
       let cofix_def = contract_cofix_use_function F cofix
       in  DOPN(MutCase,
                 concat_vect [|mia.mP;applist(cofix_def,mia.mcargs)|] mia.mlf)
   | _ -> anomaly "reduce_mind_case"
;;

let special_red_case sigma whfun cl  =
    let P = cl.(0)
    and c = cl.(1)
    and lf = sub_vect cl 2 ((vect_length cl) - 2) in redrec c []
    where rec redrec c l = 
    let constr,cargs = whfun c l in 
    match constr with 
           (DOPN(Const _,_)) as G -> 
              if (evaluable_const sigma G)
              then let Gvalue = (const_value sigma G) 
                   in if reducible_mind_case  Gvalue
                      then reduce_mind_case_use_function G
                          {mP=P; mconstr=Gvalue; mcargs=cargs; mlf=lf}
                      else redrec Gvalue cargs
              else raise Redelimination
           | _ -> if reducible_mind_case constr 
                  then reduce_mind_case
                          {mP=P; mconstr=constr; mcargs=cargs; mlf=lf}
                   else raise Redelimination
;;

(* F is convertible to DOPN(Fix(recindices,bodynum),bodyvect) make 
the reduction using this extra information *)

let contract_fix_use_function F fix =
let (DOPN(Fix(recindices,bodynum),bodyvect)) = fix in
let nbodies = vect_length recindices in
let make_Fi j = DOPN(Fix(recindices,j),bodyvect) in
let lbodies = list_assign (tabulate_list make_Fi nbodies) bodynum F in
    SAPPViList bodynum (last_vect bodyvect) lbodies
;;

let reduce_fix_use_function F whfun fix stack =
try
let (DOPN(Fix(recindices,bodynum),bodyvect)) = fix in
    (match fix_recarg fix with
     NONE -> (false,(fix,stack))
   | SOME recargnum ->
     if recargnum < 0 then (false,(fix,stack)) else
     let recarg = item stack recargnum in
     let (recarg'hd,_ as recarg')= whfun recarg [] in
     let stack' = list_assign stack recargnum (applist recarg') in
         (match recarg'hd with
          DOPN(MutConstruct _,_) -> (true,(contract_fix_use_function F fix,stack'))
        | _ -> (false,(fix,stack'))))
with Failure _ -> (false,(fix,stack))
;;


(* Check that c is an "elimination constant"
    [xn:An]..[x1:A1](<P>MutCase (Rel i) of f1..fk end g1 ..gp)
or  [xn:An]..[x1:A1](Fix(f|t) (Rel i1) ..(Rel ip)) 
    with i1..ip distinct variables not occuring in t 
keep relevenant information ([i1,Ai1;..;ip,Aip],n,b)
with b = true in case of a fixpoint in order to compute 
an equivalent of Fix(f|t)[xi<-ai] as 
[yip:Bip]..[yi1:Bi1](F bn..b1) 
    == [yip:Bip]..[yi1:Bi1](Fix(f|t)[xi<-ai] (Rel 1)..(Rel p))
with bj=aj if j<>ik and bj=(Rel c) and Bic=Aic[xn..xic-1 <- an..aic-1]
   *)

let compute_is_elim c = srec 0 [] c
where rec srec n labs c =  match whd_betadeltaeta_stack mt_evc c [] with 
   (DOP2(Lambda,T,DLAM(_,G)),[])  -> srec (n+1) (T::labs) G
 | (DOPN(Fix((nv,i)),bodies),l)   -> 
        if (length l) > n then raise Elimconst 
        else let li = 
        map (function Rel k -> if for_all_vect (noccurn k) bodies 
                               then (k,nth labs k) else raise Elimconst
                    | _ -> raise Elimconst) l
                            in if (distinct (map fst li))  then (li,n,true) 
                            else raise Elimconst
 | (DOPN(MutCase,cl),lapp) -> (match cl.(1) with Rel _ -> ([],n,false)
                                                | _ -> raise Elimconst)
 | _            -> raise Elimconst;;

let compute_consteval B =
    try SOME(compute_is_elim B) with Elimconst -> NONE
;;

let is_elim sigma (DOPN(Const sp,cl) as c) =
if (evaluable_const sigma c) & (defined_const sigma c) & not(is_existential c) then
    let (_,cb) = const_of_path sp
    in ((match cb.CONSTEVAL with
         SOME _ -> ()
       | NONE -> (match cb.CONSTBODY with
                  SOME{v=COOKED b} ->
                  cb.CONSTEVAL <- SOME(compute_consteval b)
                | SOME{v=RECIPE _} ->
                  anomalylabstrm "reduction__is_elim"
                  [< 'S"Found an uncooked transparent constant," ; 'SPC ;
                     'S"which is supposed to be impossible" >]));
        (match (cb.CONSTEVAL) with
         SOME (SOME x) -> x
       | SOME NONE -> raise Elimconst))
else raise Elimconst
;;

(* takes the fn first elements of the list and gives them back lifted by ln 
   and in reverse order *)

let rev_firstn_liftn fn ln = rfprec fn [] where rec rfprec p res l = 
   if p = 0 then res 
 else match l with [] -> raise (Invalid_argument "reduction__rev_firstn_liftn")
            | a::rest -> rfprec (p-1) ((lift ln a)::res) rest;;

let make_elim_fun sigma F largs =
  try let (lv,n,b) = is_elim sigma F and ll = length largs
      in if ll < n then raise Redelimination else
      if b then
      let labs,_ = chop_list n largs in
      let p = length lv in
        let la' = map_i (fun q aq ->
                    try (Rel (p+1-(index (n+1-q) (map fst lv)))) 
                    with Failure _ -> aq) 1
                  (map (lift p) labs) in
        it_list_i (fun i c (k,A) -> 
                   DOP2(Lambda,(substl (rev_firstn_liftn (n-k) (-i) la') A),
                   DLAM(Name(id_of_string"x"),c))) 0 (applistc F la') lv
      else F
  with Elimconst | Failure _ -> raise Redelimination;;

let rec red_elim_const sigma K largs =
if not(evaluable_const sigma K) then raise Redelimination else
    let F = make_elim_fun sigma K largs in
    match whd_betadeltaeta_stack sigma (const_value sigma K) largs with
  (DOPN(MutCase,cl),lrest) -> (special_red_case sigma (construct_const sigma) cl), lrest

| (DOPN(Fix _,_) as fix,lrest) -> 
     let (b,(c,rest)) = reduce_fix_use_function F (construct_const sigma) fix lrest
     in if b then (strong whd_beta c,rest) else raise Redelimination

 | _ -> anomaly "red_elim_const"

and construct_const sigma c stack = hnfstack stack c
    where rec hnfstack stack = function
      (DOPN(Const _,_)) as K  -> (try let (c',lrest) = red_elim_const sigma K stack
                            in hnfstack lrest c'
                        with Redelimination ->
                                  if evaluable_const sigma K 
                                  then let C = const_value sigma K
                                       in  (match C with
                                               DOPN (CoFix _,_) -> K,stack
                                             | _                -> hnfstack stack C) 
                                  else raise Redelimination)

    | (DOPN(Abst _,_) as A) -> if evaluable_abst A then hnfstack stack (abst_value A)
                               else raise Redelimination

    | DOP2(Cast,c,_)       -> hnfstack stack c
    | DOPN(AppL,cl)    -> hnfstack (app_tl_vect cl stack) (hd_vect cl)
    | DOP2(Lambda,_,DLAM(_,c)) -> (match stack with 
                           [] -> anomaly "Cannot be a constructor"
                        |  c'::rest -> stacklam hnfstack [c'] rest c)
    | DOPN(MutCase,cl) -> 
               hnfstack stack  (special_red_case sigma (construct_const sigma) cl)    
    | DOPN(MutConstruct _,_) as C -> C,stack
    | DOPN(CoFix_,_) as C -> C,stack
    | DOPN(Fix (_) ,_) as fix -> 
             let (reduced,(fix,stack')) = reduce_fix 
             (fun c st -> hnfstack st c) fix stack
             in if reduced then hnfstack stack' fix else raise Redelimination
    | _             -> raise Redelimination;;

(* linear substitution (following pretty-printer) of the value of name in c *)
(* at the occurrences of occ_list *)
(* occ_list must be non empty *)

let substlin sigma occ_list name c =
  let rec substrec n ol = function
   (DOPN(Const _,_) as co)
                   -> if name=(path_of_const co) then
                           (if (hd ol)=n then
     if evaluable_const sigma co then 
                    let cs = (const_value sigma co) in ((n+1),(tl ol),cs)
     else errorlabstrm"substlin" 
          [< pTERM co; 'S " is not a defined constant" >]
                              else ((n+1),ol,co))
     
                        else (n,ol,co)

 | (DOPN(Abst _,_) as co)
                   -> if name=(path_of_abst co) then
                           (if (hd ol)=n then
                                 let cs = (abst_value co) in ((n+1),(tl ol),cs)
                              else ((n+1),ol,co))
                        else (n,ol,co)

(* INEFFICIENT: OPTIMIZE *)
 | DOPN(AppL,tl) -> let c1 = hd_vect tl and cl = tl_vect tl in
   it_vect (fun (n1,ol1,c1') c2 ->
                (match ol1 with 
                 [] -> (n1,[],applist(c1',[c2]))
               | _  -> let (n2,ol2,c2') = substrec n1 ol1 c2
                       in (n2,ol2,applist(c1',[c2']))))
           (substrec n ol c1) cl

 | DOP2(Lambda,c1,DLAM(na,c2)) -> let (n1,ol1,c1') = substrec n ol c1
                        in (match ol1 with 
                               [] -> (n1,[],DOP2(Lambda,c1',DLAM(na,c2)))
                             | _  -> let (n2,ol2,c2') = substrec n1 ol1 c2
                                       in (n2,ol2,DOP2(Lambda,c1',DLAM(na,c2'))))
 | DOP2(Prod,c1,DLAM(na,c2))   -> let (n1,ol1,c1') = substrec n ol c1
                        in (match ol1 with 
                               [] -> (n1,[],DOP2(Prod,c1',DLAM(na,c2)))
                             | _  -> let (n2,ol2,c2') = substrec n1 ol1 c2
                                       in (n2,ol2,DOP2(Prod,c1',DLAM(na,c2'))))
 | DOPN(MutCase,cl)   ->
   (match list_of_vect cl with
    P::d::llf ->
    let rec substlist nn oll =
        function
        [] -> (nn,oll,[])
      | f::lfe -> (let (nn1,oll1,f') = substrec nn oll f
                   in (match oll1 with
                       [] -> (nn1,[],f'::lfe)
                     | _  -> let (nn2,oll2,lfe') = substlist nn1 oll1 lfe 
                             in (nn2,oll2,f'::lfe')))
    in (match (substlist n ol (P::d::llf)) with
        (n2,ol2,P'::d'::lf') -> (n2,ol2,DOPN(MutCase,vect_of_list(P'::d'::lf')))
      | _ -> anomaly "substlist")
  | _ -> invalid_arg "substlist")

 | DOP2(Cast,c1,c2)   -> let (n1,ol1,c1') = substrec n ol c1
                        in (match ol1 with 
                               [] -> (n1,[],DOP2(Cast,c1',c2))
                             | _  -> let (n2,ol2,c2') = substrec n1 ol1 c2
                                       in (n2,ol2,DOP2(Cast,c1',c2')))
 | (DOPN(Fix _,_) as co) -> 
           (warning "do not consider occurrences inside fixpoints"; (n,ol,co))

 | (DOPN(CoFix _,_) as co) -> 
           (warning "do not consider occurrences inside cofixpoints"; (n,ol,co))

 | co               -> (n,ol,co)
  in
 let (nn,oll,co) = (substrec 1 (sort (fun (x,y)-> (x<y)) occ_list) c)
   in match oll with
       [] -> strong whd_betaiota co
     | l  -> if nn=1 then error ((string_of_path name)^" does not occur")
                     else error ("bad occurrence numbers of "^(string_of_path name));;

(* unfoldn : (int list -> string -> constr -> constr) *)
(* unfolds a constant in a term following a list of occurrences *)
let unfoldn sigma = function
   []  -> error "Empty list of occurrences"
 | l   -> substlin sigma l;;


(* Same as simplify, but also reduces elimination constants *)
let whd_nf_stack sigma stack c = nf_app stack c
where rec nf_app stack c =
    match c with
    DOP2(Lambda,c1,DLAM(name,c2))    ->
    (match stack with
     [] -> (c,[])
   | a1::rest -> stacklam nf_app [a1] rest c2)

  | DOPN(AppL,cl)      -> nf_app (app_tl_vect cl stack) (hd_vect cl)
  | DOP2(Cast,c,_)           -> nf_app stack c
  | (DOPN(Const _,_) as x)     ->
    (try let (c',lrest) = red_elim_const sigma x stack
         in nf_app lrest c'
     with Redelimination -> x,stack)
  | (DOPN(MutCase,cl)) -> (try nf_app stack 
                                  (special_red_case sigma (fun c st -> nf_app st c) cl)
                           with Redelimination -> c,stack)
  | (DOPN(Fix _,_) as fix) ->
    let (reduced,(fix,rest)) = reduce_fix (fun c st -> nf_app st c) fix stack
    in if reduced then nf_app rest fix
       else (fix,stack)

  | x                   -> x,stack;;

let whd_nf sigma c = applist(whd_nf_stack sigma [] c);;

let nf sigma c = strong (whd_nf sigma) c;;

let one_step_reduce sigma = redrec []
where rec redrec largs = function
   DOP2(Lambda,t,DLAM(n,c))  -> (match largs with
                      []      -> error "Not reducible 1"
                    | a::rest -> applistc (subst1 a c) rest)
 | DOPN(AppL,cl)     -> redrec (app_tl_vect cl largs) (hd_vect cl)
 | (DOPN(Const _,_)) as k   -> (try let (c',l) = red_elim_const sigma k largs
                          in applistc c' l
                      with Redelimination ->
                                    if evaluable_const sigma k then
                                        applistc (const_value sigma k) largs
                                    else error "Not reductible 1")
 | DOPN(Abst _,_) as A -> if evaluable_abst A then applistc (abst_value A) largs
                          else error "Not reducible 0"

 | DOPN(MutCase,cl)  -> (try  applistc (special_red_case sigma (whd_betadeltaiota_stack sigma) cl) largs 
                         with Redelimination -> error "Not reducible 2")
 | (DOPN(Fix _,_) as fix) ->
    let (reduced,(fix,stack)) = reduce_fix (whd_betadeltaiota_stack sigma) fix largs
    in if reduced then applistc fix stack else error "Not reducible 3"

 | DOP2(Cast,c,_)   -> redrec largs c
 | _              -> error "Not reducible 3";;

let reduce sigma = redrec [] 
where rec redrec largs x = match x with
        DOP2(Lambda,t,DLAM(n,c)) -> (match largs with
                                []      -> applist(x,largs)
                              | a::rest -> stacklam redrec [a] rest c)
      | DOPN(AppL,cl)    -> redrec (app_tl_vect cl largs) (hd_vect cl)
      | (DOPN(Const _,_)) as k      -> (try let c',lrest = red_elim_const sigma k largs
                              in redrec lrest c'
                          with Redelimination ->
                                        if evaluable_const sigma k 
                                        then let C = (const_value sigma k)
                                             in match C with 
                                                   DOPN(CoFix _,_) -> k
                                                 | _ ->  redrec largs C
                                        else applist(k,largs))
      | DOPN(Abst _,_) as A -> if evaluable_abst A then redrec largs (abst_value A)
                               else applist(A,largs)

      | DOP2(Cast,c,_)       -> redrec largs c
      | DOPN(MutCase,cl)  -> (try redrec largs 
                                (special_red_case sigma (whd_betadeltaiota_stack sigma) cl)
                         with Redelimination -> applist(x,largs))
     | (DOPN(Fix _,_)) ->
       let (reduced,(fix,stack)) = reduce_fix (whd_betadeltaiota_stack sigma) x largs
       in if reduced then redrec stack fix else applist(x,largs)

      | _         -> applist(x,largs);;

let hnf_constr sigma c = reduce sigma c;;

let is_info_sort s = 
let (DOP0(Sort(s))) = whd_betadeltaiota mt_evc s in 
          match s with Prop(Pos) | Type(Pos,_) -> true
                      | _ -> false;;


(* $Id: reduction.ml,v 1.8 1996/02/12 15:48:19 cpaulin Exp $ *)
