(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                refiner.ml                                *)
(****************************************************************************)

#infix "o";;
#open "std";;
#open "initial";;
#open "names";;
#open "vectops";;
#open "generic";;
#open "term";;
#open "reduction";;
#open "himsg";;
#open "termenv";;
#open "mach";;
#open "more_util";;
#open "proof";;
#open "proof_trees";;
#open "proof_constraints";;
#open "pp";;
#open "stdpp";;
#open "logic";;


(* hypotheses : 'a goalty -> constr signature *)
let hypotheses gl = gl.hyps;;
(* conclusion : 'a goalty -> constr *)
let conclusion gl = gl.concl;;

(* sig_it : 'a sigma -> 'a*)
let sig_it x = x.it;;
(* sig_sig : 'a sigma -> global_constraints *)
let sig_sig x = x.sigma;;

(*let Project gls =
    abs_rc(ts_mk {focus = gls.it.lc;
                  sign = gls.it.hyps;
                  decls = ts_it (repr_gc gls.sigma)});;
*)


(* Projects takes a goal sigma { it=(gls:proof__goal) ; sigma = (gc:proof__global_constraints) }
and gives readable_constraints, where
focus = local constraints of gls
sign = signature of gls
decls = existential var declarations of gc
*)
let Project gls =
    proof__rc_of_gc (gls.sigma) (gls.it)
;;

let pf_status pf = pf.status;;

let is_complete pf = (COMPLETE_PROOF = (pf_status pf));;

let on_open_proofs f pf = if is_complete pf then pf else f pf;;

let rec and_status = function
    [] -> COMPLETE_PROOF
  | COMPLETE_PROOF::l -> and_status l
  | _ -> INCOMPLETE_PROOF
;;


(* mapshape [ l1 ; ... ; lk ] [ v1 ; ... ; vk ] [ p_1 ; .... ; p_(l1+...+lk) ]
gives
[ (v1 [p_1 ... p_l1]) ; (v2 [ p_(l1+1) ... p_(l1+l2) ]) ; ... ;
(vk [ p_(l1+...+l(k-1)+1) ... p_(l1+...lk) ])
]
*)
let rec mapshape nl (fl:((proof list) -> proof) list) (l:proof list) =
    match nl with
    [] -> []
  | h::t ->
    let m,l = chop_list h l
    in (hd fl m)::(mapshape t (tl fl) l)
;;

(* given a proof p, frontier p gives (l,v) where l is the list of goals
to be solved to complete the proof, and v is the corresponding validation *)
let rec frontier p =
    match p.ref with
    NONE -> ([p.goal],
                fun [p'] ->
                    if p'.goal = p.goal then p'
                    else errorlabstrm "refiner__frontier"
                         [< 'S "frontier was handed back a ill-formed proof." >])
  | SOME(r,pfl) ->
    let gll,vl = split(map frontier pfl)
    in (flat gll,
        fun retpfl ->
            let pfl' = mapshape (map length gll) vl retpfl
            in {status = and_status (map pf_status pfl');
                subproof=p.subproof;
                goal = p.goal;
                ref = SOME(r,pfl')})
;;

(* list_pf p is the lists of goals to be solved in order to complete the
proof p *)
let list_pf p = fst(frontier p);;

(* leaf g is the canonical incomplete proof of a goal g *)
let leaf g = {status=INCOMPLETE_PROOF;
              subproof=NONE;
              goal=g;
              ref=NONE};;

(* the identity tactic *)
let IDTAC gls = {it=[gls.it];sigma=gls.sigma},
    (fun [pf] -> pf | _ -> errorlabstrm "refiner__IDTAC"
         [< 'S "IDTAC validation is applicable only to a one-proof list">] );;

(* unTAC : tactic -> goal sigma -> proof sigma *)
let unTAC tac g =
let (gl_sigma,v) = tac g
in {it=v(map leaf gl_sigma.it); sigma=gl_sigma.sigma};;

let unpackage glsig = (ref (glsig.sigma)),glsig.it;;
let repackage r v = {it=v;sigma = !r};;
let apply_sig_tac r tac g =
    let glsigma,v = tac (repackage r g)
    in (r := glsigma.sigma;
        (glsigma.it,v))
;;

(* THEN tac1 tac2 gls applies the tactic tac1 to gls and applies tac2 to
every resulting subgoals *)
let THEN (tac1:tactic) (tac2:tactic) (gls:goal sigma) =
    let (sigr,g) = unpackage gls in
    let gl,p = apply_sig_tac sigr tac1 g in
    let gll,pl = split(map (apply_sig_tac sigr tac2) gl) in
    (repackage sigr (flat gll)) ,  (p o mapshape(map length gll)pl)
;;
#infix "THEN";;

(* THEN_i tac1 tac2 n gls applies the tactic tac1 to gls and applies
tac2 (i+n-1) to the i_th resulting subgoal *)
let THEN_i (tac1:tactic) (tac2:(int -> tactic)) n (gls:goal sigma) =
    let (sigr,g) = unpackage gls in
    let gl,p = apply_sig_tac sigr tac1 g in
    let gll,pl = split(map_i (fun n -> apply_sig_tac sigr (tac2 n)) n gl) in
    (repackage sigr (flat gll)) , (p o mapshape(map length gll)pl)
;;

(* THENS tac1 [t1 ; ... ; tn] gls applies the tactic tac1 to gls and applies
t1,..., tn to the n resulting subgoals. Raises an error if the number of
resulting subgoals is not n *)
let THENS (tac1:tactic) (tac2l:tactic list) (gls:goal sigma) =
   let (sigr,g) = unpackage gls in
   let gl,p = apply_sig_tac sigr tac1 g  in
   let tac2gl = try combine(tac2l,gl)
                with Invalid_argument _ -> errorlabstrm "refiner__THENS"
                                          [<'S "Wrong number of tactics.">]
    in
   let gll,pl = split(map (fun (tac2,g) -> apply_sig_tac sigr tac2 g) tac2gl)
   in
       (repackage sigr (flat gll)) , (p o mapshape(map length gll)pl)
;;
#infix "THENS";;

(* THENS tac1 [t1 ; ... ; tn] gls applies the tactic tac1 to gls and tac2
to the last resulting subgoal *)
let THENL (tac1:tactic) (tac2:tactic) (gls:goal sigma) =
let (sigr,g) = unpackage gls in
let gl,p = apply_sig_tac sigr tac1 g
in if length gl > 0 then
    let g,rest = sep_last gl in
    let tac2l = (map (fun _ -> IDTAC) rest)@[tac2] in
    let tac2gl = try combine(tac2l,gl)
                 with Invalid_argument _ -> errorlabstrm "refiner__THENL"
                                       [<'S "Wrong number of tactics.">]
     in
    let gll,pl = split(map (fun (tac2,g) -> apply_sig_tac sigr tac2 g) tac2gl)
    in
        (repackage sigr (flat gll)) , (p o mapshape(map length gll)pl)
   else errorlabstrm "refiner__THENL" [< 'S"No resulting subgoal.">]
;;
#infix "THENL";;

(* PROGRESS tac ptree applies tac to the goal ptree and fails if tac leaves
the goal unchanged *)
let PROGRESS (tac:tactic) (ptree : goal sigma) =
  let rslt = tac ptree
  in if (length (fst rslt).it)=1 then
      let subgoal=(hd (fst rslt).it)
      in if (hypotheses subgoal) = (hypotheses ptree.it) &
            (conclusion subgoal) = (conclusion ptree.it) &
            (subgoal.ctxt = ptree.it.ctxt) &
            (subgoal.lc = ptree.it.lc) &
            ts_eq (repr_gc ptree.sigma) (repr_gc (fst rslt).sigma) then
            errorlabstrm "refiner__PROGRESS" [< 'S"Failed to progress.">]
         else rslt
     else rslt
;;

(* ORELSE f1 f2 tries to apply f1 and if it fails, applies f2 *)
let ORELSE (f1:tactic) (f2:tactic) (g:goal sigma) =
    try (PROGRESS f1) g
    with UserError _ -> f2 g ;;
#infix "ORELSE";;

(* TRY f tries to apply f, and if it fails, leave the goal unchanged *)
let TRY (f:tactic) = f ORELSE IDTAC;;

let THENTRY (f:tactic) (g:tactic) = f THEN (TRY g);;
#infix "THENTRY";;

let COMPLETE (tac:tactic) =
    fun goal ->
        let achievement = tac goal
        in if null (fst achievement).it then
            achievement
           else errorlabstrm "refiner__COMPLETE"
                             [< 'S "Proof is not complete" >]
;;

(* Beware: call by need of CAML, g is needed *)

(*
let rec REPEAT (f:tactic) (g:goal) =
  (((PROGRESS f) THEN REPEAT f) ORELSE IDTAC ) g ;;
*)
let rec REPEAT = fun T g ->
     ((T THEN (REPEAT T)) ORELSE IDTAC) g;;

(* Try the first tactic that does not fail in a list of tactics *)

let rec FIRST = fun tacl g ->
 match tacl with
    []      -> errorlabstrm "refiner__FIRST" [< 'S"No applicable tactic.">]
 |  T::rest -> (try T g with UserError _ -> FIRST rest g);;

let TRY T = T ORELSE IDTAC;;

let AT_LEAST_ONE T = T THEN REPEAT T;;

let (FAILTAC:tactic) = fun _ -> errorlabstrm "refiner__FAILTAC"
                                          [< 'S"Failtac always fails.">]
;;

(* Iteration tactical *)
let DO n t = dorec n 
   where rec dorec k = 
          if k < 1 then errorlabstrm "refiner__DO"
                        [<'S"Wrong argument : Do needs a positive integer.">]
          else if k = 1 then t else t THEN (dorec (k-1));;


(* Tactic table *)
let tac_tab =
    (mhm__new 17 : (string,
                        (tactic_arg list -> tactic)
                        * (readable_constraints -> goal ->
                              tactic_expression -> std_ppcmds)) mhm__t);;
let add_tactic s t =
if mhm__in_dom tac_tab s
  then
    errorlabstrm "refiner__add_tactic" [<'S "Cannot redeclare a tactic.">]
  else
    (mhm__add tac_tab (s,t))
;;

let overwriting_add_tactic s t =
    if mhm__in_dom tac_tab s then
        (mhm__rmv tac_tab s;
         warning ("Overwriting definition of tactic "^s));
    mhm__add tac_tab (s,t)
;;

let lookup_tactic s =
    try fst(mhm__map tac_tab s)
    with Not_found -> errorlabstrm "refiner__lookup_tactic"
                      [< 'S"The tactic " ; 'S s ; 'S" is not installed" >]
;;

let lookup_tacpr s =
    try snd(mhm__map tac_tab s)
    with Not_found -> errorlabstrm "refiner__lookup_tacpr"
                      [< 'S"The tactic " ; 'S s ; 'S" is not installed" >]
;;

let pr_tactic sigma gl ((s,_) as texp) =
    lookup_tacpr s sigma gl texp;;

(* refiner r is a tactic applying the rule r *)
let refiner r goal_sigma =
    match r with
    (PRIM pr) ->
    let sgl = 
        (try prim_refiner pr (Project goal_sigma) goal_sigma.it
         with UserError _ as e -> raise e
            | e -> (MSG(errors__explain_user_exn e);
                    errors__reraise_user_exn e))

    in ({it=sgl; sigma = goal_sigma.sigma},
        (fun pfl ->
             if for_all2eq (fun g pf -> g = pf.goal) sgl pfl then
                 {status = and_status (map pf_status pfl);
                  goal = goal_sigma.it;
                  subproof = NONE;
                  ref = SOME(r,pfl)}
             else errorlabstrm "refiner__refiner"
              [< 'S"Bad subproof in validation.">]))


  | ((TACTIC(s,targs)) as r) ->
    let goal = goal_sigma.it in
    let tacfun = lookup_tactic s in
    let sgl_sigma,v = tacfun targs goal_sigma in
    let hidden_proof = v (map leaf sgl_sigma.it)
    in (sgl_sigma,
        fun spfl ->
            if for_all2eq (fun g pf -> g = pf.goal) sgl_sigma.it spfl then
                {status=and_status (map pf_status spfl);
                 goal=goal_sigma.it;
                 subproof=SOME hidden_proof;
                 ref=SOME(r,spfl)}
            else errorlabstrm "refiner__refiner"
                  [< 'S"Bad subproof in validation.">])

   | ((CONTEXT ctxt) as r) ->
     let gl = goal_sigma.it in
     let sg = mkGOAL gl.lc ctxt gl.hyps gl.concl
     in ({it=[sg];sigma=goal_sigma.sigma},
         fun [pf] ->
             if pf.goal = sg then
                 {status=pf.status;
                  goal=gl;
                  subproof=NONE;
                  ref=SOME(r,[pf])}
             else errorlabstrm "refiner__refiner"
                   [< 'S "Bad subproof in validation.">])

   | ((LOCAL_CONSTRAINTS lc) as r) ->
     let gl = goal_sigma.it in
     let sg = mkGOAL lc gl.ctxt gl.hyps gl.concl
     in ({it=[sg];sigma=goal_sigma.sigma},
         fun [pf] ->
             if pf.goal = sg then
                 {status=pf.status;
                  goal=gl;
                  subproof=NONE;
                  ref=SOME(r,[pf])}
             else errorlabstrm "refiner__refiner"
                   [<'S"Bad subproof in validation.">])
;;

(* rc_of_pfsigma : proof sigma -> readable_constraints *)
let rc_of_pfsigma sigma = rc_of_gc sigma.sigma sigma.it.goal;;
(* rc_of_glsigma : proof sigma -> readable_constraints *)
let rc_of_glsigma sigma = rc_of_gc sigma.sigma sigma.it;;

type walking_constraints == readable_constraints idstamped;;

(* Local_Constraints lc is a tactic that makes the local constraints be lc *)
let Local_Constraints lc pf = refiner (LOCAL_CONSTRAINTS lc) pf;;

(* extract_open_proof : constr signature -> proof
-> constr * (int * constr) list
takes a constr signature corresponding to global definitions
and a (not necessarly complete) proof
and gives a pair (pfterm,obl)
where pfterm is the constr corresponding to the proof
and obl is an int*constr list [ (m1,c1) ; ... ; (mn,cn)]
where the mi are metavariables numbers, and ci are their types.
Their proof should be completed in order to complete the initial proof
*)

let extract_open_proof sign pf =

let open_obligations = ref [] in

let rec proof_extractor vl = function

    {ref=SOME(PRIM _,_);_} as pf ->
    prim_extractor proof_extractor vl pf

   | {ref=SOME(TACTIC _,spfl);
      subproof=SOME hidden_proof;_} ->
     let sgl,v = frontier hidden_proof in
     let flat_proof = v spfl
     in proof_extractor vl flat_proof

   | {ref=SOME(CONTEXT ctxt,[pf]);_} -> (proof_extractor vl) pf

   | {ref=SOME(LOCAL_CONSTRAINTS lc,[pf]);_} -> (proof_extractor vl) pf

   | {ref=NONE;goal=goal;_} ->
     let rel_env = get_rels vl in
     let n_rels = length rel_env in
     let visible_rels =
         map_succeed
         (fun id ->
              match lookup_id id vl with
              GLOBNAME _ -> failwith "caught"
            | RELNAME(n,_) -> (n,id))
         (ids_of_sign goal.hyps) in
     let sorted_rels = sort (fun ((n1,_),(n2,_)) -> gt(n1,n2)) visible_rels in

     let abs_concl =
         list_it
         (fun (_,id) concl -> produit id (snd(lookup_sign id goal.hyps)) concl)
         sorted_rels
         goal.concl in

     let mv = newMETA()
     in open_obligations := (mv,abs_concl):: !open_obligations;
        applist(DOP0(Meta mv),map (fun (n,_) -> Rel n) sorted_rels) in

let pfterm = proof_extractor (GLOB sign) pf in

    (pfterm,rev !open_obligations)
;;

(* extracts a constr from a proof, and raises an error if the proof is
incomplete *)
let extract_proof sign pf =
    match extract_open_proof sign pf with
    T,[] -> T
  | _ ->
    errorlabstrm "extract_proof"
    [< 'S "Attempt to save an incomplete proof" >]
;;

(* whd_evar goes through existential variables when they are defined *)
let rec whd_evar sigma = function
    DOPN(Const sp,_) as K ->
    if evc__is_evar sigma sp & evaluable_const sigma K then
        whd_evar sigma (const_value sigma K)
    else K
;;

(* expanses every existential constant and compute the normal form through
beta iota reduction *)
let rec expand_evars sigma c =
    strong whd_betaiota (strong (whd_evar sigma) c)
;;  

let Context ctxt pf = refiner (CONTEXT ctxt) pf;;
let Tactic texp gl = refiner (TACTIC texp) gl;;

(* register_tactic s tac pr registers a tactic s under the name s *)
let register_tactic s tac pr =
    (add_tactic s (tac,pr);
     (fun args -> Tactic(s,args)))
;;

(* overwriting_register_tactic s tac pr registers a tactic s under the name
s even if a tactic of the same name is already registered *)
let overwriting_register_tactic s tac pr =
    (overwriting_add_tactic s (tac,pr);
     (fun args -> Tactic(s,args)))
;;

(* solve_subgoal n tac pf_sigma applies the tactic tac at the nth subgoal of
pf_sigma *)
let solve_subgoal n tac pf_sigma =
let (sigr,pf) = unpackage pf_sigma in
let gl,p = frontier pf in
if gl = [] then errorlabstrm "solve_subgoal" [< 'S"No more subgoals.">] 
else if length gl < n then errorlabstrm 
 "solve_subgoal" [< 'S"Trying to apply a tactic to a non existent goal" >]
else
let tac2 i = if i = n then tac else IDTAC in
let gll,pl = split(map_i (fun n -> apply_sig_tac sigr (tac2 n)) 1 gl)
in repackage sigr ((p o mapshape(map length gll)pl) (map leaf (flat gll)))
;;

(* The type of proofs and a few utilities *)
type pfstate = {pf : proof ; pfsigma : global_constraints};;
let proof_of_pfstate pfs = pfs.pf;;
let evc_of_pfstate pfs = pfs.pfsigma;;
let cursor_of_pfstate pfs = [];;

(* gives a new proof (a leaf) of a goal gl *)
let mk_pfstate gl = {pf = leaf gl ; pfsigma = abs_gc(ts_mk(lavm__new sp_ord))};;

(* solve the nth subgoal with tactic tac *)
let solve_pf n tac pfs =
let rslts = solve_subgoal n tac {it = pfs.pf;sigma = pfs.pfsigma}
in {pf = rslts.it;
    pfsigma = rslts.sigma}
;;

(* extracts a constr from a proof ; raises an error if the proof is
not complete *)
let extract_pfstate pfs =
    strong (whd_betadeltatiota (rc_of_gc pfs.pfsigma pfs.pf.goal))
           (extract_proof pfs.pf.goal.hyps pfs.pf)
;;

type pftreestate = {tpf: proof ;
                    tpfsigma : global_constraints;
                    tstack : (int * validation) list}
;;
let proof_of_pftreestate pts = pts.tpf;;
let cursor_of_pftreestate pts = map fst pts.tstack;;
let evc_of_pftreestate pts = pts.tpfsigma;;

let descend n p =
    match p.ref with
    NONE -> error "descend: leaf"
  | SOME(r,pfl) ->
    if list_length pfl >= n then
        let (left,wanted::right) = chop_list (n-1) pfl
        in [wanted],
            (fun pfl' ->
                 if list_length pfl' = 1 & (hd pfl').goal = wanted.goal then
                     let pf' = hd pfl' in
                     let spfl = left@(pf'::right) in
                     let newstatus = and_status (map pf_status spfl)
                     in {status = newstatus;
                         goal = p.goal;
                         subproof = p.subproof;
                         ref = SOME(r,spfl)}

                 else error "descend: validation")
    else error "descend: too few subproofs"
;;

let traverse = fun
    (* go to the parent *)
    0 pts ->
    (match  pts.tstack with
     [] -> error "traverse: no ancestors"
   | (_,v)::tl ->
     {tpf = v [pts.tpf];
      tstack = tl;
      tpfsigma = pts.tpfsigma})

    (* go to the hidden tactic-proof, if any, otherwise fail *)
  | (-1) pts ->
    (match pts.tpf.subproof with
     NONE -> error "traverse: not a tactic-node"
   | SOME spf ->
     let v = (fun pfl -> pts.tpf)
     in {tpf = spf;
         tstack = (-1,v)::pts.tstack;
         tpfsigma = pts.tpfsigma})

    (* when n>0, go to the nth child *)
  | n pts ->
    let [npf],v = descend n pts.tpf
    in {tpf = npf;
        tpfsigma = pts.tpfsigma;
        tstack = (n,v):: pts.tstack}
;;

let solve_pftreestate tac pts =
let pf = leaf pts.tpf.goal in
let {pf=npf;pfsigma=npfsigma} = solve_pf 1 tac {pf= pf;
                                                pfsigma=pts.tpfsigma}
in {tpf = npf;
    tpfsigma = npfsigma;
    tstack = pts.tstack}
;;

let solve_nth_pftreestate n tac pts =
let pf = pts.tpf in
let {pf=npf;pfsigma=npfsigma} = solve_pf n tac {pf= pf;
                                                pfsigma=pts.tpfsigma}
in {tpf = npf;
    tpfsigma = npfsigma;
    tstack = pts.tstack}
;;

let mk_pftreestate g =
let pfs = mk_pfstate g
in {tpf = pfs.pf;
    tstack = [];
    tpfsigma = pfs.pfsigma}
;;

let extract_pftreestate pts =
    if pts.tstack = [] then
        extract_pfstate {pf = pts.tpf; pfsigma = pts.tpfsigma}
    else errorlabstrm "extract_pftreestate"
        [< 'S"Cannot extract from a proof-tree in which we have descended;" ;
           'SPC ; 'S"Please ascend to the root" >]
;;

let rec first_unproven pts =
let pf = (proof_of_pftreestate pts)
in if is_complete_proof pf then
    errorlabstrm "first_unproven"
    [< 'S"No unproven subgoals" >]
   else if is_leaf_proof pf then
       pts
   else
       try let childnum =
       try_find_i 
       (fun n pf -> if not(is_complete_proof pf) then n else failwith "caught")
       1
       (children_of_proof pf)
           in first_unproven (traverse childnum pts)
       with e -> errorlabstrm "first_unproven"
                 [< errors__explain_sys_exn e >]
;;

let rec last_unproven pts =
let pf = (proof_of_pftreestate pts)
in if is_complete_proof pf then
    errorlabstrm "last_unproven"
    [< 'S"No unproven subgoals" >]
   else if is_leaf_proof pf then
       pts
   else
       try let children = (children_of_proof pf) in
           let nchilds = length children in
           let childnum =
               try_find_i 
               (fun n pf ->
                    if not(is_complete_proof pf) then n else failwith "caught")
               1
               (rev children)
           in last_unproven(traverse (nchilds-childnum+1) pts)
       with e -> errorlabstrm "last_unproven"
                 [< errors__explain_sys_exn e >]
;;

let rec node_prev_unproven loc pts =
let pf = proof_of_pftreestate pts
in match cursor_of_pftreestate pts with
    [] -> last_unproven pts
  | n::l ->
    if is_complete_proof pf or loc = 1 then
        node_prev_unproven n (traverse 0 pts)
   else if is_complete_proof (nth (children_of_proof pf) (loc - 1)) then
       node_prev_unproven (loc - 1) pts
   else first_unproven (traverse (loc - 1) pts)
;;


let rec node_next_unproven loc pts =
let pf = proof_of_pftreestate pts
in match cursor_of_pftreestate pts with
    [] -> first_unproven pts
  | n::l ->
    if is_complete_proof pf or loc = (length (children_of_proof pf)) then
        node_next_unproven n (traverse 0 pts)
   else if is_complete_proof (nth (children_of_proof pf) (loc + 1)) then
       node_next_unproven (loc + 1) pts
   else last_unproven(traverse (loc + 1) pts)
;;

let next_unproven pts =
let pf = proof_of_pftreestate pts
in if is_leaf_proof pf then
    match cursor_of_pftreestate pts with
    [] -> error "next_unproven"
  | n::_ -> node_next_unproven n (traverse 0 pts)
   else node_next_unproven (length (children_of_proof pf)) pts
;;

let prev_unproven pts =
let pf = proof_of_pftreestate pts
in if is_leaf_proof pf then
    match cursor_of_pftreestate pts with
    [] -> error "prev_unproven"
  | n::_ -> node_prev_unproven n (traverse 0 pts)
   else node_prev_unproven 1 pts
;;

let rec top_of_tree pts =
    match cursor_of_pftreestate pts with
    [] -> pts
  | n::_ -> top_of_tree(traverse 0 pts)
;;


(* $Id: refiner.ml,v 1.4 1996/02/12 15:48:23 cpaulin Exp $ *)
