(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                termenv.ml                                *)
(****************************************************************************)
#infix "o";;
#open "std";;
#open "pp";;
#open "stdpp";;
#open "names";;
#open "more_util";;
#open "vectops";;
#open "generic";;
#open "term";;
#open "constrtypes";;
#open "environ";;
#open "proof";;
#open "proof_constraints";;
#open "proof_trees";;

(* Variables *)

let initial_sign = vartab__initial_sign;;
let initial_fsign = vartab__initial_fsign;;
let initial_assumptions = vartab__initial_assumptions;;
let auto_save_variables = vartab__auto_save_variables;;


let global_operator sp id =
  match lib__leaf_object_tag (objsp_of sp) with
    "CONSTANT" ->
    let (cmap,_,_) = outConstant(lib__map_leaf(objsp_of sp)) in
    let cb = listmap__map cmap (kind_of_path sp)
    in (Const sp,cb.CONSTHYPS)

  | "MUTUALINDUCTIVE" ->
    let (mimap,_) = outMutualInductive (lib__map_leaf (objsp_of sp)) in
    let mib = listmap__map mimap (kind_of_path sp)
    in mind_oper_of_id sp mib id

  | _ -> raise Not_found
;;

(* id_of_global gives the name of the given sort oper *)
let id_of_global = function
    (Const sp) -> basename sp

  | (MutInd(sp,tyi)) -> if tyi=0 then basename sp else
    let (_,mib) = mind_of_path sp in
    let mip = mind_nth_type_packet mib tyi
    in mip.MINDTYPENAME

  | (MutConstruct((sp,tyi),i)) ->
    let (_,mib) = mind_of_path sp in
    let mip = mind_nth_type_packet mib tyi
    in if i <= vect_length mip.MINDCONSNAMES & i > 0 then
        mip.MINDCONSNAMES.(i-1)
       else failwith "id_of_global"
;;

let informative_construction ty =
    match contents_of_kind (cast_type ty) with
         Null -> false
       | _ -> true
;;

let informative_reference = function
    (DOPN(Const sp,_)) ->
    (match kind_of_path sp with
     FW -> true
   | CCI ->
     let (_,cb) = const_of_path sp
     in informative_construction cb.CONSTTYPE
   | OBJ ->
     anomalylabstrm "termenv__informative_reference"
     [< 'S"Found a constant with a ill-formed path-kind" >])

  | (DOPN(MutInd (sp,tyi),_)) ->
    (match kind_of_path sp with
     FW -> true
   | CCI ->
     let (_,mib) = mind_of_path sp in
     let ar = mib.MINDPACKETS.(tyi).MINDARITY
     in informative_construction ar
   | OBJ ->
     anomalylabstrm "termenv__informative_reference"
     [< 'S"Found a constant with a ill-formed path-kind" >])
   | x -> errorlabstrm "termenv__informative_reference"
          [< 'S"Malformed reference" >]
;;

let informative_constant = function
    (DOPN(Const _,_) as K) ->
    informative_reference K
  | _ -> invalid_arg "termenv__informative_constant"
;;


let inf_app f = function Inf (c) -> Inf (f c) | Logic -> Logic;;

let inst_constr (ids,c) args =
    if for_all2eq (fun id (VAR id') -> id = id' | _ _ -> false) ids args then
        c
    else if length ids <> length args then
        invalid_arg "Hum - a constant with a wrong number of arguments"
    else replace_vars (combine(ids,map make_substituend args)) c
;;

let inst_fterm (ids,ft) fargs =
    if for_all2eq (fun id (VAR id') -> id = id'
                       | _ _ -> false) ids fargs then
        ft
    else if length ids <> length fargs then
        invalid_arg "Hum - an fconstant with a wrong number of arguments"
    else replace_vars (combine(ids,map make_substituend fargs)) ft
;;

(* Constants and Existentials *)

(* A constant is an existential if the atomic part of its name is "?"*)
(* is_existential : 'a oper *)
let is_existential_oper = function
    (Const sp) ->
    let bn = basename sp
    in "?" = SA__un_intern(atompart_of_id bn)
  | _ -> false
;;

(* is_existential constr -> bool *)
let is_existential = function
    DOPN(oper,_) -> is_existential_oper oper
  | _ -> false
;;

(* defined_existential : readable_constraints -> constr -> bool *)
let defined_existential sigma = function
    (DOPN(Const sp,_)) ->
    (try (match (evc__map sigma sp).body with
         EVAR_DEFINED _ -> true
       | _ -> false)
     with Not_found -> false)
  | _ -> invalid_arg "termenv__defined_existential"
;;

(* defined_constant : readable_constraints -> constr -> bool *)
let defined_constant (sigma : readable_constraints) = function
    (DOPN(Const sp,_) as K) ->
    (try let (_,cb) = const_of_path sp
         in (match cb.CONSTBODY with
             NONE -> false
           | SOME _ -> true)
     with Not_found -> false)
  | _ -> invalid_arg "termenv__defined_constant"
;;

(* opaque_constant : readable_constraints -> constr -> bool *)
let opaque_constant (sigma : readable_constraints) = function
    (DOPN(Const sp,_) as K) ->
    (try let (_,cb) = const_of_path sp
         in (match cb.CONSTBODY with
             NONE -> true
           | SOME _ -> cb.CONSTOPAQUE)
     with Not_found -> false)
  | _ -> invalid_arg "termenv__opaque_constant"
;;

(* A const is defined if it is a defined existential or a constant
with a non empty body *)
(* defined_const : readable_constraints -> constr -> bool *)
let defined_const (sigma : readable_constraints) = function
    (DOPN(Const sp,_) as K) ->
    (is_existential K & (defined_existential sigma K))
    or (defined_constant sigma K)
  | _ -> invalid_arg "termenv__defined_const"
;;

(* A const is opaque if it is a non-defined existential or
a non-existential opaque constant *)
let opaque_const (sigma : readable_constraints) = function
    (DOPN(Const sp,_) as K) ->
    (is_existential K & (not (defined_existential sigma K)))
    or ((not is_existential K) & (opaque_constant sigma K))
  | _ -> invalid_arg "termenv__opaque_const"
;;

(* A const is evaluable if it is defined and not opaque *)
let evaluable_const (sigma : readable_constraints) K =
    try (not (opaque_const sigma K)) & (defined_const sigma K)
    with Not_found -> false
;;

(* A translucent const is a defined existential *)
let translucent_const (sigma : readable_constraints) K =
    is_existential K & defined_existential sigma K
;;

(* A cookable constant is a defined constant. Raises an error if sp is not
a constant *)
let cookable_constant sp =
try let cb = snd (const_of_path sp) 
 in (match cb.CONSTBODY with
                    SOME _ -> true
                  | NONE -> false)
with Invalid_argument _ -> errorlabstrm "termenv__cookable_constant"
                            [< 'S"Can not cook a non constant" >]
;;

(* exitential_value gives the value of a defined existential *)
let existential_value sigma = function
    (DOPN(Const sp,args) as K) ->
    if defined_const sigma K then
        let evd = evc__map sigma sp
        in match evd.body with
            EVAR_DEFINED c -> inst_constr (ids_of_sign evd.evgoal.hyps,c) (list_of_vect args)
          | _ -> anomalylabstrm "termenv__existential_value"
                 [< 'S"The existential variable code just registered a" ;
                    'SPC ; 'S"grave internal error." >]
    else failwith "undefined existential"
  | _ -> invalid_arg "termenv__existential_value"
;;

(* constant_value gives the value of a defined constant *)
let constant_value (sigma : readable_constraints) = function
    (DOPN(Const sp,args)) as K ->
    let (_,cb) = const_of_path sp
    in if (not cb.CONSTOPAQUE) & defined_const sigma K then
        (match cb.CONSTBODY with
         SOME{v=COOKED body} -> 
         inst_constr (ids_of_sign cb.CONSTHYPS,body) (list_of_vect args)
       | SOME{v=RECIPE _} ->
         anomalylabstrm "termenv__constant_value"
          [< 'S"a transparent constant which was not cooked">]
       | NONE -> anomalylabstrm "termenv__constant_value"
          [< 'S"a defined constant as no body.">])
       else failwith "opaque"
  | _ -> invalid_arg "const_value called with invalid argument";;

(* existential_type gives the type of an existential *)
let existential_type sigma (DOPN(Const sp,args) as K) =
if evc__ctxt_access sigma sp then
    let evd = evc__map sigma sp
    in inst_constr (ids_of_sign evd.evgoal.hyps,evd.evgoal.concl) (list_of_vect args)
else errorlabstrm "termenv__existential_type" [< 'S(string_of_path sp) ; 'SPC;
                                                 'S" not declared" >]
;;

(* constant_type gives the type of a constant *)
let constant_type (sigma : readable_constraints) = function
    (DOPN(Const sp,args)) ->
    let (_,cb) = const_of_path sp
    in inst_constr (ids_of_sign cb.CONSTHYPS,
                    cb.CONSTTYPE) (list_of_vect args)
  | _ -> invalid_arg "constant_type called with invalid argument";;


(* gives the value of a const *)
let const_value sigma (DOPN(Const sp,_) as K) =
    if is_existential K then
        existential_value sigma K
    else constant_value sigma K
;;

(* gives the type of a const *)
let const_type sigma (DOPN(Const sp,_) as K) =
    if is_existential K then
        existential_type sigma K
    else constant_type sigma K
;;

(* Mutual Inductives *)

let mind_path = 
 function
    (DOPN(MutInd (sp,tyi),_)) -> if tyi = 0 then sp else
    let (_,mib) = mind_of_path sp in
    let mip = mind_nth_type_packet mib tyi
    in let (pa,_,k) = repr_path sp 
    in make_path pa (mip.MINDTYPENAME) k 
  | (DOPN(MutConstruct ((sp,tyi),ind),_)) -> 
    let (_,mib) = mind_of_path sp in
    let mip = mind_nth_type_packet mib tyi
    in let (pa,_,k) = repr_path sp 
    in make_path pa (mip.MINDCONSNAMES.(ind-1)) k 

  | _ -> invalid_arg "mind_path called with invalid argument"
;;

let mind_specif_of_mind = function 
      (DOPN(MutInd (sp,tyi),args)) -> let (_,mib) = mind_of_path sp in
   {sp=sp;mib=mib;tyi=tyi;args=args;mip=mind_nth_type_packet mib tyi}
    | _ ->  invalid_arg "mind_specif_of_mind called with invalid argument";;

let mis_singl mispec = mispec.mib.MINDSINGL;;

let mis_lc_arity mispec = 
    let idhyps = ids_of_sign mispec.mib.MINDHYPS 
    and largs = list_of_vect mispec.args in
    (inst_constr (idhyps,mispec.mip.MINDLC) largs,
       inst_constr (idhyps,mispec.mip.MINDARITY) largs);;

let mis_lc mispec = 
    inst_constr (ids_of_sign mispec.mib.MINDHYPS,
                    mispec.mip.MINDLC) (list_of_vect mispec.args);;


let mis_arity mispec = 
    let idhyps = ids_of_sign mispec.mib.MINDHYPS 
    and largs = list_of_vect mispec.args in 
    inst_constr (idhyps,mispec.mip.MINDARITY) largs;;

let mind_arity = mis_arity o mind_specif_of_mind;;


let mis_nparams mispec = mispec.mib.MINDNPARAMS;;
let mis_ntypes mispec = mispec.mib.MINDNTYPES;;
let mis_consnames mispec = mispec.mip.MINDCONSNAMES;;
let mis_nconstr mispec = vect_length (mispec.mip.MINDCONSNAMES);;
let mis_typename mispec = mispec.mip.MINDTYPENAME;;

let mis_kd mispec = mispec.mip.MINDKD;;
let mis_kn mispec = mispec.mip.MINDKN;;

let mis_lamexp mispec = 
    match mispec.mip.MINDLAMEXP with
        NONE -> NONE
      | SOME c -> SOME(inst_constr (ids_of_sign mispec.mib.MINDHYPS,c)
                                   (list_of_vect mispec.args));;

let mind_lamexp = mis_lamexp o mind_specif_of_mind;;

let mis_recargs mispec =  
    map_vect (fun mip -> mip.MINDLISTREC) mispec.mib.MINDPACKETS;;

let mis_recarg mispec =  mispec.mip.MINDLISTREC;;

let mind_recargs = mis_recargs o mind_specif_of_mind;;
let mind_nparams = mis_nparams o mind_specif_of_mind;;

(* functions working for both notions of inductive definitions *)

(* TODO: Implement opaque abstractions *)
(* TODO: Understand what opaque abstractions _mean_! *)
let evaluable_abst = function
    (DOPN(Abst _,_)) -> true
  | _ -> invalid_arg "termenv__evaluable_abst";;

let translucent_abst = function
    (DOPN(Abst _,_)) -> false
  | _ -> invalid_arg "termenv__translucent_abst";;


let abst_value A = abstraction__contract_abstraction A;;
let global_abst sp args = abstraction__make_abstraction sp args;;

let search_abst pk id args =
let sp = nametab__sp_of_id pk id
in if abstraction__is_abstraction sp then
    global_abst sp args
   else failwith "search_abst"
;;


(* $Id: termenv.ml,v 1.8 1996/02/12 15:48:32 cpaulin Exp $ *)
