(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                typing.ml                                 *)
(****************************************************************************)

(* A type-checker of the full system with a predicative hierarchy of types *)
(* Three flavors of Prop and Type for Christine's extraction algorithm     *)
(* Inductive types                                                         *)

#infix "o";;
#open "std";;
#open "names";;
#open "more_util";;
#open "vectops";;
#open "generic";;
#open "impuniv";;
#open "environ";;
#open "pp";;
#open "stdpp";;
#open "term";;
#open "termenv";;
#open "constrtypes";;
#open "reduction";;
#open "himsg";;
#open "proof";;

(********************************************************************)

(*                      Inductive types                 *)

(********************************************************************)

(* gives the vector of constructors and of 
   types of constructors of an inductive definition 
   correctly instanciated *)

let mis_type_mconstructs mispec =
let specif = mis_lc mispec
and ntypes = mis_ntypes mispec
and nconstr = mis_nconstr mispec in
let make_Ik k = DOPN(MutInd(mispec.sp,k),mispec.args) 
and make_Ck k = DOPN(MutConstruct((mispec.sp,mispec.tyi),k+1),mispec.args) in
(tabulate_vect make_Ck nconstr, SAPPVList specif (tabulate_list make_Ik ntypes));;

let type_mconstructs sigma  mind =
let redmind  = check_mrectype_spec sigma mind in
let mispec = mind_specif_of_mind redmind in 
mis_type_mconstructs mispec;;


(* gives the type of one constructor of an inductive definition *)

let type_mconstruct sigma i mind =
let (DOPN(MutInd(sp,tyi),args) as redmind) 
    = check_mrectype_spec sigma mind in
let mispec = mind_specif_of_mind redmind in 
let specif = mis_lc mispec
and ntypes = mis_ntypes mispec
and nconstr = mis_nconstr mispec in
let make_Ik k = DOPN(MutInd(sp,k),args) in 
 if i > nconstr then error "Not enough constructors in the type" else 
 SAPPViList (i-1) specif (tabulate_list make_Ik ntypes);;

(* compute the type of a well-formed expression *)

let is_correct_arity_dep sigma kd = srec where rec srec ind (Pt,t) =
  match whd_betadeltaiota sigma Pt,whd_betadeltaiota sigma t with
    DOP2(Prod,A1,DLAM(_,A2)),DOP2(Prod,A1',DLAM(_,A2')) -> conv_x sigma A1 A1' &
                                     srec (applist(lift 1 ind,[Rel 1])) (A2,A2')
  | DOP2(Prod,A1,DLAM(_,A2)),_               ->
    (match whd_betadeltaiota sigma A2 with
     DOP0(Sort s) -> conv_x sigma A1 ind & exists (sort_eq_x s) kd
   | _ -> false)
  | _                             -> false;;


let is_correct_arity_nodep sigma kd = srec where rec srec (Pt,t) =
  match whd_betadeltaiota sigma Pt,whd_betadeltaiota sigma t with
    DOP2(Prod,A1,DLAM(_,A2)),DOP2(Prod,A1',DLAM(_,A2')) -> conv_x sigma A1 A1' & srec (A2,A2')
  | DOP2(Prod,A1,DLAM(_,A2)),_               -> false
  | (_,DOP2(Prod,_,_))               -> false
  | (DOP0(Sort s),_)                         -> exists (sort_eq_x s) kd
  | _ -> false;;

let make_arity_dep k = mrec 
  where rec mrec c M = match whd_betadeltaiota mt_evc c with 
    (DOP0(Sort _))          -> arrow M k
  | DOP2(Prod,B,DLAM(n,C))  -> prod_name(n,B,mrec C (applist(lift 1 M,[Rel 1])))
  | _                       -> error "make_arity_dep: not an arity";;

let make_arity_nodep k = mrec 
 where rec mrec c = match whd_betadeltaiota mt_evc c with 
       (DOP0(Sort _)) -> k
     | DOP2(Prod,B,DLAM(x,C))         -> DOP2(Prod,B,DLAM(x,mrec C))
     | _                   -> error "make_arity_nodep: not an arity";;


let message_arity_non_con env kd kn (c,P) (ind,t,Pt) = 
  let listaritydep = map (fun s -> make_arity_dep (DOP0(Sort s)) t ind) kd 
and listaritynodep = map (fun s -> make_arity_nodep (DOP0(Sort s)) t) kn 
in (HOV 0 
[<  'S "Incorrect elimination of"; 'WS(1); pTERMINENV(env,c); 'WS(1);
        'S "in the inductive type "; 
        pTERMINENV(env,ind); 'FNL;
   'S "The elimination predicate "; pTERMINENV(env,P); 'WS(1);
       'S "has type ";  pTERMINENV (env,Pt); 'FNL;
       'S "It should be one of : "; 'BRK(1,0) ;
      (HOV 0 (prlist_with_sep (fun () -> [<'S","; 'BRK(1,0)>]) 
        (fun a -> pTERMINENV(env,a)) 
       (listaritydep@listaritynodep))); 'FNL >]) ;;

let message_badsort kp ki = 
let explanation =
  if (is_info_sort kp) & not (is_info_sort ki) 
  then "non-informative objects may not construct informative ones."
  else match (kp,ki) with 
       (DOP0(Sort (Type(Pos,_))),DOP0(Sort (Prop(_)))) -> 
    "strong informative elimination is not compatible with extraction."
     | (DOP0(Sort (Type(Null,_))),DOP0(Sort (Prop(_))))
   -> "strong elimination on non-small inductive types leads to paradoxes."
in (HOV 0 
  [< 'FNL; 
     'S "Elimination of an inductive object of sort : "; pTERM ki; 'BRK(1,0);
     'S "is not allowed on a predicate in sort : "; pTERM kp ;'FNL;
     'S "because"; 'WS(1); 'S explanation >]) ;;

exception Arity of (constr*constr) option;;

let is_correct_arity env sigma kd kn (c,P) = srec where rec srec ind (Pt,t) =
 try (match whd_betadeltaiota sigma Pt,whd_betadeltaiota sigma t with
    DOP2(Prod,A1,DLAM(_,A2)),DOP2(Prod,A1',DLAM(_,A2')) 
          -> if conv sigma A1 A1' 
             then srec (applist(lift 1 ind,[Rel 1])) (A2,A2') 
             else raise (Arity NONE)
  | DOP2(Prod,A1,DLAM(_,A2)),ki
          -> let k = whd_betadeltaiota sigma A2 in 
             let ksort = (match k with DOP0(Sort s) -> s 
                    | _ -> raise (Arity NONE)) in
             if (conv sigma A1 ind)&(exists (sort_eq_x ksort) kd)  then 
                                        true,k
             else  raise (Arity (SOME(k,ki)))
  | k,DOP2(Prod,_,_) ->  raise (Arity NONE)
  | k,ki   -> let ksort = (match k with DOP0(Sort s) -> s 
                         | _ ->  raise (Arity NONE)) in
               if exists (sort_eq_x ksort) kn then false,k
               else raise (Arity (SOME(k,ki))))
with (Arity (SOME(k,ki)))  -> errorlabstrm "is_correct_arity"
                        [< (message_arity_non_con env kd kn (c,P) (ind,t,Pt));
                           (message_badsort k ki) >]
  |  (Arity NONE) -> errorlabstrm "is_correct_arity"
                        [<(message_arity_non_con env kd kn (c,P) (ind,t,Pt))>]
;;


(******** Check if the case analysis for c of type typc 
          is compatible with the type typP of the predicate *****)

let find_case_dep_mis env sigma mispec (c,P) (mind,largs) typP = 
 let nparams = mis_nparams mispec and kd = mis_kd mispec 
  and kn = mis_kn mispec and t = mis_arity mispec in 
  let (globargs,la) = chop_list nparams largs in
  let glob_t = hnf_prod_applist sigma "find_elim_boolean" t globargs in
  fst(is_correct_arity env sigma kd kn (c,P) (applist(mind,globargs)) (typP,glob_t))
;;

let find_case_dep_nparams env sigma (c,P) typc typP =    
  let (mind,largs) = find_mrectype sigma typc in
  let mispec = mind_specif_of_mind mind in 
   find_case_dep_mis env sigma mispec (c,P) (mind,largs) typP, mis_nparams mispec;;


(****************************************************)
(* Type of elimination                              *)
(****************************************************)

(********* Case construction ***********)

let type_one_branch_dep (sigma,nparams,globargs,P) co t 
      = typrec 0 (hnf_prod_applist sigma "type_case" t globargs)
      where rec typrec n c = match whd_betadeltaiota sigma c with
       DOP2(Prod,A1,DLAM(x,A2)) -> prod_name (x,A1,typrec (n+1) A2)
     | x -> let (DOPN(AppL,LAV)) = ensure_appl x in 
           let (_,Vargs) = chop_vect (nparams+1) LAV
           in applist 
              (appvect ((lift n P),Vargs),
              [applist(co,((map (lift n) globargs)@(rel_list 0 n)))]);;

let type_one_branch_nodep (sigma,nparams,globargs,P) t 
      = typrec 0 (hnf_prod_applist sigma "type_case" t globargs)
  where rec typrec n c = match whd_betadeltaiota sigma c with 
          DOP2(Prod,A1,DLAM(x,A2)) -> DOP2(Prod,A1,DLAM(x,typrec (n+1) A2))
        | x -> let (DOPN(AppL,LAV)) = ensure_appl x in
               let (_,Vargs) = chop_vect (nparams+1) LAV in 
               appvect (lift n P,Vargs);;

let type_case_branches env sigma ct Pt P c =
  try let (MI,largs) = find_mrectype sigma ct
      and dep,nparams = find_case_dep_nparams env sigma (c,P) ct Pt 
                       (* also check the arity is correct *)
  in  let (lconstruct,ltypconstr) = type_mconstructs sigma MI in
      let (globargs,la) = chop_list nparams largs in
  if dep then (map2_vect (type_one_branch_dep (sigma,nparams,globargs,P))
                            lconstruct ltypconstr,
                beta_applist (P,(la@[c])))
     else (map_vect (type_one_branch_nodep (sigma,nparams,globargs,P)) 
                    ltypconstr,
           beta_applist (P,la))
  with Induc -> errorlabstrm "Case on non inductive type"
       (HOV 0 [< 'S "In Case expression "; pTERMINENV(env,c); 'CUT ; 'S "has type"; 
              'WS 1; pTERMINENV(env,ct); 'WS 1; 'CUT; 
              'S "which is not an inductive definition">])
               ;;

let type_case env sigma ct Pt P c = 
  try let (MI,largs) = find_mrectype sigma ct
      and dep,nparams = find_case_dep_nparams env sigma (c,P) ct Pt 
                       (* also check the arity is correct *)
  in  let (globargs,la) = chop_list nparams largs in
  if dep then beta_applist (P,(la@[c]))
     else beta_applist (P,la)
  with Induc -> errorlabstrm "Case on non inductive type"
       (HOV 0 [< 'S "In Case expression "; pTERMINENV(env,c); 'CUT ; 'S "has type"; 
              'WS 1; pTERMINENV(env,ct); 'WS 1; 'CUT; 
              'S "which is not an inductive definition">])
               ;;

let check_branches_message env sigma (c,ct) (explft,lft) = 
let n = vect_length explft and expn = vect_length lft in
if n<>expn then 
errorlabstrm "Case with wrong number of cases"
 (HOV 0 [< 'S "Case on term "; pTERMINENV(env,c); 'CUT ; 'S "of type"; 'WS 1;
           pTERMINENV(env,ct); 'WS 1; 'CUT; 'S "expects " ; 'INT expn; 'S " branches" >]) 
else
let rec check_conv i = 
    if i = n then () else
    if not (conv sigma lft.(i) (explft.(i))) then 
errorlabstrm "Ill-formed branches" 
 (HOV 0 [< 'S "In Case expression on term "; pTERMINENV(env,c); 
           'SPC; 'CUT; 'S "the branch  " ; 'INT (i+1); 'S " has type";  'SPC;
         pTERMINENV(env,lft.(i)); 'CUT; 
         'S " which should be :"; 'WS 1; pTERM explft.(i) >]) 
    else check_conv (i+1) 
in check_conv 0;;



let type_case_elim env sigma ct Pt lft P c =
let (bty,rslty) = type_case_branches env sigma ct Pt P c in
check_branches_message env sigma (c,ct) (bty,lft); rslty;;



let rec sort_of_arity sigma c =
     match whd_betadeltaiota sigma c with
     DOP0(Sort( _)) as c'      -> c'
   | DOP2(Prod,_,DLAM(_,c2)) -> sort_of_arity sigma c2
   | _            -> error "sort_of_arity: Not an arity";;


(* B7| Prod intro and quantification *)

(* Cumulativity: slide type t up if product with second type imposes it *)
let adjust t = function
    DOP0(Sort(Prop(_)))    -> t
  | DOP0(Sort(Type(_,u'))) -> (match t with 
          DOP0(Sort(Type(c,u))) -> let v = sup(u,u') 
                                   in if eq(u,v) then t else DOP0(Sort(Type(c,v)))
       | _ -> invalid_arg "adjust")
  | _          -> invalid_arg "level_of_kind/adjust";;

let type_of_product sigma domtyp rangtyp =
    match whd_betadeltaiota sigma rangtyp with
    DOP0(Sort(Prop _)) -> rangtyp
  | (DOP0(Sort(Type _))) as typ' -> adjust typ' domtyp
  | _ -> error "type_of_product: Not a sort"
;;

let type_of_type = function
    DOP0(Sort(Type(cts,u))) -> DOP0(Sort(Type(cts,Super u)))
  | DOP0(Sort(Prop cts)) -> DOP0(Sort(Type(cts,prop_univ)))
  | _ -> error "type_of_type: Not a sort"
;;

(* Fix introduction *)


(*******************************************************************)
(* Checking correctness of recursives declarations                 *)
(*******************************************************************)

exception NotFound;;

(* find an integer in a sorted list *)
let find_sorted_assoc p = findrec where rec findrec = function 
   ((a,ta)::l) -> if a < p then findrec l 
             else if a = p then ta else raise NotFound
  | _     -> raise NotFound;;

let map_lift_fst = map (function (n,t)->(n+1,t));;

(* mind_recarg is a vector giving the list of recargs for each type in
the mutually inductive definition *)

(* propagate checking for F,incorporating recursive arguments *)
let check_term mind_recvec F = crec where rec crec n l (lrec,c) = match
  (lrec,strip_outer_cast c) with
   (Norec::lr,DOP2(Lambda,_,DLAM(_,b))) -> 
                           let l' = map_lift_fst l in crec (n+1) l' (lr,b)
 | (Mrec(i)::lr,DOP2(Lambda,_,DLAM(_,b)))  -> let l' = map_lift_fst l in
                                crec (n+1) ((1,mind_recvec.(i))::l') (lr,b)
 | (Imbr(lc)::lr,DOP2(Lambda,_,DLAM(_,b))) -> let l' = map_lift_fst l in
                                crec (n+1) ((1,lc)::l') (lr,b)
 | _,f -> F n l f;;

let is_inst_var sigma k c = match whd_betadeltaiota_stack sigma c [] with 
     (Rel n,_) -> n=k
    | _        -> false;;

(* Check if t is a subterm of Rel n, and gives its specification, 
assuming lst already gives index of
subterms with corresponding specifications of recursive arguments *)

let case_branches v = sub_vect v 2 (vect_length v-2);;

(* A powerful notion of subterm *)

let is_subterm_specif sigma lcx mind_recvec = crec where rec 
crec n lst c = 
   let (x,_) = whd_betadeltaiota_stack sigma c [] in match x with 
     (Rel k)         -> find_sorted_assoc k lst
|  DOPN(MutCase,vargs)  -> let c = vargs.(1) and br = case_branches vargs in
       if vect_length br = 0 then [||] else
       let lcv = (try if is_inst_var sigma n c then lcx else crec n lst c 
                  with NotFound -> (make_vect (vect_length br) []))
       in if vect_length br <> vect_length lcv then anomaly "is_subterm"
          else let stl = map2_vect 
            (fun lc a -> check_term mind_recvec crec n lst (lc,a)) lcv br 
               in stl.(0)
|  DOP2(Lambda,_,DLAM(_,b)) -> let lst' = map_lift_fst lst in
                                crec (n+1) lst' b
(***** Experimental change *************************)
|  DOP0(Meta _)             -> [||]
(***************************************************)
|  _                        -> raise NotFound;;

let is_subterm sigma lcx mind_recvec n lst c = 
try is_subterm_specif sigma lcx mind_recvec n lst c; true 
with NotFound -> false;;



(* Checking function for terms containing existential variables.
 The function noccur_with_meta consideres the fact that
 each existential variable (as well as each isevar)
 in the term appears applied to its local context,
 which may contain the CoFix variables. These occurrences of CoFix variables
 are not considered *)

  let noccur_with_meta sigma n m term = let rec occur_rec n = function
    Rel(p)        -> if n<=p & p<n+m then raise Occur
  | VAR _         -> ()
  | DOPN(op,cl)   -> (match op with
                         AppL ->( match strip_outer_cast (cl.(0)) with
                                     DOP0 (Meta _) -> ()
                                   | _             -> do_vect (occur_rec n) cl)
                       | (Const sp) -> if evc__in_dom sigma sp
                                       then ()
                                       else do_vect (occur_rec n) cl
                       |  _    -> do_vect (occur_rec n) cl)
  | DOPL(_,cl)    -> do_list (occur_rec n) cl
  | DOP0(_)       -> ()
  | DOP1(_,c)     -> occur_rec n c
  | DOP2(_,c1,c2) -> occur_rec n c1; occur_rec n c2
  | DLAM(_,c)     -> occur_rec (n+1) c
  | DLAMV(_,v)    -> do_vect (occur_rec (n+1)) v
  | _             -> ()
  in try (occur_rec n term; true) with Occur -> false;;


(*******Experimental change ***************************************)
(* noccur_bet ==> noccur_with_meta + an entry for Meta in the case*)
(******************************************************************)

let check_subterm_rec_meta sigma vectn k def = 
if k < 0 then true else
let nfi = vect_length vectn in 
 (* check fi does not appear in the k+1 first abstractions, 
      gives the type of the k+1-eme abstraction  *)
let rec check_occur n def = match strip_outer_cast def with
      DOP2(Lambda,A,DLAM(_,B)) -> if noccur_with_meta sigma n nfi A then
                      if n = k+1 then (A,B) else check_occur (n+1) B
                      else error "Bad occurrence of recursive call"
           | _ -> error "Not enough abstractions in the definition"
in let (C,d) = check_occur 1 def 
in let (DOPN(MutInd(sp,tyi),_) as MI,largs) = 
  try find_minductype sigma C 
  with Induc -> error "Recursive definition on a non inductive type" in

let mind_recvec = mind_recargs MI in 
let lcx = mind_recvec.(tyi) in

let rec check_rec_call n lst t = 
            (* n gives the index of the recursive variable *)
if noccur_with_meta sigma (n+k+1) nfi t then true (* no recursive call in the term *)
else match whd_betadeltaiota_stack sigma t [] with 
     (Rel p,l) -> 
       if n+k+1 <= p & p < n+k+nfi+1 (* recursive call *)
       then let glob = nfi+n+k-p in  (* the index of the recursive call *) 
            let np = vectn.(glob) in (* the decreasing arg of the rec call *)
            if length l > np then 
                  let (la,(z::lrest)) = chop_list np l in 
                  if (is_subterm sigma lcx mind_recvec n lst z) 
                  then for_all (check_rec_call n lst) (la@lrest)
                  else error "Recursive call applied to an illegal term"
            else error  "Not enough arguments for the recursive call"
       else for_all (check_rec_call n lst) l        
    | (DOPN(MutCase,vargs),l) -> let P = vargs.(0) and c = vargs.(1) 
                                 and lrest = case_branches vargs in 
          let lc = try if is_inst_var sigma n c then lcx 
                        else is_subterm_specif sigma lcx mind_recvec n lst c
                   with NotFound -> make_vect (vect_length lrest) []
          in (for_all2eq_vect
              (fun c a -> check_term mind_recvec check_rec_call n lst (c,a))
              lc lrest) & (for_all (check_rec_call n lst) (c::P::l)) 
    | (DOP2(_,A,DLAM(_,B)),[]) ->   (check_rec_call n lst A)
                          & (check_rec_call (n+1) (map_lift_fst lst) B)
    | (DOP2(_,A,B),l) ->   (check_rec_call n lst A)
                          & (check_rec_call n lst B)
                          & (for_all (check_rec_call n lst) l)
    | (DOPN(_,la),l) ->   (for_all_vect (check_rec_call n lst) la)
                          & (for_all (check_rec_call n lst) l)
    | (DOP0 (Meta _),l) -> true
    | (_,l)    ->   for_all (check_rec_call n lst) l

in check_rec_call 1 [] d;;    
 

(* vargs is supposed to be built from A1;..Ak;[f1]..[fk][|d1;..;dk|]
and vdeft is [|t1;..;tk|] such that f1:A1,..,fk:Ak |- di:ti
nvect is [|n1;..;nk|] which gives for each recursive definition 
the inductive-decreasing index 
the function checks the convertibility of ti with Ai *)

let check_fix sigma (DOPN(Fix(nvect,j),vargs)) = 
 let nbfix = let nv = vect_length vargs in 
            if nv < 2 then error "Ill-formed recursive definition" else nv-1 in
  let varit = sub_vect vargs 0 nbfix in
  let ldef = last_vect vargs in
  let ln = vect_length nvect and la = vect_length varit in
  if ln <> la then error "Ill-formed fix term"
  else let (lna,vdefs) = decomp_DLAMV_name ln ldef in 
       let vlna = vect_of_list lna in
       let check_type i = 
         try check_subterm_rec_meta sigma nvect nvect.(i) vdefs.(i) 
         with UserError (s,str) ->
           let ass = assumptions_for_print lna in
     errorlabstrm "typing__check_fix"
     (HOV 0 [< str; 'FNL; 
       'S"The " ; if ln = 1 then [<>] else [<'INT (i+1); 'S "-th ">];
       'S"recursive definition "; 'CUT ; pTERMINENV (ass,vdefs.(i)); 'CUT;
       'S " is not well-formed">])
           in for i = 0 to ln-1 do check_type i done;;

let type_fix sigma (DOPN(Fix(_,j),vargs)) vdeft = 
let lt = vect_length vdeft in 
if vect_length vargs = lt+1 then 
for i = 0 to lt-1 do 
    if not (conv_leq sigma vdeft.(i) (lift lt vargs.(i)))
    then let lc = last_vect vargs in 
    let (lna,vdefs) = decomp_DLAMV_name lt lc in 
    let ass = assumptions_for_print lna in
    errorlabstrm "Ill typed fix-point" 
    (HOV 0 [< 'S"The " ; if lt = 1 then [<>] else [<'INT (i+1); 'S "-th">];
       'S"recursive definition " ; 'CUT; pTERMINENV (ass,vdefs.(i));
       'S "has type";'WS 1; pTERMINENV (ass,vdeft.(i));'WS 1;'CUT;
        'S "it should be"; 'WS 1; pTERM vargs.(i) >])
done; vargs.(j);;


(*********************************************)
(* Checking correctness of CoFix definitions *)
(*********************************************)


(* The function check_guard_rec checks the guarded condition for one 
   definition of the block.
   The argument  nbfix is the number of functions defined in the block.
   The argument  def is the definition of certain function of the block.
   The argument deftype is the type of def  *)


(* Checking function without existential variables *)

let check_guard_rec sigma nbfix def deftype = 

  let rec codomain_is_coind C  =
     match strip_outer_cast C with
           DOP2(Prod,A,DLAM(_,B)) ->  codomain_is_coind B 
         | B  -> try find_mcoinductype sigma B
                 with Induc -> error "The codomain is not a coinductive type"
         | _ -> error "Type of Cofix function not as expected"   in

  let (DOPN(MutInd(sp,tyi),_) as MI,_) = codomain_is_coind deftype  in
  let vlra = (mind_recargs MI).(tyi) in  

  let rec check_rec_call alreadygrd n vlra  t = 
  if noccur_bet n nbfix t then true (* no recursive call in the term *)
  else match whd_betadeltaiota_stack sigma t [] with 

      (Rel p,l) -> 
              if n <= p & p < n+nbfix (* recursive call *)
             then if alreadygrd
                  then if for_all (noccur_bet n nbfix) l 
                       then true
                       else error "Nested recursive occurrences"
                  else error "Unguarded  recursive call"
             else  error "Variable applied to a recursive call "

    | (DOPN ((MutConstruct(spj,i)),l), args)  ->
         let lra =vlra.(i-1) in 
         let MI = DOPN(MutInd(spj),l) in
         let _,realargs = chop_list (mind_nparams MI) args in
         let rec process_args_of_constr l lra   =
                     match l with
                         []    -> true 
                       | t::lr ->( match lra  with 
                                    [] -> error "Anomaly: a constructor with an empty list of recargs is being applied" 
                                 |  (Mrec _)::lrar -> 
                                         (check_rec_call true n  vlra t) &
                                         (process_args_of_constr lr lrar)
                
                                 |  Norec::lrar  -> 
                                         if    (noccur_bet n nbfix t) 
                                         then  (process_args_of_constr lr lrar)
                                         else  error "Recursive call inside a non-recursive argument of constructor"

                                 |  (Imbr vlra)::lrar  -> 
                                            (check_rec_call true n vlra t ) &
                                            (process_args_of_constr lr lrar)) 
         in (process_args_of_constr realargs lra)


    | (DOP2(Lambda,A,DLAM(_,B)),[]) -> 
               if (noccur_bet n nbfix A)
               then (check_rec_call alreadygrd (n+1)  vlra B)
               else error "Recursive call in the type of an abstraction"
       
    | (DOPN(MutCase,vargs),l) -> 
          let P = vargs.(0)  and 
             c = vargs.(1)  and
             vrest = case_branches vargs in 
              if (noccur_with_meta sigma n nbfix P) 
              then if (noccur_with_meta sigma n nbfix c) 
                   then if (for_all (noccur_with_meta sigma n nbfix) l)
                        then (for_all_vect (check_rec_call alreadygrd n vlra) vrest)
                        else error "Recursive call in the argument of a function defined by cases"        
                   else error "Recursive call in the argument of a case expression"
              else error "Recursive call in the type of a Case expression" 


    | _    ->   error "Definition not in guarded form"

in check_rec_call false 1 vlra def;;


let check_guard_rec_meta sigma nbfix def deftype = 

  let rec codomain_is_coind C  =
     match strip_outer_cast C with
           DOP2(Prod,A,DLAM(_,B)) ->  codomain_is_coind B 
         | B  -> try find_mcoinductype sigma B
                 with Induc -> error "The codomain is not a coinductive type"
         | _ -> error "Type of Cofix function not as expected"   in

  let (DOPN(MutInd(sp,tyi),_) as MI,_) = codomain_is_coind deftype  in
  let vlra = (mind_recargs MI).(tyi) in  

  let rec check_rec_call alreadygrd n vlra  t = 
    if (noccur_with_meta sigma n nbfix t) 
    then true 
    else match whd_betadeltaiota_stack sigma t [] with 
     
      (DOP0 (Meta _), l) -> true

    |  (Rel p,l) -> 
              if n <= p & p < n+nbfix (* recursive call *)
             then if alreadygrd
                  then if for_all (noccur_with_meta sigma n nbfix) l 
                       then true
                       else error "Nested recursive occurrences"
                  else error "Unguarded  recursive call"
             else  error "Va "

    | (DOPN ((MutConstruct(spj,i)),l), args)  ->
         let lra =vlra.(i-1) in 
         let MI = DOPN(MutInd(spj),l) in
         let _,realargs = chop_list (mind_nparams MI) args in
         let rec process_args_of_constr l lra   =
                     match l with
                         []    -> true 
                       | t::lr ->( match lra  with 
                                    [] -> error "Anomaly: a constructor with an empty list of recargs is being applied" 
                                 |  (Mrec _)::lrar -> 
                                         (check_rec_call true n  vlra t) &
                                         (process_args_of_constr lr lrar)
                
                                 |  Norec::lrar  -> 
                                         if    (noccur_with_meta sigma n nbfix t) 
                                         then  (process_args_of_constr lr lrar)
                                         else  error "Recursive call inside a non-recursive argument of constructor"

                                 |  (Imbr vlra)::lrar  -> 
                                            (check_rec_call true n vlra t ) &
                                            (process_args_of_constr lr lrar)) 
         in (process_args_of_constr realargs lra)


    | (DOP2(Lambda,A,DLAM(_,B)),[]) -> 
               if (noccur_with_meta sigma n nbfix A)
               then (check_rec_call alreadygrd (n+1)  vlra B)
               else error "Recursive call in the type of an abstraction"


    | (DOPN(MutCase,vargs),l) -> 
         let P = vargs.(0)  and 
             c = vargs.(1)  and
             vrest = case_branches vargs in 
              if (noccur_with_meta sigma n nbfix P) 
              then if (noccur_with_meta sigma n nbfix c) 
                   then if (for_all (noccur_with_meta sigma n nbfix) l)
                        then (for_all_vect (check_rec_call alreadygrd n vlra) vrest)
                        else error "Recursive call in the argument of a function defined by cases"        
                   else error "Recursive call in the argument of a case expression"
              else error "Recursive call in the type of a Case expression" 

    | _    -> error "Definition not in guarded form"

in check_rec_call false 1 vlra def;;


(* The  function which checks that the whole block of definitions 
   satisfies the guarded condition *)

let check_cofix sigma (DOPN(CoFix(j),vargs)) = 
  let nbfix = let nv = vect_length vargs in 
                  if nv < 2 
                  then error "Ill-formed recursive definition" 
                  else nv-1 in
  let varit = sub_vect vargs 0 nbfix in
  let ldef  = last_vect vargs in
  let la    = vect_length varit in
  let (lna,vdefs) = decomp_DLAMV_name la ldef in
  let vlna = vect_of_list lna in
  let check_type i = 
         try check_guard_rec_meta sigma nbfix vdefs.(i)  varit.(i)
         with UserError (s,str) ->
         let ass = assumptions_for_print lna in 
     errorlabstrm "typing__check_cofix"
     (HOV 0 [< str; 'FNL; 
       'S" The " ; if la = 1 then [<>] else [<'INT (i+1); 'S "-th ">];
       'S"recursive definition "; 'CUT ; pTERMINENV (ass,vdefs.(i)); 'CUT;
       'S " is not well-formed">])
  in for i = 0 to nbfix-1 do check_type i done;;


let type_cofix sigma (DOPN(CoFix(j),vargs)) vdeft = 
let lt = vect_length vdeft in 
if vect_length vargs = lt+1 then 
for i = 0 to lt-1 do 
    if not (conv_leq sigma vdeft.(i) (lift lt vargs.(i)))
    then let lc = last_vect vargs in 
    let (lna,vdefs) = decomp_DLAMV_name lt lc in 
    let ass = assumptions_for_print lna in
    errorlabstrm "Ill typed fix-point" 
   (HOV 0 [< 'S"The " ; if lt = 1 then [<>] else [<'INT (i+1); 'S "-th">];
       'S"recursive definition " ; 'CUT; pTERMINENV (ass,vdefs.(i));
       'S "has type";'WS 1; pTERMINENV (ass,vdeft.(i));'WS 1;'CUT;
        'S "it should be"; 'WS 1; pTERM vargs.(i) >])
done; vargs.(j);;


(* A function which checks that a term well typed verifies both
   syntaxic conditions *)


let control_only_guard sigma = control_rec where rec control_rec = function
    Rel(p)        -> ()
  | VAR _         -> ()
  | DOP0(_)       -> ()
  | DOPN(CoFix(_),cl) as cofix  -> (check_cofix sigma cofix) ; 
                                   (do_vect control_rec cl)
  | DOPN(Fix(_),cl) as fix  -> (check_fix sigma fix) ; 
                               (do_vect control_rec cl)
  | DOPN(_,cl)    -> do_vect control_rec cl
  | DOPL(_,cl)    -> do_list control_rec cl
  | DOP1(_,c)     -> control_rec c
  | DOP2(_,c1,c2) -> control_rec c1; control_rec c2
  | DLAM(_,c)     -> control_rec c
  | DLAMV(_,v)    -> do_vect control_rec v
;;



(* $Id: typing.ml,v 1.10 1996/02/12 15:48:36 cpaulin Exp $ *)
