(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                 mltop.ml                                 *)
(****************************************************************************)
(* Code to hook Coq into the ML toplevel -- depends on having the
   caml-light compiler mostly visible.

   The functions implemented here are:

   [read_ml_module name]

   Reads in the ML module fname from the current ML path.

   [import_ml_module name]

   Does a "#open" in the toplevel environment of the named module.

   [freeze_ml_env ()]

   returns a list of the ML modules which are "open" in the current
   environment.

   [restore_ml_env ()]

   Restores the specified list of ML modules to the current
   environment, killing off anything which was there already.

   [use_string ()]

   Parses, compiles, and executes the given string

   ----------------------------------------------------------------

   How to build an ML module interface with these functions

   ----------------------------------------------------------------

   The idea is that the ML directory path is like the Coq directory
   path.  So we can maintain the two in parallel.

   In the same way, we can use the "ml_env" as a kind of ML
   environment, which we freeze, unfreeze, and add things to just like
   to the other environments.

   Finally, we can create an object which is an ML module, and require
   that the "caching" of the ML module cause the loading of the
   associated ML file, if that file has not been yet loaded.  Of
   course, the problem is how to record dependences between ML
   modules.

   I do not know of a solution to this problem, other than to
   put all the needed names into the ML Module object.

   *)

#open "unix";;
#open "std";;
#open "pp";;
#open "stdpp";;
#open "std";;
#open "lomsg";;
#open "libobject";;
#open "library";;
#open "more_util";;
#open "system";;
#open "vernacinterp";;
#open "tacmach";;

let (no_load_object : string -> unit) =
  fun s -> errorlabstrm "mltop__no_load_object"
                        [< 'S"Loading of ML object file forbidden" >]
;;

let (no_include : string -> unit) =
  fun s -> warning "Can not access the ML compiler"
;;

let l_o = ref no_load_object;;

let incl = ref no_include;;

let do_directive (d,s) =
  let name = "/tmp/coq_directive" ^ (string_of_int (getpid())) ^ ".ml" in
  let f = open_out name in
  do_list (output_string f) ["#";d;" \"";s;"\";;\n"];
  close_out f;
  !incl name;
  sys__remove name
;;

let use_string s =
  let name = "/tmp/coq_string" ^ (string_of_int (getpid())) ^ ".ml" in
  let f = open_out name in
  do_list (output_string f) [s;"\n"];
  close_out f;
  !incl name;
  sys__remove name
;;

(* ML Search Path *)
let add_ml_path s =
    do_directive("directory",s)
;;

let del_ml_path s =
    error "del_ml_path is NOT implemented"
;;

(*
NEW_MAIN:
let get_ml_path () = !misc__load_path;;
*)

let get_ml_path () = system__search_paths();;
(* ML Modules *)

let do_open_ml_module s =
    do_directive("open",s)
;;

let do_close_ml_module s =
    do_directive("close",s)
;;

(* NEW_MAIN 
let get_used_modules () = map (fun m -> m.mod_name) !used_modules;;
let module_used_p s = exists (fun m -> s = m.mod_name) !used_modules;;
*)
let get_used_modules () = ([]:string list);;
let module_used_p s = false;;

let orig_used_modules = ref (get_used_modules());;
let update_init_state() = orig_used_modules := (get_used_modules());;

let add_used_module s =
    if not(module_used_p s) then do_open_ml_module s;;

let remove_used_module s =
    if not(listset__memb !orig_used_modules s) then
        do_close_ml_module s
;;

let init_ml_env () =
    (* used_modules := []; : NEW_MAIN *)
    do_listRL add_used_module !orig_used_modules
;;

let freeze_ml_env () = get_used_modules();;
let read_ml_env () = get_used_modules();;

let restore_ml_env l =
    init_ml_env();
    do_listRL add_used_module l
;;

type ml_module_object =
    {mnames : string list};;

(* known_loaded_module contains (when updated) the names of the loaded
 * ML modules (linked or loaded with load_object).
 *)

let known_loaded_modules = (mls__mt() : string mls__t);;

let newly_loaded_modules = ref ([] : string list);;
let get_newly_loaded_modules () = !newly_loaded_modules;;
let loaded_modules = ref ([]:string list);;
let get_loaded_modules () = !loaded_modules;;
let orig_loaded_modules = ref (get_loaded_modules ());;

let update_loaded_modules () =
  do_list (fun m -> mls__add known_loaded_modules m)
          (get_newly_loaded_modules ());
  newly_loaded_modules := []
;;

let unloaded_modules modnames =
let updated = ref false
in filter (neg(fun s ->
               mls__memb known_loaded_modules s or
               (if (not !updated) then
                    (update_loaded_modules();
                     updated := true;
                     mls__memb known_loaded_modules s)
                else false)))
    modnames
;;


let load_object mname =
update_loaded_modules ();
if mls__memb known_loaded_modules mname
     then
      ()
     else
      try
        begin
        !l_o mname;
        newly_loaded_modules := mname :: (!newly_loaded_modules);
        loaded_modules := mname :: (!loaded_modules)
        end

     with
      (UserError ("mltop__no_load_object",_) as u) -> raise u
      |  _ -> anomalylabstrm "mltop__load_object"
                             [< 'S"Anomaly during loading of ml-object";
                                'SPC ; 'S mname >]
;;

let init_ml_modules () = loaded_modules := !orig_loaded_modules
;;

let cache_ml_module_object (_,{mnames=mnames}) =
let toload = unloaded_modules mnames in
    do_list
    (fun mname ->
         try (PP [< 'S"[Loading ML file " ; 'S mname ; 'S" ..." >];
             flush std_out;
             load_object mname;
             PPNL [< 'S"done]" >];
             flush std_out;
             mls__add known_loaded_modules mname)
        with
           (UserError ("mltop__no_load_object",_) as u) -> raise u
           | _ -> (PPNL [< 'S" failed]" >];
             error "load_ml_object"))
   toload
;;

let reintern_ml_module_object mo = mo;;

let specification_ml_module_object x = x;;

let (inMLModule,outMLModule) =
    declare_object("ML-MODULE",
                   {cache_function = cache_ml_module_object;
                    reintern_function = reintern_ml_module_object;
                    specification_function = specification_ml_module_object});;

let restore_ml_modules x =
  let toload = unloaded_modules x
  in
  do_list
    (fun mname ->
         try load_object mname;
             mls__add known_loaded_modules mname
        with
             (UserError("mltop__no_load_object",_) as u) ->
              himsg__WARN [< 'S"Can not restore ML object files"; 'SPC;
                           'S"whereas trying to restore" ; 'SPC ; 'S mname >]
            | _ -> (PPNL [< 'S" failed]"  >]; error "load_ml_object"))
   toload
;;

summary__declare_summary "ML-MODULES"
{summary__freeze_function = (fun () -> rev (get_loaded_modules()));
 summary__unfreeze_function = (fun x -> restore_ml_modules x);
 summary__init_function = (fun () -> init_ml_modules ());
 summary__sync_function = (fun () -> false)}
;;


type ml_import_object =
    {import_name : string};;

let cache_ml_import_object (_,{import_name=import_name}) =
    try add_used_module import_name
    with Not_found -> error ("Cannot import module "^import_name)
;;

let reintern_ml_import_object mo = mo;;

let specification_ml_import_object x = x;;

let (inMLImport,outMLImport) =
    declare_object("ML-IMPORT",
                   {cache_function = (fun x -> cache_ml_import_object x);
                    reintern_function = reintern_ml_import_object;
                    specification_function = specification_ml_import_object});;

summary__declare_summary "ML-IMPORTS"
{summary__freeze_function = (fun x -> freeze_ml_env x);
 summary__unfreeze_function = (fun x -> restore_ml_env x);
 summary__init_function = (fun x -> init_ml_env x);
 summary__sync_function = (fun () -> false)}
;;

let declare_ml_modules l =
    add_anonymous_object (inMLModule {mnames=l})
;;

let declare_ml_import s =
    add_anonymous_object (inMLImport{import_name=s})
;;

vinterp_add("AddMLPath",
    fun  [VARG_STRING s] ->
    (fun () ->
         add_ml_path (glob s)));;

let print_ml_path () =
    let l = get_ml_path () in
        PP [< 'S"ML Load Path: " ;
            HOV 0 (prlist_with_sep (fun () -> [< 'FNL >]) (fun s -> [< 'S s >]) l) >]
;;

vinterp_add("PrintMLPath",
    fun [] -> fun () -> print_ml_path ());;

let print_ml_modules () =
    let l = read_ml_env () in
        PP [< 'S"Open ML Modules : " ;
            HOV 0 (prlist_with_sep (fun () -> [< 'FNL >]) (fun s -> [< 'S s >]) l) >]
;;

vinterp_add("PrintMLModules",
    fun [] -> fun () -> print_ml_modules ());;

vinterp_add("DeclareMLModule",
    fun  l ->
    let sl = map (fun (VARG_STRING s) -> s) l in
    (fun () -> declare_ml_modules sl));;

vinterp_add("ImportMLModule",
    fun  [VARG_STRING s] ->
    (fun () -> declare_ml_import s));;

vinterp_add("RequireMLModule",
    fun  l ->
    let sl = map (fun (VARG_STRING s) -> s) l in
    (fun () -> (declare_ml_modules sl;
                do_list declare_ml_import sl)));;

let tacfun = ref (NONE : tactic option);;

let ml_tac s =
(try use_string ("mltop__tacfun := more_util__SOME(fun gls -> ("^s^") gls);;")
 with _ ->
     error "use_string"
     [< 'S"The string " ; 'S s ; 'S"failed to compile" >]);
match !tacfun with
    NONE ->
    anomalylabstrm "mltop__ml_tac"
    [< 'S"The caml-light compiler didn't put a piece of " ; 'SPC ;
       'S"compiled tactic code where it was supposed to be." >]
  | SOME tac -> tac
;;

let vernac_ml_tac = register_string_tactic "MLTACTIC" ml_tac;;


(* $Id: mltop.ml,v 1.4 1996/02/12 15:49:10 cpaulin Exp $ *)
