(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                somatch.ml                                *)
(****************************************************************************)
#open "std";;
#open "names";;
#open "vectops";;
#open "more_util";;
#open "pp";;
#open "stdpp";;
#open "generic";;
#open "term";;

(* Given a term with second-order variables in it, represented by Meta's,
   and possibly applied using XTRA[$SOAPP] to terms, this function will
   perform second-order, binding-preserving, matching, in the case where
   the pattern is a pattern in the sense of Dale Miller.

   ALGORITHM:

   Given a pattern, we decompose it, flattening Cast's and apply's,
   recursing on all operators, and pushing the name of the binder each
   time we descend a binder.

   When we reach a first-order variable, we ask that the corresponding
   term's free-rels all be higher than the depth of the current stack.

   When we reach a second-order application, we ask that the
   intersection of the free-rels of the term and the current stack be
   contained in the arguments of the application, and in that case, we
   construct a DLAM with the names on the stack.

 *)

let is_soapp_operator = function
    (DOPN(XTRA("$SOAPP",[]),cl)) -> true
  | (DOP2(XTRA("$SOAPP",[]),_,_)) -> true
  | _ -> false;;

let dest_soapp_operator = function
    (DOPN(XTRA("$SOAPP",[]),cl)) -> (hd_vect cl,list_of_tl_vect cl)
  | (DOP2(XTRA("$SOAPP",[]),c1,c2)) -> (c1,[c2])
  | _ -> failwith "dest_soapp_operator"
;;

let constrain ((n:int),(M:constr)) sigma =
    if listmap__in_dom sigma n then
        if eq_constr(M,listmap__map sigma n) then sigma
        else error "somatch"
    else listmap__add sigma (n,M)
;;

let build_dlam toabstract stk (M:constr) = buildrec M 1 stk
    where rec buildrec M = fun
    _ [] -> M
  | n (na::tl) -> (if listset__memb toabstract n then
                      buildrec (DLAM(na,M)) (n+1) tl
                   else buildrec (pop M) (n+1) tl)
;;

let memb_metavars = fun
    NONE _ -> true
  | (SOME mvs) n -> listset__memb mvs n
;;

let somatch metavars = sorec [] []
    where rec sorec stk sigma P T =
    let cP = whd_castapp P and
        cT = whd_castapp T in

    if is_soapp_operator cP then
        (match dest_soapp_operator cP with
         (DOP0(Meta n),l) ->
         if not(memb_metavars metavars n) then error "somatch" else
         let ok_args = map (fun (Rel i) -> i | _ -> error "somatch") l in
         let frels = free_rels cT in
             if listset__subset frels ok_args then
                 constrain (n,build_dlam ok_args stk cT) sigma
             else error "somatch"

       | _ -> error "somatch")
    else
    match (cP,cT) with

    (DOP0(Meta n),M) ->
    if not(memb_metavars metavars n) then
        (match M with
         DOP0(Meta m) -> if n=m then sigma else error "somatch"
       | _ -> error "somatch")
    else
    let depth = length stk in
        if for_all (fun i -> i > depth) (free_rels M) then
            constrain (n,lift (-depth) M) sigma
        else error "somatch"

  | (VAR v1,VAR v2) ->
    if v1 = v2 then sigma else error "somatch"

  | (Rel n1,Rel n2) ->
    if n1 = n2 then sigma else error "somatch"

  | (DOP0 op1,DOP0 op2) ->
    if op1 = op2 then sigma else error "somatch"

  | (DOP1(op1,c1),
     DOP1(op2,c2)) ->
    if op1 = op2 then sorec stk sigma c1 c2
    else error "somatch"
    
  | (DOP2(op1,c1,d1),
     DOP2(op2,c2,d2)) ->
    if op1 = op2 then
        sorec stk (sorec stk sigma c1 c2) d1 d2
    else error "somatch"

  | (DOPN(op1,cl1),
     DOPN(op2,cl2)) ->
    if op1 = op2 & vect_length cl1 = vect_length cl2 then
        it_vect2 (sorec stk) sigma cl1 cl2
    else error "somatch"

  | (DOPL(op1,cl1),
     DOPL(op2,cl2)) ->
    if op1 = op2 & length cl1 = length cl2 then
        it_list2 (sorec stk) sigma cl1 cl2
    else error "somatch"

  | (DLAM(_,c1),
     DLAM(na,c2)) -> sorec (na::stk) sigma c1 c2

  | (DLAMV(_,cl1),
     DLAMV(na,cl2)) ->
    if vect_length cl1 = vect_length cl2 then
        it_vect2 (sorec (na::stk)) sigma cl1 cl2
    else error "somatch"

  | _ -> error "somatch"
;;

(* $Id: somatch.ml,v 1.2 1996/02/12 15:49:18 cpaulin Exp $ *)
