(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                               tactics1.ml                                *)
(****************************************************************************)

#infix "o";;
#open "std";;
#open "initial";;
#open "names";;
#open "generic";;
#open "term";;
#open "more_util";;
#open "pp";;
#open "stdpp";;
#open "tacmach";;
#open "proof_trees";;
#open "proof";;
#open "clenv";;
#open "tactics";;
#open "pattern";;

#infix "ORELSE";;
#infix "NEXT";;
#infix "THEN";;
#infix "THENF";;
#infix "THENL";;
#infix "THENS";;

let error_first_n n l =
    try first_n n l
    with Failure s -> error s
;;

let error_lookup_sign id sign =
    try lookup_sign id sign
    with Invalid_argument s -> error ("Invalid argument: "^s)
;;

let THEN_i1 tac1 tac2 = THEN_i tac1 tac2 1;;
#infix "THEN_i1";;

let nth_clause n gl =
    if n = 0 then NONE
    else if n < 0 then
        let id = nth (ids_of_sign (pf_hyps gl)) (-n)
        in SOME id
    else let id = nth (ids_of_sign (pf_hyps gl)) n
         in SOME id
;;

let clause_type cls gl =
    match cls with
    NONE -> pf_concl gl
  | SOME id -> snd(lookup_sign id (pf_hyps gl));;

let h cls gls = clause_type cls gls;;

let matches gls N pat =
let M = get_pat pat in
let (wc,_) = StartWalk gls
in try (clenv__unify [] wc M N; true)
    with UserError _ -> false
;;

let dest_match gls N pat =
let M = get_pat pat in
let mvs = collect_metas M in
let (wc,_) = StartWalk gls in
let (mvb,_) = clenv__unify [] wc M N in
    map (listmap__map mvb) mvs
;;

let somatches N pat =
let M = sopattern__get_pat pat in
    try (somatch__somatch NONE M N; true)
    with UserError _ -> false
;;

let dest_somatch N pat =
let M = sopattern__get_pat pat in
let mvs = collect_metas M in
let mvb = somatch__somatch (SOME mvs) M N in
    map (listmap__map mvb) mvs
;;

let soinstance pat arglist =
let M = sopattern__get_pat pat in
let mvs = collect_metas M in
let mvb = combine(mvs,arglist) in
    sosub__soexecute (reduction__strong (reduction__whd_meta mvb) M)
;;

(* [OnCL clausefinder clausetac]
 * executes the clausefinder to find the clauses, and then executes the
   clausetac on the list so obtained.
 *)
let OnCL (cfind: goal sigma -> clause list) (cltac: clause list -> tactic) =
((function gl ->
let cll = cfind gl
in cltac cll gl) : tactic)
;;

(* OR-branch *)
let TryClauses tac = firstrec
    where rec firstrec = function
    [] -> FAILTAC
  | cls::tl -> tac cls ORELSE firstrec tl
;;

(* AND-branch *)
let OnClauses tac = onrec
    where rec onrec = function
    [] -> IDTAC
  | cls::tl -> tac cls THEN onrec tl
;;

let AllHyps gl = (map inSOME (ids_of_sign (pf_hyps gl)));;

let LastHyp gl = let (id,T) = hd_sign (pf_hyps gl)
                 in [(SOME id)];;

let NLastHyps n gl =
    let ids = error_first_n n (ids_of_sign (pf_hyps gl))
    in map inSOME ids
;;

let AllClauses gl = let ids = ids_of_sign(pf_hyps gl)
                    in  (NONE::(map inSOME ids))
;;

let ApplyToAHyp tac = OnCL AllHyps (TryClauses tac);;

let TryAllClauses tac = OnCL AllClauses (TryClauses tac);;

let OnAllClauses tac = OnCL AllClauses (OnClauses tac);;

let OnClause T cls gl = T cls gl;;

let OnAllClausesLR T = OnCL (rev o AllClauses) (OnClauses T);;

let OnLastHyp tac = OnCL LastHyp (OnClauses tac);;

let OnNthLastHyp n tac gls = tac (nth_clause n gls) gls;;

let OnNLastHyps n tac = OnCL (NLastHyps n) (OnClauses tac);;

let ClauseTacThen tac continuation =
    (fun cls -> tac cls THEN continuation)
;;

let If pred (tac1:tactic) tac2 gl =
    if pred gl then tac1 gl else tac2 gl
;;

let IfOnClause pred tac1 tac2 cls gl =
    if pred (cls,clause_type cls gl) then
        tac1 cls gl
    else tac2 cls gl
;;

let rec RepeatFor n tac gl =
    if n = 0 then IDTAC gl
    else if n < 0 then error "RepeatFor"
    else (tac THEN (RepeatFor (n-1) tac)) gl
;;


(* OR-branch *)
let TryL tac = firstrec
    where rec firstrec = function
    [] -> FAILTAC
  | cls::tl -> tac cls ORELSE firstrec tl
;;

(* AND-branch *)
let OnL tac = onrec
    where rec onrec = function
    [] -> IDTAC
  | cls::tl -> tac cls THEN onrec tl
;;

(* $Id: tactics1.ml,v 1.2 1996/02/12 15:49:24 cpaulin Exp $ *)
