(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                 caml.ml                                  *)
(****************************************************************************)

(**************************************************************************)
(*        production of camllight syntax out of an ML environment         *)
(**************************************************************************)

#open "pp_control";;
#open "stdpp";;
#open "pp";;
#open "std";;
#open "more_util";;
#open "system";;
#open "names";;
#open "vernacinterp";;
#open "ml_import";;
#open "mlterm";;

let open_par = function true -> [< 'S"(" >] | false -> [< >];;
let close_par = function true -> [< 'S")" >] | false -> [< >];;


(**************************************************************************)
(*          translation of type expressions and inductive types           *)
(**************************************************************************)

let print_typeid = function
    TYPEparam id -> [< 'S"'" ; 'S(string_of_id id) >]
  | TYPEname id  -> [< 'S(string_of_id id) >]
;;


(* uncurry_list : ('a -> std_pcmds) -> 'a list -> std_ppcmds
 * formats a list [x1;...;xn] in its uncurried form (x1,...,xn). *)

let uncurry_list f = function
    []  -> [< >]
  | [x] -> [< (f x) >]
  | xl  -> [< 'S"(" ;
      	      prlist_with_sep (fun () -> [< 'S", " >]) (fun x -> (f x)) xl ;
	      'S")"
      	    >]
;;

let uncurry_list2 f = function
    []  -> [< >]
  | [x] -> [< 'S" " ; (f x) >]
  | xl  -> [< 'S"(" ;
      	      prlist_with_sep (fun () -> [< 'S"," >]) (fun x -> (f x)) xl ;
	      'S")"
      	    >]
;;


(* caml_of_type : bool -> MLtype -> std_ppcmds
 * [caml_of_type b t] formats the type expression t.
 * b=true if we need parentheses around the result. *)

let rec caml_of_type paren_p = function
     TYarr(t1,t2) ->
       	[< open_par paren_p ;
	   caml_of_type true t1 ; 'S" -> " ;
           caml_of_type false t2 ;
	   close_par paren_p
        >]

  | TYapp cl ->
      	let n = length cl in
	if n=1 then 
	  caml_of_type false (hd cl)
        else
	  [< open_par paren_p ;
	     uncurry_list (fun c -> caml_of_type false c)
      	       	       	  (tl cl) ; 'S" " ;
             caml_of_type false (hd cl) ;
	     close_par paren_p
	  >]

  | TYvar tid ->
      	[< print_typeid tid >]

  | TYglob id ->
      	[< 'S(string_of_id id) >]
;;


(* caml_of_one_inductive : MLind -> std_ppcmds
 * [caml_of_one_inductive mi] formats the declaration of mutual
 * inductive mi. *)

let caml_of_one_inductive (pl,name,cl) =
  let fmt_constructor (id,l) =
    [< 'S(string_of_id id) ;
       match l with
         [] -> [< >] 
       | _  -> [< 'S" of " ;
      	       	  prlist_with_sep (fun () -> [< 'S" * " >]) 
      	       	       	          (fun s -> caml_of_type true s) l
	       >]
    >] in

  [< uncurry_list (fun id -> [< 'S"'" ; 'S(string_of_id id) >]) pl ;
     if pl=[] then [< >] else [< 'S" " >] ;
     'S(string_of_id name) ; 'S" " ;
     V 0 [< 'S"= " ;
            prlist_with_sep (fun () -> [< 'FNL ; 'S"| ">])
                            (fun c -> fmt_constructor c)
		            cl
         >]
  >]
;;

let caml_of_inductive il =
  [< 'S"type " ;
     prlist_with_sep 
      	  (fun () -> [< 'FNL ; 'S"and " >])
       	  (fun i -> caml_of_one_inductive i)
	  il ;
     'FNL
  >]
;;


(**************************************************************************)
(*                  production of caml syntax for terms                   *)
(**************************************************************************)

let collect_lambda = collect []
  where rec collect acc = function
     MLlam(id,t) -> collect (id::acc) t
   | x           -> acc,x
;;

let abst = function
    [] -> [< >]
  | l  -> [< 'S"fun " ;
             prlist_with_sep (fun  ()-> [< 'S" " >])
      	       	       	     (fun id -> [< 'S(string_of_id id) >]) l ;
             'S" -> " >]
;;

let pr_binding = function
    [] -> [< >]
  | l  -> [< 'S" " ; prlist_with_sep (fun () -> [< 'S" " >])
      	       	       	(fun id -> [< 'S(string_of_id id) >]) l >]
;;


(* pp_mlast : identifier list -> bool -> std_ppcmds list
 *	      -> MLast -> std_ppcmds
 * [pp_mlast idl b args t] formats the Ml term (t a1...am)
 * in the de Bruijn environment idl, where args=[a1...am].
 * b=true if we need parentheses around the result. *)

let rec pp_mlast idl paren_p args = 

  let apply st = match args with
     [] -> st
   | _  -> HOV 2 [< open_par paren_p ; st ; 'S" " ;
                    prlist_with_sep (fun () -> [< 'SPC >]) (fun s -> s) args ;
                    close_par paren_p
                 >] in

  function
    MLrel n ->
      	 apply [< 'S(string_of_id (nth idl n)) >]

  | MLapp (h, args') ->
      	 let stl = map (fun t -> pp_mlast idl true [] t) args' in
         pp_mlast idl paren_p (stl@args) h

  | MLlam _ as t ->
      	 let fl,t' = collect_lambda t in
	 let st = [< abst (rev fl) ; pp_mlast (fl@idl) false [] t' >] in
	 if args=[] then
           [< open_par paren_p ; st ; close_par paren_p >]
         else
           apply [< 'S"(" ; st ; 'S")" >]

  | MLglob id ->
      	 apply [< 'S(string_of_id id) >]
	
  | MLcons (_,id,args') ->
      	 if args'=[] then
	   [< 'S(string_of_id id) >]
	 else
	   [< open_par paren_p ; 'S(string_of_id id) ;
	      uncurry_list2 (fun t -> pp_mlast idl true [] t) args' ;
	      close_par paren_p
           >]

  | MLcase (t, pv) ->
      	 apply
      	 [< if args<>[] then [< 'S"(" >]  else open_par paren_p ;
      	    V 0 [< 'S"match " ; pp_mlast idl false [] t ; 'S" with" ;
		   'FNL ; 'S"  " ;
		   pp_mlpat idl pv >] ;
	    if args<>[] then [< 'S")" >] else close_par paren_p >]

  | MLfix f ->
      	 pp_mlfix idl paren_p f args

  | MLexn id ->
      	 [< open_par paren_p ; 'S"failwith \"" ; print_id id ; 'S"\"" ;
	    close_par paren_p >]

and pp_mlfix idl paren_p (j,in_p,fid,bl) args =
  let idl' = (rev fid)@idl in
  [< open_par paren_p ; V 0 [< 'S"let rec " ;
       prlist_with_sep
      	  (fun () -> [< 'FNL ; 'S"and " >])
	  (fun (fi,ti) -> pp_mlfunction idl' fi ti)
	  (combine (fid,bl)) ;
       'FNL ;
       if in_p then 
      	 HOV 2 [< 'S"in " ; 'S(string_of_id (nth fid (j+1))) ;
                  if args<>[] then
                    [< 'S" "; prlist_with_sep (fun () -> [<'S" ">])
                                 (fun s -> s) args >]
                  else [< >]
      	       >]
       else 
         [< >] >] ;
     close_par paren_p >]

and pp_mlfunction idl f t =
  let bl,t' = collect_lambda t in
  let is_function pv =
    let ktl = map_vect_list (fun (_,l,t0) -> (list_length l,t0)) pv in
    not(exists (fun (k,t0) -> occurs (k+1) t0) ktl) in

  match t' with 
    MLcase(MLrel 1,pv) ->
       if is_function pv then
	 [< 'S(string_of_id f) ; pr_binding (rev (tl bl)) ;
       	    'S" = function" ; 'FNL ;
	    V 0 [< 'S"  " ; pp_mlpat (bl@idl) pv >] >]
       else
         [< 'S(string_of_id f) ; pr_binding (rev bl) ; 'S" = match " ;
	    'S(string_of_id (hd bl)) ; 'S" with" ; 'FNL ;
	    V 0 [< 'S"  " ; pp_mlpat (bl@idl) pv >] >]
	   
  | _ -> [< 'S(string_of_id f) ; pr_binding (rev bl) ;
	    'S" =" ; 'FNL ; 'S"  " ;
	    HOV 2 (pp_mlast (bl@idl) false [] t') >]

and pp_mlpat idl pv =
  let pp_one_pat (name,ids,t) =
      let paren_p = match t with
                      MLlam _  -> true
                    | MLcase _ -> true
                    | _        -> false in

      HOV 2 [< 'S(string_of_id name) ;
      	       uncurry_list2 (fun id -> [< 'S(string_of_id id) >]) (rev ids) ;
	       'S" ->" ; 'SPC ; pp_mlast (ids@idl) paren_p [] t
            >]

  in [< prvect_with_sep (fun () -> [< 'FNL ; 'S"| " >])
                        (fun p -> pp_one_pat p)
	                pv >]
;;


(* caml_of_decl : MLdecl -> std_ppcmds
 * [caml_of_decl d] formats the declaration d. *)

let caml_of_decl = function
    DECLtype il -> caml_of_inductive il

  | DECLabbrev (id, idl, t) ->
          [< 'S"type" ; if idl<>[] then [< 'S" " >] else [< >] ;
	     uncurry_list (fun id -> [< 'S"'" ; 'S(string_of_id id) >]) idl ;
      	     'S" " ; 'S(string_of_id id) ; 'S" == " ;
	      caml_of_type false t >]

(***  Incorrect car id peut disparaitre au profit du nom lie dans dans f
alors qu'il apparaitrait dans la suite ...
Ne semble pas tres satisfaisant dans le cas de mutuellement inductifs...

| DECLglob (id, MLfix f) ->
          [<  HOV 2 (pp_mlfix [] false f []) ; 'FNL >]
****)
 
  | DECLglob (id, MLfix (n,b,[id'],l)) ->
          [<  HOV 2 (pp_mlfix [] false (n,false,[id],l) []) ; 'FNL >]
  
  | DECLglob (id, t) ->
      	  [< 'S"let " ; pp_mlfunction [] id t ; 'FNL >]       
;;


(**************************************************************************)
(*             translation of an environment into caml syntax.            *)
(**************************************************************************)

let caml_of_env env =
  prlist (fun d -> [< caml_of_decl d ; 'S";;" ; 'FNL ; 'FNL >]) env
;;


(* make_caml_file : MLdecl list -> string -> unit
 * [make_caml_file env s] formats the ML envirnoment env in caml syntax in
 * the file s.ml *)

let make_file env s =
  let chan = open_trapping_failure open_out s ".ml" in
  let fp = copy_fp std_fp in
  let gp = dflt_gp in
  
  with_output_to fp chan;
  fully_pp gp;
  PP_WITH gp fp (caml_of_env env);
  close_out chan
;;


(**************************************************************************)
(*            Write Caml File filename [ ident1 ... identn ].             *)
(**************************************************************************)

let write_caml_file file idl =
  let spl = map fwsp_of_id idl in
  let fwenv = fw_env__minimal_env spl in
  let mlenv = fwtoml__mlenv_of_fwenv (true,false) fwenv in
  let env = optimise__optimise idl mlenv in
  make_file env file
;;

(*** TODO: remove overwriting ***)
overwriting_vinterp_add("WRITECAMLFILE",
  fun [VARG_STRING file ; VARG_VARGLIST l] ->
      	      let idl = map (fun (VARG_IDENTIFIER id) -> id) l in
	      fun () -> write_caml_file file idl
    | _ -> anomaly "WRITECAMLFILE called with bad arguments.")
;;


(* $Id: caml.ml,v 1.7 1996/02/12 17:36:42 cpaulin Exp $ *)
