(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                fw_env.ml                                 *)
(****************************************************************************)

#open "stdpp";;
#open "lomsg";;
#open "std";;
#open "more_util";;
#open "names";;
#open "lib";;
#open "generic";;
#open "term";;
#open "environ";;
#open "termenv";;
#open "constants";;
#open "constrtypes";;
#open "ml_import";;

(**************************************************************************)
(*                   recuperer un environnement FW minimal                *)
(**************************************************************************)

(* collect : constr -> section_path list
 * [collect t] collects the section paths appearing in the term t. *)

let collect = collect_rec []
  where rec collect_rec acc = function
    DOP1(_,c)  			   -> collect_rec acc c
  | DOP2(_,c1,c2)                  -> let l1 = collect_rec acc c1 in
      	       	                      collect_rec l1 c2
  | DOPN(Const sp,_) 		   -> if mem sp acc then acc else sp::acc
  | DOPN(MutInd(sp,_),_) 	   -> if mem sp acc then acc else sp::acc
  | DOPN(MutConstruct((sp,_),_),_) -> if mem sp acc then acc else sp::acc
  | DOPN(_,cv) 			   -> it_vect collect_rec acc cv
  | DOPL(_,cl) 			   -> it_list collect_rec acc cl
  | DLAM(_,c) 			   -> collect_rec acc c
  | DLAMV(_,cv) 		   -> it_vect collect_rec acc cv
  | _ 				   -> acc
;;


(* List of the needed section paths, with corresponding operations. *)

let NSP = (hashtbl__new 17 : (section_path,unit) hashtbl__t);;
let reset_needed () = hashtbl__clear NSP;;
let add_needed_sp sp = hashtbl__add NSP sp ();;
let is_needed sp = try hashtbl__find NSP sp; true with Not_found -> false;;


(* The FW optimized environment (in reverse order). *)

let FWenv = ref ([] : section_path list);;
let reset_fwenv () = FWenv := [];;
let add_fwenv sp = FWenv := sp :: !FWenv;;


(* traverse_object : object -> unit
 * If the object is a constant or an inductive then look for its sp
 * and adds them to the needed list. *)

let traverse_object sp obj =
  match (libobject__object_tag obj) with
    "CONSTANT" ->
      	    let (cdecl,_,_) = outConstant obj in
      	    let cbody = try listmap__map cdecl FW
		          with Not_found -> anomaly "traverse_object: FW" in
      	    if cookable_constant sp then begin
      	      cook_constant sp;		  (* cook the constant if needed *)
	      (match cbody.CONSTBODY with
	         SOME {v=COOKED c} -> let spl = collect c in
       	         	       	      do_list add_needed_sp spl
	       | _ -> anomaly "traverse_object: should be cooked !")
	    end else
	      let spl = collect cbody.CONSTTYPE in
	      do_list add_needed_sp spl

  | "MUTUALINDUCTIVE" ->
      	    let (midecl,_) = outMutualInductive obj in
	    let mibody = try listmap__map midecl FW
		         with Not_found -> anomaly "traverse_object: FW" in
	    do_vect (fun mipacket -> let spl = collect mipacket.MINDLC in
	                             do_list add_needed_sp spl)
      	       	    mibody.MINDPACKETS
      	    
  | _ -> anomaly "leaf needed which is not a constant or an inductive"
;;


(* traverse_stack : library_segment -> unit
 * travels (recursively) in a stack and applies traverse_object on objects
 * when they are needed. *)

let rec traverse_stack = function
   (sp,ClosedDir(_,{module_p=true;_},_,ctxt))::stk ->
       traverse_stack (rev ctxt); traverse_stack stk

 | (sp,ClosedDir(_,_,_,ctxt))::stk ->
       traverse_stack stk

 | (sp,LEAF obj)::stk ->
       let sp' = fwsp_of sp in
       if (is_needed sp') & not(is_ml_import sp') then begin
          add_fwenv sp;
	  traverse_object sp' obj
       end;
       traverse_stack stk

 | _::stk -> traverse_stack stk

 | [] -> ()
;;


(* optimize : unit -> section_path list
 * [optimize ()] builds an optimized environment for ML programs *)

let optimize () =
  let initial_segment = contents_after NONE in
  reset_fwenv();
  traverse_stack initial_segment;
  !FWenv
;;


(* minimal_env : section_path list -> section_path list
 * [optimized_env spl] returns a minimal environment for the FW section
 * paths list spl. *)

let minimal_env spl =
  reset_needed();
  do_list (fun sp -> add_needed_sp sp) spl;
  optimize()
;;

(* $Id: fw_env.ml,v 1.2 1996/02/12 17:36:45 cpaulin Exp $ *)
