(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                fwtoml.ml                                 *)
(****************************************************************************)

#open "pp";;
#open "stdpp";;
#open "std";;
#open "vectops";;
#open "more_util";;
#open "system";;
#open "generic";;
#open "names";;
#open "term";;
#open "vartab";;
#open "constrtypes";;
#open "machops";;
#open "himsg";;
#open "mach";;
#open "reduction";;
#open "proof";;
#open "environ";;
#open "ml_import";;
#open "mlterm";;
#open "vernacinterp";;

(* Added for the contribution by J.-F. Monin to extract non 
   ML typable functions *) 

let LOOSE_EXTRACTION = ref false;;

let loose_extraction () = !LOOSE_EXTRACTION;;

let set_loose b = LOOSE_EXTRACTION:= b;;

overwriting_vinterp_add("LOOSEF", function
   [] -> (fun () -> set_loose false));;

overwriting_vinterp_add("LOOSET", function
   [] -> (fun () -> set_loose true));;


let cofix = ref false;;

let execute rc env t =
  try execute_rec rc env t
  with UserError("cannot execute Implicit",_) -> errorlabstrm
           "fwtoml__execute" [< 'S"Cannot extract terms with Implicit" >]
;;

(**************************************************************************)
(*                    informations on inductive types:                    *)
(*      we keep them in a table in order to get quickly constructor       *)
(*                           names and arities                            *)
(**************************************************************************)

type packet_info = 
  { PACKET_name    : identifier ;
    PACKET_ncons   : int ;
    PACKET_cnames  : identifier vect ;
    PACKET_arities : int vect }
;;

let mind_table =
  (hashtbl__new 17 : (section_path * int , packet_info) hashtbl__t)
;;

let store_mind_info key info =
  hashtbl__add mind_table key info
;;

let get_mind_info key =
  hashtbl__find mind_table key
;;


(**************************************************************************)
(*          translation of type expressions and inductive types           *)
(**************************************************************************)

(* mltype_of_constr : typeid list -> constr -> MLtype
 * [mltype_of_constr dbenv c] translates the type expression c into
 * an ML type expression, in a de Bruijn environment dbenv. *)

let rec mltype_of_constr dbenv c = match (strip_outer_cast c) with
    DOP2(Prod,c1,DLAM(name,c2)) ->
      	let avoid = map (function TYPEparam id -> id
      	       	       	       	| TYPEname id -> id) dbenv in
      	let id = caml_name_of avoid name in
	TYarr(mltype_of_constr dbenv c1,
              mltype_of_constr ((TYPEname id)::dbenv) c2)

  | DOPN(AppL,cv) ->
        TYapp(map (fun c -> mltype_of_constr dbenv c) (list_of_vect cv))

  | Rel n ->
      	TYvar(nth dbenv n)

  | DOPN(MutInd (sp,i),_) ->
      	if is_ml_import sp then
	  let mlid = find_ml_import (IMPORT_mind (sp,i)) in
	  TYglob(id_of_mlid mlid)
	else
	  let info = get_mind_info (sp,i) in
      	  TYglob (info.PACKET_name)

  | DOPN(Const sp,_) ->
      	if is_ml_import sp then
	  let mlid = find_ml_import (IMPORT_const sp) in
	  TYglob(id_of_mlid mlid)
	else
          TYglob(basename sp)

(* added for a contribution of Lannion *)

  | DOP0(Sort (Prop Pos)) -> if loose_extraction () then
     (warning "extracting a too polymorphic function for the ML type-checker"; 
          TYglob(basename (path_of_string "#Datatypes#unit.fw")))
     else failwith "not an ML type"


  | _ -> failwith "not an ML type"
;;

let rec isSet = function
    DOP0(Sort(Prop(Pos))) -> true
  | DOP2(Cast,c,_)        -> isSet c
  | _			  -> false
;;

let abbrev_of_constr c =
  let rec get_params acc = function
    DOP2(Lambda, typ, DLAM(name,t)) ->
      if isSet typ then 
      	 let id = caml_name_of acc name in
      	 get_params (id::acc) t
      else
      	 failwith "not an ML type"
  | t -> acc,t  in

  let idl,t = get_params [] c in
  let env = map (fun id -> TYPEparam id) idl in
  rev idl, mltype_of_constr env t
;;

(* decl_of_inductive : section_path -> mutual_inductive_body -> MLdecl
 * [decl_of_inductive sp cb] translates an inductive definition into
 * an ML type declaration. *)

let rec type_of_c idl env acc = function
      DOP2(Cast,t,_) ->
      	    type_of_c idl env acc t

    | DOP2(Prod,t1,DLAM(name,t2)) -> 
      	    let j = execute mt_evc env t1 in
	    let c = cast_of_judgement j in
	    let env' = add_rel (name,c) env in
	    let avoid = map (function TYPEparam id -> id
      	         	       	    | TYPEname id -> id) idl in
   	    let idl' = (TYPEname (caml_name_of avoid name))::idl in

      	    if istype j.TYP then
	      type_of_c idl' env' acc t2
	    else
	      type_of_c idl' env' ((mltype_of_constr idl t1)::acc) t2

    | _ ->  rev acc
;;

let decl_of_one_inductive (sp,i) env np packet =
  let name = packet.MINDTYPENAME in

  let rec params_names_and_env t acc env' = function
     0 -> acc,env'
   | n -> (match (strip_outer_cast t) with
             DOP2(Prod,T,DLAM(nm,t')) -> 
	         let cj = cast_of_judgement (execute mt_evc env' T) in
      	       	 let id = caml_name_of acc nm in
      	       	 params_names_and_env t' (id::acc) (add_rel (nm,cj) env') (n-1)
	   | _ -> anomaly "decl_of_one_inductive: should be a product!") in

  let rec remove_params t = function
     0 -> t
   | n -> (match (strip_outer_cast t) with
             DOP2(Prod,_,DLAM(_,t')) -> remove_params t' (n-1)
	   | _ -> anomaly "decl_of_one_inductive: should be a product!") in

  let namel,consv = decomp_all_DLAMV_name packet.MINDLC in
  let nbcons = vect_length consv in
  let arl = map_vect (fun c -> remove_params c np) consv in
  let pl,env' = params_names_and_env consv.(0) [] env np in
  let idl = (map (fun id -> (TYPEparam id)) pl)
      	   @(map (function n -> (TYPEname (caml_name_of [] n))) namel) in

  let info = { PACKET_name    = name ;
      	       PACKET_ncons   = nbcons ;
      	       PACKET_cnames  = packet.MINDCONSNAMES ;
	       PACKET_arities = make_vect nbcons 0 } in

  let fmt_constructor j =
    let l = type_of_c idl env' [] arl.(j) in
    info.PACKET_arities.(j) <- length l ;
    (packet.MINDCONSNAMES.(j), l)
  in

  store_mind_info (sp,i) info;
  (rev pl, name, map (fun i -> fmt_constructor i) (list0n (nbcons-1)))
;;

let decl_of_inductive sp mib =
  let nparams = mib.MINDNPARAMS in
  let env = it_vect (fun e p -> add_rel (Name p.MINDTYPENAME, p.MINDARITY) e)
      	       	    (GLOB(initial_sign()))
      	       	    mib.MINDPACKETS in
  let ipl = combine ((list0n (mib.MINDNTYPES - 1)),
      	       	     (list_of_vect mib.MINDPACKETS)) in

  DECLtype(
   map (fun (i,packet) -> decl_of_one_inductive (sp,i) env nparams packet) ipl)
;;


(**************************************************************************)
(*    translation of constants : type abbreviations and declarations      *)
(**************************************************************************)

let collect_n_lambda = collect []
  where rec collect acc = fun
    0 t              -> acc,0,t
  | n (MLlam(id,t')) -> collect (id::acc) (n-1) t'
  | n t              -> acc,n,t
;;

(* mlast_of_constr : identifier list -> constr assumptions -> constr -> MLast
 * [mlast_of_constr idl env c] give an MLast for the term c in the
 * environment env. idl is the list of the identifiers of env. *)

let rec mlast_of_constr idl env args =
  let apply t = match args with
     [] -> t
   | _  -> MLapp(t,args) in

  function c -> match strip_outer_cast c with
    DOPN(Fix _,_) as t ->
      	 apply (mlast_of_fix idl env true t)

  | DOPN(CoFix _,_) as t ->
      	 if !cofix then
      	   apply (mlast_of_fix idl env true t)
	 else
	   failwith "cofix"

  | DOPN(MutCase,cv) ->
      	 let cv' = tl_vect cv in
	 let c = cv'.(0) and tv = tl_vect cv' in
	 let j = execute mt_evc env c in
	 let typ = whd_betadeltaiota mt_evc j.TYP in
	 let mi = match tactics__head_constr typ with
	                DOPN(MutInd k,_) -> k
		      | _ -> anomaly "mlast_of_constr: bad type in case" in
	 let h = mlast_of_constr idl env [] c in
	 let args = map_vect (fun c -> mlast_of_constr idl env [] c) tv in
	 let pl = mlpat_of_constr idl mi args in
	 apply (MLcase(h, pl))

  | DOPN(MutConstruct((sp,i),j),_) ->
      	 let info = get_mind_info (sp,i) in
	 let cname =
      	       if is_ml_import sp then
      	       	 id_of_mlid (find_ml_import (IMPORT_construct ((sp,i),j)))
      	       else
      	       	 info.PACKET_cnames.(j-1) in
	 let n = info.PACKET_arities.(j-1) in
	 let d = n - (length args) in
	 if d>0 then
	   let vl = new_vars idl d in
	   let rels = map (fun i -> MLrel i) (rev (tl (list0n d))) in
	   let args' = map (fun t -> ml_lift d t) args in
	   let argl = args'@rels in
      	   it_list (fun t id -> MLlam(id,t)) (MLcons(j, cname, argl)) vl
         else
	   let (l,m) = chop_list n args in
	   if m=[] then MLcons(j, cname, l)
	           else MLapp(MLcons(j, cname, l), m)

  | DOPN(Const sp,_) ->
      	 let id = if is_ml_import sp then
	            id_of_mlid (find_ml_import (IMPORT_const sp))
		  else
		    basename sp in
      	 apply (MLglob id)

  | DOPN(AppL,cv) ->
      	 if vect_length cv = 0 then
	   anomaly "mlast_of_constr: application without argument!"
	 else
	   let args' = it_vect 
      	      (fun l c -> let j = execute mt_evc env c in
	                  if isprop j.TYP then
      	       	       	    l
      	       	       	  else
      	       	       	    (mlast_of_constr idl env [] c)::l)
              [] (tl_vect cv) in
	   mlast_of_constr idl env ((rev args')@args) cv.(0) 
  
  | DOP2(Lambda,T,DLAM(name,c')) ->
      	 let id = caml_name_of idl name in
	 let j = execute mt_evc env T in
	 let cT = cast_of_judgement j in
	 if istype j.TYP then
	   (* apply (mlast_of_constr (id::idl) (add_rel (name,cT) env) [] c')*)
	   let t = mlast_of_constr (id::idl) (add_rel (name,cT) env) [] c' in
	   apply (ml_pop t)
	 else
	   apply (MLlam (id,
      	       	    mlast_of_constr (id::idl) (add_rel (name,cT) env) [] c'))

  | Rel n ->
      	 apply (MLrel n)

  | _ -> anomaly "mlast_of_constr: called with bad args"

and mlast_of_fix idl env in_p c = match strip_outer_cast c with
   (DOPN(Fix _,_) | DOPN(CoFix _,_)) as t ->
      	  let (j,cv) = match t with
      	    DOPN(Fix (_,j),cv) -> j,cv
	  | DOPN(CoFix j,cv)   -> j,cv in
      	  let k = (vect_length cv)-1 in
	  let fiv,tiv = decomp_all_DLAMV_name (cv.(k)) in
	  let Aiv = sub_vect cv 0 k in
      	  let cAl = map (fun A ->
      	     assumption_of_judgement mt_evc (execute mt_evc env A))
	     (list_of_vect Aiv) in
	  let env',_ = it_list2 (fun (acc,i) name cA ->
	     (add_rel (name,lift i cA) acc),(succ i)) (env,0) (rev fiv) cAl in
	  let fid = it_list (fun l name -> let id = caml_name_of idl name in
	     id::l) [] fiv in
	  let idl' = (rev fid)@idl in
      	  let tv = map (fun c -> mlast_of_constr idl' env' [] c)
      	       	       (list_of_vect tiv) in

      	  MLfix (j, in_p, fid, tv)

 | _ -> anomaly "mlast_of_fix: called with bad args"

and mlpat_of_constr idl mi cl =
  let info = get_mind_info mi in
  let one_pat j =
    let c = cl.(j) in
    let name = try  id_of_mlid (find_ml_import (IMPORT_construct (mi,j+1)))
      	       with Not_found -> info.PACKET_cnames.(j) in
    let n = info.PACKET_arities.(j) in
    let ids,r,t = collect_n_lambda n c in
    if r=0 then
      (name, ids, t)
    else
      let ids' = new_vars (ids@idl) r in
      let rels = map (fun i -> MLrel i) (rev (tl (list0n r))) in
      (name, ids'@ids, MLapp(ml_lift r t, rels))

  in map_vect (fun i -> one_pat i)
      	      (vect_of_list (list0n (info.PACKET_ncons - 1)))
;;

let rec is_rec = function
    DOPN(Fix _,_)  -> true
  | DOP2(Cast,c,_) -> is_rec c
  | _		   -> false
;;

(* decl_of_constant : section_path -> constant_body -> MLdecl
 * [decl_of_constant sp cb] returns the ML declaration for constant cb
 * of section path sp (OBJ). *)

let axiom_type_msg id =
  errorlabstrm "decl_of_constant"
  [< 'S"You must link axiom " ; print_id id ; 'S" first !" >]
;;

let not_an_ML_type_msg id =
  errorlabstrm "decl_of_constant"
  [< 'S"Constant " ; print_id id ; 'S" does not correspond to an ML type" >]
;;

let decl_of_constant axioms sp cb =
  let id = basename sp in
  let typ = cb.CONSTTYPE in
  let j = execute mt_evc (GLOB(initial_fsign())) typ in
  if istype j.TYP then
    let c = match cb.CONSTBODY with
       	      SOME {v=COOKED c} -> c
	    | _ 		-> axiom_type_msg id in
    let idl,t = try abbrev_of_constr c
      	       	with (Failure "not an ML type") -> not_an_ML_type_msg id in
    DECLabbrev(id, idl, t)
  else
    let b = cb.CONSTBODY in match b with
      SOME {v=COOKED c} ->
        let t = if is_rec c then
	          mlast_of_fix [] (GLOB(initial_sign())) false c
      	        else
	          mlast_of_constr [] (GLOB(initial_sign())) [] c
	in DECLglob(id, t)	  

    | _ -> (MSGNL [< 'S"The axiom " ; print_id id ;
                     'S" is translated into an exception." >] ;
	    DECLglob(id, MLexn id) )
;;


(**************************************************************************)
(*           translation of an environment into ML declarations           *)
(*            traverse the environment and distinguish between            *)
(*                  constants and inductive definitions                   *)
(**************************************************************************)

(* decl_of_sp : bool -> section_path -> MLdecl
 * [decl_of_env ax spl] returns the declaration corresponding to
 * sp (OBJ). ax=true if axioms are allowed. *)

let decl_of_sp axioms sp =
  let lib__LEAF obj = lib__map sp in
  match libobject__object_tag obj with
    "CONSTANT" ->
      	  let (cdecl,_,_) = outConstant obj in
      	  (try decl_of_constant axioms sp (listmap__map cdecl FW)
	   with Failure "cofix" -> errorlabstrm "decl_of_sp"
	     [< 'S"Co-recursive definition in " ;
      	        'S(string_of_id (basename sp)) ; 'S": extract into Gofer !" >])

  | "MUTUALINDUCTIVE" ->
      	  let (midecl,_) = outMutualInductive obj in
      	  let fsp = fwsp_of sp in
      	  decl_of_inductive fsp (listmap__map midecl FW)

  | _  -> anomaly "caml_of_sp: not a constant or an inductive"
;;

(*****
let rec rename_globals renl = function
    DECLtype(pl,id,cl)   -> let id' = caml_name_of id [] in
      	       	       	    if id' <> id then
  | DECLabbrev(id,idl,t) -> 
  | DECLglob(id,t)       -> 
  | [] -> []
;;
******)

let mlenv_of_fwenv (axioms,b) env =
  reset_caml_names () ;
  cofix := b;
  do_list (fun sp -> decl_of_sp axioms (objsp_of sp)) (rev !ML_INDUCTIVES) ;
  map (fun sp -> decl_of_sp axioms sp) env  
;;

(* $Id: fwtoml.ml,v 1.7 1996/02/12 17:36:47 cpaulin Exp $ *)
