(****************************************************************************)
(*                 The Calculus of InductiveConstructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                 gofer.ml                                 *)
(****************************************************************************)

(**************************************************************************)
(*        production of gofer syntax out of an ML environment             *)
(**************************************************************************)

#open "pp_control";;
#open "stdpp";;
#open "pp";;
#open "std";;
#open "more_util";;
#open "system";;
#open "names";;
#open "vernacinterp";;
#open "ml_import";;
#open "mlterm";;
#open "caml";;

let open_par = function true -> [< 'S"(" >] | false -> [< >];;
let close_par = function true -> [< 'S")" >] | false -> [< >];;

let upper c =
  let i = int_of_char c in (i>64) & (i<91)
;;

let lower c = 
  let i = int_of_char c in (i>96) & (i<123)
;;

let gofer_of_id id =
  let s = string_of_id id in
  let c = nth_char s 0 in
  if upper c then "ren"^s else s
;;

let Gofer_of_id id =
  let s = string_of_id id in
  let c = nth_char s 0 in
  if lower c then "Ren"^s else s
;;


(**************************************************************************)
(*          translation of type expressions and inductive types           *)
(**************************************************************************)

let print_typeid = function
    TYPEname id  -> [< 'S(Gofer_of_id id) >]
  | TYPEparam id -> [< 'S(gofer_of_id id) >]
;;

(* gofer_of_type : bool -> MLtype -> std_ppcmds
 * [gofer_of_type b t] formats the type expression t.
 * b=true if we need parentheses around the result. *)

let rec gofer_of_type paren_p = function
     TYarr(t1,t2) ->
       	[< open_par paren_p ;
	   gofer_of_type true t1 ; 'S" -> " ;
           gofer_of_type false t2 ;
	   close_par paren_p
        >]

  | TYapp cl ->
      	let n = length cl in
	if n=1 then 
	  gofer_of_type false (hd cl)
        else
	  [< open_par paren_p ;
             gofer_of_type false (hd cl) ; 'S" " ;
	     prlist_with_sep (fun () -> [< 'S" " >])
      	       	       	     (fun c -> gofer_of_type true c) (tl cl) ;
	     close_par paren_p
	  >]

  | TYvar tid ->
      	[< print_typeid tid >]

  | TYglob id ->
      	[< 'S(Gofer_of_id id) >]
;;


(* gofer_of_one_inductive : MLind -> std_ppcmds
 * [gofer_of_one_inductive mi] formats the declaration of mutual
 * inductive mi. *)

let gofer_of_one_inductive (pl,name,cl) =
  let fmt_constructor (id,l) =
    [< 'S(Gofer_of_id id) ;
       match l with
         [] -> [< >] 
       | _  -> [< 'S" " ;
      	       	  prlist_with_sep (fun () -> [< 'S" " >]) 
      	       	       	          (fun s -> gofer_of_type true s) l
	       >]
    >] in

  [< 'S(Gofer_of_id name) ; 'S" " ;
     prlist_with_sep (fun () -> [< 'S" " >])
      	       	     (fun id -> [< 'S(gofer_of_id id) >]) pl ;
     if pl=[] then [< >] else [< 'S" " >] ;
     
     V 0 [< 'S"= " ;
            prlist_with_sep (fun () -> [< 'FNL ; 'S"| ">])
                            (fun c -> fmt_constructor c)
		            cl
         >]
  >]
;;

let gofer_of_inductive il =
  [< 'S"data " ;
     prlist_with_sep 
      	  (fun () -> [< 'FNL >])
       	  (fun i -> gofer_of_one_inductive i)
	  il ;
     'FNL
  >]
;;


(**************************************************************************)
(*                  production of gofer syntax for terms                   *)
(**************************************************************************)

let abst = function
    [] -> [< >]
  | l  -> [< 'S"\\" ;
             prlist_with_sep (fun  ()-> [< 'S" " >])
      	       	       	     (fun id -> [< 'S(gofer_of_id id) >]) l ;
             'S" -> " >]
;;

let pr_binding = function
    [] -> [< >]
  | l  -> [< 'S" " ; prlist_with_sep (fun () -> [< 'S" " >])
      	       	       	(fun id -> [< 'S(gofer_of_id id) >]) l >]
;;

(* pp_mlast : identifier list -> bool -> std_ppcmds list
 *	      -> MLast -> std_ppcmds
 * [pp_mlast idl b args t] formats the Ml term (t a1...am)
 * in the de Bruijn environment idl, where args=[a1...am].
 * b=true if we need parentheses around the result. *)

let rec pp_mlast idl paren_p args = 

  let apply st = match args with
     [] -> st
   | _  -> HOV 2 [< open_par paren_p ; st ; 'S" " ;
                    prlist_with_sep (fun () -> [< 'SPC >]) (fun s -> s) args ;
                    close_par paren_p
                 >] in

  function
    MLrel n ->
      	 apply [< 'S(gofer_of_id (nth idl n)) >]

  | MLapp (h, args') ->
      	 let stl = map (fun t -> pp_mlast idl true [] t) args' in
         pp_mlast idl paren_p (stl@args) h

  | MLlam _ as t ->
      	 let fl,t' = collect_lambda t in
	 let st = [< abst (rev fl) ; pp_mlast (fl@idl) false [] t' >] in
	 if args=[] then
           [< open_par paren_p ; st ; close_par paren_p >]
         else
           apply [< 'S"(" ; st ; 'S")" >]

  | MLglob id ->
      	 apply [< 'S(gofer_of_id id) >]
	
  | MLcons (_,id,args') ->
      	 if args'=[] then
	   [< 'S(Gofer_of_id id) >]
	 else
	   [< open_par paren_p ; 'S(Gofer_of_id id) ; 'S" " ;
	      prlist_with_sep (fun () -> [< 'S" " >])
	                      (fun t -> pp_mlast idl true [] t) args' ;
	      close_par paren_p
           >]

  | MLcase (t, pv) ->
      	 apply
      	 [< if args<>[] then [< 'S"(" >]  else open_par paren_p ;
      	    V 0 [< 'S"case " ; pp_mlast idl false [] t ; 'S" of" ;
		   'FNL ; 'S"  " ;
		   pp_mlpat idl pv >] ;
	    if args<>[] then [< 'S")" >] else close_par paren_p >]

  | MLfix f ->
      	 pp_mlfix idl paren_p f args

  | MLexn id ->
      	 [< open_par paren_p ; 'S"error \"" ; print_id id ; 'S"\"" ;
	    close_par paren_p >]

and pp_mlfix idl paren_p (j,in_p,fid,bl) args =
  let idl' = fid@idl in
  [< open_par paren_p ; V 0 [< 'S"let { " ;
       prlist_with_sep
      	  (fun () -> [< 'S";" ; 'FNL >])
	  (fun (fi,ti) -> pp_mlfunction idl' fi ti)
	  (combine (fid,bl)) ;
       'S" }" ; 'FNL ;
       if in_p then 
      	 HOV 2 [< 'S"in " ; 'S(gofer_of_id (nth fid (j+1))) ;
                  if args<>[] then
                    [< 'S" "; prlist_with_sep (fun () -> [<'S" ">])
                                 (fun s -> s) args >]
                  else [< >]
      	       >]
       else 
         [< >] >] ;
     close_par paren_p >]

and pp_mlfunction idl f t =
  let bl,t' = collect_lambda t in

 (***TO REMOVE
  let is_function pv =
    let ktl = map_vect_list (fun (_,l,t0) -> (list_length l,t0)) pv in
    not(exists (fun (k,t0) -> occurs (k+1) t0) ktl) in

  match t' with 
    MLcase(MLrel 1,pv) ->
       if is_function pv then
	 [< 'S(gofer_of_id f) ; pr_binding (rev (tl bl)) ;
       	    'S" = \\" ; 'FNL ;
	    V 0 [< 'S"  " ; pp_mlpat (bl@idl) pv >] >]
       else
         [< 'S(gofer_of_id f) ; pr_binding (rev bl) ; 'S" = case " ;
	    'S(gofer_of_id (hd bl)) ; 'S" of" ; 'FNL ;
	    V 0 [< 'S"  " ; pp_mlpat (bl@idl) pv >] >]
	   
  | _ -> ***)

  [< 'S(gofer_of_id f) ; pr_binding (rev bl) ;
     'S" =" ; 'FNL ; 'S"  " ;
     HOV 2 (pp_mlast (bl@idl) false [] t') >]

and pp_mlpat idl pv =
  let pp_one_pat (name,ids,t) =
      let paren_p = match t with
                      MLlam _  -> true
                    | MLcase _ -> true
                    | _        -> false in

      HOV 2 [< 'S(Gofer_of_id name) ;
      	       if ids=[] then [< >]
      	       else [< 'S" " ; prlist_with_sep (fun () -> [<'S" ">])
      	       	       (fun id -> [< 'S(gofer_of_id id) >]) (rev ids) >] ;
	       'S" ->" ; 'SPC ; pp_mlast (ids@idl) paren_p [] t
            >]

  in [< prvect_with_sep (fun () -> [< 'FNL ; 'S"  " >])
                        (fun p -> pp_one_pat p)
	                pv >]
;;


(* gofer_of_decl : MLdecl -> std_ppcmds
 * [gofer_of_decl d] formats the declaration d. *)

let gofer_of_decl = function
    DECLtype il -> gofer_of_inductive il

  | DECLabbrev (id, idl, t) ->
          [< 'S"type " ; 'S(Gofer_of_id id) ; 'S" " ;
	     prlist_with_sep (fun () -> [< 'S" " >])
	                     (fun id -> [< 'S(gofer_of_id id) >]) idl;
      	     if idl<>[] then [< 'S" " >] else [< >] ;'S"= " ;
	     gofer_of_type false t ; 'FNL >]

  | DECLglob (id , MLfix(_,_,idl,fl)) ->
      	  prlist_with_sep (fun () -> [< 'FNL >])
	  (fun (id,f) -> pp_mlfunction (rev idl) id f)
	  (combine (idl,fl))

  | DECLglob (id, t) ->
      	  [< pp_mlfunction [] id t ; 'FNL >]       
;;


(**************************************************************************)
(*             translation of an environment into gofer syntax.            *)
(**************************************************************************)

let gofer_of_env env =
  prlist (fun d -> [< gofer_of_decl d ; 'FNL >]) env
;;


(* make_gofer_file : MLdecl list -> string -> unit
 * [make_gofer_file env s] formats the ML envirnoment env in gofer syntax
 * in the file s.ml *)

let make_file env s =
  let chan = open_trapping_failure open_out s ".gs" in
  let fp = copy_fp std_fp in
  let gp = dflt_gp in
  
  with_output_to fp chan;
  fully_pp gp;
  PP_WITH gp fp (gofer_of_env env);
  close_out chan
;;


(**************************************************************************)
(*            Write Gofer File filename [ ident1 ... identn ].            *)
(**************************************************************************)

let write_gofer_file file idl =
  let spl = map fwsp_of_id idl in
  let fwenv = fw_env__minimal_env spl in
  let mlenv = fwtoml__mlenv_of_fwenv (true,true) fwenv in
  make_file mlenv file
;;

(*** TODO: remove overwriting ***)
overwriting_vinterp_add("WRITEGOFERFILE",
  fun [VARG_STRING file ; VARG_VARGLIST l] ->
      	      let idl = map (fun (VARG_IDENTIFIER id) -> id) l in
	      fun () -> write_gofer_file file idl
    | _ -> anomaly "WRITEGOFERFILE called with bad arguments.")
;;



(* $Id: gofer.ml,v 1.4 1996/02/12 17:36:49 cpaulin Exp $ *)
