(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                               ml_import.ml                               *)
(****************************************************************************)

#open "pp";;
#open "stdpp";;
#open "std";;
#open "more_util";;
#open "summary";;
#open "libobject";;
#open "names";;
#open "reintern0";;
#open "generic";;
#open "library";;
#open "libobject";;
#open "term";;
#open "constrtypes";;
#open "vartab";;
#open "impuniv";;
#open "environ";;
#open "indtypes";;
#open "himsg";;
#open "reduction";;
#open "dischcore";;
#open "proof";;
#open "mach";;
#open "declare";;
#open "command";;
#open "constants";;
#open "pretty";;
#open "trad";;
#open "vernacinterp";;
#open "vernacentries";;


(* TODO : move this function to another file, like more_util. *)

let list0n = gen []
  where rec gen acc = function
    -1 -> acc
  | n  -> gen (n::acc) (n-1)
;;


(**************************************************************************)
(*           Importing ML objects, and linking axioms to terms.           *)
(**************************************************************************)

(* ML names. *)

type ml_identifier == string option * identifier;;

let mlid_of_id id = (NONE, id)
;;

let id_of_mlid = function
    NONE  ,id -> id
  | SOME f,id -> id_of_string (f ^ (string_of_id id))
;;

let rec next_global_ident id =
  if not(is_global id) then id
  else next_global_ident (lift_ident id)
;;


(* A table to keep the ML imports. *)

type ml_import = IMPORT_const     of section_path
      	       | IMPORT_mind      of section_path * int
	       | IMPORT_construct of (section_path * int) * int
;;

let ml_import_tab =
  (hashtbl__new 17 : (ml_import, ml_identifier) hashtbl__t)
;;

let ML_INDUCTIVES = ref ([] : section_path list);;

let add_ml_inductive_import sp =
  if not(mem sp !ML_INDUCTIVES) then
    ML_INDUCTIVES := sp :: !ML_INDUCTIVES
;;

let add_ml_import imp qid =
  hashtbl__add ml_import_tab imp qid ;
  match imp with
    IMPORT_mind (sp,_) -> add_ml_inductive_import sp
  | _ 		       -> ()
;;

let find_ml_import imp =
  hashtbl__find ml_import_tab imp
;;

let is_ml_import sp =
  try find_ml_import (IMPORT_const sp) ; true
  with Not_found ->
    (try find_ml_import (IMPORT_mind (sp,0)) ; true
     with Not_found ->
       (try find_ml_import (IMPORT_construct ((sp,0),0)) ; true
        with Not_found -> false))
;;

let reintern_ml_import = function
    (IMPORT_const sp),(s,id)    -> (IMPORT_const (ri_sp sp)),(s,ri_id id)
  | (IMPORT_mind (sp,i)),(s,id) -> (IMPORT_mind (ri_sp sp,i)),(s,ri_id id)
  | (IMPORT_construct ((sp,i),j)),(s,id)
      	       	       -> (IMPORT_construct ((ri_sp sp,i),j)),(s,ri_id id)
;;

let inMLImport,outMLImport =
  declare_object ("MLIMPORT",
     { cache_function         = (fun (_,(imp,qid)) -> add_ml_import imp qid) ;
       reintern_function      = reintern_ml_import ;
       specification_function = (fun x -> x) })
;;

let freeze () =
  (hashtbl__freeze ml_import_tab, !ML_INDUCTIVES)
;;

let unfreeze (ft,stk) =
  hashtbl__unfreeze ft ml_import_tab ;
  ML_INDUCTIVES := stk
;;

declare_summary "MLIMPORT-TABLE"
  { freeze_function   = freeze ;
    unfreeze_function = unfreeze ;
    init_function     = (fun () -> hashtbl__clear ml_import_tab ;
      	       	       	       	   ML_INDUCTIVES := []) ;
    sync_function     = (fun () -> false) }
;;


(* Replace CCI section_path with FW section_path in a term. *)

let whd_fwify = function
    DOPN(Const sp,cl)               -> DOPN(Const (fwsp_of sp),cl)
  | DOPN(MutInd(sp,i),cl)           -> DOPN(MutInd (fwsp_of sp,i),cl)
  | DOPN(MutConstruct((sp,j),i),cl) -> DOPN(MutConstruct((fwsp_of sp,j),i),cl)
  | x                               -> x
;;
 
let fwify = strong whd_fwify;;


let fwsp_of_id id =
  try nametab__fw_sp_of_id id
  with Not_found -> errorlabstrm "fwsp_of_id"
	  [< 'S(string_of_id id) ; 'S" is not a defined informative object" >]
;;


(**************************************************************************)
(*                      ML Import file__name : type.                      *)
(**************************************************************************)

(***** constants__make_parameter_body *****)
let make_fw_parameter_body (hyps,j) =
  match hnf_constr mt_evc j.TYP with
    DOP0(Sort(Prop Pos)) | DOP0(Sort(Type(Pos,_))) ->
      let typ = j_val j in
      {CONSTKIND = FW;
       CONSTHYPS=hyps;
       CONSTBODY=NONE;
       CONSTEVAL = NONE;
       CONSTOPAQUE = true;
       CONSTTYPE=(match typ with
                    DOP2(Cast,c,t) -> DOP2(Cast,c,hnf_constr mt_evc t)
                  | _ -> typ)}
  | _ -> errorlabstrm "make_fw_parameter_body" 
         [< pFTERM (j_val j); 'S " is not an informative term." >]
;;

(***** constants__infexecute_parameter *****)
let fw_infexecute_parameter fhyps T =
  let T = instantiate_universes T in
  let j = execute_rec mt_evc (GLOB fhyps) T in
  let ids = auto_save_variables() in
  let fhyps' = thin_to_level fhyps j.TYP in
  [FW,make_fw_parameter_body (fhyps',j)]
;;

(***** declare__machine_parameter *****)
let fw_machine_parameter fhyps (name,cty) =
  let sp = lib__make_path OBJ name in
  let (u',cmap) = with_universes (fw_infexecute_parameter fhyps)
      	       	       	         (sp,initial_universes,cty) in
  add_named_object (name,OBJ) (inConstant(cmap,NeverDischarge,u'))
;;

(***** command__parameter_def_var *****)
let fw_parameter id com =
  let Tcci = fconstr_of_com mt_evc (GLOB(initial_fsign())) com in
  let Tfw = fwify Tcci in
  let _,fhyps = initial_assumptions() in
  fw_machine_parameter fhyps (id,Tfw)
;;

let ml_import qid com =
  let (f,id) = mlid_of_id qid in
  let id' = next_global_ident id in
  fw_parameter id' com ;

  let sp = lib__make_path FW id' in
  add_anonymous_object (inMLImport (IMPORT_const sp,(f,id))) ;
  MSGNL [< 'S(string_of_id id) ; 'S" imported" ;
           if id=id' then [< 'S"." >] 
      	       	     else [< 'S", renamed as " ; 'S(string_of_id id') ; 'S".">]
        >]
;;

(*** TODO: remove overwriting ***)
overwriting_vinterp_add("ML_IMPORT", function
   [VARG_IDENTIFIER id; VARG_COMMAND com] -> (fun () -> ml_import id com)
 | _ -> anomaly "ML_IMPORT called with bad arguments.")
;;


(**************************************************************************)
(*                         Link ident := Fw-term.                         *)
(**************************************************************************)

let not_an_axiom id =
  errorlabstrm "ml_import__fw_link"
  [< 'S(string_of_id id) ; 'S" is not an FW axiom." >]
;;

let not_have_type env cb' cb =
  let c = match cb'.CONSTBODY with
      SOME {v=COOKED x} -> x
    | NONE -> anomaly "ml_import__not_have_type: should be a constant" in
  errorlabstrm "ml_import__not_convertible"
  [< 'S"In environment " ; pENV env ; 'FNL ;
     'S"The term: " ; pFTERMINENV(env,c) ; 'SPC ; 
     'S"does not have type" ; 'SPC ; pFTERMINENV(env,cb.CONSTTYPE) ;
     'S"." ; 'FNL ;
     'S"Actually, it has type" ; 'SPC ; pFTERMINENV(env,cb'.CONSTTYPE) >]
;;

let fw_infexecute_constant fhyps c =
  let c = (instantiate_universes c) in
  let j = execute_rec mt_evc (GLOB fhyps) c in
  let ids = auto_save_variables() in
  let fhyps' = thin_to_level fhyps j.KIND in
  [FW, make_constant_body FW false (fhyps',j)]
;;

let fw_link id com =

  (*** First we check that id corresponds to an informative axiom. ***)
  let sp = fwsp_of_id id in
  let osp = objsp_of sp in
  let (cmap,s,u) = match lib__leaf_object_tag osp with
                     "CONSTANT" -> outConstant (lib__map_leaf osp)
		   | _ -> not_an_axiom id in
  let cb = try listmap__map cmap FW
           with Not_found -> not_an_axiom id in
  if cb.CONSTBODY <> NONE then not_an_axiom id ;

  (*** Then we execute the term com. ***)
  let Tcci = fconstr_of_com mt_evc (GLOB(initial_fsign())) com in
  let Tfw = fwify Tcci in
  let _,fhyps = initial_assumptions() in
  let sp' = lib__make_path OBJ id in
  let (u',cmap') = with_universes (fw_infexecute_constant fhyps)
      	       	       	          (sp',initial_universes,Tfw) in
  let cb' = listmap__map cmap' FW in

  (*** We check that the type of com is convertible with cb.CONSTTYPE ***)
  if not(conv mt_evc cb.CONSTTYPE cb'.CONSTTYPE) then
    not_have_type (GLOB fhyps) cb' cb ;

  (*** At last, we transform the original axiom into a constant. ***)
(**lib__remap (osp,lib__LEAF (inConstant(listmap__remap cmap FW cb',s,u))); **)
  cb.CONSTBODY <- cb'.CONSTBODY ;
  cb.CONSTOPAQUE <- false ;

  let c = match cb'.CONSTBODY with
      SOME {v=COOKED x} -> x
    | NONE -> anomaly "ml_import__fw_link: not possible!" in
  MSGNL [< 'S"Constant " ; print_id id ; 'S" linked to" ; 'SPC ;
           (* pFTERMINENV (GLOB fhyps,c) *) pFTERM c >]
;;

(*** TODO: remove overwriting ***)
overwriting_vinterp_add("LINK", function
   [VARG_IDENTIFIER id; VARG_COMMAND com] -> (fun () -> fw_link id com)
 | _ -> anomaly "ML_IMPORT called with bad arguments.")
;;


(**************************************************************************)
(*       ML Inductive name [ c1 ... cn ] == <Inductive Definition>.       *)
(**************************************************************************)
(*                     ML Inductive name1 [ c1,1 ... c1,n1 ]              *)
(*                                  ...                                   *)
(*                    	       	    namek [ ck,1 ... ck,nk ]              *)
(*                     == <Mutual Inductive Definition>.                  *)
(**************************************************************************)

(***** declare__machine_minductive *****)
let fw_machine_minductive fhyps nparams mispecvec lfinite = 
  if vect_length mispecvec = 0 then anomaly "fw_machine_minductive" 
  else 
   let mindidvec = 
     map_vect (fun  (indid,_,_,_,_) -> indid) mispecvec
   and arlcvec = 
     map_vect (fun  (_,_,indarity,indconstructors,_) ->
      	       	       	 (indarity,indconstructors)) mispecvec
   and namesvec = 
     map_vect (fun  (indid,indstamp,_,_,consnames) -> 
      	       	       	 (indstamp,indid,consnames)) mispecvec in

  let sp = lib__make_path OBJ mindidvec.(0) in
  let (u',mimap) = with_universes 
      (execute_minductive fhyps nparams namesvec lfinite)
                               (sp,initial_universes,arlcvec) in

  (***  declare_minductive (mindidvec,mimap,u')  ***)
  add_named_object (mindidvec.(0),OBJ) (inMutualInductive (mimap,u'))
  (***  do_vect declare_eliminations mindidvec  ***)
;;

(***** trad__fconstruct *****)
let fw_fconstruct sigma fsign com =
  let c = fconstr_of_com sigma (GLOB fsign) com in
  mach__fexecute sigma fsign c
;;

(***** command__build_mutual *****)
let fw_build_mutual lparams lnamearconstrs =

  let lrecnames = list__map (fun (_,x,_,_)->x)  lnamearconstrs
  and lfinite = vect_of_list (list__map (fun (x,_,_,_)->x) lnamearconstrs)
  and nparams = length lparams
  and sign0 = initial_fsign() in

  let (ind_sign,arityl) = it_list 
    (fun (sign,arl) (_,recname,arityc,_) -> 
      let arityj = fw_fconstruct mt_evc sign0 (mkProdCit lparams arityc) in
      let arity = j_val arityj in
      (add_sign (recname,arity) sign, (arity::arl)))
      (sign0,[]) lnamearconstrs in

  let mispecvec = vect_of_list 
    (list__map2 (fun ar (_,name,_,lname_constr) -> 
    let consconstrl = list__map 
     (fun (_,constr) -> 
      	let c = fconstr_of_com_sign mt_evc ind_sign (mkProdCit lparams constr)
      	in fwify c) lname_constr in
    (name,Anonymous,ar,
        put_DLAMSV_subst (rev lrecnames) (vect_of_list consconstrl),
              vect_of_list (list__map fst lname_constr)))
    (rev arityl) lnamearconstrs) in

  let _,fhyps = initial_assumptions() in
  fw_machine_minductive fhyps nparams mispecvec lfinite
;;

let not_same_number_msg () =
  errorlabstrm "ml_import__ml_import_inductive"
  [< 'S"Not the same number of ML and FW types." >]
;;

let ml_import_inductive names lparams lnamearconstrs =

  let assoc_list =
    if (list_length names) <> (list_length lnamearconstrs) then
      not_same_number_msg ()
    else
      list_it2 (fun (id,l) (_,id',_,l') acc ->
      	  if (list_length l)<>(list_length l') then
	    errorlabstrm "ml_import__ml_import_inductive"
	    [< 'S"The ML type" ; 'SPC ; print_id id ; 'S" and the FW type" ;
	       'SPC ; print_id id' ; 'SPC ; 'S"do not have the same number" ;
      	       'SPC ; 'S"of constructors." >]
	  else
	    ((id,id'),combine(l,map fst l'))::acc) names lnamearconstrs [] in

  fw_build_mutual lparams lnamearconstrs ;

  do_list
      (fun i ->
      	  let ((mlid,id),l) = nth assoc_list (i+1) in
      	  let qid = mlid_of_id mlid in
      	  let sp = lib__make_path FW id in
	  add_anonymous_object (inMLImport (IMPORT_mind (sp,i),qid)) ;
	  do_list (fun j ->
      	      let (mlid,id) = nth l j in
      	      let qid = mlid_of_id mlid in
      	      add_anonymous_object
      	       	       	      (inMLImport (IMPORT_construct ((sp,i),j),qid)) )
	  (tl (list0n (list_length l))) )
      (list0n ((list_length assoc_list) - 1)) ;

  PPNL [< 'S"ML inductive type(s) " ;
      	  prlist_with_sep (fun () -> [< 'S"," ; 'SPC >])
	                  (fun (id,_) -> [< 'S(string_of_id id) >]) names ;
         'S" imported." >]
;;  

(*** TODO: remove overwriting ***)
overwriting_vinterp_add("ML_ONEINDUCTIVE",
    fun [VARG_VARGLIST nl;
         VARG_IDENTIFIER s;
         VARG_COMMAND c;
         VARG_BINDERLIST binders;
         VARG_BINDERLIST constructors] ->
        if list_length nl <> 1 then not_same_number_msg () ;
	let [VARG_VARGLIST (VARG_IDENTIFIER id::l)] = nl in
	let cnames = map (fun (VARG_IDENTIFIER id) -> id) l in
        let la = join_binders binders in
        let li = list__map (fun ([id],c) -> (id,c)) constructors in
        fun () -> ml_import_inductive [id,cnames] la [(true,s,c,li)]
      | _ -> anomaly "ML_ONEINDUCTIVE called with bad arguments.")
;;

(*** TODO: remove overwriting ***)
overwriting_vinterp_add("ML_MUTUALINDUCTIVE",
    fun [VARG_VARGLIST l;
      	 VARG_BINDERLIST binders;
      	 VARG_VARGLIST indl] ->
	let names = map (fun (VARG_VARGLIST (VARG_IDENTIFIER id::ll)) ->
      	       	         id,map (fun (VARG_IDENTIFIER id) -> id) ll) l in
        let la = join_binders binders in
        let lnamearconstructs = 
          map (fun (VARG_VARGLIST [VARG_IDENTIFIER arid;
      	       	       	       	   VARG_COMMAND arc;
      	       	       	       	   VARG_BINDERLIST lconstr])
                -> (true,arid,arc,list__map (fun ([id],c) -> (id,c)) lconstr))
      	      indl in
        fun () -> ml_import_inductive names la lnamearconstructs
      | _ -> anomaly "ML_MUTUALINDUCTIVE called with bad arguments.")
;;


(**************************************************************************)
(*                            Fw Print ident.                             *)
(*                     To see FW constants or axioms.                     *)
(**************************************************************************)

let fprint_recipe = function
    SOME{v=COOKED c} -> pFTERM c
  | SOME{v=RECIPE _} -> [< 'S"<recipe>" >]
  | NONE -> [< 'S"<uncookable>" >]
;;

let fprint_name id =
  try let sp = nametab__sp_of_id FW id in
      let obj = lib__map_leaf (objsp_of sp) in
      (match object_tag obj with
       	 "CONSTANT" ->
      	     let (cmap,_,_) = outConstant obj in
	     let {CONSTBODY=val;CONSTTYPE=typ;_} = listmap__map cmap FW in
	     (match val with
	        NONE -> [< 'S"*** [ "; 'S(string_of_id id);  'S " : "; 'CUT ;
                           pFTERM typ ; 'S" ]"; 'FNL >] 
	      | _    -> [< 'S(string_of_id id); 'S" = " ; fprint_recipe val ;
	                   'FNL ; 'S "     : "; pFTERM typ ; 'FNL >] )

       | "MUTUALINDUCTIVE" ->
       	     let (cmap,_) = outMutualInductive obj in
	     [< print_mutual FW (listmap__map cmap FW) ; 'FNL >]
  
       | _ -> invalid_arg "fprint_name" )
  with Not_found | Invalid_argument _ -> errorlabstrm "ml_import__fprint_name"
                    [< 'S(string_of_id id) ; 'S" is not an FW constant." >]
;;

(*** TODO: remove overwriting ***)
overwriting_vinterp_add("FWPRINT",
    fun [VARG_IDENTIFIER id] -> fun () -> MSG(fprint_name id)
      | _ -> anomaly "FWPRINT called with bad arguments.")
;;

(* $Id: ml_import.ml,v 1.5 1996/02/12 17:36:51 cpaulin Exp $ *)
