(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                mlterm.ml                                 *)
(****************************************************************************)

#open "more_util";;
#open "names";;

(* functions to get correct caml names out of identifiers
 * TODO : identifiers like _foo and foo_ are not accepted by caml !
 * TODO : deal with caml reserved keywords, like true,match... *)

let name_cpt = ref 0;;

let reset_caml_names () = name_cpt := 0;;

let caml_name_of_int n = make_ident (SA__intern (gen n)) (-1)
  where rec gen n = 
    let su = string_of_char (char_of_int (97+(n mod 26))) in
    match n / 26 with
      0 -> su
    | d -> (gen d)^su
;;

let caml_new_name () =
  name_cpt := !name_cpt + 1;
  caml_name_of_int !name_cpt
;;

let underscore = id_of_string "_";;

let caml_reserved = ["and"; "as"; "begin"; "do"; "done"; 
                     "else"; "end"; "exception"; "fun"; "function"; 
                     "if"; "in"; "let"; "match"; 
                     "mutable"; "not"; "of"; "or"; "prefix";
                     "raise"; "rec"; "then"; "try"; "type"; "value";
                     "where"; "while"; "with" ];;

let is_caml_keyword =
  let t = (hashtbl__new 17 : (string, unit) hashtbl__t) in
  do_list (fun s -> hashtbl__add t s ()) caml_reserved ;
  function s -> try hashtbl__find t s ; true
      	       	with Not_found -> false
;;

let caml_name_of avoid = function
    Anonymous -> next_ident_away (caml_new_name()) avoid
  | Name id   -> let s = string_of_id id in
      	       	 if (nth_char s 0) = `_` then 
      	       	   next_ident_away (caml_new_name()) avoid
      	       	 else if is_caml_keyword s then
		   next_ident_away (id_of_string (s^"_ren")) avoid
		 else
      	       	   next_ident_away id avoid
;;

let rec new_vars avoid = function
    0 -> []
  | n -> let id = make_ident (SA__intern "x") (-1) in
      	 let f = next_ident_away id avoid in
	 f::(new_vars (f::avoid) (n-1))
;;


(* rename_list : identifier list -> identifier list -> identifier list
 * [rename_list V l] renames the list l with names out of V. *)

let rename_list V ids = 
  let rec ren1 acc = function
    []    -> rev acc
  | a::l1 -> 
       if mem a V then
      	 let nv = caml_name_of (union l1 (union V acc)) (Name a) in
	 ren1 (nv::acc) l1
       else
       	 ren1 (a::acc) l1 
  in ren1 [] ids
;;


(* lifting on terms. ml_lift : int -> MLast -> MLast
 * [ml_lift k M] lifts the binding depth of M across k bindings. *)

let ml_liftn k n c = if k=0 then c else liftrec n c
  where rec liftrec n = function
    MLrel i as c      -> if i<n then c else MLrel(i+k)
  | MLapp(t,args)     -> MLapp(liftrec n t, map (liftrec n) args)
  | MLlam(id,t)       -> MLlam(id,liftrec (n+1) t)
  | MLcons(j,id,args) -> MLcons(j,id,map (liftrec n) args)
  | MLcase(t,pl)      -> MLcase(liftrec n t,
      	       	    map_vect (fun (id,idl,p) -> let k = list_length idl in
			              (id,idl,liftrec (n+k) p)) pl)
  | x                 -> x
;;

let ml_lift k c = ml_liftn k 1 c;;

let ml_pop c = ml_lift (-1) c;;

let ml_liftn_branch k (id,ids,v) =
  (id,ids,ml_liftn k (list_length ids+1) v)
;;

(* ml_subst1 : identifier list -> MLast -> MLast -> MLast
 * [ml_subst1 V M t] substitutes M for (Rel 1) in t, renaming bound variables
 * out of V. *)

let rec ml_subst1 V =
  let rec subst n M = function
    MLrel i           -> if i=n then M
                         else if i<n then MLrel i else MLrel (i-1)
  | MLapp(t,argl)     -> MLapp(subst n M t, map (subst n M) argl)
  | MLlam(id,t)       -> if mem id V then
      	       	           let id' = caml_name_of V (Name id) in
      	       	           MLlam(id',subst (n+1) (ml_lift 1 M) t)
		         else
		           MLlam(id,subst (n+1) (ml_lift 1 M) t)
  | MLcons(i,id,argl) -> MLcons(i,id,map (subst n M) argl)
  | MLcase(t,pv)      -> MLcase(subst n M t, map_vect (fun (id,ids,t) ->
      	       	           let k = list_length ids in
		           let ids' = rename_list V ids in
      	       	           (id,ids',subst (n+k) (ml_lift k M) t)) pv)
  | MLfix(i,b,ids,cl) -> let k = list_length ids in
      	       	       	 let ids' = rename_list V ids in
      	       	       	 MLfix(i,b,ids',map (subst (n+k) (ml_lift k M)) cl)
  | x                 -> x
  in subst 1
;;


(* occurs : int -> MLast -> bool
 * [occurs k M] returns true if (Rel k) occurs in M. *)

let rec occurs k = function
    MLrel i          -> i=k
  | MLapp(t,argl)    -> (occurs k t) or (occurs_list k argl)
  | MLlam(_,t)       -> occurs (k+1) t
  | MLcons(_,_,argl) -> occurs_list k argl
  | MLcase(t,pv)     -> (occurs k t) or
      	       	       	(exists (fun (k',t') -> occurs (k+k') t')
      	       	            (map_vect_list (fun (_,l,t') -> 
      	       	             let k' = list_length l in (k',t')) pv))
  | MLfix(_,_,l,cl)  -> let k' = list_length l in occurs_list (k+k') cl
  | _                -> false
and occurs_list k l =
  exists (fun t -> occurs k t) l
;;



(* $Id: mlterm.ml,v 1.4 1996/02/12 17:36:54 cpaulin Exp $ *)
