(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                               optimise.ml                                *)
(****************************************************************************)

#open "std";;
#open "names";;
#open "mlterm";;

(* free_var : identifier list -> MLast -> identifier list
 * [free_var env t] returns the list of free variables of t in
 * de Bruijn environment env. *)

let free_var = free 0
  where rec free n idl = function
    MLrel i           -> if i>n then [nth idl i] else []
  | MLapp(t,args)     -> it_list union [] (map (free n idl) (t::args))
  | MLlam(id,t)       -> free (n+1) (id::idl) t
  | MLcons(_,_,args)  -> it_list union [] (map (free n idl) args)
  | MLcase(t,pv)      -> union (free n idl t)
      	       	          (it_vect union [] (map_vect (fun (_,ids,t) ->
			   let k = length ids in free (n+k) (ids@idl) t) pv))
  | MLfix(_,_,ids,cl) -> let k = length ids in
      	       	       	 it_list union [] (map (free (n+k) (ids@idl)) cl)
  | _ -> []
;;

let free_var_list idl l = it_list union [] (map (free_var idl) l);;


(* ml_subst_glob : identifier -> MLast -> MLast -> MLast
 * [ml_subst_glob id M t] substitutes M for (MLglob id) in t. *)

let ml_subst_glob id M = substrec
  where rec substrec = function
    MLglob id' as t -> if id=id' then M else t
  | MLapp(t,argl)   -> MLapp(substrec t, map substrec argl)
  | MLlam(id,t)     -> MLlam(id, substrec t)
  | MLcons(i,id,l)  -> MLcons(i,id,map substrec l)
  | MLcase(t,v)     -> MLcase(substrec t,
      	       	       	   map_vect (fun (n,l,t) -> (n,l,substrec t)) v)
  | MLfix(i,b,l,f)  -> MLfix(i,b,l, map substrec f)
  | x		    -> x
;;


(* norm : identifier list -> MLast -> MLast
 * [norm env M] normalises the term M in the de Bruijn environment env. *)

let rec norm idl = function
    MLapp(t1,argl) ->
      	let t1' = norm idl t1 in
	(match t1' with
	   MLlam(id,t4) ->
	        let M = hd argl in 
	        let V = union (free_var idl M) (free_var (id::idl) t4) in
	        (match tl argl with
		   [] -> norm idl (ml_subst1 V M t4)
		 | l  -> norm idl (MLapp(ml_subst1 V M t4, l)))
         | MLcase(t4,pv) ->
	        let argl' = map (norm idl) argl in
		let V = free_var_list idl argl' in
	        let pv' = map_vect 
      	       	       	   (fun (n,l,t) -> 
      	       	       	      let V' = union V (free_var (l@idl) t) in
      	       	       	      let l' = rename_list V' l in
			      let k = list_length l in
			      let argl'' = map (ml_lift k) argl' in
      	       	       	      (n,l',norm (l'@idl) (MLapp(t,argl'')))) pv in
      	       	norm idl (MLcase(t4,pv'))
	 | _ -> MLapp(t1', map (norm idl) argl))

  | MLlam(id,t) -> MLlam(id, norm (id::idl) t)

  | MLcase(t,v) as X -> 
      	 let t2 = norm idl t in
	 (match t2 with
	    MLcons(j,_,argl) -> 
	         let (_,ids,c) = v.(j-1) in
		 let V = union (it_list union [] (map (free_var idl) argl))
		               (free_var (ids@idl) c) in
		 let c' = it_list (fun a M -> ml_subst1 V M a) c (rev argl) in
		 norm idl c'

      	  | MLcase(t',v') ->
	         let VX = free_var idl X in
		 let c = MLcase(t',map_vect (fun (id,ids,t0) ->
                       let ids' = rename_list VX ids in
                       let k = length ids' in
                       let v'' = map_vect (ml_liftn_branch k) v in
	 	       (id,ids',MLcase(t0,v''))) v') in
		 norm idl c

      	  | _ -> MLcase(t2,map_vect (fun (id,ids,c) -> 
      	       	       	       	       	     (id,ids,norm (ids@idl) c)) v) )

  | MLfix(i,b,ids,l) -> MLfix(i,b,ids,map (norm (ids@idl)) l)

  | MLcons(i,id,argl) -> MLcons(i,id,map (norm idl) argl)

  | t -> t
;;


(* is_strict : MLast -> bool
 * [is_strict M] tells us if the term M must NOT be expanded. *)

let rec ml_size = function
    MLapp(t,l)     -> (length l) + (ml_size t) + (ml_size_list l)
  | MLlam(_,t)     -> 1 + (ml_size t)
  | MLcons(_,_,l)  -> ml_size_list l
  | MLcase(t,pv)   -> 1 + (ml_size t) + (it_list (prefix +) 0
      	       	              (map_vect_list (fun (_,_,t) -> ml_size t) pv))
  | MLfix(_,_,_,f) -> ml_size_list f
  | _              -> 0
and ml_size_list l =
  it_list (prefix +) 0 (map ml_size l)
;;

let rec hd_var = function
    MLapp(t,_) -> hd_var t
  | MLrel _    -> true
  | _	       -> false
;;

let rec ml_abs_var = function
    MLlam(id,t)  -> add_set id (ml_abs_var t)
  | MLcase(_,pv) -> list_it union (map_vect_list (fun (_,l,t) ->
		       subtract (ml_abs_var t) l) pv) []
  | _            -> []
;;

let strict_var = strict []
  where rec strict idl = function
    MLapp(MLglob _,l) -> strict_list idl l
  | MLapp(MLrel n, _) -> [nth idl n]
  | MLapp(t1,l)       -> if hd_var t1 then strict idl t1
      	       	       	 else union (strict idl t1) (strict_list idl l)
  | MLlam(id,t)       -> strict (id::idl) t
  | MLcons(_,_,l)     -> strict_list idl l
  | MLcase(t,pv)      -> union (strict idl t)
      	       	       	     (list_it intersect (map_vect_list (fun (_,l,c) ->
			      subtract (strict (l@idl) c) l) pv) idl)
  | MLfix(_,_,ids,f)  -> strict_list (ids@idl) f
  | MLrel n           -> [nth idl n]
  | _                 -> []
  and strict_list idl l =
    it_list union [] (map (strict idl) l)
;;

let is_strict t =
  (  (ml_size t>4)
   & ([] = subtract (ml_abs_var t) (strict_var t)) )
  or (match t with MLfix _ -> true | _ -> false)
;;

let rec is_constr = function
    MLcons _   -> true
  | MLlam(_,t) -> is_constr t
  | _          -> false
;;


(* optimise : MLdecl list -> MLdecl list
 * [optimise mlenv] performs partial evaluation on the ML environment
 * mlenv (ie) some global terms are expanded in the following terms. *)

let expand (id,t) = function
    DECLglob(id',t') -> DECLglob(id', ml_subst_glob id t t')
  | d		     -> d
;;  

let rec optimise needed = opt_rec
  where rec opt_rec = function
    [] -> []

  | (DECLtype _ |DECLabbrev _) as d::l ->
      	d::(opt_rec l)

  | DECLglob(id,(MLexn _ as t)) as d::l ->
        let l' = map (expand (id,t)) l in
	opt_rec l'

  | DECLglob(id,t) as d::l ->
      	let t' = norm [] t in
	if (not (is_constr t)) & (is_strict t') or l=[] or (mem id needed) then
	  (DECLglob(id,t'))::(opt_rec l)
      	else begin
	  lomsg__warning ("The constant "^(string_of_id id)^" is expanded.");
	  let l' = map (expand (id,t')) l in
          opt_rec l'
	end
;;

(* $Id: optimise.ml,v 1.5 1996/02/12 17:36:58 cpaulin Exp $ *)
