(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                ccidpc.ml                                 *)
(****************************************************************************)

#open "dpctypes";;

#open "names";;

#open "pp";;
#open "stdpp";;
#open "std";;
#open "more_util";;
#open "generic";;
#open "term";;
#open "environ";;
#open "termenv";;
#open "himsg";;
#open "tacmach";;
#open "reduction";;
#open "proof";;
#open "clenv";;
#open "tactics";;
#open "trad";;
#open "pattern";;
#open "tactics0";;
#open "tactics1";;
#open "tactics2";;
#open "tactics3";;

#infix "ORELSE";;
#infix "NEXT";;
#infix "THEN";;
#infix "THENF";;
#infix "THENL";;
#infix "THENS";;

#infix "THEN_i1";;

let mmk = make_module_marker ["#Prelude.obj"];;

let and_pattern = put_pat mmk "(and ? ?)";;
let or_pattern = put_pat mmk "(or ? ?)";;
let not_pattern = put_pat mmk "(not ?)";;
let exist_pattern = put_pat mmk "(ex ? ?)";;

let and_head = put_pat mmk "and";;
let or_head = put_pat mmk "or";;
let not_head = put_pat mmk "not";;
let exist_head = put_pat mmk "ex";;

let is_imp_term T =
    match T with
    DOP2(Prod,_,DLAM(_,B)) -> not(dependent (Rel 1) B)
  | _ -> false
;;

let is_forall_term T =
    match whd_betaiota T with
    DOP2(Prod,_,DLAM(_,B)) -> dependent (Rel 1) B
  | _ -> false
;;

let ctr = ref 0;;
let gen_ident id = make_ident (atompart_of_id id) (incr ctr;!ctr);;
let gen_name A na = 
    let (Name id) = (named_hd A na)
    in gen_ident id;;

let dpc_of_cci_term lid = tradrec
    where rec tradrec cciterm =
    let (hd,args) = whd_betaiota_stack cciterm [] in
    let dpcargs = map tradrec args
    in (match hd with
        VAR id -> if dpcargs=[] then Var id 
                  else if mem id lid 
                       then error "dpc_of_cci_term (not first order)"
                       else App (Var id :: dpcargs)

      | ( DOPN(Const _,_) | DOPN(MutInd _,_) | DOPN(MutConstruct _,_)) as t 
      	   -> if dpcargs=[] then Glob (CN t)
	                    else App (Glob (CN t) :: dpcargs)

      | _ -> errorlabstrm "dpc_of_cci_term"
                [< 'S"Not a first order formula" >] )
;;


let cci_of_dpc_term tradsign sign = tradrec
    where rec tradrec = function
    Var id -> if mem_sign tradsign id then VAR(snd(lookup_sign id tradsign))
              else VAR id
  | Glob (ID id) -> VAR id
  | Glob (CN t)  -> t
  | App (t::dpcargs) ->
      let t' = tradrec t in
      applist(t', map tradrec dpcargs)
;;

let dpc_of_cci_fmla gls cciterm = tradrec [] (whd_betaiota cciterm)
    where rec tradrec lid cciterm =
    if matches gls cciterm and_pattern then
        let [A;B] = dest_match gls cciterm and_pattern
        in Conj(tradrec lid A,tradrec lid B)
    else if matches gls cciterm or_pattern then
        let [A;B] = dest_match gls cciterm or_pattern
        in Disj(tradrec lid A,tradrec lid B)
    else if matches gls cciterm not_pattern then
        let [A] = dest_match gls cciterm not_pattern
        in Neg(tradrec lid A)
    else if matches gls cciterm exist_pattern then
        let [_;B] = dest_match gls cciterm exist_pattern
        in (match B with
            DOP2(Lambda,A,(DLAM(na,_) as B)) ->
            let id = gen_name A na
            in Exists(id,tradrec (id::lid) (SAPP B (VAR id)))
          | _ -> error "dpc_of_cci_fmla (exists)")
    else if is_imp_term cciterm then
        let (DOP2(Prod,A,DLAM(_,B))) = cciterm
        in Imp(tradrec lid A,tradrec lid (pop B))
    else if is_forall_term cciterm then
        let (DOP2(Prod,A,(DLAM(na,_) as B))) = cciterm in
        let id = gen_name A na
        in ForAll(id,tradrec (id::lid) (SAPP B (VAR id)))
    else let (hd,args) = whd_betaiota_stack cciterm [] in
         let dpcargs = map (dpc_of_cci_term lid) args
         in (match hd with
             VAR id -> if mem id lid 
                   then errorlabstrm "dpc_of_cci_fmla"
                          [< 'S"Quantification over a predicate" >]
                   else Atom((ID id,0),dpcargs)

           | (DOPN(MutInd _,_) | DOPN(MutConstruct _,_) | DOPN(Const _,_)) as t
      	        -> Atom( (CN t,0) , dpcargs)

           | _ -> errorlabstrm "dpc_of_cci_flma"
                     [< pTERMINENV(GLOB(pf_hyps gls),hd); 'SPC ;
                        'S"is not an atomic predicate" >] )
;;

let rec alpha_term bl1 bl2 = fun
    (Var id1) (Var id2) ->
      if (pure_index(id1,bl1) = -1) then (id1=id2)
      else pure_index(id1,bl1) = pure_index(id2,bl2)
  | (Glob t1) (Glob t2) -> t1=t2
  | (App al1) (App al2) ->
      for_all2eq (alpha_term bl1 bl2) al1 al2
  | _ _ -> false
;;

let ForAllI gls =
    if is_forall_term (pf_concl gls) then
        tactics__intro gls
    else FAILTAC gls;;

let ForAllE id t gls =
    match pf_whd_betadeltaiota gls (pf_type_of gls (VAR id)) with
    DOP2(Prod,A,(DLAM _ as B)) as T ->
    if is_forall_term T then
        (Cut (SAPP B t)
         THENS [IDTAC;
                Exact (applist(VAR id,[t]))]) gls
    else FAILTAC gls
;;  

let ExistE id gls =
    if matches gls (snd(lookup_sign id (pf_hyps gls))) exist_pattern then
        (elimination [] (VAR id) THEN
         DO 2 (DImp NONE)) gls
    else FAILTAC gls
;;

let NegE id gls = 
    if matches gls (snd(lookup_sign id (pf_hyps gls))) not_pattern then
        (elimination [] (VAR id)) gls
    else FAILTAC gls
;;

let exist_intro_head = put_pat mmk "ex_intro";;

let ExistI t gls =
    let (wc,KONT) = StartWalk gls in
    let clause = mk_clenv_hnf_constr_type_of wc (get_pat exist_intro_head) in
    let clause' = clenv_constrain_missing_args [t] clause
    in res_pf KONT clause' gls
;;

let rec alpha_fmla bl1 bl2 = fun
    (Atom((cn1,_),al1)) (Atom((cn2,_),al2)) ->
    cn1=cn2 & for_all2eq (alpha_term bl1 bl2) al1 al2
  | (Conj(A1,B1))(Conj(A2,B2)) ->
    alpha_fmla bl1 bl2 A1 A2 & alpha_fmla bl1 bl2 B1 B2

  | (Disj(A1,B1))(Disj(A2,B2)) ->
    alpha_fmla bl1 bl2 A1 A2 & alpha_fmla bl1 bl2 B1 B2

  | (Imp(A1,B1))(Imp(A2,B2)) ->
    alpha_fmla bl1 bl2 A1 A2 & alpha_fmla bl1 bl2 B1 B2

  | (Neg(A1)) (Neg(A2)) -> alpha_fmla bl1 bl2 A1 A2

  | (ForAll(x1,A1)) (ForAll(x2,A2)) ->
    alpha_fmla (x1::bl1) (x2::bl2) A1 A2

  | (Exists(x1,A1)) (Exists(x2,A2)) ->
    alpha_fmla (x1::bl1) (x2::bl2) A1 A2

  | _ _ -> false;;

let alpha_eq (kspine,M) (jspine,N) = alpha_fmla kspine jspine M N;;

let first_pred P = firstrec
    where rec firstrec = function
    [] -> error "first_pred"
  | h::t -> if P h then h else firstrec t
;;

let find_fmla_left (kspine,f) (jspine,gl) =
    let sign = pf_hyps gl in
    let ids = ids_of_sign sign
    in first_pred 
         (fun id -> ( try alpha_eq (kspine,f)
                        (jspine,dpc_of_cci_fmla gl (snd(lookup_sign id sign)))
             with _ -> false) 
         ) ids
;;

let OnNthClause tac n gls =
    let cls = nth_clause n gls
    in OnClause tac cls gls
;;

let ElimTypeFalse gls =
  (elimination_type (pf_global gls (id_of_string "False"))) gls
;;

let rec tradpf kspine jspine dpcpf gls =
    match dpcpf with

    Proof2(_,_,Axiom2 f) ->
    let id = find_fmla_left (kspine,f) (jspine,gls)
    in Exact (VAR id) gls

  | Proof2(_,_,LWeakening2(_,pf)) -> trad kspine jspine pf gls

  | Proof2(_,_,RWeakening2(_,pf)) ->
     ( ElimTypeFalse THEN tradpf kspine jspine pf ) gls

  | Proof2(_,_,RConj2(f1,pf1,f2,pf2)) ->
    (DAnd NONE THENS [trad kspine jspine pf1;
                      trad kspine jspine pf2]) gls

  | Proof2(_,_,LConj2(f1,f2,pf)) ->
    let id = find_fmla_left (kspine,Conj(f1,f2)) (jspine,gls)
    in (DAnd (SOME id) THEN (trad kspine jspine pf)) gls

  | Proof2(_,_,LDisj2(f1,pf1,f2,pf2)) ->
    let id = find_fmla_left (kspine,Disj(f1,f2)) (jspine,gls)
    in (match pf1 with 
          Proof2(_,[],_) -> (DOr 0 (SOME id) THENS 
                            [ (ElimTypeFalse THEN trad kspine jspine pf1);
                              trad kspine jspine pf2]) gls
        | _              -> (DOr 0 (SOME id) THENS 
                            [ trad kspine jspine pf1;
                              (ElimTypeFalse THEN trad kspine jspine pf2)]) gls
       )

  | Proof2(_,_,RImp2(f1,f2,pf)) ->
    (DImp NONE THEN (trad kspine jspine pf)) gls

  | Proof2(_,_,LImp2(f1,pf1,f2,pf2)) ->
    let id = find_fmla_left (kspine,Imp(f1,f2)) (jspine,gls)
    in (DImp (SOME id) THENS [trad kspine jspine pf2;
                              trad kspine jspine pf1]) gls

  | Proof2(_,_,RForAll2(kid,f,pf)) ->
    (ForAllI THEN
     (OnLastHyp (fun (SOME jid) ->
                     trad (kid::kspine) (jid::jspine) pf))) gls

  | Proof2(_,_,LForAll2(kid,kterm,f,pf)) ->
    let jterm = cci_of_dpc_term (kspine,jspine) (pf_hyps gls) kterm in
    let id = find_fmla_left (kspine,ForAll(kid,f)) (jspine,gls)
    in (ForAllE id jterm THEN
        trad kspine jspine pf) gls

  | Proof2(_,_,LExists2(kid,f,pf)) ->
    let id = find_fmla_left (kspine,Exists(kid,f)) (jspine,gls)
    in (ExistE id THEN
        (OnNthClause (fun (SOME jid) ->
                          trad (kid::kspine) (jid::jspine) pf)
         (-2))) gls

  | Proof2(_,_,RExists2(kid,kterm,f,pf)) ->
    let jterm = cci_of_dpc_term (kspine,jspine) (pf_hyps gls) kterm
    in (ExistI jterm THEN tradpf kspine jspine pf) gls

  | Proof2(_,_,RDisj2(f1,f2,Proof2(_,_,RWeakening2(f3,pf)))) ->
    if alpha_eq (kspine,f1) (kspine,f3) then
        (DOr 2 NONE THEN tradpf kspine jspine pf) gls
    else if alpha_eq (kspine,f2) (kspine,f3) then
        (DOr 1 NONE THEN tradpf kspine jspine pf) gls
    else error "Not Intuitionistic, eh?"

  | Proof2(_,_,RNeg2(f,pf)) -> 
      (red THEN intro THEN tradpf kspine jspine pf) gls

  | Proof2(_,_,LNeg2(f,pf)) -> 
     let id = find_fmla_left (kspine,Neg(f)) (jspine,gls)
     in (NegE id THEN tradpf kspine jspine pf) gls

  | _ -> error "tradpf : Not an intuitionistic proof !"

and trad kspine jspine dpcpf gls =
    tradpf kspine jspine dpcpf gls
;;

(* $Id: ccidpc.ml,v 1.3 1996/02/12 17:37:07 cpaulin Exp $ *)
