(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                 unif.ml                                  *)
(****************************************************************************)

          (* UNIFICATION : Martelli & Montanari's algorithm *)
          (* ---------------------------------------------- *)

(* Terms are of the type :
 *     type term = Var of identifier
 *		 | Glob of constname
 * 		 | App of term list
 *     
 * Unification of two atomic formulae :
 *     unif_atoms : atom_id * term list -> atom_id * term list 
 *                  -> (term*term) list  
 *
 *  Unification of two terms :
 *     unif_terms : term -> term -> (term*term) list
 *
 *)

#open "names";;
#open "general";;
#open "dpctypes";;
#open "subst";;

let rec vars_of_list l = match l with
    t::ll -> let vll = vars_of_list ll
      	       in let vt = vars_of_term t
	       in union vt vll
  | []    -> []
and vars_of_term t = match t with
      Var x  -> [Var(x)]
    | Glob _ -> []
    | App lt -> vars_of_list lt;;

let rec nb_occ_term x t = match t with
    Var y  -> if x=Var(y) then 1 else 0
  | App lt -> nb_occ_list x lt
  | _      -> 0
and nb_occ_list x l = match l with
    t::ll -> (nb_occ_term x t)+(nb_occ_list x ll)
  | []    -> 0;;

let cpt = ref 0;;

let w_atom = SA__intern "-";;

let new_id () = 
  cpt := !cpt+1;
  make_ident w_atom !cpt;;
 
let rec initU l1 l2 = match (l1,l2) with
     ((a1::ll1),(a2::ll2)) -> let w = new_id()
       	       	       	      in let equ = ([Var w],[a1; a2])
			      in equ::(initU ll1 ll2)
   | ([],[]) -> []
   | _ -> raise Impossible_case;;

let initT l1 l2 =
  let lt =l1@l2
  in map (fun x -> ([x],[])) (vars_of_list lt);;
  
let rec CF (t1,t2) = match t1 with 
    Var(x1) -> 
      	(match t2 with
	    Var _ -> ([t1] , [([t1; t2],[])])
	  | _     -> ([] , [([t1],[t2])])
      	)
  | Glob _ -> (match t2 with
                 Glob _ -> if t1=t2 then ([t1] , []) else raise Not_unifiable
	       | Var _  -> ([] , [([t2],[t1])])
	       | _      -> raise Not_unifiable)
  | App lt1 -> 
      	(match t2 with
	    Var _   -> ([] , [([t2],[t1])])
	  | Glob _  -> raise Not_unifiable
      	  | App lt2 -> 
      	       if (list_length lt1)=(list_length lt2)
	       then    let lcf = map CF (combine (lt1,lt2))
		    in let lc  = map (fun ([x],_) -> x
		                        | ([],e) -> hd (fst (hd e))
      	       	       	       	       	| _ -> failwith "Imp.CF") lcf
		    in let llf = map (fun (_,x) -> x) lcf
		    in  ([App lc] , glue llf)
	       else raise Not_unifiable
	);;

let assocT x T =
  let rec assocT_rec = function 
      ((lv , _) as equ::TT) -> if mem x lv then equ
      	       	       	       	       	   else assocT_rec TT
    | [] -> raise Not_found
  in assocT_rec T;;					     

let rec sub_elem_list x = function
    (a::l) -> if a=x then l else a::(sub_elem_list x l)
  | [] -> [];;

let newUT1 x1 x2 U T = 
     let (lv1,lt1) as equ1 = assocT x1 T
  in let (lv2,lt2) as equ2 = assocT x2 T	
  in let nT = sub_elem_list equ1 (sub_elem_list equ2 T)
  in if equ1=equ2 
     then (U , T)
     else    if lt1=[] or lt2=[] 
      	     then (U , (lv1@lv2 , lt1@lt2)::nT)
	     else let t1 = hd lt1 and t2 = hd lt2 
		  in let (C,F) = CF (t1,t2)
		  in (F@U , (lv1@lv2 , C)::nT);; 

let newUT2 x1 t1 U T =
     let (lv1,lt1) as equ1 = assocT x1 T
  in let nT = sub_elem_list equ1 T
  in match lt1 with
       [t2] -> let (C,F) = CF (t1,t2) 
               in (F@U , (lv1 , C)::nT)
     | [] -> (U , (lv1 , [t1])::nT)
     | _ -> failwith "Imp.newUT2";;

let newUT3 w t1 t2 U T =
     let (C,F) = CF (t1,t2)
  in match t1 with
       Var(_) as x1 -> let (lv1,lt1) as equ1 = assocT x1 T  
       	       	       in let nT = sub_elem_list equ1 T
		       in let TT = (w::lv1,lt1)::nT
      	       	       in (match t2 with
		              Var(_) as x2 -> (([x1;x2],[])::U , TT)
                            | _ -> (([x1],[t2])::U , TT)
      	       	       	  )
     | _ -> (match t2 with
       	       	 Var(_) as x2 -> let (lv2,lt2) as equ2 = assocT x2 T
		                 in let nT = sub_elem_list equ2 T
				 in let TT = (w::lv2,lt2)::nT
				 in (([x2],[t1])::U , TT)
	       | _ -> let (C,F) = CF (t1,t2)
	              in (F@U , ([w],C)::T)
      	    );;

  (* mm : Compute the derivation of (U,T) until U=[] ***********)

let rec mm U T = match U with
    ((v,s)::UU) -> 
      	(match s with 
	    []      -> (match v with
	                  [x1;x2] -> let (nU,nT) = newUT1 x1 x2 UU T
			             in mm nU nT
			| _ -> failwith "Imp.cas 1"
      	       	       )
	  | [t1]    -> (match v with
	                  [x1] -> let (nU,nT) = newUT2 x1 t1 UU T
			          in mm nU nT
			| _ -> failwith "Imp.cas 2"
      	       	       )
	  | [t1;t2] -> (match v with
	                  [x1] -> let (nU,nT) = newUT3 x1 t1 t2 UU T
			          in mm nU nT
			| _ -> failwith "Imp.cas 3"
      	       	       )
	  | _ -> raise Impossible_case
	) 
  | [] -> T;;


let nb_occ_list_list lv ltt = 
  let rec nb_occ_ll_rec = function 
      (v::lv1) -> (nb_occ_list v ltt)+(nb_occ_ll_rec lv1)
    | [] -> 0
  in nb_occ_ll_rec lv;;

let rec gro_aux = function
    (Var x)::ll -> if (atompart_of_id x) = w_atom then gro_aux ll
      	       	       	       	       	   else (Var x)::(gro_aux ll)
  | h::ll       -> h::(gro_aux ll)
  | []          -> [];; 

let rec gro_aux_T = function 
    ((lv,lt)::TT) -> 
      	 let lv0 = gro_aux lv 
         in (match lv0 with 	       	 
      	       [] -> gro_aux_T TT
	     | _ -> (lv0,lt)::(gro_aux_T TT)
	    )
  | [] -> [];;

(* From now on, the equations of T are associated with the number of
   occurences of their variables in the right equations *********)

let rec find_null = function
    ((r,(_,_)) as equ::TT) -> if !r=0 then equ else find_null TT
  | [] -> raise Not_unifiable;;

let rec recompute T1 t = match T1 with
    ((r,(lv,lt))::TT) -> let n = nb_occ_list_list lv [t]
       	       	       	 in (ref (!r-n),(lv,lt))::(recompute TT t)
  | [] -> [];;

let rec sorted_Tc T sT = 
  if T=[] then sT else
       let (_,(lv,lt)) as equ = find_null T
    in let T1 = sub_elem_list equ T
    in match lt with
         []  -> sorted_Tc T1 (equ::sT)
       | [t] -> sorted_Tc (recompute T1 t) (equ::sT)
       | _   -> raise Impossible_case;;

let rec apply_subst s = function
    Var(_) as v -> if mem_assoc v s then assoc v s
      	       	       	       	    else v
  | Glob _ as x -> x				   
  | App lt      -> App (map (apply_subst s) lt);;

let rec unif_from_sTc Un = function
     ((_,(lv,lt))::TT) -> 
           (match lt with
	        []  ->    let x1 = hd lv and lv1 = tl lv
		       in let U0 = map (fun x->(x,x1)) lv1
		       in unif_from_sTc (Un@U0) TT
	      | [t] ->    let t0 = apply_subst Un t
      	       	       in let U0 = map (fun x->(x,t0)) lv
		       in unif_from_sTc (Un@U0) TT
	      | _   -> raise Impossible_case
      	   ) 
   | [] -> Un;;

let unif_from_T T0 =
     let T = gro_aux_T T0
  in let llt = map (fun (_,x) -> x) T
  in let ltt  = glue llt
  in let rec comp_nb T1 = match T1 with
           ((lv,lt)::TT1) -> (ref (nb_occ_list_list lv ltt),(lv,lt)):: 
	                         (comp_nb TT1)
         | _ -> []
  in let Tc = comp_nb T
  in let sTc = sorted_Tc Tc []
  in unif_from_sTc [] sTc;;

(* unif_atoms : atom -> atom -> unifier ******)

let unif_atoms (p1,l1) (p2,l2) = 
  if (fst p1)<>(fst p2) or (list_length l1)<>(list_length l2)
  then raise Not_unifiable
  else cpt := 0;
       let T = mm (initU l1 l2) (initT l1 l2)
       in unif_from_T T;;

(* unif_terms : term -> term -> unifier *******)

let unif_terms t1 t2 = 
  cpt := 0;
  let l1 = [t1] and l2 =[t2]
  in let T = mm (initU l1 l2) (initT l1 l2)
  in unif_from_T T;;

(* assoc_unif : unifier -> variable -> term ******)

let assoc_unif v u = 
  try assoc v u
  with Not_found -> v;;

(* apply_unif : unifier -> term -> term ******)

let apply_unif u t = 
  let rec au_rec = function
      Var _  as v -> assoc_unif v u
    | Glob _ as x -> x
    | App lt      -> App (map au_rec lt)
  in au_rec t;;

(* appear_var_term : variable -> term -> bool ********)

let appear_var_term v t = 
  let rec avt_rec  = function
      Var _ as v1 -> v1=v
    | Glob _      -> false
    | App lt      -> it_list (fun x t1 -> x or (avt_rec t1)) false lt
  in avt_rec t;;

(* up u1 u2 : returns the least unifier greater than u1 and u2
              If no such unifier exists, it raises Up_error   *******)

let rec up u1 u2 = match u2 with
    (((Var s1) as y1,t1) as eq::uu2) -> 
      	if mem_assoc y1 u1
	then    let M1 = assoc y1 u1
	     in let t1' = apply_unif u1 t1
	     in let u0 = try unif_terms M1 t1'
	                 with Not_unifiable -> raise Up_error
	     in let u1' = map (fun (x,t) -> (x,apply_unif u0 t)) u1
	     in up (u1'@u0) uu2
	else    let t1' = apply_unif u1 t1
	     in if t1' = y1 
      	       	then u1
	        else if appear_var_term y1 t1'
		     then raise Up_error
		     else let u1' = 
      	       	       	     map (fun (x,t) -> (x,subst_term s1 t1' t)) u1
      	       	       	  in up ((y1,t1')::u1') uu2 
  | [] -> u1
  | _  -> failwith "unif__up: impossible case";;

(* $Id: unif.ml,v 1.3 1996/02/12 17:37:37 cpaulin Exp $ *)
