(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                 dhyp.ml                                  *)
(****************************************************************************)

(* Programmable destruction of hypotheses and conclusions.

   The idea here is that we are going to store patterns.  These
   patterns look like:

   TYP=<patern>
   SORT=<pattern>

   and from these patterns, we will be able to decide which tactic to
   execute.

   For hypotheses, we have a vector of 4 patterns:

   HYP[TYP] HYP[SORT] CONCL[TYP] CONCL[SORT]

   and for conclusions, we have 2:

   CONCL[TYP] CONCL[SORT]

   If the user doesn't supply some of these, they are just replaced
   with empties.

   The process of matching goes like this:

   We use a discrimination net to look for matches between the pattern
   for HYP[TOP] (CONCL[TOP]) and the type of the chosen hypothesis.
   Then, we use this to look for the right tactic to apply, by
   matching the rest of the slots.  Each match is tried, and if there
   is more than one, this fact is reported, and the one with the
   lowest priority is taken.  The priority is a parameter of the
   tactic input.

   The tactic input is an expression to hand to the
   tactic-interpreter, and its priority.

   For most tactics, the priority should be the number of subgoals
   generated.

   Matching is compatible with second-order matching of sopattern.

   SYNTAX:

   Hint DHyp <hyp-pattern> pri <tac-pattern>
   with $i:=blabl $y:=lskjdljb etc.

   and

   Hint DConcl <concl-pattern> pri <tac-pattern>
   with $i:=blabl $y:=lskjdljb etc.

   The bindings at the end allow us to transfer information from the
   patterns on terms into the patterns on tactics in a safe way - we
   will perform second-order normalization and conversion to an AST
   before substitution into the tactic-expression.

   WARNING: The binding mechanism is NOT intended to facilitate the
   transfer of large amounts of information from the terms to the
   tactic.  This should be done in a special-purpose tactic.

 *)

#infix "o";;
#open "std";;
#open "more_util";;
#open "initial";;
#open "names";;
#open "vectops";;
#open "generic";;
#open "term";;
#open "reduction";;
#open "himsg";;
#open "mach";;
#open "proof";;
#open "proof_trees";;
#open "trad";;
#open "tacmach";;
#open "tactics";;
#open "pp";;
#open "stdpp";;
#open "clenv";;
#open "termenv";;
#open "tactics";;
#open "reintern0";;
#open "reintern1";;
#open "libobject";;
#open "library";;
#open "vernacinterp";;
#open "sopattern";;
#open "tactics1";;
#open "termast";;
#open "ast";;

#infix "THEN";;

(* two patterns - one for the type, and one for the type of the type *)
type destructor_pattern =
    {d_typ: constr; d_sort: constr};;

type ('a,'b) location = HYP of 'a | CONCL of 'b;;

(* hypothesis patterns might need to do matching on the conclusion, too.
 * conclusion-patterns only need to do matching on the hypothesis *)
type located_destructor_pattern ==
    (bool * destructor_pattern * destructor_pattern,
     destructor_pattern) location
;;

type destructor_data =
    {d_pat : located_destructor_pattern;
     d_bindings : (identifier * constr) list;
     d_pri : int;
     d_code : ast__spat} (* should be of phylum tactic *)
;;

let ri_dpat dp = {d_typ = ri_constr dp.d_typ;
                  d_sort = ri_constr dp.d_sort}
;;

let ri_ldpat = function
    HYP(b,p1,p2) -> HYP(b,ri_dpat p1,ri_dpat p2)
  | CONCL p -> CONCL(ri_dpat p)
;;

let ri_dd dd =
    {d_pat = ri_ldpat dd.d_pat;
     d_bindings = map (fun (id,c) -> (ri_id id,ri_constr c)) dd.d_bindings;
     d_pri = dd.d_pri;
     d_code = ast__ri_spat dd.d_code}
;;

type t == (identifier,destructor_data) nbtermdn__t;;
type frozen_t == (identifier,destructor_data) nbtermdn__frozen_t;;

let tactab = (nbtermdn__new {nbtermdn__name_ord = id_ord} : t);;

let lookup pat = nbtermdn__lookup tactab pat;;

let init () = nbtermdn__empty tactab;;

let freeze () = nbtermdn__freeze tactab;;
let unfreeze fs = nbtermdn__unfreeze fs tactab;;

let rollback f x =
    let fs = freeze()
    in try f x
       with e -> (unfreeze fs; raise e)
;;

let add (na,dd) =
let pat = match dd.d_pat with
    HYP(_,p,_) -> p.d_typ
  | CONCL p -> p.d_typ
in
    if nbtermdn__in_dn tactab na then
        (MSGNL [< 'S"Warning [Overriding Destructor Entry " ; 'S(string_of_id na) ; 'S"]" >];
         nbtermdn__remap tactab na (pat,dd))
    else nbtermdn__add tactab (na,(pat,dd));;

library__declare_summary "destruct-hyp-concl"
{summary__freeze_function = freeze;
 summary__unfreeze_function = unfreeze;
 summary__init_function = init;
 summary__sync_function = (fun () -> false)}
;;

let cache_dd (_,(na,dd)) =
    (try add (na,dd)
    with _ -> anomalylabstrm "dhyp__add"
              [< 'S"The code which adds destructor hints broke;" ;
                 'SPC ; 'S"this is not supposed to happen" >])
;;

let reintern_dd (na,dd) = (ri_id na,ri_dd dd);;

let specification_dd x = x;;

type destructor_data_object == identifier * destructor_data;;

let ((inDD:destructor_data_object->object),
      (outDD:object->destructor_data_object)) =
    declare_object ("DESTRUCT-HYP-CONCL-DATA",
                    {cache_function = cache_dd;
                     reintern_function = reintern_dd;
                     specification_function = specification_dd});;

let add_destructor_hint na pat pri code bindings =
    add_anonymous_object
    (inDD(na,{d_pat = pat;
              d_bindings = bindings;
              d_pri=pri;
              d_code=code}))
;;

vinterp_add
("HintDestruct",
fun ((VARG_IDENTIFIER na)::
     (VARG_ANY location)::
     (VARG_COMMAND patcom)::
     (VARG_NUMBER pri)::
     (VARG_SPAT tacexp)::
     bindings) ->
let bindings = map (fun (VARG_TACTIC_ARG(BINDING(Dep id,c))) -> (id,c)) bindings in
let loc =
    match location with
    ASTnode("CONCL",[]) -> CONCL()
  | ASTnode("DiscardableHYP",[]) -> HYP true
  | ASTnode("PreciousHYP",[]) -> HYP false in

fun () ->

let pat = raw_sopattern_of_compattern (initial_sign()) patcom in

let bindings =
    map (fun (id,c) ->
             (id,raw_sopattern_of_compattern (initial_sign()) c))
    bindings in

    add_destructor_hint na
    (match loc with
     HYP b ->
     HYP(b,{d_typ=pat;d_sort=DOP0(Meta(newMETA()))},
         {d_typ=DOP0(Meta(newMETA()));d_sort=DOP0(Meta(newMETA()))})
   | CONCL () ->
     CONCL({d_typ=pat;d_sort=DOP0(Meta(newMETA()))}))
    pri
    tacexp
    bindings)
;;

let match_dpat dp cls gls =
let cltyp = clause_type cls gls in
    match (cls,dp) with
    (SOME id,HYP(_,hypd,concld)) ->
    (somatch__somatch NONE hypd.d_typ cltyp)@
    (somatch__somatch NONE hypd.d_sort (pf_type_of gls cltyp))@
    (somatch__somatch NONE concld.d_typ (pf_concl gls))@
    (somatch__somatch NONE concld.d_sort (pf_type_of gls (pf_concl gls)))
  | (NONE,CONCL concld) ->
    (somatch__somatch NONE concld.d_typ (pf_concl gls))@
    (somatch__somatch NONE concld.d_sort (pf_type_of gls (pf_concl gls)))
  | _ -> error "ApplyDestructor"
;;

let ApplyDestructor cls discard dd gls =
let mvb = match_dpat dd.d_pat cls gls in

let astb =
    map
    (fun (id,c) ->
         let sub_c = (reduction__strong (reduction__whd_meta mvb) c) in
         let red_c = sosub__soexecute sub_c in
             (id,FRAGast(bdize (GLOB (pf_hyps gls)) red_c)))
    dd.d_bindings in

let astb = (match cls with
            SOME id -> (id_of_string "$0",FRAGast(ASTnvar id))::astb
          | NONE -> astb) in

let tcom = ast__ssub astb dd.d_code in

let Discard =
    match (cls,dd.d_pat) with
    (SOME id,HYP(discardable,_,_)) ->
    if discard & discardable then thin [id] else IDTAC
  | (NONE,CONCL _) -> IDTAC
  | _ -> error "ApplyDestructor" in

    match tcom with
    FRAGast ast -> (tacentries__vernac_interp ast THEN Discard) gls
  | _ -> errorlabstrm "ApplyDestructor"
    [< 'S"Malformed user tactic text" >]
;;

(* [DHyp id gls]

   will take an identifier, get its type, look it up in the
   discrimination net, get the destructors stored there, and then try
   them in order of priority.
 *)

let DestructHyp discard id gls =
let hyptyp = clause_type (SOME id) gls in
let ddl = map snd (lookup hyptyp) in
let sorted_ddl = sort (fun (dd1,dd2) -> dd1.d_pri > dd2.d_pri) ddl in
    FIRST (map (ApplyDestructor (SOME id) discard) sorted_ddl) gls
;;

let CDHyp id gls = DestructHyp true id gls;;
let DHyp id gls = DestructHyp false id gls;;

#open "tacinterp";;

tacinterp_add("DHyp",(fun [IDENTIFIER id] -> DHyp id));;
tacinterp_add("CDHyp",(fun [IDENTIFIER id] -> CDHyp id));;

(* [DConcl gls]

   will take a goal, get its concl, look it up in the
   discrimination net, get the destructors stored there, and then try
   them in order of priority.
 *)

let DConcl gls =
let ddl = map snd (lookup (pf_concl gls)) in
let sorted_ddl = sort (fun (dd1,dd2) -> dd1.d_pri > dd2.d_pri) ddl in
    FIRST (map (ApplyDestructor NONE false) sorted_ddl) gls
;;

tacinterp_add("DConcl",(fun [] -> DConcl));;

(* $Id: dhyp.ml,v 1.2 1996/02/12 15:55:57 cpaulin Exp $ *)
