(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                 gelim.ml                                 *)
(****************************************************************************)
#infix "o";;
#open "std";;
#open "initial";;
#open "more_util";;
#open "pp";;
#open "stdpp";;
#open "names";;
#open "generic";;
#open "term";;
#open "printer";;
#open "termenv";;
#open "reduction";;
#open "mach";;
#open "trad";;
#open "tacmach";;
#open "proof";;
#open "proof_trees";;
#open "clenv";;
#open "tactics";;
#open "lexer";;
#open "pattern";;
#open "tactics0";;
#open "tactics1";;
#open "tactics2";;
#open "tactics3";;
#open "equality";;

#infix "ORELSE";;
#infix "NEXT";;
#infix "THEN";;
#infix "THENF";;
#infix "THENL";;
#infix "THENS";;

#infix "THEN_i1";;

(* [make_inv_predicate (ity,args) C]
  
   is given the inductive type, its arguments, both the global
   parameters and its local arguments, and is expected to produce a
   predicate P such that if largs is the "local" part of the
   arguments, then (P largs) will be convertible with a conclusion of
   the form:

   <A1>a1=a1-><A2>a2=a2 ... -> C

   Algorithm: suppose length(largs)=n

   (1) Push the entire arity, [xbar:Abar], carrying along largs and
   the conclusion

   (2) Pair up each ai with its respective Rel version: a1==(Rel n),
   a2==(Rel n-1), etc.

   (3) For each pair, ai,Rel j, if the Ai is dependent - that is, the
   type of [Rel j] is an open term, then we construct the iterated
   tuple, [make_iterated_tuple] does it, and use that for our equation

   Otherwise, we just use <Ai>ai=Rel j

 *)

let named_push_and_liftl n env T l = pushrec n T (env,l)
    where rec pushrec = fun
    0 T (env,l) -> (env,T,l)
  | n (DOP2(Prod,T,DLAM(na,B))) (env,l) 
                -> pushrec (n-1) B (push_and_lift (named_hd T na,T) env l)
  | n (DOP2(Cast,T,_)) (env,l) -> pushrec n T (env,l)
  | _ _ _ -> error "push_and_liftl"
;;

let mmk = make_module_marker ["#Prelude.obj"];;
let eq_constant = put_pat mmk "eq";;

let make_inv_predicate sigma sign (ity,args) C =
let nparams = mind_nparams ity in
let (globargs,largs) = chop_list nparams args in
let arity = hnf_prod_applist sigma "make_inv_predicate" (mind_arity ity) globargs in
let len = length largs in
let env = [] and largs = (map insert_lifted largs) in
let env,_,largs = named_push_and_liftl len env arity largs in

(* Now the arity is pushed, and we need to construct the pairs
 * ai,Rel(n-i+1)
 *)
let larg_var_list = map_i (fun i ai -> insert_lifted(DOP2(Implicit,extract_lifted ai,Rel(len-i+1)))) 1 largs in

(* Now, we can recurse down this list, for each ai,(Rel k) whether to
   push <Ai>(Rel k)=ai (when   Ai is closed).

   In any case, we carry along the rest of larg_var_list *)

let rec build_concl (env,l) =
    match l with
    [] ->
    let neqns = ((length env)-len) in
    let env,C,[] = prod_and_popl neqns env C [] in
    let _,C,[] = lam_and_popl (length env) env C []
    in (C,neqns)

  | t::restlist ->
    let (DOP2(Implicit,ai,Rel k)) = extract_lifted t in
    let Tk = type_of_rel mt_evc (castify_env mt_evc (ENVIRON(sign,env))) (Rel k) in
    let (lhs,eqnty,rhs) =
        if closed Tk then
            (Rel k,Tk,ai)
        else
            make_iterated_tuple mt_evc (castify_env mt_evc (ENVIRON(sign,env)))
                                (ai,type_of mt_evc sign ai)
                                (Rel k,Tk) in
    let eqn = applist(get_pat eq_constant,[eqnty;lhs;rhs])

    in build_concl(push_and_lift (Anonymous,eqn) env restlist)

in

let (Predicate,neqns) = build_concl(env,larg_var_list) in

(* OK - this predicate should now be usable by res_elimination_then to
   do elimination on the conclusion. *)
              
    (Predicate,neqns)
;;

let Apply c gl =
    let (wc,KONT) = StartWalk gl in
    let clause = mk_clenv_hnf_constr_type_of wc c
    in res_pf KONT clause gl
;;

let Auto n = auto__automatic n;;

let res_elimination_then tac indbinding c gl =
let (wc,KONT) = StartWalk gl in

let t = strong_prodspine (pf_whd_betadeltaiota gl) (pf_type_of gl c) in

let indclause = mk_clenv_from wc (c,t) in

let indclause' = clenv_constrain_with_bindings indbinding indclause in

let newc = clenv_instance_template indclause' in

let (ity,args) = decomp_app (clenv_instance_template_type indclause') in

let (elim_predicate,neqns) =
    make_inv_predicate (Project gl) (pf_hyps gl) (ity,args) (pf_concl gl) in

let nparams = mind_nparams ity in

let largs = snd(chop_list nparams args) in

let cut_concl = applist(elim_predicate,largs) in

    (Cut cut_concl THENS
     [OnLastHyp (fun (SOME id) ->
                     ApplyUsing (applist(VAR id,
                                         tabulate_list (fun _ -> DOP0(Meta (newMETA()))) neqns)) THEN Auto 5);
      elimination_then_using (tac neqns) (SOME elim_predicate) ([],[]) newc]) gl
;;
  
let res_elim_then tac id = res_elimination_then tac [] (VAR id) ;;

(* The result of the elimination is a bunch of goals like:

           |- (cibar:Cibar)Equands->C

   where the cibar are either dependent or not.  We are fed a
   signature, with "true" for every recursive argument, and false for
   every non-recursive one.  So we need to do the
   sign_branch_len(sign) intros, thinning out all recursive
   assumptions.  This leaves us with exactly length(sign) assumptions.

   We save their names, and then do introductions for all the equands
   (there are some number of them, which is the other argument of the
   tactic)

   This gives us the #neqns equations, whose names we get also, and
   the #length(sign) arguments.

   Suppose that #nodep of these arguments are non-dependent.
   Generalize and thin them.

   This gives us #dep = #length(sign)-#nodep arguments which are
   dependent.

   Now, we want to take each of the equations, and do all possible
   injections to get the left-hand-side to be a variable.  At the same
   time, if we find a lhs/rhs pair which are different, we can
   discriminate them to prove false and finish the branch.

   Then, we thin away the equations, and do the introductions for the
   #nodep arguments which we generalized before.
 *)

let last_n_hyp_clauses n gl =
    tabulate_list (fun i -> nth_clause (-(i+1)) gl) n
;;

let var_occurs_in_pf gl id =
    occur_var id (pf_concl gl) or
    exists_sign (fun _ T -> occur_var id T) (pf_hyps gl)
;;

let split_dep_and_nodep idl gl =
    (filter (var_occurs_in_pf gl) idl,
     filter (neg (var_occurs_in_pf gl)) idl)
;;

(* Called after the case-assumptions have been killed off, and all the
   intros have been done.  Given that the clause in question is an
   equality (if it isn't we fail), we are responsible for projecting
   the equality, using Injection and Discriminate, and applying it to
   the concusion *)

let mmk = make_module_marker ["#Prelude.obj"];;
let eq_pattern = put_pat mmk "(eq ? ? ?)";;

let constructor_heads_match gls M N =
    match (pf_whd_betadeltaiota_stack gls M [],
           pf_whd_betadeltaiota_stack gls N []) with
    ((DOPN(MutConstruct(_,i),_),_),
     (DOPN(MutConstruct(_,j),_),_)) ->
    i=j 
  | _ -> error "ProjectAndApply"
;;

(* invariant: ProjectAndApply is responsible for erasing the clause
   which it is given as input *)
let ProjectAndApply cls gls =

let [T;t1;t2] = dest_match gls (clause_type cls gls) eq_pattern in

if (match strip_outer_cast t1 with VAR _ -> true | _ -> false) or
    (match strip_outer_cast t2 with VAR _ -> true | _ -> false) then
    ((TRY ((HypSubst (outSOME cls) NONE) ORELSE
           (RevHypSubst (outSOME cls) NONE))) THEN (ThinClause cls)) gls

else
let deq_trailer neqns =
    RepeatFor neqns (Intro THEN
                     OnLastHyp (fun (SOME id) -> TRY(HypSubst id NONE)) THEN
                     OnLastHyp ThinClause)

in (DEqThen deq_trailer cls THEN ThinClause cls) gls
;;

let case_trailer neqns ba gl =
let nargs = length ba.assums and
    argids = ba.assums in
let depids = filter ((var_occurs_in_pf gl)) argids and
    nodepids = filter ((neg (var_occurs_in_pf gl))) argids in
    (RepeatFor neqns Intro THEN
     BringHyps (map inSOME nodepids) THEN
     ThinClauses (map inSOME nodepids) THEN
     OnCL (rev o (NLastHyps neqns)) (BringHyps) THEN
     OnCL (NLastHyps neqns) ThinClauses THEN
     RepeatFor neqns
     (Intro THEN OnLastHyp (fun cls -> TRY (ProjectAndApply cls))) THEN
     RepeatFor (length nodepids) Intro THEN
     OnL (fun id -> TRY (ThinClause (SOME id))) depids) gl
;;

let Inv id gls =
    (res_elim_then (fun neqns -> IntroElimAssumsThen (case_trailer neqns)) id THEN
     ThinClause (SOME id)) gls;;

let Inv_tac = register_ident_tactic "GElim" (fun id gls -> Inv id gls);;

let half_case_trailer neqns ba gl =
let nargs = length ba.assums and
    argids = ba.assums in
let depids = filter ((var_occurs_in_pf gl)) argids and
    nodepids = filter ((neg (var_occurs_in_pf gl))) argids in
    (RepeatFor neqns Intro THEN
     BringHyps (map inSOME nodepids) THEN
     ThinClauses (map inSOME nodepids) THEN
     OnCL (rev o (NLastHyps neqns)) (BringHyps) THEN
     OnCL (NLastHyps neqns) ThinClauses THEN
     RepeatFor neqns Intro) gl
;;
let HalfInv id gls =
    (res_elim_then (fun neqns -> IntroElimAssumsThen (half_case_trailer neqns)) id) gls
;;

let HalfInv_tac = register_ident_tactic "HalfGElim" HalfInv;;

let IntrosReplacing ids gls = introrec ids gls
    where rec introrec = function
    [] -> IDTAC
  | id::tl -> ( IntroReplacing id
               ORELSE
                (IntroErasing id ORELSE IntroUsing id)
          )
              THEN
              introrec tl
;;

(* InvIn will bring the specified clauses into the conclusion, and then
 * perform inversion on the named hypothesis.  After, it will intro them
 * back to their places in the hyp-list.
 *)
let InvIn id ids gls =
    (BringHyps (map inSOME ids) THEN
     Inv id THEN
     IntrosReplacing ids) gls
;;

let InvIn_tac =
let gentac = register_tactic "GElimIn"
    (fun ((IDENTIFIER id)::hl) ->
         InvIn id (map (fun (IDENTIFIER id) -> id) hl))
    (fun sigma goal (_,(IDENTIFIER id)::hl) ->
         let hypl = map (fun (IDENTIFIER id) -> id) hl in
         [< 'S"Inv" ; 'SPC ; 'S(string_of_id id) ; 'SPC ; 'S"in" ; 'SPC ;
            prlist_with_sep (fun () -> [< 'SPC >])
                            (fun id -> [< 'S(string_of_id id) >]) hypl >])
in fun id hl -> gentac ((IDENTIFIER id)::(map (fun id -> (IDENTIFIER id)) hl))
;;


(* $Id: gelim.ml,v 1.3 1996/02/12 15:56:14 cpaulin Exp $ *)
