(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                  inv.ml                                  *)
(****************************************************************************)
#infix "o";;
#open "std";;
#open "initial";;
#open "more_util";;
#open "pp";;
#open "stdpp";;
#open "names";;
#open "generic";;
#open "term";;
#open "printer";;
#open "termenv";;
#open "reduction";;
#open "mach";;
#open "trad";;
#open "tacmach";;
#open "proof";;
#open "proof_trees";;
#open "clenv";;
#open "tactics";;
#open "lexer";;
#open "pattern";;
#open "tactics0";;
#open "tactics1";;
#open "tactics2";;
#open "tactics3";;
#open "equality";;

#infix "ORELSE";;
#infix "NEXT";;
#infix "THEN";;
#infix "THENF";;
#infix "THENL";;
#infix "THENS";;

#infix "THEN_i1";;


(* [make_inv_predicate (ity,args) C]
  
   is given the inductive type, its arguments, both the global
   parameters and its local arguments, and is expected to produce a
   predicate P such that if largs is the "local" part of the
   arguments, then (P largs) will be convertible with a conclusion of
   the form:

   <A1>a1=a1-><A2>a2=a2 ... -> C

   Algorithm: suppose length(largs)=n

   (1) Push the entire arity, [xbar:Abar], carrying along largs and
   the conclusion

   (2) Pair up each ai with its respective Rel version: a1==(Rel n),
   a2==(Rel n-1), etc.

   (3) For each pair, ai,Rel j, if the Ai is dependent - that is, the
   type of [Rel j] is an open term, then we construct the iterated
   tuple, [make_iterated_tuple] does it, and use that for our equation

   Otherwise, we just use <Ai>ai=Rel j

 *)

let named_push_and_liftl n env T l = pushrec n T (env,l)
    where rec pushrec = fun
    0 T (env,l) -> (env,T,l)
  | n (DOP2(Prod,T,DLAM(na,B))) (env,l) 
                -> pushrec (n-1) B (push_and_lift (named_hd T na,T) env l)
  | n (DOP2(Cast,T,_)) (env,l) -> pushrec n T (env,l)
  | _ _ _ -> error "push_and_liftl"
;;




let named_push_lambda_and_liftl n env T l = pushrec n T (env,l)
    where rec pushrec = fun
    0 T (env,l) -> (env,T,l)
  | n (DOP2(Lambda,T,DLAM(na,B))) (env,l) 
                -> pushrec (n-1) B (push_and_lift (named_hd T na,T) env l)
  | n (DOP2(Cast,T,_)) (env,l) -> pushrec n T (env,l)
  | _ _ _ -> error "push_and_liftl"
;;

let mmk = make_module_marker ["#Prelude.obj"];;
let mmk = make_module_marker ["#Logic_Type.obj"];;



let eq_constant = put_pat mmk "eq";;
let eq_pattern = put_pat mmk "(eq ? ? ?)";;

  
let eqT_constant = put_pat mmk "eqT";;
let eqT_pattern = put_pat mmk "(eqT ? ? ?)";;

let is_prop c = (whd_castapp c)=(DOP0 (Sort (Prop Null)));;

let is_set c = (whd_castapp c)=DOP0 (Sort (Prop Pos));;

let is_type c = match (whd_castapp c) with 
                 (DOP0 (Sort (Type(Null,_)))) -> true 
                | _ -> false 
;;

let find_eq_pattern ty = 
      if is_set ty
       then  eq_constant
       else  if is_type ty
              then eqT_constant
              else  errorlabstrm "make_inv_predicate" [< 'S "no primitive equality on proofs" >]
;;

let dest_match_eq gls eqn =
try dest_match gls eqn eq_pattern
 with _ -> dest_match gls eqn eqT_pattern
;;
(********** To generate dependent inversion 
let make_inv_predicate sigma sign (ity,args) C =
let nparams = mind_nparams ity in
let (globargs,largs) = chop_list nparams args in
let arity = hnf_prod_applist sigma "make_inv_predicate" (mind_arity ity) globargs in
let len = length largs in
let env = [] and largs = (map insert_lifted largs) in
let env,_,largs = named_push_and_liftl len env arity largs in

(* Now the arity is pushed, and we need to construct the pairs
 * ai,Rel(n-i+1)
 *)
let larg_var_list = map_i (fun i ai -> insert_lifted(DOP2(Implicit,extract_lifted ai,Rel(len-i+1)))) 1 largs in

(* Now, we can recurse down this list, for each ai,(Rel k) whether to
   push <Ai>(Rel k)=ai (when   Ai is closed).

   In any case, we carry along the rest of larg_var_list *)

let rec build_concl (env,l) =
    match l with
    [] ->
    let neqns = ((length env)-len) in
    let env,C,[] = prod_and_popl neqns env C [] in
    let _,C,[] = lam_and_popl (length env) env C []
    in (C,neqns)

  | t::restlist ->
    let (DOP2(Implicit,ai,Rel k)) = extract_lifted t in
    let Tk = type_of_rel mt_evc (castify_env mt_evc (ENVIRON(sign,env))) (Rel k) in
    let (lhs,eqnty,rhs) =
        if closed Tk then
            (Rel k,Tk,ai)
        else
            make_iterated_tuple mt_evc (castify_env mt_evc (ENVIRON(sign,env)))
                                (ai,type_of mt_evc sign ai)
                                (Rel k,Tk) in
    let type_type_rhs = type_of sigma sign (type_of sigma sign rhs) in
    let eqn = applist (get_pat (find_eq_pattern type_type_rhs),[eqnty;lhs;rhs])
    in build_concl(push_and_lift (Anonymous,eqn) env restlist)

in

let (Predicate,neqns) = build_concl(env,larg_var_list) in

(* OK - this predicate should now be usable by res_elimination_then to
   do elimination on the conclusion. *)
              
    (Predicate,neqns)
;;
**************)
let type_of_predicate_argument gls ity globargs =
  let sigma = Project gls in 
  let sort = sort_of_goal gls  in
  let elim = indrec__make_case_gen sigma ity sort in
  let type_elim = type_of  sigma (pf_hyps gls) elim in
  let nparams = mind_nparams ity in
  let (env,predicate,_) = named_push_and_liftl nparams [] type_elim [] in
  let (_,predicate,_) = lam_and_popl nparams env predicate [] in
  let prod = whd_beta (applist (predicate,globargs)) 
  in match prod with
      (DOP2(Prod,ty,_)) -> ty
;;


let make_inv_predicate  (ity,args) c dflt_concl dep_option gls  =
let sigma= Project gls in
let sign=pf_hyps gls in
let C=(pf_concl gls) in
let (VAR id)= c in 
let nparams = mind_nparams ity in
let (globargs,largs_init) = chop_list nparams args in
let arity = hnf_prod_applist sigma "make_inv_predicate" (mind_arity ity) globargs in
let len = length largs_init in
let env = [] and largs = (map insert_lifted largs_init) in
let (env,larg_var_list,C,dephyp) =                   
 if not dep_option (* (dependent (VAR id) C) *)
 then (* We push de arity and leave C unchanged *)
      let env_ar,_,largs_ar = named_push_and_liftl len env arity largs in
      let larg_var_list = map_i 
                (fun i ai -> insert_lifted(DOP2(Implicit,extract_lifted ai,Rel(len-i+1)))) 1 largs
      in  (env_ar,larg_var_list,C,0)
 else 
 if  not (dependent (VAR id) C) 
  then errorlabstrm "make_inv_predicate" [< 'S "Current goal does not depend on "; print_id id >]
  else   
      (* We abstract the conclusion of goal with respect to args and c to be C 
       *  in order to rewrite and have c also rewritten when the case will be done
       *)
      let P=type_of_predicate_argument gls ity globargs in
      let C2 = (match dflt_concl with
                 NONE -> abstract_list_all gls P C (largs_init@[c])  
               | (SOME concl) -> concl) in  
      let env,_,largs = named_push_lambda_and_liftl (nb_lam C2) env C2 largs in     
      let C3 = whd_beta (applist (C2,list_of_vect (rel_vect (length largs) (nb_prod arity +1)))) in
      
     let larg_var_list = map_i (fun i ai->insert_lifted(DOP2(Implicit,extract_lifted ai,Rel(len-i+2)))) 1 largs 
     in (env,larg_var_list,C3,1)

(* Now the arity is pushed, and we need to construct the pairs
 * ai,Rel(n-i+1)
 *)
(* Now, we can recurse down this list, for each ai,(Rel k) whether to
   push <Ai>(Rel k)=ai (when   Ai is closed).

   In any case, we carry along the rest of larg_var_list *)

in let rec build_concl (env,l) =
    match l with
    [] ->
    let neqns = ((length env-dephyp)-len) in
    let env,C,[] = prod_and_popl neqns env C [] in
    let _,C,[] = lam_and_popl (length env) env C []
    in (C,neqns)

  | t::restlist ->
    let (DOP2(Implicit,ai,Rel k)) = extract_lifted t in
    let Tk = type_of_rel mt_evc (castify_env mt_evc (ENVIRON(sign,env))) (Rel k) in
    let (lhs,eqnty,rhs) =
        if closed Tk then
            (Rel k,Tk,ai)
        else
            make_iterated_tuple mt_evc (castify_env mt_evc (ENVIRON(sign,env)))
                                (ai,type_of mt_evc sign ai)
                                (Rel k,Tk) in
    let type_type_rhs = type_of sigma sign (type_of sigma sign rhs) in
    let eqn = applist (get_pat (find_eq_pattern type_type_rhs),[eqnty;lhs;rhs])
    in build_concl(push_and_lift (Anonymous,eqn) env restlist)

in

let (Predicate,neqns) = build_concl(env,larg_var_list) in

(* OK - this predicate should now be usable by res_elimination_then to
   do elimination on the conclusion. *)
              
    (Predicate,neqns)
;;

let Apply c gl =
    let (wc,KONT) = StartWalk gl in
    let clause = mk_clenv_hnf_constr_type_of wc c
    in res_pf KONT clause gl
;;

let Auto n = auto__automatic n;;
(*** => A disappear in the future **
let res_elimination_then tac indbinding c gl =
let (wc,KONT) = StartWalk gl in

let t = strong_prodspine (pf_whd_betadeltaiota gl) (pf_type_of gl c) in

let indclause = mk_clenv_from wc (c,t) in

let indclause' = clenv_constrain_with_bindings indbinding indclause in

let newc = clenv_instance_template indclause' in

let (ity,args) = decomp_app (clenv_instance_template_type indclause') in

let (elim_predicate,neqns) =
    make_inv_predicate (Project gl) (pf_hyps gl) (ity,args) (pf_concl gl) in

let nparams = mind_nparams ity in

let largs = snd(chop_list nparams args) in

let cut_concl = applist(elim_predicate,largs) in

    (Cut cut_concl THENS
     [OnLastHyp (fun (SOME id) ->
                     ApplyUsing (applist(VAR id,
                                         tabulate_list (fun _ -> DOP0(Meta (newMETA()))) neqns)) THEN Auto 5);
      elimination_then_using (tac neqns) (SOME elim_predicate) ([],[]) newc]) gl
;;
  
let res_elim_then tac id = res_elimination_then tac [] (VAR id) ;;
**********)


let res_case_then tac indbinding c dflt_concl dep_option gl =
let (wc,KONT) = StartWalk gl in

let t = strong_prodspine (pf_whd_betadeltaiota gl) (pf_type_of gl c) in

let indclause = mk_clenv_from wc (c,t) in

let indclause' = clenv_constrain_with_bindings indbinding indclause in

let newc = clenv_instance_template indclause' in

let (ity,args) = decomp_app (clenv_instance_template_type indclause') in

let (elim_predicate,neqns) = make_inv_predicate (ity,args) c dflt_concl dep_option gl in

let nparams = mind_nparams ity in

let largs = snd(chop_list nparams args) in

let (cut_concl,case_tac) = if dep_option & (dependent c (pf_concl gl))
   then applist(elim_predicate,largs@[c]),case_then_using 
   else applist(elim_predicate,largs),case_nodep_then_using 
in
    (Cut cut_concl THENS
     [OnLastHyp (fun (SOME id) ->
                     ApplyUsing (applist(VAR id,
                                         tabulate_list (fun _ -> DOP0(Meta (newMETA()))) neqns)) THEN Auto 5);
      case_tac (tac neqns) (SOME elim_predicate) ([],[]) newc]) gl
;;

let case_then dflt_concl dep_option tac id  = res_case_then tac [] (VAR id) dflt_concl dep_option ;;




(* The result of the elimination is a bunch of goals like:

           |- (cibar:Cibar)Equands->C

   where the cibar are either dependent or not.  We are fed a
   signature, with "true" for every recursive argument, and false for
   every non-recursive one.  So we need to do the
   sign_branch_len(sign) intros, thinning out all recursive
   assumptions.  This leaves us with exactly length(sign) assumptions.

   We save their names, and then do introductions for all the equands
   (there are some number of them, which is the other argument of the
   tactic)

   This gives us the #neqns equations, whose names we get also, and
   the #length(sign) arguments.

   Suppose that #nodep of these arguments are non-dependent.
   Generalize and thin them.

   This gives us #dep = #length(sign)-#nodep arguments which are
   dependent.

   Now, we want to take each of the equations, and do all possible
   injections to get the left-hand-side to be a variable.  At the same
   time, if we find a lhs/rhs pair which are different, we can
   discriminate them to prove false and finish the branch.

   Then, we thin away the equations, and do the introductions for the
   #nodep arguments which we generalized before.
 *)

let last_n_hyp_clauses n gl =
    tabulate_list (fun i -> nth_clause (-(i+1)) gl) n
;;

let var_occurs_in_pf gl id =
    occur_var id (pf_concl gl) or
    exists_sign (fun _ T -> occur_var id T) (pf_hyps gl)
;;

let split_dep_and_nodep idl gl =
    (filter (var_occurs_in_pf gl) idl,
     filter (neg (var_occurs_in_pf gl)) idl)
;;

(* Called after the case-assumptions have been killed off, and all the
   intros have been done.  Given that the clause in question is an
   equality (if it isn't we fail), we are responsible for projecting
   the equality, using Injection and Discriminate, and applying it to
   the concusion *)

let constructor_heads_match gls M N =
    match (pf_whd_betadeltaiota_stack gls M [],
           pf_whd_betadeltaiota_stack gls N []) with
    ((DOPN(MutConstruct(_,i),_),_),
     (DOPN(MutConstruct(_,j),_),_)) ->
    i=j
  | _ -> error "ProjectAndApply"
;;


let rec dependent_hyps id idlist sign= dep_rec idlist 
where rec dep_rec =fun
  [] -> []
| (id1::l) -> let id1ty= (snd (lookup_sign id1 sign))
              in  if occur_var id id1ty then id1::dep_rec l else dep_rec l 
;;


let IntrosReplacing ids gls = introrec ids gls
    where rec introrec = function
    [] -> IDTAC
  | id::tl -> ( IntroReplacing id
               ORELSE
                (IntroErasing id ORELSE IntroUsing id)
          )
              THEN
              introrec tl
;;

(* Computes the subset of hypothesis in the local context whose
   type depends on t (should be of the form (VAR id)), then
   it generalizes them, applies tac to rewrite all occurrencies of  t, and introduces
   generalized hypotheis.
   Precondition: t=(VAR id)
*)
let GeneralizeRewriteIntros  tac depids t gls = 
  let dids=dependent_hyps (id_of_var t) depids (pf_hyps gls)
   in (BringHyps (map inSOME dids)   THEN 
       tac THEN
       IntrosReplacing dids) gls ;;


(* invariant: ProjectAndApply is responsible for erasing the clause
   which it is given as input
   It simplifies the clause (an equality) to use it as a rewrite rule and then
   erases the result of the simplification.
*)

let ProjectAndApply cls depids gls =

let [T;t1;t2] = dest_match_eq gls (clause_type cls gls)  in

let subst_hyp_LR cls =  (TRY(HypSubst (outSOME cls)  NONE)) in
let subst_hyp_RL cls  = (TRY(RevHypSubst (outSOME cls) NONE)) in

let is_var t  = (match strip_outer_cast t with VAR _ -> true | _ -> false) in

match (is_var t1), (is_var t2) with
 true,  _  -> GeneralizeRewriteIntros (subst_hyp_LR cls THEN ThinClause cls) depids t1 gls
| _ , true -> GeneralizeRewriteIntros (subst_hyp_RL cls THEN ThinClause cls) depids t2 gls
| _ ,  _   -> 
   let subst_hyp gls = OnLastHyp orient_rule   gls 
     where  orient_rule  cls  = 
         let [T;t1;t2] = dest_match_eq gls (clause_type cls gls)  
         in if is_var t1  then  GeneralizeRewriteIntros  (subst_hyp_LR cls) depids t1   
             else  if is_var t2   
                     then GeneralizeRewriteIntros  (subst_hyp_RL cls) depids t2 
                     else subst_hyp_RL cls
in 

   let deq_trailer neqns =
      RepeatFor neqns (Intro THEN
                       subst_hyp  THEN
                       OnLastHyp ThinClause)

    in  (TRY (DEqThen deq_trailer cls) THEN ThinClause cls) gls
;;





(* invariant: ProjectAndApplyNoThining simplifies the clause (an equality) .
   If it can discriminate then the goal is proved, if not tries to use it as
   a rewrite rule. It erases the clause which is given as input
   
*)
let ProjectAndApplyNoThining cls depids gls =

let [T;t1;t2] = dest_match_eq gls (clause_type cls gls)  in

let subst_hyp_LR cls  = TRY(HypSubst (outSOME cls)  NONE) in
let subst_hyp_RL cls  = TRY(RevHypSubst (outSOME cls) NONE) in

let is_var t  = (match strip_outer_cast t with VAR _ -> true | _ -> false) in

match (is_var t1), (is_var t2) with
 true,  _  -> GeneralizeRewriteIntros (subst_hyp_LR cls) depids t1 gls
| _ , true -> GeneralizeRewriteIntros (subst_hyp_RL cls) depids t2 gls
| _ ,  _   -> 
   let subst_hyp gls = OnLastHyp orient_rule  gls 
     where  orient_rule  cls = 
         let [T;t1;t2] = dest_match_eq gls (clause_type cls gls) 
         in if is_var t1  then  GeneralizeRewriteIntros (subst_hyp_LR cls) depids t1 
             else if is_var t2 then GeneralizeRewriteIntros (subst_hyp_RL cls) depids t2
                   else subst_hyp_RL cls
 in 
   let deq_trailer neqns = RepeatFor neqns (Intro THEN subst_hyp)
   in  (DEqThen deq_trailer cls THEN ThinClause cls) gls
;;


(*
let case_trailer neqns ba gl =
let nargs = length ba.assums and
    argids = ba.assums in
let depids = filter ((var_occurs_in_pf gl)) argids and
    nodepids = filter ((neg (var_occurs_in_pf gl))) argids in
  (*  prgls gl; *)
    ((RepeatFor neqns Intro THEN 
      BringHyps (map inSOME nodepids) THEN (fun gls -> prgls gls ; IDTAC gls)) THEN
     ThinClauses (map inSOME nodepids) THEN 
     OnCL (rev o (NLastHyps neqns)) (BringHyps) THEN
     OnCL (NLastHyps neqns) ThinClauses THEN
     RepeatFor neqns
     (Intro THEN (fun gls -> prgls gls; IDTAC gls) THEN OnLastHyp (fun cls->TRY (ProjectAndApply cls depids))) THEN 
     (fun gls -> (PP [<'INT length depids>]; prgls gls;
     RepeatFor (length nodepids) Intro  gls ))THEN
     OnL (fun id -> TRY (ThinClause (SOME id))) depids) gl
;;
 *)

(* *)

let case_trailer neqns ba gl =
let nargs = length ba.assums and
    argids = ba.assums in
let depids = filter ((var_occurs_in_pf gl)) argids and
    nodepids = filter ((neg (var_occurs_in_pf gl))) argids in
    (RepeatFor neqns Intro THEN
     BringHyps (map inSOME nodepids) THEN
     ThinClauses (map inSOME nodepids) THEN
     OnCL (rev o (NLastHyps neqns)) (BringHyps) THEN
     OnCL (NLastHyps neqns) ThinClauses THEN
     RepeatFor neqns
     (Intro THEN OnLastHyp (fun cls -> TRY (ProjectAndApply cls depids))) THEN
     RepeatFor (length nodepids) Intro THEN
     OnL (fun id -> TRY (ThinClause (SOME id))) depids) gl
;;
(* *)


let case_trailer_nothining neqns ba gl =
let nargs = length ba.assums and
    argids = ba.assums in
let depids = filter ((var_occurs_in_pf gl)) argids and
    nodepids = filter ((neg (var_occurs_in_pf gl))) argids in
    (RepeatFor neqns Intro THEN
     BringHyps (map inSOME nodepids) THEN
     ThinClauses (map inSOME nodepids) THEN
     OnCL (rev o (NLastHyps neqns)) (BringHyps) THEN
     OnCL (NLastHyps neqns) ThinClauses THEN
     RepeatFor neqns
     (Intro THEN OnLastHyp (fun cls -> TRY (ProjectAndApplyNoThining cls depids))) THEN
     RepeatFor (length nodepids) Intro THEN
     OnL (fun id -> TRY (ThinClause (SOME id))) depids) gl
;;


let not_found_message  ids =
     [<'S "During the relocation of global references, the variable(s) ["; 
        'SPC ; prlist (fun id -> [<'S (string_of_id id) ; 'SPC >]) ids;
        'S" ] were not found in the current environment" >]
;;


let no_generalisation = [< 'S "Cannot find a well typed generalisation of the goal">];;



let dep_prop_prop_message id =
 [< 'S "Inversion on "; print_id id ; 'S " needs dependent elimination Prop-Prop and it is not available" >]
;; 


let Inv dflt_concl dep_option id gls =
 try((case_then dflt_concl dep_option 
             (fun neqns -> IntroCaseAssumsThen (case_trailer neqns)) id 
      THEN (TRY (ThinClause (SOME id))))  gls)

 with UserError ("abstract_list_all",_) -> errorlabstrm "Inv" no_generalisation
 |   UserError ("Case analysis",s) -> errorlabstrm "Inv needs Nodep Prop Set" s
 |   UserError (a,b)  -> errorlabstrm "Inv" b
 |   Invalid_argument "it_list2" -> 
           errorlabstrm "Inv" (dep_prop_prop_message id) 
 |   Not_found  ->  errorlabstrm "Inv" (not_found_message [id]) 
;;


(* Inv generates nodependent inversion lemma *)

let Inv_tac = register_ident_tactic "Inv" (fun id gls -> Inv NONE false id gls);;

let DInv_tac = register_ident_tactic "DInv" (fun id gls -> Inv NONE true id gls);;



(* Generates a dependent inversion lemma with a certain generalisation 
  of the goal  
*)

let Inv_with = 
 let gentac= 
  register_tactic "DInvWith"
  (fun [IDENTIFIER id;COMMAND com] gls ->
             Inv (SOME (pf_constr_of_com gls com)) true id gls)
  (fun sigma gl (_,[IDENTIFIER id;COMMAND com]) ->
     [< 'S"Dependent Inversion_clear" ; 'SPC ; print_id id ; 'SPC ;'S "with"; 'SPC; pr_com sigma gl com >])
in fun id com -> gentac [IDENTIFIER id;COMMAND com]
;;


(* Behaves as Inv but does not erase the derived equalities *)

let InvNoThining dflt_concl dep_option id gls =

try((case_then dflt_concl dep_option
       (fun neqns -> IntroCaseAssumsThen (case_trailer_nothining neqns)) id )
       gls)
with UserError ("abstract_list_all",_) -> errorlabstrm "Inv" no_generalisation
 |   UserError ("Case analysis",s) -> errorlabstrm "Inv needs Nodep Prop Set" s
 |   UserError (a,b) -> errorlabstrm "InvNoThining" b
 |   Invalid_argument "it_list2" -> errorlabstrm "InvNoThining" (dep_prop_prop_message id) 
 |   Not_found  ->  errorlabstrm "Inv" (not_found_message [id])
;;


(* generates a non dependent inversion lemma *)


let InvNoThining_tac = 
    register_ident_tactic "InvNoThining" (fun id gls -> InvNoThining  NONE false id gls);;

(* performs dependent inversion trying to generalise the current goal *)
let DInvNoThining_tac = 
    register_ident_tactic "DInvNoThining" (fun id gls -> InvNoThining  NONE true id gls);;


(* generates a dependent inversion using a given generalisation of the goal *)
let DInvNoThining_with = 
 let gentac= 
  register_tactic "DInvNoThiningWith"
  (fun [IDENTIFIER id;COMMAND com] gls ->
             InvNoThining(SOME (pf_constr_of_com gls com)) true id gls)
  (fun sigma gl (_,[IDENTIFIER id;COMMAND com]) ->
     [< 'S"Dependent Inversion" ; 'SPC ; print_id id ; 'SPC ;'S "with"; 'SPC; pr_com sigma gl com >])
in fun id com -> gentac [IDENTIFIER id;COMMAND com]
;;





let half_case_trailer neqns ba gl =
let nargs = length ba.assums and
    argids = ba.assums in
let depids = filter ((var_occurs_in_pf gl)) argids and
    nodepids = filter ((neg (var_occurs_in_pf gl))) argids in
    (RepeatFor neqns Intro THEN
     BringHyps (map inSOME nodepids) THEN
     ThinClauses (map inSOME nodepids) THEN
     OnCL (rev o (NLastHyps neqns)) (BringHyps) THEN
     OnCL (NLastHyps neqns) ThinClauses THEN
     RepeatFor neqns Intro) gl
;;


let HalfInv id gls =
try ((case_then NONE false 
             (fun neqns -> IntroCaseAssumsThen (half_case_trailer neqns)) id) 
             gls )
with UserError ("Case analysis",s) -> errorlabstrm "Inv needs Nodep Prop Set" s
 |   Invalid_argument "it_list2" -> 
       errorlabstrm "HalfInv" (dep_prop_prop_message id) 
 |   UserError (a,b) -> error "HalfInv" b
 |   Not_found  ->  errorlabstrm "HalfInv" (not_found_message [id])
;;

let HalfInv_tac = register_ident_tactic "HalfInv" HalfInv;;

(* InvIn will bring the specified clauses into the conclusion, and then
 * perform inversion on the named hypothesis.  After, it will intro them
 * back to their places in the hyp-list.
 *)

let length_sign gls = length (fst (pf_hyps gls));; 

let InvIn id ids gls =
  let nb_prod_init = nb_prod (pf_concl gls) in
   let intros_replace_ids gls =
      let nb_of_new_hyp  = nb_prod (pf_concl gls) - (length ids + nb_prod_init)
      in if nb_of_new_hyp < 1  then IntrosReplacing ids gls
            else (DO nb_of_new_hyp Intro THEN IntrosReplacing ids) gls
   in try 
      ((BringHyps (map inSOME ids) THEN
        Inv NONE false id THEN
        intros_replace_ids) gls)
      with  Not_found -> errorlabstrm "InvIn" (not_found_message ids)
         |  UserError ("Case analysis",s) -> 
                   errorlabstrm "Inv needs Nodep Prop Set" s
         |  UserError(a,b) -> errorlabstrm "InvIn" b  
         |  Invalid_argument "it_list2" -> 
                   errorlabstrm "InvIn" (dep_prop_prop_message id) 

;;

let InvIn_tac =
let gentac = register_tactic "InvIn"
    (fun ((IDENTIFIER id)::hl) ->
         InvIn id (map (fun (IDENTIFIER id) -> id) hl))
    (fun sigma goal (_,(IDENTIFIER id)::hl) ->
         let hypl = map (fun (IDENTIFIER id) -> id) hl in
         [< 'S"Inv" ; 'SPC ; 'S(string_of_id id) ; 'SPC ; 'S"in" ; 'SPC ;
            prlist_with_sep (fun () -> [< 'SPC >])
                            (fun id -> [< 'S(string_of_id id) >]) hypl >])
in fun id hl -> gentac ((IDENTIFIER id)::(map (fun id -> (IDENTIFIER id)) hl))
;;


let InvInNoThining id ids gls =
 let nb_prod_init = nb_prod (pf_concl gls) in
 let intros_replace_ids gls =
   let nb_of_new_hyp  = nb_prod (pf_concl gls) - (length ids + nb_prod_init)
        in if nb_of_new_hyp < 1  then IntrosReplacing ids gls
            else (DO nb_of_new_hyp Intro THEN IntrosReplacing ids) gls
   in try 
       ((BringHyps (map inSOME ids) THEN
         InvNoThining NONE false id THEN
         intros_replace_ids) gls)
      with Not_found -> errorlabstrm "InvInNoThinining" (not_found_message ids)
         |  UserError ("Case analysis",s) -> 
                  errorlabstrm "Inv needs Nodep Prop Set" s 
         |  UserError(a,b) -> errorlabstrm "InvInNoThining" b  
         |  Invalid_argument "it_list2" -> errorlabstrm "InvInNoThining" (dep_prop_prop_message id) 
          
;;

let InvInNoThining_tac =
let gentac = register_tactic "InvInNoThining"
    (fun ((IDENTIFIER id)::hl) ->
         InvInNoThining id (map (fun (IDENTIFIER id) -> id) hl))
    (fun sigma goal (_,(IDENTIFIER id)::hl) ->
         let hypl = map (fun (IDENTIFIER id) -> id) hl in
         [< 'S"InvNoThining" ; 'SPC ; 'S(string_of_id id) ; 'SPC ; 'S"in" ; 'SPC ;
            prlist_with_sep (fun () -> [< 'SPC >])
                            (fun id -> [< 'S(string_of_id id) >]) hypl >])
in fun id hl -> gentac ((IDENTIFIER id)::(map (fun id -> (IDENTIFIER id)) hl))
;;











(* $Id: inv.ml,v 1.10 1996/02/12 15:56:32 cpaulin Exp $ *)
