(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                leminv.ml                                 *)
(****************************************************************************)
#infix "o";;
#open "std";;
#open "initial";;
#open "more_util";;
#open "pp";;
#open "stdpp";;
#open "names";;
#open "generic";;
#open "term";;
#open "printer";;
#open "termenv";;
#open "reduction";;
#open "mach";;
#open "trad";;
#open "tacmach";;
#open "proof";;
#open "proof_trees";;
#open "clenv";;
#open "declare";;
#open "tactics";;
#open "lexer";;
#open "pattern";;
#open "tactics0";;
#open "tactics1";;
#open "tactics2";;
#open "tactics3";;
#open "equality";;
#open "inv";;

#infix "ORELSE";;
#infix "NEXT";;
#infix "THEN";;
#infix "THENF";;
#infix "THENL";;
#infix "THENS";;

#infix "THEN_i1";;


let prop_sort = (DOP0 (Sort (Prop Null)));;

let not_found_message  ids =
     [<'S "During the relocation of global references, the variable(s) ["; 
        'SPC ; prlist (fun id -> [<'S (string_of_id id) ; 'SPC >]) ids;
        'S" ] were not found in the current environment" >]
;;


let not_work_message = "tactic fails to build the inversion lemma, may be because the predicate has arguments that depend on other arguments";;


(* Inversion stored in lemmas *)

(* ALGORITHM:

   An inversion stored in a lemma is computed from a term-pattern, in
   a signature, as follows:

   Suppose we have an inductive relation, (I abar), in a signature Gamma:

       Gamma |- (I abar)

   Then we compute the free-variables of abar.  Suppose that Gamma is
   thinned out to only include these.

   [We need technically to require that all free-variables of the
    types of the free variables of abar are themselves free-variables
    of abar.  This needs to be checked, but it should not pose a
    problem - it is hard to imagine cases where it would not hold.]

   Now, we pose the goal:

       (P:(Gamma)Prop)(Gamma)(I abar)->(P vars[Gamma]).

   We execute the tactic:

   REPEAT Intro THEN (OnLastHyp (Inv NONE false o outSOME))

   This leaves us with some subgoals.  All the assumptions after "P"
   in these subgoals are new assumptions.  I.e. if we have a subgoal,

       P:(Gamma)Prop, Gamma, Hbar:Tbar |- (P ybar)

   then the assumption we needed to have was

      (Hbar:Tbar)(P ybar)

   So we construct all the assumptions we need, and rebuild the goal
   with these assumptions.  Then, we can re-apply the same tactic as
   above, but instead of stopping after the inversion, we just apply
   the respective assumption in each subgoal.

 *)


let thin_hyps_to_term (hyps,T) =
let vars = (global_vars T) in
    rev_sign(fst(it_sign (fun ((hyps,globs) as sofar) id a ->
                         if mem id globs then
                             (add_sign (id,a) hyps,(global_vars a)@globs)
                         else sofar) (nil_sign,vars) hyps))
;;
(********* to parameterize the sort
let compute_first_inversion_scheme sign I =
let (ity,largs) = find_Inductype mt_evc I in
let fv = global_vars I in
let thin_sign = thin_hyps_to_term (sign,I) in

if not(listset__equal fv (ids_of_sign thin_sign)) then
    errorlabstrm "lemma_inversion"
    [< 'S"Cannot compute lemma inversion when there are" ; 'SPC ;
       'S"free variables in the types of an inductive" ; 'SPC ;
       'S"which are not free in its instance" >];

let [P] = get_new_ids [id_of_string "P"] sign in

let Pty = it_sign (fun B id ty -> produit id ty B)
                  (DOP0(Sort (Prop Null))) thin_sign in

let PHead = applist(VAR P,rev(map (fun id -> VAR id) (ids_of_sign thin_sign))) in

let Goal = arrow I PHead

in (prepend_sign thin_sign (add_sign (P,Pty) nil_sign),
    Goal)
;;
*********)
let rel_of_env env = rel_rec env 1
where rec rel_rec  = fun 
  [ ]  _ -> []
| (_::env) n -> (Rel n)::(rel_rec env (n+1))
;;

let build_app op env = applist (op, rev (rel_of_env env));;

(* similar to prod_and_pop, but gives [na:T]B intead of (na:T)B *)

let prod_and_pop_named = fun
    [] body l acc_ids -> error "lam_and_pop"
  | ((na,T)::tlenv) body l acc_ids -> 
     let (Name id)=if na=Anonymous 
                    then Name(next_ident_away (id_of_string "a") acc_ids)
                    else na
     in (tlenv,DOP2(Prod,T,DLAM((Name id),body)),
         map (fun (0,x) -> (0,lift (-1) x)
              | (n,x) -> (n-1,x)) l,
         (id::acc_ids))
         
;;

(* similar to prod_and_popl but gives [nan:Tan]...[na1:Ta1]B instead of
 * (nan:Tan)...(na1:Ta1)B  it generates new names with respect to l
    whenever nai=Anonymous *)

let prod_and_popl_named  n env T l = poprec n (env,T,l,[])
    where rec poprec = fun
    0 (env,B,l,_) -> (env,B,l)
  | n ([],_,_,_) -> error "lam_and_popl"
  | n (env,B,l,acc_ids) -> poprec (n-1) (prod_and_pop_named env B l acc_ids)
;;

(* [dep_option] indicates wether the inversion lemma is dependent or not.
   If it is dependent and I is of the form (x_bar:T_bar)(I t_bar) then
   the stated goal will be (x_bar:T_bar)(H:(I t_bar))(P t_bar H) 
   where P:(x_bar:T_bar)(H:(I t_bar))[sort]            .
   The generalisation of such a goal at the moment of the dependent case should
   be easy

   If it is non dependent, then if [I]=(I t_bar) and (x_bar:T_bar) are thte
   variables occurring in [I], then the stated goal will be:
    (x_bar:T_bar)(I t_bar)->(P x_bar) 
   where P: P:(x_bar:T_bar)(H:(I t_bar)->[sort]     
*)
let compute_first_inversion_scheme sign I sort dep_option =
let (ity,largs) = find_Inductype mt_evc I in
let ar = mind_arity ity in
let fv = global_vars I in
let thin_sign = thin_hyps_to_term (sign,I) in

if not(listset__equal fv (ids_of_sign thin_sign)) then
    errorlabstrm "lemma_inversion"
    [< 'S"Cannot compute lemma inversion when there are" ; 'SPC ;
       'S"free variables in the types of an inductive" ; 'SPC ;
       'S"which are not free in its instance" >];

let [P] = get_new_ids [id_of_string "P"] sign in

let (Pty,Goal) =
 if dep_option   
  then let (env,_,_) = push_and_liftl (nb_prod ar) []  ar [] in
       let [H] = get_new_ids [id_of_string "H"] sign in 
       let (env1,_)= push_and_lift (Name H, (build_app ity env)) env [] in
       let (_,Pty,_) = prod_and_popl_named (length env1) env1 sort [] in 
       let PHead= applist(VAR P, largs@[Rel 1]) 
       in  (Pty, prod_name(Name H,I,PHead))

  else let PHead= applist(VAR P, rev(map (fun id -> VAR id) (ids_of_sign thin_sign))) 
       in (it_sign (fun B id ty -> produit id ty B) sort thin_sign,
            arrow I PHead) 
    
in (prepend_sign thin_sign (add_sign (P,Pty) nil_sign),
    Goal)
;;





(* [inversion_scheme sign I]

   Given a local signature, [sign], and an instance of an inductive
   relation, [I], inversion_scheme will prove the associated inversion
   scheme on sort [sort]. Depending on the value of [dep_option] it will
   build a dependent lemma or a non-dependent one

 *)

let inversion_scheme sign I sort dep_option inv_op =
let (I,sign) = add_prods_sign mt_evc (I,sign) in
let (InvSign,InvGoal) = compute_first_inversion_scheme sign I sort dep_option in

let InvSign = castify_sign mt_evc InvSign in

if not(listset__subset (global_vars InvGoal) (ids_of_sign InvSign)) then
    errorlabstrm "lemma_inversion"
    [< 'S"Computed inversion goal was not closed in initial signature" >];

let InvGoalj = fexecute mt_evc InvSign InvGoal in

let pfs = mk_pftreestate (mkGOAL mt_lc mt_ctxt InvSign (j_val InvGoalj)) in

let pfs = solve_pftreestate (Intro THEN (OnLastHyp (inv_op o outSOME))) pfs in

let pf = proof_of_pftreestate pfs in

let (pfterm,meta_types) = refiner__extract_open_proof pf.goal.hyps pf in

let InvSign =
    sign_it
    (fun id ty sign ->
         if mem_sign (initial_sign()) id then sign
         else add_sign (id,ty) sign)
    InvSign
    nil_sign in

let (InvSign,mvb) =
    it_list
    (fun (sign,mvb) (mv,mvty) ->
         let [H] = get_new_ids [id_of_string "H"] sign
         in (add_sign (H,mvty) sign,
             (mv,((VAR H):constr))::mvb))
    (InvSign,[])
    meta_types in

let InvProof =
    it_sign (fun B id ty -> lambda id ty B)
            (strong (whd_meta mvb) pfterm) InvSign in

InvProof

;;

#open "discharge";;
#open "libobject";;

let add_inversion_lemma name sign I sort dep_option inv_op =
 let InvProof = inversion_scheme sign I sort dep_option inv_op
 in machine_constant_verbose (initial_assumptions())
                      ((name,false,NeverDischarge),InvProof)
;;



#open "pfedit";;

(* inv_op = Inv (derives de complete inv. lemma)
 * inv_op = InvNoThining (derives de semi inversion lemma)
 *)

let inversion_lemma_from_goal n na id sort dep_option inv_op =
let pts = get_pftreestate() in
let pf = proof_of_pftreestate pts in
let gll,_ = frontier pf in
let gl = nth gll n in
   add_inversion_lemma na gl.hyps (snd(lookup_sign id gl.hyps)) sort dep_option inv_op
;;

#open "vernacinterp";;

vinterp_add
("MakeInversionLemmaFromHyp",
 fun [VARG_NUMBER n;
      VARG_IDENTIFIER na;
      VARG_IDENTIFIER id] ->
 fun () ->
     inversion_lemma_from_goal n na id prop_sort false (Inv NONE false))
;;

(*
vinterp_add
("MakeInversionLemma",
 fun [VARG_IDENTIFIER na;
      VARG_COMMAND com] ->
 fun () ->
     add_inversion_lemma na (initial_sign())
                            (constr_of_com mt_evc (GLOB(initial_sign())) com) prop_sort false (Inv NONE false))
;;
*)
vinterp_add
("MakeInversionLemma",
 fun [VARG_IDENTIFIER na;
      VARG_COMMAND com;
      VARG_COMMAND sort] ->
 fun () ->
     add_inversion_lemma na (initial_sign())
                            (constr_of_com mt_evc (GLOB(initial_sign())) com) 
                            (constr_of_com mt_evc (GLOB(initial_sign())) sort)
                            false (Inv NONE false))
;;

vinterp_add
("MakeSemiInversionLemmaFromHyp",
 fun [VARG_NUMBER n;
      VARG_IDENTIFIER na;
      VARG_IDENTIFIER id] ->
 fun () ->
     inversion_lemma_from_goal n na id prop_sort false (InvNoThining NONE false))
;;

(*
vinterp_add
("MakeSemiInversionLemma",
 fun [VARG_IDENTIFIER na;
      VARG_COMMAND com;] ->
 fun () ->
     add_inversion_lemma na (initial_sign())
                            (constr_of_com mt_evc (GLOB(initial_sign())) com) prop_sort  false (InvNoThining NONE false))
;;
*)
vinterp_add
("MakeSemiInversionLemma",
 fun [VARG_IDENTIFIER na;
      VARG_COMMAND com;
      VARG_COMMAND sort] ->
 fun () ->
     add_inversion_lemma na (initial_sign())
                            (constr_of_com mt_evc (GLOB(initial_sign())) com) 
                            (constr_of_com mt_evc (GLOB(initial_sign())) sort)
                            false (InvNoThining NONE false))
;;


vinterp_add
("MakeDependentInversionLemma",
 fun [VARG_IDENTIFIER na;
      VARG_COMMAND com;
      VARG_COMMAND sort] ->
 fun () ->
     add_inversion_lemma na (initial_sign())
                            (constr_of_com mt_evc (GLOB(initial_sign())) com) 
                            (constr_of_com mt_evc (GLOB(initial_sign())) sort)  
                            true (Inv NONE true))
;;


vinterp_add
("MakeDependentSemiInversionLemma",
 fun [VARG_IDENTIFIER na;
      VARG_COMMAND com;
      VARG_COMMAND sort] ->
 fun () ->
     add_inversion_lemma na (initial_sign())
                            (constr_of_com mt_evc (GLOB(initial_sign())) com) 
                            (constr_of_com mt_evc (GLOB(initial_sign())) sort)  
                            true (InvNoThining NONE true))
;;


(* ================================= *)
(* Applying a given inversion lemma  *)
(* ================================= *)


let LemInv id c gls =
try (
let (wc,KONT) = StartWalk gls in
let clause = mk_clenv_type_of wc c in
let clause = clenv_constrain_with_bindings [(ABS (-1),VAR id)] clause
in res_pf KONT clause gls )
with Not_found  ->  errorlabstrm "LemInv" (not_found_message [id])
  |  UserError (a,b) -> errorlabstrm "LemInv" [< 'S "Cannot refine current goal with the lemma "; term0 (GLOB (initial_sign())) c  >] 
;;


let UseInversionLemma =
let gentac =
    register_tactic "UseInversionLemma"
    (fun [IDENTIFIER id;COMMAND com] gls ->
         LemInv id (pf_constr_of_com gls com) gls)
    (fun sigma gl (_,[IDENTIFIER id;COMMAND com]) ->
         [< 'S"UseInv" ; 'SPC ; print_id id ; 'SPC ; pr_com sigma gl com >])
in fun id com -> gentac [IDENTIFIER id;COMMAND com]
;;



let LemInvIn id c ids gls =
     let intros_replace_ids gls =
        let nb_of_new_hyp  = nb_prod (pf_concl gls) - length ids
        in if nb_of_new_hyp < 1  then IntrosReplacing ids gls
            else (DO nb_of_new_hyp Intro THEN IntrosReplacing ids) gls
     in try 
         ((BringHyps (map inSOME ids) THEN
           LemInv id c THEN
           intros_replace_ids) gls)
        with Not_found -> errorlabstrm "LemInvIn" (not_found_message ids)
          |  UserError(a,b) -> errorlabstrm "LemInvIn" b  
          
;;




let UseInversionLemmaIn =
let gentac = register_tactic "UseInversionLemmaIn"
    (fun ((IDENTIFIER id)::(COMMAND com)::hl) gls ->
         LemInvIn id (pf_constr_of_com gls com)
                     (map (fun (IDENTIFIER id) -> id) hl) gls)
    (fun sigma gl (_,(IDENTIFIER id)::(COMMAND com)::hl) ->
         let hypl = map (fun (IDENTIFIER id) -> id) hl in
         [< 'S"Inv" ; 'SPC ; 'S(string_of_id id) ; 'SPC ; pr_com sigma gl com ; 'SPC ; 'S"in" ; 'SPC ;
            prlist_with_sep (fun () -> [< 'SPC >])
                            (fun id -> [< 'S(string_of_id id) >]) hypl >])
in fun id com hl -> gentac ((IDENTIFIER id)::(COMMAND com)::(map (fun id -> (IDENTIFIER id)) hl))
;;











(* $Id: leminv.ml,v 1.7 1996/02/12 15:56:40 cpaulin Exp $ *)
