(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                mimick.ml                                 *)
(****************************************************************************)
#infix "o";;
#open "initial";;
#open "std";;
#open "more_util";;
#open "reintern0";;
#open "reintern1";;
#open "names";;
#open "generic";;
#open "term";;
#open "proof_constraints";;
#open "proof_trees";;
#open "proof";;
#open "tacmach";;
#open "pfedit";;
#open "himsg";;
#open "pp";;
#open "stdpp";;
#open "ast";;

#infix "THENS";;
#infix "THEN";;
#infix "ORELSE";;



let ri_BindOcc = function
    Dep id -> Dep(ri_id id)
  | NoDep n -> NoDep n
;;

let ri_prim_rule pr =
    {name = pr.name;
     hypspecs = map ri_id pr.hypspecs;
     newids = map ri_id pr.newids;
     params = map ri_tok pr.params;
     terms = map ri_constr pr.terms}
;;

let ri_tactic_arg = function
    COMMAND ast -> COMMAND(ri_ast ast)
  | CONSTR constr -> CONSTR(ri_constr constr)
  | IDENTIFIER id -> IDENTIFIER (ri_id id)
  | INTEGER n -> INTEGER n
  | BINDING (bo,ast) -> BINDING(ri_BindOcc bo,ri_ast ast)
  | PATTERN (nl,ast) -> PATTERN(nl,ri_ast ast)
  | UNFOLD (nl,id) -> UNFOLD(nl,ri_id id)
  | QUOTED_STRING s -> QUOTED_STRING s
  | TACEXP ast -> TACEXP(ri_ast ast)
  | FIXEXP (f,n,ast) -> FIXEXP(ri_id f,n,ri_ast ast)
  | COFIXEXP (f,ast) -> COFIXEXP(ri_id f,ri_ast ast)
;;

let ri_tactic_expression (s,argl) = (s,map ri_tactic_arg argl);;

let rec ri_proofty ri_lc pf =
    {status = pf.status;
     goal = ri_goalty ri_lc pf.goal;
     subproof = option_app (ri_proofty ri_lc) pf.subproof;
     ref = option_app (fun (r,pfl) -> (ri_rulety ri_lc r,
                                       map (ri_proofty ri_lc) pfl)) pf.ref}

and ri_goalty ri_lc g =
    {ctxt = ri_ctxtty ri_lc g.ctxt;
     lc = ri_lc g.lc;
     hyps = ri_sign ri_constr g.hyps;
     concl = ri_constr g.concl}

and ri_rulety ri_lc = function
    PRIM pr -> PRIM(ri_prim_rule pr)
  | TACTIC texp -> TACTIC(ri_tactic_expression texp)
  | CONTEXT ctxt -> CONTEXT(ri_ctxtty ri_lc ctxt)
  | LOCAL_CONSTRAINTS lc -> LOCAL_CONSTRAINTS(ri_lc lc)

and ri_ctxtty ri_lc ctxt =
    {pgm = option_app ri_constr ctxt.pgm;
     mimick = option_app (ri_proofty ri_lc) ctxt.mimick}
;;

let ri_lc = abs_lc o ri_spavs o repr_lc;;

let ri_proof pf = ri_proofty ri_lc pf;;

let (outProof,inProof) =

let (raw_outProof,raw_inProof) = system__extern_intern(1003,".pf") in

let outProof filename (pf:proof) =
    raw_outProof filename (prepare_extern pf) and

    inProof filename = (do_reintern ri_proof true (raw_inProof filename) : proof) in

    (outProof,inProof)
;;

let SaveProof s =
    let pts = get_pftreestate() in
    let pf = proof_of_pftreestate pts in
        outProof s pf
;;


let ReadProof s = inProof s;;

let AssociateProof pf gls = 
    Context (set_mimick (SOME pf) (pf_ctxt gls)) gls
;;

let EraseProof gls =
    match (pf_ctxt gls).mimick with
    NONE -> IDTAC gls
  | SOME _ -> Context (set_mimick NONE (pf_ctxt gls)) gls
;;

let eq_goal g1 g2 =
    eq_constr(g1.concl,g2.concl) &
    for_all2eq (curry eq_constr) (vals_of_sign g1.hyps) (vals_of_sign g2.hyps) &
    for_all2eq (prefix =) (ids_of_sign g1.hyps) (ids_of_sign g2.hyps)
;;


let rec Mimick pf gls =
    (match pf.ref with
     NONE -> IDTAC gls
   | SOME (r,spfl) ->
     ((TRY (refiner r))
      THEN
      (fun gls ->
           match first_result (fun pf -> if eq_goal (sig_it gls) pf.goal then pf
                                         else failwith "caught") spfl with
           SOME pf -> Mimick pf gls
         | NONE -> IDTAC gls))
     gls)
;;

let flat1 pf =
    (match (pf.subproof,pf.ref) with
     (NONE,_) -> error "flat1"
   | (SOME subpf,SOME(_,spfl)) ->
     let (cl,v) = frontier subpf
     in v spfl)
;;

(* for all x in S1 there exists y in S2 such that P(x,y) *)
let for_all_exists P S1 S2 =
    for_all (fun x -> exists (fun y -> P x y) S2) S1
;;

(* [Tactic Mimick] will mimick a proof, and whenever a step fails to
    produce children which are equal to the ones in the original
    proof, it will revert to the proof which is behind the current
    proof, failing if it doesn't find such a proof.

    The rules which we can handle are PRIM, CONTEXT, and TACTIC.  The
    others cause immediate failure.
 *)
let rec TacticMimick pf gls =
    (match pf.ref with
     NONE -> IDTAC gls
   | SOME (r,spfl) ->
     (let (cls,v) = (TRY (refiner r) gls) in
      let cl = sig_it cls
      in if not(for_all_exists (fun g spf -> eq_goal g spf.goal) cl spfl) then
          begin
          MSGNL [< 'S"Mimick step " ; 'CUT ; print_rule mt_evc pf.goal r ;
                   'S" failed on goal " ; 'CUT ; pSEQ pf.goal >];
          (TacticMimick (flat1 pf) ORELSE IDTAC) gls
          end
         else
     ((TRY (refiner r)) THEN SelectMimickSubgoal spfl) gls))

and SelectMimickSubgoal spfl gls =
    match first_result (fun pf -> if eq_goal (sig_it gls) pf.goal then pf
                                  else failwith "caught") spfl with
    SOME pf -> TacticMimick pf gls
  | NONE -> IDTAC gls
;;


let rec mapshape nl (fl:((proof list) -> proof) list) (l:proof list) =
    match nl with
    [] -> []
  | h::t ->
    let m,l = chop_list h l
    in (hd fl m)::(mapshape t (tl fl) l)
;;

let TTHEN (tac1:ttactic) (tac2:ttactic) g =
    let gl,p = tac1 g in
    let gll,pl = split(map tac2 gl) in
    (flat gll) ,  (p o mapshape(map length gll)pl)
;;

#infix "TTHEN";;

let TIDTAC p = [p],(fun [p] -> p | _ -> error "TIDTAC");;

let rec frontier p =
    match p.ref with
    NONE -> ([p],
                fun [p'] ->
                    if p'.goal = p.goal then p'
                    else error "frontier was handed back a ill-formed proof")
  | SOME(r,pfl) ->
    let gll,vl = split(map frontier pfl)
    in (flat gll,
        fun retpfl ->
            let pfl' = mapshape (map length gll) vl retpfl
            in {status = refiner__and_status (map refiner__pf_status pfl');
                subproof=p.subproof;
                goal = p.goal;
                ref = SOME(r,pfl')})
;;

let leaf g = {status=INCOMPLETE_PROOF;
              subproof=NONE;
              goal=g;
              ref=NONE};;

let LEAF p = TIDTAC (leaf p.goal);;

let unTTAC ttac p = let (cl,v) = ttac p in v cl;;

let descend p =
    match p.ref with
    NONE -> error "descend"
  | SOME(r,pfl) ->
    (pfl,
     (fun retpfl ->
          if for_all2eq (fun p1 p2 -> p1.goal = p2.goal) pfl retpfl then
              {status = refiner__and_status (map refiner__pf_status retpfl);
                subproof=p.subproof;
                goal = p.goal;
                ref = SOME(r,retpfl)}
          else error "descend"))
;;

let expand p =
    match p.ref with
    NONE -> error "expand"
  | SOME(TACTIC _,spfl) ->
    let (_,v) = frontier (outSOME p.subproof)
    in (spfl,v)
  | _ -> error "expand"
;;

let ElideTactics l = eliderec
    where rec eliderec pf =
    match pf.ref with
    NONE -> TIDTAC pf
  | SOME (r,_) ->
    (match r with
     TACTIC(s,_) ->
     if listset__memb l s then
         (expand TTHEN eliderec) pf
     else (descend TTHEN eliderec) pf
   | _ -> (descend TTHEN eliderec) pf)
;;

let show_mimick_script () =
    let pts = get_pftreestate () in
    let pf = proof_of_pftreestate pts and
        evc = evc_of_pftreestate pts in

    let mimick_names = ["Mimick";"TacticMimick"]@
                       ["DumpMimick";"DumpTacticMimick"]@
                       ["DumpMimickWith";"DumpTacticMimickWith"] in
    let pf = unTTAC (ElideTactics (mimick_names@["Interpret"])) pf in
    let pf = unTTAC (ElideTactics mimick_names) pf
    in MSG(print_script true (rc_of_gc evc pf.goal) pf.goal.hyps pf)
;;

#open "vernacinterp";;
#open "libobject";;
vinterp_add
("WriteMimick",
 fun [VARG_STRING filename] ->
 fun () -> SaveProof filename)
;;

vinterp_add
("ShowMimick",
 fun [] ->
 fun () -> show_mimick_script())
;;

let ReadDumpedProof () =
let ident = id_of_string(get_proof()) in
    let sp = lib__make_path OBJ ident in
    let fname = "$MIMICKDIR/"^(string_of_path sp) in
        ReadProof fname
;;

let ReadDumpedProofFrom ident =
    let sp = lib__make_path OBJ ident in
    let fname = "$MIMICKDIR/"^(string_of_path sp) in
        ReadProof fname
;;


register_string_tactic "Mimick" (fun s -> Mimick (ReadProof s));;
register_string_tactic "TacticMimick" (fun s -> TacticMimick (ReadProof s));;
register_atomic_tactic "DumpMimick" (fun gls -> Mimick (ReadDumpedProof()) gls);;
register_atomic_tactic "DumpTacticMimick" (fun gls -> TacticMimick (ReadDumpedProof()) gls);;
register_ident_tactic "DumpMimickWith" (fun id -> Mimick (ReadDumpedProofFrom id));;
register_ident_tactic "DumpTacticMimickWith" (fun id -> TacticMimick (ReadDumpedProofFrom id));;

let save_proof_to_standard_mimick_spot ident =
    let pf = proof_of_pftreestate(get_pftreestate()) in
    if not(is_complete_proof pf) then
        error "Cannot save incomplete proof of mimicking";
    let sp = lib__make_path OBJ ident in
    let fname = "$MIMICKDIR/"^(string_of_path sp) in
        SaveProof fname
;;

#open "command";;
#open "discharge";;
#open "vernacentries";;

overwriting_vinterp_add("SaveNamed",
    fun [] ->
    fun () -> show_script() ;
              save_proof_to_standard_mimick_spot (id_of_string(get_proof()));
              save_named true);;

overwriting_vinterp_add("SaveAnonymousThm",
    fun [VARG_IDENTIFIER id] ->
    fun () -> show_script() ;
              save_proof_to_standard_mimick_spot id;
              save_anonymous_thm true (string_of_id id));;
overwriting_vinterp_add("SaveAnonymousRmk",
    fun [VARG_IDENTIFIER id] ->
    fun () -> show_script() ;
              save_proof_to_standard_mimick_spot id;
              save_anonymous_remark true (string_of_id id));;

overwriting_vinterp_add("TheoremProof",
    fun [VARG_STRING kind;VARG_IDENTIFIER s;VARG_COMMAND c;
         VARG_VARGLIST coml] ->
        let calls = (map (fun (VCALL(na,l)) -> (na,l)) coml) in
        let (stre,opacity) = (match kind with
                 "THEOREM" -> (NeverDischarge,true)
               | "LEMMA" -> (make_strength(safe_cdddr (library__cwd())),true)
               | "FACT" -> (make_strength(safe_cdr (library__cwd())),true)
               | "REMARK" -> (make_strength(library__cwd()),true)
               | "DEFINITION" -> (NeverDischarge,false)
               | "LET" -> (make_strength(safe_cddr (library__cwd())),false)
               | "LOCAL" -> (make_strength(library__cwd()),false))
        in fun () ->
        try
        library__with_heavy_rollback
        (fun () ->
            (start_proof (string_of_id s) stre c;
             show_open_subgoals();
             do_listLR (vernacinterp__call NONE) calls;
             show_script();
             save_proof_to_standard_mimick_spot (id_of_string(get_proof()));
             save_named opacity))
        ()
        with e ->
            MSGNL [< 'S"Error during checking of theorem " ; print_id s ;
                          'SPC ; errors__explain_sys_exn e ; 
                          'SPC ; 'S"... trying to mimick" >];
            restart();
            show_open_subgoals();
            solve_nth 1 (TacticMimick (ReadDumpedProof()));
            show_script();
            if not(is_complete_proof(proof_of_pftreestate(get_pftreestate()))) then
                error "Mimick failed"
            else save_named opacity)
;;

(* $Id: mimick.ml,v 1.5 1996/02/12 15:56:46 cpaulin Exp $ *)
