(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                 point.ml                                 *)
(****************************************************************************)

#infix "o";;
#open "std";;
#open "more_util";;
#open "initial";;
#open "names";;
#open "vectops";;
#open "generic";;
#open "term";;
#open "reduction";;
#open "himsg";;
#open "mach";;
#open "proof";;
#open "proof_trees";;
#open "trad";;
#open "tacmach";;
#open "tactics";;
#open "pp";;
#open "stdpp";;
#open "clenv";;
#open "termenv";;
#open "tactics";;
#open "reintern0";;
#open "reintern1";;
#open "libobject";;
#open "library";;
#open "vernacinterp";;
#open "sopattern";;
#open "tactics1";;
#open "termast";;
#open "ast";;

#infix "THEN";;

(* two patterns - one for the type, and one for the type of the type *)
type destructor_pattern =
    {d_typ: constr; d_sort: constr};;

type ('a,'b) location = HYP of 'a | CONCL of 'b;;

(* hypothesis patterns might need to do matching on the conclusion, too.
 * conclusion-patterns only need to do matching on the hypothesis *)
type located_destructor_pattern ==
    (bool * destructor_pattern * destructor_pattern,
     destructor_pattern) location
;;

type destructor_data =
    {d_pat : located_destructor_pattern;
     d_mark : int; (* number of meta-variable containing mark *)
     d_bindings : (identifier * constr) list;
     d_pri : int;
     d_code : ast__spat; (* should be of phylum tactic *)
     d_result : (int,unit) location option list} (* the result of the tactic *)
;;

let ri_dpat dp = {d_typ = ri_constr dp.d_typ;
                  d_sort = ri_constr dp.d_sort}
;;

let ri_ldpat = function
    HYP(b,p1,p2) -> HYP(b,ri_dpat p1,ri_dpat p2)
  | CONCL p -> CONCL(ri_dpat p)
;;

let ri_dd dd =
    {d_pat = ri_ldpat dd.d_pat;
     d_mark = dd.d_mark;
     d_bindings = map (fun (id,c) -> (ri_id id,ri_constr c)) dd.d_bindings;
     d_pri = dd.d_pri;
     d_code = ast__ri_spat dd.d_code;
     d_result = dd.d_result}
;;

type t == (identifier,destructor_data) nbtermdn__t;;
type frozen_t == (identifier,destructor_data) nbtermdn__frozen_t;;

let tactab = (nbtermdn__new {nbtermdn__name_ord = id_ord} : t);;

let lookup pat = nbtermdn__lookup tactab pat;;

let init () = nbtermdn__empty tactab;;

let freeze () = nbtermdn__freeze tactab;;
let unfreeze fs = nbtermdn__unfreeze fs tactab;;

let rollback f x =
    let fs = freeze()
    in try f x
       with e -> (unfreeze fs; raise e)
;;

let add (na,dd) =
let pat = match dd.d_pat with
    HYP(_,p,_) -> p.d_typ
  | CONCL p -> p.d_typ
in
    if nbtermdn__in_dn tactab na then
        (MSGNL [< 'S"Warning [Overriding Destructor Entry " ; 'S(string_of_id na) ; 'S"]" >];
         nbtermdn__remap tactab na (pat,dd))
    else nbtermdn__add tactab (na,(pat,dd));;

library__declare_summary "point-hyp-concl"
{summary__freeze_function = freeze;
 summary__unfreeze_function = unfreeze;
 summary__init_function = init;
 summary__sync_function = (fun () -> false)}
;;

let cache_dd (_,(na,dd)) =
    (try add (na,dd)
    with _ -> anomalylabstrm "dhyp__add"
              [< 'S"The code which adds destructor hints broke;" ;
                 'SPC ; 'S"this is not supposed to happen" >])
;;

let reintern_dd (na,dd) = (ri_id na,ri_dd dd);;

let specification_dd x = x;;

type destructor_data_object == identifier * destructor_data;;

let ((inDD:destructor_data_object->object),
      (outDD:object->destructor_data_object)) =
    declare_object ("POINT-HYP-CONCL-DATA",
                    {cache_function = cache_dd;
                     reintern_function = reintern_dd;
                     specification_function = specification_dd});;

let add_destructor_hint na pat (mark,rl) pri code bindings =
    add_anonymous_object
    (inDD(na,{d_pat = pat;
              d_mark = mark;
              d_bindings = bindings;
              d_pri=pri;
              d_code=code;
              d_result = rl}))
;;

vinterp_add
("HintPoint",
fun ((VARG_IDENTIFIER na)::
     (VARG_ANY location)::
     (VARG_COMMAND patcom)::
     (VARG_NUMBER pri)::
     (VARG_NUMBER subterm)::
     (VARG_ANY (ASTnode("result_list",results)))::
     (VARG_SPAT tacexp)::
     bindings) ->
let bindings = map (fun (VARG_TACTIC_ARG(BINDING(Dep id,c))) -> (id,c)) bindings in
let loc =
    match location with
    ASTnode("CONCL",[]) -> CONCL()
  | ASTnode("DiscardableHYP",[]) -> HYP true
  | ASTnode("PreciousHYP",[]) -> HYP false in

let rl =
    map (fun (ASTnode("CONCL",[])) -> SOME(CONCL())
           | (ASTnode("HYP",[ASTprim(TOKnum n)])) -> SOME(HYP n)
           | (ASTnode("NONE",[])) -> NONE) results in

fun () ->

let pat = raw_sopattern_of_compattern (initial_sign()) patcom in

let bindings =
    map (fun (id,c) ->
             (id,raw_sopattern_of_compattern (initial_sign()) c))
    bindings in

    add_destructor_hint na
    (match loc with
     HYP b ->
     HYP(b,{d_typ=pat;d_sort=DOP0(Meta(newMETA()))},
         {d_typ=DOP0(Meta(newMETA()));d_sort=DOP0(Meta(newMETA()))})
   | CONCL () ->
     CONCL({d_typ=pat;d_sort=DOP0(Meta(newMETA()))}))
    (subterm,rl)
    pri
    tacexp
    bindings)
;;

let match_dpat dp cls gls =
let cltyp = clause_type cls gls in
    match (cls,dp) with
    (SOME id,HYP(_,hypd,concld)) ->
    (somatch__somatch NONE hypd.d_typ cltyp)@
    (somatch__somatch NONE hypd.d_sort (pf_type_of gls cltyp))@
    (somatch__somatch NONE concld.d_typ (pf_concl gls))@
    (somatch__somatch NONE concld.d_sort (pf_type_of gls (pf_concl gls)))
  | (NONE,CONCL concld) ->
    (somatch__somatch NONE concld.d_typ (pf_concl gls))@
    (somatch__somatch NONE concld.d_sort (pf_type_of gls (pf_concl gls)))
  | _ -> error "ApplyDestructor"
;;

let mark_occurs M =
    occur_opern (Abst (path_of_string"#Point#mark.cci")) M;;

let ApplyDestructor cls discard dd gls =
let mvb = match_dpat dd.d_pat cls gls in

(try
    if not(mark_occurs (listmap__map mvb dd.d_mark)) then
        error "ApplyDestructor"
with Not_found -> error "ApplyDestructor");

let astb =
    map
    (fun (id,c) ->
         let sub_c = (reduction__strong (reduction__whd_meta mvb) c) in
         let red_c = sosub__soexecute sub_c in
             (id,FRAGast(bdize (GLOB (pf_hyps gls)) red_c)))
    dd.d_bindings in

let astb = (match cls with
            SOME id -> (id_of_string "$0",FRAGast(ASTnvar id))::astb
          | NONE -> astb) in

let tcom = ast__ssub astb dd.d_code in

let Discard =
    match (cls,dd.d_pat) with
    (SOME id,HYP(discardable,_,_)) ->
    if discard & discardable then thin [id] else IDTAC
  | (NONE,CONCL _) -> IDTAC
  | _ -> error "ApplyDestructor" in

    match tcom with
    FRAGast ast -> (tacentries__vernac_interp ast THEN Discard) gls
  | _ -> errorlabstrm "ApplyDestructor"
    [< 'S"Malformed user tactic text" >]
;;

(* [PointHyp id gls]

   will take an identifier, get its type, look it up in the
   discrimination net, get the destructors stored there, and then try
   them in order of priority.
 *)

let PointHyp discard id gls =
let hyptyp = clause_type (SOME id) gls in
let ddl = map snd (lookup hyptyp) in
let sorted_ddl = sort (fun (dd1,dd2) -> dd1.d_pri > dd2.d_pri) ddl in
    FIRST (map (ApplyDestructor (SOME id) discard) sorted_ddl) gls
;;

let CPHyp id gls = PointHyp true id gls;;
let PHyp id gls = PointHyp false id gls;;

#open "tacinterp";;

tacinterp_add("PHyp",(fun [IDENTIFIER id] -> PHyp id));;
tacinterp_add("CPHyp",(fun [IDENTIFIER id] -> CPHyp id));;

(* [PConcl gls]

   will take a goal, get its concl, look it up in the
   discrimination net, get the destructors stored there, and then try
   them in order of priority.
 *)

let PConcl gls =
let ddl = map snd (lookup (pf_concl gls)) in
let sorted_ddl = sort (fun (dd1,dd2) -> dd1.d_pri > dd2.d_pri) ddl in
    FIRST (map (ApplyDestructor NONE false) sorted_ddl) gls
;;

tacinterp_add("PConcl",(fun [] -> PConcl));;

#open "tactics2";;
let PointStart cls c gl = ChangeClause c cls gl;;

let PointEnd gls =
let unfolder = Unfolds [(id_of_string "mark")] in
    OnAllClauses
    (IfOnClause (fun (cls,T) -> mark_occurs T) unfolder (fun _ -> IDTAC)) gls;;

let PointStart_tac =
let gentac =
    register_tactic "PointStart"
    (fun [INTEGER n;COMMAND c] ->
         (fun gl -> PointStart (nth_clause n gl) (pf_constr_of_com gl c) gl))
    (fun sigma goal (_,[INTEGER n;COMMAND c]) ->
         [< 'S"PointStart " ; 'INT n ; 'SPC ; pr_com sigma goal c >])
in (fun n c -> gentac [INTEGER n;COMMAND c])
;;

let PointEnd_tac = register_atomic_tactic "PointEnd" PointEnd;;

let rec FIRSTWITH = fun tacl g ->
 match tacl with
    []      -> error "FIRSTWITH : no applicable tactic"
 |  T::rest -> (try T g with UserError _ -> FIRSTWITH rest g);;

let rec Point discard cls gls =
let ddl = map snd (lookup (clause_type cls gls)) in
let sorted_ddl = sort (fun (dd1,dd2) -> dd1.d_pri > dd2.d_pri) ddl in

match
    FIRSTWITH
    ((map (fun dd gls ->
              SOME(ApplyDestructor cls discard dd gls,dd)) sorted_ddl)
     @[(fun gls -> NONE)])
    gls
with
  NONE -> IDTAC gls
| SOME((cl,v),dd) ->
(* now we have a successful destructor, so we use the results field to
   continue *)

 THEN_i (fun _ -> (cl,v))
 (fun i g ->
      match nth dd.d_result i with
      NONE -> IDTAC g
    | SOME(CONCL()) -> Point true NONE g
    | SOME(HYP n) -> Point true (nth_clause n g) g)
 1
 gls
;;

let GenPoint discard cls c gls =
    (PointStart cls c THEN (Point discard cls) THEN PointEnd) gls;;

let Point = GenPoint false;;
let CPoint = GenPoint true;;

let Point_tac =
let gentac =
    register_tactic "Point"
    (fun [IDENTIFIER id;COMMAND c] ->
         (fun gl -> Point (SOME id) (pf_constr_of_com gl c) gl)
       | [COMMAND c] ->
         (fun gl -> Point NONE (pf_constr_of_com gl c) gl))
    (fun sigma goal (_,[IDENTIFIER id;COMMAND c]) ->
         [< 'S"Point " ; pr_com sigma goal c ; 'S" in " ; 'S(string_of_id id) >]
       | sigma goal (_,[COMMAND c]) ->
         [< 'S"Point " ; pr_com sigma goal c >])
in (fun (SOME id) c -> gentac [IDENTIFIER id;COMMAND c]
      | NONE c -> gentac [COMMAND c])
;;

let CPoint_tac =
let gentac =
    register_tactic "CPoint"
    (fun [IDENTIFIER id;COMMAND c] ->
         (fun gl -> CPoint (SOME id) (pf_constr_of_com gl c) gl)
       | [COMMAND c] ->
         (fun gl -> CPoint NONE (pf_constr_of_com gl c) gl))
    (fun sigma goal (_,[IDENTIFIER id;COMMAND c]) ->
         [< 'S"CPoint " ; pr_com sigma goal c ; 'S" in " ; 'S(string_of_id id) >]
       | sigma goal (_,[COMMAND c]) ->
         [< 'S"CPoint " ; pr_com sigma goal c >])
in (fun (SOME id) c -> gentac [IDENTIFIER id;COMMAND c]
      | NONE c -> gentac [COMMAND c])
;;

(* $Id: point.ml,v 1.2 1996/02/12 15:56:50 cpaulin Exp $ *)
