(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                               progmach.ml                                *)
(****************************************************************************)
#infix "o";;
#open "std";;
#open "names";;
#open "vectops";;
#open "generic";;
#open "term";;
#open "reduction";;
#open "typing";;
#open "termenv";;
#open "stdpp";;
#open "machops";;
#open "mach";;
#open "tacmach";;
#open "environ";;
#open "termenv";;
#open "termast";;
#open "astterm";;
#open "pp";;
#open "stdpp";;
#open "himsg";;
#open "more_util";;
#open "lomsg";;
#open "impuniv";;
#open "ast";;
#open "proof_trees";;
#open "proof_constraints";;
#open "proof";;
#open "constrtypes";;

let isVAR = function VAR _ -> true | _ -> false;;
let isRel = function Rel _ -> true | _ -> false;;

(* gives a new section path each time it is called *)
let gensym_exsp =
let existential_ctr = ref 0
in
fun () ->
let n = (incr existential_ctr; !existential_ctr)
in make_path(PA__intern["isevar"]) (make_ident(SA__intern"?") n) CCI;;

(* domain of a product *)
let prod_dom_tycon sigma = function
    NONE -> NONE
  | SOME c -> (match hnf_constr sigma c with
               DOP2(Prod,c1,_) -> SOME c1
             | _ -> NONE)
;;

(* range of a product *)
let prod_rng_tycon sigma = function
    NONE -> NONE
  | SOME c -> (match hnf_constr sigma c with
               DOP2(Prod,_,DLAM(_,c2)) -> SOME c2
             | _ -> NONE)
;;

(* domain of a lamdba *)
let lambda_dom_tycon = prod_dom_tycon
;;

let app_rng_tycon sigma var = function
    NONE -> NONE
  | SOME c -> (match hnf_constr sigma c with
               DOP2(Prod,_,B) -> SOME (SAPP B var)
             | _ -> NONE)
;;

let isevars = ref mt_evc;;

let set_ise evc = isevars := evc;;
let get_ise () = !isevars;;

(* declare a new existential of section path sp, type typ, with the initial
signature sign *)
let ise_declare sp sign typ =
    isevars := evc__add !isevars (sp,{evgoal=mkGOAL mt_lc mt_ctxt sign typ;
                                      body = EVAR_EMPTY})
;;

(* declare a new existential in an env env with type typ *)
let new_isevar env typ =
let (ENVIRON(sign,dbenv)) = env in
let t = it_list (fun B (na,D)-> DOP2(Lambda,D,DLAM(na,B))) typ dbenv in
let rec aux sign = fun
    0 T -> (sign,T)
  | n (DOP2(Lambda,D,(DLAM(na,B)))) ->
    let id = next_name_away na (ids_of_sign sign)
    in aux (add_sign (id,D) sign) (n-1) (SAPP (DLAM(na,B)) (VAR id))
  | _ _ -> anomalylabstrm "trad__new_isevar" [< 'S"Bug in" ; 'SPC ;
                                              'S"trad__new_isevar" ;'SPC;
                                              'S"code." >]
in
let (sign',typ') = aux sign (length dbenv) t in
let newsp = gensym_exsp()
in (ise_declare newsp sign' typ';
    DOPN(Const newsp,
         vect_of_list ((rev(rel_list 0 (length dbenv)))
                       @(map (fun id -> VAR id) (ids_of_sign sign)))))
;;

let ise_define sp body =
let evd = evc__map !isevars sp
in match evd.body with
    EVAR_EMPTY ->
    isevars := evc__remap !isevars sp {evgoal = evd.evgoal;
                                       body = EVAR_DEFINED body}
  | _ -> anomaly "cannot define an isevar twice"
;;

let ise_in_dom sp = evc__in_dom !isevars sp;;

let ise_defined (DOPN(Const sp,_) as K) =
    not(ise_in_dom sp) or
    defined_const (get_ise()) K;;

let ise_map sp = evc__map !isevars sp;;

let ise_pred P a =
let e = get_ise()
in try if P a then true
       else (set_ise e;false)
   with reraise -> (set_ise e;raise reraise)
;;

let ise_or P Q =
let u = get_ise()
in try if P() then true
       else
           (set_ise u;
            if Q() then true
            else (set_ise u;false))
   with reraise -> (set_ise u;raise reraise)
;;

type conv_pb = CONV | CONV_LEQ | CONV_X;;

let conversion_problems = ref ([] : (conv_pb * constr * constr) list);;
let solved_problems = ref ([] : (section_path * evar_declaration) list);;

let add_conv_pb pb = (conversion_problems := pb::!conversion_problems);;

let sort_increasing_snd = sort (fun ((_,Rel m),(_,Rel n)) -> m < n);;

let head_is_exist = hrec where rec hrec = function
    DOPN(Const sp,_) as K -> ise_in_dom sp & not(ise_defined K)
  | DOPN(AppL,cl) -> hrec (hd_vect cl)
  | DOP2(Cast,c,_) -> hrec c
  | _ -> false
;;

let rec is_eliminator = function
    DOPN(AppL,_) -> true
  | DOPN(MutCase,_) -> true
  | DOP2(Cast,c,_) -> is_eliminator c
  | _ -> false
;;

let head_is_embedded_exist c =
    (head_is_exist c) & (is_eliminator c)
;;

let headconstant = hrec where rec hrec = function
    DOPN(Const sp,_) -> sp
  | DOPN(MutCase,cl) -> hrec (hd_vect cl)
  | DOPN(AppL,cl) -> hrec (hd_vect cl)
  | DOP2(Cast,c,_)-> hrec c
  | _ -> failwith "headconstant"
;;

let status_changed sp (pbty,t1,t2) =
    try (sp = headconstant t1) or (sp = headconstant t2)
    with Failure _ ->
    try (sp = headconstant t2)
    with Failure _ -> false
;;

let vect_lift = map_i_vect lift 0;;

let whd_propify = function
  DOPN(Const sp,cl) -> DOPN(Const (ccisp_of sp),cl)
| DOPN(MutInd(sp,i),cl) -> DOPN(MutInd (ccisp_of sp,i),cl)
| DOPN(MutConstruct((sp,j),i),cl) -> DOPN(MutConstruct((ccisp_of sp,j),i),cl)
| x -> x
;;

let propify = strong whd_propify;;

let execute_pure sigma sign pg = 
  fexecute sigma sign (propify pg) ;;

let execute_meta_pure sigma sign metamap pg = 
  fexemeta sigma metamap sign (propify pg) ;;

let execute_pure_rec sigma sign pg = 
  mach__execute_rec sigma sign (propify pg) ;;

let prog_relative sigma n env = 
try
   (let (_,typ) = lookup_rel n env in
    match strip_outer_cast typ with
        DOP0(Implicit) -> error "Cannot use a logical assumption in a program"
      | _ ->
    let typtyp = cast_type typ
    in  {VAL=Rel(n);
         (* We cannot assume, as in the normal machine operation, that the 
          types of variables are closed terms, and thus we must lift them *)
         TYP=lift n typ;
         KIND = lift n (whd_betadeltaiota sigma typtyp)})
with Not_found -> errorlabstrm "prog_relative"
                  [< 'S"Tried to typecheck an open term in a environment which did not define all of its free references" >]
;;

let traceme (s:string) (j:judgement) = ();;

let j_val = machops__j_val;;

let assumption_of_judgement = machops__assumption_of_judgement ;;

let assumption_of_judgement_prog sigma j =
let typ = whd_betadeltaiota sigma j.TYP in 
match typ with DOP0(Sort(_)) -> DOP2(Cast,j.VAL,typ) 
             | _ -> DOP2(Cast,j.VAL,DOP0(Implicit))
;;

let make_Kn n = makerec n
    where rec makerec = function
    0 -> (Rel (n+1))
  | n -> DOP2(Lambda,DOP0 Implicit,DLAM(Anonymous,makerec (n-1)))
;;

(****
let red_mind sigma ((nparams,sp,ar,DLAMV(na,lc)),la) =
let num_to_eat = min nparams (length la) in
let eaten,rest = chop_list num_to_eat la in
let X_arg_eater = make_Kn num_to_eat in
let rest_nparams = nparams - num_to_eat
in ((rest_nparams,sp,
    hnf_prod_applist sigma "red_ind" ar eaten,
    DLAMV(na,map_vect
          (fun C ->
               let C' = strong whd_beta (subst1 X_arg_eater C)
               in hnf_prod_applist sigma "red_mind" C' (map (lift 1) eaten)) lc)),
    rest)
;;
****)
let red_mind sigma ((nparams,sp,ar,lc),la) =
let num_to_eat = min nparams (length la) in
let eaten,rest = chop_list num_to_eat la in
let X_arg_eater = make_Kn num_to_eat in
let rest_nparams = nparams - num_to_eat
in ((rest_nparams,sp,
    hnf_prod_applist sigma "red_ind" ar eaten,
    map_vect
          (fun C ->
               let C' = strong whd_beta (subst1 X_arg_eater C)
               in hnf_prod_applist sigma "red_mind" C' (map (lift 1) eaten)) lc),
    rest)
;;

(****
let conv1_minductive sigma convfun (mind1,l1) (mind2,l2) =
    let (DOPN(MutInd(sp1,i1),cl1)) = mind1 and
        (DOPN(MutInd(sp2,i2),cl2)) = mind2  in
    let ar1 = mind_arity mind1 and
        ar2 = mind_arity mind2 and
        mis1 = mind_specif_of_mind mind1 and
        mis2 = mind_specif_of_mind mind2 in
    let (DLAMV(_,lc1) as llc1) = mis_lc mis1 and
        (DLAMV(_,lc2) as llc2) = mis_lc mis2 and
        nparams1 = mis_nparams mis1 and
        nparams2 = mis_nparams mis2 in
    let ((nparams1,sp1,ar1,DLAMV(_,lc1)),l1) =
        red_mind sigma ((nparams1,sp1,ar1,llc1),l1) and
        ((nparams2,sp2,ar2,DLAMV(_,lc2)),l2) =
        red_mind sigma ((nparams2,sp2,ar2,llc2),l2)
    in  nparams1 = nparams2 &
        i1=i2 &
        convfun ar1 ar2 
        & for_all2eq_vect convfun lc1 lc2
        & for_all2eq convfun l1 l2
;;
****)
let conv1_minductive sigma convfun (mind1,l1) (mind2,l2) =
    let (DOPN(MutInd(sp1,i1),cl1)) = mind1 and
        (DOPN(MutInd(sp2,i2),cl2)) = mind2  in
    let ar1 = mind_arity mind1 and
        ar2 = mind_arity mind2 and
        mis1 = mind_specif_of_mind mind1 and
        mis2 = mind_specif_of_mind mind2 in
    let nparams1 = mis_nparams mis1 and
        nparams2 = mis_nparams mis2 in
    let mib1 = mis1.mib and mib2 = mis2.mib in
    let n1 = mib1.MINDNTYPES and n2 = mib2.MINDNTYPES in
    let mpack1 = mib1.MINDPACKETS and mpack2 = mib2.MINDPACKETS in
    let conv_pack mip1 mip2 = 
        let (_,lc1) = decomp_all_DLAMV_name (mip1.MINDLC) and
            (_,lc2) = decomp_all_DLAMV_name (mip2.MINDLC) in
        let ((nparams1,_,ar1,lc1),l1) =
              red_mind sigma ((nparams1,sp1,ar1,lc1),l1) and
            ((nparams2,_,ar2,lc2),l2) =
              red_mind sigma ((nparams2,sp2,ar2,lc2),l2) in
         nparams1=nparams2
         & convfun ar1 ar2
         & for_all2eq_vect convfun lc1 lc2
         & for_all2eq convfun l1 l2
    in  n1=n2
      & i1=i2
      & for_all2eq_vect conv_pack mpack1 mpack2
;;

let conv1_mconstructor sigma convfun
    (DOPN(MutConstruct(indsp1,i1),cl1),l1) (DOPN(MutConstruct(indsp2,i2),cl2),l2) =
    i1 = i2
    & (indsp1 = indsp2
       or conv1_minductive sigma convfun (DOPN(MutInd indsp1,cl1),l1)
                                  (DOPN(MutInd indsp2,cl2),l2))
;;


(* equality of terms without universe adjustment for synthesis *)
let rec conv1_x sigma term1 term2 =
      term1 = term2 or eqappr_x sigma (hnf sigma term1,hnf sigma term2)
and eqappr_x sigma = function
       ((Rel(n),l1),(Rel(m),l2)) ->
           (n=m) & (length(l1) = length(l2)) & (for_all2 (conv1_x sigma) l1 l2)
     | ((DOP2(Cast,c,_),l),appr2) -> eqappr_x sigma ((c,l),appr2)
     | (appr1,(DOP2(Cast,c,_),l)) -> eqappr_x sigma (appr1,(c,l))
     | ((VAR _ as v1,l1),(VAR _ as v2,l2)) ->
           (eq(id_of_var v1,id_of_var v2) & (length(l1) = length(l2)) & (for_all2 (conv1_x sigma) l1 l2))
     | ((DOP0(Meta(n)),l1),(DOP0(Meta(m)),l2)) ->
           ((n=m) & (length(l1) = length(l2)) & (for_all2 (conv1_x sigma) l1 l2))
     | ((DOP0(Sort s1),[]),(DOP0(Sort s2),[])) -> sort_eq_x s1 s2
     | ((DOP2(Lambda,c1,DLAM(_,c2)),[]),(DOP2(Lambda,c'1,DLAM(_,c'2)),[])) -> 
                                          (conv1_x sigma) c1 c'1 & (conv1_x sigma) c2 c'2
     | ((DOP2(Prod,c1,DLAM(_,c2)),[]),(DOP2(Prod,c'1,DLAM(_,c'2)),[])) -> 
                                          (conv1_x sigma) c1 c'1 & (conv1_x sigma) c2 c'2

     | ((DOPN(Const sp1,al1),l1),(DOPN(Const sp2,al2),l2)) ->
           (eq(sp1,sp2) & eq(al1,al2) & (length(l1) = length(l2)) & (for_all2 (conv1_x sigma) l1 l2))
        or ((evaluable_const sigma (DOPN(Const sp2,al2)))
            & eqappr_x sigma  ((DOPN(Const sp1,al1),l1), apprec sigma l2 (const_value sigma (DOPN(Const sp2,al2)))))
        or ((not (evaluable_const sigma (DOPN(Const sp2,al2))))
            & (evaluable_const sigma (DOPN(Const sp1,al1)))
            & eqappr_x sigma  (apprec sigma l1 (const_value sigma (DOPN(Const sp1,al1))),(DOPN(Const sp2,al2),l2)))
     | ((DOPN(Const sp1,al1),l1),p2) ->  
       (evaluable_const sigma (DOPN(Const sp1,al1)))
       & eqappr_x sigma (apprec sigma l1 (const_value sigma (DOPN(Const sp1,al1))),p2)
     | (p1,(DOPN(Const sp2,al2),l2))  -> 
       (evaluable_const sigma (DOPN(Const sp2,al2)))
       & eqappr_x sigma (p1,apprec sigma l2 (const_value sigma (DOPN(Const sp2,al2))))

     | ((DOPN(Abst sp1,al1),l1),(DOPN(Abst sp2,al2),l2)) ->
           (eq(sp1,sp2) & eq(al1,al2) & (length(l1) = length(l2)) & (for_all2 (conv1_x sigma) l1 l2))
        or ((evaluable_abst (DOPN(Abst sp2,al2)))
            & eqappr_x sigma  ((DOPN(Abst sp1,al1),l1), apprec sigma l2 (abst_value (DOPN(Abst sp2,al2)))))
        or ((not (evaluable_abst (DOPN(Abst sp2,al2))))
            & (evaluable_abst (DOPN(Abst sp1,al1)))
            & eqappr_x sigma  (apprec sigma l1 (abst_value (DOPN(Abst sp1,al1))),(DOPN(Abst sp2,al2),l2)))
     | ((DOPN(Abst sp1,al1),l1),p2) ->  
       (evaluable_abst (DOPN(Abst sp1,al1)))
       & eqappr_x sigma (apprec sigma l1 (abst_value (DOPN(Abst sp1,al1))),p2)
     | (p1,(DOPN(Abst sp2,al2),l2))  -> 
       (evaluable_abst (DOPN(Abst sp2,al2)))
       & eqappr_x sigma (p1,apprec sigma l2 (abst_value (DOPN(Abst sp2,al2))))
     | ((DOPN(MutInd indsp1,cl1) as mind1,l'1),
        (DOPN(MutInd indsp2,cl2) as mind2,l'2)) ->
       ((indsp1 = indsp2 & for_all2eq_vect (conv1_x sigma) cl1 cl2)
        & for_all2eq (conv1_x sigma) l'1 l'2)
        or conv1_minductive sigma (conv1_x sigma) (mind1,l'1) (mind2,l'2)
         
     | ((DOPN(MutConstruct i1,cl1) as mconstr1,l1),
        (DOPN(MutConstruct i2,cl2) as mconstr2,l2)) ->
        (((i1=i2) & for_all2eq_vect (conv1_x sigma) cl1 cl2)
         & for_all2eq (conv1_x sigma) l1 l2)
         or conv1_mconstructor sigma (conv1_x sigma) (mconstr1,l1) (mconstr2,l2)

     | ((DOPN(MutCase,cl1),l'1),(DOPN(MutCase,cl2),l'2)) -> 
       (for_all2eq_vect (conv1_x sigma) cl1 cl2) &
       (for_all2eq (conv1_x sigma) l'1 l'2)

     | ((DOPN(Fix i1,cl1),l'1),(DOPN(Fix i2,cl2),l'2)) -> 
       i1 = i2 &
       (for_all2eq_vect (conv1_x sigma) cl1 cl2) &
       (for_all2eq (conv1_x sigma) l'1 l'2)

     | ((DOPN(CoFix i1,cl1),l'1),(DOPN(CoFix i2,cl2),l'2)) -> 
       i1 = i2 &
       (for_all2eq_vect (conv1_x sigma) cl1 cl2) &
       (for_all2eq (conv1_x sigma) l'1 l'2)

     | (DOP0(Implicit),[]),(DOP0(Implicit),[]) -> true
     | (DLAM(_,c1),[]),(DLAM(_,c2),[]) -> conv1_x sigma c1 c2
     | (DLAMV(_,vc1),[]),(DLAMV(_,vc2),[]) -> for_all2eq_vect (conv1_x sigma) vc1 vc2
     | _ -> false;;

let pf_conv1_x gls M N = conv1_x (Project gls) M N;;

let has_undefined_isevars c =
    (try (hasrec c ; false) with Failure "caught" -> true)
    where rec hasrec = function
    DOPN(Const sp,cl) as K ->
    if ise_in_dom sp & not(ise_defined K) then failwith "caught"
    else do_vect hasrec cl

  | DOP1(_,c) -> hasrec c
  | DOP2(_,c1,c2) -> (hasrec c1; hasrec c2)
  | DOPL(_,l) -> do_list hasrec l
  | DOPN(_,cl) -> do_vect hasrec cl
  | DLAM(_,c) -> hasrec c
  | DLAMV(_,cl) -> do_vect hasrec cl
  | (VAR _|Rel _|DOP0 _) -> ()
;;

let evar_hnf c =
    match hnf_constr (get_ise()) c with
    DOPN(AppL,cl) -> (hd_vect cl,list_of_tl_vect cl)
  | c -> (c,[])
;;

let evar_apprec l c = evar_hnf (applistc c l);;

let rec rename_rels curidx sofar = function
    [] -> sofar
  | ((id,Rel n)::tl as l) ->
    if curidx = n then rename_rels (curidx+1) (subst1 (VAR id) sofar) tl
    else rename_rels (curidx+1) (subst1 (DOP0 Implicit) sofar) l
;;

let clean_rhs rhs worklist =
let workvars = filter (isVAR o snd) worklist in
let rhs' = replace_vars (map (fun (id',VAR id) -> (id,CLOSEDTERM (VAR id'))) workvars) rhs in
let workrels = filter (isRel o snd) worklist in
let workrels' = sort_increasing_snd workrels
in rename_rels 1 rhs' workrels'
;;

let rec do_solve_pb1(sp,args,rhs) =
(let evd = ise_map sp in
let hyps= evd.evgoal.hyps in
let worklist = combine (ids_of_sign hyps,args) in
let rhs'' = clean_rhs rhs worklist in
if not(closed rhs'') then error "cannot solve this unification problem";
(* rhs'' should be the completely clean term. *)
ise_define sp rhs'';
solved_problems := (sp,ise_map sp)::!solved_problems;
let pbs = !conversion_problems in
   conversion_problems := [];
   for_all (function pb ->
            if status_changed sp pb then
                rerun_problem1 pb
            else (add_conv_pb pb;true)) pbs)

and evar_conv1_x term1 term2 =
let term1 = strip_outer_cast term1 and
    term2 = strip_outer_cast term2 
in if term1 = term2 then true
   else if not(has_undefined_isevars term1) & not(has_undefined_isevars term2) then
       conv1_x (get_ise()) term1 term2
    else if (match (term1,term2) with
        (DOPN(Const sp1,_),DOPN(Const sp2,_)) ->
        (ise_in_dom sp1 & not(ise_defined term1))
     or (ise_in_dom sp2 & not(ise_defined term2))
      | (DOPN(Const sp1,_),_) -> ise_in_dom sp1 & not(ise_defined term1)
      | (_,DOPN(Const sp2,_)) -> ise_in_dom sp2 & not(ise_defined term2)
      | _ -> false) then
        solve_pb1(CONV_X,term1,term2)
    else 
        let (t1,l1) = evar_hnf term1
        and (t2,l2) = evar_hnf term2
        in if (head_is_embedded_exist t1 & not(is_eliminator t2))
            or (head_is_embedded_exist t2 & not(is_eliminator t1)) then
            (add_conv_pb (CONV_X,applist(t1,l1),applist(t2,l2));true)
        else
            evar_eqappr1_x ((t1,l1),(t2,l2))

and solve_pb1 ((pbty,t1,t2) as pb) =
    if t1 = t2 then true else
    match (t1,t2) with
    (DOPN(Const sp1,args1),DOPN(Const sp2,args2)) ->
    if ise_in_dom sp1 & ise_in_dom sp2 
        & not(ise_defined t1) & ise_defined t2 then
        try_solve_pb1(sp1,args1,t2)
    else if ise_in_dom sp2 & not(ise_defined t2) & ise_defined t1 then
        try_solve_pb1(sp2,args2,t1)
    else (add_conv_pb pb;true)

  | (DOPN(Const sp1,args1),t2) ->try_solve_pb1(sp1,args1,t2)

  | (t1,DOPN(Const sp2,args2)) -> try_solve_pb1(sp2,args2,t1)
   
and try_solve_pb1(sp,argsv,t2) =
let args = list_of_vect argsv
in if for_all (fun (VAR _) -> true | (Rel _) -> true
                 | (DOP0 Implicit) -> true | _ -> false) args
    & distinct (filter (fun (DOP0 Implicit) -> false | _ -> true) args) then
    (* We can instanciate the existential *)
    do_solve_pb1(sp,args,t2)
   else error "We found an existential that we could not instanciate"

and rerun_problem1 = function
  (CONV_X,t1,t2) -> evar_conv1_x t1 t2

and evar_eqappr1_x = function
       ((DOPN(Const sp1,al1),l1),(DOPN(Const sp2,al2),l2)) ->
       solve_pb1(CONV_X,DOPN(Const sp1,al1),DOPN(Const sp2,al2))
       & for_all2eq evar_conv1_x l1 l2

     | ((DOPN(Const sp1,al1),l1),(t2,l2)) ->  
       solve_pb1(CONV_X,DOPN(Const sp1,al1),t2)
       & for_all2eq evar_conv1_x l1 l2

     | ((t1,l1),(DOPN(Const sp2,al2),l2))  -> 
       solve_pb1(CONV_X,t1,DOPN(Const sp2,al2))
       & for_all2eq evar_conv1_x l1 l2

     | ((DOPN(Abst sp1,al1),l1),(DOPN(Abst sp2,al2),l2)) ->
           (eq(sp1,sp2) & eq(al1,al2) & (length(l1) = length(l2)) & (for_all2 (evar_conv1_x) l1 l2))
        or ((evaluable_abst (DOPN(Abst sp2,al2)))
            & evar_eqappr1_x ((DOPN(Abst sp1,al1),l1), evar_apprec l2 (abst_value (DOPN(Abst sp2,al2)))))
        or ((not (evaluable_abst (DOPN(Abst sp2,al2))))
            & (evaluable_abst (DOPN(Abst sp1,al1)))
            & evar_eqappr1_x  (evar_apprec l1 (abst_value (DOPN(Abst sp1,al1))),(DOPN(Abst sp2,al2),l2)))

     | ((DOPN(Abst sp1,al1),l1),p2) ->  
       (evaluable_abst (DOPN(Abst sp1,al1)))
       & evar_eqappr1_x (evar_apprec l1 (abst_value (DOPN(Abst sp1,al1))),p2)

     | (p1,(DOPN(Abst sp2,al2),l2))  -> 
       (evaluable_abst (DOPN(Abst sp2,al2)))
       & evar_eqappr1_x (p1,evar_apprec l2 (abst_value (DOPN(Abst sp2,al2))))

     | ((Rel(n),l1),(Rel(m),l2)) ->
           (n=m) & (length(l1) = length(l2)) & (for_all2 evar_conv1_x l1 l2)
     | ((DOP2(Cast,c,_),l),appr2) -> evar_eqappr1_x ((c,l),appr2)
     | (appr1,(DOP2(Cast,c,_),l)) -> evar_eqappr1_x (appr1,(c,l))
     | ((VAR _ as v1,l1),(VAR _ as v2,l2)) ->
           (eq(id_of_var v1,id_of_var v2) & (length(l1) = length(l2)) & (for_all2 evar_conv1_x l1 l2))
     | ((DOP0(Meta(n)),l1),(DOP0(Meta(m)),l2)) ->
           ((n=m) & (length(l1) = length(l2)) & (for_all2 evar_conv1_x l1 l2))
     | ((DOP0(Sort s1),[]),(DOP0(Sort s2),[])) -> sort_eq_x s1 s2
     | ((DOP2(Lambda,c1,DLAM(_,c2)),[]),(DOP2(Lambda,c'1,DLAM(_,c'2)),[])) -> 
                                          evar_conv1_x c1 c'1 & evar_conv1_x c2 c'2
     | ((DOP2(Prod,c1,DLAM(_,c2)),[]),(DOP2(Prod,c'1,DLAM(_,c'2)),[])) -> 
                                          evar_conv1_x c1 c'1 & evar_conv1_x c2 c'2
     | ((DOPN(MutInd indsp1,cl1) as ind1,l'1),(DOPN(MutInd indsp2,cl2) as ind2,l'2)) ->
        (ise_or (fun () -> (indsp1 = indsp2 & for_all2eq_vect evar_conv1_x cl1 cl2 & for_all2eq evar_conv1_x l'1 l'2))
                 (fun () -> conv1_minductive (get_ise()) evar_conv1_x (ind1,l'1) (ind2,l'2)))

     | ((DOPN(MutConstruct i1,cl1) as constr1,l1),(DOPN(MutConstruct i2,cl2) as constr2,l2)) ->
         (ise_or (fun () -> (i1=i2) & (for_all2eq_vect evar_conv1_x cl1 cl2)
          & (for_all2eq evar_conv1_x l1 l2))
           (fun () -> conv1_mconstructor (get_ise()) evar_conv1_x (constr1,l1) (constr2,l2)))

     | ((DOPN(MutCase,cl1),l'1),(DOPN(MutCase,cl2),l'2)) -> 
         (for_all2eq_vect evar_conv1_x cl1 cl2) & (for_all2eq evar_conv1_x l'1 l'2)

     | ((DOPN(Fix(i1),cl1),l1),(DOPN(Fix(i2),cl2),l2))   ->
       i1=i2 & 
       (for_all2eq_vect evar_conv1_x cl1 cl2) &
       (for_all2eq evar_conv1_x l1 l2)

     | ((DOPN(CoFix(i1),cl1),l1),(DOPN(CoFix(i2),cl2),l2))   ->
       i1=i2 & 
       (for_all2eq_vect evar_conv1_x cl1 cl2) &
       (for_all2eq evar_conv1_x l1 l2)

     | (DOP0(Implicit),[]),(DOP0(Implicit),[]) -> true
(* added to compare easily the specification of fixed points *)
     | (DLAM(_,c1),[]),(DLAM(_,c2),[]) -> evar_conv1_x c1 c2
     | (DLAMV(_,vc1),[]),(DLAMV(_,vc2),[]) -> for_all2eq_vect evar_conv1_x vc1 vc2
     | _ -> false;;

let the_conv1_x t1 t2 = conv1_x (get_ise()) t1 t2 or evar_conv1_x t1 t2;;

let evar_type_fix (DOPN(Fix(_,j),vargs)) vdeft = 
let lt = vect_length vdeft in 
let var = sub_vect vargs 0 lt in 
if for_all2eq_vect
   (fun ari ti ->  the_conv1_x (lift lt ari) ti) var vdeft then
   var.(j) 
else error "Implicit Syntax: Ill-typed case recursive call
                             (branches are malformed)";;

let evar_type_cofix (DOPN(CoFix(j),vargs)) vdeft = 
let lt  = vect_length vdeft in 
let var = sub_vect vargs 0 lt in 
if for_all2eq_vect
   (fun ari ti ->  the_conv1_x (lift lt ari) ti) var vdeft then
   var.(j) 
else error "Implicit Syntax: (cofix) Ill-typed case recursive call
                             (branches are malformed)";;

let type_case_elim_pg env sigma ct Pt lft P c =
let (bty,rslty) = type_case_branches env sigma ct Pt P c in
let n = vect_length lft and expn = vect_length bty in
if n<>expn then 
errorlabstrm "Case with wrong number of cases"
 (HOV 0 [< 'S "Case on term "; pTERMINENV(env,c); 'CUT ; 'S "has type"; 'WS 1;
           pTERMINENV(env,ct); 'WS 1; 'CUT; 'S "and expects" ; 'INT expn; 'S "n" >]) 
else
let rec check_conv i = 
    if i = n then () else
    if not (conv1_x sigma lft.(i) (bty.(i))) then 
errorlabstrm "Ill-formed branches" 
 (HOV 0 [< 'S "In case expression on term "; pTERMINENV(env,c); 
           'SPC; 'CUT; 'S "the branch  " ; 'INT (i+1); 'S " has type";  'SPC;
         pTERMINENV(env,lft.(i)); 'CUT; 'S "it should be :"; 'WS 1; 
         pTERMINENV(env,bty.(i)) >]) 
    else check_conv (i+1) 
in check_conv 0; rslty;;

let the_type_case_elim_pg env sigma ct Pt lft P c =
let (bty,rslty) = type_case_branches env sigma ct Pt P c in
let n = vect_length lft and expn = vect_length bty in
if n<>expn then 
errorlabstrm "Case with wrong number of cases"
 (HOV 0 [< 'S "Case on term "; pTERMINENV(env,c); 'CUT ; 'S "has type"; 'WS 1;
           pTERMINENV(env,ct); 'WS 1; 'CUT; 'S "and expects" ; 'INT expn; 'S "n" >]) 
else
let rec check_conv i = 
    if i = n then () else
    if not (the_conv1_x lft.(i) (bty.(i))) then 
errorlabstrm "Ill-formed branches" 
 (HOV 0 [< 'S "In case expression on term "; pTERMINENV(env,c); 
           'SPC; 'CUT; 'S "the branch  " ; 'INT (i+1); 'S " has type";  'SPC;
         pTERMINENV(env,lft.(i)); 'CUT; 'S "it should be :"; 'WS 1; 
         pTERMINENV(env,bty.(i)) >]) 
    else check_conv (i+1) 
in check_conv 0; rslty;;

let apply_rel_list_prog sigma env nocheck argjl funj =
let rec apply_rec typ = function
    [] -> {VAL=applist(j_val funj,map j_val argjl);
           TYP=typ;
           KIND = funj.KIND}
  | hj::restjl ->
  match whd_betadeltaiota sigma typ with
     DOP2(Prod,c1,DLAM(_,c2)) ->
     if nocheck or conv1_x sigma hj.TYP c1 then
         apply_rec (subst1 hj.VAL c2) restjl
        else (match strip_outer_cast hj.VAL with
                DOP1(XTRA("COMMENT",[]),c) -> apply_rec typ restjl
              | _ -> errorlabstrm "apply"
              [< 'S"Illegal application (Type Error) : "; 'CUT ;
                 apply_error env (funj.VAL,funj.TYP)
                                 (map (fun j -> (j.VAL,j.TYP)) argjl) >])
   | _ -> errorlabstrm "apply"
              [< 'S"Illegal application (Non-functional construction) : ";
                 'CUT ;
                 apply_error env (funj.VAL,funj.TYP)
                                 (map (fun j -> (j.VAL,j.TYP)) argjl) >]
in apply_rec funj.TYP argjl;;

let attempt_coercion env typ = apprec
    where rec apprec argj =
    match hnf_constr (get_ise()) argj.TYP with
    DOP2(Prod,argc1,argB) ->
    let K = new_isevar env argc1 in
    let argj = {VAL = applist(argj.VAL,[K]);
                TYP = SAPP argB K;
                KIND = argj.KIND}
    in if ise_pred (the_conv1_x (SAPP argB K))  typ then
        argj
       else apprec argj

  | _ -> failwith "attempt_coercion"
;;

let the_apply_rel_list_prog env argjl funj tycon =
let rec apply_rec acc typ = function
    [] -> (match tycon with NONE -> ()
                         | SOME typ' -> let _ = the_conv1_x typ typ' in ());
          {VAL=applist(j_val funj,map j_val (rev acc));
           TYP=typ;
           KIND = funj.KIND}
  | hj::restjl ->
  match whd_betadeltaiota (get_ise()) typ with
     DOP2(Prod,c1,DLAM(_,c2)) ->
     if the_conv1_x hj.TYP c1 then
         apply_rec (hj::acc) (subst1 hj.VAL c2) restjl
        else (match strip_outer_cast hj.VAL with
                DOP1(XTRA("COMMENT",[]),c) -> apply_rec (hj::acc) typ restjl
              | _ -> 
                (try let hj = attempt_coercion env c1 hj
                     in apply_rec (hj::acc) (subst1 hj.VAL c2) restjl
                 with UserError _ | Failure _ ->
                 errorlabstrm "apply"
                 [< 'S"Illegal application (Type Error) : "; 'CUT ;
                     apply_error env (funj.VAL,funj.TYP)
                                 (map (fun j -> (j.VAL,j.TYP)) argjl) >]))
   | _ -> errorlabstrm "apply"
              [< 'S"Illegal application (Non-functional construction) : ";
                 'CUT ;
                 apply_error env (funj.VAL,funj.TYP)
                                 (map (fun j -> (j.VAL,j.TYP)) argjl) >]
in apply_rec [] funj.TYP argjl;;

let cast_rel_prog sigma env cj tj =
    if conv1_x sigma cj.TYP tj.VAL then
        {VAL=j_val cj;
         TYP=tj.VAL;
         KIND = whd_betadeltaiota sigma tj.TYP}
    else errorlabstrm "cast"
          [< 'S"In environment " ; pENV env ; 'FNL ;
           'S"The term: " ; pTERMINENV(env,cj.VAL) ; 'SPC ; 'S"does not have type";
           'SPC ; pTERMINENV(env,tj.VAL) ; 'SPC ; 'S"Rather, it has type " ;
           'SPC ; pTERMINENV(env,cj.TYP) >]
;;

let the_cast_rel_prog env cj tj =
    if the_conv1_x cj.TYP tj.VAL then
        {VAL=j_val cj;
         TYP=tj.VAL;
         KIND = hnf_constr (get_ise()) tj.TYP}
    else (try attempt_coercion env tj.VAL cj
         with UserError _ | Failure _ ->
          errorlabstrm "cast"
          [< 'S"In environment " ; pENV env ; 'FNL ;
           'S"The term: " ; pTERMINENV(env,cj.VAL) ; 'SPC ; 
           'S"does not have type";
           'SPC ; pTERMINENV(env,tj.VAL) ; 'SPC ; 'S"Rather, it has type " ;
           'SPC ; pTERMINENV(env,cj.TYP) >])
;;

let type_of_existential sigma env (DOPN(Const sp,args) as K) =
if evc__ctxt_access sigma sp then
try let evd = evc__map sigma sp
    in if for_all2eq_vect (fun id (VAR id') -> id = id' | _ _ -> false) (vect_of_list (ids_of_sign evd.evgoal.hyps)) args
        & hyp_seq_ordered_inclusion evd.evgoal.hyps (get_globals env) then
        evd.evgoal.concl
       else 
        const_type sigma K
with Not_found -> error ((string_of_path sp) ^ " not declared")
   | Failure _ -> invalid_arg "the current environment and the constant's env of construction are incompatible"
else error ((string_of_path sp) ^ " not declared")
;;

let type_of_const sigma env (DOPN(Const sp,_) as K) =
let bn = basename sp
in if "?" = SA__un_intern(atompart_of_id bn) then
    type_of_existential sigma env K
   else type_of_constant sigma env K
;;

exception NotFound;;

let find_sorted_assoc p = findrec where rec findrec = function 
   ((a,ta)::l) -> if a < p then findrec l 
             else if a = p then ta else raise NotFound
  | _     -> raise NotFound;;

let case_branches v = sub_vect v 2 (vect_length v-2);;

let is_inst_var sigma k c = match whd_betadeltaiota_stack sigma c [] with 
     (Rel n,_) -> n=k
    | _        -> false;;

let map_lift_fst = map (function (n,t)->(n+1,t));;

let check_term mind_recvec F = crec where rec crec n l (lrec,c) = match
  (lrec,strip_outer_cast c) with
   (Norec::lr,DOP2(Lambda,_,DLAM(_,b))) -> 
                           let l' = map_lift_fst l in crec (n+1) l' (lr,b)
 | (Mrec(i)::lr,DOP2(Lambda,_,DLAM(_,b)))  -> let l' = map_lift_fst l in
                                crec (n+1) ((1,mind_recvec.(i))::l') (lr,b)
 | (Imbr(lc)::lr,DOP2(Lambda,_,DLAM(_,b))) -> let l' = map_lift_fst l in
                                crec (n+1) ((1,lc)::l') (lr,b)
 | _,f -> F n l f;;

let is_subterm_specif sigma lcx mind_recvec = crec where rec 
crec n lst c = 
   let (x,_) = whd_betadeltaiota_stack sigma c [] in match x with 
     (Rel k)         -> find_sorted_assoc k lst
|  DOPN(MutCase,vargs)  -> let c = vargs.(1) and br = case_branches vargs in
       if vect_length br = 0 then [||] else
       let lcv = (try if is_inst_var sigma n c then lcx else crec n lst c 
                  with NotFound -> (make_vect (vect_length br) []))
       in if vect_length br <> vect_length lcv then anomaly "is_subterm"
          else let stl = map2_vect 
            (fun lc a -> check_term mind_recvec crec n lst (lc,a)) lcv br 
               in stl.(0)
|  DOP2(Lambda,_,DLAM(_,b)) -> let lst' = map_lift_fst lst in
                                crec (n+1) lst' b
(***** Experimental change *************************)
|  DOP0(Meta _)             -> [||]
(***************************************************)
|  _                        -> raise NotFound;;

let is_subterm sigma lcx mind_recvec n lst c = 
try is_subterm_specif sigma lcx mind_recvec n lst c; true 
with NotFound -> false;;

  let noccur_with_meta n m term = let rec occur_rec n = function
    Rel(p)        -> if n<=p & p<n+m then raise Occur
  | VAR _         -> ()
  | DOPN(op,cl)   -> (match op with
                         AppL ->( match strip_outer_cast (cl.(0)) with
                                    DOP0 (Meta _) -> ()
                                   | _             -> do_vect (occur_rec n) cl)
                        | _    -> do_vect (occur_rec n) cl)
  | DOPL(_,cl)    -> do_list (occur_rec n) cl
  | DOP0(_)       -> ()
  | DOP1(_,c)     -> occur_rec n c
  | DOP2(_,c1,c2) -> occur_rec n c1; occur_rec n c2
  | DLAM(_,c)     -> occur_rec (n+1) c
  | DLAMV(_,v)    -> do_vect (occur_rec (n+1)) v
  | _             -> ()
  in try (occur_rec n term; true) with Occur -> false;;

let check_subterm_rec_meta_prog sigma vectn k def = 
if k < 0 then true else
let nfi = vect_length vectn in 
 (* check fi does not appear in the k+1 first abstractions, 
      gives the type of the k+1-eme abstraction  *)
let rec check_occur n def = match strip_outer_cast def with
      DOP2(Lambda,A,DLAM(_,B)) -> if noccur_with_meta n nfi A then
                      if n = k+1 then (A,B) else check_occur (n+1) B
                      else error "Bad occurrence of recursive call"
    | DOP2(XTRA("ANNOT",[]),c,_) -> check_occur n c
    | _ -> error "Not enough abstractions in the definition"
in let (C,d) = check_occur 1 def 
in let (DOPN(MutInd(sp,tyi),_) as MI,largs) = 
  try find_minductype sigma C 
  with Induc -> error "Recursive definition on a non inductive type" in

let mind_recvec = mind_recargs MI in 
let lcx = mind_recvec.(tyi) in

let rec check_rec_call n lst t = 
            (* n gives the index of the recursive variable *)
if noccur_with_meta (n+k+1) nfi t then true (* no recursive call in the term *)
else match whd_betadeltaiota_stack sigma t [] with 
     (Rel p,l) -> 
       if n+k+1 <= p & p < n+k+nfi+1 (* recursive call *)
       then let glob = nfi+n+k-p in  (* the index of the recursive call *) 
            let np = vectn.(glob) in (* the decreasing arg of the rec call *)
            if length l > np then 
                  let (la,(z::lrest)) = chop_list np l in 
                  if (is_subterm sigma lcx mind_recvec n lst z) 
                  then for_all (check_rec_call n lst) (la@lrest)
                  else error "Recursive call applied to an illegal term"
            else error  "Not enough arguments for the recursive call"
       else for_all (check_rec_call n lst) l        
    | (DOPN(MutCase,vargs),l) -> let P = vargs.(0) and c = vargs.(1) 
                                 and lrest = case_branches vargs in 
          let lc = try if is_inst_var sigma n c then lcx 
                        else is_subterm_specif sigma lcx mind_recvec n lst c
                   with NotFound -> make_vect (vect_length lrest) []
          in (for_all2eq_vect
              (fun c a -> check_term mind_recvec check_rec_call n lst (c,a))
              lc lrest) & (for_all (check_rec_call n lst) (c::P::l)) 
    | (DOP2(_,A,DLAM(_,B)),[]) ->   (check_rec_call n lst A)
                          & (check_rec_call (n+1) (map_lift_fst lst) B)
    | (DOP2(_,A,B),l) ->   (check_rec_call n lst A)
                          & (check_rec_call n lst B)
                          & (for_all (check_rec_call n lst) l)
    | (DOPN(_,la),l) ->   (for_all_vect (check_rec_call n lst) la)
                          & (for_all (check_rec_call n lst) l)
    | (DOP0 (Meta _),l) -> true
    | (_,l)    ->   for_all (check_rec_call n lst) l

in check_rec_call 1 [] d;;    
 

let check_fix_prog sigma (DOPN(Fix(nvect,j),vargs)) = 
 let nbfix = let nv = vect_length vargs in 
            if nv < 2 then error "Ill-formed recursive definition" else nv-1 in
  let varit = sub_vect vargs 0 nbfix in
  let ldef = last_vect vargs in
  let ln = vect_length nvect and la = vect_length varit in
  if ln <> la then error "Ill-formed fix term"
  else let (lna,vdefs) = decomp_DLAMV_name ln ldef in 
       let vlna = vect_of_list lna in
       let check_type i = 
         try check_subterm_rec_meta_prog sigma nvect nvect.(i) vdefs.(i) 
         with UserError (s,str) ->
           let ass = assumptions_for_print lna in
     errorlabstrm "check_fix_prog"
     (HOV 0 [< str; 'FNL; 
       'S"The " ; if ln = 1 then [<>] else [<'INT (i+1); 'S "-th ">];
       'S"recursive definition "; 'CUT ; pTERMINENV (ass,vdefs.(i)); 'CUT;
       'S " is not well-formed">])
           in for i = 0 to ln-1 do check_type i done;;

let unsafe_fmachine_prog metamap sign comsign constr =
    exemeta_rec (NONE,NONE) sign comsign constr
where rec exemeta_rec (valcon,tycon) env comenv =  function
  DOP0(XTRA("ISEVAR",[])) ->
  (match valcon with
   SOME val -> exemeta_rec (NONE,NONE) env comenv val
 | NONE ->
  (match tycon with
   NONE -> error "unconstrained isevar"
 | SOME ty -> let tyj = exemeta_rec (NONE,NONE) env comenv ty in
              let K = new_isevar env (j_val tyj)
              in exemeta_rec (NONE,NONE) env comenv K))
| DOP0(Meta n) -> (try (match assoc n metamap with
                        DOP2(Cast,typ,kind) ->
                        {VAL=DOP0(Meta n);
                         TYP=typ;
                         KIND=kind}
                      | typ ->
                        {VAL=DOP0(Meta n);
                         TYP=typ;
                         KIND=hnf_constr (get_ise()) (exemeta_rec (NONE,NONE) (GLOB(get_globals env)) (GLOB(get_globals comenv)) typ).TYP})
                    with Not_found -> error "Metavariable wasn't in the metamap")

|  Rel(n)                 -> prog_relative (get_ise()) n env

| (DOP2(Lambda,DOP1(XTRA("COMMENT",[]),C),DLAM(na,B))) ->
  (* [x:{C}]B --> Btype
     when C is well-typed in comenv and is purely logical

     [x:{C}]B --> Btype
     when C is wel-typed in comenv and is informative, and irrelevant
     because of levels

     [x:{C}]B --> (x:extC)Btype
     when C is well-typed in comenv and is informative and relevant
     the extraction of C is extC

     B is well-typed in the environments:
     -- add the extraction of C to env (Implicit if C is logical)
     -- add C to comenv
   *)

  let (Cj,Cinf) = mach__infexecute_rec (get_ise()) (comenv,env) C in
  let relevant = (match Cinf with
                  Logic -> false
                | Inf Cinfj ->
                  anomaly "Unimplemented(progmach)") in
      if not relevant then
          let env' = (add_rel (na,assumption_of_judgement (get_ise()) Cj) env) and
            comenv' = (add_rel (na,assumption_of_judgement (get_ise()) Cj) comenv) in
          let Bj = exemeta_rec (NONE,NONE) env' comenv' B
          in {VAL = (DOP2(Lambda,
                          DOP1(XTRA("COMMENT",[]),j_val Cj),
                          DLAM(na,j_val Bj)));
              TYP = pop Bj.TYP;
              KIND = Bj.KIND}
      else anomaly "Unimplemented(progmach)"

| (DOP2(XTRA("ANNOT",[]),c,T)) ->
  (* c::{T} --> ctype
     when T is well-typed in comenv and extraction_T is convertible
     with ctype
   *)
  let cj = exemeta_rec (valcon,tycon) env comenv c and
      ise_list = get_ise() and
      Tjcom = trad__is_fmachine false comenv T in
  set_ise ise_list;
  let (Tj,Tinf) = mach__infexecute_rec (get_ise()) (comenv,env) Tjcom in
      (match Tinf with
       Logic -> error "Annotation must be of computational nature"
     | Inf Tinfj ->
       traceme "cj" cj;
       traceme "Tinfj" Tinfj;
       if the_conv1_x Tinfj.VAL cj.TYP then
           {VAL=DOP2(XTRA("ANNOT",[]),j_val cj,j_val Tj);
            TYP = cj.TYP;
            KIND = cj.KIND}
       else error "Mistyped annotation in program")

| (DOP1(XTRA("COMMENT",[]),c)) ->
  (* c::{T} --> ctype
     when T is well-typed in comenv and extraction_T is convertible
     with ctype
   *)
  let (Tj,Tinf) = mach__infexecute_rec (get_ise()) (comenv,env) c in
      (match Tinf with
       Inf Tinfj -> error "must be a logical argument"
     | Logic ->
           {VAL=(DOP1(XTRA("COMMENT",[]),j_val Tj));
            TYP = Tj.TYP;
            KIND = Tj.KIND})

| (DOPN(XTRA("REC",[]),cl)) ->  
  let P = cl.(0)
  and c = cl.(1)
  and lf = sub_vect cl 2 ((vect_length cl) - 2) in
  let {TYP=ct;_} as cj = exemeta_rec (NONE,NONE) env comenv c
  and Pj = exemeta_rec (NONE,NONE) env comenv P in
  let P = Pj.VAL and
      c = cj.VAL in
  let (bty,rsty) = indrec__type_rec_branches true (get_ise()) env ct Pj.TYP P c in
  let lfj = map2_vect (fun tyc f -> exemeta_rec (NONE,SOME tyc) env comenv f) bty lf in
  let lft = (map_vect (fun j -> j.TYP) lfj) in
     let REC = concat_vect [|P; c|] (map_vect j_val lfj) in
      {VAL=indrec__transform_rec env (get_ise()) REC (cj.TYP,Pj.TYP);
       TYP = rsty;
       KIND = sort_of_arity (get_ise()) Pj.TYP}

| (VAR id as val) ->
  let (DOP2(Cast,typ,kind)) = snd(lookup_glob id env)
  and str = id_of_var val
  in {VAL=val;
      TYP=typ;
      KIND=kind}

| (DOPN(Abst _,_) as val) ->
   if evaluable_abst val then
       exemeta_rec (NONE,NONE) env comenv (abst_value val)
   else error "Cannot typecheck an unevaluable abstraction"

| (DOPN(Const sp,_) as val) ->
    (match type_of_const (get_ise()) env val with
      (DOP2(Cast,typ,kind)) -> {VAL=val;TYP=typ;KIND = kind}
 | typ -> {VAL=val;TYP=typ;
           KIND= hnf_constr (get_ise()) (exemeta_rec (NONE,NONE) env comenv typ).TYP})

| DOPN(MutInd _,_) as val ->
  let (DOP2(Cast,typ,kind)) = type_of_mind (get_ise()) env val
  in {VAL=val;
      TYP=typ;
      KIND=kind}
 
| DOPN(MutConstruct _,_) as val          -> 
  let (DOP2(Cast,typ,kind)) = type_of_mconstr (get_ise()) env val
  in {VAL=val;
      TYP=typ;
      KIND=kind}

| DOP0(Sort(Prop(c)))                 -> fcn_proposition c
| DOP0(Sort(Type(c,u)))               -> fcn_type_with_universe (c,u)
| DOPN(AppL,tl)           -> 
  let j = exemeta_rec (NONE,NONE) env comenv (hd_vect tl) in
  let jl = rev(snd(it_list (fun (tycon,jl) c ->
                            let dtc = prod_dom_tycon (get_ise()) tycon in
                            let cj = exemeta_rec (NONE,dtc) env comenv c in
                            let rtc = app_rng_tycon (get_ise()) cj.VAL tycon
                            in (rtc,cj::jl)) (SOME j.TYP,[]) (list_of_tl_vect tl)))
  in
      the_apply_rel_list_prog env jl j tycon

| DOP2(Lambda,c1,DLAM(name,c2))      -> 
  let jfw = exemeta_rec (lambda_dom_tycon (get_ise()) tycon,NONE) env comenv c1 in
  let jcci = try execute_pure_rec (get_ise()) comenv jfw.VAL 
             with UserError _ -> jfw in
  let assumfw = assumption_of_judgement (get_ise()) jfw in
  let varfw = (name,assumfw) in
  let assumcci = assumption_of_judgement_prog (get_ise()) jcci in
  let varcci = (name,assumcci) in
  let j' = exemeta_rec (NONE,prod_rng_tycon (get_ise()) tycon) 
                       (add_rel varfw env) (add_rel varcci comenv) c2 in 
      abs_rel (get_ise()) name assumfw j'

| DOP2(Prod,c1,DLAM(name,c2))        ->
  let jfw = exemeta_rec (NONE,NONE) env comenv c1 in
 let jcci = try execute_pure_rec (get_ise()) comenv jfw.VAL 
             with UserError _ -> jfw in
  let assumfw = assumption_of_judgement (get_ise()) jfw in
  let varfw = (name,assumfw) in
  let assumcci = assumption_of_judgement_prog (get_ise()) jcci in
  let varcci = (name,assumcci) in
  let j' = exemeta_rec (NONE,NONE) (add_rel varfw env) 
                                   (add_rel varcci comenv) c2 in
      gen_rel (get_ise()) env name assumfw j'

| (DOPN(MutCase,cl) as val) ->
  let P = cl.(0)
  and c = cl.(1)
  and lf = sub_vect cl 2 ((vect_length cl) - 2) in
  (let {TYP=ct;_} as cj = exemeta_rec (NONE,NONE) env comenv c in
   let {TYP=Pt;VAL=P} as Pj = exemeta_rec (NONE,NONE) env comenv P in
   let (bty,rsty) = type_case_branches env (get_ise()) ct Pt P c in
   let lfj = map2_vect (fun tyc f -> exemeta_rec (NONE,SOME tyc) env comenv f) bty lf in
   let lft = (map_vect (fun j -> j.TYP) lfj)
   in {VAL=DOPN(MutCase,concat_vect [|j_val Pj; j_val cj|] (map_vect j_val lfj));
        TYP=the_type_case_elim_pg env (get_ise()) ct Pj.TYP lft P c;
        KIND = sort_of_arity (get_ise()) Pj.TYP})

| (DOPN(Fix(vn,i),cl) as val) ->
  if  exists_vect (fun n -> n < 0) vn then 
      error "General Fixpoints not allowed";
  let nbfix = let nv = vect_length cl in 
           if nv < 2 then error "Ill-formed recursive definition" else nv-1 in
  let lar = sub_vect cl 0 nbfix in
  let ldef = last_vect cl in
  let (lfi,vdef) = decomp_DLAMV_name nbfix ldef in
  let newenv,newcomenv = it_vect2 
     (fun (env,comenv) name ar -> 
        let arfw = exemeta_rec (NONE,NONE) env comenv ar in
        let assumfw = assumption_of_judgement (get_ise()) arfw in
        let arcci = try execute_pure_rec (get_ise()) comenv arfw.VAL 
                    with UserError _ -> arfw in
        let assumcci = assumption_of_judgement_prog (get_ise()) arcci
        in (add_rel (name,assumfw) env),(add_rel (name,assumcci) comenv))
     (env,comenv) (vect_of_list (rev lfi)) (vect_lift lar) in
  let larj = map_vect (exemeta_rec (NONE,NONE) env comenv) lar in
  let vdefj = map_i_vect (fun i def -> exemeta_rec (NONE,SOME(larj.(i).VAL)) newenv newcomenv def) 0 vdef in
  let vdefv = put_DLAMSV lfi (map_vect j_val vdefj) in
  let larv = map_vect j_val larj in
  let fix = DOPN(Fix(vn,i),concat_vect larv [|vdefv|])
  in check_fix_prog (get_ise()) fix;
  {VAL= fix;
   TYP=evar_type_fix fix (map_vect (fun j -> j.TYP) vdefj);
   KIND=larj.(i).TYP}

| (DOPN(CoFix i,cl) as val) ->
(*  if nofix  
  then error "General Fixpoints not allowed"
  else *)
       let nbfix = let nv = vect_length cl in 
                       if nv < 2 
                       then error "Ill-formed recursive definition" 
                       else nv-1 in
       let lar = sub_vect cl 0 nbfix in
       let ldef = last_vect cl in
       let (lfi,vdef) = decomp_DLAMV_name nbfix ldef in
       let newenv,newcomenv = it_vect2 
            (fun (env,comenv) name ar -> 
               let arfw = exemeta_rec (NONE,NONE) env comenv ar in
               let assumfw = assumption_of_judgement (get_ise()) arfw in
               let arcci = try execute_pure_rec (get_ise()) comenv arfw.VAL 
                           with UserError _ -> arfw in
               let assumcci = assumption_of_judgement_prog (get_ise()) arcci in
                 (add_rel (name,assumfw) env),(add_rel (name,assumcci) comenv))
           (env,comenv) (vect_of_list (rev lfi)) lar in
       let larj  = map_vect (exemeta_rec (NONE,NONE) env comenv) lar in
       let vdefj = map_i_vect (fun i def -> exemeta_rec (NONE,SOME(larj.(i).VAL)) newenv newcomenv def) 0 vdef in
       let vdefv = put_DLAMSV lfi (map_vect j_val vdefj) in
       let larv  = map_vect j_val larj in
       let cofix = DOPN(CoFix i,concat_vect larv [|vdefv|])
       in check_cofix (get_ise()) cofix;
          {VAL= cofix;
           TYP=evar_type_cofix cofix (map_vect (fun j -> j.TYP) vdefj);
           KIND=larj.(i).TYP}

| DOP2(Cast,c,t) ->
   let tj = exemeta_rec (NONE,NONE) env comenv t in
   let cj = exemeta_rec (NONE,SOME (assumption_of_judgement (get_ise()) tj)) env comenv c in
      the_cast_rel_prog env cj tj

(*|  DOP0(XTRA("ISEVAR",[])) -> errorlabstrm "unsafe_machine_prog"
            [< 'S"No implicit syntax; all type arguments are needed" >] *)

| _ -> error "Cannot be executed";;

let rec whd_ise = function
    (DOPN(Const sp,_) as K) ->
    if ise_defined K & ise_in_dom sp then
        whd_ise (const_value (get_ise()) K)
    else if ise_in_dom sp then
        failwith "whd_ise"
    else K
  | DOP2(Cast,c,_) -> whd_ise c
  | DOP0(Sort(Type(c,_))) -> DOP0(Sort(Type(c,Dummy_univ)))
  | c -> c
;;

let is_fmachine_prog metamap (comsign,sign) c =
conversion_problems := [];
solved_problems := [];
set_ise mt_evc;
let j = unsafe_fmachine_prog metamap sign comsign c
in strong whd_ise j.VAL;;


let unsafe_machine_prog (nocheck,noverify) (sigma,metamap) sign comsign constr =
    exemeta_rec sign comsign constr
where rec exemeta_rec env comenv =  function
  DOP0(Meta n) -> (try (match assoc n metamap with
                        DOP2(Cast,typ,kind) ->
                        {VAL=DOP0(Meta n);
                         TYP=typ;
                         KIND=kind}
                      | typ ->
                        {VAL=DOP0(Meta n);
                         TYP=typ;
                         KIND=hnf_constr sigma (exemeta_rec (GLOB(get_globals env)) (GLOB(get_globals comenv)) typ).TYP})
                    with Not_found -> error "Metavariable wasn't in the metamap")

|  Rel(n)                 -> prog_relative sigma n env

| (DOP2(Lambda,DOP1(XTRA("COMMENT",[]),C),DLAM(na,B))) ->
  (* [x:{C}]B --> Btype
     when C is well-typed in comenv and is purely logical

     [x:{C}]B --> Btype
     when C is wel-typed in comenv and is informative, and irrelevant
     because of levels

     [x:{C}]B --> (x:extC)Btype
     when C is well-typed in comenv and is informative and relevant
     the extraction of C is extC

     B is well-typed in the environments:
     -- add the extraction of C to env (Implicit if C is logical)
     -- add C to comenv
   *)

  let (Cj,Cinf) = mach__infexecute_rec sigma (comenv,env) C in
  let relevant = (match Cinf with
                  Logic -> false
                | Inf Cinfj ->
                  anomaly "Unimplemented(progmach)") in
      if not relevant then
          let env' = (add_rel (na,assumption_of_judgement sigma Cj) env) and
            comenv' = (add_rel (na,assumption_of_judgement sigma Cj) comenv) in
          let Bj = exemeta_rec env' comenv' B
          in {VAL = (DOP2(Lambda,
                          DOP1(XTRA("COMMENT",[]),j_val Cj),
                          DLAM(na,j_val Bj)));
              TYP = pop Bj.TYP;
              KIND = Bj.KIND}
      else anomaly "Unimplemented(progmach)"

| (DOP2(XTRA("ANNOT",[]),c,T)) ->
  (* c::{T} --> ctype
     when T is well-typed in comenv and extraction_T is convertible
     with ctype
   *)
  let cj = exemeta_rec env comenv c and
      (Tj,Tinf) = mach__infexecute_rec sigma (comenv,env) T in
      (match Tinf with
       Logic -> error "Annotation must be of computational nature"
     | Inf Tinfj ->
       traceme "cj" cj;
       traceme "Tinfj" Tinfj;
       if conv1_x sigma Tinfj.VAL cj.TYP then
           {VAL=DOP2(XTRA("ANNOT",[]),j_val cj,j_val Tj);
            TYP = cj.TYP;
            KIND = cj.KIND}
       else error "Mistyped annotation in program")

| (DOP1(XTRA("COMMENT",[]),c)) ->
  (* c::{T} --> ctype
     when T is well-typed in comenv and extraction_T is convertible
     with ctype
   *)
  let (Tj,Tinf) = mach__infexecute_rec sigma (comenv,env) c in
      (match Tinf with
       Inf Tinfj -> error "must be a logical argument"
     | Logic ->
           {VAL=(DOP1(XTRA("COMMENT",[]),j_val Tj));
            TYP = Tj.TYP;
            KIND = Tj.KIND})

| (DOPN(XTRA("REC",[]),cl)) ->  let P = cl.(0)
  and c = cl.(1)
  and lf = sub_vect cl 2 ((vect_length cl) - 2) in
  let {TYP=ct;_} as cj = exemeta_rec env comenv c
  and {TYP=Pt;_} as Pj = exemeta_rec env comenv P in
        exemeta_rec env comenv (indrec__transform_rec env sigma cl (ct,Pt)) 

| (VAR id as val) ->
  let (DOP2(Cast,typ,kind)) = snd(lookup_glob id env)
  and str = id_of_var val
  in {VAL=val;
      TYP=typ;
      KIND=kind}

| (DOPN(Abst _,_) as val) ->
   if evaluable_abst val then
       exemeta_rec env comenv (abst_value val)
   else error "Cannot typecheck an unevaluable abstraction"

| (DOPN(Const sp,_) as val) ->
  let (DOP2(Cast,typ,kind)) = type_of_const sigma env val
  in {VAL=val;
      TYP=typ;
      KIND = kind}

| DOPN(MutInd _,_) as val ->
  let (DOP2(Cast,typ,kind)) = type_of_mind sigma env val
  in {VAL=val;
      TYP=typ;
      KIND=kind}
 
| DOPN(MutConstruct _,_) as val          -> 
  let (DOP2(Cast,typ,kind)) = type_of_mconstr sigma env val
  in {VAL=val;
      TYP=typ;
      KIND=kind}

| DOP0(Sort(Prop(c)))                 -> fcn_proposition c
| DOP0(Sort(Type(c,u)))               -> fcn_type_with_universe (c,u)
| DOPN(AppL,tl)           -> 
  let j = exemeta_rec env comenv (hd_vect tl) in
  let jl = map_vect_list (exemeta_rec env comenv) (tl_vect tl) in
      apply_rel_list_prog sigma env nocheck jl j

| DOP2(Lambda,c1,DLAM(name,c2))      -> 
  let jfw = exemeta_rec env comenv c1 in
  let jcci = try execute_pure_rec sigma comenv jfw.VAL 
             with UserError _ -> jfw in
  let assumfw = assumption_of_judgement sigma jfw in
  let varfw = (name,assumfw) in
  let assumcci = assumption_of_judgement_prog sigma jcci in
  let varcci = (name,assumcci) in
  let j' = exemeta_rec (add_rel varfw env) (add_rel varcci comenv) c2 in 
      abs_rel sigma name assumfw j'

| DOP2(Prod,c1,DLAM(name,c2))        ->
  let jfw = exemeta_rec env comenv c1 in
 let jcci = try execute_pure_rec sigma comenv jfw.VAL 
             with UserError _ -> jfw in
  let assumfw = assumption_of_judgement sigma jfw in
  let varfw = (name,assumfw) in
  let assumcci = assumption_of_judgement_prog sigma jcci in
  let varcci = (name,assumcci) in
  let j' = exemeta_rec (add_rel varfw env) (add_rel varcci comenv) c2 in
      gen_rel sigma env name assumfw j'

| (DOPN(MutCase,cl) as val) ->
  let P = cl.(0)
  and c = cl.(1)
  and lf = sub_vect cl 2 ((vect_length cl) - 2) in
  (let {TYP=ct;_} as cj = exemeta_rec env comenv c in
   let Pj = exemeta_rec env comenv P in
   let lfj = map_vect (exemeta_rec env comenv) lf in
   let lft = (map_vect (fun j -> j.TYP) lfj)
   in  {VAL=DOPN(MutCase,concat_vect [|j_val Pj; j_val cj|] (map_vect j_val lfj));
        TYP=type_case_elim_pg env sigma ct Pj.TYP lft P c;
        KIND = sort_of_arity sigma Pj.TYP})

| (DOPN(Fix(vn,i),cl) as val) ->
  if  exists_vect (fun n -> n < 0) vn then 
      error "General Fixpoints not allowed";
  let nbfix = let nv = vect_length cl in 
           if nv < 2 then error "Ill-formed recursive definition" else nv-1 in
  let lar = sub_vect cl 0 nbfix in
  let ldef = last_vect cl in
  let (lfi,vdef) = decomp_DLAMV_name nbfix ldef in
  let newenv,newcomenv = it_vect2 
     (fun (env,comenv) name ar -> 
        let arfw = exemeta_rec env comenv ar in
        let assumfw = assumption_of_judgement sigma arfw in
        let arcci = try execute_pure_rec sigma comenv arfw.VAL 
                    with UserError _ -> arfw in
        let assumcci = assumption_of_judgement_prog sigma arcci
        in (add_rel (name,assumfw) env),(add_rel (name,assumcci) comenv))
     (env,comenv) (vect_of_list (rev lfi)) (vect_lift lar) in
  let larj = map_vect (exemeta_rec env comenv) lar in
  let vdefj = map_vect (exemeta_rec newenv newcomenv) vdef in
  let vdefv = put_DLAMSV lfi (map_vect j_val vdefj) in
  let larv = map_vect j_val larj in
  let fix = DOPN(Fix(vn,i),concat_vect larv [|vdefv|])
  in check_fix_prog sigma fix;
  {VAL= fix;
   TYP=type_fix sigma fix (map_vect (fun j -> j.TYP) vdefj);
   KIND=larj.(i).TYP}

| (DOPN(CoFix i,cl) as val) ->
(*  if nofix  
  then error "General Fixpoints not allowed"
  else *)
       let nbfix = let nv = vect_length cl in 
                       if nv < 2 
                       then error "Ill-formed recursive definition" 
                       else nv-1 in
       let lar = sub_vect cl 0 nbfix in
       let ldef = last_vect cl in
       let (lfi,vdef) = decomp_DLAMV_name nbfix ldef in
       let newenv,newcomenv = it_vect2 
            (fun (env,comenv) name ar -> 
               let arfw = exemeta_rec env comenv ar in
               let assumfw = assumption_of_judgement sigma arfw in
               let arcci = try execute_pure_rec sigma comenv arfw.VAL 
                           with UserError _ -> arfw in
               let assumcci = assumption_of_judgement_prog sigma arcci in
                 (add_rel (name,assumfw) env),(add_rel (name,assumcci) comenv))
           (env,comenv) (vect_of_list (rev lfi)) (vect_lift lar) in
       let larj  = map_vect (exemeta_rec env comenv) lar in
       let vdefj = map_vect (exemeta_rec newenv newcomenv) vdef in
       let vdefv = put_DLAMSV lfi (map_vect j_val vdefj) in
       let larv  = map_vect j_val larj in
       let cofix = DOPN(CoFix i,concat_vect larv [|vdefv|])
       in check_cofix sigma cofix;
          {VAL= cofix;
           TYP=type_cofix sigma cofix (map_vect (fun j -> j.TYP) vdefj);
           KIND=larj.(i).TYP}

| DOP2(Cast,c,t) ->
  let cj = exemeta_rec env comenv c in
  let tj = exemeta_rec env comenv t
  in cast_rel_prog sigma env cj tj

|  DOP0(XTRA("ISEVAR",[])) -> errorlabstrm "unsafe_machine_prog"
            [< 'S"No implicit syntax; all type arguments are needed" >]

| _ -> error "Cannot be executed";;

let core_machine_prog nocheck = unsafe_machine_prog (nocheck,false);;

(* WITHOUT INFORMATION *)
let machine_prog nocheck metamap sign comsign constr =
let j = core_machine_prog nocheck metamap sign comsign constr
in {VAL=j.VAL;TYP=strong strip_outer_cast j.TYP;KIND=j.KIND};;

let exemeta_prog sigma metamap = machine_prog false (sigma,metamap);;
let execute_rel sigma = machine_prog false (sigma,[]);;
let fexecute_prog sigma sign comsign = machine_prog false (sigma,[]) (GLOB sign) (GLOB comsign);;

let absfun_fw =
(fun sigma env id ->
    if (atompart_of_id id = SA__intern "?") then
        error "existentials in fterms not implemented"
    else (try VAR(fst(lookup_glob id env))
          with Not_found ->
          (try search_freference env id
           with UserError _ -> (try search_synconst FW id
                             with Not_found -> error((string_of_id id)^"  not declared")))))
;;

let absolutize_prog sigma env c = 
let fsign = fsign_of_sign sigma (get_globals env)
and global_env = env
in (absrec (GLOB fsign) global_env c
    where rec absrec env globenv = function
    VAR id -> absfun_fw sigma env id
  | Rel _ as t -> t

  | (DOP2(Lambda,DOP1(XTRA("COMMENT",[]),C),DLAM(na,B))) ->
    DOP2(Lambda,DOP1(XTRA("COMMENT",[]),absolutize_cci sigma globenv C),
         DLAM(na,absrec (add_rel (na,()) env) (add_rel (na,()) globenv) B))
  | (DOP2(XTRA("ANNOT",[]),c,T)) ->
    (DOP2(XTRA("ANNOT",[]),(absrec env globenv) c,
          absolutize_cci sigma globenv T))

  | (DOP1(XTRA("COMMENT",[]),c)) ->
    (DOP1(XTRA("COMMENT",[]),absolutize_cci sigma globenv c))

  | DOP0 _ as t -> t
  | DOP1(oper,c) -> DOP1(oper,absrec env globenv c)
  | DOP2(oper,c1,c2) -> DOP2(oper,absrec env globenv c1,absrec env globenv c2)
  | DOPN(oper,cl) -> DOPN(oper,map_vect (absrec env globenv) cl)
  | DOPL(oper,cl) -> DOPL(oper,map (absrec env globenv) cl)
  | DLAM(na,c) -> DLAM(na,absrec (add_rel (na,()) env) (add_rel (na,()) globenv) c)
  | DLAMV(na,cl) -> DLAMV(na,map_vect (absrec (add_rel (na,()) env) (add_rel (na,()) globenv)) cl))
;;

let dbize_prog sigma env c =
    let t = (dbize env c) in
    try absolutize_prog sigma env t
    with Not_found -> 
errorlabstrm "absolutize_prog"
            [< 'S"During the relocation of global references," ; 'SPC ;
               'S"the variable(s) [" ; 'SPC ;
               prlist (fun id ->
                           try (absolutize_prog sigma env (VAR id);
                                [< >])
                           with _ -> [< 'S (string_of_id id) ; 'SPC >])
               (global_vars t) ;
               'S"]" ; 'SPC ; 'S"were not found in the current environment" >]
      | e ->
            errorlabstrm "absolutize_prog"
            [< 'S"During the relocation of global references," ; 'SPC ;
               'S"The error " ; errors__explain_sys_exn e ;
               'S"was raised. Perhaps a term of the Realizer is not an FW term and you then have to replace it by its extraction." >]
;;

(*
let raw_prog_of_progcom sigma sign c =
    sosub__soexecute(dbize_prog sigma (GLOB sign) c)
;;
*)

let execute_prog_rel sigma (env,fenv) c = execute_rel sigma fenv env c;;
let execute_meta_prog_rel sigma (env,fenv) metamap c = exemeta_prog sigma metamap fenv env c;;
let type_of_prog_rel sigma (env,fenv) c = (execute_prog_rel sigma (env,fenv) c).TYP;;
let type_meta_of_prog_rel sigma (env,fenv) metamap c = (execute_meta_prog_rel sigma (env,fenv) metamap c).TYP;;
let kind_of_prog_rel sigma (env,fenv) c = (execute_prog_rel sigma (env,fenv) c).KIND;;

let execute_prog sigma signp c = execute_prog_rel sigma (FGLOB signp) c;;
let execute_meta_prog sigma signp metamap c = execute_meta_prog_rel sigma (FGLOB signp) metamap c;;
let type_of_prog sigma signp c = type_of_prog_rel sigma (FGLOB signp) c;;
let type_meta_of_prog sigma signp metamap c = type_meta_of_prog_rel sigma (FGLOB signp) metamap c;;

(*
let raw_prog_of_progcom sigma sign c =
    (execute_prog sigma (FSIGN sigma sign) (sosub__soexecute(dbize_prog sigma (GLOB sign) c))).VAL
;;
*)
let raw_prog_of_progcom sigma sign c =
    is_fmachine_prog [] (FGLOB (FSIGN sigma sign)) (sosub__soexecute(dbize_prog sigma (GLOB sign) c)) ;;

(* $Id: progmach.ml,v 1.12 1996/02/12 15:56:53 cpaulin Exp $ *)
