(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                program.ml                                *)
(****************************************************************************)
#infix "o";;
#open "std";;
#open "pp";;
#open "stdpp";;
#open "names";;
#open "vectops";;
#open "generic";;
#open "term";;
#open "reduction";;
#open "termenv";;
#open "more_util";;
#open "mach";;
#open "machops";;
#open "proof_trees";;
#open "refiner";;
#open "proof";;
#open "tacmach";;
#open "progmach";;
#open "constrtypes";;
#open "clenv";;
#open "initial";;
#open "tactics";;
#open "indtypes";;
#open "printer";;
#open "pfedit";;
#open "cofix";;
#open "himsg";;

#open "tactics1";;
#open "tactics2";;
#open "tactics3";;

#infix "THEN";;
#infix "THENL";;
#infix "THENS";;
#infix "ORELSE";;

let METAMAP = ref ([]:((int*constr) list));;
let clear_METAMAP = METAMAP:=[];;

(* ====================== *)
(* List of new exceptions *)
(* ====================== *)

(* raised when specif_of_annot fails *)
exception NotProgram ;;

(* raised when conclusion needs an expansion *)
exception Expand ;;

(* raised when program needs an expansion *)
exception Generalize of int ;;

(* raised when there are not two imbricated eliminations *)
exception Elim;;

(* raised when a program must not be modified by the modif_app_ind function *)
exception NoModif ;;
exception Rewrite;;

(* ================================== *)
(* List of different useful functions *)
(* ================================== *)

(* A more appropriated function to check the dependency between two terms *) 
let dependant =
 let rec deprec M = function t ->
     (eq_term M t)
     or (match t with
     VAR _       -> false
   | DOP1(_,c)   -> deprec M c
   | DOP2(_,c,t) -> deprec M c or deprec M t
   | DOPN(Const _,_) -> eq_term M t
   | DOPN(_,cl)  -> exists_vect (deprec M) cl
   | DOPL(_,cl)  -> exists (deprec M) cl
   | DLAM(_,c)   -> deprec (lift 1 M) c
   | DLAMV(_,v)  -> exists_vect (deprec (lift 1 M)) v
   | _ -> false)
 in deprec;;

(* Checks if a term t = (MutInd() a1 ... an) *)
let rec is_mind = function
  DOPN(MutInd _,_) -> true
| DOP2(Cast,t,_) -> is_mind t
| DOPN(AppL,args) -> is_mind args.(0)
| _ -> false ;;

(* Tries to execute the term t as a pure term of the CCI.
   If it fails executes t as a program 
   Entries : sigma = ex.var. 
             sign = signature
             t = term
   Output : judgement *)
let try_execute sigma sign t = 
  try execute_pure sigma sign t
  with UserError _ -> execute_meta_prog sigma (FSIGN sigma sign) !METAMAP t ;;

(* Suppression of the outermost annotation of a term if it exists *)
let rec without_annot = function
  (DOP2(XTRA("ANNOT",[]),c,T))  -> c
| DOP2(Cast,t,_) -> without_annot t
| t -> t ;;

(* Extraction of the annotation of a term *)
let rec keep_only_annot = function
  (DOP2(XTRA("ANNOT",[]),c,T))  -> T
| DOP2(Cast,t,_) -> keep_only_annot t
| t -> t ;;

(* List of functions to create a clausal environment with a program *)
let ctxt_type_of_prog evc sign c = 
  type_of_prog evc (FSIGN evc sign) c;;

let w_type_of_prog wc c = ctxt_type_of_prog (w_Underlying wc) (w_hyps wc) c;;

let mk_clenv_hnf_constr_type_of_prog wc t =
    mk_clenv_from wc (t,w_hnf_constr wc (w_type_of_prog wc t));;

(* Entries : wc = walking context
             arg = program
             gls = goal *)
let make_clenv wc arg gls = 
  try mk_clenv_hnf_constr_type_of wc 
    (try_execute (Project gls) (pf_hyps gls) (keep_only_annot arg)).VAL 
  with UserError _ -> mk_clenv_hnf_constr_type_of_prog wc 
    (try_execute (Project gls) (pf_hyps gls) (keep_only_annot arg)).VAL ;;

(* Suppression of the outermost comment of a term if it exists *)
let rec without_comment = function
  (DOP1(XTRA("COMMENT",[]),c))  -> c
| DOP2(Cast,t,_) -> without_comment t
| t -> t ;;

(* Checks if a sort is informative or not *)
let rec informative_sort = function
  DOP0(Sort(Prop Pos)) -> true
| DOP0(Sort(Type(_))) -> true
| DOP0(Sort(Prop Null)) -> false
| DOP2(Prod,_,DLAM(_,B)) -> informative_sort B
| DOP2(Cast,t,_) -> informative_sort t
| _ -> anomaly"informative_sort" ;;

let pf_informative_type gls t = 
    is_info_cast_type (Project gls) (DOP2(Cast,t,pf_type_of gls t));;

let clenv_is_info_type clenv c t = 
    is_info_cast_type (w_Underlying clenv.hook) (DOP2(Cast,c,t));;

(* Extracts the kind of a term 
   Entries : sigma = ex. var.
             sign = signature
             t = term 
   Output : kind *)
let kind_of sigma sign t = (fexecute sigma sign t).KIND ;;

(* The same with different entries
   Entries : gls = goal
             t = term *)
let pf_kind_of gls t = kind_of (Project gls) (pf_hyps gls) t ;;

(* Extraction of a term
   Entries : gls = goal
             t = term *)
let pf_extraction gls t =
  match snd(infexecute (Project gls) (FSIGN (Project gls) (pf_hyps gls))  t) with
  Logic -> error "pf_extraction : not informative"
| Inf j -> j.VAL
;;

(* Checks if a term is logical or not
   Entries : l = list of bounded var.
             P = term to be checked
             gls = goal *)
let is_logical l P gls = isrec l P where rec isrec lv P =
let sigma = Project gls and sign = pf_hyps gls in
 match P with
  DOP2(Lambda,A,DLAM(_,B)) -> isrec ((isrec lv A)::lv) B
| DOP2(Cast,_,DOP0(Sort(Prop c))) -> c=Null
| DOP2(Cast,c,_) -> isrec lv c
| DOP2(Prod,A,DLAM(_,B)) -> isrec ((isrec lv A)::lv) B
| DOPN(AppL,args) -> isrec lv (args.(0))
| DOPN(Const _,_) -> let j = try fexecute sigma sign P
                             with UserError _ -> execute_meta_prog sigma (FSIGN sigma sign) !METAMAP P in
                      not(is_info_judge sigma j)
| VAR _           -> let j = try fexecute sigma sign P
                             with UserError _ -> execute_meta_prog sigma (FSIGN sigma sign) !METAMAP P in
                      not(is_info_judge sigma j)
| DOPN(MutInd _,_)   -> let j = try fexecute sigma sign P 
                             with UserError _ -> execute_meta_prog sigma (FSIGN sigma sign) !METAMAP P in
                      not(is_info_judge sigma j)
| Rel(n) -> nth lv n
| DOP0(Sort(Prop c)) -> (c=Null)
| DOP0(Sort(Type(_))) -> false
| DOPN(MutCase,args) -> isrec lv (args.(1))
| DOPN(Fix _,args) -> isrec lv (args.(0))
| DOPN(CoFix _,args) -> isrec lv (args.(0))
| DOP2(XTRA("ANNOT",[]),c,_) -> isrec lv c
| DOP1(XTRA("COMMENT",[]),c) -> true
| _ -> error"is_logical" ;;

(* Split a term with a head and its arguments
   Output : (hd,args) = head, list of arguments *)
let list_arg gls = function
  DOPN(AppL,args) -> lrec (list_of_vect args), tl(list_of_vect args)
                     where rec lrec l = 
                    (match l with
                       x::rest -> 
                        let sigma = Project gls and sign = pf_hyps gls in
                        let j = execute_meta_prog sigma (FSIGN sigma sign) !METAMAP x in
                          if not(is_info_judge sigma j)
                          then lrec rest
                          else x::(lrec rest)
                     | [] -> [])
| x -> [x],[] ;;

(* Gives the number of informative arguments corresponding to a type *)
let nb_info_arg gls = 
  let rec nrec ni l = function
    DOP2(Prod,A,DLAM(na,B)) -> if (is_logical l A gls) then
                                nrec ni (true::l) B
                               else
                                nrec (ni+1) (false::l) B
  | DOP2(Cast,c,_) -> nrec ni l c                                
  | _ -> ni
in nrec 0 [] ;;

(* Checks if a term is a constructor or a mutconstructor or not *)
let rec is_construct = function DOPN(MutConstruct _,_) -> true 
                          | DOP2(Cast,t,_) -> is_construct t
                          | _ -> false ;;

let rec is_construct_head = function DOPN(MutConstruct _,_) -> true 
                          | DOP2(Cast,t,_) -> is_construct_head t
                          | DOPN(AppL,args) -> is_construct_head args.(0)
                          | _ -> false ;;

(* Reduces completely a type
   Output : (b,t) = boolean, term
   the boolean is true if the head of the type is bounded
   the term is the initial type completely reduced *)
let rec red_all_lie gls = function
  DOP2(Prod,c1,DLAM(na,c2)) -> let (b,c2') = red_all_lie gls c2 in 
                                  b, DOP2(Prod,c1,DLAM(na,c2'))
| DOP2(Cast,c,_) -> red_all_lie gls c
| t -> let (x,l) = pf_whd_betadeltaiota_stack gls t [] in  
        match strip_outer_cast x with
          DOP2(Prod,_,_) -> red_all_lie gls x
        | Rel(_) -> true, applist(x,l)
        | _ -> false, applist(x,l) ;;

(* Extracts only the reduced type from the previous function *)
let red_all gls = snd o (red_all_lie gls) ;;

(* Gives the list of the types of the constructors of an inductive type
   Entry : ity = inductive type
   Output : list of the types of the constructors *)
let mconstructor_types ity =
let lc = mis_lc (mind_specif_of_mind ity)
in SAPPV lc ity
;;

(* Suppresses as many products in a constructor type as parameters
   Entries : nparams = number of parameters
             t = type where the products have to be suppressed
   Output : new type *)            
let without_nparams nparams t = crec nparams t where rec crec n = function
  DOP2(Cast,c,_) -> crec n c
| (DOP2(Prod,A,DLAM(_,B)) as C) -> if n=0 then C else crec (n-1) B
| c -> c ;;

(* Checks if a mutual inductive type ity is a singleton 
   type with respect to a given type typ
   Entries : sigma = ex. var.
             env = signature
             typ = underlying type of the singleton type
             ity = inductive type
             largs = list of arguments of the inductive type
             gls = goal
   Output : boolean *)
(*****
let is_singleton sigma env typ (ity,largs) =
let (constructors,nparams) = 
     (mconstructor_types ity,mind_nparams ity)
in if vect_length constructors <> 1 then false
else
  let constructor_type = constructors.(0) in
  let globargs,_ = chop_list nparams largs in
  let glob_Cty = hnf_prod_applist sigma "is_singleton" constructor_type globargs in
  match snd(infexecute_rec sigma (FENV sigma env) glob_Cty) with
    Logic -> false
  | Inf {VAL=inf_Cty;_} -> conv_x sigma (without_nparams nparams inf_Cty) typ
;;****)

let is_singleton sigma env typ (ity,largs) =
    match mis_singl (mind_specif_of_mind ity) with 
    NONE -> false 
  | SOME _ -> let (_,inf) = infexecute_rec sigma (FENV sigma env) 
                  (applist(ity,largs)) in 
              match inf with 
        Inf {VAL=inf_Cty;_} -> conv_x sigma inf_Cty typ
      | _ -> anomaly "program__is_singleton";;

(* Checks if a  mutual inductive type is a tuple *)
let is_tuple ity =
  let mis = mind_specif_of_mind ity in
    mis.mib.MINDNTYPES =1 & vect_length (mis_consnames mis) = 1
;;

(* Weak-head beta-delta-iota reduction from a clenv *)
let clenv_whd_betadeltaiota_stack clenv c stack =
   whd_betadeltaiota_stack (w_Underlying clenv.hook) c stack ;;

let clenv_whd_betadeltaiota clenv c =
   whd_betadeltaiota (w_Underlying clenv.hook) c ;;

(* signature of a clenv *)
let clenv_sign clenv = w_hyps clenv.hook ;;
(* ex. var. of a clenv *)
let clenv_evc clenv = w_Underlying clenv.hook ;;

(* Changes the head of a clenv with (templ,templty) *)
let clenv_change_head (templ,templty) clenv =
  {templval = mk_freelisted templ;
  templtyp = mk_freelisted templty;
  namenv = clenv.namenv;
  env = clenv.env;
  hook = clenv.hook}
;;

let clenv_dunbind dflt T B clenv =
let (sign',B') = dunbind dflt (clenv_sign clenv) T B in
let clenv' = clenv_add_sign (hd_sign sign') clenv in
  (B',clenv')
;;

(* Safe item function *)
let safe_item l n =
  try item l n
  with Failure _ -> invalid_arg "safe_item"
;;

(* Applies the refine tactic and connects the corresponding clenv 
   by applying the continuation KONT. Applies the tactic mvtac to 
   the list of generated subgoals.
   Entries : KONT : continuation
             clenv : clenv to be connected
             mvtac : tactic to applies after the resolution
             gls : current goal *)
let res_THEN KONT clenv mvtac gls =
let clenv' = clenv_unify (clenv_instance_template_type clenv) (pf_concl gls) clenv in
let mvs = collect_metas (clenv_instance_template clenv')
in (THEN_i (clenv_refine KONT clenv')
           (fun i -> mvtac (safe_item mvs i))
           0) gls
;;

(* Applies the precedent res_THEN function with a tactic thingtac 
   taking arguments in the assocmap. In practice, thingtac is Realizer 
   and assocmap the association list between metavariables and programs.
   Entries : KONT : continuation
             clenv : clenv to be connected
             assocmap : association list between meta variables and terms
             thingtac : tactic to be applied
             gls : current goal *)
let assoc_THEN KONT clenv assocmap thingtac gls =
  (res_THEN KONT clenv
  (fun mv -> let arg = try SOME(listmap__map assocmap mv)
                       with Not_found -> NONE
             in thingtac arg))
  gls
;;

(* Lists of function to do a resolution as before but for the second order.
   The final used function is assoc_THEN2 with the same principle as 
   assoc_THEN *)
let collect_metas_with_safe_item mvtac =
(fun clenv i -> 
let mvs = collect_metas (clenv_instance_template clenv) in
 mvtac (safe_item mvs i)) ;;

let FlexHeadSubstitution gl p oplist clause =
    let (clause',cllist) = constrain_clenv_using_subterm_list false clause oplist (pf_concl gl) in
    let typp = clenv_instance_type clause' p
    in clenv_unify (DOP0(Meta p)) (abstract_list_all gl typp (pf_concl gl) cllist) clause'
;;

let abstract_then_resolve_THEN KONT clause mvtac gl =
  match decomp_app (clenv_instance_template_type clause) with
   (DOP0(Meta(p)),oplist) ->
    let clause' = FlexHeadSubstitution gl p oplist clause
    in res_pf_THEN_i KONT clause' mvtac 0 gl
  | _ -> error "abstract_then_resolve"
;;

let refinew_scheme_THEN KONT clause mvtac gl =
  let mvtac' = collect_metas_with_safe_item mvtac in
  ((res_pf_THEN_i KONT clause mvtac' 0)
    ORELSE (abstract_then_resolve_THEN KONT clause mvtac')) gl
;;

let assoc_THEN2 KONT clenv assocmap thingtac gls =
  (refinew_scheme_THEN KONT clenv
  (fun mv -> let arg = try SOME(listmap__map assocmap mv)
                       with Not_found -> NONE
             in thingtac arg))
  gls
;;

(* Counts the number of informative products in a term depending on a 
   list of arguments. 
   Entries : args : list of arguments
             gls : current goal
             term
   Output : number of counted products *)             
let nb_intros_liees_list args gls  =
  let rec nrec ni = function
    DOP2(Prod,A,DLAM(_,B)) -> if (pf_informative_type gls A)
                                 or (exists (fun t -> dependent (strip_outer_cast t) A) args)
                                 or (exists (fun t -> not(closed t)) args)
                              then ni
                              else nrec (ni+1) B
  | DOP2(Cast,c,_) -> nrec ni c
  | _ -> ni
in nrec 0 ;;                     

(* The same for dependence w.r.t. only one term *)
let nb_intros_liees c = nb_intros_liees_list [c] ;;

(* Counts the number of informative products in a term.
   Entries : gls : current goal
             term
   Output : number of counted products *)  
let nb_intros gls  =
  let rec nrec ni = function
    DOP2(Prod,A,DLAM(_,B)) -> if not (pf_informative_type gls A)
                              then nrec (ni+1) B
                              else ni
  | DOP2(Cast,c,_) -> nrec ni c
  | _ -> ni
in nrec 0 ;;                     

(* =============================== *)
(* Associating a program to a goal *)
(* =============================== *)

let Realizer pg gls =
let sign = (pf_hyps gls) in
let metamap = !METAMAP in
let ty = type_meta_of_prog (Project gls) (FSIGN (Project gls) (pf_hyps gls)) metamap pg in
let infconcl = (match snd(pf_infexecute gls (pf_concl gls)) with
                Logic -> error "Cannot attach a realizer to a logical goal"
              | Inf conclj -> conclj.VAL) in
if pf_conv1_x gls infconcl ty then
    let octxt = pf_ctxt gls in
    let nctxt = set_pgm (SOME pg) octxt in
        Context nctxt gls
else error "Type of program and informative extraction of goal do not coincide"
;;

let Realizer_top pg gls =
let pg' = match (strip_outer_cast pg) with
            DOPN(Const _,_) -> (try strong whd_beta (pf_one_step_reduce gls pg) 
                                with UserError _ -> pg)
          | _ -> pg in
let ty = type_of_prog (Project gls) (FSIGN (Project gls) (pf_hyps gls)) pg' in
let infconcl = (match snd(pf_infexecute gls (pf_concl gls)) with
                Logic -> error "Cannot attach a realizer to a logical goal"
              | Inf conclj -> conclj.VAL) in
if pf_conv1_x gls infconcl ty then
    let octxt = pf_ctxt gls in
    let nctxt = set_pgm (SOME pg') octxt in
        clear_METAMAP; Context nctxt gls
else error "Type of program and informative extraction of goal do not coincide"
;;

let Realizer_tac = overwriting_register_comarg_tactic "Realizer"
(fun c gl -> Realizer_top (raw_prog_of_progcom (Project gl) (pf_hyps gl) c) gl)
;;

(* ============================= *)
(* Finding a program from a goal *)
(* ============================= *)

let extract_pg gls = 
  let ctxt = pf_ctxt gls in
  let exist_pg = get_pgm ctxt in
  match exist_pg with
    SOME pg -> pg
  | NONE -> error"no program associated to this subgoal" ;;

(* ================ *)
(* Abstraction case *)
(* ================ *)

(* makes a lambda-abstraction [xbar:Tbar]body from
  the arguments xbar, Tbar, body, assuming that the
  variables are bound by name (VAR) NOT by Rel.
 *)
let lambda_list ids types body =
  list_it2 (fun id T body -> lambda id T body) ids types body
;;

(* Precondition:
   Goal: Gamma [x:A] |- Phi
   Program: (y:progA)P[y]

   Want the assumption x to be pure, otherwise want to
   eliminate it and find the new program to associate to the subgoal.
 *)

let FinishIntroOnProgram id pg gl =
let A = pf_type_of gl (VAR id) in
let hd,largs = decomp_app A in
  (match pg with
    DOP2(Lambda,progA,B) ->
      (match strip_outer_cast hd with
      (DOPN(MutInd _,_) as hdm) ->
        let sign = pf_hyps gl and sigma = Project gl in
        let A' = try propify (execute_pure sigma sign progA).VAL with 
                    UserError _ -> progA in
        if not(pf_conv1_x gl A' A) & is_singleton sigma (GLOB sign) progA (hdm,largs) then
          (case_analysis [] (VAR id) THEN Realizer pg) gl
        else if is_tuple hd then
          try (simple_elimination_then
       (IntroCaseAssumsThen
        (fun ba gls ->
              let cargs = list__map mkVAR ba.cargs in
              let (DOPN(MutInd(sp,i),cl)) = ba.ba.ity in
              let chd = DOPN(MutConstruct((sp,i),ba.ba.branchnum),cl) in
              let y' = applist(chd,cargs) in
              let subst_pg = (SAPP B y') in
              let carg_types = list__map (pf_type_of gls) cargs in
              let lam_pg = lambda_list ba.cargs carg_types subst_pg in
              let clsl = list__map inSOME ba.assums in
                (BringHyps clsl THEN ThinClauses clsl THEN
         Realizer lam_pg) gls))
       (VAR id) gl)
          with UserError _ -> Realizer (SAPP B (VAR id)) gl
        else Realizer (SAPP B (VAR id)) gl
    | _ -> Realizer (SAPP B (VAR id)) gl)
 | _ -> error "FinishIntroOnProgram")
;;

(* Precondition : the program is a lambda-abstraction.
   If it is a logical one, introduces as many hypotheses as needed 
   until arriving to the first compatible one.
   If it is an informative, introduces as many hypotheses as needed
   until arriving to the first informative compatible one *)
let rec deroule_intro pg gls = 
  let concl = pf_concl gls in
  match (pg,strip_outer_cast concl) with
    (DOP2(Lambda,DOP1(XTRA("COMMENT",[]),A),(DLAM(Name(ide),_) as B)),
     DOP2(Prod,A',_)) ->
      if (pf_informative_type gls A') then
        error "invalid form of goal"
      else
        if pf_conv_x gls A A' then
          (intro_using ide THEN OnLastHyp (fun (SOME id) -> Realizer (SAPP B (VAR id)))) gls
        else
          (try (intro THEN deroule_intro pg) gls
           with UserError _ -> ((cut_tac A) THENL (auto__trivial) THEN Realizer B) gls)
  | (DOP2(Lambda,A,(DLAM(Name(ide),_) as B)),DOP2(Prod,A',_)) ->
      if pf_informative_type gls A' then
        (intro_using ide THEN OnLastHyp (fun (SOME id) -> FinishIntroOnProgram id pg)) gls
      else
        (intro THEN deroule_intro pg) gls
  | _ -> (red THEN deroule_intro pg) gls ;;
   
(* ========================================= *)
(* Retrieving the specification of a program *)
(* ========================================= *)

(* Finds the inductive type at the end of a conclusion 
   Entries : clenv : clenv in which type checking can be executed
             term of the conclusion
   Ouput : Inductive type if success *)
let rec clenv_find_ind clenv = function
    (DOP2(Prod,c1,DLAM(_,c2))) ->  clenv_find_ind clenv c2
  | (DOP2(Cast,c,_))           ->  clenv_find_ind clenv c
  | (DOPN(Const _,_) as val)   ->  
      let newval = one_step_reduce (clenv_evc clenv) val in
      if eq_constr(val,newval) then error"find_ind"
      else clenv_find_ind clenv newval
  | (DOPN(AppL,cl) as val)     ->  
      (match clenv_whd_betadeltaiota_stack clenv val [] with
        (DOPN(MutInd _,_) as ind),args -> (ind,fst(chop_list(mind_nparams ind) args))    
      | (DOP2(Prod,c1,DLAM(_,c2))),_ -> clenv_find_ind clenv c2              
      | _ -> error"find_ind")
  | (DOPN(MutInd _,_) as ind) -> find_minductype (clenv_evc clenv) ind
  | _ -> error"find_ind" ;;

(* Tries to match the head product of hdty with argty.
   Creates a proof term in three parts arg, list and clenv' 
   by adding metas until the good logical argument arg is found.
   Entries : clenv : clenv in which proof type checking can be executed 
             head : term in the CoC
             hdty : type of this term
             arg : corresponding program term
             argty : type of this program term
   Outputs : arg : proof term correspongind to arg
            list : list of metavariables introduced as logical arguments,
                   the head is head (with eventually new metavariables)
            clenv' : new clenv containing introduced metavariables *) 
let clausal_form_logic clenv head hdty arg argty = clrec [head] clenv hdty
where rec clrec argstack clenv = function
  DOP2(Cast,c,_) -> clrec argstack clenv c
| DOP2(Prod,c1,(DLAM(na,_) as c2)) -> 
    (try let clenv' = clenv_unify c1 argty clenv in
          arg,list__map (clenv_instance_term clenv') (rev argstack),clenv'
     with UserError _ -> 
        if not(clenv_is_info_type clenv c1 (clenv_type_of clenv c1)) then 
          let mv = newMETA() in
          let clenv' = clenv_pose (na,mv,c1) clenv in
          let (head,args,clenv'') = clrec (DOP0(Meta(mv))::argstack) clenv' 
                                          (SAPP c2 (DOP0(Meta(mv)))) in
            (head,(DOP0(Meta(mv)))::args,clenv'')
        else error"bad logical argument")
| _ -> error"bad logical argument" ;;

(* Tests if a type is the one of a program or not *)
let rec is_type_of_program = function
  Inf{VAL=DOP0(Sort(Prop _));_} -> true
| _ -> false ;;

(* Tests if a type is an arity or not *)
let rec is_arity clenv P = match clenv_whd_betadeltaiota clenv P with
  DOP0(Sort(Prop _)) -> true
| DOP2(Prod,_,DLAM(_,B)) -> is_arity clenv B
| DOP2(Cast,c,_) -> is_arity clenv c
| _ -> false ;;

(* Tries to unify arg1 and arg2 in clenv.
   Reduces the terms if necessary
   Entries : arg1, arg2, clenv
   Ouput : new clenv with metavariables instanciated *)
let rec unify_red arg1 arg2 clenv = 
let sigma = clenv_evc clenv in
  try clenv_typed_unify arg1 arg2 clenv
  with UserError(s1,_) -> 
   try let arg1' = (strong whd_beta (one_step_reduce sigma arg1)) in
        if not(eq_constr(arg1',arg1)) then
                 unify_red arg2 arg1' clenv
        else error s1
   with UserError(s2,_) -> 
    let arg2' = (strong whd_beta (one_step_reduce sigma arg2)) in
      if not(eq_constr(arg2',arg2)) then
       unify_red arg1 arg2' clenv 
      else error s2;;
                
(* Creates a list of metas corresponding to the logical products in a type.
   Returns the head of the type, the list of metavariables and the new clenv.
   Entries : clenv : clenv in which proof type checking can be executed 
             typ : type in the CoC
   Outputs : concl : conclusion of the type including new metavariables
            list : list of metavariables introduced as logical arguments
            clenv' : new clenv containing introduced metavariables *) 
let rec clausal_form_info clenv = function
  DOP2(Cast,c,_) -> clausal_form_info clenv c
| (DOP2(Prod,c1,(DLAM(na,_) as c2)) as head) -> 
    let T = clenv_type_of clenv c1 
    in let (_,inf) = infexecute (clenv_evc clenv) 
                             (FSIGN (clenv_evc clenv) (clenv_sign clenv)) T
    in if is_type_of_program inf then (head,[],clenv)
       else let mv = newMETA() in
            let clenv' = clenv_pose (na,mv,c1) clenv in
            if not(clenv_is_info_type clenv c1 T) then
              let (head',args,clenv') = clausal_form_info clenv' 
                                                  (SAPP c2 (DOP0(Meta(mv)))) in
                (head',(DOP0(Meta(mv)))::args,clenv')
            else ((SAPP c2 (DOP0(Meta(mv)))),[DOP0(Meta(mv))],clenv')
| head -> (head,[],clenv) ;;


(* Creates a list of metas corresponding to the logical products in the type
   argty. Tries then to match the current conclusion with the head of argty
   containing metavariables.
   Then instanciates the current proof term head and the current argument.
   Returns the new head, the new list of arguments and a new clenv.
   Entries : clenv : clenv in which proof type checking can be executed 
             head : list containing the current head and eventually its 
                    logical arguments
             concl : conclusion
             arg : proof term in CoC
             argty : type of this proof term
   Outputs : head' : new head of the constructed proof term
            arg' : list of proof terms arguments to be applied to head'
            clenv' : new clenv containing introduced metavariables *) 
let try_clausal_form_match clenv head concl arg argty =
 clrec clenv arg argty where rec clrec clenv arg argty =
  try let clenv' = unify_red concl argty clenv in
 (list__map (clenv_instance_term clenv') head),(clenv_instance_term clenv' arg),clenv'
  with UserError _ -> 
    (match strip_outer_cast argty with
      DOP2(Prod,c1,(DLAM(na,_) as c2)) -> 
        if not(clenv_is_info_type clenv c1 (clenv_type_of clenv c1)) then
          let mv = newMETA() in
          let clenv' = clenv_pose (na,mv,c1) clenv in
            clrec clenv' (DOPN(AppL,vect_of_list(arg::[DOP0(Meta(mv))]))) 
                  (SAPP c2 (DOP0(Meta(mv))))
        else error"try_clausal_form_match"
    | _ -> error"try_clausal_form_match") ;;

(* Forgets arities in head in a type and replaces them by metavariables.
   Entries : clenv : clenv in which type checking is possible 
             headfw : head of program term corresponding to the current type
             argstack : list of arguments for this program term
             typ : current type
   Outputs : concl : new type with eventually metavariables to replace arities
            argstack' : new list with added metavariables at the end
            clenv' : new clenv *)
let rec without_arity_in_head clenv headfw argstack = function
 (DOP2(Prod,c1,(DLAM(na,_) as c2)) as prod) -> 
 if is_arity clenv c1 then
  (match strip_outer_cast headfw with
   DOPN(AppL,cl) ->
     let _::[arg] = list_of_vect cl in
     let mv = newMETA() in
     let clenv' = clenv_pose (Anonymous,mv,c1) clenv in
            ((SAPP c2 (DOP0(Meta(mv)))),argstack@[DOP0(Meta(mv))],clenv')
 | _ -> (prod,argstack,clenv))
 else (prod,argstack,clenv)
| DOP2(Cast,c,_) -> without_arity_in_head clenv headfw argstack c
| cl -> (cl,argstack,clenv) ;;

(* Find the index of the first false place in a list of boolean.
   Entry : list of booleans
   Output : number *)
let find_info_prod = frec 1 where rec frec n = function
  true::l -> frec (n+1) l
| false::_ -> n
| [] -> error"find_info_prod";;

(* Checks if a type is a dependent type or not *)
let rec is_dependent_type = function
  DOP2(Cast,c,_) -> is_dependent_type c 
| DOP2(Prod,_,DLAM(_,B)) -> if noccurn 1 B then 
                              is_dependent_type B
                            else (true,B)
| x -> false,x ;;

(* For a singleton type, construction of the corresponding lambda-expression
   to appear as an argument of a match in an elimination
   on this singleton type.
   Entries : lc : type of the constructor of the singleton type
             ind : singleton type
             nparams : number of parameters
             head : term to be placed at the end of the lambda-expression.
                    It is a list; if the list is empty, this is to say
                    that there is no term to place at the end.
   Output : lambda-expression *)
let construct_arg_sig lc ind nparams head gls = 
crec (1+nparams) [] lc where rec crec n l a = 
match a with
  DOP2(Cast,c,_) -> crec n l c
| DOP2(Prod,A,DLAM(na,B))  -> 
    if noccurn n A then
      DOP2(Lambda,A,DLAM(named_hd A na,crec (n+1) ((is_logical l A gls)::l) B))
    else 
      DOP2(Lambda,(subst1 ind A),DLAM(named_hd A na,DOP2(Lambda,subst1 ind a,DLAM(named_hd (subst1 ind a) Anonymous,crec (n+1) ((is_logical l (subst1 ind a) gls)::((is_logical l (subst1 ind A) gls)::l)) B))))
| _ -> if head = [] then
        Rel(find_info_prod l)
       else applist(hd head,[Rel(find_info_prod l)]) ;;

(* From a list of integer i, creates a list of Rel(n-i).
   Entries : n : integer
             l : list of integers
   Ouput : list of Rel(n-i) *)
let rec make_rel_list n = function
  a::l -> Rel(n-a)::(make_rel_list n l)
| [] -> [] ;;

(* Same as construct_arg_sig but for a tuple and not a singleton.
   Entries : lc : type of the constructor of the singleton type
             ind : singleton type
             nparams : number of parameters
             head : term to be placed at the end of the lambda-expression.
                    It is a list; if the list is empty, this is to say
                    that there is no term to place at the end.
             concl : current conclusion
   Output : lambda-expression *)

let construct_arg_ind lc ind nparams head concl gls = 
let (headconcl,_) = decomp_app concl in
try 
let ((DOPN(MutInd(sp,i),cl) as ind'),largs) = 
              find_minductype mt_evc headconcl in
let mis = mind_specif_of_mind ind' in
let (_,lc_of_cl) = decomp_all_DLAMV_name (mis_lc mis) in
drec (1+nparams) [] lc (hd_vect lc_of_cl) 
where rec drec n l lc lc_of_cl =
let nparams = mis_nparams mis in 
let (globargs,_) = chop_list nparams largs in
  (match (lc,lc_of_cl) with
    (DOP2(Cast,c,_),_) -> drec n l c lc_of_cl
  | (_,DOP2(Cast,c,_)) -> drec n l lc c
  | (DOP2(Prod,B,DLAM(na,C)),(DOP2(Prod,B',DLAM(_,C')))) -> 
       if noccurn n B then
         if pf_conv_x gls B B' then
            DOP2(Lambda,B,DLAM(na,drec (n+1) (n::l) C C'))
         else 
            DOP2(Lambda,B,DLAM(na,drec (n+1) l C lc_of_cl))
       else 
         if pf_conv_x gls B B' then
            DOP2(Lambda,subst1 ind B,DLAM(na,DOP2(Lambda,subst1 ind lc,DLAM(named_hd (subst1 ind lc) Anonymous,drec (n+1) (n::l) C C'))))
         else
            DOP2(Lambda,subst1 ind B,DLAM(na,DOP2(Lambda,subst1 ind lc,DLAM(named_hd (subst1 ind lc) Anonymous,drec (n+1) l C lc_of_cl))))
  | _ -> if head = [] then
                applist(DOPN(MutConstruct((sp,i),1),cl),globargs@(rev (make_rel_list n l)))
         else 
              applist(DOPN(MutConstruct((sp,i),1),cl),globargs@[applist(hd head,(rev (make_rel_list n l)))]))
with Induc -> error "construct_arg_ind" ;;

(* Combination of the previous functions to eliminate an argument
   those type is a singleton or a tuple when needed.
   Entries : concl : current conclusion
             ccihead : proof term to be eliminated
             cciheadty : its type
             cciarg : program term to be eliminated
             cciargty : its type
             gls : current goal
   Output : new proof term transformed in an elimination *)
let elim_sig_arg concl ccihead cciheadty cciarg cciargty gls = 
  let sigma = Project gls and sign = pf_hyps gls in
  let (head,args) = let (head,args) = decomp_app cciargty in
                        (match head with
                           DOPN(Const _,_) -> decomp_app(strong whd_beta (pf_one_step_reduce gls cciargty))
                     | _ -> (head,args)) in
  let argty = fst(whd_betadeltaiota_stack sigma cciargty []) in
  try 
    let (T,largs,lc,nparams) =
           let (T,largs) = find_minductype sigma cciargty in
           let mis = mind_specif_of_mind T in
           let (_,lc) = decomp_all_DLAMV_name(mis_lc mis)
           in (T,largs,lc,mis_nparams mis)
in  if is_singleton sigma (GLOB sign) (pf_extraction gls concl) (T,largs) then
    let lf = whd_beta (applist(construct_arg_sig (hd_vect lc) T nparams [ccihead] gls,largs)) in
    let (b,P) = is_dependent_type cciheadty in
    let typ = if b then 
             let lf' = whd_beta (applist(construct_arg_sig (hd_vect lc) T nparams [] gls,largs)) in
             let proj = DOPN(MutCase,vect_of_list((concl::[Rel(1)])@[lf']))
               in 
                DOP2(Lambda,cciargty,DLAM(named_hd cciargty Anonymous,
                                          (subst1 proj P)))
              else (match cciheadty with
                      DOP2(Prod,_,DLAM(_,B)) -> B
                    | _ -> error"elim_sig_arg") in
         DOPN(MutCase,vect_of_list((typ::[cciarg])@[lf]))
  
  else if is_tuple head then
        let lf = whd_beta (applist(construct_arg_ind (hd_vect lc) T nparams [ccihead] concl gls,largs)) in
        let (b,P) = is_dependent_type cciheadty in
        let typ = if b then 
              let lf' = whd_beta (applist(construct_arg_ind (hd_vect lc) T nparams [] concl gls,largs)) in
              let proj = 
                    DOPN(MutCase,vect_of_list((concl::[Rel(1)])@[lf'])) in
           DOP2(Lambda,cciargty,DLAM(named_hd cciargty Anonymous,(subst1 proj P)))
                  else (match cciheadty with
                          DOP2(Prod,_,DLAM(_,B)) -> B
                        | _ -> error"elim_sig_arg") in
             DOPN(MutCase,vect_of_list((typ::[cciarg])@[lf]))
       else error"elim_sig_arg" 
  with Induc | Not_found | Invalid_argument _ -> error"elim_sig_arg";;

(* Same as construct_arg_sig.
   But this function eliminates the head of an application and no more 
   an argument.
   Entries : lc : type of the constructor of the singleton type
             ind : singleton type
             nparams : number of parameters
             arg : term to be placed at the end of the lambda-expression,
                    It is a list; if the list is empty, this is to say
                    that there is no term to place at the end.
   Output : lambda-expression *)
let construct_head_sig lc ind nparams arg gls = 
drec (1+nparams) [] lc where rec drec n l a =
match a with
  DOP2(Cast,c,_) -> drec n l c
| DOP2(Prod,A,DLAM(na,B)) -> 
      if noccurn n A then
        DOP2(Lambda,A,DLAM(named_hd A na,drec (n+1) ((is_logical l A gls)::l) B))
      else 
        DOP2(Lambda,subst1 ind A,DLAM(named_hd (subst1 ind A) na,DOP2(Lambda,subst1 ind a,DLAM(named_hd (subst1 ind a) Anonymous,drec (n+1) ((is_logical l (subst1 ind a) gls)::((is_logical l (subst1 ind A) gls)::l)) B))))
| _ -> if arg = [] then
        Rel(find_info_prod l)
       else applist(Rel(find_info_prod l),arg) ;;

(* Gives the type of the tail of a lambda-expression.
   Entries : sigma : metavariables environment
             sign : current signature
             term : lambda-expression
   Output : typ : type of the tail *)
let rec type_tail_lf sigma sign = function
  DOP2(Cast,c,_) -> type_tail_lf sigma sign c 
| DOP2(Lambda,A,(DLAM(Name(x),B) as C)) -> type_tail_lf sigma (add_sign (x,A) sign) (SAPP C (VAR x))
| DOP2(Lambda,A,DLAM(_,B)) -> type_tail_lf sigma sign B
| x -> type_of_rel sigma (GLOB sign) x ;;

(* Combination of the two previous functions to eliminate an head of an
   application if needed.
   Entries : ccival : proof term to be transformed (head of an application)
             ccityp : its type
             arg : argument applied to the head
             gls : current goal
   Output : new proof term with an elimination *)
let elim_sig_head ccival ccityp arg gls = 
  let sigma = Project gls and sign = pf_hyps gls in
  try 
  let (head,largs) = let (head,args) = decomp_app ccityp in
                      (match head with
                         DOPN(Const _,_) -> decomp_app(strong whd_beta (pf_one_step_reduce gls ccityp))
                       | _ -> (head,args)) in
  let tyarg = pf_extraction gls ccityp in
  let (T,largs) = find_minductype sigma ccityp in
  if is_singleton sigma (GLOB sign) tyarg (T,largs) then
    let (T,largs,lc,nparams) = 
        let mis = mind_specif_of_mind T in
        let (_,lc) = decomp_all_DLAMV_name(mis_lc mis) in 
          (T,largs,lc,mis_nparams mis) in
    let lf = propify(whd_beta (applist(construct_head_sig (hd_vect lc) T nparams [arg] gls,largs))) in
    let typP = type_tail_lf sigma sign lf in
      DOPN(MutCase,vect_of_list((typP::[ccival])@[lf]))
  else error"elim_sig_head"
 with Induc | Not_found | Invalid_argument _ -> error"elim_sig_head" ;;

let elim_sig ccival ccityp gls = 
  let sigma = Project gls and sign = pf_hyps gls in
  try 
  let (head,largs) = let (head,args) = decomp_app ccityp in
                      (match  head with
                         DOPN(Const _,_) -> decomp_app(strong whd_beta (pf_one_step_reduce gls ccityp))
                       | _ -> (head,args)) in
  let tyarg = pf_extraction gls ccityp in
   let (T,largs) = find_minductype sigma ccityp in
    if is_singleton sigma (GLOB sign) tyarg (T,largs) then
        let mis = mind_specif_of_mind T in 
        let (_,lc) = decomp_all_DLAMV_name(mis_lc mis) in 
        let nparams = mis_nparams mis in
    let lf = propify(whd_beta (applist(construct_head_sig (hd_vect lc) T nparams [] gls,largs))) in
    let typP = type_tail_lf sigma sign lf in
      DOPN(MutCase,vect_of_list((typP::[ccival])@[lf]))
  else error"elim_sig"
 with Induc | Not_found | Invalid_argument _ -> error"elim_sig" ;;

(* Basic function to find the specification of a program.
   Entries : gls : current goal
             clenv : current clausal environment
             pmap : association list between metavariables and programs
             cl : current conclusion
             pg : current program
             PG : program associated to the goal  (not necessarily the same)
   Output : proof term with metavariables correspondint to pg *)
let specif_of_annot_part gls (clenv,pmap) cl pg PG = 
let rec specrec k (clenv,pmap) cl pg  = 
  match telescope_appl (strip_outer_cast pg) with
      (VAR id as val) -> (val,snd(lookup_sign id (clenv_sign clenv)),
                          (clenv,pmap))
    | (DOPN(Const sp,_) as val) -> 
        let ccival=global_reference (GLOB (clenv_sign clenv)) (ccisp_of sp) (basename sp) in
        (ccival,clenv_type_of clenv ccival,(clenv,pmap))
    | (DOP0(Meta p) as val) -> (val, cl, (clenv,pmap))
    | (DOPN(MutConstruct((sp,j),i),largs) as val) -> 
      (match cl with
        DOP0(Sort(Prop Pos)) ->
                let j = execute_pure (clenv_evc clenv) (clenv_sign clenv) val in
                  (j.VAL,j.TYP,(clenv,pmap))
      | _ ->
        if occur_meta cl then
          let ccival = DOPN(MutConstruct((ccisp_of_fwsp sp,j),i),largs) in
            (ccival,clenv_type_of clenv ccival,(clenv,pmap))
        else
        let (DOPN(MutInd(sp',num),cl') as ind,globargs) = clenv_find_ind clenv cl in
        let base = if k=0 then val else PG in
        if (is_singleton (Project gls) (GLOB (pf_hyps gls)) (type_meta_of_prog (Project gls) (FSIGN (Project gls) (pf_hyps gls)) !METAMAP base) (ind,globargs)) then
          let nparams = mind_nparams(DOPN(MutInd(sp,j),largs)) in
          if (nparams = 0) or (k>1+nparams) then
          let ccival = DOPN(MutConstruct((ccisp_of_fwsp sp,j),i),largs) in
            (ccival,clenv_type_of clenv ccival,(clenv,pmap))
          else
          let ccival=applist(DOPN(MutConstruct((sp',num),1),cl'),globargs) in 
          let ccityp = clenv_type_of clenv ccival in 
          let (clenv',pmap',tail_typ,args) = 
              one_arg (clenv,pmap,ccityp) k val in
          let (_,args',clenv'') = clausal_form_info clenv' tail_typ in
          let ccival' = applist(DOPN(MutConstruct((sp',num),1),cl'),
                                   globargs@args@args') in
            (ccival',clenv_type_of clenv'' ccival',(clenv'',pmap'))
        else
            let val=applist(DOPN(MutConstruct((sp',num),i),cl'),globargs) in 
            let T = clenv_type_of clenv val in
            let (_,args,clenv') = clausal_form_info clenv T in
            let ccival = applist(val,args) in
              (ccival,clenv_type_of clenv' ccival,(clenv',pmap)))
    | (DOP2(XTRA("ANNOT",[]),c,T)) -> 
        let mv = newMETA() in
          let clenv' = clenv_pose (Anonymous,mv,T) clenv in
            (DOP0(Meta(mv)),T,(clenv',listmap__add pmap (mv,c)))
    | (DOP2(Lambda,c1,(DLAM _ as B))) -> 
      let j=(try execute_pure (clenv_evc clenv) (clenv_sign clenv) c1 
             with UserError _ -> raise NotProgram) in
      let c1assum = j_val j in
      let (B',clenv') = clenv_dunbind (id_of_string "H") c1assum B clenv in
      let (ccival,ccityp,(clenv'',pmap')) = specrec k (clenv',pmap) cl B'
      in let ccival' = clenv_instance_term clenv'' ccival and
             ccityp' = clenv_instance_term clenv'' ccityp
      in if occur_meta ccival' then
           raise NotProgram
         else let (ty,B') = dbind (clenv_sign clenv'') ccival' in
              let (_,tB') = dbind (clenv_sign clenv'') ccityp' in
                (DOP2(Lambda,ty,B'),DOP2(Prod,ty,tB'),(clenv'',pmap'))
    | (DOPN(AppL,args)) ->
      let t1::[t2] = list_of_vect args in
      let (ccivalt1,ccitypt1,(clenv',pmap')) = specrec (k+1) (clenv,pmap) cl t1
      in let sigma = clenv_evc clenv' 
      in let j = execute_meta_prog sigma (FSIGN sigma (clenv_sign clenv')) !METAMAP t2 in
      if not(is_info_judge sigma j) then
        let (arg,t1'::mvlist,clenv'') = 
                    clausal_form_logic clenv' ccivalt1 ccitypt1 j.VAL j.TYP in
        let ccival = DOPN(AppL,vect_of_list((t1'::mvlist)@[without_comment arg])) in
          (ccival,clenv_type_of clenv ccival,(clenv'',pmap'))
      else
        let (concl,args,clenv'') = clausal_form_info clenv'  
                                               (strong whd_beta ccitypt1) in
        if is_arity clenv j.TYP then
          let ccival = DOPN(AppL,vect_of_list(ccivalt1::args)) in
            (ccival,clenv_type_of clenv'' ccival,(clenv'',pmap'))
        else
          (try match_with_cl (clenv'',pmap') ccivalt1 t1 concl args t2 k
           with Expand -> 
            (try match_with_cl (clenv'',pmap') ccivalt1 t1
                   (strong whd_beta (one_step_reduce (clenv_evc clenv'') concl))
                   args t2 k
             with UserError _ -> raise (Generalize k)))
    | _ -> raise NotProgram

and one_arg = fun  (clenv,pmap,t) k c -> 
   let (concl,args,clenv') = clausal_form_info clenv t in
   let sigma = clenv_evc clenv' in
   if is_arity clenv (execute_meta_prog sigma (FSIGN sigma (clenv_sign clenv)) !METAMAP c).TYP
   then (clenv',pmap,concl,args) 
   else (match concl with
          DOP2(Prod,c1,DLAM(na,c2)) ->
           try 
            let (ccival',ccityp',(clenv'',pmap')) = specrec k (clenv',pmap) c1 c
            in let (c2'::L',c',clenv''') = try_clausal_form_match 
                                           clenv'' (c2::args) c1 ccival' ccityp'
            in (clenv''',pmap',(subst1 c' c2'),L'@[c'])
           with NotProgram | UserError _ ->
                  let mv = newMETA() in
                  let clenv'' = clenv_pose (na,mv,c1) clenv' in
                  let pmap' = listmap__add pmap (mv,c) in
                     (clenv'',pmap',concl,args@[DOP0(Meta(mv))])
        | _ -> anomaly"specif") 

and match_with_cl (clenv,pmap) head headfw concl argstack arg k =
let (concl',argstack',clenv1) = without_arity_in_head clenv headfw argstack 
                                                      concl
in match concl' with
  DOP2(Prod,c1,DLAM(na,c2)) ->
  (try let (ccival,ccityp,(clenv2,pmap')) = specrec 0 (clenv1,pmap) c1 arg
       in let (_,argstack'',clenv3) = clausal_form_info clenv2 ccityp
       in let ccival'= applist(ccival,argstack'')
       in let ccityp' = clenv_type_of clenv3 ccival'
       in (try let (head',arg',clenv4) = try_clausal_form_match clenv3 
                                   [applist(head,argstack')] c1 ccival' ccityp'
               in let ccival = DOPN(AppL,vect_of_list(head'@[arg']))
               in (ccival,clenv_type_of clenv4 ccival,(clenv4,pmap'))
           with UserError _ -> 
                let val = applist(head,argstack') in 
                let newpg = elim_sig_arg c1 val (clenv_type_of clenv1 val)
                                         ccival' ccityp' gls in 
                  (newpg,clenv_type_of clenv3 newpg,(clenv3,pmap')))
   with (NotProgram | UserError _ | Failure _) -> 
       let mv = newMETA() in
       let clenv2 = clenv_pose (na,mv,c1) clenv1 in
       let pmap' = listmap__add pmap (mv,arg) in
       let ccival = applist(head,argstack'@[DOP0(Meta(mv))]) in
        (ccival,clenv_type_of clenv2 ccival,(clenv2,pmap')))
| _ -> try let newpg = elim_sig_head head (clenv_type_of clenv head) arg gls in 
              (newpg,clenv_type_of clenv newpg,(clenv,pmap))
       with UserError _ -> raise Expand

in specrec 1 (clenv,pmap) cl pg;;

(* Same as before when pg and PG are the same *)
let specif_of_annot gls (clenv,pmap) cl pg = specif_of_annot_part gls (clenv,pmap) cl pg pg ;;

(* ================ *)
(* Application Case *)
(* ================ *)

(* Checks if the head of a type is bound or not and gives this head.
   Entries : gls : current goal
             term : proof type
   Output : (b,t) : (boolean,head of the type) *)             
let rec find_bound gls = function
  DOP2(Prod,_,DLAM(_,B)) -> find_bound gls B
| DOP2(Cast,c,_) -> find_bound gls c
| concl -> let (x,args) = pf_whd_betadeltaiota_stack gls concl [] in
            match strip_outer_cast x with
              Rel(_) -> (true,concl)
            | DOP0(Meta _) -> (true,concl)
            | _ -> (false, concl) ;;

(* Checks if the head of an application is bound or not. *)
let rec is_head_rel = function
  DOPN(AppL,args) -> is_head_rel (hd_vect args)
| DOP2(Cast,c,_) -> is_head_rel c
| Rel(_) -> true
| DOP0(Meta _) -> true
| _ -> false ;;

(* Repeats a tactic t n times. *)
let DO n t = dorec n 
   where rec dorec k = 
          if k < 1 then IDTAC
          else if k = 1 then t else t THEN (dorec (k-1)) ;;

(* Generalizes a goal gls n times with respect to the n arguments of a program
   if the program is an application with at least n arguments.
   Associated to this new goal the program without the n generalized arguments.
   Applied in the case of a Rec or a MutCase at the head of the program.
   Entries : n : number of generalizations
             pg : program
             gls : current goal
   Output : generalized goal *)
let rec generalize_term n pg gls = 
  if n=0 then IDTAC gls else 
  let (wc,KONT) = StartWalk gls in  
  (genrec pg THEN (fun gl -> let pg = extract_pg gl in
                              generalize_term (n-1) pg gl)) gls
  where genrec pg gls =
  match telescope_appl(strip_outer_cast pg) with
    DOPN(AppL,args) -> 
      let [t1;t2] = list_of_vect args in
      try let (ccival,ccityp,(clenv,pmap)) = 
        specif_of_annot gls (make_clenv wc t2 gls,[]) (DOP0(Sort(Prop Pos))) t2
          in let newconcl = DOP2(Prod,ccityp,DLAM(named_hd ccityp Anonymous,subst_term ccival (pf_concl gls)))
          in (KONT clenv.hook THEN (cut_tac newconcl) THENS
              [intro THEN
              (fun gls ->
              let (id,T) = hd_sign (pf_hyps gls) in
              let (DOP2(Prod,_,B)) = pf_whd_betadeltaiota gls T in
              let (wc,KONT) = StartWalk gls in
              let clenv' = connect_clenv wc clenv in
              let clenv'' = clenv_change_head (applist(VAR id,[ccival]),
                                                  (SAPP B ccival)) clenv' in
              assoc_THEN KONT clenv'' pmap
                         (fun (SOME pg) gl -> Realizer pg gl
                            | NONE gl -> IDTAC gl) gls);
              Realizer t1]) gls
      with NotProgram -> error"insufficient specification of the program"
  | _ -> anomaly"generalize_term" ;;        

(* Checks if an application has no more than one argument. *)
let rec with_max_one_arg = function
  DOP2(Cast,t,_) -> with_max_one_arg t
| DOPN(AppL,args) -> ((vect_length args)<= 2)
| _ -> true ;;

(* Creates a list with the names of logical hypotheses in the current context
   that depend on c (c has to be a pure program term).
   Entries : c : program term
             gls : current goal 
   Output : list of hypotheses names *)
let one_generalize c gls = 
  try let sign = pf_hyps gls and sigma = Project gls in
      let j = execute_pure sigma sign c in
      let ccival = j.VAL in
      let rec fullrec sign n = 
        if n=0 then [] else
        let (signid,signval) = hd_sign sign in
        if (dependent (strip_outer_cast ccival) signval) & (with_max_one_arg (collapse_appl signval)) & (is_logical [] signval gls) then
          (global (GLOB sign) signid)::(fullrec (tl_sign sign) (n-1))
        else fullrec (tl_sign sign) (n-1)
      in fullrec sign (sign_length sign)
  with UserError _ -> [] ;;

(* Treats a list of hypotheses names and suppresses all the ones whose
   type already appears in the current conclusion.
   Entries : concl : current conclusion
             l : list of hypotheses names
             gls : current goal
   Output : list of hypotheses names (passed through a filter) *)
let rec keep_only_not_in_concl concl l gls = match l with
    x::l' -> let t = pf_type_of gls x in
             if dependent t concl then
              keep_only_not_in_concl concl l' gls
             else
              x::(keep_only_not_in_concl concl l' gls)
  | [] -> [] ;;

(* Generalizes a conclusion first with respect to a list of hypotheses
   and then with respect to a value val of type typ.
   Entries : val : proof value
             typ : its type
             l : list of hypotheses names
             concl : current conclusion
             gls : current goal 
   Output : new generalized conclusion *)
let rec create_new_concl val typ l concl gls = 
let rec crec l = (match l with
  x::l' -> let t = pf_type_of gls x in
            DOP2(Prod,t, DLAM(named_hd t Anonymous,(crec l')))
| [] -> concl)
in DOP2(Prod,typ,DLAM(named_hd typ Anonymous,subst_term val (crec l))) ;;

(* Gives the type of the nth product of a type t.
   Entries : n : number
             t : type
   Output : type *)
let rec good_type n t = 
  match t with
    DOP2(Prod,_,DLAM(_,A)) -> if n=0 then A else good_type (n-1) A
  | _ -> anomaly "shouldn't appear" ;;

(* Generalizes a goal gls n times with respect to the n arguments of a program
   if the program is an application with at least n arguments.
   Associated to this new goal the program without the n generalized arguments.
   This is only applied in the case of a fixpoint at the head of the program.
   Entries : n : number of generalizations
             pg : program
             gls : current goal
   Output : generalized goal *)
let rec generalize_fix_term n pg gls = 
  if n=0 then IDTAC gls else 
  let (wc,KONT) = StartWalk gls in  
  (genrec pg THEN (fun gl -> let pg = extract_pg gl in
                              generalize_fix_term (n-1) pg gl)) gls
  where genrec pg gls =
  match telescope_appl(strip_outer_cast pg) with
    DOPN(AppL,args) -> 
      let [t1;t2] = list_of_vect args in
      try let (ccival,ccityp,(clenv,pmap)) = 
        specif_of_annot gls (make_clenv wc t2 gls,[]) (DOP0(Sort(Prop Pos))) t2
          in let l = one_generalize t2 gls
          in let l' = keep_only_not_in_concl (pf_concl gls) l gls
          in let newconcl = create_new_concl ccival ccityp l' (pf_concl gls) gls
          in (KONT clenv.hook THEN (cut_tac newconcl) THENS
              [intro THEN
              (fun gls ->
              let (id,T) = hd_sign (pf_hyps gls) in
              let len = (length l') in
              let typ = good_type len (pf_whd_betadeltaiota gls T) in
              let (wc,KONT) = StartWalk gls in
              let clenv' = connect_clenv wc clenv in
              let clenv'' = clenv_change_head (applist(VAR id,[ccival]@l'),
                                 (subst1 ccival (lift (-len) typ))) clenv' in
              assoc_THEN KONT clenv'' pmap
                         (fun (SOME pg) gl -> Realizer pg gl
                            | NONE gl -> IDTAC gl) gls);
              Realizer t1]) gls
      with NotProgram -> error"insufficient specification of the program"
  | _ -> anomaly"generalize_fix_term" ;;        

(* Does as many preliminary logical introductions as necessary in the case of
   a program corresponding to an application.
   Entries : concltyp : head of the conclusion
             concl : conclusion
             gls : current goal
   Output : new goal with introductions done *)
let refinew_intros concltyp concl gls =
  let ni = if is_head_rel concltyp then
            let (_,args) = pf_whd_betadeltaiota_stack gls concltyp [] in
              SOME(nb_intros_liees_list args gls concl)
           else NONE
  in match ni with
      NONE -> intros gls
    | SOME(ni) -> DO ni intros gls ;;

(* Refines the current goal with the proof term ccival of type ccityp.
   clenv is the current clausal environment.
   pmap the association list between metas and programs.
   The head of the clenv is transformed to e the good one.
   Then, a first order refinement is tried and then a second order if it fails.
   Entries : ccival : proof term to refine
             ccityp : its type
             clenv : current clausal environment
             pmap : current association list between metas and programs
             gls : current goal
   Output : new refined goal(s) with program(s) associated to *)
let apply_tac KONT (ccival,ccityp) (clenv,pmap) gls = 
  let clenv' = clenv_change_head (ccival,ccityp) clenv in   
  let clenv'' = clenv_apply_n_times (nb_prod ccityp) clenv' in
  let templtyp = clenv_instance_template_type clenv'' in
  let clenv''' = clenv_apply_n_times (nb_prod templtyp) clenv'' in
    try assoc_THEN KONT clenv''' pmap (fun (SOME pg) gl -> Realizer pg gl
                                      | NONE gl -> IDTAC gl) gls
    with UserError _ ->
        assoc_THEN2 KONT clenv''' pmap (fun (SOME pg) gl -> Realizer pg gl
                                      | NONE gl -> IDTAC gl) gls
;;

(* Applies the previous tactics but first decides the number of logical
   introductions to be done.
   The current program is head applied to args. The corresponding proof term
   is calculated.
   Entries : head : head of the current program
             args : args of the current program
             gls : current goal
   Output : new refined goal(s) with program(s) associated to *)
let refine_pg head args gls = 
  let (wc,KONT) = StartWalk gls in
  let concl = pf_concl gls in
  let (ccival,ccityp,(clenv,pmap)) = 
   specif_of_annot gls (make_clenv wc head gls,[]) concl (applist(head,args)) in
  let (b,concltyp) = find_bound gls ccityp in
  if b then
      (KONT clenv.hook THEN (refinew_intros concltyp concl) THEN 
        (fun gl -> let (wc,KONT) = StartWalk gl in
                   let clenv' = connect_clenv wc clenv in
                      apply_tac KONT (ccival,ccityp) (clenv',pmap) gl)) gls
  else let ni = nb_intros gls concl in
       (KONT clenv.hook THEN (RepeatFor ni intro)
        THEN (fun gl -> let (wc,KONT) = StartWalk gl in
                        let clenv' = connect_clenv wc clenv in
                        apply_tac KONT (ccival,ccityp) (clenv',pmap) gl)) gls ;;

(* General tactics to refine a goal with a program which is an application.
   Applies the previous tactics by deciding if generalizations are needed.
   The current program is head applied to args. The corresponding proof term
   is calculated.
   Entries : head : head of the current program
             args : args of the current program
             gls : current goal
   Output : new refined goal(s) with program(s) associated to *)
let program_app head args gls = 
  let (wc,KONT) = StartWalk gls in
  let concl = pf_concl gls in
  match (strip_outer_cast head) with
    DOPN(MutCase,_) -> generalize_term (length args) (applist(head,args)) gls
  | DOPN(Fix _,_) -> generalize_fix_term (length args) (applist(head,args)) gls
  | DOPN(CoFix _,_) -> generalize_fix_term (length args) (applist(head,args)) gls
  | _ -> 
    let (_,ccityp,_) = specif_of_annot_part gls 
                     (make_clenv wc head gls,[]) concl head (applist(head,args))
         in let (b,concltyp) = find_bound gls ccityp
         in try if b then
                  ((refinew_intros concltyp concl) 
                  THEN (refine_pg head args)) gls
                else 
                  let concl' = (red_all gls concl) in
                  ((change_pf concl') THEN 
                    (fun gl -> let ni = nb_intros gl concl'
                               in RepeatFor ni intro gl)
                  THEN (refine_pg head args)) gls
            with Generalize k -> generalize_term k (applist(head,args)) gls
 ;;

(* ============== *)
(* Recursion Case *)
(* ============== *)

(* Checks if a term is a variable or not. *)
let rec is_var = function
  VAR _ -> true
| DOP2(Cast,c,_) -> is_var c
| DOPN(AppL,args) -> is_var (hd_vect args) 
| _ -> false ;;

(* Checks if a term is a constant or not. *)
let rec is_const = function
  DOPN(Const _,_) -> true
| DOP2(Cast,c,_) -> is_const c
| _ -> false ;;

(* Checks if a term is a reducible constant or not. *)
let is_const_red t args gls = 
  let b = try let t' = strong (whd_betadeltaiota (Project gls)) (pf_one_step_reduce gls (applist(t,args))) in 
              is_construct_head t'
          with UserError _ -> false in
  (is_const t) & b ;;

(* Calculates a list of integers.
   This list represents for each constructor type of the number of 
   corresponding lambda-abstraction. Its depends on the fact that
   the type is recursive or not.
   Entries : recursive : if the type is recursive or not
             l : list of constructors types
   Output : list of number of lamba-expression for each constructor *)
let rec search_nb_lam recursive l = 
  match l with
    S::l' -> let rec mrec n m S = 
                match strip_outer_cast S with
                  DOP2(Lambda,_,DLAM(_,B)) -> mrec (n+1) m B
                | DOP2(Prod,A,DLAM(_,B))   -> if noccurn (n+1) A then
                                                   mrec (n+1) m B
                                              else (match B with
                                                    DOP2(Prod,_,DLAM(_,B')) -> mrec (n+2) (m+1) B'
                                                  | _ -> n-m)
                | _ -> if recursive then n-2 else n-m-1
             in (match S with
                  DLAMV(_,l) -> (map (mrec 1 0) (list_of_vect l))
                | DLAM(_) -> let (_,l) = decomp_all_DLAMV_name S in
                                (map (mrec 1 0) (list_of_vect l))
                | _ -> (mrec 0 0 S)::search_nb_lam recursive l')
 | [] -> [] ;;

(* Adds a lambda-expression [name:val]... after the nth abstraction
   of the lambda-expression f.
   Entries : n : number of abstractions to pass
             f : lambda-expression
             name : name to put in the new abstraction
             val : type of name
   Output : new lambda-expression *)
let rec change_lam_at n f name val =
  if n=0 then DOP2(Lambda,val,DLAM(Name(name),subst_var name (lift 1 f)))
  else match strip_outer_cast f with
         DOP2(Lambda,A,DLAM(name',B)) -> 
              DOP2(Lambda,A,DLAM(name',change_lam_at (n-1) B name val))
       | _ -> anomaly"shouldn't appear" ;;

(* Applies the precedent function on a list of number and a list of
   lambda-expressions.
   Entries : name : name to be add in each lambda-expression
             ln : list of numbers
             lf : list of lambda-expressions
             val : type of name
   Output : list of new lambda-expressions *)
let rec modif_lf name ln lf val  =
  match (lf,ln) with
    (f::lf',n::ln') -> (change_lam_at n f name val)::modif_lf name ln' lf' val
  | ([],[]) -> [] ;;

(* Suppress the integer n on each number element of the input list.
   Entries : n : number
             l : list of integers
   Output : list of integers *)
let rec suppress_params n = function
  m::l -> (m-n)::(suppress_params n l)
| [] -> [] ;;

(* Looks in the current signature if they are hypotheses that depend
   on the term c. If yes, adds these hypotheses as lambda-abstractions
   in each element of the list lf at the right place by using the precedent
   functions.
   Entries : recursive : if the type of c is recursive or not
             c : term on which hypotheses have to depend
             lf : list of lambda-expressions
             gls : current goal
   Output : (l',lf') : (list of dependent hypotheses names, modified lf) *)
let full_generalize recursive c lf gls =
  try let sign = pf_hyps gls and sigma = Project gls in
      let j = execute_pure sigma sign c in
      let ccival = j.VAL in
      let rec fullrec sign n l = 
        if n=0 then [],l else
        let (signid,signval) = hd_sign sign in
        if (dependent (strip_outer_cast ccival) signval) then
          let inf = snd(infexecute sigma (FSIGN sigma sign) signval) in
          match inf with
            Logic -> (global (GLOB sign) signid)::(fst (fullrec (tl_sign sign) (n-1) l)),l
          | Inf j -> let T = (execute_meta_prog sigma (FSIGN sigma sign) !METAMAP c).TYP
                     in let m = (match T with
                                    DOPN(MutInd _,_) ->
                                      search_nb_lam recursive [mis_lc(mind_specif_of_mind T)]
                                  | _ -> 
                                    let (T',largs) = find_minductype sigma T
                                    in let m' = search_nb_lam recursive [mis_lc(mind_specif_of_mind T')] 
                                    in suppress_params (length largs) m')
                     in let l' = modif_lf signid m lf j.VAL in
                        (global (GLOB sign) signid)::(fst(fullrec (tl_sign sign) (n-1) l')),l'
        else fullrec (tl_sign sign) (n-1) l
      in fullrec sign (sign_length sign) lf
  with UserError _ -> [],lf ;;

(* Pass through all the lambda-abstractions of a term and gives the end. *)
let rec split_lf  = function
  DOP2(Lambda,_,DLAM(_,B)) -> split_lf B
| DOP2(Cast,c,_) -> split_lf c
| x -> x ;;

(* Replaces the end of succession of lambda-abstractions by a term. *)
let rec replace_in_lf term = function
  DOP2(Lambda,A,DLAM(na,B)) -> DOP2(Lambda,A,DLAM(na,replace_in_lf term B))
| DOP2(Cast,c,_) -> replace_in_lf term c
| x -> term ;;

let rec replace_in_lf_lift term n = function
  DOP2(Lambda,A,DLAM(na,B)) -> DOP2(Lambda,A,DLAM(na,replace_in_lf_lift term (n+1) B))
| DOP2(Cast,c,_) -> replace_in_lf_lift term n c
| x -> lift (n-1) term ;;

(* Same but with a list of terms to be replaced in a list of terms. *)
let rec replace_in_lf_one term l gls = function
  DOP2(Lambda,A,DLAM(na,B)) -> DOP2(Lambda,A,DLAM(na,replace_in_lf_one term l gls B))
| DOPN(MutCase,args) -> 
    let P = args.(0) and c = args.(1) 
    and lf = list_of_vect(sub_vect args 2 ((vect_length args) - 2)) in
    let sigma = Project gls and sign = pf_hyps gls in
      DOPN(MutCase,vect_of_list((execute_meta_prog sigma (FSIGN sigma sign) !METAMAP term).TYP::(c::(replace_in_lf_list l lf gls))))
| DOPN(MutConstruct(_,i),_) -> nth l i
| DOP2(Cast,c,_) -> replace_in_lf_one term l gls c
| x -> try let x' = strong whd_beta (pf_one_step_reduce gls x) in
            replace_in_lf_one term l gls x'
       with UserError _ -> term
and replace_in_lf_list l l' gls = lrec (l,l') where rec lrec = function
  a::m,b::n -> (replace_in_lf_one a l gls b)::lrec(m,n)
| [],[] -> [] ;;

(* Checks if a term is pure or not. If yes, returns the corresponding proof
   term. If no, returns the same program term.
   Entries : term : term to be checked
             gls : current goal
   Output : (b,val) = (boolean,pure or not pure term) *)
let is_pure term gls = 
  let sigma = Project gls and sign = pf_hyps gls in
  try let j = execute_pure sigma sign term in
      let inf = snd(infexecute sigma (FSIGN sigma sign) j.VAL) in
      match inf with
        Logic -> false, term
      | Inf{TYP=typ;_} -> let j' = execute_pure sigma sign typ in
                          if pf_conv_x gls j.TYP j'.VAL then
                              true, j.VAL
                          else false, term
  with UserError _ -> false, term ;; 

(* Same but in the context of a clenv and not a goal. *)
let clenv_is_pure term sign clenv gls = 
  let sigma = w_Underlying clenv.hook in
  let metamap =
    map (fun (n,CLVAL(_,typ)) -> (n,(typ.rebus,Logic))
           | (n,CLTYP typ) -> (n,(typ.rebus,Logic))) (intavm__toList clenv.env) in
  let metamap2 = 
    map (fun (n,CLVAL(_,typ)) -> 
            if occur_meta typ.rebus then (n,(typ.rebus,Logic))
            else
            (n,(typ.rebus,snd(infexemeta sigma metamap (FGLOB (FSIGN sigma sign)) typ.rebus)))
           | (n,CLTYP typ) -> 
            if occur_meta typ.rebus then (n,(typ.rebus,Logic))
            else (n,(typ.rebus,snd(infexemeta sigma metamap (FGLOB (FSIGN sigma sign)) typ.rebus)))) (intavm__toList clenv.env) in
  try let (j,_) = infexemeta sigma metamap2 (FGLOB (FSIGN sigma sign)) (propify term) in
      let (_,inf) = infexemeta sigma metamap2 (FGLOB (FSIGN sigma sign)) j.VAL in
      match inf with
        Logic -> false, term
      | Inf{TYP=typ;_} -> let (j',_) = infexemeta sigma metamap2 (FGLOB (FSIGN sigma sign)) (propify typ) in
                          if pf_conv_x gls j.TYP j'.VAL then
                              true, j.VAL
                          else false, term
  with UserError _ -> false, term ;; 

(* Register to store the indexes of informtive subgoals. *)
let REG_INF = ref ([]:(int list));;
(* Clears this register in a given goal. *)
let clear_REG_INF gls = REG_INF:=[]; IDTAC gls;;

(* Associates a list of programs to a list of subgoals *)
let assoc_arg lf = 
  (fun i gls -> 
          (match snd(pf_infexecute gls (pf_concl gls)) with
            Logic -> IDTAC gls
          | _ -> REG_INF := !REG_INF@[i];
                 let j = index i !REG_INF in
                 try let f = item lf (j-1) in
                      Realizer f gls
                 with Failure _ -> IDTAC gls)) ;;

(* Applies Elim or Case with ccival and associates the lf to the subgoals.
   Entries : recursive : boolean
             ccival : value to be eliminated
             lf : list of programs to be associated
             gls : current goal
   Output : list of subgoals *)
let elim_pg recursive ccival lf gls = 
  if recursive then
    ((THEN_i (elimination [] ccival) (assoc_arg lf) 0) THEN clear_REG_INF) gls
  else 
    ((THEN_i (SimpleCaseInd ccival) (assoc_arg lf) 0) THEN clear_REG_INF) gls 
;;

(* Applies ElimType with ccityp and associates the lf to the subgoals.
   Entries : recursive : boolean
             ccityp : type to be eliminated
             lf : list of programs to be associated
             gls : current goal
   Output : list of subgoals *)
let elim_type_pg ccityp lf gls = 
((THEN_i (elimination_type ccityp) (assoc_arg lf) 0) THEN clear_REG_INF) gls ;;

(* Applies Case for mutual inductive types 
   with ccival and associates the lf to the subgoals.
   Entries : ccival : value to be eliminated
             lf : list of programs to be associated
             gls : current goal
   Output : list of subgoals *)
let case_pg ccival lf gls = 
  ((THEN_i (case_analysis [] ccival) (assoc_arg lf) 0) THEN clear_REG_INF) gls ;;

(* Applies CaseType for mutual inductive types
   with ccityp and associates the lf to the subgoals.
   Entries : ccityp : type to be eliminated
             lf : list of programs to be associated
             gls : current goal
   Output : list of subgoals *)
let case_type_pg ccityp lf gls = 
 ((THEN_i (case_analysis_type ccityp) (assoc_arg lf) 0) THEN clear_REG_INF) gls ;;

(* Transforms a goal Gamma |- G into Gamma |- (x:ccityp)(ccival=x)G *)
let elim_pattern ccival ccityp gls =
  let prop = applist(pf_global gls (id_of_string "eq"),
                     ccityp::(ccival::[ccival])) in 
    ((cut_tac prop) THENL (auto__trivial) THEN (pattern_list [[2],ccival])) gls ;;

(* Suppress all the products at the head of a term *)
let rec cut_prod = function
  DOP2(Prod,_,DLAM(_,B)) -> cut_prod B
| DOP2(Cast,t,_) -> cut_prod t
| x -> x ;;

(* Same as pf_extraction but in the context of a clenv.
   Entries : ce : clenv
             sign : signature
             t : term to be extracted
   Output : extracted term *)
let rec clenv_extraction ce sign t =
let metamap =
    map (fun (n,CLVAL(_,typ)) -> (n,(typ.rebus,Logic))
           | (n,CLTYP typ) -> (n,(typ.rebus,Logic))) (intavm__toList ce.env) in
let metamap2 = 
    map (fun (n,CLVAL(_,typ)) -> 
            if occur_meta typ.rebus then (n,(typ.rebus,Logic))
            else
            (n,(typ.rebus,snd(infexemeta (w_Underlying ce.hook) metamap (FGLOB (FSIGN (w_Underlying ce.hook) sign)) typ.rebus)))
           | (n,CLTYP typ) -> 
            if occur_meta typ.rebus then (n,(typ.rebus,Logic))
            else (n,(typ.rebus,snd(infexemeta (w_Underlying ce.hook) metamap (FGLOB (FSIGN (w_Underlying ce.hook) sign)) typ.rebus)))) (intavm__toList ce.env) in
  match snd(infexemeta (w_Underlying ce.hook) metamap2 (FGLOB (FSIGN (w_Underlying ce.hook) sign)) t) with
  Logic -> error "pf_extraction : not informative"
| Inf j -> j.VAL
;;

(* Replaces the metas of a term with the corresponding program in pmap.
   Entries : pmap : association list between metas and programs
             t : term where metas have to be replaced
   Output : new term *)
let replace_meta_with_prog pmap t = prec t where rec prec = function
  DOP0(Meta n) -> assoc n pmap 
| DOP2(Lambda,A,DLAM(na,B)) -> DOP2(Lambda,A,DLAM(na,prec B))
| DOP2(Cast,c,_) -> prec c
| DOP2(Prod,A,DLAM(na,B)) ->  DOP2(Prod,A,DLAM(na,prec B))
| DOPN(AppL,args) -> DOPN(AppL,map_vect prec args)
| DOPN(MutCase,args) -> DOPN(MutCase,map_vect prec args)
| DOPN(Fix(vn,i),args) -> DOPN(Fix(vn,i),map_vect prec args)
| DOPN(CoFix(i),args) -> DOPN(CoFix(i),map_vect prec args)
| c -> c ;;

(* Suppress all the logical products of a term. *)
let tail_type gls  t = trec t [] where rec trec t l =
  match t with
    DOP2(Cast,c,_) -> trec c l
  | DOP2(Prod,A,DLAM(_,B)) -> if (is_logical l A gls) then
                                  trec B (true::l)
                              else B
  | _ -> error"tail_type" ;;

(* Applies a term at the end of a succession of lambda-abstractions.
   [xi:Ti]t -> [xi:Ti](t arg).
   Entries : lf : term with lambda-abstractions
             arg : argument to be applied
   Ouput : new term *)
let rec move_arg_in_lf lf arg = 
  match strip_outer_cast lf with
    DOP2(Lambda,A,DLAM(x,B)) -> DOP2(Lambda,A,DLAM(x,move_arg_in_lf B arg))
  | x -> applist(x,[arg]) ;;

(* Applies a term at the end of a succession of lambda-abstractions.
   [xi:Ti]t -> [xi:Ti](arg t).
   Entries : lf : term with lambda-abstractions
             arg : argument to be applied
   Ouput : new term *)
let rec move_arg_in_lf2 lf arg = 
  match strip_outer_cast lf with
    DOP2(Lambda,A,DLAM(x,B)) -> DOP2(Lambda,A,DLAM(x,move_arg_in_lf2 B arg))
  | x -> applist(arg,[x]) ;;

(* Checks if a term contains a MutCase. *)
let rec is_rec = function
  DOP2(Cast,c,_) -> is_rec c
| DOPN(AppL,args) -> let l = list_of_vect args in
                     if (length l = 1) then (is_rec (hd l))
                     else
                      (is_rec (hd l)) or (is_rec (hd(tl l)))
| DOP2(Lambda,c,DLAM(_,c')) -> (is_rec c) or (is_rec c')
| DOP2(Prod,c,DLAM(_,c')) -> (is_rec c) or (is_rec c')
| DOPN(MutInd _,c) -> (exists is_rec (list_of_vect c))
| DOPN(MutConstruct _,c) -> (exists is_rec (list_of_vect c))
| DOPN(MutCase,_) -> true
| DOPN(Fix _,args) -> is_rec (args.(0))
| DOPN(CoFix _,args) -> is_rec (args.(0))
| DOP2(XTRA("ANNOT",_),c,t) -> (is_rec c) or (is_rec t)
| _ -> false ;;

(* Moves a MutCase in an application back to the head in order
   to have a MutCase and no more an application.
   Entries : clenv : local clenv in which typechecking is possible
             gls : current goal
   Output : new term *)
let rec move_rec clenv gls = mrec where rec mrec = function
  (DOPN(AppL,args) as val) -> let t1::[t2] = list_of_vect args in
      (match t1,t2 with
        DOPN(MutCase,args),_ ->
          let lf = list_of_vect(sub_vect args 2 ((vect_length args) - 2))
          and c = args.(1) and P = args.(0) in
          (match lf with
            [f] -> let P' = (match P with
                              DOP2(Prod,_,DLAM(_,B)) -> B
                            | DOP2(Lambda,_,DLAM(_,B)) -> B
                            | _ -> error"move_rec") in
                   let lf' = move_arg_in_lf f t2 in
                    mrec (DOPN(MutCase,vect_of_list((P'::[c])@[lf'])))
          | _ -> val)
     
      |  _,DOPN(MutCase,args) ->
          let lf = list_of_vect(sub_vect args 2 ((vect_length args) - 2))
          and c = args.(1) and P = args.(0) in
          (match lf with
            [f] -> let B = tail_type gls (clenv_type_of clenv t1) in
                   let arg = subst1 (DOPN(MutCase,vect_of_list((P::[Rel(1)])@[f]))) B in
                   let T = clenv_type_of clenv c in
                   let P' = DOP2(Lambda,T,DLAM(named_hd T Anonymous,arg)) in
                   let lf' = move_arg_in_lf2 f t1 in
                    mrec (DOPN(MutCase,vect_of_list((P'::[c])@[lf'])))
          | _ -> val)
      | _,_ -> if is_rec t1 then 
                  if is_rec t2 then 
                      mrec(telescope_appl(applist(mrec t1,[mrec t2])))
                  else
                      mrec(telescope_appl(applist(mrec t1,[t2])))
               else if is_rec t2 then
                      mrec(telescope_appl(applist(t1,[mrec t2])))
                    else
                      applist(t1,[t2]))
| x -> x ;;

(* Suppress all the logical products of a term. *)
let remove_logical_products prod gls = crec [] prod where rec crec l prod =
  match prod with
    DOP2(Cast,c,_) -> crec l c
  | DOP2(Prod,A,(DLAM(Name(x),B) as C)) -> if (is_logical l A gls) then
                                      crec (l@[true]) (SAPP C (VAR x))
                                    else (SAPP C (VAR x))
  | DOP2(Prod,A,DLAM(_,B)) -> if (is_logical l A gls) then
                                      crec (l@[true]) B
                              else B
  | x -> x ;;

(* Same as before but for mutual inductive types. No recursive parameters,
   since an elimination is always non recursive in this case. *)
let rec case_app P c lf gls = 
  let (wc,KONT) = StartWalk gls in
  let concl = pf_concl gls and sign = pf_hyps gls and sigma = Project gls in
  let (b,ccival) = is_pure c gls in
  if b then
    if dependant (strip_outer_cast ccival) concl then 
      let ni = nb_intros_liees ccival gls concl in
        (RepeatFor ni intro THEN (case_pg ccival lf)) gls
    else
      let ni = nb_intros gls concl in
        (RepeatFor ni intro THEN (elim_pattern ccival (pf_type_of gls ccival))
                            THEN (case_pg ccival lf)) gls
  else 
      let (ccival',ccityp,(clenv,pmap)) = 
        specif_of_annot gls (mk_clenv_hnf_constr_type_of wc concl,[]) concl c in
       match ccival' with
        DOPN(MutCase,cl) -> 
          let newc = clenv_extraction clenv sign cl.(1) 
          and lf' = hd_vect(sub_vect cl 2 ((vect_length cl) - 2)) in
          let newlf' = clenv_extraction clenv sign lf' in
          let fin = replace_meta_with_prog pmap (split_lf newlf') in
          let term = DOPN(MutCase,vect_of_list(P::(fin::(map (subst_term newc) lf)))) in
          let newlf = replace_in_lf term newlf' in
            case_app P newc [newlf] gls
      | _ -> if is_rec ccival' then
              let term = move_rec clenv gls ccival' in
               (match term with
                  DOPN(MutCase,args) ->
                    let c_of_c = clenv_extraction clenv sign args.(1) in
                    let lf_of_c = hd_vect(sub_vect args 2 ((vect_length args) - 2)) in
                    let lf_of_c' = clenv_extraction clenv sign lf_of_c in
                    let fin = replace_meta_with_prog pmap (split_lf lf_of_c') in
                    let term = DOPN(MutCase,vect_of_list(P::(fin::lf))) in
                    let newlf = lift (-1) (subst_term c_of_c (replace_in_lf term lf_of_c')) in
                      case_app P c_of_c [newlf] gls
                | _ -> error"case_app")
             else 
              (try 
                let j = execute_meta_prog sigma (FSIGN sigma sign) !METAMAP c in
              if (pf_conv_x gls ccityp (execute_pure sigma sign j.TYP).VAL) then
                  (try let newc = strong whd_beta (pf_one_step_reduce gls c) in
                    (match newc with
                      DOPN(MutCase,args) ->
                        let newP = args.(0) and newc' = args.(1) 
                        and newlf = list_of_vect(sub_vect args 2 
                                      ((vect_length args) - 2)) in
                        let args' = replace_in_lf_list lf newlf gls in
                          case_app P newc' args' gls
                    | _ -> error "not a rec")
                  with UserError _ -> raise Elim)
                else raise Elim
            with Elim | UserError _ ->
             (try
              if occur_meta ccival' then
              let head = without_annot c in              
              let ni = nb_intros gls concl in
                (RepeatFor ni intro THEN (case_type_pg (cut_prod ccityp) ([head]@lf))) gls
             else 
              let ni = nb_intros gls concl in
                (RepeatFor ni intro THEN (case_pg ccival' lf)) gls 
              with UserError(s,_) -> 
                  if (is_singleton sigma (GLOB sign) (clenv_extraction clenv sign ccityp) (decomp_app (remove_logical_products ccityp gls))) then
                       let newpg = elim_sig ccival' (remove_logical_products ccityp gls) gls in
                       (match newpg with
                          DOPN(MutCase,args) ->
                            let newP = args.(0) and newc' = args.(1) 
                            and newlf = list_of_vect(sub_vect args 2 
                                      ((vect_length args) - 2)) in
                            let args' = replace_in_lf_lift (DOPN(MutCase,vect_of_list (P::(Rel 1)::lf))) 0 (clenv_extraction clenv sign (hd newlf)) in
                              case_app P newc' [args'] gls
                        | _ -> error s)
          else error s));;

(* Same as before but for mutual inductive types. *)
let program_case P lf c gls =
  let sigma = Project gls and sign = pf_hyps gls in
  let j = execute_meta_prog sigma (FSIGN sigma sign) !METAMAP c in
  if (is_var c) then
    let (l',lf') = full_generalize false c lf gls in
      ((generalize_tac l') THEN (case_app P c lf')) gls
  else case_app P c lf gls ;;

(* ========= *)
(* Injection *)
(* ========= *)

(* Checks if a term is specified or not. *)
let rec non_specified = function
  DOP2(Cast,c,_) -> non_specified c
| DOPN(AppL,args) -> non_specified (hd_vect args)
| DOP2(Lambda,_,DLAM(_,c)) -> non_specified c
| DOP2(XTRA("ANNOT",[]),_,_) -> false
| _ -> true ;;

(* Constructs the ith constructor of ind applied to its parameters.
   Entries : i : number of the constructor
             ind : inductive type
             largs : list of arguments to be applied
   Output : constructor. *)

let val_mconstruct i ind largs = 
  match strip_outer_cast ind with
    DOPN(MutInd(sp,j),cl) -> let nparams = mind_nparams ind in
                          let (globargs,_) = chop_list nparams largs in
                            applist(DOPN(MutConstruct((sp,j),i),cl),globargs)
  | _ -> error "val_mconstruct" ;;

(* Same as execute_pure but in the context of a clenv *)
let clenv_execute_pure ce sign t =
  let metamap =
    map (fun (n,CLVAL(_,typ)) -> (n,typ.rebus)
           | (n,CLTYP typ) -> (n,typ.rebus)) (intavm__toList ce.env) in
 fexemeta (w_Underlying ce.hook) metamap (GLOB sign) (propify t) ;;

(* Exception raised to trap particular cases of injection.
   Ie when [x:T]y needs to be transformed into (exists T' [x:T]y),
   for example. *)
exception Particular;;

(* Transforms a term ccival into (Construct(1,ind) args ccival).
   Corresponds to an injection in the good singleton type.
   Entries : n : integer to trap errors
             ccival : value to be injected
             ccityp : its type
             clenv : local clenv
             pmap : association list of metas and programs
             gls : current goal
   Output : new goal with good program associated to *)
let injection n (ccival,ccityp) (clenv,pmap) gls =
  (change_pf (red_all gls (pf_concl gls)) THEN intros THEN
    (fun gls ->
        let sigma = Project gls and sign = pf_hyps gls 
        and concl = pf_concl gls in
       let (wc,KONT) = StartWalk gls in 
       let head,largs = let (head,args) = decomp_app concl in
            (match head with
                DOPN(Const _,_) -> decomp_app (strong whd_beta (pf_one_step_reduce gls concl))
              | _ -> (head,args)) in
        (try if is_singleton sigma (GLOB sign) (clenv_extraction clenv sign ccityp) (head,largs) then
             if (n=0) & (occur_meta ccityp) then
              raise Particular
             else
               let val = 
                    let (T,largs) = find_minductype sigma concl in
                       val_mconstruct 1 T largs in
               let (_,args,clenv') = clausal_form_info clenv (pf_type_of gls val) in
               let j = clenv_execute_pure clenv' sign
                        (strong whd_beta (applist(val,args@[ccival]))) in
               let clenv'' = connect_clenv wc clenv' in
                 apply_tac KONT (j.VAL,j.TYP) (clenv'',pmap) gls
             else error"injection"
        with UserError _ | Invalid_argument _ -> if n=0 then error"injection"
                            else let pg = clenv_extraction clenv sign ccival
                                 in Realizer pg gls))) gls ;;

(* Takes a program on the form (head args). 
   Calculates the corresponding proof term ccival.
   Looks if transformations are first needed on ccival (to be injected),
   that is if MutCase need to be moved first. Moves them if necessary
   and applies the appropriated tactics. Then, branches on the previous
   injection function.
   Entries : head : head of the program
             args : list of arguments of the program
             gls : current goal
   Output : new subgoal *)
let program_inj head args gls = 
  let (wc,KONT) = StartWalk gls in
  let concl = pf_concl gls and sign = pf_hyps gls in
  let (ccival,ccityp,(clenv,pmap)) = 
    specif_of_annot gls (make_clenv wc head gls,[]) concl (applist(head,args))
  in prec 0 ccival clenv gls where rec prec n ccival clenv gls =
  match telescope_appl(strip_outer_cast ccival) with
     DOPN(MutCase,args) -> 
      let lf = list_of_vect(sub_vect args 2 ((vect_length args) - 2))
      and c = args.(1) in
      let cty = clenv_type_of clenv c in
        (match lf with
          [f] -> if occur_meta c then
                  let cfw = clenv_extraction clenv sign c in
                  (case_analysis_type cty 
                   THENS [Realizer (replace_meta_with_prog pmap cfw);(fun gl ->
                          let (wc,_) = StartWalk gl in
                          let clenv' = connect_clenv wc clenv in
                           prec (n+1) f clenv' gl)]) gls
                 else (case_analysis [] c
                       THEN (fun gl ->
                         let (wc,_) = StartWalk gl in
                         let clenv' = connect_clenv wc clenv in
                          prec (n+1) f clenv' gl)) gls
        | _ -> injection n (ccival,ccityp) (clenv,pmap) gls)
  | DOP2(Lambda,_,B) -> 
      (intro THEN (fun gl -> let (wc,_) = StartWalk gl in
                             let clenv' = connect_clenv wc clenv in
               OnLastHyp (fun (SOME id) -> prec (n+1) (SAPP B (VAR id)) clenv')
             gl)) gls
  | DOPN(AppL,args) ->
      let t1::[t2] = list_of_vect args in
      (match t1,t2 with
         DOPN(MutCase,args),_ ->
          let lf = list_of_vect(sub_vect args 2 ((vect_length args) - 2))
          and c = args.(1) and P = args.(0) in
          (match lf with
            [f] -> let P' = (match P with
                              DOP2(Prod,_,DLAM(_,B)) -> B
                            | _ -> error"program_inj") in
                   let lf' = move_arg_in_lf f t2 in
                   let term = DOPN(MutCase,vect_of_list((P'::[c])@[lf'])) in
                    prec (n+1) term clenv gls
          | _ -> injection n (ccival,ccityp) (clenv,pmap) gls)
      | _,DOPN(MutCase,args) ->           
          let lf = list_of_vect(sub_vect args 2 ((vect_length args) - 2))
          and c = args.(1) and P = args.(0) in
          (match lf with
            [f] -> let B = tail_type gls (clenv_type_of clenv t1) in
                   let arg = subst1 (DOPN(MutCase,vect_of_list((P::[Rel(1)])@[f]))) B in
                   let T = clenv_type_of clenv c in
                   let P' = DOP2(Lambda,T,DLAM(named_hd T Anonymous,arg)) in
                   let lf' = move_arg_in_lf2 f t1 in
                   let term = DOPN(MutCase,vect_of_list((P'::[c])@[lf'])) in
                    prec (n+1) term clenv gls
          | _ -> injection n (ccival,ccityp) (clenv,pmap) gls)
      | _,_ -> if (is_rec t1) then
                  let term = move_rec clenv gls t1 in
                    prec (n+1) (applist(term,[t2])) clenv gls
               else if (is_rec t2) then
                      let term = move_rec clenv gls t2 in
                        prec (n+1) (applist(t1,[term])) clenv gls
                    else injection n (ccival,ccityp) (clenv,pmap) gls)
  | _ -> injection n (ccival,ccityp) (clenv,pmap) gls ;;

(* Calls the previous function but first looks if previous logical
   introductions are needed. *)
let program_inj_intros head args gls =
  let (wc,KONT) = StartWalk gls in
  let concl = pf_concl gls in
  let (ccival,ccityp,(clenv,pmap)) = 
   specif_of_annot gls (make_clenv wc head gls,[]) concl (applist(head,args)) in
  let ni = match telescope_appl(strip_outer_cast ccival) with
           DOPN(MutCase,args) -> 
              let lf = list_of_vect(sub_vect args 2 ((vect_length args) - 2)) in
               (match lf with
                  [f] -> nb_intros gls concl
                | _   -> 0)
          | DOPN(AppL,args) -> 
              (match list_of_vect args with
                _::[DOPN(MutCase,args')] ->
              let lf = list_of_vect(sub_vect args' 2 ((vect_length args') - 2))
              in (match lf with
                    [f] -> nb_intros gls concl
                  | _   -> 0)
               | _ -> 0)
          | _ -> 0
  in (RepeatFor ni intros THEN program_inj head args) gls ;;
          

(* ================================================ *)
(* Simplification : b <-> if b then true else false *)
(* ================================================ *)

(* Constructs a list of (MutConstruct(i,P) params) for 1<=i<=n. *)

let mconstruct_lf n P largs = 
  make_list_ind n [] where rec make_list_ind k l = 
    if k=0 then l else make_list_ind (k-1) ((val_mconstruct k P largs)::l) ;;

(* Calls the previous mutconstruct_lf with n = length of constructors
   and P = ind. *)
let construct_abs_mconstr ind largs = 
   let l = mis_consnames(mind_specif_of_mind ind) in
   let n = length (list_of_vect l) in
          mconstruct_lf n ind largs ;;

(* Checks if a MutCase is hidden behind a constant. *)
let is_hidden_rec gls t = 
   try let t' = strong whd_beta (pf_one_step_reduce gls t) in
          (match strip_outer_cast t' with
             DOPN(MutCase,_) -> true
            | _ -> false)
   with UserError _ -> false ;;

(* Checks if a term is a modifiable.
   A term is modifiable 
   (i.e. b can be transformed into if b then true else false)
   if it has no hidden MutCase or if it is a variable. 
   It must not be a reducible constant too. 
   The term has to be (head args). *)
let is_modifiable head args gls = 
  (not(exists (is_hidden_rec gls) args) or (is_var head)) & not(is_const_red head args gls);;

(* Checks if a product is anonymous or not. *)
let rec is_anonymous_prod gls = function
  DOP2(Prod,_,DLAM(Anonymous,_))-> true
| DOP2(Prod,A,DLAM(_,B)) -> not(is_logical [] A gls) & noccurn 1 B
| DOP2(Cast,c,_) -> is_anonymous_prod gls c
| _ -> false ;;

(* Creates a list of n functions applying IDTAC to a goal.
   The first function of the list is first_tac. *)
let make_IDTAC_list first_tac n = 
let rec mrec n = 
  if n=0 then []
  else (fun gl -> IDTAC gl)::(mrec (n-1)) 
in first_tac::(mrec n);;

(* Compares two conclusion concl1 and concl2.
   If they just differs in some arguments, these ones are kept in a list
   of the form (argconcl1=argconcl2).
   Entries : concl1, concl2 : two conclusions
             gls : current goal
   Output : list of arguments.*)
let rec compare_two_concl concl1 concl2 gls = 
  match (collapse_appl concl1,collapse_appl concl2) with
    (DOPN(AppL,args1),DOPN(AppL,args2)) -> 
      if eq_constr(args1.(0),args2.(0)) then
        let rec mrec l = function
          x::l1,y::l2 -> if eq_constr(x,y) then mrec l (l1,l2)
                         else 
                          let prop = applist(pf_global gls (id_of_string "eq"),
                                      (pf_type_of gls x)::(x::[y])) in 
                           mrec (prop::l) (l1,l2)
        | [],[] -> l
        in mrec [] (list_of_vect (tl_vect args1),list_of_vect (tl_vect args2))
      else error"compare_two_concl"
 |  (DOP2(Cast,c1,_),DOP2(Cast,c2,_)) -> compare_two_concl c1 c2 gls
 |  (DOP2(Cast,c1,_),_) -> compare_two_concl c1 concl2 gls
 |  (_,DOP2(Cast,c2,_)) -> compare_two_concl concl1 c2 gls
 |  _ -> error"compare_two_concl" ;;

(* Checks if a term is a constructor or a mutconstructor or an application
   beginning with a constructor or a mutconstructor. *)
let rec is_a_construct gls = function
  DOP2(Cast,c,_) -> is_a_construct gls c
| DOPN(Const _,_) as c -> let c' = strong whd_beta (pf_one_step_reduce gls c) in
                            is_a_construct gls c'
| DOP2(Lambda,_,DLAM(_,B)) -> is_a_construct gls B
| DOPN(MutConstruct _,_) -> true
| DOPN(AppL,args) -> is_a_construct gls args.(0)
| _ -> false ;;

(* Creates a list of functions from a list whose elements have the form
   (arg1=arg2). Each function cuts an element (arg1=arg2) in the
   current goal, introduces this new hypothesis and eliminates it. *)
let rec make_cut_list list_cut gls  = match list_cut with
  a::l -> 
    ((fun gl -> (cut_tac a 
                 THENS [(intro THEN (fun gl -> OnLastHyp 
                          (fun (SOME id) -> 
                                elimination [] (VAR id)) gl));auto__auto NONE]) gl)
     THEN (fun gl -> make_cut_list l gl)) gls 
| [] -> IDTAC gls;;

(* Global function with modifies a program b into if b then true else false.
   The program has the form (head args). *)
let modif_app_ind head args gls = 
  let pg = applist(head,args) in
  let concl = pf_concl gls in
  let ni = nb_intros gls concl in
    (RepeatFor ni intro THEN 
      (fun gls ->
        let (wc,KONT) = StartWalk gls in
        let concl = pf_concl gls and sigma = Project gls 
        and sign = pf_hyps gls in
        let (ccival,ccityp,(clenv,pmap)) = specif_of_annot gls (make_clenv wc head gls,[]) concl pg in
        let typg = (execute_meta_prog sigma (FSIGN sigma sign) !METAMAP pg).TYP in
        let head',largs = 
          let (head,args) = decomp_app (remove_logical_products ccityp gls) in
              (match head with
                DOPN(Const _,_) -> decomp_app (strong whd_beta (pf_one_step_reduce gls (remove_logical_products ccityp gls)))
              | _ -> (head,args)) in
        try let lf = if is_singleton sigma (GLOB sign) typg (head',largs) then
                    [DOP2(Lambda,typg,DLAM(named_hd typg Anonymous,Rel(1)))]
                 else
                  if (is_modifiable head args gls) & not(fst(clenv_is_pure ccival sign clenv gls)) then
                    let (T,largs) = find_minductype sigma typg in
                    construct_abs_mconstr T largs
                  else raise Rewrite in
        if occur_meta ccival then
          let first_tac = (fun gl -> let concl = pf_concl gl in
                               if is_anonymous_prod gl concl then raise NoModif
                               else IDTAC gl) in
          let IDTAC_list = make_IDTAC_list first_tac (length lf) in 
            (KONT clenv.hook 
            THEN case_type_pg (cut_prod (applist(head',largs))) ([pg]@lf) 
            THENS IDTAC_list) gls
        else
            (KONT clenv.hook THEN case_pg ccival lf) gls
        with NoModif -> error"modif_app_ind"
            | Invalid_argument _ | Rewrite | UserError _-> 
                         let concl1 = remove_logical_products ccityp gls in
                         let concl2 = remove_logical_products concl gls in
                         let list_cut = compare_two_concl concl1 concl2 gls in
                          ((THEN_i (make_cut_list list_cut) (assoc_arg [pg]) 0) THEN clear_REG_INF) gls)) gls ;;

(* Injection of a program which can either not be specified or 
   not be injected with the standard method. *)
let inject_bad_program pg gls = 
      let (wc,KONT) = StartWalk gls and sigma = Project gls 
      and sign = pf_hyps gls and concl = pf_concl gls in
      try let j = execute_pure sigma sign pg in
          let (ccival,ccityp,(clenv,pmap)) = 
          specif_of_annot gls (mk_clenv_hnf_constr_type_of wc j.VAL,[]) concl pg
          in injection 0 (ccival,ccityp) (clenv,pmap) gls 
      with UserError _ ->
      let j = execute_meta_prog sigma (FSIGN sigma sign) !METAMAP pg in
      let clenv = mk_clenv_from wc (j.VAL,j.TYP) in
      let mv = newMETA() in
      let T = pf_extraction gls (pf_concl gls) in
      let clenv' = clenv_pose (Anonymous,mv,T) clenv in
      let ccival = DOP0(Meta(mv)) in
      let pmap = [(mv,pg)] in
      let ccityp = clenv_type_of clenv' ccival
      in injection 0 (ccival,ccityp) (clenv',pmap) gls ;;

(* Checks if the head of an application if a meta or not. *)
let rec is_head_ex = function
  DOP0(Meta _) -> true
| DOP2(Cast,c,_) -> is_head_ex c
| DOPN(AppL,args) -> is_head_ex args.(0)
| _ -> false;;

(*===============*)
(* Fixpoint case *)
(*===============*)

(* Looks for the good index of the decreasing argument in a fixpoint. 
   Entries : i : index from the program point of view.
             concl : current conclusion
             gls : current goal
   Output : new index *)
let real_decreasing_number i concl gls = nrec 1 1 [] concl 
where rec nrec n m l concl =
  match concl with
    DOP2(Prod,A,DLAM(_,B)) ->   if (is_logical l A gls) then
                                  nrec n (m+1) (true::l) B
                                else if n=i+1 then m-1 
                                     else nrec (n+1) (m+1) (false::l) B
  | DOP2(Cast,c,_) -> nrec n m l c
  | _ -> m-1 ;;

(* Finds the ith products in a term.
   Returns the term without the ith first products and a list associating
   de Bruijn indexes and booleans (the boolean says if the free variable
   corresponding to the de Bruijn index is logical or not.
   Entries : i : number
             concl : conclusion
   Ouput : (A,l) = (term, boolean list) *)
let find_ind_arg i concl = nrec 0 [] concl 
where rec nrec n l concl =
  match concl with
    DOP2(Prod,A,DLAM(_,B)) -> if n=i then
                                A,l
                              else nrec (n+1) (false::l) B
  | DOP2(Cast,c,_) -> nrec n l c
  | _ -> concl,l ;;

let eq_constr2 x y = eq_constr(x,y) ;;

(* Generalizes the current goal with respect to the type of arg and 
   associates pg to the new goal.
   Entries : arg : argument to generalize
             pg : program to be associated
             gls : current goal
   Output : new goal *)
let generalize_concl arg pg gls = 
 let (wc,KONT) = StartWalk gls in 
 let typ = pf_type_of gls arg in
 let newconcl = DOP2(Prod,typ,DLAM(named_hd typ Anonymous,subst_term arg (pf_concl gls))) in
    ((cut_tac newconcl) THENS [auto__auto NONE;Realizer pg]) gls ;;

let substn_meta laml typ kind n =
    substrec n
    where rec substrec depth = function 
     (Rel k as x)     -> if k<=depth then x
                         else if k-depth <= 1 then
                                 let p = newMETA() in
                                 METAMAP := !METAMAP@[(p,DOP2(Cast,typ,kind))];
                                 lift_substituend depth (make_substituend (applist(laml,[DOP2(Cast,DOP0(Meta(p)),typ)])))
                         else Rel(k-1)
   | VAR id           -> VAR id
   | DOPN(oper,cl)    -> DOPN(oper,map_vect (substrec depth) cl)
   | DOPL(oper,cl)    -> DOPL(oper,map (substrec depth) cl)
   | DOP1(oper,c)     -> DOP1(oper,substrec depth c)
   | DOP2(oper,c1,c2) -> DOP2(oper,substrec depth c1,substrec depth c2)
   | DLAM(na,c)       -> DLAM(na,substrec (depth+1) c)
   | DLAMV(na,v)      -> DLAMV(na,map_vect (substrec (depth+1)) v)
   | x                -> x;;

let subst1_meta laml typ kind = substn_meta laml typ kind 0;;

let substn laml n = substn_many [make_substituend laml] (n-1);;

(* Generalizes the current goal as many times as they are logical arguments
   in the list largs which are not parameters and are closed.
   Entries : globargs : parameters
             largs : list of arguments
             lrel : list of boolean to typecheck
             pg : program to be associated
             gls : current goal
   Output : new goal *)
let simple_count_needed globargs largs lrel pg gls = 
let sigma = Project gls and sign = pf_hyps gls in
crec largs pg gls where rec crec l pg gls = 
match l with
  a::l' -> if not(exists (eq_constr2 a) globargs) & (closed a) then
            if (is_logical lrel (pf_type_of gls a) gls) then
              (generalize_concl a pg THEN crec l' pg) gls
            else
              let j1 = execute_meta_prog sigma (FSIGN sigma sign) !METAMAP a in
              let typ = j1.TYP and kind = j1.KIND in
              let newpg = match pg with
                DOPN(Fix(vn,i),cl) -> 
                  let j = execute_meta_prog sigma (FSIGN sigma sign) !METAMAP
                            (DOP2(Prod,typ,DLAM(named_hd typ Anonymous,
                                  lift 1 (strip_outer_cast cl.(0))))) in
                  let newcl0 = DOP2(Cast,j.VAL,j.TYP) in
                  let newvn = [| vn.(0)+1 |] in
                  let newcl1 = match cl.(1) with
                               DLAMV(name,clvect) ->
                                 let newtyp = DOP2(Cast,j1.TYP,j1.KIND) in
  DLAMV(name,[| DOP2(Cast,DOP2(Lambda,newtyp,
      DLAM(named_hd newtyp Anonymous,
          (lift 1 (subst1_meta (Rel(1)) typ kind (strip_outer_cast (hd_vect clvect)))))),
        j.VAL) |])
                            | _ -> error"count_needed"  in
                   DOPN(Fix(newvn,i),[| newcl0;newcl1 |])
              | _ -> error "count_needed" in
                (generalize_concl a newpg THEN crec l' newpg) gls
          else crec l' pg gls
| [] -> IDTAC gls ;;

let count_needed globargs largs lrel pg gls = 
let sigma = Project gls and sign = pf_hyps gls in
crec largs pg gls where rec crec l pg gls = 
match l with
  a::l' -> if not(exists (eq_constr2 a) globargs) & (closed a) then
            if (is_logical lrel (pf_type_of gls a) gls) then
              (generalize_concl a pg THEN crec l' pg) gls
            else
              let j1 = execute_meta_prog sigma (FSIGN sigma sign) !METAMAP a in
              let typ = j1.TYP and kind = j1.KIND in
              let newpg = match pg with
                DOPN(Fix(vn,i),cl) -> 
                  let j = execute_meta_prog sigma (FSIGN sigma sign) !METAMAP
                            (DOP2(Prod,typ,DLAM(named_hd typ Anonymous,
                                  lift 1 (strip_outer_cast cl.(i))))) in
                  let newcl = vect_of_list(except (last_vect cl) (list_of_vect cl)) in
                  let newcl0 = map_i_vect 
                                 (fun k t -> if k=i then
                                              DOP2(Cast,j.VAL,j.TYP) 
                                             else t) 0 newcl in
                  let newvn = map_i_vect 
                                 (fun k t -> if k=i then
                                              t+1 
                                             else t) 0 vn in
                  let newcl1 = 
                      let rec nrec cl = match cl with
                        DLAM(name,rest) -> DLAM(name,nrec rest)
                      | DLAMV(name,clvect) ->
                          let newtyp = DOP2(Cast,j1.TYP,j1.KIND) in
  DLAMV(name,map_i_vect 
                (fun k t -> if k=i then
                              DOP2(Cast,DOP2(Lambda,newtyp,
                                   DLAM(named_hd newtyp Anonymous,
                                          (lift 1 (subst1_meta (Rel(1)) typ kind (strip_outer_cast t))))),j.VAL)
                            else t) 0 clvect)
                      | _ -> error"count_needed"  in
                        nrec (last_vect cl) in
                   DOPN(Fix(newvn,i),concat_vect newcl0 [|newcl1|])
              | _ -> error "count_needed" in
                (generalize_concl a newpg THEN crec l' newpg) gls
          else crec l' pg gls
| [] -> IDTAC gls ;;


(* Same but does not the generalizations but counts the numbers that have 
   to be done. *)
let count_added globargs largs lrel gls = 
crec 0 largs where rec crec n l = 
match l with
  a::l' -> if not(exists (eq_constr2 a) globargs) & (closed a) then
            crec (n+1) l'
          else crec n l' 
| [] -> n ;;

(* Applies count_needed with the good parameters and then applies
   the fixpoint tactic with the good arguments.
   Entries : f : identifier of the fixpoint
             i : program number of the decreasing argument
             l : list of booleans to typecheck
             concl : current conclusion
             pg : program to be associated
             gls : current goal
   Output : new goal *)
let rec simple_nb_needed_generalize f i l concl pg gls = 
  match collapse_appl(strip_outer_cast concl) with
    DOPN(AppL,args) -> let t1::largs = list_of_vect args in
                          (match t1 with
                           DOPN(MutInd _,_) -> 
                            let nparams = mind_nparams t1 in
                            let (globargs,_) = chop_list nparams largs in
                            let k = count_added globargs largs l gls in
                             (simple_count_needed globargs largs l pg THEN fix f (i+k+1)) gls
                          | DOPN(Const _,_) -> simple_nb_needed_generalize f i l (strong whd_beta (pf_one_step_reduce gls concl)) pg gls
                          | _ -> anomaly"simple_nb_needed_generalize")
  | _ -> fix f (i+1) gls;;

let rec nb_needed_generalize lf ln ltyp lrel i concl pg gls = 
  match collapse_appl(strip_outer_cast concl) with
    DOPN(AppL,args) -> let t1::largs = list_of_vect args in
                          (match t1 with
                           DOPN(MutInd _,_) -> 
                            let nparams = mind_nparams t1 in
                            let (globargs,_) = chop_list nparams largs in
                            let k = count_added globargs largs lrel gls in
                            let newln = map_i (fun j t -> if j=i then t+k
                                                          else t) 1 ln in
                             (count_needed globargs largs lrel pg THEN mutual_fix lf newln ltyp) gls
                          | DOPN(Const _,_) -> nb_needed_generalize lf ln ltyp lrel i (strong whd_beta (pf_one_step_reduce gls concl)) pg gls
                          | _ -> anomaly"nb_needed_generalize")
  | _ -> mutual_fix lf ln ltyp gls;;

(* Global function to computes a program fixpoint.
   Entries : i, vn, cl : arguments of the fixpoint
             pg : program to be associated
             gls : current goal
   Output : new goal *)
let apply_simple_fixpoint vn cl pg gls = 
  let k = real_decreasing_number vn.(0) (pf_concl gls) gls in
  let ldef = last_vect cl in
  let f = match ldef with
                 DLAMV(Name(f),_) -> f
              | _ -> anomaly "shouldn't appear" in
  let newf =  let avoid = ids_of_sign (pf_hyps gls) in 
              (next_ident_away_from f avoid) in
  let (A,lrel) = find_ind_arg k (pf_concl gls) in
    (simple_nb_needed_generalize newf k lrel A pg THEN 
        (fun gl -> let pg = try extract_pg gl 
                            with UserError _ -> pg in
                   let t = match strip_outer_cast pg with
                            DOPN(Fix(_),cl) -> 
                                (match last_vect cl with
                                    DLAMV(_,t) -> t
                                  | _ -> anomaly "shouldn't appear")
                          | _ -> anomaly "shouldn't appear" in
     OnLastHyp (fun (SOME id) -> Realizer (subst1 (VAR id) (hd_vect t))) gl)) gls ;;

let get_annot l = grec 0 l where rec grec n = function
  a::l -> (match a with
              (DOP2(XTRA("ANNOT",[]),c,T))  -> T::(grec (n+1) l)
            | _ -> if n=0 then 
                     grec (n+1) l
                   else 
                     error "Each non-first function of a mutual program (co)fixpoint must be annotated")
| [] -> [];;

let get_back_fix_annot pg = grec [] pg where rec grec l pg = match pg with
  DOP2(Lambda,A,DLAM(_,B)) -> grec l B
| DOP2(Cast,c,_) -> grec l c
| DOPN(AppL,args) -> grec l (args.(0))
| DOPN(Fix _,args) -> grec l (last_vect args) 
| DOPN(CoFix _,args) -> grec l (last_vect args) 
| DLAM(_,A) -> grec l A
| DLAMV(_,restv) -> get_annot (list_of_vect restv)
| DOP2(XTRA("ANNOT",[]),c,_) -> grec l c
| _ -> l ;;

let apply_fixpoint vn i cl pg gls = 
match (list_of_vect vn) with
  [a] -> apply_simple_fixpoint vn cl pg gls
|  _  ->
  let ltyp = get_back_fix_annot pg in
  let vnlist = list_of_vect vn in
  let ltyp_util = let rec nrec n l = match l with
                                  [] -> if n=i+2 then [pf_concl gls] else []
                                | a::l' -> if n=i+1 then
                                              (pf_concl gls)::a::l'
                                           else a::(nrec (n+1) l') in
                    nrec 1 ltyp in
  let ln = map (fun (n,t) -> real_decreasing_number n t gls) (combine(vnlist,ltyp_util)) in
  let k = nth ln (i+1) in
  let realln = map (fun t -> t+1) ln in
  let (lna,_) = decomp_all_DLAMV_name (last_vect cl) in
  let avoid = ids_of_sign (pf_hyps gls) in
  let lf = map (fun (Name(f)) -> next_ident_away_from f avoid | _ -> anomaly "shouldn't appear") (rev lna) in
  let (A,lrel) = find_ind_arg k (pf_concl gls) in
    (THEN_i (nb_needed_generalize lf realln ltyp lrel k A pg)
            (fun i gl -> let pg = try extract_pg gl 
                                with UserError _ -> pg in
                         let t = 
                            match strip_outer_cast pg with
                              DOPN(Fix(_),cl) -> 
                               let (_,t) = decomp_all_DLAMV_name (last_vect cl)
                               in t
                            | _ -> anomaly "shouldn't appear" in
                         let n = length realln in
                         let newpg = 
                              let rec srec j l t = match l with
                                   a::l' -> srec (j+1) l' (substn (VAR a) (n-j+1) t)
                                 | [] -> t
                              in srec 1 lf (without_annot t.(i)) in
                           Realizer newpg gl) 0) gls ;;

(* Global function to computes a program cofixpoint.
   Entries : cl : argument of the cofixpoint
             pg : program to be associated
             gls : current goal
   Output : new goal *)
let apply_simple_cofixpoint cl pg gls = 
  let ldef = last_vect cl in
  let f = match ldef with
                 DLAMV(Name(f),_) -> f
              | _ -> anomaly "shouldn't appear" in
  let newf =  let avoid = ids_of_sign (pf_hyps gls) in 
              (next_ident_away_from f avoid) in
    (cofix newf THEN 
        (fun gl -> let t = match strip_outer_cast pg with
                            DOPN(CoFix(_),cl) -> 
                                (match last_vect cl with
                                    DLAMV(_,t) -> t
                                  | _ -> anomaly "shouldn't appear")
                          | _ -> anomaly "shouldn't appear" in
     OnLastHyp (fun (SOME id) -> Realizer (subst1 (VAR id) (hd_vect t))) gl)) gls ;;

let apply_cofixpoint cl pg gls = 
let (lna,_) = decomp_all_DLAMV_name (last_vect cl) in
match lna with
  [a] -> apply_simple_cofixpoint cl pg gls
|  _  ->
  let ltyp = get_back_fix_annot pg in
  let avoid = ids_of_sign (pf_hyps gls) in
  let lf = map (fun (Name(f)) -> next_ident_away_from f avoid | _ -> anomaly "shouldn't appear") (rev lna) in
    (THEN_i (mutual_cofix lf ltyp)
            (fun i gl -> let t = match strip_outer_cast pg with
                              DOPN(CoFix(_),cl) -> 
                               let (_,t) = decomp_all_DLAMV_name (last_vect cl)
                               in t
                            | _ -> anomaly "shouldn't appear" in
                         let n = length lf in
                         let newpg = 
                              let rec srec j l t = match l with
                                   a::l' -> srec (j+1) l' (substn (VAR a) (n-j+1) t)
                                 | [] -> t
                              in srec 1 lf (without_annot t.(i)) in
                           Realizer newpg gl) 0) gls ;;

(* ============= *)
(* Global tactic *)
(* ============= *)

let rec program_app_tac head args gls =
  let (wc,KONT) = StartWalk gls and concl = pf_concl gls in
  try let (ccival,ccityp,(clenv,pmap)) = 
     specif_of_annot gls (make_clenv wc head gls,[]) concl (applist(head,args))
     in if is_head_ex ccival then
          (cut_tac (clenv_type_of clenv (DOP0(Meta (fst (hd pmap))))) 
          THENS [(try Realizer (snd (hd (tl pmap))) THEN auto__auto NONE
                 with Failure _ -> auto__auto NONE);
                 Realizer (snd(hd pmap))]) gls
      else 
      try let lf = list__map snd pmap in
            (exact ccival THENS (list__map Realizer lf)) gls
      with UserError _ ->
        try program_app head args gls
        with UserError _ -> 
         try if non_specified ccival then 
                try program_inj_intros head args gls
                with Particular -> inject_bad_program (applist(head,args)) gls
             else error"program_app_tac" 
         with UserError _ -> 
          let pg = applist(head,args) in
          try if is_construct pg then error"program_app_tac"
              else let pg' = strong whd_beta (pf_one_step_reduce gls pg) in
                let (head',args') =  
                 (match pg' with
                      (DOPN(MutConstruct _,_) as head) -> head,vect_of_list []
                    | _ -> error "program_app_tac") in
                  program_app_tac head' (list_of_vect args') gls
          with UserError _ -> modif_app_ind head args gls  
  with (NotProgram | Generalize _ | UserError _) -> program_app head args gls;;

let program_app_ext pg gls =
  let sigma = (Project gls) and sign = (pf_hyps gls) in
  let (linfargs,allargs)= list_arg gls (collapse_appl pg) in
  let head = hd linfargs and infargs = tl linfargs in
  let j = try_execute sigma sign head in
  let n = nb_info_arg gls j.TYP in
  if (n>(length infargs)) then
    match red_all gls (execute_meta_prog sigma (FSIGN sigma sign) !METAMAP pg).TYP with
      DOP2(Prod,A,DLAM((Name(_) as na),B)) ->
        let pg' = DOP2(Lambda,A,DLAM(na,DOPN(AppL,vect_of_list(pg::[Rel(1)]))))
        in deroule_intro pg' gls
    | DOP2(Prod,A,DLAM(_,B)) ->
        let pg' = DOP2(Lambda,A,DLAM(named_hd A Anonymous,
                                     DOPN(AppL,vect_of_list(pg::[Rel(1)]))))
        in deroule_intro pg' gls
    | _ -> program_app_tac head allargs gls
  else program_app_tac head allargs gls ;;

let rec program_tactic pg gls = 
  let go_program_app_ext pg gls =
    try let j = execute_pure (Project gls) (pf_hyps gls) pg in 
        (try exact j.VAL gls
         with UserError _ -> let concl = pf_concl gls in
                            (match concl with
                                (DOPN(MutInd(sp,i),cl)) ->
                                  let c1 = (DOPN(MutConstruct((sp,i),1),cl)) in
                                    (match (pf_type_of gls c1) with
                                      DOP2(Prod,A,_) -> 
                                        if (pf_conv_x gls j.TYP A) 
                                        then exact (DOPN(AppL,vect_of_list(c1::[j.VAL]))) gls
                                        else error "program_tactic"
                                    | _ -> error "program_tactic")
                            | _ -> error"program_tactic"))
    with UserError _ | Induc ->
      (try program_app_ext pg gls
      with UserError(s,_) -> 
        let pg' = strong whd_beta (pf_one_step_reduce gls pg) in
        if eq_constr(pg,pg') then error s 
        else program_tactic pg' gls) in
match pg with
  DOP2(Lambda,_,_) -> 
    (try deroule_intro pg gls
     with UserError _ -> inject_bad_program pg gls)
| DOPN(MutCase,args) -> 
    let P = args.(0) and c = args.(1) 
    and lf = list_of_vect(sub_vect args 2 ((vect_length args) - 2)) in
    (try program_case P lf c gls
    with NotProgram -> error"Insufficient specification of the program")
| DOPN(Fix(vn,i),cl)  -> apply_fixpoint vn i cl pg gls
| DOPN(CoFix(_),cl) -> apply_cofixpoint cl pg gls
| DOP2(Cast,t,_) -> program_tactic t gls
| _ -> go_program_app_ext pg gls ;;

let program gls = 
  let pg = extract_pg gls in
  ((Context (set_pgm NONE (pf_ctxt gls))) THEN program_tactic pg) gls ;;

let program_tac = overwriting_register_atomic_tactic "Program" program ;;

let program_expand gls = 
  let pg = extract_pg gls in
  ((Context (set_pgm NONE (pf_ctxt gls))) THEN 
    Realizer (strong whd_beta (pf_one_step_reduce gls pg))) gls ;;

let program_expand_tac = overwriting_register_atomic_tactic "Program_Expand" program_expand
 ;;

let program_all gls = 
  ((REPEAT program_tac) THEN (TRY (auto__auto NONE))) gls ;;

let program_all_tac = overwriting_register_atomic_tactic "Program_all" program_all ;;

let show_nth_program n =
    let pf = proof_of_pftreestate (get_pftreestate ()) in
    let gllist = fst(frontier pf) in
       try pr_ctxt (nth gllist n).ctxt
       with Failure _ -> [< 'S "No such program" >] ;;

#open"vernacinterp" ;;

vinterp_add("ShowProgram",
    fun [] -> (fun () -> PPNL(show_nth_program 1))
      | [VARG_NUMBER n] -> (fun () -> PPNL(show_nth_program n)));;


(* $Id: program.ml,v 1.14 1996/03/01 11:48:41 cpaulin Exp $ *)
