(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                record.ml                                 *)
(****************************************************************************)

#infix "o";;
#open "std";;
#open "initial";;
#open "names";;
#open "system";;
#open "term";;
#open "mach";;
#open "discharge";;
#open "tacmach";;
#open "pfedit";;
#open "command";;
#open "pretty";;
#open "lexer";;
#open "vkey";;
#open "parsing";;
#open "himsg";;
#open "more_util";;
#open "pp";;
#open "stdpp";;
#open "proof";;
#open "proof_trees";;
#open "tactics";;
#open "library";;
#open "libobject";;
#open "environ";;
#open "termenv";;
#open "states";;
#open "trad";;
#open "ast";;
#open "vernacinterp";;
#open "vernacentries";;
#open "termast";;

let mkAppLC(f,l) = OP(("APPLIST",[]),f::l);;
let mkLambdaC(x,A,B) = OP(("LAMBDA",[]),[A;SLAM(SOME x,B)]);;
let mkLambdaCit = list_it (fun (x,A) B -> mkLambdaC(x,A,B));;

let make_constructor fields appc =
 let rec aux fields =
  match fields with
     [] -> appc
   | (id,ty)::l -> ASTnode("PROD",[ty; ASTslam ( SOME id, aux l)]) 
 in aux fields;;

(* free_vars existe de'ja` dans ast.ml *)

let rec drop = function
    [] -> []
  | (NONE::t) -> drop t
  | ((SOME id)::t) -> id::(drop t)
;;

let fold curriedf l base = list_it (fun a -> fun b -> curriedf(a,b)) l base
;;

let free_vars t = aux(t,[]) where rec aux = function
  (ASTnvar id,accum) -> listset__add accum id
| (ASTnode(_,tl),accum) -> fold aux tl accum
| (ASTslam(ido,body),accum) ->
        (listset__diff (aux(body,[])) (drop [ido]))@accum
| (ASTprim _,accum) -> accum
;;

let free_in_asts id l = 
  let rec aux = function [] -> true
                       | a::l -> (not (mem id (free_vars a))) & (aux l) in
  aux l;;
 
let all_vars t = aux(t,[]) where rec aux = function
  (ASTnvar id,accum) -> listset__add accum id
| (ASTnode(_,tl),accum) -> fold aux tl accum
| (ASTslam(SOME id,body),accum) -> aux (body,(listset__add accum id))
| (ASTslam(NONE,body),accum) -> aux (body,accum)
| (ASTprim _,accum) -> accum
;;
        
let repl_in_ast ids pars T =
    aux T where rec aux = function
    (ASTnvar id) -> if mem id ids then mkAppLC(ASTnvar id,pars)
                                  else (ASTnvar id)
  | (ASTnode(opn,tl)) -> ASTnode(opn,map aux tl)
  | (ASTslam(ido,body) as t) -> ASTslam(ido,aux body)
  | (ASTprim _ as ast) -> ast
;;

let print_id_list l = [< 'S "[" ; prlist (fun id -> [< 'S (string_of_id id) >]) l; 'S "]" >];; 

let definition_record namerec params s name_constructor fields =
 let idparams = list__map fst params in      
 let typarams = list__map snd params in
 let namefields = list__map fst fields in
 let tyfields = list__map snd fields in
 let app_constructor = ASTnode("APPLIST",(ASTnvar namerec)::
                               list__map (fun id -> ASTnvar id) idparams) in
 let type_constructor = make_constructor fields app_constructor in 
 if not (free_in_asts namerec tyfields)
    then message "Error: A record cannot be recursive"
    else (build_mutual params [(true,namerec,s,[(name_constructor,type_constructor)])];
          let x = next_ident_away (id_of_string "x") (it_list (fun l ty -> listset__union (all_vars ty) l) 
                                                             idparams tyfields) in
          it_list (fun (ids_ok,ids_not_ok) (id,ty) -> 
                       let fv_ty = free_vars ty in
                       let bad_projs = (intersect ids_not_ok fv_ty) in
                       if not (listset__is_empty bad_projs)
                         then (PPNL [< 'S"Warning: "; 'S(string_of_id id); 
                                       'S" cannot be defined. The projections ";
                                      print_id_list bad_projs; 'S " were not defined" >];
                               (ids_ok,id::ids_not_ok))
                         else let typres = if listset__is_empty (intersect ids_ok fv_ty)
                                then ty
                                else let newty = repl_in_ast ids_ok 
                                                 (map (fun id -> ASTnvar id) (idparams@[x])) ty in
                               mkLambdaC(x,app_constructor,newty) in
                        let branch = mkLambdaCit fields (ASTnvar id) in         
                        (try (definition_body id NeverDischarge 
                              (mkLambdaCit params (mkLambdaC (x,app_constructor,
                                 ASTnode ("MUTCASE",[typres;ASTnvar(x);branch]))));
                              message ((string_of_id id) ^ " is defined");
                              (id::ids_ok,ids_not_ok))
                         with UserError(s,pps) -> 
                      (PPNL [<'S"Warning: ";'S (string_of_id id); 'S" cannot be defined. "; pps >];
                                    (ids_ok,id::ids_not_ok)))) ([],[]) fields;());;
         
add("RECORD1",
    fun [VARG_IDENTIFIER namerec;
         VARG_BINDERLIST binders; 
         VARG_COMMAND s; 
         VARG_IDENTIFIER name_constructor; 
         VARG_BINDERLIST fs] ->
        let params = join_binders binders in
        let fields = list__map (fun ([id],c) -> (id,c)) fs in  
        fun () -> definition_record namerec params s name_constructor fields);;

add("RECORD",
    fun [VARG_IDENTIFIER namerec;
         VARG_BINDERLIST binders; 
         VARG_COMMAND s; 
         VARG_BINDERLIST fs] ->
        let params = join_binders binders in
        let fields = list__map (fun ([id],c) -> (id,c)) fs in  
        let name_constructor = id_of_string("Build_" ^ (string_of_id namerec)) in
        fun () -> definition_record namerec params s name_constructor fields);;


(* $Id: record.ml,v 1.3 1996/02/12 15:57:05 cpaulin Exp $ *)
