(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                tauto.mli                                 *)
(****************************************************************************)

(* Mars 1993 *)
(* Autor: Cesar A. Munnoz H *)

#open "tacmach";;
#open "term";;

(* Prototipo *)
(* Estructuras de Datos *)
type Formula =
       FVar  of string 
     | FAnd  of Formula*Formula
     | FOr   of Formula*Formula
     | FImp  of Formula*Formula
     | FEqu  of Formula*Formula
     | FNot  of Formula
     | FEq   of Formula*Formula*Formula
     | FPred of constr (* Predicado proposicional *)
     | FFalse
     | FTrue
     (* La siguiente no puede aparecer en una formula de entrada *)
     (* Representa una formula atomica cuando aparece en un principal de una
	regla *)
     | FLis of Formula list (* Lista de formulas *)
     | FAto of string       (* Formula atomica *)
     | FLisfor of string;;  (* Variable para una lista de formulas *)
	                    (* En el antecedente se llama GAMA, 
			       en el sucedente DELTA *)

(* Terminos en calculo lambda *)
type Termino =
       TVar of string
     | TApl of Formula*Formula*Termino*Termino
     | TFun of string*Formula*Termino
     | TPar of Formula*Formula*Termino*Termino 
     | TInl of Formula*Formula*Termino 
     | TInr of Formula*Formula*Termino
     | TFst of Formula*Formula*Termino
     | TSnd of Formula*Formula*Termino 
     | TCase of Formula list * Termino list 
     | TZ of Formula * Termino 
     | TExi of string
     | TRefl of Formula * Formula (*Reflexividad de la igualdad *) 
     | TSim  of Formula * Formula * Formula * Termino
                                  (*Simetria de la igualdad *)
     | TTrue
    (* Los siguientes terminos se usan solamente en las sustituciones *)
     | TSum of Termino*Termino (* Suma de terminos *)
     | TLis of Termino list  (* Lista de terminos *) 
     | TLister of string     (* Variable para una lista de terminos *)
			     (* En el antecendete se llama Gama, 
			        en el sucedente Delta *)
     | TZero of Formula;;    (* Milagro               *)

(* Es una formula asociada con un termino del calculo lambda, o los 
   multiconjuntos Gama y Delta *)
type FormulaT == Termino*Formula;;

(* La primera componente es el antecedente, la segunda es sucedente *)
type Sequente == FormulaT list * FormulaT list;;

(* Substitucion de variable por una formula *)
type SubsF == (string*Formula) list;;

(* Substitucion de variable por un lambda termino *)
type SubsT == (string*Termino) list;;

type Regla = 
     { Tip: string;  (* Tipo de la formula principal *)
       Heu: bool;    (* Si es una regla heuristica o no *)
       Ant: bool;    (* Si principal es antecedente o sucedente *)
       Pri: FormulaT;(* Formula principal de la regla *)
       Pre: Sequente list; (* Premisas de la regla *)
       Res: Sequente;(* Restricciones para la aplicacion de una regla*)
       Def: SubsT;   (* Definicion de los terminos del lado derecho *)
       Sub: SubsT;   (* Substitucion que se aplica al lado derecho de la
			conclusion para obetener el lambda termino *)
       Ren: string list; (* Variables que se deben renombrar *)
       Vardelta:bool;    (* Si se usa la variable proposicional DELTA *) 
       Ssi:bool;         (* Si la regla es reversible o no *)
       Rendelta: string list};; (* Renombramientos de delta *)
       (* Note que si Res = A |- B, entonces la conclusion de la regla es
          A,Gama,Pri' |- B, Pri'',Delta 
	  Si ant = true Pri'= Pri
	  Si ant =false Pri''=Pri*)

(* Substitucion Formula Termino para aplicar una regla *)
type SFT    =
     { SReg:Regla ref; (*Apuntador a la regla *)
       SFor:SubsF;     (*Substitucion de Formulas *)
       SGam:FormulaT list;   (* Lista de formulas de Gamma *)
       SDel:FormulaT list;   (* Lista de formulas de Delta *)    
       SRen:(string*SubsT) list; (* Renombramientos de variables *)
       STer:SubsT;   (* Susbstitucion de terminos  *)
       SDef:SubsT};; (* Definicion de terminos     *) 

type SubsFT = SNil | SCons of SFT;;

type ReglaSub = RNil | RCons of (SFT*Regla list*FormulaT list*Sequente);;

(* De un arbol de demostracion *)
type Nodo = 
     { Seq: Sequente ref; (* Sequente que se resuelve *)
       Reg: Regla ref;    (* Regla usada para resolver *)
       Sd:  SubsT;        (* Substitucion que define los lambda terminos *)
       St:  SubsT};;      (* Substitucion que calcula el lambda termino  *)

(* Arbol de demostracion *)  
type Arbol = Nil | Cons of Arbol * Nodo * Arbol;;

(* Demostracion *)
(* Si el secuente es valido Arb es un arbol de demostracion y Lisbut
   es vacio, sino Lisbut es un contexto en el cual Arb es valido *)
type Demostracion = {Arb:Arbol;Lisbut:FormulaT list};;

(* Definicion de excepcion para rescribir terminos *)
exception NoAplica;;
exception TacticFailure;; 

value Tauto_tac : tactic;;
value Intuition:tactic;;
value Intuition_tac:tactic;;
value Tauto:tactic;;


(* $Id: tauto.mli,v 1.4 1996/02/12 15:57:10 cpaulin Exp $ *)
