(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                tauto2.ml                                 *)
(****************************************************************************)

#infix "o";;
#open "std";;
#open "initial";;
#open "names";;
#open "generic";;
#open "term";;
#open "more_util";;
#open "pp";;
#open "stdpp";;
#open "tacmach";;
#open "proof_trees";;
#open "clenv";;
#open "tactics";;
#open "sopattern";;
#open "tactics1";;
#open "tactics2";;

#infix "ORELSE";;
#infix "NEXT";;
#infix "THEN";;
#infix "THENF";;
#infix "THENL";;
#infix "THENS";;

let hlset_subset hls1 hls2 =
    for_all (fun e -> exists (fun e' -> eq_constr (e,e')) hls2) hls1
;;

let hlset_eq hls1 hls2 =
    hlset_subset hls1 hls2 & hlset_subset hls2 hls1
;;  

let eq_gls g1 g2 =
    eq_constr (pf_concl g1,pf_concl g2)
 &  (let hl1 = snd(pf_hyps g1) and
         hl2 = snd(pf_hyps g2)
     in hlset_eq hl1 hl2)
;;

let gls_memb BTS g = exists (eq_gls g) BTS;;

let gls_add g BTS =
    if gls_memb BTS g then error "backtrack in tauto"
    else g::BTS
;;

let Classically cltac = function
    (SOME _ as cls) -> cltac cls THEN ThinClause cls
  | NONE -> cltac NONE
;;

let somatch M pat = somatch__somatch NONE (get_pat pat) M;;

let module_mark = ["#Logic.obj"];;
let pi_pattern = put_pat mmk "(x:?)((?)@[x])";;
let and_pattern = put_pat mmk "(and ?1 ?2)";;
let or_pattern = put_pat mmk "(or ?1 ?2)";;
let not_pattern = put_pat mmk "(not ?1)";;
let imply_pattern = put_pat mmk "?1->?2";;
let mkIMP A B = soinstance imply_pattern [A;B];;

let is_atomic M =
    not(exists (somatches M) [pi_pattern;and_pattern;or_pattern;not_pattern])
;;
let Hypothesis (SOME id) = Exact (VAR id);;
let Assumption gls = ApplyToAHyp Hypothesis gls;;

(* Steps of the procedure *)

(* 1. A,Gamma |- A *)
let dyck_hypothesis =
    register_ident_tactic "dyck_hypothesis" (Hypothesis o inSOME);;

(* 2. False,Gamma |- G *)
let dyck_absurdity_elim =
    register_ident_tactic "dyck_absurdity_elim" Absurdity;;

(*3. A,B,Gamma |- G
  ---------------
  A/\B,Gamma |- G
 *)
let dyck_and_elim =
    register_ident_tactic "dyck_and_elim" ((Classically DAnd) o inSOME);;

(*4. Gamma |- A  Gamma |- B
  -----------------------
  Gamma |- A /\ B
 *)
let dyck_and_intro =
    register_atomic_tactic "dyck_and_intro" (DAnd NONE);;


(*5. A,Gamma |- G    B,Gamma|- G
  ---------------------------
  A\/B,Gamma |- G
 *)
let dyck_or_elim =
    register_ident_tactic "dyck_or_elim" ((Classically (DOr 0)) o inSOME);;

(*6. Gamma |- A
  ----------
  Gamma |- A\/B
 *)
let dyck_or_introleft =
    register_atomic_tactic "dyck_or_introleft" (DOr 1 NONE)
;;

(*7. Gamma |-B
  ---------
  Gamma |- A\/B
 *)
let dyck_or_introright =
    register_atomic_tactic "dyck_or_introright" (DOr 2 NONE)
;;

(*8. A,Gamma |- B
  --------------
  Gamma |- A -> B
 *)
let dyck_imply_intro =
    register_atomic_tactic "dyck_imply_intro" (DImp NONE)
;;

(*9.
    B,A,Gamma |- G
    --------------
    A->B,A,Gamma |- G  (A Atomique)
 *)
let atomic_imply_bot_pattern = put_pat mmk "?1->?2";;
let atomic_imply_step cls gls =
let mvb = somatch (clause_type cls gls) atomic_imply_bot_pattern in

if not(is_atomic (listmap__map mvb 1)) then
    error "atomic_imply_step";

(DImp cls THENS [ThinClause cls;ApplyToAHyp Hypothesis]) gls
;;

let dyck_atomic_imply_elim =
    register_ident_tactic "dyck_atomic_imply_elim" 
    (atomic_imply_step o inSOME)
;;


(*10.
    C ->(D-> B),Gamma |- G
    -----------------------
    (C/\D)->B,Gamma |- G
 *)

let and_imply_bot_pattern = put_pat mmk "(and ?1 ?2)->?3";;
let and_imply_step cls gls =
let mvb = somatch (clause_type cls gls) and_imply_bot_pattern in
let C = listmap__map mvb 1 and
    D = listmap__map mvb 2 and
    B = listmap__map mvb 3 in

(Cut (mkIMP C (mkIMP  D B))
 THENS [ThinClause cls ;
        (RepeatFor 2 Intro) THEN DImp cls THENS
        [Assumption;
         DAnd NONE THEN Assumption]]) gls
;;

let dyck_and_imply_elim =
    register_ident_tactic "dyck_and_imply_elim" (and_imply_step o inSOME);;

(*11.
    C->B,D->B,Gamma |-G
    --------------------
    (C\/D)->B,Gamma |- G
*)

let or_imply_bot_pattern = put_pat mmk "(or ?1 ?2)->?3";;
let or_imply_step cls gls =
let mvb = somatch (clause_type cls gls) or_imply_bot_pattern in
let C = listmap__map mvb 1 and
    D = listmap__map mvb 2 and
    B = listmap__map mvb 3 in

(ParallelCut [mkIMP C B;mkIMP D B]
 THENS [ThinClause cls;
        Intro THEN DImp cls THENS [Assumption ; DOr 1 NONE THEN Assumption];
        Intro THEN DImp cls THENS [Assumption ; DOr 2 NONE THEN Assumption]]) gls
;;

let dyck_or_imply_elim =
    register_ident_tactic "dyck_or_imply_elim" (or_imply_step o inSOME);;

(*12.
B,Gamma|- G   D->B,Gamma |- C->D
----------------------------------
(C->D)->B,Gamma |- G
*)

let back_thru_2 id =
    applist(VAR id,[DOP0(Meta(newMETA()));DOP0(Meta(newMETA()))])
;;

let back_thru_1 id =
    applist(VAR id,[DOP0(Meta(newMETA()))])
;;

let imply_imply_bot_pattern = put_pat mmk "(?1->?2)->?3";;
let imply_imply_step cls gls =
let (SOME H0) = cls in (* (C->D)->B *)
let mvb = somatch (clause_type cls gls) imply_imply_bot_pattern in
let C = listmap__map mvb 1 and
    D = listmap__map mvb 2 and
    B = listmap__map mvb 3 in

(Cut B THENS
 [ThinClause cls; (* B |- G *)
  Cut (mkIMP (mkIMP D B) (mkIMP C D)) THENS
  [OnLastHyp (fun (SOME H1) (*(D->B)->(C->D)*) ->
   DImp cls THENS [Assumption ; IDTAC] THEN Intro THEN
   OnLastHyp (fun (SOME H2) (* C *) ->
   Refine (back_thru_2 H1) THENS
   [Intro THEN Refine (back_thru_1 H0) THENS
    [Intro THEN Assumption];
    Assumption]));
   ThinClause cls THEN Intro]]) gls
;;              

let dyck_imply_imply_elim =
    register_ident_tactic "dyck_imply_imply_elim" 
    (imply_imply_step o inSOME);;

let rec Prove g =
    COMPLETE
    ((ApplyToAHyp (ClauseTacThen (dyck_hypothesis o outSOME) Prove)) 
     ORELSE
     (ApplyToAHyp (ClauseTacThen (dyck_absurdity_elim o outSOME) Prove)) 
     ORELSE
     (ApplyToAHyp (ClauseTacThen (dyck_and_elim o outSOME) Prove)) 
     ORELSE
     (ApplyToAHyp (ClauseTacThen (dyck_or_elim o outSOME) Prove)) 
     ORELSE
     (ApplyToAHyp (ClauseTacThen (dyck_atomic_imply_elim o outSOME) Prove)) 
     ORELSE
     (ApplyToAHyp (ClauseTacThen (dyck_and_imply_elim o outSOME) Prove)) 
     ORELSE
     (ApplyToAHyp (ClauseTacThen (dyck_or_imply_elim o outSOME) Prove)) 
     ORELSE
     (ApplyToAHyp (ClauseTacThen (dyck_imply_imply_elim o outSOME) Prove)) 
     ORELSE
     (dyck_and_intro THEN Prove) 
     ORELSE
     (dyck_or_introleft THEN Prove) 
     ORELSE
     (dyck_or_introright THEN Prove) 
     ORELSE
     (dyck_imply_intro THEN Prove)) g
;;

let iffnot_ids = [id_of_string "not";id_of_string"iff"];;

let Tauto2 gls =
    (OnAllClausesLR (Unfolds [id_of_string "not";id_of_string"iff"]) 
      THEN Prove) gls;;

let Tauto2_tac = register_atomic_tactic "Tauto2" Tauto2;;


(* $Id: tauto2.ml,v 1.4 1996/02/12 15:57:12 cpaulin Exp $ *)
