(*****************************************************************************)
(*          Projet Coq - Calculus of Inductive Constructions V5.10           *)
(*****************************************************************************)
(*                                                                           *)
(*          Categories                                                       *)
(*                                                                           *)
(*          Peter Aczel, Dec. 92                                             *)
(*          rev. Gerard Huet,  Aug. 1993                                     *)
(*          rev. Amokrane Saibi,  May 1994                                   *)
(*                                                                           *)
(*****************************************************************************)

Require Export Map.

(*  The type of categories *)

Section cat.

Variable Ob  : Type. (* Objects *)
Variable Hom : Ob->Ob->Setoid. (* Hom Setoid *)

Variable Op_comp : (a,b,c:Ob)|((Hom a b) => ((Hom b c) => (Hom a c)))|.

Definition Cat_comp := [a,b,c:Ob](Ap2 (Op_comp a b c)).

Syntax constr Cat_comp <<(Cat_comp $1 $2 $3 $f $g)>> 10
		[<hov 0> <$f:"Term":L> " o " <$g:"Term":L>].
Grammar command command2 := [ command1($f) " o " command2($g) ] -> 
                            [$0 = <<(Cat_comp ? ? ? $f $g)>>].

Definition assoc_law := (a,b,c,d:Ob)(f:|(Hom a b)|)(g:|(Hom b c)|)(h:|(Hom c d)|)
                        (f o (g o h)) =%S ((f o g) o h).

Variable Id : (a:Ob)|(Hom a a)|.

Definition idl_law := (a,b:Ob)(f:|(Hom a b)|)
                      ((Id a) o f) =%S f.

Definition idr_law := (a,b:Ob)(f:|(Hom b a)|) 
                      f =%S (f o (Id ?)).

End cat.


Syntactic Definition Assoc_law := assoc_law | 2.
Syntax constr assoc_law <<(assoc_law $1 $2 $o)>> 10
	     [<hov 0> "Assoc_law " <$o:"term":L>].

Syntactic Definition Idl_law := idl_law | 2.
Syntax constr idl_law <<(idl_law $1 $2 $o $i)>> 10
	     [<hov 0> "Idl_law " <$o:"term":L> " " <$i:"term":L>].

Syntactic Definition Idr_law := idr_law | 2.
Syntax constr idr_law <<(idr_law $1 $2 $o $i)>> 10
	     [<hov 0> "Idr_law " <$o:"term":L> " " <$i:"term":L>].

Structure Category : Type := 
 {Ob      : Type;
  hom     : Ob -> Ob -> Setoid;
  Op_comp : (a,b,c:Ob)|((hom a b) => ((hom b c) => (hom a c)))|;
  id      : (a:Ob)|(hom a a)|;
  Prf_ass : (Assoc_law Op_comp);
  Prf_idl : (Idl_law Op_comp id);
  Prf_idr : (Idr_law Op_comp id)}.

(* Projections *)

Syntactic Definition Hom := hom | 1.
Syntax constr Hom <<(hom $1 $a $b)>> 10
	     [<hov 0> "Hom " <$a:"term":L> " " <$b:"term":L>].

Syntactic Definition Id := id | 1.
Syntax constr Id <<(id $1 $a)>> 10
		[<hov 0> "Id " <$a:"Term":L>].

Definition Comp := [C:Category][a,b,c:(Ob C)]
                   (Cat_comp (Ob C) (hom C) (Op_comp C) a b c).

Syntax constr Comp <<(Comp $1 $2 $3 $4 $f $g)>> 10
		[<hov 0> <$f:"Term":L> " o " <$g:"Term":L>].

Grammar command command2 := [ command1($f) " o " command2($g) ] -> 
                            [$0 = <<(Comp ? ? ? ? $f $g)>>].

(* construire un ope'rateur de composition a` partir d'une
   fonction de composition *)

Section composition_to_operator.

Variable Ob       : Type.
Variable Hom      : Ob->Ob->Setoid.
Variable Comp_fun : (a,b,c:Ob)|(Hom a b)|->|(Hom b c)|->|(Hom a c)|.

Definition Congl_law := (a,b,c:Ob)(f,g:|(Hom b c)|)(h:|(Hom a b)|)
           f =%S g -> (Comp_fun a b c h f) =%S (Comp_fun a b c h g). 

Definition Congr_law := (a,b,c:Ob)(f,g:|(Hom a b)|)(h:|(Hom b c)|)
           f =%S g -> (Comp_fun a b c f h) =%S (Comp_fun a b c g h). 

Definition Cong_law := (a,b,c:Ob)(f,f':|(Hom a b)|)(g,g':|(Hom b c)|)
           f =%S f' -> g =%S g' -> (Comp_fun a b c f g) =%S (Comp_fun a b c f' g'). 

Hypothesis pcgl : Congl_law.
Hypothesis pcgr : Congr_law.

Variable a, b, c : Ob.

Lemma Comp1_map_law : (f:|(Hom a b)|)(Map_law (Comp_fun a b c f)).
Proof.
Unfold map_law; Auto.
Qed.

Definition Comp1_map := [f:|(Hom a b)|]
      (Build_Map (Comp_fun a b c f) (Comp1_map_law f)).

Lemma Comp_map_law : (map_law (Hom a b) (Hom b c)=>(Hom a c) Comp1_map).
Proof.
Unfold map_law; Simpl.
Unfold Ext Comp1_map; Simpl.
Auto.
Qed.

Definition Build_Comp := (build_Map (Hom a b) (Hom b c)=>(Hom a c) Comp1_map Comp_map_law).

End composition_to_operator.


Section cat_cong.

Variable C : Category.

Lemma Prf_congl : (Congl_law ? ? (Comp C)).
Unfold Congl_law; Intros a b c f g h H0.
Unfold Comp Cat_comp ap2.
Apply (Pres ? ? (Ap (Op_comp C a b c) h)).
Auto.
Qed.

Lemma Prf_congr : (Congr_law ? ? (Comp C)). 
Unfold Congr_law; Intros a b c f g h H0.
Unfold Comp Cat_comp ap2.
Apply (Pres ? ? (Op_comp C a b c) f g H0).
Qed.

Lemma Prf_cong : (Cong_law ? ? (Comp C)).
Unfold Cong_law; Intros a b c f f' g g' H0 H1.
(* *) Apply (Prf_trans (Hom a c)) with y:= (f' o g).
Apply Prf_congr; Assumption.
Apply Prf_congl; Assumption.
Qed.

End cat_cong.


(* $Id: Category.v,v 1.2 1996/02/12 15:23:44 cpaulin Exp $ *)
