(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                 Logic.v                                  *)
(****************************************************************************)

Require PreludeSyntax.

Inductive True : Prop := I : True. 

(* Absurdity *)

Inductive False : Prop := . 

(* Negation *)

Definition not := [A:Prop] A->False.

Syntax constr not  <<(not $t1)>>  2 
       [<hov 0>  "~" <$t1:"term":E> ].

(* Conjunction *)

(* Pairing / Introduction *)

Inductive and [A,B:Prop] : Prop := conj : A -> B -> A/\B.

Section Projections.

Variables A,B : Prop.

Theorem proj1 : A/\B -> A.
Proof.
Induction 1; Trivial.
Qed.
    
Theorem proj2 : A/\B -> B.
Proof.
Induction 1; Trivial.
Qed.

End Projections.

Syntax constr and  <<(and  $t1  $t2)>>  6 
       [<hov 0>  <$t1:"term":L> [0 0] "/\\" <$t2:"term":E>].

Syntax constr conj <<(conj $t1 $t2 $t3 $t4)>>  1
       [<hov 0> "<" <$t1:"term":L> "," <$t2:"term":L> ">" 
        "{" <$t3:"term":L> "," <$t4:"term":L> "}"].

(* Disjunction *)

Inductive or [A,B:Prop] : Prop
    := or_introl : A -> A\/B 
     | or_intror : B -> A\/B.

Syntax constr or <<(or  $t1  $t2)>> 7
        [<hov 0>  <$t1:"term":L> [0 0]  "\\/" <$t2:"term":E>].

(* Equivalence *)

Definition iff := [P,Q:Prop] (P->Q) /\ (Q->P).

Syntax constr iff <<(iff  $t1  $t2)>> 8
       [<hov 0>  <$t1:"term":L> [1 0] [<hov 0> "<->" [1 1] <$t2:"term":E>]].

Definition IF := [P,Q,R:Prop] (P/\Q) \/ (~P/\R).

Syntax constr if  <<(IF $c $t $e)>> 1 
      [<hov 0> "if" [1 1] <$c:"term":E>
                [<hov 0> [1 1] "then" [1 1] <$t:"term":E>]
                [<hov 0> [1 1] "else" [1 1] <$e:"term":E>]].


(* First-order quantifiers *)

Definition all := [A:Set][P:A->Prop](x:A)(P x). 

Syntax constr all2 <<(all $_ $t2)>> 10
           [<hov 0> "All "   <$t2:"term":L> ].


Inductive ex [A:Set;P:A->Prop] : Prop 
    := ex_intro : (x:A)(P x)->(ex A P).

Syntax constr ex_imp <<(ex $_  $t1)>> 10
          [<hov 0> "Ex "  <$t1:"term":L> ].


Inductive ex2 [A:Set;P,Q:A->Prop] : Prop
    := ex_intro2 : (x:A)(P x)->(Q x)->(ex2 A P Q).

Syntax constr ex2_imp <<(ex2 $_  $t1 $t2)>> 10
     [<hov 0> "Ex2" [1 1] [<hv 0> <$t1:"term":L> [1 0] <$t2:"term":L> ]].


(* Equality *)

Inductive eq [A:Set;x:A] : A->Prop
       := refl_equal : (eq A x x).


Syntax constr equal <<(eq $_  $t1  $t2)>> 1
           [<hov 0>  <$t1:"term":E> [0 1]  "=" <$t2:"term":E>].

Syntactic Definition All := (all ?).
Syntactic Definition Ex := (ex ?).
Syntactic Definition Ex2 := (ex2 ?).

Hint I conj or_introl or_intror refl_equal.

Section Logic_lemmas.

Theorem absurd : (A:Prop)(C:Prop) A -> ~A -> C.
Proof.
Unfold not; Intros A C h1 h2.
Elim (h2 h1).
Qed.

Section equality.
 Variable A,B : Set.
 Variable f   : A->B.
 Variable x,y,z : A.

 Theorem sym_equal : x=y -> y=x.
 Proof.
 Induction 1; Trivial.
 Qed.

 Theorem trans_equal : x=y -> y=z -> x=z.
 Proof.
 Induction 2; Trivial.
 Qed.

 Theorem f_equal : x=y -> (f x)=(f y).
 Proof.
 Induction 1; Trivial.
 Qed.

 Theorem sym_not_equal : ~(x=y) -> ~(y=x).
 Proof.
 Red; Intros h1 h2; Apply h1; Elim h2; Trivial.
 Qed.

End equality.

Definition eq_ind_r : (A:Set)(x:A)(P:A->Prop)(P x)-> (y:A) y=x -> (P y).
Intros A x P H y H0; Case sym_equal with 1:= H0; Trivial.
Defined.

Definition eq_rec_r : (A:Set)(x:A)(P:A->Set)(P x)->(y:A)y=x->(P y).
Intros A x P H y H0; Case sym_equal with 1:= H0; Trivial.
Defined.


End Logic_lemmas.

Immediate sym_equal sym_not_equal.


(* $Id: Logic.v,v 1.3 1996/02/12 15:28:29 cpaulin Exp $ *)
