(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                               TheoryList.v                               *)
(****************************************************************************)

(* Some programs and results about lists following CAML Manual *)

Chapter Lists.

Variable  A:Set.

Inductive list : Set := nil : list | cons : A -> list -> list.

(*********************)
(* The null function *)
(*********************)

Definition Isnil : list -> Prop := [l:list]nil=l.

Lemma Isnil_nil : (Isnil nil).
Red; Auto.
Save.

Lemma not_Isnil_cons : (a:A)(l:list)~(Isnil (cons a l)).
Unfold Isnil.
Intros; Discriminate.
Save.

Hint Isnil_nil not_Isnil_cons.

Lemma Null : (l:list){(Isnil l)}+{~(Isnil l)}.
Realizer [l:list]<bool>Case l of
                      (* nil *) true
                 (* cons a m *) [a,m]false
		 end.
Program_all.
Save.

(***********************)
(* The Uncons function *)
(***********************)

Lemma Uncons : (l:list){a : A & { m: list | (cons a m)=l}}+{Isnil l}.
Realizer [l:list]<(Exc A*list)>Case l of 
                      (* nil *) Error
                 (* cons a m *) [a,m](Value (a,m))
		 end.
Program_all.
Save.

(********************************)
(* The head function            *)
(********************************)

Lemma Hd : (l:list){a : A | Ex([m: list](cons a m)=l)}+{Isnil l}.
Realizer [l:list]<(Exc A)>Case l of
                      (* nil *) Error
                 (* cons a m *) [a,m](Value a)
		 end.
Program_all.
Exists m; Auto.
Save.

Lemma Tl : (l:list){m:list|(Ex [a: A] (cons a m)=l)
                         \/ ((Isnil l) /\ (Isnil m)) }.
Realizer [l:list]<list>Case l of
                      (* nil *) nil
                 (* cons a m *) [a,m]m
		 end.
Program_all.
Left; Exists a; Auto.
Save.

(****************************************)
(* Length of lists                      *)
(****************************************)

Fixpoint length [l:list] : nat 
   := (<nat>Case l of (* nil *) O
                 (* cons a m *) [a,m](S (length m)) end).

Fixpoint Length_l [l:list] : nat -> nat 
  :=  [n:nat]<nat>Case l of
                      (* nil *) n
                 (* cons a m *) [a,m](Length_l m (S n))
		 end.

(* A tail recursive version *)
Lemma Length_l_pf : (l:list)(n:nat){m:nat|(plus n (length l))=m}.
Realizer Length_l.
Program_all.
Elim e; Simpl; Auto.
Save.

Lemma Length : (l:list){m:nat|(length l)=m}.
Realizer [l:list](Length_l_pf l O).
Program_all.
Save.


(*****************************)
(* The append function       *)
(*****************************)

Fixpoint app [l:list] : list -> list 
   := [m:list]<list>Case l of (* nil *) m 
                         (* cons a m *) [a,l1](cons a (app l1 m)) end.


Lemma app_nil_end : (l:list)l=(app l nil).
Intro l ; Elim l ; Simpl ; Auto.
Intros a y h ; Elim h ; Auto.
Save.
Hint app_nil_end.

Lemma app_assoc_l : (l,m,n : list)(app (app l m) n)=(app l (app m n)).
Intros l m n ; Elim l ; Simpl ; Auto.
Intros a y h ; Elim h ; Auto.
Save.
Hint app_assoc_l.

(*******************************)
(* Members of lists            *)
(*******************************)

Inductive In [a:A] : list -> Prop 
   := in_hd : (l:list)(In a (cons a l))
   | in_tl : (b:A)(l:list)(In a l)->(In a (cons b l)).
Hint in_hd in_tl.

Lemma In_hd_eq : (a,b:A)(l:list)(a=b)->(In a (cons b l)).
Induction 1; Auto.
Save.
Immediate In_hd_eq.

Inductive AllS [P:A->Prop] : list -> Prop 
   := allS_nil : (AllS P nil)
   | allS_cons : (a:A)(l:list)(P a)->(AllS P l)->(AllS P (cons a l)).
Hint allS_nil allS_cons.

Axiom eqA_dec : (a,b:A){a=b}+{~a=b}.

Lemma mem : (a:A)(l:list){(In a l)}+{(AllS [b:A]~b=a l)}.
Realizer [a:A][l:list]<bool>Match l with 
		(* nil *) false
	   (* cons b m *) [b,m,H]<bool>if (eqA_dec a b) then true
                         	         else H
		      end.
Program_all.
Save.

(**********************************)
(* Index of elements              *)
(**********************************)

Require Le.
Require Lt.

Inductive nth_spec : list->nat->A->Prop :=
  nth_spec_O : (a:A)(l:list)(nth_spec (cons a l) (S O) a)
| nth_spec_S : (n:nat)(a,b:A)(l:list)
           (nth_spec l n a)->(nth_spec (cons b l) (S n) a).
Hint nth_spec_O nth_spec_S.

Inductive fst_nth_spec : list->nat->A->Prop :=
  fst_nth_O : (a:A)(l:list)(fst_nth_spec (cons a l) (S O) a)
| fst_nth_S : (n:nat)(a,b:A)(l:list)(~a=b)->
           (fst_nth_spec l n a)->(fst_nth_spec (cons b l) (S n) a).
Hint fst_nth_O fst_nth_S.

Lemma fst_nth_nth : (l:list)(n:nat)(a:A)(fst_nth_spec l n a)->(nth_spec l n a).
Induction 1; Auto.
Save.
Immediate fst_nth_nth.

Lemma nth_lt_O : (l:list)(n:nat)(a:A)(nth_spec l n a)->(lt O n).
Induction 1; Auto.
Save.

Lemma nth_le_length : (l:list)(n:nat)(a:A)(nth_spec l n a)->(le n (length l)).
Induction 1; Simpl; Auto.
Save.

Fixpoint Nth_func [l:list] : nat -> (Exc A) 
  := [n:nat]<(Exc A)>Case l of 
     (* nil *) (error A)
(* cons a m *) [a,l']
                      <(Exc A)>Case n of
                           (* O *) (error A)
                         (* S m *) [m]<(Exc A)>Case m of
                                	(* O *) (value A a)
                              	      (* S p *) [p](Nth_func l' (S p))
					  end
			 end
	end.

Lemma Nth : (l:list)(n:nat)
            {a:A|(nth_spec l n a)}+{(n=O)\/(lt (length l) n)}.
Realizer Nth_func.
Program_all.
Simpl; Elim n; Auto.
(Elim o; Intro); [Absurd ((S p)=O); Auto | Auto].
Save.

Lemma Item : (l:list)(n:nat){a:A|(nth_spec l (S n) a)}+{(le (length l) n)}.
Realizer [l:list][n:nat](Nth l (S n)).
Program_all.
(Elim o; Intro); [Absurd ((S n)=O); Auto | Auto].
Save.

Require Minus.

Lemma Index_p : (a:A)(l:list)(p:nat)
     {n:nat|(fst_nth_spec l (minus (S n) p) a)}+{(AllS [b:A]~a=b l)}.
Realizer
 [a:A][l:list]
 <nat->(Exc nat)>Match l with 
      (* nil *) [p]Error
 (* cons b m *) [b,m,H,p]<(Exc nat)>if (eqA_dec a b) then (Value p)
                                      else (H (S p))
 end.
Program_all.
Elim e; Elim minus_Sn_m; Trivial; Elim minus_n_n; Auto.
Elim minus_Sn_m; Auto.
Apply lt_le_weak; Apply lt_O_minus_lt; Apply nth_lt_O with l1 a; Auto.
Save.

Lemma Index : (a:A)(l:list)
     {n:nat|(fst_nth_spec l n a)}+{(AllS [b:A]~a=b l)}.
Realizer [a:A][l:list](Index_p a l (S O)).
Program_all.
Rewrite (minus_n_O n0); Auto.
Save.

Section Find_sec.
Variable R,P : A -> Prop.

Inductive InR : list -> Prop 
   := inR_hd : (a:A)(l:list)(R a)->(InR (cons a l))
   | inR_tl : (a:A)(l:list)(InR l)->(InR (cons a l)).
Hint inR_hd inR_tl.

Definition InR_inv := 
       [l:list]<Prop>Case l of 
                       (* nil *) False 
                  (* cons a m *) [b:A][m:list](R b)\/(InR m) 
               end.

Lemma InR_INV : (l:list)(InR l)->(InR_inv l).
Induction 1; Simpl; Auto.
Save.

Lemma InR_cons_inv : (a:A)(l:list)(InR (cons a l))->((R a)\/(InR l)).
Intros a  l; Exact (InR_INV (cons a l)).
Save.

Lemma InR_or_app : (l,m:list)((InR l)\/(InR m))->(InR (app l m)).
Induction 1.
Induction 1; Simpl; Auto.
Intro; Elim l; Simpl; Auto.
Save.

Lemma InR_app_or : (l,m:list)(InR (app l m))->((InR l)\/(InR m)).
Intros l m; Elim l; Simpl; Auto.
Intros b l' Hrec IAc; Elim (InR_cons_inv b (app l' m)); Auto.
Intros; Elim Hrec; Auto.
Save.

Hypothesis RS_dec : (a:A){(R a)}+{(P a)}.

Lemma Find : (l:list){a:A | (In a l) & (R a)}+{(AllS P l)}.
Realizer [l:list]
          <(Exc A)>Match l with 
               (* nil *) Error
          (* cons a m *) [a,m,H]<(Exc A)>if (RS_dec a) then (Value a)
                                           else H
	  end.
Program_all.
Save.

Variable B:Set.
Variable T : A -> B -> Prop.

Variable TS_dec : (a:A){c:B| (T a c)}+{(P a)}.

Lemma Try_find : (l:list){c:B|(Ex2 [a:A](In a l) [a:A](T a c))}+{(AllS P l)}.
Realizer [l:list]
          <(Exc B)>Match l with 
               (* nil *) Error
          (* cons a m *) [a,m,H]<(Exc B)>if (TS_dec a) then
                                             [c](Value c)
                                           else H
	  end.
Program_all.
Exists a; Auto.
Elim e; Intros a1 H1 H2.
Exists a1; Auto.
Save.

End Find_sec.

End Lists.



(* $Id: TheoryList.v,v 1.4 1996/02/12 15:29:02 cpaulin Exp $ *)
