(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                                 Format.v                                 *)
(****************************************************************************)

    
    (**********************************************************************)
    (* A list of words separated by white characters (space or line feed) *)
    (* is given. The program produces a formated list, that is the same   *)
    (* list of words separated by just one line-feed or space in order to *)
    (* put the maximum number of words on the same line.                  *)
    (**********************************************************************)


Chapter Format.

(* Environment *)

(****************************************************************************)
(* The non-white characters (letters) are indexed by a type Ind.            *)
(* The white spaces (separators) are                                        *)
(* indexed by the type bool (space = true and linefeed = false)             *)
(* The data structure is a type of lists with two constructors for cons of  *)
(* letters and cons of separators                                           *)
(****************************************************************************)


 Parameter      Induct :   Set.   
                                        
 Inductive l_ch : Set := nil : l_ch 
                     | consbl : bool->l_ch->l_ch 
                     | consltr : Induct->l_ch->l_ch.

 Definition l_ch_it := 
  [C:Set][x:C][f1:bool->C->C][f2:Induct->C->C]
     Fix F {F/1:l_ch -> C :=
	[l:l_ch]<C>Case l of 
			x
			[b:bool][l0:l_ch](f1 b (F l0))
                     	[i:Induct][l0:l_ch](f2 i (F l0)) end} :
      (C:Set)C -> (bool->C->C) -> (Induct->C->C)->l_ch ->C.

 Definition nat_it := 
  [C:Set][x:C][f:C->C]
     Fix F{F/1:nat->C :=
	[n:nat]<C>Case n of x [m:nat](f (F m)) end} :
      (C:Set)C -> (C -> C) -> nat -> C.

 Local conssp (consbl true). 
 Local conslf (consbl false). 

(* case definition *)

Section Casebl_section.

 Variable  A : Set.                                                    
 Variable c  : l_ch.
 Variable xnil : A.
 Variables xbl,xltr:A->A.
 Definition  Casebl := (l_ch_it A xnil [b:bool]xbl [i:Induct]xltr c) : A.

End Casebl_section.

(* Length function *)

 Definition lgth := [l:l_ch](Casebl nat l O S S) : l_ch->nat.              

(* Append function *)

 Definition app := [l,m:l_ch](l_ch_it l_ch m consbl consltr l) : 
      l_ch->l_ch->l_ch.   

 Local  appsp [l,m:l_ch](app l (conssp m)).                        

 Local  applf [l:l_ch][m:l_ch](app l (conslf m)).          

(* Maximal allowed length of a line *)

 Parameter  max : nat.                                                   

(* Predicate that recognize a word : sequenz (eventually empty) of letters *) 

Inductive word : l_ch -> Prop
    := wordnil : (word nil)
    | wordltr : (j:Induct)(l:l_ch)(word l)->(word (consltr j l)).

(* Predicate that recognize a non-empty sequenz of separators *)

Inductive space  :  l_ch->Prop
     :=  space_bl    : (b:bool)(space (consbl b nil))
       |space_co_bl : (b:bool)(l:l_ch)(space l)->(space (consbl b l)).

(* We give an axiomatisation of predicates inf and sup on integers.     *)
(* We will cut  a line according these predicates.                      *)
(* inf n means  n < max and sup n means n >= max                        *)

 Parameter inf, sup : nat->Prop.                                    
 Axiom ax1     :   (n:nat)({(inf n)}+{(sup n)}).                       

 Local  inflg   [m:l_ch](inf (lgth m)).                            

 Local  suplg   [m:l_ch](sup (lgth m)).                            

 Local plus_l_ch := [m:l_ch][n:nat](nat_it nat n S (lgth m)) : l_ch->nat->nat.

(* The words must be of length less than max. We define a predicate valide. *)
(* A proof of (valide l) gives a decomposition of the list of characters    *)
(* in words of length less than max, separated by spaces                    *)

Inductive valide  :l_ch->Set
    := vword : (m:l_ch)(word m)->(inflg m)->(valide m)
     |vapp  : (m1,m2,p:l_ch)(word m1)->(inflg m1)->(space m2)->(valide p)
            ->(valide (app m1 (app m2 p))).

(* The empty list is valide if 0<max *)

 Hypothesis infmax : (inf O).                                         

(* A list will be non-valid if it contains a too long word *)

Inductive NV : l_ch->Prop
    := NVword :  (l:l_ch)(word l)->(suplg l)->(NV l)
    |NVapp1 :  (l,m:l_ch)(NV l)->(NV (app l m))
    |NVapp2 :  (l,m:l_ch)(NV l)->(NV (app m l)).

(* We show that every list satisfies valide or NV, this is an informative *)
(* proof (a preprocessing of the text).                                    *)

Inductive val_or_no [l:l_ch] : Set
    := inval : (valide l)->(val_or_no l)
     |inNV : (NV l) -> (val_or_no l).

(* To add a space to a valid list is easy *)

Remark valbl :  (b:bool)(l:l_ch)(valide l)->(valide (consbl b l)).
Intros; Apply (vapp nil (consbl b nil) l); Auto.
Apply wordnil.
Apply space_bl.
Save.

(* For adding a letter to a valid list, we need to test whether max *)
(* is overheaded.                                                   *) 

Remark valltr : (i:Induct)(l:l_ch)(valide l)->(val_or_no (consltr i l)).
Intros; Elim H; Intros.
Elim (ax1 (S (lgth m))); Intros.
Apply (inval (consltr i m)).
Apply vword; Auto.
Apply wordltr; Auto.
Apply (inNV (consltr i m)).
Apply NVword; Auto.
Apply wordltr; Auto.
Elim (ax1 (S (lgth m1))); Intros.
Apply (inval (consltr i (app m1 (app m2 p)))).
Apply (vapp (consltr i m1) m2 p) ; Auto.
Apply wordltr; Auto.
Apply (inNV (consltr i (app m1 (app m2 p)))).
Apply (NVapp1 (consltr i m1) (app m2 p)).
Apply NVword; Auto.
Apply wordltr; Auto.
Save.

(* The complete proof is by induction on the structure of the list *)

Lemma preproc : (l:l_ch)(val_or_no l).
Intro; Elim l; Intros.
Apply (inval nil).
Apply vword; Auto.
Apply wordnil.
Elim H; Intros.
Apply (inval (consbl b l0)).
Apply valbl; Auto.
Apply (inNV (consbl b l0)).
Apply (NVapp2 l0 (consbl b nil)); Auto.
Elim H; Intros.
Apply valltr; Auto.
Apply (inNV (consltr i l0)).
Apply (NVapp2 l0 (consltr i nil)); Auto.
Save.

(* We define the relation "two lists are equivalent", the intended meaning *)
(* is that they represent the same list of words.                          *)

Section Equivalence.

Inductive Eq  : l_ch->l_ch->Prop
    := Eq_nil : (Eq nil nil)
    | Eq_bl_nil : (b:bool)(l:l_ch)(Eq l nil)->(Eq (consbl b l) nil)
    | Eq_co_bl  : (b,c:bool)(l,m:l_ch)(Eq l m)->(Eq (consbl b l) (consbl c m))
    | Eq_co_ltr : (i:Induct)(l,m:l_ch)
            (Eq l m)->(Eq (consltr i l) (consltr i m))
    | Eq_bl_bl  : (b,c:bool)(l,m:l_ch)
            (Eq l (consbl b m))->(Eq (consbl c l)(consbl b m))
    | Eq_tran : (l,m,n:l_ch)(Eq l m)->(Eq m n)->(Eq l n).

(* Properties of equivalence *)

(* Reflexivity *)

 Theorem Eq_re :   (l:l_ch)(Eq l l).
 Intro; Elim l; Intros.
 Apply Eq_nil.
 Apply Eq_co_bl; Auto.
 Apply Eq_co_ltr; Auto.
 Save.

(* Eq stability with respect to append *)

 Theorem Eq_app :  (l,m,n:l_ch)(Eq m n)->(Eq (app l m) (app l n)).
 Intros; Elim l ; Intros; Auto.
 Apply (Eq_co_bl b b (app l0 m) (app l0 n)); Auto.
 Apply (Eq_co_ltr i (app l0 m) (app l0 n)); Auto.
 Save.

(* particular case n=nil, directly proved *)

 Theorem Eq_app_nil : (l,m:l_ch)(Eq m nil)->(Eq (app l m) l).
 Intros; Elim l; Intros; Auto.
 Apply (Eq_co_bl b b (app l0 m) l0); Auto.
 Apply (Eq_co_ltr i (app l0 m) l0); Auto.
 Save.

(* Adding more than one separator does not change anything *)
 Theorem Eq_space_bl : (b:bool)(l,m,n:l_ch)(space l)->(Eq m (consbl b n))->(Eq (app l m) (consbl b n)).
 Intros.
 Elim H; Intros.
 Apply (Eq_bl_bl b b0 m n); Auto.
 Apply (Eq_bl_bl b b0 (app l0 m) n); Auto.
 Save.

(* Adding separators at the beginning of the list is equivalent to nil *)
 Theorem Eq_space_nil : (m:l_ch)(space m)->(Eq (app m nil) nil).
 Intros; Elim H; Intros.
 Apply (Eq_bl_nil b nil Eq_nil); Auto.
 Apply (Eq_bl_nil b (app l nil) H1); Auto.
 Save.

(* Adding separators is equivalent to add a space *)
 Theorem Eq_space : (l,m,n:l_ch)(space l)->(Eq m n)->(Eq (app l m) (conssp n)).
 Intros; Elim H; Intros.
 Apply (Eq_co_bl b true m n); Auto.
 Apply (Eq_bl_bl true b (app l0 m) n); Auto.
 Save.

End Equivalence.

(* Definition of the predicate formn : nat->l_ch->Prop. (formn n l) means that
 l is a formated list with n characters on its last line *)

Inductive formn :nat->l_ch->Prop
    := fword : (m:l_ch)(word m)->(inflg m)->(formn (lgth m) m)
    | fwordinf : (p:nat)(m1:l_ch)(b:bool)(m2:l_ch)
             (inf (plus_l_ch m1 p))->(word m1)->(formn p (consbl b m2))
              ->(formn (plus_l_ch m1 p) (appsp m1 m2))
    | fwordsup : (p:nat)(m1:l_ch)(b:bool)(m2:l_ch)
         (sup (plus_l_ch m1 p))->(word m1)->(inflg m1)->
         (formn p (consbl b m2)) ->(formn (lgth m1) (applf m1 m2))
    | fsp : (p:nat)(i:Induct)(m:l_ch)
            (formn p (consltr i m))->(formn (S p) (conssp (consltr i m))).

(* Program's specification for a list l : there exists n : nat and m : list *)
(* such that (formn n m) and (Eq l m)                                       *)

 Inductive SigFormat [l:l_ch] : Set
    := ExFormat : (n:nat)(m:l_ch)(formn n m)->(Eq l m)->(SigFormat l).

(* Two lemmas *)

Lemma Lem1 : (b:bool)(p,m1,m2:l_ch)(n:nat)(h1:(word m1))(h2:(inflg m1))
             (t1:(formn n (consbl b m2)))
			 (t2:(Eq p (consbl b m2)))
             (SigFormat (app m1 p)).
Intros.
Elim (ax1 (plus_l_ch m1 n)); Intros.
Apply (ExFormat (app m1 p) (plus_l_ch m1 n) (appsp m1 m2)).
Apply (fwordinf n m1 b m2); Auto.
Apply (Eq_app m1 p (conssp m2)); Auto.
Apply (Eq_tran p (consbl b m2) (conssp m2)); Auto.
Apply (Eq_co_bl b true m2 m2); Auto.
Apply Eq_re.
Apply (ExFormat (app m1 p) (lgth m1) (applf m1 m2)).
Apply (fwordsup n m1 b m2); Auto.
Apply (Eq_app m1 p (conslf m2)); Auto.
Apply (Eq_tran p (consbl b m2) (conslf m2)); Auto.
Apply (Eq_co_bl b false m2 m2); Auto.
Apply Eq_re.
Save.

Local prop2  [m1,m2,p:l_ch][n:nat][u:l_ch]
    (formn n u)->(Eq p u)->(SigFormat (app m1 (app m2 p))).             

Lemma Lem2 :(m1,m2,p:l_ch)
			(h1:(word m1))
			(h2:(inflg m1))
			(h3:(space m2))
			(n:nat)
			(p0:l_ch)
			(prop2 m1 m2 p n p0).
Intros.
Elim p0; Intros.
Red; Intros.
Apply (ExFormat (app m1 (app m2 p)) (lgth m1) m1).
Apply (fword m1); Auto.
Apply (Eq_app_nil m1 (app m2 p)); Auto.
Apply (Eq_tran (app m2 p) (app m2 nil) nil); Auto.
Apply (Eq_app m2 p nil); Auto.
Apply (Eq_space_nil m2); Auto.
Red; Intros.
Apply (Lem1 b (app m2 p) m1 l n); Auto.
Apply (Eq_space_bl b m2 p l); Auto.
Red; Intros.
Apply (Lem1 true (app m2 p) m1 (consltr i l) (S n)); Auto.
Apply (fsp n i l); Auto.
Apply (Eq_space m2 p (consltr i l)); Auto.
Save.

(* Formatting of a valid list *)

 Theorem format :  (l:l_ch)(valide l)->(SigFormat l).
 Intros.
 Elim H; Intros.
 Apply (ExFormat  m (lgth m) m).
 Apply (fword m); Auto.
 Apply Eq_re.
 Elim H0; Intros.
 Apply (Lem2 m1 m2 p w i s n m); Auto.
 Save.

(* Formatting of a list : this function will detect if the list is not valid *)

 Theorem format_all : (l:l_ch)(SigFormat l)+{(NV l)}.
 Intro; Elim (preproc l) ; Intros.
 Left; Apply format; Auto.
 Right; Auto.
 Save.

End Format.


(* $Id: Format.v,v 1.2 1996/02/12 15:31:07 cpaulin Exp $ *)
