(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                               Manna_prog.v                               *)
(****************************************************************************)
    (*******************************************************************)
    (* f : nat->nat is an increasing and unbounded function            *)
    (* lambo(f)(n) = inf {m:nat| f(m)>n}                               *)
    (* is computed using the following algorithm :                     *)
    (* lambo f n = if f(0) > n the 0 else limbo(1)                     *)
    (*        where limbo(i) = if f(i) > n then 0                      *)
    (*                         else let j = limbo(2i) in               *)
    (*                  if f(j+i) > n then j else j+i                  *)
    (*******************************************************************)

Require Relation_Definitions.

Definition one    : nat    := (S O).

(* Addition and multiplication by 2 *)
Fixpoint add [n:nat] : nat->nat :=
 [m:nat]<nat>Case n of
	m
	[p:nat](S (add p m))
 end.

Definition twice : nat->nat := [n:nat](add n n).

Definition peano := 
[P:nat->Prop][b:(P O)][i:((u:nat)(P u)->(P (S u)))]
 Fix F {F/1 : (y:nat)(P y) :=
   [n:nat]<P>Case n of b [x:nat](i x (F x)) end}
         : (P:nat->Prop)(P O)->((u:nat)(P u)->(P (S u)))->(n:nat)(P n).

(* Operator for pattern-matching *)
 Definition     nat_match 
     :    (n:nat)(P:nat->Prop)(P O)->((m:nat)(P (S m)))->(P n)
     :=    [n:nat][P:nat->Prop][h1:(P O)][h2:(m:nat)(P (S m))]
        (peano P h1 [m:nat][ind:(P m)](h2 m) n).

(* Properties of addition                         *)
Theorem add_nO :   (n:nat)<nat>(add n O)=n.
Intro; Symmetry; Apply plus_n_O.
Save.

Theorem add_ass :  (u,v,w:nat)<nat>(add (add u v) w)=(add u (add v w)).
Intros.
Elim u; Auto.
Intros; Simpl.
Elim H; Auto.
Save.

(* Induction de contenu positif                        *)
Definition peano_set := 
[P:nat->Set][b:(P O)][i:((u:nat)(P u)->(P (S u)))]
 Fix F {F/1 : (y:nat)(P y) :=
   [n:nat]<P>Case n of b [x:nat](i x (F x)) end}
         : (P:nat->Set)(P O)->((u:nat)(P u)->(P (S u)))->(n:nat)(P n).

(* Absurdity *)

 Theorem abs :   (A:Prop)(False -> A).
 Intros; Elim H; Auto.
 Save.

(* An axiomatized order relation on natural numbers *)

 Parameter     inf : nat->nat->Prop.

 Axiom tran_inf    :     (transitive nat inf).
 Axiom infOO    :     (inf O O).
 Axiom infS_inf    :    (n,m:nat)(inf (S n) (S m))->(inf n m).
 Axiom inf_infS    :    (n,m:nat)(inf n m)->(inf (S n) (S m)).
 Axiom infS    :    (n:nat)(inf n (S n)).
 Axiom absO    :    (inf one O)->False.

 Definition sup    : nat->nat->Prop    := [n,m:nat](inf (S m) n).

(* Properties of this order *)

 Lemma re_inf :   (n:nat)(inf n n).
 Intro; Elim n.
 Apply infOO.
 Intros; Apply inf_infS; Auto.
 Save.

 Lemma infO :    (n:nat)(inf O n).
 Intro; Elim n.
 Apply infOO.
 Intros; Apply (tran_inf O n0 (S n0)); Auto.
 Apply infS.
 Save.

 Lemma infS_O :  (n:nat)(inf (S n) O)->False.
 Intro; Elim n; Intros.
 Apply absO; Auto.
 Elim H.
 Apply (tran_inf (S n0) (S (S n0)) O); Auto.
 Apply infS.
 Save.

 Lemma infSn_n :   (n:nat)(inf (S n) n)->False.
 Intro; Elim n; Intros.
 Apply absO; Auto.
 Elim H.
 Apply infS_inf; Auto.
 Save.

 Lemma inf_sup_abs :  (n,m:nat)(inf n m)->(sup n m)->False.
 Intros; Apply (infSn_n n).
 Apply (tran_inf (S n) (S m) n); Auto.
 Apply inf_infS; Auto.
 Save.

(* A lemma : (sup n m)->(sup (twice n) (S m)) *)

 Lemma inf_add : (n,m:nat)(inf m (add n m)).
 Intros; Elim n.
 Apply re_inf.
 Intros; Apply (tran_inf m (add n0 m) (add (S n0) m)); Auto.
 Simpl; Apply infS.
 Save.

 Lemma sup_twice : (n,m:nat)(sup n m)->(sup (twice n) (S m)).
 Intros n m; Elim n; Intros.
 Apply abs.
 Apply (infS_O m H).
 Apply (inf_infS (S m) (add n0 (S n0))).
 Apply (tran_inf (S m) (S n0) (add n0 (S n0))); Auto.
 Apply inf_add.
 Save.

(* Program Lambo *)

(* Hypotheses *)
Variable f    : nat->nat.

(* f is unbounded *)
Hypothesis Unbound    
    :    (n:nat){y:nat|(sup (f y) n)}.

(* f is increasing *)
Hypothesis Increas    : (n:nat)(m:nat)(inf n m)->(inf (f n) (f m)).

(* There is a procedure to decide if inf or sup hold *)

Axiom inf_sup    : (x:nat)(y:nat) {(inf x y)}+{(sup x y)}.
Axiom inf_sup0    : (x:nat)(y:nat)(inf x y)\/(sup x y).

(* We give n : nat and try to compute lambo(n) *)
Variable n : nat.

 Definition Inf : nat->Prop   := [m:nat](inf (f m) n).
 Definition Sup : nat->Prop   := [m:nat](sup (f m) n).

(* We only use the following properties of Inf and Sup *)

 Definition bound : {y:nat|(sup (f y) n)}
           := (Unbound n).

 Definition Inf_Sup : (u:nat){(Inf u)}+{(Sup u)}
             := [u:nat](inf_sup (f u) n).

 Definition Inf_Sup_abs : (u:nat)(Inf u)->(Sup u)->False
             := [u:nat](inf_sup_abs (f u) n).

(* F is increasing is used this way *)

 Lemma infInf :     (u,v:nat)(inf u v)->(Inf v)->(Inf u).
 Intros; Apply (tran_inf (f u) (f v) n); Auto.
 Apply Increas; Auto.
 Save.

(* Setifications *)
 Definition Small : nat->Prop := [m:nat](i:nat)(inf (S i) m)->(Inf i).

 Lemma SmallO:  (Small O) .
 Red; Intros; Apply abs; Apply (infS_O i H).
 Save.

(* The initial specification *)
 Definition Lambo : Set := {m:nat|(Sup m)&(Small m)}.

 Fact Lem1:  (m:nat)(Inf m)->(Small (S m)).
 Red; Intros; Apply (infInf i m); Auto.
 Apply infS_inf; Auto.
 Save.

(* Transformation of the specification *)
 Definition Lambo1 : Set 
              := ({m:nat|(Inf m)&(Sup (S m))})+{(Sup O)}.

 Lemma Reduct1:  Lambo1->Lambo.
 Realizer [h:Lambo1]
             <Lambo>Match h with
                 [h1:?](S h1)
                 O
	     end.
 Program_all.
 Apply Lem1; Auto.
 Exact SmallO.
 Save.
 
(* Intermediate function *)
 Definition Limbo : nat -> Set 
     := [i:nat]
     (sup i O)-> ({m:nat|(Inf m)&(Sup (add i m))})+{(Sup O)}.

(* (add one m)=(S m) by Beta reduction *)

 Lemma Reduct2:  (Limbo one)->Lambo1 .
 Intro; Apply H.
 Apply (re_inf one).
 Save.

(* Termination *)

(* A parametrized order *)
 Definition bd : nat->nat->nat->Prop 
          := [y:nat][u:nat][v:nat]((sup u v)/\(inf v y)).

(* Property to be well-formed *)
 Definition wf_bd 
     : nat->nat->Set
     := [y:nat][i:nat]
     (P:nat->Set)((v:nat)((u:nat)(bd y u v)->(P u))->(P v))->(P i).

(* Some property of bd *)

 Fact Lem2:  (y1,y2,u,v:nat)(bd y1 u v)->(inf v y2)->(bd y2 u v).
 Intros; Elim H; Intros.
 Red; Auto.
 Save.

 Fact Lem3:  (y,u,v,w:nat)(bd y u v)->(bd y v w)->(sup y w).
 Intros; Apply (tran_inf (S w) v y); Auto.
 Elim H0; Auto.
 Elim H; Auto.
 Save.

 Lemma Term:  (i,y:nat)(sup i y)->(wf_bd y i).
 Realizer [i,y:nat][P:Set][q:nat->(nat->P)->P](q i [h:nat](except P)).
 Program_all.
 Apply (inf_sup_abs i y); Auto. 
 Elim H0; Auto.
 Save.

(* Proof of (wf_bd y) by induction on y *)

 Lemma cas_base: (i:nat)(sup i O)->(wf_bd O i).
 Realizer [i:nat][P:Set][q:nat->(nat->P)->P](q i [h:nat](except P)).
 Program_all.
 Apply (inf_sup_abs i O); Auto.
 Elim H0; Auto.
 Save.

 Lemma cas_ind:  (i,y:nat)(wf_bd y i)->(wf_bd (S y) i).
 Realizer [i,y:nat][ind:wf_bd][P:Set][q:nat->(nat->P)->P]
            (ind P [v:nat][g:nat->P]
                     (q v [u:nat]
                            <P>Match (inf_sup v y) with
				(g u)
                                (q u [h:nat](except P))
			     end)).
 Program_all.
 Apply (Lem2 (S y) y u v); Auto.
 Apply (infSn_n v).
 Apply (tran_inf (S v) (S y) v) ; Auto.
 Apply (Lem3 (S y) h u v) ; Auto.
 Save.

Theorem Wf1:  (y,i:nat)(sup i O)->(wf_bd y i).
Intros; Elim y.
Apply cas_base; Auto.
Intros; Apply cas_ind; Auto.
Save.

(* Actually we will use a simpler induction scheme *)

 Definition Induct : nat->nat->(nat->Set)->Set
             := [y:nat][i:nat][P:nat->Set]
          ((k:nat)((inf k y)->(P (twice k)))->(P k))->(P i).

(* Proof of the induction scheme *)
Theorem wf_Ind: (y:nat)(i:nat)(sup i O)->(P:nat->Set)(Induct y i P).
Realizer [y,i:nat][P:Set]
           <(Induct P)>Match y with
		 [q:?](q i [{Hyp:(inf i O)}](except P))
                 [u,ind,q:?]
                    (ind [k:nat][r:P]
                           (q k (<P>if (inf_sup k u) then
				      r
				 else
				   (q (twice k) 
				     [{Hyp:(inf (twice k) (S u))}](except P)))))
	   end.
Program_all.
Apply (inf_sup_abs i O); Auto.
Apply (inf_sup_abs (twice k) (S n1)); Auto.
Apply sup_twice; Auto.
Save.

(* Proof of the Limbo's program  *)

 Fact Lem4:  (u,v:nat)(Sup u)->(Inf v)->(inf v u).
 Intros.
 Elim (inf_sup0 v u); Intros.
 Apply H1.
 Apply abs.
 Apply (Inf_Sup_abs u); Auto.
 Apply (infInf u v); Auto.
 Apply (tran_inf u (S u) v); Auto.
 Apply infS.
 Save.

 Definition LimboSig : nat->Set 
            := [i:nat]{m:nat|(Inf m)&(Sup (add i m))}.

 Lemma LimboLem:  (i:nat)(sup i O)->(Inf O)->(LimboSig i).
 Realizer [i:nat]
            (wf_Ind bound i nat
                    [k,hk:nat]
                      <nat>if (Inf_Sup (add k O)) then
                          <nat>if (Inf_Sup (add k hk)) then
                                  (add k hk)
                               else hk

                           else O).
 Program_all.
 Apply Lem4.
 Elim bound; Auto.
 Elim (add_nO k); Auto.
 Unfold twice in s.
 Elim add_ass; Auto.
 Save.

 Lemma Prog:  (i:nat)(Limbo i).
 Realizer  [i:nat]
          <Exc nat>if (Inf_Sup O) then
             (value nat (LimboLem i))
             else (error nat).
 Program_all.
 Save.

(* Final proof *)
Theorem LamboProg:  Lambo .
Realizer (Reduct1 (Reduct2 (Prog one))).
Program_all.
Save.



(* $Id: Manna_prog.v,v 1.6 1996/02/12 15:31:13 cpaulin Exp $ *)
