(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                               Floyd_prog.v                               *)
(****************************************************************************)

Require Program.
Require Arith.
Require Compare.
Require SetTheory.

(* Some arithmetical lemma *)

Lemma lemme1 : (p,q,n,m:nat) (le (plus p q) n)->(le (plus p m) (plus n m)).
Intros p q n m H.
Apply le_reg_r.
Apply le_trans with (plus p q); Auto.
Save.

Lemma lemme2 : (p,q,m,n:nat) (le (plus p q) n)->(le (plus p (plus q m)) (plus n m)).
Intros p q m H; Elim (plus_assoc_r p q m); Auto.
Save.
Hint lemme2.

(************************************************************************)
(*  On met une relation logique entre deux sommets,                     *)
(*  et il y a une et une seule arete entre deux sommets                 *)
(************************************************************************)

Axiom E : V->V->nat->Prop.

Axiom Edge_un : (x,y:V) (n,m:nat) (E x y n)->(E x y m)->(n=m).

(************************************************************************)
(*  Deux sommets sont relies soit par un arc de cout n                  *)
(*  soit par un un arc de cout infini <--> non relies                   *)
(************************************************************************)

Inductive binding [x,y:V] : Set
   := Poids : (n:nat) (E x y n) -> (binding x y)
   | Infini: ((n:nat) ~(E x y n))-> (binding x y).
Hint (*Poids*) Infini.

(************************************************************************)
(*  Pour tout x,y on a binding x y                                      *)
(************************************************************************)

Axiom dec_E : (x,y:V) (binding x y).
Hint dec_E.

(************************************************************************)
(*  Definition d un chemin de cout n entre deux sommets                 *)
(************************************************************************)

Inductive path [W:set;x:V] : V->nat->Set
   := Edge : (y:V) (n:nat) (E x y n) -> (path W x y n)
   | Composed : (y,z:V) (n,m:nat) (path W x y n)->(In W y)->(E y z m)->
                                         (path W x z (plus n m)).
Hint Edge.

(************************************************************************)
(* 	Si path W x y n  &  W incl W' then path W' x y n                *)
(************************************************************************)


Definition Edgec := [W:set][x:V][y:V][n:nat](Edge W x y n).
Definition Composedc := [W:set][x:V][y,z:V][n,m:nat](Composed W x y z n m).

Lemma Path_inclusion : (W,W':set) (x,y:V) (n:nat)(incl W W')->(path W x y n)->(path W' x y n).
Realizer [x,y:V][n:nat][H0:path]
                       <path>Match H0 with
			  [y0,n0](Edgec x y0 n0)
                          [z,t,m,p,P1,P2](Composedc x z t m p P2) end.
Program_all.
Save.

(************************************************************************)
(* 	Si path W x y n alors path (W+u) x y n                          *)
(************************************************************************)

Lemma Path_add : (W:set) (x,y,u:V) (n:nat) (path W x y n)->(path (add W u) x y n).
Realizer [x,y,u:V][n:nat][H:path](Path_inclusion x y n H).
Program_all.
Save.
Hint Path_add.

(************************************************************************)
(* 	Si path False x y n alors E x y n                               *)
(************************************************************************)

Lemma Path_empty : (x,y:V) (n:nat)(path empty x y n)->(E x y n).
Intros x y n H.
Elim H; Auto.
Intros z t m p P1 E' I; Elim I.
Save.
Immediate Path_empty.

(************************************************************************)
(*   Si path W x y n et path W y z m et in W y alors path W x z n+m     *)
(************************************************************************)

Lemma Path_transitive : (W:set) (x,y,z:V) (n,m:nat)
     (path W y z m)->(path W x y n)->(In W y)->(path W x z (plus n m)).
Realizer [x,y,z:V][n,m:nat][H:path]
             <path->path>Match H with
                  [t,p,P](Composedc x y t n p P)
                  [t,u,p,q,P,H0,P0](Composedc x t u (plus n p) q (H0 P0))
             end.
Program_all.
Save.

(************************************************************************)
(*  Si path (W+y) x z n alors path W x z n ou                           *)
(*              (path W x y p  &  path W y z q & p+q<=n)                *)
(************************************************************************)

Inductive step [W:set;y,x,z:V;n:nat] : Set
    := Direct : (p:nat) (path W x z p)->(le p n)-> (step W y x z n)
    | Indirect : (p,q:nat) (path W x y p)->(path W y z q)->(le (plus p q) n)->
                 (step W y x z n).

Definition Directc := [W:set][y,x,z:V][n:nat][p:nat](Direct W y x z n p).
Definition Indirectc :=
	 [W:set][y,x,z:V][n:nat][p,q:nat](Indirect W y x z n p q).

Lemma Step_edge : (W:set) (y,x,z:V) (n:nat) (E x z n)->(step W y x z n).
Realizer [x,y,z:V][n:nat](Directc x y z n n (Edgec y z n)).
Program_all.
Save.
Hint Step_edge.

Lemma Path_step : (W:set) (y,x,z:V) (n:nat) (path (add W y)  x z n)->(step W y x z n).
Realizer [{W:set}][y,x,z:V][n:nat][H:path]
            <step>Match H with
		[y0,n0](Step_edge y x y0 n0)
                [t,u,m,p,P,S]
                	<step>Match S with
                               [q,P0]
                                 <step>if (add_null t y) 
			:: :: { ((In (add W y) t) -> ({(y=t)}+{(In W t)})) }
                                    then (Indirectc t x u (plus m p) q p
                                          		   P0 (Edgec t u p))
                                    else (Directc y x u (plus m p) (plus q p)
                                       		 (Composedc x t u q p P0))
                               [r,s,P0,P1]
                                   <step>if (add_null t y) 
			:: :: { ((In (add W y) t) -> ({(y=t)}+{(In W t)})) }
                                     then (Indirectc y x u (plus m p) r p P0
                                         	     (Edgec t u p))
                                     else (Indirectc y x u (plus m p) r
                                         	     (plus s p) P0
                                         	     (Composedc y t u s p P1))
		 	end
	    end.
Program_all.
Rewrite y1; Auto.
Apply lemme1 with s ; Auto.
Save.
Hint Path_step.

(************************************************************************)
(*  Setification                                                        *)
(************************************************************************)

Inductive floyd [W:set;x,y:V] : Set 
 := Min_path : (n:nat) (path W x y n)->((m:nat) (path W x y m)->(le n m))->
                      (floyd W x y)
|  No_path : ((n:nat) (path W x y n)->False)->(floyd W x y).

(************************************************************************)
(* On utilise un ensemble F fini de sommets                             *)
(************************************************************************)

Definition Min_pathc := [W:set][x,y:V][n:nat](Min_path W x y n).
Definition No_pathc := [W:set][x,y:V](No_path W x y).
Variable F : set.

(* cas de base *)
Lemma floyd_empty : (x,y:V)(floyd empty x y).
Realizer [x,y:V]<floyd>if (dec_E x y)
		   then [n](Min_pathc x y n (Edgec x y n)) 
		   else (No_pathc x y).
Program_all.
Intros m H; Replace m with n; Auto.
Apply (Edge_un x y); Auto.
Intros; Elim (n n0); Auto.
Save.
Hint floyd_empty.

Lemma floyd_induction : (y:V)(Q:set)(In F y)->(~(In Q y))->
     ((x,z:V)(In F x)->(In F z)->(floyd Q x z))->
      (x,z:V)(In F x)->(In F z)->(floyd (add Q y) x z).
Realizer [y:V][H0:V->V->floyd][x,z:V]
                  <floyd>Case (H0 x z) of
                     [n,Pxz]
                       <floyd>Case (H0 x y) of
                          [m,Pxy]
                            <floyd>Case (H0 y z) of
                               [p,Pyz]
                                 <floyd>if (le_dec n (plus m p))
                                 then (Min_pathc x z n (Path_add x z y n Pxz))
                                 else (Min_pathc x z (plus m p)
                                       		(Path_transitive x y z m p
                                          		(Path_add y z y p Pyz)
                                          		(Path_add x y y m Pxy)))
                               (Min_pathc x z n (Path_add x z y n Pxz)) 
			    end
                          (Min_pathc x z n (Path_add x z y n Pxz)) 
			end
                     <floyd>Case (H0 x y) of
                        [m,Pxy]
                          <floyd>Case (H0 y z) of
                             [p,Pyz]
                               (Min_pathc x z (plus m p)
                                  (Path_transitive x y z m p
                                     (Path_add y z y p Pyz)
                                     (Path_add x y y m Pxy)))
                             (No_pathc x z) 
			  end
                        (No_pathc x z) 
		     end 
		  end.
Program_all.
(* (m:nat)(path (add Q y) x z m)->(le n m) *)
Intros q Padd.
Elim (Path_step Q y x z q Padd).
Intros r P le_rq.
Apply le_trans with r ; Auto.
Intros r s Pxyr Pyzs le.
Apply le_trans with (plus r s); Auto.
Apply le_trans with (plus m p); Auto.
Apply le_plus_plus; Auto.

(* (m0:nat)(path (add Q y) x z m0)->(le (plus m p) m0) *)
Intros q Padd.
Elim (Path_step Q y x z q Padd).
Intros r P le_rq.
Apply le_trans with r ; Auto.
Apply le_trans with n ; Auto.
Intros r s Pxyr Pyzs le.
Apply le_trans with (plus r s); Auto.
Apply le_plus_plus; Auto.

(* (m:nat)(path (add Q y) x z m)->(le n m) *)
Intros q Padd.
Elim (Path_step Q y x z q Padd).
Intros r P le_rq.
Apply le_trans with r ; Auto.
Intros r s Pxyr Pyzs.
Elim (f s); Auto.

(* (m:nat)(path (add Q y) x z m)->(le n m) *)
Intros q Padd.
Elim (Path_step Q y x z q Padd).
Intros r P le_rq.
Apply le_trans with r ; Auto.
Intros r s Pxyr.
Elim (f r); Auto.

(* (m0:nat)(path (add Q y) x z m0)->(le (plus m p) m0) *)
Intros q Padd.
Elim (Path_step Q y x z q Padd).
Intros r P le_rq.
Elim (f r); Auto.
Intros r s Pxyr Pyzs le.
Apply le_trans with (plus r s); Auto.
Apply le_plus_plus; Auto.

(* (n:nat)(path (add Q y) x z n)->False *)
Intros q Padd.
Elim (Path_step Q y x z q Padd).
Intros r P le_rq.
Apply (f r); Auto.
Intros r s Pxyr Pyzs le.
Apply (f0 s); Auto.

(* (n:nat)(path (add Q y) x z n)->False *)
Intros q Padd.
Elim (Path_step Q y x z q Padd).
Intros r P le_rq.
Apply (f r); Auto.
Intros r s Pxyr Pyzs le.
Apply (f0 r); Auto.

Save.
Hint floyd_induction.

(* Preuve pour tout ensemble fini par recurrence *)
Axiom F_finit : (enumerate F).

(* the induction principle *)
Lemma F_induction : (P:set->Set)(P empty)->
         ((y:V)(Q:set)(In F y)->(~(In Q y))->(P Q)->(P (add Q y)))
         ->(P F).
Realizer [P:Set][Pemp:P]
                  <(V->P->P)->P>Match F_finit with
		     [_]Pemp
                     [y,y0,H,H0](H0 y (H [y1:V][H3:P](H0 y1 H3))) 
		  end.
Program_all.
Save.

(*********************************************************)
(*  Preuve de Floyd                                      *)
(*********************************************************)

Lemma Floyd_1 : (x,y:V)(In F x)->(In F y)->(floyd F x y).
Pattern 3 F.
Realizer (F_induction V->V->floyd 
	       [x,y:V](floyd_empty x y)
               [y:V][H1:V->V->floyd][x,y0:V](floyd_induction y H1 x y0)).
Program_all.
Save.


(************************************)
(*      Arrays on V	            *)
(************************************)

Section Array.
Variable A: V->Set.

Inductive array : set->Set :=
    empty_array : (array empty)
   | add_array  : (P:set)(y:V)(A y)->(array P)->(array (add P y)).

End Array.
Hint empty_array add_array.

(***************)
(* Access      *)
(***************)

Lemma acces : (A:V->Set)(P:set)(array A P)->(x:V)(In P x)->(A x).
Realizer [A:Set][H:(array A)]
	   (array_rec A V->A
	       [x:V](False_rec A)
	       [{P:set}][y:V][ay:A][Arr:(array A)][accP:V->A][x:V]
  		    <A>if (add_null x y) 
			:: :: { ((In (add P y) x)->({(y=x)}+{(In P x)})) }
		    then ay
		    else (accP x) 
	      H).
Program_all.
Save.


Lemma acces_F : (A:V->Set)(array A F)->(x:V)(In F x)->(A x).
Realizer [A:Set][H:(array A)][x:V](acces A H x).
Program_all.
Save.

(*******************)
(* Initialisation  *)
(*******************)

Lemma array_init : (A:V->Set)((y:V)(In F y)->(A y))->(array A F).
Realizer [A:Set]<(V->A)->(array A)>Match F_finit with
                   [H](empty_array ?)
                   [y,y0,H,H0](add_array ? y (H0 y) (H [y1:V](H0 y1))) 
		 end.
Program_all.
Save.
Hint array_init.

(*********************************************************)
(* Representation de Floyd par une matrice               *)
(*********************************************************)

Definition floyd_matrix := [Q:set]
    (array [x:V](array (floyd Q x) F) F).


Theorem build_floyd_matrix : 
       (Q:set)((x,y:V)(In F x)->(In F y)->(floyd Q x y))->(floyd_matrix Q).
Realizer [H:V->V->floyd]
           (array_init (array floyd)
               [y:V](array_init floyd [y0:V](H y y0))).
Program_all.
Save.
Hint build_floyd_matrix.

Theorem acces_floyd_matrix : 
    (Q:set)(floyd_matrix Q)->(x,y:V)(In F x)->(In F y)->(floyd Q x y).
Realizer [H:(array (array floyd))]
           [x,y:V](acces floyd (acces (array floyd) H x) y).
Program_all.
Save.
Hint acces_floyd_matrix.

Theorem Floyd_2 : (floyd_matrix F).
Realizer (F_induction floyd_matrix
              (build_floyd_matrix [x:V][y:V](floyd_empty x y))
              [y:V][H1:floyd_matrix]
                 (build_floyd_matrix
                    [x,y0:V](floyd_induction y
				 [x0,z:V](acces_floyd_matrix H1 x0 z) x y0))).
Program_all.
Save.





(* $Id: Floyd_prog.v,v 1.4 1996/02/12 15:29:50 cpaulin Exp $ *)
