(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                              Quick2_prog.v                               *)
(****************************************************************************)

Require List.
Require Gt.

(* Chapter Quicksort *)

(*****************************)
(* A decidable relation on A *)
(*****************************)

Variable   inf : A -> A -> Prop.
Definition sup := [x,y:A]~(inf x y).

Hypothesis inf_sup : (x,y:A){(inf x y)}+{(sup x y)}.


(******************************)
(* mil a l m = app l a.m      *)
(******************************)

Definition mil := [a:A][l,m:list](app l (cons a m)) : A->list->list->list.

Lemma mil_app : (a:A)(l,m,n:list)(mil a l (app m n))=(app (mil a l m) n).
	Intros.
	Unfold mil.
	Elim (ass_app l (cons a m) n).
	Auto.
Save.

(*****************************)
(* Equivalence of two lists  *)
(*****************************)

Inductive permut : list->list->Prop :=
    permut_nil  : (permut nil nil)
   |permut_tran : (l,m,n:list)(permut l m)->(permut m n)->(permut l n)
   |permut_cmil : (a:A)(l,m,n:list)
         (permut l (app m n))->(permut (cons a l) (mil a m n))
   |permut_milc : (a:A)(l,m,n:list)
         (permut (app m n) l)->(permut (mil a m n) (cons a l)).

Hint permut_nil permut_cmil permut_milc.

(******************************)
(* if l eq m then a.l eq a.m  *)
(******************************)


Lemma permut_cons : (a:A)(l,m:list)(permut l m)->(permut (cons a l) (cons a m)).
	Intros.
	Change (permut (cons a l) (mil a nil m)).
	Auto.
Save.
Hint permut_cons.

(*****************************)
(* l eq l                    *)
(*****************************)

Lemma permut_refl : (l:list)(permut l l).
	Induction l ; Auto.
Save.
Hint permut_refl.

(**********************************)
(* if l eq m then m eq l      *)
(**********************************)

Lemma permut_sym : (l,m:list)(permut l m)->(permut m l).
	Intros l m h1 ; Elim h1 ; Auto.
	Intros l0 m0 n h2 h3 h4 h5.
	Apply permut_tran with m0 ; Auto.
Save.
Immediate permut_sym.


(***************************************************************)
(*  if m1 eq n1  and m2 eq n2 then (app m1 m2) eq (app n1 n2)  *)
(*  and if m eq (app m1 m2) then m eq (app n1 n2)              *)
(***************************************************************)

Lemma permut_app1 : (m1,n1,l:list)(permut m1 n1)->(permut (app l m1) (app l n1)).
	Intros ; Elim l ; Simpl ; Auto.
Save.
Hint permut_app1.

(* A technical lemma *)
Lemma permut_app_mil : (a:A)(l1,m1,l2,m2,n2:list)
     (permut (app l1 m1) (app (app l2 m2) n2))
		-> (permut (app (cons a l1) m1) (app (mil a l2 m2) n2)).
	Intros ; Simpl.
	Elim (mil_app a l2 m2 n2).
	Apply permut_cmil.
	Elim (app_ass l2 m2 n2) ; Auto.
Save.
Hint permut_app_mil.

Lemma permut_app_app : (m1,m2,n1,n2 :list)
     (permut m1 n1)->(permut m2 n2)->(permut (app m1 m2) (app n1 n2)).
	Intros m1 m2 n1 n2 h1 h2.
	Elim h1 ; Intros.
	Exact h2.
	Apply permut_tran with (app m n2) ; Auto.
	Apply permut_tran with (app m m2) ; Auto.
	Auto.
	Apply permut_sym ; Auto.
Save.
Hint permut_app_app.

Lemma permut_app : (m,m1,m2,n1,n2 :list)(permut m1 n1)->(permut m2 n2)->
     (permut m (app m1 m2))->(permut m (app n1 n2)).
	Intros.
	Apply permut_tran with (app m1 m2) ; Auto.
	Save.

(*****************************************************)
(* Each element of a list satisfies a given relation *)
(*****************************************************)

Section Rlist_.

Variable R : A->Prop.

Inductive Rlist  : list -> Prop :=
    Rnil : (Rlist nil)
   |Rcons : (x:A)(l:list)(R x)->(Rlist l)->(Rlist (cons x l)).

Hint Rnil Rcons.

Lemma Rlist_app : (m,n:list)(Rlist m)->(Rlist n)->(Rlist (app m n)).
	Intros m n h1 h2 ; Elim h1 ; Simpl ; Auto.
Save.
Hint Rlist_app.

Section Rlist_cons_.

Variable a : A.
Variable l : list.
Hypothesis Rc : (Rlist (cons a l)).

Lemma Rlist_cons : (R a)/\(Rlist l).
    Inversion_clear Rc; Auto.
Save.

End Rlist_cons_.

Section Rlist_append.
Variable n,m : list.

Lemma Rlist_appd : (Rlist (app n m))->((Rlist n)/\(Rlist m)).
Elim n ; Simpl; Auto.
Intros a y h1 h2.
Elim (Rlist_cons a (app y m)) ; Auto.
Intros h3 h4; Elim h1 ; Auto.
Save.

End Rlist_append.
Hint Rlist_appd.

(****************************************)
(* if l eq m then (Rlist l)=>(Rlist m)  *)
(****************************************)

Lemma Rpermut : (m,n:list)(permut m n)->(Rlist m)->(Rlist n).
	Intros m n h1 ; Elim h1 ; Unfold mil ; Auto.
	Intros a l m0 n0 h2 h3 h4.
	Elim (Rlist_cons a l); Auto.
        Intros h5 h6; Elim (Rlist_appd m0 n0); Auto.
	Intros a l m0 n0 h2 h3 h4.
	Elim (Rlist_appd m0 (cons a n0)) ; Auto.
        Intros h5 h6; Elim (Rlist_cons a n0) ; Auto.
Save.

End Rlist_.
Hint Rnil Rcons Rlist_app.

Section Inf_Sup.
(******************************************)
(* (inf_list x l) == (inf x a) for a in l *)
(* (sup_list x l) == (sup x a) for a in l *)
(******************************************)

Hypothesis x : A.
Hypothesis l : list.

    Definition inf_list := (Rlist (inf x) l).
    Definition sup_list := (Rlist (sup x) l).

End Inf_Sup.

Hint Unfold  inf_list sup_list.

(**************************************)
(* Definition for a list to be sorted *)
(**************************************)

Inductive sort : list->Prop :=
     sort_nil : (sort nil)
   | sort_mil : (a:A)(l,m:list)(sup_list a l)->(inf_list a m)
        ->(sort l)->(sort m)->(sort (mil a l m)).

Hint sort_nil sort_mil.

Lemma permutapp : (a0:A)(y,l1,l2:list)(permut y (app l1 l2))->(permut (cons a0 y) (app l1 (cons a0 l2))).
Intros.
Exact (permut_cmil a0 y l1 l2 H).
Save.
Hint permutapp.

Lemma sortmil : (a:A)(x,x0,l1,l2:list)(sup_list a l1)->(inf_list a l2)->(sort x)->(sort x0)->(permut l1 x)->(permut l2 x0)->(sort (mil a x x0)).
Intros.
Apply sort_mil ; Auto.
Unfold sup_list ; Apply Rpermut with l1 ; Auto. 
Unfold inf_list ; Apply Rpermut with l2 ; Auto.
Save.
(*Hint sortmil.*)


Lemma permutmil : (a:A)(m,x,x0,l1,l2:list)(permut l1 x)->(permut l2 x0)->(permut m (app l1 l2))->(permut (cons a m) (mil a x x0)).
Intros.
Apply permut_cmil.
Apply permut_app with l1 l2 ; Auto.
Save.
(*Hint permutmil.*)

Definition ltl := [l,m:list](gt (length m) (length l)).

Hint Unfold ltl.

Lemma ltl_cons : (a,a0:A)(l1,y:list)(ltl l1 (cons a y))->(ltl l1 (cons a (cons a0 y))).
Unfold ltl; Simpl; Auto.
Save.
Hint ltl_cons.

Lemma ltl_cons_cons : (a,a0:A)(l2,y:list)(ltl l2 (cons a y))->(ltl (cons a0 l2) (cons a (cons a0 y))).
Unfold ltl; Simpl; Auto.
Save.
Hint ltl_cons_cons.

Require Wf_nat.

Inductive Splitting_spec [a:A; l:list] : Set :=
  Split_intro : (l1,l2:list)
                (sup_list a l1)->(inf_list a l2)->(permut l (app l1 l2))
                ->(ltl l1 (cons a l))->(ltl l2 (cons a l))
                ->(Splitting_spec a l).

Lemma Splitting : (a:A)(l:list)(Splitting_spec a l).
Realizer
 [a:A][l:list]
   <list*list>Match l with 
       (* nil *) (nil,nil)
  (* cons b m *) [b,m,ll]<list*list>let (l1,l2) = ll in
                           <list*list>if (inf_sup a b) then
                          (* inf a b *) (l1,(cons b l2))
			              else
                          (* sup a b *) ((cons b l1),l2)
   end.
Program_all.
Simpl; Auto.
Save.

Theorem Quicksort : (l:list){m:list|(sort m)&(permut l m)}.
Realizer <list>rec quick :: :: { ltl }
	   	[l:list]
		 <list>Case l of
             (* nil *)  nil
      	(* cons a m *)  [a,m]<list>let (l1,l2) = (Splitting a m) in
  	    	          (mil a (quick l1) (quick l2)) 
		 end.
Program_all.
Exact (well_founded_gtof list length).
Apply (sortmil a0 x0 x l1 l2) ; Auto.
Apply (permutmil a0 m x0 x l1 l2) ; Auto.
Save.



(* $Id: Quick2_prog.v,v 1.5 1996/02/12 15:30:15 cpaulin Exp $ *)
