(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.10                                 *)
(*                              Nov 25th 1994                               *)
(*                                                                          *)
(****************************************************************************)
(*                               Tauto_def.v                                *)
(****************************************************************************)

Require Bool.
Hint diff_true_false.

Require Peano_dec.

(*************************************************************)
(* Formula of propositional logic : definition and semantics *)
(*************************************************************)

Inductive PropForm : Set := 
          Fvar : nat -> PropForm
        | Or   : PropForm -> PropForm -> PropForm
        | And  : PropForm -> PropForm -> PropForm
        | Impl : PropForm -> PropForm -> PropForm
        | Neg  : PropForm -> PropForm
        | Bot  : PropForm.

(* Assignment of truth values to propositional variables *)

Definition Assign := nat -> bool.

(* The semantics of a propositional formula under an assignment *)

Fixpoint Prop_sem [A:Assign;F:PropForm] : bool :=
    <bool>Case F of
        (* Fvar n *) [n:nat](A n)
        (* Or F G *) [F,G:PropForm](orb (Prop_sem A F) (Prop_sem A G))
       (* And F G *) [F,G:PropForm](andb (Prop_sem A F) (Prop_sem A G))
      (* Impl F G *) [F,G:PropForm](implb (Prop_sem A F) (Prop_sem A G))
         (* Neg F *) [F:PropForm](negb (Prop_sem A F))
           (* Bot *) false 
	    end.

(***********************************************************)
(*  IF expressions : definition and semantics              *)
(***********************************************************)

Inductive IFExpr : Set := 
          Var  : nat -> IFExpr
        | Tr   : IFExpr
        | Fa   : IFExpr
        | IfE   : IFExpr -> IFExpr -> IFExpr -> IFExpr.

(* The semantics of an IF formula under an assignment *)

Fixpoint IF_sem [A:Assign;F:IFExpr] : bool :=
<bool> Case F of
    (* Var n *) [n:nat](A n)
       (* Tr *) true
       (* Fa *) false
(* IfE F G H *) [F,G,H:IFExpr](ifb (IF_sem A F) (IF_sem A G) (IF_sem A H)) 
end.

(*****************************************************************)
(* 1- Transformation of a propositional formula into an          *)
(*    equivalent IF-expression.				         *)
(*****************************************************************)

Definition Prop_IF_Equiv : PropForm -> IFExpr -> Prop
  := [F:PropForm][I:IFExpr](A:Assign)(Prop_sem A F)=(IF_sem A I).

Lemma PropForm_IFExpr : (F:PropForm){I:IFExpr|(Prop_IF_Equiv F I)}.
Induction F.
	(* case (Fvar n) *)
Intro n; Exists (Var n).
Unfold Prop_IF_Equiv; Simpl; Auto.
	(* case (Or G H) *)
Intros G HG H HH.
Elim HG; Intros IG SG.
Elim HH; Intros IH SH.
Exists (IfE IG Tr IH).
Unfold Prop_IF_Equiv; Simpl.
Intro A; Elim SG; Elim SH; Trivial.
	(* case (And G H) *)
Intros G HG H HH.
Elim HG; Intros IG SG.
Elim HH; Intros IH SH.
Exists (IfE IG IH Fa).
Unfold Prop_IF_Equiv; Simpl.
Intro A; Elim SG; Elim SH; Trivial.
	(* case (Impl G H) *)
Intros G HG H HH.
Elim HG; Intros IG SG.
Elim HH; Intros IH SH.
Exists (IfE IG IH Tr).
Unfold Prop_IF_Equiv; Simpl.
Intro A; Elim SG; Elim SH; Trivial.
	(* case (Neg G) *)
Intros G HG.
Elim HG; Intros IG SG.
Exists (IfE IG Fa Tr).
Unfold Prop_IF_Equiv; Simpl.
Intro A; Elim SG; Trivial.
        (* case Bot *)
Exists Fa; Unfold Prop_IF_Equiv; Simpl; Trivial.
Save.

(****************************************************************************)
(* Definition of Normal IF expressions.                                     *)
(* The conditional test is restricted to be a propositional variable and    *)
(* not an arbitrary IF expression                                           *)
(****************************************************************************)

Inductive Normal : IFExpr -> Prop
  := NVar : (n:nat)(Normal (Var n))
  | NTr  : (Normal Tr)
  | NFa  : (Normal Fa)
  | NIf  : (n:nat)(F,G:IFExpr)
           (Normal F)->(Normal G)->(Normal (IfE (Var n) F G)).

Hint NVar NTr NFa NIf.

(*******************************************)
(* Properties of Normal                    *)
(*******************************************)

(* The inverse properties for normal *)

Definition Norminv := [F:IFExpr]
     <Prop>Case F of
         (* Var n *) [n:nat]True
            (* Tr *) True
            (* Fa *) True
     (* IfE G H I *) [G,H,I:IFExpr]
                     <Prop>Case G of
              (* Var m *) [m:nat]((Normal H)/\(Normal I))
                 (* Tr *) False
	         (* Fa *) False
       (* IfE G' H' I' *) [G',H',I':IFExpr]False
		     end 
     end.

Lemma Normal_inv : (F:IFExpr)(Normal F) -> (Norminv F).
Induction 1; Simpl; Auto.
Save.

Lemma not_Norm_If_Tr : (F,G:IFExpr)~(Normal (IfE Tr F G)).
Intros F G; Exact (Normal_inv (IfE Tr F G)).
Save.

Lemma not_Norm_If_Fa : (F,G:IFExpr)~(Normal (IfE Fa F G)).
Intros F G; Exact (Normal_inv (IfE Fa F G)).
Save.

Lemma not_Norm_If_If : (F,G,H,I,J:IFExpr)~(Normal (IfE (IfE F G H) I J)).
Intros F G H I J;  Exact (Normal_inv (IfE (IfE F G H) I J)).
Save.

Lemma Norm_If_Var : (n:nat)(F,G:IFExpr)(Normal (IfE (Var n) F G))->((Normal F)/\(Normal G)).
Intros n F G; Exact (Normal_inv (IfE (Var n) F G)).
Save.

Hint not_Norm_If_Tr not_Norm_If_Fa not_Norm_If_If.

(******************************************************)
(* The proof of Normal corresponds to a partial match *)
(******************************************************)

Lemma Normal_rec : (P:IFExpr->Set)
        ((n:nat)(P (Var n)))->(P Tr)->(P Fa)->
        ((n:nat)(F:IFExpr)(G:IFExpr)(Normal F)->(P F)->
                (Normal G)->(P G)->(P (IfE (Var n) F G)))
        ->(i:IFExpr)(Normal i)->(P i).
Induction i; Trivial.
Intros F PF G PG I PI.
Elim F; Intros.
Elim (Norm_If_Var n G I); Auto.
Absurd (Normal (IfE Tr G I)); Auto.
Absurd (Normal (IfE Fa G I)); Auto.
Absurd (Normal (IfE (IfE i0 i1 i2) G I)); Auto.
Save.

(************************************************************************)
(* Representation of partial assignment with two lists of propositional *)
(* variables  whose values will be set either to true or false          *)
(************************************************************************)

Inductive Constraints : Set := 
  Emp :  Constraints | Cons : nat -> Constraints -> Constraints.

(* Assigned n P means that n occurs in P *)

Inductive Assigned [n:nat] : Constraints -> Prop 
  := Assign_hd : (PA:Constraints)(Assigned n (Cons n PA))
  | Assign_tl : (m:nat)(PA:Constraints)(Assigned n PA)->
                (Assigned n (Cons m PA)).

Hint Assign_hd Assign_tl.

Lemma Assign_eq : (n,m:nat)(PA:Constraints)(<nat>m=n)->(Assigned n (Cons m PA)).
Induction 1; Auto.
Save.
Hint Assign_eq.

(****************************************)
(* Properties of Assigned               *)
(****************************************)

Definition Assign_inv := [n:nat][PA:Constraints]
 (<Prop>Case PA of
         (* Emp *) False
   (* Cons m PA *) [m:nat][PA:Constraints]
                   (Assigned n PA)\/m=n end).

Lemma Assigned_inv : (n:nat)(PA:Constraints)(Assigned n PA)->(Assign_inv n PA).
Induction 1; Simpl; Auto.
Save.

Lemma not_Ass_emp : (n:nat)~(Assigned n Emp).
Intro n; Exact (Assigned_inv n Emp).
Save.
Hint not_Ass_emp.

Lemma Assign_cons : (n,m:nat)(PA:Constraints)
     (Assigned n (Cons m PA))->((Assigned n PA)\/m=n).
Intros n m PA; Exact (Assigned_inv n (Cons m PA)).
Save.

Lemma not_Ass_cons_intro : (n,m:nat)(PA:Constraints)
     (~m=n)->(~(Assigned n PA))->~(Assigned n (Cons m PA)).
Red; Intros n m PA MN NPA AC.
Elim (Assign_cons n m PA); Trivial.
Save.
Hint not_Ass_cons_intro.

Lemma not_Ass_cons_elim : (n,m:nat)(PA:Constraints)
     (~(Assigned n (Cons m PA)))->((~m=n)/\~(Assigned n PA)).
Unfold not; Auto.
Save.
Hint not_Ass_cons_elim.

(**************************************)
(* The relation Assigned is decidable *)
(**************************************)

Lemma Assign_dec : (n:nat)(PA:Constraints){(Assigned n PA)}+{~(Assigned n PA)}.
Induction PA.
Auto.
Intros m PA' HPA; Elim HPA.
Auto.
Elim (eq_nat_dec m n).
Induction 1; Auto.
Auto.
Save.

(**********************************************************************)
(* An assignment A is compatible with a constraint P for the boolean  *)
(* value b if A is b for each variable in P                           *)
(**********************************************************************)

Inductive Compatible [A:Assign;b:bool] : Constraints -> Prop 
 := Compat_emp  : (Compatible A b Emp)
 | Compat_cons : (n:nat)(lb:Constraints)(b=(A n))
                  ->(Compatible A b lb)->(Compatible A b (Cons n lb)).

Hint Compat_emp Compat_cons.

(**********************************************)
(* Properties of the compatibility predicate  *)
(**********************************************)

Definition Compat_inv := [A:Assign][b:bool][lb:Constraints]
( <Prop>Case lb  of
       (* Emp *)  True
(* Cons n lb' *)  [n:nat][lb':Constraints]
                  ((Compatible A b lb')/\b=(A n)) end).

Lemma Compatible_inv : (A:Assign)(b:bool)(PA:Constraints)
     (Compatible A b PA)->(Compat_inv A b PA).
Induction 1; Simpl; Auto.
Save.

Lemma Compat_cons_elim : (A:Assign)(b:bool)(n:nat)(lb:Constraints)
     (Compatible A b (Cons n lb))
   ->((Compatible A b lb)/\(b=(A n))).
Intros A b n lb; Exact (Compatible_inv A b (Cons n lb)).
Save.

Lemma Compat_assign : (lb:Constraints)(b:bool)(A:Assign)(n:nat)
     (Compatible A b lb)->(Assigned n lb)->b=(A n).
Induction 1.
Intro; Absurd (Assigned n Emp); Auto.
Intros; Elim (Assign_cons n n0 lb0); Auto.
Induction 1; Auto.
Save.

(* Two partial assignments are coherent if they do not contains 
   the same variables *)
Definition Coherent := 
  [ltrue,lfalse:Constraints]
  (n:nat)(Assigned n ltrue)->~(Assigned n lfalse).

Lemma Coherent_sym : (l,m:Constraints)(Coherent l m)->(Coherent m l).
Red; Intros.
Red; Intros.
Exact (H n H1 H0).
Save.
Immediate Coherent_sym.

Lemma Coh_emp_r : (l:Constraints)(Coherent l Emp).
Red; Red; Intros.
Absurd (Assigned n Emp); Auto.
Save.
Hint Coh_emp_r.

Lemma Coherent_not_Ass_true : (ltrue,lfalse:Constraints)
     (n:nat)(Coherent ltrue lfalse)->(~(Assigned n ltrue))
     ->(Coherent ltrue (Cons n lfalse)).
Unfold Coherent; Intros.
Apply not_Ass_cons_intro; Auto.
Red; Intro; Absurd (Assigned n0 ltrue); Auto.
Elim H2; Auto.
Save.
Hint Coherent_not_Ass_true.

Lemma Coherent_not_Ass_false : (ltrue,lfalse:Constraints)
     (n:nat)(Coherent ltrue lfalse)->(~(Assigned n lfalse))
     ->(Coherent (Cons n ltrue) lfalse).
Intros; Apply Coherent_sym; Auto.
Save.
Hint Coherent_not_Ass_false.

Lemma Coherent_cons_false_elim : (ltrue,lfalse:Constraints)(n:nat)
	(Coherent ltrue (Cons n lfalse))
        ->((Coherent ltrue lfalse)/\~(Assigned n ltrue)).
Unfold Coherent; Intros; Split.
Intros m Am; Elim (not_Ass_cons_elim m n lfalse); Auto.
Red; Intros; Absurd (Assigned n (Cons n lfalse)); Auto.
Save.

(* Proof of if we have two coherent partial assignments
   ltrue and lfalse, we can find an assignment which is compatible
   with ltrue for the value true and lfalse for the value false *)

(* a constant assignment *)

Lemma Compat_Ass_const : (lb:Constraints)(b:bool)(Compatible [n:nat]b b lb).
Induction lb; Auto.
Save.
Hint Compat_Ass_const.

(* Updating one value in an assignment *)
Definition upd_assign : Assign -> nat -> bool -> Assign :=
    [A:Assign][n:nat][b:bool][m:nat]
    (<bool> Case (eq_nat_dec n m) of
	(* Inl H *) [H:n=m]b
        (* Inr H *) [H:~n=m](A m) end).

Lemma upd_assign_val : (A:Assign)(n:nat)(b:bool)(m:nat)
     (n=m)->b=(upd_assign A n b m).
Unfold upd_assign.
Intros; Elim (eq_nat_dec n m); Simpl; Auto.
Intro; Absurd n=m; Auto.
Save.
Hint upd_assign_val.

Lemma upd_assign_rest : (A:Assign)(n:nat)(b:bool)(m:nat)
     (~n=m)->(A m)=(upd_assign A n b m).
Unfold upd_assign.
Intros; Elim (eq_nat_dec n m); Simpl; Auto.
Intro; Absurd n=m; Auto.
Save.
Hint upd_assign_rest.

Lemma upd_assign_same : (A:Assign)(n:nat)(b:bool)(m:nat)
     (b=(A m))->b=(upd_assign A n b m).
Intros; Elim (eq_nat_dec n m); Simpl; Auto.
Intro; Elim upd_assign_rest; Auto.
Save.
Hint upd_assign_same.

Lemma Compat_upd_assign_neg : (lb:Constraints)(A:Assign)(n:nat)(b,b':bool)
     (~b=b')->(Compatible A b lb)
     ->(~(Assigned n lb))->(Compatible (upd_assign A n b') b lb).
Induction 2; Intros; Trivial.
Elim (not_Ass_cons_elim n n0 lb0); Trivial.
Intros; Apply Compat_cons; Auto.
Elim upd_assign_rest; Auto.
Save.
Hint Compat_upd_assign_neg.

Lemma Compat_upd_assign_Cons : (lb:Constraints)(A:Assign)(n:nat)(b:bool)
     (Compatible A b lb) -> (Compatible (upd_assign A n b) b (Cons n lb)).
Induction 1; Intros.
Auto.
Elim (Compat_cons_elim (upd_assign A n b) b n lb0); Auto.
Save.
Hint Compat_upd_assign_Cons.

Definition Extends : Constraints -> Constraints -> Assign -> Prop
  := [ltrue,lfalse:Constraints][A:Assign]
    (Compatible A true ltrue) /\ (Compatible A false lfalse).

Lemma Extends_intro : (ltrue,lfalse:Constraints)(A:Assign)
     (Compatible A true ltrue)->(Compatible A false lfalse)
     ->(Extends ltrue lfalse A).
Red; Auto.
Save.
Hint Extends_intro.

Lemma Extends_cons_false : (ltrue,lfalse:Constraints)(A:Assign)(n:nat)
     (Extends ltrue lfalse A)->(false=(A n))
     ->(Extends ltrue (Cons n lfalse) A).
Induction 1; Auto.
Save.
Hint Extends_cons_false.

Lemma Extends_cons_true : (ltrue,lfalse:Constraints)(A:Assign)(n:nat)
     (Extends ltrue lfalse A)->(true=(A n))
     ->(Extends (Cons n ltrue) lfalse A).
Induction 1; Auto.
Save.
Hint Extends_cons_true.

Lemma Extends_cons_elim_false : (ltrue,lfalse:Constraints)(A:Assign)(n:nat)
     (Extends ltrue (Cons n lfalse) A) ->
     ((Extends ltrue lfalse A) /\ false=(A n)).
Induction 1; Intros.
Elim (Compat_cons_elim A false n lfalse); Auto.
Save.

Lemma Extends_cons_elim_true : (ltrue,lfalse:Constraints)(A:Assign)(n:nat)
     (Extends (Cons n ltrue) lfalse A) ->
     ((Extends ltrue lfalse A) /\ true=(A n)).
Induction 1; Intros.
Elim (Compat_cons_elim A true n ltrue); Auto.
Save.

Lemma Extends_assign_false : (ltrue,lfalse:Constraints)(A:Assign)(n:nat)
     (Extends ltrue lfalse A) -> (~(Assigned n ltrue)) 
     -> (Extends ltrue (Cons n lfalse) (upd_assign A n false)).
Red; Induction 1; Auto.
Save.
Hint Extends_assign_false.

Lemma Coherent_Compat : (ltrue,lfalse:Constraints)(Coherent ltrue lfalse)
      ->{A:Assign|(Extends ltrue lfalse A)}.
Induction lfalse.
Intro; Exists [n:nat]true.
Red; Auto.
Intros n lfalse' Hf Cl.
Elim (Coherent_cons_false_elim ltrue lfalse' n); Trivial.
Intros; Elim Hf; Trivial.
Intros A EA; Exists (upd_assign A n false); Auto.
Save.
Hint Coherent_Compat.

Definition Sem : Assign -> IFExpr -> bool -> Prop
               := [A:Assign][F:IFExpr][b:bool]b=(IF_sem A F).

Lemma Sem_Tr : (A:Assign)(Sem A Tr true).
Red; Auto.
Save.
Hint Sem_Tr.

Lemma Sem_Fa : (A:Assign)(Sem A Fa false).
Red; Auto.
Save.
Hint Sem_Fa.

Lemma Sem_If : (A:Assign)(F,G:IFExpr)(n:nat)(b:bool)
        (((true=(A n)) -> (Sem A F b)) 
          /\ ((false=(A n)) -> (Sem A G b)))
     -> (Sem A (IfE (Var n) F G) b).
Red; Simpl; Induction 1.
Elim (A n); Simpl; Intros.
Red in H0; Auto.
Red in H1; Auto.
Save.
Hint Sem_If.


Definition Tautology : IFExpr -> Prop
           := [F:IFExpr](A:Assign)(Sem A F true).

Definition Refutable : IFExpr -> Prop
           := [F:IFExpr](Ex [A:Assign](Sem A F false)).

Definition Part_Tauto : IFExpr -> Constraints -> Constraints -> Prop
           := [F:IFExpr][ltrue,lfalse:Constraints]
             (A:Assign)(Extends ltrue lfalse A)->(Sem A F true).

Definition Part_Refut : IFExpr -> Constraints -> Constraints -> Prop
           := [F:IFExpr][ltrue,lfalse:Constraints]
            (Ex [A:Assign]((Sem A F false)/\(Extends ltrue lfalse A))).

Lemma Part_Refut_intro : (F:IFExpr)(ltrue,lfalse,ltrue',lfalse':Constraints)
     (Coherent ltrue lfalse)->
     ((A:Assign)(Extends ltrue lfalse A)->
                ((Sem A F false)/\(Extends ltrue' lfalse' A)))
     -> (Part_Refut F ltrue' lfalse').
Intros; Elim (Coherent_Compat ltrue lfalse); Trivial.
Intros A EA; Red; Exists A; Auto.
Save.

(* Useful lemmas *)

(* (Part_Equiv F G ltrue lfalse) means that the two formulas F and G
   are equivalent under all assignemnts which extend ltrue,lfalse *)

Definition Part_Equiv := 
     [F,G:IFExpr][ltrue,lfalse:Constraints]
     (A:Assign)(Extends ltrue lfalse A)->(IF_sem A F)=(IF_sem A G).

Lemma Part_Equiv_Sem : (F,G:IFExpr)(ltrue,lfalse:Constraints)
     (Part_Equiv F G ltrue lfalse)->(A:Assign)(Extends ltrue lfalse A)
     ->(b:bool)(Sem A F b)->(Sem A G b).
Red; Intros.
Elim H; Auto.
Save.

(* If F and G are equivalent under all assignment which extends
   ltrue, lfalse then the partial tautology problem is the same
   for both formulas *)

Lemma Part_Equiv_Tauto : (F,G:IFExpr)(ltrue,lfalse:Constraints)
     (Part_Equiv F G ltrue lfalse)
     ->({Part_Tauto F ltrue lfalse}+{Part_Refut F ltrue lfalse})
     ->({Part_Tauto G ltrue lfalse}+{Part_Refut G ltrue lfalse}).
Induction 2; Intros.
Left; Red; Intros.
Apply Part_Equiv_Sem with F ltrue lfalse; Auto.
(*Apply a; Auto.*)
Right; Elim y; Intros.
Red; Exists x.
Elim H1; Split; Trivial.
Apply Part_Equiv_Sem with F ltrue lfalse; Auto.
Save.

(* Some equivalent formulas *)

Lemma Part_Equiv_Var_true : (n:nat)(ltrue,lfalse:Constraints)
     (Assigned n ltrue)->(Part_Equiv Tr (Var n) ltrue lfalse).
Red; Intros; Simpl.
Apply Compat_assign with ltrue; Auto.
Elim H0; Auto.
Save.
Hint Part_Equiv_Var_true.

Lemma Part_Equiv_If_true : (n:nat)(F,G:IFExpr)(ltrue,lfalse:Constraints)
     (Assigned n ltrue)->(Part_Equiv F (IfE (Var n) F G) ltrue lfalse).
Red; Intros; Simpl.
Replace (A n) with true; Auto.
Apply Compat_assign with ltrue; Auto.
Elim H0; Auto.
Save.
Hint Part_Equiv_If_true.

Lemma Part_Equiv_If_false : (n:nat)(F,G:IFExpr)(ltrue,lfalse:Constraints)
     (Assigned n lfalse)->(Part_Equiv G (IfE (Var n) F G) ltrue lfalse).
Red; Intros; Simpl.
Replace (A n) with false; Auto.
Apply Compat_assign with lfalse; Auto.
Elim H0; Auto.
Save.
Hint Part_Equiv_If_false.

(* We solve the problem for all normal formulas and 
   coherent partial assignments *)

Lemma Partial_Tauto : (F:IFExpr)(Normal F)->
     (ltrue,lfalse:Constraints)
     (Coherent ltrue lfalse)
     ->({Part_Tauto F ltrue lfalse}+{Part_Refut F ltrue lfalse}).
(* We only look at the cases corresponding to normal forms *)
Induction 1.

(* Case F = Var n *)
Intros n ltrue lfalse Cltlf; Elim (Assign_dec n ltrue).
   (* Subcase n is assigned to true *)
Intro; Apply Part_Equiv_Tauto with Tr; Auto.
Left; Red ; Auto.
   (* Subcase n is not assigned to true, so ltrue and (Cons n lfalse)
      are coherent, a compatible assignment is a refutation *)
Right.
Apply Part_Refut_intro with ltrue (Cons n lfalse); Auto.
Intros; Elim (Extends_cons_elim_false ltrue lfalse A n); Auto.

(* Case F = Tr *)
Left; Red; Auto.

(* Case F = Fa *)
Right.
Apply Part_Refut_intro with ltrue lfalse; Auto.

(* Case F = (IF (Var n) F' G') *)
Intros n F' G' NF HF NG HG ltrue lfalse Cltlf.
Elim (Assign_dec n ltrue); Intro.
	(* Subcase n is assigned to true, (Var n)=T *)
Apply Part_Equiv_Tauto with F'; Auto.
        (* Subcase n is not assigned to true *)
Elim (Assign_dec n lfalse); Intro.
         	(* Subsubcase n is assigned to false (Var n)=F *)
Apply Part_Equiv_Tauto with G'; Auto.
                (* Subsubcase n is not assigned to false *)
Elim (HF (Cons n ltrue) lfalse); Auto.
		(* Case F' is a tautology under (Var n)=T *)
Intro; Elim (HG ltrue (Cons n lfalse)); Auto.
			(* Subcase G' is a tautology under (Var n)=F *)
Left; Red.  
Intros; Apply Sem_If.
Split; Intro.
Apply y1; Auto.
Apply y2; Auto.
                        (* Subcase G' is refutable  under (Var n)=F *)
Right.
Elim y2; Intros A SA.
Red; Exists A.
Elim SA; Intros.
Elim (Extends_cons_elim_false ltrue lfalse A n); Trivial.
Split; Trivial.
Apply Sem_If.
Split; Trivial.
Elim H3; Intro; Absurd true=false; Auto.
		(* Case F' is refutable  under (Var n)=T *)
Right.
Elim y1; Intros A SA.
Red; Exists A.
Elim SA; Intros.
Elim (Extends_cons_elim_true ltrue lfalse A n); Trivial.
Split; Trivial.
Apply Sem_If.
Split; Trivial.
Elim H3; Intro; Absurd true=false; Auto.
Save.

(* The final program corresponds to the case ltrue=lfalse=empty *)
Lemma Norm_Tautology : (F:IFExpr)(Normal F)->{(Tautology F)}+{(Refutable F)}.
Intros F NF; Elim (Partial_Tauto F NF Emp Emp); Trivial.
Intro; Left; Red.
Intro; Apply y; Auto.
Intro; Right; Red.
Elim y; Intros A EA.
Exists A; Elim EA; Auto.
Save.

(****************************************************)
(* Each IF formula admits an equivalent Normal form *)
(****************************************************)

Definition Equiv := [F,G:IFExpr](A:Assign)(IF_sem A F)=(IF_sem A G).

Lemma Equiv_If_Tr : (F,G:IFExpr)(Equiv (IfE Tr F G) F).
Unfold Equiv; Auto.
Save.

Lemma Equiv_If_Fa : (F,G:IFExpr)(Equiv (IfE Fa F G) G).
Unfold Equiv; Auto.
Save.

Lemma Equiv_If_If : (F,G,H,F',G':IFExpr)(Equiv (IfE (IfE F G H) F' G')
                                (IfE F (IfE G F' G') (IfE H F' G'))).
Unfold Equiv; Simpl; Intros.
Elim (IF_sem A F); Unfold 2 3 ifb; Auto.
Save.
Hint Equiv_If_If.

Lemma Equiv_refl : (F:IFExpr)(Equiv F F).
Unfold Equiv;Auto.
Save.
Hint Equiv_refl.

Lemma Equiv_trans : (F,G,H:IFExpr)(Equiv F G)->(Equiv G H)->(Equiv F H).
Unfold Equiv; Intros.
Transitivity (IF_sem A G); Auto.
Save.

Lemma Equiv_congr : (F,G,H,F',G',H':IFExpr)
     (Equiv F F')->(Equiv G G')->(Equiv H H')->
     (Equiv (IfE F G H) (IfE F' G' H')).
Unfold Equiv; Intros; Simpl.
Elim H0; Elim H1; Elim H2; Auto.
Save.
Hint Equiv_congr.

(* The well-founded relation should satisfy :
                                G < IfE Tr G H
                                H < IfE Fa G H
                                G,H < (IfE Var G H)
   (IfE F (IfE G F' G') (IfE H F' G')) < (IfE (IfE F G H) F' G')
The measure is m : IFExpr -> nat
   such that m(Tr)=m(Fa)=m(Var(n))=1
        and  m(IfE F G H)=m(F)*(S(m(G)+m(H)))

*)

Lemma IFExpr_norm_rec1 : (P:IFExpr->Set)
     (P Tr) -> (P Fa) -> ((n:nat)(P (Var n)))
     -> ((Y,Z:IFExpr)(P Y)->(P (IfE Tr Y Z)))
     -> ((Y,Z:IFExpr)(P Z)->(P (IfE Fa Y Z)))
     -> ((n:nat)(Y,Z:IFExpr)(P Y)->(P Z)->(P (IfE (Var n) Y Z)))
     -> ((X,Y,Z,T,U:IFExpr)(P (IfE X (IfE T Y Z) (IfE U Y Z)))
                           ->(P (IfE (IfE X T U) Y Z)))
     -> (X:IFExpr)(P X).
Induction X; Trivial.
Intros X' PX'; Elim X'; Auto.
Save.

Lemma IFExpr_norm_ind : (P:IFExpr->Prop)
     (P Tr) -> (P Fa) -> ((n:nat)(P (Var n)))
     -> ((Y,Z:IFExpr)(P Y)->(P (IfE Tr Y Z)))
     -> ((Y,Z:IFExpr)(P Z)->(P (IfE Fa Y Z)))
     -> ((n:nat)(Y,Z:IFExpr)(P Y)->(P Z)->(P (IfE (Var n) Y Z)))
     -> ((X,Y,Z,T,U:IFExpr)(P (IfE X (IfE T Y Z) (IfE U Y Z)))
                           ->(P (IfE (IfE X T U) Y Z)))
     -> (X:IFExpr)(P X).
Induction X; Trivial.
Intros X' PX'; Elim X'; Auto.
Save.

Definition norm_order : IFExpr->IFExpr->Prop :=
     [X,Y:IFExpr]
     <Prop>Case Y of
	(* Var n *) [n:nat]False
           (* Tr *) False
           (* Fa *) False
    (* IfE T U V *) [T,U,V:IFExpr]
                    <Prop>Case T of
                     (* Var n *) [n:nat](U=X) \/ (V=X)
                        (* Tr *) U=X
                        (* Fa *) V=X
                 (* IfE P Q R *) [P,Q,R:IFExpr]
                                   (IfE P (IfE Q U V) (IfE R U V))=X
	            end 
      end.

Lemma Wf_norm_order : (well_founded IFExpr norm_order).
Red;Intro.
Pattern 1 a.
Apply IFExpr_norm_ind; Intros; Apply Acc_intro; Simpl.
Induction 1.
Induction 1.
Induction 1.
Induction 1; Trivial.
Induction 1; Trivial.
Induction 1; Induction 1; Trivial.
Induction 1; Trivial.
Save.
Hint Wf_norm_order.

Lemma IFExpr_match : (P:IFExpr->Set)
        ((n:nat)(P (Var n)))->(P Tr)->(P Fa)->((X,Y,Z:IFExpr)(P (IfE X Y Z)))
      ->(I:IFExpr)(P I).
Induction I; Auto.
Save.

Lemma IFExpr_norm_rec2 : (P:IFExpr->Set)
     (P Tr) -> (P Fa) -> ((n:nat)(P (Var n)))
     -> ((Y,Z:IFExpr)(P Y)->(P (IfE Tr Y Z)))
     -> ((Y,Z:IFExpr)(P Z)->(P (IfE Fa Y Z)))
     -> ((n:nat)(Y,Z:IFExpr)(P Y)->(P Z)->(P (IfE (Var n) Y Z)))
     -> ((X,Y,Z,T,U:IFExpr)(P (IfE X (IfE T Y Z) (IfE U Y Z)))
                           ->(P (IfE (IfE X T U) Y Z)))
     -> (X:IFExpr)(P X).
Intros; Pattern X; Apply (well_founded_induction IFExpr norm_order); Trivial.
Intros F'; Pattern F'; Apply IFExpr_match; Trivial.
Intros Y Z T; Pattern Y; Apply IFExpr_match; Simpl; Auto.
Save.

Variable IFExpr_norm_rec :
(P:IFExpr->Set)
     (P Tr) -> (P Fa) -> ((n:nat)(P (Var n)))
     -> ((Y,Z:IFExpr)(P Y)->(P (IfE Tr Y Z)))
     -> ((Y,Z:IFExpr)(P Z)->(P (IfE Fa Y Z)))
     -> ((n:nat)(Y,Z:IFExpr)(P Y)->(P Z)->(P (IfE (Var n) Y Z)))
     -> ((X,Y,Z,T,U:IFExpr)(P (IfE X (IfE T Y Z) (IfE U Y Z)))
                           ->(P (IfE (IfE X T U) Y Z)))
     -> (X:IFExpr)(P X).



(* $Id: Tauto_def.v,v 1.2 1996/02/12 15:30:30 cpaulin Exp $ *)
