(****************************************************************************)
(*                 The Calculus of Inductive Constructions                  *)
(*                                                                          *)
(*                                Projet Coq                                *)
(*                                                                          *)
(*                     INRIA                        ENS-CNRS                *)
(*              Rocquencourt                        Lyon                    *)
(*                                                                          *)
(*                                Coq V5.11                                 *)
(*                              Feb 2nd 1996                                *)
(*                                                                          *)
(****************************************************************************)
(*                               Schroeder.v                                *)
(****************************************************************************)


(*     If A is of cardinal less than B and conversely, then A and B          *)
(*       are equipollent                                                     *)
(*     In other words, if there is an injective map from A to B and          *)
(*   an injective map from B to A then there exists a map from A onto B.     *)

(*                (d'apres une demonstration de Fraenkel)                    *)

Require Ensembles.      (* Ensemble, In, Included, Setminus *)
Require Relations_1.    (* Relation, Transitive *)
Require Powerset.       (* Inclusion_is_transitive *)
Require Classical_Prop. (* classic *)

Require Setminus_fact.
Require Sums.
Require Functions.
Require Equipollence.

Section Schroeder_Bernstein.


(*****************************************************************************)
(* The only use of classical arguments is through the following instance of  *)
(* the excluded-middle axiom                                                 *)

Definition in_or_not_in := [U:Type][x:U][A:U->Prop](classic (In U A x)).


(*****************************************************************************)
(*  A and B are sets of elements in the univers U                            *)


Variable U:Type.

Local SU (Ensemble U).

Variables A,B:SU.  (* A et B sont des ensembles d'elements de l'univers U  *)


  Section Bijection.

  (**************************************************************************)
  (* On montre dans ce paragraphe que si f et g sont des injections resp    *)
  (* de A dans B et de B dans A alors on peut trouver un sous-ensemble J de *)
  (* A tq h qui est f sur J et g sur A\J est une bijection de A dans B      *)

  Variables f,g:(Relation U).  (* f et g sont des relations *)

  Hypothesis f_inj:(injection U A B f). (* f et g sont des injections *)
  Hypothesis g_inj:(injection U B A g).

  Local Imf : (Ensemble U) -> (Ensemble U) := (Im U f).
  Local Img : (Ensemble U) -> (Ensemble U) := (Im U g).

  (* Construction de J tq g(B\f(J))=A\J *)

    (* (Setminus U A C) designe la difference A\C         *)
    (* (Included U A C) signifie que A est inclus dans C  *)

    Local F [C:SU](Setminus U A (Img (Setminus U B (Imf C)))).

    Local D [C:SU](Included U C (F C)).

    Local J (Sum U D).


  (*  On va montrer que J correspond a ce que l'on cherche *)

    (* J correspond exactement au point fixe de Tarski pour F, *)
    (* fonction croissante relativement a l'inclusion          *)

    (* Lemma F est croissante *)

      Remark F_croissante :
      (C,C':SU)(Included U C C')->(Included U (F C) (F C')).
        Intros; Unfold F.
        Apply Setminus_contravariant.
        Unfold Img.
        Apply Im_stable_par_incl.
        Apply Setminus_contravariant.
        Unfold Imf.
        Apply Im_stable_par_incl.
        Assumption.
      Qed.

    (* On va montrer que F(J)=A\Img(B\Imf(J))=J *)

       (* D'abord l'inclusion dans un sens *)

         (*  Lemma J_dans_FJ (Included U J (F J))  *)

         Remark J_dans_FJ : (Included U J (F J)).
           Unfold J.
           Apply Sum_is_majoring.
           Intros C C_in_D.
           Cut (Transitive (Ensemble U) (Included U)).
             2:Apply Inclusion_is_transitive.
           Intro Incl_is_trans.
           Unfold Transitive in Incl_is_trans.
           Apply Incl_is_trans with (F C).
           (* Que C est inclus dans (F C) *)
             Assumption.
           (* Que (F C) inclus dans (F (Sum U D)) *)
             Apply F_croissante.
             Apply Sum_is_minoring.
             Assumption.
         Qed.

       (* Puis dans l'autre sens *)

         (*  Lemma FJ_dans_J (Included U (F J) J)  *)

         Remark FJ_dans_J : (Included U (F J) J).
           Unfold J.
           Apply Sum_is_minoring.
           Red.
           Red.
           Apply F_croissante.
           Exact J_dans_FJ.
         Qed.


  (* On montre que h qui est f sur J et g ailleurs est une bijection *)

    Inductive h [x,y:U] : Prop :=
       hl_intro : (In U J x)->(f x y)->(h x y)
     | hr_intro : (Setminus U B (Imf J) y)->(g y x)->(h x y).


  (*  Theorem h_bij (bijection U A B h)     *)

  Theorem h_bij : (bijection U A B h).


    (* h est de A dans B *)
    Remark h1: (Rel U A B h).
	Goal.
        Apply Rel_intro; Do 2 Intro; Intro h_x_y.
        (* h est sur A *) 
          Elim h_x_y.
          (* sur J : f est de A dans B *)
            Elim f_inj.
            Intro f_Rel; Intros.
            Elim f_Rel.
            Intros f_sur_A f_sur_B.
            Apply f_sur_A with y; Assumption.

        (* sur A\J: g est de B dans A *)
          Elim g_inj.
          Intro g_Rel; Intros.
          Elim g_Rel.
          Intros g_sur_B g_sur_A.
          Apply g_sur_A with y; Assumption.

      (* h est sur B *) 
        Elim h_x_y.
        (* sur J : f est de A dans B *)
          Elim f_inj.
          Intro f_Rel; Intros.
          Elim f_Rel.
          Intros f_sur_A f_sur_B.
          Apply f_sur_B with x; Assumption.

        (* sur A\J: g est de B dans A *)
        Elim g_inj.
        Intro g_Rel; Intros.
        Elim g_Rel.
        Intros g_sur_B g_sur_A.
        Apply g_sur_B with x; Assumption.

    Qed.


    (* h verifie au_plus_une_im *)
    Remark h2: (au_plus_une_im U h).
	Goal.
      Red; Intros x y z h_x_y h_x_z.
      Elim h_x_y.

      (* sur J *)
        Elim h_x_z.
        (* cas ou (h x y) et (h x z) se comporte comme f : correct *)
          Elim f_inj.
          Unfold au_plus_une_im; Intros f_Rel f_au_plus_1_im; Intros.
          Apply f_au_plus_1_im with x; Assumption.

        (* Cas ou (h x y) se comporte comme f et
                  (h x z) comme g : contradiction *)
          Do 2 Intro; Intro x_in_J; Intro.
          Cut (Included U J (F J)).
            Unfold Included; Unfold F; Unfold Setminus; Intro Hyp.
            Elim (Hyp x x_in_J).
            Intros x_in_A x_in_non_Img.
            Elim x_in_non_Img.
            Red.
            Red.
            Apply Im_intro with z; Assumption.
          Exact J_dans_FJ.

      (* sur A\J *)
        Elim h_x_z.
        (* Cas ou (h x y) se comporte comme g et
                  (h x z) comme f : contradiction *)
          Intro x_in_J; Intros.
          Cut (Included U J (F J)).
            Unfold Included; Unfold F; Unfold Setminus; Intro Hyp.
            Elim (Hyp x x_in_J).
            Intros x_in_A x_in_non_Img.
            Elim x_in_non_Img.
            Red.
            Red.
            Apply Im_intro with y; Assumption.
          Exact J_dans_FJ.


        (* cas ou (h x y) et (h x z) se comporte comme g : correct *) 
          Elim g_inj.
          Unfold au_plus_un_ant; Do 3 Intro; Intro g_au_plus_1_ant; Intros.
          Apply g_au_plus_1_ant with x; Assumption.

    Qed.


    (* h verifie au_moins_une_im *)
    Remark h3: (au_moins_une_im U A h).
	Goal.
      Red.
      Intros.
      Elim (in_or_not_in U x (Img (Setminus U B (Imf J)))).

      (* sur A\J *)
      Unfold Img; Intro x_in_Img.
      Elim x_in_Img.
      Intros y g_y_x H1.
      Exists y.
      Apply hr_intro; Assumption.

      (* sur J *)
      Intros.
        (* De f fonction, on deduit f verifie au_moins_une_im *)
        Elim f_inj.
        Unfold au_moins_une_im; Do 2 Intro; Intro f_au_moins_1_im; Intro.
        Elim (f_au_moins_1_im x H).
        Intros y f_x_y.
        Exists y.
        Apply hl_intro.
          Apply FJ_dans_J.
          Red; Red; Red.
          Split; Assumption.
        Assumption.

    Qed.


    (* h verifie au_plus_un_ant *)
    Remark h4: (au_plus_un_ant U h).
	Goal.
      Red; Do 3 Intro; Intros h_x_z h_y_z.
      Elim h_x_z.

      (* sur J *)
        Elim h_y_z.
        (* cas ou (h x y) et (h x z) se comporte comme f : correct *)
          Elim f_inj.
          Intros.
          Cut (x,y,z:U)(f x z)->(f y z)->(x==y).
          Intro Hyp; Apply Hyp with z; Assumption.
          Assumption.

        (* Montrer qu'on ne peut avoir (f x z) et (g z y) avec x dans J et
            z hors de (Imf J) sans contradiction *)
          Unfold Setminus; Intro z_in_Setminus_B_Imf_J; Intros.
          Elim z_in_Setminus_B_Imf_J.
          Intros z_in_B z_in_non_Imf_J.
          Elim z_in_non_Imf_J.
          Red.
          Red.
          Apply Im_intro with x; Assumption.

      (* sur A\J *)
        Elim h_y_z.
        (* Montrer qu'on ne peut avoir (f y z) et (g z x) avec x dans J et
            z hors de (Imf J) sans contradiction *)
          Unfold Setminus; Do 2 Intro; Intro z_in_Setminus_B_Imf_J; Intros.
          Elim z_in_Setminus_B_Imf_J.
          Intros z_in_B z_in_non_Imf_J.
          Elim z_in_non_Imf_J.
          Red.
          Red.
          Apply Im_intro with y; Assumption.

        (* De g fonction on deduit g verifie au_plus_une_im c'est-a-dire
                                              au_plus_un_ant pour h *)
          Elim g_inj.
          Intros.
          Cut (z,x,y:U)(g z x)->(g z y)->(x==y).
          Intro Hyp; Apply Hyp with z; Assumption.
          Assumption.

    Qed.


    (* h verifie au_moins_un_ant *)
    Remark h5: (au_moins_un_ant U B h).
	Goal.
      Red.
      Intros.
      Elim (in_or_not_in U y (Imf J)).

      (* sur J *)
      Unfold Imf; Intro y_in_Imf.
        (* De f injective on deduit f verifie au_moins_un_ant *)
          Elim y_in_Imf.
          Intros x f_x_y; Intro.
          Exists x.
          Apply hl_intro; Assumption.

      (* sur A\J *)
        Intros.
        (* De g injective on deduit g verifie au_moins_une_im c'est-a-dire
                                              au_moins_un_ant pour h *)
          Elim g_inj.
          Unfold au_moins_une_im; Do 2 Intro; Intro g_au_moins_1_im; Intro.
          Elim (g_au_moins_1_im y H).
          Intros x g_y_x.
          Exists x.
          Apply hr_intro.
          Red.
          Split; Assumption.
          Assumption.

    Qed.

	Resume h_bij.
    Proof (bijection_intro U A B h h1 h2 h3 h4 h5).

  End Bijection.


(*    Le theoreme de Schroeder-Bernstein-Cantor     *)

Theorem Schroeder : (inf_card U A B)->(inf_card U B A)->(equipollence U A B).

  Intros A_inf_B B_inf_A.
  Elim A_inf_B.
  Intros.
  Elim B_inf_A.
  Intros.
  Apply equipollence_intro with (h f f0).
  Apply h_bij; Assumption.

Qed.


End Schroeder_Bernstein.


                           (* The end *)


(* $Id: Schroeder.v,v 1.2 1996/02/12 15:32:47 cpaulin Exp $ *)
